(function(/*! Brunch !*/) {
  'use strict';

  var globals = typeof window !== 'undefined' ? window : global;
  if (typeof globals.require === 'function') return;

  var modules = {};
  var cache = {};

  var has = function(object, name) {
    return ({}).hasOwnProperty.call(object, name);
  };

  var expand = function(root, name) {
    var results = [], parts, part;
    if (/^\.\.?(\/|$)/.test(name)) {
      parts = [root, name].join('/').split('/');
    } else {
      parts = name.split('/');
    }
    for (var i = 0, length = parts.length; i < length; i++) {
      part = parts[i];
      if (part === '..') {
        results.pop();
      } else if (part !== '.' && part !== '') {
        results.push(part);
      }
    }
    return results.join('/');
  };

  var dirname = function(path) {
    return path.split('/').slice(0, -1).join('/');
  };

  var localRequire = function(path) {
    return function(name) {
      var dir = dirname(path);
      var absolute = expand(dir, name);
      return globals.require(absolute);
    };
  };

  var initModule = function(name, definition) {
    var module = {id: name, exports: {}};
    definition(module.exports, localRequire(name), module);
    var exports = cache[name] = module.exports;
    return exports;
  };

  var require = function(name) {
    var path = expand(name, '.');

    if (has(cache, path)) return cache[path];
    if (has(modules, path)) return initModule(path, modules[path]);

    var dirIndex = expand(path, './index');
    if (has(cache, dirIndex)) return cache[dirIndex];
    if (has(modules, dirIndex)) return initModule(dirIndex, modules[dirIndex]);

    throw new Error('Cannot find module "' + name + '"');
  };

  var define = function(bundle, fn) {
    if (typeof bundle === 'object') {
      for (var key in bundle) {
        if (has(bundle, key)) {
          modules[key] = bundle[key];
        }
      }
    } else {
      modules[bundle] = fn;
    }
  };

  globals.require = require;
  globals.require.define = define;
  globals.require.register = define;
  globals.require.brunch = true;
})();

window.require.register("app", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  // Application bootstrapper
  require('utils/ember_reopen');
  var stringUtils = require('utils/string_utils');

  module.exports = Em.Application.create({
    name: 'Ambari Web',
    rootElement: '#wrapper',

    store: DS.Store.create({
      revision: 4,
      adapter: DS.FixtureAdapter.create({
        simulateRemoteResponse: false
      })
    }),
    isAdmin: false,
    /**
     * return url prefix with number value of version of HDP stack
     */
    stackVersionURL:function(){
      var stackVersion = this.get('currentStackVersion') || this.get('defaultStackVersion');
      if(stackVersion.indexOf('HDPLocal') !== -1){
        return '/stacks/HDPLocal/version/' + stackVersion.replace(/HDPLocal-/g, '');
      }
      return '/stacks/HDP/version/' + stackVersion.replace(/HDP-/g, '');
    }.property('currentStackVersion'),
    
    /**
     * return url prefix with number value of version of HDP stack
     */
    stack2VersionURL:function(){
      var stackVersion = this.get('currentStackVersion') || this.get('defaultStackVersion');
      if(stackVersion.indexOf('HDPLocal') !== -1){
        return '/stacks2/HDPLocal/versions/' + stackVersion.replace(/HDPLocal-/g, '');
      }
      return '/stacks2/HDP/versions/' + stackVersion.replace(/HDP-/g, '');
    }.property('currentStackVersion'),

    falconServerURL: function () {
      var falconService = this.Service.find().findProperty('serviceName', 'FALCON');
      if (falconService) {
        return falconService.get('hostComponents').findProperty('componentName', 'FALCON_SERVER').get('host.hostName');
      }
    }.property().volatile(),

    clusterName: null,
    clockDistance:null, // server clock - client clock
    currentStackVersion: '',
    currentStackVersionNumber: function(){
      return this.get('currentStackVersion').replace(/HDP(Local)?-/, '');
    }.property('currentStackVersion'),
    isHadoop2Stack: function(){
      return (stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.0") === 1 ||
        stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.0") === 0)
    }.property('currentStackVersionNumber'),
    isHadoop21Stack: function(){
      return (stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.1") === 1 ||
        stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.1") === 0)
    }.property('currentStackVersionNumber'),

    /**
     * If High Availability is enabled
     * Based on <code>clusterStatus.isInstalled</code>, stack version, <code>SNameNode</code> availability
     *
     * @type {Boolean}
     */
    isHaEnabled: function() {
      if (!this.get('isHadoop2Stack')) return false;
      return !this.HostComponent.find().someProperty('componentName', 'SECONDARY_NAMENODE');
    }.property('router.clusterController.isLoaded'),

    /**
     * List of disabled components for the current stack with related info.
     * Each element has followed structure:
     * @type {Em.Object}
     *   @property componentName {String} - name of the component
     *   @property properties {Object} - mapped properties by site files,
     *    for example:
     *      properties: { global_properties: [], site_properties: [], etc. }
     *   @property reviewConfigs {Ember.Object} - reference review_configs.js
     *   @property serviceComponent {Object} - reference service_components.js
     *
     * @type {Array}
     */
    stackDependedComponents: [],

    /**
     * Restore component data that was excluded from stack.
     *
     * @param component {Ember.Object} - #stackDependedComponents item
     */
    enableComponent: function(component) {
      var propertyFileNames = ['global_properties', 'site_properties'];
      var requirePrefix = this.get('isHadoop2Stack') ? 'data/HDP2/' : 'data/';
      // add component to service_components list
      require('data/service_components').push(component.get('serviceComponent'));
      // add properties
      propertyFileNames.forEach(function(fileName) {
        require(requirePrefix + fileName).configProperties = require(requirePrefix + fileName).configProperties.concat(component.get('properties.'+fileName));
      });
      var reviewConfigsService = require('data/review_configs')
        .findProperty('config_name', 'services').config_value
        .findProperty('service_name', component.get('serviceComponent.service_name'));
      reviewConfigsService.get('service_components').pushObject(component.get('reviewConfigs'));
    },
    /**
     * Disabling component. Remove related data from lists such as
     * properties, review configs, service components.
     *
     * @param component {Object} - component info reference service_components.js
     *
     * @return {Ember.Object} - item of <code>stackDependedComponents</code> property
     */
    disableComponent: function(component) {
      var componentCopy, propertyFileNames;
      propertyFileNames = ['global_properties', 'site_properties'];
      componentCopy = Em.Object.create({
        componentName: component.component_name,
        properties: {},
        reviewConfigs: {},
        configCategory: {},
        serviceComponent: {}
      });
      componentCopy.set('serviceComponent', require('data/service_components').findProperty('component_name', component.component_name));
      // remove component from service_components list
      require('data/service_components').removeObject(componentCopy.get('serviceComponent'));
      var serviceConfigsCategoryName, requirePrefix, serviceConfig;
      // get service category name related to component
      serviceConfig = require('data/service_configs').findProperty('serviceName', component.service_name);
      serviceConfig.configCategories = serviceConfig.configCategories.filter(function(configCategory) {
        if (configCategory.get('hostComponentNames')) {
          serviceConfigsCategoryName = configCategory.get('name');
          if (configCategory.get('hostComponentNames').contains(component.component_name)) {
            componentCopy.set('configCategory', configCategory);
          }
        }
        return true;
      });
      requirePrefix = this.get('isHadoop2Stack') ? 'data/HDP2/' : 'data/';
      var propertyObj = {};
      propertyFileNames.forEach(function(propertyFileName) {
        propertyObj[propertyFileName] = [];
      });
      // remove config properties related to this component
      propertyFileNames.forEach(function(propertyFileName) {
        var properties = require(requirePrefix + propertyFileName);
        properties.configProperties = properties.configProperties.filter(function(property) {
          if (property.category == serviceConfigsCategoryName) {
            propertyObj[propertyFileName].push(property);
            return false;
          } else {
            return true;
          }
        });
      });
      componentCopy.set('properties', propertyObj);
      // remove component from review configs
      var reviewConfigsService = require('data/review_configs')
        .findProperty('config_name', 'services').config_value
        .findProperty('service_name', component.service_name);
      reviewConfigsService.set('service_components', reviewConfigsService.get('service_components').filter(function (serviceComponent) {
        if (serviceComponent.get('component_name') != component.component_name) {
          return true;
        } else {
          componentCopy.set('reviewConfigs', serviceComponent);
          return false;
        }
      }));
      return componentCopy;
    },
    /**
     * Resolve dependency in components. Check forbidden/allowed components and
     * remove/restore related data.
     */
    handleStackDependedComponents: function() {
      // need for unit testing and test mode
      if (this.get('handleStackDependencyTest') || this.testMode) return;
      var stackVersion, stackDependedComponents;
      stackVersion = this.get('currentStackVersionNumber');
      stackDependedComponents = [];
      // disable components
      require('data/service_components').filterProperty('stackVersions').forEach(function(component) {
        if (!component.stackVersions.contains(stackVersion))
          stackDependedComponents.push(this.disableComponent(component));
      }, this);
      // enable components
      if (this.get('stackDependedComponents').length > 0) {
        this.get('stackDependedComponents').forEach(function(component) {
          if (component.get('serviceComponent').stackVersions.contains(this.get('currentStackVersionNumber'))) {
            this.enableComponent(component);
            stackDependedComponents = this.get('stackDependedComponents').removeObject(component);
          }
        }, this);
      }
      this.set('stackDependedComponents', this.get('stackDependedComponents').concat(stackDependedComponents));
    }.observes('currentStackVersionNumber'),

    /**
     * List of components with allowed action for them
     * @type {Em.Object}
     */
    components: Ember.Object.create({
      reassignable: ['NAMENODE', 'SECONDARY_NAMENODE', 'JOBTRACKER', 'RESOURCEMANAGER'],
      restartable: function() {
        return this.get('masters').concat(this.get('slaves'));
      }.property('masters'),
      deletable: ['SUPERVISOR', 'HBASE_MASTER', 'DATANODE', 'TASKTRACKER', 'NODEMANAGER', 'HBASE_REGIONSERVER'],
      rollinRestartAllowed: ["DATANODE", "TASKTRACKER", "NODEMANAGER", "HBASE_REGIONSERVER", "SUPERVISOR"],
      decommissionAllowed: ["DATANODE", "TASKTRACKER", "NODEMANAGER", "HBASE_REGIONSERVER"],
      addableToHost: ["DATANODE", "TASKTRACKER", "NODEMANAGER", "HBASE_REGIONSERVER", "HBASE_MASTER", "ZOOKEEPER_SERVER", "SUPERVISOR"],
      slaves: function() {
        return require('data/service_components').filter(function(component){
          return !component.isClient && !component.isMaster;
        }).mapProperty('component_name').uniq().without("DASHBOARD").without("MYSQL_SERVER");
      }.property().cacheable(),

      masters: function() {
        return require('data/service_components').filterProperty('isMaster', true)
          .mapProperty('component_name').concat(['MYSQL_SERVER']).uniq();
      }.property().cacheable(),
      clients: function() {
        return require('data/service_components').filterProperty('isClient', true).mapProperty('component_name').uniq();
      }.property().cacheable()
    })
  });
});
window.require.register("classes/job_class", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var date = require('utils/date');
  var misc = require('utils/misc');

  App.Job2 = Ember.Object.extend({

    id: "", //string
    jobName: "", //string
    workflowEntityName: "", //string
    maps: 0, //number
    reduces: 0, //number
    status: "", //string
    input: 0, //number
    output: 0, //number
    elapsed_time: 0, //number

    duration: function() {
      return date.timingFormat(parseInt(this.get('elapsed_time')));
    }.property('elapsed_time'),

    inputFormatted: function () {
      return misc.formatBandwidth(this.get('input'));
    }.property('input'),

    outputFormatted: function () {
      return misc.formatBandwidth(this.get('output'));
    }.property('output')

  });
  
});
window.require.register("classes/run_class", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var date = require('utils/date');
  var misc = require('utils/misc');

  App.Run2 = Ember.Object.extend({
    id: null, //string
    appName: null, //string
    userName: null, //string
    numJobsTotal: 0, //number
    numJobsCompleted: 0, //number
    startTime: 0, //number
    elapsedTime: 0, //number
    workflowContext: null, //string
    input: 0, //number
    output: 0, //number

    /**
     * Will set to true when we load all jobs related to this run
     */
    loadAllJobs : false,

    /**
     * runId  short part
     */
    idFormatted: function() {
      return this.get('id').substr(0, 20);
    }.property('id'),

    /**
     * Run duration
     */
    duration: function() {
      return date.timingFormat(this.get('elapsedTime'));
    }.property('elapsedTime'),

    /**
     * Status of running jobs
     */
    isRunning: function () {
      return !this.get('numJobsTotal') == this.get('numJobsCompleted');
    }.property('numJobsTotal', 'numJobsCompleted'),

    /**
     * Sum of input bandwidth for all jobs with appropriate measure
     */
    inputFormatted: function () {
      return misc.formatBandwidth(this.get('input'));
    }.property('input'),

    /**
     *  Sum of output bandwidth for all jobs with appropriate measure
     */
    outputFormatted: function () {
      return misc.formatBandwidth(this.get('output'));
    }.property('output'),

    lastUpdateTime: function() {
      return this.get('startTime') + this.get('elapsedTime');
    }.property('elapsedTime', 'startTime'),

    lastUpdateTimeFormatted: function() {
      return date.dateFormat(this.get('lastUpdateTime'));
    }.property('lastUpdateTime'),

    lastUpdateTimeFormattedShort: function(){
      return date.dateFormatShort(this.get('lastUpdateTime'));
    }.property('lastUpdateTime'),

    /**
     * Type value based on first part of id
     */
    type: function() {
      if (this.get('id').indexOf('pig_') === 0) {
        return 'Pig';
      }
      if (this.get('id').indexOf('hive_') === 0) {
        return 'Hive';
      }
      if (this.get('id').indexOf('mr_') === 0) {
        return 'MapReduce';
      }
      return 'Undefined';
    }.property('id')
  });
});
window.require.register("config", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.version = '1.5.1.110'; // filled out by set-ambari-version.sh script
  App.testMode = (location.port == '3333'); // test mode is automatically enabled if running on brunch server
  App.testModeDelayForActions = 10000;
  App.skipBootstrap = false;
  App.alwaysGoToInstaller = false;
  App.testEnableSecurity = true; // By default enable security is tested; turning it false tests disable security
  App.testNameNodeHA = true;
  App.apiPrefix = '/api/v1';
  App.defaultStackVersion = 'HDP-2.0.5';
  App.defaultLocalStackVersion = 'HDPLocal-2.0.5';
  App.defaultJavaHome = '/usr/jdk/jdk1.6.0_31';
  App.timeout = 180000; // default AJAX timeout
  App.maxRetries = 3; // max number of retries for certain AJAX calls
  App.bgOperationsUpdateInterval = 6000;
  App.componentsUpdateInterval = 6000;
  App.contentUpdateInterval = 15000;
  App.maxRunsForAppBrowser = 500;
  App.pageReloadTime=3600000;
  App.singleNodeInstall = false;
  App.singleNodeAlias = document.location.hostname;
  App.minDiskSpace = 2.0; // minimum disk space required for '/' for each host before install, unit GB
  App.minDiskSpaceUsrLib = 1.0; // minimum disk space for '/usr/lib' for each host before install, unit GB
  App.mirroringDatasetNamePrefix = 'ambari-mirroring-'; // special prefix to be added for each Falcon feed name created with Ambari
  App.healthIconClassGreen = 'icon-ok-sign'; // bootstrap icon class for healthy/started service/host/host-component
  App.healthIconClassRed = 'icon-warning-sign'; // bootstrap icon class for master down/stopped service/host/host-component
  App.healthIconClassOrange = 'icon-minus-sign'; // bootstrap icon class for slave down/decommissioned host/host-component
  App.healthIconClassYellow = 'icon-question-sign'; // bootstrap icon class for heartbeat lost service/host/host-component

  // experimental features are automatically enabled if running on brunch server
  App.enableExperimental = false;

  App.supports = {
    addServices: true,
    hostOverrides: true,
    hostOverridesInstaller: true,
    hostOverridesHost: true,
    mirroring: false,
    secureCluster: true,
    secureClusterProceedPopup: false,
    reassignMaster: true,
    stackUpgrade: false,
    capacitySchedulerUi: false,
    startStopAllServices: true,
    hiveOozieExtraDatabases: true,
    multipleHBaseMasters: true,
    addMasters: false,
    customizeSmokeTestUser: true,
    hue: false,
    falcon: true,
    ldapGroupMapping: false,
    localRepositories: true,
    highAvailability: true,
    deleteHost: true,
    autoRollbackHA: false,
    appTimelineServer: true,
    jobs: true,
    ubuntu: false,
    views: false
  };

  if (App.enableExperimental) {
    for (var support in App.supports) {
      App.supports[support] = true;
    }
  }

  // this is to make sure that IE does not cache data when making AJAX calls to the server
  if (!$.mocho) {
    $.ajaxSetup({
      cache: false,
      headers: {"X-Requested-By": "X-Requested-By"}
    });
  }

  /**
   * Test Mode values
   */
  App.test_hostname = 'hostname';
  
});
window.require.register("controllers", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  // load all controllers here

  require('controllers/application');
  require('controllers/login_controller');
  require('controllers/wizard');
  require('controllers/installer');
  require('controllers/global/background_operations_controller');
  require('controllers/main');
  require('controllers/main/admin');
  require('controllers/main/admin/highAvailability_controller');
  require('controllers/main/admin/highAvailability/wizard_controller');
  require('controllers/main/admin/highAvailability/progress_controller');
  require('controllers/main/admin/highAvailability/rollback_controller');
  require('controllers/main/admin/highAvailability/step1_controller');
  require('controllers/main/admin/highAvailability/step2_controller');
  require('controllers/main/admin/highAvailability/step3_controller');
  require('controllers/main/admin/highAvailability/step4_controller');
  require('controllers/main/admin/highAvailability/step5_controller');
  require('controllers/main/admin/highAvailability/step6_controller');
  require('controllers/main/admin/highAvailability/step7_controller');
  require('controllers/main/admin/highAvailability/step8_controller');
  require('controllers/main/admin/highAvailability/step9_controller');
  require('controllers/main/admin/rollbackHA/step1_controller');
  require('controllers/main/admin/rollbackHA/step2_controller');
  require('controllers/main/admin/rollbackHA/step3_controller');
  require('controllers/main/admin/rollbackHA/rollback_wizard_controller');
  require('controllers/main/admin/cluster');
  require('controllers/main/admin/stack_upgrade_controller');
  require('controllers/main/admin/user');
  require('controllers/main/admin/misc_controller');
  require('controllers/main/admin/access_controller');
  require('controllers/main/admin/user/edit');
  require('controllers/main/admin/user/create');
  require('controllers/main/admin/advanced');
  require('utils/polling');
  require('controllers/main/admin/security');
  require('controllers/main/admin/security/security_progress_controller');
  require('controllers/main/admin/security/disable');
  require('controllers/main/admin/security/add/addSecurity_controller');
  require('controllers/main/admin/security/add/step1');
  require('controllers/main/admin/security/add/step2');
  require('controllers/main/admin/security/add/step3');
  require('controllers/main/admin/security/add/step4');
  require('controllers/main/admin/authentication');
  require('controllers/main/alerts_controller');
  require('controllers/main/service');
  require('controllers/main/service/item');
  require('controllers/main/service/info/summary');
  require('controllers/main/service/info/configs');
  require('controllers/main/service/info/audit');
  require('controllers/main/service/add_controller');
  require('controllers/main/service/reassign_controller');
  require('controllers/main/service/reassign/step1_controller');
  require('controllers/main/service/reassign/step2_controller');
  require('controllers/main/service/reassign/step3_controller');
  require('controllers/main/service/reassign/step4_controller');
  require('controllers/main/service/reassign/step5_controller');
  require('controllers/main/service/reassign/step6_controller');
  require('controllers/main/service/manage_config_groups_controller');
  require('controllers/main/host');
  require('controllers/main/host/details');
  require('controllers/main/host/configs_service');
  require('controllers/main/host/add_controller');
  require('controllers/main/host/addHost/step4_controller');
  require('controllers/main/charts');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_processrun');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_diskspaceused');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_cpuWaitIO');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_memoryused');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_bytesread');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_byteswritten');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_gctime');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_memHeapUsed');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_gctime');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_mapsRunning');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_reducesRunning');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_memHeapUsed');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_gctime');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_memHeapUsed');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_ResourceUsed');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_readrequest');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_writerequest');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_compactionqueue');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_regions');
  require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_memstoresize');
  require('controllers/main/charts/heatmap');
  require('controllers/main/apps_controller');
  require('controllers/main/jobs_controller');
  require('controllers/main/jobs/hive_job_details_controller');
  require('controllers/main/apps/item_controller');
  require('controllers/main/mirroring_controller');
  require('controllers/main/mirroring/edit_dataset_controller');
  require('controllers/main/mirroring/datasets_controller');
  require('controllers/main/mirroring/jobs_controller');
  require('controllers/main/mirroring/manage_clusters_controller');
  require('controllers/main/views');
  require('controllers/wizard/slave_component_groups_controller');
  require('controllers/wizard/step0_controller');
  require('controllers/wizard/step1_controller');
  require('controllers/wizard/step2_controller');
  require('controllers/wizard/step3_controller');
  require('controllers/wizard/step4_controller');
  require('controllers/wizard/step5_controller');
  require('controllers/wizard/step6_controller');
  require('controllers/wizard/step7_controller');
  require('controllers/wizard/step8_controller');
  require('controllers/wizard/step9_controller');
  require('controllers/wizard/step10_controller');
  require('controllers/wizard/stack_upgrade/step1_controller');
  require('controllers/wizard/stack_upgrade/step2_controller');
  require('controllers/wizard/stack_upgrade/step3_controller');
  require('controllers/global/cluster_controller');
  require('controllers/global/update_controller');
  require('controllers/global/configuration_controller');
  
});
window.require.register("controllers/application", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ApplicationController = Em.Controller.extend(App.UserPref, {

    name: 'applicationController',

    clusterName: function () {
      return (App.router.get('clusterController.clusterName') || 'My Cluster');
    }.property('App.router.clusterController.clusterName'),

    ambariVersion: function () {
      return (App.router.get('clusterController.ambariVersion') || Em.I18n.t('common.notAvailable'));
    }.property('App.router.clusterController.ambariVersion'),

    clusterDisplayName: function () {
      var name = this.get('clusterName');
      return name.length > 13 ? name.substr(0, 10) + "..." : name;
    }.property('clusterName'),

    isClusterDataLoaded: function() {
      return App.router.get('clusterController.isLoaded') && App.router.get('loggedIn');
    }.property('App.router.clusterController.isLoaded','App.router.loggedIn'),

    init: function(){
      this._super();
    },

    dataLoading: function () {
      var dfd = $.Deferred();
      var self = this;
      this.getUserPref(this.persistKey()).done(function () {
        var curPref = self.get('currentPrefObject');
        self.set('currentPrefObject', null);
        dfd.resolve(curPref);
      });
      return dfd.promise();
    },
    persistKey: function (loginName) {
      if (App.get('testMode')) {
        return 'admin_settings_show_bg';
      }
      if (!loginName)
        loginName = App.router.get('loginName');
      return 'admin-settings-show-bg-' + loginName;
    },
    currentPrefObject: null,

    getUserPrefSuccessCallback: function (response, request, data) {
      if (response != null) {
        console.log('Got persist value from server with key ' + data.key + '. Value is: ' + response);
        this.set('currentPrefObject', response);
        return response;
      }
    },
    getUserPrefErrorCallback: function (request, ajaxOptions, error) {
      // this user is first time login
      if (request.status == 404) {
        console.log('Persist did NOT find the key');
        this.set('currentPrefObject', true);
        this.postUserPref(this.persistKey(), true);
        return true;
      }
    },

    showSettingsPopup: function() {
      // Settings only for admins
      if (!App.get('isAdmin')) return;

      var self = this;
      var curValue = null;
      this.dataLoading().done(function (initValue) {
        App.ModalPopup.show({
          header: Em.I18n.t('common.userSettings'),
          bodyClass: Em.View.extend({
            templateName: require('templates/common/settings'),
            isNotShowBgChecked: !initValue,
            updateValue: function () {
              curValue = !this.get('isNotShowBgChecked');
            }.observes('isNotShowBgChecked')
          }),
          primary: Em.I18n.t('common.save'),
          onPrimary: function() {
            if (curValue == null) {
              curValue = initValue;
            }
            var key = self.persistKey();
            if (!App.testMode) {
              self.postUserPref(key, curValue);
            }
            this.hide();
          }
        })
      });
    },

    showAboutPopup: function() {

      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('common.aboutAmbari'),
        secondary: false,
        bodyClass: Em.View.extend({
          templateName: require('templates/common/about'),
          ambariVersion: this.get('ambariVersion')
        })
      });    
    }

  });
});
window.require.register("controllers/global/background_operations_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.BackgroundOperationsController = Em.Controller.extend({
    name: 'backgroundOperationsController',

    /**
     * Whether we need to refresh background operations or not
     */
    isWorking : false,

    allOperationsCount : 0,

    /**
     * For host component popup
     */
    services:[],
    serviceTimestamp: null,
    /**
     * Possible levels:
     * REQUESTS_LIST
     * HOSTS_LIST
     * TASKS_LIST
     * TASK_DETAILS
     */
    levelInfo: Em.Object.create({
      name: 'REQUESTS_LIST',
      requestId: null,
      taskId: null,
      sync: false
    }),

    /**
     * Start polling, when <code>isWorking</code> become true
     */
    startPolling: function(){
      if(this.get('isWorking')){
        this.requestMostRecent();
        App.updater.run(this, 'requestMostRecent', 'isWorking', App.bgOperationsUpdateInterval);
      }
    }.observes('isWorking'),

    /**
     * Get requests data from server
     * @param callback
     */
    requestMostRecent: function (callback) {
      var queryParams = this.getQueryParams();
      App.ajax.send({
        'name': queryParams.name,
        'sender': this,
        'success': queryParams.successCallback,
        'callback': callback,
        'data': queryParams.data
      });
    },
    /**
     * construct params of ajax query regarding displayed level
     */
    getQueryParams: function () {
      var levelInfo = this.get('levelInfo');
      var result = {
        name: 'background_operations.get_most_recent',
        successCallback: 'callBackForMostRecent',
        data: {}
      };
      if (levelInfo.get('name') === 'TASK_DETAILS' && !App.testMode) {
        result.name = 'background_operations.get_by_task';
        result.successCallback = 'callBackFilteredByTask';
        result.data = {
          'taskId': levelInfo.get('taskId'),
          'requestId': levelInfo.get('requestId'),
          'sync': levelInfo.get('sync')
        };
      } else if (levelInfo.get('name') === 'TASKS_LIST' || levelInfo.get('name') === 'HOSTS_LIST') {
        result.name = 'background_operations.get_by_request';
        result.successCallback = 'callBackFilteredByRequest';
        result.data = {
          'requestId': levelInfo.get('requestId'),
          'sync': levelInfo.get('sync')
        };
      }
      levelInfo.set('sync', false);
      return result;
    },

    /**
     * Push hosts and their tasks to request
     * @param data
     * @param ajaxQuery
     * @param params
     */
    callBackFilteredByRequest: function (data, ajaxQuery, params) {
      var requestId = data.Requests.id;
      var request = this.get('services').findProperty('id', requestId);
      var hostsMap = {};
      var previousTaskStatusMap = request.get('previousTaskStatusMap');
      var currentTaskStatusMap = {};
      data.tasks.forEach(function (task) {
        var host = hostsMap[task.Tasks.host_name];
        task.Tasks.request_id = requestId;
        if (host) {
          host.logTasks.push(task);
          host.isModified = (host.isModified) ? true : previousTaskStatusMap[task.Tasks.id] !== task.Tasks.status;
        } else {
          hostsMap[task.Tasks.host_name] = {
            name: task.Tasks.host_name,
            publicName: task.Tasks.host_name,
            logTasks: [task],
            isModified: previousTaskStatusMap[task.Tasks.id] !== task.Tasks.status
          };
        }
        currentTaskStatusMap[task.Tasks.id] = task.Tasks.status;
      }, this);
      /**
       * sync up request progress with up to date progress of hosts on Host's list,
       * to avoid discrepancies while waiting for response with latest progress of request
       * after switching to operation's list
       */
      if (request.get('isRunning')) {
        request.set('progress', App.HostPopup.getProgress(data.tasks));
        request.set('status', App.HostPopup.getStatus(data.tasks)[0]);
        request.set('isRunning', (request.get('progress') !== 100));
      }
      request.set('previousTaskStatusMap', currentTaskStatusMap);
      request.set('hostsMap', hostsMap);
      this.set('serviceTimestamp', App.dateTime());
    },
    /**
     * Update task, with uploading two additional properties: stdout and stderr
     * @param data
     * @param ajaxQuery
     * @param params
     */
    callBackFilteredByTask: function (data, ajaxQuery, params) {
      var request = this.get('services').findProperty('id', data.Tasks.request_id);
      var host = request.get('hostsMap')[data.Tasks.host_name];
      var task = host.logTasks.findProperty('Tasks.id', data.Tasks.id);
      task.Tasks.status = data.Tasks.status;
      task.Tasks.stdout = data.Tasks.stdout;
      task.Tasks.stderr = data.Tasks.stderr;
      this.set('serviceTimestamp', App.dateTime());
    },

    /**
     * Prepare, received from server, requests for host component popup
     * @param data
     */
    callBackForMostRecent: function (data) {
      var runningServices = 0;
      var self = this;
      var currentRequestIds = [];

      data.items.forEach(function (request) {
        var rq = self.get("services").findProperty('id', request.Requests.id);
        var isRunning = (request.Requests.task_count -
          (request.Requests.aborted_task_count + request.Requests.completed_task_count + request.Requests.failed_task_count
           + request.Requests.timed_out_task_count - request.Requests.queued_task_count)) > 0;
        var requestParams = this.parseRequestContext(request.Requests.request_context);
        var inputs = null;
        if (request.Requests.inputs) {
          inputs = JSON.parse(request.Requests.inputs);
        }
        var oneHost = false;
        if (inputs && inputs.included_hosts) {
          var hosts = inputs.included_hosts.split(',');
          oneHost = (hosts.length < 2);
        }
        if(request.Requests.request_schedule && oneHost && /Recommission/.test(requestParams.requestContext)){
          request.Requests.request_schedule.schedule_id = null;
        }
        currentRequestIds.push(request.Requests.id);
        if (rq) {
          rq.set('progress', Math.ceil(request.Requests.progress_percent));
          rq.set('status', request.Requests.request_status);
          rq.set('isRunning', isRunning);
          rq.set('startTime', request.Requests.start_time);
          rq.set('endTime', request.Requests.end_time);
        } else {
          rq = Em.Object.create({
            id: request.Requests.id,
            name: requestParams.requestContext,
            displayName: requestParams.requestContext,
            progress: Math.floor(request.Requests.progress_percent),
            status: request.Requests.request_status,
            isRunning: isRunning,
            hostsMap: {},
            tasks: [],
            startTime: request.Requests.start_time,
            endTime: request.Requests.end_time,
            dependentService: requestParams.dependentService,
            sourceRequestScheduleId: request.Requests.request_schedule && request.Requests.request_schedule.schedule_id,
            previousTaskStatusMap: {},
            contextCommand: requestParams.contextCommand
          });
          self.get("services").unshift(rq);
        }
        runningServices += ~~isRunning;
      }, this);
      //remove old request if it's absent in API response
      self.get('services').forEach(function(service, index, services){
        if(!currentRequestIds.contains(service.id)) {
          services.splice(index, 1);
        }
      });
      self.set("allOperationsCount", runningServices);
      self.set('serviceTimestamp', App.dateTime());
    },

    /**
     * parse request context and if keyword "_PARSE_" is present then format it
     * @param requestContext
     * @return {Object}
     */
    parseRequestContext: function (requestContext) {
      var parsedRequestContext;
      var service;
      var contextCommand;
      if (requestContext) {
        if (requestContext.indexOf(App.BackgroundOperationsController.CommandContexts.PREFIX) !== -1) {
          var contextSplits = requestContext.split('.');
          contextCommand = contextSplits[1];
          service = contextSplits[2];
          switch(contextCommand){
          case "STOP":
          case "START":
            if (service === 'ALL_SERVICES') {
              parsedRequestContext = Em.I18n.t("requestInfo." + contextCommand.toLowerCase()).format(Em.I18n.t('common.allServices'));
            } else {
              parsedRequestContext = Em.I18n.t("requestInfo." + contextCommand.toLowerCase()).format(App.Service.DisplayNames[service]);
            }
            break;
          case "ROLLING-RESTART":
            parsedRequestContext = Em.I18n.t("rollingrestart.rest.context").format(App.format.role(service), contextSplits[3], contextSplits[4]);
            break;
          }
        } else {
          parsedRequestContext = requestContext;
        }
      } else {
        parsedRequestContext = Em.I18n.t('requestInfo.unspecified');
      }
      return {
        requestContext: parsedRequestContext,
        dependentService: service,
        contextCommand: contextCommand
      }
    },

    popupView: null,

    /**
     * Onclick handler for background operations number located right to logo
     */
    showPopup: function(){
      // load the checkbox on footer first, then show popup.
      var self = this;
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        App.updater.immediateRun('requestMostRecent');
        if(self.get('popupView') && App.HostPopup.get('isBackgroundOperations')){
          self.set ('popupView.isNotShowBgChecked', !initValue);
          self.set('popupView.isOpen', true);
          $(self.get('popupView.element')).appendTo('#wrapper');
        } else {
          self.set('popupView', App.HostPopup.initPopup("", self, true));
          self.set ('popupView.isNotShowBgChecked', !initValue);
        }
      });
    }

  });

  /**
   * Each background operation has a context in which it operates.
   * Generally these contexts are fixed messages. However, we might
   * want to associate semantics to this context - like showing, disabling
   * buttons when certain operations are in progress.
   *
   * To make this possible we have command contexts where the context
   * is not a human readable string, but a pattern indicating the command
   * it is running. When UI shows these, they are translated into human
   * readable strings.
   *
   * General pattern of context names is "_PARSE_.{COMMAND}.{ID}[.{Additional-Data}...]"
   */
  App.BackgroundOperationsController.CommandContexts = {
    PREFIX : "_PARSE_",
    /**
     * Stops all services
     */
    STOP_ALL_SERVICES : "_PARSE_.STOP.ALL_SERVICES",
    /**
     * Starts all services
     */
    START_ALL_SERVICES : "_PARSE_.START.ALL_SERVICES",
    /**
     * Starts service indicated by serviceID.
     * @param {String} serviceID Parameter {0}. Example: HDFS
     */
    START_SERVICE : "_PARSE_.START.{0}",
    /**
     * Stops service indicated by serviceID.
     * @param {String} serviceID Parameter {0}. Example: HDFS
     */
    STOP_SERVICE : "_PARSE_.STOP.{0}",
    /**
     * Performs rolling restart of componentID in batches.
     * This context is the batchNumber batch out of totalBatchCount batches.
     * @param {String} componentID Parameter {0}. Example "DATANODE"
     * @param {Number} batchNumber Parameter {1}. Batch number of this batch. Example 3.
     * @param {Number} totalBatchCount Parameter {2}. Total number of batches. Example 10.
     */
    ROLLING_RESTART : "_PARSE_.ROLLING-RESTART.{0}.{1}.{2}"
  }
  
});
window.require.register("controllers/global/cluster_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ClusterController = Em.Controller.extend({
    name:'clusterController',
    cluster:null,
    isLoaded:false,
    ambariProperties: null,
    ambariVersion: null,
    clusterDataLoadedPercent: 'width:0', // 0 to 1
    /**
     * Whether we need to update statuses automatically or not
     */
    isWorking: false,
    updateLoadStatus:function (item) {
      var loadList = this.get('dataLoadList');
      var loaded = true;
      var numLoaded = 0;
      var loadListLength = 0;
      loadList.set(item, true);
      for (var i in loadList) {
        if (loadList.hasOwnProperty(i)) {
          loadListLength++;
          if(!loadList[i] && loaded){
            loaded = false;
          }
        }
        // calculate the number of true
        if (loadList.hasOwnProperty(i) && loadList[i]){
          numLoaded++;
        }
      }
      this.set('isLoaded', loaded);
      this.set('clusterDataLoadedPercent', 'width:' + (Math.floor(numLoaded / loadListLength * 100)).toString() + '%');
    },

    dataLoadList:Em.Object.create({
      'hosts':false,
      'serviceMetrics':false,
      'services': false,
      'cluster':false,
      'clusterStatus':false,
      'racks':false,
      'users':false,
      'componentConfigs': false
    }),

    /**
     * load cluster name
     */
    loadClusterName:function (reload) {
      if (this.get('clusterName') && !reload) {
        return;
      }

      App.ajax.send({
        name: 'cluster.load_cluster_name',
        sender: this,
        success: 'loadClusterNameSuccessCallback',
        error: 'loadClusterNameErrorCallback'
      });

      if(!App.get('currentStackVersion')){
        App.set('currentStackVersion', App.defaultStackVersion);
      }
    },

    loadClusterNameSuccessCallback: function (data) {
      this.set('cluster', data.items[0]);
      App.set('clusterName', data.items[0].Clusters.cluster_name);
      App.set('currentStackVersion', data.items[0].Clusters.version);
    },

    loadClusterNameErrorCallback: function (request, ajaxOptions, error) {
      console.log('failed on loading cluster name');
      this.set('isLoaded', true);
    },

    /**
     * load current server clock in milli-seconds
     */
    loadClientServerClockDistance: function () {
      var dfd = $.Deferred();
      this.getServerClock().done(function () {
        dfd.resolve();
      });
      return dfd.promise();
    },

    getServerClock: function(){
      return App.ajax.send({
        name: 'ambari.service.load_server_clock',
        sender: this,
        success: 'getServerClockSuccessCallback',
        error: 'getServerClockErrorCallback'
      });
    },
    getServerClockSuccessCallback: function (data) {
      var clientClock = new Date().getTime();
      var serverClock = (data.RootServiceComponents.server_clock).toString();
      serverClock = serverClock.length < 13? serverClock+ '000': serverClock;
      App.set('clockDistance', serverClock - clientClock);
      App.set('currentServerTime', parseInt(serverClock));
      console.log('loading ambari server clock distance');
    },
    getServerClockErrorCallback: function () {
      console.log('Cannot load ambari server clock');
    },

    getUrl:function (testUrl, url) {
      return (App.testMode) ? testUrl : App.apiPrefix + '/clusters/' + this.get('clusterName') + url;
    },

    /**
     * Provides the URL to use for Ganglia server. This URL
     * is helpful in populating links in UI.
     *
     * If null is returned, it means GANGLIA service is not installed.
     */
    gangliaUrl: function () {
      if (App.testMode) {
        return 'http://gangliaserver/ganglia/?t=yes';
      } else {
        // We want live data here
        var svcs = App.Service.find();
        var gangliaSvc = svcs.findProperty("serviceName", "GANGLIA");
        if (gangliaSvc) {
          var svcComponents = gangliaSvc.get('hostComponents');
          if (svcComponents) {
            var gangliaSvcComponent = svcComponents.findProperty("componentName", "GANGLIA_SERVER");
            if (gangliaSvcComponent) {
              var hostName = gangliaSvcComponent.get('host.hostName');
              if (hostName) {
                var host = App.Host.find(hostName);
                if (host) {
                  hostName = host.get('publicHostName');
                }
                return this.get('gangliaWebProtocol') + "://" + (App.singleNodeInstall ? App.singleNodeAlias + ":42080" : hostName) + "/ganglia";
              }
            }
          }
        }
        return null;
      }
    }.property('App.router.updateController.isUpdated', 'dataLoadList.hosts','gangliaWebProtocol'),

    /**
     * Provides the URL to use for NAGIOS server. This URL
     * is helpful in getting alerts data from server and also
     * in populating links in UI.
     *
     * If null is returned, it means NAGIOS service is not installed.
     */
    nagiosUrl:function () {
      if (App.testMode) {
        return 'http://nagiosserver/nagios';
      } else {
        // We want live data here
        var svcs = App.Service.find();
        var nagiosSvc = svcs.findProperty("serviceName", "NAGIOS");
        if (nagiosSvc) {
          var svcComponents = nagiosSvc.get('hostComponents');
          if (svcComponents) {
            var nagiosSvcComponent = svcComponents.findProperty("componentName", "NAGIOS_SERVER");
            if (nagiosSvcComponent) {
              var hostName = nagiosSvcComponent.get('host.hostName');
              if (hostName) {
                var host = App.Host.find(hostName);
                if (host) {
                  hostName = host.get('publicHostName');
                }
                return this.get('nagiosWebProtocol') + "://" + (App.singleNodeInstall ? App.singleNodeAlias + ":42080" : hostName) + "/nagios";
              }
            }
          }
        }
        return null;
      }
    }.property('App.router.updateController.isUpdated', 'dataLoadList.serviceMetrics', 'dataLoadList.hosts','nagiosWebProtocol'),

    nagiosWebProtocol: function () {
      var properties = this.get('ambariProperties');
      if (properties && properties.hasOwnProperty('nagios.https') && properties['nagios.https']) {
        return "https";
      } else {
        return "http";
      }
    }.property('ambariProperties'),

    gangliaWebProtocol: function () {
      var properties = this.get('ambariProperties');
      if (properties && properties.hasOwnProperty('ganglia.https') && properties['ganglia.https']) {
        return "https";
      } else {
        return "http";
      }
    }.property('ambariProperties'),

    isNagiosInstalled:function () {
      return !!App.Service.find().findProperty('serviceName', 'NAGIOS');
    }.property('App.router.updateController.isUpdated', 'dataLoadList.serviceMetrics'),

    isGangliaInstalled:function () {
      return !!App.Service.find().findProperty('serviceName', 'GANGLIA');
    }.property('App.router.updateController.isUpdated', 'dataLoadList.serviceMetrics'),

    /**
     * Send request to server to load components updated statuses
     * @param callback Slave function, should be called to fire delayed update.
     * @param isInitialLoad
     * Look at <code>App.updater.run</code> for more information
     * @return {Boolean} Whether we have errors
     */
    loadUpdatedStatus: function (callback, isInitialLoad) {
      if (!this.get('clusterName')) {
        callback();
        return false;
      }
      App.set('currentServerTime', App.get('currentServerTime') + App.componentsUpdateInterval);
      var testUrl = App.get('isHadoop2Stack') ? '/data/hosts/HDP2/hc_host_status.json' : '/data/dashboard/services.json';
      var statusUrl = '/hosts?fields=Hosts/host_status,Hosts/maintenance_state,host_components/HostRoles/state,host_components/HostRoles/maintenance_state,alerts/summary&minimal_response=true';
      if (isInitialLoad) {
        testUrl = '/data/hosts/HDP2/hosts_init.json';
        statusUrl = '/hosts?fields=Hosts/host_name,Hosts/maintenance_state,Hosts/public_host_name,Hosts/cpu_count,Hosts/ph_cpu_count,Hosts/total_mem,' +
          'Hosts/host_status,Hosts/last_heartbeat_time,Hosts/os_arch,Hosts/os_type,Hosts/ip,host_components/HostRoles/state,host_components/HostRoles/maintenance_state,' +
          'Hosts/disk_info,metrics/disk,metrics/load/load_one,metrics/cpu/cpu_system,metrics/cpu/cpu_user,' +
          'metrics/memory/mem_total,metrics/memory/mem_free,alerts/summary&minimal_response=true';
      }
      //desired_state property is eliminated since calculateState function is commented out, it become useless
      statusUrl = this.getUrl(testUrl, statusUrl);

      App.HttpClient.get(statusUrl, App.statusMapper, {
        complete: callback
      });
      return true;
    },

    /**
     * Run <code>loadUpdatedStatus</code> with delay
     * @param delay
     */
    loadUpdatedStatusDelayed: function(delay){
      setTimeout(function(){
        App.updater.immediateRun('loadUpdatedStatus');
      }, delay);
    },

    /**
     * Start polling, when <code>isWorking</code> become true
     */
    startPolling: function(){
      if(!this.get('isWorking')){
        return false;
      }
      App.updater.run(this, 'loadUpdatedStatus', 'isWorking', App.componentsUpdateInterval); //update will not run it immediately
      return true;
    }.observes('isWorking'),
    /**
     *
     *  load all data and update load status
     */
    loadClusterData:function () {
      var self = this;
      this.loadAmbariProperties();
      if (!this.get('clusterName')) {
        return;
      }

      if(this.get('isLoaded')) { // do not load data repeatedly
        App.router.get('mainController').startPolling();
        return;
      }
      var clusterUrl = this.getUrl('/data/clusters/cluster.json', '?fields=Clusters');
      var usersUrl = App.testMode ? '/data/users/users.json' : App.apiPrefix + '/users/?fields=*';
      var racksUrl = "/data/racks/racks.json";

      App.HttpClient.get(racksUrl, App.racksMapper, {
        complete:function (jqXHR, textStatus) {
          self.updateLoadStatus('racks');
        }
      }, function (jqXHR, textStatus) {
        self.updateLoadStatus('racks');
      });

      App.HttpClient.get(clusterUrl, App.clusterMapper, {
        complete:function (jqXHR, textStatus) {
          self.updateLoadStatus('cluster');
        }
      }, function (jqXHR, textStatus) {
          self.updateLoadStatus('cluster');
      });

      if (App.testMode) {
        self.updateLoadStatus('clusterStatus');
      } else {
        App.clusterStatus.updateFromServer(true).complete(function() {
          self.updateLoadStatus('clusterStatus');
        });
      }
      
      App.HttpClient.get(usersUrl, App.usersMapper, {
        complete:function (jqXHR, textStatus) {
          self.updateLoadStatus('users');
        }
      }, function (jqXHR, textStatus) {
          self.updateLoadStatus('users');
      });

      /**
       * Order of loading:
       * 1. request for services
       * 2. put services in cache
       * 3. request for hosts and host-components (single call)
       * 4. request for service metrics
       * 5. load host-components to model
       * 6. load hosts to model
       * 7. load services from cache with metrics to model
       * 8. update stale_configs of host-components (depends on App.supports.hostOverrides)
       */
      App.router.get('updateController').updateServices(function () {
        self.updateLoadStatus('services');
        self.loadUpdatedStatus(function () {
          self.updateLoadStatus('hosts');
          if (App.supports.hostOverrides) {
            App.router.get('updateController').updateComponentConfig(function () {
              self.updateLoadStatus('componentConfigs');
            });
          } else {
            self.updateLoadStatus('componentConfigs');
          }
        }, true);
        App.router.get('updateController').updateServiceMetric(function () {}, true);
      });
    },
    /**
     * json from serviceMetricsMapper on initial load
     */
    serviceMetricsJson: null,
    /**
     * control that services was loaded to model strictly after hosts and host-components
     * regardless which request was completed first
     * @param json
     */
    deferServiceMetricsLoad: function (json) {
      if (json) {
        if (this.get('dataLoadList.hosts')) {
          App.serviceMetricsMapper.map(json, true);
          this.updateLoadStatus('serviceMetrics');
        } else {
          this.set('serviceMetricsJson', json);
        }
      } else if (this.get('serviceMetricsJson')) {
        json = this.get('serviceMetricsJson');
        this.set('serviceMetricsJson', null);
        App.serviceMetricsMapper.map(json, true);
        this.updateLoadStatus('serviceMetrics');
      }
    },

    requestHosts: function(realUrl, callback){
      var testHostUrl =  App.get('isHadoop2Stack') ? '/data/hosts/HDP2/hosts.json':'/data/hosts/hosts.json';
      var url = this.getUrl(testHostUrl, realUrl);
      App.HttpClient.get(url, App.hostsMapper, {
        complete: callback
      }, callback)
    },

    loadAmbariProperties: function() {
      App.ajax.send({
        name: 'ambari.service',
        sender: this,
        success: 'loadAmbariPropertiesSuccess',
        error: 'loadAmbariPropertiesError'
      });
      return this.get('ambariProperties');
    },

    loadAmbariPropertiesSuccess: function(data) {
      console.log('loading ambari properties');
      this.set('ambariProperties', data.RootServiceComponents.properties);
      this.set('ambariVersion', data.RootServiceComponents.component_version);
    },

    loadAmbariPropertiesError: function() {
      console.warn('can\'t get ambari properties');
    },

    clusterName:function () {
      return (this.get('cluster')) ? this.get('cluster').Clusters.cluster_name : null;
    }.property('cluster'),
    
    updateClusterData: function () {
      var testUrl = App.get('isHadoop2Stack') ? '/data/clusters/HDP2/cluster.json':'/data/clusters/cluster.json';
      var clusterUrl = this.getUrl(testUrl, '?fields=Clusters');
      App.HttpClient.get(clusterUrl, App.clusterMapper, {
        complete:function(){}
      });
    }
  });
  
});
window.require.register("controllers/global/configuration_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ConfigurationController = Em.Controller.extend({
    name: 'configurationController',

    getConfigsByTags: function (tags) {
      var storedTags = [];
      App.db.getConfigs().forEach(function (site) {
        storedTags.push({
          siteName: site.type,
          tagName: site.tag
        })
      });
      if (this.checkTagsChanges(tags, storedTags)) {
        return this.loadFromServer(tags);
      } else {
        return this.loadFromDB(tags.mapProperty('siteName'));
      }
    },
    /**
     * check whether tag versions have been changed
     * if they are different then return true
     * otherwise false
     * @param tags
     * @param storedTags
     * @return {Boolean}
     */
    checkTagsChanges: function (tags, storedTags) {
      var isDifferent = false;
      var i = 0;
      while (i < tags.length && !isDifferent) {
        var storedTag = storedTags.findProperty('siteName', tags[i].siteName);
        isDifferent = (!storedTag || storedTag.tagName !== tags[i].tagName);
        i++;
      }
      return isDifferent;
    },
    loadFromDB: function (siteNames) {
      var configs = App.db.getConfigs();
      return configs.filter(function (site) {
        return (siteNames.contains(site.type));
      })
    },
    /**
     * load configs from server
     * and update them in local DB
     * @param tags
     * @return {Array}
     */
    loadFromServer: function (tags) {
      var loadedConfigs = App.config.loadConfigsByTags(tags);
      var storedConfigs = App.db.getConfigs();
      loadedConfigs.forEach(function (loadedSite) {
        var storedSite = storedConfigs.findProperty('type', loadedSite.type);
        if (storedSite) {
          storedSite.tag = loadedSite.tag;
          storedSite.properties = loadedSite.properties;
        } else {
          storedConfigs.push(loadedSite);
        }
      });
      App.db.setConfigs(storedConfigs);
      return loadedConfigs;
    }


  });
  
});
window.require.register("controllers/global/update_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.UpdateController = Em.Controller.extend({
    name:'updateController',
    isUpdated:false,
    cluster:null,
    isWorking: false,
    timeIntervalId: null,
    clusterName:function () {
      return App.router.get('clusterController.clusterName');
    }.property('App.router.clusterController.clusterName'),

    getUrl:function (testUrl, url) {
      return (App.testMode) ? testUrl : App.apiPrefix + '/clusters/' + this.get('clusterName') + url;
    },

    /**
     * Start polling, when <code>isWorking</code> become true
     */
    updateAll:function(){
      if(this.get('isWorking')) {
        App.updater.run(this, 'updateServices', 'isWorking');
        App.updater.run(this, 'updateHostConditionally', 'isWorking');
        App.updater.run(this, 'updateServiceMetricConditionally', 'isWorking', App.componentsUpdateInterval);
        App.updater.run(this, 'graphsUpdate', 'isWorking');
        if (App.supports.hostOverrides) {
          App.updater.run(this, 'updateComponentConfig', 'isWorking');
        }
      }
    }.observes('isWorking'),
    /**
     * Update hosts depending on which page is open
     * Make a call only on follow pages:
     * /main/hosts
     * /main/hosts/*
     * /main/charts/heatmap
     * @param callback
     */
    updateHostConditionally: function (callback) {
      var location = App.router.get('location.lastSetURL');
      if (/\/main\/(hosts|charts\/heatmap).*/.test(location)) {
        this.updateHost(callback);
      } else {
        callback();
      }
    },
    /**
     * Update service metrics depending on which page is open
     * Make a call only on follow pages:
     * /main/dashboard
     * /main/services/*
     * @param callback
     */
    updateServiceMetricConditionally: function(callback){
      var location = App.router.get('location.lastSetURL');
      if (/\/main\/(dashboard|services).*/.test(location)) {
        this.updateServiceMetric(callback);
      } else {
        callback();
      }
    },

    updateHost:function(callback) {
      var testUrl = App.get('isHadoop2Stack') ? '/data/hosts/HDP2/hosts.json' : '/data/hosts/hosts.json';
      var hostsUrl = this.getUrl(testUrl, '/hosts?fields=Hosts/host_name,Hosts/host_status,Hosts/last_heartbeat_time,Hosts/disk_info,Hosts/maintenance_state,' +
        'metrics/disk,metrics/load/load_one,metrics/cpu/cpu_system,metrics/cpu/cpu_user,metrics/memory/mem_total,metrics/memory/mem_free'+
        '&minimal_response=true');
      App.HttpClient.get(hostsUrl, App.hostsMapper, {
        complete: callback
      });
    },
    graphs: [],
    graphsUpdate: function (callback) {
        var existedGraphs = [];
        this.get('graphs').forEach(function (_graph) {
          var view = Em.View.views[_graph.id];
          if (view) {
            existedGraphs.push(_graph);
            //console.log('updated graph', _graph.name);
            view.loadData();
            //if graph opened as modal popup update it to
            if($(".modal-graph-line .modal-body #" + _graph.popupId + "-container-popup").length) {
              view.loadData();
            }
          }
        });
      callback();
      this.set('graphs', existedGraphs);
    },

    /**
     * Updates the services information. 
     *
     * @param callback
     * @param isInitialLoad  If true, only basic information is loaded.
     */
    updateServiceMetric: function (callback, isInitialLoad) {
      var self = this;
      self.set('isUpdated', false);

      var conditionalFields = [];
      var initialFields = [];
      var serviceSpecificParams = {
        'FLUME': "host_components/metrics/flume/flume",
        'YARN': "host_components/metrics/yarn/Queue," +
                "ServiceComponentInfo/rm_metrics/cluster/activeNMcount," +
                "ServiceComponentInfo/rm_metrics/cluster/unhealthyNMcount," +
                "ServiceComponentInfo/rm_metrics/cluster/rebootedNMcount," +
                "ServiceComponentInfo/rm_metrics/cluster/decommissionedNMcount",
        'HBASE': "host_components/metrics/hbase/master/IsActiveMaster," +
                 "ServiceComponentInfo/MasterStartTime," +
                 "ServiceComponentInfo/MasterActiveTime," +
                 "ServiceComponentInfo/AverageLoad," +
                 "ServiceComponentInfo/Revision," +
                 "ServiceComponentInfo/RegionsInTransition",
        'MAPREDUCE': "ServiceComponentInfo/AliveNodes," +
                     "ServiceComponentInfo/GrayListedNodes," +
                     "ServiceComponentInfo/BlackListedNodes," +
                     "ServiceComponentInfo/jobtracker/*,",
        'STORM': "metrics/api/cluster/summary,"
      };
      var services = App.cache['services'];
      services.forEach(function (service) {
        var urlParams = serviceSpecificParams[service.ServiceInfo.service_name];
        if (urlParams) {
          conditionalFields.push(urlParams);
        }
      });
      var conditionalFieldsString = conditionalFields.length > 0 ? ',' + conditionalFields.join(',') : '';
      var initialFieldsString = initialFields.length > 0 ? ',' + initialFields.join(',') : '';
      var testUrl = App.get('isHadoop2Stack') ? '/data/dashboard/HDP2/master_components.json':'/data/dashboard/services.json';

      var realUrl = '/components/?ServiceComponentInfo/category=MASTER&fields=' +
        'ServiceComponentInfo/Version,' +
        'ServiceComponentInfo/StartTime,' +
        'ServiceComponentInfo/HeapMemoryUsed,' +
        'ServiceComponentInfo/HeapMemoryMax,' +
        'ServiceComponentInfo/service_name,' +
        'host_components/HostRoles/host_name,' +
        'host_components/HostRoles/state,' +
        'host_components/HostRoles/maintenance_state,' +
        'host_components/HostRoles/stale_configs,' +
        'host_components/metrics/jvm/memHeapUsedM,' +
        'host_components/metrics/jvm/HeapMemoryMax,' +
        'host_components/metrics/jvm/HeapMemoryUsed,' +
        'host_components/metrics/jvm/memHeapCommittedM,' +
        'host_components/metrics/mapred/jobtracker/trackers_decommissioned,' +
        'host_components/metrics/cpu/cpu_wio,' +
        'host_components/metrics/rpc/RpcQueueTime_avg_time,' +
        'host_components/metrics/dfs/FSNamesystem/*,' +
        'host_components/metrics/dfs/namenode/Version,' +
        'host_components/metrics/dfs/namenode/DecomNodes,' +
        'host_components/metrics/dfs/namenode/TotalFiles,' +
        'host_components/metrics/dfs/namenode/UpgradeFinalized,' +
        'host_components/metrics/dfs/namenode/Safemode,' +
        'host_components/metrics/runtime/StartTime' +
        conditionalFieldsString +
        '&minimal_response=true';

      var servicesUrl = isInitialLoad ? this.getUrl(testUrl, realUrl + initialFieldsString) : this.getUrl(testUrl, realUrl);
      callback = callback || function () {
        self.set('isUpdated', true);
      };
      App.HttpClient.get(servicesUrl, App.serviceMetricsMapper, {
        complete: function(){
          callback();
        }
      });
    },
    updateServices: function (callback) {
      var testUrl = '/data/services/HDP2/services.json';
      var componentConfigUrl = this.getUrl(testUrl, '/services?fields=alerts/summary,ServiceInfo/state,ServiceInfo/maintenance_state&minimal_response=true');
      App.HttpClient.get(componentConfigUrl, App.serviceMapper, {
        complete: callback
      });
    },
    updateComponentConfig: function (callback) {
      var testUrl = '/data/services/host_component_stale_configs.json';
      var componentConfigUrl = this.getUrl(testUrl, '/host_components?fields=HostRoles/host_name&HostRoles/stale_configs=true&minimal_response=true');
      App.HttpClient.get(componentConfigUrl, App.componentConfigMapper, {
        complete: callback
      });
    }

  });
  
});
window.require.register("controllers/installer", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.InstallerController = App.WizardController.extend({

    name: 'installerController',

    totalSteps: 11,

    content: Em.Object.create({
      cluster: null,
      installOptions: null,
      hosts: null,
      services: null,
      slaveComponentHosts: null,
      masterComponentHosts: null,
      serviceConfigProperties: null,
      advancedServiceConfig: null,
      configGroups: [],
      slaveGroupProperties: null,
      stacks: null,
      controllerName: 'installerController'
    }),

    init: function () {
      this._super();
      this.get('isStepDisabled').setEach('value', true);
      this.get('isStepDisabled').pushObject(Ember.Object.create({
        step: 0,
        value: false
      }));
    },
    /**
     * redefined connectOutlet method to avoid view loading by unauthorized user
     * @param view
     * @param content
     */
    connectOutlet: function(view, content) {
      if(App.db.getAuthenticated()) {
        this._super(view, content);
      }
    },

    getCluster: function(){
      return jQuery.extend({}, this.get('clusterStatusTemplate'));
    },

    getInstallOptions: function(){
      return jQuery.extend({}, this.get('installOptionsTemplate'));
    },

    getHosts: function(){
      return [];
    },

    /**
     * Remove host from model. Used at <code>Confirm hosts(step2)</code> step
     * @param hosts Array of hosts, which we want to delete
     */
    removeHosts: function (hosts) {
      var dbHosts = this.getDBProperty('hosts');
      hosts.forEach(function (_hostInfo) {
        var host = _hostInfo.hostName;
        delete dbHosts[host];
      });
      this.setDBProperty('hosts', dbHosts);
    },

    /**
     * Load confirmed hosts.
     * Will be used at <code>Assign Masters(step5)</code> step
     */
    loadConfirmedHosts: function () {
      this.set('content.hosts', this.getDBProperty('hosts') || []);
    },

    /**
     * Load services data. Will be used at <code>Select services(step4)</code> step
     */
    loadServices: function () {
      var servicesInfo = this.getDBProperty('service');
      if(servicesInfo && servicesInfo.length) {
        servicesInfo.forEach(function (item, index) {
          servicesInfo[index] = Em.Object.create(item);
          servicesInfo[index].isInstalled = false;
        });
        this.set('content.services', servicesInfo);
        console.log('installerController.loadServices: loaded data ', JSON.stringify(servicesInfo));
        console.log('selected services ', servicesInfo.filterProperty('isSelected', true).mapProperty('serviceName'));
      } else {
        console.log("Failed to load Services");
     }
    },

    /**
     * total set of hosts registered to cluster, analog of App.Host model,
     * used in Installer wizard until hosts are installed
     */
    allHosts: function () {
      var rawHosts = this.get('content.hosts');
      var masterComponents = this.get('content.masterComponentHosts');
      var slaveComponents = this.get('content.slaveComponentHosts');
      var hosts = [];
      masterComponents.forEach(function (component) {
        var host = rawHosts[component.hostName];
        if (host.hostComponents) {
          host.hostComponents.push(Em.Object.create({
            componentName: component.component,
            displayName: component.display_name
          }));
        } else {
          rawHosts[component.hostName].hostComponents = [
            Em.Object.create({
              componentName: component.component,
              displayName: component.display_name
            })
          ]
        }
      });
      slaveComponents.forEach(function (component) {
        component.hosts.forEach(function (rawHost) {
          var host = rawHosts[rawHost.hostName];
          if (host.hostComponents) {
            host.hostComponents.push(Em.Object.create({
              componentName: component.componentName,
              displayName: component.displayName
            }));
          } else {
            rawHosts[rawHost.hostName].hostComponents = [
              Em.Object.create({
                componentName: component.componentName,
                displayName: component.displayName
              })
            ]
          }
        });
      });

      for (var hostName in rawHosts) {
        var host = rawHosts[hostName];
        var disksOverallCapacity = 0;
        var diskFree = 0;
        host.disk_info.forEach(function (disk) {
          disksOverallCapacity += parseFloat(disk.size);
          diskFree += parseFloat(disk.available);
        });
        hosts.pushObject(Em.Object.create({
            id: host.name,
            ip: host.ip,
            osType: host.os_type,
            osArch: host.os_arch,
            hostName: host.name,
            publicHostName: host.name,
            cpu: host.cpu,
            memory: host.memory,
            diskInfo: host.disk_info,
            diskTotal: disksOverallCapacity / (1024 * 1024),
            diskFree: diskFree / (1024 * 1024),
            hostComponents: host.hostComponents
          }
        ))
      }
      return hosts;
    }.property('content.hosts'),

    stacks: [],

    /**
     * Load stacks data from server or take exist data from local db
     */
    loadStacks: function () {
      var stacks = App.db.getStacks();
      if (stacks && stacks.length) {
        var convertedStacks = [];
        stacks.forEach(function (stack) {
          convertedStacks.pushObject(Ember.Object.create(stack));
        });
        App.set('currentStackVersion', convertedStacks.findProperty('isSelected').get('name'));
        this.set('content.stacks', convertedStacks);
      } else {
        App.ajax.send({
          name: 'wizard.stacks',
          sender: this,
          success: 'loadStacksSuccessCallback',
          error: 'loadStacksErrorCallback'
        });
      }
    },

    /**
     * Send queries to load versions for each stack
     */
    loadStacksSuccessCallback: function (data) {
      var stacks = data.items;
      var result;
      this.get('stacks').clear();
      stacks.forEach(function (stack) {
        App.ajax.send({
          name: 'wizard.stacks_versions',
          sender: this,
          data: {
            stackName: stack.Stacks.stack_name
          },
          success: 'loadStacksVersionsSuccessCallback',
          error: 'loadStacksVersionsErrorCallback'
        });
      }, this);
      result = this.get('stacks');
      if (!result.length) {
        console.log('Error: therea are no active stacks');
      } else {
        var defaultStackVersion = result.findProperty('name', App.defaultStackVersion);
        if (defaultStackVersion) {
          defaultStackVersion.set('isSelected', true)
        } else {
          result.objectAt(0).set('isSelected', true);
        }
      }
      App.db.setStacks(result);
      this.set('content.stacks', result);
    },

    /**
     * onError callback for loading stacks data
     */
    loadStacksErrorCallback: function () {
      console.log('Error in loading stacks');
    },

    /**
     * Parse loaded data and create array of stacks objects
     */
    loadStacksVersionsSuccessCallback: function (data) {
      var result = [];
      var stackVersions = data.items.filterProperty('Versions.active');
      stackVersions.sortProperty('Versions.stack_version').reverse().forEach(function (version) {
        /*
         * operatingSystems:[
         *  {
         *    osType: 'centos5',
         *    baseUrl: 'http://...',
         *    originalBaseUrl: 'http://...',
         *    defaultBaseUrl: 'http://...',
         *    latestBaseUrl: 'http://...',
         *    mirrorsList: '';
         *  },
         *  {
         *    osType: 'centos6',
         *    baseUrl: 'http://...',
         *    originalBaseUrl: 'http://...',
         *    defaultBaseUrl: 'http://...',
         *    latestBaseUrl: 'http://...',
         *    mirrorsList: '';
         *  },
         * ]
         */
        var oses = [];
        if (version.operatingSystems) {
          version.operatingSystems.forEach(function (os) {
            if (os.repositories) {
              os.repositories.forEach(function (repo) {
                if(repo.Repositories.repo_name == version.Versions.stack_name) {
                  var defaultBaseUrl = repo.Repositories.default_base_url || repo.Repositories.base_url;
                  var latestBaseUrl = repo.Repositories.latest_base_url || defaultBaseUrl;
                  if (!App.supports.ubuntu && os.OperatingSystems.os_type == 'ubuntu12') return; // @todo: remove after Ubuntu support confirmation
                  oses.push({
                    osType: os.OperatingSystems.os_type,
                    baseUrl: latestBaseUrl,
                    latestBaseUrl: latestBaseUrl,
                    originalLatestBaseUrl: latestBaseUrl,
                    originalBaseUrl: repo.Repositories.base_url,
                    defaultBaseUrl: defaultBaseUrl,
                    mirrorsList: repo.Repositories.mirrors_list
                  });
                }
              });
            }
          });
        }
        result.push(
            Em.Object.create({
              name: version.Versions.stack_name + "-" + version.Versions.stack_version,
              isSelected: false,
              operatingSystems: oses
            })
        );
      }, this);
      this.get('stacks').pushObjects(result);
    },

    /**
     * onError callback for loading stacks data
     */
    loadStacksVersionsErrorCallback: function () {
      console.log('Error in loading stacks');
    },

    /**
     * check server version and web client version
     */
    checkServerClientVersion: function () {
      var dfd = $.Deferred();
      var self = this;
      if (App.get('version')) {
        self.getServerVersion().done(function () {
          dfd.resolve();
        });
      } else {
        this.set('isServerClientVersionMismatch', false);
        dfd.resolve();
      }
      return dfd.promise();
    },
    getServerVersion: function(){
      return App.ajax.send({
        name: 'ambari.service.load_server_version',
        sender: this,
        success: 'getServerVersionSuccessCallback',
        error: 'getServerVersionErrorCallback'
      });
    },
    getServerVersionSuccessCallback: function (data) {
      var clientVersion = App.get('version');
      var serverVersion = (data.RootServiceComponents.component_version).toString();
      this.set('versionConflictAlertBody', Em.I18n.t('app.versionMismatchAlert.body').format(serverVersion, clientVersion));
      this.set('isServerClientVersionMismatch', clientVersion != serverVersion);
    },
    getServerVersionErrorCallback: function () {
      console.log('ERROR: Cannot load Ambari server version');
    },

    /**
     * Save data to model
     * @param stepController App.WizardStep4Controller
     */
    saveServices: function (stepController) {
      var serviceNames = [];
      this.setDBProperty('service', stepController.get('content'));
      stepController.filterProperty('isSelected', true).forEach(function (item) {
        serviceNames.push(item.serviceName);
      });
      this.set('content.selectedServiceNames', serviceNames);
      this.setDBProperty('selectedServiceNames', serviceNames);
      console.log('installerController.saveServices: saved data ', serviceNames);
    },

    /**
     * Save Master Component Hosts data to Main Controller
     * @param stepController App.WizardStep5Controller
     */
    saveMasterComponentHosts: function (stepController) {
      var obj = stepController.get('selectedServicesMasters');

      var masterComponentHosts = [];
      obj.forEach(function (_component) {
        masterComponentHosts.push({
          display_name: _component.get('display_name'),
          component: _component.get('component_name'),
          hostName: _component.get('selectedHost'),
          serviceId: _component.get('serviceId'),
          isInstalled: false
        });
      });

      console.log("installerController.saveMasterComponentHosts: saved hosts ", masterComponentHosts);
      this.setDBProperty('masterComponentHosts', masterComponentHosts);
      this.set('content.masterComponentHosts', masterComponentHosts);
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadMasterComponentHosts: function () {
      var masterComponentHosts = this.getDBProperty('masterComponentHosts') || [];
      this.set("content.masterComponentHosts", masterComponentHosts);
      console.log("InstallerController.loadMasterComponentHosts: loaded hosts ", masterComponentHosts);
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadSlaveComponentHosts: function () {
      var slaveComponentHosts = this.getDBProperty('slaveComponentHosts') || null;
      this.set("content.slaveComponentHosts", slaveComponentHosts);
      console.log("InstallerController.loadSlaveComponentHosts: loaded hosts ", slaveComponentHosts);
    },

    /**
     * Load serviceConfigProperties to model
     */
    loadServiceConfigProperties: function () {
      var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
      this.set('content.serviceConfigProperties', serviceConfigProperties);
      console.log("InstallerController.loadServiceConfigProperties: loaded config ", serviceConfigProperties);

      this.set('content.advancedServiceConfig', this.getDBProperty('advancedServiceConfig'));
    },

    /**
     * Load information about hosts with clients components
     */
    loadClients: function () {
      var clients = this.getDBProperty('clientInfo');
      this.set('content.clients', clients);
      console.log("InstallerController.loadClients: loaded list ", clients);
    },

    /**
     * Generate clients list for selected services and save it to model
     * @param stepController step4WizardController
     */
    saveClients: function (stepController) {
      var clients = [];
      var serviceComponents = require('data/service_components');

      stepController.get('content').filterProperty('isSelected', true).forEach(function (_service) {
        var client = serviceComponents.filterProperty('service_name', _service.serviceName).findProperty('isClient', true);
        if (client) {
          clients.pushObject({
            component_name: client.component_name,
            display_name: client.display_name,
            isInstalled: false
          });
        }
      }, this);

      this.setDBProperty('clientInfo', clients);
      this.set('content.clients', clients);
      console.log("InstallerController.saveClients: saved list ", clients);
    },

    /**
     * Save stacks data to local db
     * @param stepController step1WizardController
     */
    saveStacks: function (stepController) {
      var stacks = stepController.get('content.stacks');
      if (stacks.length) {
        App.set('currentStackVersion', stacks.findProperty('isSelected').get('name'));
      } else {
        App.set('currentStackVersion', App.defaultStackVersion);
      }
      App.db.setStacks(stacks);
      this.set('content.stacks', stacks);
    },

    /**
     * Check validation of the customized local urls
     * @param stepController step1WizardController
     */
    checkRepoURL: function (stepController) {
      var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
      selectedStack.set('reload', true);
      var nameVersionCombo = selectedStack.name;
      var stackName = nameVersionCombo.split('-')[0];
      var stackVersion = nameVersionCombo.split('-')[1];
      if (selectedStack && selectedStack.operatingSystems) {
        this.set('validationCnt', selectedStack.get('operatingSystems').filterProperty('selected', true).length);
        this.set('invalidCnt', 0);
        selectedStack.operatingSystems.forEach(function (os) {
          os.errorTitle = null;
          os.errorContent = null;
          if (os.skipValidation) {
            this.set('validationCnt', 0);
          }
          if (os.selected && !os.skipValidation) {
            os.validation = 'icon-repeat';
            selectedStack.set('reload', !selectedStack.get('reload'));
            App.ajax.send({
              name: 'wizard.advanced_repositories.valid_url',
              sender: this,
              data: {
                stackName: stackName,
                stackVersion: stackVersion,
                nameVersionCombo: nameVersionCombo,
                osType: os.osType,
                data: {
                  'Repositories': {
                    'base_url': os.baseUrl
                  }
                }
              },
              success: 'checkRepoURLSuccessCallback',
              error: 'checkRepoURLErrorCallback'
            });
          }
        }, this);
      }
    },
    setInvalidUrlCnt: function () {
      var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
      selectedStack.set('invalidCnt', this.get('invalidCnt'));
    }.observes('invalidCnt'),
    /**
     * onSuccess callback for check Repo URL.
     */
    checkRepoURLSuccessCallback: function (response, request, data) {
      console.log('Success in check Repo URL. data osType: ' + data.osType );
      var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
      if (selectedStack && selectedStack.operatingSystems) {
        var os = selectedStack.operatingSystems.findProperty('osType', data.osType);
        os.validation = 'icon-ok';
        selectedStack.set('reload', !selectedStack.get('reload'));
        this.set('validationCnt', this.get('validationCnt') - 1);
      }
    },

    /**
     * onError callback for check Repo URL.
     */
    checkRepoURLErrorCallback: function (request, ajaxOptions, error, data) {
      console.log('Error in check Repo URL. The baseURL sent is:  ' + data.data);
      var osType = data.url.split('/')[8];
      var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
      if (selectedStack && selectedStack.operatingSystems) {
        var os = selectedStack.operatingSystems.findProperty('osType', osType);
        os.validation = 'icon-exclamation-sign';
        os.errorTitle = request.status + ":" + request.statusText;
        os.errorContent = $.parseJSON(request.responseText) ? $.parseJSON(request.responseText).message : "";
        selectedStack.set('reload', !selectedStack.get('reload'));
        this.set('validationCnt', this.get('validationCnt') - 1);
        this.set('invalidCnt', this.get('invalidCnt') + 1);
      }
    },

    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '10':
        case '9':
        case '8':
        case '7':
          this.loadServiceConfigGroups();
          this.loadServiceConfigProperties();
        case '6':
          this.loadSlaveComponentHosts();
          this.loadClients();
        case '5':
          this.loadMasterComponentHosts();
          this.loadConfirmedHosts();
        case '4':
          this.loadServices();
        case '3':
          this.loadConfirmedHosts();
        case '2':
          this.load('installOptions');
        case '1':
          this.loadStacks();
        case '0':
          this.load('cluster');
      }
    },
    /**
     * Clear all temporary data
     */
    finish: function () {
      this.setCurrentStep('0');
      this.clearStorageData();
      var persists = App.router.get('applicationController').persistKey();
      App.router.get('applicationController').postUserPref(persists,true);
    },

    setStepsEnable: function () {
      for (var i = 0; i <= this.totalSteps; i++) {
        var step = this.get('isStepDisabled').findProperty('step', i);
        if (i <= this.get('currentStep')) {
          step.set('value', false);
        } else {
          step.set('value', true);
        }
      }
    }.observes('currentStep'),

    setLowerStepsDisable: function (stepNo) {
      for (var i = 0; i < stepNo; i++) {
        var step = this.get('isStepDisabled').findProperty('step', i);
        step.set('value', true);
      }
    }
  });

  
});
window.require.register("controllers/login_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.LoginController = Em.Object.extend({

    name: 'loginController',

    loginName: '',
    password: '',

    errorMessage: '',

    submit: function (e) {
      this.set('errorMessage', '');

      var self = this;

      App.get('router').login();
    },

    postLogin: function (isConnected, isAuthenticated) {
      if (!isConnected) {
        console.log('Failed to connect to Ambari Server');
        this.set('errorMessage', Em.I18n.t('login.error.bad.connection'));
      } else if (!isAuthenticated) {
        console.log('Failed to login as: ' + this.get('loginName'));
        this.set('errorMessage', Em.I18n.t('login.error.bad.credentials'));
      }
    }

  });
});
window.require.register("controllers/main", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  require('models/background_operation');

  App.MainController = Em.Controller.extend({
    name: 'mainController',

    updateTitle: function(){
      var name = App.router.get('clusterController.clusterName');
      if (name) {
        name = name.length > 13 ? name.substr(0, 10) + "..." : name;
      } else {
        name = Em.I18n.t('common.loading');
      }
      $('title').text('Ambari - ' + name);
    }.observes('App.router.clusterController.clusterName'),

    isClusterDataLoaded: function(){
      return App.router.get('clusterController.isLoaded');
    }.property('App.router.clusterController.isLoaded'),

    clusterDataLoadedPercent: function(){
      return App.router.get('clusterController.clusterDataLoadedPercent');
    }.property('App.router.clusterController.clusterDataLoadedPercent'),
    /**
     * run all processes and cluster's data loading
     */
    initialize: function(){
      App.router.get('clusterController').loadClusterData();
    },

    dataLoading: function () {
      var self = this;
      var dfd = $.Deferred();
      if (App.router.get('clusterController.isLoaded')) {
        dfd.resolve();
      } else {
        var interval = setInterval(function () {
          if (self.get('isClusterDataLoaded')) {
            dfd.resolve();
            clearInterval(interval);
          }
        }, 50);
      }
      return dfd.promise();
    },

    startPolling: function () {
      if (App.router.get('clusterController.isLoaded')) {
        App.router.get('updateController').set('isWorking', true);
        App.router.get('backgroundOperationsController').set('isWorking', true);
        App.router.get('clusterController').set('isWorking', true);
      }
    }.observes('App.router.clusterController.isLoaded'),
    stopPolling: function(){
      App.router.get('updateController').set('isWorking', false);
      App.router.get('backgroundOperationsController').set('isWorking', false);
      App.router.get('clusterController').set('isWorking', false);
    },

    reloadTimeOut: null,

    pageReload: function () {

      clearTimeout(this.get("reloadTimeOut"));

      this.set('reloadTimeOut',
          setTimeout(function () {
            if (App.clusterStatus.get('isInstalled')) {
              location.reload();
            }
          }, App.pageReloadTime)
      );
    }.observes("App.router.location.lastSetURL", "App.clusterStatus.isInstalled"),

    scRequest: function(request) {
      return App.router.get('mainServiceController').get(request);
    },

    isAllServicesInstalled: function() {
      return this.scRequest('isAllServicesInstalled');
    }.property('App.router.mainServiceController.content.content.@each',
        'App.router.mainServiceController.content.content.length'),

    isStartAllDisabled: function() {
      return this.scRequest('isStartAllDisabled');
    }.property('App.router.mainServiceController.isStartStopAllClicked',
        'App.router.mainServiceController.content.@each.healthStatus'),

    isStopAllDisabled: function() {
      return this.scRequest('isStopAllDisabled');
    }.property('App.router.mainServiceController.isStartStopAllClicked',
        'App.router.mainServiceController.content.@each.healthStatus'),

    gotoAddService: function() {
      App.router.get('mainServiceController').gotoAddService();
    },

    startAllService: function(event){
      App.router.get('mainServiceController').startAllService(event);
    },

    stopAllService: function(event){
      App.router.get('mainServiceController').stopAllService(event);
    },

    /**
     * check server version and web client version
     */
    checkServerClientVersion: function () {
      var dfd = $.Deferred();
      var self = this;
      if (App.get('version')) {
        self.getServerVersion().done(function () {
          dfd.resolve();
        });
      } else {
        this.set('isServerClientVersionMismatch', false);
        dfd.resolve();
      }
      return dfd.promise();
    },
    getServerVersion: function(){
      return App.ajax.send({
        name: 'ambari.service.load_server_version',
        sender: this,
        success: 'getServerVersionSuccessCallback',
        error: 'getServerVersionErrorCallback'
      });
    },
    getServerVersionSuccessCallback: function (data) {
      var clientVersion = App.get('version');
      var serverVersion = (data.RootServiceComponents.component_version).toString();
      this.set('versionConflictAlertBody', Em.I18n.t('app.versionMismatchAlert.body').format(serverVersion, clientVersion));
      this.set('isServerClientVersionMismatch', clientVersion != serverVersion);
    },
    getServerVersionErrorCallback: function () {
      console.log('ERROR: Cannot load Ambari server version');
    }

  });
  
});
window.require.register("controllers/main/admin", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminController = Em.Controller.extend({
    name: 'mainAdminController',
    category: 'user',
    /**
     * Check if access page available.
     * Turn on if YARN service is installed with Application Timeline Server component and TEZ installed too.
     *
     * @type {Boolean}
     */
    isAccessAvailable: function() {
      var dependencies = {
        services: ['YARN', 'TEZ'],
        components: ['APP_TIMELINE_SERVER']
      };
      var serviceNames = App.Service.find().mapProperty('serviceName')
        .filter(function(serviceName) {
          return dependencies.services.contains(serviceName);
        });
      var componentNames = App.get('stackDependedComponents').mapProperty('componentName')
        .filter(function(componentName) {
          return dependencies.components.contains(componentName);
        });
      var isAppTimelineServerAvailable = App.HostComponent.find().findProperty('componentName','APP_TIMELINE_SERVER');
      return (dependencies.services.length == serviceNames.length && componentNames.length == 0 && isAppTimelineServerAvailable);
    }.property()
  });
});
window.require.register("controllers/main/admin/access_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAccessController = Em.Controller.extend(App.UserPref, {
    name:'mainAdminAccessController',

    /**
     * Show jobs by default
     * @type {bool}
     */
    showJobs: true,

    /**
     * User pref request should be sync
     * @type {bool}
     */
    makeRequestAsync: false,

    /**
     * User pref key
     * @type {string}
     */
    persistKey: 'showJobsForNonAdmin',

    /**
     * Handle Save button click event
     */
    save: function() {
      this.postUserPref(this.get('persistKey'), this.get('showJobs'));
    },

    loadShowJobsForUsers: function () {
      var dfd = $.Deferred();
      this.getUserPref(this.get('persistKey')).done(function (value) {
        dfd.resolve(value);
      }).fail(function(value) {
        dfd.resolve(value);
      });
      return dfd.promise();
    },

    getUserPrefSuccessCallback: function (data) {
      this.set('showJobs', data);
      return data;
    },

    getUserPrefErrorCallback: function () {
      if (App.get('isAdmin')) {
        this.postUserPref(this.get('persistKey'), true);
      }
      this.set('showJobs', true);
      return true;
    }

  });
});
window.require.register("controllers/main/admin/advanced", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAdvancedController = Em.Controller.extend({
    name:'mainAdminAdvancedController',
    uninstall: function(event){
      var params = event.context;
      App.ModalPopup.show({
        uninstallParams: params,
        header: Em.I18n.t('admin.advanced.popup.header'),
        bodyClass: App.MainAdminAdvancedPasswordView.reopen({}), // layout: Em.Handlebars.compile()
        onPrimary: function(){
          var form = this.getForm();
          if(form) {
            if(form.isValid()) {
              console.warn("TODO: request for cluster uninstall");
            }
          }
          this.onClose();
        },
        onSecondary: function(){
          this.onClose();
        },

        getForm: function(){
          var form = false;
          $.each(this.get('_childViews'), function(){
            if(this.get('path') == "bodyClass") {
              return form = this.get('_childViews')[0];
            }
          });

          return form;
        }
      })
    }
  });
});
window.require.register("controllers/main/admin/authentication", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAuthenticationController = Em.Controller.extend({
    name:'mainAdminAuthenticationController',
    /**
     * save user form after editing
     * @param event
     */
    save:function (event) {
      var form = event.context;
      if (form.isValid()) {
        form.save();
        App.ModalPopup.show({
          header:Em.I18n.t('admin.authentication.form.testConfiguration'),
          body:form.get('resultText'),
          secondary:false,
          onPrimary:function () {
            this.hide();
          }
        });
      }
    },
    content:App.Authentication.find(1)
  });
});
window.require.register("controllers/main/admin/cluster", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.MainAdminClusterController = Em.Controller.extend({
    name:'mainAdminClusterController',
    services: [],
    repositories: [],
    upgradeVersion: '',
    /**
     * get the newest version of HDP from server
     */
    updateUpgradeVersion: function(){
      if(App.router.get('clusterController.isLoaded')){
        App.ajax.send({
          name: 'cluster.update_upgrade_version',
          sender: this,
          success: 'updateUpgradeVersionSuccessCallback',
          error: 'updateUpgradeVersionErrorCallback'
        });
      }
    }.observes('App.router.clusterController.isLoaded', 'App.currentStackVersion','App.router.mainServiceController.content.length'),

    updateUpgradeVersionSuccessCallback: function(data) {
      var upgradeVersion = this.get('upgradeVersion') || App.defaultStackVersion;
      var currentVersion = App.get('currentStackVersionNumber');
      upgradeVersion = upgradeVersion.replace(/HDP-/, '');
      data.items.mapProperty('Versions.stack_version').forEach(function(version){
        upgradeVersion = (stringUtils.compareVersions(upgradeVersion, version) === -1) ? version : upgradeVersion;
      });
      var currentStack = data.items.findProperty('Versions.stack_version', currentVersion);
      var upgradeStack = data.items.findProperty('Versions.stack_version', upgradeVersion);
      var minUpgradeVersion = upgradeStack.Versions.min_upgrade_version;
      if(minUpgradeVersion && (stringUtils.compareVersions(minUpgradeVersion, currentVersion) === 1)){
        upgradeVersion = currentVersion;
        upgradeStack = currentStack;
      }
      upgradeVersion = 'HDP-' + upgradeVersion;
      this.set('upgradeVersion', upgradeVersion);
      if(currentStack && upgradeStack) {
        this.parseServicesInfo(currentStack, upgradeStack);
      }
      else {
        console.log('HDP stack doesn\'t have services with defaultStackVersion');
      }
    },

    updateUpgradeVersionErrorCallback: function(request, ajaxOptions, error) {
      console.log('Error message is: ' + request.responseText);
      console.log('HDP stack doesn\'t have services with defaultStackVersion');
    },

    /**
     * get the installed repositories of HDP from server
     */
    loadRepositories: function(){
      if(App.router.get('clusterController.isLoaded')){
        var nameVersionCombo = App.get('currentStackVersion');
        var stackName = nameVersionCombo.split('-')[0];
        var stackVersion = nameVersionCombo.split('-')[1];
        App.ajax.send({
          name: 'cluster.load_repositories',
          sender: this,
          data: {
            stackName: stackName,
            stackVersion: stackVersion
          },
          success: 'loadRepositoriesSuccessCallback',
          error: 'loadRepositoriesErrorCallback'
        });
      }
    }.observes('App.router.clusterController.isLoaded'),

    loadRepositoriesSuccessCallback: function (data) {
      var allRepos = [];
      data.items.forEach(function(os) {
        var repo = Em.Object.create({
          baseUrl: os.repositories[0].Repositories.base_url,
          osType: os.repositories[0].Repositories.os_type
        });
        allRepos.push(repo);
      }, this);
      allRepos.stackVersion = App.get('currentStackVersionNumber');
      this.set('repositories', allRepos);
    },

    loadRepositoriesErrorCallback: function(request, ajaxOptions, error) {
      console.log('Error message is: ' + request.responseText);
    },

    /**
     * parse services info(versions, description) by version
     */
    parseServicesInfo: function (currentStack, upgradeStack) {
      var result = [];
      var installedServices = App.Service.find().mapProperty('serviceName');
      var displayOrderConfig = require('data/services');
      if(currentStack.stackServices.length && upgradeStack.stackServices.length){
        // loop through all the service components
        for (var i = 0; i < displayOrderConfig.length; i++) {
          var entry = currentStack.stackServices.
            findProperty("StackServices.service_name", displayOrderConfig[i].serviceName);
          if (entry) {
            entry = entry.StackServices;
            if (installedServices.contains(entry.service_name)) {
              var myService = Em.Object.create({
                serviceName: entry.service_name,
                displayName: displayOrderConfig[i].displayName,
                isDisabled: displayOrderConfig[i].isDisabled,
                isSelected: true,
                isInstalled: false,
                isHidden: displayOrderConfig[i].isHidden,
                description: entry.comments,
                version: entry.service_version,
                newVersion: ''
              });
              // it's possible that there is no corresponding service in the new stack
              var matchedService = upgradeStack.stackServices.findProperty("StackServices.service_name", displayOrderConfig[i].serviceName);
              if (matchedService) {
                myService.newVersion = matchedService.StackServices.service_version;
              }
              //From 1.3.0 for Hive we display only "Hive" (but it install HCat and WebHCat as well)
              if (this.get('upgradeVersion').replace(/HDP-/, '') >= '1.3.0' && displayOrderConfig[i].serviceName == 'HIVE') {
                myService.set('displayName', 'Hive');
              }
              result.push(myService);
            }
          }
          else {
            console.warn('Service not found - ', displayOrderConfig[i].serviceName);
          }
        }
      }
      this.set('services', result);
    }
  });
  
});
window.require.register("controllers/main/admin/highAvailability/progress_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityProgressPageController = App.HighAvailabilityWizardController.extend({

    name: 'highAvailabilityProgressPageController',

    status: 'IN_PROGRESS',
    clusterDeployState: 'HIGH_AVAILABILITY_DEPLOY',
    tasks: [],
    commands: [],
    currentRequestIds: [],
    logs: [],
    currentTaskId: null,
    POLL_INTERVAL: 4000,
    isSubmitDisabled: true,
    serviceTimestamp: null,
    isRollback: false,

    loadStep: function () {
      console.warn('func: loadStep');
      this.clearStep();
      this.initializeTasks();
      this.loadTasks();
      this.addObserver('tasks.@each.status', this, 'onTaskStatusChange');
      this.onTaskStatusChange();
    },

    clearStep: function () {
      console.warn('func: clearStep');
      this.set('isSubmitDisabled', true);
      this.set('tasks', []);
      this.set('logs', []);
      this.set('currentRequestIds', []);
    },

    initializeTasks: function() {
      console.warn('func: initializeTasks');
      var commands = this.get('commands');
      var currentStep = App.router.get('highAvailabilityWizardController.currentStep');
      for (var i = 0; i < commands.length; i++) {
        this.get('tasks').pushObject(Ember.Object.create({
          title: Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.task' + i + '.title'),
          status: 'PENDING',
          id: i,
          command: commands[i],
          showRetry: false,
          showRollback: false,
          name: Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.task' + i + '.title'),
          displayName: Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.task' + i + '.title'),
          progress: 0,
          isRunning: false,
          hosts: []
        }));
      }
    },

    services: function(){
      return this.get('tasks');
    }.property('tasks'),

    loadTasks: function () {
      console.warn('func: loadTasks');
      var self = this;
      var loadedStauses = this.get('content.tasksStatuses');
      var loadedLogs = this.get('content.logs');
      if (loadedStauses && loadedLogs && loadedStauses.length === this.get('tasks').length) {
        this.get('tasks').forEach(function(task,i){
          self.setTaskStatus(task.get('id'), loadedStauses[i]);
          self.restoreTaskLog(task.get('id'), loadedLogs[i]);
        });
        if (loadedStauses.contains('IN_PROGRESS')) {
          var curTaskId = this.get('tasks')[loadedStauses.indexOf('IN_PROGRESS')].get('id');
          this.set('currentRequestIds', this.get('content.requestIds'));
          this.set('currentTaskId', curTaskId);
          this.doPolling();
        }else if (loadedStauses.contains('QUEUED')){
          var curTaskId = this.get('tasks')[loadedStauses.indexOf('QUEUED')].get('id');
          this.set('currentTaskId', curTaskId);
          this.runTask(curTaskId);
        }
      }
    },

    setTaskStatus: function (taskId, status) {
      console.warn('func: setTaskStatus');
      this.get('tasks').findProperty('id', taskId).set('status', status);
    },

    restoreTaskLog: function (taskId, log) {
      console.warn('func: restoreTaskLog');
      this.get('tasks').findProperty('id', taskId).set('hosts', log);
    },

    setTaskLogs: function (taskId, tasks) {
      console.warn('func: setTaskLogs');
      var hosts = [];
      var uniqHosts = tasks.mapProperty('Tasks.host_name').uniq();
      uniqHosts.forEach(function (host) {
        var curHostTasks = tasks.filterProperty('Tasks.host_name', host);
        hosts.push(
         {
            name: host,
            publicName: host,
            logTasks: curHostTasks
          }
        );
      });
      this.get('tasks').findProperty('id', taskId).set('hosts', hosts);
      this.set('serviceTimestamp', App.dateTime());
    },

    retryTask: function () {
      console.warn('func: retryTask');
      var task = this.get('tasks').findProperty('status', 'FAILED');
      task.set('showRetry', false);
      task.set('showRollback', false);
      task.set('status', 'PENDING');
    },

    manualRollback: function () {
      console.warn('func: manualRollback');
      App.ModalPopup.show({
        header: Em.I18n.t('admin.highAvailability.confirmRollbackHeader'),
        primary: Em.I18n.t('yes'),
        showCloseButton: false,
        onPrimary: function () {
          var controller = App.router.get('highAvailabilityWizardController');
          controller.clearTasksData();
          controller.clearStorageData();
          controller.setCurrentStep('1');
          App.router.get('updateController').set('isWorking', true);
          App.clusterStatus.setClusterStatus({
            clusterName: App.router.get('content.cluster.name'),
            clusterState: 'DEFAULT',
            wizardControllerName: App.router.get('highAvailabilityRollbackController.name'),
            localdb: App.db.data
          });
          this.hide();
          App.router.transitionTo('main.admin.index');
          location.reload();
        },
        secondary : Em.I18n.t('no'),
        onSecondary: function(){
          this.hide();
        },
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile( Em.I18n.t('admin.highAvailability.confirmManualRollbackBody'))
        })
      });
    },

    rollback: function () {
      console.warn('func: rollback');
      var task = this.get('tasks').findProperty('status', 'FAILED');
      App.router.get(this.get('content.controllerName')).saveFailedTask(task);
      App.ModalPopup.show({
        header: Em.I18n.t('admin.highAvailability.confirmRollbackHeader'),
        primary: Em.I18n.t('common.confirm'),
        showCloseButton: false,
        onPrimary: function () {
          App.router.get('highAvailabilityWizardController').clearTasksData();
          App.router.transitionTo('main.admin.highAvailabilityRollback');
          this.hide();
        },
        secondary : Em.I18n.t('common.cancel'),
        body: Em.I18n.t('admin.highAvailability.confirmRollbackBody')
      });
    },

    onTaskStatusChange: function () {
      console.warn('func: onTaskStatusChange1');
      var statuses = this.get('tasks').mapProperty('status');
      var logs = this.get('tasks').mapProperty('hosts');
      var requestIds = this.get('currentRequestIds');
      console.warn('func: onTaskStatusChange5',statuses, logs, requestIds);
      // save task info
      App.router.get(this.get('content.controllerName')).saveTasksStatuses(statuses);
      App.router.get(this.get('content.controllerName')).saveRequestIds(requestIds);
      App.router.get(this.get('content.controllerName')).saveLogs(logs);
      // call saving of cluster status asynchronous
      // synchronous executing cause problems in Firefox
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: this.get('clusterDeployState'),
        wizardControllerName: this.get('content.controllerName'),
        localdb: App.db.data
      }, { async: true, success: 'statusChangeCallback', sender: this });
    },
    /**
     * Method that called after saving persist data to server.
     * Switch task according its status.
     */
    statusChangeCallback: function() {
      if (!this.get('tasks').someProperty('status', 'IN_PROGRESS') && !this.get('tasks').someProperty('status', 'QUEUED') && !this.get('tasks').someProperty('status', 'FAILED')) {
        var nextTask = this.get('tasks').findProperty('status', 'PENDING');
        if (nextTask) {
          console.warn('func: onTaskStatusChange2');
          this.set('status', 'IN_PROGRESS');
          this.setTaskStatus(nextTask.get('id'), 'QUEUED');
          this.set('currentTaskId', nextTask.get('id'));
          this.runTask(nextTask.get('id'));
        } else {
          console.warn('func: onTaskStatusChange3');
          this.set('status', 'COMPLETED');
          this.set('isSubmitDisabled', false);
        }
      } else if (this.get('tasks').someProperty('status', 'FAILED')) {
        console.warn('func: onTaskStatusChange4');
        this.set('status', 'FAILED');
        this.get('tasks').findProperty('status', 'FAILED').set('showRetry', true);
        if(App.supports.autoRollbackHA){
          this.get('tasks').findProperty('status', 'FAILED').set('showRollback', true);
        }
      }
      this.get('tasks').filterProperty('status','COMPLETED').setEach('showRetry', false);
      this.get('tasks').filterProperty('status','COMPLETED').setEach('showRollback', false);
    },

    /**
     * Run command of appropriate task
     */
    runTask: function (taskId) {
      console.warn('func: runTask',taskId);
      this[this.get('tasks').findProperty('id', taskId).get('command')]();
    },

    onTaskError: function () {
      console.warn('func: onTaskError');
      this.setTaskStatus(this.get('currentTaskId'), 'FAILED');
    },

    onTaskCompleted: function () {
      console.warn('func: onTaskCompleted');
      this.setTaskStatus(this.get('currentTaskId'), 'COMPLETED');
    },

    createComponent: function (componentName, hostName) {
      console.warn('func: createComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      var hostComponents = [];
      for (var i = 0; i < hostName.length; i++) {
        hostComponents = App.HostComponent.find().filterProperty('componentName', componentName);
        if (!hostComponents.length || !hostComponents.mapProperty('host.hostName').contains(hostName[i])) {
          App.ajax.send({
            name: 'admin.high_availability.create_component',
            sender: this,
            data: {
              hostName: hostName[i],
              componentName: componentName,
              taskNum: hostName.length
            },
            success: 'onCreateComponent',
            error: 'onCreateComponentError'
          });
        } else {
          // Simulates format returned from ajax.send
          this.onCreateComponent(null, null, {hostName: hostName[i], componentName: componentName, taskNum: hostName.length});
        }
      }
    },

    onCreateComponent: function () {
      console.warn('func: onCreateComponent');
      var hostName = arguments[2].hostName;
      var componentName = arguments[2].componentName;
      var taskNum = arguments[2].taskNum;
      this.installComponent(componentName, hostName, taskNum);
    },

    onCreateComponentError: function (error) {
      if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.ResourceAlreadyExistsException') !== -1) {
        this.onCreateComponent();
      } else {
        this.onTaskError();
      }
    },

    installComponent: function (componentName, hostName, taskNum) {
      console.warn('func: installComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.install_component',
          sender: this,
          data: {
            hostName: hostName[i],
            componentName: componentName,
            displayName: App.format.role(componentName),
            taskNum: taskNum || hostName.length
          },
          success: 'startPolling',
          error: 'onTaskError'
        });
      }
    },

    startComponent: function (componentName, hostName) {
      console.warn('func: startComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.start_component',
          sender: this,
          data: {
            hostName: hostName[i],
            componentName: componentName,
            displayName: App.format.role(componentName),
            taskNum: hostName.length
          },
          success: 'startPolling',
          error: 'onTaskError'
        });
      }
    },

    stopComponent: function (componentName, hostName) {
      console.warn('func: stopComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.stop_component',
          sender: this,
          data: {
            hostName: hostName[i],
            componentName: componentName,
            displayName: App.format.role(componentName),
            taskNum: hostName.length
          },
          success: 'startPolling',
          error: 'onTaskError'
        });
      }
    },

    startPolling: function (data) {
      if (data) {
        console.warn('func: startPolling1');
        this.get('currentRequestIds').push(data.Requests.id);
        var tasksCount = arguments[2].taskNum || 1;
        if (tasksCount === this.get('currentRequestIds').length) {
          console.warn('func: startPolling2');
          this.doPolling();
        }
      } else {
        console.warn('func: startPolling3');
        this.setTaskLogs(this.get('currentTaskId'), []);
        this.onTaskCompleted();
      }
    },

    doPolling: function () {
      console.warn('func: doPolling');
      this.setTaskStatus(this.get('currentTaskId'), 'IN_PROGRESS');
      var requestIds = this.get('currentRequestIds');
      this.set('logs', []);
      for (var i = 0; i < requestIds.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.polling',
          sender: this,
          data: {
            requestId: requestIds[i]
          },
          success: 'parseLogs',
          error: 'onTaskError'
        });
      }
    },

    parseLogs: function (logs) {
      console.warn('func: parseLogs');
      this.get('logs').push(logs.tasks);
      if (this.get('currentRequestIds').length === this.get('logs').length) {
        var tasks = [];
        this.get('logs').forEach(function (logs) {
          tasks.pushObjects(logs);
        }, this);
        var self = this;
        var currentTaskId = this.get('currentTaskId');
        this.setTaskLogs(currentTaskId, tasks);
        if (!tasks.someProperty('Tasks.status', 'PENDING') && !tasks.someProperty('Tasks.status', 'QUEUED') && !tasks.someProperty('Tasks.status', 'IN_PROGRESS')) {
          this.set('currentRequestIds', []);
          if (tasks.someProperty('Tasks.status', 'FAILED')  || tasks.someProperty('Tasks.status', 'TIMEDOUT') || tasks.someProperty('Tasks.status', 'ABORTED')) {
            this.setTaskStatus(currentTaskId, 'FAILED');
          } else {
            this.setTaskStatus(currentTaskId, 'COMPLETED');
          }
        } else {
          var actionsPerHost = tasks.length;
          var completedActions = tasks.filterProperty('Tasks.status', 'COMPLETED').length
            + tasks.filterProperty('Tasks.status', 'FAILED').length
            + tasks.filterProperty('Tasks.status', 'ABORTED').length
            + tasks.filterProperty('Tasks.status', 'TIMEDOUT').length;
          var queuedActions = tasks.filterProperty('Tasks.status', 'QUEUED').length;
          var inProgressActions = tasks.filterProperty('Tasks.status', 'IN_PROGRESS').length;
          var progress = Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsPerHost * 100);
          this.get('tasks').findProperty('id', currentTaskId).set('progress', progress);
          window.setTimeout(function () {
            self.doPolling()
          }, self.POLL_INTERVAL);
        }
      }
    },

    done: function () {
      if (!this.get('isSubmitDisabled')) {
        this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
        App.router.send('next');
      }
    }
  });

  
});
window.require.register("controllers/main/admin/highAvailability/rollback_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  require('controllers/main/admin/highAvailability/progress_controller');

  App.HighAvailabilityRollbackController = App.HighAvailabilityProgressPageController.extend({

    name: "highAvailabilityRollbackController",

    failedTask: null,
    configsSaved: false,
    deletedHdfsClients: 0,
    numOfDelOperations: 0,
    isRollback: true,
    hostsToPerformDel: [],

    commands: [
      'stopAllServices',
      'restoreHBaseConfigs',
      'stopFailoverControllers',
      'deleteFailoverControllers',
      'stopStandbyNameNode',
      'stopNameNode',
      'restoreHDFSConfigs',
      'enableSecondaryNameNode',
      'stopJournalNodes',
      'deleteJournalNodes',
      'deleteAdditionalNameNode',
      'startAllServices'
    ],

    loadStep: function () {
      console.warn('func: loadStep');
      this.initData();
      this.clearStep();
      this.loadTasks();
      this.addObserver('tasks.@each.status', this, 'onTaskStatusChange');
      this.onTaskStatusChange();
    },

    initData: function () {
      console.warn('func: initData');
      this.loadMasterComponentHosts();
      this.loadFailedTask();
      this.loadHdfsClientHosts();
    },

    setCommandsAndTasks: function(tmpTasks) {
      console.warn('func: setCommandsAndTasks');
      var fTask = this.get('failedTask');
      var index = [
        'deleteSNameNode',
        'startAllServices',
        'reconfigureHBase',
        'startZKFC',
        'installZKFC',
        'startSecondNameNode',
        'startNameNode',
        'startZooKeeperServers',
        'reconfigureHDFS',
        'disableSNameNode',
        'startJournalNodes',
        'installJournalNodes',
        'installNameNode',
        'stopAllServices'
      ].indexOf(fTask.command);

      if(index > 6){
        --index;
      }
      var newCommands = this.get('commands').splice(index);
      this.set('commands', newCommands);
      var newTasks = tmpTasks.splice(index);
      for (var i = 0; i < newTasks.length; i++) {
        newTasks[i].id = i;
      }
      this.set('tasks', newTasks);
      var hbaseTask = this.get('tasks').findProperty('command', 'restoreHBaseConfigs');
      if (!App.Service.find().someProperty('serviceName', 'HBASE') && hbaseTask) {
        this.get('tasks').splice(hbaseTask.get('id'), 1);
      }
    },

    clearStep: function () {
      console.warn('func: clearStep');
      this.set('isSubmitDisabled', true);
      this.set('tasks', []);
      this.set('logs', []);
      this.set('currentRequestIds', []);
      var commands = this.get('commands');
      var tmpTasks = [];
      for (var i = 0; i < commands.length; i++) {
        tmpTasks.pushObject(Ember.Object.create({
          title: Em.I18n.t('admin.highAvailability.rollback.task' + i + '.title'),
          status: 'PENDING',
          id: i,
          command: commands[i],
          showRetry: false,
          showRollback: false,
          showSkip: false,
          name: Em.I18n.t('admin.highAvailability.rollback.task' + i + '.title'),
          displayName: Em.I18n.t('admin.highAvailability.rollback.task' + i + '.title'),
          progress: 0,
          isRunning: false,
          hosts: []
        }));
      }
      this.setCommandsAndTasks(tmpTasks);
    },

    onTaskStatusChange: function () {
      console.warn('func: onTaskStatusChange');
      if (!this.get('tasks').someProperty('status', 'IN_PROGRESS') && !this.get('tasks').someProperty('status', 'QUEUED') && !this.get('tasks').someProperty('status', 'FAILED')) {
        var nextTask = this.get('tasks').findProperty('status', 'PENDING');
        if (nextTask) {
          console.warn('func: onTaskStatusChange1');
          this.set('status', 'IN_PROGRESS');
          this.setTaskStatus(nextTask.get('id'), 'QUEUED');
          this.set('currentTaskId', nextTask.get('id'));
          this.runTask(nextTask.get('id'));
        } else {
          console.warn('func: onTaskStatusChange2');
          this.set('status', 'COMPLETED');
          this.set('isSubmitDisabled', false);
        }
      } else if (this.get('tasks').someProperty('status', 'FAILED')) {
        console.warn('func: onTaskStatusChange3');
        this.set('status', 'FAILED');
        this.get('tasks').findProperty('status', 'FAILED').set('showRetry', true);
        this.get('tasks').findProperty('status', 'FAILED').set('showSkip', true);
      }
      this.get('tasks').filterProperty('status','COMPLETED').setEach('showRetry', false);
      this.get('tasks').filterProperty('status','COMPLETED').setEach('showSkip', false);

      var statuses = this.get('tasks').mapProperty('status');
      var logs = this.get('tasks').mapProperty('hosts');
      var requestIds = this.get('currentRequestIds');
      console.warn('func: onTaskStatusChange4',statuses,logs,requestIds);
      this.saveTasksStatuses(statuses);
      this.saveRequestIds(requestIds);
      this.saveLogs(logs);
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: 'HIGH_AVAILABILITY_ROLLBACK',
        wizardControllerName: 'highAvailabilityRollbackController',
        localdb: App.db.data
      });
    },

    skipTask: function () {
      console.warn('func: skipTask');
      var task = this.get('tasks').findProperty('status', 'FAILED');
      task.set('showRetry', false);
      task.set('showSkip', false);
      task.set('status', 'COMPLETED');
    },

    retryTask: function () {
      console.warn('func: retryTask');
      var task = this.get('tasks').findProperty('status', 'FAILED');
      task.set('showRetry', false);
      task.set('showSkip', false);
      task.set('status', 'PENDING');
    },

    onTaskCompleted: function () {
      console.warn('func: onTaskCompleted');
      var curTaskStatus = this.getTaskStatus(this.get('currentTaskId'));
      if (curTaskStatus != 'FAILED' && curTaskStatus != 'TIMEDOUT' && curTaskStatus != 'ABORTED') {
        this.setTaskStatus(this.get('currentTaskId'), 'COMPLETED');
      }
    },

    getTaskStatus: function (taskId) {
      console.warn('func: getTaskStatus');
      return this.get('tasks').findProperty('id', taskId).get('status');
    },

    loadFailedTask: function(){
      console.warn('func: loadFailedTask');
      var failedTask = App.db.getHighAvailabilityWizardFailedTask();
      this.set('failedTask', failedTask);
    },

    done: function () {
      if (!this.get('isSubmitDisabled')) {
        this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
        this.popup.proceedOnClose();
      }
    },

    stopAllServices: function(){
      console.warn('func: stopAllServices');
      App.ajax.send({
        name: 'admin.high_availability.stop_all_services',
        sender: this,
        success: 'startPolling',
        error: 'onTaskError'
      });
    },
    restoreHBaseConfigs: function(){
      console.warn('func: restoreHBaseConfigs');
      this.loadConfigTag("hbaseSiteTag");
      var hbaseSiteTag = this.get("content.hbaseSiteTag");
      App.ajax.send({
        name: 'admin.high_availability.load_hbase_configs',
        sender: this,
        data: {
          hbaseSiteTag: hbaseSiteTag
        },
        success: 'onLoadHbaseConfigs',
        error: 'onTaskError'
      });
    },

    stopFailoverControllers: function(){
      console.warn('func: stopFailoverControllers');
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
      this.stopComponent('ZKFC', hostNames);
    },
    deleteFailoverControllers: function(){
      console.warn('func: deleteFailoverControllers');
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
      this.checkBeforeDelete('ZKFC', hostNames);
    },
    stopStandbyNameNode: function(){
      console.warn('func: stopStandbyNameNode');
      var hostName = this.get('content.masterComponentHosts').findProperty('isAddNameNode', true).hostName;
      this.stopComponent('NAMENODE', hostName);
    },
    stopNameNode: function(){
      console.warn('func: stopNameNode');
      var hostNames = this.get('content.masterComponentHosts').findProperty('isCurNameNode').hostName;
      this.stopComponent('NAMENODE', hostNames);
    },
    restoreHDFSConfigs: function(){
      console.warn('func: restoreHDFSConfigs');
      this.unInstallHDFSClients();
    },
    enableSecondaryNameNode: function(){
      console.warn('func: enableSecondaryNameNode');
      var hostName = this.get('content.masterComponentHosts').findProperty('component', 'SECONDARY_NAMENODE').hostName;
      this.installComponent('SECONDARY_NAMENODE', hostName, hostName.length);
    },
    stopJournalNodes: function(){
      console.warn('func: stopJournalNodes');
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
      this.stopComponent('JOURNALNODE', hostNames);
    },
    deleteJournalNodes: function(){
      console.warn('func: deleteJournalNodes');
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
      this.unInstallComponent('JOURNALNODE', hostNames);
    },
    deleteAdditionalNameNode: function(){
      console.warn('func: deleteAdditionalNameNode');
      var hostNames = this.get('content.masterComponentHosts').filterProperty('isAddNameNode', true).mapProperty('hostName');
      this.unInstallComponent('NAMENODE', hostNames);
    },
    startAllServices: function(){
      console.warn('func: startAllServices');
      App.ajax.send({
        name: 'admin.high_availability.start_all_services',
        sender: this,
        success: 'startPolling',
        error: 'onTaskError'
      });
    },

    onLoadHbaseConfigs: function (data) {
      console.warn('func: onLoadHbaseConfigs');
      var hbaseSiteProperties = data.items.findProperty('type', 'hbase-site').properties;
      App.ajax.send({
        name: 'admin.high_availability.save_configs',
        sender: this,
        data: {
          siteName: 'hbase-site',
          properties: hbaseSiteProperties
        },
        success: 'onTaskCompleted',
        error: 'onTaskError'
      });
    },

    stopComponent: function (componentName, hostName) {
      console.warn('func: stopComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.stop_component',
          sender: this,
          data: {
            hostName: hostName[i],
            componentName: componentName,
            displayName: App.format.role(componentName),
            taskNum: hostName.length
          },
          success: 'startPolling',
          error: 'onTaskError'
        });
      }
    },

    onDeletedHDFSClient: function () {
      console.warn('func: onDeletedHDFSClient');
      var deletedHdfsClients = this.get('deletedHdfsClients');
      var hostName = this.get("content.hdfsClientHostNames");
      var notDeletedHdfsClients = hostName.length - deletedHdfsClients;
      if (notDeletedHdfsClients > 1 && hostName.length != 1 ) {
        this.set('deletedHdfsClients', deletedHdfsClients+1);
        return;
      }
      this.loadConfigTag("hdfsSiteTag");
      this.loadConfigTag("coreSiteTag");
      var hdfsSiteTag = this.get("content.hdfsSiteTag");
      var coreSiteTag = this.get("content.coreSiteTag");
      App.ajax.send({
        name: 'admin.high_availability.load_configs',
        sender: this,
        data: {
          hdfsSiteTag: hdfsSiteTag,
          coreSiteTag: coreSiteTag
        },
        success: 'onLoadConfigs',
        error: 'onTaskError'
      });
    },

    onLoadConfigs: function (data) {
      console.warn('func: onLoadConfigs');
      this.set('configsSaved', false);
      App.ajax.send({
        name: 'admin.high_availability.save_configs',
        sender: this,
        data: {
          siteName: 'hdfs-site',
          properties: data.items.findProperty('type', 'hdfs-site').properties
        },
        success: 'onHdfsConfigsSaved',
        error: 'onTaskError'
      });
      App.ajax.send({
        name: 'admin.high_availability.save_configs',
        sender: this,
        data: {
          siteName: 'core-site',
          properties: data.items.findProperty('type', 'core-site').properties
        },
        success: 'onHdfsConfigsSaved',
        error: 'onTaskError'
      });
    },

    onHdfsConfigsSaved: function () {
      console.warn('func: onHdfsConfigsSaved');
      if (!this.get('configsSaved')) {
        this.set('configsSaved', true);
        return;
      }
      this.onTaskCompleted();
    },

    unInstallHDFSClients: function () {
      console.warn('func: unInstallHDFSClients');
      var hostName = this.get("content.hdfsClientHostNames");
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.delete_component',
          sender: this,
          data: {
            componentName: 'HDFS_CLIENT',
            hostName: hostName[i]
          },
          success: 'onDeletedHDFSClient',
          error: 'onTaskError'
        });
      }
    },

    unInstallComponent: function (componentName, hostName) {
      console.warn('func: unInstallComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.maintenance_mode',
          sender: this,
          data: {
            hostName: hostName[i],
            componentName: componentName,
            taskNum: hostName.length,
            callback: 'checkBeforeDelete'
          },
          success: 'checkResult',
          error: 'checkResult'
        });
      }
    },

    checkBeforeDelete: function (componentName, hostName){
      console.warn('func: checkBeforeDelete');
      this.set('hostsToPerformDel', []);
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.high_availability.getHostComponent',
          sender: this,
          data: {
            componentName: componentName,
            hostName: hostName[i],
            taskNum: hostName.length,
            callback: 'deleteComponent'
          },
          success: 'checkResult',
          error: 'checkResult'
        });
      }
    },

    checkResult: function () {
      console.warn('func: checkResult');
      var callback = arguments[2].callback;
      var hostName = arguments[2].hostName;
      var componentName = arguments[2].componentName;
      var taskNum = arguments[2].taskNum;
      var hostsToPerformDel = this.get('hostsToPerformDel');
      if(arguments[1] != 'error'){
        hostsToPerformDel.push({
          hostName: hostName,
          isOnHost: true
        });
      }else{
        hostsToPerformDel.push({
          hostName: 'error',
          isOnHost: false
        });
      }
      if(hostsToPerformDel.length == taskNum){
        var hostsForDel = hostsToPerformDel.filterProperty('isOnHost', true).mapProperty('hostName');
        this.set('hostsToPerformDel', []);
        if(hostsForDel.length == 0){
          this.onTaskCompleted();
          return;
        }
        this[callback](componentName, hostsForDel);
      }
    },

    deleteComponent: function (componentName, hostName) {
      console.warn('func: deleteComponent');
      if (!(hostName instanceof Array)) {
        hostName = [hostName];
      }
      this.set('numOfDelOperations', hostName.length);
      for (var i = 0; i < hostName.length; i++) {
        App.ajax.send({
          name: 'admin.delete_component',
          sender: this,
          data: {
            componentName: componentName,
            hostName: hostName[i]
          },
          success: 'onDeleteComplete',
          error: 'onTaskError'
        });
      }
    },

    onDeleteComplete: function () {
      console.warn('func: onDeleteComplete');
      var leftOp = this.get('numOfDelOperations');
      if(leftOp > 1){
        this.set('numOfDelOperations', leftOp-1);
        return;
      }
      this.onTaskCompleted();
    }

  });
  
});
window.require.register("controllers/main/admin/highAvailability/step1_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  require('controllers/main/admin/misc_controller');

  App.HighAvailabilityWizardStep1Controller = Em.Controller.extend({
    name: "highAvailabilityWizardStep1Controller",

    isNameServiceIdValid: function () {
      return /^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])$/.test(this.get('content.nameServiceId'));
    }.property('content.nameServiceId'),

    next: function () {
      if (this.get('isNameServiceIdValid')) {
        App.router.send('next');
      }
    }

  });

  
});
window.require.register("controllers/main/admin/highAvailability/step2_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  require('controllers/wizard/step5_controller');

  App.HighAvailabilityWizardStep2Controller = App.WizardStep5Controller.extend({

    name:"highAvailabilityWizardStep2Controller",

    /**
     * master components which could be assigned to multiple hosts
     */
    multipleComponents: ['NAMENODE', 'JOURNALNODE'],

    /**
     * Load services info to appropriate variable and return masterComponentHosts
     * @return Array
     */
    masterHostMapping:function () {
      var mapping = [], mappingObject, self = this, mappedHosts, hostObj, hostInfo;
      //get the unique assigned hosts and find the master services assigned to them


      mappedHosts = this.get("selectedServicesMasters").mapProperty("selectedHost").uniq();

      mappedHosts.forEach(function (item) {

        hostObj = self.get("hosts").findProperty("host_name", item);
        console.log("Name of the host is: " + hostObj.host_name);

        var masterServices = self.get("selectedServicesMasters").filterProperty("selectedHost", item);
        masterServices.forEach(function(item){
          if(this.get('multipleComponents').contains(item.component_name)){
            item.set('color','green');
          }else{
            item.set('color','grey');
          }
        }, this);

        mappingObject = Ember.Object.create({
          host_name:item,
          hostInfo:hostObj.host_info,
          masterServices:masterServices
        });

        mapping.pushObject(mappingObject);
      }, this);

      return mapping.sortProperty('host_name');
    }.property("selectedServicesMasters.@each.selectedHost"),

    /**
     * Put master components to <code>selectedServicesMasters</code>, which will be automatically rendered in template
     * @param masterComponents
     */
    renderComponents:function (masterComponents) {
      var services = this.get('content.services')
        .filterProperty('isInstalled', true).filterProperty('isInstalled', true).mapProperty('serviceName'); //list of shown services

      var result = [];

      var curNameNode = masterComponents.findProperty('component_name',"NAMENODE");
      curNameNode.isCurNameNode = true;
      curNameNode.zId = 0;

      //Create JOURNALNODE
      for (var index = 0; index < 3; index++) {
        masterComponents.push(
          {
            component_name: "JOURNALNODE",
            display_name: "JournalNode",
            isHiveCoHost: false,
            isInstalled: false,
            selectedHost: this.get("hosts")[index].get("host_name"),
            serviceId: "HDFS",
            zId: index
          }
        )
      }
      //Create Additional NameNode
      masterComponents.push(
        {
          component_name: "NAMENODE",
          display_name: "NameNode",
          isHiveCoHost: false,
          isInstalled: false,
          selectedHost: this.get("hosts").mapProperty('host_name').without(curNameNode.selectedHost)[0],
          serviceId: "HDFS",
          isAddNameNode: true,
          zId: 1
        }
      );

      masterComponents.forEach(function (item) {
        var componentObj = Ember.Object.create(item);
        result.push(componentObj);
      }, this);

      this.set("selectedServicesMasters", result);

      var components = result.filterProperty('component_name',"NAMENODE");
      components.push.apply(components, result.filterProperty('component_name',"JOURNALNODE"));

      this.set('servicesMasters', components);
      this.set('componentToRebalance', "NAMENODE");
      this.incrementProperty('rebalanceComponentHostsCounter');
      this.set('componentToRebalance', "JOURNALNODE");
      this.incrementProperty('rebalanceComponentHostsCounter');
    }

  });

  
});
window.require.register("controllers/main/admin/highAvailability/step3_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityWizardStep3Controller = Em.Controller.extend({
    name: "highAvailabilityWizardStep3Controller",
    selectedService: null,
    stepConfigs: [],
    serverConfigData: {},
    haConfig: $.extend(true, {}, require('data/HDP2/ha_properties').haConfig),
    once: false,

    clearStep: function () {
      this.get('stepConfigs').clear();
      this.serverConfigData = {};
    },

    loadStep: function () {
      this.clearStep();
      this.loadConfigsTags();
    },

    loadConfigsTags: function () {
      App.ajax.send({
        name: 'config.tags',
        sender: this,
        success: 'onLoadConfigsTags',
        error: 'onTaskError'
      });
    },


    onLoadConfigsTags: function (data) {
      var urlParams = [];
      var hdfsSiteTag = data.Clusters.desired_configs['hdfs-site'].tag;
      var coreSiteTag = data.Clusters.desired_configs['core-site'].tag;
      urlParams.push('(type=hdfs-site&tag=' + hdfsSiteTag + ')');
      urlParams.push('(type=core-site&tag=' + coreSiteTag + ')');
      this.set("hdfsSiteTag", {name : "hdfsSiteTag", value : hdfsSiteTag});
      this.set("coreSiteTag", {name : "coreSiteTag", value : coreSiteTag});

      if (App.Service.find().someProperty('serviceName', 'HBASE')) {
        var hbaseSiteTag = data.Clusters.desired_configs['hbase-site'].tag;
        urlParams.push('(type=hbase-site&tag=' + hbaseSiteTag + ')');
        this.set("hbaseSiteTag", {name : "hbaseSiteTag", value : hbaseSiteTag});
      }
      App.ajax.send({
        name: 'admin.get.all_configurations',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'onLoadConfigs',
        error: 'onTaskError'
      });
    },

    onLoadConfigs: function (data) {
      this.set('serverConfigData',data);
      this.tweakServiceConfigs(this.get('haConfig.configs'));
      this.renderServiceConfigs(this.get('haConfig'));
    },


    tweakServiceConfigs: function(configs) {
      var nameServiceId = this.get('content.nameServiceId');
      var nameServiceConfig = configs.findProperty('name','dfs.nameservices');
      this.setConfigInitialValue(nameServiceConfig,nameServiceId);
      var defaultFsConfig = configs.findProperty('name','fs.defaultFS');
      this.setConfigInitialValue(defaultFsConfig, "hdfs://" + nameServiceId);
      this.tweakServiceConfigNames(configs,nameServiceId);
      this.tweakServiceConfigValues(configs,nameServiceId);
    },

    tweakServiceConfigNames: function(configs,nameServiceId) {
      var regex = new RegExp("\\$\\{dfs.nameservices\\}","g");
      configs.forEach(function(config) {
        if (config.name.contains("${dfs.nameservices}")) {
          config.name = config.name.replace(regex,nameServiceId);
          config.displayName = config.displayName.replace(regex,nameServiceId);
        }
      }, this);
    },

    tweakServiceConfigValues: function(configs,nameServiceId) {
      var currentNameNodeHost = this.get('content.masterComponentHosts').findProperty('isCurNameNode').hostName;
      var newNameNodeHost = this.get('content.masterComponentHosts').findProperty('isAddNameNode').hostName;
      var journalNodeHosts = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
      var zooKeeperHosts = this.get('content.masterComponentHosts').filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
      var config = configs.findProperty('name','dfs.namenode.rpc-address.' + nameServiceId + '.nn1');
      this.setConfigInitialValue(config,currentNameNodeHost + ':8020');
      config = configs.findProperty('name','dfs.namenode.rpc-address.' + nameServiceId + '.nn2');
      this.setConfigInitialValue(config,newNameNodeHost + ':8020');
      config = configs.findProperty('name','dfs.namenode.http-address.' + nameServiceId + '.nn1');
      this.setConfigInitialValue(config,currentNameNodeHost + ':50070');
      config = configs.findProperty('name','dfs.namenode.http-address.' + nameServiceId + '.nn2');
      this.setConfigInitialValue(config,newNameNodeHost + ':50070');
      config = configs.findProperty('name','dfs.namenode.https-address.' + nameServiceId + '.nn1');
      this.setConfigInitialValue(config,currentNameNodeHost + ':50470');
      config = configs.findProperty('name','dfs.namenode.https-address.' + nameServiceId + '.nn2');
      this.setConfigInitialValue(config,newNameNodeHost + ':50470');
      config = configs.findProperty('name','dfs.namenode.shared.edits.dir');
      this.setConfigInitialValue(config,'qjournal://' + journalNodeHosts[0] + ':8485;' + journalNodeHosts[1] + ':8485;' + journalNodeHosts[2] + ':8485/' + nameServiceId);
      config = configs.findProperty('name','ha.zookeeper.quorum');
      this.setConfigInitialValue(config,zooKeeperHosts[0] + ':2181,' + zooKeeperHosts[1] + ':2181,' + zooKeeperHosts[2] + ':2181');
      config = configs.findProperty('name','hbase.rootdir');
      if (App.Service.find().someProperty('serviceName', 'HBASE')) {
       var value = this.get('serverConfigData.items').findProperty('type', 'hbase-site').properties['hbase.rootdir'].replace(/\/\/[^\/]*/, '//' + nameServiceId);
       this.setConfigInitialValue(config,value);
      }
    },

    setConfigInitialValue: function(config,value) {
      config.value = value;
      config.defaultValue = value;
    },

    renderServiceConfigs: function (_serviceConfig) {
      var serviceConfig = App.ServiceConfig.create({
        serviceName: _serviceConfig.serviceName,
        displayName: _serviceConfig.displayName,
        configCategories: [],
        showConfig: true,
        configs: []
      });

      _serviceConfig.configCategories.forEach(function (_configCategory) {
        if (App.Service.find().someProperty('serviceName', _configCategory.name)) {
          serviceConfig.configCategories.pushObject(_configCategory);
        }
      }, this);

      this.loadComponentConfigs(_serviceConfig, serviceConfig);

      this.get('stepConfigs').pushObject(serviceConfig);
      this.set('selectedService', this.get('stepConfigs').objectAt(0));
      this.once = true;
    },

    /**
     * Load child components to service config object
     * @param _componentConfig
     * @param componentConfig
     */
    loadComponentConfigs: function (_componentConfig, componentConfig) {
      _componentConfig.configs.forEach(function (_serviceConfigProperty) {
        var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
        componentConfig.configs.pushObject(serviceConfigProperty);
        serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
        serviceConfigProperty.validate();
      }, this);
    }

  });

  
});
window.require.register("controllers/main/admin/highAvailability/step4_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  require('controllers/main/admin/misc_controller');

  App.HighAvailabilityWizardStep4Controller = Em.Controller.extend({

    name:"highAvailabilityWizardStep4Controller",

    POLL_INTERVAL: 1000,

    isNextEnabled: false,

    pullCheckPointStatus: function () {
      var hostName = this.get('content.masterComponentHosts').findProperty('isCurNameNode', true).hostName;
      App.ajax.send({
        name: 'admin.high_availability.getNnCheckPointStatus',
        sender: this,
        data: {
          hostName: hostName
        },
        success: 'checkNnCheckPointStatus'
      });
    },

    checkNnCheckPointStatus: function (data) {
      var self = this;
      var journalTransactionInfo =  $.parseJSON(data.metrics.dfs.namenode.JournalTransactionInfo);
      var isInSafeMode = (data.metrics.dfs.namenode.Safemode != "");
      journalTransactionInfo = parseInt(journalTransactionInfo.LastAppliedOrWrittenTxId) - parseInt(journalTransactionInfo.MostRecentCheckpointTxId);
      if(journalTransactionInfo <= 1 && isInSafeMode){
        this.set("isNextEnabled", true);
        return;
      }

      window.setTimeout(function () {
        self.pullCheckPointStatus()
      }, self.POLL_INTERVAL);
    },

    done: function () {
      if (this.get('isNextEnabled')) {
        App.router.send('next');
      }
    }

  });

  
});
window.require.register("controllers/main/admin/highAvailability/step5_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityWizardStep5Controller = App.HighAvailabilityProgressPageController.extend({

    name:"highAvailabilityWizardStep5Controller",

    isHA: true,

    commands: ['stopAllServices', 'installNameNode', 'installJournalNodes', 'reconfigureHDFS', 'startJournalNodes', 'disableSNameNode'],

    hdfsSiteTag : "",
    coreSiteTag : "",

    stopAllServices: function () {
      App.ajax.send({
        name: 'admin.high_availability.stop_all_services',
        sender: this,
        success: 'startPolling',
        error: 'onTaskError'
      });
    },

    installNameNode: function () {
      var hostName = this.get('content.masterComponentHosts').findProperty('isAddNameNode').hostName;
      this.createComponent('NAMENODE', hostName);
    },

    installJournalNodes: function () {
      App.ajax.send({
        name: 'admin.high_availability.create_journalnode',
        sender: this,
        success: 'onJournalNodeCreate',
        error: 'onJournalNodeCreate'
      });
    },

    onJournalNodeCreate: function () {
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
      this.createComponent('JOURNALNODE', hostNames);
    },

    startJournalNodes: function () {
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
      this.startComponent('JOURNALNODE', hostNames);
    },

    disableSNameNode: function () {
      var hostName = this.get('content.masterComponentHosts').findProperty('component', 'SECONDARY_NAMENODE').hostName;
      App.ajax.send({
        name: 'admin.high_availability.maintenance_mode',
        sender: this,
        data: {
          hostName: hostName,
          componentName: 'SECONDARY_NAMENODE'
        },
        success: 'onTaskCompleted',
        error: 'onTaskError'
      });
    },

    reconfigureHDFS: function () {
      var data = this.get('content.serviceConfigProperties');
      var hdfsSiteProperties = data.items.findProperty('type', 'hdfs-site').properties;
      var coreSiteProperties = data.items.findProperty('type', 'core-site').properties;
      var self = this;
      App.ajax.send({
        name: 'admin.high_availability.save_configs',
        sender: this,
        data: {
          siteName: 'hdfs-site',
          properties: hdfsSiteProperties
        },
        error: 'onTaskError'
      }).done(function() {
        App.ajax.send({
          name: 'admin.high_availability.save_configs',
          sender: self,
          data: {
            siteName: 'core-site',
            properties: coreSiteProperties
          },
          error: 'onTaskError',
          success: 'installHDFSClients'
        });
      });
    },

    installHDFSClients: function () {
      var nnHostNames = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
      var jnHostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
      var hostNames = nnHostNames.concat(jnHostNames).uniq();
      this.createComponent('HDFS_CLIENT', hostNames);
      App.router.get(this.get('content.controllerName')).saveHdfsClientHosts(hostNames);
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: 'HIGH_AVAILABILITY_DEPLOY',
        wizardControllerName: this.get('content.controllerName'),
        localdb: App.db.data
      });
    }
  });

  
});
window.require.register("controllers/main/admin/highAvailability/step6_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityWizardStep6Controller = Em.Controller.extend({

    name:"highAvailabilityWizardStep6Controller",

    POLL_INTERVAL: 1000,

    isNextEnabled: function(){
      //only 3 JournalNodes could be installed
      return (this.get('initJnCounter') === 3);
    }.property('initJnCounter'),

    initJnCounter: 0,

    pullCheckPointStatus: function () {
      this.set('initJnCounter', 0);
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', "JOURNALNODE").mapProperty('hostName');
      hostNames.forEach(function (hostName) {
        this.pullEachJnStatus(hostName);
      }, this);
    },

    pullEachJnStatus: function(hostName){
      App.ajax.send({
        name: 'admin.high_availability.getJnCheckPointStatus',
        sender: this,
        data: {
          hostName: hostName
        },
        success: 'checkJnCheckPointStatus'
      });
    },

    checkJnCheckPointStatus: function (data) {
      var self = this;
      var journalStatusInfo;
      if (data.metrics && data.metrics.dfs) {
        journalStatusInfo = $.parseJSON(data.metrics.dfs.journalnode.journalsStatus);
        if (journalStatusInfo[this.get('content.nameServiceId')] && journalStatusInfo[this.get('content.nameServiceId')].Formatted === "true") {
          this.set("initJnCounter", (this.get('initJnCounter') + 1));
          return;
        }
      }

      window.setTimeout(function () {
        self.pullEachJnStatus(data.HostRoles.host_name);
      }, self.POLL_INTERVAL);
    },

    done: function () {
      if (this.get('isNextEnabled')) {
        App.router.send('next');
      }
    }

  });

  
});
window.require.register("controllers/main/admin/highAvailability/step7_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityWizardStep7Controller = App.HighAvailabilityProgressPageController.extend({

    name:"highAvailabilityWizardStep7Controller",

    isHA: true,

    commands: ['startZooKeeperServers', 'startNameNode'],

    startZooKeeperServers: function () {
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
      this.startComponent('ZOOKEEPER_SERVER', hostNames);
    },

    startNameNode: function () {
      var hostName = this.get('content.masterComponentHosts').findProperty('isCurNameNode').hostName;
      this.startComponent('NAMENODE', hostName);
    }
  });

  
});
window.require.register("controllers/main/admin/highAvailability/step8_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityWizardStep8Controller = Em.Controller.extend({

    name:"highAvailabilityWizardStep8Controller"

  });

  
});
window.require.register("controllers/main/admin/highAvailability/step9_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HighAvailabilityWizardStep9Controller = App.HighAvailabilityProgressPageController.extend({

    name:"highAvailabilityWizardStep9Controller",

    isHA: true,

    commands: ['startSecondNameNode', 'installZKFC', 'startZKFC', 'reconfigureHBase', 'deleteSNameNode', 'startAllServices'],

    hbaseSiteTag: "",

    initializeTasks: function () {
      this._super();
      if (!App.Service.find().someProperty('serviceName', 'HBASE')) {
        this.get('tasks').splice(this.get('tasks').findProperty('command', 'reconfigureHBase').get('id'), 1);
      }
    },

    startSecondNameNode: function () {
      var hostName = this.get('content.masterComponentHosts').findProperty('isAddNameNode', true).hostName;
      this.startComponent('NAMENODE', hostName);
    },

    installZKFC: function () {
      App.ajax.send({
        name: 'admin.high_availability.create_zkfc',
        sender: this,
        success: 'onZKFCCreate',
        error: 'onZKFCCreate'
      });
    },

    onZKFCCreate: function () {
      var hostName = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
      this.createComponent('ZKFC', hostName);
    },

    startZKFC: function () {
      var hostName = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
      this.startComponent('ZKFC', hostName);
    },

    reconfigureHBase: function () {
      var data = this.get('content.serviceConfigProperties');
      var hbaseSiteProperties = data.items.findProperty('type', 'hbase-site').properties;

      App.ajax.send({
        name: 'admin.high_availability.save_configs',
        sender: this,
        data: {
          siteName: 'hbase-site',
          properties: hbaseSiteProperties
        },
        success: 'saveConfigTag',
        error: 'onTaskError'
      });
    },

    saveConfigTag: function () {
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: 'HIGH_AVAILABILITY_DEPLOY',
        wizardControllerName: this.get('content.controllerName'),
        localdb: App.db.data
      });
      this.onTaskCompleted();
    },

    startAllServices: function () {
      App.ajax.send({
        name: 'admin.high_availability.start_all_services',
        sender: this,
        success: 'startPolling',
        error: 'onTaskError'
      });
    },

    deleteSNameNode: function () {
      var hostName = this.get('content.masterComponentHosts').findProperty('component', 'SECONDARY_NAMENODE').hostName;
      App.ajax.send({
        name: 'admin.delete_component',
        sender: this,
        data: {
          componentName: 'SECONDARY_NAMENODE',
          hostName: hostName
        },
        success: 'onTaskCompleted',
        error: 'onTaskError'
      });
    }

  });

  
});
window.require.register("controllers/main/admin/highAvailability/wizard_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardController = App.WizardController.extend({

    name: 'highAvailabilityWizardController',

    totalSteps: 9,

    /**
     * Used for hiding back button in wizard
     */
    hideBackButton: true,

    content: Em.Object.create({
      controllerName: 'highAvailabilityWizardController',
      cluster: null,
      hosts: null,
      services: null,
      slaveComponentHosts: null,
      masterComponentHosts: null,
      serviceConfigProperties: [],
      serviceName: 'MISC',
      hdfsUser:"hdfs",
      nameServiceId: '',
      failedTask : null
    }),

    setCurrentStep: function (currentStep, completed) {
      this._super(currentStep, completed);
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: 'HIGH_AVAILABILITY_DEPLOY',
        wizardControllerName: 'highAvailabilityWizardController',
        localdb: App.db.data
      });
    },

    /**
     * return new object extended from clusterStatusTemplate
     * @return Object
     */
    getCluster: function(){
      return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
    },

    /**
     * Load services data from server.
     */
    loadServicesFromServer: function() {
      var displayOrderConfig = require('data/services');
      var apiUrl = App.get('stack2VersionURL');
      var apiService = this.loadServiceComponents(displayOrderConfig, apiUrl);
      //
      apiService.forEach(function(item, index){
        apiService[index].isSelected = App.Service.find().someProperty('id', item.serviceName);
        apiService[index].isDisabled = apiService[index].isSelected;
        apiService[index].isInstalled = apiService[index].isSelected;
      });
      this.set('content.services', apiService);
      App.db.setService(apiService);
    },

    /**
     * Load confirmed hosts.
     * Will be used at <code>Assign Masters(step5)</code> step
     */
    loadConfirmedHosts: function(){
      var hosts = App.db.getHosts();
      if(!hosts || !hosts.length){
        var hosts = {};

        App.Host.find().forEach(function(item){
          hosts[item.get('id')] = {
            name: item.get('id'),
            cpu: item.get('cpu'),
            memory: item.get('memory'),
            disk_info: item.get('diskInfo'),
            bootStatus: "REGISTERED",
            isInstalled: true
          };
        });
        App.db.setHosts(hosts);
      }

      this.set('content.hosts', hosts);
      console.log('ReassignMasterController.loadConfirmedHosts: loaded hosts', hosts);
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadMasterComponentHosts: function () {
      var masterComponentHosts = App.db.getMasterComponentHosts();
      if(!masterComponentHosts){
        masterComponentHosts = [];
        App.HostComponent.find().filterProperty('isMaster', true).forEach(function(item){
          masterComponentHosts.push({
            component: item.get('componentName'),
            hostName: item.get('host.hostName'),
            isInstalled: true
          })
        });

      }
      this.set("content.masterComponentHosts", masterComponentHosts);
      console.log("ReassignMasterController.loadMasterComponentHosts: loaded hosts ", masterComponentHosts);
    },

    /**
     * save status of the cluster.
     * @param clusterStatus object with status,requestId fields.
     */
    saveClusterStatus: function (clusterStatus) {
      var oldStatus = this.toObject(this.get('content.cluster'));
      clusterStatus = jQuery.extend(oldStatus, clusterStatus);
      if (clusterStatus.requestId) {
        clusterStatus.requestId.forEach(function (requestId) {
          if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
            clusterStatus.oldRequestsId.push(requestId)
          }
        }, this);
      }
      this.set('content.cluster', clusterStatus);
      this.save('cluster');
    },

    /**
     * Save Master Component Hosts data to Main Controller
     * @param stepController App.WizardStep5Controller
     */
    saveMasterComponentHosts: function (stepController) {
      var obj = stepController.get('selectedServicesMasters');
      var masterComponentHosts = [];
      obj.forEach(function (_component) {
        masterComponentHosts.push({
          display_name: _component.get('display_name'),
          component: _component.get('component_name'),
          hostName: _component.get('selectedHost'),
          serviceId: _component.get('serviceId'),
          isCurNameNode: _component.get('isCurNameNode'),
          isAddNameNode: _component.get('isAddNameNode'),
          isInstalled: true
        });
      });
      App.db.setMasterComponentHosts(masterComponentHosts);
      this.set('content.masterComponentHosts', masterComponentHosts);
    },

    saveHdfsUser: function () {
      App.db.setHighAvailabilityWizardHdfsUser(this.get('content.hdfsUser'));
    },

    saveTasksStatuses: function(statuses){
      App.db.setHighAvailabilityWizardTasksStatuses(statuses);
      this.set('content.tasksStatuses', statuses);
    },

    saveConfigTag: function(tag){
      App.db.setHighAvailabilityWizardConfigTag(tag);
      this.set('content.'+[tag.name], tag.value);
    },

    saveHdfsClientHosts: function(hostNames){
      App.db.setHighAvailabilityWizardHdfsClientHosts(hostNames);
      this.set('content.hdfsClientHostNames', hostNames);
    },

      /**
       * Save config properties
       * @param stepController HighAvailabilityWizardStep3Controller
       */
    saveServiceConfigProperties: function(stepController) {
      var serviceConfigProperties = [];
      var data = stepController.get('serverConfigData');

      var _content = stepController.get('stepConfigs')[0];
      _content.get('configs').forEach(function (_configProperties) {
        var siteObj = data.items.findProperty('type', _configProperties.get('filename'));
        if (siteObj) {
          siteObj.properties[_configProperties.get('name')] = _configProperties.get('value');
        }
      }, this);
      this.setDBProperty('serviceConfigProperties', data);
      this.set('content.serviceConfigProperties', data);
    },

    loadHdfsClientHosts: function(){
      var hostNames = App.db.getHighAvailabilityWizardHdfsClientHosts();
      if (!(hostNames instanceof Array)) {
        hostNames = [hostNames];
      }
      this.set('content.hdfsClientHostNames', hostNames);
    },

    loadConfigTag: function(tag){
      var tagVal = App.db.getHighAvailabilityWizardConfigTag(tag);
      this.set('content.'+tag, tagVal);
    },


    loadHdfsUser: function(){
      var hdfsUser = App.db.getHighAvailabilityWizardHdfsUser();
      this.set('content.hdfsUser', hdfsUser);
    },

    loadTasksStatuses: function(){
      var statuses = App.db.getHighAvailabilityWizardTasksStatuses();
      this.set('content.tasksStatuses', statuses);
    },

    /**
     * Load serviceConfigProperties to model
     */
    loadServiceConfigProperties: function () {
      var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
      this.set('content.serviceConfigProperties', serviceConfigProperties);
    },

    saveRequestIds: function(requestIds){
      App.db.setHighAvailabilityWizardRequestIds(requestIds);
      this.set('content.requestIds', requestIds);
    },

    saveLogs: function(logs){
      App.db.setHighAvailabilityWizardLogs(logs);
      this.set('content.logs', logs);
    },

    loadRequestIds: function(){
      var requestIds = App.db.getHighAvailabilityWizardRequestIds();
      this.set('content.requestIds', requestIds);
    },

    loadLogs: function(){
      var logs = App.db.getHighAvailabilityWizardLogs();
      this.set('content.logs', logs);
    },

    saveNameServiceId: function(nameServiceId){
      App.db.setHighAvailabilityWizardNameServiceId(nameServiceId);
      this.set('content.nameServiceId', nameServiceId);
    },

    loadNameServiceId: function(){
      var nameServiceId = App.db.getHighAvailabilityWizardNameServiceId();
      this.set('content.nameServiceId', nameServiceId);
    },

    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '9':
        case '8':
        case '7':
        case '6':
        case '5':
          this.loadTasksStatuses();
          this.loadRequestIds();
          this.loadLogs();
        case '4':
        case '3':
          this.loadNameServiceId();
          this.loadServiceConfigProperties();
        case '2':
          this.loadServicesFromServer();
          this.loadMasterComponentHosts();
          this.loadConfirmedHosts();
          this.loadHdfsUser();
        case '1':
          this.load('cluster');
      }
    },

    /**
     * Remove all loaded data.
     * Created as copy for App.router.clearAllSteps
     */
    clearAllSteps: function () {
      this.clearInstallOptions();
      // clear temporary information stored during the install
      this.set('content.cluster', this.getCluster());
    },

    clearTasksData: function () {
      this.saveTasksStatuses(undefined);
      this.saveRequestIds(undefined);
      this.saveLogs(undefined);
    },

    /**
     * Clear all temporary data
     */
    finish: function () {
      App.db.data.HighAvailabilityWizard = {};
      App.db.data.Installer = {};
      App.router.get('updateController').updateAll();
    }
  });
  
});
window.require.register("controllers/main/admin/highAvailability_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminHighAvailabilityController = Em.Controller.extend({
    name: 'mainAdminHighAvailabilityController',

    securityEnabled: false,

    tag: null,

    dataIsLoaded: false,

    enableHighAvailability: function () {
      var message = [];
      //Prerequisite Checks
      if (this.get('securityEnabled')) {
        this.showErrorPopup(Em.I18n.t('admin.highAvailability.error.security'));
        return;
      } else {
        if (App.HostComponent.find().findProperty('componentName', 'NAMENODE').get('workStatus') !== 'STARTED') {
          message.push(Em.I18n.t('admin.highAvailability.error.namenodeStarted'));
        }
        if (App.HostComponent.find().filterProperty('componentName', 'ZOOKEEPER_SERVER').length < 3) {
          message.push(Em.I18n.t('admin.highAvailability.error.zooKeeperNum'));
        }

        if (App.Host.find().content.length < 3) {
          message.push(Em.I18n.t('admin.highAvailability.error.hostsNum'));
        }
        if (message.length > 0) {
          this.showErrorPopup(message);
          return;
        }
      }
      App.router.transitionTo('enableHighAvailability');
    },

    disableHighAvailability: function () {
      App.router.transitionTo('rollbackHighAvailability');
    },

    setSecurityStatus: function () {
      if (App.testMode) {
        this.set('securityEnabled', !App.testEnableSecurity);
        this.set('dataIsLoaded', true);
      } else {
        //get Security Status From Server
        App.ajax.send({
          name: 'admin.security_status',
          sender: this,
          success: 'getSecurityStatusFromServerSuccessCallback',
          error: 'errorCallback'
        });
      }
    },

    errorCallback: function () {
      this.showErrorPopup(Em.I18n.t('admin.security.status.error'));
    },

    getSecurityStatusFromServerSuccessCallback: function (data) {
      var configs = data.Clusters.desired_configs;
      if ('global' in configs) {
        this.set('tag', configs['global'].tag);
        this.getServiceConfigsFromServer();
      }
      else {
        this.showErrorPopup(Em.I18n.t('admin.security.status.error'));
      }
    },

    getServiceConfigsFromServer: function () {
      var tags = [
        {
          siteName: "global",
          tagName: this.get('tag')
        }
      ];
      var data = App.router.get('configurationController').getConfigsByTags(tags);
      var configs = data.findProperty('tag', this.get('tag')).properties;
      if (configs && (configs['security_enabled'] === 'true' || configs['security_enabled'] === true)) {
        this.set('securityEnabled', true);
      } else {
        this.set('securityEnabled', false);
      }
      this.set('dataIsLoaded', true);
    },

    showErrorPopup: function (message) {
      if(Array.isArray(message)){
        message = message.join('<br/>');
      } else {
        message = '<p>' + message + '</p>';
      }
      App.ModalPopup.show({
        header: Em.I18n.t('common.error'),
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile(message)
        }),
        secondary: false
      });
    }
  });
  
});
window.require.register("controllers/main/admin/misc_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  require('controllers/main/service/info/configs');

  App.MainAdminMiscController = App.MainServiceInfoConfigsController.extend({
    name:'mainAdminMiscController',
    users: null,
    content: {
      serviceName: 'MISC'
    },
    loadUsers: function() {
      this.set('selectedService', this.get('content.serviceName'));
      this.loadServiceConfig();
    },
    loadServiceConfig: function() {
      App.ajax.send({
        name: 'config.tags',
        sender: this,
        data: {
          serviceName: this.get('content.serviceName'),
          serviceConfigsDef: this.get('serviceConfigs').findProperty('serviceName', this.get('content.serviceName'))
        },
        success: 'loadServiceTagSuccess'
      });
    },
    loadServiceTagSuccess: function(data, opt, params) {
      var installedServices = App.Service.find().mapProperty("serviceName");
      var serviceConfigsDef = params.serviceConfigsDef;
      var serviceName = this.get('content.serviceName');
      var loadedClusterSiteToTagMap = {};

      for ( var site in data.Clusters.desired_configs) {
        if (serviceConfigsDef.sites.indexOf(site) > -1) {
          loadedClusterSiteToTagMap[site] = data.Clusters.desired_configs[site]['tag'];
        }
      }
      this.setServiceConfigTags(loadedClusterSiteToTagMap);
      var configGroups = App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags'));
      var configSet = App.config.mergePreDefinedWithLoaded(configGroups, [], this.get('serviceConfigTags'), serviceName);

      var misc_configs = configSet.globalConfigs.filterProperty('serviceName', this.get('selectedService')).filterProperty('category', 'Users and Groups').filterProperty('isVisible', true);

      misc_configs = App.config.miscConfigVisibleProperty(misc_configs, installedServices);

      var sortOrder = this.get('configs').filterProperty('serviceName', this.get('selectedService')).filterProperty('category', 'Users and Groups').filterProperty('isVisible', true).mapProperty('name');

      var sorted = [];

      //stack, with version lower than 2.1, doesn't have Falcon service
      var proxyUserGroup = misc_configs.findProperty('name', 'proxyuser_group');
      if (proxyUserGroup && stringUtils.compareVersions(App.get('currentStackVersionNumber'), "2.1") !== 1) {
        proxyUserGroup.set('displayName', "Proxy group for Hive, WebHCat and Oozie");
      }

      if(sortOrder) {
        sortOrder.forEach(function(name) {
          var user = misc_configs.findProperty('name', name);
          if (user) {
            sorted.push({
              isVisible: user.get('isVisible'),
              displayName: user.get('displayName'),
              value: user.get('value')
            });
          }
        });
        this.set('users', sorted);
      }
      else {
        this.set('users', misc_configs);
      }
      if(this.get("content.hdfsUser")){
        this.get('content').set('hdfsUser', misc_configs.findProperty('name','hdfs_user').get("value"));
      }
      if (this.get("content.group")) {
        this.get('content').set('group', misc_configs.findProperty('name','user_group').get("value"));
      }
      this.set('dataIsLoaded', true);
    }
  });
  
});
window.require.register("controllers/main/admin/rollbackHA/rollback_wizard_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.RollbackHighAvailabilityWizardController = App.WizardController.extend({

    name: 'rollbackHighAvailabilityWizardController',

    totalSteps: 3,

    /**
     * Used for hiding back button in wizard
     */
    hideBackButton: true,

    content: Em.Object.create({
      controllerName: 'RollbackHighAvailabilityWizardController',
      cluster: null,
      masterComponentHosts: null,
      serviceName: 'MISC',
      hdfsUser:"hdfs",
      nameServiceId: '',
      selectedAddNNHost : null,
      selectedSNNHost : null,
      activeNNHost: null
    }),

    setCurrentStep: function (currentStep, completed) {
      this._super(currentStep, completed);
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: 'ROLLBACK_HIGH_AVAILABILITY',
        wizardControllerName: 'rollbackHighAvailabilityWizardController',
        localdb: App.db.data
      });
    },

    /**
     * return new object extended from clusterStatusTemplate
     * @return Object
     */
    getCluster: function(){
      return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
    },


    /**
     * save status of the cluster.
     * @param clusterStatus object with status,requestId fields.
     */
    saveClusterStatus: function (clusterStatus) {
      var oldStatus = this.toObject(this.get('content.cluster'));
      clusterStatus = jQuery.extend(oldStatus, clusterStatus);
      if (clusterStatus.requestId) {
        clusterStatus.requestId.forEach(function (requestId) {
          if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
            clusterStatus.oldRequestsId.push(requestId)
          }
        }, this);
      }
      this.set('content.cluster', clusterStatus);
      this.save('cluster');
    },

    saveTasksStatuses: function(statuses){
      App.db.setRollbackHighAvailabilityWizardTasksStatuses(statuses);
      this.set('content.tasksStatuses', statuses);
    },

    saveRequestIds: function(requestIds){
      App.db.setRollbackHighAvailabilityWizardRequestIds(requestIds);
      this.set('content.requestIds', requestIds);
    },

    saveLogs: function(logs){
      App.db.setRollbackHighAvailabilityWizardLogs(logs);
      this.set('content.logs', logs);
    },

    saveSelectedSNN: function(addNN){
      App.db.setRollBackHighAvailabilityWizardSelectedSNN(addNN);
      this.set('content.selectedAddNN', addNN);
    },

    saveSelectedAddNN: function(sNN){
      App.db.setRollBackHighAvailabilityWizardSelectedAddNN(sNN);
      this.set('content.selectedSNN', sNN);
    },

    loadAddNNHost: function () {
      var addNNHost = App.db.getRollBackHighAvailabilityWizardAddNNHost();
      this.set('content.addNNHost', addNNHost);
    },

    loadSNNHost: function () {
      var sNNHost = App.db.getRollBackHighAvailabilityWizardSNNHost();
      this.set('content.sNNHost', sNNHost);
    },

    loadTasksStatuses: function(){
      var sNNHost = App.db.getRollbackHighAvailabilityWizardTasksStatuses();
      this.set('content.tasksStatuses', sNNHost);
    },

    loadRequestIds: function(){
      var requestIds = App.db.getRollbackHighAvailabilityWizardRequestIds();
      this.set('content.requestIds', requestIds);
    },

    loadLogs: function(){
      var logs = App.db.getRollbackHighAvailabilityWizardLogs();
      this.set('content.logs', logs);
    },

    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '3':
        case '2':
        case '1':
          this.loadSNNHost();
          this.loadAddNNHost();
          this.load('cluster');
      }
    },

    /**
     * Remove all loaded data.
     * Created as copy for App.router.clearAllSteps
     */
    clearAllSteps: function () {
      this.clearInstallOptions();
      // clear temporary information stored during the install
      this.set('content.cluster', this.getCluster());
    },

    clearTasksData: function () {
      this.saveTasksStatuses(undefined);
      this.saveRequestIds(undefined);
      this.saveLogs(undefined);
    },

    /**
     * Clear all temporary data
     */
    finish: function () {
      this.setCurrentStep('1');
      this.clearAllSteps();
      App.router.get('updateController').updateAll();
    }
  });
  
});
window.require.register("controllers/main/admin/rollbackHA/step1_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.RollbackHighAvailabilityWizardStep1Controller = Em.Controller.extend({

    name:"rollbackHighAvailabilityWizardStep1Controller"

  });

  
});
window.require.register("controllers/main/admin/rollbackHA/step2_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.RollbackHighAvailabilityWizardStep2Controller = App.HighAvailabilityWizardStep4Controller.extend({
    name:"rollbackHighAvailabilityWizardStep2Controller",

    pullCheckPointStatus: function () {
      var hostName = this.get('content.activeNNHost');
      App.ajax.send({
        name: 'admin.high_availability.getNnCheckPointStatus',
        sender: this,
        data: {
          hostName: hostName
        },
        success: 'checkNnCheckPointStatus'
      });
    }
  });

  
});
window.require.register("controllers/main/admin/rollbackHA/step3_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.RollbackHighAvailabilityWizardStep3Controller = Em.Controller.extend({
    name:"rollbackHighAvailabilityWizardStep3Controller"
  });
  
});
window.require.register("controllers/main/admin/security", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.MainAdminSecurityController = Em.Controller.extend({
    name: 'mainAdminSecurityController',
    isSubmitDisabled: false,
    securityEnabled: false,
    dataIsLoaded: false,
    serviceUsers: [],
    tag: {},
    getAddSecurityWizardStatus: function () {
      return App.db.getSecurityWizardStatus();
    },
    setAddSecurityWizardStatus: function (status) {
      App.db.setSecurityWizardStatus(status);
    },

    setDisableSecurityStatus: function (status) {
      App.db.setDisableSecurityStatus(status);
    },
    getDisableSecurityStatus: function (status) {
      return App.db.getDisableSecurityStatus();
    },

    notifySecurityOff: false,
    notifySecurityAdd: false,

    stepConfigs: [],
    desiredConfigs: [],
    securityUsers: [],
    serviceConfigTags: [],
    selectedService: null,
    isNotEditable: true,
    services: function(){
      var secureServices;
      var services = [];
      if(App.get('isHadoop2Stack')) {
        secureServices = $.extend(true, [], require('data/HDP2/secure_configs'));
      } else {
        secureServices = $.extend(true, [], require('data/secure_configs'));
      }

      var installedServices = App.Service.find().mapProperty('serviceName');
      //General (only non service tab) tab is always displayed
      services.push(secureServices.findProperty('serviceName', 'GENERAL'));
      installedServices.forEach(function (_service) {
        var secureService = secureServices.findProperty('serviceName', _service);
        if (secureService) {
          services.push(secureService);
        }
      }, this);
      return services;
    }.property(),

    loadStep: function(){
      var step2Controller = App.router.get('mainAdminSecurityAddStep2Controller');
      var services = this.get('services');
      this.get('stepConfigs').clear();
      this.get('securityUsers').clear();
      this.get('serviceConfigTags').clear();
      this.loadSecurityUsers();
      //loadSecurityUsers - desired configs fetched from server
      step2Controller.addUserPrincipals(services, this.get('securityUsers'));
      step2Controller.addMasterHostToGlobals(services);
      step2Controller.addSlaveHostToGlobals(services);
      this.renderServiceConfigs(services);
      step2Controller.changeCategoryOnHa(services, this.get('stepConfigs'));

      services.forEach(function (_secureService) {
        this.setServiceTagNames(_secureService, this.get('desiredConfigs'));
      }, this);
      var serverConfigs = App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags'));
      this.setConfigValuesFromServer(this.get('stepConfigs'), serverConfigs);

      this.set('installedServices', App.Service.find().mapProperty('serviceName'));
    },

    /**
     * get actual values of configurations from server
     * @param stepConfigs
     * @param serverConfigs
     */
    setConfigValuesFromServer: function(stepConfigs, serverConfigs){
      var allConfigs = {};
      serverConfigs.mapProperty('properties').forEach(function(_properties){
        allConfigs = $.extend(allConfigs, _properties);
      }, this);
      // for all services`
      stepConfigs.forEach(function (_content) {
        //for all components
        _content.get('configs').forEach(function (_config) {

          var componentVal = allConfigs[_config.get('name')];
          //if we have config for specified component
          if (componentVal) {
            //set it
            _config.set('value', componentVal);
          }
        }, this);
      }, this);

    },

    /**
     * set tag names according to installed services and desired configs
     * @param secureService
     * @param configs
     * @return {Object}
     */
    setServiceTagNames: function (secureService, configs) {
      //var serviceConfigTags = this.get('serviceConfigTags');
      for (var index in configs) {
        if (secureService.sites && secureService.sites.contains(index)) {
          var serviceConfigObj = {
            siteName: index,
            tagName: configs[index].tag,
            newTagName: null,
            configs: {}
          };
          console.log("The value of serviceConfigTags[index]: " + configs[index]);
          this.get('serviceConfigTags').pushObject(serviceConfigObj);
        }
      }
      return serviceConfigObj;
    },

    loadSecurityUsers: function () {
      var securityUsers = this.get('serviceUsers');
      if (!securityUsers || securityUsers.length < 1) { // Page could be refreshed in middle
        if (App.testMode) {
          securityUsers.pushObject({id: 'puppet var', name: 'hdfs_user', value: 'hdfs'});
          securityUsers.pushObject({id: 'puppet var', name: 'mapred_user', value: 'mapred'});
          securityUsers.pushObject({id: 'puppet var', name: 'hbase_user', value: 'hbase'});
          securityUsers.pushObject({id: 'puppet var', name: 'hive_user', value: 'hive'});
          securityUsers.pushObject({id: 'puppet var', name: 'smokeuser', value: 'ambari-qa'});
        } else {
          this.setSecurityStatus();
          securityUsers = this.get('serviceUsers');
        }
      }
      this.set('securityUsers', securityUsers);
    },
    /**
     * fill config with hosts of component
     * @param service
     * @param configName
     * @param componentName
     */
    setHostsToConfig: function (service, configName, componentName) {
      if (service) {
        var hosts = service.configs.findProperty('name', configName);
        if (hosts) {
          hosts.defaultValue = App.Service.find(service.serviceName)
            .get('hostComponents')
            .filterProperty('componentName', componentName)
            .mapProperty('host.hostName');
        }
      }
    },

    /**
     * Load child components to service config object
     * @param _componentConfig
     * @param componentConfig
     */
    loadComponentConfigs: function (_componentConfig, componentConfig) {
      _componentConfig.configs.forEach(function (_serviceConfigProperty) {
        var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
        componentConfig.configs.pushObject(serviceConfigProperty);
        serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
        serviceConfigProperty.validate();
      }, this);
    },

    /**
     * Render configs for active services
     * @param serviceConfigs
     */
    renderServiceConfigs: function (serviceConfigs) {
      serviceConfigs.forEach(function (_serviceConfig) {

        var serviceConfig = App.ServiceConfig.create({
          filename: _serviceConfig.filename,
          serviceName: _serviceConfig.serviceName,
          displayName: _serviceConfig.displayName,
          configCategories: _serviceConfig.configCategories,
          showConfig: true,
          configs: []
        });

        this.loadComponentConfigs(_serviceConfig, serviceConfig);

        console.log('pushing ' + serviceConfig.serviceName, serviceConfig);

        this.get('stepConfigs').pushObject(serviceConfig);
      }, this);
      this.set('selectedService', this.get('stepConfigs').filterProperty('showConfig', true).objectAt(0));
    },

    notifySecurityOffPopup: function () {
      var self = this;
      if (!this.get('isSubmitDisabled')) {
        App.ModalPopup.show({
          header: Em.I18n.t('popup.confirmation.commonHeader'),
          primary: Em.I18n.t('ok'),
          onPrimary: function () {
            App.db.setSecurityDeployCommands(undefined);
            self.setDisableSecurityStatus("RUNNING");
            App.router.transitionTo('disableSecurity');
            this.hide();
          },
          bodyClass: Ember.View.extend({
            isMapReduceInstalled: App.Service.find().mapProperty('serviceName').contains('MAPREDUCE'),
            templateName: require('templates/main/admin/security/notify_security_off_popup')
          })
        })
      }
    },

    getUpdatedSecurityStatus: function () {
      this.setSecurityStatus();
      return this.get('securityEnabled');
    },

    setSecurityStatus: function () {
      if (App.testMode) {
        this.set('securityEnabled', !App.testEnableSecurity);
        this.set('dataIsLoaded', true);
      } else {
        //get Security Status From Server
        App.ajax.send({
          name: 'admin.security_status',
          sender: this,
          success: 'getSecurityStatusFromServerSuccessCallback',
          error: 'errorCallback'
        });
      }
    },

    errorCallback: function () {
      this.set('dataIsLoaded', true);
      this.showSecurityErrorPopup();
    },

    getSecurityStatusFromServerSuccessCallback: function (data) {
      var configs = data.Clusters.desired_configs;
      this.set('desiredConfigs', configs);
      if ('global' in configs && 'hdfs-site' in configs) {
        this.set('tag.global', configs['global'].tag);
        this.set('tag.hdfs-site', configs['hdfs-site'].tag);
        this.getServiceConfigsFromServer();
      }
      else {
        this.showSecurityErrorPopup();
      }
    },

    getServiceConfigsFromServer: function () {
      var tags = [
        {
          siteName: "global",
          tagName: this.get('tag.global')
        },
        {
          siteName: "hdfs-site",
          tagName: this.get('tag.hdfs-site')
        }
      ];

      var data = App.router.get('configurationController').getConfigsByTags(tags);
      var configs = data.findProperty('tag', this.get('tag.global')).properties;
      if (configs && (configs['security_enabled'] === 'true' || configs['security_enabled'] === true)) {
        this.set('securityEnabled', true);
      }
      else {
        this.set('securityEnabled', false);
        var hdfsConfigs = data.findProperty('tag', this.get('tag.hdfs-site')).properties;
        this.setNnHaStatus(hdfsConfigs);
      }
      this.loadUsers(configs);
      this.set('dataIsLoaded', true);
    },

    setNnHaStatus: function(hdfsConfigs) {
      var nnHaStatus = hdfsConfigs && hdfsConfigs['dfs.nameservices'];
      var namenodesKey;
      if (nnHaStatus) {
        namenodesKey = 'dfs.ha.namenodes.' + hdfsConfigs['dfs.nameservices'];
      }
      if(nnHaStatus && hdfsConfigs[namenodesKey]) {
        App.db.setIsNameNodeHa('true');
      } else {
        App.db.setIsNameNodeHa('false');
      }
    },

    loadUsers: function (configs) {
      this.setUserName('hdfs_user',configs, 'hdfs');
      this.setUserName('yarn_user',configs, 'yarn');
      this.setUserName('mapred_user',configs, 'mapred');
      this.setUserName('hbase_user',configs, 'hbase');
      this.setUserName('hive_user',configs, 'hive');
      this.setUserName('proxyuser_group',configs, 'users');
      this.setUserName('smokeuser',configs, 'ambari-qa');
      this.setUserName('zk_user',configs, 'zookeeper');
      this.setUserName('oozie_user',configs, 'oozie');
      this.setUserName('nagios_user',configs, 'nagios');
      this.setUserName('user_group',configs, 'hadoop');
      this.setUserName('storm_user',configs, 'storm');
      this.setUserName('falcon_user',configs,'falcon');
      App.db.setSecureUserInfo(this.get('serviceUsers'));
    },

    /**
     *
     * @param name
     * @param configs
     * @param defaultValue
     */
    setUserName: function(name,configs,defaultValue) {
      var serviceUsers = this.get('serviceUsers');
      serviceUsers.pushObject({
        id: 'puppet var',
        name: name,
        value: configs[name] ? configs[name] : defaultValue
      });
    },

    showSecurityErrorPopup: function () {
      App.ModalPopup.show({
        header: Em.I18n.t('common.error'),
        secondary: false,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<p>{{t admin.security.status.error}}</p>')
        })
      });
    }
  });


  
});
window.require.register("controllers/main/admin/security/add/addSecurity_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.AddSecurityController = App.WizardController.extend({

    name: 'addSecurityController',
    securityEnabled: false,

    totalSteps: 4,

    content: Em.Object.create({
      services: [],
      isNnHa: 'false',
      serviceConfigProperties: null,
      controllerName: 'addSecurityController'
    }),

    installedServices: function() {
      return App.Service.find().mapProperty('serviceName');
    }.property(),

    /**
     * Loads all prior steps on refresh
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '4':
        case '3':
        case '2':
          this.loadServiceConfigs();
        case '1':
          this.loadServices();
          this.loadNnHastatus();
      }
    },

    clearServices: function () {
      if (this.get('content.services')) {
        this.get('content.services').clear();
      }
    },

    /**
     * Loads all installed services
     */
    loadServices: function () {
      this.clearServices();
      var secureServices;
      if(App.get('isHadoop2Stack')) {
        secureServices = $.extend(true, [], require('data/HDP2/secure_configs'));
      } else {
        secureServices = $.extend(true, [], require('data/secure_configs'));
      }

      var installedServices = this.get('installedServices');
      //General (only non service tab) tab is always displayed
      this.get('content.services').push(secureServices.findProperty('serviceName', 'GENERAL'));
      installedServices.forEach(function (_service) {
        var secureService = secureServices.findProperty('serviceName', _service);
        if (secureService) {
          this.get('content.services').push(secureService);
        }
      }, this);

    },

    loadNnHastatus: function() {
      var isNnHa = App.db.getIsNameNodeHa();
      this.set('content.isNnHa', isNnHa);
    },

    saveServiceConfigProperties: function (stepController) {
      var serviceConfigProperties = [];
      stepController.get('stepConfigs').forEach(function (_content) {
        _content.get('configs').forEach(function (_configProperties) {
          _configProperties.set('value', App.config.trimProperty(_configProperties,true));
          var overrides = _configProperties.get('overrides');
          var overridesArray = [];
          if(overrides!=null){
            overrides.forEach(function(override){
              var overrideEntry = {
                value: override.get('value'),
                hosts: []
              };
              override.get('selectedHostOptions').forEach(function(host){
                overrideEntry.hosts.push(host);
              });
              overridesArray.push(overrideEntry);
            });
          }
          overridesArray = (overridesArray.length) ? overridesArray : null;
          var configProperty = {
            id: _configProperties.get('id'),
            name: _configProperties.get('name'),
            value: _configProperties.get('value'),
            defaultValue: _configProperties.get('defaultValue'),
            serviceName: _configProperties.get('serviceName'),
            domain:  _configProperties.get('domain'),
            filename: _configProperties.get('filename'),
            unit: _configProperties.get('unit'),
            components: _configProperties.get('components'),
            component: _configProperties.get('component'),
            overrides: overridesArray
          };
          serviceConfigProperties.push(configProperty);
        }, this);
      }, this);
      App.db.setSecureConfigProperties(serviceConfigProperties);
      this.set('content.serviceConfigProperties', serviceConfigProperties);
    },

    /**
     * Loads all service config properties
     */

    loadServiceConfigs: function () {
      var serviceConfigProperties = App.db.getSecureConfigProperties();
      this.set('content.serviceConfigProperties', serviceConfigProperties);
    }
  });

  
});
window.require.register("controllers/main/admin/security/add/step1", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.MainAdminSecurityAddStep1Controller = Em.Controller.extend({

    name: 'mainAdminSecurityAddStep1Controller',
    isATSInstalled: function() {
      return this.get('content.services').findProperty('serviceName', 'YARN') &&
        App.Service.find('YARN').get('hostComponents').someProperty('componentName', 'APP_TIMELINE_SERVER')
    }
  });
});
window.require.register("controllers/main/admin/security/add/step2", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.MainAdminSecurityAddStep2Controller = Em.Controller.extend({

    name: 'mainAdminSecurityAddStep2Controller',
    stepConfigs: [],
    installedServices: [],
    selectedService: null,
    securityUsers: [],

    isSubmitDisabled: function () {
      return !this.stepConfigs.filterProperty('showConfig', true).everyProperty('errorCount', 0);
    }.property('stepConfigs.@each.errorCount'),

    clearStep: function () {
      this.get('stepConfigs').clear();
      this.get('securityUsers').clear();
    },


    /**
     *  Function is called whenever the step is loaded
     */
    loadStep: function () {
      console.log("TRACE: Loading addSecurity step2: Configure Services");
      this.clearStep();
      this.loadUsers();
      this.addUserPrincipals(this.get('content.services'), this.get('securityUsers'));
      this.addMasterHostToGlobals(this.get('content.services'));
      this.addSlaveHostToGlobals(this.get('content.services'));
      this.renderServiceConfigs(this.get('content.services'));
      this.changeCategoryOnHa(this.get('content.services'), this.get('stepConfigs'));
      var storedServices = this.get('content.serviceConfigProperties');
      if (storedServices) {
        var configs = new Ember.Set();

        // for all services`
        this.get('stepConfigs').forEach(function (_content) {
          //for all components
          _content.get('configs').forEach(function (_config) {

            var componentVal = storedServices.findProperty('name', _config.get('name'));
            //if we have config for specified component
            if (componentVal) {
              //set it
              _config.set('value', componentVal.value);
            }

          }, this);
        }, this);

      }
      //
      this.set('installedServices', App.Service.find().mapProperty('serviceName'));
      console.log("The services are: " + this.get('installedServices'));
      //
    },

    /**
     * Render configs for active services
     * @param serviceConfigs
     */
    renderServiceConfigs: function (serviceConfigs) {
      serviceConfigs.forEach(function (_serviceConfig) {

        var serviceConfig = App.ServiceConfig.create({
          filename: _serviceConfig.filename,
          serviceName: _serviceConfig.serviceName,
          displayName: _serviceConfig.displayName,
          configCategories: _serviceConfig.configCategories,
          showConfig: true,
          configs: []
        });

        this.loadComponentConfigs(_serviceConfig, serviceConfig);

        this.get('stepConfigs').pushObject(serviceConfig);
      }, this);
      this.set('selectedService', this.get('stepConfigs').filterProperty('showConfig', true).objectAt(0));
    },

    /**
     * Load child components to service config object
     * @param _componentConfig
     * @param componentConfig
     */
    loadComponentConfigs: function (_componentConfig, componentConfig) {
      _componentConfig.configs.forEach(function (_serviceConfigProperty) {
        var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
        componentConfig.configs.pushObject(serviceConfigProperty);
        serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
        serviceConfigProperty.validate();
      }, this);
    },

    /**
     * fill config with hosts of component
     * @param service
     * @param configName
     * @param componentName
     */
    setHostsToConfig: function (service, configName, componentName) {
      if (service) {
        var hosts = service.configs.findProperty('name', configName);
        if (hosts) {
          hosts.defaultValue = App.Service.find(service.serviceName)
            .get('hostComponents')
            .filterProperty('componentName', componentName)
            .mapProperty('host.hostName');
        }
      }
    },

    /**
     * fill principal _HOST part with actual hostname of component
     * @param service
     * @param hostConfigName
     * @param principalConfigName
     * @param defaultPrimaryName
     */
    setHostToPrincipal: function (service, hostConfigName, principalConfigName, defaultPrimaryName) {
      if (service) {
        var host = service.configs.findProperty('name', hostConfigName);
        var principal = service.configs.findProperty('name', principalConfigName);
        if (host && principal) {
          if (host.defaultValue instanceof Array) {
            host.defaultValue = host.defaultValue[0];
          }
          principal.defaultValue = defaultPrimaryName + host.defaultValue.toLowerCase();
        }
      }
    },


    loadUsers: function () {
      var securityUsers = App.router.get('mainAdminSecurityController').get('serviceUsers');
      if (!securityUsers || securityUsers.length < 1) { // Page could be refreshed in middle
        if (App.testMode) {
          securityUsers.pushObject({id: 'puppet var', name: 'hdfs_user', value: 'hdfs'});
          securityUsers.pushObject({id: 'puppet var', name: 'mapred_user', value: 'mapred'});
          securityUsers.pushObject({id: 'puppet var', name: 'hbase_user', value: 'hbase'});
          securityUsers.pushObject({id: 'puppet var', name: 'hive_user', value: 'hive'});
          securityUsers.pushObject({id: 'puppet var', name: 'smokeuser', value: 'ambari-qa'});
        } else {
          securityUsers = App.db.getSecureUserInfo();
        }
      }
      this.set('securityUsers', securityUsers);
    },

    addUserPrincipals: function (serviceConfigs, securityUsers) {
      var smokeUser = securityUsers.findProperty('name', 'smokeuser');
      var hdfsUser = securityUsers.findProperty('name', 'hdfs_user');
      var hbaseUser = securityUsers.findProperty('name', 'hbase_user');
      var generalService = serviceConfigs.findProperty('serviceName', 'GENERAL');
      var smokeUserPrincipal = generalService.configs.findProperty('name', 'smokeuser_principal_name');
      var hdfsUserPrincipal = generalService.configs.findProperty('name', 'hdfs_principal_name');
      var hbaseUserPrincipal = generalService.configs.findProperty('name', 'hbase_principal_name');
      var hbaseUserKeytab = generalService.configs.findProperty('name', 'hbase_user_keytab');
      var hbaseService = serviceConfigs.findProperty('serviceName', 'HBASE');
      if (smokeUser && smokeUserPrincipal) {
        smokeUserPrincipal.defaultValue = smokeUser.value;
      }
      if (hdfsUser && hdfsUserPrincipal) {
        hdfsUserPrincipal.defaultValue = hdfsUser.value;
      }
      if (hbaseService && hbaseUser && hbaseUserPrincipal) {
        hbaseUserPrincipal.defaultValue = hbaseUser.value;
        hbaseUserPrincipal.isVisible = true;
        hbaseUserKeytab.isVisible = true;
      }
    },

    addSlaveHostToGlobals: function (serviceConfigs) {
      var serviceComponentMap = [
        {
          serviceName: 'HDFS',
          configName: 'datanode_hosts',
          component: 'DATANODE'
        },
        {
          serviceName: 'MAPREDUCE',
          configName: 'tasktracker_hosts',
          component: 'TASKTRACKER'
        },
        {
          serviceName: 'YARN',
          configName: 'nodemanager_host',
          component: 'NODEMANAGER'
        },
        {
          serviceName: 'HBASE',
          configName: 'regionserver_hosts',
          component: 'HBASE_REGIONSERVER'
        },
        {
          serviceName: 'STORM',
          configName: 'supervisor_hosts',
          component: 'SUPERVISOR'
        }
      ];
      serviceComponentMap.forEach(function(service) {
        this.setHostsToConfig(serviceConfigs.findProperty('serviceName', service.serviceName), service.configName, service.component);
      }, this);
    },

    addMasterHostToGlobals: function (serviceConfigs) {
      var oozieService = serviceConfigs.findProperty('serviceName', 'OOZIE');
      var hiveService = serviceConfigs.findProperty('serviceName', 'HIVE');
      var webHcatService = serviceConfigs.findProperty('serviceName', 'WEBHCAT');
      var nagiosService = serviceConfigs.findProperty('serviceName', 'NAGIOS');
      var hbaseService = serviceConfigs.findProperty('serviceName', 'HBASE');
      var zooKeeperService = serviceConfigs.findProperty('serviceName', 'ZOOKEEPER');
      var hdfsService = serviceConfigs.findProperty('serviceName', 'HDFS');
      var mapReduceService = serviceConfigs.findProperty('serviceName', 'MAPREDUCE');
      var mapReduce2Service = serviceConfigs.findProperty('serviceName', 'MAPREDUCE2');
      var yarnService = serviceConfigs.findProperty('serviceName', 'YARN');
      var stormService = serviceConfigs.findProperty('serviceName', 'STORM');
      var falconService = serviceConfigs.findProperty('serviceName', 'FALCON');


      this.setHostsToConfig(oozieService, 'oozie_servername', 'OOZIE_SERVER');
      this.setHostsToConfig(hiveService, 'hive_metastore', 'HIVE_SERVER');
      this.setHostsToConfig(webHcatService, 'webhcatserver_host', 'WEBHCAT_SERVER');
      this.setHostsToConfig(nagiosService, 'nagios_server', 'NAGIOS_SERVER');
      this.setHostsToConfig(hdfsService, 'namenode_host', 'NAMENODE');
      this.setHostsToConfig(hdfsService, 'snamenode_host', 'SECONDARY_NAMENODE');
      this.setHostsToConfig(hdfsService, 'journalnode_hosts', 'JOURNALNODE');
      this.setHostsToConfig(mapReduceService, 'jobtracker_host', 'JOBTRACKER');
      this.setHostsToConfig(mapReduceService, 'jobhistoryserver_host', 'HISTORYSERVER');
      this.setHostsToConfig(mapReduce2Service, 'jobhistoryserver_host', 'HISTORYSERVER');
      this.setHostsToConfig(yarnService, 'resourcemanager_host', 'RESOURCEMANAGER');
      this.setHostsToConfig(hbaseService, 'hbasemaster_host', 'HBASE_MASTER');
      this.setHostsToConfig(zooKeeperService, 'zookeeperserver_hosts', 'ZOOKEEPER_SERVER');
      this.setHostsToConfig(falconService, 'falcon_server_host', 'FALCON_SERVER');
      if (stormService) {
        var stormComponents = ['STORM_UI_SERVER','NIMBUS','SUPERVISOR'];
        var stormHosts = [];
        stormComponents.forEach(function(componentName) {
          stormHosts.pushObjects(App.Service.find(stormService.serviceName)
            .get('hostComponents')
            .filterProperty('componentName', componentName)
            .mapProperty('host.hostName'));
        }, this);
        var hosts = stormService.configs.findProperty('name', 'storm_host');
        hosts.defaultValue  = stormHosts.uniq();
      }

      // Oozie, Falcon, WebHcat and Nagios does not support _HOST in the principal name. Actual hostname should be set instead of _HOST

      this.setHostToPrincipal(oozieService, 'oozie_servername','oozie_principal_name','oozie/');
      this.setHostToPrincipal(oozieService, 'oozie_servername','oozie_http_principal_name','HTTP/');
      this.setHostToPrincipal(falconService, 'falcon_server_host','falcon_principal_name','falcon/');
      this.setHostToPrincipal(falconService, 'falcon_server_host','falcon_http_principal_name','HTTP/');
      this.setHostToPrincipal(webHcatService, 'webhcatserver_host','webHCat_http_principal_name','HTTP/');
      this.setHostToPrincipal(nagiosService, 'nagios_server','nagios_principal_name','nagios/');
    },

    changeCategoryOnHa: function (serviceConfigs, stepConfigs) {
      var hdfsService = serviceConfigs.findProperty('serviceName', 'HDFS');
      if (hdfsService) {
        var hdfsProperties = stepConfigs.findProperty('serviceName', 'HDFS').get('configs');
        var configCategories = hdfsService.configCategories;
        if ((App.testMode && App.testNameNodeHA) || (this.get('content.isNnHa') === 'true')) {
          hdfsProperties.filterProperty('category', 'SNameNode').forEach(function (_snConfig) {
            _snConfig.set('isVisible', false);
          }, this);
          var snCategory = configCategories.findProperty('name', 'SNameNode');
          if (snCategory) {
            configCategories.removeObject(snCategory);
          }
        } else {
          hdfsProperties.filterProperty('category', 'JournalNode').forEach(function (_jnConfig) {
            _jnConfig.set('isVisible', false);
          }, this);
          var jnCategory = configCategories.findProperty('name', 'JournalNode');
          if (jnCategory) {
            configCategories.removeObject(jnCategory);
          }
        }
      }
    },

    /**
     *  submit and move to step3
     */

    submit: function () {
      if (!this.get('isSubmitDisabled')) {
        App.router.send('next');
      }
    }

  });
  
});
window.require.register("controllers/main/admin/security/add/step3", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.MainAdminSecurityAddStep3Controller = Em.Controller.extend({
    name: 'mainAdminSecurityAddStep3Controller',
    hostComponents: [],
    doDownloadCsv: function () {
      if ($.browser.msie && $.browser.version < 10) {
        this.openInfoInNewTab();
      } else {
        try {
          var blob = new Blob([stringUtils.arrayToCSV(this.get('hostComponents'))], {type: "text/csv;charset=utf-8;"});
          saveAs(blob, "host-principal-keytab-list.csv");
        } catch(e) {
           this.openInfoInNewTab();
        }
      }
    },
    openInfoInNewTab: function () {
      var newWindow = window.open('');
      var newDocument = newWindow.document;
      newDocument.write(stringUtils.arrayToCSV(this.get('hostComponents')));
      newWindow.focus();
    },
    loadStep: function(){
      var configs = this.get('content.serviceConfigProperties');
      var hosts = App.Host.find();
      var result = [];
      var componentsToDisplay = ['NAMENODE', 'SECONDARY_NAMENODE', 'DATANODE', 'JOBTRACKER', 'ZOOKEEPER_SERVER', 'HIVE_SERVER', 'TASKTRACKER',
        'OOZIE_SERVER', 'NAGIOS_SERVER', 'HBASE_MASTER', 'HBASE_REGIONSERVER','HISTORYSERVER','RESOURCEMANAGER','NODEMANAGER','JOURNALNODE',
        'SUPERVISOR', 'NIMBUS', 'STORM_UI_SERVER','FALCON_SERVER'];
      var securityUsers = [];
      if (!securityUsers || securityUsers.length < 1) { // Page could be refreshed in middle
        securityUsers = this.getSecurityUsers();
      }
      var isHbaseInstalled = App.Service.find().findProperty('serviceName', 'HBASE');
      var isStormInstalled = App.Service.find().findProperty('serviceName', 'STORM');
      var generalConfigs = configs.filterProperty('serviceName', 'GENERAL');
      var hdfsConfigs = configs.filterProperty('serviceName', 'HDFS');
      var realm = generalConfigs.findProperty('name', 'kerberos_domain').value;
      var smokeUserId = securityUsers.findProperty('name', 'smokeuser').value;
      var hdfsUserId = securityUsers.findProperty('name', 'hdfs_user').value;
      var hbaseUserId = securityUsers.findProperty('name', 'hbase_user').value;
      var mapredUserId = securityUsers.findProperty('name', 'mapred_user').value;
      var yarnUserId =  securityUsers.findProperty('name', 'yarn_user').value;
      var hiveUserId = securityUsers.findProperty('name', 'hive_user').value;
      var zkUserId = securityUsers.findProperty('name', 'zk_user').value;
      var oozieUserId = securityUsers.findProperty('name', 'oozie_user').value;
      var nagiosUserId = securityUsers.findProperty('name', 'nagios_user').value;
      var hadoopGroupId = securityUsers.findProperty('name', 'user_group').value;
      var stormUserId = securityUsers.findProperty('name', 'storm_user').value;
      var falconUserId =  securityUsers.findProperty('name', 'falcon_user').value;

      var smokeUser = smokeUserId + '@' + realm;
      var hdfsUser = hdfsUserId + '@' + realm;
      var hbaseUser = hbaseUserId + '@' + realm;
      var stormUser = stormUserId + '@' + realm;

      var smokeUserKeytabPath = generalConfigs.findProperty('name', 'smokeuser_keytab').value;
      var hdfsUserKeytabPath = generalConfigs.findProperty('name', 'hdfs_user_keytab').value;
      var hbaseUserKeytabPath = generalConfigs.findProperty('name', 'hbase_user_keytab').value;

      var hadoopHttpPrincipal = hdfsConfigs.findProperty('name', 'hadoop_http_principal_name');
      var hadoopHttpKeytabPath = hdfsConfigs.findProperty('name', 'hadoop_http_keytab').value;
      var componentToOwnerMap = {
        'NAMENODE': hdfsUserId,
        'SECONDARY_NAMENODE': hdfsUserId,
        'DATANODE': hdfsUserId,
        'JOURNALNODE': hdfsUserId,
        'TASKTRACKER': mapredUserId,
        'JOBTRACKER': mapredUserId,
        'HISTORYSERVER': mapredUserId,
        'RESOURCEMANAGER':yarnUserId,
        'NODEMANAGER':yarnUserId,
        'ZOOKEEPER_SERVER': zkUserId,
        'HIVE_SERVER': hiveUserId,
        'OOZIE_SERVER': oozieUserId,
        'NAGIOS_SERVER': nagiosUserId,
        'HBASE_MASTER': hbaseUserId,
        'HBASE_REGIONSERVER': hbaseUserId,
        'SUPERVISOR': stormUserId,
        'NIMBUS': stormUserId,
        'STORM_UI_SERVER': stormUserId,
        'FALCON_SERVER': falconUserId
      };

      var addedPrincipalsHost = {}; //Keys = host_principal, Value = 'true'

      hosts.forEach(function (host) {
        result.push({
          host: host.get('hostName'),
          component: Em.I18n.t('admin.addSecurity.user.smokeUser'),
          principal: smokeUser,
          keytabFile: stringUtils.getFileFromPath(smokeUserKeytabPath),
          keytab: stringUtils.getPath(smokeUserKeytabPath),
          owner: smokeUserId,
          group: hadoopGroupId,
          acl: '440'
        });
        result.push({
          host: host.get('hostName'),
          component: Em.I18n.t('admin.addSecurity.user.hdfsUser'),
          principal: hdfsUser,
          keytabFile: stringUtils.getFileFromPath(hdfsUserKeytabPath),
          keytab: stringUtils.getPath(hdfsUserKeytabPath),
          owner: hdfsUserId,
          group: hadoopGroupId,
          acl: '440'
        });
        if (isHbaseInstalled) {
          result.push({
            host: host.get('hostName'),
            component: Em.I18n.t('admin.addSecurity.user.hbaseUser'),
            principal: hbaseUser,
            keytabFile: stringUtils.getFileFromPath(hbaseUserKeytabPath),
            keytab: stringUtils.getPath(hbaseUserKeytabPath),
            owner: hbaseUserId,
            group: hadoopGroupId,
            acl: '440'
          });
        }

        this.setComponentConfig(result,host,'NAMENODE','HDFS','hadoop_http_principal_name','hadoop_http_keytab',Em.I18n.t('admin.addSecurity.hdfs.user.httpUser'),hadoopGroupId);
        this.setComponentConfig(result,host,'SECONDARY_NAMENODE','HDFS','hadoop_http_principal_name','hadoop_http_keytab',Em.I18n.t('admin.addSecurity.hdfs.user.httpUser'),hadoopGroupId);
        this.setComponentConfig(result,host,'JOURNALNODE','HDFS','hadoop_http_principal_name','hadoop_http_keytab',Em.I18n.t('admin.addSecurity.hdfs.user.httpUser'),hadoopGroupId);
        this.setComponentConfig(result,host,'WEBHCAT_SERVER','WEBHCAT','webHCat_http_principal_name','webhcat_http_keytab',Em.I18n.t('admin.addSecurity.webhcat.user.httpUser'),hadoopGroupId);
        this.setComponentConfig(result,host,'OOZIE_SERVER','OOZIE','oozie_http_principal_name','oozie_http_keytab',Em.I18n.t('admin.addSecurity.oozie.user.httpUser'),hadoopGroupId);
        this.setComponentConfig(result,host,'FALCON_SERVER','FALCON','falcon_http_principal_name','falcon_http_keytab',Em.I18n.t('admin.addSecurity.falcon.user.httpUser'),hadoopGroupId);
        //Derive Principal name and Keytabs only if its HDP-2 stack
        if (App.get('isHadoop2Stack')) {
          this.setComponentConfig(result,host,'HISTORYSERVER','MAPREDUCE2','jobhistory_http_principal_name','jobhistory_http_keytab',Em.I18n.t('admin.addSecurity.historyServer.user.httpUser'),hadoopGroupId);
          this.setComponentConfig(result,host,'RESOURCEMANAGER','YARN','resourcemanager_http_principal_name','resourcemanager_http_keytab',Em.I18n.t('admin.addSecurity.rm.user.httpUser'),hadoopGroupId);
          this.setComponentConfig(result,host,'NODEMANAGER','YARN','nodemanager_http_principal_name','nodemanager_http_keytab',Em.I18n.t('admin.addSecurity.nm.user.httpUser'),hadoopGroupId);
        }

        host.get('hostComponents').forEach(function(hostComponent){
          if(componentsToDisplay.contains(hostComponent.get('componentName'))){
            var serviceConfigs = configs.filterProperty('serviceName', hostComponent.get('service.serviceName'));
            var principal, keytab;
            serviceConfigs.forEach(function (config) {
              if (config.component && config.component === hostComponent.get('componentName')) {
                if (config.name.endsWith('_principal_name')) {
                  principal = config.value.replace('_HOST', host.get('hostName').toLowerCase()) + config.unit;
                } else if (config.name.endsWith('_keytab') || config.name.endsWith('_keytab_path')) {
                  keytab = config.value;
                }
              } else if (config.components && config.components.contains(hostComponent.get('componentName'))) {
                if (config.name.endsWith('_principal_name')) {
                  principal = config.value.replace('_HOST', host.get('hostName').toLowerCase()) + config.unit;
                } else if (config.name.endsWith('_keytab') || config.name.endsWith('_keytab_path')) {
                  keytab = config.value;
                }
              }
            });
            var displayName = this.changeDisplayName(hostComponent.get('displayName'));
            var key = host.get('hostName') + "--" + principal;
            if (!addedPrincipalsHost[key]) {
              var owner = componentToOwnerMap[hostComponent.get('componentName')];
              if(!owner){
                owner = '';
              }
              result.push({
                host: host.get('hostName'),
                component: displayName,
                principal: principal,
                keytabFile: stringUtils.getFileFromPath(keytab),
                keytab: stringUtils.getPath(keytab),
                owner: owner,
                group: hadoopGroupId,
                acl: '400'
              });
              addedPrincipalsHost[key] = true;
            }
          }
        },this);
      },this);
      this.set('hostComponents', result);
    },

    getSecurityUsers: function() {
      var securityUsers = [];
      if (App.testMode) {
        securityUsers.pushObject({id: 'puppet var', name: 'hdfs_user', value: 'hdfs'});
        securityUsers.pushObject({id: 'puppet var', name: 'mapred_user', value: 'mapred'});
        securityUsers.pushObject({id: 'puppet var', name: 'yarn_user', value: 'yarn'});
        securityUsers.pushObject({id: 'puppet var', name: 'hbase_user', value: 'hbase'});
        securityUsers.pushObject({id: 'puppet var', name: 'hive_user', value: 'hive'});
        securityUsers.pushObject({id: 'puppet var', name: 'falcon_user', value: 'falcon'});
        securityUsers.pushObject({id: 'puppet var', name: 'smokeuser', value: 'ambari-qa'});
        securityUsers.pushObject({id: 'puppet var', name: 'zk_user', value: 'zookeeper'});
        securityUsers.pushObject({id: 'puppet var', name: 'oozie_user', value: 'oozie'});
        securityUsers.pushObject({id: 'puppet var', name: 'nagios_user', value: 'nagios'});
        securityUsers.pushObject({id: 'puppet var', name: 'user_group', value: 'hadoop'});
      } else {
        securityUsers = App.db.getSecureUserInfo();
      }
      return securityUsers;
    },

    setComponentConfig: function(hostComponents,host,componentName,serviceName,principal,keytab,displayName,groupId) {
      if (host.get('hostComponents').someProperty('componentName', componentName)) {
        var result = {};
        var configs = this.get('content.serviceConfigProperties');
        var serviceConfigs = configs.filterProperty('serviceName', serviceName);
        var servicePrincipal = serviceConfigs.findProperty('name', principal);
        var serviceKeytabPath = serviceConfigs.findProperty('name', keytab).value;
        result.host = host.get('hostName');
        result.component = displayName;
        result.principal = servicePrincipal.value.replace('_HOST', host.get('hostName').toLowerCase()) + servicePrincipal.unit;
        result.keytabfile = stringUtils.getFileFromPath(serviceKeytabPath);
        result.keytab = stringUtils.getPath(serviceKeytabPath);
        result.owner = 'root';
        result.group = groupId;
        result.acl = '440';
        hostComponents.push(result);
      }
    },

    changeDisplayName: function (name) {
      if (name === 'HiveServer2') {
        return 'Hive Metastore and HiveServer2';
      } else {
        return name;
      }
    }
  });
  
});
window.require.register("controllers/main/admin/security/add/step4", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.MainAdminSecurityAddStep4Controller = App.MainAdminSecurityProgressController.extend({

    name: 'mainAdminSecurityAddStep4Controller',

    serviceUsersBinding: 'App.router.mainAdminSecurityController.serviceUsers',

    secureServices: function() {
      return  this.get('content.services');
    }.property('content.services'),

    isBackBtnDisabled: function () {
      return !this.get('commands').someProperty('isError', true);
    }.property('commands.@each.isCompleted'),

    isOozieSelected: function () {
      return this.get('secureServices').someProperty('serviceName', 'OOZIE');
    }.property('secureServices'),

    isHiveSelected: function () {
      return this.get('secureServices').someProperty('serviceName', 'HIVE');
    }.property('secureServices'),

    isNagiosSelected: function () {
      return this.get('secureServices').someProperty('serviceName', 'NAGIOS');
    }.property('secureServices'),

    isZkSelected: function () {
      return this.get('secureServices').someProperty('serviceName', 'ZOOKEEPER');
    }.property('secureServices'),

    isWebHcatSelected: function () {
      var installedServices = App.Service.find().mapProperty('serviceName');
      return installedServices.contains('WEBHCAT');
    },

    isSecurityApplied: function () {
      return this.get('commands').someProperty('name', 'START_SERVICES') && this.get('commands').findProperty('name', 'START_SERVICES').get('isSuccess');
    }.property('commands.@each.isCompleted'),

    clearStep: function () {
      this.set('commands',[]);
      this.set('isSubmitDisabled', true);
      this.set('isBackBtnDisabled', true);
      this.get('serviceConfigTags').clear();
    },

    loadCommands: function () {
      this._super();
      // no need to remove ATS component if YARN and ATS are not installed
      if (this.get('secureServices').findProperty('serviceName', 'YARN') && App.Service.find('YARN').get('hostComponents').someProperty('componentName', 'APP_TIMELINE_SERVER')) {
        this.get('commands').splice(2, 0, App.Poll.create({name: 'DELETE_ATS', label: Em.I18n.translations['admin.addSecurity.apply.delete.ats'], isPolling: false}));
        this.set('totalSteps', 4);
      }
      this.setIndex(this.get('commands'));
    },

    loadStep: function () {
      this.set('secureMapping', require('data/secure_mapping').slice(0));
      this.clearStep();
      var commands = App.db.getSecurityDeployCommands();
      this.prepareSecureConfigs();
      if (commands && commands.length > 0) {
        commands.forEach(function (_command, index) {
          commands[index] = App.Poll.create(_command);
        }, this);
        if (commands.someProperty('isError', true)) {
          this.get('commands').pushObjects(commands);
          this.addObserver('commands.@each.isSuccess', this, 'onCompleteCommand');
          return;
        } else if (commands.filterProperty('isStarted', true).someProperty('isCompleted', false)) {
          var runningCommand = commands.filterProperty('isStarted', true).findProperty('isCompleted', false);
          runningCommand.set('isStarted', false);
          this.get('commands').pushObjects(commands);
        } else {
          this.get('commands').pushObjects(commands);
        }
      } else {
        this.loadCommands();
        this.addInfoToCommands();
        var runningOperations = App.router.get('backgroundOperationsController.services').filterProperty('isRunning');
        var stopAllOperation = runningOperations.findProperty('name', 'Stop All Services');
        var stopCommand = this.get('commands').findProperty('name', 'STOP_SERVICES');
        if (stopCommand.get('name') === 'STOP_SERVICES' && stopAllOperation) {
          stopCommand.set('requestId', stopAllOperation.get('id'));
        }
      }
      this.addObserver('commands.@each.isSuccess', this, 'onCompleteCommand');
      this.moveToNextCommand();
    },

    enableSubmit: function () {
      var addSecurityController = App.router.get('addSecurityController');
      if (this.get('commands').someProperty('isError', true) || this.get('commands').everyProperty('isSuccess', true)) {
        this.set('isSubmitDisabled', false);
        if (this.get('commands').someProperty('isError', true)) {
          addSecurityController.setStepsEnable();
        }
      } else {
        this.set('isSubmitDisabled', true);
        addSecurityController.setLowerStepsDisable(4);
      }
    }.observes('commands.@each.isCompleted'),

    loadUiSideConfigs: function () {
      var uiConfig = [];
      var configs = this.get('secureMapping').filterProperty('foreignKey', null);
      configs.forEach(function (_config) {
        var value = _config.value;
        if (_config.hasOwnProperty('dependedServiceName')) {
          value = this.checkServiceForConfigValue(value, _config.dependedServiceName);
        }
        value = this.getGlobConfigValue(_config.templateName, value, _config.name);
        uiConfig.pushObject({
          "id": "site property",
          "name": _config.name,
          "value": value,
          "filename": _config.filename
        });
      }, this);
      var dependentConfig = this.get('secureMapping').filterProperty('foreignKey');
      dependentConfig.forEach(function (_config) {
        if (App.Service.find().mapProperty('serviceName').contains(_config.serviceName)) {
          this.setConfigValue(uiConfig, _config);
          uiConfig.pushObject({
            "id": "site property",
            "name": _config._name || _config.name,
            "value": _config.value,
            "filename": _config.filename
          });
        }
      }, this);
      return uiConfig;
    },


    checkServiceForConfigValue: function (value, serviceNames) {
      serviceNames.forEach(function (_serviceName) {
        if (!App.Service.find().mapProperty('serviceName').contains(_serviceName.name)) {
          value = value.replace(_serviceName.replace, '');
        }
      }, this);

      return value;
    },


    /**
     * Set all site property that are derived from other puppet-variable
     */

    getGlobConfigValue: function (templateName, expression, name) {
      var express = expression.match(/<(.*?)>/g);
      var value = expression;
      if (express == null) {
        return expression;
      }
      express.forEach(function (_express) {
        //console.log("The value of template is: " + _express);
        var index = parseInt(_express.match(/\[([\d]*)(?=\])/)[1]);
        var globValue = this.get('globalProperties').findProperty('name', templateName[index]);
        if (globValue) {
          console.log('The template value of templateName ' + '[' + index + ']' + ': ' + templateName[index] + ' is: ' + globValue);
          if (value !== null) {   // if the property depends on more than one template name like <templateName[0]>/<templateName[1]> then don't proceed to the next if the prior is null or not found in the global configs
            value = value.replace(_express, globValue.value);
          }
        } else {
          /*
           console.log("ERROR: The variable name is: " + templateName[index]);
           console.log("ERROR: mapped config from secureMapping file has no corresponding variable in " +
           "content.serviceConfigProperties. Two possible reasons for the error could be: 1) The service is not selected. " +
           "and/OR 2) The service_config metadata file has no corresponding global var for the site property variable");
           */
          value = null;
        }
      }, this);
      return value;
    },

    /**
     * Set all site property that are derived from other site-properties
     */
    setConfigValue: function (uiConfig, config) {
      if (config.value == null) {
        return;
      }
      var fkValue = config.name.match(/<(foreignKey.*?)>/g);
      if (fkValue) {
        fkValue.forEach(function (_fkValue) {
          var index = parseInt(_fkValue.match(/\[([\d]*)(?=\])/)[1]);
          var globalValue;
          if (uiConfig.someProperty('name', config.foreignKey[index])) {
            globalValue = uiConfig.findProperty('name', config.foreignKey[index]).value;
            config._name = config.name.replace(_fkValue, globalValue);
          } else if (this.get('globalProperties').someProperty('name', config.foreignKey[index])) {
            globalValue = this.get('globalProperties').findProperty('name', config.foreignKey[index]).value;
            config._name = config.name.replace(_fkValue, globalValue);
          }
        }, this);
      }
      //For properties in the configMapping file having foreignKey and templateName properties.

      var templateValue = config.value.match(/<(templateName.*?)>/g);
      if (templateValue) {
        templateValue.forEach(function (_value) {
          var index = parseInt(_value.match(/\[([\d]*)(?=\])/)[1]);
          var globValue = this.get('globalProperties').findProperty('name', config.templateName[index]);
          if (globValue) {
            config.value = config.value.replace(_value, globValue.value);
          } else {
            config.value = null;
          }
        }, this);
      }
    },

    prepareSecureConfigs: function () {
      this.loadGlobals();
      var storedConfigs = this.get('content.serviceConfigProperties').filterProperty('id', 'site property');
      var uiConfigs = this.loadUiSideConfigs();
      this.set('configs', storedConfigs.concat(uiConfigs));
    },

    loadGlobals: function () {
      var globals = this.get('content.serviceConfigProperties').filterProperty('id', 'puppet var');
      this.set('globalProperties', globals);
      this.loadStaticGlobal(); //Hack for properties which are declared in global_properties.js and not able to retrieve values declared in secure_properties.js
      this.loadUsersToGlobal();
      this.loadHostNamesToGlobal();
      this.loadPrimaryNamesToGlobals();
    },

    loadUsersToGlobal: function () {
      if (!this.get('serviceUsers').length) {
        this.loadUsersFromServer();
      }
      App.router.get('mainAdminSecurityController.serviceUsers').forEach(function (_user) {
        this.get('globalProperties').pushObject(_user);
      }, this);
    },

    loadHostNamesToGlobal: function () {
      var oozieHostComponent = App.Service.find('OOZIE').get('hostComponents').findProperty('componentName', 'OOZIE_SERVER');
      if (this.get('isOozieSelected') && oozieHostComponent) {
        var oozieHostName = oozieHostComponent.get('host.hostName');
        this.get('globalProperties').pushObject({
          id: 'puppet var',
          name: 'oozieserver_host',
          value: oozieHostName
        });
      }
      var hiveHostComponent = App.Service.find('HIVE').get('hostComponents').findProperty('componentName', 'HIVE_METASTORE');
      if (this.get('isHiveSelected') && hiveHostComponent) {
        var hiveHostName = hiveHostComponent.get('host.hostName');
        this.get('globalProperties').pushObject({
          id: 'puppet var',
          name: 'hivemetastore_host',
          value: hiveHostName
        });
      }
      var webHcatComponent = App.Service.find('WEBHCAT').get('hostComponents').findProperty('componentName', 'WEBHCAT_SERVER');
      if (this.isWebHcatSelected() && webHcatComponent) {
        var webHcatHostName = webHcatComponent.get('host.hostName');
        this.get('globalProperties').pushObject({
          id: 'puppet var',
          name: 'webhcat_server',
          value: webHcatHostName
        });
      }
    },

    loadStaticGlobal: function () {
      var globalProperties = this.get('globalProperties');
      this.get('globalProperties').forEach(function (_property) {
        switch (_property.name) {
          case 'security_enabled':
            _property.value = 'true';
            break;
        }
      }, this);
    },

    loadPrimaryNamesToGlobals: function () {
      var principalProperties = this.getPrincipalNames();
      principalProperties.forEach(function (_principalProperty) {
        var name = _principalProperty.name.replace('principal', 'primary');
        var value = _principalProperty.value.split('/')[0];
        this.get('globalProperties').pushObject({name: name, value: value});
      }, this);
    },

    getPrincipalNames: function () {
      var principalNames = [];
      var allPrincipalNames = [];
      this.get('globalProperties').forEach(function (_globalProperty) {
        if (/principal_name?$/.test(_globalProperty.name)) {
          principalNames.pushObject(_globalProperty);
        }
      }, this);
      this.get('secureProperties').forEach(function (_secureProperty) {
        if (/principal_name?$/.test(_secureProperty.name)) {
          var principalName = principalNames.findProperty('name', _secureProperty.name);
          if (!principalName) {
            _secureProperty.value = _secureProperty.defaultValue;
            principalNames.pushObject(_secureProperty);
          }
        }
      }, this);
      return principalNames;
    },

    loadUsersFromServer: function () {
      if (App.testMode) {
        var serviceUsers = this.get('serviceUsers');
        serviceUsers.pushObject({id: 'puppet var', name: 'hdfs_user', value: 'hdfs'});
        serviceUsers.pushObject({id: 'puppet var', name: 'mapred_user', value: 'mapred'});
        serviceUsers.pushObject({id: 'puppet var', name: 'hbase_user', value: 'hbase'});
        serviceUsers.pushObject({id: 'puppet var', name: 'hive_user', value: 'hive'});
      } else {
        App.router.set('mainAdminSecurityController.serviceUsers', App.db.getSecureUserInfo());
      }
    },


    manageSecureConfigs: function () {
      try {
        this.get('serviceConfigTags').forEach(function (_serviceConfigTags) {
          _serviceConfigTags.newTagName = 'version' + (new Date).getTime();
          if (_serviceConfigTags.siteName === 'global') {
            var realmName = this.get('globalProperties').findProperty('name', 'kerberos_domain');
            if (this.get('isNagiosSelected')) {
              var nagiosPrincipalName = this.get('globalProperties').findProperty('name', 'nagios_principal_name');
              nagiosPrincipalName.value = nagiosPrincipalName.value + '@' + realmName.value;
            }
            if (this.get('isZkSelected')) {
              var zkPrincipalName = this.get('globalProperties').findProperty('name', 'zookeeper_principal_name');
              zkPrincipalName.value = zkPrincipalName.value + '@' + realmName.value;
            }
            if (this.get('secureServices').someProperty('serviceName', 'STORM')) {
              var stormPrincipalName = this.get('globalProperties').findProperty('name', 'storm_principal_name');
              stormPrincipalName.value = stormPrincipalName.value + '@' + realmName.value;
            }
            this.get('globalProperties').forEach(function (_globalProperty) {
              if (!/_hosts?$/.test(_globalProperty.name)) {
                _serviceConfigTags.configs[_globalProperty.name] = _globalProperty.value;
              }
            }, this);
          }
          else {
            this.get('configs').filterProperty('id', 'site property').filterProperty('filename', _serviceConfigTags.siteName + '.xml').forEach(function (_config) {
              _serviceConfigTags.configs[_config.name] = _config.value;
            }, this);
          }
        }, this);
      } catch (err) {
        var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
        command.set('isSuccess', false);
        command.set('isError', true);
        if (err) {
          console.log("Error: Error occurred while applying secure configs to the server. Error message: " + err);
        }
        this.onJsError();
        return false;
      }
      return true;
    },

    deleteComponents: function(componentName, hostName) {
      App.ajax.send({
        name: 'admin.delete_component',
        sender: this,
        data: {
          componentName: componentName,
          hostName: hostName
        },
        success: 'onDeleteComplete',
        error: 'onDeleteError'
      });
    },

    onDeleteComplete: function () {
      var deleteAtsCommand = this.get('commands').findProperty('name', 'DELETE_ATS');
      console.warn('APP_TIMELINE_SERVER doesn\'t support security mode. It has been removed from YARN service ');
      deleteAtsCommand.set('isError', false);
      deleteAtsCommand.set('isSuccess', true);
    },

    onDeleteError: function () {
      console.warn('Error: Can\'t delete APP_TIMELINE_SERVER');
    },

    onJsError: function () {
      App.ModalPopup.show({
        header: Em.I18n.t('common.error'),
        secondary: false,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<p>{{t admin.security.apply.configuration.error}}</p>')
        })
      });
    }

  });
  
});
window.require.register("controllers/main/admin/security/disable", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.MainAdminSecurityDisableController = App.MainAdminSecurityProgressController.extend({

    name: 'mainAdminSecurityDisableController',
    secureServices: [],

    clearStep: function () {
      this.get('commands').clear();
      this.get('secureServices').clear();
      this.get('serviceConfigTags').clear();
    },

    loadStep: function () {
      this.clearStep();
      var commands = App.db.getSecurityDeployCommands();
      if (commands && commands.length > 0) {
        commands.forEach(function (_command, index) {
          commands[index] = App.Poll.create(_command);
        }, this);
        if (commands.someProperty('isError', true)) {
          this.get('commands').pushObjects(commands);
          this.loadSecureServices();
          this.addObserver('commands.@each.isSuccess', this, 'onCompleteCommand');
          return;
        } else if (commands.filterProperty('isStarted', true).someProperty('isCompleted', false)) {
          var runningCommand = commands.filterProperty('isStarted', true).findProperty('isCompleted', false);
          runningCommand.set('isStarted', false);
          this.get('commands').pushObjects(commands);
        } else {
          this.get('commands').pushObjects(commands);
        }
      } else {
        this.loadCommands();
        this.addInfoToCommands();
        var runningOperations = App.router.get('backgroundOperationsController.services').filterProperty('isRunning');
        var stopAllOperation = runningOperations.findProperty('name', 'Stop All Services');
        var stopCommand = this.get('commands').findProperty('name', 'STOP_SERVICES');
        if (stopCommand.get('name') === 'STOP_SERVICES' && stopAllOperation) {
          stopCommand.set('requestId', stopAllOperation.get('id'));
        }
      }
      this.loadSecureServices();
      this.addObserver('commands.@each.isSuccess', this, 'onCompleteCommand');
      this.moveToNextCommand();
    },


    enableSubmit: function () {
      if (this.get('commands').someProperty('isError', true) || this.get('commands').everyProperty('isSuccess', true)) {
        this.set('isSubmitDisabled', false);
      } else {
        this.set('isSubmitDisabled', true);
      }
    }.observes('commands.@each.isCompleted'),


    loadSecureServices: function () {
      var secureServices = App.get('isHadoop2Stack')?require('data/HDP2/secure_configs'):require('data/secure_configs');
      var installedServices = App.Service.find().mapProperty('serviceName');
      this.get('secureServices').pushObject(secureServices.findProperty('serviceName', 'GENERAL'));
      //General (only non service tab) tab is always displayed
      installedServices.forEach(function (_service) {
        var secureService = secureServices.findProperty('serviceName', _service);
        if (secureService) {
          this.get('secureServices').pushObject(secureService);
        }
      }, this);
    },


    manageSecureConfigs: function () {
      try {
        this.get('serviceConfigTags').forEach(function (_serviceConfigTags, index) {
          _serviceConfigTags.newTagName = 'version' + (new Date).getTime();
          if (_serviceConfigTags.siteName === 'global') {
            this.get('secureProperties').forEach(function (_config) {
              if (_config.name in _serviceConfigTags.configs) {
                delete _serviceConfigTags.configs[_config.name];
              }
            }, this);
            _serviceConfigTags.configs.security_enabled = 'false';
          } else {
            this.get('secureMapping').filterProperty('filename', _serviceConfigTags.siteName + '.xml').forEach(function (_config) {
              var configName = _config.name;
              if (configName in _serviceConfigTags.configs) {
                switch (configName) {
                  case 'dfs.datanode.address':
                    _serviceConfigTags.configs[configName] = '0.0.0.0:50010';
                    break;
                  case 'dfs.datanode.http.address':
                    _serviceConfigTags.configs[configName] = '0.0.0.0:50075';
                    break;
                  case 'mapred.task.tracker.task-controller':
                    _serviceConfigTags.configs[configName] = 'org.apache.hadoop.mapred.DefaultTaskController';
                    break;
                  case 'yarn.nodemanager.container-executor.class':
                    _serviceConfigTags.configs[configName] = 'org.apache.hadoop.yarn.server.nodemanager.DefaultContainerExecutor';
                    break;
                  case 'hbase.security.authentication':
                    _serviceConfigTags.configs[configName] = 'simple';
                    break;
                  case 'hbase.rpc.engine':
                    _serviceConfigTags.configs[configName] = 'org.apache.hadoop.hbase.ipc.WritableRpcEngine';
                    break;
                  case 'hbase.security.authorization':
                    _serviceConfigTags.configs[configName] = 'false';
                    break;
                  case 'zookeeper.znode.parent':
                    _serviceConfigTags.configs[configName] = '/hbase-unsecure';
                    break;
                  case 'hive.security.authorization.enabled':
                    _serviceConfigTags.configs[configName] = 'false';
                    break;
                  default:
                    delete _serviceConfigTags.configs[configName];
                }
              }
              console.log("Not Deleted" + _config.name);
            }, this);
          }
        }, this);
      } catch (err) {
        var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
        command.set('isSuccess', false);
        command.set('isError', true);
        if (err) {
          console.log("Error: Error occurred while applying secure configs to the server. Error message: " + err);
        }
        return false;
      }
      return true;
    }

  });
  
});
window.require.register("controllers/main/admin/security/security_progress_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.MainAdminSecurityProgressController = Em.Controller.extend({

    name: 'mainAdminSecurityProgressController',
    secureMapping: function () {
      if (App.get('isHadoop2Stack')) {
        return require('data/HDP2/secure_mapping');
      } else {
        return require('data/secure_mapping');
      }
    }.property(App.isHadoop2Stack),
    secureProperties: function () {
      if (App.get('isHadoop2Stack')) {
        return require('data/HDP2/secure_properties').configProperties;
      } else {
        return require('data/secure_properties').configProperties;
      }
    }.property(App.isHadoop2Stack),
    commands: [],
    configs: [],
    serviceConfigTags: [],
    globalProperties: [],
    totalSteps: 3,
    isSubmitDisabled: true,


    hasHostPopup: true,
    services: [],
    serviceTimestamp: null,


    retry: function () {
      var failedCommand = this.get('commands').findProperty('isError', true);
      if (failedCommand) {
        failedCommand.set('isStarted', false);
        failedCommand.set('isError', false);
        this.startCommand(failedCommand);
      }
    },

    updateServices: function () {
      this.services.clear();
      var services = this.get("services");
      this.get("commands").forEach(function (command) {
        var newService = Ember.Object.create({
          name: command.label,
          hosts: []
        });
        if (command && command.get("polledData")) {
          var hostNames = command.get("polledData").mapProperty('Tasks.host_name').uniq();
          hostNames.forEach(function (name) {
            newService.hosts.push({
              name: name,
              publicName: name,
              logTasks: command.polledData.filterProperty("Tasks.host_name", name)
            });
          });
          services.push(newService);
        }
      });
      this.set('serviceTimestamp', App.dateTime());
    }.observes('commands.@each.polledData'),

    loadCommands: function () {
      this.get('commands').pushObjects([
        App.Poll.create({name: 'STOP_SERVICES', label: Em.I18n.translations['admin.addSecurity.apply.stop.services'], isPolling: true }),
        App.Poll.create({name: 'APPLY_CONFIGURATIONS', label: Em.I18n.translations['admin.addSecurity.apply.save.config'], isPolling: false }),
        App.Poll.create({name: 'START_SERVICES', label: Em.I18n.translations['admin.addSecurity.apply.start.services'], isPolling: true })
      ]);
      this.setIndex(this.get('commands'));
    },

    setIndex: function(commandArray) {
      commandArray.forEach(function(command,index){
        command.set('index',index+1);
      },this);
    },

    startCommand: function (commnad) {
      if (this.get('commands').length === this.totalSteps) {
        if (!commnad) {
          var startedCommand = this.get('commands').filterProperty('isStarted', true);
          commnad = startedCommand.findProperty('isCompleted', false);
        }
        if (commnad && commnad.get('isPolling') === true) {
          commnad.set('isStarted', true);
          commnad.start();
        } else if (commnad && commnad.get('name') === 'APPLY_CONFIGURATIONS') {
          commnad.set('isStarted', true);
          if (App.testMode) {
            commnad.set('isError', false);
            commnad.set('isSuccess', true);
          } else {
            this.loadClusterConfigs();
          }
        } else if (commnad && commnad.get('name') === 'DELETE_ATS') {
          commnad.set('isStarted', true);
          if (App.testMode) {
            commnad.set('isError', false);
            commnad.set('isSuccess', true);
          } else {
            var timeLineServer = App.Service.find('YARN').get('hostComponents').findProperty('componentName', 'APP_TIMELINE_SERVER');
            this.deleteComponents('APP_TIMELINE_SERVER', timeLineServer.get('host.hostName'));
          }
        }
      }
    },


    onCompleteCommand: function () {
      if (this.get('commands').length === this.totalSteps) {
        var index = this.get('commands').filterProperty('isSuccess', true).length;
        if (index > 0) {
          var lastCompletedCommandResult = this.get('commands').objectAt(index - 1).get('isSuccess');
          if (lastCompletedCommandResult) {
            var nextCommand = this.get('commands').objectAt(index);
            this.moveToNextCommand(nextCommand);
          }
        }
      }
    },

    moveToNextCommand: function (nextCommand) {
      if (!nextCommand) {
        nextCommand = this.get('commands').findProperty('isStarted', false);
      }
      if (nextCommand) {
        this.startCommand(nextCommand);
      }
    },

    addInfoToCommands: function () {
      this.addInfoToStopService();
      this.addInfoToStartServices();
    },


    addInfoToStopService: function () {
      var command = this.get('commands').findProperty('name', 'STOP_SERVICES');
      var url = (App.testMode) ? '/data/wizard/deploy/2_hosts/poll_1.json' : App.apiPrefix + '/clusters/' + App.router.getClusterName() + '/services';
      var data = '{"RequestInfo": {"context" :"' + Em.I18n.t('requestInfo.stopAllServices') + '"}, "Body": {"ServiceInfo": {"state": "INSTALLED"}}}';
      command.set('url', url);
      command.set('data', data);
    },

    addInfoToStartServices: function () {
      var command = this.get('commands').findProperty('name', 'START_SERVICES');
      var url = (App.testMode) ? '/data/wizard/deploy/2_hosts/poll_1.json' : App.apiPrefix + '/clusters/' + App.router.getClusterName() + '/services?params/run_smoke_test=true';
      var data = '{"RequestInfo": {"context": "' + Em.I18n.t('requestInfo.startAllServices') + '"}, "Body": {"ServiceInfo": {"state": "STARTED"}}}';
      command.set('url', url);
      command.set('data', data);
    },

    loadClusterConfigs: function () {
      App.ajax.send({
        name: 'admin.security.add.cluster_configs',
        sender: this,
        success: 'loadClusterConfigsSuccessCallback',
        error: 'loadClusterConfigsErrorCallback'
      });
    },

    loadClusterConfigsSuccessCallback: function (data) {
      var self = this;
      //prepare tags to fetch all configuration for a service
      this.get('secureServices').forEach(function (_secureService) {
        self.setServiceTagNames(_secureService, data.Clusters.desired_configs);
      },this);
      this.getAllConfigurations();
    },

    loadClusterConfigsErrorCallback: function (request, ajaxOptions, error) {
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      command .set('isSuccess', false);
      command .set('isError', true);
      console.log("TRACE: error code status is: " + request.status);
    },

    /**
     * set tagnames for configuration of the *-site.xml
     */
    setServiceTagNames: function (secureService, configs) {
      //var serviceConfigTags = this.get('serviceConfigTags');
      for (var index in configs) {
        if (secureService.sites && secureService.sites.contains(index)) {
          var serviceConfigObj = {
            siteName: index,
            tagName: configs[index].tag,
            newTagName: null,
            configs: {}
          };
          this.get('serviceConfigTags').pushObject(serviceConfigObj);
        }
      }
      return serviceConfigObj;
    },

    applyConfigurationsToCluster: function () {
      var configData = [];
      this.get('serviceConfigTags').forEach(function (_serviceConfig) {
        var Clusters = {
          Clusters: {
            desired_config: {
              type: _serviceConfig.siteName,
              tag: _serviceConfig.newTagName,
              properties: _serviceConfig.configs
            }
          }
        };
        configData.pushObject(JSON.stringify(Clusters));
      }, this);

      var data = {
        configData: '[' + configData.toString() + ']'
      };

      App.ajax.send({
        name: 'admin.security.apply_configurations',
        sender: this,
        data: data,
        success: 'applyConfigurationToClusterSuccessCallback',
        error: 'applyConfigurationToClusterErrorCallback'
      });
    },

    applyConfigurationToClusterSuccessCallback: function (data) {
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      command.set('isSuccess', true);
      command.set('isError', false);
    },

    applyConfigurationToClusterErrorCallback: function (request, ajaxOptions, error) {
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      command.set('isSuccess', false);
      command.set('isError', true);
    },

    /**
     * gets site config properties from server and sets it for every configuration
     */
    getAllConfigurations: function () {
      var urlParams = [];
      this.get('serviceConfigTags').forEach(function (_tag) {
        urlParams.push('(type=' + _tag.siteName + '&tag=' + _tag.tagName + ')');
      }, this);
      if (urlParams.length > 0) {
        App.ajax.send({
          name: 'admin.get.all_configurations',
          sender: this,
          data: {
            urlParams: urlParams.join('|')
          },
          success: 'getAllConfigurationsSuccessCallback',
          error: 'getAllConfigurationsErrorCallback'
        });
      }
    },

    getAllConfigurationsSuccessCallback: function (data) {
      console.log("TRACE: In success function for the GET getServiceConfigsFromServer call");
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      this.get('serviceConfigTags').forEach(function (_tag) {
        if (!data.items.someProperty('type', _tag.siteName)) {
          console.log("Error: Metadata for secure services (secure_configs.js) is having config tags that are not being retrieved from server");
          command.set('isSuccess', false);
          command.set('isError', true);
        }
        _tag.configs = data.items.findProperty('type', _tag.siteName).properties;
      }, this);
      if (this.manageSecureConfigs()) {
        this.escapeXMLCharacters(this.get('serviceConfigTags'));
        this.applyConfigurationsToCluster();
      }
    },

    getAllConfigurationsErrorCallback: function (request, ajaxOptions, error) {
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      command.set('isSuccess', false);
      command.set('isError', true);
      console.log("TRACE: In error function for the getServiceConfigsFromServer call");
      console.log("TRACE: error code status is: " + request.status);
    },

    /*
     Iterate over keys of all configurations and escape xml characters in their values
     */
    escapeXMLCharacters: function (serviceConfigTags) {
      serviceConfigTags.forEach(function (_serviceConfigTags) {
        var configs = _serviceConfigTags.configs;
        for (var key in configs) {
          configs[key] = this.setServerConfigValue(key, configs[key]);
        }
      }, this);
    },

    setServerConfigValue: function(configName, value) {
      switch (configName) {
        case 'storm.zookeeper.servers':
          return value;
        default:
          return App.config.escapeXMLCharacters(value);
      }
    },

    saveCommandsOnRequestId: function () {
      this.saveCommands();
    }.observes('commands.@each.requestId'),

    saveCommandsOnCompleted: function () {
      this.saveCommands();
    }.observes('commands.@each.isCompleted'),

    saveCommands: function () {
      var commands = [];
      if (this.get('commands').length === this.totalSteps) {
        this.get('commands').forEach(function (_command) {
          var command = {
            name: _command.get('name'),
            label: _command.get('label'),
            isPolling: _command.get('isPolling'),
            isVisible:  _command.get('isVisible'),
            isStarted: _command.get('isStarted'),
            requestId: _command.get('requestId'),
            isSuccess: _command.get('isSuccess'),
            isError: _command.get('isError'),
            url: _command.get('url'),
            polledData: _command.get('polledData'),
            data: _command.get('data')
          };
          commands.pushObject(command);
        }, this);
        App.db.setSecurityDeployCommands(commands);
        if (!App.testMode) {
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'ADD_SECURITY_STEP_4',
            wizardControllerName: App.router.get('addSecurityController.name'),
            localdb: App.db.data
          });
        }
      }
    }
  });
  
});
window.require.register("controllers/main/admin/stack_upgrade_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.StackUpgradeController = App.WizardController.extend({
    name: 'stackUpgradeController',

    totalSteps: 3,

    /**
     * All wizards data will be stored in this variable
     *
     * cluster - cluster name
     * upgradeOptions - upgrade options
     */
    content: Em.Object.create({
      cluster: null,
      upgradeOptions: null,
      servicesInfo: function(){
        return App.router.get('mainAdminClusterController.services');
      }.property('App.router.mainAdminClusterController.services'),
      upgradeVersion: function(){
        return App.router.get('mainAdminClusterController.upgradeVersion');
      }.property('App.router.mainAdminClusterController.upgradeVersion'),
      controllerName: 'stackUpgradeController'
    }),

    /**
     * return new object extended from clusterStatusTemplate
     * @return Object
     */
    getCluster: function(){
      return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
    },
    /**
     * return new object extended from upgradeOptionsTemplate
     * @return Object
     */
    getUpgradeOptions: function(){
      return jQuery.extend({}, this.get('upgradeOptionsTemplate'));
    },
    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      this.load('cluster');
      this.load('upgradeOptions');
    },
    upgradeOptionsTemplate:{
      localRepo: false
    },
    clear: function () {
      this.set('content', Ember.Object.create({
        servicesInfo: function(){
          return App.router.get('mainAdminClusterController.services');
        }.property('App.router.mainAdminClusterController.services'),
        upgradeVersion: function(){
          return App.router.get('mainAdminClusterController.upgradeVersion');
        }.property('App.router.mainAdminClusterController.upgradeVersion'),
        'controllerName': 'stackUpgradeController'
      }));
      this.set('currentStep', 0);
      this.clearStorageData();
    },
    clearStorageData: function(){
      App.db.setCluster(undefined);
      App.db.setUpgradeOptions(undefined);
    },
    /**
     * Finish upgrade
     */
    finish: function () {
      this.clear();
      this.setCurrentStep('1');
    }
  });
  
});
window.require.register("controllers/main/admin/user", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserController = Em.Controller.extend({
    name:'mainAdminUserController',

    /**
     * send request to the server to delete user if selected user is not current user
     * @param event
     */
    deleteRecord:function (event) {
      var self = this;
      if (event.context.get('userName') == App.get('router').getLoginName()) {
        App.ModalPopup.show({
          header:Em.I18n.t('admin.users.delete.yourself.header'),
          body:Em.I18n.t('admin.users.delete.yourself.message'),
          onPrimary:function (event) {
            this.hide();
          },
          secondary:false
        });

        return;
      }

      App.ModalPopup.show({
        header:Em.I18n.t('admin.users.delete.header').format(event.context.get('userName')),
        body:Em.I18n.t('question.sure').format(''),
        primary:Em.I18n.t('yes'),
        secondary:Em.I18n.t('no'),

        onPrimary:function () {
          self.sendCommandToServer('/users/' +  event.context.get("userName"), "DELETE" ,{},
            function (success) {

              if (!success) {
                return;
              }

              event.context.deleteRecord();

              try {
                App.store.commit();
              } catch (err) {

              }
            });
          this.hide();
        },
        onSecondary:function () {
          this.hide();
        }
      });
    },

    /**
     * send request to the server and call callback function with true if request was success or false if request was failed
     * @param url
     * @param method
     * @param postData
     * @param callback
     */
    sendCommandToServer : function(url, method, postData, callback){
      if (App.testMode) {
        url = '/data/users/users.json';
        method = 'GET';
        postData = undefined;
      } else {
        url = App.apiPrefix + url;
      }

      $.ajax({
        type: method,
        url: url,
        data: JSON.stringify(postData),
        dataType: 'json',
        timeout: App.timeout,
        success: function(data){
            callback(true, '');
        },

        error: function (request, ajaxOptions, error) {
          var message = $.parseJSON(request.responseText).message;
          message = message.substr(message.lastIndexOf(':') + 1);
          callback(false, message);
          console.log(message);
        },

        statusCode: require('data/statusCodes')
      });
    }
  });
});
window.require.register("controllers/main/admin/user/create", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserCreateController = Em.Controller.extend({
    name:'mainAdminUserCreateController',
    content:false
  });
  
});
window.require.register("controllers/main/admin/user/edit", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserEditController = App.MainAdminUserController.extend({
    name:'mainAdminUserEditController',
    content:false
  });
});
window.require.register("controllers/main/alerts_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAlertsController = Em.Controller.extend({
    name: 'mainAlertsController',

    alerts: [],
    isLoaded: false,
    isUpdating: false,
    // name of resource(hostName or ServiceName)
    resourceName: null,
    //"HOST" or "SERVICE"
    resourceType: null,
    updateTimer: null,

    /**
     * load alerts for service or host
     * @param name
     * @param type
     */
    loadAlerts: function (name, type) {
      this.set('isLoaded', false);
      this.set('resourceName', name);
      this.set('resourceType', type);
      this.getFromServer();
    },
    /**
     * update alerts
     */
    update: function () {
      var self = this;
      if (this.get('isUpdating')) {
        this.set('updateTimer', setTimeout(function () {
          self.getFromServer();
          self.update();
        }, App.componentsUpdateInterval));
      } else {
        clearTimeout(this.get('updateTimer'));
      }
    }.observes('isUpdating'),

    /**
     * ask alerts from server by type
     */
    getFromServer: function () {
      if (App.router.get('clusterController.isNagiosInstalled')) {
        if (this.get('resourceType') === "SERVICE") {
          this.getAlertsByService();
        } else if (this.get('resourceType') === "HOST") {
          this.getAlertsByHost();
        } else {
          console.warn("GET Alerts error: unknown resourceType");
        }
      } else {
        this.set('isLoaded', true);
      }
    },

    getAlertsByHost: function () {
      if (this.get('resourceName')) {
        App.ajax.send({
          name: 'alerts.get_by_host',
          sender: this,
          data: {
            hostName: this.get('resourceName')
          },
          success: 'getAlertsSuccessCallback',
          error: 'getAlertsErrorCallback'
        })
      } else {
        console.warn('GET Alerts error: hostName parameter is missing');
      }
    },

    getAlertsByService: function () {
      if (this.get('resourceName')) {
        App.ajax.send({
          name: 'alerts.get_by_service',
          sender: this,
          data: {
            serviceName: this.get('resourceName')
          },
          success: 'getAlertsSuccessCallback',
          error: 'getAlertsErrorCallback'
        })
      } else {
        console.warn('GET Alerts error: serviceName parameter is missing');
      }
    },

    /**
     * map to associate old status format with and maintain sorting
     */
    statusNumberMap: {
      "OK" : "0",
      "WARNING": "1",
      "CRITICAL": "2",
      "PASSIVE": "3"
    },

    getAlertsSuccessCallback: function (json) {
      var alerts = [];
      if (json && json.alerts && json.alerts.detail) {
        json.alerts.detail.forEach(function (_alert) {
          alerts.pushObject(App.Alert.create({
            title: _alert.description,
            serviceType: _alert.service_name,
            lastTime: _alert.status_time,
            status: this.get('statusNumberMap')[_alert.status] || "4",
            message: _alert.output,
            hostName: _alert.host_name,
            lastCheck: _alert.last_status_time
          }));
        }, this);
      }
      this.set('alerts', alerts.sortProperty('status','date').reverse());
      this.set('isLoaded', true);
    },

    getAlertsErrorCallback: function(){
      this.set('isLoaded', true);
    }
  });
  
});
window.require.register("controllers/main/apps/item_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAppsItemController = Em.Controller.extend({
    name:'mainAppsItemController',
    /**
     * Was set outside in App.MainAppsView.
     * It's instance of App.Run model
     */
    content: [],
    jobsLoaded:false,

    gettingJobs:function(){
      var currentId = this.get('content.id');
      if(this.get('content.loadAllJobs')){
        return;
      }
      var self = this;

      var url = App.testMode ? '/data/apps/jobs/'+ currentId +'.json' :
        App.apiPrefix + "/jobhistory/job?workflowId=" + currentId;

      var mapper = App.jobsMapper;
      mapper.set('controller', this);
      App.HttpClient.get(url, mapper,{
        complete:function(jqXHR, textStatus) {
          self.set('content.loadAllJobs', true);
        }
      });
    }.observes('content')

  });
  
});
window.require.register("controllers/main/apps_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var misc = require('utils/misc');
  var date = require('utils/date');

  App.MainAppsController = Em.ArrayController.extend({

    name:'mainAppsController',
    content: [],

    loaded : false,
    loading : false,

    /**
     * List of users.
     * Will be used for filtering in user column.
     * Go to App.MainAppsView.userFilterView for more information
     */
    users: function () {
      return this.get('content').mapProperty("userName").uniq().map(function(userName){
        return {
          name: userName,
          checked: false
        };
      });
    }.property('content.length'),

    loadRuns:function () {

      this.set('loading', true);
      var self = this;

      //var runsUrl = App.testMode ? "/data/apps/runs.json" : App.apiPrefix + "/jobhistory/workflow?orderBy=startTime&sortDir=DESC&limit=" + App.maxRunsForAppBrowser;
      var runsUrl = App.testMode ? "/data/apps/runs.json" : App.apiPrefix + this.get("runUrl");

      App.HttpClient.get(runsUrl, App.runsMapper, {
        complete:function (jqXHR, textStatus) {
          self.set('loading', false);
          self.set('loaded', true);
        }
      });
    },

    //Pagination Object

    paginationObject:{
      iTotalDisplayRecords :0,
      iTotalRecords:0,
      startIndex:0,
      endIndex:0
    },

    /*
     Set number of filtered jobs when switching to all jobs
     */
    iTotalDisplayRecordsObserver:function(){
      if(this.get("filterObject.allFilterActivated")){
        this.set("filterObject.allFilterActivated", false);
      }else{
        this.set("filterObject.filteredDisplayRecords",this.get("paginationObject.iTotalDisplayRecords"));
      }
    }.observes("paginationObject.iTotalDisplayRecords"),


    //Filter object

    filterObject : Ember.Object.create({
      sSearch_0:"",
      sSearch_1:"",
      sSearch_2:"",
      sSearch_3:"",
      minJobs:"",
      maxJobs:"",
      minInputBytes:"",
      maxInputBytes:"",
      minOutputBytes:"",
      maxOutputBytes:"",
      minDuration:"",
      maxDuration:"",
      minStartTime:"",
      maxStartTime:"",
      sSearch:"",
      iDisplayLength:"",
      iDisplayStart:"",
      iSortCol_0:"",
      sSortDir_0:"",

      allFilterActivated:false,
      filteredDisplayRecords:null,

      viewType:"all",
      viewTypeClickEvent:false,

      /**
       * Direct binding to job filter field
       */
      runType:"",
      onRunTypeChange:function(){
        if(this.runType == "MapReduce"){
          this.set("sSearch_2","mr");
        }else if(this.runType == "Hive"){
          this.set("sSearch_2","hive");
        }else if(this.runType == "Pig"){
          this.set("sSearch_2","pig");
        }else{
          this.set("sSearch_2","");
        }
      }.observes("runType"),

      /**
       * Direct binding to job filter field
       */
      jobs:"",
      onJobsChange:function(){
        var minMaxTmp = this.parseNumber(this.jobs);
        this.set("minJobs", minMaxTmp.min);
        this.set("maxJobs", minMaxTmp.max);
      }.observes("jobs"),

      /**
       * Direct binding to Input filter field
       */
      input:"",
      onInputChange:function(){
        var minMaxTmp = this.parseBandWidth(this.input);
        this.set("minInputBytes", minMaxTmp.min);
        this.set("maxInputBytes", minMaxTmp.max);
      }.observes("input"),

      /**
       * Direct binding to Output filter field
       */
      output:"",
      onOutputChange:function(){
        var minMaxTmp = this.parseBandWidth(this.output);
        this.set("minOutputBytes", minMaxTmp.min);
        this.set("maxOutputBytes", minMaxTmp.max);
      }.observes("output"),

      /**
       * Direct binding to Duration filter field
       */
      duration:"",
      onDurationChange:function(){
        var minMaxTmp = this.parseDuration(this.duration);
        this.set("minDuration", minMaxTmp.min);
        this.set("maxDuration", minMaxTmp.max);
      }.observes("duration"),

      /**
       * Direct binding to Run Date filter field
       */
      runDate:"",
      onRunDateChange:function(){
        var minMaxTmp = this.parseDate(this.runDate);
        this.set("minStartTime", minMaxTmp.min);
        this.set("maxStartTime", minMaxTmp.max);
      }.observes("runDate"),

      parseDuration:function(value){
        var tmp={
          min:"",
          max:""
        };

        var compareChar = isNaN(value.charAt(0)) ? value.charAt(0) : false;
        var compareScale = value.match(/s|m|h/);
        compareScale = compareScale ? compareScale[0] : "";
        var compareValue = compareChar ? parseFloat(value.substr(1, value.length)) : parseFloat(value.substr(0, value.length));
        if(isNaN(compareValue)){
          return tmp;
        }
        switch (compareScale) {
          case 'h':
          tmp.min = Math.ceil((parseFloat(compareValue)-0.0001)*1000*60*60);
          tmp.max = Math.floor((parseFloat(compareValue)+0.0001)*1000*60*60);
          break;
          case 'm':
          tmp.min = Math.ceil((parseFloat(compareValue)-0.001)*1000*60);
          tmp.max = Math.floor((parseFloat(compareValue)+0.001)*1000*60);
          break;
          case 's':
          tmp.min = Math.ceil((parseFloat(compareValue)-0.01)*1000);
          tmp.max = Math.floor((parseFloat(compareValue)+0.01)*1000);
          break;
          default:
            tmp.min = Math.ceil((parseFloat(compareValue)-0.01)*1000);
            tmp.max = Math.floor((parseFloat(compareValue)+0.01)*1000);
        }
        switch (compareChar) {
          case '<':
            tmp.min="";
            break;
          case '>':
            tmp.max="";
            break;
        }
        return tmp;
      },

      parseDate:function(value){
        var tmp={
          min:"",
          max:""
        };
        var nowTime = App.dateTime();

        switch (value){
          case 'Any':
            break;
          case 'Past 1 Day':
            tmp.min= nowTime - 86400000;
            break;
          case 'Past 2 Days':
            tmp.min= nowTime - 172800000;
            break;
          case 'Past 7 Days':
            tmp.min= nowTime - 604800000;
            break;
          case 'Past 14 Days':
            tmp.min= nowTime - 1209600000;
            break;
          case 'Past 30 Days':
            tmp.min= nowTime - 2592000000;
            break;
          case 'Running Now':
            tmp.min= nowTime;
            break;
        }
        return tmp;
      },

      parseBandWidth:function(value){
        var tmp={
          min:"",
          max:""
        };

        var compareChar = isNaN(value.charAt(0)) ? value.charAt(0) : false;
        var compareScale = value.match(/kb|k|mb|m|gb|g/);
        compareScale = compareScale ? compareScale[0] : "";
        var compareValue = compareChar ? parseFloat(value.substr(1, value.length)) : parseFloat(value.substr(0, value.length));
        if(isNaN(compareValue)){
          return tmp;
        }
        switch (compareScale) {
          case 'g': case 'gb':
            tmp.min = Math.max(1073741824,Math.ceil((compareValue-0.005)*1073741824));
            tmp.max = Math.floor((compareValue+0.005)*1073741824);
            break;
          case 'm': case 'mb':
            tmp.min = Math.max(1048576,Math.ceil((compareValue-0.05)*1048576));
            tmp.max = Math.min(1073741823,Math.floor((compareValue+0.05)*1048576));
            break;
          case 'k': case 'kb':
            tmp.min = Math.max(1024,Math.ceil((compareValue-0.05)*1024));
            tmp.max = Math.min(1048575,Math.floor((compareValue+0.05)*1024));
            break;
          default:
            tmp.min = Math.max(1024,Math.ceil((compareValue-0.05)*1024));
            tmp.max = Math.min(1048575,Math.floor((compareValue+0.05)*1024));
        }
        switch (compareChar) {
          case '<':
            tmp.min="";
            break;
          case '>':
            tmp.max="";
            break;
        }
        return tmp;
      },
      parseNumber:function(value){
        var tmp={
          min:"",
          max:""
        };
        switch (value.charAt(0)) {
          case '<':
            tmp.max=value.substr(1);
            break;
          case '>':
            tmp.min=value.substr(1);
            break;
          case '=':
            tmp.min=value.substr(1);
            tmp.max=value.substr(1);
            break;
          default:
            tmp.min=value;
            tmp.max=value;
        }
        return tmp;
      },

      /**
       * Create link for server request
       * @return {String}
       */
      createAppLink:function(){
        var link = "/jobhistory/datatable?";


        var arr = [
          "sSearch_0", "sSearch_1", "sSearch_2", "sSearch_3", "minJobs",
          "maxJobs", "minInputBytes", "maxInputBytes", "minOutputBytes",
          "maxOutputBytes", "minDuration", "maxDuration", "minStartTime",
          "maxStartTime", "sSearch", "iDisplayLength", "iDisplayStart",
          "iSortCol_0", "sSortDir_0"
        ];

        var notFilterFields = ["iDisplayLength", "iDisplayStart", "iSortCol_0", "sSortDir_0"];

        var filtersUsed = false;

        for (var n=0; n<arr.length;n++) {
          if(this.get(arr[n])){
            link += arr[n] + "=" + this.get(arr[n]) + "&";
            if (!notFilterFields.contains(arr[n])) {
              filtersUsed = true;
            }
          }
        }

        link = link.slice(0,link.length-1);

        if(!this.get("viewTypeClickEvent")) {
          this.set('viewType', filtersUsed?'filtered':'all');
        }

        return link;
      }
    }),

    /**
     * reset all filters in table
     *
     */
    clearFilters: function () {
      var obj=this.get("filterObject");
      obj.set("sSearch","");
      obj.set("sSearch_0","");
      obj.set("sSearch_1","");
      obj.set("sSearch_2","");
      obj.set("sSearch_3","");
      obj.set("runType","Any");
      obj.set("jobs","");
      obj.set("input","");
      obj.set("output","");
      obj.set("duration","");
      obj.set("runDate","Any");
    },


    runUrl : "/jobhistory/datatable",
    runTimeout : null,

    valueObserver: function(){
      if(this.get('filterObject.iDisplayLength') > this.get('content.length')) {
        this.set('filterObject.iDisplayStart', 0);
      }
      var link = this.get('filterObject').createAppLink();

      if(this.get("filterObject.viewType") == "filtered"){
        this.set("runUrl", link);
      }else{
        this.set("runUrl", "/jobhistory/datatable?iDisplayStart=" + this.get('filterObject.iDisplayStart') + "&iDisplayLength=" + this.get('filterObject.iDisplayLength') +
           '&iSortCol_0=' + this.get('filterObject.iSortCol_0') + '&sSortDir_0=' + this.get('filterObject.sSortDir_0'));
      }

      var timeout = this.get('runTimeout');
      var self = this;

      clearTimeout(timeout);
      timeout = setTimeout(function(){
        console.log(self.get("runUrl"));
        self.loadRuns();
      }, 300);

      this.set('runTimeout', timeout);

    }.observes(
        'filterObject.sSearch_0',
        'filterObject.sSearch_1',
        'filterObject.sSearch_2',
        'filterObject.sSearch_3',
        'filterObject.minJobs',
        'filterObject.maxJobs',
        'filterObject.minInputBytes',
        'filterObject.maxInputBytes',
        'filterObject.minOutputBytes',
        'filterObject.maxOutputBytes',
        'filterObject.minDuration',
        'filterObject.maxDuration',
        'filterObject.minStartTime',
        'filterObject.maxStartTime',
        'filterObject.sSearch',
        'filterObject.iDisplayLength',
        'filterObject.iDisplayStart',
        'filterObject.iSortCol_0',
        'filterObject.sSortDir_0',
        'filterObject.viewType'
    ),

    serverData: "",
    summary: null,

    /**
     * Observer for summary data from server
     */
    summaryInfo: function(){
      var tmp;
      var summary = this.get('serverData');
      if(!summary){
        tmp = {
          'jobs': {
            'avg': '-',
            'min': '-',
            'max': '-'
          },
          'input': {
            'avg': '-',
            'min': '-',
            'max': '-'
          },
          'output': {
            'avg': '-',
            'min': '-',
            'max': '-'
          },
          'duration': {
            'avg': '-',
            'min': '-',
            'max': '-'
          },
          'times': {
            'oldest': '-',
            'youngest': '-'
          }
        };
      }else{
        tmp = {
          'jobs': {
            'avg': summary.jobs.avg.toFixed(2),
            'min': summary.jobs.min,
            'max': summary.jobs.max
          },
          'input': {
            'avg': misc.formatBandwidth(summary.input.avg),
            'min': misc.formatBandwidth(summary.input.min),
            'max': misc.formatBandwidth(summary.input.max)
          },
          'output': {
            'avg': misc.formatBandwidth(summary.output.avg),
            'min': misc.formatBandwidth(summary.output.min),
            'max': misc.formatBandwidth(summary.output.max)
          },
          'duration': {
            'avg': date.timingFormat(Math.round(summary.duration.avg)),
            'min': date.timingFormat(summary.duration.min),
            'max': date.timingFormat(summary.duration.max)
          },
          'times': {
            'oldest': new Date(summary.times.oldest).toDateString(),
            'youngest': new Date(summary.times.youngest).toDateString()
          }
        };
      }
      this.set("summary",tmp);
    }.observes('serverData'),


    columnsName: Ember.ArrayController.create({
      content: [
        { name: Em.I18n.t('apps.table.column.appId'), index: 0 },
        { name: Em.I18n.t('common.name'), index: 1 },
        { name: Em.I18n.t('common.type'), index: 2 },
        { name: Em.I18n.t('common.user'), index: 3 },
        { name: Em.I18n.t('apps.avgTable.jobs'), index: 4 },
        { name: Em.I18n.t('apps.avgTable.input'), index: 5 },
        { name: Em.I18n.t('apps.avgTable.output'), index: 6 },
        { name: Em.I18n.t('apps.avgTable.duration'), index: 7 },
        { name: Em.I18n.t('apps.table.column.runDate'), index: 8 }
      ]
    })

  });
  
});
window.require.register("controllers/main/charts", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainChartsController = Em.ArrayController.extend({
    name:'mainChartsController'
  });
});
window.require.register("controllers/main/charts/heatmap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MainChartsHeatmapController = Em.Controller.extend({
    name: 'mainChartsHeatmapController',
    modelRacks: App.Rack.find(),
    racks: function () {
      var racks = [];
      this.get('modelRacks').forEach(function (rack) {
        racks.push(Em.Object.create({
          name: rack.get('name'),
          hosts: rack.get('hosts'),
          isLoaded: false
        }));
      });
      return racks;
    }.property('modelRacks.@each.isLoaded'),

    allMetrics: function () {
      var metrics = [
        Em.Object.create({
          label: Em.I18n.t('charts.heatmap.category.host'),
          category: 'host',
          items: [
            App.MainChartHeatmapDiskSpaceUsedMetric.create(),
            App.MainChartHeatmapMemoryUsedMetric.create(),
            App.MainChartHeatmapCpuWaitIOMetric.create()
            /*, App.MainChartHeatmapProcessRunMetric.create()*/
          ]
        }),
        Em.Object.create({
          label: Em.I18n.t('charts.heatmap.category.hdfs'),
          category: 'hdfs',
          items: [
            App.MainChartHeatmapDFSBytesReadMetric.create(),
            App.MainChartHeatmapDFSBytesWrittenMetric.create(),
            App.MainChartHeatmapDFSGCTimeMillisMetric.create(),
            App.MainChartHeatmapDFSMemHeapUsedMetric.create()
          ]
        })
      ];

      if (App.MapReduceService.find().get('length')) {
        metrics.push(
          Em.Object.create({
            label: Em.I18n.t('charts.heatmap.category.mapreduce'),
            category: 'mapreduce',
            items: [
              App.MainChartHeatmapMapreduceMapsRunningMetric.create(),
              App.MainChartHeatmapMapreduceReducesRunningMetric.create(),
              App.MainChartHeatmapMapreduceGCTimeMillisMetric.create(),
              App.MainChartHeatmapMapreduceMemHeapUsedMetric.create()
            ]
          })
        );
      }

      if (App.YARNService.find().get('length')) {
        metrics.push(
          Em.Object.create({
            label: Em.I18n.t('charts.heatmap.category.yarn'),
            category: 'yarn',
            items: [
              App.MainChartHeatmapYarnGCTimeMillisMetric.create(),
              App.MainChartHeatmapYarnMemHeapUsedMetric.create(),
              App.MainChartHeatmapYarnResourceUsedMetric.create()
            ]
          })
        );
      }

      if (App.HBaseService.find().get('length')) {
        metrics.push(
          Em.Object.create({
            label: Em.I18n.t('charts.heatmap.category.hbase'),
            category: 'hbase',
            items: [
              App.MainChartHeatmapHbaseReadReqCount.create(),
              App.MainChartHeatmapHbaseWriteReqCount.create(),
              App.MainChartHeatmapHbaseCompactionQueueSize.create(),
              App.MainChartHeatmapHbaseRegions.create(),
              App.MainChartHeatmapHbaseMemStoreSize.create()
            ]
          })
        );
      }
      return metrics;
    }.property(),

    selectedMetric: null,

    inputMaximum: '',

    validation: function () {
      if (this.get('selectedMetric')) {
        if (/^\d+$/.test(this.get('inputMaximum'))) {
          $('#inputMaximum').removeClass('error');
          this.set('selectedMetric.maximumValue', this.get('inputMaximum'));
        } else {
          $('#inputMaximum').addClass('error');
        }
      }
    }.observes('inputMaximum'),

    showHeatMapMetric: function (event) {
      var metricItem = event.context;
      if (metricItem) {
        this.set('selectedMetric', metricItem);
      }
    },

    hostToSlotMap: function () {
      return this.get('selectedMetric.hostToSlotMap');
    }.property('selectedMetric.hostToSlotMap'),

    loadMetrics: function () {
      var selectedMetric = this.get('selectedMetric');
      if (selectedMetric) {
        selectedMetric.refreshHostSlots();
      }
      this.set('inputMaximum', this.get('selectedMetric.maximumValue'));
    }.observes('selectedMetric'),

    /**
     * return class name for to be used for containing each rack.
     *
     * @this App.MainChartsHeatmapController
     */
    rackClass: function () {
      var rackCount = this.get('racks.length');
      if (rackCount < 2) {
        return "span12";
      } else if (rackCount == 2) {
        return "span6";
      } else {
        return "span4";
      }
    }.property('racks.length')
  });
  
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');

  var heatmap = require('utils/heatmap');

  /**
   * Base class for any heatmap metric.
   * 
   * This class basically provides the following for each heatmap metric.
   * <ul>
   * <li> Provides number of slots in which temperature can fall.
   * <li> Maintains the maximum value so as to scale slot ranges.
   * <li> Gets JSON data from server and maps response for all hosts into above
   * slots.
   * </ul>
   * 
   */
  App.MainChartHeatmapMetric = Em.Object.extend(heatmap.mappers, {
    /**
     * Name of this metric
     */
    name: null,

    /**
     * Number of slots this metric will be mapped into. When changing this value,
     * the color count in 'slotColors' should also be changed.
     */
    numberOfSlots: 5,

    /**
     * Colors for the each of the number of slots defined above. When changing the
     * number of slots, the number of colors also should be updated.
     * 
     * @type {Array}
     */
    slotColors: [ {
      r: 0x00,
      g: 0xcc,
      b: 0x00
    }, // Green
    {
      r: 0x9f,
      g: 0xee,
      b: 0x00
    }, {
      r: 0xff,
      g: 0xff,
      b: 0x00
    }, // Yellow
    {
      r: 0xff,
      g: 0xc0,
      b: 0x00
    }, // Orange
    {
      r: 0xff,
      g: 0x00,
      b: 0x00
    } ],// Red

    /**
     * Minimum value of this metric. Default is 0.
     */
    minimumValue: 0,

    /**
     * Maximum value of this metric. This has to be specified by extending classes
     * so that the range from 'minimumValue' to 'maximumValue' can be split among
     * 'numberOfSlots'. It is recommended that this value be a multiple of
     * 'numberOfSlots'.
     */
    maximumValue: 100,

    /**
     * Units of the maximum value which is shown in UI {String}
     */
    units: '',

    /**
     * Indicates whether this metric is currently loading data from the server.
     * {Boolean}
     */
    loading: false,

    /**
     * Provides following information about slots in an array of objects.
     * <ul>
     * <li> from: {number} Slot starts from this value
     * <li> to: {number} Slot ends at this value (inclusive)
     * <li> label: {String} Slot name to be shown
     * <li> cssStyle: {String} style to be embedded on hosts which fall into this
     * slot.
     * </ul>
     * 
     * Slot count will be the same as specified in 'numberOfSlots'. Slot
     * definitions will be given in increasing temperature from 'minimumValue' to
     * 'maximumValue'.
     * 
     */
    slotDefinitions: function () {
      var min = this.get('minimumValue');
      var max = parseFloat(this.get('maximumValue'));
      var slotCount = this.get('numberOfSlots');
      var labelSuffix = this.get('slotDefinitionLabelSuffix');
      var delta = (max - min) / slotCount;
      var defs = [];
      var slotColors = this.get('slotColors');
      var slotColorIndex = 0;
      for ( var c = 0; c < slotCount - 1; c++) {
        var from = this.formatLegendNumber(c * delta);
        var to = this.formatLegendNumber((c + 1) * delta);
        var label;
        if ($.trim(labelSuffix) == 'ms') {
        	label = date.timingFormat(from, 'zeroValid') + " - " + date.timingFormat(to, 'zeroValid');
        } else {
  	      label = from + labelSuffix + " - " + to + labelSuffix;
        }
        var slotColor = slotColors[slotColorIndex++];
        defs.push(Em.Object.create({
          from: from,
          to: to,
          label: label,
          cssStyle: "background-color:rgb(" + slotColor.r + "," + slotColor.g + "," + slotColor.b + ")"
        }));
      }
      from = this.formatLegendNumber((slotCount - 1) * delta);
      to = this.formatLegendNumber(max);

      if ($.trim(labelSuffix) == 'ms') {
        label = date.timingFormat(from, 'zeroValid') + " - " + date.timingFormat(to, 'zeroValid');
      } else {
        label = from + labelSuffix + " - " + to + labelSuffix;
      }

      slotColor = slotColors[slotColorIndex++];
      defs.push(Em.Object.create({
        from: from,
        to: to,
        label: label,
        cssStyle: "background-color:rgb(" + slotColor.r + "," + slotColor.g + "," + slotColor.b + ")"
      }));
      var hatchStyle = "background-color:rgb(135, 206, 250)";
      if(jQuery.browser.webkit){
        hatchStyle = "background-image:-webkit-repeating-linear-gradient(-45deg, #FF1E10, #FF1E10 3px, #ff6c00 3px, #ff6c00 6px)";
      }else if(jQuery.browser.mozilla){
        hatchStyle = "background-image:repeating-linear-gradient(-45deg, #FF1E10, #FF1E10 3px, #ff6c00 3px, #ff6c00 6px)";
      }else if(jQuery.browser.msie && jQuery.browser.version){
        var majorVersion =  parseInt(jQuery.browser.version.split('.')[0]);
        if(majorVersion>9){
          hatchStyle = "background-image:repeating-linear-gradient(-45deg, #FF1E10, #FF1E10 3px, #ff6c00 3px, #ff6c00 6px)";
        }
      }
      defs.push(Em.Object.create({
        from: NaN,
        to: NaN,
        label: Em.I18n.t('charts.heatmap.label.invalidData'),
        cssStyle: hatchStyle
      }));
      defs.push(Em.Object.create({
        from: -1,
        to: -1,
        label: Em.I18n.t('charts.heatmap.label.notApplicable'),
        cssStyle: "background-color:rgb(200, 200, 200)"
      }));
      return defs;
    }.property('minimumValue', 'maximumValue', 'numberOfSlots'),

    /**
     * In slot definitions this value is used to construct the label by appending
     * it to slot min-max values. For example giving '%' here would result in slot
     * definition label being '0% - 10%'.
     */
    slotDefinitionLabelSuffix: '',

    defaultMetric: '',

    /**
     * Name in the <code>App.ajax</code>
     * @type {String}
     */
    ajaxIndex: 'hosts.metrics',

    /**
     * Additional data for ajax-request
     * May be redeclared in child-objects
     * @type {Object}
     */
    ajaxData: {},

    /**
     * Maps server JSON into an object where keys are hostnames and values are the
     * true metric values. This function by default will map 'defaultMetric' into
     * its corresponding value.
     * 
     * @Function
     */
    metricMapper: function (json) {
      var hostToValueMap = {};
      var metricName = this.get('defaultMetric');
      if (json.items) {
        var props = metricName.split('.');
        json.items.forEach(function (item) {
          var value = item;
          props.forEach(function (prop) {
            if (value != null && prop in value) {
              value = value[prop];
            } else {
              value = null;
            }
          });
          if (value != null) {
            var hostName = item.Hosts.host_name;
            hostToValueMap[hostName] = value;
          }
        });
      }
      return hostToValueMap;
    },

    hostToValueMap: null,

    hostToSlotMap: function(){
      var hostToValueMap = this.get('hostToValueMap');
      var slotDefs = this.get('slotDefinitions');
      var allHosts = App.Host.find();
      var hostToSlotMap = {};
      if (hostToValueMap && allHosts) {
        allHosts.forEach(function(host){
          var slot = -1;
          var key = host.get('hostName');
          if (key in hostToValueMap) {
            var value = hostToValueMap[key];
            if (isNaN(value)) {
              slot = slotDefs.length - 2;
            } else {
              for ( var slotIndex = 0; slotIndex < slotDefs.length - 2; slotIndex++) {
                var slotDef = slotDefs[slotIndex];
                if (value >= slotDef.from && value <= slotDef.to) {
                  slot = slotIndex;
                }
              }
              if(slot < 0){
                // Assign it to the last legend
                slot = slotDefs.length - 3;
              }
            }
          } else {
            slot = slotDefs.length - 1;
          }
          if (slot > -1) {
            hostToSlotMap[key] = slot;
          }
        });
      }
      return hostToSlotMap;
    }.property('hostToValueMap', 'slotDefinitions'),

    /**
     * Determines which slot each host falls into. This information is given to
     * the callback's #map(hostnameToSlotObject) method. The
     * 'hostnameToSlotObject' has key as hostname, and the slot index as value.
     */
    refreshHostSlots: function () {
      this.set('loading', true);
      var fixedMetricName = this.get('defaultMetric');
      fixedMetricName = fixedMetricName.replace(/\./g, "/");
      var ajaxData = {
        metricName: fixedMetricName
      };
      jQuery.extend(ajaxData, this.get('ajaxData'));

      App.ajax.send({
        name: this.get('ajaxIndex'),
        sender: this,
        data: ajaxData,
        success: 'refreshHostSlotsSuccessCallback',
        error: 'refreshHostSlotsErrorCallback'
      });
    },

    refreshHostSlotsSuccessCallback: function (data) {
      var hostToValueMap = this.metricMapper(data);
      this.set('hostToValueMap', hostToValueMap);
      this.set('loading', false);
    },

    refreshHostSlotsErrorCallback: function () {
      this.set('loading', false);
    },

    /**
     * Turns numbers into displayable values. For example 24.345432425 into 24.3
     * etc.
     * 
     * @private
     */
    formatLegendNumber: function (num) {
      var fraction = num % 1;
      if (fraction > 0) {
        return parseFloat(num.toFixed(1));
      }
      return num;
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_cpuWaitIO", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapCpuWaitIOMetric = App.MainChartHeatmapMetric.extend({
    name: Em.I18n.t('charts.heatmap.metrics.cpuWaitIO'),
    maximumValue: 100,
    defaultMetric: 'metrics.cpu.cpu_wio',
    units: '%',
    slotDefinitionLabelSuffix: '%',
    metricMapper: function (json) {
      var map = this._super(json);
      for ( var host in map) {
        if (host in map) {
          var val = map[host];
          map[host] = (val * 100).toFixed(1);
        }
      }
      return map;
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Base class for any HDFS metric.
   */
  App.MainChartHeatmapDFSMetrics = App.MainChartHeatmapMetric.extend({

    ajaxIndex: 'hosts.metrics.host_component',

    ajaxData: {
      serviceName: 'HDFS',
      componentName: 'DATANODE'
    },

    /**
     * Custom mapper for DFS metrics
     */
    metricMapper: function(json) {
      return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
    },
    /**
     * Utility function which allows extending classes to transform the value
     * assigned to a host.
     * 
     * @Function
     */
    transformValue: null
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_bytesread", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapDFSBytesReadMetric = App.MainChartHeatmapDFSMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.bytesRead'),
    maximumValue: 1024, // 1GB
    defaultMetric: 'metrics.dfs.datanode.bytes_read',
    units: 'MB',
    slotDefinitionLabelSuffix: 'MB',
    transformValue: function (value) {
      return value / (1024 * 1024); // bytes divided by 1MB.
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_byteswritten", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapDFSBytesWrittenMetric = App.MainChartHeatmapDFSMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.bytesWritten'),
    maximumValue: 1024, // 1GB
    defaultMetric: 'metrics.dfs.datanode.bytes_written',
    units: 'MB',
    slotDefinitionLabelSuffix: 'MB',
    transformValue: function (value) {
      return value / (1024 * 1024); // bytes divided by 1MB.
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_gctime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapDFSGCTimeMillisMetric = App.MainChartHeatmapDFSMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.DFSGarbageCollection'),
    maximumValue: 10000,
    defaultMetric: 'metrics.jvm.gcTimeMillis',
    units: ' ms',
    slotDefinitionLabelSuffix: ' ms'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_memHeapUsed", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapDFSMemHeapUsedMetric = App.MainChartHeatmapDFSMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.DFSMemHeapUsed'),
    maximumValue: 512,
    defaultMetric: 'metrics.jvm.memHeapUsedM',
    units: 'MB',
    slotDefinitionLabelSuffix: 'MB'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_diskspaceused", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapDiskSpaceUsedMetric = App.MainChartHeatmapMetric.extend({
    name: Em.I18n.t('charts.heatmap.metrics.diskSpaceUsed'),
    maximumValue: 100,
    defaultMetric: 'metrics.disk',
    units: '%',
    slotDefinitionLabelSuffix: '%',
    metricMapper: function (json) {
      var hostToValueMap = {};
      var self = this;
      var metricName = this.get('defaultMetric');
      if (json.items) {
        var props = metricName.split('.');
        json.items.forEach(function (item) {
          var value = item;
          props.forEach(function (prop) {
            if (value != null && prop in value) {
              value = value[prop];
            } else {
              value = null;
            }
          });
          if (value != null) {
            value = self.diskUsageFormatted(value.disk_total - value.disk_free, value.disk_total);
            var hostName = item.Hosts.host_name;
            hostToValueMap[hostName] = value;
          }
        });
      }
      return hostToValueMap;
    },

    /**
     * Format percent disk usage to float with 2 digits
     */
    diskUsageFormatted: function(diskUsed, diskTotal) {
      var diskUsage = (diskUsed) / diskTotal * 100;
      if (isNaN(diskUsage) || diskUsage < 0 || diskUsage > 100) {
        return Em.I18n.t('charts.heatmap.unknown');
      }
      var s = Math.round(diskUsage * Math.pow(10, 2)) / Math.pow(10, 2);
      return isNaN(s) ? 0 : s;
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Base class for any HDFS metric.
   */
  App.MainChartHeatmapHbaseMetrics = App.MainChartHeatmapMetric.extend({

    ajaxIndex: 'hosts.metrics.host_component',

    ajaxData: {
      serviceName: 'HBASE',
      componentName: 'HBASE_REGIONSERVER'
    },

    /**
     * Custom mapper for HBase metrics
     */
    metricMapper: function(json) {
      return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
    },

    /**
     * Utility function which allows extending classes to transform the value
     * assigned to a host.
     *
     * @Function
     */
    transformValue: null
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_compactionqueue", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapHbaseCompactionQueueSize = App.MainChartHeatmapHbaseMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerCompactionQueueSize'),
    maximumValue: 10,
    defaultMetric: 'metrics.hbase.regionserver.compactionQueueSize',
    units: '',
    slotDefinitionLabelSuffix: ''
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_memstoresize", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapHbaseMemStoreSize = App.MainChartHeatmapHbaseMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerMemStoreSize'),
    maximumValue: 100,
    defaultMetric: 'metrics.hbase.regionserver.memstoreSizeMB',
    units: 'MB',
    slotDefinitionLabelSuffix: 'MB'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_readrequest", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapHbaseReadReqCount = App.MainChartHeatmapHbaseMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerReadCount'),
    maximumValue: 200,
    defaultMetric: 'metrics.hbase.regionserver.readRequestsCount',
    units: '',
    slotDefinitionLabelSuffix: ''
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_regions", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapHbaseRegions = App.MainChartHeatmapHbaseMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerRegions'),
    maximumValue: 10,
    defaultMetric: 'metrics.hbase.regionserver.regions',
    units: '',
    slotDefinitionLabelSuffix: ''
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_writerequest", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapHbaseWriteReqCount = App.MainChartHeatmapHbaseMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerWriteCount'),
    maximumValue: 200,
    defaultMetric: 'metrics.hbase.regionserver.writeRequestsCount',
    units: '',
    slotDefinitionLabelSuffix: ''
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Base class for any MapReduce metric.
   */
  App.MainChartHeatmapMapreduceMetrics = App.MainChartHeatmapMetric.extend({

    ajaxIndex: 'hosts.metrics.host_component',

    ajaxData: {
      serviceName: 'MAPREDUCE',
      componentName: 'TASKTRACKER'
    },

    /**
     * Custom mapper for MapReduce metrics
     */
    metricMapper: function(json) {
      return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
    },

    /**
     * Utility function which allows extending classes to transform the value
     * assigned to a host.
     * 
     * @Function
     */
    transformValue: null
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_gctime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapMapreduceGCTimeMillisMetric = App.MainChartHeatmapMapreduceMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.MapReduceGCTime'),
    maximumValue: 10000,
    defaultMetric: 'metrics.jvm.gcTimeMillis',
    units: ' ms',
    slotDefinitionLabelSuffix: ' ms'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_mapsRunning", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapMapreduceMapsRunningMetric = App.MainChartHeatmapMapreduceMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.mapsRunning'),
    maximumValue: 100,
    defaultMetric: 'metrics.mapred.tasktracker.maps_running',
    units: ' maps',
    slotDefinitionLabelSuffix: ' Maps'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_memHeapUsed", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapMapreduceMemHeapUsedMetric = App.MainChartHeatmapMapreduceMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.MRMemHeapUsed'),
    maximumValue: 512,
    defaultMetric: 'metrics.jvm.memHeapUsedM',
    units: 'MB',
    slotDefinitionLabelSuffix: 'MB'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_mapreduce_reducesRunning", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * 
   */
  App.MainChartHeatmapMapreduceReducesRunningMetric = App.MainChartHeatmapMapreduceMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.reducesRunning'),
    maximumValue: 100,
    defaultMetric: 'metrics.mapred.tasktracker.reduces_running',
    units: ' reduces',
    slotDefinitionLabelSuffix: ' Reduces'
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_memoryused", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Base class for any heatmap metric.
   * 
   * This class basically provides the following for each heatmap metric.
   * <ul>
   * <li> Provides number of slots in which temperature can fall.
   * <li> Maintains the maximum value so as to scale slot ranges.
   * <li> Gets JSON data from server and maps response for all hosts into above
   * slots.
   * </ul>
   * 
   */
  App.MainChartHeatmapMemoryUsedMetric = App.MainChartHeatmapMetric.extend({
    name: Em.I18n.t('charts.heatmap.metrics.memoryUsed'),
    maximumValue: 100,
    defaultMetric: 'metrics.memory',
    units: '%',
    slotDefinitionLabelSuffix: '%',
    metricMapper: function (json) {
      var hostToValueMap = {};
      var metricName = this.get('defaultMetric');
      if (json.items) {
        var props = metricName.split('.');
        json.items.forEach(function (item) {
          var value = item;
          props.forEach(function (prop) {
            if (value != null && prop in value) {
              value = value[prop];
            } else {
              value = null;
            }
          });
          if (value != null) {
            var total = value.mem_total;
            var used = value.mem_total - value.mem_free - value.mem_cached;
            value = ((used * 100) / total).toFixed(1);
            var hostName = item.Hosts.host_name;
            hostToValueMap[hostName] = value;
          }
        });
      }
      return hostToValueMap;
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_processrun", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Base class for any heatmap metric.
   * 
   * This class basically provides the following for each heatmap metric.
   * <ul>
   * <li> Provides number of slots in which temperature can fall.
   * <li> Maintains the maximum value so as to scale slot ranges.
   * <li> Gets JSON data from server and maps response for all hosts into above
   * slots.
   * </ul>
   * 
   */
  App.MainChartHeatmapProcessRunMetric = App.MainChartHeatmapMetric.extend({
    name: Em.I18n.t('charts.heatmap.metrics.processRun'),
    maximumValue: 1,
    defaultMetric: 'metrics.process.proc_run',
    units: 'Processes',
    metricMapper: function (json) {
      var map = this._super(json);
      for ( var host in map) {
        if (host in map) {
          var val = map[host];
          map[host] = val.toFixed(1);
        }
      }
      return map;
    }
  });
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Base class for any YARN metric.
   */
  App.MainChartHeatmapYarnMetrics = App.MainChartHeatmapMetric.extend({

    ajaxIndex: 'hosts.metrics.host_component',

    ajaxData: {
      serviceName: 'YARN',
      componentName: 'NODEMANAGER'
    },

    /**
     * Custom mapper for YARN metrics
     */
    metricMapper: function(json) {
      return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
    },

    /**
     * Utility function which allows extending classes to transform the value
     * assigned to a host.
     *
     * @Function
     */
    transformValue: null
  });
  
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_ResourceUsed", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapYarnResourceUsedMetric = App.MainChartHeatmapYarnMetrics.extend({
    metricUrlTemplate: '/clusters/{clusterName}/services/YARN/components/RESOURCEMANAGER?fields=ServiceComponentInfo/rm_metrics',
    name: Em.I18n.t('charts.heatmap.metrics.YarnMemoryUsed'),
    maximumValue: 100,
    defaultMetric: 'ServiceComponentInfo.rm_metrics.cluster.nodeManagers',
    units: ' %',
    slotDefinitionLabelSuffix: ' %',
    metricMapper: function (json) {
      var hostToValueMap = {};
      var metricName = this.get('defaultMetric');
      var props = metricName.split('.');
      var value = json;
      props.forEach(function (prop) {
        if (value != null && prop in value) {
          value = value[prop];
        } else {
          value = null;
        }
      });
      if (value != null) {
        value = JSON.parse(value);
        if (value instanceof Array) {
          value.forEach(function(host) {
            hostToValueMap[host.HostName] = (host.UsedMemoryMB / host.AvailableMemoryMB * 100).toFixed(1);
          });
        }
      }
      return hostToValueMap;
    }
  });
  
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_gctime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapYarnGCTimeMillisMetric = App.MainChartHeatmapYarnMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.YarnGCTime'),
    maximumValue: 10000,
    defaultMetric: 'metrics.jvm.gcTimeMillis',
    units: ' ms',
    slotDefinitionLabelSuffix: ' ms'
  });
  
});
window.require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_memHeapUsed", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   *
   */
  App.MainChartHeatmapYarnMemHeapUsedMetric = App.MainChartHeatmapYarnMetrics.extend({
    name: Em.I18n.t('charts.heatmap.metrics.YarnMemHeapUsed'),
    maximumValue: 512,
    defaultMetric: 'metrics.jvm.memHeapUsedM',
    units: 'MB',
    slotDefinitionLabelSuffix: 'MB'
  });
  
});
window.require.register("controllers/main/host", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');
  var componentHelper = require('utils/component');
  var batchUtils = require('utils/batch_scheduled_requests');

  App.MainHostController = Em.ArrayController.extend({
    name:'mainHostController',
    content: App.Host.find(),

    clearFilters: null,

    /**
     * Components which will be shown in component filter
     * @returns {Array}
     */
    componentsForFilter:function() {
      var installedComponents = componentHelper.getInstalledComponents();
      installedComponents.setEach('checkedForHostFilter', false);
      return installedComponents;
    }.property('App.router.clusterController.isLoaded'),

    /**
     * Master components
     * @returns {Array}
     */
    masterComponents:function () {
      return this.get('componentsForFilter').filterProperty('isMaster', true);
    }.property('componentsForFilter'),

    /**
     * Slave components
     * @returns {Array}
     */
    slaveComponents:function () {
      return this.get('componentsForFilter').filterProperty('isSlave', true);
    }.property('componentsForFilter'),

    /**
     * Client components
     * @returns {Array}
     */
    clientComponents: function() {
      return this.get('componentsForFilter').filterProperty('isClient', true);
    }.property('componentsForFilter'),

    /**
     * Filter hosts by componentName of <code>component</code>
     * @param {App.HostComponent} component
     */
    filterByComponent:function (component) {
      if(!component)
        return;
      var id = component.get('componentName');
      var column = 6;
      this.get('componentsForFilter').setEach('checkedForHostFilter', false);

      var filterForComponent = {
        iColumn: column,
        value: id,
        type: 'multiple'
      };
      App.db.setFilterConditions(this.get('name'), [filterForComponent]);
    },
    /**
     * On click callback for delete button
     */
    deleteButtonPopup:function () {
      var self = this;
      App.showConfirmationPopup(function(){
        self.removeHosts();
      });
    },

    showAlertsPopup: function (event) {
      var host = event.context;
      App.router.get('mainAlertsController').loadAlerts(host.get('hostName'), "HOST");
      App.ModalPopup.show({
        header: this.t('services.alerts.headingOfList'),
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/host/alerts_popup'),
          controllerBinding: 'App.router.mainAlertsController',
          alerts: function () {
            return this.get('controller.alerts');
          }.property('controller.alerts'),

          closePopup: function () {
            this.get('parentView').hide();
          }
        }),
        primary: Em.I18n.t('common.close'),
        secondary : null,
        didInsertElement: function () {
          this.$().find('.modal-footer').addClass('align-center');
          this.$().children('.modal').css({'margin-top': '-350px'});
        }
      });
      event.stopPropagation();
    },

    /**
     * remove selected hosts
     */
    removeHosts:function () {
      var hosts = this.get('content');
      var selectedHosts = hosts.filterProperty('isChecked', true);
      selectedHosts.forEach(function (_hostInfo) {
        console.log('Removing:  ' + _hostInfo.hostName);
      });
      this.get('fullContent').removeObjects(selectedHosts);
    },

    /**
     * remove hosts with id equal host_id
     * @param {String} host_id
     */
    checkRemoved:function (host_id) {
      var hosts = this.get('content');
      var selectedHosts = hosts.filterProperty('id', host_id);
      this.get('fullContent').removeObjects(selectedHosts);
    },

    /**
     * Bulk operation wrapper
     * @param {Object} operationData - data about bulk operation (action, hosts or hostComponents etc)
     * @param {Array} hosts - list of affected hosts
     */
    bulkOperation: function(operationData, hosts) {
      if (operationData.componentNameFormatted) {
        if (operationData.action === 'RESTART') {
          this.bulkOperationForHostComponentsRestart(operationData, hosts);
        }
        else {
          if (operationData.action.indexOf('DECOMMISSION') != -1) {
            this.bulkOperationForHostComponentsDecommission(operationData, hosts);
          }
          else {
            this.bulkOperationForHostComponents(operationData, hosts);
          }
        }
      }
      else {
        if (operationData.action === 'RESTART') {
          this.bulkOperationForHostsRestart(operationData, hosts);
        }
        else {
          if (operationData.action === 'PASSIVE_STATE') {
            this.bulkOperationForHostsPassiveState(operationData, hosts);
          }
          else {
            this.bulkOperationForHosts(operationData, hosts);
          }
        }
      }
    },

    /**
     * Bulk operation (start/stop all) for selected hosts
     * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
     * @param {Array} hosts - list of affected hosts
     */
    bulkOperationForHosts: function(operationData, hosts) {
      var query = [];
      hosts.forEach(function(host) {
        var subQuery = '(HostRoles/component_name.in(%@)&HostRoles/host_name=' + host.get('hostName') + ')';
        var components = [];
        host.get('hostComponents').forEach(function(hostComponent) {
          if (hostComponent.get('passiveState') == 'OFF') {
            if (hostComponent.get('isMaster') || hostComponent.get('isSlave')) {
              if (hostComponent.get('workStatus') === operationData.actionToCheck) {
                components.push(hostComponent.get('componentName'));
              }
            }
          }
        });
        if (components.length) {
          query.push(subQuery.fmt(components.join(',')));
        }
      });
      if (query.length) {
        query = query.join('|');
        App.ajax.send({
          name: 'bulk_request.hosts.all_components',
          sender: this,
          data: {
            query: query,
            state: operationData.action,
            requestInfo: operationData.message
          },
          success: 'bulkOperationForHostComponentsSuccessCallback'
        });
      }
      else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('rolling.nothingToDo.body').format(Em.I18n.t('hosts.host.maintainance.allComponents.context')),
          secondary: false
        });
      }
    },

    /**
     * Bulk restart for selected hosts
     * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
     * @param {Ember.Enumerable} hosts - list of affected hosts
     */
    bulkOperationForHostsRestart: function(operationData, hosts) {
      var hostComponents = [];
      hosts.forEach(function(host) {
        hostComponents.pushObjects(host.get('hostComponents').filterProperty('passiveState','OFF').toArray());
      });
      batchUtils.restartHostComponents(hostComponents, Em.I18n.t('rollingrestart.context.allOnSelectedHosts'));
    },

    /**
     * Bulk turn on/off passive state for selected hosts
     * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
     * @param {Array} hosts - list of affected hosts
     */
    bulkOperationForHostsPassiveState: function(operationData, hosts) {
      var affectedHosts = [];
      hosts.forEach(function(host) {
        if (host.get('passiveState') !== operationData.state) {
          affectedHosts.push(host);
        }
      });
      if (affectedHosts.length) {
        App.ajax.send({
          name: 'bulk_request.hosts.passive_state',
          sender: this,
          data: {
            hosts: affectedHosts,
            hostNames: affectedHosts.mapProperty('hostName').join(','),
            passive_state: operationData.state,
            requestInfo: operationData.message
          },
          success: 'updateHostPassiveState'
        });
      }
      else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('hosts.bulkOperation.passiveState.nothingToDo.body'),
          secondary: false
        });
      }
    },

    updateHostPassiveState: function(data, opt, params) {
      params.hosts.setEach('passiveState', params.passive_state);
      App.router.get('clusterController').loadUpdatedStatus(function(){
        batchUtils.infoPassiveState(params.passive_state);
      });
    },
    /**
     * Bulk operation for selected hostComponents
     * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
     * @param {Array} hosts - list of affected hosts
     */
    bulkOperationForHostComponents: function(operationData, hosts) {
      var service = App.Service.find(operationData.serviceName);
      var components = service.get('hostComponents').filter(function(hc) {
        if (hc.get('componentName') != operationData.componentName || hc.get('passiveState') != 'OFF' ) {
          return false;
        }
        if(hc.get('workStatus') == operationData.action) {
          return false;
        }
        return hosts.contains(hc.get('host'));
      });

      if (components.length) {
        var hostsWithComponentInProperState = components.mapProperty('host.hostName');
        App.ajax.send({
          name: 'bulk_request.host_components',
          sender: this,
          data: {
            hostNames: hostsWithComponentInProperState.join(','),
            state: operationData.action,
            requestInfo: operationData.message + ' ' + operationData.componentNameFormatted,
            componentName: operationData.componentName
          },
          success: 'bulkOperationForHostComponentsSuccessCallback'
        });
      }
      else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('rolling.nothingToDo.body').format(operationData.componentNameFormatted),
          secondary: false
        });
      }
    },

    /**
     * Bulk decommission/recommission for selected hostComponents
     * @param {Object} operationData
     * @param {Array} hosts
     */
    bulkOperationForHostComponentsDecommission: function(operationData, hosts) {
      var service = App.Service.find(operationData.serviceName);
      var components = service.get('hostComponents').filter(function(hc) {
        if (hc.get('componentName') != operationData.realComponentName || hc.get('passiveState') != 'OFF' ) {
          return false;
        }
        return hosts.contains(hc.get('host'));
      });

      if (components.length) {
        var hostsWithComponentInProperState = components.mapProperty('host.hostName');
        var turn_off = operationData.action.indexOf('OFF') !== -1;
        var svcName = operationData.serviceName;
        var masterName = operationData.componentName;
        var slaveName = operationData.realComponentName;
        var hostNames = hostsWithComponentInProperState.join(',');
        if (turn_off) {
          // For recommession
          if (svcName === "YARN" || svcName === "HBASE" || svcName === "HDFS") {
            App.router.get('mainHostDetailsController').doRecommissionAndStart(hostNames, svcName, masterName, slaveName);
          }
          else if (svcName === "MAPREDUCE") {
            App.router.get('mainHostDetailsController').doRecommissionAndRestart(hostNames, svcName, masterName, slaveName);
          }
        } else {
          hostsWithComponentInProperState = components.filterProperty('workStatus','STARTED').mapProperty('host.hostName');
          //For decommession
          if (svcName == "HBASE") {
            // HBASE service, decommission RegionServer in batch requests
            App.router.get('mainHostDetailsController').doDecommissionRegionServer(hostNames, svcName, masterName, slaveName);
          } else {
            var parameters = {
              "slave_type": slaveName
            };
            var contextString = turn_off? 'hosts.host.' + slaveName.toLowerCase() + '.recommission':
              'hosts.host.' + slaveName.toLowerCase() + '.decommission';
            if (turn_off) {
              parameters['included_hosts'] = hostsWithComponentInProperState.join(',')
            }
            else {
              parameters['excluded_hosts'] = hostsWithComponentInProperState.join(',');
            }
            App.ajax.send({
              name: 'bulk_request.decommission',
              sender: this,
              data: {
                context: Em.I18n.t(contextString),
                serviceName: service.get('serviceName'),
                componentName: operationData.componentName,
                parameters: parameters
              },
              success: 'bulkOperationForHostComponentsSuccessCallback'
            });
          }
        }
      }
      else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('rolling.nothingToDo.body').format(operationData.componentNameFormatted),
          secondary: false
        });
      }
    },

    /**
     * Bulk restart for selected hostComponents
     * @param {Object} operationData
     * @param {Array} hosts
     */
    bulkOperationForHostComponentsRestart: function(operationData, hosts) {
      var service = App.Service.find(operationData.serviceName);
      var components = service.get('hostComponents').filter(function(hc) {
        if (hc.get('componentName') != operationData.componentName || hc.get('passiveState') != 'OFF') {
          return false;
        }
        return hosts.contains(hc.get('host'));
      });

      if (components.length) {
        batchUtils.showRollingRestartPopup(components.objectAt(0).get('componentName'), service.get('displayName'), service.get('passiveState') === "ON", false, components);
      }
      else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('rolling.nothingToDo.body').format(operationData.componentNameFormatted),
          secondary: false
        });
      }
    },

    updateHostComponentsPassiveState: function(data, opt, params) {
      App.router.get('clusterController').loadUpdatedStatus(function(){
        batchUtils.infoPassiveState(params.passive_state);
      });
    },
    /**
     * Show BO popup after bulk request
     */
    bulkOperationForHostComponentsSuccessCallback: function() {
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        if (initValue) {
          App.router.get('backgroundOperationsController').showPopup();
        }
      });
    }

  });
  
});
window.require.register("controllers/main/host/addHost/step4_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.AddHostStep4Controller= Em.Controller.extend({

    name:"addHostStep4Controller",

    serviceConfigGroups: null,
    isConfigGroupLoaded: false,

    loadConfigGroups: function () {
      App.ajax.send({
        name: 'config_groups.all_fields',
        sender: this,
        success: 'successLoadingConfigGroup'
      });
    },

    successLoadingConfigGroup: function (data) {
      App.router.get('addHostController.content').set('configGroups',data.items);
      this.set('isConfigGroupLoaded', true);
    },

    configGroupsLoading: function () {
      var df = $.Deferred();
      if (this.get('isConfigGroupLoaded')) {
        df.resolve();
      } else {
        this.interval = setInterval(function () {
          if (this.get('isConfigGroupLoaded')) {
            df.resolve();
            clearInterval(this.interval);
          }
        }.bind(this), 55);
      }
      return df.promise();
    }

  });

  
});
window.require.register("controllers/main/host/add_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.AddHostController = App.WizardController.extend({

    name: 'addHostController',

    totalSteps: 7,

    /**
     * Used for hiding back button in wizard
     */
    hideBackButton: true,

    /**
     * All wizards data will be stored in this variable
     *
     * cluster - cluster name
     * hosts - hosts, ssh key, repo info, etc.
     * services - services list
     * hostsInfo - list of selected hosts
     * slaveComponentHosts, hostSlaveComponents - info about slave hosts
     * masterComponentHosts - info about master hosts
     * serviceConfigGroups - info about selected config group for service
     * configGroups - all config groups
     * config??? - to be described later
     */
    content: Em.Object.create({
      cluster: null,
      hosts: null,
      installOptions: null,
      services: null,
      slaveComponentHosts: null,
      masterComponentHosts: null,
      serviceConfigProperties: null,
      advancedServiceConfig: null,
      controllerName: 'addHostController',
      serviceConfigGroups: null,
      configGroups: null
    }),

    components:require('data/service_components'),

    setCurrentStep: function (currentStep, completed) {
      this._super(currentStep, completed);
      App.clusterStatus.setClusterStatus({
        wizardControllerName: this.get('name'),
        localdb: App.db.data
      });
    },

    /**
     * return new object extended from clusterStatusTemplate
     * @return Object
     */
    getCluster: function(){
      return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
    },

    /**
     * return new object extended from installOptionsTemplate
     * @return Object
     */
    getInstallOptions: function(){
      return jQuery.extend({}, this.get('installOptionsTemplate'));
    },

    /**
     * return empty hosts array
     * @return Array
     */
    getHosts: function(){
      return [];
    },

    /**
     * Remove host from model. Used at <code>Confirm hosts</code> step
     * @param hosts Array of hosts, which we want to delete
     */
    removeHosts: function (hosts) {
      var dbHosts = this.getDBProperty('hosts');
      hosts.forEach(function (_hostInfo) {
        var host = _hostInfo.hostName;
        delete dbHosts[host];
      });
      this.setDBProperty('hosts', dbHosts);
    },

    /**
     * Load services data from server.
     */
    loadServicesFromServer: function() {
      var displayOrderConfig = require('data/services');
      var apiUrl = App.get('stack2VersionURL');
      var apiService = this.loadServiceComponents(displayOrderConfig, apiUrl);
      //
      apiService.forEach(function(item, index){
        apiService[index].isSelected = App.Service.find().someProperty('id', item.serviceName);
        apiService[index].isDisabled = apiService[index].isSelected;
        apiService[index].isInstalled = apiService[index].isSelected;
      });
      this.set('content.services', apiService);
      this.setDBProperty('service', apiService);
    },

    /**
     * Load services data. Will be used at <code>Select services(step4)</code> step
     */
    loadServices: function () {
      var servicesInfo = this.getDBProperty('service');
      servicesInfo.forEach(function (item, index) {
        servicesInfo[index] = Em.Object.create(item);
      });
      this.set('content.services', servicesInfo);
      console.log('AddHostController.loadServices: loaded data ', servicesInfo);
      var serviceNames = servicesInfo.filterProperty('isSelected', true).mapProperty('serviceName');
      console.log('selected services ', serviceNames);
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadMasterComponentHosts: function () {
      var masterComponentHosts = this.getDBProperty('masterComponentHosts');
      if (!masterComponentHosts) {
        masterComponentHosts = [];
        App.HostComponent.find().filterProperty('isMaster', true).forEach(function (item) {
          masterComponentHosts.push({
            component: item.get('componentName'),
            hostName: item.get('host.hostName'),
            isInstalled: true,
            serviceId: item.get('service.id'),
            display_name: item.get('displayName')
          })
        });
        this.setDBProperty('masterComponentHosts', masterComponentHosts);
      }
      this.set("content.masterComponentHosts", masterComponentHosts);
      console.log("AddHostController.loadMasterComponentHosts: loaded hosts ", masterComponentHosts);
    },

    /**
     * Save HBase and ZooKeeper to main controller
     * @param stepController
     */
    saveHbZk: function(stepController) {
      var self = this;
      var hosts = stepController.get('hosts');
      var headers = stepController.get('headers');
      var masterComponentHosts = this.getDBProperty('masterComponentHosts');

      headers.forEach(function(header) {
        var rm = masterComponentHosts.filterProperty('component', header.get('name'));
        if(rm) {
          masterComponentHosts.removeObjects(rm);
        }
      });

      headers.forEach(function(header) {
        var component = self.get('components').findProperty('component_name', header.get('name'));
        hosts.forEach(function(host) {
          if (host.get('checkboxes').findProperty('title', component.display_name).checked) {
            masterComponentHosts .push({
              display_name: component.display_name,
              component: component.component_name,
              hostName: host.get('hostName'),
              serviceId: component.service_name,
              isInstalled: false
            });
          }
        });
      });

      console.log("installerController.saveMasterComponentHosts: saved hosts ", masterComponentHosts);
      this.setDBProperty('masterComponentHosts', masterComponentHosts);
      this.set('content.masterComponentHosts', masterComponentHosts);
    },

    /**
     * return slaveComponents bound to hosts
     * @return {Array}
     */
    getSlaveComponentHosts: function () {
      return this._super().filter(function(component){
        return component.isInstalled;
      });
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadSlaveComponentHosts: function () {
      var slaveComponentHosts = this.getDBProperty('slaveComponentHosts');
      if (!slaveComponentHosts) {
        slaveComponentHosts = this.getSlaveComponentHosts();
      }
      this.set("content.slaveComponentHosts", slaveComponentHosts);
      console.log("AddHostController.loadSlaveComponentHosts: loaded hosts ", slaveComponentHosts);
    },

    /**
     * Load information about hosts with clients components
     */
    loadClients: function () {
      var clients = this.getDBProperty('clientInfo');
      this.set('content.clients', clients);
      console.log("AddHostController.loadClients: loaded list ", clients);
    },

    /**
     * Generate clients list for selected services and save it to model
     */
    saveClients: function () {
      var clients = [];
      var serviceComponents = require('data/service_components');
      var hostComponents = App.HostComponent.find();

      this.get('content.services').filterProperty('isSelected', true).forEach(function (_service) {
        var client = serviceComponents.filterProperty('service_name', _service.serviceName).findProperty('isClient', true);
        if (client) {
          clients.pushObject({
            component_name: client.component_name,
            display_name: client.display_name,
            isInstalled: hostComponents.filterProperty('componentName', client.component_name).length > 0
          });
        }
      }, this);

      this.setDBProperty('clientInfo', clients);
      this.set('content.clients', clients);
      console.log("AddHostController.saveClients: saved list ", clients);
    },

    /**
     *  Apply config groups from step4 Configurations
     */
    applyConfigGroup: function () {
      var serviceConfigGroups = this.get('content.serviceConfigGroups');
      serviceConfigGroups.forEach(function (group){
        if (group.configGroups.someProperty('ConfigGroup.group_name', group.selectedConfigGroup)) {
          var configGroup = group.configGroups.findProperty('ConfigGroup.group_name', group.selectedConfigGroup);
          group.hosts.forEach(function(host){
            configGroup.ConfigGroup.hosts.push({
              host_name: host
            });
          },this);
          delete configGroup.href;
          App.ajax.send({
            name: 'config_groups.update_config_group',
            sender: this,
            data: {
              id: configGroup.ConfigGroup.id,
              configGroup: configGroup
            }
          });
        }
      },this);
    },

    /**
     * Load information about selected config groups
     */
    getServiceConfigGroups: function () {
      var serviceConfigGroups = this.getDBProperty('serviceConfigGroups');
      this.set('content.serviceConfigGroups', serviceConfigGroups);
    },

    /**
     * Save information about selected config groups
     */
    saveServiceConfigGroups: function () {
      this.setDBProperty('serviceConfigGroups', this.get('content.serviceConfigGroups'));
      this.set('content.serviceConfigGroups', this.get('content.serviceConfigGroups'));
    },

    /**
     * Set content.serviceConfigGroups for step4
     */
    loadServiceConfigGroups: function () {
      var slaveComponentHosts = this.get('content.slaveComponentHosts');
      var selectedServices = [];
      var selectedClientHosts = slaveComponentHosts.findProperty('componentName', 'CLIENT').hosts.mapProperty('hostName');
      var componentServiceMap = App.QuickDataMapper.componentServiceMap();

      slaveComponentHosts.forEach(function (slave) {
        if (slave.hosts.length > 0) {
          if (slave.componentName != "CLIENT") {
            var service = componentServiceMap[slave.componentName];
            var configGroups = this.get('content.configGroups').filterProperty('ConfigGroup.tag', service);
            var configGroupsNames = configGroups.mapProperty('ConfigGroup.group_name');
            var defaultGroupName = App.Service.DisplayNames[service] + ' Default';
            configGroupsNames.unshift(defaultGroupName);
            selectedServices.push({
              serviceId: service,
              displayName: App.Service.DisplayNames[service],
              hosts: slave.hosts.mapProperty('hostName'),
              configGroupsNames: configGroupsNames,
              configGroups: configGroups,
              selectedConfigGroup: defaultGroupName
            });
          }
        }
      }, this);
      if (selectedClientHosts.length > 0) {
        this.loadClients();
        var clients = this.get('content.clients');
        clients.forEach(function (client) {
          var service = componentServiceMap[client.component_name];
          var serviceMatch = selectedServices.findProperty('serviceId', service);
          if (serviceMatch) {
            serviceMatch.hosts = serviceMatch.hosts.concat(selectedClientHosts).uniq();
          } else {
            var configGroups = this.get('content.configGroups').filterProperty('ConfigGroup.tag', service);
            var configGroupsNames = configGroups.mapProperty('ConfigGroup.group_name').sort();
            var defaultGroupName = App.Service.DisplayNames[service] + ' Default';
            configGroupsNames.unshift(defaultGroupName);
            selectedServices.push({
              serviceId: service,
              displayName: App.Service.DisplayNames[service],
              hosts: selectedClientHosts,
              configGroupsNames: configGroupsNames,
              configGroups: configGroups,
              selectedConfigGroup: defaultGroupName
            });
          }
        }, this);
      }
      selectedServices.forEach(function(selectedService){
        selectedService.configGroups.sort(function(cfgA, cfgB){
          return cfgA.ConfigGroup.group_name >= cfgB.ConfigGroup.group_name;
        });
      });
      this.set('content.serviceConfigGroups', selectedServices);
    },

    loadServiceConfigProperties: function () {
      var serviceConfigProperties = App.db.get('AddService', 'serviceConfigProperties');
      if (!serviceConfigProperties || !serviceConfigProperties.length) {
        serviceConfigProperties = App.db.get('Installer', 'serviceConfigProperties');
      }
      this.set('content.serviceConfigProperties', serviceConfigProperties);
      console.log("AddHostController.loadServiceConfigProperties: loaded config ", serviceConfigProperties);
    },
    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '7':
        case '6':
        case '5':
          this.loadServiceConfigProperties();
          this.getServiceConfigGroups();
        case '4':
        case '3':
          this.loadClients();
          this.loadServices();
          this.loadMasterComponentHosts();
          this.loadSlaveComponentHosts();
          this.load('hosts');
        case '2':
          this.loadServices();
        case '1':
          this.load('hosts');
          this.load('installOptions');
          this.load('cluster');
      }
    },

    /**
     * Remove all loaded data.
     * Created as copy for App.router.clearAllSteps
     */
    clearAllSteps: function () {
      this.clearInstallOptions();
      // clear temporary information stored during the install
      this.set('content.cluster', this.getCluster());
    },

    /**
     * Clear all temporary data
     */
    finish: function () {
      this.setCurrentStep('1');
      this.clearAllSteps();
      this.clearStorageData();
      App.router.get('updateController').updateAll();
      App.updater.immediateRun('updateHost');
    },

    installServices: function (isRetry) {
      this.set('content.cluster.oldRequestsId', []);
      var clusterName = this.get('content.cluster.name');
      var data;
      var name;
      var hostnames = [];
      for (var hostname in this.getDBProperty('hosts')) {
        hostnames.push(hostname);
      }

      if (isRetry) {
        name = 'wizard.install_services.add_host_controller.is_retry';
      }
      else {
        name = 'wizard.install_services.add_host_controller.not_is_retry';
      }
      data = {
        "RequestInfo": {
          "context": Em.I18n.t('requestInfo.installComponents'),
          "query": "HostRoles/host_name.in(" + hostnames.join(',') + ")"
        },
        "Body": {
          "HostRoles": {"state": "INSTALLED"}
        }
      };
      data = JSON.stringify(data);
      App.ajax.send({
        name: name,
        sender: this,
        data: {
          data: data,
          cluster: clusterName
        },
        success: 'installServicesSuccessCallback',
        error: 'installServicesErrorCallback'
      });
    }

  });
  
});
window.require.register("controllers/main/host/configs_service", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MainHostServiceConfigsController = App.MainServiceInfoConfigsController.extend({
    name: 'mainHostServiceConfigsController',
    host: null,
    isHostsConfigsPage: true,
    /**
     * On load function
     */
    loadStep: function () {
      var content = this.get('content');
      this.set('host', content.host);
      this._super();
    },

    /**
     * Removes categories which are not valid for this host. Ex: Remove JOBTRACKER
     * category on host which does not have it installed.
     */
    renderServiceConfigs: function (serviceConfigs) {
      var newServiceConfigs = jQuery.extend({}, serviceConfigs);
      var hostHostComponentNames = [];
      var hostComponents = this.get('host.hostComponents');
      if (hostComponents) {
        hostComponents.forEach(function (hc) {
          var name = hc.get('componentName');
          if (!hostHostComponentNames.contains(name)) {
            hostHostComponentNames.push(name);
          }
        });
      }
      newServiceConfigs.configCategories = serviceConfigs.configCategories.filter(function (category) {
        var hcNames = category.get('hostComponentNames');
        if (hcNames != null && hcNames.length > 0) {
          var show = false;
          hcNames.forEach(function (name) {
            if (hostHostComponentNames.contains(name)) {
              show = true;
            }
          });
          return show;
        }
        return true;
      });
      this._super(newServiceConfigs);
    },

    typeTagToHostMap: null,

    configKeyToConfigMap: null,

    /**
     * This method will *not load* the overridden properties. However it will
     * replace the value shown for properties which this host has override for.
     */
    loadServiceConfigHostsOverrides: function (serviceConfigs, loadedGroupToOverrideSiteToTagMap, configGroups) {
      var thisHostName = this.get('host.hostName');
      var configKeyToConfigMap = {};
      serviceConfigs.forEach(function (item) {
        configKeyToConfigMap[item.name] = item;
      });
      var typeTagToGroupMap = {};
      var urlParams = [];
      for (var group in loadedGroupToOverrideSiteToTagMap) {
        var groupObj = configGroups.findProperty('name', group);
        if (groupObj.get('hosts').contains(thisHostName)) {
          var overrideTypeTags = loadedGroupToOverrideSiteToTagMap[group];
          for (var type in overrideTypeTags) {
            var tag = overrideTypeTags[type];
            typeTagToGroupMap[type + "///" + tag] = groupObj;
            urlParams.push('(type=' + type + '&tag=' + tag + ')');
          }
        }
      }
      this.set('typeTagToGroupMap', typeTagToGroupMap);
      this.set('configKeyToConfigMap', configKeyToConfigMap);
      if (urlParams.length > 0) {
        App.ajax.send({
          name: 'host.service_config_hosts_overrides',
          sender: this,
          data: {
            urlParams: urlParams.join('|')
          },
          success: 'loadServiceConfigHostsOverridesSuccessCallback',
          error: 'loadServiceConfigHostsOverridesErrorCallback'
        });
      }
    },
    loadServiceConfigHostsOverridesSuccessCallback: function (data) {
      var typeTagToGroupMap = this.get('typeTagToGroupMap');
      var configKeyToConfigMap = this.get('configKeyToConfigMap');
      data.items.forEach(function (config) {
        var group = typeTagToGroupMap[config.type + "///" + config.tag];
        var properties = config.properties;
        for ( var prop in properties) {
          var serviceConfig = configKeyToConfigMap[prop];
          var hostOverrideValue = properties[prop];
          if (serviceConfig && serviceConfig.displayType === 'int') {
            if (/\d+m$/.test(hostOverrideValue)) {
              hostOverrideValue = hostOverrideValue.slice(0, hostOverrideValue.length - 1);
            }
          } else if (serviceConfig && serviceConfig.displayType === 'checkbox') {
            switch (hostOverrideValue) {
              case 'true':
                hostOverrideValue = true;
                break;
              case 'false':
                hostOverrideValue = false;
                break;
            }
          }
          if (serviceConfig) {
            // Value of this property is different for this host.
            console.log("loadServiceConfigHostsOverrides(" + group + "): [" + group + "] OVERRODE(" + serviceConfig.name + "): " + serviceConfig.value + " -> " + hostOverrideValue);
            serviceConfig.value = hostOverrideValue;
            serviceConfig.defaultValue = hostOverrideValue;
            serviceConfig.isOriginalSCP = false;
            serviceConfig.group = group;
          }
        }
      });
      console.log("loadServiceConfigHostsOverrides(" + this.get('host.hostName') + "): Finished loading.");
    },
    loadServiceConfigHostsOverridesErrorCallback: function (request, ajaxOptions, error) {
      console.log("TRACE: error code status is: " + request.status);
    },
    /**
     * invoke dialog for switching group of host
     */
    switchHostGroup: function () {
      var self = this;
      App.config.launchSwitchConfigGroupOfHostDialog(this.get('selectedConfigGroup'), this.get('configGroups'), this.get('host.hostName'), function(newGroup){
        self.set('selectedConfigGroup', newGroup);
      })
    }
  });
  
});
window.require.register("controllers/main/host/details", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var batchUtils = require('utils/batch_scheduled_requests');

  App.MainHostDetailsController = Em.Controller.extend({
    name: 'mainHostDetailsController',
    content: null,
    isFromHosts: false,

    /**
     * path to page visited before
     */
    referer: '',
    /**
     * open dashboard page
     */
    routeHome: function () {
      App.router.transitionTo('main.dashboard');
    },

    /**
     * open summary page of the selected service
     * @param event
     */
    routeToService: function(event){
      var service = event.context;
      App.router.transitionTo('main.services.service.summary',service);
    },

    serviceActiveComponents: function() {
      return this.get('content.hostComponents').filterProperty('service.isInPassive',false);
    }.property('content.hostComponents'),

    serviceNonClientActiveComponents: function() {
      return this.get('serviceActiveComponents').filterProperty('isClient',false);
    }.property('serviceActiveComponents'),


    /**
     * Send specific command to server
     * @param url
     * @param _method
     * @param postData
     * @param callback
     */
    sendCommandToServer : function(url, postData, _method, callback){
      var url =  (App.testMode) ?
        '/data/wizard/deploy/poll_1.json' : //content is the same as ours
        App.apiPrefix + '/clusters/' + App.router.getClusterName() + url;

      var method = App.testMode ? 'GET' : _method;

      $.ajax({
        type: method,
        url: url,
        data: JSON.stringify(postData),
        dataType: 'json',
        timeout: App.timeout,
        success: function(data){
          if(data && data.Requests){
            callback(data.Requests.id);
          } else{
            callback(null);
            console.log('cannot get request id from ', data);
          }
        },

        error: function (request, ajaxOptions, error) {
          //do something
          console.log('error on change component host status');
          App.ajax.defaultErrorHandler(request, url, method);
        },

        statusCode: require('data/statusCodes')
      });
    },

    /**
     * send command to server to start selected host component
     * @param event
     */
    startComponent: function (event) {
      var self = this;
      App.showConfirmationPopup(function() {
        var component = event.context;
        var context = Em.I18n.t('requestInfo.startHostComponent') + " " + component.get('displayName');
        self.sendStartComponentCommand(component, context);
      });
    },
    
    /**
     * PUTs a command to server to start a component. If no 
     * specific component is provided, all components are started.
     * @param component  When <code>null</code> all startable components are started. 
     * @param context  Context under which this command is beign sent. 
     */
    sendStartComponentCommand: function(components, context) {
      var url = Em.isArray(components) ?
          '/hosts/' + this.get('content.hostName') + '/host_components' :
          '/hosts/' + this.get('content.hostName') + '/host_components/' + components.get('componentName').toUpperCase();
      var dataToSend = {
        RequestInfo : {
          "context" : context
        },
        Body:{
          HostRoles:{
            state: 'STARTED'
          }
        }
      };
      if (Em.isArray(components)) {
        dataToSend.RequestInfo.query = "HostRoles/component_name.in(" + components.mapProperty('componentName').join(',') + ")";
      }
      this.sendCommandToServer(url, dataToSend, 'PUT',
        function(requestId){

        if(!requestId){
          return;
        }

        console.log('Send request for STARTING successfully');

        if (App.testMode) {
          if(Em.isArray(components)){
            var allComponents = this.get('content.hostComponents');
            allComponents.forEach(function(component){
              component.set('workStatus', App.HostComponentStatus.stopping);
              setTimeout(function(){
                component.set('workStatus', App.HostComponentStatus.stopped);
              },App.testModeDelayForActions);
            });
          } else {
            components.set('workStatus', App.HostComponentStatus.starting);
            setTimeout(function(){
              components.set('workStatus', App.HostComponentStatus.started);
            },App.testModeDelayForActions);
          }
        } else {
          App.router.get('clusterController').loadUpdatedStatusDelayed(500);
        }
        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });

      });
    },

    /**
     * send command to server to delete selected host component
     *
     */
    deleteComponent: function (event) {
      var self = this;
      var component = event.context;
      var componentName = component.get('componentName');
      var displayName = component.get('displayName');
      var isLastComponent = (App.HostComponent.find().filterProperty('componentName', componentName).get('length') === 1);
      App.ModalPopup.show({
        header: Em.I18n.t('popup.confirmation.commonHeader'),
        primary: Em.I18n.t('hosts.host.deleteComponent.popup.confirm'),
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/host/details/deleteComponentPopup')
        }),
        isChecked: false,
        disablePrimary: function () {
          return !this.get('isChecked');
        }.property('isChecked'),
        lastComponent: function() {
          if (isLastComponent) {
            this.set('isChecked', false);
            return true;
          } else {
            this.set('isChecked', true);
            return false;
          }
        }.property(),
        lastComponentError:  Em.View.extend({
          template: Ember.Handlebars.compile(Em.I18n.t('hosts.host.deleteComponent.popup.warning').format(displayName))
        }),
        restartNagiosMsg: Em.View.extend({
          template: Ember.Handlebars.compile(Em.I18n.t('hosts.host.deleteComponent.popup.msg2').format(displayName))
        }),
        deleteComponentMsg: function() {
          return Em.I18n.t('hosts.host.deleteComponent.popup.msg1').format(displayName);
        }.property(),
        onPrimary: function () {
          self._doDeleteHostComponent(component);
          self.set('redrawComponents', true);
          this.hide();
        }
      });

    },

    /**
     * Trigger to reset list of master/slaves components on the view
     * @type {bool}
     */
    redrawComponents: false,

    /**
     * Deletes the given host component, or all host components.
     * 
     * @param component  When <code>null</code> all host components are deleted.
     * @return  <code>null</code> when components get deleted.
     *          <code>{xhr: XhrObj, url: "http://", method: "DELETE"}</code> 
     *          when components failed to get deleted. 
     */
    _doDeleteHostComponent: function(component) {
      var url = component !== null ? 
          '/hosts/' + this.get('content.hostName') + '/host_components/' + component.get('componentName').toUpperCase() : 
          '/hosts/' + this.get('content.hostName') + '/host_components';
      url = App.apiPrefix + '/clusters/' + App.router.getClusterName() + url;
      var deleted = null;
      $.ajax({
        type: 'DELETE',
        url: url,
        timeout: App.timeout,
        async: false,
        success: function (data) {
          deleted = null;
        },
        error: function (xhr, textStatus, errorThrown) {
          console.log('Error deleting host component');
          console.log(textStatus);
          console.log(errorThrown);
          deleted = {xhr: xhr, url: url, method: 'DELETE'};
        },
        statusCode: require('data/statusCodes')
        });
      return deleted;
    },

    /**
     * send command to server to upgrade selected host component
     * @param event
     */
    upgradeComponent: function (event) {
      var self = this;
      var component = event.context;
      App.showConfirmationPopup(function() {
        self.sendCommandToServer('/hosts/' + self.get('content.hostName') + '/host_components/' + component.get('componentName').toUpperCase(),{
              RequestInfo : {
                "context" : Em.I18n.t('requestInfo.upgradeHostComponent') + " " + component.get('displayName')
              },
              Body:{
                HostRoles:{
                  stack_id: 'HDP-1.2.2',
                  state: 'INSTALLED'
                }
              }
            }, 'PUT',
            function(requestId){
              if(!requestId){
                return;
              }

              console.log('Send request for UPGRADE successfully');

              if (App.testMode) {
                component.set('workStatus', App.HostComponentStatus.starting);
                setTimeout(function(){
                  component.set('workStatus', App.HostComponentStatus.started);
                },App.testModeDelayForActions);
              } else {
                App.router.get('clusterController').loadUpdatedStatusDelayed(500);
              }

              // load data (if we need to show this background operations popup) from persist
              App.router.get('applicationController').dataLoading().done(function (initValue) {
                if (initValue) {
                  App.router.get('backgroundOperationsController').showPopup();
                }
              });

            });
      });
    },
    /**
     * send command to server to stop selected host component
     * @param event
     */
    stopComponent: function (event) {
      var self = this;
      App.showConfirmationPopup(function() {
        var component = event.context;
        var context = Em.I18n.t('requestInfo.stopHostComponent')+ " " + component.get('displayName');
        self.sendStopComponentCommand(component, context);
      });
    },
    
    /**
     * PUTs a command to server to stop a component. If no 
     * specific component is provided, all components are stopped.
     * @param component  When <code>null</code> all components are stopped. 
     * @param context  Context under which this command is beign sent. 
     */
    sendStopComponentCommand: function(components, context){
      var url = Em.isArray(components) ?
          '/hosts/' + this.get('content.hostName') + '/host_components' :
          '/hosts/' + this.get('content.hostName') + '/host_components/' + components.get('componentName').toUpperCase();
      var dataToSend = {
        RequestInfo : {
          "context" : context
        },
        Body:{
          HostRoles:{
            state: 'INSTALLED'
          }
        }
      };
      if (Em.isArray(components)) {
        dataToSend.RequestInfo.query = "HostRoles/component_name.in(" + components.mapProperty('componentName').join(',') + ")";
      }
      this.sendCommandToServer( url, dataToSend, 'PUT',
        function(requestId){
        if(!requestId){
          return;
        }

        console.log('Send request for STOPPING successfully');

        if (App.testMode) {
          if(Em.isArray(components)){
            components.forEach(function(component){
              component.set('workStatus', App.HostComponentStatus.stopping);
              setTimeout(function(){
                component.set('workStatus', App.HostComponentStatus.stopped);
              },App.testModeDelayForActions);
            });
          } else {
            components.set('workStatus', App.HostComponentStatus.stopping);
            setTimeout(function(){
              components.set('workStatus', App.HostComponentStatus.stopped);
            },App.testModeDelayForActions);
          }

        } else {
          App.router.get('clusterController').loadUpdatedStatusDelayed(500);
        }

        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      });
    },

    restartComponent: function(event) {
      var self = this;
      var component = event.context;
      App.showConfirmationPopup(function(){
        batchUtils.restartHostComponents([component], Em.I18n.t('rollingrestart.context.selectedComponentOnSelectedHost').format(component.get('displayName')));
      });
    },
    /**
     * send command to server to install selected host component
     * @param event
     */
    addComponent: function (event) {
      var self = this;
      var component = event.context;
      var componentName = component.get('componentName').toUpperCase().toString();
      var subComponentNames = component.get('subComponentNames');
      var displayName = component.get('displayName');

      var securityEnabled = App.router.get('mainAdminSecurityController').getUpdatedSecurityStatus();

      if (componentName === 'ZOOKEEPER_SERVER') {
        App.showConfirmationPopup(function() {
          self.primary(component);
        }, Em.I18n.t('hosts.host.addComponent.addZooKeeper'));
      }
      else {
        if (securityEnabled && component.get('componentName') !== 'CLIENTS') {
          App.showConfirmationPopup(function() {
            self.primary(component);
          }, Em.I18n.t('hosts.host.addComponent.securityNote').format(componentName,self.get('content.hostName')));
        }
        else {
          var dn = displayName;
          if (subComponentNames !== null && subComponentNames.length > 0) {
            var dns = [];
            subComponentNames.forEach(function(scn){
              dns.push(App.format.role(scn));
            });
            dn += " ("+dns.join(", ")+")";
          }
          App.ModalPopup.show({
            primary: Em.I18n.t('hosts.host.addComponent.popup.confirm'),
            header: Em.I18n.t('popup.confirmation.commonHeader'),
            addComponentMsg: function() {
              return Em.I18n.t('hosts.host.addComponent.msg').format(dn);
            }.property(),
            bodyClass: Ember.View.extend({
              templateName: require('templates/main/host/details/addComponentPopup')
            }),
            restartNagiosMsg : Em.View.extend({
              template: Ember.Handlebars.compile(Em.I18n.t('hosts.host.addComponent.note').format(dn))
            }),
            onPrimary: function () {
              this.hide();
              if (component.get('componentName') === 'CLIENTS') {
                // Clients component has many sub-components which
                // need to be installed.
                var scs = component.get('subComponentNames');
                scs.forEach(function (sc, index) {
                  var c = Em.Object.create({
                    displayName: App.format.role(sc),
                    componentName: sc
                  });
                  self.primary(c, scs.length - index === 1);
                });
              } else {
                self.primary(component, true);
              }
            }
          });
        }
      }
    },

    primary: function(component, showPopup) {
      var self = this;
      var componentName = component.get('componentName').toUpperCase().toString();
      var displayName = component.get('displayName');

      self.sendCommandToServer('/hosts?Hosts/host_name=' + self.get('content.hostName'), {
          RequestInfo: {
            "context": Em.I18n.t('requestInfo.installHostComponent') + " " + displayName
          },
          Body: {
            host_components: [
              {
                HostRoles: {
                  component_name: componentName
                }
              }
            ]
          }
        },
        'POST',
        function (requestId) {

          console.log('Send request for ADDING NEW COMPONENT successfully');

          self.sendCommandToServer('/host_components?HostRoles/host_name=' + self.get('content.hostName') + '\&HostRoles/component_name=' + componentName + '\&HostRoles/state=INIT', {
              RequestInfo: {
                "context": Em.I18n.t('requestInfo.installNewHostComponent') + " " + displayName
              },
              Body: {
                HostRoles: {
                  state: 'INSTALLED'
                }
              }
            },
            'PUT',
            function (requestId) {
              if (!requestId) {
                return;
              }

              console.log('Send request for INSTALLING NEW COMPONENT successfully');

              if (App.testMode) {
                component.set('workStatus', App.HostComponentStatus.installing);
                setTimeout(function () {
                  component.set('workStatus', App.HostComponentStatus.stopped);
                }, App.testModeDelayForActions);
              } else {
                App.router.get('clusterController').loadUpdatedStatusDelayed(500);
              }

              // load data (if we need to show this background operations popup) from persist
              App.router.get('applicationController').dataLoading().done(function (initValue) {
                if (initValue) {
                  App.router.get('backgroundOperationsController').showPopup();
                }
                if (componentName === 'ZOOKEEPER_SERVER') {
                  self.set('zkRequestId', requestId);
                  self.addObserver('App.router.backgroundOperationsController.serviceTimestamp', self, self.checkZkConfigs);
                  self.checkZkConfigs();
                }
              });
            });
        });
    },
    /**
     * Load tags
     */
    checkZkConfigs: function() {
      var bg = App.router.get('backgroundOperationsController.services').findProperty('id', this.get('zkRequestId'));
      if (!bg) return;
      if (!bg.get('isRunning')) {
        this.loadConfigs();
      }
    },
    loadConfigs: function() {
      this.removeObserver('App.router.backgroundOperationsController.serviceTimestamp', this, this.checkZkConfigs);
      App.ajax.send({
        name: 'config.tags',
        sender: this,
        success: 'loadConfigsSuccessCallback'
      });
    },
    /**
     * Load needed configs
     * @param data
     */
    loadConfigsSuccessCallback: function(data) {
      var urlParams = [];
      if (App.get('isHaEnabled')) {
        urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
      }
      if (App.Service.find().someProperty('serviceName', 'HBASE')) {
        urlParams.push('(type=hbase-site&tag=' + data.Clusters.desired_configs['hbase-site'].tag + ')');
      }
      if (App.Service.find().someProperty('serviceName', 'HIVE')) {
        urlParams.push('(type=webhcat-site&tag=' + data.Clusters.desired_configs['webhcat-site'].tag + ')');
      }
      if (App.Service.find().someProperty('serviceName', 'STORM')) {
        urlParams.push('(type=storm-site&tag=' + data.Clusters.desired_configs['storm-site'].tag + ')');
      }
      if (!urlParams.length) {
        return;
      }
      App.ajax.send({
        name: 'reassign.load_configs',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'setNewZkConfigs'
      });
    },
    /**
     * Set new values for some configs (based on available ZooKeeper Servers)
     * @param data
     */
    setNewZkConfigs: function(data) {
      var configs = [];
      data.items.forEach(function (item) {
        configs[item.type] = item.properties;
      }, this);

      var zks = this.getZkServerHosts();
      var zks_with_port = '';
      zks.forEach(function(zk) {
        zks_with_port += zk + ':2181,';
      });
      zks_with_port = zks_with_port.slice(0,-1);

      if (App.get('isHaEnabled')) {
        configs['core-site']['ha.zookeeper.quorum'] = zks_with_port;
      }
      if (configs['hbase-site']) {
        configs['hbase-site']['hbase.zookeeper.quorum'] = zks.join(',');
      }
      if (configs['webhcat-site']) {
        configs['webhcat-site']['templeton.zookeeper.hosts'] = zks_with_port;
      }
      if (configs['storm-site']) {
        configs['storm-site']['storm.zookeeper.servers'] = JSON.stringify(zks).replace(/"/g, "'");
      }
      for (var site in configs) {
        if (!configs.hasOwnProperty(site)) continue;
        App.ajax.send({
          name: 'reassign.save_configs',
          sender: this,
          data: {
            siteName: site,
            properties: configs[site]
          }
        });
      }
    },

    /**
     * Is deleteHost action id fired
     */
    fromDeleteHost: false,

    getZkServerHosts: function() {
      var zks = App.HostComponent.find().filterProperty('componentName', 'ZOOKEEPER_SERVER').mapProperty('host.hostName');
      if (this.get('fromDeleteHost')) {
        this.set('fromDeleteHost', false);
        return zks.without(this.get('content.hostName'));
      }
      return zks;
    },

    /**
     * send command to server to install selected host component
     * @param event
     * @param context
     */
    installComponent: function (event, context) {
      var self = this;
      var component = event.context;
      var componentName = component.get('componentName').toUpperCase().toString();
      var displayName = component.get('displayName');

      App.ModalPopup.show({
        primary: Em.I18n.t('hosts.host.installComponent.popup.confirm'),
        header: Em.I18n.t('popup.confirmation.commonHeader'),
        installComponentMessage: function(){
          return Em.I18n.t('hosts.host.installComponent.msg').format(displayName);
        }.property(),
        restartNagiosMsg : Em.View.extend({
          template: Ember.Handlebars.compile(Em.I18n.t('hosts.host.addComponent.note').format(displayName))
        }),
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/host/details/installComponentPopup')
        }),
        onPrimary: function () {
          this.hide();
          self.sendCommandToServer('/hosts/' + self.get('content.hostName') + '/host_components/' + component.get('componentName').toUpperCase(), {
              RequestInfo: {
                "context": Em.I18n.t('requestInfo.installHostComponent') + " " + displayName
              },
              Body: {
                HostRoles: {
                  state: 'INSTALLED'
                }
              }
            },
            'PUT',
            function (requestId) {
              if (!requestId) {
                return;
              }

              console.log('Send request for REINSTALL COMPONENT successfully');

              if (App.testMode) {
                component.set('workStatus', App.HostComponentStatus.installing);
                setTimeout(function () {
                  component.set('workStatus', App.HostComponentStatus.stopped);
                }, App.testModeDelayForActions);
              } else {
                App.router.get('clusterController').loadUpdatedStatusDelayed(500);
              }

              // load data (if we need to show this background operations popup) from persist
              App.router.get('applicationController').dataLoading().done(function (initValue) {
                if (initValue) {
                  App.router.get('backgroundOperationsController').showPopup();
                }
              });
            });
        }
      });
    },

    /**
     * send command to server to run decommission on DATANODE, TASKTRACKER, NODEMANAGER, REGIONSERVER
     * @param component
     */
    decommission: function(component){
      var self = this;
      App.showConfirmationPopup(function(){
        var svcName = component.get('service.serviceName');
        var hostName = self.get('content.hostName');
        // HDFS service, decommission DataNode
        if (svcName === "HDFS") {
          self.doDecommission(hostName, svcName, "NAMENODE", "DATANODE");
        }
        // YARN service, decommission NodeManager
        if (svcName === "YARN") {
          self.doDecommission(hostName, svcName, "RESOURCEMANAGER", "NODEMANAGER");
        }
        // MAPREDUCE service, decommission TaskTracker
        if (svcName === "MAPREDUCE") {
          self.doDecommission(hostName, svcName, "JOBTRACKER", "TASKTRACKER");
        }
        // HBASE service, decommission RegionServer
        if (svcName === "HBASE") {
          self.doDecommissionRegionServer(hostName, svcName, "HBASE_MASTER", "HBASE_REGIONSERVER");
        }

        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      });
    },
    
    /**
     * send command to server to run recommission on DATANODE, TASKTRACKER, NODEMANAGER
     * @param component
     */
    recommission: function(component){
      var self = this;
      App.showConfirmationPopup(function(){
        var svcName = component.get('service.serviceName');
        var hostName = self.get('content.hostName');
        // HDFS service, Recommission datanode
        if (svcName === "HDFS") {
          self.doRecommissionAndStart(hostName, svcName, "NAMENODE", "DATANODE");
        }
        // YARN service, Recommission nodeManager
        if (svcName === "YARN") {
          self.doRecommissionAndStart(hostName, svcName, "RESOURCEMANAGER", "NODEMANAGER");
        }
        // MAPREDUCE service, Recommission taskTracker
        if (svcName === "MAPREDUCE") {
          self.doRecommissionAndRestart(hostName, svcName, "JOBTRACKER", "TASKTRACKER");
        }
        // HBASE service, Recommission RegionServer
        if (svcName === "HBASE") {
          self.doRecommissionAndStart(hostName, svcName, "HBASE_MASTER", "HBASE_REGIONSERVER");
        }

        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      });
    },

    /**
     * Performs Decommission (for DN, TT and NM)
     */
    doDecommission: function(hostName, serviceName, componentName, slaveType){
      var contextNameString = 'hosts.host.' + slaveType.toLowerCase() + '.decommission';
      var context = Em.I18n.t(contextNameString);
      App.ajax.send({
        name: 'host.host_component.decommission_slave',
        sender: this,
        data: {
          context: context,
          command: 'DECOMMISSION',
          hostName: hostName,
          serviceName: serviceName ,
          componentName: componentName,
          slaveType: slaveType
        },
        success: 'decommissionSuccessCallback',
        error: 'decommissionErrorCallback'
      });
    },

    /**
     * Performs Decommission (for RegionServer)
     */
    doDecommissionRegionServer: function(hostNames, serviceName, componentName, slaveType){

      App.ajax.send({
        name: 'host.host_component.recommission_and_restart',
        sender: this,
        data: {
          intervalTimeSeconds: 1,
          tolerateSize : 0,
          batches:[
            {
              "order_id" : 1,
              "type" : "POST",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  "context" : Em.I18n.t('hosts.host.regionserver.decommission.batch1'),
                  "command" : "DECOMMISSION",
                  "parameters" : {
                    "slave_type": slaveType,
                    "excluded_hosts": hostNames
                  }
                },
                "Requests/resource_filters": [{"service_name" : serviceName, "component_name" : componentName}]
              }
            },
            {
              "order_id": 2,
              "type": "PUT",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/host_components",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  context: Em.I18n.t('hosts.host.regionserver.decommission.batch2'),
                  query: 'HostRoles/component_name=' + slaveType + '&HostRoles/host_name.in(' + hostNames + ')&HostRoles/maintenance_state=OFF'
                },
                "Body": {
                  HostRoles: {
                    state: "INSTALLED"
                  }
                }
              }
            },
            {
              "order_id" : 3,
              "type" : "POST",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  "context" : Em.I18n.t('hosts.host.regionserver.decommission.batch3'),
                  "command" : "DECOMMISSION",
                  "service_name" : serviceName,
                  "component_name" : componentName,
                  "parameters" : {
                    "slave_type": slaveType,
                    "excluded_hosts": hostNames,
                    "mark_draining_only": "true"
                  }
                }
              }
            }
          ]
        },
        success: 'decommissionSuccessCallback',
        error: 'decommissionErrorCallback'
      });
    },

    decommissionErrorCallback: function (request, ajaxOptions, error) {
      console.log('ERROR: '+ error);
    },
    /**
     * Success ajax response for Recommission/Decommission slaves
     * @param data
     * @param ajaxOptions
     */
    decommissionSuccessCallback: function(data, ajaxOptions) {
      if(data && (data.Requests || data.resources[0].RequestSchedule) ) {
        if (!App.testMode) {
          App.router.get('clusterController').loadUpdatedStatusDelayed(500);
        }
        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      }
      else {
        console.log('cannot get request id from ', data);
      }
    },

    doRecommissionAndStart:  function(hostNames, serviceName, componentName, slaveType){
      var contextNameString_1 = 'hosts.host.' + slaveType.toLowerCase() + '.recommission';
      var context_1 = Em.I18n.t(contextNameString_1);
      var contextNameString_2 = 'requestInfo.startHostComponent.' + slaveType.toLowerCase();
      var startContext = Em.I18n.t(contextNameString_2);
      App.ajax.send({
        name: 'host.host_component.recommission_and_restart',
        sender: this,
        data: {
          intervalTimeSeconds: 1,
          tolerateSize : 1,
          batches:[
            {
              "order_id" : 1,
              "type" : "POST",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  "context" : context_1,
                  "command" : "DECOMMISSION",
                  "parameters" : {
                    "slave_type": slaveType,
                    "included_hosts": hostNames
                  }
                },
                "Requests/resource_filters": [{"service_name" : serviceName, "component_name" : componentName}]
              }
            },
            {
              "order_id": 2,
              "type": "PUT",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/host_components",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  context: startContext,
                  query: 'HostRoles/component_name=' + slaveType + '&HostRoles/host_name.in(' + hostNames + ')&HostRoles/maintenance_state=OFF'
                },
                "Body": {
                  HostRoles: {
                    state: "STARTED"
                  }
                }
              }
            }
          ]
        },
        success: 'decommissionSuccessCallback',
        error: 'decommissionErrorCallback'
      });
    },
    doRecommissionAndRestart:  function(hostNames, serviceName, componentName, slaveType){
      var contextNameString_1 = 'hosts.host.' + slaveType.toLowerCase() + '.recommission';
      var context_1 = Em.I18n.t(contextNameString_1);
      var contextNameString_2 = 'hosts.host.' + slaveType.toLowerCase() + '.restart';
      var context_2 = Em.I18n.t(contextNameString_2);
      App.ajax.send({
        name: 'host.host_component.recommission_and_restart',
        sender: this,
        data: {
          intervalTimeSeconds: 1,
          tolerateSize : 1,
          batches:[
            {
              "order_id" : 1,
              "type" : "POST",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  "context" : context_1,
                  "command" : "DECOMMISSION",
                  "parameters" : {
                    "slave_type": slaveType,
                    "included_hosts": hostNames
                  }
                },
                "Requests/resource_filters": [{"service_name" : serviceName, "component_name" : componentName}]
              }
            },
            {
              "order_id" : 2,
              "type" : "POST",
              "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
              "RequestBodyInfo" : {
                "RequestInfo" : {
                  "context" : context_2,
                  "command" : "RESTART",
                  "service_name" : serviceName,
                  "component_name" : slaveType,
                  "hosts" : hostNames
                }
              }
            }
          ]
        },
        success: 'decommissionSuccessCallback',
        error: 'decommissionErrorCallback'
      });
    },

    doAction: function(option) {
      switch (option.context.action) {
        case "deleteHost":
          this.validateAndDeleteHost();
          break;
        case "startAllComponents":
          if (!this.get('content.isNotHeartBeating')) this.doStartAllComponents();
          break;
        case "stopAllComponents":
          if (!this.get('content.isNotHeartBeating')) this.doStopAllComponents();
          break;
        case "restartAllComponents":
          if (!this.get('content.isNotHeartBeating')) this.doRestartAllComponents();
          break;
        case "onOffPassiveModeForHost":
          this.onOffPassiveModeForHost(option.context);
          break;
        default:
          break;
      }
    },

    onOffPassiveModeForHost: function(context) {
      var state = context.active ? 'ON' : 'OFF';
      var self = this;
      var message = context.label + ' for host';
      App.showConfirmationPopup(function() {
            self.hostPassiveModeRequest(state, message)
          },
          Em.I18n.t('hosts.passiveMode.popup').format(context.active ? 'On' : 'Off',this.get('content.hostName'))
      );
    },

    hostPassiveModeRequest: function(state,message) {
      App.ajax.send({
        name: 'bulk_request.hosts.passive_state',
        sender: this,
        data: {
          hostNames: this.get('content.hostName'),
          passive_state: state,
          requestInfo: message
        },
        success: 'updateHost'
      });
    },

    updateHost: function(data, opt, params) {
      this.set('content.passiveState', params.passive_state);
      App.router.get('clusterController').loadUpdatedStatus(function(){
        batchUtils.infoPassiveState(params.passive_state);
      });
   },

    doStartAllComponents: function() {
      var self = this;
      var components = this.get('serviceNonClientActiveComponents');
      var componentsLength = components == null ? 0 : components.get('length');
      if (componentsLength > 0) {
        App.showConfirmationPopup(function() {
          self.sendStartComponentCommand(components, Em.I18n.t('hosts.host.maintainance.startAllComponents.context'));
        });
      }
    },
    
    doStopAllComponents: function() {
      var self = this;
      var components = this.get('serviceNonClientActiveComponents');
      var componentsLength = components == null ? 0 : components.get('length');
      if (componentsLength > 0) {
        App.showConfirmationPopup(function() {
          self.sendStopComponentCommand(components, Em.I18n.t('hosts.host.maintainance.stopAllComponents.context'));
        });
      }
    },

    doRestartAllComponents: function() {
      var self = this;
      var components = this.get('serviceActiveComponents');
      var componentsLength = components == null ? 0 : components.get('length');
      if (componentsLength > 0) {
        App.showConfirmationPopup(function() {
          batchUtils.restartHostComponents(components, Em.I18n.t('rollingrestart.context.allOnSelectedHost').format(self.get('content.hostName')));
        });
      }
    },

    /**
     * Deletion of hosts not supported for this version
     */
    validateAndDeleteHost: function () {
      if (!App.supports.deleteHost) {
        return;
      }
      var stoppedStates = [App.HostComponentStatus.stopped,
                           App.HostComponentStatus.install_failed,
                           App.HostComponentStatus.upgrade_failed,
                           App.HostComponentStatus.init,
                           App.HostComponentStatus.unknown];
      var masterComponents = [];
      var runningComponents = [];
      var unknownComponents = [];
      var nonDeletableComponents = [];
      var lastComponents = [];
      var componentsOnHost = this.get('content.hostComponents');
      var allComponents = App.HostComponent.find();
      var zkServerInstalled = false;
      if (componentsOnHost && componentsOnHost.get('length') > 0) {
        componentsOnHost.forEach(function (cInstance) {
          if (cInstance.get('componentName') === 'ZOOKEEPER_SERVER') {
            zkServerInstalled = true;
          }
          if (allComponents.filterProperty('componentName', cInstance.get('componentName')).get('length') === 1) {
            lastComponents.push(cInstance.get('displayName'));
          }
          var workStatus = cInstance.get('workStatus');
          if (cInstance.get('isMaster') && !cInstance.get('isDeletable')) {
            masterComponents.push(cInstance.get('displayName'));
          }
          if (stoppedStates.indexOf(workStatus) < 0) {
            runningComponents.push(cInstance.get('displayName'));
          }
          if (!cInstance.get('isDeletable')) {
            nonDeletableComponents.push(cInstance.get('displayName'));
          }
          if (workStatus === App.HostComponentStatus.unknown) {
            unknownComponents.push(cInstance.get('displayName'));
          }
        });
      }
      if (masterComponents.length > 0) {
        this.raiseDeleteComponentsError(masterComponents, 'masterList');
        return;
      } else if (nonDeletableComponents.length > 0) {
        this.raiseDeleteComponentsError(nonDeletableComponents, 'nonDeletableList');
        return;
      } else if (runningComponents.length > 0) {
        this.raiseDeleteComponentsError(runningComponents, 'runningList');
        return;
      }
      if (zkServerInstalled) {
        var self = this;
        App.showConfirmationPopup(function() {
          self._doDeleteHost(unknownComponents, lastComponents);
        }, Em.I18n.t('hosts.host.addComponent.deleteHostWithZooKeeper'));
      }
      else {
        this._doDeleteHost(unknownComponents, lastComponents);
      }
    },
    
    raiseDeleteComponentsError: function (components, type) {
      App.ModalPopup.show({
        header: Em.I18n.t('hosts.cant.do.popup.title'),
        type: type,
        showBodyEnd: function() {
          return this.get('type') === 'runningList' || this.get('type') === 'masterList';
        }.property(),
        components: components,
        componentsStr: function() {
          return this.get('components').join(", ");
        }.property(),
        componentsBody: function() {
          return Em.I18n.t('hosts.cant.do.popup.'+type+'.body').format(this.get('components').length);
        }.property(),
        componentsBodyEnd: function() {
          if (this.get('showBodyEnd')) {
            return Em.I18n.t('hosts.cant.do.popup.'+type+'.body.end');
          }
          return '';
        }.property(),
        bodyClass: Em.View.extend({
          templateName: require('templates/main/host/details/raiseDeleteComponentErrorPopup')
        }),
        secondary: null
      })
    },

    /**
     * show confirmation popup to delete host
     */
    _doDeleteHost: function(unknownComponents,lastComponents) {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('hosts.delete.popup.title'),
        deletePopupBody: function() {
          return Em.I18n.t('hosts.delete.popup.body').format(self.get('content.publicHostName'));
        }.property(),
        lastComponent: function() {
           if (lastComponents && lastComponents.length) {
             this.set('isChecked', false);
             return true;
           } else {
             this.set('isChecked', true);
             return false;
           }
        }.property(),
        disablePrimary: function () {
          return !this.get('isChecked');
        }.property('isChecked'),
        isChecked: false,
        lastComponentError:  Em.View.extend({
          template: Ember.Handlebars.compile(Em.I18n.t('hosts.delete.popup.body.msg4').format(lastComponents))
        }),
        unknownComponents: function() {
          if (unknownComponents && unknownComponents.length) {
            return unknownComponents.join(", ");
          }
          return '';
        }.property(),
        bodyClass: Em.View.extend({
          templateName: require('templates/main/host/details/doDeleteHostPopup')
        }),
        onPrimary: function() {
          self.set('fromDeleteHost', true);
          var allComponents = self.get('content.hostComponents');
          var deleteError = null;
          allComponents.forEach(function(component){
            if (!deleteError) {
              deleteError = self._doDeleteHostComponent(component);
            }
          });
          if (!deleteError) {
            App.ajax.send({
              name: 'host.delete',
              sender: this,
              data: {
                hostName: self.get('content.hostName')
              },
              success: 'deleteHostSuccessCallback',
              error: 'deleteHostErrorCallback'
            });

          }
          else {
            this.hide();
            deleteError.xhr.responseText = "{\"message\": \"" + deleteError.xhr.statusText + "\"}";
            App.ajax.defaultErrorHandler(deleteError.xhr, deleteError.url, deleteError.method, deleteError.xhr.status);
          }
        },
        deleteHostSuccessCallback: function(data) {
          var dialogSelf = this;
          App.router.get('updateController').updateHost(function(){
            self.loadConfigs();
            dialogSelf.hide();
            App.router.transitionTo('hosts.index');
          });
        },
        deleteHostErrorCallback: function (xhr, textStatus, errorThrown, opt) {
          console.log('Error deleting host.');
          console.log(textStatus);
          console.log(errorThrown);
          xhr.responseText = "{\"message\": \"" + xhr.statusText + "\"}";
          self.loadConfigs();
          this.hide();
          App.ajax.defaultErrorHandler(xhr, opt.url, 'DELETE', xhr.status);
        }
      })
    },

    restartAllStaleConfigComponents: function() {
      var self = this;
      App.showConfirmationPopup(function () {
        var staleComponents = self.get('content.componentsWithStaleConfigs');
        batchUtils.restartHostComponents(staleComponents, Em.I18n.t('rollingrestart.context.allWithStaleConfigsOnSelectedHost').format(self.get('content.hostName')));
      });
    },

    /**
     * open Reassign Master Wizard with selected component
     * @param event
     */
    moveComponent: function (event) {
      App.showConfirmationPopup(function() {
        var component = event.context;
        var reassignMasterController = App.router.get('reassignMasterController');
        reassignMasterController.saveComponentToReassign(component);
        reassignMasterController.getSecurityStatus();
        reassignMasterController.setCurrentStep('1');
        App.router.transitionTo('reassign');
      });
    },

    /**
     * Restart clients host components to apply config changes
     */
    refreshConfigs: function(event) {
      var self = this;
      var components = event.context.filter(function(component) {
        return component.get('staleConfigs');
      });
      if (components.get('length') > 0) {
        App.showConfirmationPopup(function() {
          batchUtils.restartHostComponents(components, Em.I18n.t('rollingrestart.context.allClientsOnSelectedHost').format(self.get('content.hostName')));
        });
      }
    }

  });
  
});
window.require.register("controllers/main/jobs/hive_job_details_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var jobsUtils = require('utils/jobs');

  App.MainHiveJobDetailsController = Em.Controller.extend({
    name : 'mainHiveJobDetailsController',

    content : null,
    loaded : false,
    loadTimeout: null,
    job: null,
    sortingColumn: null,

    loadJobDetails : function() {
      var self = this;
      var timeout = this.get('loadTimeout');
      var yarnService = App.YARNService.find().objectAt(0);
      if (yarnService != null) {
        var content = this.get('content');
        if (content != null) {
          jobsUtils.refreshJobDetails(content, function() {
            self.set('content', self.get('job'));
            self.set('loaded', true);
          }, function(errorId) {
            switch (errorId) {
            case 'job.dag.noId':
              App.showAlertPopup(Em.I18n.t('jobs.hive.tez.dag.error.noDagId.title'), Em.I18n.t('jobs.hive.tez.dag.error.noDagId.message'));
              break;
            case 'job.dag.noname':
              App.showAlertPopup(Em.I18n.t('jobs.hive.tez.dag.error.noDag.title'), Em.I18n.t('jobs.hive.tez.dag.error.noDag.message'));
              break;
            case 'job.dag.id.noDag':
              App.showAlertPopup(Em.I18n.t('jobs.hive.tez.dag.error.noDagForId.title'), Em.I18n.t('jobs.hive.tez.dag.error.noDagForId.message'));
              break;
            case 'job.dag.id.loaderror':
            case 'job.dag.name.loaderror':
              break;
            default:
              break;
            }
            self.routeToJobs();
          });
        }
      } else {
        clearTimeout(timeout);
        timeout = setTimeout(function() {
          self.loadJobDetails();
        }, 300);
      }
    },

    /**
     * path to page visited before
     */
    referer : '',
    /**
     * open dashboard page
     */
    routeHome : function() {
      App.router.transitionTo('main.dashboard');
    },

    /**
     * open jobs page
     *
     * @param event
     */
    routeToJobs : function() {
      App.router.transitionTo('main.jobs.index');
    }
  });
  
});
window.require.register("controllers/main/jobs_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainJobsController = Em.Controller.extend({
  /*
   * https://github.com/emberjs/ember.js/issues/1221 prevents this controller
   * from being an Ember.ArrayController. Doing so will keep the UI flashing
   * whenever any of the 'sortProperties' or 'sortAscending' properties are set.
   * 
   *  To bypass this issue this controller will be a regular controller. Also,
   *  for memory-leak issues and sorting purposes, we are decoupling the backend
   *  model and the UI model. There will be simple Ember POJOs for the UI which
   *  will be periodically updated from backend Jobs model. 
   */
    
    name:'mainJobsController',

    /**
     * Unsorted ArrayProxy
     */
    content: App.HiveJob.find(),
    
    /**
     * Sorted ArrayProxy
     */
    sortedContent: [],

    contentAndSortObserver : function() {
      Ember.run.once(this, 'contentAndSortUpdater');
      if(App.HiveJob.find().findProperty('id', this.get('lastJobID'))) {
        this.set('hasNewJobs', false);
      }
    }.observes('content.length', 'content.@each.id', 'content.@each.startTime', 'content.@each.endTime', 'sortProperties', 'sortAscending'),
    
    contentAndSortUpdater: function() {
      this.set('sortingDone', false);
      var content = this.get('content');
      var sortedContent = content.toArray();
      var sortProperty = this.get('sortProperty');
      var sortAscending = this.get('sortAscending');
      sortedContent.sort(function(r1, r2) {
        var r1id = r1.get(sortProperty);
        var r2id = r2.get(sortProperty);
        if (r1id < r2id)
          return sortAscending ? -1 : 1;
        if (r1id > r2id)
          return sortAscending ? 1 : -1;
        return 0;
      });
      var sortedArray = this.get('sortedContent');
      var count = 0;
      sortedContent.forEach(function(sortedJob){
        if(sortedArray.length <= count) {
          sortedArray.pushObject(Ember.Object.create());
        }
        sortedArray[count].set('failed', sortedJob.get('failed'));
        sortedArray[count].set('hasTezDag', sortedJob.get('hasTezDag'));
        sortedArray[count].set('queryText', sortedJob.get('queryText'));
        sortedArray[count].set('name', sortedJob.get('name'));
        sortedArray[count].set('user', sortedJob.get('user'));
        sortedArray[count].set('id', sortedJob.get('id'));
        sortedArray[count].set('startTimeDisplay', sortedJob.get('startTimeDisplay'));
        sortedArray[count].set('endTimeDisplay', sortedJob.get('endTimeDisplay'));
        sortedArray[count].set('durationDisplay', sortedJob.get('durationDisplay'));
        count ++;
      });
      if(sortedArray.length > count) {
        for(var c = sortedArray.length-1; c >= count; c--){
          sortedArray.removeObject(sortedArray[c]);
        }
      }
      sortedContent.length = 0;
      this.set('sortingDone', true);
    },

    navIDs: {
      backIDs: [],
      nextID: ''
    },
    lastJobID: '',
    hasNewJobs: false,
    loaded : false,
    loading : false,
    resetPagination: false,
    loadJobsTimeout: null,
    loadTimeout: null,
    jobsUpdateInterval: 6000,
    jobsUpdate: null,
    sortingColumn: null,
    sortProperty: 'id',
    sortAscending: true,
    sortingDone: true,

    sortingColumnObserver: function () {
      if(this.get('sortingColumn')){
        this.set('sortProperty', this.get('sortingColumn').get('name'));
        this.set('sortAscending', this.get('sortingColumn').get('status') == "sorting_desc" ? false : true );
      }
    }.observes('sortingColumn.name','sortingColumn.status'),

    updateJobsByClick: function () {
      this.set('navIDs.backIDs', []);
      this.set('navIDs.nextID', '');
      this.get('filterObject').set('nextFromId', '');
      this.get('filterObject').set('backFromId', '');
      this.get('filterObject').set('fromTs', '');
      this.set('hasNewJobs', false);
      this.set('resetPagination', true);
      this.loadJobs();
    },

    updateJobs: function (controllerName, funcName) {
      clearInterval(this.get('jobsUpdate'));
      var self = this;
      var interval = setInterval(function () {
        App.router.get(controllerName)[funcName]();
      }, this.jobsUpdateInterval);
      this.set('jobsUpdate', interval);
    },

    totalOfJobs: 0,
    setTotalOfJobs: function () {
      if(this.get('totalOfJobs') < this.get('content.length')){
        this.set('totalOfJobs', this.get('content.length'));
      }
    }.observes('content.length'),

    filterObject: Ember.Object.create({
      id: "",
      isIdFilterApplied: false,
      jobsLimit: 10,
      user: "",
      windowStart: "",
      windowEnd: "",
      nextFromId: "",
      backFromId: "",
      fromTs: "",
      isAnyFilterApplied: false,

      onApplyIdFilter: function () {
        if(this.get('id') == ""){
          this.set('isIdFilterApplied', false);
        }else{
          this.set('isIdFilterApplied', true);
        }
      }.observes('id'),

      /**
       * Direct binding to startTime filter field
       */
      startTime: "",
      onStartTimeChange:function(){
        var time = "";
        var curTime = new Date().getTime();
        switch (this.get('startTime')) {
          case 'Past 1 hour':
            time = curTime - 3600000;
            break;
          case 'Past 1 Day':
            time = curTime - 86400000;
            break;
          case 'Past 2 Days':
            time = curTime - 172800000;
            break;
          case 'Past 7 Days':
            time = curTime - 604800000;
            break;
          case 'Past 14 Days':
            time = curTime - 1209600000;
            break;
          case 'Past 30 Days':
            time = curTime - 2592000000;
            break;
          case 'Custom':
            this.showCustomDatePopup();
            break;
          case 'Any':
            time = "";
            break;
        }
        if(this.get('startTime') != "Custom"){
          this.set("windowStart", time);
          this.set("windowEnd", "");
        }
      }.observes("startTime"),

      // Fields values from Select Custom Dates form
      customDateFormFields: Ember.Object.create({
        startDate: null,
        hoursForStart: null,
        minutesForStart: null,
        middayPeriodForStart: null,
        endDate: null,
        hoursForEnd: null,
        minutesForEnd: null,
        middayPeriodForEnd: null
      }),

      errors: Ember.Object.create({
        isStartDateError: false,
        isEndDateError: false
      }),

      errorMessages: Ember.Object.create({
        startDate: '',
        endDate: ''
      }),

      showCustomDatePopup: function () {
        var self = this;
        var windowEnd = "";
        var windowStart = "";
        App.ModalPopup.show({
          header: Em.I18n.t('jobs.table.custom.date.header'),
          onPrimary: function () {
            self.validate();
            if(self.get('errors.isStartDateError') || self.get('errors.isEndDateError')){
              return false;
            }

            var windowStart = self.createCustomStartDate();
            var windowEnd = self.createCustomEndDate();

            self.set("windowStart", windowStart.getTime());
            self.set("windowEnd", windowEnd.getTime());
            this.hide();
          },
          onSecondary: function () {
            self.set('startTime','Any');
            this.hide();
          },
          bodyClass: App.JobsCustomDatesSelectView.extend({
            controller: self
          })
        });
      },

      createCustomStartDate : function () {
        var startDate = this.get('customDateFormFields.startDate');
        var hoursForStart = this.get('customDateFormFields.hoursForStart');
        var minutesForStart = this.get('customDateFormFields.minutesForStart');
        var middayPeriodForStart = this.get('customDateFormFields.middayPeriodForStart');
        if (startDate && hoursForStart && minutesForStart && middayPeriodForStart) {
          return new Date(startDate + ' ' + hoursForStart + ':' + minutesForStart + ' ' + middayPeriodForStart);
        }
        return null;
      },

      createCustomEndDate : function () {
        var endDate = this.get('customDateFormFields.endDate');
        var hoursForEnd = this.get('customDateFormFields.hoursForEnd');
        var minutesForEnd = this.get('customDateFormFields.minutesForEnd');
        var middayPeriodForEnd = this.get('customDateFormFields.middayPeriodForEnd');
        if (endDate && hoursForEnd && minutesForEnd && middayPeriodForEnd) {
          return new Date(endDate + ' ' + hoursForEnd + ':' + minutesForEnd + ' ' + middayPeriodForEnd);
        }
        return null;
      },

      clearErrors: function () {
        var errorMessages = this.get('errorMessages');
        Em.keys(errorMessages).forEach(function (key) {
          errorMessages.set(key, '');
        }, this);
        var errors = this.get('errors');
        Em.keys(errors).forEach(function (key) {
          errors.set(key, false);
        }, this);
      },

      // Validation for every field in customDateFormFields
      validate: function () {
        var formFields = this.get('customDateFormFields');
        var errors = this.get('errors');
        var errorMessages = this.get('errorMessages');
        this.clearErrors();
        // Check if feild is empty
        Em.keys(errorMessages).forEach(function (key) {
          if (!formFields.get(key)) {
            errors.set('is' + key.capitalize() + 'Error', true);
            errorMessages.set(key, Em.I18n.t('jobs.customDateFilter.error.required'));
          }
        }, this);
        // Check that endDate is after startDate
        var startDate = this.createCustomStartDate();
        var endDate = this.createCustomEndDate();
        if (startDate && endDate && (startDate > endDate)) {
          errors.set('isEndDateError', true);
          errorMessages.set('endDate', Em.I18n.t('jobs.customDateFilter.error.date.order'));
        }
      },

      /**
       * Create link for server request
       * @return {String}
       */
      createJobsFiltersLink: function() {
        var link = "?fields=events,primaryfilters,otherinfo&secondaryFilter=tez:true";
        var numberOfAppliedFilters = 0;

        if(this.get("id") !== "") {
          link = "/" + this.get("id") + link;
          numberOfAppliedFilters++;
        }

        link += "&limit=" + (parseInt(this.get("jobsLimit")) + 1);

        if(this.get("user") !== ""){
          link += "&primaryFilter=user:" + this.get("user");
          numberOfAppliedFilters++;
        }
        if(this.get("backFromId") != ""){
          link += "&fromId=" + this.get("backFromId");
        }
        if(this.get("nextFromId") != ""){
          link += "&fromId=" + this.get("nextFromId");
        }
        if(this.get("fromTs") != ""){
          link += "&fromTs=" + this.get("fromTs");
        }
        if(this.get("startTime") !== "" && this.get("startTime") !== "Any"){
          link += this.get("windowStart") !== "" ? ("&windowStart=" + this.get("windowStart")) : "";
          link += this.get("windowEnd") !== "" ? ("&windowEnd=" + this.get("windowEnd")) : "";
          numberOfAppliedFilters++;
        }

        if(numberOfAppliedFilters > 0){
          this.set('isAnyFilterApplied', true);
        }else{
          this.set('isAnyFilterApplied', false);
        }

        return link;
      }
    }),

    columnsName: Ember.ArrayController.create({
      content: [
        { name: Em.I18n.t('jobs.column.id'), index: 0 },
        { name: Em.I18n.t('jobs.column.user'), index: 1 },
        { name: Em.I18n.t('jobs.column.start.time'), index: 2 },
        { name: Em.I18n.t('jobs.column.end.time'), index: 3 },
        { name: Em.I18n.t('jobs.column.duration'), index: 4 }
      ]
    }),

    lastIDSuccessCallback: function(data, jqXHR, textStatus) {
      var lastReceivedID = data.entities[0].entity;
      if(this.get('lastJobID') == '') {
        this.set('lastJobID', lastReceivedID);
      } else if (this.get('lastJobID') !== lastReceivedID) {
        this.set('lastJobID', lastReceivedID);
        if(!App.HiveJob.find().findProperty('id', lastReceivedID)) {
          this.set('hasNewJobs', true);
        }
      }
    },

    lastIDErrorCallback: function(data, jqXHR, textStatus) {
      console.debug(jqXHR);
    },

    loadJobs : function() {
      var self = this;
      var timeout = this.get('loadTimeout');
      var yarnService = App.YARNService.find().objectAt(0);
      if (yarnService != null) {
        this.set('loading', true);
        var historyServerHostName = yarnService.get('appTimelineServerNode.hostName');
        var filtersLink = this.get('filterObject').createJobsFiltersLink();
        var hiveQueriesUrl = App.testMode ? "/data/jobs/hive-queries.json" : "/proxy?url=http://" + historyServerHostName
          + ":" + yarnService.get('ahsWebPort') + "/ws/v1/timeline/HIVE_QUERY_ID" + filtersLink;
        App.ajax.send({
          name: 'jobs.lastID',
          sender: self,
          data: {
            historyServerHostName: historyServerHostName,
            ahsWebPort: yarnService.get('ahsWebPort')
          },
          success: 'lastIDSuccessCallback',
          error : 'lastIDErrorCallback'
        }),
        App.HttpClient.get(hiveQueriesUrl, App.hiveJobsMapper, {
          complete : function(data, jqXHR, textStatus) {
            self.set('loading', false);
            if(self.get('loaded') == false || self.get('resetPagination') == true){
              self.initializePagination();
              self.set('resetPagination', false);
            }
            self.set('loaded', true);
          }
        }, function (jqXHR, textStatus) {
          App.hiveJobsMapper.map({entities : []});
        });
      }else{
        clearTimeout(timeout);
        timeout = setTimeout(function(){
          self.loadJobs();
        }, 300);
      }
    },

    initializePagination: function() {
      var back_link_IDs = this.get('navIDs.backIDs.[]');
      if(!back_link_IDs.contains(this.get('lastJobID'))) {
        back_link_IDs.push(this.get('lastJobID'));
      }
      this.set('filterObject.backFromId', this.get('lastJobID'));
      this.get('filterObject').set('fromTs', App.dateTime());
    },

    navigateNext: function() {
      this.set("filterObject.backFromId", '');
      var back_link_IDs = this.get('navIDs.backIDs.[]');
      var lastBackID = this.get('navIDs.nextID');
      if(!back_link_IDs.contains(lastBackID)) {
        back_link_IDs.push(lastBackID);
      }
      this.set('navIDs.backIDs.[]', back_link_IDs);
      this.set("filterObject.nextFromId", this.get('navIDs.nextID'));
      this.set('navIDs.nextID', '');
      this.loadJobs();
    },

    navigateBack: function() {
      this.set("filterObject.nextFromId", '');
      var back_link_IDs = this.get('navIDs.backIDs.[]');
      back_link_IDs.pop();
      var lastBackID = back_link_IDs[back_link_IDs.length - 1]
      this.set('navIDs.backIDs.[]', back_link_IDs);
      this.set("filterObject.backFromId", lastBackID);
      this.loadJobs();
    },

    refreshLoadedJobs : function() {
      var timeout = this.get('loadJobsTimeout');
      var self = this;

      clearTimeout(timeout);
      timeout = setTimeout(function(){
        self.loadJobs();
      }, 300);

      this.set('loadJobsTimeout', timeout);
    }.observes(
        'filterObject.id',
        'filterObject.jobsLimit',
        'filterObject.user',
        'filterObject.windowStart',
        'filterObject.windowEnd'
    )
  })
  
});
window.require.register("controllers/main/mirroring/datasets_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainDatasetsController = Em.ArrayController.extend({
    name: 'mainDatasetsController'
  });
  
});
window.require.register("controllers/main/mirroring/edit_dataset_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  App.MainMirroringEditDataSetController = Ember.Controller.extend({
    name: 'mainMirroringEditDataSetController',

    isEdit: false,

    datasetIdToEdit: null,

    // Fields values from Edit DataSet form
    formFields: Ember.Object.create({
      datasetName: null,
      datasetTargetClusterName: null,
      datasetSourceDir: null,
      datasetTargetDir: null,
      datasetStartDate: null,
      hoursForStart: null,
      minutesForStart: null,
      middayPeriodForStart: null,
      datasetEndDate: null,
      hoursForEnd: null,
      minutesForEnd: null,
      middayPeriodForEnd: null,
      datasetFrequency: null,
      repeatOptionSelected: null
    }),

    // Messages for errors occurred during Edit DataSet form validation
    errorMessages: Ember.Object.create({
      name: '',
      sourceDir: '',
      targetDir: '',
      startDate: '',
      endDate: '',
      frequency: '',
      targetClusterName: ''
    }),

    errors: Ember.Object.create({
      isNameError: false,
      isSourceDirError: false,
      isTargetDirError: false,
      isStartDateError: false,
      isEndDateError: false,
      isFrequencyError: false,
      isTargetClusterNameError: false
    }),

    clearStep: function () {
      var formFields = this.get('formFields');
      Em.keys(formFields).forEach(function (key) {
        formFields.set(key, null);
      }, this);
      this.clearErrors();
    },

    clearErrors: function () {
      var errorMessages = this.get('errorMessages');
      Em.keys(errorMessages).forEach(function (key) {
        errorMessages.set(key, '');
      }, this);
      var errors = this.get('errors');
      Em.keys(errors).forEach(function (key) {
        errors.set(key, false);
      }, this);
    },

    showAddPopup: function () {
      this.showPopup(Em.I18n.t('mirroring.dataset.newDataset'));
      this.set('isEdit', false);
    },

    showEditPopup: function (dataset) {
      this.set('datasetIdToEdit', dataset.get('id'));
      this.showPopup(Em.I18n.t('mirroring.dataset.editDataset'));
      this.set('isEdit', true);
    },

    showPopup: function (header) {
      var self = this;
      var popup = App.ModalPopup.show({
        classNames: ['sixty-percent-width-modal'],
        header: header,
        primary: Em.I18n.t('mirroring.dataset.save'),
        secondary: Em.I18n.t('common.cancel'),
        showCloseButton: false,
        isSaving: false,
        saveDisabled: function () {
          return self.get('saveDisabled');
        }.property('App.router.' + self.get('name') + '.saveDisabled'),
        disablePrimary: function () {
          return this.get('saveDisabled') || this.get('isSaving');
        }.property('saveDisabled', 'isSaving'),
        onPrimary: function () {
          // Apply form validation for first click
          if (!this.get('primaryWasClicked')) {
            this.toggleProperty('primaryWasClicked');
            self.applyValidation();
            if (this.get('saveDisabled')) {
              return false;
            }
          }
          self.save();
          App.router.transitionTo('main.mirroring.index');
        },
        primaryWasClicked: false,
        onSecondary: function () {
          this.hide();
          App.router.send('gotoShowJobs');
        },
        bodyClass: App.MainMirroringEditDataSetView.extend({
          controller: self
        })
      });
      this.set('popup', popup);
    },

    // Set observer to call validate method if any property from formFields will change
    applyValidation: function () {
      Em.keys(this.get('formFields')).forEach(function (key) {
        this.addObserver('formFields.' + key, this, 'validate');
      }, this);
      this.validate();
    },

    // Return date object calculated from appropriate fields
    scheduleStartDate: function () {
      var startDate = this.get('formFields.datasetStartDate');
      var hoursForStart = this.get('formFields.hoursForStart');
      var minutesForStart = this.get('formFields.minutesForStart');
      var middayPeriodForStart = this.get('formFields.middayPeriodForStart');
      if (startDate && hoursForStart && minutesForStart && middayPeriodForStart) {
        return new Date(startDate + ' ' + hoursForStart + ':' + minutesForStart + ' ' + middayPeriodForStart);
      }
      return null;
    }.property('formFields.datasetStartDate', 'formFields.hoursForStart', 'formFields.minutesForStart', 'formFields.middayPeriodForStart'),

    // Return date object calculated from appropriate fields
    scheduleEndDate: function () {
      var endDate = this.get('formFields.datasetEndDate');
      var hoursForEnd = this.get('formFields.hoursForEnd');
      var minutesForEnd = this.get('formFields.minutesForEnd');
      var middayPeriodForEnd = this.get('formFields.middayPeriodForEnd');
      if (endDate && hoursForEnd && minutesForEnd && middayPeriodForEnd) {
        return new Date(endDate + ' ' + hoursForEnd + ':' + minutesForEnd + ' ' + middayPeriodForEnd);
      }
      return null;
    }.property('formFields.datasetEndDate', 'formFields.hoursForEnd', 'formFields.minutesForEnd', 'formFields.middayPeriodForEnd'),


    // Validation for every field in Edit DataSet form
    validate: function () {
      var formFields = this.get('formFields');
      var errors = this.get('errors');
      var errorMessages = this.get('errorMessages');
      this.clearErrors();
      // Check if feild is empty
      Em.keys(errorMessages).forEach(function (key) {
        if (!formFields.get('dataset' + key.capitalize())) {
          errors.set('is' + key.capitalize() + 'Error', true);
          errorMessages.set(key, Em.I18n.t('mirroring.required.error'));
        }
      }, this);
      // Check that endDate is after startDate
      var scheduleStartDate = this.get('scheduleStartDate');
      var scheduleEndDate = this.get('scheduleEndDate');
      if (scheduleStartDate && scheduleEndDate && (scheduleStartDate > scheduleEndDate)) {
        errors.set('isEndDateError', true);
        errorMessages.set('endDate', Em.I18n.t('mirroring.dateOrder.error'));
      }
      // Check that startDate is after current date
      if (!this.get('isEdit') && new Date(App.dateTime()) > scheduleStartDate) {
        errors.set('isStartDateError', true);
        errorMessages.set('startDate', Em.I18n.t('mirroring.startDate.error'));
      }
      // Check that repeat field value consists only from digits
      if (isNaN(this.get('formFields.datasetFrequency'))) {
        errors.set('isFrequencyError', true);
        errorMessages.set('frequency', Em.I18n.t('mirroring.required.invalidNumberError'));
      }
    },

    // Add '0' for numbers less than 10
    addZero: function (number) {
      return ('0' + number).slice(-2);
    },

    // Convert date to TZ format
    toTZFormat: function (date) {
      return date.toISOString().replace(/\:\d{2}\.\d{3}/,'');
    },

    // Converts hours value from 24-hours format to AM/PM format
    toAMPMHours: function (hours) {
      var result = hours % 12;
      result = result ? result : 12;
      return this.addZero(result);
    },

    save: function () {
      this.set('popup.isSaving', true);
      var datasetNamePrefix = App.mirroringDatasetNamePrefix;
      var datasetName = this.get('formFields.datasetName');
      var prefixedDatasetName = datasetNamePrefix + datasetName;
      var sourceCluster = App.get('clusterName');
      var targetCluster = this.get('formFields.datasetTargetClusterName');
      var sourceDir = this.get('formFields.datasetSourceDir');
      var targetDir = this.get('formFields.datasetTargetDir');
      var datasetFrequency = this.get('formFields.datasetFrequency');
      var repeatOptionSelected = this.get('formFields.repeatOptionSelected');
      var startDate = this.get('scheduleStartDate');
      var endDate = this.get('scheduleEndDate');
      var scheduleStartDateFormatted = this.toTZFormat(startDate);
      var scheduleEndDateFormatted = this.toTZFormat(endDate);

      // Compose XML data, that will be sended to server
      var dataToSend = '<?xml version="1.0"?><feed description="" name="' + prefixedDatasetName + '" xmlns="uri:falcon:feed:0.1"><frequency>' + repeatOptionSelected + '(' + datasetFrequency + ')' +
          '</frequency><clusters><cluster name="' + sourceCluster + '" type="source"><validity start="' + scheduleStartDateFormatted + '" end="' + scheduleEndDateFormatted +
          '"/><retention limit="days(7)" action="delete"/></cluster><cluster name="' + targetCluster + '" type="target"><validity start="' + scheduleStartDateFormatted + '" end="' + scheduleEndDateFormatted +
          '"/><retention limit="months(1)" action="delete"/><locations><location type="data" path="' + targetDir + '" /></locations></cluster></clusters><locations><location type="data" path="' +
          sourceDir + '" /></locations><ACL owner="hue" group="users" permission="0755" /><schema location="/none" provider="none"/></feed>';
      if (this.get('isEdit')) {
        App.ajax.send({
          name: 'mirroring.update_entity',
          sender: this,
          data: {
            name: prefixedDatasetName,
            type: 'feed',
            entity: dataToSend,
            falconServer: App.get('falconServerURL')
          },
          success: 'onSaveSuccess',
          error: 'onSaveError'
        });
      } else {
        // Send request to server to create dataset
        App.ajax.send({
          name: 'mirroring.create_new_dataset',
          sender: this,
          data: {
            dataset: dataToSend,
            falconServer: App.get('falconServerURL')
          },
          success: 'onSaveSuccess',
          error: 'onSaveError'
        });
      }
    },

    onSaveSuccess: function () {
      this.set('popup.isSaving', false);
      this.get('popup').hide();
      App.router.get('mainMirroringController').loadData();
    },

    onSaveError: function (response) {
      this.set('popup.isSaving', false);
      if (response && response.responseText) {
        var errorMessage = /(?:\<message\>)((.|\n)+)(?:\<\/message\>)/.exec(response.responseText);
        if (errorMessage.length > 1) {
          App.showAlertPopup(Em.I18n.t('common.error'), Em.I18n.t('mirroring.manageClusters.error') + ': ' + errorMessage[1]);
        }
      }
    },

    saveDisabled: function () {
      var errors = this.get('errors');
      return errors.get('isNameError') || errors.get('isSourceDirError') || errors.get('isTargetDirError') || errors.get('isStartDateError') || errors.get('isEndDateError') || errors.get('isFrequencyError') || errors.get('isTargetClusterNameError');
    }.property('errors.isNameError', 'errors.isSourceDirError', 'errors.isTargetDirError', 'errors.isStartDateError', 'errors.isEndDateError', 'errors.isFrequencyError', 'errors.isTargetClusterNameError')
  });
  
});
window.require.register("controllers/main/mirroring/jobs_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainDatasetJobsController = Em.Controller.extend({
    name: 'mainDatasetJobsController',

    suspend: function () {
      App.ajax.send({
        name: 'mirroring.suspend_entity',
        sender: this,
        data: {
          name: this.get('content.prefixedName'),
          type: 'feed',
          falconServer: App.get('falconServerURL')
        },
        success: 'onSuspendSuccess',
        error: 'onError'
      });
    },

    onSuspendSuccess: function() {
      this.set('content.status', 'SUSPENDED');
      this.get('content.datasetJobs').filterProperty('status', 'RUNNING').setEach('status', 'SUSPENDED');
    },

    resume: function () {
      App.ajax.send({
        name: 'mirroring.resume_entity',
        sender: this,
        data: {
          name: this.get('content.prefixedName'),
          type: 'feed',
          falconServer: App.get('falconServerURL')
        },
        success: 'onResumeSuccess',
        error: 'onError'
      });
    },

    onResumeSuccess: function() {
      this.set('content.status', 'RUNNING');
      this.get('content.datasetJobs').filterProperty('status', 'SUSPENDED').setEach('status', 'RUNNING');
    },

    schedule: function () {
      App.ajax.send({
        name: 'mirroring.schedule_entity',
        sender: this,
        data: {
          name: this.get('content.prefixedName'),
          type: 'feed',
          falconServer: App.get('falconServerURL')
        },
        success: 'onScheduleSuccess',
        error: 'onError'
      });
    },

    onScheduleSuccess: function() {
      this.set('content.status', 'RUNNING');
    },


    delete: function () {
      var self = this;
      App.showConfirmationPopup(function () {
        App.ajax.send({
          name: 'mirroring.delete_entity',
          sender: self,
          data: {
            name: self.get('content.prefixedName'),
            type: 'feed',
            falconServer: App.get('falconServerURL')
          },
          success: 'onDeleteSuccess',
          error: 'onError'
        });
      });
    },

    onDeleteSuccess: function() {
      var dataset = this.get('content');
      dataset.deleteRecord();
      App.store.commit();
      dataset.get('stateManager').transitionTo('loading');
      this.set('content', null);
      App.router.get('mainMirroringController').set('selectedDataset', null);
      App.router.send('gotoShowJobs');
    },

    suspendInstance: function (event) {
      App.ajax.send({
        name: 'mirroring.suspend_instance',
        sender: this,
        data: {
          feed: this.get('content.prefixedName'),
          name: event.context.get('name'),
          job: event.context,
          falconServer: App.get('falconServerURL')
        },
        success: 'onSuspendInstanceSuccess',
        error: 'onError'
      });
    },

    onSuspendInstanceSuccess: function () {
      this.get('content.datasetJobs').filterProperty('name', arguments[2].name).setEach('status', 'SUSPENDED');
    },

    resumeInstance: function (event) {
      App.ajax.send({
        name: 'mirroring.resume_instance',
        sender: this,
        data: {
          feed: this.get('content.prefixedName'),
          name: event.context.get('name'),
          job: event.context,
          falconServer: App.get('falconServerURL')
        },
        success: 'onResumeInstanceSuccess',
        error: 'onError'
      });
    },

    onResumeInstanceSuccess: function () {
      this.get('content.datasetJobs').filterProperty('name', arguments[2].name).setEach('status', 'RUNNING');
    },

    killInstance: function (event) {
      App.ajax.send({
        name: 'mirroring.kill_instance',
        sender: this,
        data: {
          feed: this.get('content.prefixedName'),
          name: event.context.get('name'),
          job: event.context,
          falconServer: App.get('falconServerURL')
        },
        success: 'onKillInstanceSuccess',
        error: 'onError'
      });
    },

    onKillInstanceSuccess: function () {
      this.get('content.datasetJobs').filterProperty('name', arguments[2].name).setEach('status', 'KILLED');
    },

    onError: function () {
      App.showAlertPopup(Em.I18n.t('common.error'), arguments[2]);
    },

    downloadEntity: function () {
      var xml = this.formatDatasetXML(this.get('content'));
      if ($.browser.msie && $.browser.version < 10) {
        this.openInfoInNewTab(xml);
      } else {
        try {
          var blob = new Blob([xml], {type: 'text/xml;charset=utf-8;'});
          saveAs(blob, Em.I18n.t('mirroring.dataset.entity') + '.xml');
        } catch (e) {
          this.openInfoInNewTab(xml);
        }
      }
    },

    openInfoInNewTab: function (xml) {
      var newWindow = window.open('');
      var newDocument = newWindow.document;
      newDocument.write('<pre>' + App.config.escapeXMLCharacters(xml, true) + '</pre>');
      newWindow.focus();
    },

    formatDatasetXML: function (dataset) {
      return '<?xml version="1.0"?>\n' + '<feed description="" name="' + dataset.get('name') + '" xmlns="uri:falcon:feed:0.1">\n' +
          '<frequency>' + dataset.get('frequencyUnit') + '(' + dataset.get('frequency') + ')' + '</frequency>\n' +
          '<clusters>\n<cluster name="' + dataset.get('sourceClusterName') + '" type="source">\n' +
          '<validity start="' + dataset.get('scheduleStartDate') + '" end="' + dataset.get('scheduleEndDate') + '"/>\n' +
          '<retention limit="days(7)" action="delete"/>\n</cluster>\n<cluster name="' + dataset.get('targetClusterName') +
          '" type="target">\n<validity start="' + dataset.get('scheduleStartDate') + '" end="' + dataset.get('scheduleEndDate') + '"/>\n' +
          '<retention limit="months(1)" action="delete"/>\n<locations>\n<location type="data" path="' + dataset.get('targetDir') + '" />\n' +
          '</locations>\n</cluster>\n</clusters>\n<locations>\n<location type="data" path="' + dataset.get('sourceDir') + '" />\n' +
          '</locations>\n<ACL owner="hue" group="users" permission="0755" />\n<schema location="/none" provider="none"/>\n</feed>';
    }
  });
  
});
window.require.register("controllers/main/mirroring/manage_clusters_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainMirroringManageClustersController = Em.ArrayController.extend({
    name: 'mainMirroringManageClustersController',

    // link to popup object
    popup: null,

    executeTooltip: Em.I18n.t('mirroring.manageClusters.executeTooltip'),
    readonlyTooltip: Em.I18n.t('mirroring.manageClusters.readonlyTooltip'),
    workflowTooltip: Em.I18n.t('mirroring.manageClusters.workflowTooltip'),
    writeTooltip: Em.I18n.t('mirroring.manageClusters.writeTooltip'),

    clusters: [],

    newCluster: null,

    isLoaded: function () {
      return App.router.get('mainMirroringController.isLoaded');
    }.property('App.router.mainMirroringController.isLoaded'),

    onLoad: function () {
      if (this.get('isLoaded')) {
        var clusters = [];
        App.TargetCluster.find().forEach(function (cluster) {
          var newCluster = {
            name: cluster.get('name'),
            execute: cluster.get('execute'),
            workflow: cluster.get('workflow'),
            write: cluster.get('write'),
            readonly: cluster.get('readonly'),
            staging: cluster.get('staging'),
            working: cluster.get('working'),
            temp: cluster.get('temp')
          };
          // Source cluster should be shown on top
          if (cluster.get('name') === App.get('clusterName')) {
            clusters.unshift(Ember.Object.create(newCluster));
          } else {
            clusters.push(Ember.Object.create(newCluster));
          }
        }, this);
        this.set('clusters', clusters);
      }
    }.observes('isLoaded'),

    selectedCluster: null,

    // Disable input fields for already created clusters
    isEditDisabled: function () {
      return !this.get('clustersToCreate').mapProperty('name').contains(this.get('selectedCluster.name'));
    }.property('selectedCluster.name', 'clustersToCreate.@each.name'),

    addCluster: function () {
      var self = this;
      var newClusterPopup = App.ModalPopup.show({
        header: Em.I18n.t('mirroring.manageClusters.create.cluster.popup'),
        bodyClass: Em.View.extend({
          controller: self,
          templateName: require('templates/main/mirroring/create_new_cluster')
        }),
        classNames: ['create-target-cluster-popup'],
        primary: Em.I18n.t('common.save'),
        secondary: Em.I18n.t('common.cancel'),
        onPrimary: function () {
          this.set('disablePrimary', true);
          self.createNewCluster();
        },
        willInsertElement: function () {
          var clusterName = App.get('clusterName');
          var newCluster = Ember.Object.create({
            name: '',
            execute: '',
            workflow: '',
            write: '',
            readonly: '',
            staging: '/apps/falcon/<cluster-name>/staging',
            working: '/apps/falcon/<cluster-name>/working',
            temp: '/tmp'
          });
          self.set('newCluster', newCluster);
        },
        didInsertElement: function () {
          this._super();
          this.fitHeight();
        }
      });
      this.set('newClusterPopup', newClusterPopup);
    },

    removeCluster: function () {
      var self = this;
      var selectedClusterName = self.get('selectedCluster.name');
      App.showConfirmationPopup(function () {
        App.ajax.send({
          name: 'mirroring.delete_entity',
          sender: self,
          data: {
            name: selectedClusterName,
            type: 'cluster',
            falconServer: App.get('falconServerURL')
          },
          success: 'onRemoveClusterSuccess',
          error: 'onError'
        });
      }, Em.I18n.t('mirroring.manageClusters.remove.confirmation').format(selectedClusterName));
    },

    onRemoveClusterSuccess: function () {
      this.set('clusters', this.get('clusters').without(this.get('selectedCluster')));
    },

    onError: function (response) {
      if (response && response.responseText) {
        var errorMessage = /(?:\<message\>)((.|\n)+)(?:\<\/message\>)/.exec(response.responseText);
        if (errorMessage.length > 1) {
          App.showAlertPopup(Em.I18n.t('common.error'), Em.I18n.t('mirroring.manageClusters.error') + ': ' + errorMessage[1]);
        }
      }
    },

    createNewCluster: function () {
      App.ajax.send({
        name: 'mirroring.submit_entity',
        sender: this,
        data: {
          type: 'cluster',
          entity: this.formatClusterXML(this.get('newCluster')),
          falconServer: App.get('falconServerURL')
        },
        success: 'onCreateClusterSuccess',
        error: 'onCreateClusterError'
      });
    },

    onCreateClusterSuccess: function () {
      this.get('clusters').pushObject(this.get('newCluster'));
      this.get('newClusterPopup').hide();
    },

    onCreateClusterError: function (response) {
      this.set('newClusterPopup.disablePrimary', false);
      this.onError(response);
    },

    /**
     * Return XML-formatted string made from cluster object
     * @param {Object} cluster - object with cluster data
     * @return {String}
     */
    formatClusterXML: function (cluster) {
      return '<?xml version="1.0"?><cluster colo="local" description="" name="' + cluster.get('name') +
          '" xmlns="uri:falcon:cluster:0.1"><interfaces><interface type="readonly" endpoint="' + cluster.get('readonly') +
          '" version="2.2.0" /><interface type="execute" endpoint="' + cluster.get('execute') +
          '" version="2.2.0" /><interface type="workflow" endpoint="' + cluster.get('workflow') +
          '" version="4.0.0" />' + '<interface type="messaging" endpoint="tcp://' + App.get('falconServerURL') + ':61616?daemon=true" version="5.1.6" />' +
          '<interface type="write" endpoint="' + cluster.get('write') + '" version="2.2.0" />' +
          '</interfaces><locations><location name="staging" path="' + cluster.get('staging') +
          '" /><location name="temp" path="' + cluster.get('temp') +
          '" /><location name="working" path="' + cluster.get('working') +
          '" /></locations></cluster>';
    }
  });
  
});
window.require.register("controllers/main/mirroring_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var misc = require('utils/misc');

  App.MainMirroringController = Em.ArrayController.extend({
    name: 'mainMirroringController',

    datasetsData: [],

    // formatted data for targetClusterMapper
    clustersData: {},

    // counter for datasets load queries
    datasetCount: 0,

    // counter for target cluster load queries
    clusterCount: 0,

    selectedDataset: null,

    isDatasetsLoaded: false,

    isTargetClustersLoaded: false,

    isRequiredServicesStarted: false,

    isDatasetLoadingError: false,

    actionsDisabled: function () {
      return !this.get('isRequiredServicesStarted') || this.get('isDatasetLoadingError');
    }.property('isRequiredServicesStarted', 'isDatasetLoadingError'),

    isLoaded: function () {
      return this.get('isDatasetsLoaded') && this.get('isTargetClustersLoaded');
    }.property('isDatasetsLoaded', 'isTargetClustersLoaded'),

    datasets: App.Dataset.find(),

    loadData: function () {
      var isRequiredServicesStarted = App.Service.find().findProperty('serviceName', 'OOZIE').get('workStatus') == 'STARTED' && App.Service.find().findProperty('serviceName', 'FALCON').get('workStatus') == 'STARTED';
      this.set('isRequiredServicesStarted', isRequiredServicesStarted);
      if (isRequiredServicesStarted) {
        this.set('isDatasetLoadingError', false);
        this.get('datasetsData').clear();
        this.set('clustersData', {});
        this.set('datasetCount', 0);
        this.set('clusterCount', 0);
        this.loadDatasets();
        this.loadClusters();
      } else {
        this.set('isDatasetLoadingError', true);
      }
    },

    loadDatasets: function () {
      App.ajax.send({
        name: 'mirroring.get_all_entities',
        sender: this,
        data: {
          type: 'feed',
          falconServer: App.get('falconServerURL')
        },
        success: 'onLoadDatasetsListSuccess',
        error: 'onLoadDatasetsListError'
      });
    },

    onLoadDatasetsListSuccess: function (data) {
      var parsedData = misc.xmlToObject(data);
      var datasets = parsedData.entities.entity;
      if (datasets) {
        datasets = Em.isArray(datasets) ? datasets : [datasets];
        datasets = datasets.filter(function (dataset) {
          return dataset.name['#text'].indexOf(App.mirroringDatasetNamePrefix) === 0;
        });
      }
      if (datasets && datasets.length) {
        this.set('datasetCount', datasets.length);
        datasets.forEach(function (dataset) {
          App.ajax.send({
            name: 'mirroring.get_definition',
            sender: this,
            data: {
              name: dataset.name['#text'],
              type: 'feed',
              status: dataset.status['#text'],
              falconServer: App.get('falconServerURL')
            },
            success: 'onLoadDatasetDefinitionSuccess',
            error: 'onLoadDatasetDefinitionError'
          });
        }, this);
      } else {
        this.set('isDatasetsLoaded', true);
      }
    },

    onLoadDatasetsListError: function () {
      this.set('isDatasetLoadingError', true);
      console.error('Failed to load datasets list.');
    },

    onLoadDatasetDefinitionSuccess: function (data) {
      var parsedData = misc.xmlToObject(data);
      var clusters = parsedData.feed.clusters;
      var targetCluster, sourceCluster;

      if (clusters.cluster[0].locations) {
        targetCluster = clusters.cluster[0];
        sourceCluster = clusters.cluster[1];
      } else {
        targetCluster = clusters.cluster[1];
        sourceCluster = clusters.cluster[0];
      }
      var datasetName = parsedData.feed['@attributes'].name.replace(App.mirroringDatasetNamePrefix, '');
      this.get('datasetsData').push(
          Ember.Object.create({
            name: datasetName,
            status: arguments[2].status,
            sourceClusterName: sourceCluster['@attributes'].name,
            targetClusterName: targetCluster['@attributes'].name,
            sourceDir: parsedData.feed.locations.location['@attributes'].path,
            targetDir: targetCluster.locations.location['@attributes'].path,
            frequency: parsedData.feed.frequency['#text'].match(/\d+/)[0],
            frequencyUnit: parsedData.feed.frequency['#text'].match(/\w+(?=\()/)[0],
            scheduleEndDate: sourceCluster.validity['@attributes'].end,
            scheduleStartDate: sourceCluster.validity['@attributes'].start,
            instances: []
          })
      );
      var currentDate = new Date(App.dateTime());
      if (currentDate > new Date(sourceCluster.validity['@attributes'].start)) {
        App.ajax.send({
          name: 'mirroring.dataset.get_all_instances',
          sender: this,
          data: {
            dataset: parsedData.feed['@attributes'].name,
            formattedDatasetName: datasetName,
            start: sourceCluster.validity['@attributes'].start,
            end: App.router.get('mainMirroringEditDataSetController').toTZFormat(currentDate),
            falconServer: App.get('falconServerURL')
          },
          success: 'onLoadDatasetInstancesSuccess',
          error: 'onLoadDatasetsInstancesError'
        });
      } else {
        this.saveDataset();
      }
    },

    onLoadDatasetDefinitionError: function () {
      this.set('isDatasetLoadingError', true);
      console.error('Failed to load dataset definition.');
    },

    onLoadDatasetInstancesSuccess: function (data, sender, opts) {
      var datasetsData = this.get('datasetsData');
      if (data && data.instances) {
        var datasetJobs = [];
        data.instances.forEach(function (instance) {
          if (instance.cluster == App.get('clusterName')) {
            datasetJobs.push({
              dataset: opts.formattedDatasetName,
              id: instance.instance + '_' + opts.dataset,
              name: instance.instance,
              status: instance.status,
              endTime: new Date(instance.endTime).getTime() || 0,
              startTime: new Date(instance.startTime).getTime() || 0
            });
          }
        }, this);
        datasetsData.findProperty('name', opts.formattedDatasetName).set('instances', datasetJobs);
      }
      this.saveDataset();
    },

    saveDataset: function () {
      this.set('datasetCount', this.get('datasetCount') - 1);
      if (this.get('datasetCount') < 1) {
        App.dataSetMapper.map(this.get('datasetsData'));
        this.set('isDatasetsLoaded', true);
        if (App.router.get('currentState.name') === 'index' && App.router.get('currentState.parentState.name') === 'mirroring') {
          App.router.send('gotoShowJobs');
        }
      }
    },

    onLoadDatasetsInstancesError: function () {
      console.error('Failed to load dataset instances.');
      this.saveDataset();
    },

    loadClusters: function () {
      App.ajax.send({
        name: 'mirroring.get_all_entities',
        sender: this,
        data: {
          type: 'cluster',
          falconServer: App.get('falconServerURL')
        },
        success: 'onLoadClustersListSuccess',
        error: 'onLoadClustersListError'
      });
    },

    onLoadClustersListSuccess: function (data) {
      var clustersData = this.get('clustersData');
      clustersData.items = [];
      var parsedData = misc.xmlToObject(data);
      var clusters = parsedData.entities.entity;
      if (data && clusters) {
        clusters = Em.isArray(clusters) ? clusters : [clusters];
        this.set('clusterCount', clusters.length);
        clusters.mapProperty('name.#text').forEach(function (cluster) {
          App.ajax.send({
            name: 'mirroring.get_definition',
            sender: this,
            data: {
              name: cluster,
              type: 'cluster',
              falconServer: App.get('falconServerURL')
            },
            success: 'onLoadClusterDefinitionSuccess',
            error: 'onLoadClusterDefinitionError'
          });
        }, this);
      } else {
        var defaultFS = this.loadDefaultFS();
        var clusterName = App.get('clusterName');
        var sourceCluster = Ember.Object.create({
          name: clusterName,
          execute: App.HostComponent.find().findProperty('componentName', 'RESOURCEMANAGER').get('host.hostName') + ':8050',
          readonly: 'hftp://' + App.HostComponent.find().findProperty('componentName', 'NAMENODE').get('host.hostName') + ':50070',
          workflow: 'http://' + App.HostComponent.find().findProperty('componentName', 'OOZIE_SERVER').get('host.hostName') + ':11000/oozie',
          write: defaultFS,
          staging: '/apps/falcon/' + clusterName + '/staging',
          working: '/apps/falcon/' + clusterName + '/working',
          temp: '/tmp'
        });
        var sourceClusterData = App.router.get('mainMirroringManageClustersController').formatClusterXML(sourceCluster);
        App.ajax.send({
          name: 'mirroring.submit_entity',
          sender: this,
          data: {
            type: 'cluster',
            entity: sourceClusterData,
            falconServer: App.get('falconServerURL')
          },
          success: 'onSourceClusterCreateSuccess',
          error: 'onSourceClusterCreateError'
        });
        clustersData.items.push(sourceCluster);
      }
    },

    /**
     * Return fs.defaultFS config property loaded from server
     * @return {String}
     */
    loadDefaultFS: function () {
      App.ajax.send({
        name: 'config.tags.sync',
        sender: this,
        success: 'onLoadConfigTagsSuccess',
        error: 'onLoadConfigTagsError'
      });
      var configs = App.router.get('configurationController').getConfigsByTags([
        {
          siteName: "core-site",
          tagName: this.get('tag')
        }
      ]);
      return configs[0].properties['fs.defaultFS'];
    },

    // Loaded core-site tag version
    tag: null,

    onLoadConfigTagsSuccess: function (data) {
      this.set('tag', data.Clusters.desired_configs['core-site'].tag);
    },

    onLoadConfigTagsError: function () {
      console.error('Error in loading fs.defaultFS');
    },

    onLoadClustersListError: function () {
      this.set('isDatasetLoadingError', true);
      console.error('Failed to load clusters list.');
    },

    onSourceClusterCreateSuccess: function () {
      App.targetClusterMapper.map(this.get('clustersData'));
      this.set('isTargetClustersLoaded', true);
    },

    onSourceClusterCreateError: function () {
      console.error('Error in creating source cluster entity.');
    },

    onLoadClusterDefinitionSuccess: function (data) {
      var parsedData = misc.xmlToObject(data);
      var clustersData = this.get('clustersData');
      var interfaces = parsedData.cluster.interfaces.interface;
      var locations = parsedData.cluster.locations.location;
      var staging = locations.findProperty('@attributes.name', 'staging');
      var working = locations.findProperty('@attributes.name', 'working');
      var temp = locations.findProperty('@attributes.name', 'temp');
      clustersData.items.push(
          {
            name: parsedData.cluster['@attributes'].name,
            execute: interfaces.findProperty('@attributes.type', 'execute')['@attributes'].endpoint,
            readonly: interfaces.findProperty('@attributes.type', 'readonly')['@attributes'].endpoint,
            workflow: interfaces.findProperty('@attributes.type', 'workflow')['@attributes'].endpoint,
            write: interfaces.findProperty('@attributes.type', 'write')['@attributes'].endpoint,
            staging: staging && staging['@attributes'].path,
            working: working && working['@attributes'].path,
            temp: temp && temp['@attributes'].path
          }
      );
      this.set('clusterCount', this.get('clusterCount') - 1);
      if (this.get('clusterCount') < 1) {
        App.targetClusterMapper.map(clustersData);
        this.set('isTargetClustersLoaded', true);
      }
    },

    onLoadClusterDefinitionError: function () {
      this.set('isDatasetLoadingError', true);
      console.error('Failed to load cluster definition.');
    },

    manageClusters: function () {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('mirroring.dataset.manageClusters'),
        classNames: ['sixty-percent-width-modal'],
        bodyClass: App.MainMirroringManageClusterstView.extend({
          controller: App.router.get('mainMirroringManageClustersController')
        }),
        primary: null,
        secondary: Em.I18n.t('common.close'),
        hide: function () {
          self.loadData();
          App.router.send('gotoShowJobs');
          this._super();
        },
        didInsertElement: function () {
          this._super();
          this.fitHeight();
        }
      });
    }
  });
  
});
window.require.register("controllers/main/service", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainServiceController = Em.ArrayController.extend({
    name:'mainServiceController',
    content: function(){
      if(!App.router.get('clusterController.isLoaded')){
        return [];
      }
      return App.Service.find();
    }.property('App.router.clusterController.isLoaded').volatile(),

    cluster: function () {
      if (!App.router.get('clusterController.isLoaded')) {
        return null;
      }
      return App.Cluster.find().objectAt(0);
    }.property('App.router.clusterController.isLoaded'),

    isAllServicesInstalled: function() {
      if (!this.get('content.content')) return false;
      var availableServices = App.db.getServices();
      if (!availableServices) {
        this.loadAvailableServices();
        availableServices = App.db.getServices();
      }
      if (!App.supports.hue) {
        availableServices = availableServices.without('HUE');
      }
      return this.get('content.content').length == availableServices.length;
    }.property('content.content.@each', 'content.content.length'),

    loadAvailableServices: function() {
      App.ajax.send({
        name: 'wizard.service_components',
        sender: this,
        data: {
          stackUrl: App.get('stack2VersionURL'),
          stackVersion: App.get('currentStackVersionNumber')
        },
        success: 'loadAvailableServicesSuccessCallback'
      });
    },

    loadAvailableServicesSuccessCallback: function(jsonData) {
      var data = [];
      var displayOrderConfig = require('data/services');
      for (var i = 0; i < displayOrderConfig.length; i++) {
        var entry = jsonData.items.findProperty("StackServices.service_name", displayOrderConfig[i].serviceName);
        if (entry) {
          data.push(entry.StackServices.service_name);
        }
      }
      App.db.setServices(data);
    },

    isStartAllDisabled: function(){
      if(this.get('isStartStopAllClicked') == true) {
        return true;
      }
      var stoppedServiceLength = this.get('content').filterProperty('healthStatus','red').filterProperty('isClientsOnly', false).length;
      return (stoppedServiceLength === 0); // all green status
    }.property('isStartStopAllClicked', 'content.@each.healthStatus'),
    isStopAllDisabled: function(){
      if(this.get('isStartStopAllClicked') == true) {
        return true;
      }
      var startedServiceLength = this.get('content').filterProperty('healthStatus','green').length;
      return (startedServiceLength === 0);
    }.property('isStartStopAllClicked', 'content.@each.healthStatus'),
    isStartStopAllClicked: function(){
      return (App.router.get('backgroundOperationsController').get('allOperationsCount') !== 0);
    }.property('App.router.backgroundOperationsController.allOperationsCount'),

    /**
     * callback for <code>start all service</code> button
     */
    startAllService: function(event){
      if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
        return;
      }
      var self = this;
      App.showConfirmationFeedBackPopup(function(query) {
        self.allServicesCall('startAllService', query);
      });
    },

    /**
     * callback for <code>stop all service</code> button
     */
    stopAllService: function(event){
      if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
        return;
      }
      var self = this;
      App.showConfirmationFeedBackPopup(function(query) {
        self.allServicesCall('stopAllService', query);
      });
    },

    allServicesCall: function(state, query) {
      var data;
      if (state == 'stopAllService') {
        data = '{"RequestInfo": {"context" :"' +
          App.BackgroundOperationsController.CommandContexts.STOP_ALL_SERVICES +
          '"}, "Body": {"ServiceInfo": {"state": "INSTALLED"}}}';
      } else {
        data = '{"RequestInfo": {"context" :"' +
          App.BackgroundOperationsController.CommandContexts.START_ALL_SERVICES +
          '"}, "Body": {"ServiceInfo": {"state": "STARTED"}}}';
      }

      App.ajax.send({
        name: 'service.start_stop',
        sender: this,
        data: {
          data: data,
          query: query
        },
        success: 'allServicesCallSuccessCallback',
        error: 'allServicesCallErrorCallback'
      });
    },

    allServicesCallSuccessCallback: function(data, xhr, params) {
      console.log("TRACE: Start/Stop all service -> In success function for the start/stop all Service call");
      console.log("TRACE: Start/Stop all service -> value of the received data is: " + data);
      var requestId = data.Requests.id;
      params.query.set('status', 'SUCCESS');
      console.log('requestId is: ' + requestId);

      // load data (if we need to show this background operations popup) from persist
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        if (initValue) {
          App.router.get('backgroundOperationsController').showPopup();
        }
      });
    },
    allServicesCallErrorCallback: function(request, ajaxOptions, error, opt, params) {
      console.log("ERROR");
      params.query.set('status', 'FAIL');
    },

    gotoAddService: function() {
      if (this.get('isAllServicesInstalled')) {
        return;
      }
      App.router.transitionTo('main.serviceAdd');
    }

  });
  
});
window.require.register("controllers/main/service/add_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  App.AddServiceController = App.WizardController.extend({

    name: 'addServiceController',

    serviceConfigs:require('data/service_configs'),

    totalSteps: 7,

    /**
     * Used for hiding back button in wizard
     */
    hideBackButton: true,

    /**
     * All wizards data will be stored in this variable
     *
     * cluster - cluster name
     * installOptions - ssh key, repo info, etc.
     * services - services list
     * hosts - list of selected hosts
     * slaveComponentHosts, - info about slave hosts
     * masterComponentHosts - info about master hosts
     * config??? - to be described later
     */
    content: Em.Object.create({
      cluster: null,
      hosts: null,
      installOptions: null,
      services: null,
      slaveComponentHosts: null,
      masterComponentHosts: null,
      serviceConfigProperties: null,
      advancedServiceConfig: null,
      controllerName: 'addServiceController',
      configGroups: []
    }),

    setCurrentStep: function (currentStep, completed) {
      this._super(currentStep, completed);
      App.clusterStatus.setClusterStatus({
        wizardControllerName: this.get('name'),
        localdb: App.db.data
      });
    },

    /**
     * return new object extended from clusterStatusTemplate
     * @return Object
     */
    getCluster: function(){
      return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
    },

    /**
     * Load confirmed hosts.
     * Will be used at <code>Assign Masters(step5)</code> step
     */
    loadConfirmedHosts: function(){
      var hosts = this.getDBProperty('hosts');
      if(!hosts){
        var hosts = {};

        App.Host.find().forEach(function(item){
          hosts[item.get('id')] = {
            name: item.get('id'),
            cpu: item.get('cpu'),
            memory: item.get('memory'),
            disk_info: item.get('diskInfo'),
            bootStatus: "REGISTERED",
            isInstalled: true
          };
        });
        this.setDBProperty('hosts', hosts);
      }

      this.set('content.hosts', hosts);
      console.log('AddServiceController.loadConfirmedHosts: loaded hosts', hosts);
    },

    /**
     * Load services data from server.
     */
    loadServicesFromServer: function() {
      if(this.getDBProperty('service')){
        return;
      }
      var displayOrderConfig = require('data/services');
      var apiUrl = App.get('stack2VersionURL');
      var apiService = this.loadServiceComponents(displayOrderConfig, apiUrl);
      //
      apiService.forEach(function(item, index){
        apiService[index].isSelected = App.Service.find().someProperty('id', item.serviceName);
        apiService[index].isDisabled = apiService[index].isSelected;
        apiService[index].isInstalled = apiService[index].isSelected;
      });
      this.set('content.services', apiService);
      this.setDBProperty('service', apiService);
    },

    /**
     * Load services data. Will be used at <code>Select services(step4)</code> step
     */
    loadServices: function () {
      var servicesInfo = this.getDBProperty('service');
      servicesInfo.forEach(function (item, index) {
        servicesInfo[index] = Em.Object.create(item);
      });
      this.set('content.services', servicesInfo);
      console.log('AddServiceController.loadServices: loaded data ', servicesInfo);

      var serviceNames = servicesInfo.filterProperty('isSelected', true).filterProperty('isDisabled', false).mapProperty('serviceName');
      console.log('selected services ', serviceNames);

      this.set('content.skipSlavesStep', !serviceNames.contains('MAPREDUCE') && !serviceNames.contains('HBASE')  && !serviceNames.contains('STORM') && !serviceNames.contains('YARN'));
      if (this.get('content.skipSlavesStep')) {
        this.get('isStepDisabled').findProperty('step', 3).set('value', this.get('content.skipSlavesStep'));
      }
    },

    /**
     * Save data to model
     * @param stepController App.WizardStep4Controller
     */
    saveServices: function (stepController) {var serviceNames = [];
      this.setDBProperty('service', stepController.get('content'));
      console.log('AddServiceController.saveServices: saved data', stepController.get('content'));
      stepController.filterProperty('isSelected', true).filterProperty('isInstalled', false).forEach(function (item) {
        serviceNames.push(item.serviceName);
      });
      this.set('content.selectedServiceNames', serviceNames);
      this.setDBProperty('selectedServiceNames',serviceNames);
      console.log('AddServiceController.selectedServiceNames:', serviceNames);

      this.set('content.skipSlavesStep', !serviceNames.contains('MAPREDUCE') && !serviceNames.contains('HBASE') && !serviceNames.contains('STORM') && !serviceNames.contains('YARN'));
      if (this.get('content.skipSlavesStep')) {
        this.get('isStepDisabled').findProperty('step', 3).set('value', this.get('content.skipSlavesStep'));
      }
    },

    /**
     * Save Master Component Hosts data to Main Controller
     * @param stepController App.WizardStep5Controller
     */
    saveMasterComponentHosts: function (stepController) {
      var obj = stepController.get('selectedServicesMasters');
      var masterComponentHosts = [];
      var installedComponents = App.HostComponent.find();

      obj.forEach(function (_component) {
          masterComponentHosts.push({
            display_name: _component.display_name,
            component: _component.component_name,
            hostName: _component.selectedHost,
            serviceId: _component.serviceId,
            isInstalled: installedComponents.someProperty('componentName', _component.component_name)
          });
      });

      console.log("AddServiceController.saveMasterComponentHosts: saved hosts ", masterComponentHosts);
      this.setDBProperty('masterComponentHosts', masterComponentHosts);
      this.set('content.masterComponentHosts', masterComponentHosts);

      this.set('content.skipMasterStep', this.get('content.masterComponentHosts').everyProperty('isInstalled', true));
      this.get('isStepDisabled').findProperty('step', 2).set('value', this.get('content.skipMasterStep'));
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadMasterComponentHosts: function () {
      var masterComponentHosts = this.getDBProperty('masterComponentHosts');
      if(!masterComponentHosts){
        masterComponentHosts = [];
        App.HostComponent.find().filterProperty('isMaster', true).forEach(function(item){
          masterComponentHosts.push({
            component: item.get('componentName'),
            hostName: item.get('host.hostName'),
            isInstalled: true
          })
        });

      }
      this.set("content.masterComponentHosts", masterComponentHosts);
      console.log("AddServiceController.loadMasterComponentHosts: loaded hosts ", masterComponentHosts);

      this.set('content.skipMasterStep', this.get('content.masterComponentHosts').everyProperty('isInstalled', true));
      this.get('isStepDisabled').findProperty('step', 2).set('value', this.get('content.skipMasterStep'));
    },

    /**
     * Does service have any configs
     * @param {string} serviceName
     * @returns {boolean}
     */
    isServiceConfigurable: function(serviceName) {
      return this.get('serviceConfigs').mapProperty('serviceName').contains(serviceName);
    },

    /**
     * Should Config Step be skipped (based on selected services list)
     * @returns {boolean}
     */
    skipConfigStep: function() {
      var skipConfigStep = true;
      var selectedServices = this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false).mapProperty('serviceName');
      selectedServices.map(function(serviceName) {
        skipConfigStep = skipConfigStep && !this.isServiceConfigurable(serviceName);
      }, this);
      return skipConfigStep;
    },

    loadServiceConfigProperties: function() {
      this._super();
      if (!this.get('content.services')) {
        this.loadServices();
      }
      if (this.get('currentStep') > 1 && this.get('currentStep') < 6) {
        this.set('content.skipConfigStep', this.skipConfigStep());
        this.get('isStepDisabled').findProperty('step', 4).set('value', this.get('content.skipConfigStep'));
      }
    },

    saveServiceConfigProperties: function(stepController) {
      this._super(stepController);
      if (this.get('currentStep') > 1 && this.get('currentStep') < 6) {
        this.set('content.skipConfigStep', this.skipConfigStep());
        this.get('isStepDisabled').findProperty('step', 4).set('value', this.get('content.skipConfigStep'));
      }
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadSlaveComponentHosts: function () {
      var slaveComponentHosts = this.getDBProperty('slaveComponentHosts');
      if(!slaveComponentHosts){
        slaveComponentHosts = this.getSlaveComponentHosts();
      }
      this.set("content.slaveComponentHosts", slaveComponentHosts);
      console.log("AddServiceController.loadSlaveComponentHosts: loaded hosts ", slaveComponentHosts);
    },

    /**
     * Load information about hosts with clients components
     */
    loadClients: function(){
      var clients = this.getDBProperty('clientInfo');
      this.set('content.clients', clients);
      console.log("AddServiceController.loadClients: loaded list ", clients);
    },

    /**
     * Generate clients list for selected services and save it to model
     * @param stepController step4WizardController
     */
    saveClients: function(stepController){
      var clients = [];
      var serviceComponents = require('data/service_components');
      var hostComponents = App.HostComponent.find();

      stepController.get('content').filterProperty('isSelected',true).forEach(function (_service) {
        var client = serviceComponents.filterProperty('service_name', _service.serviceName).findProperty('isClient', true);
        if (client) {
          clients.pushObject({
            component_name: client.component_name,
            display_name: client.display_name,
            isInstalled: hostComponents.filterProperty('componentName', client.component_name).length > 0
          });
        }
      }, this);

      this.setDBProperty('clientInfo', clients);
      this.set('content.clients', clients);
      console.log("AddServiceController.saveClients: saved list ", clients);
    },

    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '7':
        case '6':
        case '5':
          this.load('cluster');
        case '4':
          this.loadServiceConfigGroups();
          this.loadServiceConfigProperties();
        case '3':
          this.loadServices();
          this.loadClients();
          this.loadSlaveComponentHosts();//depends on loadServices
        case '2':
          this.loadMasterComponentHosts();
          this.loadConfirmedHosts();
        case '1':
          this.loadServices();
      }
    },

    /**
     * Remove all loaded data.
     * Created as copy for App.router.clearAllSteps
     */
    clearAllSteps: function () {
      this.clearInstallOptions();
      // clear temporary information stored during the install
      this.set('content.cluster', this.getCluster());
    },

    /**
     * Clear all temporary data
     */
    finish: function () {
      this.setCurrentStep('1');
      this.clearAllSteps();
      this.clearStorageData();
      App.router.get('updateController').updateAll();
    },

    installServices: function (isRetry) {
      this.set('content.cluster.oldRequestsId', []);
      var clusterName = this.get('content.cluster.name');
      var data;
      var name;
      if (isRetry) {
        this.getFailedHostComponents();
        console.log('failedHostComponents', this.get('failedHostComponents'));
        name = 'wizard.install_services.installer_controller.is_retry';
        data = {
          "RequestInfo": {
            "context" : Em.I18n.t('requestInfo.installComponents'),
            "query": "HostRoles/component_name.in(" + this.get('failedHostComponents').join(',') + ")"
          },
          "Body": {
            "HostRoles": {
              "state": "INSTALLED"
            }
          }
        };
        data = JSON.stringify(data);
      }
      else {
        name = 'wizard.install_services.installer_controller.not_is_retry';
        data = '{"RequestInfo": {"context" :"' + Em.I18n.t('requestInfo.installServices') + '"}, "Body": {"ServiceInfo": {"state": "INSTALLED"}}}';
      }
      App.ajax.send({
        name: name,
        sender: this,
        data: {
          data: data,
          cluster: clusterName
        },
        success: 'installServicesSuccessCallback',
        error: 'installServicesErrorCallback'
      });
    },

    /**
     * List of failed to install HostComponents while adding Service
     */
    failedHostComponents: [],

    getFailedHostComponents: function() {
      App.ajax.send({
        name: 'wizard.install_services.add_service_controller.get_failed_host_components',
        sender: this,
        success: 'getFailedHostComponentsSuccessCallback',
        error: 'getFailedHostComponentsErrorCallback'
      });
    },

    /**
     * Parse all failed components and filter out installed earlier components (based on selected to install services)
     * @param {Object} json
     */
    getFailedHostComponentsSuccessCallback: function(json) {
      var allFailed = json.items.filterProperty('HostRoles.state', 'INSTALL_FAILED');
      var currentFailed = [];
      var selectedServices = this.getDBProperty('service').filterProperty('isInstalled', false).filterProperty('isSelected', true).mapProperty('serviceName');
      allFailed.forEach(function(failed) {
        if (selectedServices.contains(failed.component[0].ServiceComponentInfo.service_name)) {
          currentFailed.push(failed.HostRoles.component_name);
        }
      });
      this.set('failedHostComponents', currentFailed);
    },

    getFailedHostComponentsErrorCallback: function(request, ajaxOptions, error) {
      console.warn(error);
    }

  });
  
});
window.require.register("controllers/main/service/info/audit", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainServiceInfoAuditController = Em.Controller.extend({
    name: 'mainServiceInfoAuditController'
  });
});
window.require.register("controllers/main/service/info/configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  require('controllers/wizard/slave_component_groups_controller');
  var batchUtils = require('utils/batch_scheduled_requests');

  App.MainServiceInfoConfigsController = Em.Controller.extend({
    name: 'mainServiceInfoConfigsController',
    isHostsConfigsPage: false,
    forceTransition: false,
    dataIsLoaded: false,
    stepConfigs: [], //contains all field properties that are viewed in this service
    selectedService: null,
    serviceConfigTags: null,
    selectedConfigGroup: null,
    configGroups: [],
    globalConfigs: [],
    uiConfigs: [],
    customConfig: [],
    serviceConfigsData: require('data/service_configs'),
    isApplyingChanges: false,
    // contain Service Config Property, when user proceed from Select Config Group dialog
    overrideToAdd: null,
    serviceConfigs: function () {
      return App.config.get('preDefinedServiceConfigs');
    }.property('App.config.preDefinedServiceConfigs'),
    customConfigs: function () {
      return App.config.get('preDefinedCustomConfigs');
    }.property('App.config.preDefinedCustomConfigs'),
    configMapping: function () {
      return App.config.get('configMapping');
    }.property('App.config.configMapping'),
    configs: function () {
      return  App.config.get('preDefinedGlobalProperties');
    }.property('App.config.preDefinedGlobalProperties'),

    secureConfigs: function () {
      if (App.get('isHadoop2Stack')) {
        return require('data/HDP2/secure_mapping');
      } else {
        return require('data/secure_mapping');
      }
    }.property('App.isHadoop2Stack'),

    /**
     * Map, which contains relation between group and site
     * to upload overriden properties
     */
    loadedGroupToOverrideSiteToTagMap: {},
    /**
     * During page load time the cluster level site to tag
     * mapping is stored here.
     *
     * Example:
     * {
     *  'global': 'version1',
     *  'hdfs-site': 'version1',
     *  'core-site': 'version1'
     * }
     */
    loadedClusterSiteToTagMap: {},
    /**
     * Holds the actual base service-config server data uploaded.
     * This is used by the host-override mechanism to update host
     * specific values.
     */
    savedSiteNameToServerServiceConfigDataMap: {},

    isSubmitDisabled: function () {
      return (!(this.stepConfigs.everyProperty('errorCount', 0)) || this.get('isApplyingChanges'));
    }.property('stepConfigs.@each.errorCount', 'isApplyingChanges'),

    slaveComponentGroups: null,

    /**
     * Filter text will be located here
     */
    filter: '',

    /**
     * Dropdown menu items in filter combobox
     */
    filterColumns: function () {
      var result = [];
      for (var i = 1; i < 2; i++) {
        result.push(Ember.Object.create({
          name: this.t('common.combobox.dropdown.' + i),
          selected: false
        }));
      }
      return result;
    }.property(),

    /**
     * clear and set properties to default value
     */
    clearStep: function () {
      this.set('isInit', true);
      this.set('hash', null);
      this.set('forceTransition', false);
      this.set('dataIsLoaded', false);
      this.set('filter', '');
      this.get('filterColumns').setEach('selected', false);
      this.get('stepConfigs').clear();
      this.get('globalConfigs').clear();
      this.get('uiConfigs').clear();
      this.get('customConfig').clear();
      this.set('loadedGroupToOverrideSiteToTagMap', {});
      this.set('savedSiteNameToServerServiceConfigDataMap', {});
      if (this.get('serviceConfigTags')) {
        this.set('serviceConfigTags', null);
      }
    },

    serviceConfigProperties: function () {
      return App.db.getServiceConfigProperties();
    }.property('content'),

    /**
     * "Finger-print" of the <code>stepConfigs</code>. Filled after first configGroup selecting
     * Used to determine if some changes were made (when user navigates away from this page)
     * {String}
     */
    hash: null,
    /**
     * Is this initial config group changing
     * {Boolean}
     */
    isInit: true,
    /**
     * On load function
     */
    loadStep: function () {
      console.log("TRACE: Loading configure for service");
      this.clearStep();
      this.loadServiceConfigs();
    },

    getHash: function() {
      var hash = {};
      this.get('stepConfigs')[0].configs.forEach(function(config) {
        hash[config.get('name')] = {value: config.get('value'), overrides: []};
        if (!config.get('overrides')) return;
        if (!config.get('overrides.length')) return;

        config.get('overrides').forEach(function(override) {
          hash[config.get('name')].overrides.push(override.get('value'));
        });
      });
      return JSON.stringify(hash);
    },

    /**
     * Loads the actual configuration of all host components.
     * This helps in determining which services need a restart, and also
     * in showing which properties are actually applied or not.
     * This method also compares the actual_configs with the desired_configs
     * and builds a diff structure.
     *
     * Internally it calculates an array of host-components which need restart.
     * Example:
     * [
     *  {
     *    componentName: 'DATANODE',
     *    serviceName: 'HDFS',
     *    host: 'host.name',
     *    type: 'core-site',
     *    desiredConfigTags: {tag:'version1'},
     *    actualConfigTags: {tag:'version4'. host_override:'version2'}
     *  },
     *  ...
     * ]
     *
     * From there it return the following restart-data for this service.
     * It represents the hosts, whose components need restart, and the
     * properties which require restart.
     *
     * {
     *  hostAndHostComponents: {
     *   'hostname1': {
     *     'DATANODE': {
     *       'property1': 'value1',
     *       'property2': 'value2'
     *     },
     *     'TASKTRACKER': {
     *       'prop1': 'val1'
     *     }
     *    },
     *    'hostname6': {
     *     'ZOOKEEPER': {
     *       'property1': 'value3'
     *     }
     *    }
     *  },
     *  propertyToHostAndComponent: {
     *    'property1': {
     *      'hostname1': ['DATANODE'],
     *      'hostname6': ['ZOOKEEPER']
     *    },
     *    'property2': {
     *      'hostname1': ['DATANODE']
     *    },
     *    'prop1': {
     *      'hostname1': ['TASKTRACKER']
     *    }
     *  }
     * }
     */
    loadActualConfigsAndCalculateRestarts: function () {
      var currentService = this.get('content.serviceName');
      var restartData = {
        hostAndHostComponents: {},
        propertyToHostAndComponent: {}
      };
      var self = this;
      console.log("loadActualConfigsAndCalculateRestarts(): Restart data = ", restartData);
      return restartData;
    },

    /**
     * Loads service configurations
     */
    loadServiceConfigs: function () {
      App.ajax.send({
        name: 'config.tags_and_groups',
        sender: this,
        data: {
          serviceName: this.get('content.serviceName'),
          serviceConfigsDef: this.get('serviceConfigs').findProperty('serviceName', this.get('content.serviceName'))
        },
        success: 'loadServiceTagsSuccess'
      });
    },

    loadServiceTagsSuccess: function (data, opt, params) {
      var serviceConfigsDef = params.serviceConfigsDef;
      var serviceName = this.get('content.serviceName');
      console.debug("loadServiceConfigs(): data=", data);
      // Create default configuration group
      var defaultConfigGroupHosts = App.Host.find().mapProperty('hostName');
      var selectedConfigGroup;
      var siteToTagMap = {};
      serviceConfigsDef.sites.forEach(function(siteName){
        if(data.Clusters.desired_configs[siteName]){
          siteToTagMap[siteName] = data.Clusters.desired_configs[siteName].tag;
        } else {
          siteToTagMap[siteName] = 'version1';
        }
      }, this);
      this.loadedClusterSiteToTagMap = siteToTagMap;
      //parse loaded config groups
      if (App.supports.hostOverrides) {
        var configGroups = [];
        if (data.config_groups.length) {
          data.config_groups.forEach(function (item) {
            item = item.ConfigGroup;
            if (item.tag === this.get('content.serviceName')) {
              var groupHosts = item.hosts.mapProperty('host_name');
              var newConfigGroup = App.ConfigGroup.create({
                id: item.id,
                name: item.group_name,
                description: item.description,
                isDefault: false,
                parentConfigGroup: null,
                service: App.Service.find().findProperty('serviceName', item.tag),
                hosts: groupHosts,
                configSiteTags: []
              });
              groupHosts.forEach(function (host) {
                defaultConfigGroupHosts = defaultConfigGroupHosts.without(host);
              }, this);
              item.desired_configs.forEach(function (config) {
                newConfigGroup.configSiteTags.push(App.ConfigSiteTag.create({
                  site: config.type,
                  tag: config.tag
                }));
              }, this);
              // select default selected group for hosts page
              if (!selectedConfigGroup && this.get('isHostsConfigsPage') && newConfigGroup.get('hosts').contains(this.get('host.hostName')) && this.get('content.serviceName') === item.tag) {
                selectedConfigGroup = newConfigGroup;
              }
              configGroups.push(newConfigGroup);
            }
          }, this);
        }
        this.set('configGroups', configGroups);
      }
      var defaultConfigGroup = App.ConfigGroup.create({
        name: App.Service.DisplayNames[serviceName] + " Default",
        description: "Default cluster level " + serviceName + " configuration",
        isDefault: true,
        hosts: defaultConfigGroupHosts,
        parentConfigGroup: null,
        service: this.get('content'),
        serviceName: serviceName,
        configSiteTags: []
      });
      if (!selectedConfigGroup) {
        selectedConfigGroup = defaultConfigGroup;
      }

      this.get('configGroups').sort(function(configGroupA, configGroupB){
        return (configGroupA.name > configGroupB.name);
      });
      this.get('configGroups').unshift(defaultConfigGroup);
      this.set('selectedConfigGroup', selectedConfigGroup);
    },

    onConfigGroupChange: function () {
      this.get('stepConfigs').clear();
      var selectedConfigGroup = this.get('selectedConfigGroup');
      var serviceName = this.get('content.serviceName');
      //STEP 1: handle tags from JSON data for host overrides
      this.loadedGroupToOverrideSiteToTagMap = {};
      if (App.supports.hostOverrides) {
        var configGroupsWithOverrides = selectedConfigGroup.get('isDefault') && !this.get('isHostsConfigsPage') ? this.get('configGroups') : [selectedConfigGroup];
        configGroupsWithOverrides.forEach(function (item) {
          var groupName = item.get('name');
          this.loadedGroupToOverrideSiteToTagMap[groupName] = {};
          item.get('configSiteTags').forEach(function (siteTag) {
            var site = siteTag.get('site');
            this.loadedGroupToOverrideSiteToTagMap[groupName][site] = siteTag.get('tag');
          }, this);
        }, this);
      }
      //STEP 2: Create an array of objects defining tag names to be polled and new tag names to be set after submit
      this.setServiceConfigTags(this.loadedClusterSiteToTagMap);
      //STEP 3: Load advanced configs from server
      var advancedConfigs = [];
      App.config.loadAdvancedConfig(serviceName, function (properties) {
        advancedConfigs.pushObjects(properties);
      }, true);
      //STEP 4: Load on-site config by service from server
      var configGroups = App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags'));
      //STEP 5: Merge global and on-site configs with pre-defined
      var configSet = App.config.mergePreDefinedWithLoaded(configGroups, advancedConfigs, this.get('serviceConfigTags'), serviceName);
      configSet = App.config.syncOrderWithPredefined(configSet);
      //var serviceConfigs = this.getSitesConfigProperties(advancedConfigs);
      var configs = configSet.configs;
      //put global configs into globalConfigs to save them separately
      this.set('globalConfigs', configSet.globalConfigs);

      //STEP 6: add advanced configs
      //App.config.addAdvancedConfigs(configs, advancedConfigs, serviceName);
      //STEP 7: add custom configs
      App.config.addCustomConfigs(configs);
      //put properties from capacity-scheduler.xml into one config with textarea view
      if (this.get('content.serviceName') === 'YARN' && !App.supports.capacitySchedulerUi) {
        configs = App.config.fileConfigsIntoTextarea(configs, 'capacity-scheduler.xml');
      }
      //STEP 8: add configs as names of host components
      this.addHostNamesToGlobalConfig();

      var allConfigs = this.get('globalConfigs').concat(configs);
      //STEP 9: Load and add overriden configs of group
      App.config.loadServiceConfigGroupOverrides(allConfigs, this.loadedGroupToOverrideSiteToTagMap, this.get('configGroups'));
      var restartData = this.loadActualConfigsAndCalculateRestarts();
      //STEP 10: creation of serviceConfig object which contains configs for current service
      var serviceConfig = App.config.createServiceConfig(serviceName);
      //STEP11: Make SecondaryNameNode invisible on enabling namenode HA
      if (serviceConfig.get('serviceName') === 'HDFS') {
        App.config.OnNnHAHideSnn(serviceConfig);
      }
      this.checkForRestart(serviceConfig, restartData);
      if (serviceName || serviceConfig.serviceName === 'MISC') {
        //STEP 11: render configs and wrap each in ServiceConfigProperty object
        this.loadComponentConfigs(allConfigs, serviceConfig, restartData, advancedConfigs);
        this.get('stepConfigs').pushObject(serviceConfig);
      }
      this.set('selectedService', this.get('stepConfigs').objectAt(0));
      this.checkForSecureConfig(this.get('selectedService'));
      this.set('dataIsLoaded', true);

      this.set('hash', this.getHash());
      this.set('isInit', false);

    }.observes('selectedConfigGroup'),

    /**
     * Changes format from Object to Array
     *
     * {
     *  'core-site': 'version1',
     *  'hdfs-site': 'version1',
     *  'global': 'version2',
     *  ...
     * }
     *
     * to
     *
     * [
     *  {
     *    siteName: 'core-site',
     *    tagName: 'version1',
     *    newTageName: null
     *  },
     *  ...
     * ]
     *
     * set tagnames for configuration of the *-site.xml
     */
    setServiceConfigTags: function (desiredConfigsSiteTags) {
      console.debug("setServiceConfigTags(): Trying to set ", desiredConfigsSiteTags);
      var newServiceConfigTags = [];
      for (var index in desiredConfigsSiteTags) {
        newServiceConfigTags.pushObject({
          siteName: index,
          tagName: desiredConfigsSiteTags[index],
          newTagName: null
        }, this);
      }
      console.debug("setServiceConfigTags(): Setting 'serviceConfigTags' to ", newServiceConfigTags);
      this.set('serviceConfigTags', newServiceConfigTags);
    },

    /**
     * check whether host component must be restarted
     * @param serviceConfig
     * @param restartData
     */
    checkForRestart: function (serviceConfig, restartData) {
      var hostsCount = 0;
      var hostComponentCount = 0;
      var restartHosts = Ember.A([]);
      if (restartData != null && restartData.hostAndHostComponents != null && !jQuery.isEmptyObject(restartData.hostAndHostComponents)) {
        serviceConfig.set('restartRequired', true);
        for (var host in restartData.hostAndHostComponents) {
          hostsCount++;
          var componentsArray = Ember.A([]);
          for (var component in restartData.hostAndHostComponents[host]) {
            componentsArray.push(Ember.Object.create({name: App.format.role(component)}));
            hostComponentCount++;
          }
          var hostObj = App.Host.find(host);
          restartHosts.push(Ember.Object.create({hostData: hostObj, components: componentsArray}))
        }
        serviceConfig.set('restartRequiredHostsAndComponents', restartHosts);
        serviceConfig.set('restartRequiredMessage', 'Service needs ' + hostComponentCount + ' components on ' + hostsCount + ' hosts to be restarted.')
      }
    },

    /**
     * check whether the config property is a security related knob
     * @param serviceConfig
     */
    checkForSecureConfig: function (serviceConfig) {
      serviceConfig.get('configs').forEach(function (_config) {
        this.get('secureConfigs').forEach(function (_secureConfig) {
          if (_config.get('name') === _secureConfig.name) {
            _config.set('isSecureConfig', true);
          }
        }, this)
      }, this)
    },

    /**
     * Get info about hosts and host components to configDefaultsProviders
     * @returns {{masterComponentHosts: Array, slaveComponentHosts: Array, hosts: {}}}
     */
    getInfoForDefaults: function() {

      var slaveComponentHosts = [];
      var slaves = App.HostComponent.find().filterProperty('isSlave', true).map(function(item) {
        return Em.Object.create({
          host: item.get('host.hostName'),
          componentName: item.get('componentName')
        });
      });
      slaves.forEach(function(slave) {
        var s = slaveComponentHosts.findProperty('componentName', slave.componentName);
        if (s) {
          s.hosts.push({hostName: slave.host});
        }
        else {
          slaveComponentHosts.push({
            componentName: slave.get('componentName'),
            hosts: [{hostName: slave.host}]
          });
        }
      });

      var masterComponentHosts = App.HostComponent.find().filterProperty('isMaster', true).map(function(item) {
        return {
          component: item.get('componentName'),
          serviceId: item.get('service.serviceName'),
          host: item.get('host.hostName')
        }
      });
      var hosts = {};
      App.Host.find().map(function(host) {
        hosts[host.get('hostName')] = {
          name: host.get('hostName'),
          cpu: host.get('cpu'),
          memory: host.get('memory'),
          disk_info: host.get('diskInfo')
        };
      });

      return {
        masterComponentHosts: masterComponentHosts,
        slaveComponentHosts: slaveComponentHosts,
        hosts: hosts
      };
    },

    /**
     * Load child components to service config object
     * @param configs
     * @param componentConfig
     * @param restartData
     * @param advancedConfigs
     */
    loadComponentConfigs: function (configs, componentConfig, restartData, advancedConfigs) {

      var localDB = this.getInfoForDefaults();
      var recommendedDefaults = {};
      var s = this.get('serviceConfigsData').findProperty('serviceName', this.get('content.serviceName'));
      var defaultGroupSelected = this.get('selectedConfigGroup.isDefault');
      var defaults = [];
      if (s.defaultsProviders) {
        s.defaultsProviders.forEach(function(defaultsProvider) {
          var d = defaultsProvider.getDefaults(localDB);
          defaults.push(d);
          for (var name in d) {
            var defaultValueFromStack = advancedConfigs.findProperty('name',name);
            if (!!d[name]) {
              recommendedDefaults[name] = d[name];
            } else {
             // If property default value is not declared on client, fetch it from stack definition
             // If it's not declared with any valid value in both server stack and client, then js reference error is expected to be thrown
              recommendedDefaults[name] = defaultValueFromStack.value
            }
          }
        });
      }
      if (s.configsValidator) {
        s.configsValidator.set('recommendedDefaults', recommendedDefaults);
      }


      configs.forEach(function (_serviceConfigProperty) {
        console.log("config", _serviceConfigProperty);
        if (!_serviceConfigProperty) return;
        var overrides = _serviceConfigProperty.overrides;
        // we will populate the override properties below
        _serviceConfigProperty.overrides = null;

        if (_serviceConfigProperty.isOverridable === undefined) {
          _serviceConfigProperty.isOverridable = true;
        }
        if (_serviceConfigProperty.displayType === 'checkbox') {
          switch (_serviceConfigProperty.value) {
            case 'true':
              _serviceConfigProperty.value = true;
              _serviceConfigProperty.defaultValue = true;
              break;
            case 'false':
              _serviceConfigProperty.value = false;
              _serviceConfigProperty.defaultValue = false;
              break;
          }
        }
        var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
        var propertyName = serviceConfigProperty.get('name');
        if (restartData != null && propertyName in restartData.propertyToHostAndComponent) {
          serviceConfigProperty.set('isRestartRequired', true);
          var message = '<ul>';
          for (var host in restartData.propertyToHostAndComponent[propertyName]) {
            var appHost = App.Host.find(host);
            message += "<li>" + appHost.get('publicHostName');
            message += "<ul>";
            restartData.propertyToHostAndComponent[propertyName][host].forEach(function (comp) {
              message += "<li>" + App.format.role(comp) + "</li>"
            });
            message += "</ul></li>";
          }
          message += "</ul>";
          serviceConfigProperty.set('restartRequiredMessage', message);
        }
        if (serviceConfigProperty.get('serviceName') === this.get('content.serviceName')) {

          // Do not reset values when reconfiguring.
          // This might be useful to setting better descriptions
          // or default values sometime in the future.
          // defaults.forEach(function(defaults) {
          //   for(var name in defaults) {
          //    if (serviceConfigProperty.name == name) {
          //       serviceConfigProperty.set('value', defaults[name]);
          //       serviceConfigProperty.set('defaultValue', defaults[name]);
          //     }
          //   }
          // });

          if (s.configsValidator) {
            var validators = s.configsValidator.get('configValidators');
            for (var validatorName in validators) {
              if (serviceConfigProperty.name == validatorName) {
                serviceConfigProperty.set('serviceValidator', s.configsValidator);
              }
            }
          }

          console.log("config result", serviceConfigProperty);
        } else {
          serviceConfigProperty.set('isVisible', false);
        }
        if (overrides != null) {
          overrides.forEach(function (override) {
            var newSCP = App.ServiceConfigProperty.create(_serviceConfigProperty);
            newSCP.set('value', override.value);
            newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
            newSCP.set('parentSCP', serviceConfigProperty);
            if (App.supports.hostOverrides && defaultGroupSelected) {
              newSCP.set('group', override.group);
              newSCP.set('isEditable', false);
            }
            var parentOverridesArray = serviceConfigProperty.get('overrides');
            if (parentOverridesArray == null) {
              parentOverridesArray = Ember.A([]);
              serviceConfigProperty.set('overrides', parentOverridesArray);
            }
            parentOverridesArray.pushObject(newSCP);
            console.debug("createOverrideProperty(): Added:", newSCP, " to main-property:", serviceConfigProperty)
          }, this)
        }
        if (App.get('isAdmin')) {
          if(defaultGroupSelected && !this.get('isHostsConfigsPage')){
            serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
          }else{
            serviceConfigProperty.set('isEditable', false);
          }
        } else {
          serviceConfigProperty.set('isEditable', false);
        }
        componentConfig.configs.pushObject(serviceConfigProperty);
        serviceConfigProperty.validate();
      }, this);
      var overrideToAdd = this.get('overrideToAdd');
      if (overrideToAdd) {
        overrideToAdd = componentConfig.configs.findProperty('name', overrideToAdd.name);
        if (overrideToAdd) {
          this.addOverrideProperty(overrideToAdd);
          this.set('overrideToAdd', null);
        }
      }
    },

    /**
     * Determines which host components are running on each host.
     * @param {Array} services
     * @param {String} status 'running' or 'unknown'
     * @return Returned in the following format:
     * {
     *  runningHosts: {
     *    'hostname1': 'NameNode, DataNode, JobTracker',
     *    'hostname2': 'DataNode',
     *  },
     *  runningComponentCount: 5
     * }
     */
    getHostComponentsByStatus: function (services, status) {
      var hosts = [];
      var componentCount = 0;
      var hostToIndexMap = {};
      services.forEach(function (service) {
        var hostComponents = service.get('hostComponents').filterProperty('workStatus', status);
        if (hostComponents != null) {
          hostComponents.forEach(function (hc) {
            var hostName = hc.get('host.publicHostName');
            var componentName = hc.get('displayName');
            componentCount++;
            if (!(hostName in hostToIndexMap)) {
              hosts.push({
                name: hostName,
                components: ""
              });
              hostToIndexMap[hostName] = hosts.length - 1;
            }
            var hostObj = hosts[hostToIndexMap[hostName]];
            if (hostObj.components.length > 0)
              hostObj.components += ", " + componentName;
            else
              hostObj.components += componentName;
          });
          hosts.sort(function (a, b) {
            return a.name.localeCompare(b.name);
          });
        }
      });
      return {
        hosts: hosts,
        componentCount: componentCount
      };
    },

    /**
     * open popup with appropriate message
     */
    restartServicePopup: function (event) {
      if (this.get("isSubmitDisabled")) {
        return;
      }
      var header;
      var message;
      var messageClass;
      var hasUnknown = false;
      var value;
      var flag = false;
      var runningHosts = null;
      var runningComponentCount = 0;
      var unknownHosts = null;
      var unknownComponentCount = 0;

      var dfd = $.Deferred();
      var self = this;
      var serviceName = this.get('content.serviceName');
      var displayName = this.get('content.displayName');

      if (App.supports.hostOverrides ||
        (serviceName !== 'HDFS' && this.get('content.isStopped') === true) ||
        ((serviceName === 'HDFS') && this.get('content.isStopped') === true && (!App.Service.find().someProperty('id', 'MAPREDUCE') || App.Service.find('MAPREDUCE').get('isStopped')))) {

        // warn the user if any service directories are being changed
        var dirChanged = false;

        if (serviceName === 'HDFS') {
          var hdfsConfigs = self.get('stepConfigs').findProperty('serviceName', 'HDFS').get('configs');
          if (App.get('isHadoop2Stack')) {
            if (
              hdfsConfigs.findProperty('name', 'dfs.namenode.name.dir').get('isNotDefaultValue') ||
                hdfsConfigs.findProperty('name', 'dfs.namenode.checkpoint.dir').get('isNotDefaultValue') ||
                hdfsConfigs.findProperty('name', 'dfs.datanode.data.dir').get('isNotDefaultValue')
              ) {
              dirChanged = true;
            }
          } else {
            if (
              hdfsConfigs.findProperty('name', 'dfs.name.dir').get('isNotDefaultValue') ||
                hdfsConfigs.findProperty('name', 'fs.checkpoint.dir').get('isNotDefaultValue') ||
                hdfsConfigs.findProperty('name', 'dfs.data.dir').get('isNotDefaultValue')
              ) {
              dirChanged = true;
            }
          }
        } else if (serviceName === 'MAPREDUCE') {
          var mapredConfigs = self.get('stepConfigs').findProperty('serviceName', 'MAPREDUCE').get('configs');
          if (
            mapredConfigs.findProperty('name', 'mapred.local.dir').get('isNotDefaultValue') ||
              mapredConfigs.findProperty('name', 'mapred.system.dir').get('isNotDefaultValue')
            ) {
            dirChanged = true;
          }
        }

        if (dirChanged) {
          App.showConfirmationPopup(function () {
            dfd.resolve();
          }, Em.I18n.t('services.service.config.confirmDirectoryChange').format(displayName));
        } else {
          dfd.resolve();
        }

        dfd.done(function () {
          var result = self.saveServiceConfigProperties();
          App.router.get('clusterController').updateClusterData();
          App.router.get('updateController').updateComponentConfig(function(){});
          flag = result.flag;
          if (result.flag === true) {
            header = Em.I18n.t('services.service.config.saved');
            message = Em.I18n.t('services.service.config.saved.message');
            messageClass = 'alert alert-success';
            // warn the user if any of the components are in UNKNOWN state
            var uhc;
            if (self.get('content.serviceName') !== 'HDFS' || (self.get('content.serviceName') === 'HDFS' && !App.Service.find().someProperty('id', 'MAPREDUCE'))) {
              uhc = self.getHostComponentsByStatus([self.get('content')], App.HostComponentStatus.unknown);
            } else {
              uhc = self.getHostComponentsByStatus([self.get('content'), App.Service.find('MAPREDUCE')], App.HostComponentStatus.unknown);
            }
            unknownHosts = uhc.hosts;
            unknownComponentCount = uhc.componentCount;
          } else {
            header = Em.I18n.t('common.failure');
            message = result.message;
            messageClass = 'alert alert-error';
            value = result.value;
          }
        });
      } else {
        var rhc;
        if (this.get('content.serviceName') !== 'HDFS' || (this.get('content.serviceName') === 'HDFS' && !App.Service.find().someProperty('id', 'MAPREDUCE'))) {
          rhc = this.getHostComponentsByStatus([this.get('content')], App.HostComponentStatus.started);
          header = Em.I18n.t('services.service.config.notSaved');
          message = Em.I18n.t('services.service.config.msgServiceStop');
        } else {
          rhc = this.getHostComponentsByStatus([this.get('content'), App.Service.find('MAPREDUCE')], App.HostComponentStatus.started);
          header = Em.I18n.t('services.service.config.notSaved');
          message = Em.I18n.t('services.service.config.msgHDFSMapRServiceStop');
        }
        messageClass = 'alert alert-error';
        runningHosts = rhc.hosts;
        runningComponentCount = rhc.componentCount;
        dfd.resolve();
      }

      dfd.done(function () {
        App.ModalPopup.show({
          header: header,
          primary: Em.I18n.t('ok'),
          secondary: null,
          onPrimary: function () {
            this.hide();
            if (flag) {
              self.loadStep();
            }
          },
          bodyClass: Ember.View.extend({
            flag: flag,
            message: message,
            messageClass: messageClass,
            runningHosts: runningHosts,
            runningComponentCount: runningComponentCount,
            unknownHosts: unknownHosts,
            unknownComponentCount: unknownComponentCount,
            siteProperties: value,
            getDisplayMessage: function () {
              var displayMsg = [];
              var siteProperties = this.get('siteProperties');
              if (siteProperties) {
                siteProperties.forEach(function (_siteProperty) {
                  var displayProperty = _siteProperty.siteProperty;
                  var displayNames = _siteProperty.displayNames;
                  /////////
                  if (displayNames && displayNames.length) {
                    if (displayNames.length === 1) {
                      displayMsg.push(displayProperty + Em.I18n.t('as') + displayNames[0]);
                    } else {
                      var name;
                      displayNames.forEach(function (_name, index) {
                        if (index === 0) {
                          name = _name;
                        } else if (index === siteProperties.length - 1) {
                          name = name + Em.I18n.t('and') + _name;
                        } else {
                          name = name + ', ' + _name;
                        }
                      }, this);
                      displayMsg.push(displayProperty + Em.I18n.t('as') + name);

                    }
                  } else {
                    displayMsg.push(displayProperty);
                  }
                }, this);
              }
              return displayMsg;

            }.property('siteProperties'),

            runningHostsMessage: function () {
              return Em.I18n.t('services.service.config.stopService.runningHostComponents').format(this.get('runningComponentCount'), this.get('runningHosts.length'));
            }.property('runningComponentCount', 'runningHosts.length'),

            unknownHostsMessage: function () {
              return Em.I18n.t('services.service.config.stopService.unknownHostComponents').format(this.get('unknownComponentCount'), this.get('unknownHosts.length'));
            }.property('unknownComponentCount', 'unknownHosts.length'),

            templateName: require('templates/main/service/info/configs_save_popup')
          })
        })
      });
    },

    /**
     * Save config properties
     */
    saveServiceConfigProperties: function () {
      var result = {
        flag: false,
        message: null,
        value: null
      };
      var selectedConfigGroup = this.get('selectedConfigGroup');
      var configs = this.get('stepConfigs').findProperty('serviceName', this.get('content.serviceName')).get('configs');

      if (selectedConfigGroup.get('isDefault')) {
        this.saveGlobalConfigs(configs);
        if (this.get('content.serviceName') === 'YARN' && !App.supports.capacitySchedulerUi) {
          configs = App.config.textareaIntoFileConfigs(configs, 'capacity-scheduler.xml');
        }
        this.saveSiteConfigs(configs);

        /**
         * First we put cluster configurations, which automatically creates /configurations
         * resources. Next we update host level overrides.
         */
        result.flag = this.doPUTClusterConfigurations();
      } else {
        var overridenConfigs = [];
        var groupHosts = [];
        configs.filterProperty('isOverridden', true).forEach(function (config) {
          overridenConfigs = overridenConfigs.concat(config.get('overrides'));
        });
        this.formatConfigValues(overridenConfigs);
        selectedConfigGroup.get('hosts').forEach(function(hostName){
          groupHosts.push({"host_name": hostName});
        });

        this.putConfigGroupChanges({
          ConfigGroup: {
            "id": selectedConfigGroup.get('id'),
            "cluster_name": App.get('clusterName'),
            "group_name": selectedConfigGroup.get('name'),
            "tag": selectedConfigGroup.get('service.id'),
            "description": selectedConfigGroup.get('description'),
            "hosts": groupHosts,
            "desired_configs": this.buildGroupDesiredConfigs(overridenConfigs)
          }
        });
        result.flag = this.get('isPutConfigGroupChangesSuccess');
      }
      if (!result.flag) {
        result.message = Em.I18n.t('services.service.config.failSaveConfig');
      } else {
        if (!result.flag) {
          result.message = Em.I18n.t('services.service.config.failSaveConfigHostOverrides');
        }
      }
      console.log("The result from applyCreatdConfToService is: " + result);
      return result;
    },
    /**
     * construct desired_configs for config groups from overriden properties
     * @param configs
     * @param timeTag
     * @return {Array}
     */
    buildGroupDesiredConfigs: function (configs, timeTag) {
      var sites = [];
      var time = timeTag || (new Date).getTime();
      configs.forEach(function (config) {
        var type = config.get('filename').replace('.xml', '');
        var site = sites.findProperty('type', type);
        if (site) {
          site.properties.push({
            name: config.get('name'),
            value: config.get('value')
          });
        } else {
          site = {
            type: type,
            tag: 'version' + time,
            properties: [{
              name: config.get('name'),
              value: config.get('value')
            }]
          };
          sites.push(site);
        }
      });
      sites.forEach(function(site){
        if(site.type === 'global') {
          site.properties = this.createGlobalSiteObj(site.tag, site.properties).properties;
        } else {
          site.properties = this.createSiteObj(site.type, site.tag, site.properties).properties;
        }
      }, this);
      return sites;
    },
    /**
     * persist properties of config groups to server
     * @param data
     */
    putConfigGroupChanges: function (data) {
      App.ajax.send({
        name: 'config_groups.update_config_group',
        sender: this,
        data: {
          id: data.ConfigGroup.id,
          configGroup: data
        },
        success: "putConfigGroupChangesSuccess"
      });
    },
    isPutConfigGroupChangesSuccess: false,
    putConfigGroupChangesSuccess: function () {
      this.set('isPutConfigGroupChangesSuccess', true);
    },
    /**
     * save new or change exist configs in global configs
     * @param configs
     */
    saveGlobalConfigs: function (configs) {
      var globalConfigs = this.get('globalConfigs');
      configs.filterProperty('id', 'puppet var').forEach(function (uiConfigProperty) {
        uiConfigProperty.set('value', App.config.trimProperty(uiConfigProperty));
        if (globalConfigs.someProperty('name', uiConfigProperty.name)) {
          var modelGlobalConfig = globalConfigs.findProperty('name', uiConfigProperty.name);
          modelGlobalConfig.value = uiConfigProperty.value;
        } else {
          globalConfigs.pushObject({
            name: uiConfigProperty.name,
            value: uiConfigProperty.value
          });
        }
      }, this);

      this.setHiveHostName(globalConfigs);
      this.setOozieHostName(globalConfigs);
      this.set('globalConfigs', globalConfigs);
    },

    /**
     * set hive hostnames in global configs
     * @param globals
     */
    setHiveHostName: function (globals) {
      if (globals.someProperty('name', 'hive_database')) {
        var hiveDb = globals.findProperty('name', 'hive_database');
        if (hiveDb.value === 'New MySQL Database') {
          var ambariHost = globals.findProperty('name', 'hive_ambari_host');
          if (ambariHost) {
            ambariHost.name = 'hive_hostname';
          }
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_database'));
        } else if (hiveDb.value === 'Existing MySQL Database') {
          var existingMySqlHost = globals.findProperty('name', 'hive_existing_mysql_host');
          if (existingMySqlHost) {
            existingMySqlHost.name = 'hive_hostname';
          }
          globals = globals.without(globals.findProperty('name', 'hive_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'hive_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_database'));
        } else { //existing oracle database
          var existingOracleHost = globals.findProperty('name', 'hive_existing_oracle_host');
          if (existingOracleHost) {
            existingOracleHost.name = 'hive_hostname';
          }
          globals = globals.without(globals.findProperty('name', 'hive_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'hive_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_database'));
        }

      }
    },

    /**
     * set oozie hostnames in global configs
     * @param globals
     */
    setOozieHostName: function (globals) {
      if (globals.someProperty('name', 'oozie_database')) {
        var oozieDb = globals.findProperty('name', 'oozie_database');
        if (oozieDb.value === 'New Derby Database') {
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_database'));
        } else if (oozieDb.value === 'New MySQL Database') {
          var ambariHost = globals.findProperty('name', 'oozie_ambari_host');
          if (ambariHost) {
            ambariHost.name = 'oozie_hostname';
          }
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_derby_database'));

        } else if (oozieDb.value === 'Existing MySQL Database') {
          var existingMySqlHost = globals.findProperty('name', 'oozie_existing_mysql_host');
          if (existingMySqlHost) {
            existingMySqlHost.name = 'oozie_hostname';
          }
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_derby_database'));
        } else { //existing oracle database
          var existingOracleHost = globals.findProperty('name', 'oozie_existing_oracle_host');
          if (existingOracleHost) {
            existingOracleHost.name = 'oozie_hostname';
          }
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_derby_database'));
        }

      }
    },

    /**
     * save site configs
     * @param configs
     */
    saveSiteConfigs: function (configs) {
      //storedConfigs contains custom configs as well
      var serviceConfigProperties = configs.filterProperty('id', 'site property');
      this.formatConfigValues(serviceConfigProperties);
      var storedConfigs = serviceConfigProperties.filterProperty('value');
      var mappedConfigs = App.config.excludeUnsupportedConfigs(this.get('configMapping').all(), App.Service.find().mapProperty('serviceName'));
      var allUiConfigs = this.loadUiSideConfigs(mappedConfigs);
      this.set('uiConfigs', storedConfigs.concat(allUiConfigs));
    },

    formatConfigValues: function(serviceConfigProperties){
      serviceConfigProperties.forEach(function (_config) {
        if (typeof _config.get('value') === "boolean") _config.set('value', _config.value.toString());
        _config.set('value', App.config.trimProperty(_config), true);
      });
    },

    /**
     * return configs from the UI side
     * @param configMapping array with configs
     * @return {Array}
     */
    loadUiSideConfigs: function (configMapping) {
      var uiConfig = [];
      var configs = configMapping.filterProperty('foreignKey', null);
      this.addDynamicProperties(configs);
      configs.forEach(function (_config) {
        var valueWithOverrides = this.getGlobConfigValueWithOverrides(_config.templateName, _config.value, _config.name);
        if (valueWithOverrides !== null) {
          uiConfig.pushObject({
            "id": "site property",
            "name": _config.name,
            "value": valueWithOverrides.value,
            "filename": _config.filename,
            "overrides": valueWithOverrides.overrides
          });
        }
      }, this);
      return uiConfig;
    },


    addDynamicProperties: function (configs) {
      var allConfigs = this.get('stepConfigs').findProperty('serviceName', this.get('content.serviceName')).get('configs');
      var templetonHiveProperty = allConfigs.someProperty('name', 'templeton.hive.properties');
      if (!templetonHiveProperty && this.get('content.serviceName') === 'WEBHCAT') {
        configs.pushObject({
          "name": "templeton.hive.properties",
          "templateName": ["hivemetastore_host"],
          "foreignKey": null,
          "value": "hive.metastore.local=false,hive.metastore.uris=thrift://<templateName[0]>:9083,hive.metastore.sasl.enabled=yes,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse",
          "filename": "webhcat-site.xml"
        });
      }
    },

    /**
     * return global config value
     * @param templateName
     * @param expression
     * @param name
     * @return {Object}
     * example: <code>{
     *   value: '...',
     *   overrides: {
     *    'value1': [h1, h2],
     *    'value2': [h3]
     *   }
     * }</code>
     */
    getGlobConfigValueWithOverrides: function (templateName, expression, name) {
      var express = expression.match(/<(.*?)>/g);
      var value = expression;
      if (express == null) {
        return { value: expression, overrides: {}};      // if site property do not map any global property then return the value
      }
      var overrideHostToValue = {};
      express.forEach(function (_express) {
        //console.log("The value of template is: " + _express);
        var index = parseInt(_express.match(/\[([\d]*)(?=\])/)[1]);
        if (this.get('globalConfigs').someProperty('name', templateName[index])) {
          //console.log("The name of the variable is: " + this.get('content.serviceConfigProperties').findProperty('name', templateName[index]).name);
          var globalObj = this.get('globalConfigs').findProperty('name', templateName[index]);
          var globValue = globalObj.value;
          // Hack for templeton.zookeeper.hosts
          var preReplaceValue = null;
          if (value !== null) {   // if the property depends on more than one template name like <templateName[0]>/<templateName[1]> then don't proceed to the next if the prior is null or not found in the global configs
            preReplaceValue = value;
            value = this._replaceConfigValues(name, _express, value, globValue);
          }
          if (globalObj.overrides != null) {
            for (var ov in globalObj.overrides) {
              var hostsArray = globalObj.overrides[ov];
              hostsArray.forEach(function (host) {
                if (!(host in overrideHostToValue)) {
                  overrideHostToValue[host] = this._replaceConfigValues(name, _express, preReplaceValue, ov);
                } else {
                  overrideHostToValue[host] = this._replaceConfigValues(name, _express, overrideHostToValue[host], ov);
                }
              }, this);
            }
          }
        } else {
          /*
           console.log("ERROR: The variable name is: " + templateName[index]);
           console.log("ERROR: mapped config from configMapping file has no corresponding variable in " +
           "content.serviceConfigProperties. Two possible reasons for the error could be: 1) The service is not selected. " +
           "and/OR 2) The service_config metadata file has no corresponding global var for the site property variable");
           */
          value = null;
        }
      }, this);

      var valueWithOverrides = {
        value: value,
        overrides: {}
      };
      if (!jQuery.isEmptyObject(overrideHostToValue)) {
        for (var host in overrideHostToValue) {
          var hostVal = overrideHostToValue[host];
          if (!(hostVal in valueWithOverrides.overrides)) {
            valueWithOverrides.overrides[hostVal] = [];
          }
          valueWithOverrides.overrides[hostVal].push(host);
        }
      }
      return valueWithOverrides;
    },

    _replaceConfigValues: function (name, express, value, globValue) {
      return value.replace(express, globValue);
    },

    /**
     * Saves cluster level configurations for all necessary sites.
     * PUT calls are made to /api/v1/clusters/clusterName for each site.
     *
     * @return {Boolean}
     */
    doPUTClusterConfigurations: function () {
      var result = true;
      var serviceConfigTags = this.get('serviceConfigTags');
      this.setNewTagNames(serviceConfigTags);
      var siteNameToServerDataMap = {};
      var configController = App.router.get('configurationController');
      var filenameExceptions = App.config.get('filenameExceptions');

      serviceConfigTags.forEach(function (_serviceTags) {
        var loadedProperties;
        if (_serviceTags.siteName === 'global') {
          console.log("TRACE: Inside global");
          var serverGlobalConfigs = this.createGlobalSiteObj(_serviceTags.newTagName, this.get('globalConfigs'));
          siteNameToServerDataMap['global'] = serverGlobalConfigs;
          loadedProperties = configController.getConfigsByTags([{siteName: 'global', tagName: this.loadedClusterSiteToTagMap['global']}]);
          if (loadedProperties && loadedProperties[0]) {
            loadedProperties = loadedProperties[0].properties;
          }
          if (this.isConfigChanged(loadedProperties, serverGlobalConfigs.properties)) {
            result = result && this.doPUTClusterConfigurationSite(serverGlobalConfigs);
          }
        }
        else {
          if (_serviceTags.siteName === 'core-site') {
            console.log("TRACE: Inside core-site");
            if (this.get('content.serviceName') === 'HDFS' || this.get('content.serviceName') === 'GLUSTERFS') {
              var coreSiteConfigs = this.createCoreSiteObj(_serviceTags.newTagName);
              siteNameToServerDataMap['core-site'] = coreSiteConfigs;
              loadedProperties = configController.getConfigsByTags([{siteName: 'core-site', tagName: this.loadedClusterSiteToTagMap['core-site']}]);
              if (loadedProperties && loadedProperties[0]) {
                loadedProperties = loadedProperties[0].properties;
              }
              if (this.isConfigChanged(loadedProperties, coreSiteConfigs.properties)) {
                result = result && this.doPUTClusterConfigurationSite(coreSiteConfigs);
              }
            }
          }
          else {
            var filename = (filenameExceptions.contains(_serviceTags.siteName)) ? _serviceTags.siteName : _serviceTags.siteName + '.xml';
            var siteConfigs = this.get('uiConfigs').filterProperty('filename', filename);
            var serverConfigs = this.createSiteObj(_serviceTags.siteName, _serviceTags.newTagName, siteConfigs);
            siteNameToServerDataMap[_serviceTags.siteName] = serverConfigs;
            loadedProperties = configController.getConfigsByTags([{siteName: _serviceTags.siteName, tagName: this.loadedClusterSiteToTagMap[_serviceTags.siteName]}]);
            if (loadedProperties && loadedProperties[0]) {
              loadedProperties = loadedProperties[0].properties;
            }
            if (!loadedProperties) {
              loadedProperties = {};
            }
            if (filename === 'mapred-queue-acls' && !App.supports.capacitySchedulerUi) {
              return;
            }
            if (this.isConfigChanged(loadedProperties, serverConfigs.properties)) {
              result = result && this.doPUTClusterConfigurationSite(serverConfigs);
            }
          }
        }
      }, this);
      this.savedSiteNameToServerServiceConfigDataMap = siteNameToServerDataMap;
      return result;
    },


    /**
     * Compares the loaded config values with the saving config values.
     */
    isConfigChanged: function (loadedConfig, savingConfig) {
      var changed = false;
      if (loadedConfig != null && savingConfig != null) {
        var seenLoadKeys = [];
        for (var loadKey in loadedConfig) {
          if (!loadedConfig.hasOwnProperty(loadKey)) continue;
          seenLoadKeys.push(loadKey);
          var loadValue = loadedConfig[loadKey];
          var saveValue = savingConfig[loadKey];
          if ("boolean" == typeof(saveValue)) {
            saveValue = saveValue.toString();
          }
          if (saveValue == null) {
            saveValue = "null";
          }
          if (loadValue !== saveValue) {
            changed = true;
            break;
          }
        }
        for (var saveKey in savingConfig) {
          if (seenLoadKeys.indexOf(saveKey) < 0) {
            changed = true;
            break;
          }
        }
      }
      return changed;
    },

    /**
     * Saves configuration of a particular site. The provided data
     * contains the site name and tag to be used.
     */
    doPUTClusterConfigurationSite: function (data) {
      App.ajax.send({
        name: 'config.cluster_configuration.put',
        sender: this,
        data: {
          data: JSON.stringify({
            Clusters: {
              desired_config: data
            }
          }),
          cluster: App.router.getClusterName()
        },
        success: 'doPUTClusterConfigurationSiteSuccessCallback',
        error: 'doPUTClusterConfigurationSiteErrorCallback'
      });
      return this.get('doPUTClusterConfigurationSiteResult');
    },

    /**
     * @type {bool}
     */
    doPUTClusterConfigurationSiteResult: null,

    doPUTClusterConfigurationSiteSuccessCallback: function(data) {
      console.log("applyClusterConfigurationToSite(): In success for data:", data);
      this.set('doPUTClusterConfigurationSiteResult', true);
    },

    doPUTClusterConfigurationSiteErrorCallback: function(request, ajaxOptions, error) {
      console.log('applyClusterConfigurationToSite(): ERROR:', request.responseText, ", error=", error);
      this.set('doPUTClusterConfigurationSiteResult', false);
    },

    /**
     * add newTagName property to each config in serviceConfigs
     * @param serviceConfigs
     */
    setNewTagNames: function (serviceConfigs) {
      var time = (new Date).getTime();
      serviceConfigs.forEach(function (_serviceConfigs) {
        _serviceConfigs.newTagName = 'version' + time;
      }, this);
    },

    /**
     * create global site object
     * @param {String} tagName
     * @param {Array} globalConfigs
     * @return {Object}
     */
    createGlobalSiteObj: function (tagName, globalConfigs) {
      var heapsizeException = ['hadoop_heapsize', 'yarn_heapsize', 'nodemanager_heapsize', 'resourcemanager_heapsize', 'apptimelineserver_heapsize'];
      var globalSiteProperties = {};
      globalConfigs.forEach(function (_globalSiteObj) {
        // do not pass any globalConfigs whose name ends with _host or _hosts
        if (_globalSiteObj.isRequiredByAgent !== false) {
          // append "m" to JVM memory options except for hadoop_heapsize
          if (/_heapsize|_newsize|_maxnewsize$/.test(_globalSiteObj.name) && !heapsizeException.contains(_globalSiteObj.name)) {
            _globalSiteObj.value += "m";
          }
          globalSiteProperties[_globalSiteObj.name] = App.config.escapeXMLCharacters(_globalSiteObj.value);
          //this.recordHostOverride(_globalSiteObj, 'global', tagName, this);
          //console.log("TRACE: name of the global property is: " + _globalSiteObj.name);
          //console.log("TRACE: value of the global property is: " + _globalSiteObj.value);
        }
      }, this);
      return {"type": "global", "tag": tagName, "properties": globalSiteProperties};
    },

    /**
     * create core site object
     * @param tagName
     * @return {Object}
     */
    createCoreSiteObj: function (tagName) {
      var coreSiteObj = this.get('uiConfigs').filterProperty('filename', 'core-site.xml');
      var coreSiteProperties = {};
      coreSiteObj.forEach(function (_coreSiteObj) {
        coreSiteProperties[_coreSiteObj.name] = App.config.escapeXMLCharacters(_coreSiteObj.value);
        //this.recordHostOverride(_coreSiteObj, 'core-site', tagName, this);
      }, this);
      return {"type": "core-site", "tag": tagName, "properties": coreSiteProperties};
    },

    /**
     * create site object
     * @param siteName
     * @param tagName
     * @param siteObj
     * @return {Object}
     */
    createSiteObj: function (siteName, tagName, siteObj) {
      var siteProperties = {};
      siteObj.forEach(function (_siteObj) {
        switch(siteName) {
          case 'falcon-startup.properties':
          case 'falcon-runtime.properties':
            siteProperties[_siteObj.name] = _siteObj.value;
            break;
          default:
            siteProperties[_siteObj.name] = this.setServerConfigValue(_siteObj.name, _siteObj.value);
        }
      }, this);
      return {"type": siteName, "tag": tagName, "properties": siteProperties};
    },
    /**
     * This method will be moved to config's decorators class.
     *
     * For now, provide handling for special properties that need
     * be specified in special format required for server.
     *
     * @param configName {String} - name of config property
     * @param value {Mixed} - value of config property
     *
     * @return {String} - formated value
     */
    setServerConfigValue: function (configName, value) {
      switch (configName) {
        case 'storm.zookeeper.servers':
          return JSON.stringify(value).replace(/"/g, "'");
          break;
        case 'content':
          return value;
          break;
        default:
          return App.config.escapeXMLCharacters(value);
      }
    },
    /**
     * return either specific url for request if testMode is false or testUrl
     * @param testUrl
     * @param url
     * @return {*}
     */
    getUrl: function (testUrl, url) {
      return (App.testMode) ? testUrl : App.apiPrefix + '/clusters/' + App.router.getClusterName() + url;
    },

    /**
     * Adds host name of master component to global config;
     */
    addHostNamesToGlobalConfig: function () {
      var serviceName = this.get('content.serviceName');
      var globalConfigs = this.get('globalConfigs');
      var serviceConfigs = this.get('serviceConfigs').findProperty('serviceName', serviceName).configs;
      var hostComponents = App.HostComponent.find();
      //namenode_host is required to derive "fs.default.name" a property of core-site
      var nameNodeHost = this.get('serviceConfigs').findProperty('serviceName', 'HDFS').configs.findProperty('name', 'namenode_host');
      try {
        nameNodeHost.defaultValue = hostComponents.filterProperty('componentName', 'NAMENODE').mapProperty('host.hostName');
        globalConfigs.push(nameNodeHost);
      } catch (err) {
        console.log("No NameNode Host available.  This is expected if you're using GLUSTERFS rather than HDFS.");
      }

      //zooKeeperserver_host
      var zooKeperHost = this.get('serviceConfigs').findProperty('serviceName', 'ZOOKEEPER').configs.findProperty('name', 'zookeeperserver_hosts');
      if (serviceName === 'ZOOKEEPER' || serviceName === 'HBASE' || serviceName === 'WEBHCAT') {
        zooKeperHost.defaultValue = hostComponents.filterProperty('componentName', 'ZOOKEEPER_SERVER').mapProperty('host.hostName');
        globalConfigs.push(zooKeperHost);
      }

      switch (serviceName) {
        case 'HDFS':
          if (hostComponents.someProperty('componentName', 'SECONDARY_NAMENODE') && hostComponents.findProperty('componentName', 'SECONDARY_NAMENODE').get('workStatus') != 'DISABLED') {
            var sNameNodeHost = serviceConfigs.findProperty('name', 'snamenode_host');
            sNameNodeHost.defaultValue = hostComponents.findProperty('componentName', 'SECONDARY_NAMENODE').get('host.hostName');
            globalConfigs.push(sNameNodeHost);
          }
          break;
        case 'MAPREDUCE':
          var jobTrackerHost = serviceConfigs.findProperty('name', 'jobtracker_host');
          jobTrackerHost.defaultValue = hostComponents.findProperty('componentName', 'JOBTRACKER').get('host.hostName');
          globalConfigs.push(jobTrackerHost);
        case 'MAPREDUCE2':
          var historyServerHost = serviceConfigs.findProperty('name', 'hs_host');
          historyServerHost.defaultValue = hostComponents.findProperty('componentName', 'HISTORYSERVER').get('host.hostName');
          globalConfigs.push(historyServerHost);
          break;
        case 'YARN':
          var resourceManagerHost = serviceConfigs.findProperty('name', 'rm_host');
          var ATSHost = this.getMasterComponentHostValue('APP_TIMELINE_SERVER');
          if (ATSHost) {
            var ATSProperty = serviceConfigs.findProperty('name', 'ats_host');
            ATSProperty.defaultValue = ATSHost
            globalConfigs.push(ATSProperty);
          }
          resourceManagerHost.defaultValue = hostComponents.findProperty('componentName', 'RESOURCEMANAGER').get('host.hostName');
          globalConfigs.push(resourceManagerHost);
          //yarn.log.server.url config dependent on HistoryServer host
          if (hostComponents.someProperty('componentName', 'HISTORYSERVER')) {
            historyServerHost = this.get('serviceConfigs').findProperty('serviceName', 'MAPREDUCE2').configs.findProperty('name', 'hs_host');
            historyServerHost.defaultValue = hostComponents.findProperty('componentName', 'HISTORYSERVER').get('host.hostName');
            globalConfigs.push(historyServerHost);
          }
          break;
        case 'HIVE':
          var hiveMetastoreHost = serviceConfigs.findProperty('name', 'hivemetastore_host');
          hiveMetastoreHost.defaultValue = hostComponents.findProperty('componentName', 'HIVE_SERVER').get('host.hostName');
          globalConfigs.push(hiveMetastoreHost);
          var hiveDb = globalConfigs.findProperty('name', 'hive_database').value;
          if (['Existing MySQL Database', 'Existing Oracle Database'].contains(hiveDb)) {
            globalConfigs.findProperty('name', 'hive_hostname').isVisible = true;
          }
          break;

        case 'OOZIE':
          var oozieServerHost = serviceConfigs.findProperty('name', 'oozieserver_host');
          oozieServerHost.defaultValue = hostComponents.findProperty('componentName', 'OOZIE_SERVER').get('host.hostName');
          globalConfigs.push(oozieServerHost);
          var oozieDb = globalConfigs.findProperty('name', 'oozie_database').value;
          if (['Existing MySQL Database', 'Existing Oracle Database'].contains(oozieDb)) {
            globalConfigs.findProperty('name', 'oozie_hostname').isVisible = true;
          }
          break;
        case 'HBASE':
          var hbaseMasterHost = serviceConfigs.findProperty('name', 'hbasemaster_host');
          hbaseMasterHost.defaultValue = hostComponents.filterProperty('componentName', 'HBASE_MASTER').mapProperty('host.hostName');
          globalConfigs.push(hbaseMasterHost);
          break;
        case 'HUE':
          var hueServerHost = serviceConfigs.findProperty('name', 'hueserver_host');
          hueServerHost.defaultValue = hostComponents.findProperty('componentName', 'HUE_SERVER').get('host.hostName');
          globalConfigs.push(hueServerHost);
          break;
        case 'WEBHCAT':
          var webhcatMasterHost = serviceConfigs.findProperty('name', 'webhcatserver_host');
          webhcatMasterHost.defaultValue = hostComponents.filterProperty('componentName', 'WEBHCAT_SERVER').mapProperty('host.hostName');
          globalConfigs.push(webhcatMasterHost);
          var hiveMetastoreHost = this.get('serviceConfigs').findProperty('serviceName', 'HIVE').configs.findProperty('name', 'hivemetastore_host');
          hiveMetastoreHost.defaultValue = hostComponents.findProperty('componentName', 'HIVE_SERVER').get('host.hostName');
          globalConfigs.push(hiveMetastoreHost);
          break;
        case 'STORM':
          var masterHostComponents = [
            { name: 'STORM_UI_SERVER', propertyName: 'stormuiserver_host' },
            { name: 'DRPC_SERVER', propertyName: 'drpcserver_host' },
            { name: 'STORM_REST_API', propertyName: 'storm_rest_api_host' }
          ];

          masterHostComponents.forEach(function(component) {
            var hostValue = this.getMasterComponentHostValue(component.name);
            var config = serviceConfigs.findProperty('name', component.propertyName);
            config.defaultValue = hostValue;
            globalConfigs.push(config);
          }, this);

          var supervisorHosts = hostComponents.filterProperty('componentName','SUPERVISOR').mapProperty('host.hostName');
          if (supervisorHosts.length > 0) {
            var supervisorHostsConfig = serviceConfigs.findProperty('name', 'supervisor_hosts');
            supervisorHostsConfig.defaultValue = supervisorHosts;
            globalConfigs.push(supervisorHostsConfig);
          }
          break;
      }
    },

    getMasterComponentHostValue: function(componentName) {
      var component = this.get('content.hostComponents').findProperty('componentName', componentName);
      return component ? component.get('host.hostName') : false;
    },
    /**
     * Provides service component name and display-name information for
     * the current selected service.
     */
    getCurrentServiceComponents: function () {
      var service = this.get('content');
      var components = service.get('hostComponents');
      var validComponents = Ember.A([]);
      var seenComponents = {};
      components.forEach(function (component) {
        var cn = component.get('componentName');
        var cdn = component.get('displayName');
        if (!seenComponents[cn]) {
          validComponents.push(Ember.Object.create({
            componentName: cn,
            displayName: cdn,
            selected: false
          }));
          seenComponents[cn] = cn;
        }
      });
      return validComponents;
    }.property('content'),

    doCancel: function () {
      this.loadStep();
    },

    restartAllStaleConfigComponents: function() {
      var self = this;
      var serviceDisplayName = this.get('content.displayName');
      var bodyMessage = Em.Object.create({
        confirmMsg: Em.I18n.t('services.service.restartAll.confirmMsg').format(serviceDisplayName),
        confirmButton: Em.I18n.t('services.service.restartAll.confirmButton'),
        additionalWarningMsg: this.get('content.passiveState') === 'OFF' ? Em.I18n.t('services.service.restartAll.warningMsg.turnOnMM').format(serviceDisplayName): null
      });
      App.showConfirmationFeedBackPopup(function(query) {
        var selectedService = self.get('content.id');
        batchUtils.restartAllServiceHostComponents(selectedService, true, query);
      }, bodyMessage);
    },

    rollingRestartStaleConfigSlaveComponents: function(componentName) {
      batchUtils.launchHostComponentRollingRestart(componentName.context, this.get('content.displayName'), this.get('content.passiveState') === "ON", true);
    },

    showHostsShouldBeRestarted: function() {
      var hosts = [];
      for(var hostName in this.get('content.restartRequiredHostsAndComponents')) {
        hosts.push(hostName);
      }
      var hostsText = hosts.length == 1 ? Em.I18n.t('common.host') : Em.I18n.t('common.hosts');
      hosts = hosts.join(', ');
      this.showItemsShouldBeRestarted(hosts, Em.I18n.t('service.service.config.restartService.shouldBeRestarted').format(hostsText));
    },
    showComponentsShouldBeRestarted: function() {
      var rhc = this.get('content.restartRequiredHostsAndComponents');
      var hostsComponets = [];
      var componentsObject = {};
      for(var hostName in rhc) {
        rhc[hostName].forEach(function(hostComponent) {
          hostsComponets.push(hostComponent);
          if(componentsObject[hostComponent] != undefined) {
            componentsObject[hostComponent]++;
          } else {
            componentsObject[hostComponent] = 1;
          }
        })
      }
      var componentsList = [];
      for( var obj in componentsObject) {
        var componentDisplayName = (componentsObject[obj] > 1) ? obj + 's' : obj;
        componentsList.push(componentsObject[obj] + ' ' + componentDisplayName);
      }
      var componentsText = componentsList.length == 1 ? Em.I18n.t('common.component') : Em.I18n.t('common.components');
      hostsComponets = componentsList.join(', ');
      this.showItemsShouldBeRestarted(hostsComponets, Em.I18n.t('service.service.config.restartService.shouldBeRestarted').format(componentsText));
    },

    showItemsShouldBeRestarted: function(content, header) {
      App.ModalPopup.show({
        content: content,
        header: header,
        bodyClass: Em.View.extend({
          templateName: require('templates/common/selectable_popup'),
          textareaVisible: false,
          textTrigger: function() {
            this.set('textareaVisible', !this.get('textareaVisible'));
          },
          putContentToTextarea: function() {
            var content = this.get('parentView.content');
            if (this.get('textareaVisible')) {
              var wrapper = $(".task-detail-log-maintext");
              $('.task-detail-log-clipboard').html(content).width(wrapper.width()).height(wrapper.height());
              Em.run.next(function() {
                $('.task-detail-log-clipboard').select();
              });
            }
          }.observes('textareaVisible')
        }),
        secondary: null
      });
    },

    addOverrideProperty: function(serviceConfigProperty) {
      var overrides = serviceConfigProperty.get('overrides');
      if (!overrides) {
        overrides = [];
        serviceConfigProperty.set('overrides', overrides);
      }
      // create new override with new value
      var newSCP = App.ServiceConfigProperty.create(serviceConfigProperty);
      newSCP.set('value', '');
      newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
      newSCP.set('parentSCP', serviceConfigProperty);
      newSCP.set('isEditable', true);
      console.debug("createOverrideProperty(): Added:", newSCP, " to main-property:", serviceConfigProperty);
      overrides.pushObject(newSCP);
    },

    manageConfigurationGroup: function () {
      this.manageConfigurationGroups();
    },

    manageConfigurationGroups: function (controller) {
      var serviceData = (controller && controller.get('selectedService')) || this.get('content');
      var serviceName = serviceData.get('serviceName');
      var displayName = serviceData.get('displayName');
      App.router.get('manageConfigGroupsController').set('isInstaller', !!controller);
      App.ModalPopup.show({
        header: Em.I18n.t('services.service.config_groups_popup.header').format(displayName),
        bodyClass: App.MainServiceManageConfigGroupView.extend({
          serviceName: serviceName,
          displayName: displayName,
          controllerBinding: 'App.router.manageConfigGroupsController'
        }),
        classNames: ['sixty-percent-width-modal', 'manage-configuration-group-popup'],
        primary: Em.I18n.t('common.save'),
        onPrimary: function() {
          var modifiedConfigGroups = this.get('subViewController.hostsModifiedConfigGroups');
          // Save modified config-groups
          if (!!controller) {
            controller.set('selectedService.configGroups', App.router.get('manageConfigGroupsController.configGroups'));
            controller.selectedServiceObserver();
            if (controller.get('name') == "wizardStep7Controller") {
              if (controller.get('selectedService.selected') === false && modifiedConfigGroups.toDelete.length > 0) {
                controller.setGroupsToDelete(modifiedConfigGroups.toDelete);
              }
              App.config.persistWizardStep7ConfigGroups();
              this.updateConfigGroupOnServicePage();
            }
            this.hide();
            return;
          }
          console.log("manageConfigurationGroups(): Saving modified config-groups: ", modifiedConfigGroups);
          var self = this;
          var errors = [];
          var deleteQueriesCounter = modifiedConfigGroups.toClearHosts.length + modifiedConfigGroups.toDelete.length;
          var createQueriesCounter = modifiedConfigGroups.toSetHosts.length + modifiedConfigGroups.toCreate.length;
          var deleteQueriesRun = false;
          var createQueriesRun = false;
          var runNextQuery = function () {
            if (!deleteQueriesRun && deleteQueriesCounter > 0) {
              deleteQueriesRun = true;
              modifiedConfigGroups.toClearHosts.forEach(function (cg) {
                App.config.clearConfigurationGroupHosts(cg, finishFunction, finishFunction);
              }, this);
              modifiedConfigGroups.toDelete.forEach(function (cg) {
                App.config.deleteConfigGroup(cg, finishFunction, finishFunction);
              }, this);
            } else if (!createQueriesRun && deleteQueriesCounter < 1) {
              createQueriesRun = true;
              modifiedConfigGroups.toSetHosts.forEach(function (cg) {
                App.config.updateConfigurationGroup(cg, finishFunction, finishFunction);
              }, this);
              modifiedConfigGroups.toCreate.forEach(function (cg) {
                App.config.postNewConfigurationGroup(cg, finishFunction);
              }, this);
            }
          };
          var finishFunction = function (xhr, text, errorThrown) {
            if (xhr && errorThrown) {
              var error = xhr.status + "(" + errorThrown + ") ";
              try {
                var json = $.parseJSON(xhr.responseText);
                error += json.message;
              } catch (err) {
              }
              console.error('Error updating Config Group:', error);
              errors.push(error);
            }
            if (createQueriesRun) {
              createQueriesCounter--;
            } else {
              deleteQueriesCounter--;
            }
            if (deleteQueriesCounter + createQueriesCounter < 1) {
              if (errors.length > 0) {
                console.log(errors);
                self.get('subViewController').set('errorMessage', errors.join(". "));
              } else {
                self.updateConfigGroupOnServicePage();
                self.hide();
              }
            } else {
              runNextQuery();
            }
          };
         runNextQuery();
        },
        onSecondary: function () {
          this.hide();
        },
        onClose: function () {
          this.hide();
        },
        subViewController: function () {
          return App.router.get('manageConfigGroupsController');
        }.property('App.router.manageConfigGroupsController'),

        updateConfigGroupOnServicePage: function () {
          var subViewController = this.get('subViewController');
          var selectedConfigGroup = subViewController.get('selectedConfigGroup');
          var managedConfigGroups = subViewController.get('configGroups');
          if (!controller) {
            controller = App.router.get('mainServiceInfoConfigsController');
            controller.set('configGroups', managedConfigGroups);
          } else {
            controller.set('selectedService.configGroups', managedConfigGroups);
          }
          //check whether selectedConfigGroup was selected
          if (selectedConfigGroup && controller.get('configGroups').someProperty('name', selectedConfigGroup.get('name'))) {
            controller.set('selectedConfigGroup', selectedConfigGroup);
          } else {
            controller.set('selectedConfigGroup', managedConfigGroups.findProperty('isDefault', true));
          }
        },

        updateButtons: function(){
          var modified = this.get('subViewController.isHostsModified');
          this.set('disablePrimary', !modified);
        }.observes('subViewController.isHostsModified'),
        secondary : Em.I18n.t('common.cancel'),
        didInsertElement: function () {}
      });
    },

    selectConfigGroup: function (event) {
      if (!this.get('isInit')) {
        if (this.hasUnsavedChanges()) {
          this.showSavePopup(null, event);
          return;
        }
      }
      this.set('selectedConfigGroup', event.context);
    },

    /**
     * Are some unsaved changes available
     * @returns {boolean}
     */
    hasUnsavedChanges: function() {
      return this.get('hash') != this.getHash();
    },

    /**
     * If some configs are changed and user navigates away or select another config-group, show this popup with propose to save changes
     * @param {String} path
     * @param {object} event - triggered event for selecting another config-group
     */
    showSavePopup: function (path, event) {
      var _this = this;
      App.ModalPopup.show({
        header: Em.I18n.t('common.warning'),
        body: Em.I18n.t('services.service.config.exitPopup.body'),
        footerClass: Ember.View.extend({
          templateName: require('templates/main/service/info/save_popup_footer'),
          isSaveDisabled: function() {
            return _this.get('isSubmitDisabled');
          }.property()
        }),
        primary: Em.I18n.t('common.save'),
        secondary: Em.I18n.t('common.cancel'),
        onSave: function () {
          _this.restartServicePopup();
          this.hide();
        },
        onDiscard: function () {
          if (path) {
            _this.set('forceTransition', true);
            App.router.route(path);
          } else if (event) {
            // Prevent multiple popups
            _this.set('hash', _this.getHash());
            _this.selectConfigGroup(event);
          }
          this.hide();
        },
        onCancel: function () {
          this.hide();
        }
      });
    }
  });
  
});
window.require.register("controllers/main/service/info/summary", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MainServiceInfoSummaryController = Em.Controller.extend({
    name: 'mainServiceInfoSummaryController',

    nagiosUrl: function(){
      return App.router.get('clusterController.nagiosUrl');
    }.property('App.router.clusterController.nagiosUrl'),
    
    isNagiosInstalled: function(){
      return App.router.get('clusterController.isNagiosInstalled');
    }.property('App.router.clusterController.isNagiosInstalled'),

    isGangliaInstalled: function(){
      return App.router.get('clusterController.isGangliaInstalled');
    }.property('App.router.clusterController.isGangliaInstalled')
  });
});
window.require.register("controllers/main/service/item", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var service_components = require('data/service_components');
  var batchUtils = require('utils/batch_scheduled_requests');

  App.MainServiceItemController = Em.Controller.extend({
    name: 'mainServiceItemController',

    /**
     * Callback functions for start and stop service have few differences
     *
     * Used with currentCallBack property
     */
    callBackConfig: {
      'STARTED': {
        'c': 'STARTING',
        'f': 'starting',
        'c2': 'live',
        'hs': 'started',
        's': 'start'
      },
      'INSTALLED': {
        'c': 'STOPPING',
        'f': 'stopping',
        'c2': 'dead',
        'hs': 'stopped',
        's': 'stop'
      }
    },
    /**
     * Common method for ajax (start/stop service) responses
     * @param data
     * @param ajaxOptions
     * @param params
     */
    startStopPopupSuccessCallback: function (data, ajaxOptions, params) {
      if (data && data.Requests) {
        params.query.set('status', 'SUCCESS');
        var config = this.get('callBackConfig')[(JSON.parse(ajaxOptions.data)).Body.ServiceInfo.state];
        var self = this;
        console.log('Send request for ' + config.c + ' successfully');
        if (App.testMode) {
          self.set('content.workStatus', App.Service.Health[config.f]);
          self.get('content.hostComponents').setEach('workStatus', App.HostComponentStatus[config.f]);
          setTimeout(function () {
            self.set('content.workStatus', App.Service.Health[config.c2]);
            self.get('content.hostComponents').setEach('workStatus', App.HostComponentStatus[config.hs]);
          }, App.testModeDelayForActions);
        }
        else {
          App.router.get('clusterController').loadUpdatedStatusDelayed(500);// @todo check working without param 500
        }
        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      } else {
        params.query.set('status', 'FAIL');
        console.log('cannot get request id from ', data);
      }
    },
    startStopPopupErrorCallback: function(request, ajaxOptions, error, opt, params){
      params.query.set('status', 'FAIL');
    },
    /**
     * Confirmation popup for start/stop services
     * @param event
     * @param serviceHealth - 'STARTED' or 'INSTALLED'
     */
    startStopPopup: function(event, serviceHealth) {
      if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
        return;
      }
      var self = this;
      var serviceDisplayName = this.get('content.displayName');
      var isMaintenanceOFF = this.get('content.passiveState') === 'OFF';
      var bodyMessage = Em.Object.create({
        confirmMsg: serviceHealth == 'INSTALLED'? Em.I18n.t('services.service.stop.confirmMsg').format(serviceDisplayName) : Em.I18n.t('question.sure'),
        confirmButton: serviceHealth == 'INSTALLED'? Em.I18n.t('services.service.stop.confirmButton') : Em.I18n.t('ok'),
        additionalWarningMsg:  isMaintenanceOFF && serviceHealth == 'INSTALLED'? Em.I18n.t('services.service.stop.warningMsg.turnOnMM').format(serviceDisplayName) : null
      });

      App.showConfirmationFeedBackPopup(function(query) {
        self.set('isPending', true);
        self.startStopPopupPrimary(serviceHealth, query);
      }, bodyMessage);
    },

    startStopPopupPrimary: function (serviceHealth, query) {
      var requestInfo = "";
      if (serviceHealth == "STARTED") {
        requestInfo = App.BackgroundOperationsController.CommandContexts.START_SERVICE.format(this.get('content.serviceName'));
      } else {
        requestInfo = App.BackgroundOperationsController.CommandContexts.STOP_SERVICE.format(this.get('content.serviceName'));
      }

      App.ajax.send({
        'name': 'service.item.start_stop',
        'sender': this,
        'success': 'startStopPopupSuccessCallback',
        'error': 'startStopPopupErrorCallback',
        'data': {
          'requestInfo': requestInfo,
          'serviceName': this.get('content.serviceName').toUpperCase(),
          'state': serviceHealth,
          'query': query
        }
      });
      this.set('isStopDisabled', true);
      this.set('isStartDisabled', true);
    },

    /**
     * On click callback for <code>start service</code> button
     * @param event
     */
    startService: function (event) {
      this.startStopPopup(event, App.HostComponentStatus.started);
    },

    /**
     * On click callback for <code>stop service</code> button
     * @param event
     */
    stopService: function (event) {
      this.startStopPopup(event, App.HostComponentStatus.stopped);
    },

    /**
     * On click callback for <code>run rebalancer</code> button
     * @param event
     */
    runRebalancer: function (event) {
      var self = this;
      App.showConfirmationPopup(function() {
        self.content.set('runRebalancer', true);
        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      });
    },

    /**
     * On click callback for <code>run compaction</code> button
     * @param event
     */
    runCompaction: function (event) {
      var self = this;
      App.showConfirmationPopup(function() {
        self.content.set('runCompaction', true);
        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      });
    },

    /**
     * On click callback for <code>run smoke test</code> button
     * @param event
     */
    runSmokeTest: function (event) {
      var self = this;
      if (this.get('content.serviceName') === 'MAPREDUCE2' && !App.Service.find('YARN').get('isStarted')) {
        App.showAlertPopup(Em.I18n.t('common.error'), Em.I18n.t('services.mapreduce2.smokeTest.requirement'));
        return;
      }
      App.showConfirmationFeedBackPopup(function(query) {
        self.runSmokeTestPrimary(query);
      });
    },

    restartAllHostComponents : function(serviceName) {
      var serviceDisplayName = this.get('content.displayName');
      var bodyMessage = Em.Object.create({
        confirmMsg: Em.I18n.t('services.service.restartAll.confirmMsg').format(serviceDisplayName),
        confirmButton: Em.I18n.t('services.service.restartAll.confirmButton'),
        additionalWarningMsg: this.get('content.passiveState') === 'OFF' ? Em.I18n.t('services.service.restartAll.warningMsg.turnOnMM').format(serviceDisplayName): null
       });
      App.showConfirmationFeedBackPopup(function(query) {
        batchUtils.restartAllServiceHostComponents(serviceName, false, query);
      }, bodyMessage);
    },

    turnOnOffPassive: function(label) {
      var self = this;
      var state = this.get('content.passiveState') == 'OFF' ? 'ON' : 'OFF';
      var onOff = state === 'ON' ? "On" : "Off";
      App.showConfirmationPopup(function() {
            self.turnOnOffPassiveRequest(state, label)
          },
          Em.I18n.t('hosts.passiveMode.popup').format(onOff,self.get('content.displayName'))
      );
    },

    rollingRestart: function(hostComponentName) {
      batchUtils.launchHostComponentRollingRestart(hostComponentName, this.get('content.displayName'), this.get('content.passiveState') === "ON", false, this.get('content.passiveState') === "ON");
    },

    turnOnOffPassiveRequest: function(state,message) {
      App.ajax.send({
        'name': 'service.item.passive',
        'sender': this,
        'data': {
          'requestInfo': message,
          'serviceName': this.get('content.serviceName').toUpperCase(),
          'passive_state': state
        },
        'success':'updateService'
      });
    },

    updateService: function(data, opt, params) {
      var self = this;
      App.router.get('clusterController').loadUpdatedStatus(function(){
        self.set('content.passiveState', params.passive_state);
        batchUtils.infoPassiveState(params.passive_state);
      });
    },

    runSmokeTestPrimary: function(query) {
      App.ajax.send({
        'name': 'service.item.smoke',
        'sender': this,
        'success':'runSmokeTestSuccessCallBack',
        'error':'runSmokeTestErrorCallBack',
        'data': {
          'serviceName': this.get('content.serviceName'),
          'displayName': this.get('content.displayName'),
          'actionName': this.get('content.serviceName') === 'ZOOKEEPER' ? 'ZOOKEEPER_QUORUM_SERVICE_CHECK' : this.get('content.serviceName') + '_SERVICE_CHECK',
          'query': query
        }
      });
    },

    runSmokeTestSuccessCallBack: function (data, ajaxOptions, params) {
      if (data.Requests.id) {
        // load data (if we need to show this background operations popup) from persist
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          params.query.set('status', 'SUCCESS');
          if (initValue) {
            App.router.get('backgroundOperationsController').showPopup();
          }
        });
      }
      else {
        params.query.set('status', 'FAIL');
        console.warn('error during runSmokeTestSuccessCallBack');
      }
    },
    runSmokeTestErrorCallBack: function (request, ajaxOptions, error, opt, params) {
      params.query.set('status', 'FAIL');
    },

    /**
     * On click callback for <code>Reassign <master component></code> button
     * @param hostComponent
     */
    reassignMaster: function (hostComponent) {
      var component = App.HostComponent.find().findProperty('componentName', hostComponent);
      console.log('In Reassign Master', hostComponent);
      var reassignMasterController = App.router.get('reassignMasterController');
      reassignMasterController.saveComponentToReassign(component);
      reassignMasterController.getSecurityStatus();
      reassignMasterController.setCurrentStep('1');
      App.router.transitionTo('reassign');
    },

    /**
     * On click callback for <code>action</code> dropdown menu
     * Calls runSmokeTest, runRebalancer, runCompaction or reassignMaster depending on context
     * @param event
     */
    doAction: function (event) {
      if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
        return;
      }
      var methodName = event.context.action;
      var context = event.context.context;
      if (methodName) {
        this[methodName](context);
      }
    },

    /**
     * Restart clients host components to apply config changes
     */
    refreshConfigs: function() {
      var self = this;
      if (this.get('content.isClientsOnly')) {
        App.showConfirmationFeedBackPopup(function(query) {
          batchUtils.restartHostComponents(self.get('content.hostComponents'), Em.I18n.t('rollingrestart.context.allForSelectedService').format(self.get('content.serviceName')), query);
        });
      }
    },

    setStartStopState: function () {
      var serviceName = this.get('content.serviceName');
      var backgroundOperations = App.router.get('backgroundOperationsController.services');
      if (backgroundOperations.length > 0) {
        for (var i = 0; i < backgroundOperations.length; i++) {
          if (backgroundOperations[i].isRunning &&
              (backgroundOperations[i].dependentService === "ALL_SERVICES" ||
               backgroundOperations[i].dependentService === serviceName)) {
            this.set('isPending', true);
            return;
          }
        }
        this.set('isPending', false);
      } else {
        this.set('isPending', true);
      }
    }.observes('App.router.backgroundOperationsController.serviceTimestamp'),

    isServiceRestartable: function() {
      return this.get('content.serviceName') !== "FLUME";
    }.property('content.serviceName'),

    isStartDisabled: function () {
      if(this.get('isPending')) return true;
      return !(this.get('content.healthStatus') == 'red');
    }.property('content.healthStatus','isPending'),

    isStopDisabled: function () {
      if(this.get('isPending')) return true;
      if (App.get('isHaEnabled') && this.get('content.serviceName') == 'HDFS' && this.get('content.hostComponents').filterProperty('componentName', 'NAMENODE').someProperty('workStatus', App.HostComponentStatus.started)) {
        return false;
      }
      return (this.get('content.healthStatus') != 'green');
    }.property('content.healthStatus','isPending'),

    isPending:true

  });
  
});
window.require.register("controllers/main/service/manage_config_groups_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var hostsManagement = require('utils/hosts');
  var componentHelper = require('utils/component');
  var serviceComponents = require('data/service_components');

  App.ManageConfigGroupsController = Em.Controller.extend({
    name: 'manageConfigGroupsController',

    isLoaded: false,

    isInstaller: false,

    serviceName: null,

    configGroups: [],

    originalConfigGroups: [],

    selectedConfigGroup: null,

    selectedHosts: [],

    resortConfigGroup: function() {
      var configGroups = Ember.copy(this.get('configGroups'));
      if(configGroups.length < 2){
        return;
      }
      var defaultConfigGroup = configGroups.findProperty('isDefault');
      configGroups.removeObject(defaultConfigGroup);
      var sorted = [defaultConfigGroup].concat(configGroups.sortProperty('name'));

      this.removeObserver('configGroups.@each.name', this, 'resortConfigGroup');
      this.set('configGroups', sorted);
      this.addObserver('configGroups.@each.name', this, 'resortConfigGroup');
    }.observes('configGroups.@each.name'),

    loadConfigGroups: function (serviceName) {
      if (this.get('isInstaller')) {
        this.set('serviceName', serviceName);
        var configGroups = this.copyConfigGroups(App.router.get('wizardStep7Controller.selectedService.configGroups'));
        var originalConfigGroups = this.copyConfigGroups(configGroups);
        this.set('configGroups', configGroups);
        this.set('originalConfigGroups', originalConfigGroups);
        this.set('isLoaded', true);
      } else {
        this.set('serviceName', serviceName);
        App.ajax.send({
          name: 'service.load_config_groups',
          data: {
            serviceName: serviceName
          },
          sender: this,
          success: 'onLoadConfigGroupsSuccess',
          error: 'onLoadConfigGroupsError'
        });
      }
    },

    onLoadConfigGroupsSuccess: function (data) {
      var usedHosts = [];
      var unusedHosts = [];
      var serviceName = this.get('serviceName');
      var defaultConfigGroup = App.ConfigGroup.create({
        name: App.Service.DisplayNames[serviceName] + " Default",
        description: "Default cluster level " + this.get('serviceName') + " configuration",
        isDefault: true,
        parentConfigGroup: null,
        service: this.get('content'),
        configSiteTags: [],
        serviceName: serviceName
      });
      if (data && data.items) {
        var groupToTypeToTagMap = {};
        var configGroups = [];
        data.items.forEach(function (configGroup) {
          configGroup = configGroup.ConfigGroup;
          var hostNames = configGroup.hosts.mapProperty('host_name');
          var newConfigGroup = App.ConfigGroup.create({
            id: configGroup.id,
            name: configGroup.group_name,
            description: configGroup.description,
            isDefault: false,
            parentConfigGroup: defaultConfigGroup,
            service: App.Service.find().findProperty('serviceName', configGroup.tag),
            hosts: hostNames,
            configSiteTags: [],
            properties: [],
            apiResponse: configGroup
          });
          usedHosts = usedHosts.concat(newConfigGroup.get('hosts'));
          configGroups.push(newConfigGroup);
          var newConfigGroupSiteTags = newConfigGroup.get('configSiteTags');
          configGroup.desired_configs.forEach(function (config) {
            newConfigGroupSiteTags.push(App.ConfigSiteTag.create({
              site: config.type,
              tag: config.tag
            }));
            if (!groupToTypeToTagMap[configGroup.group_name]) {
              groupToTypeToTagMap[configGroup.group_name] = {}
            }
            groupToTypeToTagMap[configGroup.group_name][config.type] = config.tag;
          });
        }, this);
        unusedHosts = App.Host.find().mapProperty('hostName');
        usedHosts.uniq().forEach(function (host) {
          unusedHosts = unusedHosts.without(host);
        }, this);
        defaultConfigGroup.set('childConfigGroups', configGroups);
        defaultConfigGroup.set('hosts', unusedHosts);
        var allGroups = [defaultConfigGroup].concat(configGroups);
        this.set('configGroups', allGroups);
        var originalGroups = this.copyConfigGroups(allGroups);
        this.set('originalConfigGroups', originalGroups);
        this.loadProperties(groupToTypeToTagMap);
        this.set('isLoaded', true);
      }
    },

    onLoadConfigGroupsError: function () {
      console.error('Unable to load config groups for service.');
    },

    loadProperties: function (groupToTypeToTagMap) {
      var typeTagToGroupMap = {};
      var urlParams = [];
      for (var group in groupToTypeToTagMap) {
        var overrideTypeTags = groupToTypeToTagMap[group];
        for (var type in overrideTypeTags) {
          var tag = overrideTypeTags[type];
          typeTagToGroupMap[type + "///" + tag] = group;
          urlParams.push('(type=' + type + '&tag=' + tag + ')');
        }
      }
      var params = urlParams.join('|');
      if (urlParams.length) {
        App.ajax.send({
          name: 'config.host_overrides',
          sender: this,
          data: {
            params: params,
            typeTagToGroupMap: typeTagToGroupMap
          },
          success: 'onLoadPropertiesSuccess'
        });
      }
    },

    onLoadPropertiesSuccess: function (data, opt, params) {
      data.items.forEach(function (configs) {
        var typeTagConfigs = [];
        App.config.loadedConfigurationsCache[configs.type + "_" + configs.tag] = configs.properties;
        var group = params.typeTagToGroupMap[configs.type + "///" + configs.tag];
        for (var config in configs.properties) {
          typeTagConfigs.push({
            name: config,
            value: configs.properties[config]
          });
        }
        this.get('configGroups').findProperty('name', group).get('properties').pushObjects(typeTagConfigs);
      }, this);
    },

    showProperties: function () {
      var properies = this.get('selectedConfigGroup.propertiesList').htmlSafe();
      if (properies) {
        App.showAlertPopup(Em.I18n.t('services.service.config_groups_popup.properties'), properies);
      }
    },
    addHosts: function () {
      if (this.get('selectedConfigGroup.isAddHostsDisabled')){
        return false;
      }
      var availableHosts = this.get('selectedConfigGroup.availableHosts');
      var popupDescription = {
        header: Em.I18n.t('hosts.selectHostsDialog.title'),
        dialogMessage: Em.I18n.t('hosts.selectHostsDialog.message').format(App.Service.DisplayNames[this.get('serviceName')])
      };
      hostsManagement.launchHostsSelectionDialog(availableHosts, [], false, this.get('componentsForFilter'), this.addHostsCallback.bind(this), popupDescription);
    },

    /**
     * add hosts callback
     */
    addHostsCallback: function (selectedHosts) {
      var group = this.get('selectedConfigGroup');
      if (selectedHosts) {
        var defaultHosts = group.get('parentConfigGroup.hosts').slice();
        var configGroupHosts = group.get('hosts');
        selectedHosts.forEach(function (hostName) {
          configGroupHosts.pushObject(hostName);
          defaultHosts.removeObject(hostName);
        });
        group.set('parentConfigGroup.hosts', defaultHosts);
      }
    },

    /**
     * delete hosts from group
     */
    deleteHosts: function () {
      if (this.get('isDeleteHostsDisabled')) {
        return;
      }
      var groupHosts = this.get('selectedConfigGroup.hosts');
      var defaultGroupHosts = this.get('selectedConfigGroup.parentConfigGroup.hosts').slice();
      this.get('selectedHosts').slice().forEach(function (hostName) {
        defaultGroupHosts.pushObject(hostName);
        groupHosts.removeObject(hostName);
      });
      this.set('selectedConfigGroup.parentConfigGroup.hosts', defaultGroupHosts);
      this.set('selectedHosts', []);
    },

    isDeleteHostsDisabled: function () {
      var selectedConfigGroup = this.get('selectedConfigGroup');
      if (selectedConfigGroup) {
        return selectedConfigGroup.isDefault || this.get('selectedHosts').length === 0;
      }
      return true;
    }.property('selectedConfigGroup', 'selectedConfigGroup.hosts.length', 'selectedHosts.length'),

    /**
     * confirm delete config group
     */
    confirmDelete : function () {
      var self = this;
      App.showConfirmationPopup(function() {
        self.deleteConfigGroup();
      });
    },
    /**
     * add hosts to group
     * @return {Array}
     */
    componentsForFilter: function () {
      return serviceComponents.filterProperty('service_name', this.get('serviceName')).map(function (component) {
        return Em.Object.create({
          displayName: component.display_name,
          componentName: component.isClient ? 'CLIENT' : component.component_name,
          selected: false
        });
      });
    }.property('serviceName'),

    /**
     * delete selected config group
     */
    deleteConfigGroup: function () {
      var selectedConfigGroup = this.get('selectedConfigGroup');
      if (this.get('isDeleteGroupDisabled')) {
        return;
      }
      //move hosts of group to default group (available hosts)
      this.set('selectedHosts', selectedConfigGroup.get('hosts'));
      this.deleteHosts();
      this.get('configGroups').removeObject(selectedConfigGroup);
      this.set('selectedConfigGroup', this.get('configGroups').findProperty('isDefault'));
    },
    /**
     * rename new config group
     */
    renameConfigGroup: function () {
      if(this.get('selectedConfigGroup.isDefault')) {
        return;
      }
      var self = this;
      this.renameGroupPopup = App.ModalPopup.show({
        primary: Em.I18n.t('ok'),
        secondary: Em.I18n.t('common.cancel'),
        header: Em.I18n.t('services.service.config_groups.rename_config_group_popup.header'),
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/service/new_config_group')
        }),
        configGroupName: self.get('selectedConfigGroup.name'),
        configGroupDesc: self.get('selectedConfigGroup.description'),
        warningMessage: '',
        isDescriptionDirty: false,
        validate: function () {
          var warningMessage = '';
          var originalGroup = self.get('selectedConfigGroup');
          if (originalGroup.get('description') !== this.get('configGroupDesc') && !this.get('isDescriptionDirty')) {
            this.set('isDescriptionDirty', true);
          }
          if (originalGroup.get('name').trim() === this.get('configGroupName').trim()) {
            if (this.get('isDescriptionDirty')) {
              warningMessage = '';
            } else {
              warningMessage = Em.I18n.t("config.group.selection.dialog.err.name.exists");
            }
          } else {
            if (self.get('configGroups').mapProperty('name').contains(this.get('configGroupName').trim())) {
              warningMessage = Em.I18n.t("config.group.selection.dialog.err.name.exists");
            }
          }
          this.set('warningMessage', warningMessage);
        }.observes('configGroupName', 'configGroupDesc'),
        disablePrimary: function () {
          return !(this.get('configGroupName').trim().length > 0 && !this.get('warningMessage'));
        }.property('warningMessage', 'configGroupName', 'configGroupDesc'),
        onPrimary: function () {
          self.set('selectedConfigGroup.name', this.get('configGroupName'));
          self.set('selectedConfigGroup.description', this.get('configGroupDesc'));
          self.get('selectedConfigGroup.properties').forEach(function(property){
            property.set('group', self.get('selectedConfigGroup'));
          });
          this.hide();
        }
      });
      this.get('renameGroupPopup').validate();
    },

    /**
     * add new config group
     */
    addConfigGroup: function (duplicated) {
      duplicated = (duplicated === true);
      var self = this;
      this.addGroupPopup = App.ModalPopup.show({
        primary: Em.I18n.t('ok'),
        secondary: Em.I18n.t('common.cancel'),
        header: Em.I18n.t('services.service.config_groups.add_config_group_popup.header'),
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/service/new_config_group')
        }),
        configGroupName: duplicated ? self.get('selectedConfigGroup.name') + ' Copy' : "",
        configGroupDesc: duplicated ? self.get('selectedConfigGroup.description') + ' (Copy)' : "",
        warningMessage: '',
        didInsertElement: function(){
          this.validate();
        },
        validate: function () {
          var warningMessage = '';
          if (self.get('configGroups').mapProperty('name').contains(this.get('configGroupName').trim())) {
            warningMessage = Em.I18n.t("config.group.selection.dialog.err.name.exists");
          }
          this.set('warningMessage', warningMessage);
        }.observes('configGroupName'),
        disablePrimary: function () {
          return !(this.get('configGroupName').trim().length > 0 && !this.get('warningMessage'));
        }.property('warningMessage', 'configGroupName'),
        onPrimary: function () {
          var defaultConfigGroup = self.get('configGroups').findProperty('isDefault');
          var properties = [];
          var newConfigGroupData = App.ConfigGroup.create({
            id: null,
            name: this.get('configGroupName').trim(),
            description: this.get('configGroupDesc'),
            isDefault: false,
            parentConfigGroup: defaultConfigGroup,
            service: Em.Object.create({id: self.get('serviceName')}),
            hosts: [],
            configSiteTags: [],
            properties: []
          });
          if (duplicated) {
            self.get('selectedConfigGroup.properties').forEach(function(property) {
              var property = App.ServiceConfigProperty.create($.extend(false, {}, property));
              property.set('group', newConfigGroupData);
              properties.push(property);
            });
            newConfigGroupData.set('properties', properties);
          } else {
            newConfigGroupData.set('properties', []);
          }
          self.get('configGroups').pushObject(newConfigGroupData);
          defaultConfigGroup.get('childConfigGroups').pushObject(newConfigGroupData);
          this.hide();
        }
      });
    },

    duplicateConfigGroup: function() {
      this.addConfigGroup(true);
    },

    hostsModifiedConfigGroups: function () {
      if (!this.get('isLoaded')) {
        return false;
      }
      var groupsToClearHosts = [];
      var groupsToDelete = [];
      var groupsToSetHosts = [];
      var groupsToCreate = [];
      var groups = this.get('configGroups');
      var originalGroups = this.get('originalConfigGroups');
      var originalGroupsNames = originalGroups.mapProperty('name').without(originalGroups.findProperty('isDefault').get('name'));
      groups.forEach(function (group) {
        if (!group.get('isDefault')) {
          var originalGroup = originalGroups.findProperty('name', group.get('name'));
          if (originalGroup) {
            if (!(JSON.stringify(group.get('hosts').slice().sort()) === JSON.stringify(originalGroup.get('hosts').sort()))) {
              groupsToClearHosts.push(group.set('id', originalGroup.get('id')));
              if (group.get('hosts').length) {
                groupsToSetHosts.push(group.set('id', originalGroup.get('id')));
              }
            } else if (group.get('description') !== originalGroup.get('description')) {
              groupsToSetHosts.push(group.set('id', originalGroup.get('id')));
            }
            originalGroupsNames = originalGroupsNames.without(group.get('name'));
          } else {
            groupsToCreate.push(group);
          }
        }
      });
      originalGroupsNames.forEach(function (groupName) {
        groupsToDelete.push(originalGroups.findProperty('name', groupName));
      }, this);
      return {
        toClearHosts: groupsToClearHosts,
        toDelete: groupsToDelete,
        toSetHosts: groupsToSetHosts,
        toCreate: groupsToCreate
      };
    }.property('selectedConfigGroup.hosts.@each', 'selectedConfigGroup.hosts.length', 'selectedConfigGroup.description', 'configGroups', 'isLoaded'),

    isHostsModified: function () {
      var modifiedGroups = this.get('hostsModifiedConfigGroups');
      if (!this.get('isLoaded')) {
        return false;
      }
      return !!(modifiedGroups.toClearHosts.length || modifiedGroups.toSetHosts.length || modifiedGroups.toCreate.length || modifiedGroups.toDelete.length);
    }.property('hostsModifiedConfigGroups'),

    /**
     * copy config groups to manage popup to give user choice whether or not save changes
     * @param originGroups
     * @return {Array}
     */
    copyConfigGroups: function (originGroups) {
      var configGroups = [];
      var result = [];
      var defaultConfigGroup = App.ConfigGroup.create($.extend(true, {}, originGroups.findProperty('isDefault')));
      originGroups.forEach(function (configGroup) {
        if (!configGroup.get('isDefault')) {
          var copiedGroup = App.ConfigGroup.create($.extend(true, {}, configGroup));
          copiedGroup.set('parentConfigGroup', defaultConfigGroup);
          configGroups.pushObject(copiedGroup);
        }
      });
      defaultConfigGroup.set('childConfigGroups', configGroups.slice());
      configGroups.pushObject(defaultConfigGroup);
      configGroups.forEach(function (group) {
        var groupCopy = {};
        for (var prop in group) {
          if (group.hasOwnProperty(prop)) {
            groupCopy[prop] = group[prop];
          }
        }
        groupCopy.properties.forEach(function(property){
          property.set('group', group);
        });
        result.push(App.ConfigGroup.create(groupCopy));
      }, this);
      return result;
    }
  });
  
});
window.require.register("controllers/main/service/reassign/step1_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ReassignMasterWizardStep1Controller = Em.Controller.extend();

  
});
window.require.register("controllers/main/service/reassign/step2_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ReassignMasterWizardStep2Controller = App.WizardStep5Controller.extend({

    currentHostId: null,
    showCurrentHost: true,

    loadStep: function() {
      this._super();
      if(this.get('content.reassign.component_name') == "NAMENODE" && !this.get('content.masterComponentHosts').findProperty('component', "SECONDARY_NAMENODE")){
        this.set('showCurrentHost', false);
        this.set('componentToRebalance', 'NAMENODE');
        this.incrementProperty('rebalanceComponentHostsCounter');
      }else{
        this.set('showCurrentHost', true);
        this.rebalanceSingleComponentHosts(this.get('content.reassign.component_name'));
      }
    },

    loadComponents: function () {
      var components = this.get('components').filterProperty('isMaster', true);
      var masterComponents = this.get('content.masterComponentHosts');
      this.set('currentHostId', this.get('content').get('reassign').host_id);
      var componentNameToReassign = this.get('content').get('reassign').component_name;
      var result = [];
      masterComponents.forEach(function (master) {
        var color = "grey";
        if(master.component == componentNameToReassign){
          color = 'green';
        }
        result.push({
          component_name: master.component,
          display_name: App.format.role(master.component),
          selectedHost: master.hostName,
          isInstalled: true,
          serviceId: App.HostComponent.find().findProperty('componentName', master.component).get('serviceName'),
          isHiveCoHost: ['HIVE_METASTORE', 'WEBHCAT_SERVER'].contains(master.component),
          color: color
        });
      }, this);
      return result;
    },

    rebalanceSingleComponentHosts:function (componentName) {
      var currentComponents = this.get("selectedServicesMasters").filterProperty("component_name", componentName),
        componentHosts = currentComponents.mapProperty("selectedHost"),
        availableComponentHosts = [],
        preparedAvailableHosts = null;
      this.get("hosts").forEach(function (item) {
        if (this.get('currentHostId') !== item.get('host_name')) {
          availableComponentHosts.pushObject(item);
        }
      }, this);
      if (availableComponentHosts.length == 0) {
        return;
      }
      currentComponents.forEach(function (item) {
        preparedAvailableHosts = availableComponentHosts.slice(0);
        if (item.get('selectedHost') == this.get('currentHostId') && item.get('component_name') == this.get('content.reassign.component_name')) {
          item.set('selectedHost', preparedAvailableHosts.objectAt(0).host_name);
        }
        item.set("availableHosts", preparedAvailableHosts.sortProperty('host_name'));
      }, this);
    }
  });

  
});
window.require.register("controllers/main/service/reassign/step3_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ReassignMasterWizardStep3Controller = Em.Controller.extend();
  
});
window.require.register("controllers/main/service/reassign/step4_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ReassignMasterWizardStep4Controller = App.HighAvailabilityProgressPageController.extend({

    isReassign: true,

    commands: ['stopServices', 'createHostComponents', 'putHostComponentsInMaintenanceMode', 'reconfigure', 'installHostComponents', 'startZooKeeperServers', 'startNameNode', 'deleteHostComponents', 'startServices'],

    clusterDeployState: 'REASSIGN_MASTER_INSTALLING',

    multiTaskCounter: 0,

    hostComponents: [],
    restartYarnMRComponents: false,

    loadStep: function () {
      if (this.get('content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
        this.set('hostComponents', ['NAMENODE', 'ZKFC']);
      } else {
        this.set('hostComponents', [this.get('content.reassign.component_name')]);
      }
      this.set('restartYarnMRComponents', ['RESOURCEMANAGER', 'JOBTRACKER'].contains(this.get('content.reassign.component_name')));
      this.set('serviceName', [this.get('content.reassign.service_id')]);
      this._super();
    },

    initializeTasks: function () {
      var commands = this.get('commands');
      var currentStep = App.router.get('reassignMasterController.currentStep');
      var hostComponentsNames = '';

      this.get('hostComponents').forEach(function (comp, index) {
        hostComponentsNames += index ? '+' : '';
        hostComponentsNames += comp === 'ZKFC' ? comp : App.format.role(comp);
      }, this);

      for (var i = 0; i < commands.length; i++) {
        var TaskLabel = i === 3 ? this.get('serviceName') : hostComponentsNames; //For Reconfigure task, show serviceName
        var title = Em.I18n.t('services.reassign.step4.task' + i + '.title').format(TaskLabel);
        this.get('tasks').pushObject(Ember.Object.create({
          title: title,
          status: 'PENDING',
          id: i,
          command: commands[i],
          showRetry: false,
          showRollback: false,
          name: title,
          displayName: title,
          progress: 0,
          isRunning: false,
          hosts: []
        }));
      }

      if (this.get('content.hasManualSteps')) {
        if (this.get('content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
          // Only for reassign NameNode with HA enabled
          this.get('tasks').splice(7, 2);
        } else {
          this.get('tasks').splice(5, 4);
        }
      } else {
        this.get('tasks').splice(5, 2);
      }
    },

    hideRollbackButton: function () {
      var failedTask = this.get('tasks').findProperty('showRollback');
      if (failedTask) {
        failedTask.set('showRollback', false)
      }
    }.observes('tasks.@each.showRollback'),

    onComponentsTasksSuccess: function () {
      this.set('multiTaskCounter', this.get('multiTaskCounter') + 1);
      if (this.get('multiTaskCounter') >= this.get('hostComponents').length) {
        this.onTaskCompleted();
      }
    },

    stopServices: function () {
      if(this.get('restartYarnMRComponents')) {
        var list = App.Service.find().mapProperty("serviceName").without("HDFS").join(',');
        var conf = {
          name: 'reassign.stop_YMR2_services',
          sender: this,
          data: {servicesList: list},
          success: 'startPolling',
          error: 'onTaskError'
        };
        App.ajax.send(conf);
      } else {
        App.ajax.send({
          name: 'reassign.stop_services',
          sender: this,
          success: 'startPolling',
          error: 'onTaskError'
        });
      }
    },

    createHostComponents: function () {
      this.set('multiTaskCounter', 0);
      var hostComponents = this.get('hostComponents');
      var hostName = this.get('content.reassignHosts.target');
      for (var i = 0; i < hostComponents.length; i++) {
        this.createComponent(hostComponents[i], hostName);
      }
    },

    onCreateComponent: function () {
      this.onComponentsTasksSuccess();
    },

    putHostComponentsInMaintenanceMode: function () {
      this.set('multiTaskCounter', 0);
      var hostComponents = this.get('hostComponents');
      var hostName = this.get('content.reassignHosts.source');
      for (var i = 0; i < hostComponents.length; i++) {
        App.ajax.send({
          name: 'reassign.maintenance_mode',
          sender: this,
          data: {
            hostName: hostName,
            componentName: hostComponents[i]
          },
          success: 'onComponentsTasksSuccess',
          error: 'onTaskError'
        });
      }
    },

    installHostComponents: function () {
      this.set('multiTaskCounter', 0);
      var hostComponents = this.get('hostComponents');
      var hostName = this.get('content.reassignHosts.target');
      for (var i = 0; i < hostComponents.length; i++) {
        this.installComponent(hostComponents[i], hostName, hostComponents.length);
      }
    },

    reconfigure: function () {
      this.loadConfigsTags();
    },

    loadConfigsTags: function () {
      App.ajax.send({
        name: 'config.tags',
        sender: this,
        success: 'onLoadConfigsTags',
        error: 'onTaskError'
      });
    },

    onLoadConfigsTags: function (data) {
      var componentName = this.get('content.reassign.component_name');
      var urlParams = [];
      switch (componentName) {
        case 'NAMENODE':
          urlParams.push('(type=hdfs-site&tag=' + data.Clusters.desired_configs['hdfs-site'].tag + ')');
          urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
          if (App.Service.find().someProperty('serviceName', 'HBASE')) {
            urlParams.push('(type=hbase-site&tag=' + data.Clusters.desired_configs['hbase-site'].tag + ')');
          }
          break;
        case 'SECONDARY_NAMENODE':
          urlParams.push('(type=hdfs-site&tag=' + data.Clusters.desired_configs['hdfs-site'].tag + ')');
          urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
          break;
        case 'JOBTRACKER':
          urlParams.push('(type=mapred-site&tag=' + data.Clusters.desired_configs['mapred-site'].tag + ')');
          break;
        case 'RESOURCEMANAGER':
          urlParams.push('(type=yarn-site&tag=' + data.Clusters.desired_configs['yarn-site'].tag + ')');
          break;
      }
      App.ajax.send({
        name: 'reassign.load_configs',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'onLoadConfigs',
        error: 'onTaskError'
      });
    },

    configsSitesCount: null,

    configsSitesNumber: null,

    onLoadConfigs: function (data) {
      var isHadoop2Stack = App.get('isHadoop2Stack');
      var securityEnabled = this.get('content.securityEnabled');
      var componentName = this.get('content.reassign.component_name');
      var targetHostName = this.get('content.reassignHosts.target');
      var sourceHostName = this.get('content.reassignHosts.source');
      var configs = {};
      var componentDir = '';
      var secureConfigs = [];
      this.set('configsSitesNumber', data.items.length);
      this.set('configsSitesCount', 0);
      data.items.forEach(function (item) {
        configs[item.type] = item.properties;
      }, this);
      switch (componentName) {
        case 'NAMENODE':
          if (isHadoop2Stack) {
            if (App.get('isHaEnabled')) {
              var nameServices = configs['hdfs-site']['dfs.nameservices'];
              if (configs['hdfs-site']['dfs.namenode.http-address.' + nameServices + '.nn1'] === sourceHostName + ':50070') {
                configs['hdfs-site']['dfs.namenode.http-address.' + nameServices + '.nn1'] = targetHostName + ':50070';
                configs['hdfs-site']['dfs.namenode.https-address.' + nameServices + '.nn1'] = targetHostName + ':50470';
                configs['hdfs-site']['dfs.namenode.rpc-address.' + nameServices + '.nn1'] = targetHostName + ':8020';
              } else {
                configs['hdfs-site']['dfs.namenode.http-address.' + nameServices + '.nn2'] = targetHostName + ':50070';
                configs['hdfs-site']['dfs.namenode.https-address.' + nameServices + '.nn2'] = targetHostName + ':50470';
                configs['hdfs-site']['dfs.namenode.rpc-address.' + nameServices + '.nn2'] = targetHostName + ':8020';
              }
            } else {
              configs['hdfs-site']['dfs.namenode.http-address'] = targetHostName + ':50070';
              configs['hdfs-site']['dfs.namenode.https-address'] = targetHostName + ':50470';
              configs['core-site']['fs.defaultFS'] = 'hdfs://' + targetHostName + ':8020';
            }
            componentDir = configs['hdfs-site']['dfs.namenode.name.dir'];
          } else {
            componentDir = configs['hdfs-site']['dfs.name.dir'];
            configs['hdfs-site']['dfs.http.address'] = targetHostName + ':50070';
            configs['hdfs-site']['dfs.https.address'] = targetHostName + ':50470';
            configs['core-site']['fs.default.name'] = 'hdfs://' + targetHostName + ':8020';
          }
          if (!App.get('isHaEnabled')) {
            if (App.Service.find().someProperty('serviceName', 'HBASE')) {
              configs['hbase-site']['hbase.rootdir'] = configs['hbase-site']['hbase.rootdir'].replace(/\/\/[^\/]*/, '//' + targetHostName + ':8020');
            }
          }
          if (securityEnabled) {
            secureConfigs.push({keytab: configs['hdfs-site']['dfs.namenode.keytab.file'], principal: configs['hdfs-site']['dfs.namenode.kerberos.principal']});
            secureConfigs.push({keytab: configs['hdfs-site']['dfs.web.authentication.kerberos.keytab'], principal: configs['hdfs-site']['dfs.web.authentication.kerberos.principal']});
          }
          break;
        case 'SECONDARY_NAMENODE':
          if (isHadoop2Stack) {
            componentDir = configs['hdfs-site']['dfs.namenode.checkpoint.dir'];
            configs['hdfs-site']['dfs.namenode.secondary.http-address'] = targetHostName + ':50090';
          } else {
            componentDir = configs['core-site']['fs.checkpoint.dir'];
            configs['hdfs-site']['dfs.secondary.http.address'] = targetHostName + ':50090';
          }
          if (securityEnabled) {
            secureConfigs.push({keytab: configs['hdfs-site']['dfs.secondary.namenode.keytab.file'], principal: configs['hdfs-site']['dfs.secondary.namenode.kerberos.principal']});
            secureConfigs.push({keytab: configs['hdfs-site']['dfs.web.authentication.kerberos.keytab'], principal: configs['hdfs-site']['dfs.web.authentication.kerberos.principal']});
          }
          break;
        case 'JOBTRACKER':
          configs['mapred-site']['mapred.job.tracker.http.address'] = targetHostName + ':50030';
          configs['mapred-site']['mapred.job.tracker'] = targetHostName + ':50300';
          if (securityEnabled) {
            secureConfigs.push({keytab: configs['mapred-site']['mapreduce.jobtracker.keytab.file'], principal: configs['mapred-site']['mapreduce.jobtracker.kerberos.principal']});
          }
          break;
        case 'RESOURCEMANAGER':
          configs['yarn-site']['yarn.resourcemanager.address'] = targetHostName + ':8050';
          configs['yarn-site']['yarn.resourcemanager.admin.address'] = targetHostName + ':8141';
          configs['yarn-site']['yarn.resourcemanager.resource-tracker.address'] = targetHostName + ':8025';
          configs['yarn-site']['yarn.resourcemanager.scheduler.address'] = targetHostName + ':8030';
          configs['yarn-site']['yarn.resourcemanager.webapp.address'] = targetHostName + ':8088';
          configs['yarn-site']['yarn.resourcemanager.hostname'] = targetHostName;
          if (securityEnabled) {
            secureConfigs.push({keytab: configs['yarn-site']['yarn.resourcemanager.keytab'], principal: configs['yarn-site']['yarn.resourcemanager.principal']});
            secureConfigs.push({keytab: configs['yarn-site']['yarn.resourcemanager.webapp.spnego-keytab-file'], principal: configs['yarn-site']['yarn.resourcemanager.webapp.spnego-principal']});
          }

          break;
      }
      if (componentDir || secureConfigs.length) {
        App.router.get(this.get('content.controllerName')).saveComponentDir(componentDir);
        App.router.get(this.get('content.controllerName')).saveSecureConfigs(secureConfigs);
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('content.cluster.name'),
          clusterState: this.get('clusterDeployState'),
          wizardControllerName: this.get('content.controllerName'),
          localdb: App.db.data
        });
      }
      for (var site in configs) {
        if (!configs.hasOwnProperty(site)) continue;
        App.ajax.send({
          name: 'reassign.save_configs',
          sender: this,
          data: {
            siteName: site,
            properties: configs[site]
          },
          success: 'onSaveConfigs',
          error: 'onTaskError'
        });
      }
    },

    onSaveConfigs: function () {
      this.set('configsSitesCount', this.get('configsSitesCount') + 1);
      if (this.get('configsSitesCount') === this.get('configsSitesNumber')) {
        this.onTaskCompleted();
      }
    },

    startZooKeeperServers: function () {
      var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
      this.startComponent('ZOOKEEPER_SERVER', hostNames);
    },

    startNameNode: function () {
      var hostName = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName').without(this.get('content.reassignHosts.target'));
      this.startComponent('NAMENODE', hostName);
    },

    startServices: function () {
      if(this.get('restartYarnMRComponents')) {
        var list = App.Service.find().mapProperty("serviceName").without("HDFS").join(',');
        var conf = {
          name: 'reassign.start_YMR2_services',
          sender: this,
          data: {servicesList: list},
          success: 'startPolling',
          error: 'onTaskError'
        };
        App.ajax.send(conf);
      } else {
        App.ajax.send({
          name: 'reassign.start_services',
          sender: this,
          success: 'startPolling',
          error: 'onTaskError'
        });
      }
    },

    deleteHostComponents: function () {
      this.set('multiTaskCounter', 0);
      var hostComponents = this.get('hostComponents');
      var hostName = this.get('content.reassignHosts.source');
      for (var i = 0; i < hostComponents.length; i++) {
        App.ajax.send({
          name: 'reassign.remove_component',
          sender: this,
          data: {
            hostName: hostName,
            componentName: hostComponents[i]
          },
          success: 'onComponentsTasksSuccess',
          error: 'onDeleteHostComponentsError'
        });
      }
    },

    onDeleteHostComponentsError: function (error) {
      if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.NoSuchResourceException') !== -1) {
        this.onComponentsTasksSuccess();
      } else {
        this.onTaskError();
      }
    },

    done: function () {
      if (!this.get('isSubmitDisabled')) {
        this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
        if (this.get('content.hasManualSteps')) {
          App.router.send('next');
        } else {
          App.router.send('complete');
        }
      }
    }
  });
  
});
window.require.register("controllers/main/service/reassign/step5_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ReassignMasterWizardStep5Controller = Em.Controller.extend({

  });
  
});
window.require.register("controllers/main/service/reassign/step6_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ReassignMasterWizardStep6Controller = App.HighAvailabilityProgressPageController.extend({

    isReassign: true,

    commands: ['deleteHostComponents', 'startServices'],

    clusterDeployState: 'REASSIGN_MASTER_INSTALLING',

    multiTaskCounter: 0,

    hostComponents: [],

    loadStep: function () {
      if (this.get('content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
        this.set('hostComponents', ['NAMENODE', 'ZKFC']);
      } else {
        this.set('hostComponents', [this.get('content.reassign.component_name')]);
      }
      this._super();
    },

    initializeTasks: function () {
      var commands = this.get('commands');
      var hostComponentsNames = '';
      this.get('hostComponents').forEach(function (comp, index) {
        hostComponentsNames += index ? '+' : '';
        hostComponentsNames += comp === 'ZKFC' ? comp : App.format.role(comp);
      }, this);
      var currentStep = App.router.get('reassignMasterController.currentStep');
      for (var i = 0; i < commands.length; i++) {
        var title = Em.I18n.t('services.reassign.step6.task' + i + '.title').format(hostComponentsNames);
        this.get('tasks').pushObject(Ember.Object.create({
          title: title,
          status: 'PENDING',
          id: i,
          command: commands[i],
          showRetry: false,
          showRollback: false,
          name: title,
          displayName: title,
          progress: 0,
          isRunning: false,
          hosts: []
        }));
      }
    },

    hideRollbackButton: function () {
      var failedTask = this.get('tasks').findProperty('showRollback');
      if (failedTask) {
        failedTask.set('showRollback', false)
      }
    }.observes('tasks.@each.showRollback'),

    onComponentsTasksSuccess: function () {
      this.set('multiTaskCounter', this.get('multiTaskCounter') + 1);
      if (this.get('multiTaskCounter') >= this.get('hostComponents').length) {
        this.onTaskCompleted();
      }
    },

    startServices: function () {
      App.ajax.send({
        name: 'reassign.start_services',
        sender: this,
        success: 'startPolling',
        error: 'onTaskError'
      });
    },

    deleteHostComponents: function () {
      this.set('multiTaskCounter', 0);
      var hostComponents = this.get('hostComponents');
      var hostName = this.get('content.reassignHosts.source');
      for (var i = 0; i < hostComponents.length; i++) {
        App.ajax.send({
          name: 'reassign.remove_component',
          sender: this,
          data: {
            hostName: hostName,
            componentName: hostComponents[i]
          },
          success: 'onComponentsTasksSuccess',
          error: 'onDeleteHostComponentsError'
        });
      }
    },

    onDeleteHostComponentsError: function (error) {
      if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.NoSuchResourceException') !== -1) {
        this.onComponentsTasksSuccess();
      } else {
        this.onTaskError();
      }
    }
  });
  
});
window.require.register("controllers/main/service/reassign_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterController = App.WizardController.extend({

    name: 'reassignMasterController',

    totalSteps: 6,

    /**
     * Used for hiding back button in wizard
     */
    hideBackButton: true,

    /**
     * All wizards data will be stored in this variable
     *
     * cluster - cluster name
     * installOptions - ssh key, repo info, etc.
     * services - services list
     * hosts - list of selected hosts
     * slaveComponentHosts, - info about slave hosts
     * masterComponentHosts - info about master hosts
     * config??? - to be described later
     */
    content: Em.Object.create({
      cluster: null,
      hosts: null,
      installOptions: null,
      services: null,
      slaveComponentHosts: null,
      masterComponentHosts: null,
      serviceConfigProperties: null,
      advancedServiceConfig: null,
      controllerName: 'reassignMasterController',
      serviceName: 'MISC',
      hdfsUser: "hdfs",
      group: "hadoop",
      reassign: null,
      componentsWithManualCommands: ['NAMENODE', 'SECONDARY_NAMENODE'],
      hasManualSteps: false,
      securityEnabled: false
    }),

    addManualSteps: function () {
      this.set('content.hasManualSteps', this.get('content.componentsWithManualCommands').contains(this.get('content.reassign.component_name')) || this.get('content.securityEnabled'));
    }.observes('content.reassign.component_name', 'content.securityEnabled'),

    getSecurityStatus: function () {
      if (App.testMode) {
        this.set('securityEnabled', !App.testEnableSecurity);
      } else {
        //get Security Status From Server
        App.ajax.send({
          name: 'config.tags.sync',
          sender: this,
          success: 'getSecurityStatusSuccessCallback',
          error: 'errorCallback'
        });
      }
    },

    errorCallback: function () {
      console.error('Cannot get security status from server');
    },

    getSecurityStatusSuccessCallback: function (data) {
      var configs = data.Clusters.desired_configs;
      if ('global' in configs) {
        this.getServiceConfigsFromServer(configs['global'].tag);
      }
      else {
        console.error('Cannot get security status from server');
      }
    },

    getServiceConfigsFromServer: function (tag) {
      var tags = [
        {
          siteName: "global",
          tagName: tag
        }
      ];
      var data = App.router.get('configurationController').getConfigsByTags(tags);
      var configs = data.findProperty('tag', tag).properties;
      var result = configs && (configs['security_enabled'] === 'true' || configs['security_enabled'] === true);
      this.saveSecurityEnabled(result);
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: 'DEFAULT',
        wizardControllerName: 'reassignMasterController',
        localdb: App.db.data
      });
    },

    /**
     * return new object extended from clusterStatusTemplate
     * @return Object
     */
    getCluster: function () {
      return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
    },

    /**
     * Load services data from server.
     */
    loadServicesFromServer: function () {
      var displayOrderConfig = require('data/services');
      var apiUrl = App.get('stack2VersionURL');
      var apiService = this.loadServiceComponents(displayOrderConfig, apiUrl);
      //
      apiService.forEach(function (item, index) {
        apiService[index].isSelected = App.Service.find().someProperty('id', item.serviceName);
        apiService[index].isDisabled = apiService[index].isSelected;
        apiService[index].isInstalled = apiService[index].isSelected;
      });
      this.set('content.services', apiService);
      App.db.setService(apiService);
    },

    /**
     * Load confirmed hosts.
     * Will be used at <code>Assign Masters(step5)</code> step
     */
    loadConfirmedHosts: function () {
      var hosts = App.db.getHosts();
      if (!hosts || !hosts.length) {
        var hosts = {};

        App.Host.find().forEach(function (item) {
          hosts[item.get('id')] = {
            name: item.get('id'),
            cpu: item.get('cpu'),
            memory: item.get('memory'),
            disk_info: item.get('diskInfo'),
            bootStatus: "REGISTERED",
            isInstalled: true
          };
        });
        App.db.setHosts(hosts);
      }

      this.set('content.hosts', hosts);
      console.log('ReassignMasterController.loadConfirmedHosts: loaded hosts', hosts);
    },

    /**
     * Load master component hosts data for using in required step controllers
     */
    loadMasterComponentHosts: function () {
      var masterComponentHosts = App.db.getMasterComponentHosts();
      if (!masterComponentHosts) {
        masterComponentHosts = [];
        App.HostComponent.find().filterProperty('isMaster', true).forEach(function (item) {
          masterComponentHosts.push({
            component: item.get('componentName'),
            hostName: item.get('host.hostName'),
            isInstalled: true
          })
        });

      }
      this.set("content.masterComponentHosts", masterComponentHosts);
      console.log("ReassignMasterController.loadMasterComponentHosts: loaded hosts ", masterComponentHosts);
    },

    /**
     * Load tasks statuses for step5 of Reassign Master Wizard to restore installation
     */
    loadTasksStatuses: function () {
      var statuses = App.db.getReassignTasksStatuses();
      this.set('content.tasksStatuses', statuses);
      console.log('ReassignMasterController.loadTasksStatuses: loaded statuses', statuses);
    },

    /**
     * save status of the cluster.
     * @param clusterStatus object with status,requestId fields.
     */
    saveClusterStatus: function (clusterStatus) {
      var oldStatus = this.toObject(this.get('content.cluster'));
      clusterStatus = jQuery.extend(oldStatus, clusterStatus);
      if (clusterStatus.requestId) {
        clusterStatus.requestId.forEach(function (requestId) {
          if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
            clusterStatus.oldRequestsId.push(requestId)
          }
        }, this);
      }
      this.set('content.cluster', clusterStatus);
      this.save('cluster');
    },

    /**
     * Save Master Component Hosts data to Main Controller
     * @param stepController App.WizardStep5Controller
     */
    saveMasterComponentHosts: function (stepController) {
      var obj = stepController.get('selectedServicesMasters');

      var masterComponentHosts = [];
      obj.forEach(function (_component) {
        masterComponentHosts.push({
          display_name: _component.get('display_name'),
          component: _component.get('component_name'),
          hostName: _component.get('selectedHost'),
          serviceId: _component.get('serviceId'),
          isInstalled: true
        });
      });
      App.db.setMasterComponentHosts(masterComponentHosts);
      this.set('content.masterComponentHosts', masterComponentHosts);
    },

    loadComponentToReassign: function () {
      var masterComponent = App.db.getMasterToReassign();
      if (masterComponent) {
        this.set('content.reassign', masterComponent);
      }
    },

    saveComponentToReassign: function (masterComponent) {
      var component = {
        component_name: masterComponent.get('componentName'),
        display_name: masterComponent.get('displayName'),
        service_id: masterComponent.get('service.serviceName'),
        host_id: masterComponent.get('host.hostName')
      };
      App.db.setMasterToReassign(component);
    },
    saveTasksStatuses: function (statuses) {
      App.db.setReassignTasksStatuses(statuses);
      this.set('content.tasksStatuses', statuses);
      console.log('ReassignMasterController.saveTasksStatuses: saved statuses', statuses);
    },

    loadRequestIds: function () {
      var requestIds = App.db.getReassignMasterWizardRequestIds();
      this.set('content.requestIds', requestIds);
    },

    saveRequestIds: function (requestIds) {
      App.db.setReassignMasterWizardRequestIds(requestIds);
      this.set('content.requestIds', requestIds);
    },

    saveLogs: function (logs) {
      App.db.setReassignMasterWizardLogs(logs);
      this.set('content.logs', logs);
    },

    loadLogs: function () {
      var logs = App.db.getReassignMasterWizardLogs();
      this.set('content.logs', logs);
    },

    saveComponentDir: function (componentDir) {
      App.db.setReassignMasterWizardComponentDir(componentDir);
      this.set('content.componentDir', componentDir);
    },

    loadComponentDir: function () {
      var componentDir = App.db.getReassignMasterWizardComponentDir();
      this.set('content.componentDir', componentDir);
    },

    saveReassignHosts: function (reassignHosts) {
      App.db.setReassignMasterWizardReassignHosts(reassignHosts);
      this.set('content.reassignHosts', reassignHosts);
    },

    loadReassignHosts: function () {
      var reassignHosts = App.db.getReassignMasterWizardReassignHosts();
      this.set('content.reassignHosts', reassignHosts);
    },


    saveSecurityEnabled: function (securityEnabled) {
      this.setDBProperty('securityEnabled', securityEnabled);
      this.set('content.securityEnabled', securityEnabled);
    },

    loadSecurityEnabled: function () {
      var securityEnabled = this.getDBProperty('securityEnabled');
      this.set('content.securityEnabled', securityEnabled);
    },

    saveSecureConfigs: function (secureConfigs) {
      this.setDBProperty('secureConfigs', secureConfigs);
      this.set('content.secureConfigs', secureConfigs);
    },

    loadSecureConfigs: function () {
      var secureConfigs = this.getDBProperty('secureConfigs');
      this.set('content.secureConfigs', secureConfigs);
    },

    /**
     * Load data for all steps until <code>current step</code>
     */
    loadAllPriorSteps: function () {
      var step = this.get('currentStep');
      switch (step) {
        case '6':
        case '5':
          this.loadSecureConfigs();
          this.loadComponentDir();
        case '4':
          this.loadTasksStatuses();
          this.loadRequestIds();
          this.loadLogs();
        case '3':
          this.loadReassignHosts();
        case '2':
          this.loadServicesFromServer();
          this.loadMasterComponentHosts();
          this.loadConfirmedHosts();
        case '1':
          this.loadComponentToReassign();
          this.load('cluster');
      }
    },

    /**
     * Remove all loaded data.
     * Created as copy for App.router.clearAllSteps
     */
    clearAllSteps: function () {
      this.clearInstallOptions();
      // clear temporary information stored during the install
      this.set('content.cluster', this.getCluster());
    },

    /**
     * Clear all temporary data
     */
    finish: function () {
      this.setCurrentStep('1');
      this.clearAllSteps();
      this.clearStorageData();
      App.router.get('updateController').updateAll();
    }

  });
  
});
window.require.register("controllers/main/views", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainViewsController = Em.Controller.extend({
    name:'mainViewsController'
  })
});
window.require.register("controllers/wizard", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  require('models/host');

  App.WizardController = Em.Controller.extend({

    isStepDisabled: null,

    init: function () {
      this.set('isStepDisabled', []);
      this.clusters = App.Cluster.find();
      this.get('isStepDisabled').pushObject(Ember.Object.create({
        step: 1,
        value: false
      }));
      for (var i = 2; i <= this.get('totalSteps'); i++) {
        this.get('isStepDisabled').pushObject(Ember.Object.create({
          step: i,
          value: true
        }));
      }
    },

    slaveComponents: function () {
      return require('data/service_components').filter(function (component) {
        return !(component.isClient || component.isMaster ||
          ["GANGLIA_MONITOR", "DASHBOARD", "MYSQL_SERVER"].contains(component.component_name));
      });
    }.property(),

    dbNamespace: function(){
      return this.get('name').capitalize().replace('Controller', "");
    }.property('name'),
    /**
     * get property from local storage
     * @param key
     * @return {*}
     */
    getDBProperty: function(key){
      return App.db.get(this.get('dbNamespace'), key);
    },
    /**
     * set property to local storage
     * @param key
     * @param value
     */
    setDBProperty: function(key, value){
      App.db.set(this.get('dbNamespace'), key, value);
    },

    allHosts: App.Host.find(),

    setStepsEnable: function () {
      for (var i = 1; i <= this.totalSteps; i++) {
        var step = this.get('isStepDisabled').findProperty('step', i);
        if (i <= this.get('currentStep')) {
          step.set('value', false);
        } else {
          step.set('value', true);
        }
      }
    }.observes('currentStep'),

    setLowerStepsDisable: function (stepNo) {
      for (var i = 1; i < stepNo; i++) {
        var step = this.get('isStepDisabled').findProperty('step', i);
        step.set('value', true);
      }
    },

    /**
     * Set current step to new value.
     * Method moved from App.router.setInstallerCurrentStep
     * @param currentStep
     * @param completed
     */
    currentStep: function () {
      return App.get('router').getWizardCurrentStep(this.get('name').substr(0, this.get('name').length - 10));
    }.property(),

    /**
     * Set current step to new value.
     * Method moved from App.router.setInstallerCurrentStep
     * @param currentStep
     * @param completed
     */
    setCurrentStep: function (currentStep, completed) {
      App.db.setWizardCurrentStep(this.get('name').substr(0, this.get('name').length - 10), currentStep, completed);
      this.set('currentStep', currentStep);
    },

    clusters: null,

    isStep0: function () {
      return this.get('currentStep') == 0;
    }.property('currentStep'),

    isStep1: function () {
      return this.get('currentStep') == 1;
    }.property('currentStep'),

    isStep2: function () {
      return this.get('currentStep') == 2;
    }.property('currentStep'),

    isStep3: function () {
      return this.get('currentStep') == 3;
    }.property('currentStep'),

    isStep4: function () {
      return this.get('currentStep') == 4;
    }.property('currentStep'),

    isStep5: function () {
      return this.get('currentStep') == 5;
    }.property('currentStep'),

    isStep6: function () {
      return this.get('currentStep') == 6;
    }.property('currentStep'),

    isStep7: function () {
      return this.get('currentStep') == 7;
    }.property('currentStep'),

    isStep8: function () {
      return this.get('currentStep') == 8;
    }.property('currentStep'),

    isStep9: function () {
      return this.get('currentStep') == 9;
    }.property('currentStep'),

    isStep10: function () {
      return this.get('currentStep') == 10;
    }.property('currentStep'),

    gotoStep: function (step) {
      if (this.get('isStepDisabled').findProperty('step', step).get('value') !== false) {
        return false;
      }
      // if going back from Step 9 in Install Wizard, delete the checkpoint so that the user is not redirected
      // to Step 9
      if (this.get('content.controllerName') == 'installerController' && this.get('currentStep') === '9' && step < 9) {
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'CLUSTER_NOT_CREATED_1',
          wizardControllerName: 'installerController',
          localdb: App.db.data
        });
      }
      if ((this.get('currentStep') - step) > 1) {
        App.ModalPopup.show({
          header: Em.I18n.t('installer.navigation.warning.header'),
          onPrimary: function () {
            App.router.send('gotoStep' + step);
            this.hide();
          },
          body: "If you proceed to go back to Step " + step + ", you will lose any changes you have made beyond this step"
        });
      } else {
        App.router.send('gotoStep' + step);
      }
      return true;
    },

    gotoStep0: function () {
      this.gotoStep(0);
    },

    gotoStep1: function () {
      this.gotoStep(1);
    },

    gotoStep2: function () {
      this.gotoStep(2);
    },

    gotoStep3: function () {
      this.gotoStep(3);
    },

    gotoStep4: function () {
      this.gotoStep(4);
    },

    gotoStep5: function () {
      this.gotoStep(5);
    },

    gotoStep6: function () {
      this.gotoStep(6);
    },

    gotoStep7: function () {
      this.gotoStep(7);
    },

    gotoStep8: function () {
      this.gotoStep(8);
    },

    gotoStep9: function () {
      this.gotoStep(9);
    },

    gotoStep10: function () {
      this.gotoStep(10);
    },

    /**
     * Initialize host status info for step9
     */
    setInfoForStep9: function () {

      var hostInfo = this.getDBProperty('hosts');
      for (var index in hostInfo) {
        hostInfo[index].status = "pending";
        hostInfo[index].message = 'Waiting';
        hostInfo[index].logTasks = [];
        hostInfo[index].tasks = [];
        hostInfo[index].progress = '0';
      }
      this.setDBProperty('hosts', hostInfo);
    },

    /**
     * Remove all data for installOptions step
     */
    clearInstallOptions: function () {
      var installOptions = jQuery.extend({}, this.get('installOptionsTemplate'));
      this.set('content.installOptions', installOptions);
      this.setDBProperty('installOptions', installOptions);
      this.set('content.hosts', []);
      this.setDBProperty('hosts', []);
    },

    toObject: function (object) {
      var result = {};
      for (var i in object) {
        if (object.hasOwnProperty(i)) {
          result[i] = object[i];
        }
      }
      return result;
    },

    /**
     * save status of the cluster. This is called from step8 and step9 to persist install and start requestId
     * @param clusterStatus object with status, isCompleted, requestId, isInstallError and isStartError field.
     */
    saveClusterStatus: function (clusterStatus) {
      var oldStatus = this.toObject(this.get('content.cluster'));
      clusterStatus = jQuery.extend(oldStatus, clusterStatus);
      if (clusterStatus.requestId &&
        clusterStatus.oldRequestsId.indexOf(clusterStatus.requestId) === -1) {
        clusterStatus.oldRequestsId.push(clusterStatus.requestId);
      }
      this.set('content.cluster', clusterStatus);
      this.setDBProperty('cluster', clusterStatus);
    },

    /**
     * Invoke installation of selected services to the server and saves the request id returned by the server.
     * @param isRetry
     */
    installServices: function (isRetry) {

      // clear requests since we are installing services
      // and we don't want to get tasks for previous install attempts
      this.set('content.cluster.oldRequestsId', []);
      var clusterName = this.get('content.cluster.name');
      var data;
      var name;
      if (isRetry) {
        name = 'wizard.install_services.installer_controller.is_retry';
        data = '{"RequestInfo": {"context" :"' + Em.I18n.t('requestInfo.installComponents') + '"}, "Body": {"HostRoles": {"state": "INSTALLED"}}}';
      }
      else {
        name = 'wizard.install_services.installer_controller.not_is_retry';
        data = '{"RequestInfo": {"context" :"' + Em.I18n.t('requestInfo.installServices') + '"}, "Body": {"ServiceInfo": {"state": "INSTALLED"}}}';
      }

      App.ajax.send({
        name: name,
        sender: this,
        data: {
          data: data,
          cluster: clusterName
        },
        success: 'installServicesSuccessCallback',
        error: 'installServicesErrorCallback'
      });
    },

    installServicesSuccessCallback: function (jsonData) {
      var installStartTime = App.dateTime();
      console.log("TRACE: In success function for the installService call");
      if (jsonData) {
        var requestId = jsonData.Requests.id;
        console.log('requestId is: ' + requestId);
        var clusterStatus = {
          status: 'PENDING',
          requestId: requestId,
          isInstallError: false,
          isCompleted: false,
          installStartTime: installStartTime
        };
        this.saveClusterStatus(clusterStatus);
      } else {
        console.log('ERROR: Error occurred in parsing JSON data');
      }
    },

    installServicesErrorCallback: function (request, ajaxOptions, error) {
      console.log("TRACE: In error function for the installService call");
      console.log("TRACE: error code status is: " + request.status);
      console.log('Error message is: ' + request.responseText);
      var clusterStatus = {
        status: 'PENDING',
        requestId: this.get('content.cluster.requestId'),
        isInstallError: true,
        isCompleted: false
      };
      this.saveClusterStatus(clusterStatus);
      App.showAlertPopup(Em.I18n.t('common.errorPopup.header'), request.responseText);
    },

    bootstrapRequestId: null,

    /*
     Bootstrap selected hosts.
     */
    launchBootstrap: function (bootStrapData) {
      App.ajax.send({
        name: 'wizard.launch_bootstrap',
        sender: this,
        data: {
          bootStrapData: bootStrapData
        },
        success: 'launchBootstrapSuccessCallback',
        error: 'launchBootstrapErrorCallback'
      });

      return this.get('bootstrapRequestId');
    },

    launchBootstrapSuccessCallback: function (data) {
      console.log("TRACE: POST bootstrap succeeded");
      this.set('bootstrapRequestId', data.requestId);
    },

    launchBootstrapErrorCallback: function () {
      console.log("ERROR: POST bootstrap failed");
      alert('Bootstrap call failed. Please try again.');
    },

    /**
     * Load <code>content.<name></code> variable from localStorage, if wasn't loaded before.
     * If you specify <code>reload</code> to true - it will reload it.
     * @param name
     * @param reload
     * @return {Boolean}
     */
    load: function (name, reload) {
      if (this.get('content.' + name) && !reload) {
        return false;
      }
      var result = this.getDBProperty(name);
      if (!result) {
        if (this['get' + name.capitalize()]) {
          result = this['get' + name.capitalize()]();
          this.setDBProperty(name, result);
          console.log(this.get('name') + ": created " + name, result);
        }
        else {
          console.debug('get' + name.capitalize(), ' not defined in the ' + this.get('name'));
        }
      }
      this.set('content.' + name, result);
      console.log(this.get('name') + ": loaded " + name, result);
    },

    save: function (name) {
      var value = this.toObject(this.get('content.' + name));
      this.setDBProperty(name, value);
      console.log(this.get('name') + ": saved " + name, value);
    },

    clear: function () {
      this.set('content', Ember.Object.create({
        'controllerName': this.get('content.controllerName')
      }));
      this.set('currentStep', 0);
      this.clearStorageData();
    },

    clusterStatusTemplate: {
      name: "",
      status: "PENDING",
      isCompleted: false,
      requestId: null,
      installStartTime: null,
      installTime: null,
      isInstallError: false,
      isStartError: false,
      oldRequestsId: []
    },

    clearStorageData: function () {
      this.setDBProperty('service',undefined); //not to use this data at AddService page
      this.setDBProperty('hosts', undefined);
      this.setDBProperty('masterComponentHosts', undefined);
      this.setDBProperty('slaveComponentHosts', undefined);
      this.setDBProperty('cluster', undefined);
      this.setDBProperty('allHostNames', undefined);
      this.setDBProperty('installOptions', undefined);
      this.setDBProperty('allHostNamesPattern', undefined);
    },

    installOptionsTemplate: {
      hostNames: "", //string
      manualInstall: false, //true, false
      useSsh: true, //bool
      javaHome: App.defaultJavaHome, //string
      localRepo: false, //true, false
      sshKey: "", //string
      bootRequestId: null, //string
      sshUser: "root" //string
    },

    loadedServiceComponents: null,

    /**
     * Generate serviceComponents as pr the stack definition  and save it to localdata
     * called form stepController step4WizardController
     */
    loadServiceComponents: function () {
      App.ajax.send({
        name: 'wizard.service_components',
        sender: this,
        data: {
          stackUrl: App.get('stack2VersionURL'),
          stackVersion: App.get('currentStackVersionNumber')
        },
        success: 'loadServiceComponentsSuccessCallback',
        error: 'loadServiceComponentsErrorCallback'
      });
      return this.get('loadedServiceComponents');
    },

    loadServiceComponentsSuccessCallback: function (jsonData) {
      var displayOrderConfig = require('data/services');
      console.log("TRACE: getService ajax call  -> In success function for the getServiceComponents call");
      console.log("TRACE: jsonData.services : " + jsonData.items);

      // Creating Model
      var Service = Ember.Object.extend({
        serviceName: null,
        displayName: null,
        isDisabled: true,
        isSelected: true,
        isInstalled: false,
        description: null,
        version: null
      });

      var data = [];
      // loop through all the service components
      for (var i = 0; i < displayOrderConfig.length; i++) {
        var entry = jsonData.items.findProperty("StackServices.service_name", displayOrderConfig[i].serviceName);
        if (entry) {
          var myService = Service.create({
            serviceName: entry.StackServices.service_name,
            displayName: displayOrderConfig[i].displayName,
            isDisabled: displayOrderConfig[i].isDisabled,
            isSelected: displayOrderConfig[i].isSelected,
            canBeSelected: displayOrderConfig[i].canBeSelected,
            isInstalled: false,
            isHidden: displayOrderConfig[i].isHidden,
            description: entry.StackServices.comments,
            version: entry.StackServices.service_version
          });

          data.push(myService);
        }
        else {
          console.warn('Service not found - ', displayOrderConfig[i].serviceName);
        }
      }

      this.set('loadedServiceComponents', data);
      console.log('TRACE: service components: ' + JSON.stringify(data));

    },

    loadServiceComponentsErrorCallback: function (request, ajaxOptions, error) {
      console.log("TRACE: STep5 -> In error function for the getServiceComponents call");
      console.log("TRACE: STep5 -> error code status is: " + request.status);
      console.log('Step8: Error message is: ' + request.responseText);
    },

    loadServicesFromServer: function () {
      var services = this.getDBProperty('service');
      var apiService = this.loadServiceComponents();
      this.set('content.services', apiService);
      this.setDBProperty('service',apiService);
    },
    /**
     * Load config groups from local DB
     */
    loadServiceConfigGroups: function () {
      var serviceConfigGroups = this.getDBProperty('serviceConfigGroups') || [];
      this.set('content.configGroups', serviceConfigGroups);
      console.log("InstallerController.configGroups: loaded config ", serviceConfigGroups);
    },

    registerErrPopup: function (header, message) {
      App.ModalPopup.show({
        header: header,
        secondary: false,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<p>{{view.message}}</p>'),
          message: message
        })
      });
    },

    /**
     * Save hosts that the user confirmed to proceed with from step 3
     * @param stepController App.WizardStep3Controller
     */
    saveConfirmedHosts: function (stepController) {
      var hostInfo = {};
      stepController.get('content.hosts').forEach(function (_host) {
        if (_host.bootStatus == 'REGISTERED') {
          hostInfo[_host.name] = {
            name: _host.name,
            cpu: _host.cpu,
            memory: _host.memory,
            disk_info: _host.disk_info,
            os_type: _host.os_type,
            os_arch: _host.os_arch,
            ip: _host.ip,
            bootStatus: _host.bootStatus,
            isInstalled: false
          };
        }
      });
      console.log('wizardController:saveConfirmedHosts: save hosts ', hostInfo);
      this.setDBProperty('hosts', hostInfo);
      this.set('content.hosts', hostInfo);
    },

    /**
     * Save data after installation to main controller
     * @param stepController App.WizardStep9Controller
     */
    saveInstalledHosts: function (stepController) {
      var hosts = stepController.get('hosts');
      var hostInfo = this.getDBProperty('hosts');

      for (var index in hostInfo) {
        hostInfo[index].status = "pending";
        var host = hosts.findProperty('name', hostInfo[index].name);
        if (host) {
          hostInfo[index].status = host.status;
          hostInfo[index].message = host.message;
          hostInfo[index].progress = host.progress;
        }
      }
      this.set('content.hosts', hostInfo);
      this.setDBProperty('hosts', hostInfo);
      console.log('wizardController:saveInstalledHosts: save hosts ', hostInfo);
    },

    /**
     * Save slaveHostComponents to main controller
     * @param stepController
     */
    saveSlaveComponentHosts: function (stepController) {

      var hosts = stepController.get('hosts');
      var headers = stepController.get('headers');

      var formattedHosts = Ember.Object.create();
      headers.forEach(function (header) {
        formattedHosts.set(header.get('name'), []);
      });

      hosts.forEach(function (host) {

        var checkboxes = host.get('checkboxes');
        headers.forEach(function (header) {
          var cb = checkboxes.findProperty('title', header.get('label'));
          if (cb.get('checked')) {
            formattedHosts.get(header.get('name')).push({
              hostName: host.hostName,
              group: 'Default',
              isInstalled: cb.get('isInstalled')
            });
          }
        });
      });

      var slaveComponentHosts = [];

      headers.forEach(function (header) {
        slaveComponentHosts.push({
          componentName: header.get('name'),
          displayName: header.get('label').replace(/\s/g, ''),
          hosts: formattedHosts.get(header.get('name'))
        });
      });

      this.setDBProperty('slaveComponentHosts', slaveComponentHosts);
      console.log('wizardController.slaveComponentHosts: saved hosts', slaveComponentHosts);
      this.set('content.slaveComponentHosts', slaveComponentHosts);
    },

    /**
     * Return true if cluster data is loaded and false otherwise.
     * This is used for all wizard controllers except for installer wizard.
     */
    dataLoading: function () {
      var dfd = $.Deferred();
      this.connectOutlet('loading');
      if (App.router.get('clusterController.isLoaded')) {
        dfd.resolve();
      } else {
        var interval = setInterval(function () {
          if (App.router.get('clusterController.isLoaded')) {
            dfd.resolve();
            clearInterval(interval);
          }
        }, 50);
      }
      return dfd.promise();
    },

    /**
     * Return true if user data is loaded via App.MainServiceInfoConfigsController
     * This function is used in reassign master wizard right now.
     */

    usersLoading: function () {
      var self = this;
      var dfd = $.Deferred();
      var miscController = App.MainAdminMiscController.create({content: self.get('content')});
      miscController.loadUsers();
      var interval = setInterval(function () {
        if (miscController.get('dataIsLoaded')) {
          if (self.get("content.hdfsUser")) {
            self.set('content.hdfsUser', miscController.get('content.hdfsUser'));
          }
          dfd.resolve();
          clearInterval(interval);
        }
      }, 10);
      return dfd.promise();
    },

    /**
     * Save cluster status before going to deploy step
     * @param name cluster state. Unique for every wizard
     */
    saveClusterState: function (name) {
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: name,
        wizardControllerName: this.get('content.controllerName'),
        localdb: App.db.data
      });
    },

    /**
     * load advanced configs from server
     */
    loadAdvancedConfigs: function (dependentController) {
      var self = this;
      var counter = this.get('content.services').filterProperty('isSelected').length;
      var loadAdvancedConfigResult = [];
      dependentController.set('isAdvancedConfigLoaded', false);
      this.get('content.services').filterProperty('isSelected').mapProperty('serviceName').forEach(function (_serviceName) {
        App.config.loadAdvancedConfig(_serviceName, function (properties) {
          loadAdvancedConfigResult.pushObjects(properties);
          counter--;
          //pass configs to controller after last call is completed
          if (counter === 0) {
            self.set('content.advancedServiceConfig', loadAdvancedConfigResult);
            self.setDBProperty('advancedServiceConfig', loadAdvancedConfigResult);
            dependentController.set('isAdvancedConfigLoaded', true);
          }
        });
      }, this);
    },
    /**
     * Load serviceConfigProperties to model
     */
    loadServiceConfigProperties: function () {
      var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
      this.set('content.serviceConfigProperties', serviceConfigProperties);
      console.log("AddHostController.loadServiceConfigProperties: loaded config ", serviceConfigProperties);
    },
    /**
     * Save config properties
     * @param stepController Step7WizardController
     */
    saveServiceConfigProperties: function (stepController) {
      var serviceConfigProperties = [];
      var updateServiceConfigProperties = [];
      stepController.get('stepConfigs').forEach(function (_content) {

        if (_content.serviceName === 'YARN' && !App.supports.capacitySchedulerUi) {
          _content.set('configs', App.config.textareaIntoFileConfigs(_content.get('configs'), 'capacity-scheduler.xml'));
        }

        _content.get('configs').forEach(function (_configProperties) {
          var configProperty = {
            id: _configProperties.get('id'),
            name: _configProperties.get('name'),
            value: _configProperties.get('value'),
            defaultValue: _configProperties.get('defaultValue'),
            description: _configProperties.get('description'),
            serviceName: _configProperties.get('serviceName'),
            domain: _configProperties.get('domain'),
            filename: _configProperties.get('filename'),
            displayType: _configProperties.get('displayType'),
            isRequiredByAgent: _configProperties.get('isRequiredByAgent')
          };
          serviceConfigProperties.push(configProperty);
        }, this);
        // check for configs that need to update for installed services
        if (stepController.get('installedServiceNames') && stepController.get('installedServiceNames').contains(_content.get('serviceName'))) {
          // get only modified configs
          var configs = _content.get('configs').filterProperty('isNotDefaultValue').filter(function(config) {
            var notAllowed = ['masterHost', 'masterHosts', 'slaveHosts', 'slaveHost'];
            return !notAllowed.contains(config.get('displayType'));
          });
          // if modified configs detected push all service's configs for update
          if (configs.length)
            updateServiceConfigProperties = updateServiceConfigProperties.concat(serviceConfigProperties.filterProperty('serviceName',_content.get('serviceName')));
          // watch for properties that are not modified but have to be updated
          if (_content.get('configs').someProperty('forceUpdate')) {
            // check for already added modified properties
            if (!updateServiceConfigProperties.findProperty('serviceName', _content.get('serviceName'))) {
              updateServiceConfigProperties = updateServiceConfigProperties.concat(serviceConfigProperties.filterProperty('serviceName',_content.get('serviceName')));
            }
          }
        }
      }, this);
      this.setDBProperty('serviceConfigProperties', serviceConfigProperties);
      this.set('content.serviceConfigProperties', serviceConfigProperties);
      this.setDBProperty('configsToUpdate', updateServiceConfigProperties);
    },
    /**
     * save Config groups
     * @param stepController
     */
    saveServiceConfigGroups: function (stepController) {
      var serviceConfigGroups = [];
      var isForUpdate = false;
      stepController.get('stepConfigs').forEach(function (service) {
        // mark group of installed service
        if (service.get('selected') === false) isForUpdate = true;
        service.get('configGroups').forEach(function (configGroup) {
          var properties = [];
          configGroup.get('properties').forEach(function (property) {
            properties.push({
              isRequiredByAgent: property.get('isRequiredByAgent'),
              name: property.get('name'),
              value: property.get('value'),
              filename: property.get('filename')
            })
          });
          //configGroup copied into plain JS object to avoid Converting circular structure to JSON
          serviceConfigGroups.push({
            id: configGroup.get('id'),
            name: configGroup.get('name'),
            description: configGroup.get('description'),
            hosts: configGroup.get('hosts'),
            properties: properties,
            isDefault: configGroup.get('isDefault'),
            isForUpdate: isForUpdate,
            service: {id: configGroup.get('service.id')}
          });
        }, this)
      }, this);
      this.setDBProperty('serviceConfigGroups', serviceConfigGroups);
      this.set('content.configGroups', serviceConfigGroups);
    },
    /**
     * return slaveComponents bound to hosts
     * @return {Array}
     */
    getSlaveComponentHosts: function () {
      var components = this.get('slaveComponents');
      var result = [];
      var installedServices = App.Service.find().mapProperty('serviceName');
      var selectedServices = this.get('content.services').filterProperty('isSelected', true).mapProperty('serviceName');
      var installedComponentsMap = {};
      var uninstalledComponents = [];

      components.forEach(function (component) {
        if (installedServices.contains(component.service_name)) {
          installedComponentsMap[component.component_name] = [];
        } else if (selectedServices.contains(component.service_name)) {
          uninstalledComponents.push(component);
        }
      }, this);
      installedComponentsMap['HDFS_CLIENT'] = [];

      App.HostComponent.find().forEach(function (hostComponent) {
        if (installedComponentsMap[hostComponent.get('componentName')]) {
          installedComponentsMap[hostComponent.get('componentName')].push(hostComponent.get('host.id'));
        }
      }, this);

      for (var componentName in installedComponentsMap) {
        var name = (componentName === 'HDFS_CLIENT') ? 'CLIENT' : componentName;
        var component = {
          componentName: name,
          displayName: App.format.role(name),
          hosts: [],
          isInstalled: true
        };
        installedComponentsMap[componentName].forEach(function (hostName) {
          component.hosts.push({
            group: "Default",
            hostName: hostName,
            isInstalled: true
          });
        }, this);
        result.push(component);
      }

      uninstalledComponents.forEach(function (component) {
        var hosts = jQuery.extend(true, [], result.findProperty('componentName', 'DATANODE').hosts);
        hosts.setEach('isInstalled', false);
        result.push({
          componentName: component.component_name,
          displayName: App.format.role(component.component_name),
          hosts: hosts,
          isInstalled: false
        })
      });

      return result;
    }
  });
  
});
window.require.register("controllers/wizard/slave_component_groups_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  /**
   * Used to manage slave component config. User could create different settings for separate group
   * @type {*}
   */
  App.SlaveComponentGroupsController = Em.ArrayController.extend({

    name: 'slaveComponentGroupsController',

    contentBinding: 'App.router.wizardStep7Controller.slaveComponentHosts',

    stepConfigsBinding: 'App.router.wizardStep7Controller.stepConfigs',

    serviceBinding: 'App.router.wizardStep7Controller.selectedService',

    servicesBinding: 'App.router.wizardStep7Controller.content.services',

    clearStep: function () {

    },

    loadStep: function () {
      this.clearStep();
      this.loadGroups();
    },

    loadGroups: function () {

      this.get('stepConfigs').forEach(function (_serviceConfig) {
        var categoryConfig = _serviceConfig.get('configCategories');
        if (categoryConfig.someProperty('isForSlaveComponent', true)) {
          var slaveCategory = categoryConfig.findProperty('isForSlaveComponent', true);
          // this.get('content') -> Output of Step 6: Mapping of each slave component and set of hosts it runs on
          if (this.get('content')) {
            if (this.get('content').someProperty('componentName', slaveCategory.get('primaryName'))) {
              // component --> each column in Step 6 is a component ( slave component )
              var component = this.get('content').findProperty('componentName', slaveCategory.get('primaryName'));
              // slaveConfigs --> originally set as null in the class App.SlaveCategory in model/service_config.js
              var slaveConfigs = slaveCategory.get('slaveConfigs');
              
              slaveCategory.set('slaveConfigs', App.SlaveConfigs.create(component));
              var slaveGroups = [];
              if (component.groups) {
                component.groups.forEach(function (_group) {
                  slaveGroups.pushObject(_group);
                }, this);
                slaveCategory.set('slaveConfigs.groups', slaveGroups);
              }
              slaveCategory.set('slaveConfigs.componentName', component.componentName);
              slaveCategory.set('slaveConfigs.displayName', component.displayName);
              /*slaveCategory.set('slaveConfigs.groups.name', component.get('name'));
               slaveCategory.set('slaveConfigs.groups.index', component.get('index'));
               slaveCategory.set('slaveConfigs.groups.type', component.get('type'));
               slaveCategory.set('slaveConfigs.groups.active', component.get('active'));*/
              if (!slaveCategory.get('slaveConfigs.groups')) {
                slaveCategory.set('slaveConfigs.groups', []);
                var componentProperties = this.componentProperties(_serviceConfig.serviceName);
                var defaultGroup = {name: 'Default', index: 'default', type: 'default', active: true, properties: componentProperties};
                slaveCategory.get('slaveConfigs.groups').pushObject(App.Group.create(defaultGroup));
              }
            }
          }
        }
      }, this);
    },

    // returns key-value pairs i.e. all fields for slave component for this specific service.
    componentProperties: function (serviceName) {

      var serviceConfigs = App.config.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName);

      var configs = [];
      var componentName = null;
      switch (serviceName) {
        case 'HDFS':
          componentName = 'DataNode';
          break;
        case 'MAPREDUCE':
          componentName = 'TaskTracker';
          break;
        case 'HBASE':
          componentName = 'RegionServer';
      }
      var slaveConfigs = serviceConfigs.configs.filterProperty('category', componentName);
      slaveConfigs.forEach(function (_serviceConfigProperty) {
        var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);

        switch(serviceConfigProperty.name){
          case 'dfs_data_dir' :
            serviceConfigProperty.initialValue();
            break;
          case 'mapred_local_dir' :
            serviceConfigProperty.initialValue();
            break;
        }
        configs.pushObject(serviceConfigProperty);
        serviceConfigProperty.validate();
      }, this);
      return configs;
    },

    selectedComponentName: function () {
      switch (App.router.get('wizardStep7Controller.selectedService.serviceName')) {
        case 'HDFS':
          return { name: 'DATANODE',
            displayName: 'DataNode'};
        case 'MAPREDUCE':
          return { name: 'TASKTRACKER',
            displayName: 'TaskTracker'};

        case 'HBASE':
          return { name: 'HBASE_REGIONSERVER',
            displayName: 'RegionServer'};
        default:
          return null;
      }

    }.property('service'),

    selectedComponentDisplayName: function() {
      return App.format.role(this.get('selectedComponentName'));
    }.property('selectedComponentName'),

    selectedSlaveComponent: function () {
      var selectedComponentName = this.get('selectedComponentName') ? this.get('selectedComponentName').displayName : null;
      var configs = null;
      if (selectedComponentName) {
        App.router.get('wizardStep7Controller.stepConfigs').forEach(function (_serviceConfig) {
          var categoryConfig = _serviceConfig.get('configCategories');
          if (categoryConfig.someProperty('name', selectedComponentName)) {
            configs = categoryConfig.findProperty('name', selectedComponentName).get('slaveConfigs');
          }
        }, this);
      }
      return configs;
    }.property('selectedComponentName', 'stepConfigs.@each.configCategories', 'stepConfigs.@each.configCategories.@each.slaveConfigs'),

    hosts: function () {
      if (this.get('selectedSlaveComponent')) {
        return this.get('selectedSlaveComponent').get('hosts');
      }
    }.property('selectedSlaveComponent'),

    groups: function () {
      var hosts = this.get('hosts');
      if(hosts){
        return hosts.mapProperty('group').uniq();
      }
      return [];
    }.property('hosts'),

    componentGroups: function () {
      var component = this.get('selectedSlaveComponent');
      if (component && component.get('groups')) {
        return component.get('groups');
      }
      return [];
    }.property('selectedSlaveComponent', 'selectedSlaveComponent.groups', 'stepConfigs.@each.configCategories.@each.slaveConfigs.groups.@each.properties.@each.value'),


    getGroupsForDropDown: function () {
      return this.get('componentGroups').getEach('name');
    }.property('selectedComponentName', 'componentGroups.@each.name'),

    activeGroup: function () {
      var componentGroups = this.get('componentGroups');
      if (componentGroups) {
        var active = componentGroups.findProperty('active', true);
        if (active){
          return active;
        }
      }
      return null;
    }.property('componentGroups.@each.active', 'componentGroups.@each.name', 'componentGroups.@each.properties.@each.value'),


    /**
     * Show slave hosts to groups popup
     * @param event
     */
    showAddSlaveComponentGroup: function (event) {
      var componentName = event.context;
      var component = this.get('selectedSlaveComponent');
      App.ModalPopup.show({
        header: componentName + Em.I18n.t('installer.controls.slaveComponentGroups'),
        bodyClass: Ember.View.extend({
          controllerBinding: 'App.router.slaveComponentGroupsController',
          header: Em.I18n.t('installer.slaveComponentHostsPopup.header').format(this.get('selectedComponentDisplayName')),
          templateName: require('templates/wizard/slave_component_hosts_popup')
        }),
        onPrimary: function (event) {
          if (component.tempSelectedGroups && component.tempSelectedGroups.length) {
            component.tempSelectedGroups.forEach(function (item) {
              var changed = component.get('hosts').filterProperty('hostName', item.hostName);
              changed.setEach('group', item.groupName);
            })
          }
          delete component.tempSelectedGroups;
          this.hide();
        },
        onSecondary: function (event) {
          delete component.tempSelectedGroups;
          this.hide();
        },
        onClose: function (event) {
          delete component.tempSelectedGroups;
          this.hide();
        }
      });
    },

    /**
     * Utility method. Save temporary info about changes in <code>slave hosts to groups</code> popup
     * @param host
     * @param groupName
     */
    changeHostGroup: function (host, groupName) {
      var component = this.get('selectedSlaveComponent');
      if (component.tempSelectedGroups === undefined) {
        component.tempSelectedGroups = [];
      }
      var values = component.tempSelectedGroups.filterProperty('hostName', host.hostName);
      if (values.length === 0)
        component.tempSelectedGroups.pushObject({hostName: host.hostName, groupName: groupName});
      else
        values.setEach('groupName', groupName);

    },

    /**
     * add new group to component(click on button)
     */
    addSlaveComponentGroup: function () {
      var component = this.get('selectedSlaveComponent');
      var newGroupName = 'New Group';
      component.get('groups').setEach('active', false);
      var newGroups = component.get('groups').filterProperty('name', newGroupName);
      if (newGroups.length === 0){
        component.newGroupIndex = 0;
      }
      else {
        component.newGroupIndex = component.newGroupIndex || 0;
        this.checkGroupName();
        newGroupName = 'New Group ' + component.newGroupIndex;
      }
      var newGroup = {name: newGroupName, index: component.newGroupIndex, type: 'new', active: true, properties: this.componentProperties(App.router.get('wizardStep7Controller.selectedService.serviceName'))};
      component.groups.pushObject(App.Group.create(newGroup));
      $('.remove-group-error').hide();
    },

    checkGroupName: function () {
      var component = this.get('selectedSlaveComponent');
      component.newGroupIndex++;
      var newGroupName = 'New Group ' + component.newGroupIndex;
      var groups = component.get('groups').filterProperty('name', newGroupName);
      if (groups.length !== 0) {
        this.checkGroupName();
      }
    },

    /**
     * Onclick handler for <code>choose hosts for slave group</code> link
     * @param event
     */
    showEditSlaveComponentGroups: function (event) {
      this.showAddSlaveComponentGroup(event);
    },

    getHostsByGroup: function (group) {
      var hosts = this.get('hosts');
      if(hosts){
        return hosts.filterProperty('group', group.name);
      }
      return [];
    },

    /**
     * Change tab
     * @param event
     */
    showSlaveComponentGroup: function (event) {
      var component = this.get('selectedSlaveComponent');
      if(!component.groups){

      }
      component.get('groups').setEach('active', false);
      var group = component.get('groups').filterProperty('name', event.context.name);
      group.setEach('active', true);
      var assignedHosts = component.get('hosts').filterProperty('group', event.context.name);
      if (assignedHosts.length === 0) {
        $('.remove-group-error').hide();
      }
    },

    /**
     * Remove tab
     * @param event
     */
    removeSlaveComponentGroup: function (event) {
      var group = event.context;
      var component = this.get('selectedSlaveComponent');
      var assignedHosts = component.get('hosts').filterProperty('group', group.name);
      if (assignedHosts.length !== 0) {
        $('.remove-group-error').show();
      } else {
        $('.remove-group-error').hide();
        var key = component.groups.indexOf(group);
        component.groups.removeObject(component.groups[key]);

        var newGroups = component.groups.filterProperty('type', 'new');
        if (newGroups.length == 0)
          component.newGroupIndex = 0;
        else {
          var lastNewGroup = newGroups[newGroups.length - 1];
          component.newGroupIndex = lastNewGroup.index;
        }
        if (group.active) {
          var lastGroup;
          if (key === component.groups.length)
            lastGroup = component.groups.slice(key - 1, key);
          else lastGroup = component.groups.slice(key, key + 1);
          lastGroup.setEach('active', true);
        }
      }
    },

    /**
     * change group name of slave component
     * @param group
     * @param newGroupName
     * @return {Boolean}
     */
    changeSlaveGroupName: function (group, newGroupName) {
      var component = this.get('selectedSlaveComponent');
      var isExist = component.get('groups').filterProperty('name', newGroupName);
      if (isExist.length !== 0)
        return true;
      else {
        var assignedHosts = component.get('hosts').filterProperty('group', group.name);
        if (assignedHosts.length !== 0){
          assignedHosts.setEach('group', newGroupName);
        }
        var groupFilter = component.get('groups').filterProperty('name', group.name);
        groupFilter.setEach('name', newGroupName);
      }
      return false;
    }

  });
  
});
window.require.register("controllers/wizard/stack_upgrade/step1_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.StackUpgradeStep1Controller = Em.Controller.extend({
    name: 'stackUpgradeStep1Controller'
  });
  
});
window.require.register("controllers/wizard/stack_upgrade/step2_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.StackUpgradeStep2Controller = Em.Controller.extend({
    name: 'stackUpgradeStep2Controller',
    /**
     * check whether all services are running
     * @return {Array}
     */
    isAllServicesRunning: function(){
      var masterComponents = App.HostComponent.find().filterProperty('isMaster', true);
      return masterComponents.everyProperty('workStatus', 'STARTED');
    },
    /**
     * callback, which save step data and route to next step
     * @param event
     */
    upgradeAction: function(event){
      if(this.isAllServicesRunning()){
        App.router.send('next');
      } else {
        this.showWarningPopup();
      }
    },
    /**
     * show warning popup if not all services are running
     */
    showWarningPopup: function(){
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step2.manualInstall.popup.header'),
        secondary: null,
        bodyClass: Ember.View.extend({
          template: Em.Handlebars.compile('{{t installer.stackUpgrade.step2.popup.body}}')
        })
      });
    }
  });
  
});
window.require.register("controllers/wizard/stack_upgrade/step3_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.StackUpgradeStep3Controller = Em.Controller.extend({
    name: 'stackUpgradeStep3Controller',

    POLL_INTERVAL: 4000,
    isPolling: false,
    /**
     * STOP_SERVICES internal statuses:
     * - PENDING
     * - IN_PROGRESS
     * - SUCCESS
     * - FAILED
     * UPGRADE_SERVICES internal statuses:
     * - PENDING
     * - IN_PROGRESS
     * - SUCCESS
     * - FAILED
     * - WARNING
     */
    processes:[
      Em.Object.create({
        name: 'STOP_SERVICES',
        displayName: Em.I18n.t('installer.stackUpgrade.step3.stop.header'),
        progress:0,
        status: 'PENDING',
        message: null,
        isRunning: false,
        hosts: [],
        isRetry: false
      }),
      Em.Object.create({
        name: 'UPGRADE_SERVICES',
        displayName: Em.I18n.t('installer.stackUpgrade.step3.upgrade.header'),
        progress: 0,
        status: 'PENDING',
        message:'',
        isRunning: false,
        hosts: [],
        isRetry: false
      })
    ],
    /**
     * pass processes as services to popup
     */
    services: function(){
      return this.get('processes');
    }.property('processes'),
    /**
     * save current requestId and clusterState
     * to localStorage and put it to server
     *
     * STOP_SERVICES cluster status:
     * - STOPPING_SERVICES,
     * UPGRADE_SERVICES cluster status:
     * - STACK_UPGRADING,
     * - STACK_UPGRADE_FAILED,
     * - STACK_UPGRADED,
     * - DEFAULT = STACK UPGRADE COMPLETED
     */
    saveClusterStatus: function(clusterStatus){
      var oldStatus = this.get('content.cluster');
      clusterStatus = jQuery.extend(oldStatus, clusterStatus);
      this.set('content.cluster', clusterStatus);
      App.router.get(this.get('content.controllerName')).save('cluster');
      if(!App.testMode){
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('content.cluster.name'),
          clusterState: clusterStatus.status,
          wizardControllerName: 'stackUpgradeController',
          localdb: App.db.data
        });
      }
    },
    // provide binding for Host Popup data
    serviceTimestamp: null,
    /**
     * load hosts for each process
     */
    loadHosts: function () {
      var hosts = [];
      var installedHosts = App.Host.find();
      this.get('processes').forEach(function(process){
        var hosts = [];
        installedHosts.forEach(function (host) {
          hosts.push(Em.Object.create({
            name: host.get('hostName'),
            publicName: host.get('publicHostName'),
            logTasks: []
          }));
        });
        process.set('hosts', hosts);
      });
    }.observes('content.servicesInfo'),
    submitButton: null,
    /**
     * restart upgrade
     * restart stop services
     * @param event
     */
    retry: function(event){
      var processName = event.context;
      var process = this.get('processes').findProperty('name', processName);
      this.resetProgress(process);
      this.resetMockConfig();
      if(processName == 'STOP_SERVICES'){
        this.stopServices();
      } else {
        this.set('submitButton', false);
        this.runUpgrade();
      }
    },
    /**
     * reset progress and status to retry
     * @param process
     */
    resetProgress: function(process){
      process.set('isRetry', false);
      process.set('status', 'PENDING');
      process.set('progress', 0);
      process.get('hosts').forEach(function(host){host.get('logTasks').clear()});
    },
    /**
     * run stop services
     */
    stopServices: function () {
      var process = this.get('processes').findProperty('name', 'STOP_SERVICES');
      process.set('isRunning', true);
      if (App.testMode) {
        this.startPolling();
        this.saveClusterStatus({
          requestId: 1,
          status: 'STOPPING_SERVICES',
          isCompleted: false
        });
      }
      else {
        var data = '{"RequestInfo": {"context": "'+ Em.I18n.t("requestInfo.stopAllServices") +'"}, "Body": {"ServiceInfo": {"state": "INSTALLED"}}}';
        App.ajax.send({
          name: 'admin.stack_upgrade.stop_services',
          sender: this,
          data: {
            data: data
          },
          success: 'stopServicesSuccessCallback',
          error: 'stopServicesErrorCallback'
        });
      }
    },
    stopServicesSuccessCallback: function (data) {
      var process = this.get('processes').findProperty('name', 'STOP_SERVICES');
      var requestId = data.Requests.id;
      var clusterStatus = {
        requestId: requestId,
        status: 'STOPPING_SERVICES',
        isCompleted: false
      };
      process.set('status', 'IN_PROGRESS');
      this.saveClusterStatus(clusterStatus);
      this.startPolling();
      console.log('Call to stop service successful')
    },
    stopServicesErrorCallback: function () {
      var process = this.get('processes').findProperty('name', 'STOP_SERVICES');
      this.finishProcess(process, 'FAILED');
      process.set('status', 'FAILED');
      console.log("Call to stop services failed");
    },
    /**
     * send request to run upgrade all services
     */
    runUpgrade: function () {
      var process = this.get('processes').findProperty('name', 'UPGRADE_SERVICES');
      process.set('isRunning', true);
      if (App.testMode) {
        this.startPolling();
        this.saveClusterStatus({
          requestId: 1,
          status: 'STACK_UPGRADING',
          isCompleted: false
        });
      }
      else {
        var data = '{"Clusters": {"version" : "' + this.get('content.upgradeVersion') + '"}}';
        App.ajax.send({
          name: 'admin.stack_upgrade.run_upgrade',
          sender: this,
          data: {
            data: data
          },
          success: 'runUpgradeSuccessCallback',
          error: 'runUpgradeErrorCallback'
        });
      }
    },
    runUpgradeSuccessCallback: function (jsonData) {
      var process = this.get('processes').findProperty('name', 'UPGRADE_SERVICES');
      var requestId = jsonData.Requests.id;
      var clusterStatus = {
        status: 'STACK_UPGRADING',
        requestId: requestId,
        isCompleted: false
      };
      process.set('status', 'IN_PROGRESS');
      this.saveClusterStatus(clusterStatus);
      this.startPolling();
    },

    runUpgradeErrorCallback: function (request, ajaxOptions, error) {
      var process = this.get('processes').findProperty('name', 'UPGRADE_SERVICES');
      this.finishProcess(process, 'FAILED');
      process.set('status', 'FAILED');
    },

    /**
     * start polling tasks for current process
     */
    startPolling: function(){
      if(!this.get('isPolling')){
        this.set('isPolling', true);
        if (App.testMode) {
          this.simulatePolling();
        } else {
          //pass an interval "1" to start poll immediately first time
          this.doPoll(1);
        }
      }
    },
    simulateAttempt:0,
    mockUrl:'',
    /**
     * simulate actual poll, using mock data
     */
    simulatePolling: function(){
      var simulateAttempt = this.get('simulateAttempt');
      var process = this.get('processes').findProperty('isRunning', true);
      var upgradeURLs = [
        '/upgrade/poll_1.json',
        '/upgrade/poll_2.json',
        '/upgrade/poll_3.json',
        '/upgrade/poll_4.json',
        '/upgrade/poll_5.json'
      ];
      var stopURLs = [
        '/stop_services/poll_1.json',
        '/stop_services/poll_2.json',
        '/stop_services/poll_3.json',
        '/stop_services/poll_4.json'
      ];
      if(process.get('name') == 'STOP_SERVICES'){
        if(simulateAttempt < 4){
          this.set('mockUrl', stopURLs[simulateAttempt]);
          this.doPoll();
          this.set('simulateAttempt', ++simulateAttempt);
        }
      } else {
        if(simulateAttempt < 5){
          this.set('mockUrl', upgradeURLs[simulateAttempt]);
          this.doPoll();
          this.set('simulateAttempt', ++simulateAttempt);
        }
      }
    },
    getUrl:function(){
      var requestId = this.get('content.cluster.requestId');
      var clusterName = this.get('content.cluster.name');
      if(App.testMode){
        return this.get('mockUrl');
      }
      return App.apiPrefix + '/clusters/' + clusterName + '/requests/' + requestId + '?fields=tasks/*';
    },
    /**
     * poll server for tasks, which contain process progress data
     * @param interval
     */
    doPoll: function(interval) {
      var self = this;
      var pollInterval = interval || self.POLL_INTERVAL;
      if (self.get('isPolling')) {
        setTimeout(function () {

          App.ajax.send({
            name: 'admin.stack_upgrade.do_poll',
            sender: self,
            data: {
              cluster: self.get('content.cluster.name'),
              requestId: self.get('content.cluster.requestId'),
              mock: self.get('mockUrl')
            },
            success: 'doPollSuccessCallback',
            error: 'doPollErrorCallback'
          }).retry({
              times: App.maxRetries,
              timeout: App.timeout
            }).then(
              null,
              function () {
                App.showReloadPopup();
                console.log('Install services all retries failed');
              });
        }, pollInterval);
      }
    },

    doPollSuccessCallback: function (data) {
      var result = this.parseTasks(data);
      if(result){
        if (App.testMode) {
          this.simulatePolling();
        }
        else {
          this.doPoll();
        }
      }
    },

    doPollErrorCallback: function () {
      console.log('ERROR: poll request failed')
    },

    /**
     * parse tasks from poll
     * change status, message, progress on services according to tasks
     * @param data
     * @return {Boolean}
     */
    parseTasks: function(data){
      var tasks = data.tasks || [];
      var process = this.get('processes').findProperty('isRunning', true);
      // if process was finished then it terminates next poll

      this.progressOnProcess(tasks, process);
      var continuePolling = this.statusOnProcess(tasks, process);
      if(process.get('hosts').length && tasks.length){
        process.get('hosts').forEach(function (host) {
          var tasksPerHost = tasks.filterProperty('Tasks.host_name', host.name);
          if (tasksPerHost.length) {
            this.setLogTasksStatePerHost(tasksPerHost, host);
          }
        }, this);
      }
      this.set('serviceTimestamp', App.dateTime());
      return continuePolling;
    },
    /**
     * calculate progress according to tasks status
     * @param actions
     * @param process
     */
    progressOnProcess: function(actions, process){
      var actionsNumber = actions.length;
      var completedActions = actions.filterProperty('Tasks.status', 'COMPLETED').length
        + actions.filterProperty('Tasks.status', 'FAILED').length
        + actions.filterProperty('Tasks.status', 'ABORTED').length
        + actions.filterProperty('Tasks.status', 'TIMEDOUT').length;
      var queuedActions = actions.filterProperty('Tasks.status', 'QUEUED').length;
      var inProgressActions = actions.filterProperty('Tasks.status', 'IN_PROGRESS').length;
      var progress = Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsNumber * 100);
      console.log('INFO: progress is: ' + progress);
      process.set('progress', progress);
    },
    /**
     * evaluate status of process according to task status
     * @param actions
     * @param process
     */
    statusOnProcess: function(actions, process){
      var status = null;
      var message = '';
      var continuePolling = true;
      var errorActions = actions.filter(function (action) {
        if (action.Tasks.status == 'FAILED' || action.Tasks.status == 'ABORTED' || action.Tasks.status == 'TIMEDOUT') {
          return true;
        }
      });
      var masterComponents = ['NAMENODE', 'SECONDARY_NAMENODE', 'SNAMENODE', 'JOBTRACKER', 'ZOOKEEPER_SERVER', 'HIVE_SERVER',
        'HIVE_METASTORE', 'MYSQL_SERVER', 'HBASE_MASTER', 'NAGIOS_SERVER', 'GANGLIA_SERVER', 'OOZIE_SERVER', 'WEBHCAT_SERVER'];
      var failedComponents = errorActions.mapProperty('Tasks.role');
      if (failedComponents.length) {
        for (var i = 0; i < failedComponents.length; i++) {
          if (masterComponents.contains(failedComponents[i])) {
            status = "FAILED";
            continuePolling = false;
            this.finishProcess(process, status);
            break;
          } else if(process.get('progress') == 100){
            status = "WARNING";
            if(process.get('name') == 'UPGRADE_SERVICES'){
              continuePolling = false;
              this.finishProcess(process, status);
            }
          }
        }
      }
      if(!status || ((status == 'WARNING') && (process.get('name') == 'STOP_SERVICES'))){
        if (actions.everyProperty('Tasks.status', 'COMPLETED')) {
          status = 'SUCCESS';
          continuePolling = false;
          this.finishProcess(process, status);
        } else {
          var activeAction = actions.findProperty('Tasks.status', 'IN_PROGRESS');
          status = 'IN_PROGRESS';
          if (activeAction === undefined || activeAction === null) {
            activeAction = actions.findProperty('Tasks.status', 'QUEUED');
            status = 'PENDING';
          }
          if (activeAction === undefined || activeAction === null) {
            activeAction = actions.findProperty('Tasks.status', 'PENDING');
            status = 'PENDING';
          }
          if (activeAction) {
            message = this.displayMessage(activeAction.Tasks);
          }
        }
      }
      console.log('INFO: status is: ' + status);
      process.set('status', status);
      process.set('message', message);
      return continuePolling;
    },
    /**
     * complete process phase
     * accept FAILED, SUCCESS, WARNING process status
     * @param process
     * @param status
     */
    finishProcess: function(process, status){
      this.set('isPolling', false);
      if(process.get('name') == 'STOP_SERVICES'){
        if(status == 'SUCCESS'){
          process.set('isRunning', false);
          this.resetMockConfig();
          this.runUpgrade();
        } else {
          process.set('isRetry', true);
        }
      }
      if(process.get('name') == 'UPGRADE_SERVICES'){
        if(status == 'SUCCESS'){
          this.set('submitButton', Em.I18n.t('common.done'));
          this.saveClusterStatus({
            status: 'STACK_UPGRADED',
            isCompleted: true
          })
        } else if(status == 'FAILED') {
          process.set('isRetry', true);
          this.set('submitButton', false);
          this.saveClusterStatus({
            status: 'STACK_UPGRADE_FAILED',
            isCompleted: false
          })
        } else if(status == 'WARNING'){
          this.set('submitButton', Em.I18n.t('installer.stackUpgrade.step3.ProceedWithWarning'));
          process.set('isRetry', true);
          this.saveClusterStatus({
            status: 'STACK_UPGRADED',
            isCompleted: true
          })
        }
      }
    },
    /**
     * set and update logTasks to each host
     * @param tasksPerHost
     * @param host
     */
    setLogTasksStatePerHost: function (tasksPerHost, host) {
      tasksPerHost.forEach(function (_task) {
        var task = host.get('logTasks').findProperty('Tasks.id', _task.Tasks.id);
        if (task) {
          host.get('logTasks').removeObject(task);
        }
        host.get('logTasks').pushObject(_task);
        //}
      }, this);
    },
    /**
     * reset mock configs to run upgrade simulation again
     */
    resetMockConfig: function(retry){
      this.set('simulateAttempt', 0);
    },
    /**
     * resume wizard on last operation
     */
    resumeStep: function () {
      var clusterStatus = this.get('content.cluster.status');
      var upgrade = this.get('processes').findProperty('name', 'UPGRADE_SERVICES');
      var stop = this.get('processes').findProperty('name', 'STOP_SERVICES');
      if(App.testMode){
        if(this.get('processes').everyProperty('isRunning', false)){
          stop.set('isRunning', true);
        }
        clusterStatus = (this.get('processes').findProperty('name', 'UPGRADE_SERVICES').get('isRunning'))?
         'STACK_UPGRADING':
         'STOPPING_SERVICES';
        upgrade.set('isRetry', false);
        stop.set('isRetry', false);
      }
      if (clusterStatus == 'STOPPING_SERVICES') {
        this.startPolling();
        stop.set('isRunning', true);
        upgrade.set('isRunning', false);
      } else if(clusterStatus != 'PENDING'){
        stop.set('status', 'SUCCESS');
        stop.set('progress', 100);
        stop.set('isRunning', false);
        upgrade.set('isRunning', true);
        if (clusterStatus == 'STACK_UPGRADING') {
          upgrade.set('status', 'IN_PROGRESS');
          this.startPolling();
        } else if (clusterStatus == 'STACK_UPGRADE_FAILED') {
          upgrade.set('status', 'FAILED');
          upgrade.set('isRetry', true);
        } else if (clusterStatus == 'STACK_UPGRADED') {
          upgrade.set('status', 'SUCCESS');
          upgrade.set('progress', 100);
          this.startPolling();
          this.set('isPolling', false);
        }
      }
    },
    /**
     * determine description of current running process
     * @param task
     * @return {*}
     */
    displayMessage: function (task) {
      var role = App.format.role(task.role);
      switch (task.command){
        case 'UPGRADE':
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.upgrade.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.upgrade.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.upgrade.inProgress') + role;
            case 'COMPLETED' :
              return Em.I18n.t('installer.step9.serviceStatus.upgrade.completed') + role;
            case 'FAILED':
              return Em.I18n.t('installer.step9.serviceStatus.upgrade.failed') + role;
          }
          break;
        case 'STOP' :
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.stop.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.stop.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.stop.inProgress') + role;
            case 'COMPLETED' :
              return role + Em.I18n.t('installer.step9.serviceStatus.stop.completed');
            case 'FAILED':
              return role + Em.I18n.t('installer.step9.serviceStatus.stop.failed');
          }
      }
    }
  });
  
});
window.require.register("controllers/wizard/step0_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.WizardStep0Controller = Em.Controller.extend({
    name: 'wizardStep0Controller',

    hasSubmitted : false,

    loadStep: function () {
      this.set('hasSubmitted',false);
      this.set('clusterNameError', '');
    },
    /**
     * validate cluster name
     * set clusterNameError if validation fails
     */
    invalidClusterName : function(){
      var clusterName = this.get('content.cluster.name');
      if (clusterName == '' && this.get('hasSubmitted')) {
        this.set('clusterNameError', Em.I18n.t('installer.step0.clusterName.error.required'));
        return true;
      } else if (/\s/.test(clusterName)) {
        this.set('clusterNameError', Em.I18n.t('installer.step0.clusterName.error.whitespaces'));
        return true;
      } else if (/[^\w\s]/gi.test(clusterName)) {
        this.set('clusterNameError', Em.I18n.t('installer.step0.clusterName.error.specialChar'));
        return true;
      } else {
        this.set('clusterNameError', '');
        return false;
      }
    }.property('hasSubmitted', 'content.cluster.name'),

    /**
     * calculates by <code>invalidClusterName</code> property
     */
    //todo: mix this and previous variables in one
    clusterNameError: '',

    /**
     * Onclick handler for <code>next</code> button
     */
    submit: function () {
      this.set('hasSubmitted', true);
      if (!this.get('invalidClusterName')) {
        App.clusterStatus.set('clusterName', this.get('content.cluster.name'));
        this.set('content.cluster.status', 'PENDING');
        this.set('content.cluster.isCompleted', false);
        App.router.send('next');
      }
    }

  });
  
});
window.require.register("controllers/wizard/step10_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.WizardStep10Controller = Em.Controller.extend({
    clusterInfo: [],

    isNagiosRestartRequired: function() {
      return this.get('content.controllerName') !== 'installerController' && App.Service.find('NAGIOS').get('isLoaded');
    }.property(),

    clearStep: function () {
      this.get('clusterInfo').clear();
    },

    loadStep: function () {
      console.log("TRACE: Loading step10: Summary Page");
      this.clearStep();
      this.loadInstalledHosts(this.loadRegisteredHosts());
      var installFlag = true;
      var startFlag = true;
      if (this.get('content.controllerName') == 'installerController') {
        installFlag = this.loadMasterComponents();
        startFlag = this.loadStartedServices();
      }
      if (installFlag && startFlag) {
        this.loadInstallTime();
      }
    },

    loadRegisteredHosts: function () {
      var masterHosts = this.get('content.masterComponentHosts').mapProperty('hostName').uniq();
      var slaveHosts = this.get('content.slaveComponentHosts');
      var hostObj = [];
      slaveHosts.forEach(function (_hosts) {
        hostObj = hostObj.concat(_hosts.hosts);
      }, this);
      slaveHosts = hostObj.mapProperty('hostName').uniq();
      var registeredHosts = App.Host.find().mapProperty('hostName').concat(masterHosts.concat(slaveHosts)).uniq();
      var registerHostsStatement = Em.I18n.t('installer.step10.hostsSummary').format(registeredHosts.length);
      var registerHostsObj = Ember.Object.create({
        id: 1,
        color: 'text-info',
        displayStatement: registerHostsStatement,
        status: []
      });
      this.get('clusterInfo').pushObject(registerHostsObj);

      return registerHostsObj;
    },

    loadInstalledHosts: function (host) {
      var hosts = this.get('content.hosts');
      var hostsInfo = [];
      for (var index in hosts) {
        hostsInfo.pushObject(hosts[index]);
      }
      var succeededHosts = hostsInfo.filterProperty('status', 'success');
      var warnedHosts = hostsInfo.filterProperty('status', 'warning').concat(hostsInfo.filterProperty('status', 'failed'));
      if (succeededHosts.length) {
        var successStatement = Em.I18n.t('installer.step10.servicesSummary').format(succeededHosts.length) + ((succeededHosts.length > 1) ? Em.I18n.t('installer.step8.hosts') : Em.I18n.t('installer.step8.host'));
        this.get('clusterInfo').findProperty('id', 1).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-success',
          displayStatement: successStatement
        }));
      }

      if (warnedHosts.length) {
        var warnStatement = warnedHosts.length + Em.I18n.t('installer.step10.warnings');
        this.get('clusterInfo').findProperty('id', 1).get('status').pushObject(Ember.Object.create({
          id: 2,
          color: 'text-warning',
          displayStatement: warnStatement,
          statements: []
        }));

        warnedHosts.forEach(function (_host) {
          var clusterState;
          console.log("Content.cluster.status is: " + this.get('content.cluster.status'));
          if (this.get('content.cluster.status') === 'INSTALL FAILED') {
            clusterState = Em.I18n.t('installer.step10.clusterState.installing');
          } else if (this.get('content.cluster.status') === 'START FAILED') {
            clusterState = Em.I18n.t('installer.step10.clusterState.starting');
          }
          var failedTasks = _host.tasks.filterProperty('Tasks.status', 'FAILED');
          failedTasks.forEach(function (_task) {
            var taskStatement = clusterState + App.format.role(_task.Tasks.role) + Em.I18n.t('installer.step10.taskStatus.failed') + _host.name;
            this.get('clusterInfo').findProperty('id', 1).get('status').findProperty('id', 2).get('statements').pushObject(Ember.Object.create({
              status: 'failed',
              color: 'text-info',
              displayStatement: taskStatement
            }));
          }, this);

          var abortedTasks = _host.tasks.filterProperty('Tasks.status', 'ABORTED');
          abortedTasks.forEach(function (_task) {
            var abortStatement = clusterState + App.format.role(_task.Tasks.role) + Em.I18n.t('installer.step10.taskStatus.aborted') + _host.name;
            this.get('clusterInfo').findProperty('id', 1).get('status').findProperty('id', 2).get('statements').pushObject(Ember.Object.create({
              status: 'aborted',
              color: 'text-info',
              displayStatement: abortStatement
            }));
          }, this);

          var timedOutTasks = _host.tasks.filterProperty('Tasks.status', 'TIMEDOUT');
          timedOutTasks.forEach(function (_task) {
            var timedOutStatement = clusterState + App.format.role(_task.Tasks.role) + Em.I18n.t('installer.step10.taskStatus.timedOut') + _host.name;
            this.get('clusterInfo').findProperty('id', 1).get('status').findProperty('id', 2).get('statements').pushObject(Ember.Object.create({
              status: 'timedout',
              color: 'text-info',
              displayStatement: timedOutStatement
            }));
          }, this);
        }, this);
      }
    },

    loadMasterComponents: function () {
      var components = this.get('content.masterComponentHosts');
      var statement;
      if (this.get('content.cluster.status') === 'INSTALL FAILED') {
        this.get('clusterInfo').pushObject(Ember.Object.create({
          id: 2,
          displayStatement: Em.I18n.t('installer.step10.installStatus.failed'),
          color: 'text-error',
          status: []
        }));
        return false;
      } else {
        this.get('clusterInfo').pushObject(Ember.Object.create({
          id: 2,
          displayStatement: Em.I18n.t('installer.step10.installStatus.installed'),
          color: 'text-success',
          status: []
        }));
      }

      console.log('STEP10 master components:  ' + JSON.stringify(components));
      components.forEach(function (_component) {
        var component = Ember.Object.create(_component);
        switch (component.component) {
          case 'NAMENODE':
            this.loadNn(component);
            break;
          case 'SECONDARY_NAMENODE':
            this.loadSnn(component);
            break;
          case 'JOBTRACKER' :
            this.loadJt(component);
            break;
          case 'HISTORYSERVER':
            this.loadHS(component);
            break;
          case 'RESOURCEMANAGER':
            this.loadRM(component);
            break;
          case 'ZOOKEEPER_SERVER' :
            // TODO: Fix this; redundant entries and wrong number
            //this.loadZk(component);
            break;
          case 'HBASE_MASTER':
            this.loadHb(component);
            break;
          case 'HIVE_SERVER':
            this.loadHiveServer(component);
            break;
          case 'OOZIE_SERVER':
            this.loadOozieServer(component);
            break;
          case 'GANGLIA_SERVER':
            this.loadGanglia(component);
            break;
          case 'NAGIOS_SERVER':
            this.loadNagios(component);
            break;
        }
      }, this);
      return true;
    },

    loadHS: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.historyServer') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to HistoryServer component');
      }
    },

    loadRM: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.resourceManager') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to ResourceManager component');
      }
    },

    loadNn: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.nameNode') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to NameNode component');
      }
    },

    loadSnn: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.secondaryNameNode') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to SecondaryNameNode component');
      }
    },

    loadJt: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.jobTracker') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to JobTracker component');
      }
    },

    loadZk: function (component) {
      var hostLength = component.get('hostName').length;
      if (hostLength) {
        var hostVal;
        if (hostLength === 1) {
          hostVal = Em.I18n.t('installer.step8.host');
        } else {
          hostVal = Em.I18n.t('installer.step8.hosts');
        }
        var statement = Em.I18n.t('installer.step10.master.zooKeeper') + component.get('hostName').length + ' ' + hostVal;
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to Zookeeper component');
      }
    },

    loadHb: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.hbase') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to HBase Master component');
      }
    },

    loadHiveServer: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.hiveMetastore') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to Hive server component');
      }
    },

    loadOozieServer: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.oozie') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to Oozie server component');
      }
    },

    loadGanglia: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.ganglia') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to Ganglia server component');
      }
    },

    loadNagios: function (component) {
      if (component.get('hostName')) {
        var statement = Em.I18n.t('installer.step10.master.nagios') + component.get('hostName');
        this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Ember.Object.create({
          id: 1,
          color: 'text-info',
          displayStatement: statement
        }));
      } else {
        console.log('ERROR: no host name assigned to Nagios server component');
      }
    },

    loadStartedServices: function (component) {
      if (this.get('content.cluster.status') === 'STARTED') {
        var statement = Em.I18n.t('installer.step10.startStatus.started');
        this.get('clusterInfo').pushObject(Ember.Object.create({
          id: 3,
          color: 'text-success',
          displayStatement: Em.I18n.t('installer.step10.startStatus.started'),
          status: []
        }));
        this.get('clusterInfo').pushObject(Ember.Object.create({
          id: 4,
          color: 'text-success',
          displayStatement: Em.I18n.t('installer.step10.startStatus.passed'),
          status: []
        }));
        return true;
      } else {
        this.get('clusterInfo').pushObject(Ember.Object.create({
          id: 3,
          color: 'text-error',
          displayStatement: Em.I18n.t('installer.step10.startStatus.failed'),
          status: []
        }));
        return false;
      }
    },

    loadInstallTime: function () {
      var statement;
      var time;
      if (this.get('content.cluster.installTime')) {
        time = this.calculateInstallTime(this.get('content.cluster.installTime'));
        if (time.minutes !== 0) {
          statement = Em.I18n.t('installer.step10.installTime.minutes').format(time.minutes, time.seconds);
        } else {
          statement = Em.I18n.t('installer.step10.installTime.seconds').format(time.seconds);
        }
        this.get('clusterInfo').pushObject(Ember.Object.create({
          id: 5,
          color: 'text-info',
          displayStatement: statement,
          status: []
        }));
      }
    },
    calculateInstallTime: function(installTime){
      var secondsPerMinute = 60;
      var minutes = Math.floor(installTime);
      var seconds = Math.floor((installTime - minutes) * secondsPerMinute);
      return {
        minutes: minutes,
        seconds: seconds
      }
    }
  });

  
});
window.require.register("controllers/wizard/step1_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.WizardStep1Controller = Em.Controller.extend({
    name: 'wizardStep1Controller'
  });
  
});
window.require.register("controllers/wizard/step2_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');
  var lazyloading = require('utils/lazy_loading');

  App.WizardStep2Controller = Em.Controller.extend({
    name: 'wizardStep2Controller',
    hostNameArr: [],
    isPattern: false,
    bootRequestId:  null,
    hasSubmitted: false,
    inputtedAgainHostNames: [],

    isInstaller: function () {
      return this.get('content.controllerName') == 'installerController';
    }.property('content.controllerName'),

    hostNames: function () {
      return this.get('content.installOptions.hostNames');
    }.property('content.installOptions.hostNames'),

    manualInstall: function () {
      return this.get('content.installOptions.manualInstall');
    }.property('content.installOptions.manualInstall'),

    sshKey: function () {
      return this.get('content.installOptions.sshKey');
    }.property('content.installOptions.sshKey'),

    sshUser: function () {
      return this.get('content.installOptions.sshUser');
    }.property('content.installOptions.sshUser'),

    installType: function () {
      return this.get('manualInstall') ? 'manualDriven' : 'ambariDriven';
    }.property('manualInstall'),

    isHostNameValid: function (hostname) {
      return validator.isHostname(hostname);
    },
    /**
     * set not installed hosts to the hostNameArr
     */
    updateHostNameArr: function(){
      this.hostNameArr = this.get('hostNames').trim().split(new RegExp("\\s+", "g"));
      this.patternExpression();
      this.get('inputtedAgainHostNames').clear();
      var installedHostNames = App.Host.find().mapProperty('hostName');
      var tempArr = [];
      for (var i = 0; i < this.hostNameArr.length; i++) {
        if (!installedHostNames.contains(this.hostNameArr[i])) {
          tempArr.push(this.hostNameArr[i]);
        } else {
          this.get('inputtedAgainHostNames').push(this.hostNameArr[i]);
        }
      }
      this.set('hostNameArr', tempArr);
    },

    invalidHostNames: [],

    /**
     * validate host names
     * @return {Boolean}
     */
    isAllHostNamesValid: function () {
      var result = true;
      this.updateHostNameArr();
      this.get('invalidHostNames').clear();
      this.hostNameArr.forEach(function(hostName){
        if (!this.isHostNameValid(hostName)) {
          this.get('invalidHostNames').push(hostName);
          result = false;
        }
      }, this);

      return result;
    },

    hostsError: null,
    /**
     * set hostsError if host names don't pass validation
     */
    checkHostError: function () {
      if (this.get('hostNames').trim() === '') {
        this.set('hostsError', Em.I18n.t('installer.step2.hostName.error.required'));
      }
      else {
        this.set('hostsError', null);
      }
    },

    checkHostAfterSubmitHandler: function() {
      if (this.get('hasSubmitted')) {
        this.checkHostError();
      }
    }.observes('hasSubmitted', 'hostNames'),

    sshKeyError: function () {
      if (this.get('hasSubmitted') && this.get('manualInstall') === false && this.get('sshKey').trim() === '') {
        return Em.I18n.t('installer.step2.sshKey.error.required');
      }
      return null;
    }.property('sshKey', 'manualInstall', 'hasSubmitted'),

    sshUserError: function(){
      if (this.get('manualInstall') === false && this.get('sshUser').trim() === '') {
        return Em.I18n.t('installer.step2.sshUser.required');
      }
      return null;
    }.property('sshUser', 'hasSubmitted', 'manualInstall'),

    /**
     * Get host info, which will be saved in parent controller
     */
    getHostInfo: function () {

      var hostNameArr = this.get('hostNameArr');
      var hostInfo = {};
      for (var i = 0; i < hostNameArr.length; i++) {
        hostInfo[hostNameArr[i]] = {
          name: hostNameArr[i],
          installType: this.get('installType'),
          bootStatus: 'PENDING'
        };
      }

      return hostInfo;
    },

    /**
     * Used to set sshKey from FileUploader
     * @param sshKey
     */
    setSshKey: function(sshKey){
      this.set("content.installOptions.sshKey", sshKey);
    },

    /**
     * Onclick handler for <code>next button</code>. Do all UI work except data saving.
     * This work is doing by router.
     * @return {Boolean}
     */
    evaluateStep: function () {
      console.log('TRACE: Entering controller:WizardStep2:evaluateStep function');

      if (this.get('isSubmitDisabled')) {
        return false;
      }

      this.set('hasSubmitted', true);

      this.checkHostError();
      if (this.get('hostsError') || this.get('sshUserError') || this.get('sshKeyError')) {
        return false;
      }

      this.updateHostNameArr();

      if (!this.hostNameArr.length) {
        this.set('hostsError', Em.I18n.t('installer.step2.hostName.error.already_installed'));
        return false;
      }

      if(this.isPattern)
      {
        this.hostNamePatternPopup(this.hostNameArr);
        return false;
      }
      if (this.get('inputtedAgainHostNames').length) {
        this.installedHostsPopup();
      } else {
        this.proceedNext();
      }
    },
    /**
     * check is there a pattern expression in host name textarea
     * push hosts that match pattern in hostNamesArr
     */
    patternExpression: function(){
      this.isPattern = false;
      var self = this;
      var hostNames = [];
      $.each(this.hostNameArr, function(e,a){
        var start, end, extra = {0:""};
        if(/\[\d*\-\d*\]/.test(a)){
          start=a.match(/\[\d*/);
          end=a.match(/\-\d*]/);

          start=start[0].substr(1);
          end=end[0].substr(1);

          if(parseInt(start) <= parseInt(end, 10) && parseInt(start, 10) >= 0){
            self.isPattern = true;

            if(start[0] == "0" && start.length > 1) {
              extra = start.match(/0*/);
            }

            for (var i = parseInt(start, 10); i < parseInt(end, 10) + 1; i++) {
              hostNames.push(a.replace(/\[\d*\-\d*\]/,extra[0].substring(0,start.length-i.toString().length)+i))
            }

          }else{
            hostNames.push(a);
          }
        }else{
          hostNames.push(a);
        }
      });
      this.hostNameArr =  hostNames;
    },
    /**
     * launch hosts to bootstrap
     * and save already registered hosts
     * @return {Boolean}
     */
    proceedNext: function(warningConfirmed){
      if (this.isAllHostNamesValid() !== true && !warningConfirmed) {
        this.warningPopup();
        return false;
      }

      if (this.get('manualInstall') === true) {
        this.manualInstallPopup();
        return false;
      }

      var bootStrapData = JSON.stringify({'verbose': true, 'sshKey': this.get('sshKey'), 'hosts': this.get('hostNameArr'), 'user': this.get('sshUser')});

      if (App.skipBootstrap) {
        this.saveHosts();
        return true;
      }

      var requestId = App.router.get(this.get('content.controllerName')).launchBootstrap(bootStrapData);
      if (requestId == '0') {
        var controller = App.router.get(App.clusterStatus.wizardControllerName);
        controller.registerErrPopup(Em.I18n.t('common.information'), Em.I18n.t('installer.step2.evaluateStep.hostRegInProgress'));
      } else if (requestId) {
        this.set('content.installOptions.bootRequestId', requestId);
        this.saveHosts();
      }
    },

    /**
     * show warning for host names without dots or IP addresses
     */
    warningPopup: function () {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('common.warning'),
        onPrimary: function () {
          this.hide();
          self.proceedNext(true);
        },
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile(Em.I18n.t('installer.step2.warning.popup.body').format(self.get('invalidHostNames').join(', ')))
        })
      });
    },

    /**
     * show popup with the list of hosts that are already part of the cluster
     */
    installedHostsPopup: function () {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('common.warning'),
        onPrimary: function () {
          self.proceedNext();
          this.hide();
        },
        bodyClass: Ember.View.extend({
          inputtedAgainHostNames: function() {
            return self.get('inputtedAgainHostNames').join(', ');
          }.property(),
          templateName: require('templates/wizard/step2_installed_hosts_popup')
        })
      });
    },

    /**
     * show popup with hosts generated by pattern
     * @param hostNames
     */
    hostNamePatternPopup: function (hostNames) {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step2.hostName.pattern.header'),
        onPrimary: function () {
          self.proceedNext();
          this.hide();
        },
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/items_list_popup'),
          items: hostNames,
          insertedItems: [],
          didInsertElement: function() {
            lazyloading.run({
              destination: this.get('insertedItems'),
              source: this.get('items'),
              context: this,
              initSize: 100,
              chunkSize: 500,
              delay: 100
            });
          }
        })
      });
    },
    /**
     * show notify that installation is manual
     * save hosts
     */
    manualInstallPopup: function () {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step2.manualInstall.popup.header'),
        onPrimary: function () {
          this.hide();
          self.saveHosts();
        },
        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step2ManualInstallPopup')
        })
      });
    },
    /**
     * warn to manually install ambari-agent on each host
     */
    manualInstallWarningPopup: function(){
      if(!this.get('content.installOptions.useSsh')){
        App.ModalPopup.show({
          header: Em.I18n.t('common.warning'),
          body: Em.I18n.t('installer.step2.manualInstall.info'),
          encodeBody: false,
          secondary: null
        });
      }
      this.set('content.installOptions.manualInstall', !this.get('content.installOptions.useSsh'));
    }.observes('content.installOptions.useSsh'),

    isSubmitDisabled: function () {
      return (this.get('hostsError') || this.get('sshKeyError') || this.get('sshUserError'))  ;
    }.property('hostsError', 'sshKeyError', 'sshUserError'),

    setAmbariJavaHome: function(){
      App.ajax.send({
        name: 'ambari.service',
        sender: this,
        success: 'onGetAmbariJavaHomeSuccess',
        error: 'onGetAmbariJavaHomeError'
      });
    },

    onGetAmbariJavaHomeSuccess: function(data) {
      this.set('content.installOptions.javaHome',data.RootServiceComponents.properties['java.home']);
    },

    onGetAmbariJavaHomeError: function() {
      console.warn('can\'t get java.home value from server');
      this.set('content.installOptions.javaHome',App.defaultJavaHome);
    },

    saveHosts: function(){
      this.set('content.hosts', this.getHostInfo());
      this.setAmbariJavaHome();
      App.router.send('next');
    }

  });
  
});
window.require.register("controllers/wizard/step3_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var lazyloading = require('utils/lazy_loading');
  var numberUtils = require('utils/number_utils');

  App.WizardStep3Controller = Em.Controller.extend({
    name: 'wizardStep3Controller',
    hosts: [],
    content: [],
    bootHosts: [],
    registeredHosts: [],
    repoCategoryWarnings: [],
    diskCategoryWarnings: [],
    registrationStartedAt: null,
    registrationTimeoutSecs: function(){
      if(this.get('content.installOptions.manualInstall')){
        return 15;
      }
      return 120;
    }.property('content.installOptions.manualInstall'),
    stopBootstrap: false,
    isSubmitDisabled: true,

    isRetryDisabled: true,
    isLoaded: false,

    navigateStep: function () {
      if(this.get('isLoaded')){
        if (!this.get('content.installOptions.manualInstall')) {
          if (!this.get('wizardController').getDBProperty('bootStatus')) {
            this.startBootstrap();
          }
        } else {
          this.set('bootHosts', this.get('hosts'));
          if (App.testMode) {
            this.getHostInfo();
            this.get('bootHosts').setEach('cpu', '2');
            this.get('bootHosts').setEach('memory', '2000000');
            this.set('isSubmitDisabled', false);
          } else {
            this.set('registrationStartedAt', null);
            this.startRegistration();
          }
        }
      }
    }.observes('isLoaded'),

    clearStep: function () {
      this.set('stopBootstrap', false);
      this.set('hosts', []);
      this.get('bootHosts').clear();
      this.get('wizardController').setDBProperty('bootStatus', false);
      this.set('isSubmitDisabled', true);
      this.set('isRetryDisabled', true);
    },

    loadStep: function () {
      console.log("TRACE: Loading step3: Confirm Hosts");
      this.set('registrationStartedAt', null);
      this.set('isLoaded', false);
      this.disablePreviousSteps();

      this.clearStep();
      this.loadHosts();
      // hosts.setEach('bootStatus', 'RUNNING');
    },

    /* Loads the hostinfo from localStorage on the insertion of view. It's being called from view */
    loadHosts: function () {
      var hostsInfo = this.get('content.hosts');
      var hosts = [];
      var bootStatus = (this.get('content.installOptions.manualInstall')) ? 'DONE' : 'PENDING';
      if (App.testMode) {
        bootStatus = 'REGISTERED';
      }

      for (var index in hostsInfo) {
        var hostInfo = App.HostInfo.create({
          name: hostsInfo[index].name,
          bootStatus: bootStatus,
          isChecked: false
        });

        hosts.pushObject(hostInfo);
      }
      this.set('hosts', hosts);
      this.set('isLoaded', true);
    },

    /**
     * Parses and updates the content based on bootstrap API response.
     * Returns true if polling should continue (some hosts are in "RUNNING" state); false otherwise
     */
    parseHostInfo: function (hostsStatusFromServer) {
      hostsStatusFromServer.forEach(function (_hostStatus) {
        var host = this.get('bootHosts').findProperty('name', _hostStatus.hostName);
        // check if hostname extracted from REST API data matches any hostname in content
        // also, make sure that bootStatus modified by isHostsRegistered call does not get overwritten
        // since these calls are being made in parallel
        if (host && !['REGISTERED', 'REGISTERING'].contains(host.get('bootStatus'))) {
          host.set('bootStatus', _hostStatus.status);
          host.set('bootLog', _hostStatus.log);
        }
      }, this);
      // if the data rendered by REST API has hosts in "RUNNING" state, polling will continue
      return this.get('bootHosts').length != 0 && this.get('bootHosts').someProperty('bootStatus', 'RUNNING');
    },

    removeHosts: function (hosts) {
      var self = this;
      App.showConfirmationPopup(function() {
        App.router.send('removeHosts', hosts);
        self.hosts.removeObjects(hosts);
        if (!self.hosts.length) {
          self.set('isSubmitDisabled', true);
        }
      },Em.I18n.t('installer.step3.hosts.remove.popup.body'));
    },

    /* Removes a single element on the trash icon click. Called from View */
    removeHost: function (hostInfo) {
      this.removeHosts([hostInfo]);
    },

    removeSelectedHosts: function () {
      if (!this.get('noHostsSelected')) {
        var selectedHosts = this.get('hosts').filterProperty('isChecked', true);
        selectedHosts.forEach(function (_hostInfo) {
          console.log('Removing:  ' + _hostInfo.name);
        });
        this.removeHosts(selectedHosts);
      }
    },

    /**
     * show popup with the list of hosts which are selected
     */
    selectedHostsPopup: function () {
      var selectedHosts = this.get('hosts').filterProperty('isChecked').mapProperty('name');
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step3.selectedHosts.popup.header'),
        onPrimary: function () {
          this.hide();
        },
        secondary: null,
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/items_list_popup'),
          items: selectedHosts,
          insertedItems: [],
          didInsertElement: function() {
            lazyloading.run({
              destination: this.get('insertedItems'),
              source: this.get('items'),
              context: this,
              initSize: 100,
              chunkSize: 500,
              delay: 100
            });
          }
        })
      });
    },

    retryHost: function (hostInfo) {
      this.retryHosts([hostInfo]);
    },

    retryHosts: function (hosts) {
      var bootStrapData = JSON.stringify({'verbose': true, 'sshKey': this.get('content.installOptions.sshKey'), 'hosts': hosts.mapProperty('name'), 'user': this.get('content.installOptions.sshUser')});
      this.numPolls = 0;
      if (this.get('content.installOptions.manualInstall') !== true) {
        var requestId = App.router.get('installerController').launchBootstrap(bootStrapData);
        this.set('content.installOptions.bootRequestId', requestId);
        this.set('registrationStartedAt', null);
        this.doBootstrap();
      } else {
        this.set('registrationStartedAt', null);
        this.get('bootHosts').setEach('bootStatus', 'DONE');
        this.startRegistration();
      }
    },

    retrySelectedHosts: function () {
      if (!this.get('isRetryDisabled')) {
        this.set('isRetryDisabled', true);
        var selectedHosts = this.get('bootHosts').filterProperty('bootStatus', 'FAILED');
        selectedHosts.forEach(function (_host) {
          _host.set('bootStatus', 'RUNNING');
          _host.set('bootLog', 'Retrying ...');
        }, this);
        this.retryHosts(selectedHosts);
      }
    },

    numPolls: 0,

    startBootstrap: function () {
      //this.set('isSubmitDisabled', true);    //TODO: uncomment after actual hookup
      this.numPolls = 0;
      this.set('registrationStartedAt', null);
      this.set('bootHosts', this.get('hosts'));
      this.doBootstrap();
    },

    isRegistrationInProgress: true,

    setRegistrationInProgressOnce: function () {
      Ember.run.once(this, 'setRegistrationInProgress');
    }.observes('bootHosts.@each.bootStatus'),

    setRegistrationInProgress: function () {
      var bootHosts = this.get('bootHosts');
      //if hosts aren't loaded yet then registration should be in progress
      var result = (bootHosts.length === 0 && !this.get('isLoaded'));
      for (var i = 0, l = bootHosts.length; i < l; i++) {
        if (bootHosts[i].get('bootStatus') !== 'REGISTERED' && bootHosts[i].get('bootStatus') !== 'FAILED') {
          result = true;
          break;
        }
      }
      this.set('isRegistrationInProgress', result);
    },

    disablePreviousSteps: function () {
      App.router.get('installerController.isStepDisabled').filter(function (step) {
        return step.step >= 0 && step.step <= 2;
      }).setEach('value', this.get('isRegistrationInProgress'));
      if (this.get('isRegistrationInProgress')) {
        this.set('isSubmitDisabled', true);
      }
    }.observes('isRegistrationInProgress'),

    doBootstrap: function () {
      if (this.get('stopBootstrap')) {
        return;
      }
      this.numPolls++;

      App.ajax.send({
        name: 'wizard.step3.bootstrap',
        sender: this,
        data: {
          bootRequestId: this.get('content.installOptions.bootRequestId'),
          numPolls: this.numPolls
        },
        success: 'doBootstrapSuccessCallback'
      }).
        retry({
          times: App.maxRetries,
          timeout: App.timeout
        }).
        then(
          null,
          function () {
            App.showReloadPopup();
            console.log('Bootstrap failed');
          }
        );
    },

    doBootstrapSuccessCallback: function (data) {
      var self = this;
      var pollingInterval = 3000;
      if (data.hostsStatus === undefined) {
        console.log('Invalid response, setting timeout');
        window.setTimeout(function () {
          self.doBootstrap()
        }, pollingInterval);
      } else {
        // in case of bootstrapping just one host, the server returns an object rather than an array, so
        // force into an array
        if (!(data.hostsStatus instanceof Array)) {
          data.hostsStatus = [ data.hostsStatus ];
        }
        console.log("TRACE: In success function for the GET bootstrap call");
        var keepPolling = this.parseHostInfo(data.hostsStatus);

        // Single host : if the only hostname is invalid (data.status == 'ERROR')
        // Multiple hosts : if one or more hostnames are invalid
        // following check will mark the bootStatus as 'FAILED' for the invalid hostname
        if (data.status == 'ERROR' || data.hostsStatus.length != this.get('bootHosts').length) {

          var hosts = this.get('bootHosts');

          for (var i = 0; i < hosts.length; i++) {

            var isValidHost = data.hostsStatus.someProperty('hostName', hosts[i].get('name'));
            if(hosts[i].get('bootStatus') !== 'REGISTERED'){
              if (!isValidHost) {
                hosts[i].set('bootStatus', 'FAILED');
                hosts[i].set('bootLog', Em.I18n.t('installer.step3.hosts.bootLog.failed'));
              }
            }
          }
        }

        if (data.status == 'ERROR' || data.hostsStatus.someProperty('status', 'DONE') || data.hostsStatus.someProperty('status', 'FAILED')) {
          // kicking off registration polls after at least one host has succeeded
          this.startRegistration();
        }
        if (keepPolling) {
          window.setTimeout(function () {
            self.doBootstrap()
          }, pollingInterval);
        }
      }
    },

    startRegistration: function () {
      if (this.get('registrationStartedAt') == null) {
        this.set('registrationStartedAt', App.dateTime());
        console.log('registration started at ' + this.get('registrationStartedAt'));
        this.isHostsRegistered();
      }
    },

    isHostsRegistered: function () {
      if (this.get('stopBootstrap')) {
        return;
      }
      App.ajax.send({
        name: 'wizard.step3.is_hosts_registered',
        sender: this,
        success: 'isHostsRegisteredSuccessCallback'
      }).
        retry({
          times: App.maxRetries,
          timeout: App.timeout
        }).
          then(
            null,
            function () {
              App.showReloadPopup();
              console.log('Error: Getting registered host information from the server');
            }
          );
    },

    isHostsRegisteredSuccessCallback: function (data) {
      console.log('registration attempt...');
      var hosts = this.get('bootHosts');
      var jsonData = data;
      if (!jsonData) {
        console.warn("Error: jsonData is null");
        return;
      }

      // keep polling until all hosts have registered/failed, or registrationTimeout seconds after the last host finished bootstrapping
      var stopPolling = true;
      hosts.forEach(function (_host, index) {
        // Change name of first host for test mode.
        if (App.testMode) {
          if (index == 0) {
            _host.set('name', 'localhost.localdomain');
          }
        }
        // actions to take depending on the host's current bootStatus
        // RUNNING - bootstrap is running; leave it alone
        // DONE - bootstrap is done; transition to REGISTERING
        // REGISTERING - bootstrap is done but has not registered; transition to REGISTERED if host found in polling API result
        // REGISTERED - bootstrap and registration is done; leave it alone
        // FAILED - either bootstrap or registration failed; leave it alone
        switch (_host.get('bootStatus')) {
          case 'DONE':
            _host.set('bootStatus', 'REGISTERING');
            _host.set('bootLog', (_host.get('bootLog') != null ? _host.get('bootLog') : '') + Em.I18n.t('installer.step3.hosts.bootLog.registering'));
            // update registration timestamp so that the timeout is computed from the last host that finished bootstrapping
            this.set('registrationStartedAt', App.dateTime());
            stopPolling = false;
            break;
          case 'REGISTERING':
            if (jsonData.items.someProperty('Hosts.host_name', _host.name)) {
              _host.set('bootStatus', 'REGISTERED');
              _host.set('bootLog', (_host.get('bootLog') != null ? _host.get('bootLog') : '') + Em.I18n.t('installer.step3.hosts.bootLog.registering'));
            } else {
              stopPolling = false;
            }
            break;
          case 'RUNNING':
            stopPolling = false;
            break;
          case 'REGISTERED':
          case 'FAILED':
          default:
            break;
        }
      }, this);

      if (stopPolling) {
        this.getHostInfo();
      } else if (hosts.someProperty('bootStatus', 'RUNNING') || App.dateTime() - this.get('registrationStartedAt') < this.get('registrationTimeoutSecs') * 1000) {
        // we want to keep polling for registration status if any of the hosts are still bootstrapping (so we check for RUNNING).
        var self = this;
        window.setTimeout(function () {
          self.isHostsRegistered();
        }, 3000);
      } else {
        // registration timed out.  mark all REGISTERING hosts to FAILED
        console.log('registration timed out');
        hosts.filterProperty('bootStatus', 'REGISTERING').forEach(function (_host) {
          _host.set('bootStatus', 'FAILED');
          _host.set('bootLog', (_host.get('bootLog') != null ? _host.get('bootLog') : '') + Em.I18n.t('installer.step3.hosts.bootLog.failed'));
        });
        this.getHostInfo();
      }
    },

    hasMoreRegisteredHosts: false,

    getAllRegisteredHosts: function() {
      App.ajax.send({
        name: 'wizard.step3.is_hosts_registered',
        sender: this,
        success: 'getAllRegisteredHostsCallback'
      });
    }.observes('bootHosts'),

    hostsInCluster: function() {
      return App.Host.find().getEach('hostName');
    }.property().volatile(),

    getAllRegisteredHostsCallback: function(hosts) {
      var registeredHosts = [];
      var hostsInCluster = this.get('hostsInCluster');
      var addedHosts = this.get('bootHosts').getEach('name');
      hosts.items.forEach(function(host){
        if (!hostsInCluster.contains(host.Hosts.host_name) && !addedHosts.contains(host.Hosts.host_name)) {
          registeredHosts.push(host.Hosts.host_name);
        }
      });
      if(registeredHosts.length) {
        this.set('hasMoreRegisteredHosts',true);
        this.set('registeredHosts',registeredHosts);
      } else {
        this.set('hasMoreRegisteredHosts',false);
        this.set('registeredHosts','');
      }
    },

    registerErrPopup: function (header, message) {
      App.ModalPopup.show({
        header: header,
        secondary: false,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<p>{{view.message}}</p>'),
          message: message
        })
      });
    },

    /**
     * Get disk info and cpu count of booted hosts from server
     */
    getHostInfo: function () {
      this.set('isWarningsLoaded', false);
      App.ajax.send({
        name: 'wizard.step3.host_info',
        sender: this,
        success: 'getHostInfoSuccessCallback',
        error: 'getHostInfoErrorCallback'
      });
    },

    getHostInfoSuccessCallback: function (jsonData) {
      var hosts = this.get('bootHosts');
      var self = this;
      this.parseWarnings(jsonData);
      var repoWarnings = [];
      var hostsContext = [];
      var diskWarnings = [];
      var hostsDiskContext = [];
      var hostsDiskNames = [];
      var hostsRepoNames = [];
      hosts.forEach(function (_host) {
        var host = (App.testMode) ? jsonData.items[0] : jsonData.items.findProperty('Hosts.host_name', _host.name);
        if (App.skipBootstrap) {
          _host.set('cpu', 2);
          _host.set('memory', ((parseInt(2000000))).toFixed(2));
          _host.set('disk_info', [{"mountpoint": "/", "type":"ext4"},{"mountpoint": "/grid/0", "type":"ext4"}, {"mountpoint": "/grid/1", "type":"ext4"}, {"mountpoint": "/grid/2", "type":"ext4"}]);
        } else if (host) {
          _host.set('cpu', host.Hosts.cpu_count);
          _host.set('memory', ((parseInt(host.Hosts.total_mem))).toFixed(2));
          _host.set('disk_info', host.Hosts.disk_info);
          _host.set('os_type', host.Hosts.os_type);
          _host.set('os_arch', host.Hosts.os_arch);
          _host.set('ip', host.Hosts.ip);

          var context = self.checkHostOSType(host.Hosts.os_type, host.Hosts.host_name);
          if(context) {
            hostsContext.push(context);
            hostsRepoNames.push(host.Hosts.host_name);
          }
          var diskContext = self.checkHostDiskSpace(host.Hosts.host_name, host.Hosts.disk_info);
          if (diskContext) {
            hostsDiskContext.push(diskContext);
            hostsDiskNames.push(host.Hosts.host_name);
          }

        }
      });
      if (hostsContext.length > 0) { // warning exist
        var repoWarning = {
          name: Em.I18n.t('installer.step3.hostWarningsPopup.repositories.name'),
          hosts: hostsContext,
          hostsNames: hostsRepoNames,
          category: 'repositories',
          onSingleHost: false
        };
        repoWarnings.push(repoWarning);
      }
      if (hostsDiskContext.length > 0) { // disk space warning exist
        var diskWarning = {
          name: Em.I18n.t('installer.step3.hostWarningsPopup.disk.name'),
          hosts: hostsDiskContext,
          hostsNames: hostsDiskNames,
          category: 'disk',
          onSingleHost: false
        };
        diskWarnings.push(diskWarning);
      }

      this.set('repoCategoryWarnings', repoWarnings);
      this.set('diskCategoryWarnings', diskWarnings);
      this.stopRegistration();
    },

    getHostInfoErrorCallback: function () {
      console.log('INFO: Getting host information(cpu_count and total_mem) from the server failed');
      this.set('isWarningsLoaded', true);
      this.registerErrPopup(Em.I18n.t('installer.step3.hostInformation.popup.header'), Em.I18n.t('installer.step3.hostInformation.popup.body'));
    },

    stopRegistration: function () {
      this.set('isSubmitDisabled', !this.get('bootHosts').someProperty('bootStatus', 'REGISTERED'));
      this.set('isRetryDisabled', !this.get('bootHosts').someProperty('bootStatus', 'FAILED'));
    },

    /**
     * Check if the customized os group contains the registered host os type. If not the repo on that host is invalid.
     */
    checkHostOSType: function (osType, hostName) {
      if(this.get('content.stacks')){
        var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
        var selectedOS = [];
        var isValid = false;
        if (selectedStack && selectedStack.operatingSystems) {
          selectedStack.get('operatingSystems').filterProperty('selected', true).forEach( function(os) {
            selectedOS.pushObject(os.osType);
            if ( os.osType == osType) {
              isValid = true;
            }
          });
        }

        if (!isValid) {
          console.log('WARNING: Getting host os type does NOT match the user selected os group in step1. ' +
            'Host Name: '+ hostName + '. Host os type:' + osType + '. Selected group:' + selectedOS);
          return Em.I18n.t('installer.step3.hostWarningsPopup.repositories.context').format(hostName, osType, selectedOS);
        } else {
          return null;
        }
      }else{
        return null;
      }
    },

    /**
     * Check if current host has enough free disk usage.
     */
    checkHostDiskSpace: function (hostName, diskInfo) {
      var minFreeRootSpace = App.minDiskSpace * 1024 * 1024; //in kilobyte
      var minFreeUsrLibSpace = App.minDiskSpaceUsrLib * 1024 * 1024; //in kilobyte
      var warningString = '';

      diskInfo.forEach( function(info) {
        switch (info.mountpoint) {
          case '/':
            warningString = info.available < minFreeRootSpace ? Em.I18n.t('installer.step3.hostWarningsPopup.disk.context2').format(App.minDiskSpace + 'GB', info.mountpoint) + ' ' + warningString : warningString;
            break;
          case '/usr':
          case '/usr/lib':
            warningString = info.available < minFreeUsrLibSpace ? Em.I18n.t('installer.step3.hostWarningsPopup.disk.context2').format(App.minDiskSpaceUsrLib + 'GB', info.mountpoint) + ' ' + warningString : warningString;
            break;
          default:
            break;
        }
      });
      if (warningString) {
        console.log('WARNING: Getting host free disk space. ' + 'Host Name: '+ hostName);
        return Em.I18n.t('installer.step3.hostWarningsPopup.disk.context1').format(hostName) + ' ' + warningString;
      } else {
        return null;
      }
    },

    submit: function () {
      if (this.get('isHostHaveWarnings')) {
        var self = this;
        App.showConfirmationPopup(
          function () {
            self.set('content.hosts', self.get('bootHosts'));
            App.router.send('next');
          },
          Em.I18n.t('installer.step3.hostWarningsPopup.hostHasWarnings'));
      }
      else {
        this.set('content.hosts', this.get('bootHosts'));
        App.router.send('next');
      }
    },

    hostLogPopup: function (event, context) {
      var host = event.context;

      App.ModalPopup.show({

        header: Em.I18n.t('installer.step3.hostLog.popup.header').format(host.get('name')),
        secondary: null,

        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step3_host_log_popup'),
          host: host,
          didInsertElement: function () {
            var self = this;
            var button = $(this.get('element')).find('.textTrigger');
            button.click(function () {
              if (self.get('isTextArea')) {
                $(this).text(Em.I18n.t('installer.step3.hostLogPopup.highlight'));
              } else {
                $(this).text(Em.I18n.t('installer.step3.hostLogPopup.copy'));
              }
              self.set('isTextArea', !self.get('isTextArea'));
            });
            $(this.get('element')).find('.content-area').mouseenter(
              function () {
                var element = $(this);
                element.css('border', '1px solid #dcdcdc');
                button.css('visibility', 'visible');
              }).mouseleave(
              function () {
                var element = $(this);
                element.css('border', 'none');
                button.css('visibility', 'hidden');
              })
          },
          isTextArea: false,
          textArea: Em.TextArea.extend({
            didInsertElement: function () {
              var element = $(this.get('element'));
              element.width($(this.get('parentView').get('element')).width() - 10);
              element.height($(this.get('parentView').get('element')).height());
              element.select();
              element.css('resize', 'none');
            },
            readOnly: true,
            value: function () {
              return this.get('content');
            }.property('content')
          })
        })
      });
    },
    /**
     * check warnings from server and put it in parsing
      */
    rerunChecks: function () {
      var self = this;
      var currentProgress = 0;
      var interval = setInterval(function () {
        currentProgress += 100000 / self.get('warningsTimeInterval');
        if (currentProgress < 100) {
          self.set('checksUpdateProgress', currentProgress);
        } else {
          clearInterval(interval);
          App.ajax.send({
            name: 'wizard.step3.rerun_checks',
            sender: self,
            success: 'rerunChecksSuccessCallback',
            error: 'rerunChecksErrorCallback'
          });
        }
      }, 1000);
    },

    rerunChecksSuccessCallback: function (data) {
      this.set('checksUpdateProgress', 100);
      this.set('checksUpdateStatus', 'SUCCESS');
      this.parseWarnings(data);
    },

    rerunChecksErrorCallback: function () {
      this.set('checksUpdateProgress', 100);
      this.set('checksUpdateStatus', 'FAILED');
      console.log('INFO: Getting host information(last_agent_env) from the server failed');
    },

    warnings: [],
    warningsByHost: [],
    warningsTimeInterval: 60000,
    isWarningsLoaded: false,
    /**
     * check are hosts have any warnings
     */
    isHostHaveWarnings: function(){
      return this.get('warnings.length') > 0;
    }.property('warnings'),

    isWarningsBoxVisible: function(){
      return (App.testMode) ? true : !this.get('isRegistrationInProgress');
    }.property('isRegistrationInProgress'),

    checksUpdateProgress:0,
    checksUpdateStatus: null,
    /**
     * filter data for warnings parse
     * is data from host in bootStrap
     * @param data
     * @return {Object}
     */
    filterBootHosts: function (data) {
      var bootHostNames = {};
      this.get('bootHosts').forEach(function (bootHost) {
        bootHostNames[bootHost.get('name')] = true;
      });
      var filteredData = {
        href: data.href,
        items: []
      };
      data.items.forEach(function (host) {
        if (bootHostNames[host.Hosts.host_name]) {
          filteredData.items.push(host);
        }
      });
      return filteredData;
    },
    /**
     * parse warnings data for each host and total
     * @param data
     */
    parseWarnings: function (data) {
      data = App.testMode ? data : this.filterBootHosts(data);
      var warnings = [];
      var warning;
      var hosts = [];
      var warningCategories = {
        fileFoldersWarnings: {},
        packagesWarnings: {},
        processesWarnings: {},
        servicesWarnings: {},
        usersWarnings: {}
      };

      data.items.sortPropertyLight('Hosts.host_name').forEach(function (_host) {
        var host = {
          name: _host.Hosts.host_name,
          warnings: []
        };
        if (!_host.Hosts.last_agent_env) {
          // in some unusual circumstances when last_agent_env is not available from the _host,
          // skip the _host and proceed to process the rest of the hosts.
          console.log("last_agent_env is missing for " + _host.Hosts.host_name + ".  Skipping _host check.");
          return;
        }

        //parse all directories and files warnings for host

        //todo: to be removed after check in new API
        var stackFoldersAndFiles = _host.Hosts.last_agent_env.stackFoldersAndFiles || [];
        stackFoldersAndFiles.forEach(function (path) {
          warning = warningCategories.fileFoldersWarnings[path.name];
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warningCategories.fileFoldersWarnings[path.name] = warning = {
              name: path.name,
              hosts: [_host.Hosts.host_name],
              category: 'fileFolders',
              onSingleHost: true
            };
          }
          host.warnings.push(warning);
        }, this);

        //parse all package warnings for host
        _host.Hosts.last_agent_env.installedPackages.forEach(function (_package) {
            warning = warningCategories.packagesWarnings[_package.name];
            if (warning) {
              warning.hosts.push(_host.Hosts.host_name);
              warning.version = _package.version;
              warning.onSingleHost = false;
            } else {
              warningCategories.packagesWarnings[_package.name] = warning = {
                name: _package.name,
                version: _package.version,
                hosts: [_host.Hosts.host_name],
                category: 'packages',
                onSingleHost: true
              };
            }
            host.warnings.push(warning);
        }, this);

        //parse all process warnings for host

        //todo: to be removed after check in new API
        var javaProcs = _host.Hosts.last_agent_env.hostHealth ? _host.Hosts.last_agent_env.hostHealth.activeJavaProcs : _host.Hosts.last_agent_env.javaProcs;
        javaProcs.forEach(function (process) {
          warning = warningCategories.processesWarnings[process.pid];
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warningCategories.processesWarnings[process.pid] = warning = {
              name: (process.command.substr(0, 35) + '...'),
              hosts: [_host.Hosts.host_name],
              category: 'processes',
              user: process.user,
              pid: process.pid,
              command: '<table><tr><td style="word-break: break-all;">' +
                  ((process.command.length < 500) ? process.command : process.command.substr(0, 230) + '...' +
                      '<p style="text-align: center">................</p>' +
                      '...' + process.command.substr(-230)) + '</td></tr></table>',
              onSingleHost: true
            };
          }
          host.warnings.push(warning);
        }, this);

        //parse all service warnings for host

        //todo: to be removed after check in new API
        if (_host.Hosts.last_agent_env.hostHealth && _host.Hosts.last_agent_env.hostHealth.liveServices) {
          _host.Hosts.last_agent_env.hostHealth.liveServices.forEach(function (service) {
            if (service.status === 'Unhealthy') {
              warning = warningCategories.servicesWarnings[service.name];
              if (warning) {
                warning.hosts.push(_host.Hosts.host_name);
                warning.onSingleHost = false;
              } else {
                warningCategories.servicesWarnings[service.name] = warning = {
                  name: service.name,
                  hosts: [_host.Hosts.host_name],
                  category: 'services',
                  onSingleHost: true
                };
              }
              host.warnings.push(warning);
            }
          }, this);
        }
        //parse all user warnings for host

        //todo: to be removed after check in new API
        if (_host.Hosts.last_agent_env.existingUsers) {
          _host.Hosts.last_agent_env.existingUsers.forEach(function (user) {
            warning = warningCategories.usersWarnings[user.userName];
            if (warning) {
              warning.hosts.push(_host.Hosts.host_name);
              warning.onSingleHost = false;
            } else {
              warningCategories.usersWarnings[user.userName] = warning = {
                name: user.userName,
                hosts: [_host.Hosts.host_name],
                category: 'users',
                onSingleHost: true
              };
            }
            host.warnings.push(warning);
          }, this);
        }

        //parse misc warnings for host
        var umask = _host.Hosts.last_agent_env.umask;
        if (umask && umask !== 18) {
          warning = warnings.filterProperty('category', 'misc').findProperty('name', umask);
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warning = {
              name: umask,
              hosts: [_host.Hosts.host_name],
              category: 'misc',
              onSingleHost: true
            };
            warnings.push(warning);
          }
          host.warnings.push(warning);
        }
        
        var firewallRunning = _host.Hosts.last_agent_env.iptablesIsRunning;
        if (firewallRunning!==null && firewallRunning) {
          var name = Em.I18n.t('installer.step3.hostWarningsPopup.firewall.name');
          warning = warnings.filterProperty('category', 'firewall').findProperty('name', name);
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warning = {
              name: name,
              hosts: [_host.Hosts.host_name],
              category: 'firewall',
              onSingleHost: true
            };
            warnings.push(warning);
          }
          host.warnings.push(warning);
        }

        hosts.push(host);
      }, this);

      for (var categoryId in warningCategories) {
        var category = warningCategories[categoryId]
        for (var warningId in category) {
          warnings.push(category[warningId]);
        }
      }

      warnings.forEach(function (warn) {
        if (warn.hosts.length < 11) {
          warn.hostsList = warn.hosts.join('<br>')
        } else {
          warn.hostsList = warn.hosts.slice(0,10).join('<br>') + '<br> ' + Em.I18n.t('installer.step3.hostWarningsPopup.moreHosts').format(warn.hosts.length - 10);
        }
      });
      hosts.unshift({
        name: 'All Hosts',
        warnings: warnings
      });
      this.set('warnings', warnings);
      this.set('warningsByHost', hosts);
      this.set('isWarningsLoaded', true);
    },
    /**
     * open popup that contain hosts' warnings
     * @param event
     */
    hostWarningsPopup: function(event){
      var self = this;
      var repoCategoryWarnings = this.get('repoCategoryWarnings');
      var diskCategoryWarnings = this.get('diskCategoryWarnings');
      App.ModalPopup.show({

        header: Em.I18n.t('installer.step3.warnings.popup.header'),
        secondary: Em.I18n.t('installer.step3.hostWarningsPopup.rerunChecks'),
        primary: Em.I18n.t('common.close'),
        onPrimary: function () {
          self.set('checksUpdateStatus', null);
          this.hide();
        },
        onClose: function(){
          self.set('checksUpdateStatus', null);
          this.hide();
        },
        onSecondary: function() {
          self.rerunChecks();
        },
        didInsertElement: function () {
          this.fitHeight();
        },

        footerClass: Ember.View.extend({
          templateName: require('templates/wizard/step3_host_warning_popup_footer'),
          classNames: ['modal-footer', 'host-checks-update'],
          footerControllerBinding: 'App.router.wizardStep3Controller',
          progressWidth: function(){
            return 'width:'+this.get('footerController.checksUpdateProgress')+'%';
          }.property('footerController.checksUpdateProgress'),
          isUpdateInProgress: function(){
            if((this.get('footerController.checksUpdateProgress') > 0) &&
               (this.get('footerController.checksUpdateProgress') < 100)){
              return true;
            }
          }.property('footerController.checksUpdateProgress'),
          updateStatusClass:function(){
            var status = this.get('footerController.checksUpdateStatus');
            if(status === 'SUCCESS'){
              return 'text-success';
            } else if(status === 'FAILED'){
              return 'text-error';
            } else {
              return null;
            }
          }.property('footerController.checksUpdateStatus'),
          updateStatus:function(){
            var status = this.get('footerController.checksUpdateStatus');
            if(status === 'SUCCESS'){
              return Em.I18n.t('installer.step3.warnings.updateChecks.success');
            } else if(status === 'FAILED'){
              return Em.I18n.t('installer.step3.warnings.updateChecks.failed');
            } else {
              return null;
            }
          }.property('footerController.checksUpdateStatus')
        }),

        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step3_host_warnings_popup'),
          classNames: ['host-check'],
          bodyControllerBinding: 'App.router.wizardStep3Controller',
          didInsertElement: function () {
            Ember.run.next(this, function () {
              App.tooltip(this.$("[rel='HostsListTooltip']"), {html: true, placement: "right"});
              App.tooltip(this.$('#process .warning-name'), {html: true, placement: "top"});
            });
          }.observes('content'),
          hostSelectView: Ember.Select.extend({
            //content has default value "All Hosts" to bind selection to category
            content: ['All Hosts'],
            hosts: function () {
              return this.get('parentView.warningsByHost').mapProperty('name');
            }.property('parentView.warningsByHost'),
            isLoaded: false,
            selectionBinding: "parentView.category",
            didInsertElement: function(){
              this.initContent();
            },
            initContent: function () {
              this.set('isLoaded', false);
              //The lazy loading for select elements supported only by Firefox and Chrome
              var isBrowserSupported = $.browser.mozilla || ($.browser.safari && navigator.userAgent.indexOf('Chrome') !== -1);
              var isLazyLoading = isBrowserSupported && this.get('hosts').length > 100;
              this.set('isLazyLoading', isLazyLoading);
              if (isLazyLoading) {
                //select need at least 30 hosts to have scrollbar
                this.set('content', this.get('hosts').slice(0, 30));
              } else {
                this.set('content', this.get('hosts'));
                this.set('isLoaded', true);
              }
            }.observes('parentView.warningsByHost'),
            /**
             * on click start lazy loading
             */
            click: function () {
              if (!this.get('isLoaded') && this.get('isLazyLoading')) {
                //filter out hosts, which already pushed in select
                var source = this.get('hosts').filter(function (_host) {
                  return !this.get('content').contains(_host);
                }, this).slice();
                lazyloading.run({
                  destination: this.get('content'),
                  source: source,
                  context: this,
                  initSize: 30,
                  chunkSize: 200,
                  delay: 50
                });
              }
            }
          }),
          warningsByHost: function () {
            return this.get('bodyController.warningsByHost');
          }.property('bodyController.warningsByHost'),
          warnings: function () {
            return this.get('bodyController.warnings');
          }.property('bodyController.warnings'),
          category: 'All Hosts',
          categoryWarnings: function () {
            return this.get('warningsByHost').findProperty('name', this.get('category')).warnings
          }.property('warningsByHost', 'category'),
          content: function () {
            var categoryWarnings = this.get('categoryWarnings');
            return [
              Ember.Object.create({
                warnings: diskCategoryWarnings,
                title: Em.I18n.t('installer.step3.hostWarningsPopup.disk'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.disk.message'),
                type: Em.I18n.t('common.issues'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.disk'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
                category: 'disk',
                isCollapsed: true
              }),
              Ember.Object.create({
                warnings: repoCategoryWarnings,
                title: Em.I18n.t('installer.step3.hostWarningsPopup.repositories'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.repositories.message'),
                type: Em.I18n.t('common.issues'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.repositories'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.invalid'),
                category: 'repositories',
                isCollapsed: true
              }),
              Ember.Object.create({
               warnings: categoryWarnings.filterProperty('category', 'firewall'),
               title: Em.I18n.t('installer.step3.hostWarningsPopup.firewall'),
               message: Em.I18n.t('installer.step3.hostWarningsPopup.firewall.message'),
               type: Em.I18n.t('common.issues'),
               emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.firewall'),
               action: Em.I18n.t('installer.step3.hostWarningsPopup.action.running'),
               category: 'firewall',
               isCollapsed: true
              }),
              Ember.Object.create({
               warnings: categoryWarnings.filterProperty('category', 'processes'),
               title: Em.I18n.t('installer.step3.hostWarningsPopup.process'),
               message: Em.I18n.t('installer.step3.hostWarningsPopup.processes.message'),
               type: Em.I18n.t('common.process'),
               emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.processes'),
               action: Em.I18n.t('installer.step3.hostWarningsPopup.action.running'),
               category: 'process',
               isCollapsed: true
              }),
              Ember.Object.create({
                warnings: categoryWarnings.filterProperty('category', 'packages'),
                title: Em.I18n.t('installer.step3.hostWarningsPopup.package'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.packages.message'),
                type: Em.I18n.t('common.package'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.packages'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.installed'),
                category: 'package',
                isCollapsed: true
              }),
              Ember.Object.create({
                warnings: categoryWarnings.filterProperty('category', 'fileFolders'),
                title: Em.I18n.t('installer.step3.hostWarningsPopup.fileAndFolder'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.fileFolders.message'),
                type: Em.I18n.t('common.path'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.filesAndFolders'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
                category: 'fileFolders',
                isCollapsed: true
              }),
              Ember.Object.create({
                warnings: categoryWarnings.filterProperty('category', 'services'),
                title: Em.I18n.t('installer.step3.hostWarningsPopup.service'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.services.message'),
                type: Em.I18n.t('common.service'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.services'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.notRunning'),
                category: 'service',
                isCollapsed: true
              }),
              Ember.Object.create({
                warnings: categoryWarnings.filterProperty('category', 'users'),
                title: Em.I18n.t('installer.step3.hostWarningsPopup.user'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.users.message'),
                type: Em.I18n.t('common.user'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.users'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
                category: 'user',
                isCollapsed: true
              }),
              Ember.Object.create({
                warnings: categoryWarnings.filterProperty('category', 'misc'),
                title: Em.I18n.t('installer.step3.hostWarningsPopup.misc'),
                message: Em.I18n.t('installer.step3.hostWarningsPopup.misc.message'),
                type: Em.I18n.t('installer.step3.hostWarningsPopup.misc.umask'),
                emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.misc'),
                action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
                category: 'misc',
                isCollapsed: true
              })
            ]
          }.property('category', 'warningsByHost'),

          showHostsPopup: function (hosts) {
            $('.tooltip').hide();
            App.ModalPopup.show({
              header: Em.I18n.t('installer.step3.hostWarningsPopup.allHosts'),
              bodyClass: Ember.View.extend({
                hosts: hosts.context,
                template: Ember.Handlebars.compile('<ul>{{#each host in view.hosts}}<li>{{host}}</li>{{/each}}</ul>')
              }),
              secondary: null
            });
          },

          onToggleBlock: function (category) {
            this.$('#' + category.context.category).toggle('blind', 500);
            category.context.set("isCollapsed", !category.context.get("isCollapsed"));
          },
          /**
           * generate number of hosts which had warnings, avoid duplicated host names in different warnings.
           */
          warningHostsNamesCount: function () {
            var hostNameMap = Ember.Object.create();
            var warningsByHost = self.get('warningsByHost').slice();
            warningsByHost.shift();
            warningsByHost.forEach( function( _host) {
              if (_host.warnings.length) {
                hostNameMap[_host.name] = true;
              }
            })
            if (repoCategoryWarnings.length) {
              repoCategoryWarnings[0].hostsNames.forEach(function (_hostName) {
                if (!hostNameMap[_hostName]) {
                  hostNameMap[_hostName] = true;
                }
              })
            }
            if (diskCategoryWarnings.length) {
              diskCategoryWarnings[0].hostsNames.forEach(function (_hostName) {
                if (!hostNameMap[_hostName]) {
                  hostNameMap[_hostName] = true;
                }
              })
            }
            var size = 0;
            for (var key in hostNameMap) {
              if (hostNameMap.hasOwnProperty(key)) size++;
            }
            return size;
          },
          warningsNotice: function () {
            var warnings = this.get('warnings');
            var issuesNumber = warnings.length + repoCategoryWarnings.length + diskCategoryWarnings.length;
            var issues = issuesNumber + ' ' + (issuesNumber.length === 1 ? Em.I18n.t('installer.step3.hostWarningsPopup.issue') : Em.I18n.t('installer.step3.hostWarningsPopup.issues'));
            var hostsCnt = this.warningHostsNamesCount();
            var hosts = hostsCnt + ' ' + (hostsCnt === 1 ? Em.I18n.t('installer.step3.hostWarningsPopup.host') : Em.I18n.t('installer.step3.hostWarningsPopup.hosts'));
            return Em.I18n.t('installer.step3.hostWarningsPopup.summary').format(issues, hosts);
          }.property('warnings', 'warningsByHost'),
          /**
           * generate detailed content to show it in new window
           */
          contentInDetails: function () {
            var content = this.get('content');
            var warningsByHost = this.get('warningsByHost').slice();
            warningsByHost.shift();
            var newContent = '';
            newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.header') + new Date;
            newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.hosts');
            newContent += warningsByHost.filterProperty('warnings.length').mapProperty('name').join(' ');
            if (content.findProperty('category', 'firewall').warnings.length) {
              newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.firewall');
              newContent += content.findProperty('category', 'firewall').warnings.mapProperty('name').join('<br>');
            }
            if (content.findProperty('category', 'fileFolders').warnings.length) {
              newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.fileFolders');
              newContent += content.findProperty('category', 'fileFolders').warnings.mapProperty('name').join(' ');
            }
            if (content.findProperty('category', 'process').warnings.length) {
              newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.process');
              content.findProperty('category', 'process').warnings.forEach(function (process, i) {
                process.hosts.forEach(function (host, j) {
                  if (!!i || !!j) {
                    newContent += ',';
                  }
                  newContent += '(' + host + ',' + process.user + ',' + process.pid + ')';
                });
              });
            }
            if (content.findProperty('category', 'package').warnings.length) {
              newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.package');
              newContent += content.findProperty('category', 'package').warnings.mapProperty('name').join(' ');
            }
            if (content.findProperty('category', 'service').warnings.length) {
              newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.service');
              newContent += content.findProperty('category', 'service').warnings.mapProperty('name').join(' ');
            }
            if (content.findProperty('category', 'user').warnings.length) {
              newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.user');
              newContent += content.findProperty('category', 'user').warnings.mapProperty('name').join(' ');
            }
            newContent += '</p>';
            return newContent;
          }.property('content', 'warningsByHost'),
          /**
           * open new browser tab with detailed content
           */
          openWarningsInDialog: function(){
            var newWindow = window.open('');
            var newDocument = newWindow.document;
            newDocument.write(this.get('contentInDetails'));
            newWindow.focus();
          }
        })
      })
    },

    registeredHostsPopup: function(){
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step3.warning.registeredHosts').format(this.get('registeredHosts').length),
        secondary: null,
        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step3_registered_hosts_popup'),
          message: Em.I18n.t('installer.step3.registeredHostsPopup'),
          registeredHosts: self.get('registeredHosts')
        })
      })
    }

  });

  
});
window.require.register("controllers/wizard/step4_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.WizardStep4Controller = Em.ArrayController.extend({

    name: 'wizardStep4Controller',
    content: [],

    isSubmitDisabled:function(){
      return this.filterProperty('isSelected', true).filterProperty('isInstalled', false).length === 0;
    }.property("@each.isSelected"),

    /**
     * Check whether all properties are selected
     */
    isAll: function () {
      return this.filterProperty('canBeSelected', true).everyProperty('isSelected', true);
    }.property('@each.isSelected'),

    /**
     * Check whether none properties(minimum) are selected
     */
    isMinimum: function () {
      return this.filterProperty('isDisabled', false).everyProperty('isSelected', false);
    }.property('@each.isSelected'),

    /**
     * Update hidden services. Make them to have the same status as master ones.
     */
    checkDependencies: function () {
      var services = {};
      this.forEach(function (service) {
        services[service.get('serviceName')] = service;
      });

      // prevent against getting error when not all elements have been loaded yet
      if (services['HBASE'] && services['ZOOKEEPER'] && services['HIVE'] && services['HCATALOG'] && services['WEBHCAT']) {
        if (services['YARN'] && services['MAPREDUCE2']) {
          services['MAPREDUCE2'].set('isSelected', services['YARN'].get('isSelected'));
        }
        services['HCATALOG'].set('isSelected', services['HIVE'].get('isSelected'));
        services['WEBHCAT'].set('isSelected', services['HIVE'].get('isSelected'));
      }
    }.observes('@each.isSelected'),

    /**
     * Onclick handler for <code>select all</code> link
     */
    selectAll: function () {
      this.filterProperty('canBeSelected', true).setEach('isSelected', true);
    },

    /**
     * onclick handler for <code>select minimum</code> link
     */
    selectMinimum: function () {
      this.filterProperty('isDisabled', false).setEach('isSelected', false);
    },

    /**
     * Check whether we should turn on <code>serviceName</code> service according to selected <code>dependentServices</code>
     * @param serviceName checked service
     * @param dependentServices list of dependent services
     * @returns {boolean}
     */
    needAddService: function(serviceName, dependentServices) {
      if (!(dependentServices instanceof Array)) {
        dependentServices = [dependentServices];
      }
      if (this.findProperty('serviceName', serviceName) && this.findProperty('serviceName', serviceName).get('isSelected') === false) {
        var ds = this.filter(function (item) {
          return dependentServices.contains(item.get('serviceName')) && item.get('isSelected');
        });
        return (ds.get('length') > 0);
      }
      return false;
    },

    /**
     * Check whether we should turn on <code>Oozie</code> service
     * @return {Boolean}
     */
    needToAddOozie: function () {
      return this.needAddService('OOZIE', ['FALCON']);
    },

    /**
     * Check whether we should turn on <code>MapReduce</code> service
     * @return {Boolean}
     */
    needToAddMapReduce: function () {
      return this.needAddService('MAPREDUCE', ['PIG', 'OOZIE', 'HIVE']);
    },
    /**
     * Check whether we should turn on <code>MapReduce2</code> service
     * @return {Boolean}
     */
    needToAddYarnMapReduce2: function() {
      return this.needAddService('YARN', ['PIG', 'OOZIE', 'HIVE','TEZ']);
    },
    /**
     * Check whether we should turn on <code>Tez</code> service
     * @return {Boolean}
     */
    needToAddTez: function() {
      return this.needAddService('TEZ', ['YARN']);
    },
    /**
     * Check whether we should turn on <code>ZooKeeper</code> service
     * @return {Boolean}
     */
    needToAddZooKeeper: function () {
      if (App.get('isHadoop2Stack')) {
        return this.findProperty('serviceName', 'ZOOKEEPER') && this.findProperty('serviceName', 'ZOOKEEPER').get('isSelected') === false;
      } else {
        return this.needAddService('ZOOKEEPER', ['HBASE', 'HIVE', 'WEBHCAT', 'STORM']);
      }
    },
    /** 
     * Check whether we should turn on <code>HDFS or GLUSTERFS</code> service
     * @return {Boolean}
     */
    noDFSs: function () {
      return (this.findProperty('serviceName', 'HDFS').get('isSelected') === false &&
      		(!this.findProperty('serviceName', 'GLUSTERFS') || this.findProperty('serviceName', 'GLUSTERFS').get('isSelected') === false));
    },

    /** 
     * Check if multiple distributed file systems were selected
     * @return {Boolean}
     */
    multipleDFSs: function () {
  	  return (this.findProperty('serviceName', 'HDFS').get('isSelected') === true &&
  	    	(this.findProperty('serviceName', 'GLUSTERFS') && this.findProperty('serviceName', 'GLUSTERFS').get('isSelected') === true));
    },

    /**
     * Check do we have any monitoring service turned on
     * @return {Boolean}
     */
    gangliaOrNagiosNotSelected: function () {
      return (this.findProperty('serviceName', 'GANGLIA').get('isSelected') === false || this.findProperty('serviceName', 'NAGIOS').get('isSelected') === false);
    },

    /**
     * Check whether user turned on monitoring service and go to next step
     */
    validateMonitoring: function () {
      if (this.gangliaOrNagiosNotSelected()) {
        this.monitoringCheckPopup();
      } else {
        App.router.send('next');
      }
    },

    /**
     * submit checks describe dependency rules between services
     * @checkCallback - callback, which check for dependency
     * @popupParams - parameters for popup
     */
    submitChecks: [
      {
        checkCallback: 'needToAddMapReduce',
        popupParams: [{serviceName: 'MAPREDUCE', selected: true}, 'mapreduceCheck']
      },
      {
        checkCallback: 'noDFSs',
        popupParams: [{serviceName:'HDFS', selected: true}, 'hdfsCheck']
      },
      {
        checkCallback: 'needToAddYarnMapReduce2',
        popupParams: [{serviceName:'YARN', selected:true}, 'yarnCheck']
      },
      {
        checkCallback: 'needToAddZooKeeper',
        popupParams: [{serviceName:'ZOOKEEPER', selected: true}, 'zooKeeperCheck']
      },
      {
        checkCallback: 'multipleDFSs',
        popupParams: [[
          {serviceName: 'HDFS', selected: true},
          {serviceName: 'GLUSTERFS', selected: false}
        ], 'multipleDFS']
      },
      {
        checkCallback: 'needToAddOozie',
        popupParams: [{serviceName:'OOZIE', selected: true}, 'oozieCheck']
      },
      {
        checkCallback: 'needToAddTez',
        popupParams: [{serviceName:'TEZ', selected: true}, 'tezCheck']
      }
    ],

    /**
     * Onclick handler for <code>Next</code> button
     */
    submit: function () {
      var submitChecks = this.get('submitChecks');
      var doValidateMonitoring = true;
      if (!this.get("isSubmitDisabled")) {
        for (var i = 0; i < submitChecks.length; i++) {
          if (this[submitChecks[i].checkCallback].call(this)) {
            doValidateMonitoring = false;
            this.needToAddServicePopup.apply(this, submitChecks[i].popupParams);
            break;
          }
        }
        if (doValidateMonitoring) {
          this.validateMonitoring();
        }
      }
    },
    
    /**
     * Select/deselect services
     * @param services array of objects
     *  <code>
     *    [
     *      {
     *        service: 'HDFS',
     *        selected: true
     *      },
     *      ....
     *    ]
     *  </code>
     * @param i18nSuffix
     */
    needToAddServicePopup: function(services, i18nSuffix) {
      if (!(services instanceof Array)) {
        services = [services];
      }
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step4.' + i18nSuffix + '.popup.header'),
        body: Em.I18n.t('installer.step4.' + i18nSuffix + '.popup.body'),
        onPrimary: function () {
          services.forEach(function(service) {
            self.findProperty('serviceName', service.serviceName).set('isSelected', service.selected);
          });
          this.hide();
          self.submit();
        }
      });
    },

    monitoringCheckPopup: function () {
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step4.monitoringCheck.popup.header'),
        body: Em.I18n.t('installer.step4.monitoringCheck.popup.body'),
        onPrimary: function () {
          this.hide();
          App.router.send('next');
        }
      });
    }
  });
  
});
window.require.register("controllers/wizard/step5_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.WizardStep5Controller = Em.Controller.extend({

    name:"wizardStep5Controller",
    title: function () {
      if (this.get('content.controllerName') == 'reassignMasterController') {
        return Em.I18n.t('installer.step5.reassign.header');
      }
      return Em.I18n.t('installer.step5.header');
    }.property('content.controllerName'),

    isReassignWizard: function () {
      return this.get('content.controllerName') == 'reassignMasterController';
    }.property('content.controllerName'),

    isAddServiceWizard: function() {
      return this.get('content.controllerName') == 'addServiceController';
    }.property('content.controllerName'),

    isReassignHive: function () {
      return this.get('servicesMasters').objectAt(0) && this.get('servicesMasters').objectAt(0).component_name == 'HIVE_SERVER' && this.get('isReassignWizard');
    }.property('isReassignWizard', 'servicesMasters'),
    /**
     * master components which could be assigned to multiple hosts
     */
    multipleComponents: ['ZOOKEEPER_SERVER', 'HBASE_MASTER'],
    /**
     * Define state for submit button. Return true only for Reassign Master Wizard and if more than one master component was reassigned.
     */
    isSubmitDisabled: function () {
      if (!this.get('isReassignWizard')) {
        return false;
      }
      var reassigned = 0;
      var arr1 = App.HostComponent.find().filterProperty('componentName', this.get('content.reassign.component_name')).mapProperty('host.hostName');
      var arr2 = this.get('servicesMasters').mapProperty('selectedHost');
      arr1.forEach(function (host) {
        if (!arr2.contains(host)) {
          reassigned++;
        }
      }, this);
      return reassigned !== 1;
    }.property('servicesMasters.@each.selectedHost'),

    hosts:[],

    componentToRebalance: null,
    rebalanceComponentHostsCounter: 0,

    servicesMasters:[],
    selectedServicesMasters:[],

    components:require('data/service_components'),

    clearStep:function () {
      this.set('hosts', []);
      this.set('selectedServicesMasters', []);
      this.set('servicesMasters', []);
    },

    loadStep:function () {
      console.log("WizardStep5Controller: Loading step5: Assign Masters");
      this.clearStep();
      this.renderHostInfo();
      this.renderComponents(this.loadComponents());

      this.updateComponent('ZOOKEEPER_SERVER');
      if(App.supports.multipleHBaseMasters){
        this.updateComponent('HBASE_MASTER');
      }

      if (!this.get("selectedServicesMasters").filterProperty('isInstalled', false).length) {
        console.log('no master components to add');
        App.router.send('next');
      }
    },

    /**
     * Used to set showAddControl flag for ZOOKEEPER_SERVER and HBASE_SERVER
     */
    updateComponent: function(componentName){
      var component = this.last(componentName);

      var services = this.get('content.services').filterProperty('isInstalled', true).mapProperty('serviceName');
      var currentService = componentName.split('_')[0];
      var showControl = !services.contains(currentService);

      if (component) {
        if(showControl){
          if (this.get("selectedServicesMasters").filterProperty("component_name", componentName).length < this.get("hosts.length") && !this.get('isReassignWizard')) {
            component.set('showAddControl', true);
          } else {
            component.set('showRemoveControl', false);
          }
        }
      }
    },

    /**
     * Load active host list to <code>hosts</code> variable
     */
    renderHostInfo:function () {

      var hostInfo = this.get('content.hosts');
      var result = [];

      for (var index in hostInfo) {
        var _host = hostInfo[index];
        if (_host.bootStatus === 'REGISTERED') {
          result.push(Ember.Object.create({
            host_name:_host.name,

            cpu:_host.cpu,
            memory:_host.memory,
            disk_info:_host.disk_info,
            host_info: Em.I18n.t('installer.step5.hostInfo').fmt(_host.name, numberUtils.bytesToSize(_host.memory, 1, 'parseFloat', 1024), _host.cpu)
          }));
        }
      }
      this.set("hosts", result);
      this.sortHosts(this.get('hosts'));
    },

    sortHosts: function (hosts) {
      hosts.sort(function (a, b) {
        if (a.get('memory') == b.get('memory')) {
          if (a.get('cpu') == b.get('cpu')) {
            return a.get('host_name').localeCompare(b.get('host_name')); // hostname asc
          }
          return b.get('cpu') - a.get('cpu'); // cores desc
        }
        return b.get('memory') - a.get('memory'); // ram desc
      });
    },

    /**
     * Load services info to appropriate variable and return masterComponentHosts
     * @return Array
     */
    loadComponents:function () {

      var services = this.get('content.services')
        .filterProperty('isSelected', true).mapProperty('serviceName'); //list of shown services

      var masterComponents = this.get('components').filterProperty('isMaster', true); //get full list from mock data
      var masterHosts = this.get('content.masterComponentHosts'); //saved to local storage info

      var resultComponents = [];

      var servicesLength = services.length;
      for (var index = 0; index < servicesLength; index++) {
        var componentInfo = masterComponents.filterProperty('service_name', services[index]);

        componentInfo.forEach(function (_componentInfo) {
          if (_componentInfo.component_name == 'ZOOKEEPER_SERVER' || _componentInfo.component_name == 'HBASE_MASTER') {
            var savedComponents = masterHosts.filterProperty('component', _componentInfo.component_name);
            if (savedComponents.length) {

              savedComponents.forEach(function (item) {
                var zooKeeperHost = {};
                zooKeeperHost.display_name = _componentInfo.display_name;
                zooKeeperHost.component_name = _componentInfo.component_name;
                zooKeeperHost.selectedHost = item.hostName;
                zooKeeperHost.serviceId = services[index];
                zooKeeperHost.isInstalled = item.isInstalled;
                zooKeeperHost.isHiveCoHost = false;
                resultComponents.push(zooKeeperHost);
              })

            } else {

              var zooHosts = this.selectHost(_componentInfo.component_name);
              zooHosts.forEach(function (_host) {
                var zooKeeperHost = {};
                zooKeeperHost.display_name = _componentInfo.display_name;
                zooKeeperHost.component_name = _componentInfo.component_name;
                zooKeeperHost.selectedHost = _host;
                zooKeeperHost.serviceId = services[index];
                zooKeeperHost.isInstalled = false;
                zooKeeperHost.isHiveCoHost = false;
                resultComponents.push(zooKeeperHost);
              });

            }
          } else {
            var savedComponent = masterHosts.findProperty('component', _componentInfo.component_name);
            var componentObj = {};
            componentObj.component_name = _componentInfo.component_name;
            componentObj.display_name = _componentInfo.display_name;
            componentObj.selectedHost = savedComponent ? savedComponent.hostName : this.selectHost(_componentInfo.component_name);   // call the method that plays selectNode algorithm or fetches from server
            componentObj.isInstalled = savedComponent ? savedComponent.isInstalled : false;
            componentObj.serviceId = services[index];
            componentObj.isHiveCoHost = ['HIVE_METASTORE', 'WEBHCAT_SERVER'].contains(_componentInfo.component_name) && !this.get('isReassignWizard');
            resultComponents.push(componentObj);
          }
        }, this);
      }

      return resultComponents;
    },

    /**
     * Put master components to <code>selectedServicesMasters</code>, which will be automatically rendered in template
     * @param masterComponents
     */
    renderComponents:function (masterComponents) {
      var self = this;
      var services = this.get('content.services')
        .filterProperty('isInstalled', true).mapProperty('serviceName'); //list of shown services
      var showRemoveControlZk = !services.contains('ZOOKEEPER') && masterComponents.filterProperty('display_name', 'ZooKeeper').length > 1;
      var showRemoveControlHb = !services.contains('HBASE') && masterComponents.filterProperty('component_name', 'HBASE_MASTER').length > 1;
      var zid = 1;
      var hid = 1;
      var nid = 1;
      var result = [];

      masterComponents.forEach(function (item) {

        if (item.component_name == 'SECONDARY_NAMENODE') {
          if (self.get('isAddServiceWizard')) {
            if (App.get('isHaEnabled')) {
              return;
            }
          }
        }

        var componentObj = Ember.Object.create(item);
        console.log("TRACE: render master component name is: " + item.component_name);

        if (item.display_name === "ZooKeeper") {
          componentObj.set('zId', zid++);
          componentObj.set("showRemoveControl", showRemoveControlZk);
        } else if (App.supports.multipleHBaseMasters && item.component_name === "HBASE_MASTER") {
          componentObj.set('zId', hid++);
          componentObj.set("showRemoveControl", showRemoveControlHb);
        }  else if (item.component_name === "NAMENODE") {
          componentObj.set('zId', nid++);
        }
        result.push(componentObj);
      }, this);

      this.set("selectedServicesMasters", result);
      if (this.get('isReassignWizard')) {
        var components = result.filterProperty('component_name', this.get('content.reassign.component_name'));
        components.setEach('isInstalled', false);
        this.set('servicesMasters', components);
      } else {
        this.set('servicesMasters', result);
      }
    },

    hasHiveServer: function () {
      return !!this.get('selectedServicesMasters').findProperty('component_name', 'HIVE_SERVER') && !this.get('isReassignWizard');
    }.property('selectedServicesMasters'),

    updateHiveCoHosts: function () {
      var hiveServer =  this.get('selectedServicesMasters').findProperty('component_name', 'HIVE_SERVER');
      var hiveMetastore = this.get('selectedServicesMasters').findProperty('component_name', 'HIVE_METASTORE');
      var webHCatServer = this.get('selectedServicesMasters').findProperty('component_name', 'WEBHCAT_SERVER');
      if (hiveServer && hiveMetastore && webHCatServer) {
        if (!this.get('isReassignHive') && this.get('servicesMasters').objectAt(0) && !(this.get('servicesMasters').objectAt(0).component_name == 'HIVE_METASTORE')) {
          this.get('selectedServicesMasters').findProperty('component_name', 'HIVE_METASTORE').set('selectedHost', hiveServer.get('selectedHost'))
        }
        this.get('selectedServicesMasters').findProperty('component_name', 'WEBHCAT_SERVER').set('selectedHost', hiveServer.get('selectedHost'));
      }
    }.observes('selectedServicesMasters.@each.selectedHost'),

    /**
     * select and return host for component by scheme
     * Scheme is an object that has keys which compared to number of hosts,
     * if key more that number of hosts, then return value of that key.
     * Value is index of host in hosts array.
     *
     * @param noOfHosts
     * @param selectionScheme
     * @return {*}
     */
    getHostForComponent: function(noOfHosts, selectionScheme){
      var hosts = this.get('hosts');
      if(hosts.length === 1 || $.isEmptyObject(selectionScheme)){
        return hosts[0];
      } else {
        for(var i in selectionScheme){
          if(window.isFinite(i)){
            if(noOfHosts < window.parseInt(i)){
              return hosts[selectionScheme[i]];
            }
          }
        }
        return hosts[selectionScheme['else']]
      }
    },

    getZooKeeperServer:function (noOfHosts) {
      var hosts = this.get('hosts');
      if (noOfHosts < 3) {
        return [hosts[0].host_name];
      } else {
        return [hosts[0].host_name, hosts[1].host_name, hosts[2].host_name];
      }
    },

    getGangliaServer:function (noOfHosts) {
      var hostNames = this.get('hosts').mapProperty('host_name');
      var hostExcAmbari = hostNames.without(location.hostname);
      if (noOfHosts > 1) {
        return hostExcAmbari[0];
      } else {
        return hostNames[0];
      }
    },

    getNagiosServer:function (noOfHosts) {
      return this.getGangliaServer(noOfHosts);
    },

    getHueServer:function (noOfHosts) {
      return this.getGangliaServer(noOfHosts);
    },

    getOozieServer:function(){
      return this.selectHost('OOZIE_SERVER');
    },

    getNimbusServer: function(noOfHosts) {
      return this.getGangliaServer(noOfHosts);
    },
    /**
     * Return hostName of masterNode for specified service
     * @param componentName
     * @return {*}
     */
    selectHost:function (componentName) {
      var noOfHosts = this.get('hosts').length;
      switch (componentName) {
        case 'KERBEROS_SERVER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 3,
            "else" : 5
          }).host_name;
        case 'NAMENODE':
          return this.getHostForComponent(noOfHosts, {
            "else" : 0
          }).host_name;
        case 'SECONDARY_NAMENODE':
          return this.getHostForComponent(noOfHosts, {
            "else" : 1
          }).host_name;
        case 'JOBTRACKER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 1,
            "else" : 2
          }).host_name;
        case 'HISTORYSERVER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 1,
            "else" : 2
          }).host_name;
        case 'RESOURCEMANAGER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 1,
            "else" : 2
          }).host_name;
        case 'HBASE_MASTER':
          return [this.getHostForComponent(noOfHosts, {
            "3" : 0,
            "6" : 0,
            "31" : 2,
            "else" : 3
          }).host_name];
        case 'OOZIE_SERVER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 2,
            "else" : 3
          }).host_name;
        case 'HIVE_SERVER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 2,
            "else" : 4
          }).host_name;
        case 'HIVE_METASTORE':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 2,
            "else" : 4
          }).host_name;
        case 'WEBHCAT_SERVER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 2,
            "else" : 4
          }).host_name;
        case 'ZOOKEEPER_SERVER':
          return this.getZooKeeperServer(noOfHosts);
        case 'GANGLIA_SERVER':
          return this.getGangliaServer(noOfHosts);
        case 'NAGIOS_SERVER':
          return this.getNagiosServer(noOfHosts);
        case 'HUE_SERVER':
          return this.getHueServer(noOfHosts);
        case 'APP_TIMELINE_SERVER':
          return this.getHostForComponent(noOfHosts, {
            "3" : 1,
            "6" : 1,
            "31" : 1,
            "else" : 2
          }).host_name;
        case 'FALCON_SERVER':
          return this.getOozieServer(noOfHosts);
        case 'STORM_UI_SERVER':
        case 'DRPC_SERVER':
        case 'STORM_REST_API':
        case 'NIMBUS':
          return this.getNimbusServer(noOfHosts);
        default:
      }
    },

    masterHostMapping: function () {
      var mapping = [], mappingObject, mappedHosts, hostObj;
      //get the unique assigned hosts and find the master services assigned to them
      mappedHosts = this.get("selectedServicesMasters").mapProperty("selectedHost").uniq();

      mappedHosts.forEach(function (item) {
        hostObj = this.get("hosts").findProperty("host_name", item);

        mappingObject = Ember.Object.create({
          host_name: item,
          hostInfo: hostObj.host_info,
          masterServices: this.get("selectedServicesMasters").filterProperty("selectedHost", item)
        });

        mapping.pushObject(mappingObject);
      }, this);

      return mapping.sortProperty('host_name');
    }.property("selectedServicesMasters.@each.selectedHost"),

    remainingHosts:function () {
      return (this.get("hosts.length") - this.get("masterHostMapping.length"));
    }.property("selectedServicesMasters.@each.selectedHost"),


    /**
     * On change callback for selects
     * @param componentName
     * @param selectedHost
     * @param zId
     */
    assignHostToMaster:function (componentName, selectedHost, zId) {
      if (selectedHost && componentName) {
        if (zId) {
          this.get('selectedServicesMasters').filterProperty('component_name', componentName).findProperty("zId", zId).set("selectedHost", selectedHost);
        } else {
          this.get('selectedServicesMasters').findProperty("component_name", componentName).set("selectedHost", selectedHost);
        }
      }
    },

    /**
     * Returns last component of selected type
     * @param componentName
     * @return {*}
     */
    last: function(componentName){
      return this.get("selectedServicesMasters").filterProperty("component_name", componentName).get("lastObject");
    },

    /**
     * Add new component to ZooKeeper Server and Hbase master
     * @param componentName
     * @return {Boolean}
     */
    addComponent:function (componentName) {
      /*
       * Logic: If ZooKeeper or Hbase service is selected then there can be
       * minimum 1 ZooKeeper or Hbase master in total, and
       * maximum 1 ZooKeeper or Hbase on every host
       */

      var maxNumMasters = this.get("hosts.length"),
        currentMasters = this.get("selectedServicesMasters").filterProperty("component_name", componentName),
        newMaster = null,
        masterHosts = null,
        suggestedHost = null,
        i = 0,
        lastMaster = null;

      if (!currentMasters.length) {
        console.log('ALERT: Zookeeper service was not selected');
        return false;
      }

      if (currentMasters.get("length") < maxNumMasters) {

        currentMasters.set("lastObject.showAddControl", false);
        currentMasters.set("lastObject.showRemoveControl", true);

        //create a new zookeeper based on an existing one
        newMaster = Ember.Object.create({});
        lastMaster = currentMasters.get("lastObject");
        newMaster.set("display_name", lastMaster.get("display_name"));
        newMaster.set("component_name", lastMaster.get("component_name"));
        newMaster.set("selectedHost", lastMaster.get("selectedHost"));
        newMaster.set("isInstalled", false);

        if (currentMasters.get("length") === (maxNumMasters - 1)) {
          newMaster.set("showAddControl", false);
        } else {
          newMaster.set("showAddControl", true);
        }
        newMaster.set("showRemoveControl", true);

        //get recommended host for the new Zookeeper server
        masterHosts = currentMasters.mapProperty("selectedHost").uniq();

        for (i = 0; i < this.get("hosts.length"); i++) {
          if (!(masterHosts.contains(this.get("hosts")[i].get("host_name")))) {
            suggestedHost = this.get("hosts")[i].get("host_name");
            break;
          }
        }

        newMaster.set("selectedHost", suggestedHost);
        newMaster.set("zId", (currentMasters.get("lastObject.zId") + 1));

        this.get("selectedServicesMasters").insertAt(this.get("selectedServicesMasters").indexOf(lastMaster) + 1, newMaster);

        this.set('componentToRebalance', componentName);
        this.incrementProperty('rebalanceComponentHostsCounter');

        return true;
      }
      return false;//if no more zookeepers can be added
    },

    /**
     * Remove component from ZooKeeper server or Hbase Master
     * @param componentName
     * @param zId
     * @return {Boolean}
     */
    removeComponent:function (componentName, zId) {
      var currentMasters = this.get("selectedServicesMasters").filterProperty("component_name", componentName);

      //work only if the Zookeeper service is selected in previous step
      if (!currentMasters.length) {
        return false;
      }

      if (currentMasters.get("length") > 1) {
        this.get("selectedServicesMasters").removeAt(this.get("selectedServicesMasters").indexOf(currentMasters.findProperty("zId", zId)));

        currentMasters = this.get("selectedServicesMasters").filterProperty("component_name", componentName);
        if (currentMasters.get("length") < this.get("hosts.length")) {
          currentMasters.set("lastObject.showAddControl", true);
        }

        if (currentMasters.get("length") === 1) {
          currentMasters.set("lastObject.showRemoveControl", false);
        }

        this.set('componentToRebalance', componentName);
        this.incrementProperty('rebalanceComponentHostsCounter');

        return true;
      }

      return false;

    },

    submit: function () {
      if (!this.get('isSubmitDisabled')){
        App.router.send('next');
      }
    }
  });



  
});
window.require.register("controllers/wizard/step6_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var db = require('utils/db');

  /**
   * By Step 6, we have the following information stored in App.db and set on this
   * controller by the router:
   *
   *   hosts: App.db.hosts (list of all hosts the user selected in Step 3)
   *   selectedServiceNames: App.db.selectedServiceNames (the services that the user selected in Step 4)
   *   masterComponentHosts: App.db.masterComponentHosts (master-components-to-hosts mapping the user selected in Step 5)
   *
   * Step 6 will set the following information in App.db:
   *   slaveComponentHosts: App.db.slaveComponentHosts (slave-components-to-hosts mapping the user selected in Step 6)
   *
   */
  App.WizardStep6Controller = Em.Controller.extend({

    hosts: [],
    /**
     * List of components info about selecting/deselecting status for components.
     *
     * @type {Array}
     *  @item {Em.Object}
     *    @property name {String} - component name
     *    @property label {String} - component display name
     *    @property allChecked {Boolean} - all checkboxes are checked
     *    @property noChecked {Boolean} - no checkboxes checked
     */
    headers: [],

    /**
     * true - assign ZK, HB
     * false - slaves and clients
     */
    isMasters: false,
    isLoaded: false,

    components: require('data/service_components'),

    isAddHostWizard: function () {
      return this.get('content.controllerName') === 'addHostController';
    }.property('content.controllerName'),

    isInstallerWizard: function () {
      return this.get('content.controllerName') === 'installerController';
    }.property('content.controllerName'),

    isAddServiceWizard: function() {
      return this.get('content.controllerName') === 'addServiceController';
    }.property('content.controllerName'),

    /**
     * verify condition that at least one checkbox of each component was checked
     */
    clearError: function () {
      var self = this;
      var isError = false;
      var err = false;
      var hosts = this.get('hosts');
      var headers = this.get('headers');
      var headersMap = {};

      headers.forEach(function (header) {
        headersMap[header.name] = true;
      });
      hosts.forEach(function (host) {
        host.get('checkboxes').forEach(function (checkbox) {
          if (headersMap[checkbox.get('component')]) {
            headersMap[checkbox.get('component')] = !checkbox.get('checked');
          }
        });
      });
      for (var i in headersMap) {
        err |= headersMap[i];
      }

      if (!err) {
        this.set('errorMessage', '');
      }

      if (this.get('isAddHostWizard')) {
        if (this.get('isMasters')) {
          this.set('errorMessage', '');
        }
        else {
          hosts.forEach(function (host) {
            isError = false;
            headers.forEach(function (header) {
              isError |= host.get('checkboxes').findProperty('title', header.get('label')).checked;
            });
            isError = !isError;
            if (!isError) {
              self.set('errorMessage', '');
            }
          });
        }
      }
    },

    clearStep: function () {
      this.set('hosts', []);
      this.set('headers', []);
      this.clearError();
      this.set('isLoaded', false);
    },

    /**
     * Enable some service for all hosts
     * @param event
     */
    selectAllNodes: function (event) {
      this.setAllNodes(event.context.name, true);
    },

    /**
     * Disable some services for all hosts
     * @param event
     */
    deselectAllNodes: function (event) {
      this.setAllNodes(event.context.name, false);
    },

    /**
     * Enable/disable some service for all hosts
     * @param {String} component - component name
     * @param {Boolean} checked - true - enable, false - disable
     */
    setAllNodes: function (component, checked) {
      this.get('hosts').forEach(function (host) {
        host.get('checkboxes').filterProperty('isInstalled', false).forEach(function (checkbox) {
          if (checkbox.get('component') === component) {
            checkbox.set('checked', checked);
          }
        });
      });
      this.checkCallback(component);
    },

    /**
     * Return whether service was selected or not
     * @param name serviceName
     * @return {*}
     */
    isServiceSelected: function (name) {
      return !!(this.get('content.services').findProperty('serviceName', name) &&
        this.get('content.services').findProperty('serviceName', name).get('isSelected'));
    },

    /**
     * Checkbox check callback
     * @param {String} component
     */
    checkCallback: function (component) {
      var header = this.get('headers').findProperty('name', component);
      var hosts = this.get('hosts');
      var allTrue = true;
      var allFalse = true;
      hosts.forEach(function (host) {
        host.get('checkboxes').forEach(function (checkbox) {
          if (checkbox.get('component') === component && !checkbox.get('isInstalled')) {
            allTrue &= checkbox.get('checked');
            allFalse &= !checkbox.get('checked');
          }
        });
      });
      header.set('allChecked', allTrue);
      header.set('noChecked', allFalse);
      this.clearError();
    },

    getComponentDisplayName: function (componentName) {
      return this.get('components').findProperty('component_name', componentName).display_name
    },

    loadStep: function () {

      var self = this;

      console.log("WizardStep6Controller: Loading step6: Assign Slaves");
      this.clearStep();

      var headers = [];

      if (this.get('isMasters')) {
        if (this.isServiceSelected('HBASE') && App.supports.multipleHBaseMasters) {
          headers.pushObject(Em.Object.create({
            name: 'HBASE_MASTER',
            label: self.getComponentDisplayName('HBASE_MASTER')
          }));
        }
        if (this.isServiceSelected('ZOOKEEPER')) {
          headers.pushObject(Em.Object.create({
            name: 'ZOOKEEPER_SERVER',
            label: self.getComponentDisplayName('ZOOKEEPER_SERVER')
          }));
        }
      }
      else {
        if (this.isServiceSelected('HDFS')) {
          headers.pushObject(Ember.Object.create({
            name: 'DATANODE',
            label: self.getComponentDisplayName('DATANODE')
          }));
        }
        if (this.isServiceSelected('MAPREDUCE')) {
          headers.pushObject(Em.Object.create({
            name: 'TASKTRACKER',
            label: self.getComponentDisplayName('TASKTRACKER')
          }));
        }
        if (this.isServiceSelected('YARN')) {
          headers.pushObject(Em.Object.create({
            name: 'NODEMANAGER',
            label: self.getComponentDisplayName('NODEMANAGER')
          }));
        }
        if (this.isServiceSelected('HBASE')) {
          headers.pushObject(Em.Object.create({
            name: 'HBASE_REGIONSERVER',
            label: self.getComponentDisplayName('HBASE_REGIONSERVER')
          }));
        }
        if (this.isServiceSelected('STORM')) {
          headers.pushObject(Em.Object.create({
            name: 'SUPERVISOR',
            label: self.getComponentDisplayName('SUPERVISOR')
          }));
        }
        headers.pushObject(Ember.Object.create({
          name: 'CLIENT',
          label: self.getComponentDisplayName('CLIENT')
        }));
      }

      headers.forEach(function (header) {
        header.setProperties({ allChecked: false, noChecked: true });
      });

      this.get('headers').pushObjects(headers);

      this.render();
      if (this.get('isMasters')) {
        if (this.get('content.skipMasterStep')) {
          App.router.send('next');
        }
      }
      else {
        if (this.get('content.skipSlavesStep')) {
          App.router.send('next');
        }
      }
    },

    /**
     * Get active host names
     * @return {Array}
     */
    getHostNames: function () {
      var hostInfo = this.get('content.hosts');
      var hostNames = [];
      for (var index in hostInfo) {
        if (hostInfo[index].bootStatus === 'REGISTERED') {
          hostNames.push(hostInfo[index].name);
        }
      }
      return hostNames;
    },

    /**
     * Load all data needed for this module. Then it automatically renders in template
     */
    render: function () {
      var hostsObj = [];
      var masterHosts = [];
      var headers = this.get('headers');
      var masterHostNames = this.get('content.masterComponentHosts').mapProperty('hostName').uniq();

      this.getHostNames().forEach(function (_hostName) {
        var hasMaster = masterHostNames.contains(_hostName);

        var obj = Em.Object.create({
          hostName: _hostName,
          hasMaster: hasMaster,
          checkboxes: []
        });

        headers.forEach(function (header) {
          obj.checkboxes.pushObject(Em.Object.create({
            component: header.name,
            title: header.label,
            checked: false,
            isInstalled: false
          }));
        });

        if (hasMaster) {
          masterHosts.pushObject(obj)
        } else {
          hostsObj.pushObject(obj);
        }
      });
      //hosts with master components should be in the beginning of list
      hostsObj.unshift.apply(hostsObj, masterHosts);

      if (this.get('isMasters')) {
        hostsObj = this.selectMasterComponents(hostsObj);
      } else {
        hostsObj = this.renderSlaves(hostsObj);
      }

      this.set('hosts', hostsObj);
      headers.forEach(function (header) {
        this.checkCallback(header.get('name'));
      }, this);
      this.set('isLoaded', true);
    },

    /**
     *
     * @param {Array} hostsObj
     * @return {Array}
     */
    renderSlaves: function (hostsObj) {
      var self = this;
      var headers = this.get('headers');
      var slaveComponents = this.get('content.slaveComponentHosts');
      if (!slaveComponents) { // we are at this page for the first time
        var client_is_set = false;
        hostsObj.forEach(function (host) {
          var checkboxes = host.get('checkboxes');
          checkboxes.setEach('checked', !host.hasMaster);
          checkboxes.setEach('isInstalled', false);
          checkboxes.findProperty('title', headers.findProperty('name', 'CLIENT').get('label')).set('checked', false);
          // First not Master should have Client (only first!)
          if (!client_is_set) {
            if (self.isServiceSelected("HDFS")) {
              var checkboxDatanode = checkboxes.findProperty('title', headers.findProperty('name', 'DATANODE').get('label'));
              if (checkboxDatanode && checkboxDatanode.get('checked')) {
                checkboxes.findProperty('title', headers.findProperty('name', 'CLIENT').get('label')).set('checked', true);
                client_is_set = true;
              }
            }
          }
        });

        if (this.get('isInstallerWizard') && hostsObj.everyProperty('hasMaster', true)) {
          var lastHost = hostsObj[hostsObj.length - 1];
          lastHost.get('checkboxes').setEach('checked', true);
        }
      }
      else {
        this.get('headers').forEach(function (header) {
          var nodes = slaveComponents.findProperty('componentName', header.get('name'));
          if (nodes) {
            nodes.hosts.forEach(function (_node) {
              var node = hostsObj.findProperty('hostName', _node.hostName);
              if (node) {
                node.get('checkboxes').findProperty('title', header.get('label')).set('checked', true);
                node.get('checkboxes').findProperty('title', header.get('label')).set('isInstalled', _node.isInstalled);
              }
            });
          }
        });
      }
      return hostsObj;
    },

    /**
     * select checkboxes which correspond to master components
     *
     * @param {Array} hostsObj
     * @return {Array}
     */
    selectMasterComponents: function (hostsObj) {
      var masterComponentHosts = this.get('content.masterComponentHosts');
      console.log('Master components selected on:', masterComponentHosts.mapProperty('hostName').uniq().join(", "));

      if (masterComponentHosts) {
        masterComponentHosts.forEach(function (item) {
          var host = hostsObj.findProperty('hostName', item.hostName);
          if (host) {
            var checkbox = host.get('checkboxes').findProperty('component', item.component);
            if (checkbox) {
              checkbox.set('checked', true);
            }
          }
        });
      }
      return hostsObj;
    },

    /**
     * Return list of master components for specified <code>hostname</code>
     * @param hostName
     * @return {*}
     */
    getMasterComponentsForHost: function (hostName) {
      return this.get('content.masterComponentHosts').filterProperty('hostName', hostName).mapProperty('component');
    },


    /**
     * Validate form. Return do we have errors or not
     * @return {Boolean}
     */
    validate: function () {

      if (this.get('isAddHostWizard')) {
        return this.validateEachHost(Em.I18n.t('installer.step6.error.mustSelectOneForHost'));
      }
      else {
        if (this.get('isInstallerWizard')) {
         return this.validateEachComponent() && this.validateEachHost(Em.I18n.t('installer.step6.error.mustSelectOneForSlaveHost'));
        }
        else {
          if(this.get('isAddServiceWizard')) {
            return this.validateEachComponent();
          }
        }
      }
    },

    /**
     * Validate all components for each host. Return do we have errors or not
     * @return {Boolean}
     */
    validateEachHost: function (errorMsg) {

      var isError = false;
      var hosts = this.get('hosts');
      var headers = this.get('headers');
      for (var i = 0; i < hosts.length; i++) {
        if (this.get('isInstallerWizard') && this.get('content.masterComponentHosts').someProperty('hostName', hosts[i].hostName)) {
          continue;
        }
        var checkboxes = hosts[i].get('checkboxes');
        isError = false;
        headers.forEach(function (header) {
          isError |= checkboxes.findProperty('title', header.get('label')).checked;
        });
        isError = !isError;
        if (isError) {
          this.set('errorMessage', errorMsg);
          break;
        }
      }
      return !isError;
    },

    /**
     * Validate a component for all hosts. Return do we have errors or not
     * @return {Boolean}
     */
    validateEachComponent: function () {
      var isError = false;
      var hosts = this.get('hosts');
      var headers = this.get('headers');
      headers.forEach(function (header) {
        var all_false = true;
        hosts.forEach(function (host) {
          var checkboxes = host.get('checkboxes');
          all_false = all_false && !checkboxes.findProperty('title', header.get('label')).checked;
        });
        isError = isError || all_false;
      });
      if (isError) {
        this.set('errorMessage', Em.I18n.t('installer.step6.error.mustSelectOne'));
      }
      return !isError;
    }

  });
  
});
window.require.register("controllers/wizard/step7_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');
  /**
   * By Step 7, we have the following information stored in App.db and set on this
   * controller by the router.
   *
   *   selectedServices: App.db.selectedServices (the services that the user selected in Step 4)
   *   masterComponentHosts: App.db.masterComponentHosts (master-components-to-hosts mapping the user selected in Step 5)
   *   slaveComponentHosts: App.db.slaveComponentHosts (slave-components-to-hosts mapping the user selected in Step 6)
   *
   */

  App.WizardStep7Controller = Em.Controller.extend({

    name: 'wizardStep7Controller',

    stepConfigs: [], //contains all field properties that are viewed in this step

    selectedService: null,

    slaveHostToGroup: null,

    secureConfigs: require('data/secure_mapping'),

    miscModalVisible: false, //If miscConfigChange Modal is shown

    gangliaAvailableSpace: null,

    gangliaMoutDir:'/',

    overrideToAdd: null,

    isInstaller: true,

    configGroups: [],

    groupsToDelete: [],

    selectedConfigGroup: null,
    /**
     * config tags of actually installed services
     */
    serviceConfigTags: [],

    serviceConfigsData: require('data/service_configs'),

    isAdvancedConfigLoaded: true,

    isSubmitDisabled: function () {
      return (!this.stepConfigs.filterProperty('showConfig', true).everyProperty('errorCount', 0) || this.get("miscModalVisible"));
    }.property('stepConfigs.@each.errorCount', 'miscModalVisible'),

    selectedServiceNames: function () {
      return this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false).mapProperty('serviceName');
    }.property('content.services').cacheable(),

    allSelectedServiceNames: function () {
      return this.get('content.services').filterProperty('isSelected').mapProperty('serviceName');
    }.property('content.services').cacheable(),

    installedServiceNames: function () {
      var serviceNames = this.get('content.services').filterProperty('isInstalled').mapProperty('serviceName');
      if(this.get('content.controllerName') !== 'installerController') {
        return serviceNames.without('SQOOP').without('HCATALOG');
      }
      return serviceNames;
    }.property('content.services').cacheable(),

    masterComponentHosts: function () {
      return this.get('content.masterComponentHosts');
    }.property('content.masterComponentHosts'),

    slaveComponentHosts: function () {
      return this.get('content.slaveGroupProperties');
    }.property('content.slaveGroupProperties', 'content.slaveComponentHosts'),

    customData: [],

    clearStep: function () {
      this.get('serviceConfigTags').clear();
      this.get('stepConfigs').clear();
      this.set('filter', '');
      this.get('filterColumns').setEach('selected', false);
    },
    /**
     *  Load config groups for installed services
     */
    loadInstalledServicesConfigGroups: function (servicesNames) {
      servicesNames.forEach(function(serviceName) {
        App.ajax.send({
          name: 'config.tags_and_groups',
          sender: this,
          data: {
            serviceName: serviceName,
            serviceConfigsDef: App.config.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName)
          },
          success: 'loadServiceTagsSuccess'
        });
      }, this);
    },

   /**
    * Load config groups success callback
    */
    loadServiceTagsSuccess: function (data, opt, params) {
      var serviceConfigsDef = params.serviceConfigsDef;
      var serviceName = params.serviceName;
      var service = this.get('stepConfigs').findProperty('serviceName', serviceName);
      // Create default configuration group
      var defaultConfigGroupHosts = this.get('wizardController.allHosts').mapProperty('hostName');
      var selectedConfigGroup;
      var siteToTagMap = {};
      for (var site in data.Clusters.desired_configs) {
        if (serviceConfigsDef.sites.indexOf(site) > -1) {
          siteToTagMap[site] = data.Clusters.desired_configs[site].tag;
        }
      }
      this.loadedClusterSiteToTagMap = siteToTagMap;
      //parse loaded config groups
      if (App.supports.hostOverrides) {
        var configGroups = [];
        if (data.config_groups.length) {
          data.config_groups.forEach(function (item) {
            item = item.ConfigGroup;
            if (item.tag === serviceName) {
              var groupHosts = item.hosts.mapProperty('host_name');
              var newConfigGroup = App.ConfigGroup.create({
                id: item.id,
                name: item.group_name,
                description: item.description,
                isDefault: false,
                parentConfigGroup: null,
                service: App.Service.find().findProperty('serviceName', item.tag),
                hosts: groupHosts,
                configSiteTags: []
              });
              groupHosts.forEach(function (host) {
                defaultConfigGroupHosts = defaultConfigGroupHosts.without(host);
              }, this);
              item.desired_configs.forEach(function (config) {
                newConfigGroup.configSiteTags.push(App.ConfigSiteTag.create({
                  site: config.type,
                  tag: config.tag
                }));
              }, this);
              configGroups.push(newConfigGroup);
            }
          }, this);
        }
      }
      var defaultConfigGroup = App.ConfigGroup.create({
        name: App.Service.DisplayNames[serviceName] + " Default",
        description: "Default cluster level " + serviceName + " configuration",
        isDefault: true,
        hosts: defaultConfigGroupHosts,
        parentConfigGroup: null,
        service: Em.Object.create({
          id: serviceName
        }),
        serviceName: serviceName,
        configSiteTags: []
      });
      if (!selectedConfigGroup) {
        selectedConfigGroup = defaultConfigGroup;
      }
      configGroups = configGroups.sortProperty('name');
      configGroups.unshift(defaultConfigGroup);
      if (App.supports.hostOverrides) {
        service.set('configGroups', configGroups);
        var loadedGroupToOverrideSiteToTagMap = {};
        if (App.supports.hostOverrides) {
          var configGroupsWithOverrides = selectedConfigGroup.get('isDefault') ? service.get('configGroups') : [selectedConfigGroup];
          configGroupsWithOverrides.forEach(function (item) {
            var groupName = item.get('name');
            loadedGroupToOverrideSiteToTagMap[groupName] = {};
            item.get('configSiteTags').forEach(function (siteTag) {
              var site = siteTag.get('site');
              loadedGroupToOverrideSiteToTagMap[groupName][site] = siteTag.get('tag');
            }, this);
          }, this);
        }
        App.config.loadServiceConfigGroupOverrides(service.get('configs'), loadedGroupToOverrideSiteToTagMap, service.get('configGroups'));
        var serviceConfig = App.config.createServiceConfig(serviceName);
        if (serviceConfig.get('serviceName') === 'HDFS') {
          App.config.OnNnHAHideSnn(serviceConfig);
        }
        service.set('selectedConfigGroup', selectedConfigGroup);
        this.loadComponentConfigs(service.get('configs'), serviceConfig, service);
      }
      service.set('configs', serviceConfig.get('configs'));
    },

    loadComponentConfigs: function (configs, componentConfig, component) {
      var localDB = App.router.get('mainServiceInfoConfigsController').getInfoForDefaults();
      var recommendedDefaults = {};
      var s = this.get('serviceConfigsData').findProperty('serviceName', component.get('serviceName'));
      var defaultGroupSelected = component.get('selectedConfigGroup.isDefault');
      var defaults = [];
      if (s.defaultsProviders) {
        s.defaultsProviders.forEach(function(defaultsProvider) {
          var d = defaultsProvider.getDefaults(localDB);
          defaults.push(d);
          for (var name in d) {
            recommendedDefaults[name] = d[name];
          }
        });
      }
      if (s.configsValidator) {
        s.configsValidator.set('recommendedDefaults', recommendedDefaults);
      }
      configs.forEach(function (serviceConfigProperty) {
        console.log("config", serviceConfigProperty);
        if (!serviceConfigProperty) return;
        var overrides = serviceConfigProperty.get('overrides');
        // we will populate the override properties below
        serviceConfigProperty.set('overrides', null);

        if (serviceConfigProperty.isOverridable === undefined) {
          serviceConfigProperty.set('isOverridable', true);
        }
        if (serviceConfigProperty.displayType === 'checkbox') {
          switch (serviceConfigProperty.value) {
            case 'true':
              serviceConfigProperty.set('value', true);
              serviceConfigProperty.set('defaultValue', true);
              break;
            case 'false':
              serviceConfigProperty.set('value', false);
              serviceConfigProperty.set('defaultValue', false);
              break;
          }
        }
        if (serviceConfigProperty.get('serviceName') === component.get('serviceName')) {
          if (s.configsValidator) {
            var validators = s.configsValidator.get('configValidators');
            for (var validatorName in validators) {
              if (serviceConfigProperty.name == validatorName) {
                serviceConfigProperty.set('serviceValidator', s.configsValidator);
              }
            }
          }
          console.log("config result", serviceConfigProperty);
        } else {
          serviceConfigProperty.set('isVisible', false);
        }
        if (overrides != null) {
          overrides.forEach(function (override) {
            var newSCP = App.ServiceConfigProperty.create(serviceConfigProperty);
            newSCP.set('value', override.value);
            newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
            newSCP.set('parentSCP', serviceConfigProperty);
            if (App.supports.hostOverrides && defaultGroupSelected) {
              var group = component.get('configGroups').findProperty('name', override.group.get('name'));
              // prevent cycle in proto object, clean link
              if (group.get('properties').length == 0)
                group.set('properties', Em.A([]));
              group.get('properties').push(newSCP);
              newSCP.set('group', override.group);
              newSCP.set('isEditable', false);
            }
            var parentOverridesArray = serviceConfigProperty.get('overrides');
            if (parentOverridesArray == null) {
              parentOverridesArray = Ember.A([]);
              serviceConfigProperty.set('overrides', parentOverridesArray);
            }
            serviceConfigProperty.get('overrides').pushObject(newSCP);
            console.debug("createOverrideProperty(): Added:", newSCP, " to main-property:", serviceConfigProperty)
          }, this);
        } else {
          serviceConfigProperty.set('overrides', Ember.A([]));
        }
        if (App.get('isAdmin')) {
          if(defaultGroupSelected && !this.get('isHostsConfigsPage')){
            serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
          } else {
            serviceConfigProperty.set('isEditable', false);
          }
        } else {
          serviceConfigProperty.set('isEditable', false);
        }
        componentConfig.get('configs').pushObject(serviceConfigProperty);
        serviceConfigProperty.validate();
      }, this);
      var overrideToAdd = this.get('overrideToAdd');
      if (overrideToAdd) {
        overrideToAdd = componentConfig.configs.findProperty('name', overrideToAdd.name);
        if (overrideToAdd) {
          this.addOverrideProperty(overrideToAdd);
          component.set('overrideToAdd', null);
        }
      }
    },
    /**
     *  Resolve dependency between configs.
     *  @param serviceName {String}
     *  @param configs {Mixed}
     */
    resolveServiceDependencyConfigs: function (serviceName, configs) {
      switch (serviceName) {
        case 'STORM':
          this.resolveStormConfigs(configs);
          break;
        default:
          break;
      }
    },

    resolveStormConfigs: function(configs) {
      var dependentConfigs, gangliaServerHost;
      dependentConfigs = ['nimbus.childopts', 'supervisor.childopts', 'worker.childopts'];
      // if Ganglia selected or installed, set ganglia host to configs
      if (this.get('installedServiceNames').contains('STORM') && this.get('installedServiceNames').contains('GANGLIA')) return;
      if (this.get('allSelectedServiceNames').contains('GANGLIA') || this.get('installedServiceNames').contains('GANGLIA')) {
        gangliaServerHost = this.get('wizardController').getDBProperty('masterComponentHosts').findProperty('component', 'GANGLIA_SERVER').hostName;
        dependentConfigs.forEach(function(configName) {
          var config = configs.findProperty('name', configName);
          var replaceStr = '.jar=host=';
          config.value = config.defaultValue = config.value.replace(replaceStr, replaceStr+gangliaServerHost);
          config.forceUpdate = true;
        }, this);
      }
    },

    /**
     * On load function
     */
    loadStep: function () {
      console.log("TRACE: Loading step7: Configure Services");
      if (!this.get('isAdvancedConfigLoaded')) {
        return;
      }
      this.clearStep();
      //STEP 1: Load advanced configs
      var advancedConfigs = this.get('content.advancedServiceConfig');
      //STEP 2: Load on-site configs by service from local DB
      var storedConfigs = this.get('content.serviceConfigProperties');
      //STEP 3: Merge pre-defined configs with loaded on-site configs
      var configs = App.config.mergePreDefinedWithStored(storedConfigs, advancedConfigs, this.get('selectedServiceNames').concat(this.get('installedServiceNames')));
      //STEP 4: Add advanced configs
      App.config.addAdvancedConfigs(configs, advancedConfigs);
      //STEP 5: Add custom configs
      App.config.addCustomConfigs(configs);
      //put properties from capacity-scheduler.xml into one config with textarea view
      if (this.get('allSelectedServiceNames').contains('YARN') && !App.supports.capacitySchedulerUi) {
        configs = App.config.fileConfigsIntoTextarea(configs, 'capacity-scheduler.xml');
      }
      this.set('groupsToDelete', this.get('wizardController').getDBProperty('groupsToDelete') || []);
      var localDB = {
        hosts: this.get('wizardController').getDBProperty('hosts'),
        masterComponentHosts: this.get('wizardController').getDBProperty('masterComponentHosts'),
        slaveComponentHosts: this.get('wizardController').getDBProperty('slaveComponentHosts')
      };
      if (this.get('wizardController.name') === 'addServiceController') {
        this.getConfigTags();
        this.setInstalledServiceConfigs(this.get('serviceConfigTags'), configs);
      }
      if (this.get('allSelectedServiceNames').contains('STORM') || this.get('installedServiceNames').contains('STORM')) {
        this.resolveServiceDependencyConfigs('STORM', configs);
      }
      //STEP 6: Distribute configs by service and wrap each one in App.ServiceConfigProperty (configs -> serviceConfigs)
      var serviceConfigs = App.config.renderConfigs(configs, storedConfigs, this.get('allSelectedServiceNames'), this.get('installedServiceNames'), localDB);
      if (this.get('wizardController.name') === 'addServiceController') {
        serviceConfigs.setEach('showConfig', true);
        serviceConfigs.setEach('selected', false);
        this.get('selectedServiceNames').forEach(function(serviceName) {
          if(!serviceConfigs.findProperty('serviceName', serviceName)) return;
          serviceConfigs.findProperty('serviceName', serviceName).set('selected', true);
        });

        // Remove SNameNode if HA is enabled
        if (App.get('isHaEnabled')) {
          configs = serviceConfigs.findProperty('serviceName', 'HDFS').configs;
          var removedConfigs = configs.filterProperty('category', 'SNameNode');
          removedConfigs.map(function(config) {
            configs = configs.without(config);
          });
          serviceConfigs.findProperty('serviceName', 'HDFS').configs = configs;
        }
      }

      this.set('stepConfigs', serviceConfigs);
      if (App.supports.hostOverridesInstaller) {
        this.loadConfigGroups(this.get('content.configGroups'));
        if (this.get('installedServiceNames').length > 0)
          this.loadInstalledServicesConfigGroups(this.get('installedServiceNames'));
      }
      this.activateSpecialConfigs();
      if (this.get('wizardController.name') === 'addServiceController') {
        this.set('selectedService', this.get('stepConfigs').filterProperty('selected', true).get('firstObject'));
      }
      else {
        this.set('selectedService', this.get('stepConfigs').filterProperty('showConfig', true).objectAt(0));
      }
      if (this.get('content.skipConfigStep')) {
        App.router.send('next');
      }
    },
    getConfigTags: function() {
      App.ajax.send({
        name: 'config.tags.sync',
        sender: this,
        success: 'getConfigTagsSuccess'
      });
    },

    getConfigTagsSuccess: function (data, opt, params) {
      var installedServiceSites = [];
      this.get('serviceConfigsData').filter(function (service) {
        if (this.get('installedServiceNames').contains(service.serviceName)){
          installedServiceSites = installedServiceSites.concat(service.sites);
        }
      }, this);
      installedServiceSites = installedServiceSites.uniq();
      var serviceConfigTags = [];
      for (var site in data.Clusters.desired_configs) {
        if (installedServiceSites.contains(site)) {
          serviceConfigTags.push({
            siteName: site,
            tagName: data.Clusters.desired_configs[site].tag,
            newTagName: null
          });
        }
      }
      this.set('serviceConfigTags', serviceConfigTags);
    },

    /**
     * set configs actual values from server
     * @param serviceConfigTags
     * @param configs
     */
    setInstalledServiceConfigs: function (serviceConfigTags, configs) {
      var configsMap = {};
      App.router.get('configurationController').getConfigsByTags(serviceConfigTags).forEach(function(configSite){
        $.extend(configsMap, configSite.properties);
      });
      configs.forEach(function (_config) {
        if (configsMap[_config.name] !== undefined) {
          // prevent overriding already edited properties
          if (_config.defaultValue != configsMap[_config.name])
            _config.value = configsMap[_config.name];
          _config.defaultValue = configsMap[_config.name];
          App.config.handleSpecialProperties(_config);
        }
      })
    },

    setGroupsToDelete: function(groups) {
      var groupsToDelete = this.get('groupsToDelete');
      groups.forEach(function(group) {
        if (group.get('id'))
          groupsToDelete.push({
            id: group.get('id')
          });
      });
      this.get('wizardController').setDBProperty('groupsToDelete', groupsToDelete);
    },
    selectedServiceObserver: function () {
      if (App.supports.hostOverridesInstaller && this.get('selectedService') && (this.get('selectedService.serviceName') !== 'MISC')) {
        var serviceGroups = this.get('selectedService.configGroups');
        serviceGroups.forEach(function (item, index, array) {
          if (item.isDefault) {
            array.unshift(item);
            array.splice(index + 1, 1);
          }
        });
        this.set('configGroups', serviceGroups);
        this.set('selectedConfigGroup', serviceGroups.findProperty('isDefault'));
      }
    }.observes('selectedService.configGroups.@each'),
    /**
     * load default groups for each service in case of initial load
     * @param serviceConfigGroups
     */
    loadConfigGroups: function (serviceConfigGroups) {
      var services = this.get('stepConfigs');
      var hosts = this.get('wizardController.allHosts').mapProperty('hostName');
      services.forEach(function (service) {
        if (service.get('serviceName') === 'MISC') return;
        var serviceRawGroups = serviceConfigGroups.filterProperty('service.id', service.serviceName);
        if (!serviceRawGroups.length) {
          service.set('configGroups', [
            App.ConfigGroup.create({
              name: App.Service.DisplayNames[service.serviceName] + " Default",
              description: "Default cluster level " + service.serviceName + " configuration",
              isDefault: true,
              hosts: Em.copy(hosts),
              service: Em.Object.create({
                id: service.serviceName
              }),
              serviceName: service.serviceName
            })
          ]);
        } else {
          var defaultGroup = App.ConfigGroup.create(serviceRawGroups.findProperty('isDefault'));
          var serviceGroups = service.get('configGroups');
          serviceRawGroups.filterProperty('isDefault', false).forEach(function (configGroup) {
            var readyGroup = App.ConfigGroup.create(configGroup);
            var wrappedProperties = [];
            readyGroup.get('properties').forEach(function(property){
              wrappedProperties.pushObject(App.ServiceConfigProperty.create(property));
            });
            wrappedProperties.setEach('group', readyGroup);
            readyGroup.set('properties', wrappedProperties);
            readyGroup.set('parentConfigGroup', defaultGroup);
            serviceGroups.pushObject(readyGroup);
          });
          defaultGroup.set('childConfigGroups', serviceGroups);
          serviceGroups.pushObject(defaultGroup);
        }
      });
    },

    selectConfigGroup: function (event) {
      this.set('selectedConfigGroup', event.context);
    },

    /**
     * rebuild list of configs switch of config group:
     * on default - display all configs from default group and configs from non-default groups as disabled
     * on non-default - display all from default group as disabled and configs from selected non-default group
     */
    switchConfigGroupConfigs: function () {
      var serviceConfigs = this.get('selectedService.configs');
      var selectedGroup = this.get('selectedConfigGroup');
      var overrideToAdd = this.get('overrideToAdd');
      var isServiceInstalled = this.get('installedServiceNames').contains(this.get('selectedService.serviceName'));
      if(!selectedGroup) return;
      var displayedConfigGroups = (selectedGroup.get('isDefault')) ?
          this.get('selectedService.configGroups').filterProperty('isDefault', false) :
          [this.get('selectedConfigGroup')];
      var overrides = [];

      displayedConfigGroups.forEach(function (group) {
        overrides.pushObjects(group.get('properties'));
      });
      serviceConfigs.forEach(function (config) {
        var configOverrides = overrides.filterProperty('name', config.get('name'));
        var isEditable = config.get('isEditable');
        if (isServiceInstalled) {
          isEditable = (!isEditable && !config.get('isReconfigurable')) ? false : selectedGroup.get('isDefault');
        } else {
          isEditable = selectedGroup.get('isDefault');
        }
        config.set('isEditable', isEditable);
        if (overrideToAdd && overrideToAdd.get('name') === config.get('name')) {
          configOverrides.push(this.addOverrideProperty(config));
          this.set('overrideToAdd', null);
        }
        configOverrides.setEach('isEditable', !selectedGroup.get('isDefault'));
        configOverrides.setEach('parentSCP', config);
        config.set('overrides', configOverrides);
      }, this);
    }.observes('selectedConfigGroup'),
    /**
     * create overriden property and push it into Config group
     * @param serviceConfigProperty
     * @return {*}
     */
    addOverrideProperty: function (serviceConfigProperty) {
      var overrides = serviceConfigProperty.get('overrides') || [];
      var newSCP = App.ServiceConfigProperty.create(serviceConfigProperty);
      var group = this.get('selectedService.configGroups').findProperty('name', this.get('selectedConfigGroup.name'));
      newSCP.set('group', group);
      newSCP.set('value', '');
      newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
      newSCP.set('parentSCP', serviceConfigProperty);
      newSCP.set('isEditable', true);
      group.get('properties').pushObject(newSCP);
      overrides.pushObject(newSCP);
      return newSCP;
    },

    manageConfigurationGroup: function () {
      App.router.get('mainServiceInfoConfigsController').manageConfigurationGroups(this);
    },
    /**
     * Filter text will be located here
     */
    filter: '',

    /**
     * Dropdown menu items in filter combobox
     */
    filterColumns: function () {
      var result = [];
      for (var i = 1; i < 2; i++) {
        result.push(Ember.Object.create({
          name: this.t('common.combobox.dropdown.' + i),
          selected: false
        }));
      }
      return result;
    }.property(),
     /**
     * make some configs visible depending on active services
     */
    activateSpecialConfigs: function () {
      var miscConfigs = this.get('stepConfigs').findProperty('serviceName', 'MISC').configs;
      miscConfigs = App.config.miscConfigVisibleProperty(miscConfigs, this.get('selectedServiceNames'));
    },

    submit: function () {
      if (!this.get('isSubmitDisabled')) {
        App.router.send('next');
      }
    }, 
    
    /**
     * Provides service component name and display-name information for 
     * the current selected service. 
     */
    getCurrentServiceComponents: function () {
      var selectedServiceName = this.get('selectedService.serviceName');
      var masterComponents = this.get('content.masterComponentHosts');
      var slaveComponents = this.get('content.slaveComponentHosts');
      var scMaps = require('data/service_components');
      
      var validComponents = Ember.A([]);
      var seenComponents = {};
      masterComponents.forEach(function(component){
        var cn = component.component;
        var cdn = component.display_name;
        if(component.serviceId===selectedServiceName && !seenComponents[cn]){
          validComponents.pushObject(Ember.Object.create({
            componentName: cn,
            displayName: cdn,
            selected: false
          }));
          seenComponents[cn] = cn;
        }
      });
      slaveComponents.forEach(function(component){
        var cn = component.componentName;
        var cdn = component.displayName;
        var componentDef = scMaps.findProperty('component_name', cn);
        if(componentDef!=null && selectedServiceName===componentDef.service_name && !seenComponents[cn]){
          validComponents.pushObject(Ember.Object.create({
            componentName: cn,
            displayName: cdn,
            selected: false
          }));
          seenComponents[cn] = cn;
        }
      });
      return validComponents;
    }.property('content')

  });
  
});
window.require.register("controllers/wizard/step8_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.WizardStep8Controller = Em.Controller.extend({
    name: 'wizardStep8Controller',

    /**
     * List of raw data about cluster that should be displayed
     * @type {Array}
     */
    rawContent: require('data/review_configs'),

    /**
     * @type {Object[]}
     */
    totalHosts: [],

    /**
     * List of data about cluster (based on formatted <code>rawContent</code>)
     * @type {Object[]}
     */
    clusterInfo: [],

    /**
     * List of services with components assigned to hosts
     * @type {Object[]}
     */
    services: [],

    /**
     * @type {Object[]}
     */
    configs: [],

    /**
     * @type {Object[]}
     */
    globals: [],

    /**
     * List of ajax-request to be executed
     * @type {Array}
     */
    ajaxQueue: [],

    /**
     * All configs
     * @type {Array}
     */
    configMapping: function(){
      return App.config.get('configMapping').all(true);
    }.property('App.config.configMapping'),

    /**
     *
     */
    slaveComponentConfig: null,

    /**
     * Should Submit button be disabled
     * @type {bool}
     */
    isSubmitDisabled: false,

    /**
     * Should Back button be disabled
     * @type {bool}
     */
    isBackBtnDisabled: false,

    /**
     * Is error appears while <code>ajaxQueue</code> executes
     * @type {bool}
     */
    hasErrorOccurred: false,

    /**
     * Are services installed
     * Used to hide Deploy Progress Bar
     * @type {bool}
     */
    servicesInstalled: false,

    /**
     * List of service config tags
     * @type {Object[]}
     */
    serviceConfigTags: [],

    /**
     * Is cluster security enabled
     * @type {bool}
     */
    securityEnabled: function() {
      return App.router.get('mainAdminSecurityController.securityEnabled');
    }.property('App.router.mainAdminSecurityController.securityEnabled'),
    /**
     * During page save time, we set the host overrides to the server.
     * The new host -> site:tag map is stored below. This will be
     * useful during save, to update the host's host components. Also,
     * it will be useful in deletion of overrides.
     *
     * Example:
     * {
     *  'hostname1': {
     *    'global': {
     *      'tagName': 'tag3187261938_hostname1',
     *      'map': {
     *        'hadoop_heapsize': '2048m'
     *      }
     *    }
     *  }
     * }
     *
     * @see loadedHostToOverrideSiteToTagMap
     */
    savedHostToOverrideSiteToTagMap: {},

    /**
     * Selected config group
     * @type {Object}
     */
    selectedConfigGroup: null,

    /**
     * List of config groups
     * @type {Object[]}
     */
    configGroups: [],

    /**
     * List of selected but not installed services
     * @type {Object[]}
     */
    selectedServices: function () {
      return this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false);
    }.property('content.services').cacheable(),

    /**
     * Clear current step data
     */
    clearStep: function () {
      this.get('services').clear();
      this.get('configs').clear();
      this.get('globals').clear();
      this.get('clusterInfo').clear();
      this.get('serviceConfigTags').clear();
      this.set('servicesInstalled', false);
    },

    /**
     * Load current step data
     */
    loadStep: function () {
      console.log("TRACE: Loading step8: Review Page");
      if (this.get('content.controllerName') != 'installerController') {
        App.router.get('mainAdminSecurityController').setSecurityStatus();
      }
      this.clearStep();
      if (this.get('content.serviceConfigProperties')) {
        this.formatProperties();
        this.loadGlobals();
        this.loadConfigs();
      }
      this.loadClusterInfo();
      this.loadServices();
      this.set('isSubmitDisabled', false);
      this.set('isBackBtnDisabled', false);
    },

    /**
     * replace whitespace character with coma between directories
     */
    formatProperties: function(){
      this.get('content.serviceConfigProperties').forEach(function(_configProperty){
          _configProperty.value = App.config.trimProperty(_configProperty,false);
      });
    },

    /**
     * Load global configs and remove some of them:
     * <ul>
     *   <li>Unused DB properties for Hive</li>
     *   <li>Unused DB properties for Oozie</li>
     * <ul>
     */
    loadGlobals: function () {
      var globals = this.get('content.serviceConfigProperties').filterProperty('id', 'puppet var');
      if (globals.someProperty('name', 'hive_database')) {
        var hiveDb = globals.findProperty('name', 'hive_database');
        var hiveDbType = {name: 'hive_database_type', value: 'mysql'};

        if (hiveDb.value === 'New MySQL Database') {
          if (globals.someProperty('name', 'hive_ambari_host')) {
            globals.findProperty('name', 'hive_hostname').value = globals.findProperty('name', 'hive_ambari_host').value;
            hiveDbType.value = 'mysql';
          }
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_database'));
        } else if (hiveDb.value === 'Existing MySQL Database'){
          globals.findProperty('name', 'hive_hostname').value = globals.findProperty('name', 'hive_existing_mysql_host').value;
          hiveDbType.value = 'mysql';
          globals = globals.without(globals.findProperty('name', 'hive_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'hive_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_oracle_database'));
        } else { //existing oracle database
          globals.findProperty('name', 'hive_hostname').value = globals.findProperty('name', 'hive_existing_oracle_host').value;
          hiveDbType.value = 'oracle';
          globals = globals.without(globals.findProperty('name', 'hive_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'hive_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'hive_existing_mysql_database'));
        }
        globals.push(hiveDbType);
      }

      if (globals.someProperty('name', 'oozie_database')) {
        var oozieDb = globals.findProperty('name', 'oozie_database');
        var oozieDbType = {name:'oozie_database_type'};

        if (oozieDb.value === 'New Derby Database') {
          globals.findProperty('name', 'oozie_hostname').value = globals.findProperty('name', 'oozie_ambari_host').value;
          oozieDbType.value = 'derby';
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_database'));
        } else if (oozieDb.value === 'Existing MySQL Database') {
          globals.findProperty('name', 'oozie_hostname').value = globals.findProperty('name', 'oozie_existing_mysql_host').value;
          oozieDbType.value = 'mysql';
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_oracle_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_derby_database'));
        } else { // existing oracle database
          globals.findProperty('name', 'oozie_hostname').value = globals.findProperty('name', 'oozie_existing_oracle_host').value;
          oozieDbType.value = 'oracle';
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_ambari_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_host'));
          globals = globals.without(globals.findProperty('name', 'oozie_existing_mysql_database'));
          globals = globals.without(globals.findProperty('name', 'oozie_derby_database'));
        }
        globals.push(oozieDbType);
      }

      this.set('globals', globals);
    },

    /**
     * Load all site properties
     */
    loadConfigs: function () {
      //storedConfigs contains custom configs as well
      var serviceConfigProperties = this.get('content.serviceConfigProperties').filterProperty('id', 'site property');
      serviceConfigProperties.forEach(function(_config){
        _config.value = (typeof _config.value === "boolean") ? _config.value.toString() : _config.value;
      });
      var storedConfigs = serviceConfigProperties.filterProperty('value');
      var mappedConfigs = App.config.excludeUnsupportedConfigs(this.get('configMapping'), this.get('selectedServices').mapProperty('serviceName'));
      var uiConfigs = this.loadUiSideConfigs(mappedConfigs);
      this.set('configs', storedConfigs.concat(uiConfigs));
    },

    /**
     * Load UI configs
     * @param {Array} configMapping
     * @return {Array}
     */
    loadUiSideConfigs: function (configMapping) {
      var uiConfig = [];
      var configs = configMapping.filterProperty('foreignKey', null);
      this.addDynamicProperties(configs);
      configs.forEach(function (_config) {
        var valueWithOverrides = this.getGlobConfigValueWithOverrides(_config.templateName, _config.value, _config.name);
        uiConfig.pushObject({
          "id": "site property",
          "name": _config.name,
          "value": valueWithOverrides.value,
          "filename": _config.filename,
          "overrides": valueWithOverrides.overrides
        });
      }, this);
      var dependentConfig = $.extend(true, [], configMapping.filterProperty('foreignKey'));
      dependentConfig.forEach(function (_config) {
        App.config.setConfigValue(uiConfig, this.get('content.serviceConfigProperties'), _config, this.get('globals'));
        uiConfig.pushObject({
          "id": "site property",
          "name": _config._name || _config.name,
          "value": _config.value,
          "filename": _config.filename
        });
      }, this);
      return uiConfig;
    },

    /**
     * Add dynamic properties to configs
     * @param {Array} configs
     */
    addDynamicProperties: function(configs) {
      var templetonHiveProperty =  this.get('content.serviceConfigProperties').someProperty('name', 'templeton.hive.properties');
      if (!templetonHiveProperty) {
        configs.pushObject({
          "name": "templeton.hive.properties",
          "templateName": ["hivemetastore_host"],
          "foreignKey": null,
          "value": "hive.metastore.local=false,hive.metastore.uris=thrift://<templateName[0]>:9083,hive.metastore.sasl.enabled=yes,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse",
          "filename": "webhcat-site.xml"
        });
      }
    },

    /**
     * Format <code>content.hosts</code> from Object to Array
     * @returns {Array}
     */
    getRegisteredHosts: function () {
      var allHosts = this.get('content.hosts');
      var hosts = [];
      for (var hostName in allHosts) {
        if (allHosts[hostName].bootStatus == 'REGISTERED') {
          allHosts[hostName].hostName = allHosts[hostName].name;
          hosts.pushObject(allHosts[hostName]);
        }
      }
      return hosts;
    },

    /**
     * Set all site property that are derived from other puppet-variable
     * @param {String} templateName
     * @param {String} expression
     * @param {String} name
     * @return {Object}
     * example: <code>{
     *   value: '...',
     *   overrides: {
     *    'value1': [h1, h2],
     *    'value2': [h3]
     *   }
     * }</code>
     */
    getGlobConfigValueWithOverrides: function (templateName, expression, name) {
      var express = expression.match(/<(.*?)>/g);
      var value = expression;
      if (express == null) {
        return { value : expression, overrides: []};      // if site property do not map any global property then return the value
      }
      var overrideHostToValue = {};
      express.forEach(function (_express) {
        //console.log("The value of template is: " + _express);
        var index = parseInt(_express.match(/\[([\d]*)(?=\])/)[1]);
        if (this.get('globals').someProperty('name', templateName[index])) {
          //console.log("The name of the variable is: " + this.get('content.serviceConfigProperties').findProperty('name', templateName[index]).name);
          var globalObj = this.get('globals').findProperty('name', templateName[index]);
          var globValue = globalObj.value;
          // Hack for templeton.zookeeper.hosts
          var preReplaceValue = null;
          if (value !== null) {   // if the property depends on more than one template name like <templateName[0]>/<templateName[1]> then don't proceed to the next if the prior is null or not found in the global configs
            preReplaceValue = value;
            value = this._replaceConfigValues(name, _express, value, globValue);
          }
          if(globalObj.overrides!=null){
            globalObj.overrides.forEach(function(override){
              var ov = override.value;
              var hostsArray = override.hosts;
              hostsArray.forEach(function(host){
                if(!(host in overrideHostToValue)){
                  overrideHostToValue[host] = this._replaceConfigValues(name, _express, preReplaceValue, ov);
                }else{
                  overrideHostToValue[host] = this._replaceConfigValues(name, _express, overrideHostToValue[host], ov);
                }
              }, this);
            }, this);
          }
        } else {
          /*
           console.log("ERROR: The variable name is: " + templateName[index]);
           console.log("ERROR: mapped config from configMapping file has no corresponding variable in " +
           "content.serviceConfigProperties. Two possible reasons for the error could be: 1) The service is not selected. " +
           "and/OR 2) The service_config metadata file has no corresponding global var for the site property variable");
           */
          value = null;
        }
      }, this);

      var valueWithOverrides = {
          value: value,
          overrides: []
      };
      var overrideValueToHostMap = {};
      if(!jQuery.isEmptyObject(overrideHostToValue)){
        for(var host in overrideHostToValue){
          var hostVal = overrideHostToValue[host];
          if(!(hostVal in overrideValueToHostMap)){
            overrideValueToHostMap[hostVal] = [];
          }
          overrideValueToHostMap[hostVal].push(host);
        }
      }
      for(var val in overrideValueToHostMap){
        valueWithOverrides.overrides.push({
          value: val,
          hosts: overrideValueToHostMap[val]
        });
      }
      return valueWithOverrides;
    },

    /**
     * replace some values in config property
     * @param {string} name
     * @param {string} express
     * @param {string} value
     * @param {string} globValue
     * @return {string}
     * @private
     */
    _replaceConfigValues: function (name, express, value, globValue) {
      return value.replace(express, globValue);
    },

    /**
     * Load all info about cluster to <code>clusterInfo</code> variable
     */
    loadClusterInfo: function () {

      //Admin name
      var admin = this.rawContent.findProperty('config_name', 'Admin');
      admin.config_value = App.db.getLoginName();
      console.log("STEP8: the value of content cluster name: " + App.db.getLoginName());
      if (admin.config_value) {
        this.get('clusterInfo').pushObject(Ember.Object.create(admin));
      }

      // cluster name
      var cluster = this.rawContent.findProperty('config_name', 'cluster');
      cluster.config_value = this.get('content.cluster.name');
      console.log("STEP8: the value of content cluster name: " + this.get('content.cluster.name'));
      this.get('clusterInfo').pushObject(Ember.Object.create(cluster));

      //hosts
      var masterHosts = this.get('content.masterComponentHosts').mapProperty('hostName').uniq();
      var slaveHosts = this.get('content.slaveComponentHosts');

      var hostObj = [];
      slaveHosts.forEach(function (_hosts) {
        hostObj = hostObj.concat(_hosts.hosts);
      }, this);

      slaveHosts = hostObj.mapProperty('hostName').uniq();

      var componentHosts = masterHosts.concat(slaveHosts).uniq();
      var totalHosts = App.Host.find().mapProperty('hostName').concat(componentHosts).uniq();
      var newHostsCount = totalHosts.length - App.Host.find().content.length;
      this.set('totalHosts', totalHosts);
      var totalHostsObj = this.rawContent.findProperty('config_name', 'hosts');
      totalHostsObj.config_value = totalHosts.length + ' (' + newHostsCount + ' new)';
      this.get('clusterInfo').pushObject(Ember.Object.create(totalHostsObj));

      //repo
      if (['addHostController', 'addServiceController'].contains(this.get('content.controllerName'))) {
        this.loadRepoInfo();
      } else { // from install wizard
        var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
        var allRepos = [];
        var supportedOs = ['redhat5', 'redhat6', 'sles11'];
        if (selectedStack && selectedStack.operatingSystems) {
          selectedStack.operatingSystems.forEach(function (os) {
            if (os.selected && supportedOs.contains(os.osType)) {
              allRepos.push(Em.Object.create({
                base_url: os.baseUrl,
                os_type: Em.I18n.t("installer.step8.repoInfo.osType." + os.osType)
              }));
            }
          }, this);
        }
        allRepos.set('display_name', Em.I18n.t("installer.step8.repoInfo.displayName"));
        this.get('clusterInfo').set('repoInfo', allRepos);
      }
    },

    /**
     * get the repositories info of HDP from server. Used only in addHost controller.
     */
    loadRepoInfo: function(){
      var nameVersionCombo = App.get('currentStackVersion');
      var stackName = nameVersionCombo.split('-')[0];
      var stackVersion = nameVersionCombo.split('-')[1];
      App.ajax.send({
        name: 'cluster.load_repositories',
        sender: this,
        data: {
          stackName: stackName,
          stackVersion: stackVersion
        },
        success: 'loadRepoInfoSuccessCallback',
        error: 'loadRepositoriesErrorCallback'
      });
    },

    loadRepoInfoSuccessCallback: function (data) {
      var allRepos = [];
      var supportedOs = ['redhat5', 'redhat6', 'sles11'];
      data.items.forEach(function (item) {
        var os = item.repositories[0].Repositories;
        if (supportedOs.contains(os.os_type)) {
          allRepos.push(Em.Object.create({
            base_url: os.base_url,
            os_type: Em.I18n.t("installer.step8.repoInfo.osType." + os.os_type)
          }));
        }
      }, this);
      allRepos.set('display_name', Em.I18n.t("installer.step8.repoInfo.displayName"));
      this.get('clusterInfo').set('repoInfo', allRepos);
    },

    loadRepoInfoErrorCallback: function(request) {
      console.log('Error message is: ' + request.responseText);
      var allRepos = [];
      allRepos.set('display_name', Em.I18n.t("installer.step8.repoInfo.displayName"));
      this.get('clusterInfo').set('repoInfo', allRepos);
    },

    /**
     * Load all info about services to <code>services</code> variable
     */
    loadServices: function () {
      var reviewService = this.rawContent.findProperty('config_name', 'services');

      this.get('selectedServices').forEach(function (_service) {
        console.log('INFO: step8: Name of the service from getService function: ' + _service.serviceName);
        var serviceObj = reviewService.config_value.findProperty('service_name', _service.serviceName);
        if (serviceObj) {
          serviceObj.get('service_components').forEach(function (_component) {
            this.assignComponentHosts(_component);
          }, this);
          this.get('services').pushObject(serviceObj);
        }
      }, this);
    },

    /**
     *
     * @param {Em.Object} component
     */
    assignComponentHosts: function (component) {
      var componentValue;
      if (component.get('customHandler')) {
        this[component.get('customHandler')].call(this, component);
      } else {
        if (component.get('isMaster')) {
          componentValue = this.get('content.masterComponentHosts')
            .findProperty('component', component.component_name).hostName;
        } else {
          var hostsLength = this.get('content.slaveComponentHosts')
            .findProperty('componentName', component.component_name)
            .hosts.length;
          componentValue = hostsLength + Em.I18n.t('installer.step8.host' + ((hostsLength > 1) ? 's' : ''));
        }
        component.set('component_value', componentValue);
      }
    },

    /**
     * Set dispalyed Hive DB value based on DB type
     * @param dbComponent
     */
    loadHiveDbValue: function (dbComponent) {
      var hiveDb = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'hive_database');
      if (hiveDb.value === 'New MySQL Database') {
        dbComponent.set('component_value', 'MySQL (New Database)');
      } else if(hiveDb.value === 'Existing MySQL Database'){
        var db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'hive_existing_mysql_database');
        dbComponent.set('component_value', db.value + ' (' + hiveDb.value + ')');
      } else { // existing oracle database
        var db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'hive_existing_oracle_database');
        dbComponent.set('component_value', db.value + ' (' + hiveDb.value + ')');
      }
    },

    /**
     * Set displayed HBase master value
     * @param {Object} hbaseMaster
     */
    loadHbaseMasterValue: function (hbaseMaster) {
      var hbaseHostName = this.get('content.masterComponentHosts').filterProperty('component', hbaseMaster.component_name);
      if (hbaseHostName.length == 1) {
        hbaseMaster.set('component_value', hbaseHostName[0].hostName);
      } else {
        hbaseMaster.set('component_value', hbaseHostName[0].hostName + Em.I18n.t('installer.step8.other').format(hbaseHostName.length - 1));
      }
    },

    /**
     * Set displayed ZooKeeper Server value
     * @param {Object} serverComponent
     */
    loadZkServerValue: function (serverComponent) {
      var zkHostNames = this.get('content.masterComponentHosts').filterProperty('component', serverComponent.component_name).length;
      var hostSuffix;
      if (zkHostNames === 1) {
        hostSuffix = Em.I18n.t('installer.step8.host');
      } else {
        hostSuffix = Em.I18n.t('installer.step8.hosts');
      }
      serverComponent.set('component_value', zkHostNames + ' ' + hostSuffix);
    },

    /**
     * Set displayed Oozie DB value based on DB type
     * @param {Object} dbComponent
     */
    loadOozieDbValue: function (dbComponent) {
      var oozieDb = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_database');
      if (oozieDb.value === 'New Derby Database'){
        var db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_derby_database');
        dbComponent.set('component_value', db.value + ' (' + oozieDb.value + ')');
      }/* else if (oozieDb.value === 'New MySQL Database') {
        dbComponent.set('component_value', 'MySQL (New Database)');
      } */else if(oozieDb.value === 'Existing MySQL Database'){
        var db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_mysql_database');
        dbComponent.set('component_value', db.value + ' (' + oozieDb.value + ')');
      } else { // existing oracle database
        var db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_oracle_database');
        dbComponent.set('component_value', db.value + ' (' + oozieDb.value + ')');
      }

    },

    /**
     * Set displayed Nagion Admin value
     * @param {Object} nagiosAdmin
     */
    loadNagiosAdminValue: function (nagiosAdmin) {
      var config = this.get('content.serviceConfigProperties');
      var adminLoginName = config.findProperty('name', 'nagios_web_login');
      var adminEmail = config.findProperty('name', 'nagios_contact');
      nagiosAdmin.set('component_value', adminLoginName.value + ' / (' + adminEmail.value + ')');
    },
    /**
     * Onclick handler for <code>next</code> button
     */
    submit: function () {
      if (this.get('isSubmitDisabled')) return;
      if ((this.get('content.controllerName') == 'addHostController') && this.get('securityEnabled')) {
        var self = this;
        App.showConfirmationPopup(function() {
          self.submitProceed();
        }, Em.I18n.t('installer.step8.securityConfirmationPopupBody'));
      }
      else {
        this.submitProceed();
      }
    },
    /**
     * Update configurations for installed services.
     *
     * @param {Array} configsToUpdate - configs need to update
     * @return {*}
     */
    updateConfigurations: function (configsToUpdate) {
      var configurationController = App.router.get('mainServiceInfoConfigsController');
      var serviceNames = configsToUpdate.mapProperty('serviceName').uniq();
      serviceNames.forEach(function(serviceName) {
        var configs = configsToUpdate.filterProperty('serviceName', serviceName);
        configurationController.setNewTagNames(configs);
        var tagName = configs.objectAt(0).newTagName;
        var siteConfigs = configs.filterProperty('id', 'site property');
        siteConfigs.mapProperty('filename').uniq().forEach(function(siteName) {
          var formattedConfigs = configurationController.createSiteObj(siteName.replace(".xml", ""), tagName, configs.filterProperty('filename', siteName));
          configurationController.doPUTClusterConfigurationSite(formattedConfigs);
        });
      });
    },

    /**
     * Prepare <code>ajaxQueue</code> and start to execute it
     */
    submitProceed: function() {
      this.set('isSubmitDisabled', true);
      this.set('isBackBtnDisabled', true);
      if (this.get('content.controllerName') == 'addHostController') {
        App.router.get('addHostController').setLowerStepsDisable(4);
      }

      // checkpoint the cluster status on the server so that the user can resume from where they left off
      switch (this.get('content.controllerName')) {
        case 'installerController':
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'CLUSTER_DEPLOY_PREP_2',
            wizardControllerName: this.get('content.controllerName'),
            localdb: App.db.data
          });
          break;
        case 'addHostController':
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'ADD_HOSTS_DEPLOY_PREP_2',
            wizardControllerName: this.get('content.controllerName'),
            localdb: App.db.data
          });
          break;
        case 'addServiceController':
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'ADD_SERVICES_DEPLOY_PREP_2',
            wizardControllerName: this.get('content.controllerName'),
            localdb: App.db.data
          });
          break;
        default:
          break;
      }
      // delete any existing clusters to start from a clean slate
      // before creating a new cluster in install wizard
      // TODO: modify for multi-cluster support
      if (this.get('content.controllerName') == 'installerController' && (!App.testMode)) {
        this.deleteClusters(this.getExistingClusterNames());
      }
      if (this.get('wizardController').getDBProperty('configsToUpdate')) {
        this.updateConfigurations(this.get('wizardController').getDBProperty('configsToUpdate'));
      }
      this.setLocalRepositories();
      this.createCluster();
      this.createSelectedServices();
      if (this.get('content.controllerName') !== 'addHostController') {
        this.createConfigurations();
        this.applyConfigurationsToCluster();
      }
      this.createComponents();
      this.registerHostsToCluster();
      if (App.supports.hostOverridesInstaller) {
        this.createConfigurationGroups();
      }
      this.createMasterHostComponents();
      this.createSlaveAndClientsHostComponents();
      this.createAdditionalHostComponents();

      this.doNextAjaxCall();
    },

    /**
     * Used in progress bar
     */
    ajaxQueueLength: function () {
      return this.get('ajaxQueue').length;
    }.property('ajaxQueue.length'),

    /**
     * Used in progress bar
     */
    ajaxQueueLeft: 0,

    clusterName: function () {
      return this.get('content.cluster.name');
    }.property('content.cluster.name'),

    clusterNames: [],

    /**
     * Get list of existing cluster names
     * @returns {string[]}
     * returns an array of existing cluster names.
     * returns an empty array if there are no existing clusters.
     */
    getExistingClusterNames: function () {
      App.ajax.send({
        name: 'wizard.step8.existing_cluster_names',
        sender: this,
        success: 'getExistingClusterNamesSuccessCallBack',
        error: 'getExistingClusterNamesErrorCallback'
      });

      return this.get('clusterNames');
    },

    /**
     * Save received list to <code>clusterNames</code>
     * @param {Object} data
     */
    getExistingClusterNamesSuccessCallBack: function (data) {
      var clusterNames = data.items.mapProperty('Clusters.cluster_name');
      console.log("Got existing cluster names: " + clusterNames);
      this.set('clusterNames', clusterNames);
    },

    /**
     * If error appears, set <code>clusterNames</code> to <code>[]</code>
     */
    getExistingClusterNamesErrorCallback: function () {
      console.log("Failed to get existing cluster names");
      this.set('clusterNames', []);
    },

    /**
     * Delete cluster by name
     * One request for one cluster!
     * @param {string[]} clusterNames
     */
    deleteClusters: function (clusterNames) {
      clusterNames.forEach(function (clusterName) {
        App.ajax.send({
          name: 'wizard.step8.delete_cluster',
          sender: this,
          data: {
            name: clusterName
          }
        });
      }, this);
    },

    /**
     * Updates local repositories for the Ambari server.
     * Base URL validation is disabled, since the checks had already run in the Select Stacks page
     * already (unless the user specifically asked to skip).
     * TODO: This function should be deleted once we modify App.InstallerController to make
     * sure that base URLs are saved even when "Skip validation" is checked.
     */
    setLocalRepositories: function () {
      if (this.get('content.controllerName') !== 'installerController' || !App.supports.localRepositories) return;
      var self = this;

      var stack = this.get('content.stacks').findProperty('isSelected', true);
      stack.operatingSystems.forEach(function (os) {
        if (os.baseUrl !== os.originalBaseUrl) {
          console.log("Updating local repository URL from " + os.originalBaseUrl + " -> " + os.baseUrl + ". ", os);
          self.addRequestToAjaxQueue({
            type: 'PUT',
            url: App.apiPrefix + App.get('stack2VersionURL') + "/operatingSystems/" + os.osType + "/repositories/" + stack.name,
            data: JSON.stringify({
              "Repositories": {
                "base_url": os.baseUrl,
                "verify_base_url": false
              }
            })
          });
        }
      });
    },


    /**
     * *******************************************************************
     * The following create* functions are called upon submitting Step 8.
     * *******************************************************************
     */

    /**
     * Create cluster using selected stack version
     * Queued request
     */
    createCluster: function () {
      if (this.get('content.controllerName') !== 'installerController') return;
      var stackVersion = (this.get('content.installOptions.localRepo')) ? App.currentStackVersion.replace(/(-\d+(\.\d)*)/ig, "Local$&") : App.currentStackVersion;
      this.addRequestToAjaxQueue({
        type: 'POST',
        url: App.apiPrefix + '/clusters/' + this.get('clusterName'),
        data: JSON.stringify({ "Clusters": {"version": stackVersion }})
      });

    },

    /**
     * Create selected to install services
     * Queued request
     * Skipped if no services where selected!
     */
    createSelectedServices: function () {
      var data = this.createSelectedServicesData();
      if (!data.length) return;
      this.addRequestToAjaxQueue({
        type: 'POST',
        url: App.apiPrefix + '/clusters/' + this.get('clusterName') + '/services',
        data: JSON.stringify(data)
      });
    },

    /**
     * Format data for <code>createSelectedServices</code> request
     * @returns {Object[]}
     */
    createSelectedServicesData: function () {
      return this.get('selectedServices').map(function (_service) {
        return {"ServiceInfo": { "service_name": _service.get('serviceName') }};
      });
    },

    /**
     * Create components for selected services
     * Queued requests
     * One request for each service!
     */
    createComponents: function () {
      var serviceComponents = require('data/service_components');
      this.get('selectedServices').forEach(function (_service) {
        var serviceName = _service.get('serviceName');
        var componentsData = serviceComponents.filterProperty('service_name', serviceName).map(function (_component) {
          return { "ServiceComponentInfo": { "component_name": _component.component_name } };
        });

        // Service must be specified in terms of a query for creating multiple components at the same time.
        // See AMBARI-1018.
        this.addRequestToAjaxQueue({
          type: 'POST',
          url: App.apiPrefix + '/clusters/' + this.get('clusterName') + '/services?ServiceInfo/service_name=' + serviceName,
          data: JSON.stringify({"components": componentsData})
        });
      }, this);

      if (this.get('content.controllerName') == 'addServiceController' && !App.testMode) {
        // Add service-components which show up in newer versions but did not
        // exist in older ones.
        var self = this;
        var newServiceComponents = {}
        if (App.get('isHadoop21Stack') && App.Service.find().findProperty('serviceName','YARN')) {
          newServiceComponents['APP_TIMELINE_SERVER'] = 'YARN';
        }
        for (var componentName in newServiceComponents) {
          var serviceName = newServiceComponents[componentName];
          // Create only if it doesnt exist
          var serviceComponentChecker = {
            success: function(){},
            error: function(){
              var componentsData = [{
                "ServiceComponentInfo": {
                  "component_name": componentName
                }
              }];
              self.addRequestToAjaxQueue({
                type: 'POST',
                url: App.apiPrefix + '/clusters/' + self.get('clusterName') + '/services?ServiceInfo/service_name=' + serviceName,
                data: JSON.stringify({"components": componentsData})
              });
            }
          }
          App.ajax.send({
            name: 'service.service_component',
            sender: serviceComponentChecker,
            data: {
              serviceName: serviceName,
              componentName: componentName,
              async: false
            },
            success: 'success',
            error: 'error'
          });
        }
      }
    },

    /**
     * Register hosts
     * Queued request
     */
    registerHostsToCluster: function () {
      var data = this.createRegisterHostData();
      if (!data.length) return;
      this.addRequestToAjaxQueue({
        type: 'POST',
        url: App.apiPrefix + '/clusters/' + this.get('clusterName') + '/hosts',
        data: JSON.stringify(data)
      });
    },

    /**
     * Format request-data for <code>registerHostsToCluster</code>
     * @returns {Object}
     */
    createRegisterHostData: function () {
      return this.getRegisteredHosts().filterProperty('isInstalled', false).map(function (host) {
        return {"Hosts": { "host_name": host.hostName}};
      });
    },

    /**
     * Register master components
     * @uses registerHostsToComponent
     */
    createMasterHostComponents: function() {
      var masterHosts = this.get('content.masterComponentHosts');
      masterHosts.mapProperty('component').uniq().forEach(function (component) {
        var hostNames = masterHosts.filterProperty('component', component).filterProperty('isInstalled', false).mapProperty('hostName');
        this.registerHostsToComponent(hostNames, component);
      }, this);
    },

    /**
     * Register slave components and clients
     * @uses registerHostsToComponent
     */
    createSlaveAndClientsHostComponents: function() {
      var masterHosts = this.get('content.masterComponentHosts');
      var slaveHosts = this.get('content.slaveComponentHosts');
      var clients = this.get('content.clients');

      /**
       * Determines on which hosts client should be installed (based on availability of master components on hosts)
       * @type {Object}
       * Format:
       * <code>
       *  {
       *    CLIENT1: Em.A([MASTER1, MASTER2, ...]),
       *    CLIENT2: Em.A([MASTER3, MASTER1, ...])
       *    ...
       *  }
       * </code>
       */
      var clientsToMasterMap = {
        HDFS_CLIENT: Em.A(['HBASE_MASTER', 'HBASE_REGIONSERVER', 'WEBHCAT_SERVER', 'HISTORYSERVER', 'OOZIE_SERVER']),
        MAPREDUCE_CLIENT: Em.A(['HIVE_SERVER', 'OOZIE_SERVER', 'NAGIOS_SERVER', 'WEBHCAT_SERVER']),
        OOZIE_CLIENT: Em.A(['NAGIOS_SERVER']),
        ZOOKEEPER_CLIENT: Em.A(['WEBHCAT_SERVER']),
        HIVE_CLIENT: Em.A(['WEBHCAT_SERVER','HIVE_SERVER']),
        HCAT: Em.A(['NAGIOS_SERVER']),
        YARN_CLIENT: Em.A(['NAGIOS_SERVER','HIVE_SERVER','OOZIE_SERVER','WEBHCAT_SERVER']),
        TEZ_CLIENT: Em.A(['NAGIOS_SERVER','HIVE_SERVER'])
      };

      slaveHosts.forEach(function (_slave) {
        if (_slave.componentName !== 'CLIENT') {
          var hostNames = _slave.hosts.filterProperty('isInstalled', false).mapProperty('hostName');
          this.registerHostsToComponent(hostNames, _slave.componentName);
        }
        else {
          clients.forEach(function (_client) {

            var hostNames = _slave.hosts.mapProperty('hostName');
            if (clientsToMasterMap[_client.component_name]) {
              clientsToMasterMap[_client.component_name].forEach(function(componentName) {
                masterHosts.filterProperty('component', componentName).filterProperty('isInstalled', false).forEach(function (_masterHost) {
                  hostNames.pushObject(_masterHost.hostName);
                });
              });
            }
            hostNames = hostNames.uniq();

            if (_client.isInstalled) {
              //check whether clients are already installed on selected master hosts!!!
              _slave.hosts.filterProperty('isInstalled', true).mapProperty('hostName').forEach(function (host) {
                if (hostNames.contains(host)) {
                  hostNames.splice(hostNames.indexOf(host), 1);
                }
              }, this);
            }

            this.registerHostsToComponent(hostNames, _client.component_name);

          }, this);
        }
      }, this);
    },

    /**
     * Register additional components
     * Based on availability of some services
     * @uses registerHostsToComponent
     */
    createAdditionalHostComponents: function() {
      var masterHosts = this.get('content.masterComponentHosts');

      // add Ganglia Monitor (Slave) to all hosts if Ganglia service is selected
      var gangliaService = this.get('content.services').filterProperty('isSelected', true).findProperty('serviceName', 'GANGLIA');
      if (gangliaService) {
        var hosts = this.getRegisteredHosts();
        if (gangliaService.get('isInstalled')) {
          hosts = hosts.filterProperty('isInstalled', false);
        }
        if (hosts.length) {
          this.registerHostsToComponent(hosts.mapProperty('hostName'), 'GANGLIA_MONITOR');
        }
      }
      // add MySQL Server if Hive is selected
      var hiveService = this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false).findProperty('serviceName', 'HIVE');
      if (hiveService) {
        var hiveDb = this.get('content.serviceConfigProperties').findProperty('name', 'hive_database');
        if(hiveDb.value == "New MySQL Database") {
          this.registerHostsToComponent(masterHosts.filterProperty('component', 'HIVE_SERVER').mapProperty('hostName'), 'MYSQL_SERVER');
        }
      }
    },

    /**
     * Register component to hosts
     * Queued request
     * @param {String[]} hostNames
     * @param {String} componentName
     */
    registerHostsToComponent: function (hostNames, componentName) {
      if (!hostNames.length) return;

      var queryStr = '';
      hostNames.forEach(function (hostName) {
        queryStr += 'Hosts/host_name=' + hostName + '|';
      });
      //slice off last symbol '|'
      queryStr = queryStr.slice(0, -1);

      var data = {
        "RequestInfo": {
          "query": queryStr
        },
        "Body": {
          "host_components": [
            {
              "HostRoles": {
                "component_name": componentName
              }
            }
          ]
        }
      };

      this.addRequestToAjaxQueue({
        type: 'POST',
        url: App.apiPrefix + '/clusters/' + this.get('clusterName') + '/hosts',
        data: JSON.stringify(data)
      });
    },

    /**
     * Create config objects for cluster and services
     */
    createConfigurations: function () {
      var self = this;
      var selectedServices = this.get('selectedServices');
      var coreSiteObject = this.createCoreSiteObj();
      if (this.get('content.controllerName') == 'installerController') {
        this.get('serviceConfigTags').pushObject(coreSiteObject);
        this.get('serviceConfigTags').pushObject(this.createSiteObj('hdfs-site',false));
        this.get('serviceConfigTags').pushObject(this.createLog4jObj('hdfs'));
      }
      var globalSiteObj = this.createGlobalSiteObj();
      if (this.get('content.controllerName') == 'addServiceController') {
        globalSiteObj.tag = 'version' + (new Date).getTime();
        coreSiteObject.tag = 'version' + (new Date).getTime();
        this.get('serviceConfigTags').pushObject(coreSiteObject);
      }
      this.get('serviceConfigTags').pushObject(globalSiteObj);

      var objMap = {
        MAPREDUCE: {site: [{filename: 'mapred-site',isXmlFile: true}], log4j: ['mapreduce']},
        MAPREDUCE2: {site: [{filename:'mapred-site',isXmlFile: true}], log4j: ['mapreduce2']},
        YARN: {site: [{filename:'yarn-site',isXmlFile: true},{filename:'capacity-scheduler',isXmlFile: true}], log4j: ['yarn']},
        HBASE: {site: [{filename:'hbase-site',isXmlFile: true}], log4j: ['hbase']},
        OOZIE: {site: [{filename:'oozie-site',isXmlFile: true}], log4j: ['oozie']},
        HIVE: {site: [{filename:'hive-site',isXmlFile: true}], log4j: ['hive','hive-exec']},
        WEBHCAT: {site: [{filename:'webhcat-site',isXmlFile: true}], log4j: []},
        HUE: {site: [{filename:'hue-site',isXmlFile: true}], log4j: []},
        PIG: {site: [], log4j: ['pig']},
        FALCON: {site: [{filename:'falcon-startup.properties',isXmlFile: false},{filename:'falcon-runtime.properties',isXmlFile: false}], log4j: []},
        TEZ: {site: [{filename:'tez-site',isXmlFile: true}], log4j: []},
        ZOOKEEPER: {site: [], log4j: ['zookeeper']}
      };

      if (App.supports.capacitySchedulerUi) {
        objMap['MAPREDUCE'].site.pushObjects([{filename:'capacity-scheduler',isXmlFile: true},{filename:'mapred-queue-acls',isXmlFile: true}]);
      }

      for(var serviceName in objMap) {
        if (objMap.hasOwnProperty(serviceName)) {
          if (selectedServices.someProperty('serviceName', serviceName)) {
            objMap[serviceName].site.forEach(function(site) {
              self.get('serviceConfigTags').pushObject(self.createSiteObj(site.filename,!site.isXmlFile));
            });
            objMap[serviceName].log4j.forEach(function(log4j) {
              self.get('serviceConfigTags').pushObject(self.createLog4jObj(log4j));
            });
          }
        }
      }

      if (selectedServices.someProperty('serviceName', 'STORM')) {
        this.get('serviceConfigTags').pushObject(this.createStormSiteObj());
      }
      if (selectedServices.someProperty('serviceName', 'ZOOKEEPER')) {
        this.get('serviceConfigTags').pushObject(this.createZooCfgObj());
      }
    },

    /**
     * Send <code>serviceConfigTags</code> to server
     * Queued request
     * One request for each service config tag
     */
    applyConfigurationsToCluster: function() {
      var configData = this.get('serviceConfigTags').map(function (_serviceConfig) {
        return JSON.stringify({
          Clusters: {
            desired_config: {
              type: _serviceConfig.type,
              tag: _serviceConfig.tag,
              properties: _serviceConfig.properties
            }
          }
        });
      }, this).toString();

      this.addRequestToAjaxQueue({
        type: 'PUT',
        url: App.apiPrefix + '/clusters/' + this.get('clusterName'),
        data: '[' + configData + ']'
      });
    },

    /**
     * Create and update config groups
     */
    createConfigurationGroups: function () {
      var configGroups = this.get('content.configGroups').filterProperty('isDefault', false);
      var clusterName = this.get('clusterName');
      var sendData = [];
      var updateData = [];
      var serviceConfigController = App.router.get('mainServiceInfoConfigsController');
      var timeTag = (new Date).getTime();
      var groupsToDelete = App.router.get(this.get('content.controllerName')).getDBProperty('groupsToDelete');
      if (groupsToDelete && groupsToDelete.length > 0) {
        this.removeInstalledServicesConfigurationGroups(groupsToDelete);
      }
      configGroups.forEach(function (configGroup) {
        var groupConfigs = [];
        var groupData = {
          "cluster_name": clusterName,
          "group_name": configGroup.name,
          "tag": configGroup.service.id,
          "description": configGroup.description,
          "hosts": [],
          "desired_configs": []
        };
        configGroup.hosts.forEach(function (hostName) {
          groupData.hosts.push({"host_name": hostName});
        });
        //wrap properties into Em.Object to make them compatible with buildGroupDesiredConfigs method
        configGroup.properties.forEach(function (property) {
          groupConfigs.push(Em.Object.create(property));
        });
        groupData.desired_configs = serviceConfigController.buildGroupDesiredConfigs.call(serviceConfigController, groupConfigs, timeTag);
        // check for group from installed service
        if (configGroup.isForUpdate === true) {
          // if group is a new one, create it
          if (!configGroup.id) {
            sendData.push({"ConfigGroup": groupData});
          } else {
            // update an existing group
            groupData.id = configGroup.id;
            updateData.push({"ConfigGroup": groupData});
          }
        } else {
          sendData.push({"ConfigGroup": groupData});
        }
        //each group should have unique tag to prevent overriding configs from common sites
        timeTag++;
      }, this);
      if (sendData.length > 0) {
        this.applyConfigurationGroups(sendData);
      }
      if (updateData.length > 0) {
        this.applyInstalledServicesConfigurationGroup(updateData);
      }
    },

    /**
     * Create new config groups request
     * Queued request
     * @param {Object[]} sendData
     */
    applyConfigurationGroups: function (sendData) {
      this.addRequestToAjaxQueue({
        type: 'POST',
        url: App.apiPrefix + '/clusters/' + this.get('clusterName') + '/config_groups',
        data: JSON.stringify(sendData)
      });
    },

    /**
     * Update existed config groups
     * @param {Object[]} updateData
     */
    applyInstalledServicesConfigurationGroup: function (updateData) {
      updateData.forEach(function(item) {
        App.router.get('mainServiceInfoConfigsController').putConfigGroupChanges(item);
      });
    },

    /**
     * Delete selected config groups
     * @param {Object[]} groupsToDelete
     */
    removeInstalledServicesConfigurationGroups: function(groupsToDelete) {
      groupsToDelete.forEach(function(item) {
        App.config.deleteConfigGroup(Em.Object.create(item));
      });
    },

    /**
     * Create Global Site object
     * @returns {{type: string, tag: string, properties: {}}}
     */
    createGlobalSiteObj: function () {
      var globalSiteProperties = {};
      var globalSiteObj = this.get('globals');
      var isGLUSTERFSSelected = this.get('selectedServices').someProperty('serviceName', 'GLUSTERFS');
      
      // screen out the GLUSTERFS-specific global config entries when they are not required
      if (!isGLUSTERFSSelected) {
        globalSiteObj = globalSiteObj.filter(function(_config) {
          return _config.name.indexOf("fs_glusterfs") < 0;
        });
      }
      
      globalSiteObj.forEach(function (_globalSiteObj) {
        var heapsizeException =  ['hadoop_heapsize','yarn_heapsize','nodemanager_heapsize','resourcemanager_heapsize', 'apptimelineserver_heapsize'];
        // do not pass any globals whose name ends with _host or _hosts
        if (_globalSiteObj.isRequiredByAgent !== false) {
          // append "m" to JVM memory options except for hadoop_heapsize
          if (/_heapsize|_newsize|_maxnewsize$/.test(_globalSiteObj.name) && !heapsizeException.contains(_globalSiteObj.name)) {
            globalSiteProperties[_globalSiteObj.name] = _globalSiteObj.value + "m";
          } else {
            globalSiteProperties[_globalSiteObj.name] = App.config.escapeXMLCharacters(_globalSiteObj.value);
          }
        }
      }, this);
      // we don't expose gmond_user to the user; it needs to be the same as gmetad_user
      globalSiteProperties['gmond_user'] = globalSiteProperties['gmetad_user'];
      return {"type": "global", "tag": "version1", "properties": globalSiteProperties};
    },

    /**
     * Create Core Site object
     * @returns {{type: string, tag: string, properties: {}}}
     */
    createCoreSiteObj: function () {
      var coreSiteObj = this.get('configs').filterProperty('filename', 'core-site.xml');
      var coreSiteProperties = {};
      // hadoop.proxyuser.oozie.hosts needs to be skipped if oozie is not selected
      var isOozieSelected = this.get('selectedServices').someProperty('serviceName', 'OOZIE');
      var oozieUser = this.get('globals').someProperty('name', 'oozie_user') ? this.get('globals').findProperty('name', 'oozie_user').value : null;
      var isHiveSelected = this.get('selectedServices').someProperty('serviceName', 'HIVE');
      var hiveUser = this.get('globals').someProperty('name', 'hive_user') ? this.get('globals').findProperty('name', 'hive_user').value : null;
      var isHcatSelected = this.get('selectedServices').someProperty('serviceName', 'WEBHCAT');
      var hcatUser = this.get('globals').someProperty('name', 'hcat_user') ? this.get('globals').findProperty('name', 'hcat_user').value : null;
      var isGLUSTERFSSelected = this.get('selectedServices').someProperty('serviceName', 'GLUSTERFS');

      // screen out the GLUSTERFS-specific core-site.xml entries when they are not needed
      if (!isGLUSTERFSSelected) {
        coreSiteObj = coreSiteObj.filter(function(_config) {
          return _config.name.indexOf("fs.glusterfs") < 0;
        });
      }

      coreSiteObj.forEach(function (_coreSiteObj) {
        if ((isOozieSelected || (_coreSiteObj.name != 'hadoop.proxyuser.' + oozieUser + '.hosts' && _coreSiteObj.name != 'hadoop.proxyuser.' + oozieUser + '.groups')) && (isHiveSelected || (_coreSiteObj.name != 'hadoop.proxyuser.' + hiveUser + '.hosts' && _coreSiteObj.name != 'hadoop.proxyuser.' + hiveUser + '.groups')) && (isHcatSelected || (_coreSiteObj.name != 'hadoop.proxyuser.' + hcatUser + '.hosts' && _coreSiteObj.name != 'hadoop.proxyuser.' + hcatUser + '.groups'))) {
          coreSiteProperties[_coreSiteObj.name] = App.config.escapeXMLCharacters(_coreSiteObj.value);
        }
        if (isGLUSTERFSSelected && _coreSiteObj.name == "fs.default.name") {
          coreSiteProperties[_coreSiteObj.name] = this.get('globals').someProperty('name', 'fs_glusterfs_default_name') ? App.config.escapeXMLCharacters(this.get('globals').findProperty('name', 'fs_glusterfs_default_name').value) : null;
        }
        if (isGLUSTERFSSelected && _coreSiteObj.name == "fs.defaultFS") {
          coreSiteProperties[_coreSiteObj.name] = this.get('globals').someProperty('name', 'glusterfs_defaultFS_name') ? App.config.escapeXMLCharacters(this.get('globals').findProperty('name', 'glusterfs_defaultFS_name').value) : null;
        }
      }, this);
      return {"type": "core-site", "tag": "version1", "properties": coreSiteProperties};
    },

    /**
     * Create siteObj for custom service with it own configs
     * @param {string} site
     * @param {Boolean} isNonXmlFile
     * @returns {{type: string, tag: string, properties: {}}}
     */
    createSiteObj: function(site,isNonXmlFile) {
      var properties = {};
      if (!!isNonXmlFile) {
        this.get('configs').filterProperty('filename', site + '.xml').forEach(function (_configProperty) {
          properties[_configProperty.name] = _configProperty.value;
        },this);
      } else {
        this.get('configs').filterProperty('filename', site + '.xml').forEach(function (_configProperty) {
          properties[_configProperty.name] = App.config.escapeXMLCharacters(_configProperty.value);
        },this);
      }
      return {"type": site, "tag": "version1", "properties": properties };
    },

    /**
     * Create log4j object for custom service with it own configs
     * @param {string} site
     * @returns {{type: string, tag: string, properties: {}}}
     */
    createLog4jObj: function (site) {
      return this.createSiteObj(site + '-log4j',true);
    },

    /**
     * Create ZooKeeper Cfg Object
     * @returns {{type: string, tag: string, properties: {}}}
     */
    createZooCfgObj: function () {
      var configs = this.get('configs').filterProperty('filename', 'zoo.cfg');
      var csProperties = {};
      configs.forEach(function (_configProperty) {
        csProperties[_configProperty.name] = App.config.escapeXMLCharacters(_configProperty.value);
      }, this);
      return {type: 'zoo.cfg', tag: 'version1', properties: csProperties};
    },

    /**
     * Create site obj for Storm
     * Some config-properties should be modified in custom way
     * @returns {{type: string, tag: string, properties: {}}}
     */
    createStormSiteObj: function () {
      var configs = this.get('configs').filterProperty('filename', 'storm-site.xml');
      var stormProperties = {};
      var specialProperties = ["storm.zookeeper.servers", "nimbus.childopts", "supervisor.childopts", "worker.childopts"];
      configs.forEach(function (_configProperty) {
        if (specialProperties.contains(_configProperty.name)) {
          if (_configProperty.name == "storm.zookeeper.servers") {
            stormProperties[_configProperty.name] = JSON.stringify(_configProperty.value).replace(/"/g, "'");
          } else {
            stormProperties[_configProperty.name] = JSON.stringify(_configProperty.value).replace(/"/g,"");
          }
        } else {
          stormProperties[_configProperty.name] = App.config.escapeXMLCharacters(_configProperty.value);
        }
      }, this);
      return {type: 'storm-site', tag: 'version1', properties: stormProperties};
    },

    /**
     * Navigate to next step after all requests are sended
     */
    ajaxQueueFinished: function () {
      console.log('everything is loaded');
      App.router.send('next');
    },

    /**
     * Do ajax-call with data in <code>ajaxQueue[0]</code> and shift <code>ajaxQueue</code>
     */
    doNextAjaxCall: function () {
      if (this.get('ajaxBusy')) return;

      var queue = this.get('ajaxQueue');
      if (!queue.length) {
        this.ajaxQueueFinished();
        return;
      }

      var first = queue[0];
      this.set('ajaxQueue', queue.slice(1));
      this.set('ajaxQueueLeft', this.get('ajaxQueue').length);
      this.set('ajaxBusy', true);

      $.ajax(first);
    },

    /**
     * We need to do a lot of ajax calls async in special order. To do this,
     * generate array of ajax objects and then send requests step by step. All
     * ajax objects are stored in <code>ajaxQueue</code>
     *
     * Each ajax-request success callback contains call of <code>doNextAjaxCall</code>
     *
     * @param {Object} params object with ajax-request parameters like url, type, data etc
     */
    addRequestToAjaxQueue: function (params) {
      if (App.testMode) return;

      var self = this;
      params = jQuery.extend({
        async: true,
        dataType: 'text',
        statusCode: require('data/statusCodes'),
        timeout: App.timeout,
        error: function () {
          console.log('Step8: In Error ');
        },
        success: function () {
          console.log("TRACE: Step8 -> In success function");
        }
      }, params);

      var success = params.success;
      var error = params.error;

      params.success = function () {
        if (success) {
          success();
        }

        self.set('ajaxBusy', false);
        self.doNextAjaxCall();
      };

      params.error = function (xhr, status, error) {
        var responseText = JSON.parse(xhr.responseText);
        var controller = App.router.get(App.clusterStatus.wizardControllerName);
        controller.registerErrPopup(Em.I18n.t('common.error'), responseText.message);
        self.set('hasErrorOccurred', true);
        // an error will break the ajax call chain and allow submission again
        self.set('isSubmitDisabled', false);
        self.set('isBackBtnDisabled', false);
        App.router.get(self.get('content.controllerName')).setStepsEnable();
        self.get('ajaxQueue').clear();
        self.set('ajaxBusy', false);
      };

      this.get('ajaxQueue').pushObject(params);
    }

  });
  
});
window.require.register("controllers/wizard/step9_controller", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  var serviceComponents = require('data/service_components');

  App.WizardStep9Controller = Em.Controller.extend({
    name: 'wizardStep9Controller',

    /**
     *  Array of host Objects that are successfully registered on "Confirm Host Options" page
     *  <code>
     *    {
     *      name: {String} host name.
     *      status: {String} Current status of the host. This field is used in the step 9 view to set the css class of the
     *             host progress bar and set the appropriate message. Possible values: info, warning, failed, heartbeat_lost and success
     *      logTasks: {Array} Tasks that are scheduled on the host for the current request.
     *      message: {String} Message to be shown in front of the host name.
     *      progress: {Int} Progress of the tasks on the host. Amount of tasks completed for all the tasks scheduled on the host.
     *      isNoTasksForInstall: {Boolean} Gets set when no task is scheduled for the host on install phase.
     *    }
     *  </code>
     */

    hosts: [],

    /**
     * overall progress of <Install,Start and Test> page shown as progress bar on the top of the page
     */
    progress: '0',

    /*
     * json file for the mock data to be used in mock mode
     */
    mockDataPrefix: '/data/wizard/deploy/5_hosts',

    /*
     * Current Request data polled from the API: api/v1/clusters/{clusterName}/requests/{RequestId}?fields=tasks/Tasks/command,
     * tasks/Tasks/exit_code,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,
     * tasks/Tasks/status&minimal_response=true
     */
    polledData: [],

    /*
     * This flag is only used in UI mock mode as a counter for number of polls.
     */
    numPolls: 1,

    // Interval in milliseconds between API calls While polling the request status for Install and, Start and Test tasks
    POLL_INTERVAL: 4000,

    /**
     * Array of objects
     * <code>
     *   {
     *     hostName: {String} Name of host that has stopped heartbeating to ambari-server
     *     componentNames: {Sting} Name of all components that are on the host
     *   }
     * </code>
     */
    hostsWithHeartbeatLost: [],

     // Flag is set in the start all services error callback function
    startCallFailed: false,

    /*
     * Status of the page. Possible values: <info, warning, failed and success>.
     * This property is used in the step-9 view for displaying the appropriate color of the overall progress bar and
     * the appropriate result message at the bottom of the page
     */
    status: 'info',

    /**
     * This computed property is used to determine if the Next button and back button on the page should be disabled. The property is
     * used in the template to grey out Next and Back buttons. Although clicking on the greyed out button do trigger the event and
     * calls submit and back function of the controller.
     */
    isSubmitDisabled: function () {
      var validStates = ['STARTED', 'START FAILED'];
      var controllerName = this.get('content.controllerName');
      if (controllerName == 'addHostController' || controllerName == 'addServiceController') {
        validStates.push('INSTALL FAILED');
      }
      return !validStates.contains(this.get('content.cluster.status'));
    }.property('content.cluster.status'),

    /**
     * This function is called when a click event happens on Next button of <Install, Start and Test> page
     */
    submit: function () {
      App.router.send('next');
    },

    /**
     * This function is called when a click event happens on back button of <Install, Start and Test> page
     */
    back: function () {
      if (!this.get('isSubmitDisabled')) {
        App.router.send('back');
      }
    },

    /**
     * Observer function: Enables previous steps link if install task failed in installer wizard.
     */
    togglePreviousSteps: function () {
      if (App.testMode) {
        return;
      } else if ('INSTALL FAILED' === this.get('content.cluster.status') && this.get('content.controllerName') == 'installerController') {
        App.router.get('installerController').setStepsEnable();
      } else {
        App.router.get('installerController').setLowerStepsDisable(9);
      }
    }.observes('content.cluster.status', 'content.controllerName'),

    /*
     * Computed property to determine if the Retry button should be made visible on the page.
     */
    showRetry: function () {
      return this.get('content.cluster.status') == 'INSTALL FAILED';
    }.property('content.cluster.status'),

    /**
     * Observer function: Calls {hostStatusUpdates} function once with change in a host status from any registered hosts.
     */
    hostStatusObserver: function () {
      Ember.run.once(this, 'updateStatus');
    }.observes('hosts.@each.status'),


    /**
     * A flag that gets set with installation failure.
     */
    installFailed: false,

    /**
     * Observer function: Updates {status} field of the controller.
     */
    updateStatus: function () {
      var status = 'info';
      if (this.get('hosts').someProperty('status', 'failed')
        || this.get('hosts').someProperty('status', 'heartbeat_lost')
        || this.get('startCallFailed')) {
        status = 'failed';
      } else if (this.get('hosts').someProperty('status', 'warning')) {
        if (this.isStepFailed()) {
          status = 'failed';
        } else {
          status = 'warning';
        }
      } else if (this.get('progress') == '100' && this.get('content.cluster.status') !== 'INSTALL FAILED') {
        status = 'success';
      }
      this.set('status', status);
    }.observes('progress'),

    /**
     * Incremental flag that triggers an event in step 9 view to change the tasks related data and icons of hosts.
     */
    logTasksChangesCounter: 0,

    /**
     * navigateStep is called by App.WizardStep9View's didInsertElement and "retry" from router.
     * content.cluster.status can be:
     * PENDING: set upon successful transition from step 1 to step 2
     * INSTALLED: set upon successful completion of install phase as well as successful invocation of start services API
     * STARTED: set up on successful completion of start phase
     * INSTALL FAILED: set up upon encountering a failure in install phase
     * START FAILED: set upon unsuccessful invocation of start services API and also upon encountering a failure
     * during start phase

     * content.cluster.isCompleted
     * set to false upon successful transition from step 1 to step 2
     * set to true upon successful start of services in this step
     * note: looks like this is the same thing as checking content.cluster.status == 'STARTED'
     */
    navigateStep: function () {
      if (App.testMode) {
        // this is for repeatedly testing out installs in test mode
        this.set('content.cluster.status', 'PENDING');
        this.set('content.cluster.isCompleted', false);
        this.set('content.cluster.requestId', 1);
      }
      var clusterStatus = this.get('content.cluster.status');
      console.log('navigateStep: clusterStatus = ' + clusterStatus);
      if (this.get('content.cluster.isCompleted') === false) {
        // the cluster has not yet successfully installed and started
        if (clusterStatus === 'INSTALL FAILED') {
          this.loadStep();
          this.loadLogData(this.get('content.cluster.requestId'));
        } else if (clusterStatus === 'START FAILED') {
          this.loadStep();
          this.loadLogData(this.get('content.cluster.requestId'));
        } else {
          // handle PENDING, INSTALLED
          this.loadStep();
          this.loadLogData(this.get('content.cluster.requestId'));
          this.startPolling();
        }
      } else {
        // handle STARTED
        // the cluster has successfully installed and started
        this.loadStep();
        this.loadLogData(this.get('content.cluster.requestId'));
        this.set('progress', '100');
      }
    },

    /**
     * This is called on initial page load, refreshes and retry event.
     * clears all in memory stale data for retry event.
     */
    clearStep: function () {
      this.get('hosts').clear();
      this.set('hostsWithHeartbeatLost', []);
      this.set('startCallFailed',false);
      this.set('status', 'info');
      this.set('progress', '0');
      this.set('numPolls', 1);
    },

    /**
     * This is called on initial page load, refreshes and retry event.
     */
    loadStep: function () {
      console.log("TRACE: Loading step9: Install, Start and Test");
      this.clearStep();
      this.loadHosts();
    },


    /**
     * Reset status and message of all hosts when retry install
     */
    resetHostsForRetry: function () {
      var hosts = this.get('content.hosts');
      for (var name in hosts) {
        hosts[name].status = "pending";
        hosts[name].message = 'Waiting';
        hosts[name].isNoTasksForInstall = false;
      }
      this.set('content.hosts', hosts);
    },

    /**
     * Sets the {hosts} array for the controller
     */
    loadHosts: function () {
      var hosts = this.get('content.hosts');
      for (var index in hosts) {
        if (hosts[index].bootStatus === 'REGISTERED') {
          var hostInfo = App.HostInfo.create({
            name: hosts[index].name,
            status: (hosts[index].status) ? hosts[index].status : 'info',
            logTasks: [],
            message: (hosts[index].message) ? hosts[index].message : 'Waiting',
            progress: 0,
            isNoTasksForInstall: false
          });
          this.get('hosts').pushObject(hostInfo);
        }
      }
    },

    /**
     *
     * @param polledData: sets the {polledData} object of the controller
     */
    replacePolledData: function (polledData) {
      this.polledData.clear();
      this.set('polledData', polledData);
    },

    /**
     *
     * @param task
     * @returns {String} The appropriate message for the host as per the running task.
     */
    displayMessage: function (task) {
      var role = App.format.role(task.role);
      switch (task.command) {
        case 'INSTALL':
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.install.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.install.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.install.inProgress') + role;
            case 'COMPLETED' :
              return Em.I18n.t('installer.step9.serviceStatus.install.completed') + role;
            case 'FAILED':
              return Em.I18n.t('installer.step9.serviceStatus.install.failed') + role;
          }
        case 'UNINSTALL':
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.uninstall.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.uninstall.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.uninstall.inProgress') + role;
            case 'COMPLETED' :
              return Em.I18n.t('installer.step9.serviceStatus.uninstall.completed') + role;
            case 'FAILED':
              return Em.I18n.t('installer.step9.serviceStatus.uninstall.failed') + role;
          }
        case 'START' :
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.start.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.start.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.start.inProgress') + role;
            case 'COMPLETED' :
              return role + Em.I18n.t('installer.step9.serviceStatus.start.completed');
            case 'FAILED':
              return role + Em.I18n.t('installer.step9.serviceStatus.start.failed');
          }
        case 'STOP' :
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.stop.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.stop.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.stop.inProgress') + role;
            case 'COMPLETED' :
              return role + Em.I18n.t('installer.step9.serviceStatus.stop.completed');
            case 'FAILED':
              return role + Em.I18n.t('installer.step9.serviceStatus.stop.failed');
          }
        case 'EXECUTE' :
        case 'SERVICE_CHECK' :
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.execute.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.execute.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.execute.inProgress') + role;
            case 'COMPLETED' :
              return role + Em.I18n.t('installer.step9.serviceStatus.execute.completed');
            case 'FAILED':
              return role + Em.I18n.t('installer.step9.serviceStatus.execute.failed');
          }
        case 'ABORT' :
          switch (task.status) {
            case 'PENDING':
              return Em.I18n.t('installer.step9.serviceStatus.abort.pending') + role;
            case 'QUEUED' :
              return Em.I18n.t('installer.step9.serviceStatus.abort.queued') + role;
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.step9.serviceStatus.abort.inProgress') + role;
            case 'COMPLETED' :
              return role + Em.I18n.t('installer.step9.serviceStatus.abort.completed');
            case 'FAILED':
              return role + Em.I18n.t('installer.step9.serviceStatus.abort.failed');
          }
      }
      return '';
    },

    /**
     * Run start/check services after installation phase.
     * Does Ajax call to start all services
     */
    launchStartServices: function () {
      var data = {
        "RequestInfo": {
          "context": Em.I18n.t("requestInfo.startServices")
        },
        "Body": {
          "ServiceInfo": { "state": "STARTED" }
        }
      };
      var name = 'wizard.step9.installer.launch_start_services';

      if (this.get('content.controllerName') === 'addHostController') {
        var hostnames = [];
        for (var hostname in this.get('wizardController').getDBProperty('hosts')) {
          hostnames.push(hostname);
        }
        data = {
          "RequestInfo": {
            "context": Em.I18n.t("requestInfo.startHostComponents"),
            "query": "HostRoles/component_name.in(" + App.get('components.slaves').join(',') + ")&HostRoles/state=INSTALLED&HostRoles/host_name.in(" + hostnames.join(',') + ")"
          },
          "Body": {
            "HostRoles": { "state": "STARTED" }
          }
        };
        name = 'wizard.step9.add_host.launch_start_services';
      }
      data = JSON.stringify(data);
      if (App.testMode) {
        this.set('numPolls', 6);
      }

      App.ajax.send({
        name: name,
        sender: this,
        data: {
          data: data,
          cluster: this.get('content.cluster.name')
        },
        success: 'launchStartServicesSuccessCallback',
        error: 'launchStartServicesErrorCallback'
      });
    },

    /**
     * Success callback function for start services task.
     * @param jsonData: {json object} Contains Request id to poll.
     */
    launchStartServicesSuccessCallback: function (jsonData) {
      var clusterStatus = {};
      if (jsonData) {
        console.log("TRACE: Step9 -> In success function for the startService call");
        console.log("TRACE: Step9 -> value of the received data is: " + jsonData);
        var requestId = jsonData.Requests.id;
        console.log('requestId is: ' + requestId);
        clusterStatus = {
          status: 'INSTALLED',
          requestId: requestId,
          isStartError: false,
          isCompleted: false
        };
        this.hostHasClientsOnly(false);
        this.saveClusterStatus(clusterStatus);
      } else {
        console.log('ERROR: Error occurred in parsing JSON data');
        this.hostHasClientsOnly(true);
        clusterStatus = {
          status: 'STARTED',
          isStartError: false,
          isCompleted: true
        };
        this.saveClusterStatus(clusterStatus);
        this.set('status', 'success');
        this.set('progress', '100');
      }
      // We need to do recovery if there is a browser crash
      App.clusterStatus.setClusterStatus({
        clusterState: 'SERVICE_STARTING_3',
        wizardControllerName: this.get('content.controllerName'),
        localdb: App.db.data
      });

      if (jsonData) {
        this.startPolling();
      }
    },

    /**
     * This function will be called for Add host wizard only.
     * @param jsonError: {boolean} Boolean is true when API to start services returns 200 ok and no json data
     */
    hostHasClientsOnly: function (jsonError) {
      this.get('hosts').forEach(function (host) {
        var OnlyClients = true;
        var tasks = host.get('logTasks');
        tasks.forEach(function (task) {
          var component = serviceComponents.findProperty('component_name', task.Tasks.role);
          if (!(component && component.isClient)) {
            OnlyClients = false;
          }
        });
        if (OnlyClients || jsonError) {
          host.set('status', 'success');
          host.set('progress', '100');
        }
      });
    },

    /**
     * Error callback function for start services task.
     */
    launchStartServicesErrorCallback: function (jqXHR) {
      console.log("ERROR");
      this.set('startCallFailed',true);
      var clusterStatus = {
        status: 'INSTALL FAILED',
        isStartError: false,
        isCompleted: false
      };
      this.saveClusterStatus(clusterStatus);
      this.get('hosts').forEach(function (host) {
        host.set('progress', '100');
      });
      this.set('progress','100');

      var params = {
        cluster: this.get('content.cluster.name')
      };

      if (this.get('content.controllerName') === 'addHostController') {
        params.name = 'wizard.step9.add_host.launch_start_services';
      } else {
        params.name = 'wizard.step9.installer.launch_start_services';
      }

      var opt = App.formatRequest.call(App.urls[params.name], params);
      App.ajax.defaultErrorHandler(jqXHR,opt.url,opt.type);
    },

    /**
     * marks a host's status as "success" if all tasks are in COMPLETED state
     */
    onSuccessPerHost: function (actions, contentHost) {
      if (actions.everyProperty('Tasks.status', 'COMPLETED') && this.get('content.cluster.status') === 'INSTALLED') {
        contentHost.set('status', 'success');
      }
    },

    /**
     * marks a host's status as "warning" if at least one of the tasks is FAILED, ABORTED, or TIMEDOUT and marks host's status as "failed" if at least one master component install task is FAILED.
     * note that if the master failed to install because of ABORTED or TIMEDOUT, we don't mark it as failed, because this would mark all hosts as "failed" and makes it difficult for the user
     * to find which host FAILED occurred on, if any
     * @param actions {Array} of tasks retrieved from polled data
     * @param contentHost {Object} A host object
     */

    onErrorPerHost: function (actions, contentHost) {
      if (!actions) return;
      if (actions.someProperty('Tasks.status', 'FAILED') || actions.someProperty('Tasks.status', 'ABORTED') || actions.someProperty('Tasks.status', 'TIMEDOUT')) {
        contentHost.set('status', 'warning');
      }
      if ((this.get('content.cluster.status') === 'PENDING' && actions.someProperty('Tasks.status', 'FAILED')) || (this.isMasterFailed(actions))) {
        contentHost.get('status') !== 'heartbeat_lost' ? contentHost.set('status', 'failed') : '';
      }
    },

    /**
     *
     * @param polledData : Json data polled from API.
     * @returns {boolean}  true if there is at least one FAILED task of master component install
     */
    isMasterFailed: function (polledData) {
      var result = false;
      polledData.filterProperty('Tasks.command', 'INSTALL').filterProperty('Tasks.status', 'FAILED').mapProperty('Tasks.role').forEach(
        function (role) {
          if (!App.get('components.slaves').contains(role)) {
            result = true;
          }
        }
      );
      return result;
    },

    /**
     * Mark a host status as in_progress if the any task on the host if either in IN_PROGRESS, QUEUED or PENDONG state.
     * @param actions {Array} of tasks retrieved from polled data
     * @param contentHost {Object} A host object
     */
    onInProgressPerHost: function (actions, contentHost) {
      var runningAction = actions.findProperty('Tasks.status', 'IN_PROGRESS');
      if (runningAction === undefined || runningAction === null) {
        runningAction = actions.findProperty('Tasks.status', 'QUEUED');
      }
      if (runningAction === undefined || runningAction === null) {
        runningAction = actions.findProperty('Tasks.status', 'PENDING');
      }
      if (runningAction !== null && runningAction !== undefined) {
        contentHost.set('status', 'in_progress');
        contentHost.set('message', this.displayMessage(runningAction.Tasks));
      }
    },

    /**
     * calculate progress of tasks per host
     * @param actions
     * @param contentHost
     * @return {Number}
     */
    progressPerHost: function (actions, contentHost) {
      var progress = 0;
      var actionsPerHost = actions.length;
      var completedActions = 0;
      var queuedActions = 0;
      var inProgressActions = 0;
      actions.forEach(function (action) {
        completedActions += +(['COMPLETED', 'FAILED', 'ABORTED', 'TIMEDOUT'].contains(action.Tasks.status));
        queuedActions += +(action.Tasks.status === 'QUEUED');
        inProgressActions += +(action.Tasks.status === 'IN_PROGRESS');
      }, this);
      /** for the install phase (PENDING), % completed per host goes up to 33%; floor(100 / 3)
       * for the start phase (INSTALLED), % completed starts from 34%
       * when task in queued state means it's completed on 9%
       * in progress - 35%
       * completed - 100%
       */
      switch (this.get('content.cluster.status')) {
        case 'PENDING':
          progress = actionsPerHost ? (Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsPerHost * 33)) : 33;
          break;
        case 'INSTALLED':
          progress = actionsPerHost ? (34 + Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsPerHost * 66)) : 100;
          break;
        default:
          progress = 100;
          break;
      }
      contentHost.set('progress', progress.toString());
      return progress;
    },

    /**
     *
     * @param polledData : Josn data retrieved from API
     * @returns {Boolean} : Has step completed successfully
     */
    isSuccess: function (polledData) {
      return polledData.everyProperty('Tasks.status', 'COMPLETED');
    },

    /**
     * return true if:
     *  1. any of the master/client components failed to install
     *  OR
     *  2. at least 50% of the slave host components for the particular service component fails to install
     */
    isStepFailed: function () {
      var failed = false;
      var polledData = this.get('polledData');
      polledData.filterProperty('Tasks.command', 'INSTALL').mapProperty('Tasks.role').uniq().forEach(function (role) {
        if (failed) {
          return;
        }
        var actionsPerRole = polledData.filterProperty('Tasks.role', role);
        if (App.get('components.slaves').contains(role)) {
          // check slave components for success factor.
          // partial failure for slave components are allowed.
          var actionsFailed = actionsPerRole.filterProperty('Tasks.status', 'FAILED');
          var actionsAborted = actionsPerRole.filterProperty('Tasks.status', 'ABORTED');
          var actionsTimedOut = actionsPerRole.filterProperty('Tasks.status', 'TIMEDOUT');
          if ((((actionsFailed.length + actionsAborted.length + actionsTimedOut.length) / actionsPerRole.length) * 100) > 50) {
            failed = true;
          }
        } else if (actionsPerRole.someProperty('Tasks.status', 'FAILED') || actionsPerRole.someProperty('Tasks.status', 'ABORTED') ||
          actionsPerRole.someProperty('Tasks.status', 'TIMEDOUT')) {
          // check non-salve components (i.e., masters and clients).  all of these must be successfully installed.
          failed = true;
        }
      }, this);
      return failed;
    },

    /**
     * polling from ui stops only when no action has 'PENDING', 'QUEUED' or 'IN_PROGRESS' status
     * Makes a state transition
     * PENDING -> INSTALLED
     * PENDING -> INSTALL FAILED
     * INSTALLED -> STARTED
     * INSTALLED -> START_FAILED
     * @param polledData json data retrieved from API
     * @returns {Boolean} true if polling should stop; false otherwise
     */
    finishState: function (polledData) {
      if (this.get('content.cluster.status') === 'INSTALLED') {
        return this.isServicesStarted(polledData);
      } else if (this.get('content.cluster.status') === 'PENDING') {
        return this.isServicesInstalled(polledData);
      } else if (this.get('content.cluster.status') === 'INSTALL FAILED' || this.get('content.cluster.status') === 'START FAILED'
        || this.get('content.cluster.status') === 'STARTED') {
        this.set('progress', '100');
        return true;
      }
      return false;
    },

    /**
     * @param polledData Josn data retrieved from API
     * @returns {boolean} Has "Start All Services" request completed successfully
     */
    isServicesStarted: function (polledData) {
      var clusterStatus = {};
      if (!polledData.someProperty('Tasks.status', 'PENDING') && !polledData.someProperty('Tasks.status', 'QUEUED') && !polledData.someProperty('Tasks.status', 'IN_PROGRESS')) {
        this.set('progress', '100');
        clusterStatus = {
          status: 'INSTALLED',
          requestId: this.get('content.cluster.requestId'),
          isCompleted: true
        };
        if (this.isSuccess(polledData)) {
          clusterStatus.status = 'STARTED';
          var serviceStartTime = App.dateTime();
          clusterStatus.installTime = ((parseInt(serviceStartTime) - parseInt(this.get('content.cluster.installStartTime'))) / 60000).toFixed(2);
        } else {
          clusterStatus.status = 'START FAILED'; // 'START FAILED' implies to step10 that installation was successful but start failed
        }
        this.saveClusterStatus(clusterStatus);
        this.saveInstalledHosts(this);
        return true;
      }
      return false;
    },

    /**
     * @param polledData Josn data retrieved from API
     * @returns {boolean} Has "Install All Services" request completed successfully
     */
    isServicesInstalled: function (polledData) {
      var clusterStatus = {};
      if (!polledData.someProperty('Tasks.status', 'PENDING') && !polledData.someProperty('Tasks.status', 'QUEUED') && !polledData.someProperty('Tasks.status', 'IN_PROGRESS')) {
        clusterStatus = {
          status: 'PENDING',
          requestId: this.get('content.cluster.requestId'),
          isCompleted: false
        };
        if (this.get('status') === 'failed') {
          clusterStatus.status = 'INSTALL FAILED';
          this.saveClusterStatus(clusterStatus);
          this.set('progress', '100');
          this.get('hosts').forEach(function (host) {
            host.set('progress', '100');
          });
          this.isAllComponentsInstalled();
        } else {
          this.set('progress', '34');
          if (this.get('content.controllerName') === 'installerController') {
            this.isAllComponentsInstalled();
          } else {
            this.launchStartServices();
          }
        }
        this.saveInstalledHosts(this);
        return true;
      }
      return false;
    },

    /**
     * This is done at HostRole level.
     * @param tasksPerHost {Array}
     * @param host {Object}
     */
    setLogTasksStatePerHost: function (tasksPerHost, host) {
      tasksPerHost.forEach(function (_task) {
        var task = host.logTasks.findProperty('Tasks.id', _task.Tasks.id);
        if (task) {
          task.Tasks.status = _task.Tasks.status;
          task.Tasks.exit_code = _task.Tasks.exit_code;
        } else {
          host.logTasks.pushObject(_task);
        }
      }, this);
    },

    /**
     * Parses the Json data retrieved from API and sets the task on the host of {hosts} array binded to template
     * @param polledData Json data retrieved from API
     * @returns {Boolean} True if stage transition is completed.
     * On true, polling will be stopped.
     */
    parseHostInfo: function (polledData) {
      console.log('TRACE: Entering host info function');
      var self = this;
      var totalProgress = 0;
      var tasksData = polledData.tasks;
      console.log("The value of tasksData is: ", tasksData);
      if (!tasksData) {
        console.log("Step9: ERROR: NO tasks available to process");
      }
      var requestId = this.get('content.cluster.requestId');
      tasksData.setEach('Tasks.request_id', requestId);
      if (polledData.Requests && polledData.Requests.id && polledData.Requests.id != requestId) {
        // We don't want to use non-current requestId's tasks data to
        // determine the current install status.
        // Also, we don't want to keep polling if it is not the
        // current requestId.
        return false;
      }
      this.replacePolledData(tasksData);
      var tasksHostMap = {};
      tasksData.forEach(function (task) {
        if (tasksHostMap[task.Tasks.host_name]) {
          tasksHostMap[task.Tasks.host_name].push(task);
        } else {
          tasksHostMap[task.Tasks.host_name] = [task];
        }
      });

      this.get('hosts').forEach(function (_host) {
        var actionsPerHost = tasksHostMap[_host.name] || []; // retrieved from polled Data
        if (actionsPerHost.length === 0) {
          if (this.get('content.cluster.status') === 'PENDING' || this.get('content.cluster.status') === 'INSTALL FAILED') {
            _host.set('progress', '33');
            _host.set('isNoTasksForInstall', true);
            _host.set('status', 'pending');
          }
          if (this.get('content.cluster.status') === 'INSTALLED' || this.get('content.cluster.status') === 'FAILED') {
            _host.set('progress', '100');
            _host.set('status', 'success');
          }
          console.log("INFO: No task is hosted on the host");
        } else {
          _host.set('isNoTasksForInstall', false);
        }
        this.setLogTasksStatePerHost(actionsPerHost, _host);
        this.onSuccessPerHost(actionsPerHost, _host);     // every action should be a success
        this.onErrorPerHost(actionsPerHost, _host);     // any action should be a failure
        this.onInProgressPerHost(actionsPerHost, _host);  // current running action for a host
        totalProgress += self.progressPerHost(actionsPerHost, _host);
        if (_host.get('progress') == '33' && _host.get('status') != 'failed' && _host.get('status') != 'warning') {
          _host.set('message', this.t('installer.step9.host.status.nothingToInstall'));
          _host.set('status', 'pending');
        }
      }, this);
      this.set('logTasksChangesCounter', this.get('logTasksChangesCounter') + 1);
      totalProgress = Math.floor(totalProgress / this.get('hosts.length'));
      this.set('progress', totalProgress.toString());
      console.log("INFO: right now the progress is: " + this.get('progress'));
      return this.finishState(tasksData);
    },

    /**
     * starts polling to the API.
     */
    startPolling: function () {
      this.set('isSubmitDisabled', true);
      this.doPolling();
    },

    /**
     *
     * @param requestId {Int} Request Id received on triggering install/start command successfully
     * @returns {string} URL to poll to track the result of the triggered command
     */
    getUrl: function (requestId) {
      var clusterName = this.get('content.cluster.name');
      var requestId = requestId || this.get('content.cluster.requestId');
      var url = App.apiPrefix + '/clusters/' + clusterName + '/requests/' + requestId + '?fields=tasks/Tasks/command,tasks/Tasks/exit_code,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,tasks/Tasks/status&minimal_response=true';
      console.log("URL for step9 is: " + url);
      return url;
    },

    /**
     * This function calls API just once to fetch log data of all tasks.
     * @param requestId {Int} Request Id received on triggering install/start command successfully
     */
    loadLogData: function (requestId) {
      var url = this.getUrl(requestId);
      var requestsId = this.get('wizardController').getDBProperty('cluster').oldRequestsId;
      if (App.testMode) {
        this.POLL_INTERVAL = 1;
      }

      requestsId.forEach(function (requestId) {
        url = this.getUrl(requestId);
        if (App.testMode) {
          this.POLL_INTERVAL = 1;
          url = this.get('mockDataPrefix') + '/poll_' + this.numPolls + '.json';
        }
        this.getLogsByRequest(url, false);
      }, this);
    },
    /**
     * {Number}
     * <code>taskId</code> of current open task
     */
    currentOpenTaskId: 0,

    /**
     * {Number}
     * <code>requestId</code> of current open task
     */
    currentOpenTaskRequestId: 0,

    /**
     * Load form server <code>stderr, stdout</code> of current open task
     */
    loadCurrentTaskLog: function () {
      var taskId = this.get('currentOpenTaskId');
      var requestId = this.get('currentOpenTaskRequestId');
      var clusterName = this.get('content.cluster.name');
      if (!taskId) {
        console.log('taskId is null.');
        return;
      }
      App.ajax.send({
        name: 'background_operations.get_by_task',
        sender: this,
        data: {
          'taskId': taskId,
          'requestId': requestId,
          'clusterName': clusterName,
          'sync': true
        },
        success: 'loadCurrentTaskLogSuccessCallback',
        error: 'loadCurrentTaskLogErrorCallback'
      });
    },

    /**
     * success callback function for getting log data of the opened task
     * @param data json object
     */
    loadCurrentTaskLogSuccessCallback: function (data) {
      var taskId = this.get('currentOpenTaskId');
      if (taskId) {
        var currentTask = this.get('hosts').findProperty('name', data.Tasks.host_name).get('logTasks').findProperty('Tasks.id', data.Tasks.id);
        if (currentTask) {
          currentTask.Tasks.stderr = data.Tasks.stderr;
          currentTask.Tasks.stdout = data.Tasks.stdout;
        }
      }
      this.set('logTasksChangesCounter', this.get('logTasksChangesCounter') + 1);
    },

    /**
     * Error callback function for getting log data of the opened task
     */
    loadCurrentTaskLogErrorCallback: function () {
      this.set('currentOpenTaskId', 0);
    },

    /**
     * Function polls the API to retrieve data for the request.
     * @param url {string} url to poll
     * @param polling  {Boolean} whether to continue polling for status or not
     */
    getLogsByRequest: function (url, polling) {
      var self = this;
      $.ajax({
        type: 'GET',
        url: url,
        async: true,
        timeout: App.timeout,
        dataType: 'text',
        success: function (data) {
          var parsedData = jQuery.parseJSON(data);
          console.log("TRACE: In success function for the GET logs data");
          console.log("TRACE: Step9 -> The value is: ", parsedData);
          var result = self.parseHostInfo(parsedData);
          if (!polling) {
            if (self.get('content.cluster.status') === 'INSTALL FAILED') {
              self.isAllComponentsInstalled();
            }
            return;
          }
          if (result !== true) {
            window.setTimeout(function () {
              if (self.get('currentOpenTaskId')) {
                self.loadCurrentTaskLog();
              }
              self.doPolling();
            }, self.POLL_INTERVAL);
          }
        },

        error: function (request, ajaxOptions, error) {
          console.log("TRACE: STep9 -> In error function for the GET logs data");
          console.log("TRACE: STep9 -> value of the url is: " + url);
          console.log("TRACE: STep9 -> error code status is: " + request.status);
        },

        statusCode: require('data/statusCodes')
      }).retry({times: App.maxRetries, timeout: App.timeout}).then(null,
        function () {
          App.showReloadPopup();
          console.log('Install services all retries failed');
        }
      );
    },

    /**
     * Delegates the function call to {getLogsByRequest} with appropriate params
     */
    doPolling: function () {
      var url = this.getUrl();

      if (App.testMode) {
        this.numPolls++;
        url = this.get('mockDataPrefix') + '/poll_' + this.get('numPolls') + '.json';

      }
      this.getLogsByRequest(url, true);
    },

    /**
     * Check that all components are in INSTALLED state before issuing start command
     */
    isAllComponentsInstalled: function () {
      if (this.get('content.controllerName') !== 'installerController') {
        return;
      }
      var name = 'wizard.step9.installer.get_host_status';
      App.ajax.send({
        name: name,
        sender: this,
        data: {
          cluster: this.get('content.cluster.name')
        },
        success: 'isAllComponentsInstalledSuccessCallback',
        error: 'isAllComponentsInstalledErrorCallback'
      });
    },

    /**
     * Success callback function for API checking host state and host_components state.
     * @param jsonData {Object}
     */
    isAllComponentsInstalledSuccessCallback: function (jsonData) {
      var clusterStatus = {
        status: 'INSTALL FAILED',
        isStartError: true,
        isCompleted: false
      };
      var hostsWithHeartbeatLost = [];
      jsonData.items.filterProperty('Hosts.host_state', 'HEARTBEAT_LOST').forEach(function (host) {
        var hostComponentObj = {hostName: host.Hosts.host_name};
        var componentArr = [];
        host.host_components.forEach(function (_hostComponent) {
          var componentName = App.format.role(_hostComponent.HostRoles.component_name);
          componentArr.pushObject(componentName);
        }, this);
        hostComponentObj.componentNames = this.getComponentMessage(componentArr);
        hostsWithHeartbeatLost.pushObject(hostComponentObj);
      }, this);
      this.set('hostsWithHeartbeatLost', hostsWithHeartbeatLost);
      if (hostsWithHeartbeatLost.length) {
        this.get('hosts').forEach(function (host) {
          if (hostsWithHeartbeatLost.someProperty(('hostName'), host.get('name'))) {
            host.set('status', 'heartbeat_lost');
          } else if (host.get('status') !== 'failed' && host.get('status') !== 'warning') {
            host.set('message', Em.I18n.t('installer.step9.host.status.startAborted'));
          }
          host.set('progress', '100');
        });
        this.set('progress', '100');
        this.saveClusterStatus(clusterStatus);
      } else if (this.get('content.cluster.status') === 'PENDING') {
        this.launchStartServices();
      }

    },

    /**
     * Error callback function for API checking host state and host_components state
     */
    isAllComponentsInstalledErrorCallback: function () {
      console.log("ERROR");
      var clusterStatus = {
        status: 'INSTALL FAILED',
        isStartError: true,
        isCompleted: false
      };
      this.set('progress', '100');
      this.get('hosts').forEach(function (host) {
        if (host.get('status') !== 'failed' && host.get('status') !== 'warning') {
          host.set('message', Em.I18n.t('installer.step9.host.status.startAborted'));
          host.set('progress', '100');
        }
      });
      this.saveClusterStatus(clusterStatus);
    },

    /**
     * @param componentArr {Array}  Array of components
     * @returns {String} Formatted string of components to display on the UI.
     */
    getComponentMessage: function (componentArr) {
      var label;
      componentArr.forEach(function (_component) {
        if (_component === componentArr[0]) {
          label = _component;
        } else if (_component !== componentArr[componentArr.length - 1]) {           // [clients.length - 1]
          label = label + ' ' + _component;
          if (_component !== componentArr[componentArr.length - 2]) {
            label = label + ',';
          }
        } else {
          label = label + ' ' + Em.I18n.t('and') + ' ' + _component;
        }
      }, this);
      return label;
    },

    /**
     * save cluster status in the parentController and localdb
     * @param clusterStatus {Object}
     */
    saveClusterStatus: function(clusterStatus) {
      if (!App.testMode) {
        App.router.get(this.get('content.controllerName')).saveClusterStatus(clusterStatus);
      } else {
        this.set('content.cluster',clusterStatus);
      }
    },

    /**
     * save cluster status in the parentController and localdb
     * @param context
     */
    saveInstalledHosts: function(context) {
      if (!App.testMode) {
        App.router.get(this.get('content.controllerName')).saveInstalledHosts(context)
      }
    }
  });
  
});
window.require.register("data/HDP2/config_mapping", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var configs = [
  /**********************************************core-site***************************************/
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["hive_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["hivemetastore_host"],
      "foreignKey": ["hive_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["oozie_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["oozieserver_host"],
      "foreignKey": ["oozie_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["webhcat_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["hivemetastore_host"],
      "foreignKey": ["webhcat_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["falcon_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable": true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": [],
      "foreignKey": ["falcon_user"],
      "value": "*",
      "filename": "core-site.xml",
      "isOverridable": true
    },

  /**********************************Oozie******************************/

    {
      "name": "oozie.service.ProxyUserService.proxyuser.<foreignKey[0]>.groups",
      "templateName": [],
      "foreignKey": ["falcon_user"],
      "value": "*",
      "filename": "oozie-site.xml",
      "isOverridable": true,
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.ProxyUserService.proxyuser.<foreignKey[0]>.hosts",
      "templateName": [],
      "foreignKey": ["falcon_user"],
      "value": "*",
      "filename": "oozie-site.xml",
      "isOverridable": true,
      "serviceName": "OOZIE"
    }
  ];

  /**
   * Configs consists of 2 types: Computed values, which cannot be modified by user
   * and overridable values, which user can modify. We provide interface how to get all of this
   * configs separately
   * @type {Object}
   */
  module.exports = {

    checkPrecondition: function () {
      return configs.filter(function (config) {
        return ((!config.precondition) || (config.precondition()));
      });
    },
    all: function (skipPreconditionCheck) {
      if (skipPreconditionCheck) {
        return configs.slice(0);
      } else {
        return this.checkPrecondition().slice(0);
      }
    },
    overridable: function (skipPreconditionCheck) {
      if (skipPreconditionCheck) {
        return configs.filterProperty('foreignKey');
      } else {
        return this.checkPrecondition().filterProperty('foreignKey');
      }
    },
    computed: function (skipPreconditionCheck) {
      if (skipPreconditionCheck) {
        return configs.filterProperty('foreignKey', null);
      } else {
        return this.checkPrecondition().filterProperty('foreignKey', null);
      }
    }
  };
  
});
window.require.register("data/HDP2/custom_configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  module.exports =
    [
      {
        "id": "site property",
        "name": "yarn.scheduler.capacity.root.<queue-name>.capacity",
        "displayName": "Capacity",
        "value": '',
        "defaultValue": '',
        "description": "Percentage of the number of slots in the cluster that are made to be available for jobs in this queue. The sum of capacities for all queues should be less than or equal 100.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "",
        "category": "CapacityScheduler",
        "unit": "%",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 1
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.capacity.root.<queue-name>.maximum-capacity",
        "displayName": "Max Capacity",
        "value": '',
        "defaultValue": '100',
        "displayType": "int",
        "description": "Defines a limit beyond which a queue cannot use the capacity of the cluster." +
          "This provides a means to limit how much excess capacity a queue can use. By default, there is no limit." +
          "The Max Capacity of a queue can only be greater than or equal to its minimum capacity. " +
          "This property could be to curtail certain jobs which are long running in nature from occupying more than a certain " +
          "percentage of the cluster, which in the absence of pre-emption, could lead to capacity guarantees of other queues being affected. " +
          "One important thing to note is that maximum-capacity is a percentage , so based on the cluster's capacity it would change. " +
          "So if large no of nodes or racks get added to the cluster, Max Capacity in absolute terms would increase accordingly. ",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "YARN",
        "category": "CapacityScheduler",
        "unit": "%",
        "valueRange": [0, 100],
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 2
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.capacity.root.<queue-name>.user-limit-factor",
        "displayName": "User Limit Factor",
        "value": '',
        "defaultValue": '1',
        "displayType": "int",
        "description": "The multiple of the queue capacity which can be configured to allow a single user to acquire more slots. " +
          "By default this is set to 1 which ensure that a single user can never take more than the queue's configured capacity " +
          "irrespective of how idle the cluster is.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "YARN",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 7
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.capacity.root.<queue-name>.state",
        "displayName": "State",
        "value": '',
        "defaultValue": 'RUNNING',
        "description": "state",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "YARN",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 8
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.capacity.root.<queue-name>.acl_submit_jobs",
        "displayName": "",
        "value": '',
        "defaultValue": '*',
        "description": "",
        "isVisible": false,
        "isRequired": true,
        "serviceName": "YARN",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml'
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.capacity.root.<queue-name>.acl_administer_jobs",
        "displayName": "",
        "value": '',
        "defaultValue": '*',
        "description": "",
        "isVisible": false,
        "isRequired": true,
        "serviceName": "YARN",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml'
      }
    ];
  
});
window.require.register("data/HDP2/global_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  /**
   * Defines service configuration properties.
   *   name:
   *     The name of the config property that is understood by Ambari server and agent.
   *     E.g., "datanode_du_reserved"
   *
   *   displayName:
   *     The human-friendly display name of the config property.
   *     E.g., "Reserved space for HDFS"
   *
   *   description:
   *     The description of the config property.
   *     E.g., "Reserved space in GB per volume"
   *
   *   defaultValue:
   *     The default value of the config property.
   *     E.g., "1"
   *
   *   isReconfigurable:
   *     Whether the config property can be reconfigured after it has been initially set and deployed.
   *     If this is unspecified, true is assumed.
   *     E.g., true, false
   *
   *   isOverridable:
   *     Whether the config property can be overridden by hosts.
   *     If this is unspecified, true is assumed.
   *
   *   isRequired:
   *     Whether the config property is required or not.
   *     If this is unspecified, true is assumed.
   *     E.g., true, false
   *
   *     isRequiredByAgent:
   *     Whether the config property is required by agent or not.
   *     If value is false then it will be not persisted in global configuration
   *
   *   displayType:
   *     How the config property is to be rendered for user input.
   *     If this is left unspecified, "string" is assumed
   *     E.g., "string", "int", "float", "checkbox", "directories", "custom", "email", "masterHost", "slaveHosts"
   *
   *   unit
   *     The unit for the config property.
   *     E.g., "ms", "MB", "bytes"
   *
   *   serviceName:
   *     The service that the config property belongs to.
   *     E.g., "HDFS", "MAPREDUCE2", "ZOOKEEPER", etc.
   *
   *   category: the category that the config property belongs to (used for grouping config properties in the UI).
   *     if unspecified, "General" is assumed.
   *     E.g., "General", "Advanced", "NameNode", "DataNode"
   *
   *   index: the sequence number in category, that point to place where config located regarding all rest in category.
   *     if unspecified, push to the end of array.
   *     E.g., 0, 1, '2'
   */

  var App = require('app');
  require('config');

  module.exports =
  {
    "configProperties": [
      /**********************************************HDFS***************************************/
      {
        "id": "puppet var",
        "name": "namenode_host",
        "displayName": "NameNode hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that has been assigned to run NameNode",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "namenode_heapsize",
        "displayName": "NameNode Java heap size",
        "description": "Initial and maximum Java heap size for NameNode (Java options -Xms and -Xmx).  This also applies to the Secondary NameNode.",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "namenode_opt_newsize",
        "displayName": "NameNode new generation size",
        "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
        "defaultValue": "200",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "snamenode_host",
        "displayName": "SNameNode host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run SecondaryNameNode",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "datanode_hosts", //not in the schema. For UI purpose
        "displayName": "DataNode hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run DataNode",
        "displayType": "slaveHosts",
        "isRequired": false,
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "domain": "datanode-global",
        "serviceName": "HDFS",
        "category": "DataNode",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "dtnode_heapsize",
        "displayName": "DataNode maximum Java heap size",
        "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "domain": "datanode-global",
        "serviceName": "HDFS",
        "category": "DataNode",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "hadoop_heapsize",
        "displayName": "Hadoop maximum Java heap size",
        "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "serviceName": "HDFS",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "hdfs_log_dir_prefix",
        "displayName": "Hadoop Log Dir Prefix",
        "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
        "defaultValue": "/var/log/hadoop",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hadoop_pid_dir_prefix",
        "displayName": "Hadoop PID Dir Prefix",
        "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
        "defaultValue": "/var/run/hadoop",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "namenode_opt_maxnewsize",
        "displayName": "NameNode maximum new generation size",
        "description": "",
        "defaultValue": "200",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode"
      },
      {
        "id": "puppet var",
        "name": "security_enabled",
        "displayName": "Hadoop Security",
        "description": "Enable hadoop security",
        "defaultValue": 'false',
        "isRequired": false,
        "displayType": "checkbox",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hcat_conf_dir",
        "displayName": "HCat conf directory",
        "description": "",
        "defaultValue": "",
        "isRequired": false,
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "lzo_enabled",
        "displayName": "LZO compression",
        "description": "LZO compression enabled",
        "defaultValue": true,
        "displayType": "checkbox",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HDFS"
      },

    /**********************************************MAPREDUCE2***************************************/
      {
        "id": "puppet var",
        "name": "hs_host",
        "displayName": "History Server",
        "description": "History Server",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE2",
        "category": "HistoryServer",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "mapred_log_dir_prefix",
        "displayName": "Mapreduce Log Dir Prefix",
        "description": "",
        "defaultValue": "/var/log/hadoop-mapreduce",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isReconfigurable": false,
        "serviceName": "MAPREDUCE2",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "mapred_pid_dir_prefix",
        "displayName": "Mapreduce PID Dir Prefix",
        "description": "",
        "defaultValue": "/var/run/hadoop-mapreduce",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isReconfigurable": false,
        "serviceName": "MAPREDUCE2",
        "category": "Advanced"
      },

    /**********************************************YARN***************************************/
      {
        "id": "puppet var",
        "name": "yarn_heapsize",
        "displayName": "YARN Java heap size",
        "description": "Max heapsize for all YARN components",
        "defaultValue": "1024",
        "isOverridable": true,
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "serviceName": "YARN",
        "category": "General",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "rm_host",
        "displayName": "ResourceManager",
        "description": "ResourceManager",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "YARN",
        "category": "ResourceManager",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "resourcemanager_heapsize",
        "displayName": "ResourceManager Java heap size",
        "description": "Max heapsize for ResourceManager",
        "defaultValue": "1024",
        "isOverridable": false,
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "serviceName": "YARN",
        "category": "ResourceManager",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "nm_hosts",
        "displayName": "NodeManager",
        "description": "List of NodeManager Hosts.",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "slaveHosts",
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "YARN",
        "category": "NodeManager",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "nodemanager_heapsize",
        "displayName": "NodeManager Java heap size",
        "description": "Max heapsize for NodeManager",
        "defaultValue": "1024",
        "isOverridable": true,
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "serviceName": "YARN",
        "category": "NodeManager",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "yarn_log_dir_prefix",
        "displayName": "YARN Log Dir Prefix",
        "description": "",
        "defaultValue": "/var/log/hadoop-yarn",
        "displayType": "directory",
        "isOverridable": false,
        "isReconfigurable": false,
        "isVisible": true,
        "serviceName": "YARN",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "yarn_pid_dir_prefix",
        "displayName": "YARN PID Dir Prefix",
        "description": "",
        "defaultValue": "/var/run/hadoop-yarn",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isReconfigurable": false,
        "serviceName": "YARN",
        "category": "Advanced"
      },
  // @todo uncomment after App Timeline service integration
  //    {
  //      "id": "puppet var",
  //      "name": "ats_host",
  //      "displayName": "App Timeline Server",
  //      "description": "Application Timeline Server Host",
  //      "defaultValue": "",
  //      "isOverridable": false,
  //      "displayType": "masterHost",
  //      "isRequiredByAgent": false,
  //      "isVisible": true,
  //      "serviceName": "YARN",
  //      "category": "AppTimelineServer",
  //      "index": 0
  //    },
  //    {
  //      "id": "puppet var",
  //      "name": "apptimelineserver_heapsize",
  //      "displayName": "AppTimelineServer Java heap size",
  //      "description": "AppTimelineServer Java heap size",
  //      "defaultValue": "1024",
  //      "isOverridable": false,
  //      "displayType": "int",
  //      "unit": "MB",
  //      "isVisible": true,
  //      "serviceName": "YARN",
  //      "category": "AppTimelineServer",
  //      "index": 1
  //    }
    /**********************************************HBASE***************************************/
      {
        "id": "puppet var",
        "name": "hbasemaster_host",
        "displayName": "HBase Master hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run HBase Master",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "hbase_master_heapsize",
        "displayName": "HBase Master Maximum Java heap size",
        "description": "Maximum Java heap size for HBase master (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": true,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "regionserver_hosts",
        "displayName": "RegionServer hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run RegionServer",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "isRequired": false,
        "domain": "regionserver-global",
        "serviceName": "HBASE",
        "category": "RegionServer",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "hbase_regionserver_heapsize",
        "displayName": "HBase RegionServers maximum Java heap size",
        "description": "Maximum Java heap size for HBase RegionServers (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "domain": "regionserver-global",
        "serviceName": "HBASE",
        "category": "RegionServer",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "hbase_log_dir",
        "displayName": "HBase Log Dir",
        "description": "Directory for HBase logs",
        "defaultValue": "/var/log/hbase",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hbase_pid_dir",
        "displayName": "HBase PID Dir",
        "description": "Directory in which the pid files for HBase processes will be created",
        "defaultValue": "/var/run/hbase",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "Advanced"
      },
      /**********************************************GLUSTERFS***************************************/
      {
        "id": "puppet var",
        "name": "glusterfs_defaultFS_name",
        "displayName": "GlusterFS default fs name",
        "description": "GlusterFS default filesystem name (glusterfs:///)",
        "defaultValue": "glusterfs:///",
        "displayType": "string",
        "isVisible": true,
        "domain": "global",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
          "id": "puppet var",
          "name": "fs_glusterfs_default_name",
          "displayName": "GlusterFS default fs name",
          "description": "GlusterFS default filesystem name (glusterfs:///)",
          "defaultValue": "glusterfs:///",
          "displayType": "string",
          "isVisible": true,
          "domain": "global",
          "serviceName": "GLUSTERFS",
          "category": "General"
        },
            {
              "id": "puppet var",
              "name": "hadoop_heapsize",
              "displayName": "Hadoop maximum Java heap size",
              "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
              "defaultValue": "1024",
              "displayType": "int",
              "unit": "MB",
              "isVisible": true,
              "domain": "global",
              "serviceName": "GLUSTERFS",
              "category": "General"
            },

    /**********************************************HIVE***************************************/
      {
        "id": "puppet var",
        "name": "hivemetastore_host",
        "displayName": "Hive Metastore host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run Hive Metastore",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 0
      },
      // for existing MySQL
      {
        "id": "puppet var",
        "name": "hive_existing_mysql_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "Using an existing MySQL database for Hive Metastore",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": false,
        "isReconfigurable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 1
      },
      // for existing Oracle
      {
        "id": "puppet var",
        "name": "hive_existing_oracle_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "Oracle",
        "description": "Using an existing Oracle database for Hive Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 1
      },
      // for new MySQL
      {
        "id": "puppet var",
        "name": "hive_ambari_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "MySQL will be installed by Ambari",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "hive_database",
        "displayName": "Hive Database",
        "value": "",
        "defaultValue": "New MySQL Database",
        "options": [
          {
            displayName: 'New MySQL Database',
            foreignKeys: ['hive_ambari_database', 'hive_ambari_host']
          },
          {
            displayName: 'Existing MySQL Database',
            foreignKeys: ['hive_existing_mysql_database', 'hive_existing_mysql_host']
          },
          {
            displayName: 'Existing Oracle Database',
            foreignKeys: ['hive_existing_oracle_database', 'hive_existing_oracle_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          }
        ],
        "description": "MySQL will be installed by Ambari",
        "displayType": "radio button",
        "isReconfigurable": false,
        "radioName": "hive-database",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "hive_hostname",
        "displayName": "Database Host",
        "description": "Specify the host on which the database is hosted",
        "defaultValue": "",
        "isReconfigurable": true,
        "displayType": "host",
        "isOverridable": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_existing_mysql_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "displayType": "host",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_existing_oracle_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "displayType": "host",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_ambari_host",
        "value": "",
        "defaultValue": "",
        "displayName": "Database Host",
        "description": "Host on which the database will be created by Ambari",
        "isReconfigurable": false,
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_metastore_port",
        "displayName": "Hive metastore port",
        "description": "",
        "defaultValue": "9083",
        "isReconfigurable": false,
        "displayType": "int",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_lib",
        "displayName": "Hive library",
        "description": "",
        "defaultValue": "/usr/lib/hive/lib/",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_dbroot",
        "displayName": "Hive db directory",
        "description": "",
        "defaultValue": "/usr/lib/hive/lib",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_log_dir",
        "displayName": "Hive Log Dir",
        "description": "Directory for Hive log files",
        "defaultValue": "/var/log/hive",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_pid_dir",
        "displayName": "Hive PID Dir",
        "description": "Directory in which the PID files for Hive processes will be created",
        "defaultValue": "/var/run/hive",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_aux_jars_path",
        "displayName": "Hive auxilary jar path",
        "description": "",
        "defaultValue": "/usr/lib/hcatalog/share/hcatalog/hcatalog-core.jar",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
    /**********************************************WEBHCAT***************************************/
      {
        "id": "puppet var",
        "name": "webhcatserver_host",
        "displayName": "WebHCat Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run WebHCat Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      {
        "id": "puppet var",
        "name": "hcat_log_dir",
        "displayName": "WebHCat Log Dir",
        "description": "Directory for WebHCat log files",
        "defaultValue": "/var/log/webhcat",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "WEBHCAT",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hcat_pid_dir",
        "displayName": "WebHCat PID Dir",
        "description": "Directory in which the PID files for WebHCat processes will be created",
        "defaultValue": "/var/run/webhcat",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "WEBHCAT",
        "category": "Advanced"
      },
    /**********************************************OOZIE***************************************/
      {
        "id": "puppet var",
        "name": "oozieserver_host",
        "displayName": "Oozie Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run Oozie Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 0
      },
      // for existing Oracle
      {
        "id": "puppet var",
        "name": "oozie_existing_oracle_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "Oracle",
        "description": "Using an existing Oracle database for Oozie Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isObserved": true,
        "isReconfigurable": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      // for new MySQL
      {
        "id": "puppet var",
        "name": "oozie_ambari_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "MySQL will be installed by Ambari",
        "displayType": "masterHost",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      // for current derby
      {
        "id": "puppet var",
        "name": "oozie_derby_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "Derby",
        "description": "Using current Derby database for Oozie Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isReconfigurable": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      // for existing MySQL oozie
      {
        "id": "puppet var",
        "name": "oozie_existing_mysql_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "Using an existing MySQL database for Oozie Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isReconfigurable": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "oozie_database",
        "displayName": "Oozie Database",
        "value": "",
        "defaultValue": "New Derby Database",
        "options": [
          {
            displayName: 'New Derby Database',
            foreignKeys: ['oozie_derby_database']
          },
          /*{
            displayName: 'New MySQL Database',
            foreignKeys: ['oozie_ambari_database', 'oozie_ambari_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          },*/
          {
            displayName: 'Existing MySQL Database',
            foreignKeys: ['oozie_existing_mysql_database', 'oozie_existing_mysql_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          },
          {
            displayName: 'Existing Oracle Database',
            foreignKeys: ['oozie_existing_oracle_database', 'oozie_existing_oracle_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          }
        ],
        "description": "Current Derby Database will be installed by Ambari",
        "displayType": "radio button",
        "isReconfigurable": false,
        "isOverridable": false,
        "radioName": "oozie-database",
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "oozie_data_dir",
        "displayName": "Oozie Data Dir",
        "description": "Data directory in which the Oozie DB exists",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/oozie/data",
        "isReconfigurable": true,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isRequired": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 7
      },
      {
        "id": "puppet var",
        "name": "oozie_hostname",
        "defaultValue": "",
        "displayName": "Database Host",
        "description": "The host where the Oozie database is located",
        "isReconfigurable": true,
        "isOverridable": false,
        "displayType": "host",
        "isVisible": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "oozie_existing_mysql_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "isOverridable": false,
        "displayType": "host",
        "isVisible": false,
        "isObserved": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_existing_oracle_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "isOverridable": false,
        "displayType": "host",
        "isVisible": false,
        "isObserved": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_ambari_host",
        "value": "",
        "defaultValue": "",
        "displayName": "Database Host",
        "description": "Host on which the database will be created by Ambari",
        "isReconfigurable": false,
        "isOverridable": false,
        "displayType": "masterHost",
        "isVisible": false,
        "isRequiredByAgent": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_log_dir",
        "displayName": "Oozie Log Dir",
        "description": "Directory for oozie logs",
        "defaultValue": "/var/log/oozie",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "oozie_pid_dir",
        "displayName": "Oozie PID Dir",
        "description": "Directory in which the pid files for oozie processes will be created",
        "defaultValue": "/var/run/oozie",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Advanced"
      },
    /**********************************************NAGIOS***************************************/
      {
        "id": "puppet var",
        "name": "nagios_web_login",
        "displayName": "Nagios Admin username",
        "description": "Nagios Web UI Admin username",
        "defaultValue": "nagiosadmin",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "NAGIOS",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "nagios_web_password",
        "displayName": "Nagios Admin password",
        "description": "Nagios Web UI Admin password",
        "defaultValue": "",
        "isReconfigurable": true,
        "displayType": "password",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "NAGIOS",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "nagios_contact",
        "displayName": "Hadoop Admin email",
        "description": "Hadoop Administrator email for alert notification",
        "defaultValue": "",
        "displayType": "email",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "NAGIOS",
        "index": 2
      },
    /**********************************************ZOOKEEPER***************************************/
      {
        "id": "puppet var",
        "name": "zookeeperserver_hosts",
        "displayName": "ZooKeeper Server hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run ZooKeeper Server",
        "displayType": "masterHosts",
        "isVisible": true,
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isRequired": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "zk_data_dir",
        "displayName": "ZooKeeper directory",
        "description": "Data directory for ZooKeeper",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/zookeeper",
        "isReconfigurable": true,
        "displayType": "directory",
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "tickTime",
        "displayName": "Length of single Tick",
        "description": "The length of a single tick in milliseconds, which is the basic time unit used by ZooKeeper",
        "defaultValue": "2000",
        "displayType": "int",
        "unit": "ms",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "initLimit",
        "displayName": "Ticks to allow for sync at Init",
        "description": "Amount of time, in ticks to allow followers to connect and sync to a leader",
        "defaultValue": "10",
        "displayType": "int",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "syncLimit",
        "displayName": "Ticks to allow for sync at Runtime",
        "description": "Amount of time, in ticks to allow followers to connect",
        "defaultValue": "5",
        "displayType": "int",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 4
      },
      {
        "id": "puppet var",
        "name": "clientPort",
        "displayName": "Port for running ZK Server",
        "description": "Port for running ZooKeeper server",
        "defaultValue": "2181",
        "displayType": "int",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 5
      },
      {
        "id": "puppet var",
        "name": "zk_log_dir",
        "displayName": "ZooKeeper Log Dir",
        "description": "Directory for ZooKeeper log files",
        "defaultValue": "/var/log/zookeeper",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "Advanced",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "zk_pid_dir",
        "displayName": "ZooKeeper PID Dir",
        "description": "Directory in which the pid files for zookeeper processes will be created",
        "defaultValue": "/var/run/zookeeper",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "Advanced",
        "index": 1
      },
    /**********************************************HUE***************************************/
      {
        "id": "puppet var",
        "name": "hueserver_host",
        "displayName": "Hue Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run Hue Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HUE",
        "category": "Hue Server"
      },
      {
        "id": "puppet var",
        "name": "hue_log_dir",
        "displayName": "HUE Log Dir",
        "description": "Directory for HUE logs",
        "defaultValue": "/var/log/hue",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HUE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hue_pid_dir",
        "displayName": "HUE Pid Dir",
        "description": "Directory in which the pid files for HUE processes will be created",
        "defaultValue": "/var/run/hue",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HUE",
        "category": "Advanced"
      },
    /**********************************************GANGLIA***************************************/
      {
        "id": "puppet var",
        "name": "ganglia_conf_dir",
        "displayName": "Ganglia conf directory",
        "description": "",
        "defaultValue": "/etc/ganglia/hdp",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "isRequiredByAgent": false,
        "serviceName": "GANGLIA",
        "category": "Advanced"
      },
    /**********************************************FALCON***************************************/
      {
        "id": "puppet var",
        "name": "falconserver_host",
        "displayName": "Falcon Server",
        "description": "The host that has been assigned to run Falcon Server",
        "defaultValue": "falcon",
        "displayType": "masterHost",
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "FALCON",
        "category": "Falcon"
      },
      {
        "id": "puppet var",
        "name": "falcon_port",
        "displayName": "Falcon server port",
        "description": "Port the Falcon Server listens on",
        "defaultValue": "15000",
        "isReconfigurable": true,
        "displayType": "int",
        "isVisible": true,
        "isRequiredByAgent": true,
        "serviceName": "FALCON",
        "category": "Falcon"
      },
      {
        "id": "puppet var",
        "name": "falcon_local_dir",
        "displayName": "Falcon data directory",
        "description": "Directory where Falcon data, such as activemq data, is stored",
        "defaultValue": "/hadoop/falcon",
        "isReconfigurable": true,
        "displayType": "directory",
        "isVisible": true,
        "isRequiredByAgent": true,
        "serviceName": "FALCON",
        "category": "Falcon"
      },
      {
        "id": "puppet var",
        "name": "falcon_store_uri",
        "displayName": "Falcon store URI",
        "description": "Directory where entity definitions are stored",
        "defaultValue": "file:///hadoop/falcon/store",
        "isReconfigurable": true,
        "displayType": "string",
        "isVisible": true,
        "isRequiredByAgent": true,
        "serviceName": "FALCON",
        "category": "Falcon"
      },
      {
        "id": "puppet var",
        "name": "falcon_log_dir",
        "displayName": "Falcon Log Dir",
        "description": "Directory for Falcon logs",
        "defaultValue": "/var/log/falcon",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": true,
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "falcon_pid_dir",
        "displayName": "Falcon PID Dir",
        "description": "Directory in which the pid files for Falcon processes will be created",
        "defaultValue": "/var/run/falcon",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": true,
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "falcon.embeddedmq",
        "displayName": "falcon.embeddedmq",
        "description": "Whether embeddedmq is enabled or not.",
        "defaultValue": "true",
        "displayType": "string",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": true,
        "isReconfigurable": true,
        "serviceName": "FALCON",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "falcon.embeddedmq.data",
        "displayName": "falcon.embeddedmq.data",
        "description": "Directory in which embeddedmq data is stored.",
        "defaultValue": "/hadoop/falcon/embeddedmq/data",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": true,
        "isReconfigurable": true,
        "serviceName": "FALCON",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "falcon.emeddedmq.port",
        "displayName": "falcon.emeddedmq.port",
        "description": "Port that embeddedmq will listen on.",
        "defaultValue": "61616",
        "displayType": "string",
        "isOverridable": false,
        "isVisible": true,
        "isRequiredByAgent": true,
        "isReconfigurable": true,
        "serviceName": "FALCON",
        "category": "Advanced"
      },
    /**********************************************STORM***************************************/
      {
        "id": "puppet var",
        "name": "storm_log_dir",
        "displayName": "storm_log_dir",
        "description": "Storm log directory",
        "defaultValue": "/var/log/storm",
        "displayType": "directory",
        "isReconfigurable": true,
        "isVisible": true,
        "isRequiredByAgent": true,
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "puppet var",
        "name": "storm_pid_dir",
        "displayName": "storm_pid_dir",
        "description": "Storm PID directory",
        "defaultValue": "/var/run/storm",
        "displayType": "directory",
        "isReconfigurable": true,
        "isVisible": true,
        "isRequiredByAgent": true,
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "puppet var",
        "name": "stormuiserver_host",
        "displayName": "Storm UI Server host",
        "description": "The host that has been assigned to run Storm UI Server",
        "defaultValue": "",
        "displayType": "masterHost",
        "isReconfigurable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "STORM",
        "category": "StormUIServer",
        "filename": "global.xml"
      },
      {
        "id": "puppet var",
        "name": "drpcserver_host",
        "displayName": "DRPC Server host",
        "description": "The host that has been assigned to run DRPC Server",
        "defaultValue": "",
        "displayType": "masterHost",
        "isReconfigurable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "STORM",
        "category": "DRPCServer",
        "filename": "global.xml"
      },
      {
        "id": "puppet var",
        "name": "supervisor_hosts",
        "displayName": "Supervisor hosts",
        "description": "The host that has been assigned to run Supervisor",
        "defaultValue": "",
        "displayType": "slaveHosts",
        "isReconfigurable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "STORM",
        "isOverridable": false,
        "category": "Supervisor",
        "filename": "global.xml"
      },
      {
        "id": "puppet var",
        "name": "storm_rest_api_host",
        "displayName": "Storm REST API host",
        "description": "The host that has been assigned to run Storm REST API Server",
        "defaultValue": "",
        "displayType": "masterHost",
        "isReconfigurable": false,
        "isVisible": true,
        "isRequiredByAgent": false,
        "serviceName": "STORM",
        "isOverridable": false,
        "category": "StormRestApi",
        "filename": "global.xml"
      },
    /**********************************************MISC***************************************/
      {
        "id": "puppet var",
        "name": "hbase_conf_dir",
        "displayName": "HBase conf dir",
        "description": "",
        "defaultValue": "/etc/hbase",
        "isRequired": false,
        "displayType": "directory",
        "isVisible": false,
        "isRequiredByAgent": false,
        "serviceName": "MISC",
        "category": "General",
        "belongsToService":[]
      },
      {
        "id": "puppet var",
        "name": "proxyuser_group",
        "displayName": "Proxy group for Hive, WebHCat, Oozie and Falcon",
        "description": "",
        "defaultValue": "users",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "filename": "core-site.xml",
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HIVE","WEBHCAT","OOZIE","FALCON"],
        "index": 18
      },
      {
        "id": "puppet var",
        "name": "ganglia_runtime_dir",
        "displayName": "Ganglia runtime directory",
        "description": "",
        "defaultValue": "/var/run/ganglia/hdp",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "MISC",
        "category": "General",
        "belongsToService":[]
      },
      {
        "id": "puppet var",
        "name": "hdfs_user",
        "displayName": "HDFS User",
        "description": "User to run HDFS as",
        "defaultValue": "hdfs",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HDFS"],
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "mapred_user",
        "displayName": "MapReduce User",
        "description": "User to run MapReduce as",
        "defaultValue": "mapred",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["MAPREDUCE2"],
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "yarn_user",
        "displayName": "YARN User",
        "description": "User to run YARN as",
        "defaultValue": "yarn",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["YARN"],
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hbase_user",
        "displayName": "HBase User",
        "description": "User to run HBase as",
        "defaultValue": "hbase",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HBASE"],
        "index": 4
      },
      {
        "id": "puppet var",
        "name": "hive_user",
        "displayName": "Hive User",
        "description": "User to run Hive as",
        "defaultValue": "hive",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HIVE"],
        "index": 5
      },
      {
        "id": "puppet var",
        "name": "hcat_user",
        "displayName": "HCat User",
        "description": "User to run HCatalog as",
        "defaultValue": "hcat",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HCATALOG"],
        "index": 6
      },
      {
        "id": "puppet var",
        "name": "webhcat_user",
        "displayName": "WebHCat User",
        "description": "User to run WebHCat as",
        "defaultValue": "hcat",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["WEBHCAT"],
        "index": 7
      },
      {
        "id": "puppet var",
        "name": "oozie_user",
        "displayName": "Oozie User",
        "description": "User to run Oozie as",
        "defaultValue": "oozie",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["OOZIE"],
        "index": 8
      },
      {
        "id": "puppet var",
        "name": "falcon_user",
        "displayName": "Falcon User",
        "description": "User to run Falcon as",
        "defaultValue": "falcon",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName":"MISC",
        "category": "Users and Groups",
        "belongsToService":["FALCON"],
        "index": 10
      },
      {
        "id": "puppet var",
        "name": "storm_user",
        "displayName": "Storm User",
        "description": "User to run Storm as",
        "defaultValue": "storm",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName":"MISC",
        "category": "Users and Groups",
        "belongsToService":["STORM"],
        "index": 9
      },
      {
        "id": "puppet var",
        "name": "zk_user",
        "displayName": "ZooKeeper User",
        "description": "User to run ZooKeeper as",
        "defaultValue": "zookeeper",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["ZOOKEEPER"],
        "index": 11
      },
      {
        "id": "puppet var",
        "name": "gmetad_user",
        "displayName": "Ganglia User",
        "description": "The user used to run Ganglia",
        "defaultValue": "nobody",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["GANGLIA"],
        "index": 12
      },
      {
        "id": "puppet var",
        "name": "gmond_user",
        "displayName": "Gmond User",
        "description": "The user used to run gmond for Ganglia",
        "defaultValue": "nobody",
        "isReconfigurable": false,
        "displayType": "advanced",
        "isOverridable": false,
        "isVisible": false,
        "serviceName":"MISC",
        "category": "Users and Groups",
        "belongsToService":[]
      },
      {
        "id": "puppet var",
        "name": "nagios_user",
        "displayName": "Nagios User",
        "description": "User to run Nagios as",
        "defaultValue": "nagios",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName":"MISC",
        "category": "Users and Groups",
        "belongsToService":["NAGIOS"],
        "index": 14
      },
      {
        "id": "puppet var",
        "name": "nagios_group",
        "displayName": "Nagios Group",
        "description": "Nagios Group",
        "defaultValue": "nagios",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName":"MISC",
        "category": "Users and Groups",
        "belongsToService":["NAGIOS"],
        "index": 13
      },
      {
        "id": "puppet var",
        "name": "smokeuser",
        "displayName": "Smoke Test User",
        "description": "The user used to run service smoke tests",
        "defaultValue": "ambari-qa",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": App.supports.customizeSmokeTestUser,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HDFS"],
        "index": 16
      },
      {
        "id": "puppet var",
        "name": "tez_user",
        "displayName": "Tez User",
        "description": "User to run Tez as",
        "defaultValue": "tez",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName":"MISC",
        "category": "Users and Groups",
        "belongsToService":["TEZ"],
        "index": 15
      },
      {
        "id": "puppet var",
        "name": "user_group",
        "displayName": "Hadoop Group",
        "description": "Group that the users specified above belong to",
        "defaultValue": "hadoop",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService":["HDFS"],
        "index": 16
      },
      {
        "id": "puppet var",
        "name": "rrdcached_base_dir",
        "displayName": "Ganglia rrdcached base directory",
        "description": "Default directory for saving the rrd files on ganglia server",
        "defaultValue": "/var/lib/ganglia/rrds",
        "displayType": "directory",
        "isReconfigurable": true,
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "GANGLIA",
        "category": "General",
        "belongsToService":["GANGLIA"]
      }

    ]
  };
  // @todo remove after App Timeline service integration
  if (App.supports.appTimelineServer) {
    module.exports.configProperties.push(
      {
        "id": "puppet var",
        "name": "ats_host",
        "displayName": "App Timeline Server",
        "description": "Application Timeline Server Host",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "YARN",
        "category": "AppTimelineServer",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "apptimelineserver_heapsize",
        "displayName": "AppTimelineServer Java heap size",
        "description": "AppTimelineServer Java heap size",
        "defaultValue": "1024",
        "isOverridable": false,
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "serviceName": "YARN",
        "category": "AppTimelineServer",
        "index": 1
      }
    );
  }
  
});
window.require.register("data/HDP2/ha_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports =
  {
    "haConfig": {
      serviceName: 'MISC',
      displayName: 'MISC',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'HDFS', displayName: 'HDFS'}),
        App.ServiceConfigCategory.create({ name: 'HBASE', displayName: 'HBase'})
      ],
      sites: ['global', 'core-site', 'hdfs-site','hbase-site'],
      configs: [
      /**********************************************HDFS***************************************/
        {
          "id": "site property",
          "name": "dfs.journalnode.edits.dir",
          "displayName": "dfs.journalnode.edits.dir",
          "description": "The Directory where the JournalNode will store its local state.",
          "isReconfigurable": true,
          "defaultValue": "/hadoop/hdfs/journal",
          "value": "/hadoop/hdfs/journal",
          "defaultDirectory": "/hadoop/hdfs/journal",
          "displayType": "directory",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "fs.defaultFS",
          "displayName": "fs.defaultFS",
          "description": "The default path prefix used by the Hadoop FS client when none is given.",
          "defaultValue": "hdfs://haCluster",
          "isReconfigurable": false,
          "value": "hdfs://haCluster",
          "category": "HDFS",
          "filename": "core-site",
          serviceName: 'MISC'
        },
        {
          "id": "site property",
          "name": "ha.zookeeper.quorum",
          "displayName": "ha.zookeeper.quorum",
          "isReconfigurable": false,
          "description": "This lists the host-port pairs running the ZooKeeper service.",
          "defaultValue": "zk1.example.com:2181,zk2.example.com:2181,zk3.example.com:2181",
          "value": "zk1.example.com:2181,zk2.example.com:2181,zk3.example.com:2181",
          "category": "HDFS",
          "filename": "core-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.nameservices",
          "displayName": "dfs.nameservices",
          "description": "Comma-separated list of nameservices.",
          "isReconfigurable": false,
          "defaultValue": "haCluster",
          "value": "haCluster",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.ha.namenodes.${dfs.nameservices}",
          "displayName": "dfs.ha.namenodes.${dfs.nameservices}",
          "description": "The prefix for a given nameservice, contains a comma-separated list of namenodes for a given nameservice.",
          "isReconfigurable": false,
          "defaultValue": "nn1,nn2",
          "value": "nn1,nn2",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.rpc-address.${dfs.nameservices}.nn1",
          "displayName": "dfs.namenode.rpc-address.${dfs.nameservices}.nn1",
          "description": "RPC address that handles all clients requests for nn1.",
          "isReconfigurable": false,
          "defaultValue": "0.0.0.0:8020",
          "value": "0.0.0.0:8020",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.rpc-address.${dfs.nameservices}.nn2",
          "displayName": "dfs.namenode.rpc-address.${dfs.nameservices}.nn2",
          "description": "RPC address that handles all clients requests for nn2.",
          "isReconfigurable": false,
          "defaultValue": "0.0.0.0:8020",
          "value": "0.0.0.0:8020",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.http-address.${dfs.nameservices}.nn1",
          "displayName": "dfs.namenode.http-address.${dfs.nameservices}.nn1",
          "description": "The fully-qualified HTTP address for nn1 NameNode.",
          "isReconfigurable": false,
          "defaultValue": "0.0.0.0:50070",
          "value": "0.0.0.0:50070",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.http-address.${dfs.nameservices}.nn2",
          "displayName": "dfs.namenode.http-address.${dfs.nameservices}.nn2",
          "description": "The fully-qualified HTTP address for nn2 NameNode.",
          "isReconfigurable": false,
          "defaultValue": "0.0.0.0:50070",
          "value": "0.0.0.0:50070",
          "category": "HDFS",
          "filename": "hdfs-site",
          serviceName: 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.https-address.${dfs.nameservices}.nn1",
          "displayName": "dfs.namenode.https-address.${dfs.nameservices}.nn1",
          "description": "The fully-qualified HTTP address for nn1 NameNode.",
          "isReconfigurable": false,
          "defaultValue": "0.0.0.0:50470",
          "value": "0.0.0.0:50470",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.https-address.${dfs.nameservices}.nn2",
          "displayName": "dfs.namenode.https-address.${dfs.nameservices}.nn2",
          "description": "The fully-qualified HTTP address for nn2 NameNode.",
          "isReconfigurable": false,
          "defaultValue": "0.0.0.0:50470",
          "value": "0.0.0.0:50470",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.client.failover.proxy.provider.${dfs.nameservices}",
          "displayName": "dfs.client.failover.proxy.provider.${dfs.nameservices}",
          "description": "The Java class that HDFS clients use to contact the Active NameNode.",
          "defaultValue": "org.apache.hadoop.hdfs.server.namenode.ha.ConfiguredFailoverProxyProvider",
          "isReconfigurable": false,
          "value": "org.apache.hadoop.hdfs.server.namenode.ha.ConfiguredFailoverProxyProvider",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.namenode.shared.edits.dir",
          "displayName": "dfs.namenode.shared.edits.dir",
          "description": " The URI which identifies the group of JNs where the NameNodes will write/read edits.",
          "isReconfigurable": false,
          "defaultValue": "qjournal://node1.example.com:8485;node2.example.com:8485;node3.example.com:8485/mycluster",
          "value": "qjournal://node1.example.com:8485;node2.example.com:8485;node3.example.com:8485/mycluster",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.ha.fencing.methods",
          "displayName": "dfs.ha.fencing.methods",
          "description": "A list of scripts or Java classes which will be used to fence the Active NameNode during a failover.",
          "isReconfigurable": false,
          "defaultValue": "shell(/bin/true)",
          "value": "shell(/bin/true)",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "dfs.ha.automatic-failover.enabled",
          "displayName": "dfs.ha.automatic-failover.enabled",
          "description": "Enable Automatic failover.",
          "isReconfigurable": false,
          "defaultValue": true,
          "value": true,
          "displayType": "checkbox",
          "category": "HDFS",
          "filename": "hdfs-site",
          "serviceName": 'MISC'
        },
        {
          "id": "site property",
          "name": "hbase.rootdir",
          "displayName": "hbase.rootdir",
          "description": "The directory shared by region servers and into which HBase persists.",
          "isReconfigurable": false,
          "defaultValue": "/hadoop/hdfs/journal",
          "value": "/hadoop/hdfs/journal",
          "category": "HBASE",
          "filename": "hbase-site",
          "serviceName": 'MISC'
        }
      ]
    }
  };
  
});
window.require.register("data/HDP2/secure_configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  require('models/service_config');
  App.SecureConfigProperties = Ember.ArrayProxy.extend({
    content: require('data/HDP2/secure_properties').configProperties
  });

  var configProperties = App.SecureConfigProperties.create();

  module.exports = [
    {
      serviceName: 'GENERAL',
      displayName: 'General',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'KERBEROS', displayName: 'Kerberos'}),
        App.ServiceConfigCategory.create({ name: 'AMBARI', displayName: 'Ambari'})
      ],
      sites: ['global'],
      configs: configProperties.filterProperty('serviceName', 'GENERAL')
    },
    {
      serviceName: 'HDFS',
      displayName: 'HDFS',
      filename: 'hdfs-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General', exclude:'HA'}),
        App.ServiceConfigCategory.create({ name: 'NameNode', displayName: 'NameNode'}),
        App.ServiceConfigCategory.create({ name: 'SNameNode', displayName: 'Secondary NameNode',exclude:'HA'}),
        App.ServiceConfigCategory.create({ name: 'JournalNode', displayName: 'JournalNode',exclude:'non-HA'}),
        App.ServiceConfigCategory.create({ name: 'DataNode', displayName: 'DataNode'})
      ],
      sites: ['core-site', 'hdfs-site'],
      configs: configProperties.filterProperty('serviceName', 'HDFS')
    },
    {
      serviceName: 'MAPREDUCE2',
      displayName: 'MapReduce2',
      filename: 'mapred-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'JobHistoryServer', displayName: 'History Server'})
      ],
      sites: ['mapred-site'],
      configs: configProperties.filterProperty('serviceName', 'MAPREDUCE2')
    },
    {
      serviceName: 'YARN',
      displayName: 'YARN',
      filename: 'yarn-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'ResourceManager', displayName : 'ResourceManager'}),
        App.ServiceConfigCategory.create({ name: 'NodeManager', displayName : 'NodeManager'})
        ],
      sites: ['yarn-site'],
      configs: configProperties.filterProperty('serviceName', 'YARN')
    },
    {
      serviceName: 'HIVE',
      displayName: 'Hive',
      filename: 'hive-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Hive Metastore', displayName: 'Hive Metastore and Hive Server 2'})
      ],
      sites: ['hive-site'],
      configs: configProperties.filterProperty('serviceName', 'HIVE')
    },
    {
      serviceName: 'WEBHCAT',
      displayName: 'WebHCat',
      filename: 'webhcat-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'WebHCat Server', displayName : 'WebHCat Server'})
      ],
      sites: ['webhcat-site'],
      configs: configProperties.filterProperty('serviceName', 'WEBHCAT')
    },
    {
      serviceName: 'HBASE',
      displayName: 'HBase',
      filename: 'hbase-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'HBase Master', displayName: 'HBase Master'}),
        App.ServiceConfigCategory.create({ name: 'RegionServer', displayName: 'RegionServer'})
      ],
      sites: ['hbase-site'],
      configs: configProperties.filterProperty('serviceName', 'HBASE')
    },
    {
      serviceName: 'ZOOKEEPER',
      displayName: 'ZooKeeper',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'ZooKeeper Server', displayName: 'ZooKeeper Server'})
      ],
      configs: configProperties.filterProperty('serviceName', 'ZOOKEEPER')

    },
    {
      serviceName: 'OOZIE',
      displayName: 'Oozie',
      filename: 'oozie-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Oozie Server', displayName:  'Oozie Server'})
      ],
      sites: ['oozie-site'],
      configs: configProperties.filterProperty('serviceName', 'OOZIE')
    },
    {
      serviceName: 'NAGIOS',
      displayName: 'Nagios',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Nagios Server', displayName:  'Nagios Server'})
      ],
      configs: configProperties.filterProperty('serviceName', 'NAGIOS')
    },
    {
      serviceName: 'STORM',
      displayName: 'Storm',
      filename: 'storm-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Storm Topology', displayName:  'Storm Topology'})
      ],
      sites: ['storm-site'],
      configs: configProperties.filterProperty('serviceName', 'STORM')
    },
    {
      serviceName: 'FALCON',
      displayName: 'Falcon',
      filename: 'falcon-startup.properties',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Falcon Server', displayName:  'Falcon Server startup properties'})
      ],
      sites: ['falcon-startup.properties'],
      configs: configProperties.filterProperty('serviceName', 'FALCON')
    }


  ];
  
});
window.require.register("data/HDP2/secure_mapping", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports = [
    {
      "name": "hadoop.security.authentication",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "core-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "hadoop.security.authorization",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "core-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "hadoop.security.auth_to_local",
      "templateName": ["resourcemanager_primary_name", "kerberos_domain", "yarn_user", "nodemanager_primary_name", "namenode_primary_name", "hdfs_user", "datanode_primary_name", "hbase_master_primary_name", "hbase_user","hbase_regionserver_primary_name","oozie_primary_name","oozie_user","jobhistory_primary_name","mapred_user","journalnode_principal_name","falcon_primary_name","falcon_user"],
      "foreignKey": null,
      "value": "RULE:[2:$1@$0](<templateName[0]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[3]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[4]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[6]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[11]>/\nRULE:[2:$1@$0](<templateName[12]>@.*<templateName[1]>)s/.*/<templateName[13]>/\nRULE:[2:$1@$0](<templateName[14]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[15]>@.*<templateName[1]>)s/.*/<templateName[16]>/\nDEFAULT",
      "filename": "core-site.xml",
      "serviceName": "HDFS",
      "dependedServiceName": [{name: "HBASE", replace: "\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/"},{name: "OOZIE",replace: "\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[11]>/"},{name: "MAPREDUCE2",replace: "\nRULE:[2:$1@$0](<templateName[12]>@.*<templateName[1]>)s/.*/<templateName[13]>/"}]
    },
    {
      "name": "dfs.namenode.kerberos.principal",
      "templateName": ["namenode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.namenode.keytab.file",
      "templateName": ["namenode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.secondary.namenode.kerberos.principal",
      "templateName": ["snamenode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml"
    },
    {
      "name": "dfs.secondary.namenode.keytab.file",
      "templateName": ["snamenode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.web.authentication.kerberos.principal",
      "templateName": ["hadoop_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.web.authentication.kerberos.keytab",
      "templateName": ["hadoop_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.kerberos.principal",
      "templateName": ["datanode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.keytab.file",
      "templateName": ["datanode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.namenode.kerberos.internal.spnego.principal",
      "templateName": [],
      "foreignKey": null,
      "value": "${dfs.web.authentication.kerberos.principal}",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.secondary.namenode.kerberos.internal.spnego.principal",
      "templateName": [],
      "foreignKey": null,
      "value": "${dfs.web.authentication.kerberos.principal}",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.journalnode.kerberos.principal",
      "templateName": ["journalnode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.journalnode.kerberos.internal.spnego.principal",
      "templateName": ["hadoop_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.journalnode.keytab.file",
      "templateName": ["journalnode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.address",
      "templateName": ["dfs_datanode_address"],
      "foreignKey": null,
      "value": "0.0.0.0:<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.http.address",
      "templateName": ["dfs_datanode_http_address"],
      "foreignKey": null,
      "value": "0.0.0.0:<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "mapreduce.jobhistory.principal",
      "templateName": ["jobhistory_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE2"
    },
    {
      "name": "mapreduce.jobhistory.keytab",
      "templateName": ["jobhistory_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE2"
    },
    {
      "name": "mapreduce.jobhistory.webapp.spnego-principal",
      "templateName": ["jobhistory_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE2"
    },
    {
      "name": "mapreduce.jobhistory.webapp.spnego-keytab-file",
      "templateName": ["jobhistory_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE2"
    },
    {
      "name": "yarn.resourcemanager.principal",
      "templateName": ["resourcemanager_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.resourcemanager.keytab",
      "templateName": ["resourcemanager_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.nodemanager.principal",
      "templateName": ["nodemanager_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.nodemanager.keytab",
      "templateName": ["nodemanager_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.resourcemanager.webapp.spnego-principal",
      "templateName": ["resourcemanager_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.resourcemanager.webapp.spnego-keytab-file",
      "templateName": ["resourcemanager_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.nodemanager.webapp.spnego-principal",
      "templateName": ["nodemanager_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.nodemanager.webapp.spnego-keytab-file",
      "templateName": ["nodemanager_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "yarn.nodemanager.container-executor.class",
      "templateName": ["yarn_nodemanager_container-executor_class"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "yarn-site.xml",
      "serviceName": "YARN"
    },
    {
      "name": "hbase.master.kerberos.principal",
      "templateName": ["hbase_master_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.master.keytab.file",
      "templateName": ["hbase_master_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.regionserver.kerberos.principal",
      "templateName": ["hbase_regionserver_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.regionserver.keytab.file",
      "templateName": ["hbase_regionserver_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hive.metastore.sasl.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.security.authorization.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.server2.authentication",
      "templateName": [],
      "foreignKey": null,
      "value": "KERBEROS",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.metastore.kerberos.principal",
      "templateName": ["hive_metastore_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.metastore.kerberos.keytab.file",
      "templateName": ["hive_metastore_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.server2.authentication.kerberos.principal",
      "templateName": ["hive_metastore_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.server2.authentication.kerberos.keytab",
      "templateName": ["hive_metastore_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "oozie.service.AuthorizationService.authorization.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.HadoopAccessorService.kerberos.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "local.realm",
      "templateName": ["kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.HadoopAccessorService.keytab.file",
      "templateName": ["oozie_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.HadoopAccessorService.kerberos.principal",
      "templateName": ["oozie_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.type",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.kerberos.principal",
      "templateName": ["oozie_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.kerberos.keytab",
      "templateName": ["oozie_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.kerberos.name.rules",
      "templateName": ["resourcemanager_primary_name", "kerberos_domain", "yarn_user", "nodemanager_primary_name", "namenode_primary_name", "hdfs_user", "datanode_primary_name", "hbase_master_primary_name", "hbase_user","hbase_regionserver_primary_name"],
      "foreignKey": null,
      "value": "RULE:[2:$1@$0](<templateName[0]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[3]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[4]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[6]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nDEFAULT",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE",
      "dependedServiceName": [{name: "HBASE", replace: "\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/"}]
    },
    {
      "name": "templeton.kerberos.principal",
      "templateName": ["webHCat_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "templeton.kerberos.keytab",
      "templateName": ["webhcat_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "templeton.kerberos.secret",
      "templateName": [""],
      "foreignKey": null,
      "value": "secret",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "templeton.hive.properties",
      "templateName": ["hivemetastore_host","hive_metastore_principal_name","kerberos_domain"],
      "foreignKey": null,
      "value": "hive.metastore.local=false,hive.metastore.uris=thrift://<templateName[0]>:9083,hive." +
        "metastore.sasl.enabled=true,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse,hive.exec.mode.local.auto=false,hive.metastore.kerberos.principal=<templateName[1]>@<templateName[2]>",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "hbase.coprocessor.master.classes",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.security.access.AccessController",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.coprocessor.region.classes",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.security.token.TokenProvider,org.apache.hadoop.hbase.security.access.SecureBulkLoadEndpoint,org.apache.hadoop.hbase.security.access.AccessController",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.security.authentication",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.security.authorization",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.coprocessor.region.classes",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.security.token.TokenProvider,org.apache.hadoop.hbase.security.access.SecureBulkLoadEndpoint,org.apache.hadoop.hbase.security.access.AccessController",
      "filename": "hbase-site.xml"
    },
    {
      "name": "hbase.bulkload.staging.dir",
      "templateName": [],
      "foreignKey": null,
      "value": "/apps/hbase/staging",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "zookeeper.znode.parent",
      "templateName": [],
      "foreignKey": null,
      "value": "/hbase-secure",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },

    /***************************************FALCON***********************************************/
    {
      "name": "*.falcon.authentication.type",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    {
      "name": "*.falcon.http.authentication.type",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    {
      "name": "*.falcon.service.authentication.kerberos.principal",
      "templateName": ["falcon_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    {
      "name": "*.falcon.service.authentication.kerberos.keytab",
      "templateName": ["falcon_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    {
      "name": "*.falcon.http.authentication.kerberos.principal",
      "templateName": ["falcon_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    {
      "name": "*.falcon.http.authentication.kerberos.keytab",
      "templateName": ["falcon_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    {
      "name": "*.dfs.namenode.kerberos.principal",
      "templateName": ["namenode_principal_name_falcon", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "falcon-startup.properties.xml",
      "serviceName": "FALCON"
    },
    /******************************************************************************************/
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["hive_metastore_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["hivemetastore_host"],
      "foreignKey": ["hive_metastore_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["oozie_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["oozieserver_host"],
      "foreignKey": ["oozie_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["webHCat_http_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["webhcat_server"],
      "foreignKey": ["webHCat_http_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "WEBHCAT"
    }
  ];

  
});
window.require.register("data/HDP2/secure_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  module.exports =
  {
    "configProperties": [
      {
        "id": "puppet var",
        "name": "security_enabled",
        "displayName": "Enable security",
        "value": "",
        "defaultValue": 'true',
        "description": "Enable kerberos security for the cluster",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "kerberos_install_type",
        "displayName": "Type of security",
        "value": "",
        "defaultValue": "MANUALLY_SET_KERBEROS",
        "description": "Type of kerberos security for the cluster",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "keytab_path",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs",
        "description": "Type of kerberos security for the cluster",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "kerberos_domain",
        "displayName": "Realm name",
        "value": "",
        "defaultValue": "EXAMPLE.COM",
        "description": "Realm name to be used for all principal names",
        "displayType": "advanced",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "kinit_path_local",
        "displayName": "Kerberos tool path",
        "value": "",
        "defaultValue": "/usr/bin",
        "description": "Directoy path to installed kerberos tools like kinit, kdestroy etc. This can have multiple comma delimited paths",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "smokeuser_principal_name",
        "displayName": "Smoke test user principal",
        "value": "",
        "defaultValue": "ambari-qa",
        "description": "This is the principal name for Smoke test user",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "smokeuser_keytab",
        "displayName": "Path to smoke test user keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/smokeuser.headless.keytab",
        "description": "Path to keytab file for smoke test user",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hdfs_principal_name",
        "displayName": "HDFS user principal",
        "value": "",
        "defaultValue": "hdfs",
        "description": "This is the principal name for HDFS user",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hdfs_user_keytab",
        "displayName": "Path to HDFS user keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hdfs.headless.keytab",
        "description": "Path to keytab file for HDFS user",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hbase_principal_name",
        "displayName": "HBase user principal",
        "value": "",
        "defaultValue": "hbase",
        "description": "This is the principal name for HBase user",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hbase_user_keytab",
        "displayName": "Path to HBase user keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hbase.headless.keytab",
        "description": "Path to keytab file for Hbase user",
        "displayType": "directory",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },

    /**********************************************HDFS***************************************/
      {
        "id": "puppet var",
        "name": "namenode_host",
        "displayName": "NameNode hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that has been assigned to run NameNode",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode"
      },
      {
        "id": "puppet var",
        "name": "namenode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nn/_HOST",
        "description": "Principal name for NameNode. _HOST will get automatically replaced with actual hostname at an instance of NameNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "NameNode",
        "components": ["NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "namenode_keytab",
        "displayName": "Path to Keytab File",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nn.service.keytab",
        "description": "Path to NameNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "NameNode",
        "components": ["NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "snamenode_host",
        "displayName": "SNameNode host",
        "value": "",
        "defaultValue": "localhost",
        "description": "The host that has been assigned to run SecondaryNameNode",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "SNameNode"
      },
      {
        "id": "puppet var",
        "name": "snamenode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nn/_HOST",
        "description": "Principal name for SNameNode. _HOST will get automatically replaced with actual hostname at an instance of SNameNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "components": ["SECONDARY_NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "snamenode_keytab",
        "displayName": "Path to Keytab File",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nn.service.keytab",
        "description": "Path to SNameNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "components": ["SECONDARY_NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "journalnode_hosts",
        "displayName": "JournalNode hosts",
        "value": "",
        "defaultValue": "localhost",
        "description": "The hosts that have been assigned to run JournalNodes",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "JournalNode"
      },
      {
        "id": "puppet var",
        "name": "journalnode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "jn/_HOST",
        "description": "Principal name for JournalNode. _HOST will get automatically replaced with actual hostname at every instance of JournalNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "JournalNode",
        "component": "JOURNALNODE"
      },
      {
        "id": "puppet var",
        "name": "journalnode_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/jn.service.keytab",
        "description": "Path to JournalNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "JournalNode",
        "component": "JOURNALNODE"
      },
      {
        "id": "puppet var",
        "name": "datanode_hosts", //not in the schema. For UI purpose
        "displayName": "DataNode hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run DataNode",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "DataNode"
      },
      {
        "id": "puppet var",
        "name": "dfs_datanode_address",
        "displayName": "Datanode address",
        "value": "",
        "defaultValue": "1019",
        "description": "Address for DataNode",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode"
      },
      {
        "id": "puppet var",
        "name": "dfs_datanode_http_address",
        "displayName": "Datanode HTTP address",
        "value": "",
        "defaultValue": "1022",
        "description": "Address for DataNode",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode"
      },
      {
        "id": "puppet var",
        "name": "datanode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "dn/_HOST",
        "description": "Principal name for DataNode. _HOST will get automatically replaced with actual hostname at every instance of DataNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode",
        "component": "DATANODE"
      },
      {
        "id": "puppet var",
        "name": "datanode_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/dn.service.keytab",
        "description": "Path to DataNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode",
        "component": "DATANODE"
      },
      {
        "id": "puppet var",
        "name": "hadoop_http_principal_name",
        "displayName": "DFS web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access for HDFS components. _HOST will get automatically replaced with actual hostname at instance of HDFS component",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "General"
      },
      {
        "id": "puppet var",
        "name": "hadoop_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for NameNode and SNameNode",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "General"
      },

    /**********************************************MAPREDUCE2***************************************/
      {
        "id": "puppet var",
        "name": "jobhistoryserver_host",
        "displayName": "History Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run History Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE2",
        "category": "JobHistoryServer"
      },
      {
        "id": "puppet var",
        "name": "jobhistory_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "jhs/_HOST",
        "description": "Principal name for History Server. _HOST will get automatically replaced with actual hostname at an instance of History Server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE2",
        "category": "JobHistoryServer",
        "component": "HISTORYSERVER"
      },
      {
        "id": "puppet var",
        "name": "jobhistory_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/jhs.service.keytab",
        "description": "Path to History Server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE2",
        "category": "JobHistoryServer",
        "component": "HISTORYSERVER"
      },
      {
        "id": "puppet var",
        "name": "jobhistory_http_principal_name",
        "displayName": "Web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access to Job History Server. _HOST will get automatically replaced with actual hostname at an instance of Job History Server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE2",
        "category": "JobHistoryServer"
      },
      {
        "id": "puppet var",
        "name": "jobhistory_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for Job History Server",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE2",
        "category": "JobHistoryServer"
      },

    /**********************************************YARN***************************************/
      {
        "id": "puppet var",
        "name": "resourcemanager_host",
        "displayName": "ResourceManager host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run ResourceManager",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "YARN",
        "category": "ResourceManager"
      },
      {
        "id": "puppet var",
        "name": "resourcemanager_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "rm/_HOST",
        "description": "Principal name for ResourceManager. _HOST will get automatically replaced with actual hostname at an instance of ResourceManager",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "ResourceManager",
        "component": "RESOURCEMANAGER"
      },
      {
        "id": "puppet var",
        "name": "resourcemanager_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/rm.service.keytab",
        "description": "Path to ResourceManager keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "ResourceManager",
        "component": "RESOURCEMANAGER"
      },
      {
        "id": "puppet var",
        "name": "resourcemanager_http_principal_name",
        "displayName": "Web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access to ResourceManager. _HOST will get automatically replaced with actual hostname at an instance of ResourceManager",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "ResourceManager"
      },
      {
        "id": "puppet var",
        "name": "resourcemanager_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for ResourceManager",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "ResourceManager"
      },
      {
        "id": "puppet var",
        "name": "nodemanager_host",
        "displayName": "NodeManager hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that has been assigned to run NodeManager",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "puppet var",
        "name": "nodemanager_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nm/_HOST",
        "description": "Principal name for NodeManager. _HOST will get automatically replaced with actual hostname at all instances of NodeManager",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "NodeManager",
        "component": "NODEMANAGER"
      },
      {
        "id": "puppet var",
        "name": "nodemanager_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nm.service.keytab",
        "description": "Path to NodeManager keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "NodeManager",
        "component": "NODEMANAGER"
      },
      {
        "id": "puppet var",
        "name": "nodemanager_http_principal_name",
        "displayName": "Web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access to NodeManager. _HOST will get automatically replaced with actual hostname at all instances of NodeManager",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "puppet var",
        "name": "nodemanager_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for NodeManager",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "puppet var",
        "name": "yarn_nodemanager_container-executor_class",
        "displayName": "yarn.nodemanager.container-executor.class",
        "value": "",
        "defaultValue": "org.apache.hadoop.yarn.server.nodemanager.LinuxContainerExecutor",
        "description": "Executor(launcher) of the containers",
        "displayType": "advanced",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "YARN",
        "category": "NodeManager"
      },

    /**********************************************WEBHCAT***************************************/
      {
        "id": "puppet var",
        "name": "webhcatserver_host",
        "displayName": "WebHCat Server host",
        "value": "",
        "defaultValue": "localhost",
        "description": "The host that has been assigned to run WebHCat Server",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      {
        "id": "puppet var",
        "name": "webHCat_http_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access for WebHCat",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      {
        "id": "puppet var",
        "name": "webhcat_http_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for WebHCat",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },

    /**********************************************HBASE***************************************/
      {
        "id": "puppet var",
        "name": "hbasemaster_host",
        "displayName": "HBase Master hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run HBase Master",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "HBase Master"
      },
      {
        "id": "puppet var",
        "name": "hbase_master_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "hbase/_HOST",
        "description": "Principal name for HBase master. _HOST will get automatically replaced with actual hostname at an instance of HBase Master",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "components": ["HBASE_MASTER"]
      },
      {
        "id": "puppet var",
        "name": "hbase_master_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hbase.service.keytab",
        "description": "Path to HBase master keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "components": ["HBASE_MASTER"]
      },
      {
        "id": "puppet var",
        "name": "regionserver_hosts",
        "displayName": "RegionServer hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run RegionServer",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "RegionServer"
      },
      {
        "id": "puppet var",
        "name": "hbase_regionserver_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "hbase/_HOST",
        "description": "Principal name for HBase RegionServer. _HOST will get automatically replaced with actual hostname at every instance of RegionServer",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "RegionServer",
        "components": ["HBASE_REGIONSERVER"]
      },
      {
        "id": "puppet var",
        "name": "hbase_regionserver_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hbase.service.keytab",
        "description": "Path to HBase RegionServer keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "RegionServer",
        "components": ["HBASE_REGIONSERVER"]
      },

    /**********************************************HIVE***************************************/
      {
        "id": "puppet var",
        "name": "hive_metastore",
        "displayName": "Hive Metastore host",
        "value": "",
        "defaultValue": "localhost",
        "description": "The host that has been assigned to run Hive Metastore and HiveServer2",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore"
      },
      {
        "id": "puppet var",
        "name": "hive_metastore_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "hive/_HOST",
        "description": "Principal name for Hive Metastore and HiveServer2. _HOST will get automatically replaced with actual hostname at an instance of Hive Metastore and HiveServer2",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "component": "HIVE_SERVER"
      },
      {
        "id": "puppet var",
        "name": "hive_metastore_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hive.service.keytab",
        "description": "Path to Hive Metastore and HiveServer2 keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "component": "HIVE_SERVER"
      },

    /**********************************************OOZIE***************************************/
      {
        "id": "puppet var",
        "name": "oozie_servername",
        "displayName": "Oozie Server host",
        "value": "",
        "defaultValue": "localhost",
        "description": "Oozie server host name",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "oozie/_HOST",
        "description": "Principal name for Oozie server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "component": "OOZIE_SERVER"
      },
      {
        "id": "puppet var",
        "name": "oozie_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/oozie.service.keytab",
        "description": "Path to Oozie server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "component": "OOZIE_SERVER"
      },
      {
        "id": "puppet var",
        "name": "oozie_http_principal_name",
        "displayName": "Web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access to Oozie",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for oozie",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },

    /**********************************************ZOOKEEPER***************************************/
      {
        "id": "puppet var",
        "name": "zookeeperserver_hosts",
        "displayName": "ZooKeeper Server hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run ZooKeeper Server",
        "displayType": "masterHosts",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server"
      },
      {
        "id": "puppet var",
        "name": "zookeeper_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "zookeeper/_HOST",
        "description": "Principal name for ZooKeeper. _HOST will get automatically replaced with actual hostname at every instance of zookeeper server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "component": "ZOOKEEPER_SERVER"
      },
      {
        "id": "puppet var",
        "name": "zookeeper_keytab_path",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/zk.service.keytab",
        "description": "Path to ZooKeeper keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "component": "ZOOKEEPER_SERVER"
      },

    /**********************************************NAGIOS***************************************/
      {
        "id": "puppet var",
        "name": "nagios_server",
        "displayName": "Nagios Server host",
        "value": "",
        "defaultValue": "localhost",
        "description": "Nagios server host",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "NAGIOS",
        "category": "Nagios Server"
      },
      {
        "id": "puppet var",
        "name": "nagios_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nagios",
        "description": "Primary name for Nagios server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "NAGIOS",
        "category": "Nagios Server",
        "component": "NAGIOS_SERVER"
      },
      {
        "id": "puppet var",
        "name": "nagios_keytab_path",
        "displayName": " Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nagios.service.keytab",
        "description": "Path to the Nagios server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "NAGIOS",
        "category": "Nagios Server",
        "component": "NAGIOS_SERVER"
      },
    /**********************************************STORM***************************************/
      {
        "id": "puppet var",
        "name": "storm_host",
        "displayName": "Storm component hosts",
        "value": "",
        "defaultValue": "",
        "description": "Storm component hosts",
        "displayType": "slaveHosts",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "STORM",
        "category": "Storm Topology"
      },
      {
        "id": "puppet var",
        "name": "storm_principal_name",
        "displayName": " Storm principal name",
        "value": "",
        "defaultValue": "storm/_HOST",
        "description": "Principal name for Supervisor. _HOST will get automatically replaced with actual hostname at an instance of every storm component.",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "STORM",
        "category": "Storm Topology",
        "components": ["SUPERVISOR", "NIMBUS", "STORM_UI_SERVER"]
      },
      {
        "id": "puppet var",
        "name": "storm_keytab",
        "displayName": "Path to Storm keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/storm.service.keytab",
        "description": "Path to the storm keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "STORM",
        "category": "Storm Topology",
        "components": ["SUPERVISOR", "NIMBUS"]
      },

    /**********************************************Falcon***************************************/
      {
        "id": "puppet var",
        "name": "falcon_server_host",
        "displayName": "Falcon server host",
        "value": "",
        "defaultValue": "",
        "description": "Falcon Server host",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "FALCON",
        "category": "Falcon Server"
      },
      {
        "id": "puppet var",
        "name": "falcon_principal_name",
        "displayName": "Falcon principal name",
        "value": "",
        "defaultValue": "falcon/_HOST",
        "description": "This is the principal name for Falcon Server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "FALCON",
        "category": "Falcon Server",
        "component": "FALCON_SERVER"
      },
      {
        "id": "puppet var",
        "name": "falcon_keytab",
        "displayName": "Path to Falcon server keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/falcon.service.keytab",
        "description": "Path to the Falcon Server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "FALCON",
        "category": "Falcon Server",
        "component": "FALCON_SERVER"
      },
      {
        "id": "puppet var",
        "name": "falcon_http_principal_name",
        "displayName": "Web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access to Falcon",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "FALCON",
        "category": "Falcon Server"
      },
      {
        "id": "puppet var",
        "name": "falcon_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for Falcon",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "FALCON",
        "category": "Falcon Server"
      },
      {
        "id": "puppet var",
        "name": "namenode_principal_name_falcon",
        "displayName": "NameNode principal name",
        "value": "",
        "defaultValue": "nn/_HOST",
        "description": "NameNode principal to talk to config store",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "FALCON",
        "category": "Falcon Server"
      }
    ]
  };
  
});
window.require.register("data/HDP2/site_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports =
  {
    "configProperties": [
    //***************************************** HDP stack **************************************
    /**********************************************HDFS***************************************/
      {
        "id": "site property",
        "name": "dfs.namenode.checkpoint.dir",
        "displayName": "SecondaryNameNode Checkpoint directory",
        "defaultDirectory": "/hadoop/hdfs/namesecondary",
        "displayType": "directory",
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "index": 1
      },
      {
        "id": "site property",
        "name": "dfs.namenode.checkpoint.period",
        "displayName": "HDFS Maximum Checkpoint Delay",
        "displayType": "int",
        "unit": "seconds",
        "category": "General",
        "serviceName": "HDFS",
        "index": 3
      },
      {
        "id": "site property",
        "name": "dfs.namenode.name.dir",
        "displayName": "NameNode directories",
        "defaultDirectory": "/hadoop/hdfs/namenode",
        "displayType": "directories",
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 1
      },
      {
        "id": "site property",
        "name": "dfs.webhdfs.enabled",
        "displayName": "WebHDFS enabled",
        "displayType": "checkbox",
        "isOverridable": false,
        "category": "General",
        "serviceName": "HDFS",
        "index": 0
      },
      {
        "id": "site property",
        "name": "dfs.datanode.failed.volumes.tolerated",
        "displayName": "DataNode volumes failure toleration",
        "displayType": "int",
        "category": "DataNode",
        "serviceName": "HDFS",
        "index": 3
      },
      {
        "id": "site property",
        "name": "dfs.datanode.data.dir",
        "displayName": "DataNode directories",
        "defaultDirectory": "/hadoop/hdfs/data",
        "displayType": "directories",
        "category": "DataNode",
        "serviceName": "HDFS",
        "index": 1
      },
      {
        "id": "site property",
        "name": "dfs.datanode.data.dir.perm",
        "displayName": "DataNode directories permission",
        "displayType": "int",
        "category": "DataNode",
        "serviceName": "HDFS"
      },
      {
        "id": "site property",
        "name": "dfs.replication",
        "displayName": "Block replication",
        "displayType": "int",
        "category": "Advanced",
        "serviceName": "HDFS"
      },
      {
        "id": "site property",
        "name": "dfs.datanode.du.reserved",
        "displayName": "Reserved space for HDFS",
        "displayType": "int",
        "unit": "bytes",
        "category": "General",
        "serviceName": "HDFS",
        "index": 2
      },
      {
        "id": "site property",
        "name": "dfs.client.read.shortcircuit",
        "displayName": "HDFS Short-circuit read",
        "displayType": "checkbox",
        "category": "Advanced",
        "serviceName": "HDFS"

      },

    /**********************************************YARN***************************************/
      {
        "id": "site property",
        "name": "yarn.acl.enable",
        "displayName": "yarn.acl.enable",
        "displayType": "checkbox",
        "serviceName": "YARN",
        "category": "ResourceManager"
      },
      {
        "id": "site property",
        "name": "yarn.admin.acl",
        "displayName": "yarn.admin.acl",
        "serviceName": "YARN",
        "category": "ResourceManager"
      },
      {
        "id": "site property",
        "name": "yarn.log-aggregation-enable",
        "displayName": "yarn.log-aggregation-enable",
        "displayType": "checkbox",
        "serviceName": "YARN",
        "category": "ResourceManager"
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.scheduler.class",
        "displayName": "yarn.resourcemanager.scheduler.class",
        "serviceName": "YARN",
        "category": "CapacityScheduler"
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.minimum-allocation-mb",
        "displayName": "yarn.scheduler.minimum-allocation-mb",
        "displayType": "int",
        "serviceName": "YARN",
        "category": "CapacityScheduler"
      },
      {
        "id": "site property",
        "name": "yarn.scheduler.maximum-allocation-mb",
        "displayName": "yarn.scheduler.maximum-allocation-mb",
        "displayType": "int",
        "serviceName": "YARN",
        "category": "CapacityScheduler"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.resource.memory-mb",
        "displayName": "yarn.nodemanager.resource.memory-mb",
        "displayType": "int",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.vmem-pmem-ratio",
        "displayName": "yarn.nodemanager.vmem-pmem-ratio",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.linux-container-executor.group",
        "displayName": "yarn.nodemanager.linux-container-executor.group",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.log-dirs",
        "displayName": "yarn.nodemanager.log-dirs",
        "defaultDirectory": "/hadoop/yarn/log",
        "displayType": "directories",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.local-dirs",
        "displayName": "yarn.nodemanager.local-dirs",
        "defaultDirectory": "/hadoop/yarn/local",
        "displayType": "directories",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.remote-app-log-dir",
        "displayName": "yarn.nodemanager.remote-app-log-dir",
        "displayType": "directory",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.remote-app-log-dir-suffix",
        "displayName": "yarn.nodemanager.remote-app-log-dir-suffix",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.aux-services",
        "displayName": "yarn.nodemanager.aux-services",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.nodemanager.log.retain-second",
        "displayName": "yarn.nodemanager.log.retain-second",
        "serviceName": "YARN",
        "category": "NodeManager"
      },
      {
        "id": "site property",
        "name": "yarn.log.server.url",
        "displayName": "yarn.log.server.url",
        "category": "Advanced",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.enabled",
        "displayName": "yarn.timeline-service.enabled",
        "category": "AppTimelineServer",
        "displayType": "checkbox",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.leveldb-timeline-store.path",
        "displayName": "yarn.timeline-service.leveldb-timeline-store.path",
        "category": "AppTimelineServer",
        "displayType": "directory",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
        "displayName": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
        "displayType": "int",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.store-class",
        "displayName": "yarn.timeline-service.store-class",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.ttl-enable",
        "displayName": "yarn.timeline-service.ttl-enable",
        "displayType": "checkbox",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.ttl-ms",
        "displayName": "yarn.timeline-service.ttl-ms",
        "displayType": "int",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.generic-application-history.store-class",
        "displayName": "yarn.timeline-service.generic-application-history.store-class",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.webapp.address",
        "displayName": "yarn.timeline-service.webapp.address",
        "displayType": "string",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.webapp.https.address",
        "displayName": "yarn.timeline-service.webapp.https.address",
        "displayType": "string",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
      {
        "id": "site property",
        "name": "yarn.timeline-service.address",
        "displayName": "yarn.timeline-service.address",
        "displayType": "string",
        "category": "AppTimelineServer",
        "serviceName": "YARN"
      },
    /**********************************************MAPREDUCE2***************************************/
      {
        "id": "site property",
        "name": "mapreduce.map.memory.mb",
        "displayName": "Default virtual memory for a job's map-task",
        "displayType": "int",
        "unit": "MB",
        "category": "General",
        "serviceName": "MAPREDUCE2"
      },
      {
        "id": "site property",
        "name": "mapreduce.reduce.memory.mb",
        "displayName": "Default virtual memory for a job's reduce-task",
        "displayType": "int",
        "unit": "MB",
        "category": "General",
        "serviceName": "MAPREDUCE2"
      },
      {
        "id": "site property",
        "name": "mapreduce.task.io.sort.mb",
        "displayName": "Map-side sort buffer memory",
        "displayType": "int",
        "unit": "MB",
        "category": "General",
        "serviceName": "MAPREDUCE2"
      },
      {
        "id": "site property",
        "name": "hadoop.security.auth_to_local",
        "displayName": "hadoop.security.auth_to_local",
        "displayType": "multiLine",
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "site property",
        "name": "yarn.app.mapreduce.am.resource.mb",
        "displayName": "yarn.app.mapreduce.am.resource.mb",
        "displayType": "int",
        "category": "Advanced",
        "serviceName": "MAPREDUCE2"
      },

    /**********************************************oozie-site***************************************/
      {
        "id": "site property",
        "name": "oozie.db.schema.name",
        "displayName": "Database Name",
        "isOverridable": false,
        "displayType": "host",
        "isObserved": true,
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.username",
        "displayName": "Database Username",
        "isOverridable": false,
        "displayType": "host",
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.password",
        "displayName": "Database Password",
        "isOverridable": false,
        "displayType": "password",
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "filename": "oozie-site.xml",
        "index": 5
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.driver", // the default value of this property is overriden in code
        "displayName": "JDBC Driver Class",
        "isOverridable": false,
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 6
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.url",
        "displayName": "Database URL",
        "isOverridable": false,
        "displayType": "advanced",
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 7
      },

    /**********************************************hive-site***************************************/
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionDriverName",  // the default value is overwritten in code
        "displayName": "JDBC Driver Class",
        "isOverridable": false,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 7
      },
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionUserName",
        "displayName": "Database Username",
        "displayType": "host",
        "isOverridable": false,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionPassword",
        "displayName": "Database Password",
        "displayType": "password",
        "isOverridable": false,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 6
      },
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionURL",
        "displayName": "Database URL",
        "displayType": "advanced",
        "isOverridable": false,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 8
      },
      {
        "id": "site property",
        "name": "ambari.hive.db.schema.name",
        "displayName": "Database Name",
        "displayType": "host",
        "isOverridable": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 4
      },
    /**********************************************tez-site*****************************************/
      {
        "id": "site property",
        "name": "tez.am.resource.memory.mb",
        "displayName": "tez.am.resource.memory.mb",
        "displayType": "int",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.am.java.opts",
        "displayName": "tez.am.java.opts",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.am.grouping.split-waves",
        "displayName": "tez.am.grouping.split-waves",
        "displayType": "float",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.am.grouping.min-size",
        "displayName": "tez.am.grouping.min-size",
        "displayType": "int",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.am.grouping.max-size",
        "displayName": "tez.am.grouping.max-size",
        "displayType": "int",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.am.log.level",
        "displayName": "tez.am.log.level",
        "displayType": "string",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.runtime.intermediate-input.compress.codec",
        "displayName": "tez.runtime.intermediate-input.compress.codec",
        "displayType": "string",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.runtime.intermediate-input.is-compressed",
        "displayName": "tez.runtime.intermediate-input.is-compressed",
        "displayType": "checkbox",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.runtime.intermediate-output.compress.codec",
        "displayName": "tez.runtime.intermediate-output.compress.codec",
        "displayType": "string",
        "category": "General",
        "serviceName": "TEZ"
      },
      {
        "id": "site property",
        "name": "tez.runtime.intermediate-output.should-compress",
        "displayName": "tez.runtime.intermediate-output.should-compress",
        "displayType": "checkbox",
        "category": "General",
        "serviceName": "TEZ"
      },

    /**********************************************hbase-site***************************************/
      {
        "id": "site property",
        "name": "hbase.tmp.dir",
        "displayName": "HBase local directory",
        "defaultDirectory": "/hadoop/hbase",
        "displayType": "directory",
        "category": "Advanced",
        "serviceName": "HBASE"

      },
      {
        "id": "site property",
        "name": "hbase.regionserver.global.memstore.upperLimit",
        "displayName": "hbase.regionserver.global.memstore.upperLimit",
        "displayType": "float",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.regionserver.global.memstore.lowerLimit",
        "displayName": "hbase.regionserver.global.memstore.lowerLimit",
        "displayType": "float",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.hstore.blockingStoreFiles",
        "displayName": "hstore blocking storefiles",
        "displayType": "int",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.hstore.compactionThreshold",
        "displayName": "HBase HStore compaction threshold",
        "displayType": "int",
        "category": "General",
        "serviceName": "HBASE",
        "index": 0
      },
      {
        "id": "site property",
        "name": "hfile.block.cache.size",
        "displayName": "HFile block cache size ",
        "displayType": "float",
        "category": "General",
        "serviceName": "HBASE",
        "index": 1
      },
      {
        "id": "site property",
        "name": "hbase.hregion.max.filesize",
        "displayName": "Maximum HStoreFile Size",
        "displayType": "int",
        "unit": "bytes",
        "category": "General",
        "serviceName": "HBASE",
        "index": 2
      },
      {
        "id": "site property",
        "name": "hbase.regionserver.handler.count",
        "displayName": "HBase RegionServer Handler",
        "displayType": "int",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 2
      },
      {
        "id": "site property",
        "name": "hbase.hregion.majorcompaction",
        "displayName": "HBase Region Major Compaction",
        "displayType": "int",
        "unit": "ms",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "hbase.hregion.memstore.block.multiplier",
        "displayName": "HBase Region Block Multiplier",
        "displayType": "int",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "hbase.hregion.memstore.mslab.enabled",
        "displayName": "hbase.hregion.memstore.mslab.enabled",
        "displayType": "checkbox",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.hregion.memstore.flush.size",
        "displayName": "HBase Region Memstore Flush Size",
        "displayType": "int",
        "unit": "bytes",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "hbase.client.scanner.caching",
        "displayName": "HBase Client Scanner Caching",
        "displayType": "int",
        "unit": "rows",
        "category": "General",
        "serviceName": "HBASE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "zookeeper.session.timeout",
        "displayName": "Zookeeper timeout for HBase Session",
        "displayType": "int",
        "unit": "ms",
        "category": "General",
        "serviceName": "HBASE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "hbase.client.keyvalue.maxsize",
        "displayName": "HBase Client Maximum key-value Size",
        "displayType": "int",
        "unit": "bytes",
        "category": "General",
        "serviceName": "HBASE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "hbase.zookeeper.quorum",
        "displayName": "hbase.zookeeper.quorum",
        "displayType": "multiLine",
        "serviceName": "HBASE",
        "category": "Advanced"
      },

    /**********************************************storm-site***************************************/
      {
        "id": "site property",
        "name": "storm.zookeeper.root",
        "displayName": "storm.zookeeper.root",
        "displayType": "directory",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.local.dir",
        "displayName": "storm.local.dir",
        "defaultDirectory": "/hadoop/storm",
        "displayType": "directory",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.servers",
        "displayName": "storm.zookeeper.servers",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.port",
        "displayName": "storm.zookeeper.port",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.session.timeout",
        "displayName": "storm.zookeeper.session.timeout",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.connection.timeout",
        "displayName": "storm.zookeeper.connection.timeout",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.retry.times",
        "displayName": "storm.zookeeper.retry.times",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.retry.interval",
        "displayName": "storm.zookeeper.retry.interval",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "General",
        "unit": "ms"
      },
      {
        "id": "site property",
        "name": "storm.zookeeper.retry.intervalceiling.millis",
        "displayName": "storm.zookeeper.retry.intervalceiling.millis",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "General",
        "unit": "ms"
      },
      {
        "id": "site property",
        "name": "storm.cluster.mode",
        "displayName": "storm.cluster.mode",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.local.mode.zmq",
        "displayName": "storm.local.mode.zmq",
        "displayType": "checkbox",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.thrift.transport",
        "displayName": "storm.thrift.transport",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "storm.messaging.transport",
        "displayName": "storm.messaging.transport",
        "serviceName": "STORM",
        "category": "General"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "General",
        "displayName": "storm.messaging.netty.buffer_size",
        "name": "storm.messaging.netty.buffer_size",
        "displayType": "int",
        "unit": "bytes"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "General",
        "displayName": "storm.messaging.netty.max_retries",
        "name": "storm.messaging.netty.max_retries",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "General",
        "displayName": "storm.messaging.netty.max_wait_ms",
        "name": "storm.messaging.netty.max_wait_ms",
        "displayType": "int",
        "unit": "ms"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "General",
        "displayName": "storm.messaging.netty.min_wait_ms",
        "name": "storm.messaging.netty.min_wait_ms",
        "displayType": "int",
        "unit": "ms"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "General",
        "displayName": "storm.messaging.netty.server_worker_threads",
        "name": "storm.messaging.netty.server_worker_threads",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "General",
        "displayName": "storm.messaging.netty.client_worker_threads",
        "name": "storm.messaging.netty.client_worker_threads",
        "displayType": "int"
      },
      {
        "id": "site property",
        "name": "nimbus.host",
        "displayName": "nimbus.host",
        "displayType": "masterHost",
        "isOverridable": false,
        "serviceName": "STORM",
        "category": "Nimbus"
      },
      {
        "id": "site property",
        "name": "nimbus.thrift.port",
        "displayName": "nimbus.thrift.port",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus"
      },
      {
        "id": "site property",
        "name": "nimbus.thrift.max_buffer_size",
        "displayName": "nimbus.thrift.max_buffer_size",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "bytes"
      },
      {
        "id": "site property",
        "name": "nimbus.childopts",
        "displayName": "nimbus.childopts",
        "displayType": "multiLine",
        "isOverridable": false,
        "serviceName": "STORM",
        "category": "Nimbus",
        "filename": "storm-site.xml"
      },
      {
        "id": "site property",
        "name": "nimbus.task.timeout.secs",
        "displayName": "nimbus.task.timeout.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.supervisor.timeout.secs",
        "displayName": "nimbus.supervisor.timeout.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.monitor.freq.secs",
        "displayName": "nimbus.monitor.freq.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.cleanup.inbox.freq.secs",
        "displayName": "nimbus.cleanup.inbox.freq.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.inbox.jar.expiration.secs",
        "displayName": "nimbus.inbox.jar.expiration.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.task.launch.secs",
        "displayName": "nimbus.task.launch.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.reassign",
        "displayName": "nimbus.reassign",
        "displayType": "checkbox",
        "isReconfigurable": true,
        "serviceName": "STORM",
        "category": "Nimbus"
      },
      {
        "id": "site property",
        "name": "nimbus.file.copy.expiration.secs",
        "displayName": "nimbus.file.copy.expiration.secs",
        "displayType": "int",
        "serviceName": "STORM",
        "category": "Nimbus",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "name": "nimbus.topology.validator",
        "displayName": "nimbus.topology.validator",
        "serviceName": "STORM",
        "category": "Nimbus"
      },
      {
        "id": "site property",
        "name": "supervisor.slots.ports",
        "displayName": "supervisor.slots.ports",
        "displayType": "string",
        "serviceName": "STORM",
        "category": "Supervisor"
      },
      {
        "id": "site property",
        "isOverrideable": false,
        "serviceName": "STORM",
        "category": "Supervisor",
        "displayName": "supervisor.childopts",
        "name": "supervisor.childopts",
        "displayType": "multiLine",
        "filename": "storm-site.xml"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "Supervisor",
        "displayName": "supervisor.worker.start.timeout.secs",
        "name": "supervisor.worker.start.timeout.secs",
        "displayType": "int",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "Supervisor",
        "displayName": "supervisor.worker.timeout.secs",
        "name": "supervisor.worker.timeout.secs",
        "displayType": "int",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "Supervisor",
        "displayName": "supervisor.monitor.frequency.secs",
        "name": "supervisor.monitor.frequency.secs",
        "displayType": "int",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "Supervisor",
        "displayName": "supervisor.heartbeat.frequency.secs",
        "name": "supervisor.heartbeat.frequency.secs",
        "displayType": "int",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "DRPCServer",
        "displayName": "drpc.port",
        "name": "drpc.port",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "DRPCServer",
        "displayName": "drpc.worker.threads",
        "name": "drpc.worker.threads",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "DRPCServer",
        "displayName": "drpc.queue.size",
        "name": "drpc.queue.size",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "DRPCServer",
        "displayName": "drpc.invocations.port",
        "name": "drpc.invocations.port",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "DRPCServer",
        "displayName": "drpc.request.timeout.secs",
        "name": "drpc.request.timeout.secs",
        "displayType": "int",
        "unit": "seconds"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "DRPCServer",
        "displayName": "drpc.childopts",
        "name": "drpc.childopts",
        "displayType": "string"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "StormUIServer",
        "displayName": "ui.port",
        "name": "ui.port",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "StormUIServer",
        "displayName": "ui.childopts",
        "name": "ui.childopts",
        "displayType": "string"
      },
      //@Todo: uncomment following properties when logviewer is treated as different section on storm service page
      /*
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "LogviewerServer",
        "displayName": "logviewer.port",
        "name": "logviewer.port",
        "displayType": "int"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "LogviewerServer",
        "displayName": "logviewer.childopts",
        "name": "logviewer.childopts",
        "displayType": "string"
      },
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "LogviewerServer",
        "displayName": "logviewer.appender.name",
        "name": "logviewer.appender.name",
        "displayType": "string"
      },
      */
      {
        "id": "site property",
        "serviceName": "STORM",
        "category": "Advanced",
        "displayName": "worker.childopts",
        "name": "worker.childopts",
        "displayType": "multiLine",
        "filename": "storm-site.xml"
      },
    /*********************************************oozie-site for Falcon*****************************/
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
        "name": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-action-create-inst",
        "name": "oozie.service.ELService.ext.functions.coord-action-create-inst",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-action-create",
        "name": "oozie.service.ELService.ext.functions.coord-action-create",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-data",
        "name": "oozie.service.ELService.ext.functions.coord-job-submit-data",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-action-start",
        "name": "oozie.service.ELService.ext.functions.coord-action-start",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-sla-submit",
        "name": "oozie.service.ELService.ext.functions.coord-sla-submit",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },
      {
        "id": "site property",
        "isReconfigurable": false,
        "serviceName": "FALCON",
        "category": "Falcon - Oozie integration",
        "displayName": "oozie.service.ELService.ext.functions.coord-sla-create",
        "name": "oozie.service.ELService.ext.functions.coord-sla-create",
        "displayType": "custom",
        "filename": "oozie-site.xml"
      },

      // Runtime properties
      {
        "id": "site property",
        "name": "*.domain",
        "displayName": "*.domain",
        "category": "FalconRuntimeSite",
        "serviceName": "FALCON",
        "filename": "falcon-runtime.properties.xml"

      },
      {
        "id": "site property",
        "name": "*.log.cleanup.frequency.minutes.retention",
        "displayName": "*.log.cleanup.frequency.minutes.retention",
        "category": "FalconRuntimeSite",
        "serviceName": "FALCON",
        "filename": "falcon-runtime.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.log.cleanup.frequency.hours.retention",
        "displayName": "*.log.cleanup.frequency.hours.retention",
        "category": "FalconRuntimeSite",
        "serviceName": "FALCON",
        "filename": "falcon-runtime.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.log.cleanup.frequency.days.retention",
        "displayName": "*.log.cleanup.frequency.days.retention",
        "category": "FalconRuntimeSite",
        "serviceName": "FALCON",
        "filename": "falcon-runtime.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.log.cleanup.frequency.months.retention",
        "displayName": "*.log.cleanup.frequency.months.retention",
        "category": "FalconRuntimeSite",
        "serviceName": "FALCON",
        "filename": "falcon-runtime.properties.xml"
      },

      //  Startup properties

      {
        "id": "site property",
        "name": "*.domain",
        "displayName": "*.domain",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.workflow.engine.impl",
        "displayName": "*.workflow.engine.impl",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.oozie.process.workflow.builder",
        "displayName": "*.oozie.process.workflow.builder",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.oozie.feed.workflow.builder",
        "displayName": "*.oozie.feed.workflow.builder",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.SchedulableEntityManager.impl",
        "displayName": "*.SchedulableEntityManager.impl",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.ConfigSyncService.impl",
        "displayName": "*.ConfigSyncService.impl",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.ProcessInstanceManager.impl",
        "displayName": "*.ProcessInstanceManager.impl",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.catalog.service.impl",
        "displayName": "*.catalog.service.impl",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.application.services",
        "displayName": "*.application.services",
        "displayType": "multiLine",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.configstore.listeners",
        "displayName": "*.configstore.listeners",
        "displayType": "multiLine",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.broker.impl.class",
        "displayName": "*.broker.impl.class",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.shared.libs",
        "displayName": "*.shared.libs",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.config.store.uri",
        "displayName": "*.config.store.uri",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.system.lib.location",
        "displayName": "*.system.lib.location",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.retry.recorder.path",
        "displayName": "*.retry.recorder.path",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.cleanup.service.frequency",
        "displayName": "*.falcon.cleanup.service.frequency",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.broker.url",
        "displayName": "*.broker.url",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.broker.ttlInMins",
        "displayName": "*.broker.ttlInMins",
        "displayType": "int",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.entity.topic",
        "displayName": "*.entity.topic",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.max.retry.failure.count",
        "displayName": "*.max.retry.failure.count",
        "displayType": "int",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.internal.queue.size",
        "displayName": "*.internal.queue.size",
        "displayType": "int",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.authentication.type",
        "displayName": "*.falcon.authentication.type",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.http.authentication.type",
        "displayName": "*.falcon.http.authentication.type",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.http.authentication.token.validity",
        "displayName": "*.falcon.http.authentication.token.validity",
        "displayType": "int",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.http.authentication.signature.secret",
        "displayName": "*.falcon.http.authentication.signature.secret",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.http.authentication.simple.anonymous.allowed",
        "displayName": "*.falcon.http.authentication.simple.anonymous.allowed",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
      {
        "id": "site property",
        "name": "*.falcon.http.authentication.kerberos.name.rules",
        "displayName": "*.falcon.http.authentication.kerberos.name.rules",
        "category": "FalconStartupSite",
        "serviceName": "FALCON",
        "filename": "falcon-startup.properties.xml"
      },
    /**********************************************webhcat-site***************************************/
      {
        "id": "site property",
        "name": "templeton.zookeeper.hosts",
        "displayName": "templeton.zookeeper.hosts",
        "displayType": "multiLine",
        "serviceName": "WEBHCAT",
        "category": "Advanced"
        /**********************************************log4j.properties***************************************/
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isRequired": false,
        "showLabel": false,
        "serviceName": "HDFS",
        "filename": "hdfs-log4j.xml",
        "category": "AdvancedHDFSLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isRequired": false,
        "showLabel": false,
        "serviceName": "YARN",
        "filename": "yarn-log4j.xml",
        "category": "AdvancedYARNLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isRequired": false,
        "showLabel": false,
        "serviceName": "HBASE",
        "filename": "hbase-log4j.xml",
        "category": "AdvancedHbaseLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isRequired": false,
        "showLabel": false,
        "serviceName": "HIVE",
        "filename": "hive-exec-log4j.xml",
        "category": "AdvancedHiveExecLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isRequired": false,
        "showLabel": false,
        "serviceName": "HIVE",
        "filename": "hive-log4j.xml",
        "category": "AdvancedHiveLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "showLabel": false,
        "serviceName": "OOZIE",
        "filename": "oozie-log4j.xml",
        "category": "AdvancedOozieLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "showLabel": false,
        "serviceName": "ZOOKEEPER",
        "filename": "zookeeper-log4j.xml",
        "category": "AdvancedZooLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isRequired": false,
        "showLabel": false,
        "serviceName": "PIG",
        "filename": "pig-log4j.xml",
        "category": "AdvancedPigLog4j"
      },



      //***************************************** GLUSTERFS stack********************************************

      {
        "id": "site property",
        "name": "fs.glusterfs.impl",
        "displayName": "GlusterFS fs impl",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.glusterfs.volname",
        "displayName": "GlusterFS volume name",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.glusterfs.mount",
        "displayName": "GlusterFS mount point",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.glusterfs.getfattrcmd",
        "displayName": "GlusterFS getfattr command",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.AbstractFileSystem.glusterfs.impl",
        "displayName": "GlusterFS Abstract Filesystem declaration",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "gluster.daemon.user",
        "displayName": "GlusterFS Daemon user",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      }

    ]
  };
  
});
window.require.register("data/config_mapping", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var configs = [
    /**********************************************HDFS***************************************/
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["hive_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable" : true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["hivemetastore_host"],
      "foreignKey": ["hive_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable" : true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["oozie_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable" : true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["oozieserver_host"],
      "foreignKey": ["oozie_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable" : true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["webhcat_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable" : true
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["hivemetastore_host"],
      "foreignKey": ["webhcat_user"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "isOverridable" : true
    }
  ];

  /**
   * Configs consists of 2 types: Computed values, which cannot be modified by user
   * and overridable values, which user can modify. We provide interface how to get all of this
   * configs separately
   * @type {Object}
   */
  module.exports = {
    all : function(){
      return configs.slice(0);
    },
    overridable: function(){
      return configs.filterProperty('foreignKey');
    },
    computed: function(){
      return configs.filterProperty('foreignKey', null);
    }
  };
  
});
window.require.register("data/custom_configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  module.exports =
    [
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.capacity",
        "displayName": "Capacity",
        "value": '',
        "defaultValue": '100',
        "description": "Percentage of the number of slots in the cluster that are made to be available for jobs in this queue. The sum of capacities for all queues should be less than or equal 100.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "unit": "%",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 1
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.maximum-capacity",
        "displayName": "Max Capacity",
        "value": '',
        "defaultValue": '100',
        "displayType": "int",
        "description": "Defines a limit beyond which a queue cannot use the capacity of the cluster." +
          "This provides a means to limit how much excess capacity a queue can use. By default, there is no limit." +
          "The Max Capacity of a queue can only be greater than or equal to its minimum capacity. " +
          "This property could be to curtail certain jobs which are long running in nature from occupying more than a certain " +
          "percentage of the cluster, which in the absence of pre-emption, could lead to capacity guarantees of other queues being affected. " +
          "One important thing to note is that maximum-capacity is a percentage , so based on the cluster's capacity it would change. " +
          "So if large no of nodes or racks get added to the cluster, Max Capacity in absolute terms would increase accordingly. ",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "unit": "%",
        "valueRange": [0, 100],
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 2
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.minimum-user-limit-percent",
        "displayName": "Min User Limit",
        "value": '',
        "defaultValue": '100',
        "displayType": "int",
        "description": "Each queue enforces a limit on the percentage of resources allocated to a user at any given time, " +
          "if there is competition for them. This user limit can vary between a minimum and maximum value. " +
          "The former depends on the number of users who have submitted jobs, and the latter is set to this property value. " +
          "For example, suppose the value of this property is 25. If two users have submitted jobs to a queue, no single user " +
          "can use more than 50% of the queue resources. If a third user submits a job, no single user can use more than 33% of " +
          "the queue resources. With 4 or more users, no user can use more than 25% of the queue's resources. " +
          "A value of 100 implies no user limits are imposed.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "unit": "%",
        "valueRange": [1, 100],
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 9
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.user-limit-factor",
        "displayName": "User Limit Factor",
        "value": '',
        "defaultValue": '1',
        "displayType": "int",
        "description": "The multiple of the queue capacity which can be configured to allow a single user to acquire more slots. " +
          "By default this is set to 1 which ensure that a single user can never take more than the queue's configured capacity " +
          "irrespective of how idle the cluster is.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 8
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.supports-priority",
        "displayName": "Supports Priority",
        "value": 'false',
        "defaultValue": 'false',
        "displayType": "checkbox",
        "description": "If true, priorities of jobs will be taken into account in scheduling decisions.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "inTable": true,
        "index": 7
      },
      {
        "id": "site property",
        "name": "mapred.queue.<queue-name>.acl-submit-job",
        "displayName": "",
        "value": '',
        "defaultValue": '*',
        "description": "",
        "isVisible": false,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'mapred-queue-acls.xml'
      },
      {
        "id": "site property",
        "name": "mapred.queue.<queue-name>.acl-administer-jobs",
        "displayName": "",
        "value": '',
        "defaultValue": '*',
        "description": "",
        "isVisible": false,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'mapred-queue-acls.xml'
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.maximum-initialized-active-tasks",
        "displayName": "Max initialized active tasks",
        "value": '',
        "defaultValue": '200000',
        "displayType": "int",
        "description": "The maximum number of tasks, across all jobs in the queue, which can be initialized concurrently. " +
          "Once the queue's jobs exceed this limit they will be queued on disk.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "index": 10
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.maximum-initialized-active-tasks-per-user",
        "displayName": "Max initialized active tasks per user",
        "value": '',
        "defaultValue": '100000',
        "displayType": "int",
        "description": "The maximum number of tasks per-user, across all the of the user's jobs in the queue, which " +
          "can be initialized concurrently. Once the user's jobs exceed this limit they will be queued on disk.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "index": 11
      },
      {
        "id": "site property",
        "name": "mapred.capacity-scheduler.queue.<queue-name>.init-accept-jobs-factor",
        "displayName": "Init accept jobs factor",
        "value": '',
        "defaultValue": '10',
        "displayType": "int",
        "description": "The multiple of (maximum-system-jobs * queue-capacity) used to determine the number of " +
          "jobs which are accepted by the scheduler. The default value is 10. If number of jobs submitted to " +
          "the queue exceeds this limit, job submission are rejected.",
        "isVisible": true,
        "isRequired": true,
        "serviceName": "MAPREDUCE",
        "category": "CapacityScheduler",
        "isQueue": true,
        "filename": 'capacity-scheduler.xml',
        "index": 12
      }
    ];
});
window.require.register("data/global_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  /**
   * Defines service configuration properties.
   *   name:
   *     The name of the config property that is understood by Ambari server and agent.
   *     E.g., "datanode_du_reserved"
   *
   *   displayName:
   *     The human-friendly display name of the config property.
   *     E.g., "Reserved space for HDFS"
   *
   *   description:
   *     The description of the config property.
   *     E.g., "Reserved space in GB per volume"
   *
   *   defaultValue:
   *     The default value of the config property.
   *     E.g., "1"
   *
   *   isReconfigurable:
   *     Whether the config property can be reconfigured after it has been initially set and deployed.
   *     If this is unspecified, true is assumed.
   *     E.g., true, false
   *
   *   isOverridable:
   *     Whether the config property can be overridden by hosts.
   *     If this is unspecified, true is assumed.
   *
   *   isRequired:
   *     Whether the config property is required or not.
   *     If this is unspecified, true is assumed.
   *     E.g., true, false
   *
   *   displayType:
   *     How the config property is to be rendered for user input.
   *     If this is left unspecified, "string" is assumed
   *     E.g., "string", "int", "float", "checkbox", "directories", "custom", "email", "masterHost", "slaveHosts"
   *
   *   unit
   *     The unit for the config property.
   *     E.g., "ms", "MB", "bytes"
   *
   *   isRequiredByAgent:
   *     Whether the config property is required by agent or not.
   *     If value is false then it will be not persisted in global configuration
   *
   *   serviceName:
   *     The service that the config property belongs to.
   *     E.g., "HDFS", "MAPREDUCE", "ZOOKEEPER", etc.
   *
   *   category: the category that the config property belongs to (used for grouping config properties in the UI).
   *     if unspecified, "General" is assumed.
   *     E.g., "General", "Advanced", "NameNode", "DataNode"
   *
   *   index: the sequence number in category, that point to place where config located regarding all rest in category.
   *     if unspecified, push to the end of array.
   *     E.g., 0, 1, '2'
   */

  var App = require('app');
  require('config');

  module.exports =
  {
    "configProperties": [
    /**********************************************HDFS***************************************/
      {
        "id": "puppet var",
        "name": "namenode_host",
        "displayName": "NameNode host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run NameNode",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "namenode_heapsize",
        "displayName": "NameNode Java heap size",
        "description": "Initial and maximum Java heap size for NameNode (Java options -Xms and -Xmx).  This also applies to the Secondary NameNode.",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "namenode_opt_newsize",
        "displayName": "NameNode new generation size",
        "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
        "defaultValue": "200",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "snamenode_host",
        "displayName": "SNameNode host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run SecondaryNameNode",
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "datanode_hosts", //not in the schema. For UI purpose
        "displayName": "DataNode hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run DataNode",
        "displayType": "slaveHosts",
        "isRequired": false,
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": true,
        "domain": "datanode-global",
        "serviceName": "HDFS",
        "category": "DataNode",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "dtnode_heapsize",
        "displayName": "DataNode maximum Java heap size",
        "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "domain": "datanode-global",
        "serviceName": "HDFS",
        "category": "DataNode",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "hadoop_heapsize",
        "displayName": "Hadoop maximum Java heap size",
        "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "serviceName": "HDFS",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "hdfs_log_dir_prefix",
        "displayName": "Hadoop Log Dir Prefix",
        "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
        "defaultValue": "/var/log/hadoop",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hadoop_pid_dir_prefix",
        "displayName": "Hadoop PID Dir Prefix",
        "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
        "defaultValue": "/var/run/hadoop",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "namenode_opt_maxnewsize",
        "displayName": "NameNode maximum new generation size",
        "description": "",
        "defaultValue": "200",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode"
      },
      {
        "id": "puppet var",
        "name": "security_enabled",
        "displayName": "Hadoop Security",
        "description": "Enable hadoop security",
        "defaultValue": 'false',
        "isRequired": false,
        "displayType": "checkbox",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HDFS",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "namenode_formatted_mark_dir",
        "displayName": "Hadoop formatted mark directory",
        "description": "",
        "defaultValue": "/var/run/hadoop/hdfs/namenode/formatted/",
        "isRequired": false,
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "HDFS",
        "category": "NameNode"
      },

    /**********************************************GLUSTERFS***************************************/
      {
          "id": "puppet var",
          "name": "fs_glusterfs_default_name",
          "displayName": "GlusterFS default fs name 1.x Hadoop",
          "description": "GlusterFS default filesystem name (glusterfs://{MasterFQDN}:9000)",
          "defaultValue": "glusterfs:///",
          "displayType": "string",
          "isVisible": true,
          "domain": "global",
          "serviceName": "GLUSTERFS",
          "category": "General"
        },
        {
            "id": "puppet var",
            "name": "glusterfs_defaultFS_name",
            "displayName": "GlusterFS default fs name 2.x Hadoop",
            "description": "GlusterFS default filesystem name (glusterfs:///)",
            "defaultValue": "glusterfs:///",
            "displayType": "string",
            "isVisible": true,
            "domain": "global",
            "serviceName": "GLUSTERFS",
            "category": "General"
          },
          {
              "id": "puppet var",
              "name": "fs_AbstractFileSystem_glusterfs_impl",
              "displayName": "GlusterFS Abstract Filesystem declaration",
              "description": "GlusterFS Abstract Filesystem declaration",
              "defaultValue": "org.apache.hadoop.fs.local.GlusterFs",
              "displayType": "string",
              "isVisible": true,
              "domain": "global",
              "serviceName": "GLUSTERFS",
              "category": "General"
            },
            {
              "id": "puppet var",
              "name": "hadoop_heapsize",
              "displayName": "Hadoop maximum Java heap size",
              "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
              "defaultValue": "1024",
              "displayType": "int",
              "unit": "MB",
  	          "isVisible": true,
  	          "serviceName": "GLUSTERFS",
              "index": 1
            },
    /**********************************************MAPREDUCE***************************************/
      {
        "id": "puppet var",
        "name": "hs_host",
        "displayName": "History Server",
        "description": "History Server",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "HistoryServer",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "jobtracker_host",
        "displayName": "JobTracker host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run JobTracker",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "jtnode_opt_newsize",
        "displayName": "JobTracker new generation size",
        "description": "Default size of Java new generation size for JobTracker in MB (Java option -XX:NewSize)",
        "defaultValue": "200",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "jtnode_opt_maxnewsize",
        "displayName": "JobTracker maximum new generation size",
        "description": "Maximum size of Java new generation for JobTracker in MB (Java option -XX:MaxNewSize)",
        "defaultValue": "200",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "jtnode_heapsize",
        "displayName": "JobTracker maximum Java heap size",
        "description": "Maximum Java heap size for JobTracker in MB (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "tasktracker_hosts",
        "displayName": "TaskTracker hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run TaskTracker",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "isRequired": false,
        "isRequiredByAgent": false,
        "domain": "tasktracker-global",
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "ttnode_heapsize",
        "displayName": "TaskTracker maximum Java heap size",
        "description": "Maximum Java heap size for TaskTracker in MB (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "lzo_enabled",
        "displayName": "LZO compression",
        "description": "LZO compression enabled",
        "defaultValue": true,
        "displayType": "checkbox",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "MAPREDUCE",
        "index": 11
      },
      {
        "id": "puppet var",
        "name": "snappy_enabled",
        "displayName": "Snappy compression",
        "description": "Snappy compression enabled",
        "defaultValue": true,
        "isReconfigurable": false,
        "displayType": "checkbox",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "MAPREDUCE",
        "index": 12
      },
      {
        "id": "puppet var",
        "name": "rca_enabled",
        "displayName": "Enable Job Diagnostics",
        "description": "Tools for tracing the path and troubleshooting the performance of MapReduce jobs",
        "defaultValue": true,
        "isReconfigurable": true,
        "isOverridable": false,
        "displayType": "checkbox",
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "index": 13
      },
    /**********************************************HBASE***************************************/
      {
        "id": "puppet var",
        "name": "hbasemaster_host",
        "displayName": "HBase Master hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run HBase Master",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "hbase_master_heapsize",
        "displayName": "HBase Master Maximum Java heap size",
        "description": "Maximum Java heap size for HBase master (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isOverridable": true,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "regionserver_hosts",
        "displayName": "RegionServer hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run RegionServer",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "isRequired": false,
        "domain": "regionserver-global",
        "serviceName": "HBASE",
        "category": "RegionServer",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "hbase_regionserver_heapsize",
        "displayName": "HBase RegionServers maximum Java heap size",
        "description": "Maximum Java heap size for HBase RegionServers (Java option -Xmx)",
        "defaultValue": "1024",
        "displayType": "int",
        "unit": "MB",
        "isVisible": true,
        "domain": "regionserver-global",
        "serviceName": "HBASE",
        "category": "RegionServer",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "hbase_log_dir",
        "displayName": "HBase Log Dir",
        "description": "Directory for HBase logs",
        "defaultValue": "/var/log/hbase",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hbase_pid_dir",
        "displayName": "HBase PID Dir",
        "description": "Directory in which the pid files for HBase processes will be created",
        "defaultValue": "/var/run/hbase",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "Advanced"
      },
    /**********************************************HIVE***************************************/
      {
        "id": "puppet var",
        "name": "hivemetastore_host",
        "displayName": "Hive Metastore host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run Hive Metastore",
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 0
      },
      // for existing MySQL
      {
        "id": "puppet var",
        "name": "hive_existing_mysql_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "Using an existing MySQL database for Hive Metastore",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": false,
        "isReconfigurable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 1
      },
      // for existing Oracle
      {
        "id": "puppet var",
        "name": "hive_existing_oracle_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "Oracle",
        "description": "Using an existing Oracle database for Hive Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 1
      },
      // for new MySQL
      {
        "id": "puppet var",
        "name": "hive_ambari_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "MySQL will be installed by Ambari",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "hive_database",
        "displayName": "Hive Database",
        "value": "",
        "defaultValue": "New MySQL Database",
        "options": [
          {
            displayName: 'New MySQL Database',
            foreignKeys: ['hive_ambari_database', 'hive_ambari_host']
          },
          {
            displayName: 'Existing MySQL Database',
            foreignKeys: ['hive_existing_mysql_database', 'hive_existing_mysql_host']
          },
          {
            displayName: 'Existing Oracle Database',
            foreignKeys: ['hive_existing_oracle_database', 'hive_existing_oracle_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          }
        ],
        "description": "MySQL will be installed by Ambari",
        "displayType": "radio button",
        "isReconfigurable": false,
        "radioName": "hive-database",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "hive_hostname",
        "displayName": "Database Host",
        "description": "Specify the host on which the database is hosted",
        "defaultValue": "",
        "isReconfigurable": true,
        "displayType": "host",
        "isOverridable": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_existing_mysql_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "displayType": "host",
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_existing_oracle_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "displayType": "host",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_ambari_host",
        "value": "",
        "defaultValue": "",
        "displayName": "Database Host",
        "description": "The host where Hive Metastore database is located",
        "isReconfigurable": false,
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "hive_metastore_port",
        "displayName": "Hive metastore port",
        "description": "",
        "defaultValue": "9083",
        "isReconfigurable": false,
        "displayType": "int",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_log_dir",
        "displayName": "Hive Log Dir",
        "description": "Directory for Hive log files",
        "defaultValue": "/var/log/hive",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_pid_dir",
        "displayName": "Hive PID Dir",
        "description": "Directory in which the PID files for Hive processes will be created",
        "defaultValue": "/var/run/hive",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "mysql_connector_url",
        "displayName": "MySQL connector url",
        "description": "",
        "defaultValue": "${download_url}/mysql-connector-java-5.1.18.zip",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hive_aux_jars_path",
        "displayName": "Hive auxilary jar path",
        "description": "",
        "defaultValue": "/usr/lib/hcatalog/share/hcatalog/hcatalog-core.jar",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "HIVE",
        "category": "Advanced"
      },
    /**********************************************WEBHCAT***************************************/
      {
        "id": "puppet var",
        "name": "webhcatserver_host",
        "displayName": "WebHCat Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run WebHCat Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      {
        "id": "puppet var",
        "name": "hcat_log_dir",
        "displayName": "WebHCat Log Dir",
        "description": "Directory for WebHCat log files",
        "defaultValue": "/var/log/webhcat",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "WEBHCAT",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hcat_pid_dir",
        "displayName": "WebHCat PID Dir",
        "description": "Directory in which the PID files for WebHCat processes will be created",
        "defaultValue": "/var/run/webhcat",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "WEBHCAT",
        "category": "Advanced"
      },
    /**********************************************OOZIE***************************************/
      {
        "id": "puppet var",
        "name": "oozieserver_host",
        "displayName": "Oozie Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run Oozie Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 0
      },
      // for existing Oracle
      {
        "id": "puppet var",
        "name": "oozie_existing_oracle_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "Oracle",
        "description": "Using an existing Oracle database for Oozie Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isObserved": true,
        "isReconfigurable": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      // for current derby
      {
        "id": "puppet var",
        "name": "oozie_derby_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "Derby",
        "description": "Using current Derby database for Oozie Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isReconfigurable": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      // for existing MySQL oozie
      {
        "id": "puppet var",
        "name": "oozie_existing_mysql_database",
        "displayName": "Database Type",
        "value": "",
        "defaultValue": "MySQL",
        "description": "Using an existing MySQL database for Oozie Metastore",
        "displayType": "masterHost",
        "isVisible": false,
        "isReconfigurable": false,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "oozie_database",
        "displayName": "Oozie Database",
        "value": "",
        "defaultValue": "New Derby Database",
        "options": [
          {
            displayName: 'New Derby Database',
            foreignKeys: ['oozie_derby_database', 'oozie_ambari_host']
          },
          {
            displayName: 'Existing MySQL Database',
            foreignKeys: ['oozie_existing_mysql_database', 'oozie_existing_mysql_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          },
          {
            displayName: 'Existing Oracle Database',
            foreignKeys: ['oozie_existing_oracle_database', 'oozie_existing_oracle_host'],
            hidden: !App.supports.hiveOozieExtraDatabases
          }
        ],
        "description": "Current Derby Database will be installed by Ambari",
        "displayType": "radio button",
        "isReconfigurable": false,
        "isOverridable": false,
        "radioName": "oozie-database",
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "oozie_hostname",
        "defaultValue": "",
        "displayName": "Database Host",
        "description": "The host where the Oozie database is located",
        "isReconfigurable": true,
        "isOverridable": false,
        "displayType": "host",
        "isVisible": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "oozie_existing_mysql_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "isOverridable": false,
        "displayType": "host",
        "isRequiredByAgent": false,
        "isVisible": false,
        "isObserved": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "oozie_existing_oracle_host",
        "displayName": "Database Host",
        "description": "Specify the host on which the existing database is hosted",
        "defaultValue": "",
        "isReconfigurable": false,
        "isOverridable": false,
        "isRequiredByAgent": false,
        "displayType": "host",
        "isVisible": false,
        "isObserved": true,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "oozie_ambari_host",
        "value": "",
        "defaultValue": "",
        "displayName": "Database Host",
        "description": "Host on which the database will be created by Ambari",
        "isReconfigurable": false,
        "isOverridable": false,
        "isRequiredByAgent": false,
        "displayType": "masterHost",
        "isVisible": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "oozie_data_dir",
        "displayName": "Oozie Data Dir",
        "description": "Data directory in which the Oozie DB exists",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/oozie/data",
        "isReconfigurable": true,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "isRequired": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "index": 8
      },

      {
        "id": "puppet var",
        "name": "oozie_log_dir",
        "displayName": "Oozie Log Dir",
        "description": "Directory for oozie logs",
        "defaultValue": "/var/log/oozie",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "oozie_pid_dir",
        "displayName": "Oozie PID Dir",
        "description": "Directory in which the pid files for oozie processes will be created",
        "defaultValue": "/var/run/oozie",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "OOZIE",
        "category": "Advanced"
      },
    /**********************************************NAGIOS***************************************/
      {
        "id": "puppet var",
        "name": "nagios_web_login",
        "displayName": "Nagios Admin username",
        "description": "Nagios Web UI Admin username",
        "defaultValue": "nagiosadmin",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "NAGIOS",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "nagios_web_password",
        "displayName": "Nagios Admin password",
        "description": "Nagios Web UI Admin password",
        "defaultValue": "",
        "isReconfigurable": true,
        "displayType": "password",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "NAGIOS",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "nagios_contact",
        "displayName": "Hadoop Admin email",
        "description": "Hadoop Administrator email for alert notification",
        "defaultValue": "",
        "displayType": "email",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "NAGIOS",
        "index": 2
      },
    /**********************************************ZOOKEEPER***************************************/
      {
        "id": "puppet var",
        "name": "zookeeperserver_hosts",
        "displayName": "ZooKeeper Server hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run ZooKeeper Server",
        "displayType": "masterHosts",
        "isRequiredByAgent": false,
        "isVisible": true,
        "isOverridable": false,
        "isRequired": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "zk_data_dir",
        "displayName": "ZooKeeper directory",
        "description": "Data directory for ZooKeeper",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/zookeeper",
        "isReconfigurable": true,
        "displayType": "directory",
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 1
      },
      {
        "id": "puppet var",
        "name": "tickTime",
        "displayName": "Length of single Tick",
        "description": "The length of a single tick in milliseconds, which is the basic time unit used by ZooKeeper",
        "defaultValue": "2000",
        "displayType": "int",
        "unit": "ms",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 2
      },
      {
        "id": "puppet var",
        "name": "initLimit",
        "displayName": "Ticks to allow for sync at Init",
        "description": "Amount of time, in ticks to allow followers to connect and sync to a leader",
        "defaultValue": "10",
        "displayType": "int",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 3
      },
      {
        "id": "puppet var",
        "name": "syncLimit",
        "displayName": "Ticks to allow for sync at Runtime",
        "description": "Amount of time, in ticks to allow followers to connect",
        "defaultValue": "5",
        "displayType": "int",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 4
      },
      {
        "id": "puppet var",
        "name": "clientPort",
        "displayName": "Port for running ZK Server",
        "description": "Port for running ZooKeeper server",
        "defaultValue": "2181",
        "displayType": "int",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "index": 5
      },
      {
        "id": "puppet var",
        "name": "zk_log_dir",
        "displayName": "ZooKeeper Log Dir",
        "description": "Directory for ZooKeeper log files",
        "defaultValue": "/var/log/zookeeper",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "Advanced",
        "index": 0
      },
      {
        "id": "puppet var",
        "name": "zk_pid_dir",
        "displayName": "ZooKeeper PID Dir",
        "description": "Directory in which the pid files for zookeeper processes will be created",
        "defaultValue": "/var/run/zookeeper",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "ZOOKEEPER",
        "category": "Advanced",
        "index": 1
      },
    /**********************************************HUE***************************************/
      {
        "id": "puppet var",
        "name": "hueserver_host",
        "displayName": "Hue Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run Hue Server",
        "displayType": "masterHost",
        "isRequiredByAgent": false,
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HUE",
        "category": "Hue Server"
      },
      {
        "id": "puppet var",
        "name": "hue_log_dir",
        "displayName": "HUE Log Dir",
        "description": "Directory for HUE logs",
        "defaultValue": "/var/log/hue",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HUE",
        "category": "Advanced"
      },
      {
        "id": "puppet var",
        "name": "hue_pid_dir",
        "displayName": "HUE Pid Dir",
        "description": "Directory in which the pid files for HUE processes will be created",
        "defaultValue": "/var/run/hue",
        "isReconfigurable": false,
        "displayType": "directory",
        "isOverridable": false,
        "isRequiredByAgent": false,
        "isVisible": true,
        "serviceName": "HUE",
        "category": "Advanced"
      },
    /**********************************************GANGLIA***************************************/
      {
        "id": "puppet var",
        "name": "ganglia_conf_dir",
        "displayName": "Ganglia conf directory",
        "description": "",
        "defaultValue": "/etc/ganglia/hdp",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "GANGLIA",
        "category": "Advanced"
      },
    /**********************************************MISC******************************************/
      {
        "id": "puppet var",
        "name": "proxyuser_group",
        "displayName": "Proxy group for Hive, WebHCat and Oozie",
        "description": "",
        "defaultValue": "users",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "filename": "core-site.xml",
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HIVE", "WEBHCAT", "OOZIE"]
      },
      {
        "id": "puppet var",
        "name": "ganglia_runtime_dir",
        "displayName": "Ganglia runtime directory",
        "description": "",
        "defaultValue": "/var/run/ganglia/hdp",
        "isReconfigurable": false,
        "displayType": "directory",
        "isVisible": false,
        "serviceName": "MISC",
        "category": "General",
        "belongsToService": []
      },
      {
        "id": "puppet var",
        "name": "hdfs_user",
        "displayName": "HDFS User",
        "description": "User to run HDFS as",
        "defaultValue": "hdfs",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HDFS"]
      },
      {
        "id": "puppet var",
        "name": "mapred_user",
        "displayName": "MapReduce User",
        "description": "User to run MapReduce as",
        "defaultValue": "mapred",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["MAPREDUCE"]
      },
      {
        "id": "puppet var",
        "name": "hbase_user",
        "displayName": "HBase User",
        "description": "User to run HBase as",
        "defaultValue": "hbase",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HBASE"]
      },
      {
        "id": "puppet var",
        "name": "hive_user",
        "displayName": "Hive User",
        "description": "User to run Hive as",
        "defaultValue": "hive",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HIVE"]
      },
      {
        "id": "puppet var",
        "name": "hcat_user",
        "displayName": "HCat User",
        "description": "User to run HCatalog as",
        "defaultValue": "hcat",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HCATALOG"]
      },
      {
        "id": "puppet var",
        "name": "webhcat_user",
        "displayName": "WebHCat User",
        "description": "User to run WebHCat as",
        "defaultValue": "hcat",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["WEBHCAT"]
      },
      {
        "id": "puppet var",
        "name": "oozie_user",
        "displayName": "Oozie User",
        "description": "User to run Oozie as",
        "defaultValue": "oozie",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["OOZIE"]
      },
      {
        "id": "puppet var",
        "name": "zk_user",
        "displayName": "ZooKeeper User",
        "description": "User to run ZooKeeper as",
        "defaultValue": "zookeeper",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["ZOOKEEPER"]
      },
      {
        "id": "puppet var",
        "name": "gmetad_user",
        "displayName": "Ganglia User",
        "description": "The user used to run Ganglia",
        "defaultValue": "nobody",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["GANGLIA"]
      },
      {
        "id": "puppet var",
        "name": "gmond_user",
        "displayName": "Gmond User",
        "description": "The user used to run gmond for Ganglia",
        "defaultValue": "nobody",
        "isReconfigurable": false,
        "displayType": "advanced",
        "isOverridable": false,
        "isVisible": false,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": []
      },
      {
        "id": "puppet var",
        "name": "nagios_user",
        "displayName": "Nagios User",
        "description": "User to run Nagios as",
        "defaultValue": "nagios",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["NAGIOS"]
      },
      {
        "id": "puppet var",
        "name": "nagios_group",
        "displayName": "Nagios Group",
        "description": "Nagios Group",
        "defaultValue": "nagios",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["NAGIOS"]
      },
      {
        "id": "puppet var",
        "name": "smokeuser",
        "displayName": "Smoke Test User",
        "description": "The user used to run service smoke tests",
        "defaultValue": "ambari-qa",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": App.supports.customizeSmokeTestUser,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HDFS"]
      },
      {
        "id": "puppet var",
        "name": "user_group",
        "displayName": "Hadoop Group",
        "description": "Group that the users specified above belong to",
        "defaultValue": "hadoop",
        "isReconfigurable": false,
        "displayType": "user",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MISC",
        "category": "Users and Groups",
        "belongsToService": ["HDFS"]
      },
      {
        "id": "puppet var",
        "name": "rrdcached_base_dir",
        "displayName": "Ganglia rrdcached base directory",
        "description": "Default directory for saving the rrd files on ganglia server",
        "defaultValue": "/var/lib/ganglia/rrds",
        "displayType": "directory",
        "isReconfigurable": true,
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "GANGLIA",
        "category": "General",
        "belongsToService": ["GANGLIA"]
      }
    ]
  };
  
});
window.require.register("data/host/categories", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  module.exports = [
    {
      value: Em.I18n.t('common.all'),
      isHealthStatus: true,
      healthStatusValue: '',
      isActive: true,
      isVisible: false
    },
    {
      value: Em.I18n.t('hosts.host.healthStatusCategory.green'),
      isHealthStatus: true,
      class: App.healthIconClassGreen,
      healthStatusValue: 'health-status-LIVE',
      observes: 'view.content.@each.healthClass'
    },
    {
      value: Em.I18n.t('hosts.host.healthStatusCategory.red'),
      isHealthStatus: true,
      class: App.healthIconClassRed,
      healthStatusValue: 'health-status-DEAD-RED',
      observes: 'view.content.@each.healthClass'
    },
    {
      value: Em.I18n.t('hosts.host.healthStatusCategory.orange'),
      isHealthStatus: true,
      class: App.healthIconClassOrange,
      healthStatusValue: 'health-status-DEAD-ORANGE',
      observes: 'view.content.@each.healthClass'
    },
    {
      value: Em.I18n.t('hosts.host.healthStatusCategory.yellow'),
      isHealthStatus: true,
      class: App.healthIconClassYellow,
      healthStatusValue: 'health-status-DEAD-YELLOW',
      observes: 'view.content.@each.healthClass'
    },
    {
      value: Em.I18n.t('hosts.host.alerts.label'),
      hostProperty: 'criticalAlertsCount',
      class: 'icon-exclamation-sign',
      isHealthStatus: false,
      healthStatusValue: 'health-status-WITH-ALERTS',
      column: 7,
      type: 'number',
      filterValue: '>0',
      observes: 'view.content.@each.criticalAlertsCount'
    },
    {
      value: Em.I18n.t('common.restart'),
      hostProperty: 'componentsWithStaleConfigsCount',
      class: 'icon-refresh',
      isHealthStatus: false,
      healthStatusValue: 'health-status-RESTART',
      column: 8,
      type: 'number',
      filterValue: '>0',
      observes: 'view.content.@each.componentsWithStaleConfigsCount'
    },
    {
      value: Em.I18n.t('common.selected'),
      hostProperty: 'selected',
      class: '',
      isHealthStatus: false,
      healthStatusValue: 'health-status-SELECTED',
      selected: true,
      column: 10,
      type: 'boolean',
      filterValue: true,
      isVisible: false,
      observes: 'view.content.@each.selected'
    },
    {
      value: Em.I18n.t('common.passive_state'),
      hostProperty: 'componentsInPassiveStateCount',
      class: 'passive-state icon-medkit',
      isHealthStatus: false,
      healthStatusValue: 'health-status-PASSIVE_STATE',
      column: 9,
      type: 'number',
      filterValue: '>0',
      observes: 'view.content.@each.componentsInPassiveStateCount'
    }
  ];
  
});
window.require.register("data/review_configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = [

    {
      config_name: 'Admin',
      display_name: 'Admin Name',
      config_value: ''
    },
    {
      config_name: 'cluster',
      display_name: 'Cluster Name',
      config_value: ''
    },
    {
      config_name: 'hosts',
      display_name: 'Total Hosts',
      config_value: ''
    },
    {
      config_name: 'Repo',
      display_name: 'Local Repository',
      config_value: ''
    },
    {
      config_name: 'services',
      display_name: 'Services',
      config_value: [
        Ember.Object.create({
          service_name: 'HDFS',
          display_name: 'HDFS',
          service_components: [
            Ember.Object.create({
              display_name: 'NameNode',
              component_name: 'NAMENODE',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'SecondaryNameNode',
              component_name: 'SECONDARY_NAMENODE',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'DataNodes',
              component_name: 'DATANODE',
              component_value: '',
              isMaster: false
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'GLUSTERFS',
          display_name: 'GLUSTERFS',
          service_components: [
            Ember.Object.create({
              display_name: 'GLUSTERFS Client',
              component_name: 'GLUSTERFS_CLIENT',
              component_value: '',
              isMaster: false
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'MAPREDUCE',
          display_name: 'MapReduce',
          service_components: [
            Ember.Object.create({
              display_name: 'JobTracker',
              component_name: 'JOBTRACKER',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'TaskTrackers',
              component_name: 'TASKTRACKER',
              component_value: '',
              isMaster: false
            }),
            Ember.Object.create({
              display_name: 'History Server',
              component_name: 'HISTORYSERVER',
              component_value: '',
              isMaster: true
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'YARN',
          display_name: 'YARN + MapReduce2',
          service_components: [
            Ember.Object.create({
              display_name: 'NodeManager',
              component_name: 'NODEMANAGER',
              component_value: '',
              isMaster: false
            }),
            Ember.Object.create({
              display_name: 'ResourceManager',
              component_name: 'RESOURCEMANAGER',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'History Server',
              component_name: 'HISTORYSERVER',
              component_value: '',
              isMaster: true
            })
            // @todo uncomment after Application Timeline Server API implementation
  //          Ember.Object.create({
  //            display_name: 'App Timeline Server',
  //            component_name: 'APP_TIMELINE_SERVER',
  //            component_value: '',
  //            isMaster: true
  //          })
          ]
        }),
        Ember.Object.create({
          service_name: 'HIVE',
          display_name: 'Hive + HCatalog',
          service_components: [
            Ember.Object.create({
              display_name: 'Hive Metastore',
              component_name: 'HIVE_METASTORE',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'Database',
              component_value: '',
              customHandler: 'loadHiveDbValue'
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'HBASE',
          display_name: 'HBase',
          service_components: [
            Ember.Object.create({
              display_name: 'Master',
              component_name: 'HBASE_MASTER',
              component_value: '',
              customHandler: 'loadHbaseMasterValue'
            }),
            Ember.Object.create({
              display_name: 'RegionServers',
              component_name: 'HBASE_REGIONSERVER',
              component_value: '',
              isMaster: false
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'ZOOKEEPER',
          display_name: 'ZooKeeper',
          service_components: [
            Ember.Object.create({
              display_name: 'Servers',
              component_name: 'ZOOKEEPER_SERVER',
              component_value: '',
              customHandler: 'loadZkServerValue'
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'OOZIE',
          display_name: 'Oozie',
          service_components: [
            Ember.Object.create({
              display_name: 'Server',
              component_name: 'OOZIE_SERVER',
              component_value: '',
              isMaster: true
            }),
            // TODO: uncomment when ready to integrate with database other than Derby
            Ember.Object.create({
              display_name: 'Database',
              component_value: '',
              customHandler: 'loadOozieDbValue'
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'NAGIOS',
          display_name: 'Nagios',
          service_components: [
            Ember.Object.create({
              display_name: 'Server',
              component_name: 'NAGIOS_SERVER',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'Administrator',
              component_value: '',
              customHandler: 'loadNagiosAdminValue'
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'GANGLIA',
          display_name: 'Ganglia',
          service_components: [
            Ember.Object.create({
              display_name: 'Server',
              component_name: 'GANGLIA_SERVER',
              component_value: '',
              isMaster: true
            })
          ]
        }),
       /* Ember.Object.create({
          service_name: 'TEZ',
          display_name: 'TEZ',
          service_components: []
        }),
        Ember.Object.create({
          service_name: 'PIG',
          display_name: 'Pig',
          service_components: []
        }),
        Ember.Object.create({
          service_name: 'SQOOP',
          display_name: 'Sqoop',
          service_components: []
        }),
        Ember.Object.create({
          service_name: 'HCATALOG',
          display_name: 'HCatalog',
          service_components: []
        }),*/
        Ember.Object.create({
          service_name: 'HUE',
          display_name: 'Hue',
          service_components: [
            Ember.Object.create({
              display_name: 'Server',
              component_name: 'HUE_SERVER',
              component_value: '',
              isMaster: true
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'FALCON',
          display_name: 'Falcon',
          service_components: [
            Ember.Object.create({
              display_name: 'Server',
              component_name: 'FALCON_SERVER',
              component_value: '',
              isMaster: true
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'STORM',
          display_name: 'Storm',
          service_components: [
            Ember.Object.create({
              display_name: 'Nimbus',
              component_name: 'NIMBUS',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'Storm REST API Server',
              component_name: 'STORM_REST_API',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'Storm UI Server',
              component_name: 'STORM_UI_SERVER',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'DRPC Server',
              component_name: 'DRPC_SERVER',
              component_value: '',
              isMaster: true
            }),
            Ember.Object.create({
              display_name: 'Supervisor',
              component_name: 'SUPERVISOR',
              component_value: '',
              isMaster: false
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'PIG',
          display_name: 'Pig',
          service_components: [
            Ember.Object.create({
              display_name: 'Clients',
              component_name: 'CLIENT',
              component_value: '',
              isMaster: false
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'SQOOP',
          display_name: 'Sqoop',
          service_components: [
            Ember.Object.create({
              display_name: 'Clients',
              component_name: 'CLIENT',
              component_value: '',
              isMaster: false
            })
          ]
        }),
        Ember.Object.create({
          service_name: 'TEZ',
          display_name: 'Tez',
          service_components: [
            Ember.Object.create({
              display_name: 'Clients',
              component_name: 'CLIENT',
              component_value: '',
              isMaster: false
            })
          ]
        })
      ]
    }
  ];

  // @todo remove after Application Timeline Server API implementation
  if (App.supports.appTimelineServer) {
    var yarnServiceComponents = module.exports.findProperty('config_name', 'services').config_value.findProperty('service_name','YARN').get('service_components');
    yarnServiceComponents.push(
      Ember.Object.create({
        display_name: 'App Timeline Server',
        component_name: "APP_TIMELINE_SERVER",
        component_value: '',
        isMaster: true
      })
    )
  }
  
});
window.require.register("data/secure_configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  require('models/service_config');
  App.SecureConfigProperties = Ember.ArrayProxy.extend({
    content: require('data/secure_properties').configProperties
  });

  var configProperties = App.SecureConfigProperties.create();

  module.exports = [
    {
      serviceName: 'GENERAL',
      displayName: 'General',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'KERBEROS', displayName: 'Kerberos'}),
        App.ServiceConfigCategory.create({ name: 'AMBARI', displayName: 'Ambari'})
      ],
      sites: ['global'],
      configs: configProperties.filterProperty('serviceName', 'GENERAL')
    },
    {
      serviceName: 'HDFS',
      displayName: 'HDFS',
      filename: 'hdfs-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'}),
        App.ServiceConfigCategory.create({ name: 'NameNode', displayName: 'NameNode'}),
        App.ServiceConfigCategory.create({ name: 'SNameNode', displayName: 'Secondary NameNode'}),
        App.ServiceConfigCategory.create({ name: 'DataNode', displayName: 'DataNode'})
      ],
      sites: ['core-site', 'hdfs-site'],
      configs: configProperties.filterProperty('serviceName', 'HDFS')
    },
    {
      serviceName: 'MAPREDUCE',
      displayName: 'MapReduce',
      filename: 'mapred-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'JobTracker', displayName: 'JobTracker'}),
        App.ServiceConfigCategory.create({ name: 'JobHistoryServer', displayName: 'History Server'}),
        App.ServiceConfigCategory.create({ name: 'TaskTracker', displayName: 'TaskTracker'})
      ],
      sites: ['mapred-site'],
      configs: configProperties.filterProperty('serviceName', 'MAPREDUCE')
    },
    {
      serviceName: 'HIVE',
      displayName: 'Hive',
      filename: 'hive-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Hive Metastore', displayName: 'Hive Metastore and Hive Server 2'})
      ],
      sites: ['hive-site'],
      configs: configProperties.filterProperty('serviceName', 'HIVE')
    },
    {
      serviceName: 'WEBHCAT',
      displayName: 'WebHCat',
      filename: 'webhcat-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'WebHCat Server', displayName : 'WebHCat Server'})
      ],
      sites: ['webhcat-site'],
      configs: configProperties.filterProperty('serviceName', 'WEBHCAT')
    },
    {
      serviceName: 'HBASE',
      displayName: 'HBase',
      filename: 'hbase-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'HBase Master', displayName: 'HBase Master'}),
        App.ServiceConfigCategory.create({ name: 'RegionServer', displayName: 'RegionServer'})
      ],
      sites: ['hbase-site'],
      configs: configProperties.filterProperty('serviceName', 'HBASE')
    },
    {
      serviceName: 'ZOOKEEPER',
      displayName: 'ZooKeeper',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'ZooKeeper Server', displayName: 'ZooKeeper Server'})
      ],
      configs: configProperties.filterProperty('serviceName', 'ZOOKEEPER')

    },
    {
      serviceName: 'OOZIE',
      displayName: 'Oozie',
      filename: 'oozie-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Oozie Server', displayName:  'Oozie Server'})
      ],
      sites: ['oozie-site'],
      configs: configProperties.filterProperty('serviceName', 'OOZIE')
    },
    {
      serviceName: 'NAGIOS',
      displayName: 'Nagios',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Nagios Server', displayName:  'Nagios Server'})
      ],
      configs: configProperties.filterProperty('serviceName', 'NAGIOS')
    }

  ];
  
});
window.require.register("data/secure_mapping", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports = [
    {
      "name": "hadoop.security.authentication",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "core-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "hadoop.security.authorization",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "core-site.xml",
      "serviceName": "HDFS"
    },

    {
      "name": "hadoop.security.auth_to_local",
      "templateName": ["jobtracker_primary_name", "kerberos_domain", "mapred_user", "tasktracker_primary_name", "namenode_primary_name", "hdfs_user", "datanode_primary_name", "hbase_master_primary_name", "hbase_user","hbase_regionserver_primary_name","oozie_primary_name","oozie_user","jobhistory_primary_name"],
      "foreignKey": null,
      "value": "RULE:[2:$1@$0](<templateName[0]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[3]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[4]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[6]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[11]>/\nRULE:[2:$1@$0](<templateName[12]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nDEFAULT",
      "filename": "core-site.xml",
      "serviceName": "HDFS",
      "dependedServiceName": [{name: "HBASE", replace: "\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/"},{name: "OOZIE",replace: "\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[11]>/"}]
    },
    {
      "name": "dfs.namenode.kerberos.principal",
      "templateName": ["namenode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.namenode.keytab.file",
      "templateName": ["namenode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.secondary.namenode.kerberos.principal",
      "templateName": ["snamenode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml"
    },
    {
      "name": "dfs.secondary.namenode.keytab.file",
      "templateName": ["snamenode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.web.authentication.kerberos.principal",
      "templateName": ["hadoop_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.web.authentication.kerberos.keytab",
      "templateName": ["hadoop_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.kerberos.principal",
      "templateName": ["datanode_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.keytab.file",
      "templateName": ["datanode_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.namenode.kerberos.internal.spnego.principal",
      "templateName": [],
      "foreignKey": null,
      "value": "${dfs.web.authentication.kerberos.principal}",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.secondary.namenode.kerberos.internal.spnego.principal",
      "templateName": [],
      "foreignKey": null,
      "value": "${dfs.web.authentication.kerberos.principal}",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.address",
      "templateName": ["dfs_datanode_address"],
      "foreignKey": null,
      "value": "0.0.0.0:<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "dfs.datanode.http.address",
      "templateName": ["dfs_datanode_http_address"],
      "foreignKey": null,
      "value": "0.0.0.0:<templateName[0]>",
      "filename": "hdfs-site.xml",
      "serviceName": "HDFS"
    },
    {
      "name": "mapreduce.jobtracker.kerberos.principal",
      "templateName": ["jobtracker_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "mapreduce.jobtracker.keytab.file",
      "templateName": ["jobtracker_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "mapreduce.jobhistory.kerberos.principal",
      "templateName": ["jobhistory_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "mapreduce.jobhistory.keytab.file",
      "templateName": ["jobhistory_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "mapreduce.tasktracker.kerberos.principal",
      "templateName": ["tasktracker_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "mapreduce.tasktracker.keytab.file",
      "templateName": ["tasktracker_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "mapred.task.tracker.task-controller",
      "templateName": ["tasktracker_task_controller"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "mapred-site.xml",
      "serviceName": "MAPREDUCE"
    },
    {
      "name": "hbase.master.kerberos.principal",
      "templateName": ["hbase_master_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.master.keytab.file",
      "templateName": ["hbase_master_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.regionserver.kerberos.principal",
      "templateName": ["hbase_regionserver_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.regionserver.keytab.file",
      "templateName": ["hbase_regionserver_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hive.metastore.sasl.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.security.authorization.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.server2.authentication",
      "templateName": [],
      "foreignKey": null,
      "value": "KERBEROS",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.metastore.kerberos.principal",
      "templateName": ["hive_metastore_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.metastore.kerberos.keytab.file",
      "templateName": ["hive_metastore_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.server2.authentication.kerberos.principal",
      "templateName": ["hive_metastore_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hive.server2.authentication.kerberos.keytab",
      "templateName": ["hive_metastore_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "hive-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "oozie.service.AuthorizationService.authorization.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.HadoopAccessorService.kerberos.enabled",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "local.realm",
      "templateName": ["kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.HadoopAccessorService.keytab.file",
      "templateName": ["oozie_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.service.HadoopAccessorService.kerberos.principal",
      "templateName": ["oozie_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.type",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.kerberos.principal",
      "templateName": ["oozie_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.kerberos.keytab",
      "templateName": ["oozie_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "oozie.authentication.kerberos.name.rules",
      "templateName": ["jobtracker_primary_name", "kerberos_domain", "mapred_user", "tasktracker_primary_name", "namenode_primary_name", "hdfs_user", "datanode_primary_name", "hbase_master_primary_name", "hbase_user","hbase_regionserver_primary_name", "jobhistory_primary_name"],
      "foreignKey": null,
      "value": "RULE:[2:$1@$0](<templateName[0]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[3]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[4]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[6]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nDEFAULT",
      "filename": "oozie-site.xml",
      "serviceName": "OOZIE",
      "dependedServiceName": [{name: "HBASE", replace: "\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/"}]
    },
    {
      "name": "templeton.kerberos.principal",
      "templateName": ["webHCat_http_principal_name", "kerberos_domain"],
      "foreignKey": null,
      "value": "<templateName[0]>@<templateName[1]>",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "templeton.kerberos.keytab",
      "templateName": ["webhcat_http_keytab"],
      "foreignKey": null,
      "value": "<templateName[0]>",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "templeton.kerberos.secret",
      "templateName": [""],
      "foreignKey": null,
      "value": "secret",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "templeton.hive.properties",
      "templateName": ["hivemetastore_host","hive_metastore_principal_name","kerberos_domain"],
      "foreignKey": null,
      "value": "hive.metastore.local=false,hive.metastore.uris=thrift://<templateName[0]>:9083,hive." +
        "metastore.sasl.enabled=true,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse,hive.exec.mode.local.auto=false,hive.metastore.kerberos.principal=<templateName[1]>@<templateName[2]>",
      "filename": "webhcat-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "hbase.coprocessor.master.classes",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.security.access.AccessController",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.coprocessor.region.classes",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.security.token.TokenProvider,org.apache.hadoop.hbase.security.access.SecureBulkLoadEndpoint,org.apache.hadoop.hbase.security.access.AccessController",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.security.authentication",
      "templateName": [],
      "foreignKey": null,
      "value": "kerberos",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.rpc.engine",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.ipc.SecureRpcEngine",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.security.authorization",
      "templateName": [],
      "foreignKey": null,
      "value": "true",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hbase.coprocessor.region.classes",
      "templateName": [],
      "foreignKey": null,
      "value": "org.apache.hadoop.hbase.security.token.TokenProvider,org.apache.hadoop.hbase.security.access.SecureBulkLoadEndpoint,org.apache.hadoop.hbase.security.access.AccessController",
      "filename": "hbase-site.xml"
    },
    {
      "name": "hbase.bulkload.staging.dir",
      "templateName": [],
      "foreignKey": null,
      "value": "/apps/hbase/staging",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "zookeeper.znode.parent",
      "templateName": [],
      "foreignKey": null,
      "value": "/hbase-secure",
      "filename": "hbase-site.xml",
      "serviceName": "HBASE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["hive_metastore_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["hivemetastore_host"],
      "foreignKey": ["hive_metastore_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "HIVE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["oozie_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["oozieserver_host"],
      "foreignKey": ["oozie_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "OOZIE"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
      "templateName": ["proxyuser_group"],
      "foreignKey": ["webHCat_http_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "WEBHCAT"
    },
    {
      "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
      "templateName": ["webhcat_server"],
      "foreignKey": ["webHCat_http_primary_name"],
      "value": "<templateName[0]>",
      "filename": "core-site.xml",
      "serviceName": "WEBHCAT"
    }
  ];

  
});
window.require.register("data/secure_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  module.exports =
  {
    "configProperties": [
      {
        "id": "puppet var",
        "name": "security_enabled",
        "displayName": "Enable security",
        "value": "",
        "defaultValue": 'true',
        "description": "Enable kerberos security for the cluster",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "kerberos_install_type",
        "displayName": "Type of security",
        "value": "",
        "defaultValue": "MANUALLY_SET_KERBEROS",
        "description": "Type of kerberos security for the cluster",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "keytab_path",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs",
        "description": "Type of kerberos security for the cluster",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "kerberos_domain",
        "displayName": "Realm name",
        "value": "",
        "defaultValue": "EXAMPLE.COM",
        "description": "Realm name to be used for all principal names",
        "displayType": "advanced",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "kinit_path_local",
        "displayName": "Kerberos tool path",
        "value": "",
        "defaultValue": "/usr/bin",
        "description": "Directoy path to installed kerberos tools like kinit, kdestroy etc. This can have multiple comma delimited paths",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "KERBEROS"
      },
      {
        "id": "puppet var",
        "name": "smokeuser_principal_name",
        "displayName": "Smoke test user principal",
        "value": "",
        "defaultValue": "ambari-qa",
        "description": "This is the principal name for Smoke test user",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "smokeuser_keytab",
        "displayName": "Path to smoke test user keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/smokeuser.headless.keytab",
        "description": "Path to keytab file for smoke test user",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hdfs_principal_name",
        "displayName": "HDFS user principal",
        "value": "",
        "defaultValue": "hdfs",
        "description": "This is the principal name for HDFS user",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hdfs_user_keytab",
        "displayName": "Path to HDFS user keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hdfs.headless.keytab",
        "description": "Path to keytab file for HDFS user",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hbase_principal_name",
        "displayName": "HBase user principal",
        "value": "",
        "defaultValue": "hbase",
        "description": "This is the principal name for HBase user",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "isReconfigurable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },
      {
        "id": "puppet var",
        "name": "hbase_user_keytab",
        "displayName": "Path to HBase user keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hbase.headless.keytab",
        "description": "Path to keytab file for Hbase user",
        "displayType": "directory",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "GENERAL",
        "category": "AMBARI"
      },

      //HDFS
      {
        "id": "puppet var",
        "name": "namenode_host",
        "displayName": "NameNode host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run NameNode",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode"
      },
      {
        "id": "puppet var",
        "name": "namenode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nn/_HOST",
        "description": "Principal name for NameNode. _HOST will get automatically replaced with actual hostname at an instance of NameNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "NameNode",
        "components": ["NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "namenode_keytab",
        "displayName": "Path to Keytab File",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nn.service.keytab",
        "description": "Path to NameNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "NameNode",
        "components": ["NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "snamenode_host",
        "displayName": "SNameNode host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run SecondaryNameNode",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "SNameNode"
      },
      {
        "id": "puppet var",
        "name": "snamenode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nn/_HOST",
        "description": "Principal name for SNameNode. _HOST will get automatically replaced with actual hostname at an instance of SNameNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "components": ["SECONDARY_NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "snamenode_keytab",
        "displayName": "Path to Keytab File",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nn.service.keytab",
        "description": "Path to SNameNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "components": ["SECONDARY_NAMENODE"]
      },
      {
        "id": "puppet var",
        "name": "datanode_hosts", //not in the schema. For UI purpose
        "displayName": "DataNode hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run DataNode",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "DataNode"
      },
      {
        "id": "puppet var",
        "name": "dfs_datanode_address",
        "displayName": "Datanode address",
        "value": "",
        "defaultValue": "1019",
        "description": "Address for DataNode",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode"
      },
      {
        "id": "puppet var",
        "name": "dfs_datanode_http_address",
        "displayName": "Datanode HTTP address",
        "value": "",
        "defaultValue": "1022",
        "description": "Address for DataNode",
        "displayType": "principal",
        "isVisible": false,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode"
      },
      {
        "id": "puppet var",
        "name": "datanode_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "dn/_HOST",
        "description": "Principal name for DataNode. _HOST will get automatically replaced with actual hostname at every instance of DataNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode",
        "component": "DATANODE"
      },
      {
        "id": "puppet var",
        "name": "datanode_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/dn.service.keytab",
        "description": "Path to DataNode keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "DataNode",
        "component": "DATANODE"
      },
      {
        "id": "puppet var",
        "name": "hadoop_http_principal_name",
        "displayName": "DFS web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access for NameNode and SNameNode. _HOST will get automatically replaced with actual hostname at instance of NameNode and SNameNode",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "General"
      },
      {
        "id": "puppet var",
        "name": "hadoop_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for NameNode and SNameNode",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HDFS",
        "category": "General"
      },
      //MAPREDUCE
      {
        "id": "puppet var",
        "name": "jobtracker_host",
        "displayName": "JobTracker host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run JobTracker",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker"
      },
      {
        "id": "puppet var",
        "name": "jobtracker_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "jt/_HOST",
        "description": "Principal name for JobTracker . _HOST will get automatically replaced with actual hostname at instance of JobTracker",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker",
        "component": "JOBTRACKER"
      },
      {
        "id": "puppet var",
        "name": "jobtracker_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/jt.service.keytab",
        "description": "Path to JobTracker keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "JobTracker",
        "component": "JOBTRACKER"
      },
      {
        "id": "puppet var",
        "name": "jobhistoryserver_host",
        "displayName": "History Server host",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run History Server",
        "displayType": "masterHost",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "JobHistoryServer"
      },
      {
        "id": "puppet var",
        "name": "jobhistory_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "jhs/_HOST",
        "description": "Principal name for History Server. _HOST will get automatically replaced with actual hostname at an instance of History Server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "JobHistoryServer",
        "component": "HISTORYSERVER"
      },
      {
        "id": "puppet var",
        "name": "jobhistory_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/jhs.service.keytab",
        "description": "Path to History Server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "JobHistoryServer",
        "component": "HISTORYSERVER"
      },
      {
        "id": "puppet var",
        "name": "tasktracker_hosts",
        "displayName": "TaskTracker hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run TaskTracker",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker"
      },
      {
        "id": "puppet var",
        "name": "tasktracker_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "tt/_HOST",
        "description": "Principal name for TaskTracker. _HOST will get automatically replaced with actual hostname at every instance of TaskTracker",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "component": "TASKTRACKER"
      },
      {
        "id": "puppet var",
        "name": "tasktracker_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/tt.service.keytab",
        "description": "Path to TaskTracker keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "component": "TASKTRACKER"
      },
      {
        "id": "puppet var",
        "name": "tasktracker_task_controller",
        "displayName": "TaskController",
        "value": "",
        "defaultValue": "org.apache.hadoop.mapred.LinuxTaskController",
        "description": "TaskController which is used to launch and manage task execution. This sets property mapred.task.tracker.task-controller",
        "displayType": "advanced",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "component": "TASKTRACKER"
      },

      //WEBHCAT
      {
        "id": "puppet var",
        "name": "webhcatserver_host",
        "displayName": "WebHCat Server host",
        "value": "",
        "defaultValue": "localhost",
        "description": "The host that has been assigned to run WebHCat Server",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      {
        "id": "puppet var",
        "name": "webHCat_http_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access for WebHCat",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      {
        "id": "puppet var",
        "name": "webhcat_http_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for WebHCat",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "WEBHCAT",
        "category": "WebHCat Server"
      },
      //HBASE
      {
        "id": "puppet var",
        "name": "hbasemaster_host",
        "displayName": "HBase Master hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run HBase Master",
        "displayType": "masterHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "HBase Master"
      },
      {
        "id": "puppet var",
        "name": "hbase_master_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "hbase/_HOST",
        "description": "Principal name for HBase master. _HOST will get automatically replaced with actual hostname at an instance of HBase Master",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "components": ["HBASE_MASTER"]
      },
      {
        "id": "puppet var",
        "name": "hbase_master_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hbase.service.keytab",
        "description": "Path to HBase master keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "HBase Master",
        "components": ["HBASE_MASTER"]
      },
      {
        "id": "puppet var",
        "name": "regionserver_hosts",
        "displayName": "RegionServer hosts",
        "value": "",
        "defaultValue": "",
        "description": "The hosts that have been assigned to run RegionServer",
        "displayType": "slaveHosts",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HBASE",
        "category": "RegionServer"
      },
      {
        "id": "puppet var",
        "name": "hbase_regionserver_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "hbase/_HOST",
        "description": "Principal name for HBase RegionServer. _HOST will get automatically replaced with actual hostname at every instance of RegionServer",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "RegionServer",
        "components": ["HBASE_REGIONSERVER"]
      },
      {
        "id": "puppet var",
        "name": "hbase_regionserver_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hbase.service.keytab",
        "description": "Path to HBase RegionServer keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HBASE",
        "category": "RegionServer",
        "components": ["HBASE_REGIONSERVER"]
      },

      //HIVE
      {
        "id": "puppet var",
        "name": "hive_metastore",
        "displayName": "Hive Metastore host",
        "value": "",
        "defaultValue": "localhost",
        "description": "The host that has been assigned to run Hive Metastore and HiveServer2",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore"
      },
      {
        "id": "puppet var",
        "name": "hive_metastore_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "hive/_HOST",
        "description": "Principal name for Hive Metastore and HiveServer2. _HOST will get automatically replaced with actual hostname at an instance of Hive Metastore and HiveServer2",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "component": "HIVE_SERVER"
      },
      {
        "id": "puppet var",
        "name": "hive_metastore_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/hive.service.keytab",
        "description": "Path to Hive Metastore and HiveServer2 keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "component": "HIVE_SERVER"
      },

      //OOZIE
      {
        "id": "puppet var",
        "name": "oozie_servername",
        "displayName": "Oozie Server host",
        "value": "",
        "defaultValue": "localhost",
        "description": "Oozie server host name",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "oozie/_HOST",
        "description": "Principal name for Oozie server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "component": "OOZIE_SERVER"
      },
      {
        "id": "puppet var",
        "name": "oozie_keytab",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/oozie.service.keytab",
        "description": "Path to Oozie server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server",
        "component": "OOZIE_SERVER"
      },
      {
        "id": "puppet var",
        "name": "oozie_http_principal_name",
        "displayName": "Web principal name",
        "value": "",
        "defaultValue": "HTTP/_HOST",
        "description": "Principal name for SPNEGO access to Oozie",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },
      {
        "id": "puppet var",
        "name": "oozie_http_keytab",
        "displayName": "Path to SPNEGO keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
        "description": "Path to SPNEGO keytab file for oozie",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "OOZIE",
        "category": "Oozie Server"
      },

      //ZooKeeper
      {
        "id": "puppet var",
        "name": "zookeeperserver_hosts",
        "displayName": "ZooKeeper Server hosts",
        "value": "",
        "defaultValue": "",
        "description": "The host that has been assigned to run ZooKeeper Server",
        "displayType": "masterHosts",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server"
      },
      {
        "id": "puppet var",
        "name": "zookeeper_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "zookeeper/_HOST",
        "description": "Principal name for ZooKeeper. _HOST will get automatically replaced with actual hostname at every instance of zookeeper server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "component": "ZOOKEEPER_SERVER"
      },
      {
        "id": "puppet var",
        "name": "zookeeper_keytab_path",
        "displayName": "Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/zk.service.keytab",
        "description": "Path to ZooKeeper keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "ZOOKEEPER",
        "category": "ZooKeeper Server",
        "component": "ZOOKEEPER_SERVER"
      },
      //NAGIOS
      {
        "id": "puppet var",
        "name": "nagios_server",
        "displayName": "Nagios Server host",
        "value": "",
        "defaultValue": "localhost",
        "description": "Nagios server host",
        "displayType": "masterHost",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "NAGIOS",
        "category": "Nagios Server"
      },
      {
        "id": "puppet var",
        "name": "nagios_principal_name",
        "displayName": "Principal name",
        "value": "",
        "defaultValue": "nagios",
        "description": "Primary name for Nagios server",
        "displayType": "principal",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "NAGIOS",
        "category": "Nagios Server",
        "component": "NAGIOS_SERVER"
      },
      {
        "id": "puppet var",
        "name": "nagios_keytab_path",
        "displayName": " Path to keytab file",
        "value": "",
        "defaultValue": "/etc/security/keytabs/nagios.service.keytab",
        "description": "Path to the Nagios server keytab file",
        "displayType": "directory",
        "isVisible": true,
        "isOverridable": false,
        "serviceName": "NAGIOS",
        "category": "Nagios Server",
        "component": "NAGIOS_SERVER"
      }
    ]
  };
  
});
window.require.register("data/service_components", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = new Ember.Set([

    {
      service_name: 'HDFS',
      component_name: 'NAMENODE',
      display_name: 'NameNode',
      isMaster: true,
      isClient: false,
      description: 'Master server that manages the file system namespace and regulates access to files by clients'
    },
    {
      service_name: 'HDFS',
      component_name: 'SECONDARY_NAMENODE',
      display_name: 'SNameNode',
      isMaster: true,
      isClient: false,
      description: 'Helper to the primary NameNode that is responsible for supporting periodic checkpoints of the HDFS metadata'
    },
    {
      service_name: 'HDFS',
      component_name: 'DATANODE',
      display_name: 'DataNode',
      isMaster: false,
      isClient: false,
      description: 'The slave for HDFS'
    },
    {
      service_name: 'HDFS',
      component_name: 'HDFS_CLIENT',
      display_name: 'HDFS Client',
      isMaster: false,
      isClient: true,
      description: 'Client component for HDFS'
    },
    {
      service_name: 'MAPREDUCE',
      component_name: 'JOBTRACKER',
      display_name: 'JobTracker',
      isMaster: true,
      isClient: false,
      description: 'Central Master service that pushes work (MR tasks) out to available TaskTracker nodes in the cluster'
    },
    {
      service_name: 'MAPREDUCE',
      component_name: 'HISTORYSERVER',
      display_name: 'History Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'MAPREDUCE',
      component_name: 'TASKTRACKER',
      display_name: 'TaskTracker',
      isMaster: false,
      isClient: false,
      description: 'The slave for MapReduce'
    },
    {
      service_name: 'MAPREDUCE',
      component_name: 'MAPREDUCE_CLIENT',
      display_name: 'MapReduce Client',
      isMaster: false,
      isClient: true,
      description: 'Client component for MapReduce'
    },
    {
      service_name: 'MAPREDUCE2',
      component_name: 'MAPREDUCE2_CLIENT',
      display_name: 'MapReduce 2 Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'MAPREDUCE2',
      component_name: 'HISTORYSERVER',
      display_name: 'History Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'TEZ',
      component_name: 'TEZ_CLIENT',
      display_name: 'Tez Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'YARN',
      component_name: 'RESOURCEMANAGER',
      display_name: 'ResourceManager',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'YARN',
      component_name: 'YARN_CLIENT',
      display_name: 'YARN Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    // @todo uncomment after Application Timeline Server API implementation
  //  {
  //    service_name: 'YARN',
  //    component_name: 'APP_TIMELINE_SERVER',
  //    display_name: 'App Timeline Server',
  //    isMaster: true,
  //    isClient: false,
  //    stackVersions: ['2.1.1'],
  //    description: ''
  //  },
    {
      service_name: 'YARN',
      component_name: 'NODEMANAGER',
      display_name: 'NodeManager',
      isMaster: false,
      isClient: false,
      description: ''
    },
    {
      service_name: 'ZOOKEEPER',
      component_name: 'ZOOKEEPER_SERVER',
      display_name: 'ZooKeeper',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'ZOOKEEPER',
      component_name: 'ZOOKEEPER_CLIENT',
      display_name: 'ZooKeeper Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'HBASE',
      component_name: 'HBASE_MASTER',
      display_name: 'HBase Master',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'HBASE',
      component_name: 'HBASE_REGIONSERVER',
      display_name: 'RegionServer',
      isMaster: false,
      isClient: false,
      description: 'The slave for HBase'
    },
    {
      service_name: 'HBASE',
      component_name: 'HBASE_CLIENT',
      display_name: 'HBase Client',
      isMaster: false,
      isClient: true,
      description: 'The slave for HBase'
    },
    {
      service_name: 'PIG',
      component_name: 'PIG',
      display_name: 'Pig',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'SQOOP',
      component_name: 'SQOOP',
      display_name: 'Sqoop',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'OOZIE',
      component_name: 'OOZIE_SERVER',
      display_name: 'Oozie Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'OOZIE',
      component_name: 'OOZIE_CLIENT',
      display_name: 'Oozie Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'HIVE',
      component_name: 'HIVE_SERVER',
      display_name: 'HiveServer2',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'HIVE',
      component_name: 'HIVE_METASTORE',
      display_name: 'Hive Metastore',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'HIVE',
      component_name: 'HIVE_CLIENT',
      display_name: 'Hive Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'HIVE',
      component_name: 'MYSQL_SERVER',
      display_name: 'MySQL Server for Hive',
      isMaster: false,
      isClient: false,
      description: ''
    },
    {
      service_name: 'HCATALOG',
      component_name: 'HCAT',
      display_name: 'HCat Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'WEBHCAT',
      component_name: 'WEBHCAT_SERVER',
      display_name: 'WebHCat Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'DASHBOARD',
      component_name: 'DASHBOARD',
      display_name: 'Monitoring Dashboard',
      isMaster: false,
      isClient: false,
      description: ''
    },
    {
      service_name: 'NAGIOS',
      component_name: 'NAGIOS_SERVER',
      display_name: 'Nagios Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'GANGLIA',
      component_name: 'GANGLIA_SERVER',
      display_name: 'Ganglia Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'GANGLIA',
      component_name: 'GANGLIA_MONITOR',
      display_name: 'Ganglia Slave',
      isMaster: false,
      isClient: false,
      description: ''
    },
    {
      service_name: 'KERBEROS',
      component_name: 'KERBEROS_SERVER',
      display_name: 'Kerberos Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    {
      service_name: 'KERBEROS',
      component_name: 'KERBEROS_ADMIN_CLIENT',
      display_name: 'Kerberos Admin Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'KERBEROS',
      component_name: 'KERBEROS_CLIENT',
      display_name: 'Kerberos Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'CLIENT',
      component_name: 'CLIENT',
      display_name: 'Client',
      isMaster: false,
      isClient: true,
      description: ''
    },
    {
      service_name: 'HUE',
      component_name: 'HUE_SERVER',
      display_name: 'Hue Server',
      isMaster: true,
      isClient: false,
      description: ''
    },
    { 
      service_name: 'GLUSTERFS',
      component_name: 'GLUSTERFS_CLIENT',
      display_name: 'GLUSTERFS Client', 
      isMaster: false, 
      isClient: true, 
      description: 'Client component for GLUSTERFS'
    },
    {
      service_name: 'FALCON',
      component_name: 'FALCON_SERVER',
      display_name: 'Falcon Server',
      isMaster: true,
      isClient: false,
      description: 'Falcon Server for mirroring'
    },
    {
      service_name: 'FALCON',
      component_name: 'FALCON_CLIENT',
      display_name: 'Falcon Client',
      isMaster: false,
      isClient: true,
      description: 'Falcon Client for mirroring'
    },
    {
      service_name: 'STORM',
      component_name: 'NIMBUS',
      display_name: 'Nimbus',
      isMaster: true,
      isClient: false,
      description: 'Master component for STORM'
    },
    {
      service_name: 'STORM',
      component_name: 'SUPERVISOR',
      display_name: 'Supervisor',
      isMaster: false,
      isClient: false,
      description: 'Slave component for STORM'
    },
    {
      service_name: 'STORM',
      component_name: 'STORM_UI_SERVER',
      display_name: 'Storm UI Server',
      isMaster: true,
      isClient: false,
      description: 'Master component for STORM'
    },
    {
      service_name: 'STORM',
      component_name: 'DRPC_SERVER',
      display_name: 'DRPC Server',
      isMaster: true,
      isClient: false,
      description: 'Master component for STORM'
    },
    {
      service_name: 'STORM',
      component_name: 'STORM_REST_API',
      display_name: 'Storm REST API Server',
      isMaster: true,
      isClient: false,
      description: 'Master component for STORM'
    }
  ]);

  // @todo remove after Application Timeline Server API implementation
  if (App.supports.appTimelineServer) {
    var appTimelineServerObj = {
      service_name: 'YARN',
      component_name: 'APP_TIMELINE_SERVER',
      display_name: 'App Timeline Server',
      isMaster: true,
      isClient: false,
      stackVersions: ['2.1'],
      description: ''
    };
    module.exports.push(appTimelineServerObj);
  }
  
});
window.require.register("data/service_configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  require('models/service_config');
  require('utils/configs/defaults_providers/yarn_defaults_provider');
  require('utils/configs/defaults_providers/tez_defaults_provider');
  require('utils/configs/defaults_providers/hive_defaults_provider');
  require('utils/configs/defaults_providers/storm_defaults_provider');
  require('utils/configs/defaults_providers/oozie_defaults_provider');
  require('utils/configs/defaults_providers/user_defaults_provider');
  require('utils/configs/validators/yarn_configs_validator');
  require('utils/configs/validators/hive_configs_validator');
  require('utils/configs/validators/tez_configs_validator');
  require('utils/configs/validators/mapreduce2_configs_validator');
  require('utils/configs/validators/storm_configs_validator');
  require('utils/configs/validators/user_configs_validator');

  module.exports = [
    {
      serviceName: 'HDFS',
      displayName: 'HDFS',
      filename: 'hdfs-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'NameNode', displayName : 'NameNode', hostComponentNames : ['NAMENODE']}),
        App.ServiceConfigCategory.create({ name: 'SNameNode', displayName : 'Secondary NameNode', hostComponentNames : ['SECONDARY_NAMENODE']}),
        App.ServiceConfigCategory.create({ name: 'DataNode', displayName : 'DataNode', hostComponentNames : ['DATANODE']}),
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedCoreSite', displayName : 'Custom core-site.xml', siteFileName: 'core-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHDFSSite', displayName : 'Custom hdfs-site.xml', siteFileName: 'hdfs-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHDFSLog4j', displayName : 'Custom log4j.properties', siteFileName: 'hdfs-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'core-site', 'hdfs-site', 'hdfs-log4j'],
      configs: []
    },
    {
      serviceName: 'GLUSTERFS',
      displayName: 'GLUSTERFS',
      filename: 'core-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'})
      ],
      sites: ['core-site'],
      configs: []
    },

    {
      serviceName: 'MAPREDUCE',
      displayName: 'MapReduce',
      filename: 'mapred-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'HistoryServer', displayName : 'History Server', hostComponentNames : ['HISTORYSERVER']}),
        App.ServiceConfigCategory.create({ name: 'JobTracker', displayName : 'JobTracker', hostComponentNames : ['JOBTRACKER']}),
        App.ServiceConfigCategory.create({ name: 'TaskTracker', displayName : 'TaskTracker', hostComponentNames : ['TASKTRACKER']}),
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedMapredSite', displayName : 'Custom mapred-site.xml', siteFileName: 'mapred-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedMapredLog4j', displayName : 'Custom log4j.properties', siteFileName: 'mapreduce-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'mapred-site', 'mapred-queue-acls', 'mapreduce-log4j'],
      configs: []
    },

    {
      serviceName: 'YARN',
      displayName: 'YARN',
      configsValidator: App.YARNConfigsValidator,
      defaultsProviders: [App.YARNDefaultsProvider.create()],
      filename: 'yarn-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'ResourceManager', displayName : 'Resource Manager', hostComponentNames : ['RESOURCEMANAGER']}),
        App.ServiceConfigCategory.create({ name: 'NodeManager', displayName : 'Node Manager', hostComponentNames : ['NODEMANAGER']}),
        App.ServiceConfigCategory.create({ name: 'AppTimelineServer', displayName : 'Application Timeline Server', hostComponentNames : ['APP_TIMELINE_SERVER']}),
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'CapacityScheduler', displayName : 'Scheduler', isCapacityScheduler : true, isCustomView: true, siteFileName: 'capacity-scheduler.xml', siteFileNames: ['capacity-scheduler.xml', 'mapred-queue-acls.xml'], canAddProperty: App.supports.capacitySchedulerUi}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedYARNSite', displayName : 'Custom yarn-site.xml', siteFileName: 'yarn-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedYARNLog4j', displayName : 'Custom log4j.properties', siteFileName: 'yarn-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'yarn-site', 'capacity-scheduler', 'yarn-log4j'],
      configs: []
    },

    {
      serviceName: 'MAPREDUCE2',
      displayName: 'MapReduce 2',
      filename: 'mapred-site',
      configsValidator: App.MapReduce2ConfigsValidator,
      defaultsProviders: [App.YARNDefaultsProvider.create()],
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'HistoryServer', displayName : 'History Server', hostComponentNames : ['HISTORYSERVER']}),
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedMapredSite', displayName : 'Custom mapred-site.xml', siteFileName: 'mapred-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedMapredLog4j', displayName : 'Custom log4j.properties', siteFileName: 'mapreduce2-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'mapred-site', 'mapred-queue-acls', 'mapreduce2-log4j'],
      configs: []
    },

    {
      serviceName: 'HIVE',
      displayName: 'Hive',
      filename: 'hive-site',
      configsValidator: App.HiveConfigsValidator,
      defaultsProviders: [App.HiveDefaultsProvider.create()],
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Hive Metastore', displayName : 'Hive Metastore'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHiveSite', displayName : 'Custom hive-site.xml', siteFileName: 'hive-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHiveLog4j', displayName : 'Custom log4j.properties', siteFileName: 'hive-log4j.xml', canAddProperty: false}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHiveExecLog4j', displayName : 'Custom hive-exec-log4j', siteFileName: 'hive-exec-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'hive-site', 'hive-log4j', 'hive-exec-log4j'],
      configs: []
    },

    {
      serviceName: 'WEBHCAT',
      displayName: 'WebHCat',
      filename: 'webhcat-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'WebHCat Server', displayName : 'WebHCat Server'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedWebHCatSite', displayName : 'Custom webhcat-site.xml', siteFileName: 'webhcat-site.xml', canAddProperty: true})
      ],
      sites: ['global', 'webhcat-site'],
      configs: []
    },

    {
      serviceName: 'HBASE',
      displayName: 'HBase',
      filename: 'hbase-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'HBase Master', displayName : 'HBase Master'}),
        App.ServiceConfigCategory.create({ name: 'RegionServer', displayName : 'RegionServer'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHbaseSite', displayName : 'Custom hbase-site.xml', siteFileName: 'hbase-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedHbaseLog4j', displayName : 'Custom log4j.properties', siteFileName: 'hbase-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'hbase-site', 'hbase-log4j'],
      configs: []
    },

    {
      serviceName: 'ZOOKEEPER',
      displayName: 'ZooKeeper',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'ZooKeeper Server', displayName : 'ZooKeeper Server'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedZooLog4j', displayName : 'Custom log4j.properties', siteFileName: 'zookeeper-log4j.xml', canAddProperty: false}),
        App.ServiceConfigCategory.create({ name: 'AdvancedZooCfg', displayName : 'Custom zoo.cfg', siteFileName: 'zoo.cfg', canAddProperty: true})
      ],
      sites: ['global', 'zookeeper-log4j', 'zoo.cfg'],
      configs: []
    },

    {
      serviceName: 'OOZIE',
      displayName: 'Oozie',
      defaultsProviders: [App.OOZIEDefaultsProvider.create()],
      filename: 'oozie-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Oozie Server', displayName : 'Oozie Server'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedOozieSite', displayName : 'Custom oozie-site.xml', siteFileName: 'oozie-site.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedOozieLog4j', displayName : 'Custom log4j.properties', siteFileName: 'oozie-log4j.xml', canAddProperty: false})
      ],
      sites: ['global', 'oozie-site', 'oozie-log4j'],
      configs: []
    },

    {
      serviceName: 'NAGIOS',
      displayName: 'Nagios',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'})
      ],
      sites: ['global'],
      configs: []
    },

    {
      serviceName: 'GANGLIA',
      displayName: 'Ganglia',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'})
      ],
      sites: ['global'],
      configs: []
    },

    {
      serviceName: 'HUE',
      displayName: 'Hue',
      filename: 'hue-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Hue Server', displayName : 'Hue Server'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'})
      ],
      sites: ['hue-site'],
      configs: []
    },

    {
      serviceName: 'PIG',
      displayName: 'Pig',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedPigLog4j', displayName : 'Custom log4j.properties', siteFileName: 'pig-log4j.xml', canAddProperty: false})
      ],
      sites: ['pig-log4j'],
      configs: []
    },
    {
      serviceName: 'FALCON',
      displayName: 'Falcon',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Falcon', displayName : 'Falcon Server'}),
        App.ServiceConfigCategory.create({ name: 'Falcon - Oozie integration', displayName : 'Falcon - Oozie integration'}),
        App.ServiceConfigCategory.create({ name: 'FalconStartupSite', displayName : 'Falcon startup.properties'}),
        App.ServiceConfigCategory.create({ name: 'FalconRuntimeSite', displayName : 'Falcon runtime.properties'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedFalconStartupSite', displayName : 'Custom startup.properties', siteFileName: 'falcon-startup.properties.xml', canAddProperty: true}),
        App.ServiceConfigCategory.create({ name: 'AdvancedFalconRuntimeSite', displayName : 'Custom runtime.properties', siteFileName: 'falcon-runtime.properties.xml', canAddProperty: true})
      ],
      sites: ['global', 'oozie-site','falcon-startup.properties', 'falcon-runtime.properties'],
      configs: []
    },

    {
      serviceName: 'STORM',
      displayName: 'Storm',
      configsValidator: App.STORMConfigsValidator,
      defaultsProviders: [App.STORMDefaultsProvider.create()],
      filename: 'storm-site',
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'Nimbus', displayName : 'Nimbus'}),
        App.ServiceConfigCategory.create({ name: 'Supervisor', displayName : 'Supervisor'}),
        App.ServiceConfigCategory.create({ name: 'StormUIServer', displayName : 'Storm UI Server'}),
        App.ServiceConfigCategory.create({ name: 'StormRestApi', displayName : 'Storm REST API Server'}),
        App.ServiceConfigCategory.create({ name: 'DRPCServer', displayName : 'DRPC Server'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedStormSite', displayName : 'Custom storm.yaml', siteFileName: 'storm-site.xml', canAddProperty: true})
      ],
      sites: ['global', 'storm-site'],
      configs: []
    },
    {
      serviceName: 'TEZ',
      displayName: 'Tez',
      filename: 'tez-site',
      configsValidator: App.TezConfigsValidator,
      defaultsProviders: [App.TezDefaultsProvider.create()],
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Advanced', displayName : 'Advanced'}),
        App.ServiceConfigCategory.create({ name: 'AdvancedTezSite', displayName : 'Custom tez-site.xml', siteFileName: 'tez-site.xml', canAddProperty: true})
      ],
      sites: ['global', 'tez-site'],
      configs: []
    },
    {
      serviceName: 'MISC',
      displayName: 'Misc',
      configsValidator: App.userConfigsValidator,
      defaultsProviders: [App.userDefaultsProvider.create()],
      configCategories: [
        App.ServiceConfigCategory.create({ name: 'General', displayName : 'General'}),
        App.ServiceConfigCategory.create({ name: 'Users and Groups', displayName : 'Users and Groups'})
      ],
      sites: ['global'],
      configs: []
    }

  ];
  
});
window.require.register("data/services", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = [
    {
      serviceName: 'HDFS',
      displayName: 'HDFS',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.hdfs.description')
    },
    {
      serviceName: 'GLUSTERFS',
      displayName: 'GLUSTERFS',
      isDisabled: false,
      isSelected: false,
      canBeSelected: true,
      description: Em.I18n.t('services.glusterfs.description')
    },
    {
      serviceName: 'MAPREDUCE',
      displayName: 'MapReduce',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.mapreduce.description')
    },
    {
      serviceName: 'MAPREDUCE2',
      displayName: 'MapReduce 2',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      isHidden: true,
      description: Em.I18n.t('services.mapreduce2.description')
    },
    {
      serviceName: 'YARN',
      displayName: 'YARN + MapReduce2',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.yarn.description')
    },
    {
      serviceName: 'TEZ',
      displayName: 'Tez',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.tez.description')
    },
    {
      serviceName: 'NAGIOS',
      displayName: 'Nagios',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.nagios.description')
    },
    {
      serviceName: 'GANGLIA',
      displayName: 'Ganglia',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.ganglia.description')
    },
    {
      serviceName: 'HIVE',
      displayName: 'Hive + HCat',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.hive.description')
    },
    {
      serviceName: 'HCATALOG',
      displayName: 'HCatalog',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      isHidden: true
    },
    {
      serviceName: 'WEBHCAT',
      displayName: 'WebHCat',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      isHidden: true
    },
    {
      serviceName: 'HBASE',
      displayName: 'HBase',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.hbase.description')
    },
    {
      serviceName: 'PIG',
      displayName: 'Pig',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.pig.description')
    },
    {
      serviceName: 'SQOOP',
      displayName: 'Sqoop',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.sqoop.description')
    },
    {
      serviceName: 'OOZIE',
      displayName: 'Oozie',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.oozie.description')
    },
    {
      serviceName: 'ZOOKEEPER',
  	  displayName: 'ZooKeeper',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      description: Em.I18n.t('services.zookeeper.description')
    },
    {
      serviceName: 'HUE',
      displayName: 'Hue',
      isDisabled: false,
      isSelected: App.supports.hue,
      canBeSelected: App.supports.hue,
      isHidden: !App.supports.hue
    },
    {
      serviceName: 'FALCON',
      displayName: 'Falcon',
      isDisabled: false,
      isSelected: App.supports.falcon,
      canBeSelected: App.supports.falcon,
      isHidden: !App.supports.falcon,
      description: Em.I18n.t('services.falcon.description')
    },
    {
      serviceName: 'STORM',
      displayName: 'Storm',
      isDisabled: false,
      isSelected: true,
      canBeSelected: true,
      isHidden: false,
      description: Em.I18n.t('services.storm.description')
    }
  ];
  
});
window.require.register("data/site_properties", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  module.exports =
  {
    "configProperties": [
    /**********************************************HDFS***************************************/
      {
        "id": "site property",
        "name": "fs.checkpoint.dir",
        "displayName": "SecondaryNameNode Checkpoint directory",
        "description": "Directory on the local filesystem where the Secondary NameNode should store the temporary images to merge",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/hdfs/namesecondary",
        "displayType": "directory",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "SNameNode",
        "index": 1
      },
      {
        "id": "site property",
        "name": "fs.checkpoint.period",
        "displayName": "HDFS Maximum Checkpoint Delay",
        "description": "Maximum delay between two consecutive checkpoints for HDFS",
        "defaultValue": "",
        "displayType": "int",
        "unit": "seconds",
        "isVisible": true,
        "category": "General",
        "serviceName": "HDFS",
        "index": 3
      },
      {
        "id": "site property",
        "name": "fs.checkpoint.size",
        "displayName": "HDFS Maximum Edit Log Size for Checkpointing",
        "description": "Maximum size of the edits log file that forces an urgent checkpoint even if the maximum checkpoint delay is not reached",
        "defaultValue": "",
        "displayType": "int",
        "unit": "bytes",
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "General",
        "index": 4
      },
      {
        "id": "site property",
        "name": "dfs.name.dir",
        "displayName": "NameNode directories",
        "description": "NameNode directories for HDFS to store the file system image",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/hdfs/namenode",
        "displayType": "directories",
        "isOverridable": false,
        "isVisible": true,
        "serviceName": "HDFS",
        "category": "NameNode",
        "index": 1
      },
      {
        "id": "site property",
        "name": "dfs.webhdfs.enabled",
        "displayName": "WebHDFS enabled",
        "description": "Whether to enable WebHDFS feature",
        "defaultValue": "",
        "displayType": "checkbox",
        "isOverridable": false,
        "category": "General",
        "serviceName": "HDFS",
        "index": 0
      },
      {
        "id": "site property",
        "name": "dfs.datanode.failed.volumes.tolerated",
        "displayName": "DataNode volumes failure toleration",
        "description": "The number of volumes that are allowed to fail before a DataNode stops offering service",
        "defaultValue": "",
        "displayType": "int",
        "isVisible": true,
        "category": "DataNode",
        "serviceName": "HDFS",
        "index": 3
      },
      {
        "id": "site property",
        "name": "dfs.data.dir",
        "displayName": "DataNode directories",
        "description": "DataNode directories for HDFS to store the data blocks",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/hdfs/data",
        "displayType": "directories",
        "isVisible": true,
        "category": "DataNode",
        "serviceName": "HDFS",
        "index": 1
      },
      {
        "id": "site property",
        "name": "dfs.datanode.data.dir.perm",
        "displayName": "DataNode directories permission",
        "description": "",
        "defaultValue": "",
        "displayType": "int",
        "isVisible": true,
        "category": "DataNode",
        "serviceName": "HDFS"
      },
      {
        "id": "site property",
        "name": "dfs.replication",
        "displayName": "Block replication",
        "description": "Default block replication.",
        "displayType": "int",
        "defaultValue": "",
        "isVisible": true,
        "category": "Advanced",
        "serviceName": "HDFS"
      },
      {
        "id": "site property",
        "name": "dfs.datanode.du.reserved",
        "displayName": "Reserved space for HDFS",
        "description": "Reserved space in bytes per volume. Always leave this much space free for non dfs use.",
        "defaultValue": "",
        "displayType": "int",
        "unit": "bytes",
        "isVisible": true,
        "category": "General",
        "serviceName": "HDFS",
        "index": 2
      },

    /******************************************MAPREDUCE***************************************/
      {
        "id": "site property",
        "name": "mapred.local.dir",
        "displayName": "MapReduce local directories",
        "description": "Directories for MapReduce to store intermediate data files",
        "defaultValue": "",
        "defaultDirectory": "/hadoop/mapred",
        "displayType": "directories",
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "index": 1
      },
      {
        "id": "site property",
        "name": "mapred.jobtracker.taskScheduler",
        "displayName": "MapReduce Capacity Scheduler",
        "description": "The scheduler to use for scheduling of MapReduce jobs",
        "defaultValue": "",
        "displayType": "advanced",
        "isOverridable": false,
        "serviceName": "MAPREDUCE",
        "index": 0
      },
      {
        "id": "site property",
        "name": "mapred.tasktracker.map.tasks.maximum",
        "displayName": "Number of Map slots per node",
        "description": "Number of slots that Map tasks that run simultaneously can occupy on a TaskTracker",
        "defaultValue": "",
        "displayType": "int",
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "index": 2
      },
      {
        "id": "site property",
        "name": "mapred.tasktracker.reduce.tasks.maximum",
        "displayName": "Number of Reduce slots per node",
        "description": "Number of slots that Reduce tasks that run simultaneously can occupy on a TaskTracker.",
        "defaultValue": "",
        "displayType": "int",
        "serviceName": "MAPREDUCE",
        "category": "TaskTracker",
        "index": 3
      },
      {
        "id": "site property",
        "name": "mapred.cluster.reduce.memory.mb",
        "displayName": "Cluster's Reduce slot size (virtual memory)",
        "description": "The virtual memory size of a single Reduce slot in the MapReduce framework",
        "defaultValue": "2048",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 2
      },
      {
        "id": "site property",
        "name": "mapred.job.map.memory.mb",
        "displayName": "Default virtual memory for a job's map-task",
        "description": "Virtual memory for single Map task",
        "defaultValue": "",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "mapred.cluster.max.map.memory.mb",
        "displayName": "Upper limit on virtual memory for single Map task",
        "description": "Upper limit on virtual memory size for a single Map task of any MapReduce job",
        "defaultValue": "",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "mapred.cluster.max.reduce.memory.mb",
        "displayName": "Upper limit on virtual memory for single Reduce task",
        "description": "Upper limit on virtual memory size for a single Reduce task of any MapReduce job",
        "defaultValue": "",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "mapred.job.reduce.memory.mb",
        "displayName": "Default virtual memory for a job's reduce-task",
        "description": "Virtual memory for single Reduce task",
        "defaultValue": "",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 6
      },
      {
        "id": "site property",
        "name": "mapred.cluster.map.memory.mb",
        "displayName": "Cluster's Map slot size (virtual memory)",
        "description": "The virtual memory size of a single Map slot in the MapReduce framework",
        "defaultValue": "1536",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 1
      },
      {
        "id": "site property",
        "name": "io.sort.mb",
        "displayName": "Map-side sort buffer memory",
        "description": "The total amount of Map-side buffer memory to use while sorting files (Expert-only configuration)",
        "defaultValue": "",
        "displayType": "int",
        "unit": "MB",
        "serviceName": "MAPREDUCE",
        "index": 7
      },
      {
        "id": "site property",
        "name": "io.sort.spill.percent",
        "displayName": "Limit on buffer",
        "description": "Percentage of sort buffer used for record collection",
        "defaultValue": "",
        "displayType": "float",
        "serviceName": "MAPREDUCE",
        "index": 8
      },
      {
        "id": "site property",
        "name": "mapred.system.dir",
        "displayName": "MapReduce system directories",
        "description": "Path on the HDFS where where the MapReduce framework stores system files",
        "defaultValue": "/mapred/system",
        "displayType": "directories",
        "serviceName": "MAPREDUCE",
        "category": "Advanced"
      },
      {
        "id": "site property",
        "name": "mapred.userlog.retain.hours",
        "displayName": "Job log retention (hours)",
        "description": "The maximum time, in hours, for which the user-logs are to be retained after the job completion.",
        "defaultValue": "",
        "displayType": "int",
        "unit": "hours",
        "serviceName": "MAPREDUCE",
        "index": 9
      },
      {
        "id": "site property",
        "name": "mapred.jobtracker.maxtasks.per.job",
        "displayName": "Maximum number tasks for a Job",
        "description": "Maximum number of tasks for a single Job",
        "defaultValue": "",
        "displayType": "int",
        "serviceName": "MAPREDUCE",
        "index": 10
      },
      {
        "id": "site property",
        "name": "mapred.hosts",
        "displayName": "mapred.hosts",
        "description": "Names a file that contains the list of nodes that may\
      connect to the jobtracker.  If the value is empty, all hosts are \
      permitted.",
        "defaultValue": "",
        "displayType": "directory",
        "category": "Advanced",
        "serviceName": "MAPREDUCE"
      },
      {
        "id": "site property",
        "name": "mapred.hosts.exclude",
        "displayName": "mapred.hosts.exclude",
        "description": " Names a file that contains the list of hosts that\
      should be excluded by the jobtracker.  If the value is empty, no\
      hosts are excluded.",
        "defaultValue": "",
        "displayType": "directory",
        "category": "Advanced",
        "serviceName": "MAPREDUCE"
      },
      {
        "id": "site property",
        "name": "mapred.child.java.opts",
        "displayName": "Java options for MapReduce tasks",
        "description": "Java options for the TaskTracker child processes.",
        "defaultValue": "",
        "displayType": "advanced",
        "category": "TaskTracker",
        "serviceName": "MAPREDUCE",
        "index": 4
      },

    /**********************************************oozie-site***************************************/
      {
        "id": "site property",
        "name": "oozie.db.schema.name",
        "displayName": "Database Name",
        "description": "Database name used for the Oozie",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "host",
        "isVisible": true,
        "isObserved": true,
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.username",
        "displayName": "Database Username",
        "description": "Database user name to use to connect to the database",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "host",
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.password",
        "displayName": "Database Password",
        "description": "Database password to use to connect to the database",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "password",
        "isVisible": true,
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "filename": "oozie-site.xml",
        "index": 5
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.driver",
        "displayName": "JDBC Driver Class",
        "defaultValue": "",
        "value": "",     // the value is overwritten in code
        "isVisible": true,
        "isOverridable": false,
        "description": "Database name used for the Oozie",
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 6
      },
      {
        "id": "site property",
        "name": "oozie.service.JPAService.jdbc.url",
        "displayName": "Database URL",
        "description": "The JDBC connection URL to the database",
        "defaultValue": "",
        "isOverridable": false,
        "displayType": "advanced",
        "category": "Oozie Server",
        "serviceName": "OOZIE",
        "index": 7
      },

    /**********************************************hive-site***************************************/
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionDriverName",
        "displayName": "JDBC Driver Class",
        "defaultValue": "",
        "value": "",     // the value is overwritten in code
        "isVisible": true,
        "isOverridable": false,
        "description": "Driver class name for a JDBC metastore",
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 7
      },
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionUserName",
        "displayName": "Database Username",
        "description": "Database user name to use to connect to the database",
        "defaultValue": "hive",
        "displayType": "host",
        "isOverridable": false,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionPassword",
        "displayName": "Database Password",
        "description": "Database password to use to connect to the metastore database",
        "defaultValue": "",
        "displayType": "password",
        "isOverridable": false,
        "isVisible": true,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 6
      },
      {
        "id": "site property",
        "name": "javax.jdo.option.ConnectionURL",
        "displayName": "Database URL",
        "value": "",
        "defaultValue": "", // set to a 'jdbc' to not include this in initial error count
        "description": "The JDBC connection URL to the database",
        "displayType": "advanced",
        "isOverridable": false,
        "isVisible": true,
        "category": "Hive Metastore",
        "serviceName": "HIVE",
        "index": 8
      },
      {
        "id": "site property",
        "name": "ambari.hive.db.schema.name",
        "displayName": "Database Name",
        "description": "Database name used as the Hive Metastore",
        "defaultValue": "",
        "isReconfigurable": true,
        "displayType": "host",
        "isOverridable": false,
        "isObserved": true,
        "serviceName": "HIVE",
        "category": "Hive Metastore",
        "index": 4
      },

    /**********************************************hbase-site***************************************/
      {
        "id": "site property",
        "name": "hbase.tmp.dir",
        "displayName": "HBase local directory",
        "description": "Temporary directory on the local filesystem",
        "defaultDirectory": "/hadoop/hbase",
        "defaultValue": "",
        "displayType": "directory",
        "isVisible": true,
        "category": "Advanced",
        "serviceName": "HBASE"

      },
      {
        "id": "site property",
        "name": "hbase.regionserver.global.memstore.upperLimit",
        "displayName": "hbase.regionserver.global.memstore.upperLimit",
        "description": "",
        "defaultValue": "",
        "displayType": "float",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.regionserver.global.memstore.lowerLimit",
        "displayName": "hbase.regionserver.global.memstore.lowerLimit",
        "defaultValue": "",
        "description": "When memstores are being forced to flush to make room in\
      memory, keep flushing until we hit this mark. Defaults to 35% of heap.\
      This value equal to hbase.regionserver.global.memstore.upperLimit causes\
      the minimum possible flushing to occur when updates are blocked due to\
      memstore limiting.",
        "displayType": "float",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.hstore.blockingStoreFiles",
        "displayName": "hstore blocking storefiles",
        "description": "If more than this number of StoreFiles in any one Store (one StoreFile is written per flush of " +
            "MemStore) then updates are blocked for this HRegion until a compaction is completed, or until " +
            "hbase.hstore.blockingWaitTime has been exceeded.",
        "defaultValue": "",
        "isRequired": true,
        "displayType": "int",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.hstore.compactionThreshold",
        "displayName": "HBase HStore compaction threshold",
        "description": "If more than this number of HStoreFiles in any one HStore then a compaction is run to rewrite all HStoreFiles files as one.",
        "defaultValue": "3",
        "displayType": "int",
        "category": "General",
        "serviceName": "HBASE",
        "index": 0
      },
      {
        "id": "site property",
        "name": "hfile.block.cache.size",
        "displayName": "HFile block cache size ",
        "description": "Percentage of maximum heap (-Xmx setting) to allocate to block cache used by HFile/StoreFile. Set to 0 to disable but it's not recommended.",
        "defaultValue": "0.40",
        "displayType": "float",
        "category": "General",
        "serviceName": "HBASE",
        "index": 1
      },
      {
        "id": "site property",
        "name": "hbase.hregion.max.filesize",
        "displayName": "Maximum HStoreFile Size",
        "description": "If any one of a column families' HStoreFiles has grown to exceed this value, the hosting HRegion is split in two.",
        "defaultValue": "",
        "displayType": "int",
        "unit": "bytes",
        "isVisible": true,
        "category": "General",
        "serviceName": "HBASE",
        "index": 2
      },
      {
        "id": "site property",
        "name": "hbase.regionserver.handler.count",
        "displayName": "HBase RegionServer Handler",
        "description": "Count of RPC Listener instances spun up on RegionServers",
        "defaultValue": "60",
        "displayType": "int",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 2
      },
      {
        "id": "site property",
        "name": "hbase.hregion.majorcompaction",
        "displayName": "HBase Region Major Compaction",
        "description": "The time between major compactions of all HStoreFiles in a region. Set to 0 to disable automated major compactions.",
        "defaultValue": "",
        "displayType": "int",
        "unit": "ms",
        "isVisible": true,
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "hbase.hregion.memstore.block.multiplier",
        "displayName": "HBase Region Block Multiplier",
        "description": "Block updates if memstore has \"Multiplier * HBase Region Memstore Flush Size\" bytes. Useful preventing runaway memstore during spikes in update traffic",
        "defaultValue": "",
        "displayType": "int",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "hbase.hregion.memstore.mslab.enabled",
        "displayName": "hbase.hregion.memstore.mslab.enabled",
        "description": "Enables the MemStore-Local Allocation Buffer,\
      a feature which works to prevent heap fragmentation under\
      heavy write loads. This can reduce the frequency of stop-the-world\
      GC pauses on large heaps.",
        "defaultValue": "",
        "displayType": "checkbox",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "hbase.hregion.memstore.flush.size",
        "displayName": "HBase Region Memstore Flush Size",
        "description": "Memstore will be flushed to disk if size of the memstore exceeds this number of bytes.",
        "defaultValue": "",
        "displayType": "int",
        "unit": "bytes",
        "category": "RegionServer",
        "serviceName": "HBASE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "hbase.client.scanner.caching",
        "displayName": "HBase Client Scanner Caching",
        "description": "Number of rows that will be fetched when calling next on a scanner if it is not served from \
      (local, client) memory. Do not set this value such that the time between invocations is greater than the scanner timeout",
        "defaultValue": "",
        "displayType": "int",
        "unit": "rows",
        "isVisible": true,
        "category": "General",
        "serviceName": "HBASE",
        "index": 3
      },
      {
        "id": "site property",
        "name": "zookeeper.session.timeout",
        "displayName": "Zookeeper timeout for HBase Session",
        "description": "HBase passes this to the zk quorum as suggested maximum time for a session",
        "defaultValue": "",
        "displayType": "int",
        "unit": "ms",
        "category": "General",
        "serviceName": "HBASE",
        "index": 4
      },
      {
        "id": "site property",
        "name": "hbase.client.keyvalue.maxsize",
        "displayName": "HBase Client Maximum key-value Size",
        "description": "Specifies the combined maximum allowed size of a KeyValue instance. It should be set to a fraction of the maximum region size.",
        "defaultValue": "",
        "displayType": "int",
        "unit": "bytes",
        "category": "General",
        "serviceName": "HBASE",
        "index": 5
      },
      {
        "id": "site property",
        "name": "dfs.client.read.shortcircuit",
        "displayName": "HDFS Short-circuit read",
        "description": "This configuration parameter turns on short-circuit local reads.",
        "defaultValue": "",
        "displayType": "checkbox",
        "category": "Advanced",
        "serviceName": "HBASE"
      },
      {
        "id": "site property",
        "name": "dfs.support.append",
        "displayName": "HDFS append support",
        "description": "HDFS append support",
        "defaultValue": "",
        "displayType": "checkbox",
        "serviceName": "HBASE",
        "category": "Advanced"
      },
      {
        "id": "site property",
        "name": "hbase.zookeeper.quorum",
        "displayName": "hbase.zookeeper.quorum",
        "defaultValue": "",
        "displayType": "multiLine",
        "serviceName": "HBASE",
        "category": "Advanced"
      },

    /**********************************************webhcat-site***************************************/
      {
        "id": "site property",
        "name": "templeton.zookeeper.hosts",
        "displayName": "templeton.zookeeper.hosts",
        "defaultValue": "",
        "displayType": "multiLine",
        "serviceName": "WEBHCAT",
        "category": "Advanced"
      },
    /**********************************************log4j.properties***************************************/

      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "HDFS",
        "filename": "hdfs-log4j.xml",
        "category": "AdvancedHDFSLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "MAPREDUCE",
        "filename": "mapreduce-log4j.xml",
        "category": "AdvancedMapredLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "HBASE",
        "filename": "hbase-log4j.xml",
        "category": "AdvancedHbaseLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "HIVE",
        "filename": "hive-exec-log4j.xml",
        "category": "AdvancedHiveExecLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "HIVE",
        "filename": "hive-log4j.xml",
        "category": "AdvancedHiveLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "OOZIE",
        "filename": "oozie-log4j.xml",
        "category": "AdvancedOozieLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "ZOOKEEPER",
        "filename": "zookeeper-log4j.xml",
        "category": "AdvancedZooLog4j"
      },
      {
        "id": "site property",
        "name": "content",
        "displayName": "content",
        "value": "",
        "defaultValue": "",
        "description": "log4j properties",
        "displayType": "content",
        "isOverridable": true,
        "isRequired": false,
        "isVisible": true,
        "showLabel": false,
        "serviceName": "PIG",
        "filename": "pig-log4j.xml",
        "category": "AdvancedPigLog4j"
      },


      //***************************************** GLUSTERFS stack********************************************
      {
        "id": "site property",
        "name": "fs.glusterfs.impl",
        "displayName": "GlusterFS fs impl",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.glusterfs.volname",
        "displayName": "GlusterFS volume name",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.glusterfs.mount",
        "displayName": "GlusterFS mount point",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      },
      {
        "id": "site property",
        "name": "fs.glusterfs.getfattrcmd",
        "displayName": "GlusterFS getfattr command",
        "displayType": "string",
        "filename": "core-site.xml",
        "serviceName": "GLUSTERFS",
        "category": "General"
      }

    ]
  };
  
});
window.require.register("data/statusCodes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  module.exports = {
  	200: function () {
  		console.log("Status code 200: Success.");
    },
    202: function () {
      console.log("Status code 202: Success for creation.");
    },
  	400: function () {
  		console.log("Error code 400: Bad Request.");
  	},
  	401: function () {
  		console.log("Error code 401: Unauthorized.");
  	},
  	402: function () {
  		console.log("Error code 402: Payment Required.");
  	},
  	403: function () {
  		console.log("Error code 403: Forbidden.");
      App.router.logOff();
  	},
  	404: function () {
  		console.log("Error code 404: URI not found.");
  	},
  	500: function () {
  		console.log("Error code 500: Internal Error on server side.");
  	},
  	501: function () {
  		console.log("Error code 501: Not implemented yet.");
  	},
  	502: function () {
  		console.log("Error code 502: Services temporarily overloaded.");
  	},
  	503: function () {
  		console.log("Error code 503: Gateway timeout.");
  	}
  };
  
});
window.require.register("ember", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports=Ember;
});
window.require.register("initialize", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  window.App = require('app');

  require('config');

  require('messages');
  require('utils/base64');
  require('utils/db');
  require('utils/helper');
  require('utils/config');
  require('mixins');
  require('models');
  require('controllers');
  require('templates');
  require('views');
  require('router');

  require('utils/ajax');
  require('utils/updater');

  require('mappers/server_data_mapper');
  require('mappers/status_mapper');
  require('mappers/hosts_mapper');
  require('mappers/cluster_mapper');
  require('mappers/jobs_mapper');
  require('mappers/runs_mapper');
  require('mappers/racks_mapper');
  require('mappers/users_mapper');
  require('mappers/service_mapper');
  require('mappers/service_metrics_mapper');
  require('mappers/target_cluster_mapper');
  require('mappers/dataset_mapper');
  require('mappers/component_config_mapper');
  require('mappers/jobs/hive_jobs_mapper');
  require('mappers/jobs/hive_job_mapper');
  require('utils/http_client');
  require('utils/host_progress_popup');

  App.initialize();

  console.log('after initialize');
  console.log('TRACE: app.js-> localStorage:Ambari.authenticated=' + localStorage.getItem('Ambari' + 'authenticated'));
  console.log('TRACE: app.js-> localStorage:currentStep=' + localStorage.getItem(App.get('router').getLoginName() + 'Installer' + 'currentStep'));
  console.log('TRACE: app.js-> router.authenticated=' + App.get('router.loggedIn'));
  
});
window.require.register("mappers/cluster_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  App.clusterMapper = App.QuickDataMapper.create({
      model : App.Cluster,
      map:function(json){
        if(!this.get('model')) {return;}
        if(json){
          var result = json;
          result = this.parseIt(result, this.config);
          App.store.load(this.get('model'), result);
          
          var cluster = App.Cluster.find(result.id);
          var clusterDesiredConfigs = [];
          // Create desired_configs_array
          if(json.Clusters.desired_configs){
            for(var site in json.Clusters.desired_configs){
              var tag = json.Clusters.desired_configs[site].tag;
              var configObj = App.ConfigSiteTag.create({
                site: site,
                tag: tag,
                hostOverrides: {}
              });
              if(json.Clusters.desired_configs[site].host_overrides!=null){
                var hostOverridesArray = {};
                json.Clusters.desired_configs[site].host_overrides.forEach(function(override){
                  var hostname = override.host_name;
                  hostOverridesArray[hostname] = override.tag;
                });
                configObj.set('hostOverrides', hostOverridesArray);
              }
              clusterDesiredConfigs.push(configObj);
            }
          }
          cluster.set('desiredConfigs', clusterDesiredConfigs);
        }
      },
      config : {
        id:'Clusters.cluster_id',
        cluster_name: 'Clusters.cluster_name',
        stack_name: 'Clusters.stack_name',
        version: 'Clusters.version',
        //$hosts: [1, 2, 3, 4],
        $racks: [1],
        max_hosts_per_rack: 'Clusters.max_hosts_per_rack'
      }
  });
  
});
window.require.register("mappers/component_config_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * The usage of previousResponse is due to detect
   * which exactly components has stale_configs changed in comparison to previous response
   */
  var previousResponse = {};

  App.componentConfigMapper = App.QuickDataMapper.create({
    map: function (json) {
      console.time('App.componentConfigMapper execution time');
      if (json.items) {
        var hostComponentRecordsMap = App.cache['hostComponentRecordsMap'];
        var staleConfigsTrue = [];
        var currentResponse = {};
        json.items.forEach(function (component) {
          var id = component.HostRoles.component_name + "_" + component.HostRoles.host_name;
          if (previousResponse[id]) {
            delete previousResponse[id];
          } else {
            staleConfigsTrue.push(id);
          }
          currentResponse[id] = true;
        });

        /**
         * if stale_configs of components became
         * true:
         *  then they will be in "staleConfigsTrue" object
         * false:
         *  then they will be in "previousResponse" object
         * if stale_configs haven't changed then both objects will be empty and components stay the same
         */
        staleConfigsTrue.forEach(function (id) {
          hostComponentRecordsMap[id].set('staleConfigs', true);
        });
        for (var id in previousResponse) {
          hostComponentRecordsMap[id].set('staleConfigs', false)
        }
        previousResponse = currentResponse;
      }
      console.timeEnd('App.componentConfigMapper execution time');
    }
  });
  
});
window.require.register("mappers/dataset_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.dataSetMapper = App.QuickDataMapper.create({
    model: App.Dataset,
    Jobs_model: App.DataSetJob,
    config: {
      id: 'name',
      name: 'name',
      status: 'status',
      source_cluster_name: 'sourceClusterName',
      target_cluster_name: 'targetClusterName',
      source_dir: 'sourceDir',
      target_dir: 'targetDir',
      frequency: 'frequency',
      frequency_unit: 'frequencyUnit',
      schedule_start_date: 'scheduleStartDate',
      schedule_end_date: 'scheduleEndDate',
      dataset_jobs_key: 'instances',
      dataset_jobs_type: 'array',
      dataset_jobs: {
        item: 'id'
      }
    },
    jobs_config: {
      id: 'id',
      name: 'name',
      status : 'status',
      start_date: 'startTime',
      end_date: 'endTime',
      dataset_id: 'dataset'
    },

    map: function (json) {
      if (!this.get('model')) {
        return;
      }
      if (json && json.length > 0) {
        var datasetResults = [];
        var dataSetJobResults = [];
        json.forEach(function (item) {
          item.instances.forEach(function (job) {
            var newInstance = this.parseIt(job, this.get('jobs_config'));
            dataSetJobResults.push(newInstance);
          }, this);
          var newitem = this.parseIt(item, this.get('config'));
              datasetResults.push(newitem);
        }, this);
        App.store.loadMany(this.get('Jobs_model'), dataSetJobResults);
        App.store.loadMany(this.get('model'), datasetResults);
      }
    }
  });
  
});
window.require.register("mappers/hosts_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  var stringUtils = require('utils/string_utils');
  /**
   * previousResponse contains only mutable properties of host
   * unlike App.cache.Hosts, which contains immutable properties
   */
  var previousResponse = {};

  App.hostsMapper = App.QuickDataMapper.create({

    model: App.Host,
    config: {
      id: 'Hosts.host_name',
      host_name: 'Hosts.host_name',
      public_host_name: 'Hosts.public_host_name',
      cluster_id: 'Hosts.cluster_name',// 1
      rack: 'Hosts.rack_info',
      host_components_key: 'host_components',
      host_components_type: 'array',
      host_components: {
        item: 'id'
      },
      critical_alerts_count: 'critical_alerts_count',
      cpu: 'Hosts.cpu_count',
      cpu_physical: 'Hosts.ph_cpu_count',
      memory: 'Hosts.total_mem',
      disk_info: 'Hosts.disk_info',
      disk_total: 'metrics.disk.disk_total',
      disk_free: 'metrics.disk.disk_free',
      health_status: 'Hosts.host_status',
      load_one: 'metrics.load.load_one',
      load_five: 'metrics.load.load_five',
      load_fifteen: 'metrics.load.load_fifteen',
      cpu_system: 'metrics.cpu.cpu_system',
      cpu_user: 'metrics.cpu.cpu_user',
      mem_total: 'metrics.memory.mem_total',
      mem_free: 'metrics.memory.mem_free',
      last_heart_beat_time: "Hosts.last_heartbeat_time",
      os_arch: 'Hosts.os_arch',
      os_type: 'Hosts.os_type',
      ip: 'Hosts.ip',
      passive_state: 'Hosts.maintenance_state'
    },
    map: function (json) {
      console.time('App.hostsMapper execution time');
      if (json.items) {
        var hostsWithFullInfo = [];
        var modifiedHosts = [];
        var hostIds = {};
        var cacheData = App.cache['Hosts'];
        var currentHostStatuses = {};
        var hostsData = {};
        var isModelLoaded = false;

        json.items.forEach(function (item) {
          var hostName = item.Hosts.host_name;
          //receive host_components when hosts were added
          item.host_components = item.host_components || [];
          item.host_components.forEach(function (host_component) {
            host_component.id = host_component.HostRoles.component_name + "_" + hostName;
          }, this);

          hostIds[hostName] = true;
          currentHostStatuses[hostName] = item.Hosts.host_status;

          var hostCache = cacheData[hostName];
          item.critical_alerts_count = hostCache.critical_alerts_count;
          var parsedItem = this.parseIt(item, this.config);

          hostsWithFullInfo.push(parsedItem);
          if (hostCache) {
            if (hostCache.is_modified) {
              modifiedHosts.push(parsedItem);
              delete hostCache.is_modified;
            } else {
              hostsData[hostName] = this.getDiscrepancies(parsedItem, previousResponse[hostName]);
            }
          }
          previousResponse[hostName] = parsedItem;
        }, this);

        App.cache['previousHostStatuses'] = currentHostStatuses;
        hostsWithFullInfo = hostsWithFullInfo.sortProperty('public_host_name');

        var clientHosts = App.Host.find();

        if (clientHosts) {
          // hosts were added
          if (clientHosts.get('length') < hostsWithFullInfo.length) {
            hostsWithFullInfo.forEach(function (host) {
              cacheData[host.id] = {
                ip: host.ip,
                os_arch: host.os_arch,
                os_type: host.os_type,
                public_host_name: host.public_host_name,
                memory: host.memory,
                cpu: host.cpu,
                cpu_physical: host.cpu_physical,
                host_components: host.host_components
              };
            });
            App.store.loadMany(this.get('model'), hostsWithFullInfo);
            isModelLoaded = true;
          } else {
            //restore properties from cache instead of asking them from server
            modifiedHosts.forEach(function (host) {
              var cacheHost = cacheData[host.id];
              if (cacheHost) {
                host.ip = cacheHost.ip;
                host.os_arch = cacheHost.os_arch;
                host.os_type = cacheHost.os_type;
                host.public_host_name = cacheHost.public_host_name;
                host.memory = cacheHost.memory;
                host.cpu = cacheHost.cpu;
                host.cpu_physical = cacheHost.cpu_physical;
                host.host_components = cacheHost.host_components;
              }
            });
            App.store.loadMany(this.get('model'), modifiedHosts);
          }
          // hosts were deleted
          if (clientHosts.get('length') > hostsWithFullInfo.length) {
            clientHosts.forEach(function (host) {
              if (host && !hostIds[host.get('hostName')]) {
                // Delete old ones as new ones will be
                // loaded by loadMany().
                this.deleteRecord(host);
                delete cacheData[host.get('id')];
              }
            }, this);
          }
        }
        if (!isModelLoaded) {
          App.Host.find().forEach(function (_host) {
            var hostData = hostsData[_host.get('id')];
            if (hostData) {
              for (var i in hostData) {
                _host.set(stringUtils.underScoreToCamelCase(i), hostData[i]);
              }
            }
          });
        }
      }
      console.timeEnd('App.hostsMapper execution time');
      if (!App.router.get('clusterController.dataLoadList.serviceMetrics')) {
        App.router.get('clusterController').deferServiceMetricsLoad();
      }
    },
    /**
     * check mutable fields whether they have been changed and if positive
     * return host object only with properties, that contains new value
     * @param current
     * @param previous
     * @return {*}
     */
    getDiscrepancies: function (current, previous) {
      var result = {};
      var fields = Em.A(['disk_total', 'disk_free', 'health_status', 'load_one', 'cpu_system', 'cpu_user', 'mem_total', 'mem_free', 'critical_alerts_count', 'passive_state']);
      if (previous) {
        fields.forEach(function (field) {
          if (current[field] != previous[field]) result[field] = current[field];
        });
        if (JSON.stringify(current.disk_info) !== JSON.stringify(previous.disk_info)) {
          result.disk_info = current.disk_info;
        }
        if (JSON.stringify(current.host_components) !== JSON.stringify(previous.host_components)) {
          result.host_components = current.host_components;
        }
        result.last_heart_beat_time = current.last_heart_beat_time;
        return result;
      }
      return current;
    }

  });
  
});
window.require.register("mappers/jobs/hive_job_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Maps a Hive job response from server into an existing Hive Job.
   *
   * This will only update an existing Hive Job and not create a new one. It will
   * populate all fields (stages, Tez DAG, etc.) except runtime information like
   * <ul>
   * <li>tezDag.id
   * <li>tezDag.vertices.state
   * <li>tezDag.vertices.startTime
   * <li>tezDag.vertices.endTime
   * <li>tezDag.vertices.tasksCount
   * <li>tezDag.vertices.file*
   * <li>tezDag.vertices.hdfs*
   * <li>tezDag.vertices.recordReadCount
   * <li>tezDag.vertices.recordWriteCount
   * </ul>
   */
  App.hiveJobMapper = App.QuickDataMapper.create({
    model : App.HiveJob,
    map : function(json) {
      var model = this.get('model');
      if (!model) {
        return;
      }
      var hiveJob = {};
      if (json && json.entity) {
        var hiveJob = {};
        hiveJob.id = json.entity;
        hiveJob.name = hiveJob.id;
        hiveJob.startTime = json.starttime;
        hiveJob.endTime = json.endtime;
        json.otherinfo.query = $.parseJSON(json.otherinfo.query);
        if (json.otherinfo.query && json.otherinfo.query.queryText) {
          hiveJob.query_text = json.otherinfo.query.queryText;
        }
        hiveJob.stages = [];
        var stagePlans = json.otherinfo.query.queryPlan["STAGE PLANS"];
        for ( var stage in stagePlans) {
          var stageValue = stagePlans[stage];
          var stageItem = {};
          stageItem.id = stage;
          stageItem.description = '. ';
          for (var item in stageValue) {
            stageItem.description += item;
          };
          hiveJob.stages.push(stageItem);
          if (stageValue.Tez != null && hiveJob.tezDag == null) {
            var dagName = stageValue.Tez['DagName:'];
            // Vertices
            var vertices = [];
            var vertexIds = [];
            var vertexIdMap = {};
            for ( var vertexName in stageValue.Tez["Vertices:"]) {
              var vertex = stageValue.Tez["Vertices:"][vertexName];
              var vertexObj = {
                id : dagName + "/" + vertexName,
                name : vertexName,
                incoming_edges : [],
                outgoing_edges : []
              };
              vertexIds.push(vertexObj.id);
              var operatorExtractor = function(obj) {
                var ops = [];
                if ($.isArray(obj)) {
                  obj.forEach(function(o) {
                    ops = ops.concat(operatorExtractor(o));
                  });
                } else {
                  for ( var key in obj) {
                    ops.push(key);
                    if (obj[key].children != null) {
                      ops = ops.concat(operatorExtractor(obj[key].children));
                    }
                  }
                }
                return ops;
              }
              if (vertex["Map Operator Tree:"] != null) {
                vertexObj.type = App.TezDagVertexType.MAP;
                vertexObj.operations = operatorExtractor(vertex["Map Operator Tree:"]);
                vertexObj.operation_plan = JSON.stringify(vertex["Map Operator Tree:"], undefined, "  ");
              } else if (vertex["Reduce Operator Tree:"] != null) {
                vertexObj.type = App.TezDagVertexType.REDUCE;
                vertexObj.operations = operatorExtractor(vertex["Reduce Operator Tree:"]);
                vertexObj.operation_plan = JSON.stringify(vertex["Reduce Operator Tree:"], undefined, "  ");
              } else if (vertex["Vertex:"] != null && vertexName==vertex['Vertex:']) {
                vertexObj.type = App.TezDagVertexType.UNION;
              }
              vertexIdMap[vertexObj.id] = vertexObj;
              vertices.push(vertexObj);
            }
            // Edges
            var edges = [];
            var edgeIds = [];
            for ( var childVertex in stageValue.Tez["Edges:"]) {
              var childVertices = stageValue.Tez["Edges:"][childVertex];
              if (!$.isArray(childVertices)) {
                // Single edge given as object instead of array
                childVertices = [ childVertices ];
              }
              childVertices.forEach(function(e) {
                var parentVertex = e.parent;
                if (e.type == 'CONTAINS') {
                  var parentVertexNode = vertexIdMap[dagName + "/" + parentVertex];
                  if (parentVertexNode != null && parentVertexNode.type == App.TezDagVertexType.UNION) {
                    // We flip the edges for Union vertices
                    var tmp = childVertex;
                    childVertex = parentVertex;
                    parentVertex = tmp;
                  }
                }
                var edgeObj = {
                  id : dagName + "/" + parentVertex + "-" + childVertex,
                  from_vertex_id : dagName + "/" + parentVertex,
                  to_vertex_id : dagName + "/" + childVertex
                };
                vertexIdMap[edgeObj.from_vertex_id].outgoing_edges.push(edgeObj.id);
                vertexIdMap[edgeObj.to_vertex_id].incoming_edges.push(edgeObj.id);
                edgeIds.push(edgeObj.id);
                switch (e.type) {
                case "BROADCAST_EDGE":
                  edgeObj.edge_type = App.TezDagEdgeType.BROADCAST;
                  break;
                case "SIMPLE_EDGE":
                  edgeObj.edge_type = App.TezDagEdgeType.SCATTER_GATHER;
                  break;
                case "CONTAINS":
                  edgeObj.edge_type = App.TezDagEdgeType.CONTAINS;
                  break;
                default:
                  break;
                }
                edges.push(edgeObj);
              });
            }
            // Create records
            var tezDag = {
              id : dagName,
              name : dagName,
              stage : stage,
              vertices : vertexIds,
              edges : edgeIds
            }
            // Once the DAG is loaded, we do not need to
            // reload as the structure does not change. Reloading
            // here causes missing data (got from other calls)
            // to propagate into UI - causing flashing.
            var newVertices = [];
            var newEdges = [];
            vertices.forEach(function(v) {
              var vertexRecord = App.TezDagVertex.find(v.id);
              if (!vertexRecord.get('isLoaded')) {
                newVertices.push(v);
              }
            });
            edges.forEach(function(e) {
              var edgeRecord = App.TezDagEdge.find(e.id);
              if (!edgeRecord.get('isLoaded')) {
                newEdges.push(e);
              }
            });
            App.store.loadMany(App.TezDagVertex, newVertices);
            App.store.loadMany(App.TezDagEdge, newEdges);
            var dagRecord = App.TezDag.find(tezDag.id);
            if (!dagRecord.get('isLoaded')) {
              App.store.load(App.TezDag, tezDag);
            }
            hiveJob.tezDag = tezDag.id;
          }
        }
        if(App.HiveJob.find().get('content').length == 0){
          App.store.load(model, hiveJob);
        }
        var hiveJobRecord = App.HiveJob.find(hiveJob.id);
        if (hiveJobRecord != null) {
          hiveJobRecord.set('stages', hiveJob.stages.sortProperty('id'));
          hiveJobRecord.set('startTime', hiveJob.startTime);
          hiveJobRecord.set('endTime', hiveJob.endTime);
          if (hiveJob.tezDag != null) {
            // Some hive queries dont use Tez
            hiveJobRecord.set('tezDag', App.TezDag.find(hiveJob.tezDag));
          }
        }
      }
    },
    config : {}
  });
});
window.require.register("mappers/jobs/hive_jobs_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.hiveJobsMapper = App.QuickDataMapper.create({
    model : App.HiveJob,
    map : function(json) {
      var model = this.get('model');
      if (!model) {
        return;
      }
      var hiveJobs = []
      if (json) {
        if(!json.entities) {
          json.entities = [];
          if(json.entity){
            json.entities = [json];
          }
        }
        var currentEntityMap = {};
        json.entities.forEach(function(entity) {
          currentEntityMap[entity.entity] = entity.entity;
          var hiveJob = {
            id : entity.entity,
            name : entity.entity,
            user : entity.primaryfilters.user
          }
          hiveJob.has_tez_dag = false;
          hiveJob.query_text = '';
          if (entity.otherinfo && entity.otherinfo.query) {
            // Explicit false match needed for when failure hook not set
            hiveJob.failed = entity.otherinfo.status===false;
            hiveJob.has_tez_dag = entity.otherinfo.query.match("\"Tez\".*\"DagName:\"");
            var queryJson = $.parseJSON(entity.otherinfo.query);
            if (queryJson && queryJson.queryText) {
              hiveJob.query_text = queryJson.queryText;
            }
          }
          if (entity.events != null) {
            entity.events.forEach(function(event) {
              switch (event.eventtype) {
              case "QUERY_SUBMITTED":
                hiveJob.start_time = event.timestamp;
                break;
              case "QUERY_COMPLETED":
                hiveJob.end_time = event.timestamp;
                break;
              default:
                break;
              }
            });
          }
          if (!hiveJob.start_time && entity.starttime > 0) {
            hiveJob.start_time = entity.starttime;
          }
          if (!hiveJob.end_time && entity.endtime > 0) {
            hiveJob.end_time = entity.endtime;
          }
          hiveJobs.push(hiveJob);
          hiveJob = null;
          entity = null;
        });

        if(hiveJobs.length > App.router.get('mainJobsController.filterObject.jobsLimit')) {
          var lastJob = hiveJobs.pop();
          if(App.router.get('mainJobsController.navIDs.nextID') != lastJob.id) {
            App.router.set('mainJobsController.navIDs.nextID', lastJob.id);
          }
          currentEntityMap[lastJob.id] = null;
        }

        // Delete IDs not seen from server
        var hiveJobsModel = model.find().toArray();
        hiveJobsModel.forEach(function(job) {
          if (job && !currentEntityMap[job.get('id')]) {
            this.deleteRecord(job);
          }
        }, this);
      }
      App.store.loadMany(model, hiveJobs);
      json = null;
      hiveJobs = null;
    },
    config : {}
  });
  
});
window.require.register("mappers/jobs_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var fillEmptyValues = function(self, obj) {
    $.each(self.config, function(field, value) {
      if (obj[value].length == 0) {
        obj[value].push({x: 0, y: 0});
      };
    });
  };

  App.jobsMapper = App.QuickDataMapper.create({
    model:App.Job,
    map:function (json) {
      if (!this.get('model')) {
        return;
      }
      if (json.jobs) {
        var result = [];
        json.jobs.forEach(function (item) {
          result.push(this.parseIt(item, this.config));
        }, this);

        var r = Ember.ArrayProxy.create({"content":[]});
        result.forEach(function(item){
          r.content.push(App.Job2.create(item));
        });

        this.set('controller.content.jobs', r.content);
      }
    },
    config:{
      id:'jobId',
      run_id:'workflowId',
      job_name:'jobName',
      workflow_entity_name:'workflowEntityName',
      user_name:'userName',
      submit_time:'submitTime',
      maps:'maps',
      reduces:'reduces',
      status:'status',
      input:'inputBytes',
      output:'outputBytes',
      elapsed_time:'elapsedTime'
    }
  });

  App.jobTimeLineMapper = App.QuickDataMapper.create({
    model: null, //model will be set outside of mapper
    config:{
      map:'map',
      shuffle:'shuffle',
      reduce:'reduce'
    },
    map:function (json) {
      var job = this.get('model'); // @model App.MainAppsItemBarView
      var parseResult = this.parseIt(json, this.config);
      var self = this;
      $.each(parseResult, function (field, value) {
        var d = self.coordinatesModify(value);
        d.reverse();
        d = self.coordinatesModify(d);
        d.reverse();
        job.set(field, d);
      });
      fillEmptyValues(this, job);
    },

    coordinatesModify: function(data) {
      var d = this.zeroAdding(data);
      d.reverse();
      d = this.zeroAdding(d);
      d.reverse();
      return d;
    },

    zeroAdding: function(data) {
      var d = [];
      var last_y = 0;
      data.forEach(function(coordinates) {
        if (coordinates.y != 0 && last_y == 0) {
          d.push({x: coordinates.x, y: 0});
        }
        d.push(coordinates);
        last_y = coordinates.y;
      });
      return d;
    }
  });

  App.taskTimeLineMapper = App.QuickDataMapper.create({
    model: null, //model will be set outside of mapper
    config:{
      allmap:'map',
      allshuffle:'shuffle',
      allreduce:'reduce'
    },
    map:function (json) {
      var job = this.get('model'); // @model App.MainAppsItemBarView
      var parseResult = this.parseIt(json, this.config);

      $.each(parseResult, function (field, value) {
        job.set(field, value);
      });
      fillEmptyValues(this, job);
    }
  });

  App.jobTasksMapper = App.QuickDataMapper.create({
    model: null, //model will be set outside of mapper
    config:{
      mapNodeLocal:'mapNodeLocal',
      mapRackLocal:'mapRackLocal',
      mapOffSwitch:'mapOffSwitch',
      reduceOffSwitch:'reduceOffSwitch',
      submit:'submitTime',
      finish:'finishTime'
    },
    map:function (json) {
      var job = this.get('model'); // @model App.MainAppsItemBarView
      var parseResult = this.parseIt(json, this.config);
      $.each(parseResult, function (field, value) {
        job.set(field, value);
      });
    }
  });
  
});
window.require.register("mappers/racks_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */


  App.racksMapper = App.QuickDataMapper.create({
    model: App.Rack,
    config: {
      id: "Racks.id",
      name: "Racks.name"
      //$hosts: ["host01", "host06", "host05"],
      //status: "Racks.status",
      //live_hosts_count: "Racks.live_hosts_count",
      //critical_hosts_count: "Racks.critical_hosts_count",
      //dead_hosts_count: "Racks.dead_hosts_count"
    }
  });
  
});
window.require.register("mappers/runs_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.runsMapper = App.QuickDataMapper.create({
    model : App.Run,
    map : function(json) {
      if(!this.get('model')) {
        return;
      }
      if(json && json.aaData) {
        var result = [];

        var pagination_info={
          iTotalDisplayRecords :json.iTotalDisplayRecords ,
          iTotalRecords:json.iTotalRecords,
          startIndex:parseInt(json.startIndex)+1,
          endIndex:parseInt(json.endIndex)+1
        };

        json.aaData.forEach(function(item, index) {
          result.push(this.generateWorkflow(item, index));
        }, this);

        var r = [];
        result.forEach(function(item){
          r.push(App.Run2.create(item));
        });

        App.router.get('mainAppsController').set('content', r);
        App.router.get('mainAppsController').set('serverData', json.summary);
        App.router.get('mainAppsController').set('paginationObject', pagination_info);
      }


    },

    generateWorkflow: function(item, index) {
      var o = this.parseIt(item, this.config);

      var r = '{dag: {';
      item.workflowContext.workflowDag.entries.forEach(function(item) {
        r += '"' + item.source + '": [';
        // if a standalone MapReduce job, there won't be any targets
        if (item.targets) {
          item.targets.forEach(function(target) {
            r += '"' + target + '",';
          });
          if(item.targets.length){
            r = r.substr(0, r.length - 1);
          }
        }
        else {
          r += item.source;
        }
        r += '],';
      });
      r = r.substr(0, r.length - 1);
      r += '}}';
      o.workflowContext = r;
      o.index = index + 1;
      return o;
    },

    config : {
      id: 'workflowId',
      appName: 'workflowName',
      numJobsTotal: 'numJobsTotal',
      numJobsCompleted: 'numJobsCompleted',
      userName:'userName',
      startTime: 'startTime',
      elapsedTime: 'elapsedTime',
      input: 'inputBytes',
      output: 'outputBytes'
    }
  });
  
});
window.require.register("mappers/server_data_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * initialize common cache container for mappers
   * App.cache contains shared data, used for syncronizing incoming server data among mappers
   */
  App.cache = {
    'Hosts': {},
    'previousHostStatuses': {},
    'previousComponentStatuses': {},
    'previousComponentPassiveStates': {},
    'hostComponentsOnService': {},
    'services': [],
    'hostComponentRecordsMap': {}
  };

  App.ServerDataMapper = Em.Object.extend({
    jsonKey: false,
    map: function (json) {
      if (json) {
        var model = this.get('model');
        var jsonKey = this.get('jsonKey');

        if (jsonKey && json[jsonKey]) { // if data come as { hdfs: {...} }
          json = json[jsonKey];
        }

        $.each(json, function (field, value) {
          model.set(field, value);
        })
      }
    }
  });


  App.QuickDataMapper = App.ServerDataMapper.extend({
    config: {},
    model: null,
    map: function (json) {
      if (!this.get('model')) {
        return;
      }

      if (json.items) {
        var result = [];

        json.items.forEach(function (item) {
          result.push(this.parseIt(item, this.config));
        }, this);

        App.store.loadMany(this.get('model'), result);
      }
    },

    parseIt: function (data, config) {
      var result = {};
      for ( var i in config) {
        if (i.substr(0, 1) === '$') {
          i = i.substr(1, i.length);
          result[i] = config['$' + i];
        } else {
          var isSpecial = false;
          if (i.substr(-5) == '_type') {
            var prefix = i.substr(0, i.length - 5);
            isSpecial = config[prefix + '_key'] != null;
          } else if (i.substr(-4) == '_key') {
            var prefix = i.substr(0, i.length - 4);
            isSpecial = config[prefix + '_type'] != null;
          }
          if (!isSpecial && typeof config[i] == 'string') {
            result[i] = this.getJsonProperty(data, config[i]);
          } else if (typeof config[i] == 'object') {
            result[i] = [];
            var _data = this.getJsonProperty(data, config[i + '_key']);
            var _type = config[i + '_type'];
            var l = _data.length;
            for ( var index = 0; index < l; index++) {
              if (_type == 'array') {
                result[i].push(this.getJsonProperty(_data[index], config[i].item));
              } else {
                result[i].push(this.parseIt(_data[index], config[i]));
              }
            }
            if(_type == 'array'){
              result[i] = result[i].sort();
            }
          }
        }
      }
      return result;
    },

    getJsonProperty: function (json, path) {
      var pathArr = path.split('.');
      var current = json;
      pathArr = this.filterDotted(pathArr);
      while (pathArr.length && current) {
        if (pathArr[0].substr(-1) == ']') {
          var index = parseInt(pathArr[0].substr(-2, 1));
          var attr = pathArr[0].substr(0, pathArr[0].length - 3);
          if (attr in current) {
            current = current[attr][index];
          }
        } else {
          current = current[pathArr[0]];
        }
        pathArr.splice(0, 1);
      }
      return current;
    },

    filterDotted: function(arr) {
      var buffer = [];
      var dottedBuffer = [];
      var dotted = false;
      arr.forEach(function(item) {
        if(/\['|\["/.test(item)) {
          dottedBuffer.push(item.substr(2, item.length));
          dotted = true;
        } else if (dotted && !/\]'|"\]/.test(item)) {
          dottedBuffer.push(item);
        } else if (/']|"]/.test(item)) {
          dottedBuffer.push(item.substr(0, item.length - 2));
          buffer.push(dottedBuffer.join('.'));
          dotted = false;
          dottedBuffer = [];
        } else {
          buffer.push(item);
        }
      });

      return buffer;
    },

    /**
     * properly delete record from model
     * @param item
     */
    deleteRecord: function (item) {
      item.deleteRecord();
      App.store.commit();
      item.get('stateManager').transitionTo('loading');
      console.log('Record with id:' + item.get('id') + ' was deleted from model');
    },
    /**
     * check mutable fields whether they have been changed and if positive
     * return host object only with properties, that contains new value
     * @param current
     * @param previous
     * @param fields
     * @return {*}
     */
    getDiscrepancies: function (current, previous, fields) {
      var result = {};
      if (previous) {
        fields.forEach(function (field) {
          if (Array.isArray(current[field])) {
            if (JSON.stringify(current[field]) !== JSON.stringify(previous[field])) {
              result[field] = current[field];
              result.isLoadNeeded = true;
            }
          } else {
            if (current[field] != previous[field]) result[field] = current[field];
          }
        });
        return result;
      }
      return current;
    },

    calculateState: function (json) {
  //    var stateEqual = (json.desired_status != json.work_status);
  //    if (stateEqual) {
  //      if (json.desired_status == 'STARTED' && json.work_status == 'INSTALLED') {
  //        json.work_status = 'STARTING';
  //      } else if (json.desired_status == 'INSTALLED' && json.work_status == 'STARTED') {
  //        json.work_status = 'STOPPING';
  //      }
  //    }
      return json;
    }
  });

  App.QuickDataMapper.componentServiceMap = function () {
    return {
      'NAMENODE': 'HDFS',
      'SECONDARY_NAMENODE': 'HDFS',
      'DATANODE': 'HDFS',
      'HDFS_CLIENT': 'HDFS',
      'JOURNALNODE': 'HDFS',
      'ZKFC': 'HDFS',
      'JOBTRACKER': 'MAPREDUCE',
      'TASKTRACKER': 'MAPREDUCE',
      'MAPREDUCE_CLIENT': 'MAPREDUCE',
      'MAPREDUCE2_CLIENT': 'MAPREDUCE2',
      'HISTORYSERVER': App.get('isHadoop2Stack') ? 'MAPREDUCE2' : 'MAPREDUCE',
      'TEZ_CLIENT': 'TEZ',
      'RESOURCEMANAGER': 'YARN',
      'YARN_CLIENT': 'YARN',
      'NODEMANAGER': 'YARN',
      'APP_TIMELINE_SERVER': 'YARN',
      'ZOOKEEPER_SERVER': 'ZOOKEEPER',
      'ZOOKEEPER_CLIENT': 'ZOOKEEPER',
      'HBASE_MASTER': 'HBASE',
      'HBASE_REGIONSERVER': 'HBASE',
      'HBASE_CLIENT': 'HBASE',
      'PIG': 'PIG',
      'SQOOP': 'SQOOP',
      'OOZIE_SERVER': 'OOZIE',
      'OOZIE_CLIENT': 'OOZIE',
      'HIVE_SERVER': 'HIVE',
      'HIVE_METASTORE': 'HIVE',
      'HIVE_CLIENT': 'HIVE',
      'MYSQL_SERVER': 'HIVE',
      'HCAT': 'HCATALOG',
      'WEBHCAT_SERVER': 'WEBHCAT',
      'NAGIOS_SERVER': 'NAGIOS',
      'GANGLIA_SERVER': 'GANGLIA',
      'GANGLIA_MONITOR': 'GANGLIA',
      'KERBEROS_SERVER': 'KERBEROS',
      'KERBEROS_ADMIN_CLIENT': 'KERBEROS',
      'KERBEROS_CLIENT': 'KERBEROS',
      'HUE_SERVER': 'HUE',
      'GLUSTERFS_CLIENT': 'GLUSTERFS',
      'FALCON_SERVER': 'FALCON',
      'FALCON_CLIENT': 'FALCON',
      'NIMBUS': 'STORM',
      'SUPERVISOR': 'STORM',
      'STORM_UI_SERVER': 'STORM',
      'DRPC_SERVER': 'STORM',
      'STORM_REST_API': 'STORM'
    }
  };
  
});
window.require.register("mappers/service_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.serviceMapper = App.QuickDataMapper.create({
    map: function (json) {
      console.time("App.serviceMapper execution time");

      var servicesData = [];
      json.items.forEach(function (service) {
        var serviceData = {
          ServiceInfo: {
            service_name: service.ServiceInfo.service_name,
            state: service.ServiceInfo.state,
            passive_state: service.ServiceInfo.maintenance_state
          },
          host_components: [],
          components: []
        };
        //check whether Nagios installed and started
        if (service.alerts) {
          serviceData.ServiceInfo.critical_alerts_count = service.alerts.summary.CRITICAL + service.alerts.summary.WARNING;
        }
        servicesData.push(serviceData);
      });

      App.cache['services'] = servicesData;

      console.timeEnd("App.serviceMapper execution time");
    }
  });
  
});
window.require.register("mappers/service_metrics_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var misc = require('utils/misc');
  var stringUtils = require('utils/string_utils');
  var previousResponse = [];

  App.serviceMetricsMapper = App.QuickDataMapper.create({

    model: App.Service,
    config: {
      id: 'ServiceInfo.service_name',
      service_name: 'ServiceInfo.service_name',
      work_status: 'ServiceInfo.state',
      passive_state: 'ServiceInfo.passive_state',
      critical_alerts_count: 'ServiceInfo.critical_alerts_count',
      $rand: Math.random(),
      $alerts: [ 1, 2, 3 ],
      host_components: 'host_components',
      tool_tip_content: 'tool_tip_content'
    },
    hdfsConfig: {
      version: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.Version',
      name_node_id: 'nameNodeComponent.host_components[0].HostRoles.host_name',
      sname_node_id: 'snameNodeComponent.host_components[0].HostRoles.host_name',
      active_name_node_id: 'active_name_node_id',
      standby_name_node_id: 'standby_name_node_id',
      standby_name_node2_id: 'standby_name_node2_id',
      journal_nodes: 'journal_nodes',
      name_node_start_time: 'nameNodeComponent.host_components[0].metrics.runtime.StartTime',
      jvm_memory_heap_used: 'nameNodeComponent.host_components[0].metrics.jvm.HeapMemoryUsed',
      jvm_memory_heap_max: 'nameNodeComponent.host_components[0].metrics.jvm.HeapMemoryMax',
      decommission_data_nodes: 'decommission_data_nodes',
      capacity_used: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CapacityUsed',
      capacity_total: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CapacityTotal',
      capacity_remaining: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CapacityRemaining',
      dfs_total_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.BlocksTotal',
      dfs_corrupt_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CorruptBlocks',
      dfs_missing_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.MissingBlocks',
      dfs_under_replicated_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.UnderReplicatedBlocks',
      dfs_total_files: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.TotalFiles',
      upgrade_status: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.UpgradeFinalized',
      safe_mode_status: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.Safemode',
      name_node_cpu: 'nameNodeComponent.host_components[0].metrics.cpu.cpu_wio',
      name_node_rpc: 'nameNodeComponent.host_components[0].metrics.rpc.RpcQueueTime_avg_time'
    },
    yarnConfig: {
      version: 'resourceManagerComponent.ServiceComponentInfo.Version',
      resource_manager_node_id: 'resourceManagerComponent.host_components[0].HostRoles.host_name',
      resource_manager_start_time: 'resourceManagerComponent.ServiceComponentInfo.StartTime',
      jvm_memory_heap_used: 'resourceManagerComponent.host_components[0].metrics.jvm.HeapMemoryUsed',
      jvm_memory_heap_max: 'resourceManagerComponent.host_components[0].metrics.jvm.HeapMemoryMax',
      containers_allocated: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AllocatedContainers',
      containers_pending: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.PendingContainers',
      containers_reserved: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.ReservedContainers',
      apps_submitted: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsSubmitted',
      apps_running: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsRunning',
      apps_pending: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsPending',
      apps_completed: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsCompleted',
      apps_killed: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsKilled',
      apps_failed: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsFailed',
      node_managers_count_active: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.activeNMcount',
      //node_managers_count_lost: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.lostNMcount',
      node_managers_count_unhealthy: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.unhealthyNMcount',
      node_managers_count_rebooted: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.rebootedNMcount',
      node_managers_count_decommissioned: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.decommissionedNMcount',
      allocated_memory: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AllocatedMB',
      available_memory: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AvailableMB',
      reserved_memory: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.ReservedMB',
      queue: 'resourceManagerComponent.queue'
    },
    mapReduce2Config: {
      version: 'jobHistoryServerComponent.ServiceComponentInfo.Version',
      job_history_server_id: 'jobHistoryServerComponent.host_components[0].HostRoles.host_name'
    },
    mapReduceConfig: {
      version: 'jobTrackerComponent.ServiceComponentInfo.Version',
      job_history_server_id: 'jobHistoryServerComponent.host_components[0].HostRoles.host_name',
      job_tracker_id: 'jobTrackerComponent.host_components[0].HostRoles.host_name',
      job_tracker_start_time: 'jobTrackerComponent.ServiceComponentInfo.StartTime',
      job_tracker_heap_used: 'jobTrackerComponent.ServiceComponentInfo.HeapMemoryUsed',
      job_tracker_heap_max: 'jobTrackerComponent.ServiceComponentInfo.HeapMemoryMax',
      alive_trackers: 'alive_trackers',
      black_list_trackers: 'black_list_trackers',
      gray_list_trackers: 'gray_list_trackers',
      map_slots: 'map_slots',
      reduce_slots: 'reduce_slots',
      jobs_submitted: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.jobs_submitted',
      jobs_completed: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.jobs_completed',
      jobs_running: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.jobs_running',
      map_slots_occupied: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.occupied_map_slots',
      map_slots_reserved: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.reserved_map_slots',
      reduce_slots_occupied: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.occupied_reduce_slots',
      reduce_slots_reserved: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.reserved_reduce_slots',
      maps_running: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.running_maps',
      maps_waiting: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.waiting_maps',
      reduces_running: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.running_reduces',
      reduces_waiting: 'jobTrackerComponent.ServiceComponentInfo.jobtracker.waiting_reduces',
      trackers_decommissioned: 'jobTrackerComponent.host_components[0].metrics.mapred.jobtracker.trackers_decommissioned',
      job_tracker_cpu: 'jobTrackerComponent.host_components[0].metrics.cpu.cpu_wio',
      job_tracker_rpc: 'jobTrackerComponent.host_components[0].metrics.rpc.RpcQueueTime_avg_time'
    },
    hbaseConfig: {
      version: 'masterComponent.ServiceComponentInfo.Version',
      master_id: 'masterComponent.host_components[0].HostRoles.host_name',
      master_start_time: 'masterComponent.ServiceComponentInfo.MasterStartTime',
      master_active_time: 'masterComponent.ServiceComponentInfo.MasterActiveTime',
      average_load: 'masterComponent.ServiceComponentInfo.AverageLoad',
      regions_in_transition: 'regions_in_transition',
      revision: 'masterComponent.ServiceComponentInfo.Revision',
      heap_memory_used: 'masterComponent.ServiceComponentInfo.HeapMemoryUsed',
      heap_memory_max: 'masterComponent.ServiceComponentInfo.HeapMemoryMax'
    },
    stormConfig: {
      total_tasks: 'restApiComponent.metrics.api.cluster.summary.["tasks.total"]',
      total_slots: 'restApiComponent.metrics.api.cluster.summary.["slots.total"]',
      free_slots: 'restApiComponent.metrics.api.cluster.summary.["slots.free"]',
      used_slots: 'restApiComponent.metrics.api.cluster.summary.["tasks.total"]',
      topologies: 'restApiComponent.metrics.api.cluster.summary.topologies',
      total_executors: 'restApiComponent.metrics.api.cluster.summary.["executors.total"]',
      nimbus_uptime: 'restApiComponent.metrics.api.cluster.summary.["nimbus.uptime"]'
    },

    model3: App.HostComponent,
    config3: {
      id: 'id',
      work_status: 'HostRoles.state',
      passive_state:'HostRoles.maintenance_state',
      desired_status: 'HostRoles.desired_state',
      component_name: 'HostRoles.component_name',
      host_id: 'HostRoles.host_name',
      stale_configs: 'HostRoles.stale_configs',
      display_name_advanced: 'display_name_advanced',
      $service_id: 'none' /* will be set outside of parse function */
    },

    map: function (json, isDefered) {
      if (!isDefered && !App.router.get('clusterController.isLoaded')) {
        App.router.get('clusterController').deferServiceMetricsLoad(json);
        return;
      }
      console.time('App.serviceMetricsMapper execution time');
      if (json.items) {

        // Host components
        var hostComponents = [];
        var services = App.cache['services'];
        var previousComponentStatuses = App.cache['previousComponentStatuses'];
        var previousComponentPassiveStates = App.cache['previousComponentPassiveStates'];
        /**
         * services contains constructed service-components structure from components array
         */
        json.items.forEach(function (component) {
          var serviceName = component.ServiceComponentInfo.service_name;
          var service = services.findProperty('ServiceInfo.service_name', serviceName);
          service.components.push(component);
          component.host_components.forEach(function (host_component) {
            host_component.id = host_component.HostRoles.component_name + "_" + host_component.HostRoles.host_name;
            previousComponentStatuses[host_component.id] = host_component.HostRoles.state;
            previousComponentPassiveStates[host_component.id] = host_component.HostRoles.maintenance_state;
            if (host_component.HostRoles.component_name == "HBASE_MASTER") {
              this.config3.ha_status = 'metrics.hbase.master.IsActiveMaster';
            }
            var comp = this.parseIt(host_component, this.config3);
            comp.service_id = serviceName;
            hostComponents.push(comp);
          }, this);
        }, this);

        this.computeAdditionalRelations(hostComponents, services);
        //load master components to model
        App.store.loadMany(this.get('model3'), hostComponents);

        var result = [];

        //parse service metrics from components
        services.forEach(function (item) {
          var finalJson = [];
          var serviceData = App.cache['hostComponentsOnService'][item.ServiceInfo.service_name];
          //service can have zero number of host-components
          if (serviceData) {
            item.host_components = serviceData.host_components;
            item.host_components.sort();
          }

          if (item && item.ServiceInfo && item.ServiceInfo.service_name == "HDFS") {
            finalJson = this.hdfsMapper(item);
            finalJson.rand = Math.random();
            result.push(finalJson);
            App.store.load(App.HDFSService, finalJson);
          } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "MAPREDUCE") {
            finalJson = this.mapreduceMapper(item);
            finalJson.rand = Math.random();
            result.push(finalJson);
            App.store.load(App.MapReduceService, finalJson);
          } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "HBASE") {
            finalJson = this.hbaseMapper(item);
            finalJson.rand = Math.random();
            result.push(finalJson);
            App.store.load(App.HBaseService, finalJson);
          } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "FLUME") {
            finalJson = this.flumeMapper(item);
            finalJson.rand = Math.random();
            result.push(finalJson);
            if (finalJson.nodeObjs) {
              finalJson.nodeObjs.forEach(function (no) {
                App.store.load(App.FlumeNode, no);
              });
            }
            App.store.load(App.FlumeService, finalJson);
          } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "YARN") {
            finalJson = this.yarnMapper(item);
            finalJson.rand = Math.random();
            result.push(finalJson);
            App.store.load(App.YARNService, finalJson);
          } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "MAPREDUCE2") {
            finalJson = this.mapreduce2Mapper(item);
            finalJson.rand = Math.random();
            result.push(finalJson);
            App.store.load(App.MapReduce2Service, finalJson);
          } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "STORM") {
            finalJson = this.stormMapper(item);
            finalJson.rand = Math.random();
            this.mapQuickLinks(finalJson, item);
            result.push(finalJson);
            App.store.load(App.StormService, finalJson);
          } else {
            finalJson = this.parseIt(item, this.config);
            finalJson.rand = Math.random();
            this.mapQuickLinks(finalJson, item);
            result.push(finalJson);
          }
        }, this);

        result = misc.sortByOrder(App.Service.servicesSortOrder, result);

        //load services to model
        if (previousResponse.length !== result.length) {
          App.store.loadMany(this.get('model'), result);
        } else {
          result.forEach(function (serviceJson) {
            var fields = ['passive_state','work_status', 'rand', 'alerts', 'quick_links', 'host_components', 'tool_tip_content', 'critical_alerts_count'];
            var service = this.get('model').find(serviceJson.id);
            var modifiedData = this.getDiscrepancies(serviceJson, previousResponse.findProperty('id', serviceJson.id), fields);
            if (modifiedData.isLoadNeeded) {
              App.store.load(this.get('model'), serviceJson);
            } else {
              for (var property in modifiedData) {
                service.set(stringUtils.underScoreToCamelCase(property), modifiedData[property]);
              }
            }
          }, this)
        }

        previousResponse = result;
      }
      console.timeEnd('App.serviceMetricsMapper execution time');
    },
    /**
     * compute display name of host-components
     * compute tooltip content of services by host-components
     * @param hostComponents
     * @param services
     */
    computeAdditionalRelations: function (hostComponents, services) {
      var isSecondaryNamenode = hostComponents.findProperty('component_name', 'SECONDARY_NAMENODE');
      services.setEach('tool_tip_content', '');
      // set tooltip for client-only services
      var clientOnlyServiceNames = Em.A(['TEZ', 'SQOOP', 'PIG']);
      clientOnlyServiceNames.forEach(function(serviceName) {
        var service = services.findProperty('ServiceInfo.service_name', serviceName);
        if (service) {
          service.tool_tip_content = Em.I18n.t('services.service.summary.clientOnlyService.ToolTip');
        }
      });
      hostComponents.forEach(function (hostComponent) {
        var service = services.findProperty('ServiceInfo.service_name', hostComponent.service_id);
        if (hostComponent) {
          // set advanced nameNode display name for HA, active or standby NameNode
          // this is useful on three places: hdfs health status hover tooltip, hdfs service summary and NN component on host detail page
          if (hostComponent.component_name === 'NAMENODE' && !isSecondaryNamenode) {
            var hdfs = this.hdfsMapper(service);
            var hostName = hostComponent.host_id;
            var activeNNText = Em.I18n.t('services.service.summary.nameNode.active');
            var standbyNNText = Em.I18n.t('services.service.summary.nameNode.standby');
            if (hdfs) {
              if (hostName == hdfs.active_name_node_id) {
                hostComponent.display_name_advanced = activeNNText;
              } else if (hostName == hdfs.standby_name_node_id || hostName == hdfs.standby_name_node2_id) {
                hostComponent.display_name_advanced = standbyNNText;
              } else {
                hostComponent.display_name_advanced = null;
              }
            }
          } else if (hostComponent.component_name === 'HBASE_MASTER') {
            if (hostComponent.work_status === 'STARTED') {
              (hostComponent.ha_status === true || hostComponent.ha_status == 'true') ?
                hostComponent.display_name_advanced = this.t('dashboard.services.hbase.masterServer.active') :
                hostComponent.display_name_advanced = this.t('dashboard.services.hbase.masterServer.standby');
            } else {
              hostComponent.display_name_advanced = null;
            }
          }

          if (hostComponent.display_name_advanced) {
            service.tool_tip_content += hostComponent.display_name_advanced + " " + App.HostComponentStatus.getTextStatus(hostComponent.work_status) + "<br/>";
          } else {
            service.tool_tip_content += App.format.role(hostComponent.component_name) + " " + App.HostComponentStatus.getTextStatus(hostComponent.work_status) + "<br/>";
          }
        }
      }, this)
    },
    /**
     * Map quick links to services:OOZIE,GANGLIA,NAGIOS,HUE
     * @param finalJson
     * @param item
     */
    mapQuickLinks: function (finalJson, item){
      if(!(item && item.ServiceInfo)) return;
      var quickLinks = {
        OOZIE: [19],
        GANGLIA: [20],
        NAGIOS: [21],
        HUE: [22],
        STORM: [31],
        FALCON: [32]
      };
      if (quickLinks[item.ServiceInfo.service_name])
        finalJson.quick_links = quickLinks[item.ServiceInfo.service_name];
    },

    hdfsMapper: function (item) {
      var finalConfig = jQuery.extend({}, this.config);
      // Change the JSON so that it is easy to map
      var hdfsConfig = this.hdfsConfig;
      item.components.forEach(function (component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "NAMENODE") {
          //enabled HA
          if ( component.host_components.length == 2) {
            var haState1;
            var haState2;
            if (component.host_components[1].metrics && component.host_components[1].metrics.dfs) {
              haState2 = component.host_components[1].metrics.dfs.FSNamesystem.HAState;
            }
            if (component.host_components[0].metrics && component.host_components[0].metrics.dfs) {
              haState1 = component.host_components[0].metrics.dfs.FSNamesystem.HAState;
            }
            var active_name_node = [];
            var standby_name_nodes = [];
            switch (haState1) {
              case "active":
                active_name_node.push(component.host_components[0].HostRoles.host_name);
                break;
              case "standby":
                standby_name_nodes.push(component.host_components[0].HostRoles.host_name);
                break;
            }
            switch (haState2) {
              case "active":
                active_name_node.push(component.host_components[1].HostRoles.host_name);
                break;
              case "standby":
                standby_name_nodes.push(component.host_components[1].HostRoles.host_name);
                break;
            }
            item.active_name_node_id = null;
            item.standby_name_node_id = null;
            item.standby_name_node2_id = null;
            switch (active_name_node.length) {
              case 1:
                item.active_name_node_id = active_name_node[0];
                break;
            }
            switch (standby_name_nodes.length) {
              case 1:
                item.standby_name_node_id = standby_name_nodes[0];
                break;
              case 2:
                item.standby_name_node_id = standby_name_nodes[0];
                item.standby_name_node2_id = standby_name_nodes[1];
                break;
            }
            // important: active nameNode always at host_components[0]; if no active, then any nameNode could work.
            if (haState2 == "active") { // change places for all model bind with host_component[0]
              var tmp = component.host_components[1];
              component.host_components[1] = component.host_components[0];
              component.host_components[0] = tmp;
            }
          }
          item.nameNodeComponent = component;
          finalConfig = jQuery.extend(finalConfig, hdfsConfig);
          // Get the live, dead & decommission nodes from string json
          if (component.host_components[0].metrics && component.host_components[0].metrics.dfs && component.host_components[0].metrics.dfs.namenode) {
            var decommissionNodesJson = App.parseJSON(component.host_components[0].metrics.dfs.namenode.DecomNodes);
          }
          item.decommission_data_nodes = [];
          for (var dcn in decommissionNodesJson) {
            item.decommission_data_nodes.push(dcn);
          }
        }
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "SECONDARY_NAMENODE") {
          item.snameNodeComponent = component;
        }
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "JOURNALNODE") {
          if (!item.journal_nodes) {
            item.journal_nodes = [];
          }
          if (component.host_components) {
            component.host_components.forEach(function (hc) {
              item.journal_nodes.push(hc.HostRoles.host_name);
            });
          }
        }
      });
      // Map
      var finalJson = this.parseIt(item, finalConfig);
      finalJson.quick_links = [1, 2, 3, 4];

      return finalJson;
    },
    yarnMapper: function (item) {
      var result = [];
      var self = this;
      var finalConfig = jQuery.extend({}, this.config);
      // Change the JSON so that it is easy to map
      var yarnConfig = this.yarnConfig;
      item.components.forEach(function (component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "RESOURCEMANAGER") {
          item.resourceManagerComponent = component;

          if (component.host_components[0].metrics && component.host_components[0].metrics.yarn) {
            var root = component.host_components[0].metrics.yarn.Queue.root;
            component.queue = JSON.stringify({
              'root': self.parseObject(root)
            });
          }
          // extend config
          finalConfig = jQuery.extend(finalConfig, yarnConfig);
        }
      });
      // Map
      var finalJson = this.parseIt(item, finalConfig);
      finalJson.quick_links = [ 23, 24, 25, 26 ];
      return finalJson;
    },

    parseObject: function(obj) {
      var res = {};
      for (var p in obj) {
        if (obj.hasOwnProperty(p)) {
          if (obj[p] instanceof Object) {
            res[p] = this.parseObject(obj[p]);
          }
        }
      }
      return res;
    },

    mapreduce2Mapper: function (item) {
      var result = [];
      var finalConfig = jQuery.extend({}, this.config);
      // Change the JSON so that it is easy to map
      var mapReduce2Config = this.mapReduce2Config;
      item.components.forEach(function (component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "HISTORYSERVER") {
          item.jobHistoryServerComponent = component;
          finalConfig = jQuery.extend(finalConfig, mapReduce2Config);
        }
      });
      // Map
      var finalJson = this.parseIt(item, finalConfig);
      finalJson.quick_links = [27, 28, 29, 30];

      return finalJson;
    },
    mapreduceMapper: function (item) {
      // Change the JSON so that it is easy to map
      var result = [];
      var finalConfig = jQuery.extend({}, this.config);
      var mapReduceConfig = this.mapReduceConfig;
      item.components.forEach(function (component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "JOBTRACKER") {
          item.jobTrackerComponent = component;
          finalConfig = jQuery.extend(finalConfig, mapReduceConfig);
          // Get the live, gray & black nodes from string json
          item.map_slots = 0;
          item.reduce_slots = 0;
          var liveNodesJson = App.parseJSON(component.ServiceComponentInfo.AliveNodes);
          var grayNodesJson = App.parseJSON(component.ServiceComponentInfo.GrayListedNodes);
          var blackNodesJson = App.parseJSON(component.ServiceComponentInfo.BlackListedNodes);
          item.alive_trackers = [];
          item.gray_list_trackers = [];
          item.black_list_trackers = [];
          if (liveNodesJson != null) {
            liveNodesJson.forEach(function (nj) {
              item.alive_trackers.push(nj.hostname);
              if (nj.slots && nj.slots.map_slots)
                item.map_slots += nj.slots.map_slots;
              if (nj.slots && nj.slots.map_slots_used)
                item.map_slots_used += nj.slots.map_slots_used;
              if (nj.slots && nj.slots.reduce_slots)
                item.reduce_slots += nj.slots.reduce_slots;
              if (nj.slots && nj.slots.reduce_slots_used)
                item.reduce_slots_used += nj.slots.reduce_slots_used;
            });
          }
          if (grayNodesJson != null) {
            grayNodesJson.forEach(function (nj) {
              item.gray_list_trackers.push(nj.hostname);
            });
          }
          if (blackNodesJson != null) {
            blackNodesJson.forEach(function (nj) {
              item.black_list_trackers.push(nj.hostname);
            });
          }
        } else if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "HISTORYSERVER") {
          item.jobHistoryServerComponent = component;
          finalConfig = jQuery.extend(finalConfig, mapReduceConfig);
        }
      });
      // Map
      var finalJson = this.parseIt(item, finalConfig);
      finalJson.quick_links = [5, 6, 7, 8, 9, 10, 11, 12];
      return finalJson;
    },
    hbaseMapper: function (item) {
      // Change the JSON so that it is easy to map
      var finalConfig = jQuery.extend({}, this.config);
      var hbaseConfig = this.hbaseConfig;
      item.components.forEach(function (component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "HBASE_MASTER") {
          item.masterComponent = component;
          finalConfig = jQuery.extend(finalConfig, hbaseConfig);
          var regionsArray = App.parseJSON(component.ServiceComponentInfo.RegionsInTransition);
          //regions_in_transition can have various type of value: null, array or int
          if (Array.isArray(regionsArray)) {
            item.regions_in_transition = regionsArray.length;
          } else {
            item.regions_in_transition = regionsArray == null ? 0 : regionsArray;
          }
        }
      });
      // Map
      var finalJson = this.parseIt(item, finalConfig);
      finalJson.average_load = parseFloat(finalJson.average_load).toFixed(2);
      finalJson.quick_links = [13, 14, 15, 16, 17, 18];
      return finalJson;
    },
    
    /**
     * Flume is different from other services, in that the important
     * data is in customizeable channels. Hence we directly transfer 
     * data into the JSON object.
     */
    flumeMapper: function (item) {
      var finalConfig = jQuery.extend({}, this.config);
      var finalJson = this.parseIt(item, finalConfig);
      item.components.forEach(function (component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "FLUME_SERVER") {
          if (!finalJson.nodes) {
            finalJson.nodes = [];
          }
          if (!finalJson.nodeObjs) {
            finalJson.nodeObjs = [];
          }
          if (component.host_components) {
            component.host_components.forEach(function (hc) {
              var fnode = {};
              fnode.id = hc.HostRoles.host_name;
              fnode.host_id = hc.HostRoles.host_name;
              fnode.channels = "";
              fnode.sources = "";
              fnode.sinks = "";
              if (hc.metrics != null && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume) {
                if (hc.metrics.flume.flume.CHANNEL) {
                  for ( var c in hc.metrics.flume.flume.CHANNEL) {
                    if (fnode.channels.length < 1) {
                      fnode.channels += c;
                    } else {
                      fnode.channels += ("," + c);
                    }
                  }
                }
                if (hc.metrics.flume.flume.SINK) {
                  for ( var c in hc.metrics.flume.flume.SINK) {
                    if (fnode.sinks.length < 1) {
                      fnode.sinks += c;
                    } else {
                      fnode.sinks += ("," + c);
                    }
                  }
                }
                if (hc.metrics.flume.flume.SOURCE) {
                  for ( var c in hc.metrics.flume.flume.SOURCE) {
                    if (fnode.sources.length < 1) {
                      fnode.sources += c;
                    } else {
                      fnode.sources += ("," + c);
                    }
                  }
                }
              }
              finalJson.nodeObjs.push(fnode);
              finalJson.nodes.push(hc.HostRoles.host_name);
            });
          }
        }
      });
      return finalJson;
    },

    /**
     * Storm mapper
     */
    stormMapper: function(item) {
      var finalConfig = jQuery.extend({}, this.config);
      var stormConfig = this.stormConfig;
      item.components.forEach(function(component) {
        if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "STORM_REST_API") {
          item.restApiComponent = component;
          finalConfig = jQuery.extend({}, finalConfig, stormConfig);
        }
      });
      return this.parseIt(item, finalConfig);
    }
  });
  
});
window.require.register("mappers/status_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.statusMapper = App.QuickDataMapper.create({
    model: App.HostComponent,
    map: function (json) {
      console.time('App.statusMapper execution time');
      if (json.items) {
        var hostsCache = App.cache['Hosts'];
        var previousHostStatuses = App.cache['previousHostStatuses'];
        var previousComponentStatuses = App.cache['previousComponentStatuses'];
        var previousComponentPassiveStates = App.cache['previousComponentPassiveStates'];
        var hostComponentRecordsMap = App.cache['hostComponentRecordsMap'];
        var servicesCache = App.cache['services'];
        var hostStatuses = {};
        var addedHostComponents = [];
        var updatedHostComponents = [];
        var componentServiceMap = App.QuickDataMapper.componentServiceMap();
        var currentComponentStatuses = {};
        var currentComponentPassiveStates = {};
        var currentHostStatuses = {};
        var hostComponentsOnService = {};

        json.items.forEach(function (host) {
          var hostName = host.Hosts.host_name;
          //update hosts, which have status changed
          if (previousHostStatuses[hostName] !== host.Hosts.host_status) {
            hostStatuses[hostName] = host.Hosts.host_status;
          }
          //preserve all hosts' status
          currentHostStatuses[hostName] = host.Hosts.host_status;
          var hostComponentsOnHost = [];
          host.host_components.forEach(function (host_component) {
            host_component.id = host_component.HostRoles.component_name + "_" + hostName;
            var existedComponent = previousComponentStatuses[host_component.id];
            var existedPassiveComponent = previousComponentPassiveStates[host_component.id];
            var service = componentServiceMap[host_component.HostRoles.component_name];

            //delete all currently existed host-components to indicate which need to be deleted from model
            delete previousComponentStatuses[host_component.id];
            delete previousComponentPassiveStates[host_component.id];

            if (existedComponent || existedPassiveComponent) {
              //update host-components, which have status changed
              if (existedComponent !== host_component.HostRoles.state || existedPassiveComponent !== host_component.HostRoles.maintenance_state) {
                updatedHostComponents.push(host_component);
              }
            } else {
              addedHostComponents.push({
                id: host_component.id,
                component_name: host_component.HostRoles.component_name,
                passive_state: host_component.HostRoles.maintenance_state,
                work_status: host_component.HostRoles.state,
                host_id: hostName,
                service_id: service
              });
              //update host-components only on adding due to Ember Data features
              if (hostsCache[hostName]) hostsCache[hostName].is_modified = true;
            }
            currentComponentStatuses[host_component.id] = host_component.HostRoles.state;
            currentComponentPassiveStates[host_component.id] = host_component.HostRoles.maintenance_state;
            //host-components to host relations
            hostComponentsOnHost.push(host_component.id);
            //host-component to service relations
            if (!hostComponentsOnService[service]) {
              hostComponentsOnService[service] = {
                host_components: []
              };
            }
            hostComponentsOnService[service].host_components.push(host_component.id);
          }, this);
          /**
           * updating relation between Host and his host-components
           */
          if (hostsCache[hostName]) {
            hostsCache[hostName].host_components = hostComponentsOnHost;
          } else {
            hostsCache[hostName] = {};
          }
          //check whether Nagios installed and started
          if (host.alerts) {
            hostsCache[hostName].critical_alerts_count = host.alerts.summary.CRITICAL + host.alerts.summary.WARNING;
          } else {
            hostsCache[hostName].critical_alerts_count = 0;
          }
        }, this);


        for (var id in previousComponentStatuses) {
          this.deleteRecord(hostComponentRecordsMap[id]);
        }

        updatedHostComponents.forEach(function (hostComponent) {
          var hostComponentRecord = hostComponentRecordsMap[hostComponent.id];
          if (hostComponentRecord) {
            hostComponentRecord.set('workStatus', hostComponent.HostRoles.state);
            hostComponentRecord.set('passiveState', hostComponent.HostRoles.maintenance_state);
          }
        }, this);

        var hostRecords = App.Host.find();
        hostRecords.forEach(function (host) {
          var status = hostStatuses[host.get('id')];
          var hostCache = hostsCache[host.get('id')];
          if (status) {
            host.set('healthStatus', status);
          }
          if (hostCache) {
            host.set('criticalAlertsCount', hostCache.critical_alerts_count);
          }
        });

        if (addedHostComponents.length) {
          App.store.loadMany(this.get('model'), addedHostComponents);
          App.HostComponent.find().forEach(function(hostComponent){
            hostComponentRecordsMap[hostComponent.get('id')] = hostComponent;
          });
        }

        // update services workStatus and passiveState
        App.Service.find().forEach(function (service) {
          var cachedServiceData = servicesCache.findProperty('ServiceInfo.service_name', service.get('serviceName'));
          if (cachedServiceData) {
            service.set('workStatus', cachedServiceData.ServiceInfo.state);
            service.set('passiveState', cachedServiceData.ServiceInfo.passive_state);
          }
        }, this);

        App.cache['previousHostStatuses'] = currentHostStatuses;
        App.cache['previousComponentStatuses'] = currentComponentStatuses;
        App.cache['previousComponentPassiveStates'] = currentComponentPassiveStates;
        App.cache['hostComponentsOnService'] = hostComponentsOnService;

      }
      console.timeEnd('App.statusMapper execution time');
      if (!App.router.get('clusterController.isLoaded')) {
        App.hostsMapper.map(json);
      }
    }
  });
  
});
window.require.register("mappers/target_cluster_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  App.targetClusterMapper = App.QuickDataMapper.create({
    model: App.TargetCluster,
    config: {
      id: 'name',
      name: 'name',
      execute: 'execute',
      workflow: 'workflow',
      write: 'write',
      readonly: 'readonly',
      staging: 'staging',
      working: 'working',
      temp: 'temp'
    },

    map: function (json) {
      var model = this.get('model');
      if (!model) {
        return;
      }

      if (json.items) {
        var result = [];
        var clustersToDelete = model.find().mapProperty('name');
        json.items.forEach(function (item) {
          result.push(this.parseIt(item, this.config));
          clustersToDelete = clustersToDelete.without(item.name);
        }, this);
        clustersToDelete.forEach(function (name) {
          this.deleteRecord(model.find().findProperty('name', name));
        }, this);
        App.store.loadMany(model, result);
      }
    }
  });
  
});
window.require.register("mappers/users_mapper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.usersMapper = App.QuickDataMapper.create({
    model : App.User,
    config : {
      id : 'Users.user_name',
      user_name : 'Users.user_name',
      roles : 'Users.roles',
      is_ldap: 'Users.ldap_user',
      admin: 'Users.admin'
    },
    map: function (json) {
      var self = this;
      json.items.forEach(function (item) {
        var result= [];
        if(!App.User.find().someProperty("userName", item.Users.user_name)) {
          item.Users.admin = self.isAdmin(item.Users.roles);
          result.push(self.parseIt(item, self.config));
          App.store.loadMany(self.get('model'), result);
        }
      });
    },
    isAdmin: function(roles) {
      return (roles.indexOf("admin") >= 0);
    }
  });
  
});
window.require.register("messages", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  Em.I18n.translations = {

    'app.name':'Ambari',
    'app.reloadPopup.link': 'Reload Page',
    'app.reloadPopup.text': 'Trying to connect to server...',
    'app.reloadPopup.header': 'Reload Page',

    'app.loadingPlaceholder': 'Loading...',
    'app.versionMismatchAlert.title': 'Ambari Server / Web Client Version Mismatch',
    'app.versionMismatchAlert.body': 'Ambari Server and Web Client versions do not match:<br> ' +
      '<br>Ambari Server: <strong>{0}</strong>' +
      '<br>Ambari Web Client: <strong>{1}</strong><br>' +
      '<br>This typically happens after upgrading Ambari due to Ambari Web Client code cached in the browser.' +
      '<br>Perform hard refresh to see if this message disappears.' +
      '<br>If you keep seeing this message, clear the browser cache completely and hit this URL again.' +
      '<br>You must resolve this error in order to continue.',
    'app.signout':'Sign out',
    'app.settings':'Settings',
    'app.aboutAmbari':'About',
    'app.settings.notshowBgOperationsPopup': 'Do not show the Background Operations dialog when starting an operation',
    'app.settings.notShowBgOperations': 'Do not show this dialog again when starting a background operation',

    'app.aboutAmbari.getInvolved': 'Get involved!',
    'app.aboutAmbari.version': 'Version',
    'app.aboutAmbari.licensed': 'Licensed under the Apache License, Version 2.0',

    'apply':'apply',
    'and':'and',
    'none':'none',
    'all':'all',
    'minimum':'minimum',
    'from':'From',
    'to':'To',
    'ok':'OK',
    'as':'as',
    'on':'on',
    'any': 'Any',
    'more':'more',
    'yes':'Yes',
    'no':'No',
    'add': 'Add',
    'ops': 'ops',


    'common.access':'Access',
    'common.learnMore':'Learn more',
    'common.showDetails':'Show Details',
    'common.back':'Back',
    'common.prev':'Prev',
    'common.next':'Next',
    'common.host':'Host',
    'common.hosts':'Hosts',
    'common.services':'Services',
    'common.group':'Group',
    'common.progress':'Progress',
    'common.status':'Status',
    'common.action':'Action',
    'common.remove':'Remove',
    'common.retry':'Retry',
    'common.skip':'Skip',
    'common.filter': 'Filter',
    'common.rollBack':'Rollback',
    'common.show':'Show',
    'common.hide':'Hide',
    'common.cancel':'Cancel',
    'common.apply':'Apply',
    'common.done':'Done',
    'common.failed':'Failed',
    'common.service': 'Service',
    'common.version':'Version',
    'common.description':'Description',
    'common.client':'Client',
    'common.zookeeper':'ZooKeeper',
    'common.hbase':'HBase',
    'common.regionServer':'RegionServer',
    'common.taskTracker':'TaskTracker',
    'common.dataNode':'DataNode',
    'common.more': 'More...',
    'common.print':'Print',
    'common.deploy':'Deploy',
    'common.message':'Message',
    'common.tasks':'Tasks',
    'common.open':'Open',
    'common.copy':'Copy',
    'common.complete':'Complete',
    'common.metrics':'Metrics',
    'common.timeRange':'Time Range',
    'common.name':'Name',
    'common.value':'Value',
    'common.ipAddress':'IP Address',
    'common.cpu':'CPU',
    'common.cores': 'Cores (CPU)',
    'common.ram':'RAM',
    'common.disk':'Disk',
    'common.diskUsage':'Disk Usage',
    'common.loadAvg':'Load Avg',
    'common.components':'Components',
    'common.component':'Component',
    'common.quickLinks':'Quick Links',
    'common.save':'Save',
    'common.servers':'Servers',
    'common.clients':'Clients',
    'common.user': 'User',
    'common.users': 'Users',
    'common.issues': 'Issues',
    'common.os':'OS',
    'common.memory':'Memory',
    'common.maximum':'Maximum',
    'common.start':'Start',
    'common.stop':'Stop',
    'common.pause':'Pause',
    'common.decommission':'Decommission',
    'common.recommission':'Recommission',
    'common.failure': 'Failure',
    'common.type': 'Type',
    'common.close': 'Close',
    'common.warning': 'Warning',
    'common.information': 'Information',
    'common.all':'All',
    'common.success': 'Success',
    'common.fail':'Fail',
    'common.error': 'Error',
    'common.loading': 'Loading',
    'common.search': 'Search',
    'common.confirm': 'Confirm',
    'common.upgrade': 'Upgrade',
    'common.reUpgrade': 'Retry Upgrade',
    'common.security':'Security',
    'common.cluster':'Cluster',
    'common.add': 'Add',
    'common.edit': 'Edit',
    'common.delete': 'Delete',
    'common.empty': 'Empty',
    'common.override':'Override',
    'common.undo':'Undo',
    'common.details':'Details',
    'common.stats':'Stats',
    'common.abort': 'Abort',
    'common.misc': 'Misc',
    'common.userSettings': 'User Settings',
    'common.aboutAmbari': 'About',
    'common.notAvailable': 'Not Available',
    'common.na': 'n/a',
    'common.operations': 'Operations',
    'common.startTime': 'Start Time',
    'common.duration': 'Duration',
    'common.reinstall': 'Re-Install',
    'common.errorPopup.header': 'An error has been encountered',
    'common.use': 'Use',
    'common.stacks': 'Stacks',
    'common.reset': 'Reset',
    'common.path': 'Path',
    'common.package': 'Package',
    'common.proceed': 'Proceed',
    'common.process': 'Process',
    'common.property': 'Property',
    'common.installed': 'Installed',
    'common.persist.error' : 'Error in persisting web client state at ambari server:',
    'common.update.error' : 'Error in retrieving web client state from ambari server',
    'common.tags': 'Tags',
    'common.important': 'Important',
    'common.allServices':'All Services',
    'common.move':'Move',
    'common.change': 'Change',
    'common.overrides': 'Overrides',
    'common.properties': 'properties',
    'common.conf.group': 'Configuration Group',
    'common.ignore': 'Ignore',
    'common.restart': 'Restart',
    'common.discard': 'Discard',
    'common.actions': 'Actions',
    'common.maintenance': 'Maintenance',
    'common.passive_state': 'Maintenance Mode',
    'common.selected': 'Selected',
    'common.password': 'Password',
    'common.url': 'URL',
    'common.advanced': 'Advanced',
    'common.download': 'Download',
    'common.current': 'Current',
    'common.additional': 'Additional',
    'common.time.start': 'Start Time',
    'common.time.end': 'End Time',
    'common.hostLog.popup.logDir.path':'/var/lib/ambari-agent/data/',
    'common.hostLog.popup.outputLog.value': 'output-{0}.txt',
    'common.hostLog.popup.errorLog.value': 'errors-{0}.txt',
    'common.maintenance.task': ' Toggle Maintenance Mode',
    'common.used': 'used',
    'common.free': 'free',
    'common.type.string': 'string',
    'common.type.number': 'number',

    'passiveState.turnOn':'Turn On Maintenance Mode',
    'passiveState.turnOff':'Turn Off Maintenance Mode',
    'passiveState.turnOnFor':'Turn On Maintenance Mode for {0}',
    'passiveState.turnOffFor':'Turn Off Maintenance Mode for {0}',

    'requestInfo.installComponents':'Install Components',
    'requestInfo.installServices':'Install Services',
    'requestInfo.startServices':'Start Services',
    'requestInfo.stopAllServices':'Stop All Services',
    'requestInfo.startAllServices':'Start All Services',
    'requestInfo.startHostComponent':'Start',
    'requestInfo.startHostComponent.datanode':'Start DataNode',
    'requestInfo.startHostComponent.nodemanager':'Start NodeManager',
    'requestInfo.startHostComponent.hbase_regionserver':'Start HBase RegionServer',
    'requestInfo.startHostComponents':'Start Components',
    'requestInfo.upgradeHostComponent':'Upgrade',
    'requestInfo.stopHostComponent':'Stop',
    'requestInfo.installHostComponent':'Install',
    'requestInfo.installNewHostComponent':'Install',
    'requestInfo.stop':'Stop {0}',
    'requestInfo.start':'Start {0}',
    'requestInfo.unspecified':'Request name not specified',

    'hostPopup.noServicesToShow':'No services to show',
    'hostPopup.noHostsToShow':'No hosts to show',
    'hostPopup.noTasksToShow':'No tasks to show',
    'hostPopup.status.category.all':'All ({0})',
    'hostPopup.status.category.pending':'Pending ({0})',
    'hostPopup.status.category.inProgress':'In Progress ({0})',
    'hostPopup.status.category.failed':'Failed ({0})',
    'hostPopup.status.category.success':'Success ({0})',
    'hostPopup.status.category.aborted':'Aborted ({0})',
    'hostPopup.status.category.timedout':'Timedout ({0})',
    'hostPopup.header.postFix':' Background Operations Running',
    'hostPopup.bgop.sourceRequestSchedule.running': 'Future operations of this batch request can be aborted',
    'hostPopup.bgop.sourceRequestSchedule.aborted': 'Future operations of this batch request have been aborted',
    'hostPopup.bgop.abort.rollingRestart': 'Abort Rolling Restart',
    'hostPopup.warning.alertsTimeOut': 'Maintenance Mode has been turned {0}. It may take a few minutes for the alerts to be {1}.',

    'question.sure':'Are you sure?',

    'popup.highlight':'click to highlight',
    'popup.confirmation.commonHeader':'Confirmation',
    'popup.prompt.commonHeader':'Prompt',
    'popup.confirmationFeedBack.sending':'Sending...',
    'popup.confirmationFeedBack.query.fail':'Request failed',

    'router.hadoopClusterNotSetUp':'Your administrator has not set up a Hadoop cluster yet.',

    'login.header':'Sign in',
    'login.username':'Username',
    'login.loginButton':'Sign in',
    'login.error.bad.credentials':'Unable to sign in. Invalid username/password combination.',
    'login.error.bad.connection':'Unable to connect to Ambari Server. Confirm Ambari Server is running and you can reach Ambari Server from this machine.',

    'graphs.noData.title': 'No Data',
    'graphs.noData.message': 'There was no data available. Possible reasons include inaccessible Ganglia service.',
    'graphs.noDataAtTime.message': 'No available data for the time period.',
    'graphs.error.title': 'Error',
    'graphs.error.message': 'There was a problem getting data for the chart ({0}: {1})',
    'graphs.timeRange.hour': 'Last 1 hour',
    'graphs.timeRange.twoHours': 'Last 2 hours',
    'graphs.timeRange.fourHours': 'Last 4 hours',
    'graphs.timeRange.twelveHours': 'Last 12 hours',
    'graphs.timeRange.day': 'Last 24 hours',
    'graphs.timeRange.week': 'Last 1 week',
    'graphs.timeRange.month': 'Last 1 month',
    'graphs.timeRange.year': 'Last 1 year',

    'users.userName.validationFail': 'Only lowercase letters and numbers are recommended; must start with a letter',
    'host.spacesValidation': 'Can\'t contain whitespaces',
    'host.trimspacesValidation': 'Can\'t contain leading or trailing whitespaces',

    'services.nagios.description':'Nagios Monitoring and Alerting system',
    'services.ganglia.description':'Ganglia Metrics Collection system',
    'services.hdfs.description':'Apache Hadoop Distributed File System',
    'services.glusterfs.description':'Apache Hadoop Compatible File System (must be installed manually)',
    'services.mapreduce.description':'Apache Hadoop Distributed Processing Framework',
    'services.sqoop.description':'Tool for transferring bulk data between Apache Hadoop and structured data stores such as relational databases',
    'services.pig.description':'Scripting platform for analyzing large datasets',
    'services.hive.description':'Data warehouse system for ad-hoc queries & analysis of large datasets and table & storage management service',
    'services.oozie.description':'System for workflow coordination and execution of Apache Hadoop jobs',
    'services.zookeeper.description':'Centralized service which provides highly reliable distributed coordination',
    'services.hbase.description':'Non-relational distributed database and centralized service for configuration management & synchronization',
    'services.hive.databaseComponent':'Database Server',
    'services.mapreduce2.description':'Apache Hadoop NextGen MapReduce (client libraries)',
    'services.yarn.description':'Apache Hadoop NextGen MapReduce (YARN)',
    'services.tez.description':'Tez is the next generation Hadoop Query Processing framework written on top of YARN',
    'services.falcon.description': 'Falcon mirroring engine',
    'services.storm.description': 'Apache Hadoop Stream processing framework',
    'services.storm.slots.metrics.title': 'Number of slots',
    'services.storm.slots.metrics.free': 'Free slots',
    'services.storm.slots.metrics.total': 'Total slots',
    'services.storm.slots.metrics.used': 'Used slots',
    'services.storm.executors.metrics.title': 'Number of executors',
    'services.storm.executors.metrics.total': 'Total executors',
    'services.storm.topology.metrics.title': 'Number of topologies',
    'services.storm.topology.metrics.total': 'Total topologies',
    'services.storm.tasks.metrics.title': 'Number of tasks',
    'services.storm.tasks.metrics.total': 'Total tasks',


    'services.alerts.head':'You have {0} critical alert notification(s).',
    'services.alerts.OK.timePrefixShort': 'OK',
    'services.alerts.WARN.timePrefixShort': 'WARN',
    'services.alerts.CRIT.timePrefixShort': 'CRIT',
    'services.alerts.MAINT.timePrefixShort': 'MAINT',
    'services.alerts.UNKNOWN.timePrefixShort': 'UNKNOWN',
    'services.alerts.OK.timePrefix': 'OK for {0}',
    'services.alerts.WARN.timePrefix': 'WARN for {0}',
    'services.alerts.CRIT.timePrefix': 'CRIT for {0}',
    'services.alerts.MAINT.timePrefix': 'MAINT for {0}',
    'services.alerts.UNKNOWN.timePrefix': 'UNKNOWN for {0}',
    'services.alerts.lastCheck': 'Last Checked {0}',
    'services.alerts.brLastCheck': '<br>Last Checked {0}',
    'services.alerts.occurredOn': 'Occurred on {0}, {1}',
    'services.alerts.headingOfList': 'Alerts and Health Checks',
    'services.alerts.goToService': 'Go to Service',
    'services.alerts.goToNagios': 'Go to Nagios Web UI',

    'topnav.logo.href':'http://ambari.apache.org/',
    'topnav.help.href':'https://cwiki.apache.org/confluence/display/AMBARI/Ambari',

    'installer.header':'Cluster Install Wizard',
    'installer.navigation.warning.header':'Navigation Warning',

    'installer.noHostsAssigned':'No host assigned',
    'installer.slaveComponentHosts.selectHosts':'select hosts for this group',
    'installer.slaveComponentHostsPopup.header':'Select which hosts should belong to which {0} group',

    'installer.controls.slaveComponentGroups':' Groups',
    'installer.controls.serviceConfigPopover.title':'{0}<br><small>{1}</small>',
    'installer.controls.serviceConfigMultipleHosts.other':'1 other',
    'installer.controls.serviceConfigMultipleHosts.others':'{0} others',
    'installer.controls.serviceConfigMasterHosts.header':'{0} Hosts',
    'installer.controls.addSlaveComponentGroupButton.title':'Add a {0} Group',
    'installer.controls.addSlaveComponentGroupButton.content':'If you need different settings on certain {0}s, you can add a {1} group.<br>All {2}s within the same group will have the same set of settings.  You can create multiple groups.',
    'installer.controls.slaveComponentChangeGroupName.error':'group with this name already exist',

    'installer.step0.header':'Welcome',
    'installer.step0.body.header':'Welcome to Apache Ambari!',
    'installer.step0.body':'Ambari makes it easy to install, manage, and monitor Hadoop clusters.<br>' +
      'We will walk you through the cluster installation process with this step-by-step wizard.',
    'installer.step0.clusterName':'Name your cluster',
    'installer.step0.clusterName.tooltip.title':'Cluster Name',
    'installer.step0.clusterName.tooltip.content':'Enter a unique cluster name. Cluster name cannot be changed later.',
    'installer.step0.clusterName.error.required':'Cluster Name is required',
    'installer.step0.clusterName.error.whitespaces':'Cluster Name cannot contain white spaces',
    'installer.step0.clusterName.error.specialChar':'Cluster Name cannot contain special characters',

    'installer.step1.header':'Select Stack',
    'installer.step1.body':'Please select the service stack that you want to use to install your Hadoop cluster.',
    'installer.step1.advancedRepo.title':'Advanced Repository Options',
    'installer.step1.advancedRepo.message':'Customize the repository Base URLs for downloading the Stack software packages. If your hosts do not have access to the internet, you will have to create a local mirror of the Stack repository that is accessible by all hosts and use those Base URLs here.',
    'installer.step1.advancedRepo.importantMassage':'<b>Important:</b> When using local mirror repositories, you only need to provide Base URLs for the Operating System you are installing for your Stack. Uncheck all other repositories.',
    'installer.step1.advancedRepo.localRepo.error.modifyUrl':'Local repository URL must be modified',
    'installer.step1.advancedRepo.localRepo.error.noUrl':'Base URL required for a local repository',
    'installer.step1.advancedRepo.localRepo.column.baseUrl':'Base URL',
    'installer.step1.advancedRepo.localRepo.label.os':'Operating System',
    'installer.step1.advancedRepo.localRepo.label.baseUrl':'Repository Base URL',
    'installer.step1.advancedRepo.localRepo.label.stack':'Stack',
    'installer.step1.advancedRepo.skipValidation.tooltip':'<b>Warning:</b> This is for advanced users only. Use this option if you want to skip validation for Repository Base URLs.',
    'installer.step1.advancedRepo.skipValidation.message':'Skip Repository Base URL validation (Advanced)',
    'installer.step1.attentionNeeded':'<b>Attention:</b> Repository URLs are REQUIRED before you can proceed.',
    'installer.step1.invalidURLAttention': '<b>Attention:</b> Please make sure all repository URLs are valid before proceeding.',
    'installer.step1.checkAtLeastOneAttention': '<b>Attention:</b> Please check at least one repository.',

    'installer.step2.header':'Install Options',
    'installer.step2.body':'Enter the list of hosts to be included in the cluster and provide your SSH key.',
    'installer.step2.targetHosts':'Target Hosts',
    'installer.step2.targetHosts.info':'Enter a list of hosts using the Fully Qualified Domain Name (FQDN), one per line',
    'installer.step2.hostPattern.tooltip.title':'Pattern Expressions',
    'installer.step2.hostPattern.tooltip.content':'You can use pattern expressions to specify a number of target hosts. For example, to specify host01.domain thru host10.domain, enter host[01-10].domain in the target hosts textarea.',
    'installer.step2.hostName.error.required':'You must specify at least one host name',
    'installer.step2.hostName.error.already_installed':'All these hosts are already part of the cluster',
    'installer.step2.hostName.error.notRequired':'Host Names will be ignored if not using SSH to automatically configure hosts',
    'installer.step2.hostName.error.invalid':'Invalid Host Name(s)',
    'installer.step2.hostName.pattern.header':'Host name pattern expressions',
    'installer.step2.sshKey':'Host Registration Information',
    'installer.step2.sshKey.error.required':'SSH Private Key is required',
    'installer.step2.passphrase.error.match':'Passphrases do not match',
    'installer.step2.manualInstall.label':'Do not use SSH to automatically configure hosts ',
    'installer.step2.manualInstall.info':'By not using SSH to connect to the target hosts, you must manually install and' +
      ' start the Ambari Agent on each host in order for the wizard to perform the necessary configurations and' +
      ' software installs.',
    'installer.step2.advancedOption':'Advanced Options',
    'installer.step2.repoConf':'Software Repository Configuration File Path',
    'installer.step2.advancedOptions.header':'Advanced Options',
    'installer.step2.localRepo.label_use':'Use a',
    'installer.step2.localRepo.label_instead':'instead of downloading software packages from the Internet',
    'installer.step2.localRepo.error.required':'Local repository file path is required',
    'installer.step2.localRepo.tooltip.title':'Local Software Repository',
    'installer.step2.localRepo.tooltip.content': 'The cluster install requires access to the Internet to fetch software ' +
      'from a remote repository. In some cases, adequate bandwidth is not available and you want to prevent downloading ' +
      'packages from the remote repository over and over again. Other times, Internet access is not available from the ' +
      'hosts in your cluster. In these situations, you must set up a version of the repository that your machines can ' +
      'access locally and this is called a <b>Local Software Repository</b>',
    'installer.step2.javaHome.label' : 'Path to 64-bit JDK',
    'installer.step2.javaHome.tooltip.title' : 'JAVA_HOME',
    'installer.step2.javaHome.tooltip.content' : 'Path to 64-bit JAVA_HOME. /usr/jdk/jdk1.6.0_31 is the default used by Ambari. You can override this to a specific path that contains the JDK. <br/> Note: the path must be valid on <b>ALL</b> hosts in your cluster.',
    'installer.step2.javaHome.tooltip.placeholder' : '/usr/jdk/jdk1.6.0_31',
    'installer.step2.useSsh.provide' : 'Provide your',
    'installer.step2.useSsh.provide_id_rsa' : ' to automatically register hosts',
    'installer.step2.useSsh.tooltip.title':'SSH Private Key',
    'installer.step2.useSsh.tooltip.content':'The <b>SSH Private Key File</b> is used to connect to the target hosts in your cluster to install the Ambari Agent.',
    'installer.step2.manualInstall.perform':'Perform',
    'installer.step2.manualInstall.perform_on_hosts':'on hosts and do not use SSH',
    'installer.step2.manualInstall.tooltip.title':'manual registration',
    'installer.step2.manualInstall.tooltip.content':'Manually registering the Ambari Agent on each host eliminates the need for SSH and should be performed prior to continuing cluster installation.',
    'installer.step2.manualInstall.popup.header':'Before You Proceed',
    'installer.step2.manualInstall.popup.body':'You must install Ambari Agents on each host you want to manage before you proceed.',
    'installer.step2.warning.popup.body':'<p>The following hostnames are not valid FQDNs:</p><p> {0} </p><p>This may cause problems during installation. Do you want to continue?</p>',
    'installer.step2.orUse':'Or use',
    'installer.step2.registerAndConfirm':'Register and Confirm',
    'installer.step2.evaluateStep.installedHosts':'These hosts are already installed on the cluster and will be ignored:',
    'installer.step2.evaluateStep.continueConfirm':'Do you want to continue?',
    'installer.step2.evaluateStep.hostRegInProgress':'Host Registration is currently in progress.  Please try again later.',
    'installer.step2.sshUser':'SSH user (root or',
    'installer.step2.sshUser.link':'passwordless sudo',
    'installer.step2.sshUser.account':'account)',
    'installer.step2.sshUser.toolTip':'An account that can execute sudo without entering a password',
    'installer.step2.sshUser.placeholder':'Enter user name',
    'installer.step2.sshUser.required':'User name is required',

    'installer.step3.header':'Confirm Hosts',
    'installer.step3.body':'Registering your hosts.<br>' +
      'Please confirm the host list and remove any hosts that you do not want to include in the cluster.',
    'installer.step3.hostLog.popup.header':'Registration log for {0}',
    'installer.step3.hosts.remove.popup.header':'Remove Hosts',
    'installer.step3.hosts.remove.popup.body':'Are you sure you want to remove the selected host(s)?',
    'installer.step3.hostInformation.popup.header':'Error in retrieving host Information',
    'installer.step3.hostInformation.popup.body' : 'All bootstrapped hosts registered but unable to retrieve cpu and memory related information',
    'installer.step3.hostOsTypeCheck.popup.body' : 'Host registered successfully, but the host operating system type NOT match the selected group in "Select Stack" step: Advanced Repository Option. You can go back to "Select Stack" step OR remove this host.' +
      'The host type is {0}, but you selected group {1} in step 1.',
    'installer.step3.hostWarningsPopup.report':'Show Report',
    'installer.step3.hostWarningsPopup.report.header': '<p style="font-family: monospace">######################################<br># Host Checks Report<br>#<br># Generated: ',
    'installer.step3.hostWarningsPopup.report.hosts': '<br>######################################<br><br>######################################<br># Hosts<br>#<br># A space delimited list of hosts which have issues.<br># Provided so that administrators can easily copy hostnames into scripts, email etc.<br>######################################<br>HOSTS<br>',
    'installer.step3.hostWarningsPopup.report.firewall': '<br><br>######################################<br># Firewall<br>#<br># A newline delimited list of firewall issues.<br>######################################<br>FIREWALL<br>',
    'installer.step3.hostWarningsPopup.report.fileFolders': '<br><br>######################################<br># Files and Folders<br>#<br># A space delimited list of files and folders which should not exist.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: rm -r /etc/hadoop /etc/hbase<br>######################################<br>FILES AND FOLDERS<br>',
    'installer.step3.hostWarningsPopup.report.process': '<br><br>######################################<br># Processes<br>#<br># A comma separated list of process tuples which should not be running.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br>######################################<br>PROCESSES<br>',
    'installer.step3.hostWarningsPopup.report.package': '<br><br>######################################<br># Packages<br>#<br># A space delimited list of software packages which should be uninstalled.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: yum remove hadoop-hdfs nagios<br>######################################<br>PACKAGES<br>',
    'installer.step3.hostWarningsPopup.report.service': '<br><br>######################################<br># Services<br>#<br># A space delimited list of services which should be up and running.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: services start ntpd httpd<br>######################################<br>SERVICES<br>',
    'installer.step3.hostWarningsPopup.report.user': '<br><br>######################################<br># Users<br>#<br># A space delimited list of users who should not exist.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: userdel hdfs<br>######################################<br>USERS<br>',
    'installer.step3.hostWarningsPopup.report.folder': '\\ /folder',
    'installer.step3.hostWarningsPopup.checks': 'Host Checks found',
    'installer.step3.hostWarningsPopup.notice':'After manually resolving the issues, click <b>Rerun Checks</b>.'+
      '<br>To manually resolve issues on <b>each host</b> run the HostCleanup script (Python 2.6 or greater is required):'+
      '<br><div class="code-snippet">python /usr/lib/python2.6/site-packages/ambari_agent/HostCleanup.py --silent --skip=users</div>' +
      '<div class="alert alert-warn"><b>Note</b>: To clean up in interactive mode, remove <b>--silent</b> option. To clean up all resources, including <i>users</i>, remove <b>--skip=users</b> option. Use <b>--help</b> for a list of available options.</div>',
    'installer.step3.hostWarningsPopup.summary':'{0} on {1}',
    'installer.step3.hostWarningsPopup.repositories':'Repository Issues',
    'installer.step3.hostWarningsPopup.repositories.name':'Repository for OS not available',
    'installer.step3.hostWarningsPopup.repositories.context':'Host ({0}) has a {1} OS type, But the repositories chosen in "Select Stack" step was {2}.',
    'installer.step3.hostWarningsPopup.repositories.message': 'The following registered hosts have different Operating System types from the available Repositories chosen in "Select Stack" step. You can go back to "Select Stack" step to select another OS repository <b>or</b> remove the host.',
    'installer.step3.hostWarningsPopup.disk':'Disk Issues',
    'installer.step3.hostWarningsPopup.disk.name':'Not enough disk space ',
    'installer.step3.hostWarningsPopup.disk.context1':'Not enough disk space on host ({0}).',
    'installer.step3.hostWarningsPopup.disk.context2':'A minimum of {0} is required for "{1}" mount.',
    'installer.step3.hostWarningsPopup.disk.message':'The following registered hosts have issues related to disk space',
    'installer.step3.hostWarningsPopup.firewall':'Firewall Issues',
    'installer.step3.hostWarningsPopup.firewall.message':'Firewall is running on the following hosts. Please configure the firewall to allow communications on the ports documented in the <i>Configuring Ports</i> section of  the <a target=\"_blank\" href=\"http://ambari.apache.org/current/installing-hadoop-using-ambari/content/\">Ambari documentation</a>',
    'installer.step3.hostWarningsPopup.firewall.name':'<i>iptables</i> Running',
    'installer.step3.hostWarningsPopup.process':'Process Issues',
    'installer.step3.hostWarningsPopup.processes.message':'The following processes should not be running',
    'installer.step3.hostWarningsPopup.fileAndFolder':'File and Folder Issues',
    'installer.step3.hostWarningsPopup.fileFolders.message':'The following files and folders should not exist',
    'installer.step3.hostWarningsPopup.package':'Package Issues',
    'installer.step3.hostWarningsPopup.packages.message':'The following packages should be uninstalled',
    'installer.step3.hostWarningsPopup.user':'User Issues',
    'installer.step3.hostWarningsPopup.users.message':'The following users should be removed',
    'installer.step3.hostWarningsPopup.service':'Service Issues',
    'installer.step3.hostWarningsPopup.services.message':'The following services should be up',
    'installer.step3.hostWarningsPopup.misc':'Misc Issues',
    'installer.step3.hostWarningsPopup.misc.message':'The following umasks should be changed',
    'installer.step3.hostWarningsPopup.misc.umask':'Umask',
    'installer.step3.hostWarningsPopup.issue':'issue',
    'installer.step3.hostWarningsPopup.issues':'issues',
    'installer.step3.hostWarningsPopup.emptyMessage':'There were no',
    'installer.step3.hostWarningsPopup.empty.filesAndFolders':'unwanted files and folders',
    'installer.step3.hostWarningsPopup.empty.processes':'unwanted processes',
    'installer.step3.hostWarningsPopup.empty.packages':'unwanted packages',
    'installer.step3.hostWarningsPopup.empty.users':'unwanted users',
    'installer.step3.hostWarningsPopup.empty.services':'unwanted services',
    'installer.step3.hostWarningsPopup.empty.misc':'issues',
    'installer.step3.hostWarningsPopup.empty.firewall':'firewalls running',
    'installer.step3.hostWarningsPopup.empty.repositories':'repositories OS type mis-match with registered hosts',
    'installer.step3.hostWarningsPopup.empty.disk':'disk space issues',
    'installer.step3.hostWarningsPopup.action.exists':'Exists on',
    'installer.step3.hostWarningsPopup.action.notRunning':'Not running on',
    'installer.step3.hostWarningsPopup.action.installed':'Installed on',
    'installer.step3.hostWarningsPopup.action.running':'Running on',
    'installer.step3.hostWarningsPopup.action.invalid':'Invalid on',
    'installer.step3.hostWarningsPopup.host':'host',
    'installer.step3.hostWarningsPopup.hosts':'hosts',
    'installer.step3.hostWarningsPopup.moreHosts':'{0} more hosts...<br>Click on link to view all hosts.',
    'installer.step3.hostWarningsPopup.allHosts':'List of hosts',
    'installer.step3.hostWarningsPopup.rerunChecks':'Rerun Checks',
    'installer.step3.hostWarningsPopup.hostHasWarnings':'Warning: Host checks failed on some of your hosts. It is highly recommended that you fix these problems first before proceeding to prevent potentially major problems with cluster installation. Are you sure you want to ignore these warnings and proceed?',
    'installer.step3.warningsWindow.allHosts':'Warnings across all hosts',
    'installer.step3.warningsWindow.warningsOn':'Warnings on ',
    'installer.step3.warningsWindow.directoriesAndFiles':'DIRECTORIES AND FILES',
    'installer.step3.warningsWindow.noWarnings':'No warnings',
    'installer.step3.hosts.noHosts':'No hosts to display',
    'installer.step3.warnings.popup.header':'Host Checks',
    'installer.step3.warnings.description':'Some warnings were encountered while performing checks against the above hosts.',
    'installer.step3.warnings.linkText':'Click here to see the warnings.',
    'installer.step3.noWarnings.linkText':'Click here to see the check results.',
    'installer.step3.warnings.noWarnings':'All host checks passed on {0} registered hosts.',
    'installer.step3.warnings.fails':'Some warnings were encountered while performing checks against the {0} registered hosts above',
    'installer.step3.warnings.someWarnings':'All host checks passed on {0} registered hosts. Note: Host checks were skipped on {1} hosts that failed to register.',
    'installer.step3.warnings.allFailed':'Host checks were skipped on {0} hosts that failed to register.',
    'installer.step3.warnings.updateChecks.success':'Host Checks successfully updated',
    'installer.step3.warnings.updateChecks.failed':'Host Checks update failed',
    'installer.step3.warnings.missingHosts':'There is no registered host',
    'installer.step3.warning.registeredHosts': '{0} Other Registered Hosts',
    'installer.step3.warning.loading':'Please wait while the hosts are being checked for potential problems...',
    'installer.step3.registeredHostsPopup': 'These are the hosts that have registered with the server, but do not appear in the list of hosts that you are adding.',
    'installer.step3.removeSelected':'Remove Selected',
    'installer.step3.retryFailed':'Retry Failed',
    'installer.step3.hosts.status.registering':'Registering',
    'installer.step3.hosts.status.installing':'Installing',
    'installer.step3.hosts.bootLog.failed':'\nRegistration with the server failed.',
    'installer.step3.hosts.bootLog.registering':'\nRegistering with the server...',
    'installer.step3.hostLogPopup.highlight':'click to highlight',
    'installer.step3.hostLogPopup.copy':'press CTRL+C',
    'installer.step3.hostsTable.selectAll':'Select All Hosts',
    'installer.step3.selectedHosts.popup.header':'Selected Hosts',

    'installer.step4.header':'Choose Services',
    'installer.step4.body':'Choose which services you want to install on your cluster.',
    'installer.step4.hdfsCheck.popup.header':'File System Required',
    'installer.step4.hdfsCheck.popup.body':'You did not select a File System but one is required. We will automatically add HDFS. Is this OK?',
    'installer.step4.multipleDFS.popup.header':'Multiple File Systems Selected',
    'installer.step4.multipleDFS.popup.body':'You selected more than one file system. We will automatically select only HDFS. Is this OK?',
    'installer.step4.mapreduceCheck.popup.header':'MapReduce Needed',
    'installer.step4.mapreduceCheck.popup.body':'You did not select MapReduce, but it is needed by other services you selected. We will automatically add MapReduce. Is this OK?',
    'installer.step4.yarnCheck.popup.header':'YARN+MapReduce2 Needed',
    'installer.step4.yarnCheck.popup.body':'You did not select YARN+MapReduce2, but it is needed by other services you selected. We will automatically add YARN+MapReduce2. Is this OK?',
    'installer.step4.zooKeeperCheck.popup.header':'ZooKeeper Needed',
    'installer.step4.zooKeeperCheck.popup.body':'You did not select ZooKeeper, but it is needed by other services you selected. We will automatically add ZooKeeper. Is this OK?',
    'installer.step4.oozieCheck.popup.header':'Oozie Needed',
    'installer.step4.oozieCheck.popup.body':'You did not select Oozie, but it is needed by other services you selected. We will automatically add Oozie. Is this OK?',
    'installer.step4.tezCheck.popup.header':'Tez Needed',
    'installer.step4.tezCheck.popup.body':'You did not select Tez, but it is needed by other services you selected. We will automatically add Tez. Is this OK?',
    'installer.step4.monitoringCheck.popup.header':'Limited Functionality Warning',
    'installer.step4.monitoringCheck.popup.body':'You did not select Nagios and/or Ganglia. If both are not selected, monitoring and alerts will not function properly. Is this OK?',

    'installer.step5.header':'Assign Masters',
    'installer.step5.reassign.header':'Select Target Host',
    'installer.step5.attention':' hosts not running master services',
    'installer.step5.body':'Assign master components to hosts you want to run them on.',
    'installer.step5.body.hive':'<i class="icon-asterisks">&#10037;</i> HiveServer2, Hive Metastore, and WebHCat Server will be hosted on the same server.',
    'installer.step5.hostInfo':'%@ (%@, %@ cores)',
    'installer.step5.hiveGroup':'HiveServer2, WebHCat Server, MySQL Server',

    'installer.step6.header':'Assign Slaves and Clients',
    'installer.step6.body':'Assign slave and client components to hosts you want to run them on.<br/>Hosts that are assigned master components are shown with <i class=icon-asterisks>&#10037;</i>. <br/>&quot;Client&quot; will install ',
    'installer.step6.error.mustSelectOne':'You must assign at least one host to each component.',
    'installer.step6.error.mustSelectOneForHost':'You must assign at least one slave/client component to each.',
    'installer.step6.wizardStep6Host.title':'master components hosted on {0}',
    'installer.step6.addHostWizard.body':'Assign HBase master and ZooKeeper server.',
    'installer.step6.error.mustSelectOneForSlaveHost': 'You must assign at least one slave/client component to each host with no master component',

    'installer.step7.header':'Customize Services',
    'installer.step7.body':'We have come up with recommended configurations for the services you selected. Customize them as you see fit.',
    'installer.step7.attentionNeeded':'<b>Attention:</b> Some configurations need your attention before you can proceed.',
    'installer.step7.config.addProperty':'Add Property',
    'installer.step7.ConfigErrMsg.header':'Custom configuration error: ',
    'installer.step7.ConfigErrMsg.message':'Error in custom configuration. Some properties entered in the box are already exposed on this page',
    'installer.step7.popup.currentValue':'Current Value',
    'installer.step7.popup.adjustedValue':'Adjusted Value',
    'installer.step7.popup.rddWarning.header':'Warning: disk space low on {0}',
    'installer.step7.popup.rddWarning.body':'A minimum of 16GB is recommended for the Ganglia Server logs but the disk mount "{0}" on {1} does not have enough space ({2} free). Go to the Misc tab and change Ganglia rrdcached base directory with more than 16GB of disk space. If you proceed without changing it, {1} will likely run out of disk space and become inoperable.',

    'installer.step8.header':'Review',
    'installer.step8.body':'Please review the configuration before installation',
    'installer.step8.deployPopup.message':'Preparing to Deploy: {0} of {1} tasks completed.',
    'installer.step8.hosts':' hosts',
    'installer.step8.host':' host',
    'installer.step8.other':' and {0} other hosts',
    'installer.step8.repoInfo.osType.redhat6':'RHEL 6/CentOS 6/Oracle Linux 6',
    'installer.step8.repoInfo.osType.redhat5':'RHEL 5/CentOS 5/Oracle Linux 5',
    'installer.step8.repoInfo.osType.sles11':'SLES 11/SUSE 11',
    'installer.step8.repoInfo.displayName':'Repositories',
    'installer.step8.securityWarning':'You are running your cluster in secure mode. You must set up the keytabs for all the hosts you are adding before you proceed.',
    'installer.step8.securityConfirmationPopupBody':'Before you proceed, please make sure that the keytabs have been set up on the hosts you are adding per the instructions on the Review page. Otherwise, the assigned components will not be able to start properly on the hosts being added.',

    'installer.step9.header':'Install, Start and Test',
    'installer.step9.body':'Please wait while the selected services are installed and started.',
    'installer.step9.status.success':'Successfully installed and started the services.',
    'installer.step9.status.warning':'Installed and started the services with some warnings.',
    'installer.step9.status.failed':'Failed to install/start the services.',
    'installer.step9.status.start.services.failed':'Start all services API call failed.',
    'installer.step9.status.install.components.failed': 'Some service components are still not known to have installed successfully. Please Retry',
    'installer.step9.status.hosts.heartbeat_lost': 'Ambari agent is not runnig on {0} host.',
    'installer.step9.host.heartbeat_lost': 'Heartbeat lost for the host',
    'installer.step9.host.heartbeat_lost_popup': 'Ambari agent process is not heartbeating on the host',
    'installer.step9.host.heartbeat_lost.header' : 'Hosts with heartbeat lost',
    'installer.step9.host.status.success':'Success',
    'installer.step9.host.status.startAborted':'Install completed. Start aborted',
    'installer.step9.host.status.warning':'Warnings encountered',
    'installer.step9.host.status.failed':'Failures encountered',
    'installer.step9.host.status.nothingToInstall':'Install complete (Waiting to start)',
    'installer.step9.hosts.status.label.inProgress':'In Progress',
    'installer.step9.hosts.status.label.warning':'Warning',
    'installer.step9.hosts.filteredHostsInfo': '{0} of {1} hosts showing',
    'installer.step9.hostLog.popup.header':'Tasks. executed on ',
    'installer.step9.hostLog.popup.categories.all':'All',
    'installer.step9.hostLog.popup.categories.pending':'Queued / Pending',
    'installer.step9.hostLog.popup.categories.in_progress':'In Progress',
    'installer.step9.hostLog.popup.categories.failed':'Failed',
    'installer.step9.hostLog.popup.categories.completed':'Success',
    'installer.step9.hostLog.popup.categories.aborted':'Cancelled',
    'installer.step9.hostLog.popup.categories.timedout':'Timed Out',
    'installer.step9.hostLog.popup.noTasksToShow':'No tasks to show',
    'installer.step9.host.status.noTasks': 'Ambari server did not schedule any tasks on the host. Either the service component on the host is already in installed state <b>OR</b> the pre-check of host reachability failed.',
    'installer.step9.overallProgress':'{0} % overall',
    'installer.step9.serviceStatus.install.pending':'Preparing to install ',
    'installer.step9.serviceStatus.install.queued':'Waiting to install ',
    'installer.step9.serviceStatus.install.inProgress':'Installing ',
    'installer.step9.serviceStatus.install.completed':'Successfully installed ',
    'installer.step9.serviceStatus.install.failed':'Failed to install ',
    'installer.step9.serviceStatus.uninstall.pending':'Preparing to uninstall ',
    'installer.step9.serviceStatus.uninstall.queued':'Waiting to uninstall ',
    'installer.step9.serviceStatus.uninstall.inProgress':'Uninstalling ',
    'installer.step9.serviceStatus.uninstall.completed':'Successfully uninstalled ',
    'installer.step9.serviceStatus.uninstall.failed':'Failed to uninstall ',
    'installer.step9.serviceStatus.start.pending':'Preparing to start ',
    'installer.step9.serviceStatus.start.queued':'Waiting to start ',
    'installer.step9.serviceStatus.start.inProgress':'Starting ',
    'installer.step9.serviceStatus.start.completed':' started successfully',
    'installer.step9.serviceStatus.start.failed':' failed to start',
    'installer.step9.serviceStatus.stop.pending':'Preparing to stop ',
    'installer.step9.serviceStatus.stop.queued':'Waiting to stop ',
    'installer.step9.serviceStatus.stop.inProgress':'Stopping ',
    'installer.step9.serviceStatus.stop.completed':' stopped successfully',
    'installer.step9.serviceStatus.stop.failed':' failed to stop',
    'installer.step9.serviceStatus.execute.pending':'Preparing to execute ',
    'installer.step9.serviceStatus.execute.queued':'Waiting to execute ',
    'installer.step9.serviceStatus.execute.inProgress':'Executing ',
    'installer.step9.serviceStatus.execute.completed':' executed successfully',
    'installer.step9.serviceStatus.execute.failed':' failed to execute',
    'installer.step9.serviceStatus.abort.pending':'Preparing to abort ',
    'installer.step9.serviceStatus.abort.queued':'Waiting to abort ',
    'installer.step9.serviceStatus.abort.inProgress':'Aborting ',
    'installer.step9.serviceStatus.abort.completed':' aborted successfully',
    'installer.step9.serviceStatus.abort.failed':' failed to abort',
    'installer.step9.serviceStatus.upgrade.pending':'Preparing to upgrade ',
    'installer.step9.serviceStatus.upgrade.queued':'Waiting to upgrade ',
    'installer.step9.serviceStatus.upgrade.inProgress':'Upgrading ',
    'installer.step9.serviceStatus.upgrade.completed':' upgraded successfully',
    'installer.step9.serviceStatus.upgrade.failed':' failed to upgrade',
    'installer.step9.components.install.failed': 'Installation Failure',

    'installer.step10.header':'Summary',
    'installer.step10.body':'Here is the summary of the install process.',
    'installer.step10.nagiosRestartRequired':'<b>Important!</b> Restarting Nagios service is required for alerts and notifications to work properly.  After clicking on the Complete button to dismiss this wizard, go to <i>Services -> Nagios</i> to restart the Nagios service.',
    'installer.step10.hostsSummary':'The cluster consists of {0} hosts',
    'installer.step10.servicesSummary':'Installed and started services successfully on {0} new ',
    'installer.step10.warnings':' warnings',
    'installer.step10.clusterState.installing':'Installing ',
    'installer.step10.clusterState.starting':'Starting ',
    'installer.step10.taskStatus.failed':' failed on ',
    'installer.step10.taskStatus.aborted':' aborted on ',
    'installer.step10.taskStatus.timedOut':' timed out on ',
    'installer.step10.installStatus.failed':'Installing master services failed',
    'installer.step10.installStatus.installed':'Master services installed',
    'installer.step10.master.nameNode':'NameNode installed on ',
    'installer.step10.master.secondaryNameNode':'SecondaryNameNode installed on ',
    'installer.step10.master.jobTracker':'JobTracker installed on ',
    'installer.step10.master.historyServer':'History Server installed on ',
    'installer.step10.master.resourceManager':'ResourceManager installed on ',
    'installer.step10.master.zooKeeper':'ZooKeeper installed on ',
    'installer.step10.master.hbase':'HBase Master installed on ',
    'installer.step10.master.hiveMetastore':'Hive Metastore installed on ',
    'installer.step10.master.oozie':'Oozie Server installed on ',
    'installer.step10.master.ganglia':'Ganglia Server installed on ',
    'installer.step10.master.nagios':'Nagios Server installed on ',
    'installer.step10.startStatus.failed':'Starting services failed',
    'installer.step10.startStatus.passed':'All tests passed',
    'installer.step10.startStatus.started':'All services started',
    'installer.step10.installTime.seconds':'Install and start completed in {0} seconds',
    'installer.step10.installTime.minutes':'Install and start completed in {0} minutes and {1} seconds',

    'addHost.step4.header':'Configurations',
    'addHost.step4.title':'Select the configuration groups to which the added hosts will belong to.',


    'installer.stackUpgrade.header':'Stack Upgrade Wizard',
    'installer.stackUpgrade.step1.newVersion':'New Version',
    'installer.stackUpgrade.step1.installedVersion':'Installed Version',
    'installer.stackUpgrade.step1.installedStackVersion':'Installed stack version',
    'installer.stackUpgrade.step1.upgradeStackVersion':'Upgrade stack version',
    'installer.stackUpgrade.step1.description':'This Stack Upgrade Wizard will walk you through the steps of upgrading the cluster to the latest available stack version.',

    'installer.stackUpgrade.step2.notice.header':'Important before you proceed, please perform following.',
    'installer.stackUpgrade.step2.notice.first':'Make sure that your NameNode is backed up.',
    'installer.stackUpgrade.step2.notice.second':'Make sure that your NameNode is backed up.',
    'installer.stackUpgrade.step2.notice.third':'Make sure that your NameNode is backed up.',
    'installer.stackUpgrade.step2.notice.complete':'Once you have completed the above, click on Upgrade to initiate the upgrade process on all hosts in your cluster',
    'installer.stackUpgrade.step2.advancedOption':'Advanced Option',
    'installer.stackUpgrade.step2.localRepository':'Use a local repository',
    'installer.stackUpgrade.step2.popup.body':'We cannot proceed since some of the master components are not currently running. Please ensure that all services are running before continuing.',
    'installer.stackUpgrade.step3.ProceedWithWarning':'Proceed with Warning',
    'installer.stackUpgrade.step3.status.success':'Successfully upgraded the cluster to {0}',
    'installer.stackUpgrade.step3.status.info':"Upgrading the cluster. \nPlease wait while we perform cluster upgrade.",
    'installer.stackUpgrade.step3.status.warning':"Upgraded the cluster to {0} with some warnings.\nYou can start using the cluster but the components that failed to upgrade will not be functional."+
    "You can click on the Retry button to retry upgrading the failed components. Alternatively you can proceed and retry upgrade on individual components in the Host Detail page.",
    'installer.stackUpgrade.step3.status.failed':"Failed to upgrade hosts. Click on each host to see what might have gone wrong.\n After fixing the problem, click the Retry button",
    'installer.stackUpgrade.step3.host.nothingToUpgrade':'Install complete (Waiting to start)',
    'installer.stackUpgrade.step3.service.upgraded':'Services upgraded',
    'installer.stackUpgrade.step3.service.upgrading':'Services upgrade in progress',
    'installer.stackUpgrade.step3.service.pending':'Services upgrade pending',
    'installer.stackUpgrade.step3.service.failedUpgrade':'Services failed to upgrade',
    'installer.stackUpgrade.step3.service.stopped':'All Services stopped',
    'installer.stackUpgrade.step3.service.stopping':'All Services stopping',
    'installer.stackUpgrade.step3.service.stopFail':'All Services failed to stop',
    'installer.stackUpgrade.step3.service.stopPending':'All Services stop pending',
    'installer.stackUpgrade.step3.retry.upgrade':'Retry Upgrade',
    'installer.stackUpgrade.step3.retry.services':'Retry stopping services',
    'installer.stackUpgrade.step3.upgrade.header':'All Services upgrade',
    'installer.stackUpgrade.step3.stop.header':'All Services stop',

    'form.create':'Create',
    'form.save':'Save',
    'form.cancel':'Cancel',
    'form.passwordRetype':'Retype Password',
    'form.saveSuccess':'Successfully saved.',
    'form.saveError':'Sorry, errors occurred.',
    'form.item.placeholders.typePassword':'Type password',

    'form.validator.invalidIp':'Please enter valid ip address',
    'form.validator.configKey':'Invalid Key. Only alphanumerics, hyphens, underscores, asterisks and periods are allowed.',

    'admin.advanced.caution':'This section is for advanced user only.<br/>Proceed with caution.',
    'admin.advanced.button.uninstallIncludingData':'Uninstall cluster including all data.',
    'admin.advanced.button.uninstallKeepData':'Uninstall cluster but keep data.',

    'admin.advanced.popup.header':'Uninstall Cluster',

    /*'admin.advanced.popup.text':'Uninstall Cluster',*/

    'admin.audit.grid.date':"Date/Time",
    'admin.audit.grid.category':"Category",
    'admin.audit.grid.operationName':"Operation",
    'admin.audit.grid.performedBy':"Performed By",
    'admin.audit.grid.service':"Category",

    'admin.authentication.form.method.database':'Use Ambari Database to authenticate users',
    'admin.authentication.form.method.ldap':'Use LDAP/Active Directory to authenticate',
    'admin.authentication.form.primaryServer':'Primary Server',
    'admin.authentication.form.secondaryServer':'Secondary Server',
    'admin.authentication.form.useSsl':'Use SSL',
    'admin.authentication.form.bind.anonymously':"Bind Anonymously",
    'admin.authentication.form.bind.useCrenedtials':"Use Credentials To Bind",
    'admin.authentication.form.bindUserDN':'Bind User DN',
    'admin.authentication.form.searchBaseDN':'Search Base DN',
    'admin.authentication.form.usernameAttribute':'Username Attribute',

    'admin.authentication.form.userDN':'User DN',
    'admin.authentication.form.configurationTest':'Configuration Test',
    'admin.authentication.form.testConfiguration':'Test Configuration',

    'admin.authentication.form.test.success':'The configuration passes the test',
    'admin.authentication.form.test.fail':'The configuration fails the test',

    'admin.highAvailability':' High Availability',
    'admin.highAvailability.button.enable':'Enable NameNode HA',
    'admin.highAvailability.button.disable':'Disable NameNode HA',
    'admin.highAvailability.disabled':'NameNode HA is disabled',
    'admin.highAvailability.enabled':'NameNode HA is enabled',
    'admin.highAvailability.confirmRollbackHeader':'Confirm Rollback',
    'admin.highAvailability.confirmRollbackBody':'This will rollback all operations that were done in HA wizard',
    'admin.highAvailability.confirmManualRollbackBody':'You are in the process of enabling NameNode HA. If you exit now, you must follow manual instructions to revert back to the non-HA setup as documented in the Ambari User Guide\'s <i>Rolling Back NameNode HA</i> section.  Are you sure you want to exit the wizard?',
    'admin.highAvailability.error.hostsNum':'You must have at least 3 hosts in your cluster to enable NameNode HA.',
    'admin.highAvailability.error.namenodeStarted':'NameNode must be running before you enable NameNode HA.',
    'admin.highAvailability.error.zooKeeperNum':'You must have at least 3 ZooKeeper Servers in your cluster to enable NameNode HA.',
    'admin.highAvailability.error.security':'You cannot enable NameNode HA via this wizard as your cluster is already secured.  First, disable security by going to Admin > Security, and then run this Enable NameNode HA wizard again.  After NameNode HA is enabled, you can go back to Admin > Security to secure the cluster.',
    'admin.highAvailability.wizard.header':'Enable NameNode HA Wizard',
    'admin.highAvailability.wizard.progressPage.notice.inProgress':'Please wait while NameNode HA is being deployed.',
    'admin.highAvailability.wizard.progressPage.notice.completed':'Please proceed to the next step.',
    'admin.highAvailability.wizard.progressPage.notice.failed':'You can click on the Retry button to retry failed tasks.',
    'admin.highAvailability.wizard.progressPage.header':'Deploy',
    'admin.highAvailability.wizard.step1.header':'Get Started',
    'admin.highAvailability.wizard.step1.nameserviceid.tooltip.title':'Nameservice ID',
    'admin.highAvailability.wizard.step1.nameserviceid.tooltip.content':'This will be the ID for the NameNode HA cluster. For example, if you set Nameservice ID to <b>mycluster</b>, the logical URI for HDFS will be <b>hdfs://mycluster</b>.',
    'admin.highAvailability.wizard.step1.nameserviceid':'Nameservice ID',
    'admin.highAvailability.wizard.step1.nameserviceid.error':'Must consist of letters, numbers, and hyphens. Cannot begin or end with a hyphen.',
    'admin.highAvailability.wizard.step2.header':'Select Hosts',
    'admin.highAvailability.wizard.step3.header':'Review',
    'admin.highAvailability.wizard.step4.header':'Create Checkpoint',
    'admin.highAvailability.wizard.step5.header':'Configure Components',
    'admin.highAvailability.wizard.step6.header':'Initialize JournalNodes',
    'admin.highAvailability.wizard.step7.header':'Start Components',
    'admin.highAvailability.wizard.step8.header':'Initialize Metadata',
    'admin.highAvailability.wizard.step9.header':'Finalize HA Setup',
    'admin.highAvailability.wizard.step4.bodyHeader':'Manual Steps Required: Create Checkpoint on NameNode',
    'admin.highAvailability.wizard.step6.bodyHeader':'Manual Steps Required: Initialize JournalNodes',
    'admin.highAvailability.wizard.step8.bodyHeader':'Manual Steps Required: Initialize NameNode HA Metadata',

    'admin.rollbackHighAvailability.wizard.step1.header':'Select Hosts page',
    'admin.rollbackHighAvailability.wizard.step2.header':'Create Checkpoint page',
    'admin.rollbackHighAvailability.wizard.step3.header':'Progress page',

    'admin.highAvailability.wizard.step5.notice.inProgress':'Please wait while the wizard configures the components.',
    'admin.highAvailability.wizard.step7.notice.inProgress':'Please wait while the wizard starts the components.',
    'admin.highAvailability.wizard.step9.notice.inProgress':'Please wait while the wizard finalizes the HA setup.',
    'admin.highAvailability.wizard.rollback.notice.inProgres':'Reverting Back to Non-HA Setup',

    'admin.highAvailability.wizard.step5.header.title':'Configure Components',
    'admin.highAvailability.wizard.step7.header.title':'Start Components',
    'admin.highAvailability.wizard.step9.header.title':'Finalize HA Setup',
    'admin.highAvailability.wizard.rollback.header.title':'Reverting Back to Non-HA Setup.',

    'admin.highAvailability.wizard.step5.task0.title':'Stop All Services',
    'admin.highAvailability.wizard.step5.task1.title':'Install Additional NameNode',
    'admin.highAvailability.wizard.step5.task2.title':'Install JournalNodes',
    'admin.highAvailability.wizard.step5.task3.title':'Reconfigure HDFS',
    'admin.highAvailability.wizard.step5.task4.title':'Start JournalNodes',
    'admin.highAvailability.wizard.step5.task5.title':'Disable Secondary NameNode',

    'admin.highAvailability.wizard.step7.task0.title':'Start ZooKeeper Servers',
    'admin.highAvailability.wizard.step7.task1.title':'Start NameNode',

    'admin.highAvailability.wizard.step9.task0.title':'Start Additional NameNode',
    'admin.highAvailability.wizard.step9.task1.title':'Install Failover Controllers',
    'admin.highAvailability.wizard.step9.task2.title':'Start Failover Controllers',
    'admin.highAvailability.wizard.step9.task3.title':'Reconfigure HBase',
    'admin.highAvailability.wizard.step9.task4.title':'Delete Secondary NameNode',
    'admin.highAvailability.wizard.step9.task5.title':'Start All Services',
    'admin.highAvailability.wizard.step9.notice.completed':'NameNode HA has been enabled successfully.',

    'admin.highAvailability.wizard.step3.curNameNode': '<b>Current NameNode:</b> ',
    'admin.highAvailability.wizard.step3.addNameNode': '<b>Additional NameNode:</b> ',
    'admin.highAvailability.wizard.step3.secNameNode': '<b>Secondary NameNode:</b> ',
    'admin.highAvailability.wizard.step3.journalNode': '<b>JournalNode:</b> ',
    'admin.highAvailability.wizard.step3.toBeInstalled': 'TO BE INSTALLED',
    'admin.highAvailability.wizard.step3.toBeDeleted': 'TO BE DELETED',
    'admin.highAvailability.wizard.step4.ckNotCreated':'Checkpoint not created yet',
    'admin.highAvailability.wizard.step4.ckCreated':'Checkpoint created',
    'admin.highAvailability.wizard.step6.jsNoInit':'JournalNodes not initialized yet',
    'admin.highAvailability.wizard.step6.jsInit':'JournalNodes initialized',
    'admin.highAvailability.wizard.step8.metaNoInit':'Metadata not initialized yet',
    'admin.highAvailability.wizard.step8.confirmPopup.body':'Please confirm that you have run the manual steps before continuing.',

    'admin.highAvailability.rollback.header':'Disable NameNode HA Wizard',
    'admin.highAvailability.rollback.task0.title':'Stop All Services',
    'admin.highAvailability.rollback.task1.title':'Restore HBase Configurations',
    'admin.highAvailability.rollback.task2.title':'Stop Failover Controllers',
    'admin.highAvailability.rollback.task3.title':'Delete Failover Controllers',
    'admin.highAvailability.rollback.task4.title':'Stop Additional NameNode',
    'admin.highAvailability.rollback.task5.title':'Stop NameNode',
    'admin.highAvailability.rollback.task6.title':'Restore HDFS Configurations',
    'admin.highAvailability.rollback.task7.title':'Enable Secondary NameNode',
    'admin.highAvailability.rollback.task8.title':'Stop JournalNodes',
    'admin.highAvailability.rollback.task9.title':'Delete JournalNodes',
    'admin.highAvailability.rollback.task10.title':'Delete Additional NameNode',
    'admin.highAvailability.rollback.task11.title':'Start All Services',
    'admin.highAvailability.rollback.notice.inProgress':'Please wait while the wizard reverts back to the non-HA setup.',

    'admin.highAvailability.rollback.step2.body':
      '<ol>' +
        '<li>Login to the NameNode host <b>{1}</b>.</li>' +
        '<li>Put the NameNode in Safe Mode (read-only mode):' +
        '<div class="code-snippet">sudo su -l {0} -c \'hdfs dfsadmin -safemode enter\'</div></li>' +
        '<li>Once in Safe Mode, create a Checkpoint:' +
        '<div class="code-snippet">sudo su -l {0} -c \'hdfs dfsadmin -saveNamespace\'</div></li>' +
        '</ol>',

    'admin.highAvailability.wizard.step8.body':
      '<div class="alert alert-info">' +
      '<ol>' +
      '<li>Login to the NameNode host <b>{1}</b>.</li>' +
      '<li>Initialize the metadata for NameNode automatic failover by running:' +
      '<div class="code-snippet">sudo su -l {0} -c \'hdfs zkfc -formatZK\'</div></li>' +
      '</div>' +
      '<div class="alert alert-info">' +
      '<ol start="3">' +
      '<li>Login to the Additional NameNode host <b>{2}</b>.<br>' +
      '<div class="alert alert-warn"><strong>Important!</strong> Be sure to login to the Additional NameNode host.<br>This is a different host from the Steps 1 and 2 above.</div>' +
      '</li>' +
      '<li>Initialize the metadata for the Additional NameNode by running:' +
      '<div class="code-snippet">sudo su -l {0} -c \'hdfs namenode -bootstrapStandby\'</div></li>' +
      '</ol>' +
      '</div>' +
      'Please proceed once you have completed the steps above.',
    'admin.highAvailability.wizard.step6.body':
      '<ol>' +
      '<li>Login to the NameNode host <b>{1}</b>.</li>' +
      '<li>Initialize the JournalNodes by running:' +
      '<div class="code-snippet">sudo su -l {0} -c \'hdfs namenode -initializeSharedEdits\'</div></li>' +
      '<li>You will be able to proceed once Ambari detects that the JournalNodes have been initialized successfully.</li>' +
      '</ol>',
    'admin.highAvailability.wizard.step4.body':
      '<ol>' +
      '<li>Login to the NameNode host <b>{1}</b>.</li>' +
      '<li>Put the NameNode in Safe Mode (read-only mode):' +
      '<div class="code-snippet">sudo su -l {0} -c \'hdfs dfsadmin -safemode enter\'</div></li>' +
      '<li>Once in Safe Mode, create a Checkpoint:' +
      '<div class="code-snippet">sudo su -l {0} -c \'hdfs dfsadmin -saveNamespace\'</div></li>' +
      '<li>You will be able to proceed once Ambari detects that the NameNode is in Safe Mode and the Checkpoint has been created successfully.</li>'+
      '<div class="alert alert-warn">If the <b>Next</b> button is enabled before you run the <b>"Step 3: Create a Checkpoint"</b> command, it means there is a recent Checkpoint already and you may proceed without running the <b>"Step 3: Create a Checkpoint"</b> command.</div>' +
      '</ol>',
    'admin.highAvailability.wizard.step3.confirm.host.body':'<b>Confirm your host selections.</b>',
    'admin.highAvailability.wizard.step3.confirm.config.body':'<div class="alert alert-info">' +
      '<b>Review Configuration Changes.</b></br>' +
      'The following lists the configuration changes that will be made by the Wizard to enable NameNode HA. This information is for <b> review only </b> and is not editable except for the  <b>dfs.journalnode.edits.dir</b> property' +
      '</div>',
    'admin.highAvailability.wizard.step2.body':'Select a host that will be running the additional NameNode.<br/> In addition,' +
      ' select the hosts to run JournalNodes, which store NameNode edit logs in a fault tolerant manner.',
    'admin.highAvailability.wizard.step1.body':'This wizard will walk you through enabling NameNode HA on your cluster.<br/>' +
      'Once enabled, you will be running a Standby NameNode in addition to your Active NameNode.<br/>' +
      'This allows for an Active-Standby NameNode configuration that automatically performs failover.<br/><br/>' +
      'The process to enable HA involves a combination of <b>automated steps</b> (that will be handled by the wizard) and ' +
      '<b>manual steps</b> (that you must perform in sequence as instructed by the wizard).<br/><br/>' +
      '<b>You should plan a cluster maintenance window and prepare for cluster downtime when enabling NameNode HA.</b>',
    'admin.highAvailability.wizard.step1.alert':'If you have HBase running, please exit this wizard and stop HBase first.',
    'admin.security.title':'Kerberos security has not been enabled',
    'admin.security.enabled': 'Kerberos security is enabled',
    'admin.security.disabled': 'Kerberos security is disabled',
    'admin.security.button.enable':'Enable Security',
    'admin.security.button.disable':'Disable Security',
    'admin.security.enable.popup.body': 'We will walk you through add security wizard',
    'admin.security.enable.popup.header': 'Add security',
    'admin.security.disable.popup.header': 'Remove security',
    'admin.security.disable.popup.body': 'Kerberos security will be disabled on the cluster',
    'admin.security.disable.popup.body.warning' : 'Note: Before proceeding, you need to manually remove all directories listed in the mapred.local.dir property in mapred-site.xml on all TaskTracker hosts; otherwise, MapReduce will not run properly after disabling security.',
    'admin.addSecurity.header': 'Enable Security Wizard',
    'admin.security.step1.header': 'Get Started',
    'admin.security.step2.header': 'Configure Services',
    'admin.security.step3.header': 'Create Principals and Keytabs',
    'admin.security.step4.header': 'Save and Apply Configuration',
    'admin.security.step1.body.header': 'Important: Before configuring Ambari to manage your Kerberos-enabled cluster, ' +
      'you must perform the following manual steps on your cluster. Be sure to record the location of the keytab files ' +
      'for each host and the principals for each Hadoop service. This information is required in order to use the wizard.',
    'admin.security.step1.body.instruction1': 'Install, configure and start your Kerberos KDC',
    'admin.security.step1.body.instruction2': 'Install and configure the Kerberos client on every host in the cluster',
    'admin.security.step1.body.instruction3': 'Create Kerberos principals for Hadoop services and hosts',
    'admin.security.step1.body.instruction4': 'Generate keytabs for each principal and place on the appropriate hosts',
    'admin.security.step1.body.instruction5': '<b>Application Timeline Server</b> component of YARN service will be <span class="text-error"><b>deleted</b></span> as part of enabling security',
    'admin.security.step2.body.header': 'Configure Kerberos security properties',
    'admin.security.step3.notice': 'You need to create the following principals and keytabs on the hosts shown.<br />'+
    'You can download the list as a CSV file and use it to create a script to generate the principals and keytabs. ' +
    'Once the principals and keytabs have been created, click on <i>Apply</i> to continue. If you need to make configuration changes, click <i>Back</i>.',
    'admin.security.step3.table.principal': 'Principal',
    'admin.security.step3.table.keytab': 'Keytab',
    'admin.security.step3.downloadCSV': 'Download CSV',
    'admin.security.step4.body.header': 'Applying kerberos security to the cluster',
    'admin.security.step4.body.success.header' : 'Kerberos-based security has been enabled on your cluster. Please wait while services are started in secure mode.',
    'admin.security.step4.body.failure.header' : 'Failed to enable Kerberos-based security on your cluster. Your cluster will keep running in non-secure mode.',
    'admin.security.apply.configuration.error' : 'Failed to apply secure configurations to the cluster. Please navigate to "Configure Services" step and make sure all services are configured with appropriate values.',
    'admin.security.disable.body.header' : 'Disabling kerberos security on the cluster',
    'admin.security.disable.body.success.header': 'Kerberos-based security has been disabled on your cluster. Please wait while services are started in non-secure mode.',
    'admin.security.disable.body.failure.header': 'Failed to disable Kerberos-based security on your cluster. Your cluster will keep running in secure mode.',
    'admin.addSecurity.apply.stop.services': 'Stop Services',
    'admin.addSecurity.apply.save.config': 'Save Configurations',
    'admin.addSecurity.apply.start.services': 'Start Services',
    'admin.addSecurity.apply.delete.ats': 'Delete ATS',
    'admin.addSecurity.user.smokeUser': 'Ambari Smoke Test User',
    'admin.addSecurity.user.hdfsUser': 'HDFS User',
    'admin.addSecurity.user.hbaseUser': 'HBase User',
    'admin.addSecurity.user.stormUser': 'Storm User',
    'admin.addSecurity.hdfs.user.httpUser': 'HDFS SPNEGO User',
    'admin.addSecurity.rm.user.httpUser': 'ResourceManager SPNEGO User',
    'admin.addSecurity.nm.user.httpUser': 'NodeManager SPNEGO User',
    'admin.addSecurity.historyServer.user.httpUser': 'History server SPNEGO User',
    'admin.addSecurity.webhcat.user.httpUser': 'WebHCat SPNEGO User',
    'admin.addSecurity.oozie.user.httpUser': 'Oozie SPNEGO User',
    'admin.addSecurity.falcon.user.httpUser': 'Falcon SPNEGO User',
    'admin.addSecurity.enable.onClose': 'You are in the process of enabling security on your cluster. ' +
      'Are you sure you want to quit? If you quit, ' +
      'you may have to re-run the security wizard from the beginning to enable security.',
    'admin.addSecurity.enable.after.stage2.onClose': 'Services are being started with the Kerberos settings you specified. '+
      'It is recommended that you wait until all the services are started to ensure that they are set up properly. ' +
      'Are you sure you want to quit?',
    'admin.addSecurity.disable.onClose': 'You are in the process of disabling security on your cluster. ' +
      'Are you sure you want to quit?',
    'admin.removeSecurity.header': 'Disable Security',
    'admin.security.applying.config.header': 'Applying Configurations',
    'admin.security.applying.config.body':'You cannot quit wizard while configurations are being applied',
    'admin.security.status.error' : 'Error in retrieving cluster security status from Ambari server',
    'admin.users.ldapAuthUsed':'LDAP Authentication is being used to authenticate users',
    'admin.users.delete.yourself.message':'You can\'t delete yourself',
    'admin.users.delete.yourself.header':'Deleting warning',

    'admin.users.delete.header':'Delete {0}',

    'admin.users.addButton':'Add Local User',
    'admin.users.editButton': 'Edit Local User',
    'admin.users.delete':'delete',
    'admin.users.edit':'edit',
    'admin.users.privileges':'Admin',
    'admin.users.type':'Type',
    'admin.users.action':'Action',
    'admin.users.passwordRetype':'Retype Password',
    'admin.users.username':'Username',
    'admin.users.createSuccess': 'User successfully created.',
    'admin.users.createError': 'Error occurred while user creating.',
    'admin.users.createError.passwordValidation': 'Passwords are different',
    'admin.users.createError.userNameExists': 'User with the same name is already exists',
    'admin.users.editError.requiredField': 'This is required',

    'admin.access.showJobs':'Enable Jobs tab for non-admin users',

    'admin.confirmUninstall':'Confirm Uninstall',
    'admin.cluster.stackVersion':'Cluster Stack Version',
    'admin.cluster.upgradeAvailable':'Upgrade available',
    'admin.cluster.upgradeUnavailable':'Upgrade unavailable',
    'admin.cluster.repositories.repositories':'Repositories',
    'admin.cluster.repositories.os':'OS',
    'admin.cluster.repositories.baseUrl':'Base URL',

    'admin.misc.header': 'Service Users and Groups',
    'admin.misc.nothingToShow': 'No user accounts to display',

    'services.service.start':'Start',
    'services.service.stop':'Stop',
    'services.nothingToAdd':'Nothing to add',
    'services.service.summary.version':'Version',
    'services.service.summary.viewHost':'View Host',
    'services.service.summary.viewHosts':'View Hosts',
    'services.service.summary.clientOnlyService.ToolTip':'Client-only service',
    'services.service.summary.DataNodesLive':'DataNodes Live',
    'services.service.summary.JournalNodesLive':'JournalNodes Live',
    'services.service.summary.mapreduce2.client':'MapReduce2 Client',
    'services.service.summary.mapreduce2.clients':'MapReduce2 Clients',
    'services.service.summary.nodeManagersLive':'NodeManagers Live',
    'services.service.summary.TrackersLive':'Trackers Live',
    'services.service.summary.RegionServersLIVE':'RegionServers Live',
    'services.service.summary.GangliaMonitorsLIVE':'Ganglia Monitors Live',
    'services.service.summary.SupervisorsLIVE':'Supervisors Live',
    'services.service.summary.nameNode':'NameNode Web UI',
    'services.service.summary.nameNodeUptime':'NameNode Uptime',
    'services.service.summary.nameNodeHeap':'NameNode Heap',
    'services.service.summary.nameNode.active':'Active NameNode',
    'services.service.summary.nameNode.standby':'Standby NameNode',
    'services.service.summary.pendingUpgradeStatus':'Upgrade Status',
    'services.service.summary.pendingUpgradeStatus.pending':'Pending upgrade',
    'services.service.summary.pendingUpgradeStatus.notPending':'No pending upgrade',
    'services.service.summary.safeModeStatus':'Safe Mode Status',
    'services.service.summary.safeModeStatus.inSafeMode':'In safe mode',
    'services.service.summary.safeModeStatus.notInSafeMode':'Not in safe mode',
    'services.service.summary.dataNodes':'DataNodes',
    'services.service.summary.diskCapacity':'HDFS Disk Capacity',
    'services.service.summary.blocksTotal':'Blocks (total)',
    'services.service.summary.blockErrors':'Block Errors',
    'services.service.summary.totalFiles':'Total Files + Dirs',
    'services.service.summary.jobTracker':'JobTracker',
    'services.service.summary.jobTrackerWebUI':'JobTracker Web UI',
    'services.service.summary.jobTrackerUptime':'JobTracker Uptime',
    'services.service.summary.trackersLiveTotal':'Trackers',
    'services.service.summary.trackersBlacklistGraylist':'Trackers',
    'services.service.summary.jobTrackerHeap':'JobTracker Heap',
    'services.service.summary.totalSlotsCapacity':'Total Slots Capacity',
    'services.service.summary.totalJobs':'Total Jobs',
    'services.service.summary.currentSlotUtiliMaps':'Map Slots',
    'services.service.summary.currentSlotUtiliReduces':'Reduce Slots',
    'services.service.summary.tasksMaps':'Tasks: Maps',
    'services.service.summary.tasksReduces':'Tasks: Reduces',
    'services.service.summary.hbaseMaster':'HBase Master Web UI',
    'services.service.summary.regionServerCount':'RegionServer Count',
    'services.service.summary.regionInTransition':'Region In Transition',
    'services.service.summary.masterStarted':'Master Started',
    'services.service.summary.masterActivated':'Master Activated',
    'services.service.summary.averageLoad':'Average Load',
    'services.service.summary.masterHeap':'Master Heap',
    'services.service.summary.moreStats':'more stats here',
    'services.service.summary.clientCount': '{0} Client Hosts',
    'services.service.summary.historyServer': 'History Server Web UI',
    'services.service.actions.run.rebalancer':'Run Rebalancer',
    'services.service.actions.run.compaction':'Run Compaction',
    'services.service.actions.run.smoke':'Run Service Check',
    'services.service.actions.reassign.master':'Move {0}',
    'services.service.actions.reassign.master.hive':'Move HiveServer2, WebHCat Server, MySQL Server',
    'services.service.actions.manage_configuration_groups':'Manage Configuration Groups...',
    'services.service.actions.manage_configuration_groups.short':'Manage Config Groups',
    'services.service.actions.serviceActions':'Service Actions',
    'services.service.summary.unknown':'unknown',
    'services.service.summary.notRunning':'Not Running',
    'services.service.summary.notAvailable':'n/a',
    'services.service.summary.diskInfoBar.used':'used',
    'services.service.summary.storm.freeslots': 'Free slots',
    'services.service.summary.storm.executors': 'Executors',
    'services.service.summary.storm.tasks': 'Tasks',
    'services.service.summary.storm.nimbus.uptime': 'Nimbus uptime',
    'services.service.summary.storm.topologies': 'Topologies',

    'services.service.info.metrics.flume.channelFillPercent':'Channel Fill Percentage',
    'services.service.info.metrics.flume.channelSize':'Channel Size',
    'services.service.info.metrics.flume.sinkDrainSuccess':'Sink Event Drain Count',
    'services.service.info.metrics.flume.sourceAccepted':'Source Event Accepted Count',
    'services.service.info.metrics.flume.sinkConnectionFailed':'Sink Connection Failed Count',
    'services.service.info.metrics.flume.gc':'Garbage Collection Time',
    'services.service.info.metrics.flume.cpu.user':'CPU (User)',
    'services.service.info.metrics.flume.jvmThreadsRunnable':'JVM Runnable Threads',
    'services.service.info.metrics.flume.jvmHeapUsed':'JVM Heap Memory Used',
    'services.service.info.metrics.flume.channelName':'Channel {0}',
    'services.service.info.metrics.flume.sinkName':'Sink {0}',
    'services.service.info.metrics.flume.sourceName':'Source {0}',
    'services.service.info.metrics.flume.hostName':'Host: {0}',

    'services.service.info.metrics.hbase.clusterRequests':'Cluster Requests',
    'services.service.info.metrics.hbase.clusterRequests.displayNames.requestCount':'Request Count',
    'services.service.info.metrics.hbase.hlogSplitSize':'HLog Split Size',
    'services.service.info.metrics.hbase.hlogSplitSize.displayNames.splitSize':'Split Size',
    'services.service.info.metrics.hbase.hlogSplitTime':'HLog Split Time',
    'services.service.info.metrics.hbase.hlogSplitTime.displayNames.splitTime':'Split Time',
    'services.service.info.metrics.hbase.regionServerQueueSize':'RegionServer Queue Size',
    'services.service.info.metrics.hbase.regionServerQueueSize.displayNames.compactionQueueSize':'Compaction Queue Size',
    'services.service.info.metrics.hbase.regionServerQueueSize.displayNames.flushQueueSize':'Flush Queue Size',
    'services.service.info.metrics.hbase.regionServerRegions':'RegionServer Regions',
    'services.service.info.metrics.hbase.regionServerRegions.displayNames.regions':'Regions',
    'services.service.info.metrics.hbase.regionServerRequests':'Cumulative Requests',
    'services.service.info.metrics.hbase.regionServerRequests.2':'RegionServer Requests',
    'services.service.info.metrics.hbase.regionServerRequests.displayNames.writeRequests':'Write Requests',
    'services.service.info.metrics.hbase.regionServerRequests.displayNames.readRequests':'Read Requests',

    'services.service.info.metrics.hdfs.blockStatus':'Block Status',
    'services.service.info.metrics.hdfs.blockStatus.displayNames.pendingReplicationBlocks':'Pending Replication Blocks',
    'services.service.info.metrics.hdfs.blockStatus.displayNames.underReplicatedBlocks':'Under Replicated Blocks',
    'services.service.info.metrics.hdfs.fileOperations':'File Operations',
    'services.service.info.metrics.hdfs.fileOperations.displayNames.fileInformationOperations':'File Information Operations',
    'services.service.info.metrics.hdfs.fileOperations.displayNames.deleteFileOperations':'Delete File Operations',
    'services.service.info.metrics.hdfs.fileOperations.displayNames.createFileOperations':'Create File Operations',
    'services.service.info.metrics.hdfs.gc':'Garbage Collection',
    'services.service.info.metrics.hdfs.gc.displayNames.gcTimeMillis':'Time',
    'services.service.info.metrics.hdfs.io':'HDFS I/O',
    'services.service.info.metrics.hdfs.io.displayNames.bytesWritten':'Bytes Written',
    'services.service.info.metrics.hdfs.io.displayNames.bytesRead':'Bytes Read',
    'services.service.info.metrics.hdfs.jvmHeap':'JVM Memory Status',
    'services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapCommittedM':'Heap Memory Committed',
    'services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapUsedM':'Non Heap Memory Used',
    'services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapUsedM':'Heap Memory Used',
    'services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapCommittedM':'Non Heap Memory Committed',
    'services.service.info.metrics.hdfs.jvmThreads':'JVM Thread Status',
    'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsBlocked':'Threads Blocked',
    'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsWaiting':'Threads Waiting',
    'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsTimedWaiting':'Threads Timed Waiting',
    'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsRunnable':'Threads Runnable',
    'services.service.info.metrics.hdfs.rpc':'RPC',
    'services.service.info.metrics.hdfs.rpc.displayNames.rpcQueueTimeAvgTime':'Queue Average Wait Time',
    'services.service.info.metrics.hdfs.spaceUtilization':'Total Space Utilization',
    'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityRemainingGB':'Capacity Remaining',
    'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityUsedGB':'Capacity Used',
    'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityTotalGB':'Capacity Total',

    'services.service.info.metrics.mapreduce.gc':'Garbage Collection',
    'services.service.info.metrics.mapreduce.gc.displayNames.gcTimeMillis':'Time',
    'services.service.info.metrics.mapreduce.jobsStatus':'Jobs Status',
    'services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsRunning':'Running',
    'services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsFailed':'Failed',
    'services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsCompleted':'Succeeded',
    'services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsPreparing':'Preparing',
    'services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsSubmitted':'Submitted',
    'services.service.info.metrics.mapreduce.jvmHeap':'JVM Memory Status',
    'services.service.info.metrics.mapreduce.jvmHeap.displayNames.memHeapCommittedM':'Heap Memory Committed',
    'services.service.info.metrics.mapreduce.jvmHeap.displayNames.memNonHeapUsedM':'Non Heap Memory Used',
    'services.service.info.metrics.mapreduce.jvmHeap.displayNames.memHeapUsedM':'Heap Memory Used',
    'services.service.info.metrics.mapreduce.jvmHeap.displayNames.memNonHeapCommittedM':'Non Heap Memory Committed',
    'services.service.info.metrics.mapreduce.jvmThreads':'JVM Thread Status',
    'services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsBlocked':'Threads Blocked',
    'services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsWaiting':'Threads Waiting',
    'services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsTimedWaiting':'Threads Timed Waiting',
    'services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsRunnable':'Threads Runnable',
    'services.service.info.metrics.mapreduce.mapSlots':'Map Slots Utilization',
    'services.service.info.metrics.mapreduce.mapSlots.displayNames.reservedMapSlots':'Map Slots Reserved',
    'services.service.info.metrics.mapreduce.mapSlots.displayNames.occupiedMapSlots':'Map Slots Occupied',
    'services.service.info.metrics.mapreduce.reduceSlots':'Reduce Slots Utilization',
    'services.service.info.metrics.mapreduce.reduceSlots.displayNames.reservedReduceSlots':'Reduce Slots Reserved',
    'services.service.info.metrics.mapreduce.reduceSlots.displayNames.occupiedReduceSlots':'Reduce Slots Occupied',
    'services.service.info.metrics.mapreduce.rpc':'RPC',
    'services.service.info.metrics.mapreduce.rpc.displayNames.RpcQueueTimeAvgTime':'Queue Average Wait Time',
    'services.service.info.metrics.mapreduce.tasksRunningWaiting':'Tasks (Running/Waiting)',
    'services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.runningMaps':'Running Map Tasks',
    'services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.runningReduces':'Running Reduce Tasks',
    'services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.waitingMaps':'Waiting Map Tasks',
    'services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.waitingReduces':'Waiting Reduce Tasks',

    'services.service.info.metrics.yarn.gc': 'Garbage Collection',
    'services.service.info.metrics.yarn.gc.displayNames.gcTimeMillis':'Time',
    'services.service.info.metrics.yarn.allocated.memory': 'Cluster Memory',
    'services.service.info.metrics.yarn.allocated.memory.displayNames.allocated': 'Allocated',
    'services.service.info.metrics.yarn.allocated.memory.displayNames.available': 'Available',
    'services.service.info.metrics.yarn.allocated.memory.displayNames.pending': 'Pending',
    'services.service.info.metrics.yarn.allocated.container': 'Containers',
    'services.service.info.metrics.yarn.allocated.container.displayNames.allocated': 'Allocated',
    'services.service.info.metrics.yarn.allocated.container.displayNames.reserved': 'Reserved',
    'services.service.info.metrics.yarn.allocated.container.displayNames.pending': 'Pending',
    'services.service.info.metrics.yarn.nodemanager.statuses':'NodeManagers',
    'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.active':'Active NodeManagers',
    'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.decommissioned':'Decommissioned NodeManagers',
    'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.lost':'Lost NodeManagers',
    'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.rebooted':'Rebooted NodeManagers',
    'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.unhealthy':'Unhealthy NodeManagers',
    'services.service.info.metrics.yarn.queueMemoryResource':'Queue Memory Used',
    'services.service.info.metrics.yarn.queueMemoryResource.displayNames.allocated':'Allocated ({0})',
    'services.service.info.metrics.yarn.queueMemoryResource.displayNames.available':'Available ({0})',
    'services.service.info.metrics.yarn.queueMemoryResource.displayName':'Queue: {0}',
    'services.service.info.metrics.yarn.apps.states.current.title': 'Current Applications',
    'services.service.info.metrics.yarn.apps.states.finished.title': 'Finished Applications',
    'services.service.info.metrics.yarn.apps.states.completed': 'Completed',
    'services.service.info.metrics.yarn.apps.states.failed': 'Failed',
    'services.service.info.metrics.yarn.apps.states.killed': 'Killed',
    'services.service.info.metrics.yarn.apps.states.pending': 'Pending',
    'services.service.info.metrics.yarn.apps.states.running': 'Running',
    'services.service.info.metrics.yarn.apps.states.submitted': 'Submitted',

    'services.service.info.menu.summary':'Summary',
    'services.service.info.menu.configs':'Configs',
    'services.service.info.summary.hostsRunningMonitor':'{0}/{1}',
    'services.service.info.summary.serversHostCount':'{0} more',
    'services.service.info.summary.nagiosWebUI':'Nagios Web UI',
    'services.service.info.summary.nagios.noAlerts':'No alerts',
    'services.service.info.summary.nagios.alerts':'Nagios service required for viewing alerts',

    'services.service.config.saved':'Saved Configuration Changes',
    'services.service.config.notSaved':'Unable to Save Configuration Changes',
    'services.service.config.restartService.TooltipMessage':'<b>Restart Service</b><br>Stale configuration used by {0} components on {1} hosts:{2}',
    'services.service.config.restartService.needToRestart':'<strong>Restart Required:</strong> ',
    'services.service.config.restartService.needToRestartEnd':'should be restarted',
    'service.service.config.restartService.shouldBeRestarted':'{0} Requiring Restart',
    'services.service.config.saved.message':'Service configuration changes saved successfully.',
    'services.service.config.msgServiceStop':'Could not save configuration changes.  Please stop the service first. You will be able to save configuration changes after all of its components are stopped.',
    'services.service.config.msgHDFSMapRServiceStop':'Could not save configuration changes.  Please stop both HDFS and MapReduce first.  You will be able to save configuration changes after all HDFS and MapReduce components are stopped.',
    'services.service.config.failCreateConfig' : 'Failure in creating service configuration',
    'services.service.config.failSaveConfig':'Failure in saving service configuration',
    'services.service.config.failSaveConfigHostOverrides':'Failure in saving service configuration overrides',
    'services.service.config.addPropertyWindow.errorMessage':'This is required',
    'services.service.config.addPropertyWindow.error.derivedKey':'This property is already defined',
    'services.service.config.addPropertyWindow.filterKeyLink' : 'Find property',
    'services.service.config.stopService.runningHostComponents':'{0} components on {1} hosts are running',
    'services.service.config.stopService.unknownHostComponents':'{0} components on {1} hosts are in unknown state.  These components might actually be running and need restarting later to make the changes effective.',
    'services.service.config.confirmDirectoryChange':'You are about to make changes to service directories that are core to {0}. Before you proceed, be absolutely certain of the implications and that you have taken necessary manual steps, if any, for the changes. Are you sure you want to proceed?',
    'services.service.config.configOverride.head':'Config Override',
    'services.service.config.configOverride.body':'Cannot override a config that has not been saved yet.',
    'services.service.config.exitPopup.body':'You have unsaved changes. Save changes or discard?',
    'services.service.config.propertyFilterPopover.title':'Properties filter',
    'services.service.config.propertyFilterPopover.content':'Enter keywords to filter properties by property name, value, or description.',

    'services.add.header':'Add Service Wizard',
    'services.reassign.header':'Move Master Wizard',
    'services.service.add':'Add Service',
    'services.service.startAll':'Start All',
    'services.service.stopAll':'Stop All',
    'services.service.stop.confirmMsg' : 'You are about to stop {0}',
    'services.service.stop.confirmButton': 'Confirm Stop',
    'services.service.stop.warningMsg.turnOnMM': 'This will generate alerts as the service is stopped. To suppress alerts, turn on Maintenance Mode for {0} prior to stopping',
    'services.service.restartAll.confirmButton': 'Confirm Restart All',
    'services.service.restartAll.confirmMsg': 'You are about to restart {0}',
    'services.service.restartAll.warningMsg.turnOnMM': 'This will trigger alerts as the service is restarted. To suppress alerts, turn on Maintenance Mode for {0} prior to running restart all',
    'services.service.config_groups_popup.header':'Manage {0} Configuration Groups',
    'services.service.config_groups_popup.notice':'You can apply different sets of {{serviceName}} configurations to groups of hosts by managing {{serviceName}} Configuration Groups and their host membership.  Hosts belonging to a {{serviceName}} Configuration Group have the same set of configurations for {{serviceName}}. Each host belongs to one {{serviceName}} Configuration Group.',
    'services.service.config_groups_popup.rename':'Rename',
    'services.service.config_groups_popup.duplicate':'Duplicate',
    'services.service.config_groups_popup.group_name_lable':'Name',
    'services.service.config_groups_popup.group_desc_lable':'Description',
    'services.service.config_groups_popup.properties':'Properties',
    'services.service.config_groups_popup.addButton':'Create new Configuration Group',
    'services.service.config_groups_popup.removeButton':'Delete Configuration Group',
    'services.service.config_groups_popup.renameButton':'Rename Configuration Group',
    'services.service.config_groups_popup.addHost':'Add hosts to selected Configuration Group',
    'services.service.config_groups_popup.addHostDisabled':'There are no available hosts to add.',
    'services.service.config_groups_popup.removeHost':'Remove hosts from selected Configuration Group',
    'services.service.config_groups_popup.duplicateButton':'Duplicate Configuration Group',
    'services.service.config_groups.add_config_group_popup.header':'Create New Configuration Group',
    'services.service.config_groups.duplicate_config_group_popup.header':'Duplicate Configuration Group',
    'services.service.config_groups.rename_config_group_popup.header':'Rename Configuration Group',
    'services.reassign.closePopup':'Move {0} wizard is in progress. You must allow the wizard to complete for Ambari to be in usable state. If you choose to quit, you must follow manual instructions to complete or revert move {0} wizard as documented in the Ambari User Guide. Are you sure you want to exit the wizard?',

    'services.reassign.step1.header':'Get Started',
    'services.reassign.step1.message1': 'This wizard will walk you through moving {0}.<br/>',
    'services.reassign.step1.message2': 'The process to reassign {0} involves a combination of <b>automated steps</b> (that will be handled by the wizard) and ' +
        '<b>manual steps</b> (that you must perform in sequence as instructed by the wizard).<br/><br/>',
    'services.reassign.step1.message3': '<br/><b>All required services will be restarted as part of the wizard. You should plan a cluster maintenance window and prepare ' +
      'for cluster downtime when moving {0}.</b>',

    'services.reassign.step2.header':'Assign Master',
    'services.reassign.step2.currentHost':'Current:',
    'services.reassign.step2.body':'Assign {0} to new host.',
    'services.reassign.step2.body.namenodeHA':'Move {0} to new host. You can move only one master component at a time.',
    'services.reassign.step3.header':'Review',
    'services.reassign.step3.body':'Please review the changes you made',
    'services.reassign.step3.targetHost':'Target Host:',
    'services.reassign.step3.sourceHost':'Source Host:',
    'services.reassign.step3.component':'Component name:',
    'services.reassign.step4.header':'Configure Component',
    'services.reassign.step4.task0.title':'Stop Required Services',
    'services.reassign.step4.task1.title':'Create {0}',
    'services.reassign.step4.task2.title':'Disable {0}',
    'services.reassign.step4.task3.title':'Reconfigure {0}',
    'services.reassign.step4.task4.title':'Install {0}',
    'services.reassign.step4.task5.title':'Start ZooKeeper Servers',
    'services.reassign.step4.task6.title':'Start NameNode',
    'services.reassign.step4.task7.title':'Delete disabled {0}',
    'services.reassign.step4.task8.title':'Start Required Services',
    'services.reassign.step4.status.success': 'Successfully moved <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host',
    'services.reassign.step4.status.success.withManualSteps': 'Proceed to the next step',
    'services.reassign.step4.status.failed': 'Failed to move <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host',
    'services.reassign.step4.status.info': 'Reassigning {0}. \nPlease wait for all tasks to be completed.',
    'services.reassign.step4.retry': 'You can click on the Retry or Abort button to retry failed task or abort changes',
    'services.reassign.step4.abortError': 'Error in aborting changes.',
    'services.reassign.step5.header': 'Manual commands',
    'services.reassign.step5.body.namenode':
        '<div class="alert alert-info">' +
        '<ol>' +
        '<li>Copy the contents of <b>{0}</b> on the source host <b>{1}</b> to <b>{0}</b> on the target host <b>{2}</b>.</li>' +
        '<li>Login to the target host <b>{2}</b> and change permissions for the NameNode dirs by running:' +
        '<div class="code-snippet">chown -R {3}:{5} /hadoop/hdfs/namenode/</div></li>' +
        '<li>Create marker directory by running:' +
        '<div class="code-snippet">mkdir -p /var/run/hadoop/hdfs/namenode/formatted</div></li>' +
        '</ol>' +
        '</div>',
    'services.reassign.step5.body.namenode_ha':
        '<div class="alert alert-info">' +
        '<ol>' +
        '<li>Login to the NameNode host <b>{4}</b>.</li>' +
        '<li>Reset automatic failover information in ZooKeeper by running:' +
        '<div class="code-snippet">sudo su -l {3} -c \'hdfs zkfc -formatZK\'</div></li>' +
        '</ol>' +
        '</div>' +
        '<div class="alert alert-info">' +
        '<ol start="3">' +
        '<li>Login to the newly installed NameNode host <b>{2}</b>.<br>' +
        '<div class="alert alert-warn"><strong>Important!</strong> Be sure to login to the newly installed NameNode host.<br>This is a different host from the Steps 1 and 2 above.</div>' +
        '</li>' +
        '<li>Initialize the metadata by running:' +
        "<div class='code-snippet'>sudo su -l {3} -c 'hdfs namenode -bootstrapStandby'</div></li>" +
        '</ol>' +
        '</div>',
    'services.reassign.step5.body.secondary_namenode':
        '<div class="alert alert-info">' +
        '<ol>' +
        '<li>Copy the contents of <b>{0}</b> on the source host <b>{1}</b> to <b>{0}</b> on the target host <b>{2}</b>.</li>' +
        '<li>Login to the target host <b>{2}</b> and change permissions for the SNameNode dirs by running:' +
        '<div class="code-snippet">chown -R {3}:{5} /hadoop/hdfs/namesecondary/</div></li>' +
        '</ol>' +
        '</div>',
    'services.reassign.step5.body.securityNotice': '<div class="alert alert-info"> <div class="alert alert-warn"> <strong>Note: </strong> Secure cluster' +
      ' requires generating necessary principals for reassigned component and creating keytab files with the principal on ' +
      'the target host. The keytab file should be accessible to the service user.</div> {0} </div>',
    'services.reassign.step5.body.securityConfigsList': 'Create keytab file <b>{0}</b> with principal <b>{1}</b> on <b>{2}</b> host.',
    'services.reassign.step5.body.proceedMsg': 'Please proceed once you have completed the steps above',
    'services.reassign.step5.confirmPopup.body': 'Please confirm that you have run the manual steps before continuing.',
    'services.reassign.step6.header': 'Start and Test services',
    'services.reassign.step6.task0.title': 'Delete disabled {0}',
    'services.reassign.step6.task1.title': 'Start All Services',
    'services.reassign.step6.status.success': 'Successfully moved <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host.',
    'services.reassign.step6.status.failed': 'Failed to move <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host.',
    'services.reassign.step6.status.info': 'Reassigning {0}. \nPlease wait for all tasks to be completed.',

    /** services page constants **/

    'service.hbase.activeMaster': 'Active Master',

    'services.hive.client': 'Hive Client',
    'services.hive.clients': 'Hive Clients',

    'services.falcon.client': 'Falcon Client',
    'services.falcon.clients': 'Falcon Clients',

    'services.oozie.client': 'Oozie Client',
    'services.oozie.clients': 'Oozie Clients',
    'services.oozie.webUi': 'Oozie Web UI',

    'services.hue.webUi': 'Hue Web UI',

    'services.ganglia.webUi': 'Ganglia Web UI',
    'services.ganglia.monitors': 'Ganglia Monitors',

    'services.mapreduce2.webUi': 'History Server UI',

    'services.zookeeper.prefix': '{0} of',
    'services.zookeeper.title': '{0} ZooKeepers',
    'services.zookeeper.postfix': 'running',
    'services.zookeeper.clients': 'ZooKeeper Clients',
    'services.zookeeper.client': 'ZooKeeper Client',

    'services.mapreduce2.history.running': 'History server is running',
    'services.mapreduce2.history.stopped': 'History server is stopped',
    'services.mapreduce2.history.unknown': 'History server status is unknown',
    'services.mapreduce2.smokeTest.requirement': 'MapReduce2 smoke test requires YARN service be started',

    'services.mapReduce.config.addQueue':'Add Queue',
    'services.mapReduce.config.editQueue':'Edit Queue',
    'services.mapReduce.config.capacitySchedulerXml':'Custom capacity-scheduler.xml',
    'services.mapReduce.config.queue.header':'Queues',
    'services.mapReduce.config.queue.name':'Queue Name',
    'services.mapReduce.config.queue.groups':'Groups',
    'services.mapReduce.config.queue.capacity':'Capacity',
    'services.mapReduce.config.queue.maxCapacity':'Max Capacity',
    'services.mapReduce.config.queue.minUserLimit':'Min User Limit',
    'services.mapReduce.config.queue.userLimitFactor':'User Limit Factor',
    'services.mapReduce.config.queue.supportsPriority': 'Supports priority',
    'services.mapReduce.config.queue.adminUsers':'Admin Users',
    'services.mapReduce.config.queue.adminGroups':'Admin Groups',
    'services.mapReduce.config.queue.maxActiveTasks':'Max active initialized tasks',
    'services.mapReduce.config.queue.maxActiveTasksPerUser':'Max active initialized tasks per user',
    'services.mapReduce.config.queue.initAcceptJobsFactor':'Init accept jobs factor',
    'services.mapReduce.extraConfig.queue.name':'Queue name',
    'services.mapReduce.description.queue.name':'Name of the queue',
    'services.mapReduce.description.queue.submit.user':"Comma separated list of usernames that are allowed to submit jobs to the queue. " +
      "If set to the special value '*', it means all users are allowed to submit jobs.",
    'services.mapReduce.description.queue.admin.user':"Comma separated list of usernames that are allowed to delete jobs or modify job's priority for " +
      "jobs not owned by the current user in the queue.  If set to the special value '*', it means all users are " +
      "allowed to do this operation.",
    'services.mapReduce.description.queue.submit.group':'Comma separated list of group names that are allowed to submit jobs to the queue.',
    'services.mapReduce.description.queue.admin.group':"Comma separated list of group names that are allowed to delete jobs or modify job's priority " +
      "for jobs not owned by the current user in the queue.",

    'services.tez.client': 'Tez client',
    'services.tez.clients': 'Tez clients',
    'services.pig.client': 'Pig client',
    'services.pig.clients': 'Pig clients',
    'services.sqoop.client': 'Sqoop client',
    'services.sqoop.clients': 'Sqoop clients',

    'services.hbase.master.error':'None of the HBase masters is active',

    'hosts.host.add':'Add New Hosts',
    'hosts.table.noHosts':'No hosts to display',

    'hosts.filters.filteredHostsInfo': '{0} of {1} hosts showing',

    'hosts.filters.selectedHostInfo': 'host selected',
    'hosts.filters.selectedHostsInfo': 'hosts selected',

    'hosts.filters.clearSelection': 'clear selection',

    'hosts.table.restartComponents.withNames':'Restart {0}',
    'hosts.table.restartComponents.withoutNames':'{0} components should be restarted',

    'hosts.table.componentsInPassiveState.withNames':'{0} in Maintenance Mode',
    'hosts.table.componentsInPassiveState.withoutNames':'{0} components in Maintenance Mode',

    'hosts.table.menu.l1.selectedHosts':'Selected Hosts',
    'hosts.table.menu.l1.filteredHosts':'Filtered Hosts',
    'hosts.table.menu.l1.allHosts':'All Hosts',
    'hosts.table.menu.l2.allComponents':'All Components',
    'hosts.table.menu.l2.restartAllComponents':'Restart All Components',

    'hosts.bulkOperation.confirmation.header':'Confirm Bulk Operation',
    'hosts.bulkOperation.confirmation.hosts':'Are you sure you want to <strong>{0}</strong> on the following {1} hosts?',
    'hosts.bulkOperation.confirmation.hostComponents':'Are you sure you want to <strong>{0} {1}</strong> on the following {2} hosts?',
    'hosts.bulkOperation.passiveState.nothingToDo.body':'All hosts that you selected are already in Maintenance Mode.',
    'hosts.bulkOperation.warningInfo.body':'Components on these hosts are stopped so decommission will be skipped.',
    'hosts.bulkOperation.host_components.passiveState.nothingToDo.body':'All host components that you selected are already in Maintenance Mode',

    'hosts.selectHostsDialog.title': 'Select Configuration Group Hosts',
    'hosts.selectHostsDialog.message': 'Select hosts that should belong to this {0} Configuration Group.  All hosts belonging to this group will have the same set of {0} configurations.',
    'hosts.selectHostsDialog.filter.placeHolder': 'Filter...',
    'hosts.selectHostsDialog.selectedHostsLink': '{0} out of {1} hosts selected',
    'hosts.selectHostsDialog.message.warning': 'Atleast one host needs to be selected.',

    'hosts.host.serviceNotAvailable': 'Service not available on this host',

    'hosts.host.metrics.dataUnavailable':'Data Unavailable',
    'hosts.host.metrics.cpu':'CPU Usage',
    'hosts.host.metrics.cpu.displayNames.cpu_wio':'CPU I/O Idle',
    'hosts.host.metrics.cpu.displayNames.cpu_idle':'CPU Idle',
    'hosts.host.metrics.cpu.displayNames.cpu_nice':'CPU Nice',
    'hosts.host.metrics.cpu.displayNames.cpu_aidle':'CPU Boot Idle',
    'hosts.host.metrics.cpu.displayNames.cpu_system':'CPU System',
    'hosts.host.metrics.cpu.displayNames.cpu_user':'CPU User',
    'hosts.host.metrics.disk':'Disk Usage',
    'hosts.host.metrics.disk.displayNames.disk_total':'Total',
    'hosts.host.metrics.disk.displayNames.disk_free':'Available',
    'hosts.host.metrics.load':'Load',
    'hosts.host.metrics.load.displayNames.load_fifteen':'15 Minute Load',
    'hosts.host.metrics.load.displayNames.load_one':'1 Minute Load',
    'hosts.host.metrics.load.displayNames.load_five':'5 Minute Load',
    'hosts.host.metrics.memory':'Memory Usage',
    'hosts.host.metrics.memory.displayNames.mem_shared':'Shared',
    'hosts.host.metrics.memory.displayNames.swap_free':'Swap',
    'hosts.host.metrics.memory.displayNames.mem_buffers':'Buffers',
    'hosts.host.metrics.memory.displayNames.mem_free':'Free',
    'hosts.host.metrics.memory.displayNames.mem_cached':'Cached',
    'hosts.host.metrics.network':'Network Usage',
    'hosts.host.metrics.network.displayNames.pkts_out':'Packets Out',
    'hosts.host.metrics.network.displayNames.bytes_in':'Bytes In',
    'hosts.host.metrics.network.displayNames.bytes_out':'Bytes Out',
    'hosts.host.metrics.network.displayNames.pkts_in':'Packets In',
    'hosts.host.metrics.processes':'Processes',
    'hosts.host.metrics.processes.displayNames.proc_total':'Total Processes',
    'hosts.host.metrics.processes.displayNames.proc_run':'Processes Run',

    'hosts.host.summary.header':'Summary',
    'hosts.host.summary.hostname':'Hostname',
    'hosts.host.summary.agentHeartbeat':'Heartbeat',
    'hosts.host.summary.hostMetrics':'Host Metrics',
    'hosts.host.summary.addComponent':'Add Component',

    'hosts.host.details.hostActions':'Host Actions',
    'hosts.host.details.needToRestart':'Host needs {0} {1} restarted',
    'hosts.host.details.needToRestart.button':'Restart',
    'hosts.host.details.deleteHost':'Delete Host',
    'hosts.host.details.startAllComponents':'Start All Components',
    'hosts.host.details.stopAllComponents':'Stop All Components',
    'hosts.host.details.restartAllComponents':'Restart All Components',
    'hosts.host.details.refreshConfigs':'Refresh configs',

    'host.host.componentFilter.master':'Master Components',
    'host.host.componentFilter.slave':'Slave Components',
    'host.host.componentFilter.client':'Client Components',
    'hosts.host.deleteComponent.popup.msg1':'Are you sure you want to delete {0}?',
    'hosts.host.deleteComponent.popup.msg2':'<b>Important:</b> After this <i>{0}</i> is deleted, go to <i>Services -> Nagios</i> to restart the Nagios service.  This is required for the alerts and notifications to work properly.',
    'hosts.host.deleteComponent.popup.warning':'<b>WARNING!</b> Delete the last <i>{0}</i> component in the cluster?</br>Deleting the last component in the cluster could result in permanent loss of service data.',
    'hosts.host.deleteComponent.popup.confirm':'Confirm Delete',
    'hosts.host.installComponent.popup.confirm':'Confirm Install',
    'hosts.host.installComponent.msg':'Are you sure you want to install {0}?',
    'hosts.host.addComponent.msg':'Are you sure you want to add {0}?',
    'hosts.host.addComponent.addZooKeeper':'Adding ZooKeeper Server may reconfigure such properties:<ul><li>ha.zookeeper.quorum</li><li>hbase.zookeeper.quorum</li><li>templeton.zookeeper.hosts</li></ul>',
    'hosts.host.addComponent.deleteHostWithZooKeeper':'Deleting host with ZooKeeper Server may reconfigure such properties:<ul><li>ha.zookeeper.quorum</li><li>hbase.zookeeper.quorum</li><li>templeton.zookeeper.hosts</li></ul>',
    'hosts.host.addComponent.note':'<b>Important:</b> After this <i>{0}</i> is installed, go to <i>Services -> Nagios</i> to restart the Nagios service.  This is required for the alerts and notifications to work properly.',
    'hosts.host.addComponent.securityNote':'You are running your cluster in secure mode. You must set up the keytab for {0} on {1} before you proceed. Otherwise, the component will not be able to start properly.',
    'hosts.host.addComponent.popup.confirm':'Confirm Add',
    'hosts.host.datanode.decommission':'Decommission DataNode',
    'hosts.host.datanode.recommission':'Recommission DataNode',
    'hosts.host.nodemanager.decommission':'Decommission NodeManager',
    'hosts.host.nodemanager.recommission':'Recommission NodeManager',
    'hosts.host.tasktracker.decommission':'Decommission TaskTracker',
    'hosts.host.tasktracker.recommission':'Recommission TaskTracker',
    'hosts.host.tasktracker.restart':'Restart TaskTracker',
    'hosts.host.regionserver.decommission.batch1':'Decommission RegionServer - Turn drain mode on',
    'hosts.host.regionserver.decommission.batch2':'Decommission RegionServer - Stop RegionServer',
    'hosts.host.regionserver.decommission.batch3':'Decommission RegionServer - Turn drain mode off',
    'hosts.host.hbase_regionserver.recommission':'Recommission HBase RegionServer',
    'hosts.host.hbase_regionserver.decommission':'Decommission HBase RegionServer',
    'hosts.host.decommissioned':'Decommissioned',
    'hosts.host.decommissioning':'Decommissioning',

    'hosts.component.passive.implied.host.mode.tooltip':'Cannot Turn Off Maintenance Mode because Host is in Maintenance Mode',
    'hosts.component.passive.implied.service.mode.tooltip':'Cannot Turn Off Maintenance Mode because {0} is in Maintenance Mode',
    'hosts.component.passive.mode':'Component is in Maintenance Mode',
    'hosts.component.passive.short.mode':'Maintenance Mode',
    'hosts.host.passive.mode':'Host is in Maintenance Mode',
    'hosts.host.alert.noAlerts':'No alerts',
    'hosts.host.alert.noAlerts.message':'There are no alerts for this host.',
    'hosts.host.healthStatus.heartBeatNotReceived':'The server has not received a heartbeat from this host for more than 3 minutes.',
    'hosts.host.healthStatus.mastersDown':"The following master components are down:\n",
    'hosts.host.healthStatus.slavesDown':"The following slave components are down:\n",
    'hosts.host.healthStatus.allUp':'All components are up',
    'hosts.host.healthStatusCategory.green': "Healthy",
    'hosts.host.healthStatusCategory.red': "Master Down",
    'hosts.host.healthStatusCategory.orange': "Slave Down",
    'hosts.host.healthStatusCategory.yellow': "Lost Heartbeat",
    'hosts.host.alerts.label': 'Alerts',
    'hosts.host.maintainance.allComponents.context': 'All Host Components',
    'hosts.host.maintainance.stopAllComponents.context': 'Stop All Host Components',
    'hosts.host.maintainance.startAllComponents.context': 'Start All Host Components',
    'hosts.host.alerts.st':'&nbsp;!&nbsp;',
    'hosts.decommission.popup.body':'Are you sure?',
    'hosts.decommission.popup.header':'Confirmation',
    'hosts.decommission.tooltip.warning':'Cannot {0} since {1} is not running',
    'hosts.delete.popup.body':'Are you sure you want to delete host <i>{0}</i>?',
    'hosts.delete.popup.body.msg1':'By removing this host, Ambari will ignore future communications from this host. Software packages will not be removed from the host. The components on the host should not be restarted. If you wish to readd this host to the cluster, be sure to clean the host.',
    'hosts.delete.popup.body.msg2':'After deleting this host, Nagios should be restarted to remove this host from Nagios monitoring. Go to the <i>Services</i> page to restart Nagios.',
    'hosts.delete.popup.body.msg3':'If this host was hosting a Zookeeper Server, the Zookeeper Service should be restarted. Go to the <i>Services</i> page.',
    'hosts.delete.popup.body.msg4':'<b>WARNING!</b> Delete the last <i>{0}</i> component[s] in the cluster?</br>Deleting the last components in the cluster could result in permanent loss of service data.',
    'hosts.delete.popup.body.msg.unknownComponents':'This host does not appear to be online and Ambari communication with the Agent has been lost.',
    'hosts.delete.popup.header':'Confirmation',
    'hosts.delete.popup.title':'Delete Host',
    'hosts.delete.popup.unknownComponents':'The following components have unknown status:',
    'hosts.cant.do.popup.title':'Unable to Delete Host',
    'hosts.cant.do.popup.masterList.body':'This host cannot be deleted since it has the following master components:',
    'hosts.cant.do.popup.masterList.body.end':'To delete this host, you must first move all the master components listed above to another host.',
    'hosts.cant.do.popup.nonDeletableList.body':'Deletion of the following {0} components is not supported. ',
    'hosts.cant.do.popup.runningList.body':'This host cannot be deleted since the following components are running:',
    'hosts.cant.do.popup.runningList.body.end':'To delete this host, you must first stop all the running components listed above. ' +
      'If this host has a DataNode, it should be decommissioned first to prevent data loss.',
    'hosts.add.header':'Add Host Wizard',
    'hosts.add.exit.header':'Exit',
    'hosts.add.exit.body':'Do you really want to exit Add Host Wizard?',
    'hosts.assignRack':'Assign Rack',
    'hosts.passiveMode.popup':'Are you sure you want to <b>Turn {0} Maintenance Mode</b> for {1}?',

    'charts.horizon.chart.showText':'show',
    'charts.horizon.chart.hideText':'hide',
    'charts.horizon.chart.attributes.cpu':'CPU',
    'charts.horizon.chart.attributes.memory':'Memory',
    'charts.horizon.chart.attributes.network':'Network',
    'charts.horizon.chart.attributes.io':'I/O',

    'charts.heatmap.selectMetric':'Select Metric...',

    'charts.heatmap.category.host':'Host',
    'charts.heatmap.item.host.memory':'Memory Used',
    'charts.heatmap.item.host.disk':'Disk Space Used',
    'charts.heatmap.item.host.process':'Total Running Processes',
    'charts.heatmap.category.hdfs':'HDFS',
    'charts.heatmap.category.mapreduce': 'MapReduce',
    'charts.heatmap.category.yarn': 'YARN',
    'charts.heatmap.category.hbase': 'HBase',
    'charts.heatmap.unknown': 'Unknown',
    'charts.heatmap.label.notApplicable' :'Not Applicable',
    'charts.heatmap.label.invalidData' :'Invalid data',
    'charts.heatmap.metrics.bytesRead' :'HDFS Bytes Read',
    'charts.heatmap.metrics.bytesWritten' :'HDFS Bytes Written',
    'charts.heatmap.metrics.DFSGarbageCollection' :'HDFS Garbage Collection Time',
    'charts.heatmap.metrics.DFSMemHeapUsed' :'HDFS JVM Heap Memory Used',
    'charts.heatmap.metrics.diskSpaceUsed' :'Host Disk Space Used %',
    'charts.heatmap.metrics.MapReduceGCTime' :'MapReduce Garbage Collection Time',
    'charts.heatmap.metrics.YarnGCTime' :'YARN Garbage Collection Time',
    'charts.heatmap.metrics.mapsRunning' :'MapReduce Maps Running',
    'charts.heatmap.metrics.MRMemHeapUsed' :'MapReduce JVM Heap Memory Used',
    'charts.heatmap.metrics.YarnMemHeapUsed' :'YARN JVM Heap Memory Used',
    'charts.heatmap.metrics.reducesRunning' :'MapReduce Reduces Running',

    'charts.heatmap.metrics.memoryUsed' :'Host Memory Used %',
    'charts.heatmap.metrics.processRun' :'Total Running Processes',
    'charts.heatmap.metrics.YarnMemoryUsed' :'YARN Memory used %',
    'charts.heatmap.metrics.cpuWaitIO':'Host CPU Wait I/O %',
    'charts.heatmap.metrics.HbaseRegionServerReadCount': 'HBase Read Request Count',
    'charts.heatmap.metrics.HbaseRegionServerWriteCount': 'HBase Write Request Count',
    'charts.heatmap.metrics.HbaseRegionServerCompactionQueueSize': 'HBase Compaction Queue Size',
    'charts.heatmap.metrics.HbaseRegionServerRegions': 'HBase Regions',
    'charts.heatmap.metrics.HbaseRegionServerMemStoreSize': 'HBase Memstore Sizes',
    'metric.notFound':'no items found',
    'metric.default':'combined',
    'metric.cpu':'cpu',
    'metric.memory':'disk used',
    'metric.network':'network',
    'metric.io':'io',
    'metric.more':'more',
    'metric.more.cpu':'CPU',
    'metric.more.disk':'Disk',
    'metric.more.load':'Load',
    'metric.more.memory':'Memory',
    'metric.more.network':'Network',
    'metric.more.process':'Process',

    'dashboard.clusterMetrics':'Cluster Metrics',

    'dashboard.clusterMetrics.cpu':'CPU Usage',
    'dashboard.clusterMetrics.cpu.displayNames.idle':'Idle',
    'dashboard.clusterMetrics.load':'Cluster Load',
    'dashboard.clusterMetrics.memory':'Memory Usage',
    'dashboard.clusterMetrics.network':'Network Usage',

    'dashboard.widgets': 'Cluster Status and Metrics',
    'dashboard.button.switch': 'Switch to classic dashboard',
    'dashboard.button.switchShort': 'Switch',
    'dashboard.button.reset': 'Reset all widgets to default ',
    'dashboard.button.gangliaLink': 'View metrics in Ganglia',

    'dashboard.widgets.NameNodeHeap': 'NameNode Heap',
    'dashboard.widgets.NameNodeCpu': 'NameNode CPU WIO',
    'dashboard.widgets.HDFSDiskUsage': 'HDFS Disk Usage',
    'dashboard.widgets.HDFSDiskUsage.DFSused': 'DFS used',
    'dashboard.widgets.HDFSDiskUsage.nonDFSused': 'non DFS used',
    'dashboard.widgets.HDFSDiskUsage.remaining': 'remaining',
    'dashboard.widgets.HDFSDiskUsage.info': '{0} ({1}%)',
    'dashboard.widgets.JobTrackerHeap': 'JobTracker Heap',
    'dashboard.widgets.JobTrackerCpu': 'JobTracker CPU WIO',
    'dashboard.widgets.JobTrackerCapacity': 'JobTracker Capacity',
    'dashboard.widgets.DataNodeUp': 'DataNodes Live',
    'dashboard.widgets.TaskTrackerUp': 'TaskTrackers Live',
    'dashboard.widgets.SuperVisorUp': 'Supervisors Live',
    'dashboard.widgets.NameNodeRpc': 'NameNode RPC',
    'dashboard.widgets.JobTrackerRpc': 'JobTracker RPC',
    'dashboard.widgets.MapReduceSlots': 'MapReduce Slots',
    'dashboard.widgets.mapSlots': 'Map Slots',
    'dashboard.widgets.reduceSlots': 'Reduce Slots',
    'dashboard.widgets.nothing': 'No Widget to Add',
    'dashboard.widgets.NameNodeUptime': 'NameNode Uptime',
    'dashboard.widgets.JobTrackerUptime': 'JobTracker Uptime',
    'dashboard.widgets.HDFSLinks': 'HDFS Links',
    'dashboard.widgets.HDFSLinks.activeNameNode': 'Active NameNode',
    'dashboard.widgets.HDFSLinks.standbyNameNode': 'Standby NameNode',
    'dashboard.widgets.HDFSLinks.standbyNameNodes': '2 Standby NameNodes',
    'dashboard.widgets.MapReduceLinks': 'MapReduce Links',
    'dashboard.widgets.HBaseLinks': 'HBase Links',
    'dashboard.widgets.HBaseAverageLoad': 'HBase Ave Load',
    'dashboard.widgets.HBaseMasterHeap': 'HBase Master Heap',
    'dashboard.widgets.HBaseRegionsInTransition': 'Region In Transition',
    'dashboard.widgets.HBaseMasterUptime': 'HBase Master Uptime',
    'dashboard.widgets.ResourceManagerHeap': 'ResourceManager Heap',
    'dashboard.widgets.ResourceManagerUptime': 'ResourceManager Uptime',
    'dashboard.widgets.NodeManagersLive': 'NodeManagers Live',
    'dashboard.widgets.YARNMemory': 'YARN Memory',

    'dashboard': {
      'widgets': {
        'popupHeader': 'Customize Widget',
        'hintInfo': {
          'common': 'Edit the percentage thresholds to change the color of current pie chart. <br />Enter two numbers between 0 to {0}',
          'hint1': 'Edit the percentage of thresholds to change the color of current widget. <br />Assume all components UP is 100, and all DOWN is 0. <br /> So enter two numbers between 0 to {0}',
          'hint2': 'Edit the thresholds to change the color of current widget.<br /><br />So enter two numbers larger than 0.',
          'hint3': 'Edit the thresholds to change the color of current widget.<br />The unit is milli-second. <br />So enter two numbers larger than 0. '
        }
      }
    },

    'dashboard.services':'Services',
    'dashboard.services.hosts':'Hosts',
    'dashboard.services.uptime':'{0}',
    'dashboard.services.hdfs.summary':'{0} of {1} nodes live, {2}% capacity used',
    'dashboard.services.hdfs.nanmenode':'NameNode',
    'dashboard.services.hdfs.snanmenode':'Secondary NameNode',
    'dashboard.services.hdfs.journalnodes':'JournalNodes',
    'dashboard.services.hdfs.capacity':'HDFS Disk Usage',
    'dashboard.services.hdfs.capacity.dfsUsed':'Disk Usage (DFS Used)',
    'dashboard.services.hdfs.capacity.nonDfsUsed':'Disk Usage (Non DFS Used)',
    'dashboard.services.hdfs.capacity.remaining':'Disk Usage (Remaining)',
    'dashboard.services.hdfs.capacityUsed':'{0} / {1} ({2}%)',
    'dashboard.services.hdfs.totalFilesAndDirs':'Total Files + Directories',
    'dashboard.services.hdfs.datanodes':'DataNodes',
    'dashboard.services.hdfs.datanodecounts':'DataNodes Status',
    'dashboard.services.hdfs.version':'Version',
    'dashboard.services.hdfs.nameNodeWebUI':'NameNode Web UI',
    'dashboard.services.hdfs.nodes.live':'live',
    'dashboard.services.hdfs.nodes.dead':'dead',
    'dashboard.services.hdfs.nodes.decom':'decommissioning',
    'dashboard.services.hdfs.nodes.uptime':'NameNode Uptime',
    'dashboard.services.hdfs.nodes.heap':'NameNode Heap',
    'dashboard.services.hdfs.nodes.heapUsed':'{0} / {1} ({2}% used)',
    'dashboard.services.hdfs.chart.label':'Capacity (Used/Total)',
    'dashboard.services.hdfs.blockErrors':'{0} corrupt / {1} missing / {2} under replicated',

    'dashboard.services.yarn.summary':'{0} of {1} nodes live',
    'dashboard.services.yarn.resourceManager':'ResourceManager',
    'dashboard.services.yarn.nodeManagers':'NodeManagers',
    'dashboard.services.yarn.nodeManager':'NodeManager',
    'dashboard.services.yarn.clients':'YARN Clients',
    'dashboard.services.yarn.client':'YARN Client',
    'dashboard.services.yarn.resourceManager.uptime':'ResourceManager Uptime',
    'dashboard.services.resourceManager.nodes.heap':'ResourceManager Heap',
    'dashboard.services.yarn.nodeManagers.status': 'NodeManagers Status',
    'dashboard.services.yarn.nodeManagers.status.msg': '{0} active / {1} lost / {2} unhealthy / {3} rebooted / {4} decommissioned',
    'dashboard.services.yarn.containers': 'Containers',
    'dashboard.services.yarn.containers.msg': '{0} allocated / {1} pending / {2} reserved',
    'dashboard.services.yarn.apps': 'Applications',
    'dashboard.services.yarn.apps.msg': '{0} submitted / {1} running / {2} pending / {3} completed / {4} killed / {5} failed',
    'dashboard.services.yarn.memory': 'Cluster Memory',
    'dashboard.services.yarn.memory.msg': '{0} used / {1} reserved / {2} available',
    'dashboard.services.yarn.queues': 'Queues',
    'dashboard.services.yarn.queues.msg': '{0} Queues',

    'dashboard.services.mapreduce.summary':'{0} of {1} trackers live, {2} jobs running',
    'dashboard.services.mapreduce.taskTrackers':'TaskTrackers',
    'dashboard.services.mapreduce.taskTrackerCounts':'TaskTrackers Status',
    'dashboard.services.mapreduce.trackers':'Trackers',
    'dashboard.services.mapreduce.nodes.blacklist':'blacklist',
    'dashboard.services.mapreduce.nodes.graylist':'graylist',
    'dashboard.services.mapreduce.slotCapacity':'Total Slots Capacity',
    'dashboard.services.mapreduce.trackersSummary':'{0}/{1}',
    'dashboard.services.mapreduce.jobs':'Total Jobs',
    'dashboard.services.mapreduce.jobsSummary':'{0} submitted / {1} completed',
    'dashboard.services.mapreduce.mapSlots':'Map Slots',
    'dashboard.services.mapreduce.mapSlotsSummary':'{0} occupied / {1} reserved',
    'dashboard.services.mapreduce.reduceSlots':'Reduce Slots',
    'dashboard.services.mapreduce.tasks.maps':'Tasks: Maps',
    'dashboard.services.mapreduce.tasks.reduces':'Tasks: Reduces',
    'dashboard.services.mapreduce.reduceSlotsSummary':'{0} occupied / {1} reserved',
    'dashboard.services.mapreduce.tasksSummary':'{0} running / {1} waiting',
    'dashboard.services.mapreduce.slotCapacitySummary':'{0} maps / {1} reduces / {2} avg per node',
    'dashboard.services.mapreduce.jobTrackerHeap':'JobTracker Heap',
    'dashboard.services.mapreduce.jobTrackerHeapSummary':'{0} of {1} ({2}% used)',
    'dashboard.services.mapreduce.jobTrackerUptime':'Job Trackers Uptime',
    'dashboard.services.mapreduce.chart.label':'Jobs Running',

    'dashboard.services.flume.summary.single':'{0} Flume Agent',
    'dashboard.services.flume.summary.multiple':'{0} Flume Agents',
    'dashboard.services.flume.agentsLabel': 'Flume Agents',
    'dashboard.services.flume.channels': 'Channels',
    'dashboard.services.flume.sources': 'Sources',
    'dashboard.services.flume.sinks': 'Sinks',


    'dashboard.services.hbase.summary':'{0} region servers with {1} average load',
    'dashboard.services.hbase.masterServer':'HBase Master',
    'dashboard.services.hbase.masterServer.active':'Active HBase Master',
    'dashboard.services.hbase.masterServer.standby':'Standby HBase Master',
    'dashboard.services.hbase.noMasterServer':'No Active Master',
    'dashboard.services.hbase.masterServerHeap':'Master Heap',
    'dashboard.services.hbase.masterServerHeap.summary':'{0} / {1} ({2}% used)',
    'dashboard.services.hbase.masterServerUptime':'Master Server Uptime',
    'dashboard.services.hbase.averageLoad':'Average Load',
    'dashboard.services.hbase.averageLoadPerServer':'{0} regions per RegionServer',
    'dashboard.services.hbase.regionServers':'RegionServers',
    'dashboard.services.hbase.regionServersSummary':'{0} live / {1} total',
    'dashboard.services.hbase.chart.label':'Request Count',
    'dashboard.services.hbase.version':'Version',
    'dashboard.services.hbase.masterWebUI':'Master Web UI',
    'dashboard.services.hbase.regions.transition':'Regions In Transition',
    'dashboard.services.hbase.masterStarted':'Master Started',
    'dashboard.services.hbase.masterActivated':'Master Activated',

    'dashboard.services.hive.clients':'Hive Clients',
    'dashboard.services.hive.client':'Hive Client',

    'dashboard.services.oozie.clients':'Oozie Clients',
    'dashboard.services.oozie.client':'Oozie Client',
    'dashboard.services.storm.supervisor': 'Supervisor',
    'dashboard.services.storm.supervisors': 'Supervisors',

    'dashboard.services.configs.popup.stopService.header':'Stop service',
    'dashboard.services.configs.popup.stopService.body' : 'Service needs to be stopped for reconfiguration',
    'dashboard.services.configs.popup.restartService.header' : 'Restart service',
    'dashboard.services.configs.popup.restartService.body' : 'Service needs to be restarted for reconfiguration',
    'timeRange.presets.1hour':'1h',
    'timeRange.presets.12hour':'12h',
    'timeRange.presets.1day':'1d',
    'timeRange.presets.1week':'1wk',
    'timeRange.presets.1month':'1mo',
    'timeRange.presets.1year':'1yr',

    'apps.item.dag.job': 'Job',
    'apps.item.dag.jobId': 'Job Id',
    'apps.item.dag.type': 'Job Type',
    'apps.item.dag.status': 'Status',
    'apps.item.dag.num_stages': 'Total Stages',
    'apps.item.dag.stages': 'Tasks per Stage',
    'apps.item.dag.maps': 'Maps',
    'apps.item.dag.reduces': 'Reduces',
    'apps.item.dag.input': 'Input',
    'apps.item.dag.output': 'Output',
    'apps.item.dag.duration': 'Duration',

    'tableView.filters.all': 'All',
    'tableView.filters.filtered': 'Filtered',
    'tableView.filters.clearFilters': 'Clear filters',
    'tableView.filters.paginationInfo': '{0} - {1} of {2}',
    'tableView.filters.clearAllFilters': 'clear filters',
    'tableView.filters.showAll': 'Show All',

    'jobs.type':'Jobs Type',
    'jobs.type.hive':'Hive',
    'jobs.show.up.to':'Show up to',
    'jobs.filtered.jobs':'{0} jobs showing',
    'jobs.filtered.clear':'clear filters',
    'jobs.column.id':'Id',
    'jobs.column.user':'User',
    'jobs.column.start.time':'Start Time',
    'jobs.column.end.time':'End Time',
    'jobs.column.duration':'Duration',
    'jobs.new_jobs.info':'New jobs available on server.',

    'apps.table.column.appId':'App ID',
    'apps.table.column.runDate': 'Run Date',
    'apps.avgTable.avg': 'Avg',
    'apps.avgTable.min': 'Min',
    'apps.avgTable.max': 'Max',
    'apps.avgTable.jobs': 'Jobs',
    'apps.avgTable.input': 'Input',
    'apps.avgTable.output': 'Output',
    'apps.avgTable.duration': 'Duration',
    'apps.avgTable.oldest': 'Oldest',
    'apps.avgTable.mostRecent': 'Most Recent',
    'apps.filters.customRunDate':'Run Date custom filter',
    'apps.filters.nothingToShow': 'No jobs to display',
    'apps.filters.filterComponents': 'Filter by <strong>Component</strong>',
    'apps.dagCharts.popup':'Job Charts',
    'apps.dagCharts.popup.job': 'Job',
    'apps.dagCharts.popup.dag':'Job Timeline',
    'apps.dagCharts.popup.tasks':'Job Tasks',
    'apps.isRunning.popup.title':'Is running',
    'apps.isRunning.popup.content':'Job is running now',

    'mirroring.dataset.dataSets':'Datasets',
    'mirroring.dataset.createDataset':'Create Dataset',
    'mirroring.dataset.editDataset':'Edit Dataset',
    'mirroring.dataset.manageClusters':'Manage Clusters',
    'mirroring.dataset.newDataset':'New Dataset',
    'mirroring.dataset.selectTargetClusters':'Select Target Cluster...',
    'mirroring.dataset.name':'Name',
    'mirroring.dataset.type':'Type',
    'mirroring.dataset.save': 'Save & Schedule',
    'mirroring.dataset.sourceDir':'Source Directory',
    'mirroring.dataset.sourceCluster':'Source Cluster',
    'mirroring.dataset.target':'Target',
    'mirroring.dataset.source':'Source',
    'mirroring.dataset.targetCluster':'Target Cluster',
    'mirroring.dataset.targetDir':'Target Directory',
    'mirroring.dataset.schedule':'Schedule',
    'mirroring.dataset.resume':'Resume',
    'mirroring.dataset.suspend':'Suspend',
    'mirroring.dataset.suspendInstance':'Suspend Instance',
    'mirroring.dataset.resumeInstance':'Resume Instance',
    'mirroring.dataset.killInstance':'Kill Instance',
    'mirroring.dataset.schedule.to':'to',
    'mirroring.dataset.schedule.repeatEvery':'Repeat every ',
    'mirroring.dataset.addTargetCluster':'Add Target Cluster',
    'mirroring.dataset.type.HDFS':'HDFS',
    'mirroring.dataset.repeat.minutes':'minutes',
    'mirroring.dataset.repeat.hours':'hours',
    'mirroring.dataset.repeat.days':'days',
    'mirroring.dataset.repeat.months':'months',
    'mirroring.dataset.middayPeriod.am':'AM',
    'mirroring.dataset.middayPeriod.pm':'PM',
    'mirroring.dataset.entity':'entity',
    'mirroring.dataset.services.not.started':'Both Falcon and Oozie should be started for correct work',
    'mirroring.dataset.dataset.loading.error':'Unable to load datasets',
    'mirroring.dataset.dataset.loading.instances':'Unable to load instances',

    'mirroring.manageClusters.ambariServer':'Ambari Server',
    'mirroring.manageClusters.interfaces':'Interfaces',
    'mirroring.manageClusters.locations':'Locations',
    'mirroring.manageClusters.specifyName':'Specify name for new target cluster:',
    'mirroring.manageClusters.execute':'Execute',
    'mirroring.manageClusters.readonly':'Readonly',
    'mirroring.manageClusters.workflow':'Workflow',
    'mirroring.manageClusters.write':'Write',
    'mirroring.manageClusters.staging':'Staging',
    'mirroring.manageClusters.working':'Working',
    'mirroring.manageClusters.temp':'Temp',
    'mirroring.manageClusters.error' :'Error in saving changes',
    'mirroring.manageClusters.create.cluster.popup' :'Create Target Cluster',
    'mirroring.manageClusters.remove.confirmation' :'Are you sure you want to delete the target cluster {0}?',
    'mirroring.manageClusters.executeTooltip' :'The "execute" interface specifies the endpoint for ResourceManager.<br>Example: resourcemanager-host:8050',
    'mirroring.manageClusters.readonlyTooltip' :'The "readonly" interface specifies the endpoint for Hadoop\'s HFTP protocol.<br>Example: hftp://namenode-host:50070',
    'mirroring.manageClusters.workflowTooltip' :'The "workflow" interface specifies the endpoint for Oozie URL.<br>Example: http://oozie-host:11000/oozie',
    'mirroring.manageClusters.writeTooltip' :'The "write" interface is the endpoint to write to HDFS. Set this to the value of fs.defaultFS (in core-site.xml) of the target cluster.<br>Example:<br>hdfs://namenode-host:8020 (non-HA); hdfs://nameservice-id:8020 (HA)',
    'mirroring.manageClusters.locationsMessage' :'Specify the following for the target cluster, not the source cluster.',

    'mirroring.table.noDatasets':'No datasets to display',
    'mirroring.table.datasetStatus':'Status',
    'mirroring.table.noJobs':'No instances to display',
    'mirroring.table.jobId':'Instance ID',
    'mirroring.table.start':'Start',
    'mirroring.table.end':'End',
    'mirroring.table.status':'Status',

    'mirroring.required.error': 'This field is required',
    'mirroring.dateOrder.error': 'End Date must be after Start Date',
    'mirroring.startDate.error': 'Start Date in the past is not allowed',
    'mirroring.required.invalidNumberError' : 'Enter valid number',

    'rollingrestart.dialog.title': 'Restart {0}s',
    'rollingrestart.dialog.primary': 'Trigger Rolling Restart',
    'rollingrestart.notsupported.hostComponent': 'Rolling restart not supported for {0} components',
    'rollingrestart.dialog.msg.restart': 'This will restart a specified number of {0}s at a time.',
    'rollingrestart.dialog.msg.noRestartHosts': 'There are no {0}s to do rolling restarts',
    'rollingrestart.dialog.msg.maintainance': 'Note: {0} {1} in Maintenance Mode will not be restarted',
    'rollingrestart.dialog.msg.maintainance.plural': 'Note: {0} {1}s in Maintenance Mode will not be restarted',
    'rollingrestart.dialog.msg.componentsAtATime': '{0}s at a time',
    'rollingrestart.dialog.msg.timegap.prefix': 'Wait ',
    'rollingrestart.dialog.msg.timegap.suffix': 'seconds between batches ',
    'rollingrestart.dialog.msg.toleration.prefix': 'Tolerate up to ',
    'rollingrestart.dialog.msg.toleration.suffix': 'restart failures',
    'rollingrestart.dialog.err.invalid.batchsize': 'Invalid restart batch size: {0}',
    'rollingrestart.dialog.err.invalid.waitTime': 'Invalid wait time between batches: {0}',
    'rollingrestart.dialog.err.invalid.toleratesize': 'Invalid failure toleration count: {0}',
    'rollingrestart.dialog.msg.serviceNotInMM':'Note: This will trigger alerts. To suppress alerts, turn on Maintenance Mode for {0} prior to triggering a rolling restart',
    'rollingrestart.dialog.msg.staleConfigsOnly': 'Only restart {0}s with stale configs',
    'rollingrestart.rest.context': 'Rolling Restart of {0}s - batch {1} of {2}',
    'rollingrestart.context.allOnSelectedHosts':'Restart all components on the selected hosts',
    'rollingrestart.context.allForSelectedService':'Restart all components for {0}',
    'rollingrestart.context.allWithStaleConfigsForSelectedService':'Restart all components with Stale Configs for {0}',
    'rollingrestart.context.allClientsOnSelectedHost':'Restart all clients on {0}',
    'rollingrestart.context.allWithStaleConfigsOnSelectedHost':'Restart components with Stale Configs on {0}',
    'rollingrestart.context.allOnSelectedHost':'Restart all components on {0}',
    'rollingrestart.context.selectedComponentOnSelectedHost':'Restart {0}',
    'rollingrestart.context.default':'Restart components',

    'rolling.command.context': 'Rolling set {0} to state "{1}" - batch {2} of {3}',
    'rolling.nothingToDo.header': 'Nothing to do',
    'rolling.nothingToDo.body': '{0} on selected hosts are already in selected state or in Maintenance Mode.',

    'restart.service.all': 'Restart All',
    'restart.service.rest.context': 'Restart {0}s',

    'menu.item.dashboard':'Dashboard',
    'menu.item.heatmaps':'Heatmaps',
    'menu.item.services':'Services',
    'menu.item.hosts':'Hosts',
    'menu.item.mirroring':'Mirroring',
    'menu.item.jobs':'Jobs',
    'menu.item.admin':'Admin',
    'menu.item.views':'Views',

    'jobs.nothingToShow': 'No jobs to display',
    'jobs.loadingTasks': 'Loading...',
    'jobs.table.custom.date.am':'AM',
    'jobs.table.custom.date.pm':'PM',
    'jobs.table.custom.date.header':'Select Custom Dates',
    'jobs.table.job.fail':'Job failed to run',
    'jobs.customDateFilter.error.required':'This field is required',
    'jobs.customDateFilter.error.date.order':'End Date must be after Start Date',
    'jobs.customDateFilter.startTime':'Start Time',
    'jobs.customDateFilter.endTime':'End Time',
    'jobs.hive.more':'show more',
    'jobs.hive.less':'show less',
    'jobs.hive.query':'Hive Query',
    'jobs.hive.stages':'Stages',
    'jobs.hive.yarnApplication':'YARN&nbsp;Application',
    'jobs.hive.tez.tasks':'Tez Tasks',
    'jobs.hive.tez.hdfs':'HDFS',
    'jobs.hive.tez.localFiles':'Local Files',
    'jobs.hive.tez.spilledRecords':'Spilled Records',
    'jobs.hive.tez.records':'Records',
    'jobs.hive.tez.reads':'{0} reads',
    'jobs.hive.tez.writes':'{0} writes',
    'jobs.hive.tez.records.count':'{0} Records',
    'jobs.hive.tez.operatorPlan':'Operator Plan',
    'jobs.hive.tez.dag.summary.metric':'Summary Metric',
    'jobs.hive.tez.dag.error.noDag.title':'No Tez Information',
    'jobs.hive.tez.dag.error.noDag.message':'This job does not identify any Tez information.',
    'jobs.hive.tez.dag.error.noDagId.title':'No Tez Information',
    'jobs.hive.tez.dag.error.noDagId.message':'No Tez information was found for this job. Either it is waiting to be run, or has exited unexpectedly.',
    'jobs.hive.tez.dag.error.noDagForId.title':'No Tez Information',
    'jobs.hive.tez.dag.error.noDagForId.message':'No details were found for the Tez ID given to this job.',
    'jobs.hive.tez.metric.input':'Input',
    'jobs.hive.tez.metric.output':'Output',
    'jobs.hive.tez.metric.recordsRead':'Records Read',
    'jobs.hive.tez.metric.recordsWrite':'Records Written',
    'jobs.hive.tez.metric.tezTasks':'Tez Tasks',
    'jobs.hive.tez.metric.spilledRecords':'Spilled Records',
    'jobs.hive.tez.edge.':'Unknown',
    'jobs.hive.tez.edge.contains':'Contains',
    'jobs.hive.tez.edge.broadcast':'Broadcast',
    'jobs.hive.tez.edge.scatter_gather':'Shuffle',

    'number.validate.empty': 'cannot be empty',
    'number.validate.notValidNumber': 'not a valid number',
    'number.validate.lessThanMinumum': 'value less than {0}',
    'number.validate.moreThanMaximum': 'value greater than {0}',

    'common.combobox.placeholder': 'Filter...',
    'common.combobox.dropdown.1': 'Overridden properties',
    //'common.combobox.dropdown.2': 'Modified properties',
    //'common.combobox.dropdown.3': 'Properties needing restart',

    'quick.links.error.label': 'Hostname is undefined',

    'contact.administrator': 'Contact System Administrator for more information!',

    'config.group.selection.dialog.title': '{0} Configuration Group',
    'config.group.selection.dialog.subtitle': 'Select or create a {0} Configuration Group where the configuration value will be overridden.',
    'config.group.selection.dialog.option.select': 'Select an existing {0} Configuration Group',
    'config.group.selection.dialog.option.select.msg': 'Overridden property will be changed for hosts belonging to the selected group.',
    'config.group.selection.dialog.option.create': 'Create a new {0} Configuration Group',
    'config.group.selection.dialog.option.create.msg': 'A new {0} Configuration Group will be created with the given name. Initially there will be no hosts in the group, with only the selected property overridden.',
    'config.group.selection.dialog.err.name.exists': 'Configuration Group with given name already exists',
    'config.group.selection.dialog.err.create': 'Error creating new Configuration Group [{0}]',
    'config.group.selection.dialog.no.groups': 'There are no existing {0} Configuration Groups.',
    'config.group.host.switch.dialog.title': 'Change Group',

    'config.group.save.confirmation.header': 'Save Configuration Group',
    'config.group.save.confirmation.msg': 'Click <em>Manage Hosts</em> to manage host membership to the configuration group',
    'config.group.save.confirmation.configGroup': 'Configuration Group',
    'config.group.save.confirmation.saved': 'has been successfully saved',
    'config.group.save.confirmation.manage.button': 'Manage Hosts',
    'config.group.description.default': 'New configuration group created on {0}',

    'utils.ajax.errorMessage': 'Error message'
  };
  
});
window.require.register("mixins", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  // load all mixins here

  require('mixins/common/userPref');
  require('mixins/main/host/details/host_components/decommissionable');
});
window.require.register("mixins/common/chart/storm_linear_time", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.StormLinearTimeChartMixin = Em.Mixin.create({
    ajaxIndex: 'service.metrics.storm.nimbus',
    metricsTemplate: 'metrics/storm/nimbus/{0}[{1},{2},{3}]',

    getDataForAjaxRequest: function() {
      var currentTime = Math.round(App.dateTime() / 1000);
      var metricTemplate = [];
      this.get('stormChartDefinition').forEach(function(chartInfo) {
        metricTemplate.push(
          this.get('metricsTemplate').format(chartInfo.field, currentTime - this.get('timeUnitSeconds'), currentTime, 15)
        );
      }, this);
      return {
        metricsTemplate: metricTemplate.join(',')
      };
    },

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var pathKeys = ['metrics','storm','nimbus'];
      var validPath = true;
      pathKeys.forEach(function(key) {
        if (!jsonData[key]) {
          validPath = false;
        } else {
          jsonData = jsonData[key];
        }
      });
      if (!validPath) {
        return seriesArray;
      }
      this.get('stormChartDefinition').forEach(function(chart){
        seriesArray.push(this.transformData(jsonData[chart.field], chart.name));
      }, this);
      return seriesArray;
    }
  });
  
});
window.require.register("mixins/common/userPref", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * Small mixin for processing user preferences
   * Provide methods to save/load some values in <code>persist</code> storage
   * When using this mixin you should redeclare methods:
   * <ul>
   *   <li>getUserPrefSuccessCallback</li>
   *   <li>getUserPrefErrorCallback</li>
   * </ul>
   * @type {Em.Mixin}
   */
  App.UserPref = Em.Mixin.create({

    /**
     * Should <code>getUserPref</code> and <code>postUserPref</code> be async
     * @type {bool}
     */
    makeRequestAsync: true,

    /**
     * Get persist value from server with persistKey
     * @param {String} key
     */
    getUserPref: function(key) {
      return App.ajax.send({
        name: 'settings.get.user_pref',
        sender: this,
        data: {
          key: key,
          async: this.get('makeRequestAsync')
        },
        success: 'getUserPrefSuccessCallback',
        error: 'getUserPrefErrorCallback'
      });
    },

    /**
     * Should be redeclared in objects that use this mixin
     * @param {*} response
     * @param {Object} request
     * @param {Object} data
     * @returns {*}
     */
    getUserPrefSuccessCallback: function (response, request, data) {},

    /**
     * Should be redeclared in objects that use this mixin
     * @param {Object} request
     * @param {Object} ajaxOptions
     * @param {String} error
     */
    getUserPrefErrorCallback: function (request, ajaxOptions, error) {},

    /**
     * Post persist key/value to server, value is object
     * @param {String} key
     * @param {Object} value
     */
    postUserPref: function (key, value) {
      var keyValuePair = {};
      keyValuePair[key] = JSON.stringify(value);
      App.ajax.send({
        'name': 'settings.post.user_pref',
        'sender': this,
        'beforeSend': 'postUserPrefBeforeSend',
        'data': {
          'async': this.get('makeRequestAsync'),
          'keyValuePair': keyValuePair
        }
      });
    },

    /**
     * Little log before post request
     * @param {Object} request
     * @param {Object} ajaxOptions
     * @param {Object} data
     */
    postUserPrefBeforeSend: function(request, ajaxOptions, data){
      console.log('BeforeSend to persist: persistKeyValues', data.keyValuePair);
    }

  });
});
window.require.register("mixins/main/host/details/host_components/decommissionable", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var uiEffects = require('utils/ui_effects');

  /**
   * Mixin for <code>App.HostComponentView</code>
   * Contains code for processing components with allowed decommission
   * @type {Em.Mixin}
   */
  App.Decommissionable = Em.Mixin.create({

    /**
     * Should be redeclared in views that use this mixin
     * @type {String}
     */
    componentForCheckDecommission: '',

    /**
     * Received from server object with data about decommission
     * @type {Object}
     */
    decommissionedStatusObject: null,

    /**
     * Received from server desired_admin_state value
     * @type {String}
     */
    desiredAdminState: null,

    /**
     * Is component in decommission process right know
     * @type {bool}
     */
    isComponentDecommissioning: null,

    /**
     * May conponent be decommissioned
     * @type {bool}
     */
    isComponentDecommissionAvailable: null,

    /**
     * May component be recommissioned
     * @type {bool}
     */
    isComponentRecommissionAvailable: null,

    /**
     * Component with stopped masters can't be docommissioned
     * @type {bool}
     */
    isComponentDecommissionDisable: function() {
      return this.get('content.service.workStatus') != App.HostComponentStatus.started;
    }.property('content.service.workStatus'),

    /**
     * Tooltip message shows if decommission/recommission is disabled
     * when masters for current component is down
     */
    decommissionTooltipMessage: function() {
      if (this.get('isComponentDecommissionDisable') && (this.get('isComponentRecommissionAvailable') || this.get('isComponentDecommissionAvailable'))) {
        var decom = this.get('isComponentRecommissionAvailable') ? Em.I18n.t('common.recommission') : Em.I18n.t('common.decommission');
        return Em.I18n.t('hosts.decommission.tooltip.warning').format(decom, App.format.role(this.get('componentForCheckDecommission')));
      }
    }.property('isComponentDecommissionDisable', 'isComponentRecommissionAvailable', 'isComponentDecommissionAvailable', 'componentForCheckDecommission'),
    /**
     * Recalculated component status based on decommission
     * @type {string}
     */
    statusClass: function () {

      //Class when install failed
      if (this.get('workStatus') === App.HostComponentStatus.install_failed) {
        return 'health-status-color-red icon-cog';
      }

      //Class when installing
      if (this.get('workStatus') === App.HostComponentStatus.installing) {
        return 'health-status-color-blue icon-cog';
      }

      if (this.get('isComponentRecommissionAvailable') && (this.get('isStart') || this.get('workStatus') == 'INSTALLED')) {
        return 'health-status-DEAD-ORANGE';
      }

      //For all other cases
      return 'health-status-' + App.HostComponentStatus.getKeyName(this.get('workStatus'));

    }.property('workStatus', 'isComponentRecommissionAvailable', 'isComponentDecommissioning'),

    /**
     * Return host component text status
     * @type {String}
     */
    componentTextStatus: function () {
      var componentTextStatus = this.get('content.componentTextStatus');
      var hostComponent = this.get('hostComponent');
      if (hostComponent) {
        componentTextStatus = hostComponent.get('componentTextStatus');
        if(this.get('isComponentRecommissionAvailable')){
          if(this.get('isComponentDecommissioning')){
            componentTextStatus = Em.I18n.t('hosts.host.decommissioning');
          } else {
            componentTextStatus = Em.I18n.t('hosts.host.decommissioned');
          }
        }
      }
      return componentTextStatus;
    }.property('workStatus','isComponentRecommissionAvailable','isComponentDecommissioning'),

    /**
     * For Stopping or Starting states, also for decommissioning
     * @type {bool}
     */
    isInProgress: function () {
      return (this.get('workStatus') === App.HostComponentStatus.stopping ||
        this.get('workStatus') === App.HostComponentStatus.starting) ||
        this.get('isDecommissioning');
    }.property('workStatus', 'isDecommissioning'),

    /**
     * Get desired_admin_state status from server
     */
    getDesiredAdminState: function(){
      return App.ajax.send({
        name: 'host.host_component.slave_desired_admin_state',
        sender: this,
        data: {
          hostName: this.get('content.host.hostName'),
          componentName: this.get('content.componentName')
        },
        success: 'getDesiredAdminStateSuccessCallback',
        error: 'getDesiredAdminStateErrorCallback'
      });
    },

    /**
     * Set received value or null to <code>desiredAdminState</code>
     * @param {Object} response
     * @returns {String|null}
     */
    getDesiredAdminStateSuccessCallback: function (response) {
      var status = response.HostRoles.desired_admin_state;
      if ( status != null) {
        this.set('desiredAdminState', status);
        return status;
      }
      return null;
    },

    /**
     * Set null to <code>desiredAdminState</code> if server returns error
     * @returns {null}
     */
    getDesiredAdminStateErrorCallback: function () {
      this.set('desiredAdminState', null);
      return null;
    },

    /**
     * Get component decommission status from server
     * @returns {$.ajax}
     */
    getDecommissionStatus: function() {
      return App.ajax.send({
        name: 'host.host_component.decommission_status',
        sender: this,
        data: {
          hostName: this.get('content.host.hostName'),
          componentName: this.get('componentForCheckDecommission'),
          serviceName: this.get('content.service.serviceName')
        },
        success: 'getDecommissionStatusSuccessCallback',
        error: 'getDecommissionStatusErrorCallback'
      });
    },

    /**
     * Set received value or null to <code>decommissionedStatusObject</code>
     * @param {Object} response
     * @returns {Object|null}
     */
    getDecommissionStatusSuccessCallback: function (response) {
      var statusObject = response.ServiceComponentInfo;
      if ( statusObject != null) {
        this.set('decommissionedStatusObject', statusObject);
        return statusObject;
      }
      return null;
    },

    /**
     * Set null to <code>decommissionedStatusObject</code> if server returns error
     * @returns {null}
     */
    getDecommissionStatusErrorCallback: function () {
      this.set('decommissionedStatusObject', null);
      return null;
    },

    /**
     * Do blinking for 1 minute
     */
    doBlinking: function () {
      var workStatus = this.get('workStatus');
      var self = this;
      var pulsate = [App.HostComponentStatus.starting, App.HostComponentStatus.stopping, App.HostComponentStatus.installing].contains(workStatus);
      if (!pulsate) {
        var component = this.get('content');
        if (component && workStatus != "INSTALLED") {
          pulsate = this.get('isDecommissioning');
        }
      }
      if (pulsate && !self.get('isBlinking')) {
        self.set('isBlinking', true);
        uiEffects.pulsate(self.$('.components-health'), 1000, function () {
          self.set('isBlinking', false);
          self.doBlinking();
        });
      }
    },

    /**
     * Start blinking when host component is starting/stopping/decommissioning
     */
    startBlinking: function () {
      this.$('.components-health').stop(true, true);
      this.$('.components-health').css({opacity: 1.0});
      this.doBlinking();
    }.observes('workStatus','isComponentRecommissionAvailable', 'isDecommissioning'),

    /**
     * Should be redeclared in views that use this mixin
     */
    loadComponentDecommissionStatus: function() {},

    didInsertElement: function() {
      this._super();
      this.loadComponentDecommissionStatus();
    },

    /**
     * Update Decommission status only one time when component was changed
     */
    updateDecommissionStatus: function() {
      Em.run.once(this, 'loadComponentDecommissionStatus');
    }.observes('content.workStatus', 'content.passiveState'),


    decommissionView: Em.View.extend({

      templateName: require('templates/main/host/decommission'),

      text: function() {
        return this.get('parentView.isComponentDecommissionAvailable') ? Em.I18n.t('common.decommission') : Em.I18n.t('common.recommission');
      }.property('parentView.isComponentDecommissionAvailable'),

      didInsertElement: function() {
        this._super();
        App.tooltip($("[rel='decommissionTooltip']"));
      },

      click: function() {
        if (!this.get('parentView.isComponentDecommissionDisable')) {
          if (this.get('parentView.isComponentDecommissionAvailable')) {
            this.get('controller').decommission(this.get('parentView.content'));
          } else {
            this.get('controller').recommission(this.get('parentView.content'));
          }
        }
      }
    })
  });
});
window.require.register("models", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  // load all models here

  require('models/form'); // should be the 1st
  require('models/authentication');
  require('models/cluster');
  require('models/cluster_states');
  require('models/hosts');
  require('models/quick_links');
  require('models/service');
  require('models/service_config');
  require('models/service_audit');
  require('models/service/hdfs');
  require('models/service/yarn');
  require('models/service/mapreduce');
  require('models/service/mapreduce2');
  require('models/service/hbase');
  require('models/service/flume');
  require('models/service/storm');
  require('models/alert');
  require('models/user');
  require('models/host');
  require('models/rack');
  require('models/job');
  require('models/run');
  require('models/app');
  require('models/background_operation');
  require('models/host_component');
  require('models/target_cluster');
  require('models/dataset');
  require('models/dataset_job');
  require('classes/run_class');
  require('classes/job_class');
  require('models/config_group');
  require('models/jobs/tez_dag');
  require('models/jobs/job');
  require('models/jobs/hive_job');
  
});
window.require.register("models/alert", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * Defines structure for App.Alert class. Keys mentioned here are for JSON data
   * which comes back from NAGIOS server.
   */
  App.Alert = Em.Object.extend({
    title: null,//service_description in ajax response
    serviceType: null,
    status: null,//current_state in ajax response
    message: null,//plugin_output in ajax response
    hostName: null,
    lastCheck: null,
    lastTime: null,

    date: function () {
      return DS.attr.transforms.date.from(this.get('lastTime'));
    }.property('lastTime'),

    /**
     * Used to show correct icon in UI
     */
    isOk: function () {
      return this.get('status') == "0";
    }.property('status'),

    /**
     * Used to show correct icon in UI
     */
    isWarning: function () {
      return this.get('status') == "1";
    }.property('status'),

    /**
     * Used to show correct icon in UI
     */
    isCritical: function() {
      return this.get('status') == '2';
    }.property('status'),

    /**
     * Used to show correct icon in UI
     */
    isPassive: function() {
      return this.get('status') == '3';
    }.property('status'),

    /**
     * Used to show only required alerts at the service level
     */
    ignoredForServices: function() {
      return ['NodeManager health', 'NodeManager process', 'TaskTracker process', 'RegionServer process', 'DataNode process', 'DataNode space', 'ZooKeeper Server process', 'Supervisors process'].contains(this.get('title'));
    }.property('title'),

    /**
     * Used to show only required alerts at the host level
     */
    ignoredForHosts: function() {
      return this.get('title').indexOf('Percent') != -1;
    }.property('title'),

    /**
     * Provides how long ago this alert happened.
     * 
     * @type {String}
     */
    timeSinceAlert: function () {
      var d = this.get('date');
      if (d) {
        var timeFormat = this.t('services.alerts.OK.timePrefix');
        switch (this.get('status')) {
          case "1":
            timeFormat = this.t('services.alerts.WARN.timePrefix');
            break;
          case "2":
            timeFormat = this.t('services.alerts.CRIT.timePrefix');
            break;
          case "3":
            timeFormat = this.t('services.alerts.MAINT.timePrefix');
            break;
          case "4":
            timeFormat = this.t('services.alerts.UNKNOWN.timePrefix');
            break;
        }
        var prevSuffix = $.timeago.settings.strings.suffixAgo;
        $.timeago.settings.strings.suffixAgo = '';
        var since = timeFormat.format($.timeago(this.makeTimeAtleastMinuteAgo(d)));
        $.timeago.settings.strings.suffixAgo = prevSuffix;
        return since;
      } else if (d == 0) {
        var timeFormat = this.t('services.alerts.OK.timePrefixShort');
        switch (this.get('status')) {
          case "1":
            timeFormat = this.t('services.alerts.WARN.timePrefixShort');
            break;
          case "2":
            timeFormat = this.t('services.alerts.CRIT.timePrefixShort');
            break;
          case "3":
            timeFormat = this.t('services.alerts.MAINT.timePrefixShort');
            break;
          case "4":
            timeFormat = this.t('services.alerts.UNKNOWN.timePrefixShort');
            break;
        }
        return timeFormat;
      } else {
        return "";
      }
    }.property('date', 'status'),
    
    makeTimeAtleastMinuteAgo: function(d){
      var diff = App.dateTime() - d.getTime();
      if (diff < 60000) {
        diff = 60000 - diff;
        return new Date(d.getTime() - diff );
      }
      return d;
    },

    /**
     * Provides more details about when this alert happened.
     * 
     * @type {String}
     */
    timeSinceAlertDetails: function () {
      var details = "";
      var date = this.get('date');
      if (date) {
        var dateString = date.toDateString();
        dateString = dateString.substr(dateString.indexOf(" ") + 1);
        dateString = Em.I18n.t('services.alerts.occurredOn').format(dateString, date.toLocaleTimeString());
        details += dateString;
      }
      var lastCheck = this.get('lastCheck');
      if (lastCheck) {
        lastCheck = new Date(lastCheck * 1000);
        details = details ? details + Em.I18n.t('services.alerts.brLastCheck').format($.timeago(lastCheck)) : Em.I18n.t('services.alerts.lastCheck').format($.timeago(lastCheck));
      }
      return details;
    }.property('lastCheck', 'date'),

    /**
     * Used to show appropriate service label in UI
     */
    serviceName: function () {
      if (this.get('serviceType')) {
        var type = this.get('serviceType').toLowerCase();
        switch (type) {
          case 'mapreduce':
            return 'MapReduce';
          case 'hdfs':
            return 'HDFS';
          case 'hbase':
            return "HBase";
          case 'zookeeper':
            return "Zookeeper";
          case 'oozie':
            return "Oozie";
          case 'hive':
            return 'Hive';
        }
      }
      return null;
    }.property('serviceType'),

    /**
     * Used to provide appropriate service link in UI
     */
    serviceLink: function () {
      if (this.get('serviceType')) {
        var type = this.get('serviceType').toLowerCase();
        switch (type) {
          case 'mapreduce':
            return '#/main/services/MAPREDUCE/summary';
          case 'hdfs':
            return '#/main/services/HDFS/summary';
          case 'hbase':
            return '#/main/services/HBASE/summary';
          case 'zookeeper':
            return '#/main/services/ZOOKEEPER/summary';
          case 'oozie':
            return '#/main/services/OOZIE/summary';
          case 'hive':
            return '#/main/services/HIVE/summary';
        }
      }
      return null;
    }.property('serviceType')
  });
  
});
window.require.register("models/app", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.App = DS.Model.extend({
    appName: DS.attr('string'),
    type: DS.attr('string'),
    numJobsTotal: DS.attr('number'),
    userName: DS.attr('string'),
    executionTime: DS.attr('string'),
    runs: DS.hasMany('App.Run')
  });

  App.App.FIXTURES = [
    {
      id: 1,
      app_name: 'pigs.sh',
      type: 'Hive',
      num_jobs_total: 5,
      user_name: 'root',
      execution_time: '1347629541543',
      runs: [1, 2, 3]
    },
    {
      id: 2,
      app_name: 'pigsm.sh',
      type: 'pig',
      num_jobs_total: 3,
      user_name: 'user1',
      execution_time: '1347656741515',
      runs: [6, 4, 5]
    },
    {
      id: 3,
      app_name: 'pigsmo.sh',
      type: 'pig',
      num_jobs_total: 4,
      user_name: 'user3',
      execution_time: '1347629587687',
      runs: [7, 8, 9, 10, 11]
    },
    {
      id: 4,
      app_name: 'pigsmok.sh',
      type: 'MapReduce',
      num_jobs_total: 0,
      user_name: 'root',
      execution_time: '134762957834',
      runs: []
    }
  ];
  
});
window.require.register("models/authentication", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  App.Authentication = DS.Model.extend({
    method:DS.attr('boolean'), // use LDAP
    primaryServer:DS.attr('string'),
    secondaryServer:DS.attr('string'),
    useSsl:DS.attr('boolean'),
    bindMethod:DS.attr('boolean'), // use credentials
    bindUser:DS.attr('string'),
    password:DS.attr('string'),
    passwordRetype:DS.attr('string'),
    searchBaseDn:DS.attr('string'),
    usernameAttribute:DS.attr('string')
  });

  App.Authentication.FIXTURES = [
    {
      id:1,
      method:0,
      primary_server:"",
      secondary_server:"",
      use_ssl:false,
      bind_method:0,
      bind_user:"",
      password:"",
      password_retype:"",
      search_base_dn:"",
      username_attribute:""
    }
  ];

  App.AuthenticationForm = App.Form.extend({
    testResult:false,
    fieldsOptions:[
      { name:"method", displayName:"", isRequired:false, displayType:"select",
        values:[
          {value:0, label:Em.I18n.t("admin.authentication.form.method.database")},
          {value:1, label:Em.I18n.t("admin.authentication.form.method.ldap")}
        ]
      },
      { name:"primaryServer", displayName:Em.I18n.t("admin.authentication.form.primaryServer"), /*validator:'ipaddress',*/
        isRequired:function () {
          return this.get('form.field.method.value');
        }.property('form.field.method.value')
      },
      { name:"secondaryServer", displayName:Em.I18n.t("admin.authentication.form.secondaryServer"), /*validator:'ipaddress',*/ isRequired:false},
      { name:"useSsl", displayName:Em.I18n.t("admin.authentication.form.useSsl"), displayType:"checkbox", isRequired:false },
      { name:"bindMethod", displayName:'', displayType:"select", isRequired:false,
        values:[
          {value:0, label:Em.I18n.t("admin.authentication.form.bind.anonymously")},
          {value:1, label:Em.I18n.t("admin.authentication.form.bind.useCrenedtials")}
        ]},
      { name:"bindUser", displayName:Em.I18n.t('admin.authentication.form.bindUserDN'), isRequired:function () {
        return this.get('form.field.bindMethod.value');
      }.property('form.field.bindMethod.value')},
      { name:"password", displayName:Em.I18n.t('common.password'), displayType:"password",
        isRequired:function () {
          return this.get('form.field.bindMethod.value');
        }.property('form.field.bindMethod.value') },
      { name:"passwordRetype", displayName:Em.I18n.t('form.passwordRetype'), displayType:"password",
        validator: "passwordRetype",
        isRequired:function () {
          return this.get('form.field.bindMethod.value');
        }.property('form.field.bindMethod.value')},
      { name:"searchBaseDn", displayName:Em.I18n.t('admin.authentication.form.searchBaseDN'),
        isRequired:function () {
          return this.get('form.field.method.value');
        }.property('form.field.method.value')
      },
      { name:"usernameAttribute", displayName:Em.I18n.t('admin.authentication.form.usernameAttribute'),
        isRequired:function () {
          return this.get('form.field.method.value');
        }.property('form.field.method.value')
      },

      { name:"userDN", displayName:Em.I18n.t('admin.authentication.form.userDN') },
      { name:"userPassword", displayName:Em.I18n.t('common.password'), displayType:'password'}
    ],
    fields:[],
    testConfiguration:function () {
      console.warn('Configuration test is randomized');
      this.set('testResult', parseInt(Math.random() * 2));
      return true;
    },
    testConfigurationMessage:function () {
      return this.get('testResult') ? Em.I18n.t('admin.authentication.form.test.success') : Em.I18n.t('admin.authentication.form.test.fail');
    }.property('testResult'),
    testConfigurationClass:function () {
      return this.get('testResult') ? "text-success" : "text-error";
    }.property('testConfigurationMessage')
  })
  ;
});
window.require.register("models/background_operation", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.BackgroundOperation = DS.Model.extend({
    operationName:DS.attr('string'),
    events: DS.hasMany('App.BackgroundOperationEvent'),
    cluster:DS.belongsTo('App.Cluster'),
    host:DS.belongsTo('App.Host'),
    operationLog: DS.attr('string')
  });

  App.BackgroundOperation.FIXTURES = [
    {
      id:1,
      operation_name:'Decommissioning host1',
      operation_log:'Decommissioning log',
      events:[1,2],
      cluster_id:1,
      host_id:1
    },
    {
      id:2,
      operation_name:'Starting DataNode on host4',
      operation_log:'Starting DataNode log',
      events:[3],
      cluster_id:1,
      host_id:1
    }
  ];

  App.BackgroundOperationEvent = DS.Model.extend({
    eventName:DS.attr('string'),
    operation:DS.belongsTo('App.BackgroundOperation'),
    eventDate: DS.attr('string')
  });

  App.BackgroundOperationEvent.FIXTURES = [
    {
      id:1,
      event_name:'Some intermediate operation',
      operation_id:1,
      event_date:'4 min ago'
    },
    {
      id:2,
      event_name:'Operation started',
      operation_id:1,
      event_date:'5 min ago'
    },
    {
      id:3,
      event_name:'Operation started',
      operation_id:2,
      event_date:'5 min ago'
    }
  ];

  
});
window.require.register("models/cluster", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.Cluster = DS.Model.extend({
    clusterName: DS.attr('string'),
    stackName: DS.attr('string'),
    version: DS.attr('string'),
    hosts: DS.hasMany('App.Host'),
    racks: DS.hasMany('App.Rack'),
    maxHostsPerRack: DS.attr('number'),
    /**
     * Array containing desired configs. New array
     * should be set by instances of class.
     */
    desiredConfigs: null
  });

  App.Cluster.FIXTURES = [/*
    {
      id: 1,
      cluster_name: 'cluster1',
      stack_name: 'HDP',
      hosts: [1, 2, 3, 4],
      racks: [1, 2, 3, 4, 5, 6],
      max_hosts_per_rack: 10
    }*/
  ];
});
window.require.register("models/cluster_states", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');

  App.clusterStatus = Ember.Object.create({
    clusterName: '',
    validStates: [
      'DEFAULT',
      'CLUSTER_NOT_CREATED_1',
      'CLUSTER_DEPLOY_PREP_2',
      'CLUSTER_INSTALLING_3',
      'SERVICE_STARTING_3',
      'CLUSTER_INSTALLED_4',
      'ADD_HOSTS_DEPLOY_PREP_2',
      'ADD_HOSTS_INSTALLING_3',
      'ADD_HOSTS_INSTALLED_4',
      'ADD_SERVICES_DEPLOY_PREP_2',
      'ADD_SERVICES_INSTALLING_3',
      'ADD_SERVICES_INSTALLED_4',
      'STOPPING_SERVICES',
      'STACK_UPGRADING',
      'STACK_UPGRADE_FAILED',
      'STACK_UPGRADED',
      'ADD_SECURITY_STEP_1',
      'ADD_SECURITY_STEP_2',
      'ADD_SECURITY_STEP_3',
      'ADD_SECURITY_STEP_4',
      'DISABLE_SECURITY',
      'HIGH_AVAILABILITY_DEPLOY',
      'ROLLBACK_HIGH_AVAILABILITY'],
    clusterState: 'CLUSTER_NOT_CREATED_1',
    wizardControllerName: null,
    localdb: null,
    key: 'CLUSTER_CURRENT_STATUS',
    isInstalled: function(){
      var notInstalledStates = ['CLUSTER_NOT_CREATED_1', 'CLUSTER_DEPLOY_PREP_2', 'CLUSTER_INSTALLING_3', 'SERVICE_STARTING_3'];
      return !notInstalledStates.contains(this.get('clusterState'));
    }.property('clusterState'),
    /**
     * get cluster data from server and update cluster status
     * @param {Boolean} isAsync set this to true if the call is to be made asynchronously.  if unspecified, false is assumed
     * @param {Boolean} overrideLocaldb
     * @return promise object for the get call
     */
    updateFromServer: function(isAsync, overrideLocaldb) {
      // if isAsync is undefined, set it to false
      isAsync = isAsync || false;
      // if overrideLocaldb is undefined, set it to true
      if(typeof overrideLocaldb == "undefined"){
        overrideLocaldb =  true;
      }
      var user = App.db.getUser();
      var login = App.db.getLoginName();
      var url = App.apiPrefix + '/persist/' + this.get('key');
      return jQuery.ajax(
        {
          url: url,
          context: this,
          async: isAsync,
          success: function (response) {
            if (response) {
              var newValue = jQuery.parseJSON(response);
              if (newValue.clusterState) {
                this.set('clusterState', newValue.clusterState);
              }
              if (newValue.clusterName) {
                this.set('clusterName', newValue.clusterName);
              }
              if (newValue.wizardControllerName) {
                this.set('wizardControllerName', newValue.wizardControllerName);
              }
              if (newValue.localdb) {
                this.set('localdb', newValue.localdb);
                // restore HAWizard data if process was started
                var isHAWizardStarted = App.get('isAdmin') && !App.isEmptyObject(newValue.localdb.HighAvailabilityWizard);
                if (overrideLocaldb || isHAWizardStarted) {
                  App.db.data = newValue.localdb;
                  App.db.setLocalStorage();
                  App.db.setUser(user);
                  App.db.setLoginName(login);
                }
              }
            } else {
              // default status already set
            }
            // this is to ensure that the local storage namespaces are initialized with all expected namespaces.
            // after upgrading ambari, loading local storage data from the "persist" data saved via an older version of
            // Ambari can result in missing namespaces that are defined in the new version of Ambari.
            App.db.mergeStorage();
          },
          error: function (xhr) {
            if (xhr.status == 404) {
              // default status already set
              console.log('Persist API did NOT find the key CLUSTER_CURRENT_STATUS');
              return;
            }
            App.ModalPopup.show({
              header: Em.I18n.t('common.error'),
              secondary: false,
              bodyClass: Ember.View.extend({
                template: Ember.Handlebars.compile('<p>{{t common.update.error}}</p>')
              })
            });
          },
          statusCode: require('data/statusCodes')
        }
      );
    },
    /**
     * update cluster status and post it on server
     * @param newValue
     * @param opt - used for additional ajax request options, by default ajax used synchronous mode
     *
     * @return {*}
     */
    setClusterStatus: function(newValue, opt){
      if(App.testMode) return false;
      var user = App.db.getUser();
      var login = App.db.getLoginName();
      var val = {clusterName: this.get('clusterName')};
      if (newValue) {
        //setter
        if (newValue.clusterName) {
          this.set('clusterName', newValue.clusterName);
          val.clusterName =  newValue.clusterName;
        }

        if (newValue.clusterState) {
          this.set('clusterState', newValue.clusterState);
          val.clusterState = newValue.clusterState;
        }
        if (newValue.wizardControllerName) {
          this.set('wizardControllerName', newValue.wizardControllerName);
          val.wizardControllerName = newValue.wizardControllerName;
        }

        if (newValue.localdb) {
          if (newValue.localdb.app && newValue.localdb.app.user)
            delete newValue.localdb.app.user;
          if (newValue.localdb.app && newValue.localdb.app.loginName)
            delete newValue.localdb.app.loginName;
          this.set('localdb', newValue.localdb);
          val.localdb = newValue.localdb;
        } else {
          delete App.db.data.app.user;
          delete App.db.data.app.loginName;
          val.localdb = App.db.data;
          App.db.setUser(user);
          App.db.setLoginName(login);
        }

        var keyValuePair = {};

        keyValuePair[this.get('key')] = JSON.stringify(val);

        var ajaxOptions = {
          name: 'cluster.state',
          sender: this,
          data: {
            keyValuePair: keyValuePair
          },
          beforeSend: 'clusterStatusBeforeSend',
          error: 'clusterStatusErrorCallBack'
        };

        if (opt) {
          ajaxOptions.async = !!opt.async;
          ajaxOptions.sender = opt.sender || this;
          ajaxOptions.success = opt.success;
          ajaxOptions.beforeSend = opt.beforeSend;
          ajaxOptions.error = opt.error;
        }

        App.ajax.send(ajaxOptions);
        return newValue;
      }
    },
    clusterStatusBeforeSend: function (keyValuePair) {
      console.log('BeforeSend: persistKeyValues', keyValuePair);
    },
    clusterStatusErrorCallBack: function(request, ajaxOptions, error, opt) {
      console.log("ERROR");
      var msg, doc;
      try {
        msg = 'Error ' + (request.status) + ' ';
        doc = $.parseXML(request.responseText);
        msg += $(doc).find("body p").text();
      } catch (e) {
        msg += JSON.parse(request.responseText).message;
      }

      App.ModalPopup.show({
        header: Em.I18n.t('common.error'),
        secondary: false,
        response: msg,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<p>{{t common.persist.error}} {{response}}</p>')
        })
      });
    },

    /**
     * general info about cluster
     */
    value: function () {
        return {
          clusterName: this.get('clusterName'),
          clusterState: this.get('clusterState'),
          wizardControllerName: this.get('wizardControllerName'),
          localdb: this.get('localdb')
        };
    }.property('clusterName', 'clusterState', 'localdb', 'wizardControllerName')

  });
  
});
window.require.register("models/config_group", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * Represents a configuration-group on the cluster.
   * A configuration-group is a collection of hosts
   * on which a collection of configurations are applied.
   *
   * Configuration group hierarchy is at 2 levels. For
   * each service there is a 'Default' configuration group
   * containing all hosts not belonging to any group of that
   * service.
   *
   * A default configuration group has child configuration
   * groups which contain configuration overrides (deltas)
   * for a bunch of hosts. This allows different configurations
   * for different hosts in a heterogeneous cluster environment.
   */
  App.ConfigGroup = Ember.Object.extend({
    id: null,
    name: null,
    description: null,
    isDefault: null,
    serviceName: null,

    /**
     * Parent configuration group for this group.
     * When {@link #isDefault} is true, this value is <code>null</code>
     * When {@link #isDefault} is false, this represents the configuration
     * deltas that are applied on the default.
     */
    parentConfigGroup: null,

    /**
     * Children configuration groups for this group.
     * When {@link #isDefault} is false, this value is <code>null</code>
     * When {@link #isDefault} is true, this represents the various
     * configuration groups that override the default.
     */
    childConfigGroups: [],

    /**
     * Service for which this configuration-group
     * is applicable.
     */
    service: null,

    /**
     * Hosts on which this configuration-group
     * is to be applied. For a service, a host can
     * belong to only one non-default configuration-group.
     *
     * When {#isDefault} is false, this contains hosts
     * for which the overrides will apply.
     *
     * When {#isDefault} is true, this value is empty, as
     * it dynamically reflects hosts not belonging to other
     * non-default groups.
     *
     */
    hosts: [],

    /**
     * In add service wizard we have installed services.
     * And on deploy step we need to update existing config groups
     * also mark it for be sure that config group data came from
     * installed service.
     *
     */
    isForUpdate: false,

    /**
     * Provides a display friendly name. This includes trimming
     * names to a certain length.
     */
    displayName: function () {
      var name = this.get('name');
      if (name && name.length>App.config.CONFIG_GROUP_NAME_MAX_LENGTH) {
        var middle = Math.floor(App.config.CONFIG_GROUP_NAME_MAX_LENGTH / 2);
        name = name.substring(0, middle) + "..." + name.substring(name.length-middle);
      }
      return name;
    }.property('name'),

    displayNameHosts: function () {
      return this.get('displayName') + ' (' + this.get('hosts.length') + ')';
    }.property('displayName', 'hosts.length'),

    apiResponse: null,

    /**
     * Provides hosts which are available for inclusion in
     * non-default configuration groups.
     */
    availableHosts: function () {
      if (this.get('isDefault')) return [];
      var unusedHostsMap = {};
      var availableHosts = [];
      var sharedHosts = [];
      //if cluster is not installed then get hosts from installerController instead of model
      if (App.clusterStatus.get('isInstalled')) {
        sharedHosts = App.Host.find();
      } else {
        sharedHosts = App.router.get('installerController.allHosts');
      }
      // parentConfigGroup.hosts(hosts from default group) - are available hosts, which don't belong to any group
      this.get('parentConfigGroup.hosts').forEach(function (hostName) {
        unusedHostsMap[hostName] = true;
      });
      sharedHosts.forEach(function (host) {
        if (unusedHostsMap[host.get('id')]) {
          availableHosts.pushObject(Ember.Object.create({
            selected: false,
            host: host,
            hostComponentNames: host.get('hostComponents').mapProperty('componentName')
          }));
        }
      });
      return availableHosts;
    }.property('isDefault', 'parentConfigGroup', 'childConfigGroups', 'parentConfigGroup.hosts.@each'),

    isAddHostsDisabled: function () {
      return (this.get('isDefault') || this.get('availableHosts.length') === 0);
    }.property('availableHosts.length'),

    /**
     * Collection of (site, tag) pairs representing properties.
     *
     * When {#isDefault} is true, this represents the
     * default cluster configurations for that service.
     *
     * When {#isDefault} is false, this represents the
     * configuration overrides on top of the cluster default for the
     * hosts identified by 'hosts'.
     */
    configSiteTags: [],

    properties: [],

    propertiesList: function () {
      var result = '';
      this.get('properties').forEach(function (item) {
        result += item.name + " : " + item.value + '<br/>';
      }, this);
      return result;
    }.property('properties.length')
  });
  
});
window.require.register("models/dataset", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.Dataset = DS.Model.extend({
    name: DS.attr('string'),
    status: DS.attr('string'),
    sourceClusterName: DS.attr('string'),
    targetClusterName: DS.attr('string'),
    sourceDir: DS.attr('string'),
    targetDir: DS.attr('string'),
    frequency: DS.attr('string'),
    frequencyUnit: DS.attr('string'),
    scheduleStartDate: DS.attr('string'),
    scheduleEndDate: DS.attr('string'),
    datasetJobs: DS.hasMany('App.DataSetJob'),

    // name with special prefix to distinguish feeds created with Ambari
    prefixedName: function () {
      return App.mirroringDatasetNamePrefix + this.get('name');
    }.property('name'),

    statusFormatted: function (){
      var status = this.get('status');
      if (status) {
        return status.toLowerCase().capitalize();
      }
    }.property('status'),

    isRunning: function () {
      return this.get('status') === 'RUNNING';
    }.property('status'),

    isSuspended: function () {
      return this.get('status') === 'SUSPENDED';
    }.property('status'),

    isSubmitted: function () {
      return this.get('status') === 'SUBMITTED';
    }.property('status'),

    //Last succeeded date. Will be calculated later.
    lastSucceededDate: function () {
      return '';
    }.property(),

    //Next instance to run. Will be calculated later.
    nextInstance: function () {
      return '';
    }.property(),

    //Class name for dataset health status indicator
    healthClass: function () {
      var jobs = this.get('datasetJobs').toArray();
      jobs = jobs.filterProperty('status', 'FAILED').concat(jobs.filterProperty('status', 'SUCCESSFUL'));
      jobs = jobs.sortProperty('endDate');
      return jobs.length && jobs[0].get('status') === 'FAILED' ? 'health-status-DEAD-RED' : 'health-status-LIVE';
    }.property('datasetJobs', 'datasetJobs.@each.status'),

    healthIconClass: function () {
      switch (this.get('healthClass')) {
        case 'health-status-LIVE':
          return App.healthIconClassGreen;
          break;
        case 'health-status-DEAD-RED':
          return App.healthIconClassRed;
          break;
        default:
          return "";
          break;
      }
    }.property('healthClass')
  });

  App.Dataset.FIXTURES = [];
});
window.require.register("models/dataset_job", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var date = require('utils/date');

  App.DataSetJob = DS.Model.extend({
    name: DS.attr('string'),
    dataset: DS.belongsTo('App.Dataset'),
    status: DS.attr('string'),
    startDate: DS.attr('number'),
    endDate: DS.attr('number'),

    statusFormatted: function () {
      return this.get('status').toLowerCase().capitalize();
    }.property('status'),

    isSuspended: function () {
      return this.get('status') === 'SUSPENDED';
    }.property('status'),

    startFormatted: function () {
      if (this.get('startDate')) {
        return $.timeago(this.get('startDate'));
      }
      return '';
    }.property('startDate'),

    endFormatted: function () {
      if (this.get('endDate')) {
        return $.timeago(this.get('endDate'));
      }
      return '';
    }.property('endDate'),

    healthClass: function () {
      var result = 'icon-question-sign';
      switch (this.get('status')) {
        case 'SUCCEEDED':
          result = 'icon-ok';
          break;
        case 'SUSPENDED':
          result = 'icon-cog';
          break;
        case 'WAITING':
          result = 'icon-time';
          break;
        case 'RUNNING':
          result = 'icon-play';
          break;
        case 'KILLED':
          result = 'icon-exclamation-sign';
          break;
        case 'FAILED':
          result = 'icon-warning-sign';
          break;
        case 'ERROR':
          result = 'icon-remove';
          break;
      }
      return result;
    }.property('status')
  });


  App.DataSetJob.FIXTURES = [];
  
});
window.require.register("models/form", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');

  // move this to models cause some errors
  App.Form = Em.View.extend({
    /**
     * generating fields from fieldsOptions
     */
    classNames:["form-horizontal"],
    attributeBindings: ['autocomplete'],
    autocomplete: 'off',
    i18nprefix:'form.',
    fields:[],
    field:{},
    messages:[],
    object:false,
    result:0, // save result var (-1 - error; 0 - init; 1 - success)
    templateName:require('templates/common/form'),
    tagName:'form',

    init:function () {
      var thisForm = this;
      if (!this.fields.length) {
        this.fieldsOptions.forEach(
          function (options) {
            var field = App.FormField.create(options);
            field.set('form', thisForm);
            thisForm.fields.push(field);
            thisForm.set("field." + field.get('name'), field);
          }
        );
      }
      this._super();
    },

    getField:function (name) {
      var field = false;
      $.each(this.fields, function () {
        if (this.get('name') == name) {
          return field = this;
        }
      });
      return field;
    },

    isValid:function () {
      var isValid = true;
      $.each(this.fields, function () {
        this.validate();
        if (!this.get('isValid')) {
          isValid = false;
          console.warn(this.get('name') + " IS INVALID : " + this.get('errorMessage'));
        }
      });

      return isValid;
    },

    updateValues:function () {
      var object = this.get('object');
      if (object instanceof Em.Object) {
        $.each(this.fields, function () {
          this.set('value', (this.get('displayType') == 'password') ? '' : object.get(this.get('name')));
        });
      } else {
        this.clearValues();
      }

    }.observes("object"),

    clearValues:function () {
      $.each(this.fields, function () {
        this.set('value', '');
      });
    },

    visibleFields:function () {
      var fields = this.get('fields');
      var visible = [];
      fields.forEach(function (field) {
        if (!field.get('isHiddenField')) {
          visible.push(field);
        }
      });
      return visible;
    }.property('fields'),

    resultText:function () {
      var text = "";
      switch (this.get('result')) {
        case -1:
          text = this.t("form.saveError");
          break;
        case 1:
          text = this.t("form.saveSuccess");
          break;
      }

      return text;
    }.property('result')
  });

  App.FormField = Em.Object.extend({ // try to realize this as view
    name:'',
    displayName:'',
  //  defaultValue:'', NOT REALIZED YET
    description:'',
    disabled:false,
    displayType:'string', // string, digits, number, directories, textarea, checkbox
    disableRequiredOnPresent:false,
    errorMessage:'',
    warnMessage:'',
    form:false,
    isRequired:true, // by default a config property is required
    unit:'',
    value:'',

    observeValue:function () {

      if (this.get('displayType') == 'hidden')
        console.warn(" FORM FIELD VALUE: ", this.get('value'));

    }.observes('value'),

    isValid:function () {
      return this.get('errorMessage') === '';
    }.property('errorMessage'),

    viewClass:function () {
      var options = {};
      var element = Em.TextField;
      switch (this.get('displayType')) {
        case 'checkbox':
          element = Em.Checkbox;
          options.checkedBinding = "value";
          break;
        case 'select':
          element = Em.Select;
          options.content = this.get('values');
          options.valueBinding = "value";
          options.optionValuePath = "content.value";
          options.optionLabelPath = "content.label";
          break;
        case 'password':
          options['type'] = 'password';
          break;
        case 'textarea':
          element = Em.TextArea;
          break;
        case 'hidden':
          options.type = "hidden";
          break;
      }

      return element.extend(options);
    }.property('displayType'),

    validate:function () {
      var value = this.get('value');
      var isError = false;
      this.set('errorMessage', '');

      if (this.get('isRequired') && (typeof value === 'string' && value.trim().length === 0)) {
        this.set('errorMessage', 'This is required');
        isError = true;
      }

      if (typeof value === 'string' && value.trim().length === 0) { // this is not to validate empty field.
        isError = true;
      }

      if (!isError) {
        this.set('errorMessage', '');
      }
    },

    isHiddenField:function () {
      return this.get('displayType') == 'hidden';
    }.property('type')
  });
  
});
window.require.register("models/host", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var misc = require('utils/misc');

  App.Host = DS.Model.extend({
    hostName: DS.attr('string'),
    publicHostName: DS.attr('string'),
    cluster: DS.belongsTo('App.Cluster'),
    hostComponents: DS.hasMany('App.HostComponent'),
    cpu: DS.attr('string'),
    cpuPhysical: DS.attr('string'),
    memory: DS.attr('string'),
    diskTotal: DS.attr('number'),
    diskFree: DS.attr('number'),
    osArch: DS.attr('string'),
    ip: DS.attr('string'),
    rack: DS.attr('string'),
    healthStatus: DS.attr('string'),
    lastHeartBeatTime: DS.attr('number'),
    osType: DS.attr("string"),
    diskInfo: DS.attr('object'),
    loadOne:DS.attr('number'),
    loadFive:DS.attr('number'),
    loadFifteen:DS.attr('number'),
    memTotal:DS.attr('number'),
    memFree:DS.attr('number'),
    cpuSystem:DS.attr('number'),
    cpuUser:DS.attr('number'),
    criticalAlertsCount: DS.attr('number'),
    passiveState: DS.attr('string'),

    /**
     * Is host checked at the main Hosts page
     */
    selected:DS.attr('boolean'),

    /**
     * Overall CPU usage (system and user)
     * @returns {Number}
     */
    cpuUsage: function () {
      if (this.get('cpuSystem') && this.get('cpuUser')) {
        return this.get('cpuSystem') + this.get('cpuUser');
      }
      return 0;
    }.property('cpuSystem', 'cpuUser'),

    /**
     * Percent value of used memory
     * @returns {Number}
     */
    memoryUsage: function () {
      if (this.get('memFree') && this.get('memTotal')) {
        var memUsed = this.get('memTotal') - this.get('memFree');
        return (100 * memUsed) / this.get('memTotal');
      }
      return 0;
    }.property('memTotal', 'memFree'),

    componentsWithStaleConfigs: function () {
      return this.get('hostComponents').filterProperty('staleConfigs', true);
    }.property('hostComponents.@each.staleConfigs'),

    /**
     * Get count of host components with stale configs
     * @returns {Number}
     */
    componentsWithStaleConfigsCount: function() {
      return this.get('componentsWithStaleConfigs').length;
    }.property('componentsWithStaleConfigs.length'),

    /**
     * Get count of host components in passive state
     * @returns {Number}
     */
    componentsInPassiveStateCount: function() {
      return this.get('hostComponents').filter(function(component) {
        return component.get('passiveState') !== 'OFF';
      }).length;
    }.property('hostComponents.@each.passiveState'),

    /**
     * Count of mounted on host disks
     * @returns {Number}
     */
    disksMounted: function() {
      return this.get('diskInfo.length');
    }.property('diskInfo.length'),

    coresFormatted: function() {
      return this.get('cpu') + ' (' + this.get('cpuPhysical') + ')';
    }.property('cpu', 'cpuPhysical'),

    /**
     * API return diskTotal and diskFree. Need to save their different
     * @returns {Number}
     */
    diskUsed: function(){
      return this.get('diskTotal') - this.get('diskFree');
    }.property('diskFree', 'diskTotal'),

    /**
     * Format diskUsed value to float with 2 digits (also convert to GB)
     * @returns {String} Format: '*** GB'
     */
    diskUsedFormatted: function() {
      return Math.round(this.get('diskUsed') * Math.pow(10, 2)) / Math.pow(10, 2) + 'GB';
    }.property('diskUsed'),

    /**
     * Format diskTotal value to float with 2 digits (also convert to GB)
     * @returns {String} Format: '*** GB'
     */
    diskTotalFormatted: function() {
      return Math.round(this.get('diskTotal') * Math.pow(10, 2)) / Math.pow(10, 2) + 'GB';
    }.property('diskTotal'),

    /**
     * Percent value of used disk space
     * @returns {Number}
     */
    diskUsage: function() {
      return (this.get('diskUsed')) / this.get('diskTotal') * 100;
    }.property('diskUsed', 'diskTotal'),

    /**
     * Format diskUsage to float with 2 digits
     * @returns {String} Format: '**.** %'
     */
    diskUsageFormatted: function() {
      if (isNaN(this.get('diskUsage')) || this.get('diskUsage') < 0) {
        return Em.I18n.t('hosts.host.metrics.dataUnavailable');
      }
      var s = Math.round(this.get('diskUsage') * Math.pow(10, 2)) / Math.pow(10, 2);
      if (isNaN(s)) {
        s = 0;
      }
      return s + '%';
    }.property('diskUsage'),

    /**
     * Formatted string with data about disk usage
     * @returns {String}
     */
    diskInfoBar: function() {
      if (isNaN(this.get('diskUsage')) || this.get('diskUsage') < 0) {
        return this.get('diskUsageFormatted');
      }
      return this.get('diskUsedFormatted') + '/' + this.get('diskTotalFormatted') + ' (' + this.get('diskUsageFormatted')
        + ' ' + Em.I18n.t('services.service.summary.diskInfoBar.used') + ')';
    }.property('diskUsedFormatted', 'diskTotalFormatted'),

    /**
     * Formatted bytes to appropriate value
     * @returns {String}
     */
    memoryFormatted: function () {
      return misc.formatBandwidth(this.get('memory') * 1024);
    }.property('memory'),

    /**
     * Return true if the host <code>healthStatus</code> is UNKNOWN
     * @returns {bool}
     */
    isNotHeartBeating : function() {
      return (App.testMode) ? false : (this.get('healthStatus') === "UNKNOWN");
    }.property('lastHeartBeatTime'),

    /**
     * Average load
     * @returns {Number}
     */
    loadAvg: function() {
      if (this.get('loadOne') != null) return this.get('loadOne').toFixed(2);
      if (this.get('loadFive') != null) return this.get('loadFive').toFixed(2);
      if (this.get('loadFifteen') != null) return this.get('loadFifteen').toFixed(2);
      return null;
    }.property('loadOne', 'loadFive', 'loadFifteen'),

    /**
     * Host health indicator
     * Based on  <code>healthStatus</code>
     * @returns {String}
     */
    healthClass: function(){
      if (this.get('passiveState')!= 'OFF') {
        return 'icon-medkit';
      }
      var statusMap = {
        'UNKNOWN': 'health-status-DEAD-YELLOW',
        'HEALTHY': 'health-status-LIVE',
        'UNHEALTHY': 'health-status-DEAD-RED',
        'ALERT': 'health-status-DEAD-ORANGE'
      };
      return statusMap[this.get('healthStatus')] || 'health-status-DEAD-YELLOW';
    }.property('healthStatus'),

    healthIconClass: function () {
      switch (this.get('healthClass')) {
        case 'health-status-LIVE':
          return App.healthIconClassGreen;
          break;
        case 'health-status-DEAD-RED':
          return App.healthIconClassRed;
          break;
        case 'health-status-DEAD-YELLOW':
          return App.healthIconClassYellow;
          break;
        case 'health-status-DEAD-ORANGE':
          return App.healthIconClassOrange;
          break;
        default:
          return "";
          break;
      }
    }.property('healthClass'),

    /**
     * Tooltip for host indicator
     * Contains affected host components names (based on <code>healthClass</code>)
     * @returns {String}
     */
    healthToolTip: function(){
      var hostComponents = this.get('hostComponents').filter(function(item){
        return item.get('workStatus') !== App.HostComponentStatus.started;
      });
      var output = '';
      if (this.get('passiveState') != 'OFF') {
        return Em.I18n.t('hosts.host.passive.mode');
      }
      switch (this.get('healthClass')){
        case 'health-status-DEAD-RED':
          hostComponents = hostComponents.filterProperty('isMaster', true);
          output = Em.I18n.t('hosts.host.healthStatus.mastersDown');
          hostComponents.forEach(function(hc, index){
            output += (index == (hostComponents.length-1)) ? hc.get('displayName') : (hc.get('displayName')+", ");
          }, this);
          break;
        case 'health-status-DEAD-YELLOW':
          output = Em.I18n.t('hosts.host.healthStatus.heartBeatNotReceived');
          break;
        case 'health-status-DEAD-ORANGE':
          hostComponents = hostComponents.filterProperty('isSlave', true);
          output = Em.I18n.t('hosts.host.healthStatus.slavesDown');
          hostComponents.forEach(function(hc, index){
            output += (index == (hostComponents.length-1)) ? hc.get('displayName') : (hc.get('displayName')+", ");
          }, this);
          break;
        case 'health-status-LIVE':
          output = Em.I18n.t('hosts.host.healthStatus.allUp');
          break;
      }
      return output;
    }.property('hostComponents.@each.workStatus','hostComponents.@each.passiveState')
  });

  App.Host.FIXTURES = [];
  
});
window.require.register("models/host_component", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HostComponent = DS.Model.extend({
    workStatus: DS.attr('string'),
    passiveState: DS.attr('string'),
    componentName: DS.attr('string'),
    haStatus: DS.attr('string'),
    displayNameAdvanced: DS.attr('string'),
    staleConfigs: DS.attr('boolean'),
    host: DS.belongsTo('App.Host'),
    service: DS.belongsTo('App.Service'),
    /**
     * Determine if component is client
     * @returns {bool}
     */
    isClient:function () {
      if(['PIG', 'SQOOP', 'HCAT', 'MAPREDUCE2_CLIENT'].contains(this.get('componentName'))){
        return true;
      }

      return Boolean(this.get('componentName').match(/_client/gi));
    }.property('componentName'),
    /**
     * Determine if component is running now
     * Based on <code>workStatus</code>
     * @returns {bool}
     */
    isRunning: function(){
      return (this.get('workStatus') == 'STARTED' || this.get('workStatus') == 'STARTING');
    }.property('workStatus'),

    /**
     * Formatted <code>componentName</code>
     * @returns {String}
     */
    displayName: function () {
      return App.format.role(this.get('componentName'));
    }.property('componentName'),

    /**
     * Determine if component is master
     * @returns {bool}
     */
    isMaster: function () {
      switch (this.get('componentName')) {
        case 'NAMENODE':
        case 'SECONDARY_NAMENODE':
        case 'SNAMENODE':
        case 'JOURNALNODE':
        case 'JOBTRACKER':
        case 'ZOOKEEPER_SERVER':
        case 'HIVE_SERVER':
        case 'HIVE_METASTORE':
        case 'MYSQL_SERVER':
        case 'HBASE_MASTER':
        case 'NAGIOS_SERVER':
        case 'GANGLIA_SERVER':
        case 'OOZIE_SERVER':
        case 'WEBHCAT_SERVER':
        case 'HUE_SERVER':
        case 'HISTORYSERVER':
        case 'FLUME_SERVER':
        case 'FALCON_SERVER':
        case 'NIMBUS':
        case 'STORM_UI_SERVER':
        case 'DRPC_SERVER':
        case 'STORM_REST_API':
        case 'RESOURCEMANAGER':
        case 'APP_TIMELINE_SERVER':
          return true;
        default:
          return false;
      }
    }.property('componentName'),

    /**
     * Determine if component is slave
     * @returns {bool}
     */
    isSlave: function(){
      switch (this.get('componentName')) {
        case 'DATANODE':
        case 'TASKTRACKER':
        case 'HBASE_REGIONSERVER':
        case 'GANGLIA_MONITOR':
        case 'NODEMANAGER':
        case 'ZKFC':
        case 'SUPERVISOR':
          return true;
        default:
          return false;
      }
    }.property('componentName'),
    /**
     * Only certain components can be deleted.
     * They include some from master components, 
     * some from slave components, and rest from 
     * client components.
     * @returns {bool}
     */
    isDeletable: function() {
      var canDelete = false;
      switch (this.get('componentName')) {
        case 'DATANODE':
        case 'TASKTRACKER':
        case 'ZOOKEEPER_SERVER':
        case 'HBASE_REGIONSERVER':
        case 'GANGLIA_MONITOR':
        case 'SUPERVISOR':
        case 'NODEMANAGER':
          canDelete = true;
          break;
        default:
      }
      if (!canDelete) {
        canDelete = this.get('isClient');
      }
      return canDelete;
    }.property('componentName', 'isClient'),
    /**
     * A host-component is decommissioning when it is in HDFS service's list of
     * decomNodes.
     * @returns {bool}
     */
    isDecommissioning: function () {
      var decommissioning = false;
      var hostName = this.get('host.hostName');
      var componentName = this.get('componentName');
      var hdfsSvc = App.HDFSService.find().objectAt(0);
      if (componentName === 'DATANODE' && hdfsSvc) {
        var decomNodes = hdfsSvc.get('decommissionDataNodes');
        var decomNode = decomNodes != null ? decomNodes.findProperty("hostName", hostName) : null;
        decommissioning = decomNode != null;
      }
      return decommissioning;
    }.property('componentName', 'host.hostName', 'App.router.clusterController.isLoaded', 'App.router.updateController.isUpdated'),
    /**
     * User friendly host component status
     * @returns {String}
     */
    isActive: function() {
      return (this.get('passiveState') == 'OFF');
    }.property('passiveState'),

    passiveTooltip: function() {
      if (!this.get('isActive')) {
        return Em.I18n.t('hosts.component.passive.mode');
      }
    }.property('isActive'),

    statusClass: function() {
      return this.get('isActive') ? this.get('workStatus') : 'icon-medkit';
    }.property('workStatus','isActive'),

    statusIconClass: function () {
      switch (this.get('statusClass')) {
        case 'STARTED':
        case 'STARTING':
          return App.healthIconClassGreen;
          break;
        case 'INSTALLED':
        case 'STOPPING':
          return App.healthIconClassRed;
          break;
        case 'UNKNOWN':
          return App.healthIconClassYellow;
          break;
        default:
          return "";
          break;
      }
    }.property('statusClass'),

    componentTextStatus: function () {
      return App.HostComponentStatus.getTextStatus(this.get("workStatus"));
    }.property('workStatus','isDecommissioning')
  });

  App.HostComponent.FIXTURES = [];

  App.HostComponentStatus = {
    started: "STARTED",
    starting: "STARTING",
    stopped: "INSTALLED",
    stopping: "STOPPING",
    install_failed: "INSTALL_FAILED",
    installing: "INSTALLING",
    upgrade_failed: "UPGRADE_FAILED",
    unknown: "UNKNOWN",
    disabled: "DISABLED",
    init: "INIT",

    /**
     * Get host component status in "machine" format
     * @param {String} value
     * @returns {String}
     */
    getKeyName:function(value){
      switch(value){
        case this.started:
          return 'started';
        case this.starting:
          return 'starting';
        case this.stopped:
          return 'installed';
        case this.stopping:
          return 'stopping';
        case this.install_failed:
          return 'install_failed';
        case this.installing:
          return 'installing';
        case this.upgrade_failed:
          return 'upgrade_failed';
        case this.disabled:
        case this.unknown:
          return 'unknown';
      }
      return 'unknown';
    },

    /**
     * Get user-friendly host component status
     * @param {String} value
     * @returns {String}
     */
    getTextStatus: function (value) {
      switch (value) {
        case this.installing:
          return 'Installing...';
        case this.install_failed:
          return 'Install Failed';
        case this.stopped:
          return 'Stopped';
        case this.started:
          return 'Started';
        case this.starting:
          return 'Starting...';
        case this.stopping:
          return 'Stopping...';
        case this.unknown:
          return 'Heartbeat lost...';
        case this.upgrade_failed:
          return 'Upgrade Failed';
        case this.disabled:
          return 'Disabled';
        case this.init:
          return 'Install Pending...';
      }
      return 'Unknown';
    },

    /**
     * Get list of possible <code>App.HostComponent</code> statuses
     * @returns {String[]}
     */
    getStatusesList: function() {
      var ret = [];
      for (var st in this) {
        if (this.hasOwnProperty(st) && Em.typeOf(this[st]) == 'string') {
          ret.push(this[st]);
        }
      }
      return ret;
    }
  };

  
});
window.require.register("models/hosts", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HostInfo = Ember.Object.extend({
    elementId: 'host',
    name: '',
    cpu: null,
    memory: null,
    message: 'Information',
    barColor: 'progress-info',
    isChecked: true,
    bootLog:null,
    bootStatus: 'PENDING',
    
    bootStatusForDisplay:function () {
      switch (this.get('bootStatus')) {
        case 'PENDING':
          return 'Preparing';
        case 'REGISTERED':
          return 'Success';
        case 'FAILED':
          return 'Failed';
        case 'RUNNING':
          return 'Installing';
        case 'DONE':
        case 'REGISTERING':
        default:
          return 'Registering';
      }
    }.property('bootStatus'),

    bootBarColor:function () {
      switch (this.get('bootStatus')) {
        case 'REGISTERED':
          return 'progress-success';
        case 'FAILED':
          return 'progress-danger';
        case 'PENDING':
        case 'RUNNING':
        case 'DONE':
        case 'REGISTERING':
        default:
          return 'progress-info';
      }
    }.property('bootStatus'),

    bootStatusColor:function () {
      switch (this.get('bootStatus')) {
        case 'REGISTERED':
          return 'text-success';
        case 'FAILED':
          return 'text-error';
        case 'PENDING':
        case 'RUNNING':
        case 'DONE':
        case 'REGISTERING':
        default:
          return 'text-info';
      }
    }.property('bootStatus'),

    isBootDone:function () {
      switch (this.get('bootStatus')) {
        case 'REGISTERED':
        case 'FAILED':
          return true;
        case 'PENDING':
        case 'RUNNING':
        case 'DONE':
        case 'REGISTERING':
        default:
          return false;
      }

    }.property('bootStatus')
  });
  
});
window.require.register("models/job", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var date = require('utils/date');
  var misc = require('utils/misc');

  App.Job = DS.Model.extend({

    run: DS.belongsTo('App.Run'),

    jobName: DS.attr('string'),
    workflowEntityName: DS.attr('string'),
    userName: DS.attr('string'),
    confPath: DS.attr('string'),
    submitTime: DS.attr('number'),
    maps: DS.attr('number'),
    reduces: DS.attr('number'),
    status: DS.attr('string'),
    input: DS.attr('number'),
    output: DS.attr('number'),
    elapsedTime: DS.attr('number'),
    duration: function () {
      return date.timingFormat(parseInt(this.get('elapsedTime')));
    }.property('elapsedTime'),
    jobTimeLine: DS.attr('string'),
    jobTaskView: DS.attr('string'),
    /**
     *  Sum of input bandwidth for all jobs with appropriate measure
     */
    inputFormatted: function () {
      var input = this.get('input');
      return misc.formatBandwidth(input);
    }.property('input'),
    /**
     *  Sum of output bandwidth for all jobs with appropriate measure
     */
    outputFormatted: function () {
      var output = this.get('output');
      return misc.formatBandwidth(output);
    }.property('output')

  });

  App.Job.FIXTURES = [];
  
});
window.require.register("models/jobs/hive_job", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.HiveJob = App.AbstractJob.extend({
    jobType : App.JobType.HIVE,
    queryText : DS.attr('string'),
    stages : DS.attr('array'),
    hasTezDag: DS.attr('boolean'),
    tezDag : DS.belongsTo('App.TezDag'),
    failed: DS.attr('boolean')
  });

  App.HiveJob.FIXTURES = [];
  
});
window.require.register("models/jobs/job", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var dateUtils = require('utils/date');

  /**
   * Base class of all Jobs.
   *
   * This class is meant to be extended and not instantiated directly.
   */
  App.AbstractJob = DS.Model.extend({
    id : DS.attr('string'),
    name : DS.attr('string'),
    user : DS.attr('string'),
    startTime : DS.attr('number'),
    endTime : DS.attr('number'),

    startTimeDisplay : function() {
      var startTime = this.get('startTime');
      if (startTime > 0) {
        return dateUtils.dateFormat(startTime);
      } else {
        return "";
      }
    }.property('startTime'),

    endTimeDisplay : function() {
      var endTime = this.get('endTime');
      if (endTime > 0) {
        return dateUtils.dateFormat(endTime);
      } else {
        return "";
      }
    }.property('endTime'),

    /**
     * Provides the duration of this job. If the job has not started, duration
     * will be given as 0. If the job has not ended, duration will be till now.
     * 
     * @return {Number} Duration in milliseconds.
     */
    duration : function() {
      var startTime = this.get('startTime');
      var endTime = this.get('endTime');
      if(endTime < startTime || endTime == undefined) {
        endTime =  App.get('currentServerTime');
      }
      return dateUtils.duration(startTime, endTime);
    }.property('startTime', 'endTime'),

    durationDisplay : function() {
      return dateUtils.timingFormat(this.get('duration'), true);
    }.property('duration'),

    /**
     * Type of this job. Should be one of constants defined in App.JobType
     */
    jobType : DS.attr('string')
  });

  App.JobType = {
    HIVE : "hive"
  };

  App.AbstractJob.FIXTURES = [];
  
});
window.require.register("models/jobs/tez_dag", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var dateUtils = require('utils/date');
  var numberUtils = require('utils/number_utils');

  App.TezDag = DS.Model.extend({
    id : DS.attr('string'),
    /**
     * When DAG is actually running on server, a unique ID is assigned.
     */
    instanceId : DS.attr('string'),
    name : DS.attr('string'),
    yarnApplicationId: DS.attr('string'),
    stage : DS.attr('string'),
    vertices : DS.hasMany('App.TezDagVertex'),
    edges : DS.hasMany('App.TezDagEdge')
  });

  App.TezDagEdge = DS.Model.extend({
    id : DS.attr('string'),
    instanceId : DS.attr('string'),
    fromVertex : DS.belongsTo('App.TezDagVertex'),
    toVertex : DS.belongsTo('App.TezDagVertex'),
    /**
     * Type of this edge connecting vertices. Should be one of constants defined
     * in 'App.TezDagEdgeType'.
     */
    edgeType : DS.attr('string')
  });

  App.TezDagVertex = DS.Model.extend({
    id : DS.attr('string'),
    /**
     * When DAG vertex is actually running on server, a unique ID is assigned.
     */
    instanceId : DS.attr('string'),
    name : DS.attr('string'),

    /**
     * State of this vertex. Should be one of constants defined in
     * App.TezDagVertexState.
     */
    state : DS.attr('string'),

    /**
     * Vertex type has to be one of the types defined in 'App.TezDagVertexType'
     * @return {string}
     */
    type : DS.attr('string'),

    /**
     * A vertex can have multiple incoming edges.
     */
    incomingEdges : DS.hasMany('App.TezDagEdge'),

    /**
     * This vertex can have multiple outgoing edges.
     */
    outgoingEdges : DS.hasMany('App.TezDagEdge'),

    startTime : DS.attr('number'),
    endTime : DS.attr('number'),

    /**
     * Provides the duration of this job. If the job has not started, duration
     * will be given as 0. If the job has not ended, duration will be till now.
     *
     * @return {Number} Duration in milliseconds.
     */
    duration : function() {
      return dateUtils.duration(this.get('startTime'), this.get('endTime'))
    }.property('startTime', 'endTime'),

    /**
     * Each Tez vertex can perform arbitrary application specific computations
     * inside. The application can provide a list of operations it has provided in
     * this vertex.
     *
     * Array of strings. [{string}]
     */
    operations : DS.attr('array'),

    /**
     * Provides additional information about the 'operations' performed in this
     * vertex. This is shown directly to the user.
     */
    operationPlan : DS.attr('string'),

    /**
     * Number of actual Map/Reduce tasks in this vertex
     */
    tasksCount : DS.attr('number'),

    tasksNumber: function () {
      return this.get('tasksCount') ? this.get('tasksCount') : 0;
    }.property('tasksCount'),

    /**
     * Local filesystem usage metrics for this vertex
     */
    fileReadBytes : DS.attr('number'),
    fileWriteBytes : DS.attr('number'),
    fileReadOps : DS.attr('number'),
    fileWriteOps : DS.attr('number'),

    /**
     * Spilled records
     */
    spilledRecords : DS.attr('number'),

    /**
     * HDFS usage metrics for this vertex
     */
    hdfsReadBytes : DS.attr('number'),
    hdfsWriteBytes : DS.attr('number'),
    hdfsReadOps : DS.attr('number'),
    hdfsWriteOps : DS.attr('number'),

    /**
     * Record metrics for this vertex
     */
    recordReadCount : DS.attr('number'),
    recordWriteCount : DS.attr('number'),

    totalReadBytes : function() {
      return this.get('fileReadBytes') + this.get('hdfsReadBytes');
    }.property('fileReadBytes', 'hdfsReadBytes'),

    totalWriteBytes : function() {
      return this.get('fileWriteBytes') + this.get('hdfsWriteBytes');
    }.property('fileWriteBytes', 'hdfsWriteBytes'),

    totalReadBytesDisplay : function() {
      return numberUtils.bytesToSize(this.get('totalReadBytes'));
    }.property('totalReadBytes'),

    totalWriteBytesDisplay : function() {
      return numberUtils.bytesToSize(this.get('totalWriteBytes'));
    }.property('totalWriteBytes'),

    durationDisplay : function() {
      return dateUtils.timingFormat(this.get('duration'), true);
    }.property('duration')
  });

  App.TezDagVertexState = {
    NEW : "NEW",
    INITIALIZING : "INITIALIZING",
    INITED : "INITED",
    RUNNING : "RUNNING",
    SUCCEEDED : "SUCCEEDED",
    FAILED : "FAILED",
    KILLED : "KILLED",
    ERROR : "ERROR",
    TERMINATING : "TERMINATING"
  };

  App.TezDagVertexType = {
    MAP: 'MAP',
    REDUCE: 'REDUCE',
    UNION: 'UNION'
  };

  App.TezDagEdgeType = {
    SCATTER_GATHER : "SCATTER_GATHER",
    BROADCAST : "BROADCAST",
    CONTAINS: "CONTAINS"
  };

  App.TezDag.FIXTURES = [];
  App.TezDagEdge.FIXTURES = [];
  App.TezDagVertex.FIXTURES = [];
});
window.require.register("models/quick_links", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var portRegex = '\\w*:(\\d+)';

  App.QuickLinks = DS.Model.extend({
    label: DS.attr('string'),
    url: DS.attr('string'),
    service_id: DS.attr('string'),
    template: DS.attr('string'),
    http_config: DS.attr('string'),
    https_config: DS.attr('string'),
    site: DS.attr('string'),
    regex: DS.attr('string'),
    default_http_port: DS.attr('number'),
    default_https_port: DS.attr('number')
  });


  App.QuickLinks.FIXTURES = [
    {
      id:1,
      label:'NameNode UI',
      url:'%@://%@:%@/dfshealth.jsp',
      service_id: 'HDFS',
      template:'%@://%@:%@/dfshealth.jsp',
      http_config: 'dfs.namenode.http-address',
      https_config: 'dfs.namenode.https-address',
      site: 'hdfs-site',
      regex: portRegex,
      default_http_port: 50070,
      default_https_port: 50470
    },
    {
      id:2,
      label:'NameNode logs',
      url:'%@://%@:%@/logs',
      service_id: 'HDFS',
      template:'%@://%@:%@/logs',
      http_config: 'dfs.namenode.http-address',
      https_config: 'dfs.namenode.https-address',
      site: 'hdfs-site',
      regex: portRegex,
      default_http_port: 50070,
      default_https_port: 50470
    },
    {
      id:3,
      label:'NameNode JMX',
      url:'%@://%@:%@/jmx',
      service_id: 'HDFS',
      template:'%@://%@:%@/jmx',
      http_config: 'dfs.namenode.http-address',
      https_config: 'dfs.namenode.https-address',
      site: 'hdfs-site',
      regex: portRegex,
      default_http_port: 50070,
      default_https_port: 50470
    },
    {
      id:4,
      label:'Thread Stacks',
      url:'%@://%@:%@/stacks',
      service_id: 'HDFS',
      template:'%@://%@:%@/stacks',
      http_config: 'dfs.namenode.http-address',
      https_config: 'dfs.namenode.https-address',
      site: 'hdfs-site',
      regex: portRegex,
      default_http_port: 50070,
      default_https_port: 50470
    },
    {
      id:5,
      label:'JobTracker UI',
      url:'%@://%@:%@/jobtracker.jsp',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/jobtracker.jsp',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:6,
      label:'Scheduling Info',
      url:'%@://%@:%@/scheduler',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/scheduler',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:7,
      label:'Running Jobs',
      url:'%@://%@:%@/jobtracker.jsp#running_jobs',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/jobtracker.jsp#running_jobs',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:8,
      label:'Retired Jobs',
      url:'%@://%@:%@/jobtracker.jsp#retired_jobs',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/jobtracker.jsp#retired_jobs',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:9,
      label:'JobHistory Server',
      url:'%@://%@:%@/jobhistoryhome.jsp',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/jobhistoryhome.jsp',
      http_config: 'mapreduce.history.server.http.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 51111
    },

    {
      id:10,
      label:'JobTracker Logs',
      url:'%@://%@:%@/logs',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/logs',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:11,
      label:'JobTracker JMX',
      url:'%@://%@:%@/jmx',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/jmx',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:12,
      label:'Thread Stacks',
      url:'%@://%@:%@/stacks',
      service_id: 'MAPREDUCE',
      template:'%@://%@:%@/stacks',
      http_config: 'mapred.job.tracker.http.address',
      https_config: 'mapred.job.tracker.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 50030,
      default_https_port: 50035
    },
    {
      id:13,
      label:'HBase Master UI',
      url:'%@://%@:%@/master-status',
      service_id: 'HBASE',
      template:'%@://%@:%@/master-status',
      http_config: 'hbase.master.info.port',
      site: 'hbase-site',
      regex: '^(\\d+)$',
      default_http_port: 60010
    },
    {
      id:14,
      label:'HBase Logs',
      url:'%@://%@:60010/logs',
      service_id: 'HBASE',
      template:'%@://%@:%@/logs',
      http_config: 'hbase.master.info.port',
      site: 'hbase-site',
      regex: '^(\\d+)$',
      default_http_port: 60010
    },
    {
      id:15,
      label:'Zookeeper Info',
      url:'%@://%@:60010/zk.jsp',
      service_id: 'HBASE',
      template:'%@://%@:%@/zk.jsp',
      http_config: 'hbase.master.info.port',
      site: 'hbase-site',
      regex: '^(\\d+)$',
      default_http_port: 60010
    },
    {
      id:16,
      label:'HBase Master JMX',
      url:'%@://%@:60010/jmx',
      service_id: 'HBASE',
      template:'%@://%@:%@/jmx',
      http_config: 'hbase.master.info.port',
      site: 'hbase-site',
      regex: '^(\\d+)$',
      default_http_port: 60010
    },
    {
      id:17,
      label:'Debug Dump',
      url:'%@://%@:%@/dump',
      service_id: 'HBASE',
      template:'%@://%@:%@/dump',
      http_config: 'hbase.master.info.port',
      site: 'hbase-site',
      regex: '^(\\d+)$',
      default_http_port: 60010
    },
    {
      id:18,
      label:'Thread Stacks',
      url:'%@://%@:%@/stacks',
      service_id: 'HBASE',
      template:'%@://%@:%@/stacks',
      http_config: 'hbase.master.info.port',
      site: 'hbase-site',
      regex: '^(\\d+)$',
      default_http_port: 60010
    },
    {
      id:19,
      label:'Oozie Web UI',
      url:'%@://%@:%@/oozie',
      service_id: 'OOZIE',
      template:'%@://%@:%@/oozie',
      http_config: 'oozie.base.url',
      site: 'oozie-site',
      regex: portRegex,
      default_http_port: 11000
    },
    {
      id:20,
      label:'Ganglia Web UI',
      url:'%@://%@/ganglia',
      service_id: 'GANGLIA',
      template:'%@://%@/ganglia'

    },
    {
      id:21,
      label:'Nagios Web UI',
      url:'%@://%@/nagios',
      service_id: 'NAGIOS',
      template:'%@://%@/nagios'
    },
    {
      id:22,
      label:'Hue Web UI',
      url:'%@://%@/hue',
      service_id: 'HUE',
      template:'%@://%@/hue'
    },
    {
      id:23,
      label:'ResourceManager UI',
      url:'%@://%@:%@',
      service_id: 'YARN',
      template:'%@://%@:%@',
      http_config: 'yarn.resourcemanager.webapp.address',
      https_config: 'yarn.resourcemanager.webapp.https.address',
      site: 'yarn-site',
      regex: portRegex,
      default_http_port: 8088,
      default_https_port: 8090

    },
    {
      id:24,
      label:'ResourceManager logs',
      url:'%@://%@:%@/logs',
      service_id: 'YARN',
      template:'%@://%@:%@/logs',
      http_config: 'yarn.resourcemanager.webapp.address',
      https_config: 'yarn.resourcemanager.webapp.https.address',
      site: 'yarn-site',
      regex: portRegex,
      default_http_port: 8088,
      default_https_port: 8090
    },
    {
      id:25,
      label:'ResourceManager JMX',
      url:'%@://%@:%@/jmx',
      service_id: 'YARN',
      template:'%@://%@:%@/jmx',
      http_config: 'yarn.resourcemanager.webapp.address',
      https_config: 'yarn.resourcemanager.webapp.https.address',
      site: 'yarn-site',
      regex: portRegex,
      default_http_port: 8088,
      default_https_port: 8090
    },
    {
      id:26,
      label:'Thread Stacks',
      url:'%@://%@:%@/stacks',
      service_id: 'YARN',
      template:'%@://%@:%@/stacks',
      http_config: 'yarn.resourcemanager.webapp.address',
      https_config: 'yarn.resourcemanager.webapp.https.address',
      site: 'yarn-site',
      regex: portRegex,
      default_http_port: 8088,
      default_https_port: 8090
    },
    {
      id:27,
      label:'JobHistory UI',
      url:'%@://%@:%@',
      service_id: 'MAPREDUCE2',
      template:'%@://%@:%@',
      http_config: 'mapreduce.jobhistory.webapp.address',
      https_config: 'mapreduce.jobhistory.webapp.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 19888
    },
    {
      id:28,
      label:'JobHistory logs',
      url:'%@://%@:%@/logs',
      service_id: 'MAPREDUCE2',
      template:'%@://%@:%@/logs',
      http_config: 'mapreduce.jobhistory.webapp.address',
      https_config: 'mapreduce.jobhistory.webapp.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 19888
    },
    {
      id:29,
      label:'JobHistory JMX',
      url:'%@://%@:%@/jmx',
      service_id: 'MAPREDUCE2',
      template:'%@://%@:%@/jmx',
      http_config: 'mapreduce.jobhistory.webapp.address',
      https_config: 'mapreduce.jobhistory.webapp.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 19888
    },
    {
      id:30,
      label:'Thread Stacks',
      url:'%@://%@:%@/stacks',
      service_id: 'MAPREDUCE2',
      template:'%@://%@:%@/stacks',
      http_config: 'mapreduce.jobhistory.webapp.address',
      https_config: 'mapreduce.jobhistory.webapp.https.address',
      site: 'mapred-site',
      regex: portRegex,
      default_http_port: 19888
    },
    {
      id:31,
      label:'Storm UI',
      url:'%@://%@:%@/',
      service_id: 'STORM',
      template:'%@://%@:%@/',
      http_config: 'nimbus.host',
      https_config: 'nimbus.host',
      site: 'storm-site',
      regex: portRegex,
      default_http_port: 8744
    },
    {
      id:32,
      label:'Falcon Web UI',
      url:'%@://%@:%@/',
      service_id: 'FALCON',
      template:'%@://%@:%@/',
      http_config: 'falcon_port',
      site: 'global',
      regex: '^(\\d+)$',
      default_http_port: 15000
    }
  ];
  
});
window.require.register("models/rack", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.Rack = DS.Model.extend({
    name: DS.attr('string'),
    status: DS.attr('string'),
    criticalHostsCount: DS.attr('number'),
    deadHostsCount: DS.attr('number'),
    hosts: function(){
      return App.Host.find();
    }.property('name'),
    liveHostsCount: function(){
      var count = 0;
      this.get('hosts').forEach(function(host){
        if(host.get('healthStatus')=="HEALTHY"){
          count++;
        }
      });
      return count;
    }.property('hosts')
  });

  App.Rack.FIXTURES = [
    //here example of data
    /*{
      id: 1,
      name: 'Rack-0',
      hosts: ['host01', 'host06', 'host05'],
      status: 'LIVE',
      live_hosts_count: 5,
      critical_hosts_count: 0,
      dead_hosts_count: 2
    }*/
  ];
});
window.require.register("models/run", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var date = require('utils/date');
  var misc = require('utils/misc');

  App.Run = DS.Model.extend({
    appName: DS.attr('string'),
    userName:DS.attr('string'),
    numJobsTotal: DS.attr('number'),
    numJobsCompleted: DS.attr('number'),
    startTime:DS.attr('string'),
    elapsedTime:DS.attr('string'),
    workflowContext:DS.attr('string'),
    input: DS.attr('number'),
    output: DS.attr('number'),

    loadAllJobs : false,

    isStared: false,
    isFiltered: false,

    /**
     * runId  short part
     */
    idFormatted: function() {
      return this.get('id').substr(0, 20);
    }.property('id'),

    /**
     * Jobs in the current run
     */
    jobs: function() {
      return App.Job.find().filterProperty('run.id', this.get('id'));
    }.property('loadAllJobs'),

    /**
     * Run duration
     */
    duration: function() {
      return date.timingFormat(parseInt(this.get('elapsedTime')));
    }.property('elapsedTime'),
    /**
     * Status of running jobs
     */
    isRunning: function () {
      return !this.get('numJobsTotal') == this.get('numJobsCompleted');
    }.property('numJobsTotal', 'numJobsCompleted'),
    /**
     * Sum of input bandwidth for all jobs with appropriate measure
     */
    inputFormatted: function () {
      var input = this.get('input');
      input = misc.formatBandwidth(input);
      return input;
    }.property('input'),

    /**
     *  Sum of output bandwidth for all jobs with appropriate measure
     */
    outputFormatted: function () {
      var output = this.get('output');
      output = misc.formatBandwidth(output);
      return output;
    }.property('output'),

    /**
     *
     */
    lastUpdateTime: function() {
      return parseInt(this.get('startTime')) + parseInt(this.get('elapsedTime'));
    }.property('elapsedTime', 'startTime'),
    /**
     *
     */
    lastUpdateTimeFormatted: function() {
      return date.dateFormat(this.get('lastUpdateTime'));
    }.property('lastUpdateTime'),
    lastUpdateTimeFormattedShort: function(){
      return date.dateFormatShort(this.get('lastUpdateTime'));
    }.property('lastUpdateTime'),
    /**
     * Type value based on first part of id
     */
    type: function() {
      if (this.get('id').indexOf('pig_') === 0) {
        return 'Pig';
      }
      if (this.get('id').indexOf('hive_') === 0) {
        return 'Hive';
      }
      if (this.get('id').indexOf('mr_') === 0) {
        return 'MapReduce';
      }
      return '';
    }.property('id')
  });

  App.Run.FIXTURES = [];
  
});
window.require.register("models/service", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  require('utils/config');

  App.Service = DS.Model.extend({

    serviceName: DS.attr('string'),
    passiveState: DS.attr('string'),
    workStatus: DS.attr('string'),
    rand: DS.attr('string'),
    toolTipContent: DS.attr('string'),
    criticalAlertsCount: DS.attr('number'),
    quickLinks: DS.hasMany('App.QuickLinks'),  // mapped in app/mappers/service_metrics_mapper.js method - mapQuickLinks
    hostComponents: DS.hasMany('App.HostComponent'),
    serviceConfigsTemplate: App.config.get('preDefinedServiceConfigs'),

    /**
     * @type {bool}
     */
    isInPassive: function() {
      return this.get('passiveState') === "ON";
    }.property('passiveState'),

    // Instead of making healthStatus a computed property that listens on hostComponents.@each.workStatus,
    // we are creating a separate observer _updateHealthStatus.  This is so that healthStatus is updated
    // only once after the run loop.  This is because Ember invokes the computed property every time
    // a property that it depends on changes.  For example, App.statusMapper's map function would invoke
    // the computed property too many times and freezes the UI without this hack.
    // See http://stackoverflow.com/questions/12467345/ember-js-collapsing-deferring-expensive-observers-or-computed-properties
    healthStatus: function(){
      switch(this.get('workStatus')){
        case 'STARTED':
          return 'green';
        case 'STARTING':
          return 'green-blinking';
        case 'INSTALLED':
          return 'red';
        case 'STOPPING':
          return 'red-blinking';
        case 'UNKNOWN':
        default:
          return 'yellow';
      }
    }.property('workStatus'),
    isStopped: function () {
      return this.get('workStatus') === 'INSTALLED';
    }.property('workStatus'),
    isStarted: function () {
      return this.get('workStatus') === 'STARTED';
    }.property('workStatus'),

    isClientsOnly: function() {
      var clientsOnly = ['SQOOP','PIG','TEZ','HCATALOG'];
      return clientsOnly.contains(this.get('serviceName'));
    }.property('serviceName'),

    isConfigurable: function () {
      var configurableServices = [
        "HDFS",
        "YARN",
        "MAPREDUCE",
        "MAPREDUCE2",
        "HBASE",
        "OOZIE",
        "HIVE",
        "WEBHCAT",
        "ZOOKEEPER",
        "PIG",
        "NAGIOS",
        "GANGLIA",
        "HUE",
        "TEZ",
        "STORM",
        "FALCON"
      ];
      return configurableServices.contains(this.get('serviceName'));
    }.property('serviceName'),

    displayName: function () {
      return App.Service.DisplayNames[this.get('serviceName')];
    }.property('serviceName'),

    /**
     * For each host-component, if the desired_configs dont match the
     * actual_configs, then a restart is required. Except for Global site
     * properties, which need to be checked with map.
     */
    isRestartRequired: function () {
      var rhc = this.get('hostComponents').filterProperty('staleConfigs', true);

      // HCatalog components are technically owned by Hive.
      if (this.get('serviceName') == 'HIVE') {
        var hcatService = App.Service.find('HCATALOG');
        if (hcatService != null && hcatService.get('isLoaded')) {
          var hcatStaleHcs = hcatService.get('hostComponents').filterProperty('staleConfigs', true);
          if (hcatStaleHcs != null) {
            rhc.pushObjects(hcatStaleHcs);
          }
        }
      }

      var hc = {};
      rhc.forEach(function(_rhc) {
        var hostName = _rhc.get('host.publicHostName');
        if (!hc[hostName]) {
          hc[hostName] = [];
        }
        hc[hostName].push(_rhc.get('displayName'));
      });
      this.set('restartRequiredHostsAndComponents', hc);
      return (rhc.length>0);

    }.property('serviceName', 'hostComponents.@each.staleConfigs'),
    
    /**
     * Contains a map of which hosts and host_components
     * need a restart. This is populated when calculating
     * #isRestartRequired()
     * Example:
     * {
     *  'publicHostName1': ['TaskTracker'],
     *  'publicHostName2': ['JobTracker', 'TaskTracker']
     * }
     */
    restartRequiredHostsAndComponents: {},
    
    /**
     * Based on the information in #restartRequiredHostsAndComponents
     */
    restartRequiredMessage: function () {
      var restartHC = this.get('restartRequiredHostsAndComponents');
      var hostCount = 0;
      var hcCount = 0;
      var hostsMsg = "<ul>";
      for(var host in restartHC){
        hostCount++;
        hostsMsg += "<li>"+host+"</li><ul>";
        restartHC[host].forEach(function(c){
          hcCount++;
          hostsMsg += "<li>"+c+"</li>";       
        });
        hostsMsg += "</ul>";
      }
      hostsMsg += "</ul>";
      return this.t('services.service.config.restartService.TooltipMessage').format(hcCount, hostCount, hostsMsg);
    }.property('restartRequiredHostsAndComponents')
  });

  App.Service.Health = {
    live: "LIVE",
    dead: "DEAD-RED",
    starting: "STARTING",
    stopping: "STOPPING",
    unknown: "DEAD-YELLOW",

    getKeyName: function (value) {
      switch (value) {
        case this.live:
          return 'live';
        case this.dead:
          return 'dead';
        case this.starting:
          return 'starting';
        case this.stopping:
          return 'stopping';
        case this.unknown:
          return 'unknown';
      }
      return 'none';
    }
  };

  App.Service.DisplayNames = {
    'HDFS': 'HDFS',
    'YARN': 'YARN',
    'MAPREDUCE': 'MapReduce',
    'MAPREDUCE2': 'MapReduce2',
    'TEZ': 'Tez',
    'HBASE': 'HBase',
    'OOZIE': 'Oozie',
    'HIVE': 'Hive',
    'HCATALOG': 'HCat',
    'ZOOKEEPER': 'ZooKeeper',
    'PIG': 'Pig',
    'SQOOP': 'Sqoop',
    'WEBHCAT': 'WebHCat',
    'GANGLIA': 'Ganglia',
    'NAGIOS': 'Nagios',
    'HUE': 'Hue',
    'FLUME': 'Flume',
    'FALCON': 'Falcon',
    'STORM': 'Storm'
  };

  App.Service.servicesSortOrder = [
    'HDFS',
    'YARN',
    'MAPREDUCE',
    'MAPREDUCE2',
    'TEZ',
    'HBASE',
    'HIVE',
    'HCATALOG',
    'WEBHCAT',
    'FLUME',
    'FALCON',
    'STORM',
    'OOZIE',
    'GANGLIA',
    'NAGIOS',
    'ZOOKEEPER',
    'PIG',
    'SQOOP',
    'HUE'
  ];

  App.Service.FIXTURES = [];
  
});
window.require.register("models/service/flume", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.FlumeService = App.Service.extend({
    version: DS.attr('string'),
    nodes: DS.hasMany('App.FlumeNode'),
    channelsCount: function () {
      var nodes = this.get('nodes');
      var count = 0;
      nodes.forEach(function (node) {
        count += node.get('channelsCount');
      });
      return count;
    }.property('nodes.@each.channelsCount'),

    sourcesCount: function () {
      var nodes = this.get('nodes');
      var count = 0;
      nodes.forEach(function (node) {
        count += node.get('sourcesCount');
      });
      return count;
    }.property('nodes.@each.sourcesCount'),

    sinksCount: function () {
      var nodes = this.get('nodes');
      var count = 0;
      nodes.forEach(function (node) {
        count += node.get('sinksCount');
      });
      return count;
    }.property('nodes.@each.sinksCount')
  });

  App.FlumeNode = DS.Model.extend({
    host: DS.belongsTo('App.Host'),

    /**
     * A comma separated list of channels.
     */
    channels: DS.attr('string'),

    /**
     * A comma separated list of sources.
     */
    sources: DS.attr('string'),

    /**
     * A comma separated list of sinks.
     */
    sinks: DS.attr('string'),

    channelsCount: function () {
      var channels = this.get('channels');
      if (!channels) {
        return 0;
      } else {
        return channels.split(',').length;
      }
    }.property('channels'),

    sourcesCount: function () {
      var sources = this.get('sources');
      if (!sources) {
        return 0;
      } else {
        return sources.split(',').length;
      }
    }.property('sources'),

    sinksCount: function () {
      var sinks = this.get('sinks');
      if (!sinks) {
        return 0;
      } else {
        return sinks.split(',').length;
      }
    }.property('sinks')
  });

  App.FlumeService.FIXTURES = [];
  App.FlumeNode.FIXTURES = [];
  
});
window.require.register("models/service/hbase", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.HBaseService = App.Service.extend({
    version: DS.attr('string'),
    master: DS.belongsTo('App.Host'),
    regionServers: function () {
      return this.get('hostComponents').filterProperty('componentName', 'HBASE_REGIONSERVER');
    }.property('hostComponents.@each'),
    masterStartTime: DS.attr('number'),
    masterActiveTime: DS.attr('number'),
    averageLoad: DS.attr('number'),
    regionsInTransition: DS.attr('number'),
    revision: DS.attr('string'),
    heapMemoryUsed: DS.attr('number'),
    heapMemoryMax: DS.attr('number')
  });

  App.HBaseService.FIXTURES = [];
  
});
window.require.register("models/service/hdfs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.HDFSService = App.Service.extend({
    version: DS.attr('string'),
    nameNode: DS.belongsTo('App.Host'),
    snameNode: DS.belongsTo('App.Host'),
    activeNameNode: DS.belongsTo('App.Host'),
    standbyNameNode: DS.belongsTo('App.Host'),
    standbyNameNode2: DS.belongsTo('App.Host'),
    dataNodes: function () {
      return this.get('hostComponents').filterProperty('componentName', 'DATANODE');
    }.property('hostComponents.@each'),
    journalNodes: function () {
      return this.get('hostComponents').filterProperty('componentName', 'JOURNALNODE');
    }.property('hostComponents.@each'),
    nameNodeStartTime: DS.attr('number'),
    jvmMemoryHeapUsed: DS.attr('number'),
    jvmMemoryHeapMax: DS.attr('number'),
    decommissionDataNodes: DS.hasMany('App.Host'),
    capacityUsed: DS.attr('number'),
    capacityTotal: DS.attr('number'),
    capacityRemaining: DS.attr('number'),
    dfsTotalBlocks: DS.attr('number'),
    dfsCorruptBlocks: DS.attr('number'),
    dfsMissingBlocks: DS.attr('number'),
    dfsUnderReplicatedBlocks: DS.attr('number'),
    dfsTotalFiles: DS.attr('number'),
    upgradeStatus: DS.attr('boolean'),
    safeModeStatus: DS.attr('string'),
    nameNodeCpu: DS.attr('number'),
    nameNodeRpc: DS.attr('number')
  });

  App.HDFSService.FIXTURES = [];
  
});
window.require.register("models/service/mapreduce", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MapReduceService = App.Service.extend({
    version: DS.attr('string'),
    jobHistoryServer: DS.belongsTo('App.Host'),
    jobTracker: DS.belongsTo('App.Host'),
    taskTrackers: function () {
      return this.get('hostComponents').filterProperty('componentName', 'TASKTRACKER');
    }.property('hostComponents.@each'),
    jobTrackerStartTime: DS.attr('number'),
    jobTrackerHeapUsed: DS.attr('number'),
    jobTrackerHeapMax: DS.attr('number'),
    aliveTrackers: DS.hasMany('App.Host'),
    blackListTrackers: DS.hasMany('App.Host'),
    grayListTrackers: DS.hasMany('App.Host'),
    mapSlots: DS.attr('number'),
    reduceSlots: DS.attr('number'),
    jobsSubmitted: DS.attr('number'),
    jobsCompleted: DS.attr('number'),
    jobsRunning: DS.attr('number'),
    mapSlotsOccupied: DS.attr('number'),
    mapSlotsReserved: DS.attr('number'),
    reduceSlotsOccupied: DS.attr('number'),
    reduceSlotsReserved: DS.attr('number'),
    mapsRunning: DS.attr('number'),
    mapsWaiting: DS.attr('number'),
    reducesRunning: DS.attr('number'),
    reducesWaiting: DS.attr('number'),
    trackersDecommissioned: DS.attr('number'),
    jobTrackerCpu: DS.attr('number'),
    jobTrackerRpc: DS.attr('number')
  });

  App.MapReduceService.FIXTURES = [];
  
});
window.require.register("models/service/mapreduce2", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MapReduce2Service = App.Service.extend({
    version: DS.attr('string'),
    jobHistoryServer: DS.belongsTo('App.Host'),
    mapReduce2Clients: function(){
      return this.get('hostComponents').filterProperty('componentName', 'MAPREDUCE2_CLIENT').mapProperty('host');
    }.property('hostComponents.length')
  });

  App.MapReduce2Service.FIXTURES = [];
  
});
window.require.register("models/service/storm", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.StormService = App.Service.extend({
    version: DS.attr('string'),
    totalTasks: DS.attr('number'),
    totalSlots: DS.attr('number'),
    usedSlots: DS.attr('number'),
    freeSlots: DS.attr('number'),
    totalExecutors: DS.attr('number'),
    topologies: DS.attr('number'),
    nimbusUptime: DS.attr('number')
  });

  App.StormService.FIXTURES = [];
  
});
window.require.register("models/service/yarn", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var objectUtils = require('utils/object_utils');

  App.YARNService = App.Service.extend({
    version: DS.attr('string'),
    resourceManagerNode: DS.belongsTo('App.Host'),
    appTimelineServerNode: function() {
      return this.get('hostComponents').filterProperty('componentName', 'APP_TIMELINE_SERVER').mapProperty('host').objectAt(0);
    }.property(),
    nodeManagerNodes: function () {
      return this.get('hostComponents').filterProperty('componentName', 'NODEMANAGER');
    }.property('hostComponents.@each'),
    nodeManagersCountActive: DS.attr('number'),
    nodeManagersCountUnhealthy: DS.attr('number'),
    nodeManagersCountRebooted: DS.attr('number'),
    nodeManagersCountDecommissioned: DS.attr('number'),
    containersAllocated: DS.attr('number'),
    containersPending: DS.attr('number'),
    containersReserved: DS.attr('number'),
    appsSubmitted: DS.attr('number'),
    appsRunning: DS.attr('number'),
    appsPending: DS.attr('number'),
    appsCompleted: DS.attr('number'),
    appsKilled: DS.attr('number'),
    appsFailed: DS.attr('number'),
    ahsWebPort: function() {
      var yarnConf = App.db.getConfigs().findProperty('type', 'yarn-site')
      if(yarnConf){
        return yarnConf.properties['yarn.timeline-service.webapp.address'].match(/:(\d+)/)[1];;
      }
      return "8188";
    }.property(),
    yarnClientNodes: function(){
      return this.get('hostComponents').filterProperty('componentName', 'YARN_CLIENT').mapProperty('host');
    }.property('hostComponents.length'),
    resourceManagerStartTime: DS.attr('number'),
    jvmMemoryHeapUsed: DS.attr('number'),
    jvmMemoryHeapMax: DS.attr('number'),
    allocatedMemory: DS.attr('number'),
    reservedMemory: DS.attr('number'),
    availableMemory: DS.attr('number'),
    queue: DS.attr('string'),
    queueFormatted: function() {
      var queue = JSON.parse(this.get('queue'));
      return objectUtils.recursiveTree(queue);
    }.property('queue'),
    queuesCount: function() {
      var queue = JSON.parse(this.get('queue'));
      return objectUtils.recursiveKeysCount(queue);
    }.property('queue'),
    allQueueNames: [],
    childQueueNames: [],
    /** 
     * Provides a flat array of queue names.
     * Example: root, root/default
     */
    maxMemory: function() {
      return this.get('allocatedMemory') + this.get('availableMemory');
    }.property('allocatedMemory','availableMemory'),

    queueNames: function () {
      var queueString = this.get('queue');
      var allQueueNames = [];
      var childQueueNames = [];
      if (queueString != null) {
        var queues = JSON.parse(queueString);
        var addQueues = function (queuesObj, path){
          var names = [];
          for ( var subQueue in queuesObj) {
            if (queuesObj[subQueue] instanceof Object) {
              var qFN = path=='' ? subQueue : path+'/'+subQueue;
              names.push(qFN);
              var subNames = addQueues(queuesObj[subQueue], qFN);
              names = names.concat(subNames);
              if (!subNames || subNames.length < 1) {
                childQueueNames.push(qFN);
              }
            }
          }
          return names;
        };
        allQueueNames = addQueues(queues, '');
      }
      this.set('allQueueNames', allQueueNames);
      this.set('childQueueNames', childQueueNames);
    }.observes('queue'),
    /**
     * ResourceManager's lost count is not accurate once RM is rebooted. Since
     * Ambari knows the total number of nodes and the counts of nodes in other
     * states, we calculate the lost count.
     */
    nodeManagersCountLost: function () {
      var allNMs = this.get('nodeManagerNodes');
      var totalCount = allNMs != null ? allNMs.get('length') : 0;
      var activeCount = this.get('nodeManagersCountActive');
      var rebootedCount = this.get('nodeManagersCountRebooted');
      var unhealthyCount = this.get('nodeManagersCountUnhealthy');
      var decomCount = this.get('nodeManagersCountDecommissioned');
      var nonLostHostsCount = activeCount + rebootedCount + decomCount + unhealthyCount;
      return totalCount >= nonLostHostsCount ? totalCount - nonLostHostsCount : 0;
    }.property('nodeManagerNodes', 'nodeManagersCountActive', 'nodeManagersCountRebooted', 'nodeManagersCountUnhealthy', 'nodeManagersCountDecommissioned')
  });

  App.YARNService.FIXTURES = [];
  
});
window.require.register("models/service_audit", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ServiceAudit = DS.Model.extend({
    date: DS.attr('date'),
    service: DS.belongsTo('App.Service'),
    operationName: DS.attr('string'),
    user: DS.belongsTo('App.User')
  });

  App.ServiceAudit.FIXTURES = [
    {
      id: 1,
      date: 'September 12, 2012 17:00',
      operation_name: 'Reconfigure',
      user_id: 2,
      service_id: 1
    },
    {
      id: 2,
      date: 'September 13, 2012 17:00',
      operation_name: 'Start',
      user_id: 1,
      service_id: 1
    },
    {
      id: 3,
      date: 'September 14, 2012 17:00',
      operation_name: 'Install',
      user_id: 1,
      service_id: 1
    },
    {
      id: 4,
      date: 'September 12, 2012 17:00',
      operation_name: 'Reconfigure',
      user_id: 2,
      service_id: 2
    },
    {
      id: 5,
      date: 'September 13, 2012 17:00',
      operation_name: 'Start',
      user_id: 1,
      service_id: 2
    },
    {
      id: 6,
      date: 'September 14, 2012 17:00',
      operation_name: 'Install',
      user_id: 1,
      service_id: 2
    }
  ];
});
window.require.register("models/service_config", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');

  App.ServiceConfig = Ember.Object.extend({
    serviceName: '',
    configCategories: [],
    configs: null,
    restartRequired: false,
    restartRequiredMessage: '',
    restartRequiredHostsAndComponents: {},
    configGroups: [],
    errorCount: function () {
      var overrideErrors = 0;
      this.get('configs').filterProperty("overrides").forEach(function (e) {
        e.overrides.forEach(function (e) {
          if (e.error) {
            overrideErrors += 1;
          }
        })
      });
      var masterErrors = this.get('configs').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
      var slaveErrors = 0;
      this.get('configCategories').forEach(function (_category) {
        slaveErrors += _category.get('slaveErrorCount');
      }, this);
      return masterErrors + slaveErrors + overrideErrors;
    }.property('configs.@each.isValid', 'configs.@each.isVisible', 'configCategories.@each.slaveErrorCount', 'configs.@each.overrideErrorTrigger')
  });

  App.ServiceConfigCategory = Ember.Object.extend({
    name: null,
    /**
     *  We cant have spaces in the name as this is being used as HTML element id while rendering. Hence we introduced 'displayName' where we can have spaces like 'Secondary Name Node' etc.
     */
    displayName: null,
    slaveConfigs: null,
    /**
     * check whether to show custom view in category instead of default
     */
    isCustomView: false,
    customView: null,
    /**
     * Each category might have a site-name associated (hdfs-site, core-site, etc.)
     * and this will be used when determining which category a particular property
     * ends up in, based on its site.
     */
    siteFileName: null,
    /**
     * Can this category add new properties. Used for custom configurations.
     */
    canAddProperty: false,
    primaryName: function () {
      switch (this.get('name')) {
        case 'DataNode':
          return 'DATANODE';
          break;
        case 'TaskTracker':
          return 'TASKTRACKER';
          break;
        case 'RegionServer':
          return 'HBASE_REGIONSERVER';
      }
      return null;
    }.property('name'),


    isForMasterComponent: function () {
      var masterServices = [ 'NameNode', 'SNameNode', 'JobTracker', 'HBase Master', 'Oozie Master',
        'Hive Metastore', 'WebHCat Server', 'ZooKeeper Server', 'Nagios', 'Ganglia' ];

      return (masterServices.contains(this.get('name')));
    }.property('name'),

    isForSlaveComponent: function () {
      var slaveComponents = ['DataNode', 'TaskTracker', 'RegionServer'];
      return (slaveComponents.contains(this.get('name')));
    }.property('name'),

    slaveErrorCount: function () {
      var length = 0;
      if (this.get('slaveConfigs.groups')) {
        this.get('slaveConfigs.groups').forEach(function (_group) {
          length += _group.get('errorCount');
        }, this);
      }
      return length;
    }.property('slaveConfigs.groups.@each.errorCount'),

    isAdvanced : function(){
      var name = this.get('name');
      return name.indexOf("Advanced") !== -1 ;
    }.property('name')
  });


  App.SlaveConfigs = Ember.Object.extend({
    componentName: null,
    displayName: null,
    hosts: null,
    groups: null
  });

  App.Group = Ember.Object.extend({
    name: null,
    hostNames: null,
    properties: null,
    errorCount: function () {
      if (this.get('properties')) {
        return this.get('properties').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
      }
    }.property('properties.@each.isValid', 'properties.@each.isVisible')
  });


  App.ServiceConfigProperty = Ember.Object.extend({

    id: '', //either 'puppet var' or 'site property'
    name: '',
    displayName: '',
    value: '',
    retypedPassword: '',
    defaultValue: '',
    defaultDirectory: '',
    description: '',
    displayType: 'string', // string, digits, number, directories, custom
    unit: '',
    category: 'General',
    isRequired: true, // by default a config property is required
    isReconfigurable: true, // by default a config property is reconfigurable
    isEditable: true, // by default a config property is editable
    isVisible: true,
    isRequiredByAgent: true, // Setting it to true implies property will be stored in global configuration
    isSecureConfig: false,
    errorMessage: '',
    warnMessage: '',
    serviceConfig: null, // points to the parent App.ServiceConfig object
    filename: '',
    isOriginalSCP : true, // if true, then this is original SCP instance and its value is not overridden value.
    parentSCP: null, // This is the main SCP which is overridden by this. Set only when isOriginalSCP is false.
    selectedHostOptions : null, // contain array of hosts configured with overridden value
    overrides : null,
    group: null, // Contain group related to this property. Set only when isOriginalSCP is false.
    isUserProperty: null, // This property was added by user. Hence they get removal actions etc.
    isOverridable: true,
    showLabel: true,
    error: false,
    warn: false,
    overrideErrorTrigger: 0, //Trigger for overrridable property error
    isRestartRequired: false,
    restartRequiredMessage: 'Restart required',
    index: null, //sequence number in category
    editDone: false, //Text field: on focusOut: true, on focusIn: false
    serviceValidator: null,
    isNotSaved: false, // user property was added but not saved

    /**
     * On Overridable property error message, change overrideErrorTrigger value to recount number of errors service have
     */
    observeErrors: function () {
      this.set("overrideErrorTrigger", this.get("overrideErrorTrigger") + 1);
    }.observes("overrides.@each.errorMessage"),
    /**
     * No override capabilities for fields which are not edtiable
     * and fields which represent master hosts.
     */
    isPropertyOverridable: function () {
      var overrideable = this.get('isOverridable');
      var editable = this.get('isEditable');
      var overrides = this.get('overrides');
      var dt = this.get('displayType');
      return overrideable && (editable || !overrides || !overrides.length) && ("masterHost" != dt);
    }.property('isEditable', 'displayType', 'isOverridable', 'overrides.length'),
    isOverridden: function() {
      var overrides = this.get('overrides');
      return (overrides != null && overrides.get('length')>0) || !this.get('isOriginalSCP');
    }.property('overrides', 'overrides.length', 'isOriginalSCP'),
    isRemovable: function() {
      var isOriginalSCP = this.get('isOriginalSCP');
      var isUserProperty = this.get('isUserProperty');
      var isEditable = this.get('isEditable');
      var hasOverrides = this.get('overrides.length') > 0;
      // Removable when this is a user property, or it is not an original property and it is editable
      return isEditable && !hasOverrides && (isUserProperty || !isOriginalSCP);
    }.property('isUserProperty', 'isOriginalSCP', 'overrides.length'),
    init: function () {
      if(this.get("displayType")=="password"){
        this.set('retypedPassword', this.get('value'));
      }
      if ((this.get('id') === 'puppet var') && this.get('value') == '') {
        this.set('value', this.get('defaultValue'));
      }
      // TODO: remove mock data
    },

    /**
     * Indicates when value is not the default value.
     * Returns false when there is no default value.
     */
    isNotDefaultValue: function () {
      var value = this.get('value');
      var dValue = this.get('defaultValue');
      var isEditable = this.get('isEditable');
      return isEditable && dValue != null && value !== dValue;
    }.property('value', 'defaultValue', 'isEditable'),

    /**
     * Don't show "Undo" for hosts on Installer Step7
     */
    cantBeUndone: function() {
      var types = ["masterHost", "slaveHosts", "masterHosts", "slaveHost","radio button"];
      var displayType = this.get('displayType');
      var result = false;
      types.forEach(function(type) {
        if (type === displayType) {
          result = true;
        }
      });
      return result;
    }.property('displayType'),

    initialValue: function (localDB) {
      var masterComponentHostsInDB = localDB.masterComponentHosts;
      //console.log("value in initialvalue: " + JSON.stringify(masterComponentHostsInDB));
      var hostsInfo = localDB.hosts; // which we are setting in installerController in step3.
      var slaveComponentHostsInDB = localDB.slaveComponentHosts;
      var isOnlyFirstOneNeeded = true;
      switch (this.get('name')) {
        case 'namenode_host':
          this.set('value', masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName'));
          break;
        case 'dfs.http.address':
          var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
          this.setDefaultValue("(\\w*)(?=:)",nnHost);
          break;
        case 'dfs.namenode.http-address':
          var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
          this.setDefaultValue("(\\w*)(?=:)",nnHost);
          break;
        case 'dfs.https.address':
          var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
          this.setDefaultValue("(\\w*)(?=:)",nnHost);
          break;
        case 'dfs.namenode.https-address':
          var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
          this.setDefaultValue("(\\w*)(?=:)",nnHost);
          break;
        case 'fs.default.name':
          var nnHost = masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName');
          this.setDefaultValue(":\/\/(\\w*)(?=:)",'://' + nnHost);
          break;
        case 'fs.defaultFS':
          var nnHost = masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName');
          this.setDefaultValue(":\/\/(\\w*)(?=:)",'://' + nnHost);
          break;
        case 'hbase.rootdir':
          var nnHost = masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName');
          this.setDefaultValue(":\/\/(\\w*)(?=:)",'://' + nnHost);
          break;
        case 'snamenode_host':
          // Secondary NameNode does not exist when NameNode HA is enabled
          var snn = masterComponentHostsInDB.findProperty('component', 'SECONDARY_NAMENODE');
          if (snn) {
            this.set('value', snn.hostName);
          }
          break;
        case 'dfs.secondary.http.address':
          var snnHost = masterComponentHostsInDB.findProperty('component', 'SECONDARY_NAMENODE');
          if (snnHost) {
            this.setDefaultValue("(\\w*)(?=:)",snnHost.hostName);
          }
          break;
        case 'dfs.namenode.secondary.http-address':
          var snnHost = masterComponentHostsInDB.findProperty('component', 'SECONDARY_NAMENODE');
          if (snnHost) {
            this.setDefaultValue("(\\w*)(?=:)",snnHost.hostName);
          }
          break;
        case 'datanode_hosts':
          this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'DATANODE').hosts.mapProperty('hostName'));
          break;
        case 'hs_host':
          this.set('value', masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName'));
          break;
        case 'yarn.log.server.url':
          var hsHost = masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName');
          this.setDefaultValue(":\/\/(\\w*)(?=:)",'://' + hsHost);
          break;
        case 'mapreduce.jobhistory.webapp.address':
          var hsHost = masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName');
          this.setDefaultValue("(\\w*)(?=:)",hsHost);
          break;
        case 'mapreduce.jobhistory.address':
          var hsHost = masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName');
          this.setDefaultValue("(\\w*)(?=:)",hsHost);
          break;
        case 'rm_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName);
          break;
        case 'ats_host':
          if (!App.supports.appTimelineServer) return; // @todo remove test mode check after App Timeline service integration
          this.set('value', masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER').hostName);
          break;
        case 'yarn.resourcemanager.hostname':
          var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
          this.set('defaultValue',rmHost);
          this.set('value',this.get('defaultValue'));
          break;
        case 'yarn.resourcemanager.resource-tracker.address':
          var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",rmHost);
          break;
        case 'yarn.resourcemanager.webapp.address':
          var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",rmHost);
          break;
        case 'yarn.resourcemanager.scheduler.address':
          var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",rmHost);
          break;
        case 'yarn.resourcemanager.address':
          var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",rmHost);
          break;
        case 'yarn.resourcemanager.admin.address':
          var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",rmHost);
          break;
        case 'yarn.timeline-service.webapp.address':
          var hsHost = masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER').hostName;
          this.setDefaultValue("(0.0.0.0)(?=:)", hsHost);
          break;
        case 'yarn.timeline-service.webapp.https.address':
          var hsHost = masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER').hostName;
          this.setDefaultValue("(0.0.0.0)(?=:)", hsHost);
          break;
        case 'nm_hosts':
          this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'NODEMANAGER').hosts.mapProperty('hostName'));
          break;
        case 'jobtracker_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'JOBTRACKER').hostName);
          break;
        case 'mapred.job.tracker':
          var jtHost = masterComponentHostsInDB.findProperty('component', 'JOBTRACKER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",jtHost);
          break;
        case 'mapred.job.tracker.http.address':
          var jtHost = masterComponentHostsInDB.findProperty('component', 'JOBTRACKER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",jtHost);
          break;
        case 'mapreduce.history.server.http.address':
          var jtHost = masterComponentHostsInDB.findProperty('component', 'HISTORYSERVER').hostName;
          this.setDefaultValue("(\\w*)(?=:)",jtHost);
          break;
        case 'tasktracker_hosts':
          this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'TASKTRACKER').hosts.mapProperty('hostName'));
          break;
        case 'hbasemaster_host':
          this.set('value', masterComponentHostsInDB.filterProperty('component', 'HBASE_MASTER').mapProperty('hostName'));
          break;
        case 'regionserver_hosts':
          this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'HBASE_REGIONSERVER').hosts.mapProperty('hostName'));
          break;
        case 'hivemetastore_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'HIVE_SERVER').hostName);
          break;
        case 'hive.metastore.uris':
          var hiveHost = masterComponentHostsInDB.findProperty('component', 'HIVE_SERVER').hostName;
          this.setDefaultValue(":\/\/(\\w*)(?=:)",'://' + hiveHost);
          break;
        case 'hive_ambari_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'HIVE_SERVER').hostName);
          break;
        case 'oozieserver_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName);
          break;
        case 'oozie.base.url':
          var oozieHost = masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName;
          this.setDefaultValue(":\/\/(\\w*)(?=:)",'://' + oozieHost);
          break;
        case 'webhcatserver_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'WEBHCAT_SERVER').hostName);
          break;
        case 'hueserver_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'HUE_SERVER').hostName);
          break;
        case 'oozie_ambari_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName);
          break;
        case 'storm.zookeeper.servers':
        case 'zookeeperserver_hosts':
          this.set('value', masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName'));
          break;
        case 'nimbus.host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'NIMBUS').hostName);
          break;
        case 'falconserver_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'FALCON_SERVER').hostName);
          break;
        case 'drpcserver_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'DRPC_SERVER').hostName);
          break;
        case 'stormuiserver_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'STORM_UI_SERVER').hostName);
          break;
        case 'storm_rest_api_host':
          this.set('value', masterComponentHostsInDB.findProperty('component', 'STORM_REST_API').hostName);
          break;
        case 'supervisor_hosts':
          this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'SUPERVISOR').hosts.mapProperty('hostName'));
          break;
        case 'hbase.zookeeper.quorum':
          var zkHosts = masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
          this.setDefaultValue("(\\w*)", zkHosts);
          break;
        case 'templeton.zookeeper.hosts':
          var zkHosts = masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
          var zkHostPort = zkHosts;
          var regex = "\\w*:(\\d+)";   //regex to fetch the port
          var portValue = this.get('defaultValue').match(new RegExp(regex));
          if (!portValue) return;
          if (portValue[1]) {
            for ( var i = 0; i < zkHosts.length; i++ ) {
              zkHostPort[i] = zkHosts[i] + ":" + portValue[1];
            }
          }
          this.setDefaultValue("(.*)", zkHostPort);
          break;
        case 'dfs.name.dir':
        case 'dfs.namenode.name.dir':
        case 'dfs.data.dir':
        case 'dfs.datanode.data.dir':
        case 'yarn.nodemanager.local-dirs':
        case 'yarn.nodemanager.log-dirs':
        case 'mapred.local.dir':
          this.unionAllMountPoints(!isOnlyFirstOneNeeded, localDB);
          break;
        case 'fs.checkpoint.dir':
        case 'dfs.namenode.checkpoint.dir':
        case 'zk_data_dir':
        case 'oozie_data_dir':
        case 'hbase.tmp.dir':
        case 'storm.local.dir':
          this.unionAllMountPoints(isOnlyFirstOneNeeded, localDB);
          break;
      }
    },

    /**
     * @param regex : String
     * @param replaceWith : String
     */
    setDefaultValue: function(regex,replaceWith) {
      var defaultValue = this.get('defaultValue');
      var re = new RegExp(regex);
      defaultValue = defaultValue.replace(re,replaceWith);
      this.set('defaultValue',defaultValue);
      this.set('value',this.get('defaultValue'));
    },

    unionAllMountPoints: function (isOnlyFirstOneNeeded, localDB) {
      var hostname = '';
      var mountPointsPerHost = [];
      var mountPointAsRoot;
      var masterComponentHostsInDB = localDB.masterComponentHosts;
      var slaveComponentHostsInDB = localDB.slaveComponentHosts;
      var hostsInfo = localDB.hosts; // which we are setting in installerController in step3.
      App.Host.find().forEach(function(item){
        if(!hostsInfo[item.get('id')]){
          hostsInfo[item.get('id')] = {
            name: item.get('id'),
            cpu: item.get('cpu'),
            memory: item.get('memory'),
            disk_info: item.get('diskInfo'),
            bootStatus: "REGISTERED",
            isInstalled: true
          };
        }
      });
      var temp = '';
      var setOfHostNames = [];
      var components = [];
      switch (this.get('name')) {
        case 'dfs.namenode.name.dir':
        case 'dfs.name.dir':
          components = masterComponentHostsInDB.filterProperty('component', 'NAMENODE');
          components.forEach(function (component) {
            setOfHostNames.push(component.hostName);
          }, this);
          break;
        case 'fs.checkpoint.dir':
        case 'dfs.namenode.checkpoint.dir':
          components = masterComponentHostsInDB.filterProperty('component', 'SECONDARY_NAMENODE');
          components.forEach(function (component) {
            setOfHostNames.push(component.hostName);
          }, this);
          break;
        case 'dfs.data.dir':
        case 'dfs.datanode.data.dir':
          temp = slaveComponentHostsInDB.findProperty('componentName', 'DATANODE');
          temp.hosts.forEach(function (host) {
            setOfHostNames.push(host.hostName);
          }, this);
          break;
        case 'mapred.local.dir':
          temp = slaveComponentHostsInDB.findProperty('componentName', 'TASKTRACKER') || slaveComponentHostsInDB.findProperty('componentName', 'NODEMANAGER');
          temp.hosts.forEach(function (host) {
            setOfHostNames.push(host.hostName);
          }, this);
          break;
        case 'yarn.nodemanager.log-dirs':
        case 'yarn.nodemanager.local-dirs':
          temp = slaveComponentHostsInDB.findProperty('componentName', 'NODEMANAGER');
          temp.hosts.forEach(function (host) {
            setOfHostNames.push(host.hostName);
          }, this);
          break;
        case 'zk_data_dir':
          components = masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER');
          components.forEach(function (component) {
            setOfHostNames.push(component.hostName);
          }, this);
          break;
        case 'oozie_data_dir':
          components = masterComponentHostsInDB.filterProperty('component', 'OOZIE_SERVER');
          components.forEach(function (component) {
            setOfHostNames.push(component.hostName);
          }, this);
          break;
        case 'hbase.tmp.dir':
          temp = slaveComponentHostsInDB.findProperty('componentName', 'HBASE_REGIONSERVER');
          temp.hosts.forEach(function (host) {
            setOfHostNames.push(host.hostName);
          }, this);
          break;
        case 'storm.local.dir':
          temp = slaveComponentHostsInDB.findProperty('componentName', 'SUPERVISOR');
          temp.hosts.forEach(function (host) {
            setOfHostNames.push(host.hostName);
          }, this);
          components = masterComponentHostsInDB.filterProperty('component', 'NIMBUS');
          components.forEach(function (component) {
            setOfHostNames.push(component.hostName);
          }, this);
          break;
      }

      // In Add Host Wizard, if we did not select this slave component for any host, then we don't process any further.
      if (setOfHostNames.length === 0) {
        return;
      }

      var allMountPoints = [];
      for (var i = 0; i < setOfHostNames.length; i++) {
        hostname = setOfHostNames[i];

        mountPointsPerHost = hostsInfo[hostname].disk_info;

        mountPointAsRoot = mountPointsPerHost.findProperty('mountpoint', '/');

        // If Server does not send any host details information then atleast one mountpoint should be presumed as root
        // This happens in a single container Linux Docker environment.
        if (!mountPointAsRoot) {
          mountPointAsRoot = {mountpoint: '/'};
        }

        mountPointsPerHost = mountPointsPerHost.filter(function (mPoint) {
          return !(['/', '/home', '/boot'].contains(mPoint.mountpoint) || ['devtmpfs', 'tmpfs', 'vboxsf'].contains(mPoint.type));
        });

        mountPointsPerHost.forEach(function (mPoint) {
          if( !allMountPoints.findProperty("mountpoint", mPoint.mountpoint)) {
            allMountPoints.push(mPoint);
          }
        }, this);
      }
      if (allMountPoints.length == 0) {
        allMountPoints.push(mountPointAsRoot);
      }
      this.set('value', '');
      if (!isOnlyFirstOneNeeded) {
        allMountPoints.forEach(function (eachDrive) {
          var mPoint = this.get('value');
          if (!mPoint) {
            mPoint = "";
          }
          if (eachDrive.mountpoint === "/") {
            mPoint += this.get('defaultDirectory') + "\n";
          } else {
            mPoint += eachDrive.mountpoint + this.get('defaultDirectory') + "\n";
          }
          this.set('value', mPoint);
          this.set('defaultValue', mPoint);
        }, this);
      } else {
        var mPoint = allMountPoints[0].mountpoint;
        if (mPoint === "/") {
          mPoint = this.get('defaultDirectory') + "\n";
        } else {
          mPoint = mPoint + this.get('defaultDirectory') + "\n";
        }
        this.set('value', mPoint);
        this.set('defaultValue', mPoint);
      }
    },

    isValid: function () {
      return this.get('errorMessage') === '';
    }.property('errorMessage'),

    viewClass: function () {
      switch (this.get('displayType')) {
        case 'checkbox':
          return App.ServiceConfigCheckbox;
        case 'password':
          return App.ServiceConfigPasswordField;
        case 'combobox':
          return App.ServiceConfigComboBox;
        case 'radio button':
          return App.ServiceConfigRadioButtons;
          break;
        case 'directories':
          return App.ServiceConfigTextArea;
          break;
        case 'content':
          return App.ServiceConfigTextAreaContent;
          break;
        case 'multiLine':
          return App.ServiceConfigTextArea;
          break;
        case 'custom':
          return App.ServiceConfigBigTextArea;
        case 'masterHost':
          return App.ServiceConfigMasterHostView;
        case 'masterHosts':
          return App.ServiceConfigMasterHostsView;
        case 'slaveHosts':
          return App.ServiceConfigSlaveHostsView;
        default:
          if (this.get('unit')) {
            return App.ServiceConfigTextFieldWithUnit;
          } else {
            return App.ServiceConfigTextField;
          }
      }
    }.property('displayType'),

    validate: function () {
      var value = this.get('value');
      var valueRange = this.get('valueRange');
      var values = [];//value split by "," to check UNIX users, groups list

      var isError = false;
      var isWarn = false;

      if (typeof value === 'string' && value.length === 0) {
        if (this.get('isRequired')) {
          this.set('errorMessage', 'This is required');
          isError = true;
        } else {
          return;
        }
      }

      if (!isError) {
        switch (this.get('displayType')) {
          case 'int':
            if (!validator.isValidInt(value)) {
              this.set('errorMessage', 'Must contain digits only');
              isError = true;
            } else {
              if(valueRange){
                if(value < valueRange[0] || value > valueRange[1]){
                  this.set('errorMessage', 'Must match the range');
                  isError = true;
                }
              }
            }
            break;
          case 'float':
            if (!validator.isValidFloat(value)) {
              this.set('errorMessage', 'Must be a valid number');
              isError = true;
            }
            break;
          case 'UNIXList':
            if(value != '*'){
              values = value.split(',');
              for(var i = 0, l = values.length; i < l; i++){
                if(!validator.isValidUNIXUser(values[i])){
                  if(this.get('type') == 'USERS'){
                    this.set('errorMessage', 'Must be a valid list of user names');
                  } else {
                    this.set('errorMessage', 'Must be a valid list of group names');
                  }
                  isError = true;
                }
              }
            }
            break;
          case 'checkbox':
            break;
          case 'directories':
          case 'directory':
            if (!validator.isValidDir(value)) {
              this.set('errorMessage', 'Must be a slash at the start');
              isError = true;
            }
            else {
              if (!validator.isAllowedDir(value)) {
                this.set('errorMessage', 'Can\'t start with "home(s)"');
                isError = true;
              }
            }
            break;
          case 'custom':
            break;
          case 'email':
            if (!validator.isValidEmail(value)) {
              this.set('errorMessage', 'Must be a valid email address');
              isError = true;
            }
            break;
          case 'host':
            var hiveOozieHostNames = ['hive_hostname','hive_existing_mysql_host','hive_existing_oracle_host','hive_ambari_host',
            'oozie_hostname','oozie_existing_mysql_host','oozie_existing_oracle_host','oozie_ambari_host'];
            if(hiveOozieHostNames.contains(this.get('name'))) {
              if (validator.hasSpaces(value)) {
                this.set('errorMessage', Em.I18n.t('host.spacesValidation'));
                isError = true;
              }
            } else {
              if (validator.isNotTrimmed(value)) {
                this.set('errorMessage', Em.I18n.t('host.trimspacesValidation'));
                isError = true;
              }
            }
            break;
          case 'advanced':
            if(this.get('name')=='javax.jdo.option.ConnectionURL' || this.get('name')=='oozie.service.JPAService.jdbc.url') {
              if (validator.isNotTrimmed(value)) {
                this.set('errorMessage', Em.I18n.t('host.trimspacesValidation'));
                isError = true;
              }
            }
            break;
          case 'password':
            // retypedPassword is set by the retypePasswordView child view of App.ServiceConfigPasswordField
            if (value !== this.get('retypedPassword')) {
              this.set('errorMessage', 'Passwords do not match');
              isError = true;
            }
        }
      }

      if (!isError) {
        // Check if this value is already in any of the overrides
        var self = this;
        var isOriginalSCP = this.get('isOriginalSCP');
        var parentSCP = this.get('parentSCP');
        if (!isOriginalSCP) {
          if (!isError && parentSCP != null) {
            if (value === parentSCP.get('value')) {
              this.set('errorMessage', 'Configuration overrides must have different value');
              isError = true;
            } else {
              var overrides = parentSCP.get('overrides');
              overrides.forEach(function (override) {
                if (self != override && value === override.get('value')) {
                  self.set('errorMessage', 'Multiple configuration overrides cannot have same value');
                  isError = true;
                }
              });
            }
          }
        }
      }
      
      var serviceValidator = this.get('serviceValidator');
      if (serviceValidator!=null) {
        var validationIssue = serviceValidator.validateConfig(this);
        if (validationIssue) {
      	this.set('warnMessage', validationIssue);
      	isWarn = true;
        }
      }

      if (!isWarn || isError) { // Errors get priority
          this.set('warnMessage', '');
          this.set('warn', false);
      } else {
          this.set('warn', true);
      }
      
      if (!isError) {
          this.set('errorMessage', '');
          this.set('error', false);
        } else {
          this.set('error', true);
        }
    }.observes('value', 'retypedPassword')

  });

  App.ConfigSiteTag = Ember.Object.extend({
    site: DS.attr('string'),
    tag: DS.attr('string'),
    /**
     * Object map of hostname->override-tag for overrides.
     * <b>Creators should set new object here.<b>
     */
    hostOverrides: null
  });
  
});
window.require.register("models/target_cluster", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.TargetCluster = DS.Model.extend({
    name: DS.attr('string'),
    execute: DS.attr('string'),
    workflow: DS.attr('string'),
    write: DS.attr('string'),
    readonly: DS.attr('string'),
    staging: DS.attr('string'),
    working: DS.attr('string'),
    temp: DS.attr('string')
  });

  App.TargetCluster.FIXTURES = [];
});
window.require.register("models/user", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');

  App.User = DS.Model.extend({
    userName:DS.attr('string'),
    id:function(){
      return this.get('userName');
    }.property('userName'),
    roles:DS.attr('string'),
    isLdap:DS.attr('boolean'),
    type: function(){
      if(this.get('isLdap')){
        return 'LDAP';
      }
      return 'Local';
    }.property('isLdap'),
    auditItems:DS.hasMany('App.ServiceAudit'),
    admin: DS.attr('boolean')
  });

  App.EditUserForm = App.Form.extend({
    className:App.User,
    object:function () {
      return App.router.get('mainAdminUserEditController.content');
    }.property('App.router.mainAdminUserEditController.content'),

    fieldsOptions:[
      { name:"userName", displayName:"Username" },
      { name:"old_password", displayName:"Current Password", displayType:"password", isRequired: false },
      { name:"new_password", displayName:"New Password", displayType:"password",  isRequired: false },
      { name:"new_passwordRetype", displayName:"Retype New Password", displayType:"password", isRequired: false },
      { name:"admin", displayName:"Admin", displayType:"checkbox", isRequired:false },
      { name:"roles", displayName:"Role", isRequired:false, isHidden:true },
      { name:"isLdap", displayName:"Type", isRequired:false, isHidden:true }
    ],
    fields:[],
    disableUsername:function () {
      this.getField("userName").set("disabled", "disabled");
    }.observes('object'),
    disableAdminCheckbox:function () {
      var object = this.get('object');
      if (object) {
        if ((object.get('userName') == App.get('router').getLoginName()) || App.supports.ldapGroupMapping && object.get("isLdap")) {
          this.getField("admin").set("disabled", true);
        } else {
          this.getField("admin").set("disabled", false);
        }
      }
    }.observes('object'),

    isValid:function () {

      var isValid = this._super();
      var thisForm = this;

      var newPass = this.get('field.new_password');
      var oldPass = this.get('field.old_password');
      var passRetype = this.get('field.new_passwordRetype');

      if (!validator.empty(newPass.get('value'))) {
        if(validator.empty(oldPass.get('value'))){
          oldPass.set('errorMessage', this.t('admin.users.editError.requiredField'));
          isValid = false;
        }
        if (newPass.get('value') != passRetype.get('value')) {
          passRetype.set('errorMessage', this.t('admin.users.createError.passwordValidation'));
          isValid = false;
        }
      }

      return isValid;
    },

    save: function () {
      var object = this.get('object');
      var formValues = {};
      $.each(this.get('fields'), function () {
        formValues[this.get('name')] = this.get('value');
      });

      $.each(formValues, function (k, v) {
        object.set(k, v);
      });

      //App.store.commit();
      this.set('result', 1);

      return true;
    }
  });
  App.CreateUserForm = App.Form.extend({
    className:App.User,
    object:function () {
      return App.router.get('mainAdminUserCreateController.content');
    }.property('App.router.mainAdminUserCreateController.content'),

    fieldsOptions:[
      { name:"userName", displayName:"Username", toLowerCase: function(){var v = this.get('value'); this.set('value', v.toLowerCase())}.observes('value') },
      { name:"password", displayName:"Password", displayType:"password", isRequired: true },
      { name:"passwordRetype", displayName:"Retype Password", displayType:"password", validator:"passwordRetype", isRequired: true },
      { name:"admin", displayName:"Admin", displayType:"checkbox", isRequired:false },
      { name:"roles", displayName:"Role", isRequired:false, isHidden:true }
    ],
    fields:[],

    isValid:function () {
      var isValid = this._super();

      var passField = this.get('field.password');
      var passRetype = this.get('field.passwordRetype');

      if (!validator.empty(passField.get('value'))) {
        if (passField.get('value') != passRetype.get('value')) {
          passRetype.set('errorMessage', this.t('admin.users.createError.passwordValidation'));
          isValid = false;
        }
      }

      if (isValid) {
        var users = App.User.find();
        var userNameField = this.getField('userName');
        var userName = userNameField.get('value');

        if (users.mapProperty('userName').contains(userName)) {
          userNameField.set('errorMessage', this.t('admin.users.createError.userNameExists'));
          return isValid = false;
        }
      }

      return isValid;
    },

    isWarn: function() {
      var isWarn = false;
      var userNameField = this.getField('userName');
      userNameField.set('warnMessage', '');
      var userName = userNameField.get('value');

      if (this.isValid() && !validator.isValidUserName(userName)) {
        userNameField.set('warnMessage', this.t('users.userName.validationFail'));
        isWarn = true;
      }
      return isWarn;
    },

    save: function () {

      var object = this.get('object');
      var formValues = {};
      $.each(this.get('fields'), function () {
        formValues[Ember.String.decamelize(this.get('name'))] = this.get('value');
      });

      if (this.get('className')) {
        App.store.load(this.get('className'), App.dateTime(), formValues);
      }
      else {
        console.log("Please define class name for your form " + this.constructor);
      }

      this.set('result', 1);

      return true;
    }
  });
  App.User.FIXTURES = [];

  
});
window.require.register("router", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.WizardRoute = Em.Route.extend({
    isRoutable: function() {
      return (typeof this.get('route') === 'string' && App.router.get('loggedIn'));
    }.property('App.router.loggedIn')
  });

  App.Router = Em.Router.extend({

    enableLogging: true,
    isFwdNavigation: true,
    backBtnForHigherStep: false,

    setNavigationFlow: function (step) {
      var matches = step.match(/\d+$/);
      var newStep;
      if (matches) {
        newStep = parseInt(matches[0]);
      }
      var previousStep = parseInt(this.getInstallerCurrentStep());
      this.set('isFwdNavigation', newStep >= previousStep);
    },


    clearAllSteps: function () {
      this.get('installerController').clear();
      this.get('addHostController').clear();
      this.get('addServiceController').clear();
      this.get('stackUpgradeController').clear();
      for (var i = 1; i < 11; i++) {
        this.set('wizardStep' + i + 'Controller.hasSubmitted', false);
        this.set('wizardStep' + i + 'Controller.isDisabled', true);
      }
    },

    /**
     * Temporary fix for getting cluster name
     * @return {*}
     */

    getClusterName: function () {
      return App.router.get('clusterController').get('clusterName');
    },


    /**
     * Get current step of Installer wizard
     * @return {*}
     */
    getInstallerCurrentStep: function () {
      return this.getWizardCurrentStep('installer');
    },

    /**
     * Get current step for <code>wizardType</code> wizard
     * @param wizardType one of <code>installer</code>, <code>addHost</code>, <code>addServices</code>
     */
    getWizardCurrentStep: function (wizardType) {
      var loginName = this.getLoginName();
      var currentStep = App.db.getWizardCurrentStep(wizardType);
      console.log('getWizardCurrentStep: loginName=' + loginName + ", currentStep=" + currentStep);
      if (!currentStep) {
        currentStep = wizardType === 'installer' ? '0' : '1';
      }
      console.log('returning currentStep=' + currentStep);
      return currentStep;
    },

    loggedIn: App.db.getAuthenticated(),

    loginName: function() {
      return this.getLoginName();
    }.property('loggedIn'),

    getAuthenticated: function () {
      var auth = App.db.getAuthenticated();
      var authResp = (auth && auth === true);
      if (authResp) {
        App.ajax.send({
          name: 'router.authentication',
          sender: this,
          success: 'onAuthenticationSuccess',
          error: 'onAuthenticationError'
        });
      } else {
        this.set('loggedIn', false);
      }
      return this.get('loggedIn');
    },

    onAuthenticationSuccess: function (data) {
      this.set('loggedIn', true);
    },

    onAuthenticationError: function (data) {
      if (data.status === 403) {
        this.set('loggedIn', false);
      } else {
        console.log('error in getAuthenticated');
      }
    },

    setAuthenticated: function (authenticated) {
      console.log("TRACE: Entering router:setAuthenticated function");
      App.db.setAuthenticated(authenticated);
      this.set('loggedIn', authenticated);
    },

    getLoginName: function () {
      return App.db.getLoginName();
    },

    setLoginName: function (loginName) {
      App.db.setLoginName(loginName);
    },

    /**
     * Set user model to local storage
     * @param user
     */
    setUser: function (user) {
      App.db.setUser(user);
    },

    /**
     * Get user model from local storage
     * @return {*}
     */
    getUser: function () {
      return App.db.getUser();
    },

    login: function () {
      var controller = this.get('loginController');
      var loginName = controller.get('loginName').toLowerCase();
      controller.set('loginName', loginName);
      var hash = window.btoa(loginName + ":" + controller.get('password'));
      var usr = '';

      if (App.testMode) {
        if (loginName === "admin" && controller.get('password') === 'admin') {
          usr = 'admin';
        } else if (loginName === 'user' && controller.get('password') === 'user') {
          usr = 'user';
        }
      }

      App.ajax.send({
        name: 'router.login',
        sender: this,
        data: {
          auth: "Basic " + hash,
          usr: usr,
          loginName: loginName
        },
        beforeSend: 'authBeforeSend',
        success: 'loginSuccessCallback',
        error: 'loginErrorCallback'
      });

    },

    authBeforeSend: function(opt, xhr, data) {
      xhr.setRequestHeader("Authorization", data.auth);
    },

    loginSuccessCallback: function(data, opt, params) {
      console.log('login success');
      var d = data;
      var isAdmin = data.Users.roles.indexOf('admin') >= 0;
      if (isAdmin) {
        App.set('isAdmin', true);
        var controller = this.get('loginController');
        this.setAuthenticated(true);
        this.setLoginName(params.loginName);
        App.usersMapper.map({"items": [data]});
        this.setUser(App.User.find(params.loginName));
        this.transitionTo(this.getSection());
        controller.postLogin(true,true);
      }
      else {
        App.ajax.send({
          name: 'router.login2',
          sender: this,
          data: {
            loginName: params.loginName,
            loginData: data
          },
          success: 'login2SuccessCallback',
          error: 'login2ErrorCallback'
        });
      }
    },

    loginErrorCallback: function(request, ajaxOptions, error, opt) {
      var controller = this.get('loginController');
      console.log("login error: " + error);
      this.setAuthenticated(false);
      if (request.status == 403) {
        controller.postLogin(true, false);
      } else {
        controller.postLogin(false, false);
      }

    },

    login2SuccessCallback: function (clusterResp, opt, params) {
      var controller = this.get('loginController');
      if (clusterResp.items.length) {
        this.setAuthenticated(true);
        this.setLoginName(params.loginName);
        App.usersMapper.map({"items": [params.loginData]});
        this.setUser(App.User.find(params.loginName));
        this.transitionTo(this.getSection());
        controller.postLogin(true,true);
      }
      else {
        controller.set('errorMessage', Em.I18n.t('router.hadoopClusterNotSetUp'));
      }
    },

    login2ErrorCallback: function (req) {
      console.log("Server not responding: " + req.statusCode);
    },

    setAmbariStacks: function () {
      App.ajax.send({
        name: 'router.set_ambari_stacks',
        sender: this,
        success: 'setAmbariStacksSuccessCallback',
        error: 'setAmbariStacksErrorCallback'
      });
    },

    setAmbariStacksSuccessCallback: function (jsonData) {
      console.log("TRACE: In success function for the setAmbariStacks call");
      var stacks = [];
      jsonData.forEach(function (_stack) {
        stacks.pushObject({
          name: _stack.name,
          version: _stack.version
        });
      }, this);
      App.db.setAmbariStacks(stacks);
      console.log('TRACEIINNGG: ambaristacks: ' + JSON.stringify(App.db.getAmbariStacks()));
    },

    setAmbariStacksErrorCallback: function (request, ajaxOptions, error) {
      console.log("TRACE: In error function for the setAmbariStacks call");
      console.log("TRACE: error code status is: " + request.status);
      console.log('Error message is: ' + request.responseText);
    },

    getSection: function () {
      if (App.testMode) {
        if (App.alwaysGoToInstaller) {
          return 'installer';
        } else {
          return 'main.dashboard';
        }
      }
      App.clusterStatus.updateFromServer(false, false);
      var clusterStatusOnServer = App.clusterStatus.get('value');
      if (!App.get('isAdmin') || clusterStatusOnServer && clusterStatusOnServer.clusterState === 'DEFAULT') {
        return 'main.dashboard';
      } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('addHostController.name')) {
        // if wizardControllerName == "addHostController", then it means someone closed the browser or the browser was crashed when we were last in Add Hosts wizard
        return 'main.hostAdd';
      } else if (clusterStatusOnServer && (clusterStatusOnServer.wizardControllerName === App.router.get('addSecurityController.name') || clusterStatusOnServer.wizardControllerName === App.router.get('mainAdminSecurityDisableController.name'))) {
        // if wizardControllerName == "addSecurityController", then it means someone closed the browser or the browser was crashed when we were last in Add Security wizard
        return 'main.admin.adminSecurity';
      } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('addServiceController.name')) {
        // if wizardControllerName == "addHostController", then it means someone closed the browser or the browser was crashed when we were last in Add Hosts wizard
        return 'main.serviceAdd';
      } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('stackUpgradeController.name')) {
        // if wizardControllerName == "stackUpgradeController", then it means someone closed the browser or the browser was crashed when we were last in Stack Upgrade wizard
        return 'main.stackUpgrade';
      } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('reassignMasterController.name')) {
        // if wizardControllerName == "reassignMasterController", then it means someone closed the browser or the browser was crashed when we were last in Reassign Master wizard
        return 'main.reassign';
      } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('highAvailabilityWizardController.name')) {
        // if wizardControllerName == "highAvailabilityWizardController", then it means someone closed the browser or the browser was crashed when we were last in NameNode High Availability wizard
        return 'main.admin.enableHighAvailability';
      }else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('rollbackHighAvailabilityWizardController.name')) {
        // if wizardControllerName == "highAvailabilityRollbackController", then it means someone closed the browser or the browser was crashed when we were last in NameNode High Availability Rollback wizard
        return 'main.admin.rollbackHighAvailability';
      } else {
        // if wizardControllerName == "installerController", then it means someone closed the browser or the browser was crashed when we were last in Installer wizard
        return 'installer';
      }
    },

    logOff: function (context) {
      $('title').text('Ambari');
      var hash = window.btoa(this.get('loginController.loginName') + ":" + this.get('loginController.password'));

      App.router.get('mainController').stopPolling();
      // App.db.cleanUp() must be called before router.clearAllSteps().
      // otherwise, this.set('installerController.currentStep, 0) would have no effect
      // since it's a computed property but we are not setting it as a dependent of App.db.
      App.db.cleanUp();
      App.set('isAdmin', false);
      this.set('loggedIn', false);
      this.clearAllSteps();
      console.log("Log off: " + App.router.getClusterName());
      this.set('loginController.loginName', '');
      this.set('loginController.password', '');
      // When logOff is called by Sign Out button, context contains event object. As it is only case we should send logoff request, we are checking context below.
      if (!App.testMode && context) {
        App.ajax.send({
          name: 'router.logoff',
          sender: this,
          data: {
            auth: "Basic " + hash
          },
          beforeSend: 'authBeforeSend',
          success: 'logOffSuccessCallback',
          error:'logOffErrorCallback'
        });
      }
      this.transitionTo('login', context);
    },

    logOffSuccessCallback: function (data) {
      console.log("invoked logout on the server successfully");
    },

    logOffErrorCallback: function (req) {
      console.log("failed to invoke logout on the server");
    },

    /**
     * initialize isAdmin if user is administrator
     */
    initAdmin: function(){
      if(App.db && App.db.getUser() && App.db.getUser().admin) {
        App.set('isAdmin', true);
        console.log('Administrator logged in');
      }
    },

    root: Em.Route.extend({
      index: Em.Route.extend({
        route: '/',
        redirectsTo: 'login'
      }),

      enter: function(router){
        router.initAdmin();
      },

      login: Em.Route.extend({
        route: '/login',

        /**
         *  If the user is already logged in, redirect to where the user was previously
         */
        enter: function (router, context) {
          if (router.getAuthenticated()) {
            Ember.run.next(function () {
              console.log(router.getLoginName() + ' already authenticated.  Redirecting...');
              router.transitionTo(router.getSection(), context);
            });
          }
        },

        connectOutlets: function (router, context) {
          $('title').text(Em.I18n.t('app.name'));
          console.log('/login:connectOutlet');
          console.log('currentStep is: ' + router.getInstallerCurrentStep());
          console.log('authenticated is: ' + router.getAuthenticated());
          router.get('applicationController').connectOutlet('login');
        }
      }),

      installer: require('routes/installer'),

      main: require('routes/main'),
      
      experimental: Em.Route.extend({
        route: '/experimental',
        enter: function (router, context) {
          
        },
        connectOutlets: function (router, context) {
          $('title').text("Ambari Experimental");
          console.log('/experimental:connectOutlet');
          router.get('applicationController').connectOutlet('experimental');
        }
      }),

      logoff: function (router, context) {
        router.logOff(context);
      }

    })
  });
  
});
window.require.register("routes/add_host_routes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: '/host/add',

    clearData: function (router) {
      App.router.get('updateController').set('isWorking', true);
      App.clusterStatus.setClusterStatus({
        clusterName: App.router.get('content.cluster.name'),
        clusterState: 'DEFAULT',
        wizardControllerName: App.router.get('addHostController.name'),
        localdb: App.db.data
      });
      router.transitionTo('hosts.index');
    },

    enter: function (router) {
      console.log('in /host/add:enter');

      var self = this;

      Ember.run.next(function () {
        var addHostController = router.get('addHostController');
        App.router.get('updateController').set('isWorking', false);
        App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header:Em.I18n.t('hosts.add.header'),
          bodyClass:  App.AddHostView.extend({
            controllerBinding: 'App.router.addHostController'
          }),
          primary:Em.I18n.t('form.cancel'),
          secondary: null,
          showFooter: false,

          onPrimary:function () {
            this.hide();
            App.router.get('updateController').set('isWorking', true);
            router.transitionTo('hosts.index');
          },
          onClose: function() {
            if (addHostController.get('currentStep') == '6') {
              App.ModalPopup.show({
                header: Em.I18n.t('hosts.add.exit.header'),
                body: Em.I18n.t('hosts.add.exit.body'),
                onPrimary: function () {
                  this.hide();
                  addHostController.finish();
                  App.router.get('updateController').set('isWorking', true);
                  self.clearData(router);
                  location.reload();
                }
              });
            } else {
              this.hide();
              self.clearData(router);
            }
          },
          didInsertElement: function(){
            this.fitHeight();
          }
        });
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus) {
          switch (currentClusterStatus.clusterState) {
            case 'ADD_HOSTS_DEPLOY_PREP_2' :
              addHostController.setCurrentStep('4');
              break;
            case 'ADD_HOSTS_INSTALLING_3' :
            case 'SERVICE_STARTING_3' :
              addHostController.setCurrentStep('5');
              break;
            case 'ADD_HOSTS_INSTALLED_4' :
              addHostController.setCurrentStep('6');
              break;
            default:
              break;
          }
        }

        router.transitionTo('step' + addHostController.get('currentStep'));
      });

    },

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        console.log('in addHost.step1:connectOutlets');
        var controller = router.get('addHostController');
        controller.setCurrentStep('1');
        controller.set('hideBackButton', true);
        controller.dataLoading().done(function () {
          controller.loadServicesFromServer();
          controller.loadAllPriorSteps();
          var wizardStep2Controller = router.get('wizardStep2Controller');
          wizardStep2Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep2', controller.get('content'));
        })
      },

      next: function (router) {
        var controller = router.get('addHostController');
        controller.save('installOptions');
        //hosts was saved to content.hosts inside wizardStep2Controller
        controller.save('hosts');
        router.transitionTo('step2');
        controller.setDBProperty('bootStatus', false);
      },
      evaluateStep: function (router) {
        console.log('in addHost.step1:evaluateStep');
        var addHostController = router.get('addHostController');
        var wizardStep2Controller = router.get('wizardStep2Controller');

        wizardStep2Controller.set('hasSubmitted', true);

        if (!wizardStep2Controller.get('isSubmitDisabled')) {
          wizardStep2Controller.evaluateStep();
        }
      }
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router) {
        console.log('in addHost.step2:connectOutlets');
        var controller = router.get('addHostController');
        controller.setCurrentStep('2');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          var wizardStep3Controller = router.get('wizardStep3Controller');
          wizardStep3Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep3', controller.get('content'));
        })
      },
      back: function(router){
        router.transitionTo('step1');
      },
      exit: function (router) {
        router.get('wizardStep3Controller').set('stopBootstrap', true);
      },
      next: function (router, context) {
        var addHostController = router.get('addHostController');
        var wizardStep3Controller = router.get('wizardStep3Controller');
        addHostController.saveConfirmedHosts(wizardStep3Controller);
        addHostController.saveClients();

        addHostController.setDBProperty('bootStatus', true);
        router.transitionTo('step3');
      },
      /**
       * Wrapper for remove host action.
       * Since saving data stored in addHostController, we should call this from router
       * @param router
       * @param context Array of hosts to delete
       */
      removeHosts: function (router, context) {
        console.log('in addHost.step2.removeHosts:hosts to delete ', context);
        var controller = router.get('addHostController');
        controller.removeHosts(context);
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router, context) {
        console.log('in addHost.step3:connectOutlets');
        var controller = router.get('addHostController');
        controller.setCurrentStep('3');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          var wizardStep6Controller = router.get('wizardStep6Controller');
          wizardStep6Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep6', controller.get('content'));
          wizardStep6Controller.set('isMasters', false);
        });
      },
      back: Em.Router.transitionTo('step2'),
      next: function (router) {
        var addHostController = router.get('addHostController');
        var wizardStep6Controller = router.get('wizardStep6Controller');

        if (wizardStep6Controller.validate()) {
          addHostController.saveSlaveComponentHosts(wizardStep6Controller);
          if(App.supports.hostOverrides){
            router.transitionTo('step4');
          }else{
            router.transitionTo('step5');
          }
        }
      }
    }),

    step4: Em.Route.extend({
      route: '/step4',
      connectOutlets: function (router, context) {
        console.log('in addHost.step4:connectOutlets');
        var controller = router.get('addHostController');
        var addHostStep4Controller = router.get('addHostStep4Controller');
        controller.setCurrentStep('4');
        addHostStep4Controller.loadConfigGroups();
        addHostStep4Controller.configGroupsLoading().done(function () {
          controller.dataLoading().done(function () {
            controller.loadAllPriorSteps();
            controller.loadServiceConfigGroups();
            addHostStep4Controller.set('wizardController', controller);
            controller.connectOutlet('addHostStep4', controller.get('content'));
          })
        })
      },
      back: function(router){
          router.transitionTo('step3');
      },
      next: function (router) {
        var addHostController = router.get('addHostController');
        addHostController.saveServiceConfigGroups();
        router.transitionTo('step5');
      }
    }),

    step5: Em.Route.extend({
      route: '/step5',
      connectOutlets: function (router, context) {
        console.log('in addHost.step5:connectOutlets');
        var controller = router.get('addHostController');
        controller.setCurrentStep('5');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.getServiceConfigGroups();
          var wizardStep8Controller = router.get('wizardStep8Controller');
          wizardStep8Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep8', controller.get('content'));
        })
      },
      back: function(router){
        if(!router.get('wizardStep8Controller.isBackBtnDisabled')) {
          if(App.supports.hostOverrides){
            router.transitionTo('step4');
          }else{
            router.transitionTo('step3');
          }
        }
      },
      next: function (router) {
        var addHostController = router.get('addHostController');
        var wizardStep8Controller = router.get('wizardStep8Controller');
        if(App.supports.hostOverrides){
          addHostController.applyConfigGroup();
        }
        addHostController.installServices();
        addHostController.setInfoForStep9();
        // We need to do recovery based on whether we are in Add Host or Installer wizard
        addHostController.saveClusterState('ADD_HOSTS_INSTALLING_3');
        wizardStep8Controller.set('servicesInstalled', true);
        router.transitionTo('step6');
      }
    }),

    step6: Em.Route.extend({
      route: '/step6',
      connectOutlets: function (router, context) {
        console.log('in addHost.step6:connectOutlets');
        var controller = router.get('addHostController');
        controller.setCurrentStep('6');
        controller.dataLoading().done(function () {
          var wizardStep9Controller = router.get('wizardStep9Controller');
          wizardStep9Controller.set('wizardController', controller);
          controller.loadAllPriorSteps();
          if (!App.testMode) {              //if test mode is ON don't disable prior steps link.
            controller.setLowerStepsDisable(6);
          }
          controller.connectOutlet('wizardStep9', controller.get('content'));
        })
      },
      back: Em.Router.transitionTo('step5'),
      retry: function(router,context) {
        var addHostController = router.get('addHostController');
        var wizardStep9Controller = router.get('wizardStep9Controller');
        wizardStep9Controller.set('wizardController', addHostController);
        if (wizardStep9Controller.get('showRetry')) {
          if (wizardStep9Controller.get('content.cluster.status') === 'INSTALL FAILED') {
            var isRetry = true;
            addHostController.installServices(isRetry);
            addHostController.setInfoForStep9();
            wizardStep9Controller.resetHostsForRetry();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            addHostController.saveClusterState('ADD_HOSTS_INSTALLING_3');
          }
          wizardStep9Controller.navigateStep();
        }
      },
      unroutePath: function() {
        return false;
      },
      next: function (router) {
        var addHostController = router.get('addHostController');
        var wizardStep9Controller = router.get('wizardStep9Controller');
        wizardStep9Controller.set('wizardController', addHostController);
        addHostController.saveInstalledHosts(wizardStep9Controller);

        // We need to do recovery based on whether we are in Add Host or Installer wizard
        addHostController.saveClusterState('ADD_HOSTS_INSTALLED_4');
        router.transitionTo('step7');
      }
    }),

    step7: Em.Route.extend({
      route: '/step7',
      connectOutlets: function (router, context) {
        console.log('in addHost.step7:connectOutlets');
        var controller = router.get('addHostController');
        controller.setCurrentStep('7');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          var wizardStep10Controller = router.get('wizardStep10Controller');
          wizardStep10Controller.set('wizardController', controller);
          if (!App.testMode) {              //if test mode is ON don't disable prior steps link.
            controller.setLowerStepsDisable(7);
          }
          controller.connectOutlet('wizardStep10', controller.get('content'));
          router.get('updateController').set('isWorking', true);
        })
      },
      back: Em.Router.transitionTo('step6'),
      complete: function (router, context) {
        var addHostController = router.get('addHostController');
        var hostsUrl = '/hosts?fields=Hosts/host_name,Hosts/public_host_name,Hosts/cpu_count,Hosts/total_mem,' +
          'Hosts/host_status,Hosts/last_heartbeat_time,Hosts/os_arch,Hosts/os_type,Hosts/ip,host_components,' +
          'metrics/disk,metrics/load/load_one,metrics/cpu/cpu_system,metrics/cpu/cpu_user,metrics/memory/mem_total,metrics/memory/mem_free';
        router.get('clusterController').requestHosts(hostsUrl, function () {
          console.log('Request for hosts, with immutable parameters')
        });
        router.get('updateController').updateAll();
        addHostController.finish();
        $(context.currentTarget).parents("#modal").find(".close").trigger('click');

        // We need to do recovery based on whether we are in Add Host or Installer wizard
        addHostController.saveClusterState('DEFAULT');

        location.reload();
      }
    }),

    backToHostsList: function (router, event) {
      App.router.get('updateController').set('isWorking', true);
      router.transitionTo('hosts.index');
    },

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3'),

    gotoStep4: Em.Router.transitionTo('step4'),

    gotoStep5: Em.Router.transitionTo('step5'),

    gotoStep6: Em.Router.transitionTo('step6'),

    gotoStep7: Em.Router.transitionTo('step7')


  });
  
});
window.require.register("routes/add_security", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: '/addSecurity',
    enter: function (router) {
      console.log('in /security/add:enter');

      Ember.run.next(function () {
        //after refresh check if the wizard is open then restore it
        if (router.get('mainAdminSecurityController').getAddSecurityWizardStatus() === 'RUNNING') {
          var mainAdminSecurityController = router.get('mainAdminSecurityController');
          var addSecurityController = router.get('addSecurityController');
          var currentStep = router.get('addSecurityController').get('currentStep');
          App.router.get('updateController').set('isWorking', false);
          App.ModalPopup.show({
              classNames: ['full-width-modal'],
              header: Em.I18n.t('admin.addSecurity.header'),
              bodyClass: App.MainAdminSecurityAddMenuView.extend({
                controllerBinding: 'App.router.addSecurityController'
              }),
              primary: Em.I18n.t('form.cancel'),
              secondary: null,
              showFooter: false,

              onClose: function () {
                var self = this;
                if (router.get('addSecurityController.currentStep') == 4) {
                  var controller = router.get('mainAdminSecurityAddStep4Controller');
                  if (!controller.get('isSubmitDisabled')) {
                    router.get('mainAdminSecurityAddStep4Controller').clearStep();
                    self.proceedOnClose();
                    return;
                  }
                  var applyingConfigCommand = router.get('mainAdminSecurityAddStep4Controller.commands').findProperty('name', 'APPLY_CONFIGURATIONS');
                  if (applyingConfigCommand) {
                    if (!applyingConfigCommand.get('isCompleted')) {
                      if (applyingConfigCommand.get('isStarted')) {
                        App.showAlertPopup(Em.I18n.t('admin.security.applying.config.header'), Em.I18n.t('admin.security.applying.config.body'));
                      } else {
                        App.showConfirmationPopup(function () {
                          self.proceedOnClose();
                        }, Em.I18n.t('admin.addSecurity.enable.onClose'));
                      }
                    } else {
                      App.showConfirmationPopup(function () {
                          self.proceedOnClose();
                        },
                        Em.I18n.t('admin.addSecurity.enable.after.stage2.onClose')
                      );
                    }
                    return;
                  }
                }
                router.get('mainAdminSecurityAddStep4Controller').clearStep();
                App.db.setSecurityDeployCommands(undefined);
                self.proceedOnClose();
              },
              proceedOnClose: function () {
                this.hide();
                router.get('mainAdminSecurityAddStep4Controller').clearStep();
                router.get('addSecurityController.content.services').clear();
                router.set('addSecurityController.content.serviceConfigProperties', null);
                App.router.get('updateController').set('isWorking', true);
                mainAdminSecurityController.setAddSecurityWizardStatus(null);
                App.db.setSecurityDeployCommands(undefined);
                router.get('addSecurityController').setCurrentStep(1);
                App.clusterStatus.setClusterStatus({
                  clusterName: router.get('content.cluster.name'),
                  clusterState: 'DEFAULT'
                });
                router.transitionTo('adminSecurity.index');
              },
              didInsertElement: function () {
                this.fitHeight();
              }
            }
          );

          App.router.transitionTo('step' + currentStep);
        } else {
          router.transitionTo('adminSecurity.index');
        }
      });
    },

    step1: Em.Route.extend({
      route: '/start',
      enter: function (router) {
        router.get('addSecurityController').setCurrentStep('1');
        if(!App.testMode){
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'ADD_SECURITY_STEP_1',
            wizardControllerName: router.get('addSecurityController.name'),
            localdb: App.db.data
          });
        }
      },

      connectOutlets: function (router) {
        console.log('in addSecurity.step1:connectOutlets');
        var controller = router.get('addSecurityController');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('mainAdminSecurityAddStep1', controller.get('content'));
        })
      },

      unroutePath: function () {
        return false;
      },

      next: function (router) {
        var addSecurityController = router.get('addSecurityController');
        addSecurityController.get('content').set('serviceConfigProperties', null);
        App.db.setSecureConfigProperties(null);
        router.transitionTo('step2');
      }
    }),

    step2: Em.Route.extend({
      route: '/configure',

      enter: function (router) {
        router.get('addSecurityController').setCurrentStep('2');
        if(!App.testMode){
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'ADD_SECURITY_STEP_2',
            wizardControllerName: router.get('addSecurityController.name'),
            localdb:  App.db.data
          });
        }
      },
      connectOutlets: function (router) {
        console.log('in addSecurity.step2:connectOutlets');
        var controller = router.get('addSecurityController');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('mainAdminSecurityAddStep2', controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      back: Em.Router.transitionTo('step1'),
      next: function (router) {
        var addSecurityController = router.get('addSecurityController');
        var addSecurityStep2Controller = router.get('mainAdminSecurityAddStep2Controller');
        addSecurityController.saveServiceConfigProperties(addSecurityStep2Controller);
        router.transitionTo('step3');
      }
    }),

    step3: Em.Route.extend({
      route: '/principal_keytab',

      enter: function (router) {
        router.get('addSecurityController').setCurrentStep('3');
        if(!App.testMode){
          App.clusterStatus.setClusterStatus({
            clusterName: this.get('clusterName'),
            clusterState: 'ADD_SECURITY_STEP_3',
            wizardControllerName: router.get('addSecurityController.name'),
            localdb:  App.db.data
          });
        }
      },
      connectOutlets: function (router) {
        console.log('in addSecurity.step3:connectOutlets');
        var controller = router.get('addSecurityController');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('mainAdminSecurityAddStep3', controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      back: Em.Router.transitionTo('step2'),
      next: function (router) {
        App.db.setSecurityDeployCommands(undefined);
        router.transitionTo('step4');
      }
    }),

    step4: Em.Route.extend({
      route: '/apply',

      enter: function (router) {
        router.get('addSecurityController').setCurrentStep('4');
      },

      connectOutlets: function (router) {
        console.log('in addSecurity.step4:connectOutlets');
        var controller = router.get('addSecurityController');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.setLowerStepsDisable(4);
          controller.connectOutlet('mainAdminSecurityAddStep4', controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      back: function (router, context) {
        var controller = router.get('mainAdminSecurityAddStep4Controller');
        if (!controller.get('isBackBtnDisabled')) {
          router.transitionTo('step3');
        }
      },
      done: function (router, context) {
        var controller = router.get('mainAdminSecurityAddStep4Controller');
        if (!controller.get('isSubmitDisabled')) {
          $(context.currentTarget).parents("#modal").find(".close").trigger('click');
        }
        location.reload();
      }
    }),

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3'),

    gotoStep4: Em.Router.transitionTo('step4')

  });

  
});
window.require.register("routes/add_service_routes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: '/service/add',

    enter: function (router) {
      console.log('in /service/add:enter');
      if (App.get('isAdmin')) {
        Em.run.next(function () {
          var addServiceController = router.get('addServiceController');
          App.router.get('updateController').set('isWorking', false);
          var popup = App.ModalPopup.show({
            classNames: ['full-width-modal'],
            header:Em.I18n.t('services.add.header'),
            bodyClass:  App.AddServiceView.extend({
              controllerBinding: 'App.router.addServiceController'
            }),
            primary:Em.I18n.t('form.cancel'),
            showFooter: false,
            secondary: null,

            onPrimary:function () {
              this.hide();
              App.router.transitionTo('main.services.index');
            },
            onClose: function() {
              this.set('showCloseButton', false); // prevent user to click "Close" many times
              App.router.get('updateController').set('isWorking', true);
              var self = this;
              App.clusterStatus.setClusterStatus({
                clusterName: App.router.get('content.cluster.name'),
                clusterState: 'DEFAULT',
                wizardControllerName: App.router.get('addServiceController.name'),
                localdb: App.db.data
              });
              App.router.get('updateController').updateServices(function(){
                App.router.get('updateController').updateServiceMetric(function(){
                  self.hide();
                });
              });
              App.router.transitionTo('main.services.index');
            },
            didInsertElement: function(){
              this.fitHeight();
            }
          });
          addServiceController.set('popup',popup);
          var currentClusterStatus = App.clusterStatus.get('value');
          if (currentClusterStatus) {
            switch (currentClusterStatus.clusterState) {
              case 'ADD_SERVICES_DEPLOY_PREP_2' :
                addServiceController.setCurrentStep('5');
                break;
              case 'ADD_SERVICES_INSTALLING_3' :
              case 'SERVICE_STARTING_3' :
                addServiceController.setCurrentStep('6');
                break;
              case 'ADD_SERVICES_INSTALLED_4' :
                addServiceController.setCurrentStep('7');
                break;
              default:
                break;
            }
          }

          router.transitionTo('step' + addServiceController.get('currentStep'));
        });
      } else {
        Em.run.next(function () {
          App.router.transitionTo('main.services');
        });
      }

    },

    /*connectOutlets: function (router) {
      console.log('in /service/add:connectOutlets');
      router.get('mainController').connectOutlet('addService');
    },*/

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        console.log('in addService.step1:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('1');
        controller.set('hideBackButton', true);
        controller.dataLoading().done(function () {
          controller.loadServicesFromServer();
          controller.loadAllPriorSteps();
          controller.connectOutlet('wizardStep4', controller.get('content.services'));
        })
      },
      next: function (router) {
        var addServiceController = router.get('addServiceController');
        var wizardStep4Controller = router.get('wizardStep4Controller');
        addServiceController.saveServices(wizardStep4Controller);
        addServiceController.saveClients(wizardStep4Controller);
        addServiceController.setDBProperty('masterComponentHosts', undefined);
        router.transitionTo('step2');
      }
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router) {
        console.log('in addService.step2:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('2');
        controller.set('hideBackButton', false);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('wizardStep5', controller.get('content'));
        })

      },
      back: Em.Router.transitionTo('step1'),
      next: function (router) {
        var addServiceController = router.get('addServiceController');
        var wizardStep5Controller = router.get('wizardStep5Controller');
        addServiceController.saveMasterComponentHosts(wizardStep5Controller);
        addServiceController.setDBProperty('slaveComponentHosts', undefined);
        router.transitionTo('step3');
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router) {
        console.log('in addService.step3:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('3');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          var wizardStep6Controller = router.get('wizardStep6Controller');
          wizardStep6Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep6', controller.get('content'));
          wizardStep6Controller.set('isMasters', false);
        })
      },
      back: function(router){
        var controller = router.get('addServiceController');
        if(!controller.get('content.skipMasterStep')){
          router.transitionTo('step2');
        } else {
          router.transitionTo('step1');
        }
      },
      next: function (router) {
        var addServiceController = router.get('addServiceController');
        var wizardStep6Controller = router.get('wizardStep6Controller');

        if (wizardStep6Controller.validate()) {
          addServiceController.saveSlaveComponentHosts(wizardStep6Controller);
          addServiceController.get('content').set('serviceConfigProperties', null);
          addServiceController.setDBProperty('serviceConfigProperties', null);
          addServiceController.setDBProperty('groupsToDelete', []);
          router.transitionTo('step4');
        }
      }
    }),

    step4: Em.Route.extend({
      route: '/step4',
      connectOutlets: function (router) {
        console.log('in addService.step4:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('4');
        controller.dataLoading().done(function () {
          var wizardStep7Controller = router.get('wizardStep7Controller');
          controller.loadAllPriorSteps();
          controller.loadAdvancedConfigs(wizardStep7Controller);
          wizardStep7Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep7', controller.get('content'));
        })
      },
      back: function(router){
        var controller = router.get('addServiceController');
        if(!controller.get('content.skipSlavesStep')){
          router.transitionTo('step3');
        } else if(!controller.get('content.skipMasterStep')) {
          router.transitionTo('step2');
        } else {
          router.transitionTo('step1');
        }
      },
      next: function (router) {
        var addServiceController = router.get('addServiceController');
        var wizardStep7Controller = router.get('wizardStep7Controller');
        addServiceController.saveServiceConfigProperties(wizardStep7Controller);
        if (App.supports.hostOverrides) {
          addServiceController.saveServiceConfigGroups(wizardStep7Controller);
        }
        router.transitionTo('step5');
      }
    }),

    step5: Em.Route.extend({
      route: '/step5',
      connectOutlets: function (router, context) {
        console.log('in addService.step5:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('5');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          var wizardStep8Controller = router.get('wizardStep8Controller');
          wizardStep8Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep8', controller.get('content'));
        })
      },
      back: function(router){
        var controller = router.get('addServiceController');
        if(!controller.get('content.skipConfigStep')) {
          router.transitionTo('step4');
        }
        else {
          if(!controller.get('content.skipSlavesStep')) {
            router.transitionTo('step3');
          }
          else {
            if(!controller.get('content.skipMasterStep')) {
              router.transitionTo('step2');
            }
            else {
              router.transitionTo('step1');
            }
          }
        }
      },
      next: function (router) {
        var addServiceController = router.get('addServiceController');
        var wizardStep8Controller = router.get('wizardStep8Controller');
        addServiceController.installServices();
        addServiceController.setInfoForStep9();

        addServiceController.saveClusterState('ADD_SERVICES_INSTALLING_3');
        wizardStep8Controller.set('servicesInstalled', true);
        router.transitionTo('step6');
      }
    }),

    step6: Em.Route.extend({
      route: '/step6',
      connectOutlets: function (router, context) {
        console.log('in addService.step6:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('6');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          var wizardStep9Controller = router.get('wizardStep9Controller');
          wizardStep9Controller.set('wizardController', controller);
          if (!App.testMode) {              //if test mode is ON don't disable prior steps link.
            controller.setLowerStepsDisable(6);
          }
          controller.connectOutlet('wizardStep9', controller.get('content'));
        })
      },
      back: Em.Router.transitionTo('step5'),
      retry: function(router,context) {
        var addServiceController = router.get('addServiceController');
        var wizardStep9Controller = router.get('wizardStep9Controller');
        if (wizardStep9Controller.get('showRetry')) {
          if (wizardStep9Controller.get('content.cluster.status') === 'INSTALL FAILED') {
            var isRetry = true;
            addServiceController.installServices(isRetry);
            addServiceController.setInfoForStep9();
            wizardStep9Controller.resetHostsForRetry();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            addServiceController.saveClusterState('ADD_SERVICES_INSTALLING_3');
          }
          wizardStep9Controller.navigateStep();
        }
      },
      unroutePath: function() {
        return false;
      },
      next: function (router) {
        var addServiceController = router.get('addServiceController');
        var wizardStep9Controller = router.get('wizardStep9Controller');
        addServiceController.saveInstalledHosts(wizardStep9Controller);

        // We need to do recovery based on whether we are in Add Host or Installer wizard
        addServiceController.saveClusterState('ADD_SERVICES_INSTALLED_4');

        router.transitionTo('step7');
      }
    }),

    step7: Em.Route.extend({
      route: '/step7',
      connectOutlets: function (router, context) {
        console.log('in addService.step7:connectOutlets');
        var controller = router.get('addServiceController');
        controller.setCurrentStep('7');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.setLowerStepsDisable(7);
          controller.connectOutlet('wizardStep10', controller.get('content'));
        })
      },
      back: Em.Router.transitionTo('step6'),
      complete: function (router, context) {
        var addServiceController = router.get('addServiceController');
        addServiceController.get('popup').onClose();
        addServiceController.finish();
        // We need to do recovery based on whether we are in Add Host or Installer wizard
        addServiceController.saveClusterState('DEFAULT');
        location.reload();
      }
    }),

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3'),

    gotoStep4: Em.Router.transitionTo('step4'),

    gotoStep5: Em.Router.transitionTo('step5'),

    gotoStep6: Em.Router.transitionTo('step6'),

    gotoStep7: Em.Router.transitionTo('step7'),

    backToServices: function (router) {
      App.router.get('updateController').set('isWorking', true);
      router.transitionTo('services');
    }

  });
  
});
window.require.register("routes/high_availability_routes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: '/highAvailability/enable',

    enter: function (router) {
      Em.run.next(function () {
        var highAvailabilityWizardController = router.get('highAvailabilityWizardController');
        App.router.get('updateController').set('isWorking', false);
        var popup = App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header: Em.I18n.t('admin.highAvailability.wizard.header'),
          bodyClass: App.HighAvailabilityWizardView.extend({
            controller: highAvailabilityWizardController
          }),
          primary: Em.I18n.t('form.cancel'),
          showFooter: false,
          secondary: null,
          hideCloseButton: function () {
            var currStep = App.router.get('highAvailabilityWizardController.currentStep');
            switch (currStep) {
              case "5" :
              case "7" :
              case "9" :
                if(App.supports.autoRollbackHA){
                  this.set('showCloseButton', false);
                }else{
                  this.set('showCloseButton', true);
                }
                break;
              default :
                this.set('showCloseButton', true);
            }
          }.observes('App.router.highAvailabilityWizardController.currentStep'),

          onClose: function () {
            var currStep = App.router.get('highAvailabilityWizardController.currentStep');
            var highAvailabilityProgressPageController = App.router.get('highAvailabilityProgressPageController');

            if(parseInt(currStep) > 4){
              if(!App.supports.autoRollbackHA){
                highAvailabilityProgressPageController.manualRollback();
              }else{
                this.hide();
                App.router.get('highAvailabilityWizardController').setCurrentStep('1');
                App.router.transitionTo('rollbackHighAvailability');
              }
            }else {
              var controller = App.router.get('highAvailabilityWizardController');
              controller.clearStorageData();
              controller.setCurrentStep('1');
              App.router.get('updateController').set('isWorking', true);
              App.clusterStatus.setClusterStatus({
                clusterName: App.router.get('content.cluster.name'),
                clusterState: 'DEFAULT',
                wizardControllerName: App.router.get('highAvailabilityWizardController.name'),
                localdb: App.db.data
              });
              this.hide();
              App.router.transitionTo('main.admin.adminHighAvailability');
            }

          },
          didInsertElement: function () {
            this.fitHeight();
          }
        });
        highAvailabilityWizardController.set('popup', popup);
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus) {
          switch (currentClusterStatus.clusterState) {
            case 'HIGH_AVAILABILITY_DEPLOY' :
              highAvailabilityWizardController.setCurrentStep(currentClusterStatus.localdb.HighAvailabilityWizard.currentStep);
              break;
            default:
              var currStep = App.router.get('highAvailabilityWizardController.currentStep');
              highAvailabilityWizardController.setCurrentStep(currStep);
              break;
          }
        }
        router.transitionTo('step' + highAvailabilityWizardController.get('currentStep'));
      });
    },

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.usersLoading().done(function () {
          controller.saveHdfsUser();
          controller.setCurrentStep('1');
          controller.dataLoading().done(function () {
            controller.loadAllPriorSteps();
            controller.connectOutlet('highAvailabilityWizardStep1', controller.get('content'));
          })
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.saveNameServiceId(router.get('highAvailabilityWizardStep1Controller.content.nameServiceId'));
        router.transitionTo('step2');
      }
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('2');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep2', controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        var highAvailabilityWizardStep2Controller = router.get('highAvailabilityWizardStep2Controller');
        var addNN = highAvailabilityWizardStep2Controller.get('selectedServicesMasters').findProperty('isAddNameNode', true).get('selectedHost');
        var sNN = highAvailabilityWizardStep2Controller.get('selectedServicesMasters').findProperty('component_name','SECONDARY_NAMENODE').get('selectedHost');
        if(addNN){
          App.db.setRollBackHighAvailabilityWizardAddNNHost(addNN);
        }
        if(sNN){
          App.db.setRollBackHighAvailabilityWizardSNNHost(sNN);
        }
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('content.cluster.name'),
          clusterState: 'HIGH_AVAILABILITY_DEPLOY',
          wizardControllerName: this.get('content.controllerName'),
          localdb: App.db.data
        });
        controller.saveMasterComponentHosts(highAvailabilityWizardStep2Controller);
        controller.get('content').set('serviceConfigProperties', null);
        controller.setDBProperty('serviceConfigProperties', null);
        router.transitionTo('step3');
      },
      back: function (router) {
        router.transitionTo('step1');
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('3');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep3',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        var stepController = router.get('highAvailabilityWizardStep3Controller');
        controller.saveServiceConfigProperties(stepController);
        controller.saveConfigTag(stepController.get("hdfsSiteTag"));
        controller.saveConfigTag(stepController.get("coreSiteTag"));
        if (App.Service.find().someProperty('serviceName', 'HBASE')) {
          controller.saveConfigTag(stepController.get("hbaseSiteTag"));
        }
        router.transitionTo('step4');
      },
      back: Em.Router.transitionTo('step2')
    }),

    step4: Em.Route.extend({
      route: '/step4',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('4');
        controller.setLowerStepsDisable(4);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep4',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.clearTasksData();
        router.transitionTo('step5');
      }
    }),

    step5: Em.Route.extend({
      route: '/step5',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('5');
        controller.setLowerStepsDisable(5);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep5',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.clearTasksData();
        router.transitionTo('step6');
      }
    }),

    step6: Em.Route.extend({
      route: '/step6',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('6');
        controller.setLowerStepsDisable(6);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep6',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        router.transitionTo('step7');
      }
    }),

    step7: Em.Route.extend({
      route: '/step7',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('7');
        controller.setLowerStepsDisable(7);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep7',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.clearTasksData();
        router.transitionTo('step8');
      }
    }),

    step8: Em.Route.extend({
      route: '/step8',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('8');
        controller.setLowerStepsDisable(8);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep8',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        App.showConfirmationPopup(function() {
          router.transitionTo('step9');
        }, Em.I18n.t('admin.highAvailability.wizard.step8.confirmPopup.body'));
      }
    }),

    step9: Em.Route.extend({
      route: '/step9',
      connectOutlets: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.setCurrentStep('9');
        controller.setLowerStepsDisable(9);
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('highAvailabilityWizardStep9',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('highAvailabilityWizardController');
        controller.clearTasksData();
        controller.finish();
        controller.get('popup').hide();
        App.clusterStatus.setClusterStatus({
          clusterName: controller.get('content.cluster.name'),
          clusterState: 'DEFAULT',
          wizardControllerName: 'highAvailabilityWizardController',
          localdb: App.db.data
        });
        router.transitionTo('main.index');
        location.reload();
      }
    }),

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3'),

    gotoStep4: Em.Router.transitionTo('step4'),

    gotoStep5: Em.Router.transitionTo('step5'),

    gotoStep6: Em.Router.transitionTo('step6'),

    gotoStep7: Em.Router.transitionTo('step7'),

    gotoStep8: Em.Router.transitionTo('step8'),

    gotoStep9: Em.Router.transitionTo('step9')
  });
  
});
window.require.register("routes/installer", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = Em.Route.extend({
    route: '/installer',
    App: require('app'),

    enter: function (router) {
      console.log('in /installer:enter');

      App.clusterStatus.set('wizardControllerName',App.router.get('installerController.name'));

      if (router.getAuthenticated()) {
        // check server/web client versions match
        App.router.get('installerController').checkServerClientVersion().done(function() {

          var name = 'Cluster Install Wizard';
          $('title').text('Ambari - ' + name);

          if (App.get('isAdmin')) {
            router.get('mainController').stopPolling();
            console.log('In installer with successful authenticated');
            console.log('current step=' + router.get('installerController.currentStep'));
            Ember.run.next(function () {
              var installerController = router.get('installerController');

              App.clusterStatus.updateFromServer();
              var currentClusterStatus = App.clusterStatus.get('value');

              if (currentClusterStatus) {
                switch (currentClusterStatus.clusterState) {
                  case 'CLUSTER_NOT_CREATED_1' :
                    router.transitionTo('step' + installerController.get('currentStep'));
                    break;
                  case 'CLUSTER_DEPLOY_PREP_2' :
                    installerController.setCurrentStep('8');
                    App.db.data = currentClusterStatus.localdb;
                    router.transitionTo('step' + installerController.get('currentStep'));
                    break;
                  case 'CLUSTER_INSTALLING_3' :
                  case 'SERVICE_STARTING_3' :
                    if (!installerController.get('isStep9')) {
                      installerController.setCurrentStep('9');
                    }
                    router.transitionTo('step' + installerController.get('currentStep'));
                    break;
                  case 'CLUSTER_INSTALLED_4' :
                    if (!installerController.get('isStep10')) {
                      installerController.setCurrentStep('10');
                    }
                    App.db.data = currentClusterStatus.localdb;
                    router.transitionTo('step' + installerController.get('currentStep'));
                    break;
                  case 'DEFAULT' :
                  default:
                    router.transitionTo('main.dashboard');
                    break;
                }
              }
            });
          } else {
            Em.run.next(function () {
              App.router.transitionTo('main.services');
            });
          }

        });
      } else {
        console.log('In installer but its not authenticated');
        console.log('value of authenticated is: ' + router.getAuthenticated());
        Ember.run.next(function () {
          router.transitionTo('login');
        });
      }
    },

    routePath: function (router, event) {
      console.log("INFO: value of router is: " + router);
      console.log("INFO: value of event is: " + event);
      router.setNavigationFlow(event);
      if (!router.isFwdNavigation) {
        this._super(router, event);
      } else {
        router.set('backBtnForHigherStep', true);

        var installerController = router.get('installerController');
        router.transitionTo('step' + installerController.get('currentStep'));
      }
    },

    connectOutlets: function (router, context) {
      console.log('in /installer:connectOutlets');
      router.get('applicationController').connectOutlet('installer');
    },

    step0: Em.Route.extend({
      route: '/step0',
      connectOutlets: function (router) {
        console.log('in installer.step0:connectOutlets');
        var controller = router.get('installerController');
        controller.setCurrentStep('0');
        controller.loadAllPriorSteps();
        controller.connectOutlet('wizardStep0', controller.get('content'));
      },

      next: function (router) {
        var installerController = router.get('installerController');
        installerController.save('cluster');
        router.transitionTo('step1');
      }
    }),

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        console.log('in installer.step1:connectOutlets');
        var controller = router.get('installerController');
        controller.setCurrentStep('1');
        controller.loadAllPriorSteps();
        controller.connectOutlet('wizardStep1', controller.get('content'));
      },
      back: Em.Router.transitionTo('step0'),
      next: function (router) {
        var wizardStep1Controller = router.get('wizardStep1Controller');
        var installerController = router.get('installerController');
        if (App.testMode) {
          installerController.set('validationCnt', 0);
          installerController.set('invalidCnt', 0);
        } else {
          installerController.checkRepoURL(wizardStep1Controller);
        }
        // make sure got all validations feedback and no invalid url, then proceed
        var myVar = setInterval(
          function(){
            var cnt = installerController.get('validationCnt');
            var invalidCnt = installerController.get('invalidCnt');
            if (cnt == 0 && invalidCnt == 0) { // all feedback exist and no invalid url
              installerController.saveStacks(wizardStep1Controller);
              installerController.setDBProperty('service', undefined);
              installerController.clearInstallOptions();
              router.transitionTo('step2');
              clearInterval(myVar);
            } else if ( cnt == 0 && invalidCnt != 0) {
              clearInterval(myVar);
            }
          },
        1000);
      }
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router, context) {
        router.setNavigationFlow('step2');

        var controller = router.get('installerController');
        controller.setCurrentStep('2');
        controller.loadAllPriorSteps();
        controller.connectOutlet('wizardStep2', controller.get('content'));
      },
      back: Em.Router.transitionTo('step1'),
      next: function (router) {
        var controller = router.get('installerController');
        controller.save('installOptions');
        //hosts was saved to content.hosts inside wizardStep2Controller
        controller.save('hosts');
        router.transitionTo('step3');
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router) {
        console.log('in installer.step3:connectOutlets');
        var controller = router.get('installerController');
        controller.setCurrentStep('3');
        controller.loadAllPriorSteps();
        var wizardStep3Controller = router.get('wizardStep3Controller');
        wizardStep3Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep3', controller.get('content'));
      },
      back: function(router){
          router.transitionTo('step2');
      },
      next: function (router, context) {
        var installerController = router.get('installerController');
        var wizardStep3Controller = router.get('wizardStep3Controller');
        installerController.saveConfirmedHosts(wizardStep3Controller);
        installerController.setDBProperty('bootStatus', true);
        installerController.loadServicesFromServer();
        router.transitionTo('step4');
      },
      exit: function (router) {
        router.get('wizardStep3Controller').set('stopBootstrap', true);
      },
      /**
       * Wrapper for remove host action.
       * Since saving data stored in installerController, we should call this from router
       * @param router
       * @param context Array of hosts to delete
       */
      removeHosts: function (router, context) {
        console.log('in installer.step2.removeHosts:hosts to delete ', context);
        var controller = router.get('installerController');
        controller.removeHosts(context);
      }
    }),

    step4: Em.Route.extend({
      route: '/step4',
      connectOutlets: function (router, context) {
        router.setNavigationFlow('step4');
        var controller = router.get('installerController');
        controller.setCurrentStep('4');
        controller.loadAllPriorSteps();
        controller.connectOutlet('wizardStep4', controller.get('content.services'));
      },
      back: Em.Router.transitionTo('step3'),

      next: function (router) {
        var controller = router.get('installerController');
        var wizardStep4Controller = router.get('wizardStep4Controller');
        controller.saveServices(wizardStep4Controller);
        controller.saveClients(wizardStep4Controller);

        controller.setDBProperty('masterComponentHosts', undefined);
        router.transitionTo('step5');
      }
    }),

    step5: Em.Route.extend({
      route: '/step5',
      connectOutlets: function (router, context) {
        router.setNavigationFlow('step5');

        var controller = router.get('installerController');
        router.get('wizardStep5Controller').set('servicesMasters', []);
        controller.setCurrentStep('5');
        controller.loadAllPriorSteps();
        controller.connectOutlet('wizardStep5', controller.get('content'));
      },
      back: Em.Router.transitionTo('step4'),
      next: function (router) {
        var controller = router.get('installerController');
        var wizardStep5Controller = router.get('wizardStep5Controller');
        controller.saveMasterComponentHosts(wizardStep5Controller);
        controller.setDBProperty('slaveComponentHosts', undefined);
        router.transitionTo('step6');
      }
    }),

    step6: Em.Route.extend({
      route: '/step6',
      connectOutlets: function (router, context) {
        router.setNavigationFlow('step6');

        var controller = router.get('installerController');
        router.get('wizardStep6Controller').set('hosts', []);
        controller.setCurrentStep('6');
        controller.loadAllPriorSteps();
        controller.connectOutlet('wizardStep6', controller.get('content'));
      },
      back: Em.Router.transitionTo('step5'),

      next: function (router) {
        var controller = router.get('installerController');
        var wizardStep6Controller = router.get('wizardStep6Controller');
        var wizardStep7Controller = router.get('wizardStep7Controller');

        if (wizardStep6Controller.validate()) {
          controller.saveSlaveComponentHosts(wizardStep6Controller);
          controller.get('content').set('serviceConfigProperties', null);
          controller.setDBProperty('serviceConfigProperties', null);
          controller.setDBProperty('advancedServiceConfig', null);
          controller.setDBProperty('serviceConfigGroups', null);
          controller.loadAdvancedConfigs(wizardStep7Controller);
          router.transitionTo('step7');
        }
      }
    }),

    step7: Em.Route.extend({
      route: '/step7',
      enter: function (router) {
        console.log('in /wizardStep7Controller:enter');
        var controller = router.get('installerController');
        controller.setCurrentStep('7');
        controller.loadAllPriorSteps();
      },
      connectOutlets: function (router, context) {
        var controller = router.get('installerController');
        var wizardStep7Controller = router.get('wizardStep7Controller');
        wizardStep7Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep7', controller.get('content'));
      },
      back: Em.Router.transitionTo('step6'),
      next: function (router) {
        var installerController = router.get('installerController');
        var wizardStep7Controller = router.get('wizardStep7Controller');
        installerController.saveServiceConfigProperties(wizardStep7Controller);
        if (App.supports.hostOverridesInstaller) {
          installerController.saveServiceConfigGroups(wizardStep7Controller);
        }
        router.transitionTo('step8');
      }
    }),

    step8: Em.Route.extend({
      route: '/step8',
      connectOutlets: function (router, context) {
        console.log('in installer.step8:connectOutlets');
        var controller = router.get('installerController');
        controller.setCurrentStep('8');
        controller.loadAllPriorSteps();
        var wizardStep8Controller = router.get('wizardStep8Controller');
        wizardStep8Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep8', controller.get('content'));
      },
      back: Em.Router.transitionTo('step7'),
      next: function (router) {
        var installerController = router.get('installerController');
        var wizardStep8Controller = router.get('wizardStep8Controller');
        // invoke API call to install selected services
        installerController.installServices();
        installerController.setInfoForStep9();
        // We need to do recovery based on whether we are in Add Host or Installer wizard
        installerController.saveClusterState('CLUSTER_INSTALLING_3');
        wizardStep8Controller.set('servicesInstalled', true);
        router.transitionTo('step9');
      }
    }),

    step9: Em.Route.extend({
      route: '/step9',
      connectOutlets: function (router, context) {
        console.log('in installer.step9:connectOutlets');
        var controller = router.get('installerController');
        controller.setCurrentStep('9');
        controller.loadAllPriorSteps();
        if (!App.testMode) {
          controller.setLowerStepsDisable(9);
        }
        var wizardStep9Controller = router.get('wizardStep9Controller');
        wizardStep9Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep9', controller.get('content'));
      },
      back: Em.Router.transitionTo('step8'),
      retry: function (router) {
        var installerController = router.get('installerController');
        var wizardStep9Controller = router.get('wizardStep9Controller');
        if (wizardStep9Controller.get('showRetry')) {
          if (wizardStep9Controller.get('content.cluster.status') === 'INSTALL FAILED') {
            var isRetry = true;
            installerController.installServices(isRetry);
            installerController.setInfoForStep9();
            wizardStep9Controller.resetHostsForRetry();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            installerController.saveClusterState('CLUSTER_INSTALLING_3');
          }
          wizardStep9Controller.navigateStep();
        }
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var installerController = router.get('installerController');
        var wizardStep9Controller = router.get('wizardStep9Controller');
        installerController.saveInstalledHosts(wizardStep9Controller);

        installerController.saveClusterState('CLUSTER_INSTALLED_4');
        router.transitionTo('step10');
      }
    }),

    step10: Em.Route.extend({
      route: '/step10',
      connectOutlets: function (router, context) {
        console.log('in installer.step10:connectOutlets');
        var controller = router.get('installerController');
        controller.setCurrentStep('10');
        controller.loadAllPriorSteps();
        if (!App.testMode) {
          controller.setLowerStepsDisable(10);
        }
        controller.connectOutlet('wizardStep10', controller.get('content'));
      },
      back: Em.Router.transitionTo('step9'),
      complete: function (router, context) {
        var controller = router.get('installerController');
        controller.finish();

        // We need to do recovery based on whether we are in Add Host or Installer wizard
        controller.saveClusterState('DEFAULT');

        router.transitionTo('main.dashboard');
      }
    }),

    gotoStep0: Em.Router.transitionTo('step0'),

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3'),

    gotoStep4: Em.Router.transitionTo('step4'),

    gotoStep5: Em.Router.transitionTo('step5'),

    gotoStep6: Em.Router.transitionTo('step6'),

    gotoStep7: Em.Router.transitionTo('step7'),

    gotoStep8: Em.Router.transitionTo('step8'),

    gotoStep9: Em.Router.transitionTo('step9'),

    gotoStep10: Em.Router.transitionTo('step10')

  });
  
});
window.require.register("routes/main", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  module.exports = Em.Route.extend({
    route: '/main',
    enter: function (router) {
      App.db.updateStorage();
        console.log('in /main:enter');
        if (router.getAuthenticated()) {
          App.router.get('mainAdminAccessController').loadShowJobsForUsers().done(function() {
            App.router.get('clusterController').loadClusterName(false);
            if(App.testMode) {
              router.get('mainController').initialize();
            }else {
              App.router.get('mainController').checkServerClientVersion().done(function() {
                App.router.get('clusterController').loadClientServerClockDistance().done(function() {
                  router.get('mainController').initialize();
                });
              });
            }
          });
          // TODO: redirect to last known state
        } else {
          Em.run.next(function () {
            router.transitionTo('login');
          });
        }
    },
    /*
     routePath: function(router,event) {
     if (router.getAuthenticated()) {
     App.router.get('clusterController').loadClusterName(false);
     router.get('mainController').initialize();
     // TODO: redirect to last known state
     } else {
     Ember.run.next(function () {
     router.transitionTo('login');
     });
     }
     }, */

    index: Ember.Route.extend({
      route: '/',
      redirectsTo: 'dashboard'
    }),


    views: Ember.Route.extend({
      route: '/views',
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainViews');
      }

    }),
    test: Em.Route.extend({
      route: '/test',
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainTest');
      }
    }),

    connectOutlets: function (router, context) {
      router.get('applicationController').connectOutlet('main');
    },

    charts: Em.Route.extend({
      route: '/charts',
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainCharts');
      },
      enter: function (router) {
        Em.run.next(function () {
          router.transitionTo('heatmap');
        });
      },
      index: Ember.Route.extend({
        route: '/',
        redirectsTo: 'heatmap'
      }),
      heatmap: Em.Route.extend({
        route: '/heatmap',
        connectOutlets: function (router, context) {
          router.get('mainChartsController').connectOutlet('mainChartsHeatmap');
        }
      }),
      horizon_chart: Em.Route.extend({
        route: '/horizon_chart',
        connectOutlets: function (router, context) {
          router.get('mainChartsController').connectOutlet('mainChartsHorizon');
        }
      }),
      showChart: function (router, event) {
        var parent = event.view._parentView;
        parent.deactivateChildViews();
        event.view.set('active', "active");
        router.transitionTo(event.context);
      }
    }),

    apps: Em.Route.extend({
      route: '/apps',
      connectOutlets: function (router) {
        if (App.get('isHadoop2Stack')) {
          Em.run.next(function () {
            router.transitionTo('main.dashboard');
          });
        } else {
          router.get('mainAppsController').loadRuns();
          router.get('mainController').connectOutlet('mainApps');
        }
      }
    }),

    jobs : Em.Route.extend({
      route : '/jobs',
      enter: function (router) {
        if(!App.router.get('mainAdminAccessController.showJobs') && !App.get('isAdmin')){
          Em.run.next(function () {
            router.transitionTo('main.dashboard');
          });
        }
      },
      exit: function(router) {
        clearInterval(router.get('mainJobsController').jobsUpdate);
      },
      index: Ember.Route.extend({
        route: '/',
        connectOutlets : function(router) {
          if (!App.get('isHadoop2Stack')) {
            Em.run.next(function() {
              router.transitionTo('main.dashboard');
            });
          } else {
            router.get('mainJobsController').loadJobs();
            router.get('mainJobsController').updateJobs('mainJobsController', 'refreshLoadedJobs');
            router.get('mainController').connectOutlet('mainJobs');
          }
        }
      }),
      jobDetails : Em.Route.extend({
        route : '/:job_id',
        connectOutlets : function(router, job) {
          if (job) {
            router.get('mainHiveJobDetailsController').set('loaded', false);
            router.get('mainController').connectOutlet('mainHiveJobDetails', job);
            router.get('mainHiveJobDetailsController').loadJobDetails();
            router.get('mainJobsController').updateJobs('mainHiveJobDetailsController', 'loadJobDetails');
          }
        },
        exit: function(router) {
          router.get('mainHiveJobDetailsController').set('loaded', false);
        }
      })
    }),

    mirroring: Em.Route.extend({
      route: '/mirroring',
      index: Ember.Route.extend({
        route: '/'
      }),

      connectOutlets: function (router) {
        router.get('mainController').connectOutlet('mainMirroring');
      },

      gotoShowJobs: function (router, context) {
        var dataset = context || router.get('mainMirroringController.selectedDataset') || App.Dataset.find().objectAt(0);
        if (dataset) {
          router.transitionTo('showDatasetJobs', dataset);
        } else {
          router.transitionTo('index');
        }
      },

      showDatasetJobs: Em.Route.extend({
        route: '/:dataset_id',
        connectOutlets: function (router, dataset) {
          router.get('mainDatasetJobsController').set('content', dataset);
          router.get('mainMirroringController').set('selectedDataset', dataset);
        }
      }),

      editDatasetRoute: Em.Route.extend({
        route: '/edit/:dataset_id',
        connectOutlets: function (router, dataset) {
          router.get('mainMirroringEditDataSetController').showEditPopup(dataset);
        }
      }),

      editDataset: function (router, event) {
        router.transitionTo('editDatasetRoute', event.view.get('dataset'));
      },

      addNewDataset: function (router) {
        router.transitionTo('addNewDatasetRoute');
      },

      addNewDatasetRoute: Em.Route.extend({
        route: '/dataset/add',
        enter: function (router) {
          var controller = router.get('mainMirroringEditDataSetController');
          controller.showAddPopup();
        }
      }),

      manageClustersRoute: Em.Route.extend({
        route: '/dataset/clusters/edit',
        enter: function (router) {
          var controller = router.get('mainMirroringController');
          controller.manageClusters();
        }
      }),

      manageClusters: function (router) {
        router.transitionTo('manageClustersRoute');
      }
    }),


    hosts: Em.Route.extend({
      route: '/hosts',
      index: Ember.Route.extend({
        route: '/',
        connectOutlets: function (router, context) {
          router.get('mainController').connectOutlet('mainHost');
        }
      }),

      hostDetails: Em.Route.extend({
        route: '/:host_id',
        connectOutlets: function (router, host) {
          router.get('mainController').connectOutlet('mainHostDetails', host);
        },

        index: Ember.Route.extend({
          route: '/',
          redirectsTo: 'summary'
        }),

        summary: Em.Route.extend({
          route: '/summary',
          connectOutlets: function (router, context) {
            router.get('mainHostDetailsController').connectOutlet('mainHostSummary');
          }
        }),

        configs: Em.Route.extend({
          route: '/configs',
          connectOutlets: function (router, context) {
            router.get('mainHostDetailsController').connectOutlet('mainHostConfigs');
          }
        }),

        metrics: Em.Route.extend({
          route: '/metrics',
          connectOutlets: function (router, context) {
            router.get('mainHostDetailsController').connectOutlet('mainHostMetrics');
          }
        }),

        audit: Em.Route.extend({
          route: '/audit',
          connectOutlets: function (router, context) {
            router.get('mainHostDetailsController').connectOutlet('mainHostAudit');
          }
        }),

        hostNavigate: function (router, event) {
          var parent = event.view._parentView;
          parent.deactivateChildViews();
          event.view.set('active', "active");
          router.transitionTo(event.context);
        }
      }),

      back: function (router, event) {
        var referer = router.get('mainHostDetailsController.referer');
        if (referer) {
          router.route(referer);
        }
        else {
          window.history.back();
        }
      },

      addHost: function (router) {
        router.transitionTo('hostAdd');
      }

    }),

    hostAdd: require('routes/add_host_routes'),

    admin: Em.Route.extend({
      route: '/admin',
      enter: function (router, transition) {
        if (!App.isAdmin) {
          Em.run.next(function () {
            router.transitionTo('main.dashboard');
          });
        }
      },

      routePath: function (router, event) {
        if (!App.isAdmin) {
          Em.run.next(function () {
            App.router.transitionTo('main.dashboard');
          });
        } else {
          var controller = router.get('mainAdminController');
          router.transitionTo('admin' + controller.get('category').capitalize());
        }
      },
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainAdmin');
      },

      index: Em.Route.extend({
        /* enter: function(router, transition){
         var controller = router.get('mainAdminController');
         router.transitionTo('admin' + controller.get('category').capitalize());
         }, */
        route: '/',
        redirectsTo: 'adminUser'
      }),


      adminUser: Em.Route.extend({
        route: '/user',
        index: Em.Route.extend({
          route: '/',
          redirectsTo: 'allUsers'
        }),
        enter: function (router) {
          router.set('mainAdminController.category', "user");
          Em.run.next(function () {
            router.transitionTo('allUsers');
          });
        },
        routePath: function (router, event) {
          router.set('mainAdminController.category', "user");
          router.transitionTo('allUsers');
          Em.run.next(function () {
            router.transitionTo('allUsers');
          });
        },
        // events
        gotoUsers: Em.Router.transitionTo("allUsers"),
        gotoCreateUser: Em.Router.transitionTo("createUser"),
        gotoEditUser: function (router, event) {
          router.transitionTo("editUser", event.context)
        },

        // states
        allUsers: Em.Route.extend({
          route: '/allUsers',
          // index: Ember.Route.extend({
          //route: '/',
          connectOutlets: function (router) {
            router.get('mainAdminController').connectOutlet('mainAdminUser');
          }
          //})
        }),

        createUser: Em.Route.extend({
          route: '/create',
          connectOutlets: function (router) {
            router.get('mainAdminController').connectOutlet('mainAdminUserCreate', {});
          }
        }),

        editUser: Em.Route.extend({
          route: '/edit/:user_id',
          connectOutlets: function (router, user) {
            router.get('mainAdminController').connectOutlet('mainAdminUserEdit', user);
          }
        })
      }),


      adminAuthentication: Em.Route.extend({
        route: '/authentication',
        connectOutlets: function (router, context) {
          router.set('mainAdminController.category', "authentication");
          router.get('mainAdminController').connectOutlet('mainAdminAuthentication');
        }
      }),

      adminHighAvailability: Em.Route.extend({
        route: '/highAvailability',
        enter: function (router) {
          Em.run.next(function () {
            router.transitionTo('adminHighAvailability.index');
          });
        },
        index: Ember.Route.extend({
          route: '/',
          connectOutlets: function (router, context) {
            router.set('mainAdminController.category', "highAvailability");
            router.get('mainAdminController').connectOutlet('mainAdminHighAvailability');
          }
        })
      }),

      enableHighAvailability: require('routes/high_availability_routes'),

      rollbackHighAvailability: require('routes/rollbackHA_routes'),



      adminSecurity: Em.Route.extend({
        route: '/security',
        enter: function (router) {
          router.set('mainAdminController.category', "security");
          var controller = router.get('mainAdminSecurityController');
          if (!(controller.getAddSecurityWizardStatus() === 'RUNNING') && !(controller.getDisableSecurityStatus() === 'RUNNING')) {
            Em.run.next(function () {
              router.transitionTo('adminSecurity.index');
            });
          } else if (controller.getAddSecurityWizardStatus() === 'RUNNING') {
            Em.run.next(function () {
              router.transitionTo('adminAddSecurity');
            });
          } else if (controller.getDisableSecurityStatus() === 'RUNNING') {
            Em.run.next(function () {
              router.transitionTo('disableSecurity');
            });
          }
        },

        index: Ember.Route.extend({
          route: '/',
          connectOutlets: function (router, context) {
            var controller = router.get('mainAdminController');
            controller.set('category', "security");
            controller.connectOutlet('mainAdminSecurity');
          }
        }),

        addSecurity: function (router, object) {
          router.get('mainAdminSecurityController').setAddSecurityWizardStatus('RUNNING');
          router.transitionTo('adminAddSecurity');
        },

        disableSecurity: Ember.Route.extend({
          route: '/disableSecurity',
          enter: function (router) {
            //after refresh check if the wizard is open then restore it
            if (router.get('mainAdminSecurityController').getDisableSecurityStatus() === 'RUNNING') {
              Ember.run.next(function () {
                App.router.get('updateController').set('isWorking', false);
                App.ModalPopup.show({
                  classNames: ['full-width-modal'],
                  header: Em.I18n.t('admin.removeSecurity.header'),
                  bodyClass: App.MainAdminSecurityDisableView.extend({
                    controllerBinding: 'App.router.mainAdminSecurityDisableController'
                  }),
                  primary: Em.I18n.t('form.cancel'),
                  secondary: null,
                  showFooter: false,

                  onClose: function () {
                    var self = this;
                    var controller = router.get('mainAdminSecurityDisableController');
                    if (!controller.get('isSubmitDisabled')) {
                      self.proceedOnClose();
                      return;
                    }
                    var applyingConfigCommand = controller.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
                    if (applyingConfigCommand && !applyingConfigCommand.get('isCompleted')) {
                      if (applyingConfigCommand.get('isStarted')) {
                        App.showAlertPopup(Em.I18n.t('admin.security.applying.config.header'), Em.I18n.t('admin.security.applying.config.body'));
                      } else {
                        App.showConfirmationPopup(function () {
                          self.proceedOnClose();
                        }, Em.I18n.t('admin.addSecurity.disable.onClose'));
                      }
                    } else {
                      self.proceedOnClose();
                    }
                  },
                  proceedOnClose: function () {
                    router.get('mainAdminSecurityDisableController').clearStep();
                    App.db.setSecurityDeployCommands(undefined);
                    App.router.get('updateController').set('isWorking', true);
                    router.get('mainAdminSecurityController').setDisableSecurityStatus(undefined);
                    App.clusterStatus.setClusterStatus({
                      clusterName: router.get('content.cluster.name'),
                      clusterState: 'DEFAULT'
                    });
                    this.hide();
                    router.transitionTo('adminSecurity.index');
                  },
                  didInsertElement: function () {
                    this.fitHeight();
                  }
                });
              });
            } else {
              router.transitionTo('adminSecurity.index');
            }
          },

          unroutePath: function () {
            return false;
          },

          done: function (router, context) {
            var controller = router.get('mainAdminSecurityDisableController');
            if (!controller.get('isSubmitDisabled')) {
              $(context.currentTarget).parents("#modal").find(".close").trigger('click');
            }
          }
        }),

        adminAddSecurity: require('routes/add_security')
      }),

      adminCluster: Em.Route.extend({
        route: '/cluster',
        connectOutlets: function (router) {
          router.set('mainAdminController.category', "cluster");
          router.get('mainAdminController').connectOutlet('mainAdminCluster');
        }
      }),
      adminAdvanced: Em.Route.extend({
        route: '/advanced',
        connectOutlets: function (router) {
          router.set('mainAdminController.category', "advanced");
          router.get('mainAdminController').connectOutlet('mainAdminAdvanced');
        }
      }),
      adminMisc: Em.Route.extend({
        route: '/misc',
        connectOutlets: function (router) {
          router.set('mainAdminController.category', "misc");
          router.get('mainAdminController').connectOutlet('mainAdminMisc');
        }
      }),
      adminAccess: Em.Route.extend({
        enter: function(router) {
          router.get('mainController').dataLoading().done(function() {
            if (!router.get('mainAdminController.isAccessAvailable')) router.transitionTo('adminUser.allUsers');
          });
        },
        route: '/access',
        connectOutlets: function (router) {
          router.set('mainAdminController.category', "access");
          router.get('mainAdminController').connectOutlet('mainAdminAccess');
        }
      }),

      adminAudit: Em.Route.extend({
        route: '/audit',
        connectOutlets: function (router) {
          router.set('mainAdminController.category', "audit");
          router.get('mainAdminController').connectOutlet('mainAdminAudit');
        }
      }),
      upgradeStack: function (router, event) {
        if (!$(event.currentTarget).hasClass('inactive')) {
          router.transitionTo('stackUpgrade');
        }
      },


      adminNavigate: function (router, object) {
        router.transitionTo('admin' + object.context.capitalize());
      },

  //events
      goToAdmin: function (router, event) {
        router.transitionTo(event.context);
      }

    }),
    stackUpgrade: require('routes/stack_upgrade'),

    dashboard: Em.Route.extend({
      route: '/dashboard',
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainDashboard');
      }
    }),

    services: Em.Route.extend({
      route: '/services',
      index: Em.Route.extend({
        route: '/',
        enter: function (router) {
          Em.run.next(function () {
            var controller = router.get('mainController');
            controller.dataLoading().done(function () {
              var service = router.get('mainServiceItemController.content');
              if (!service || !service.get('isLoaded')) {
                service = App.Service.find().objectAt(0); // getting the first service to display
              }
              if (service.get('routeToConfigs')) {
                router.transitionTo('service.configs', service);
              }
              else {
                router.transitionTo('service.summary', service);
              }
            });
          });
        }
      }),
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainService');
      },
      service: Em.Route.extend({
        route: '/:service_id',
        connectOutlets: function (router, service) {
          router.get('mainServiceController').connectOutlet('mainServiceItem', service);
          router.transitionTo('summary');
        },
        index: Ember.Route.extend({
          route: '/'
        }),
        summary: Em.Route.extend({
          route: '/summary',
          connectOutlets: function (router, context) {
            var item = router.get('mainServiceItemController.content');
            //if service is not existed then route to default service
            if (item.get('isLoaded')) {
              router.get('mainServiceItemController').connectOutlet('mainServiceInfoSummary', item);
            } else {
              router.transitionTo('services.index');
            }
          }
        }),
        metrics: Em.Route.extend({
          route: '/metrics',
          connectOutlets: function (router, context) {
            var item = router.get('mainServiceItemController.content');
            router.get('mainServiceItemController').connectOutlet('mainServiceInfoMetrics', item);
          }
        }),
        configs: Em.Route.extend({
          route: '/configs',
          connectOutlets: function (router, context) {
            var item = router.get('mainServiceItemController.content');
            //if service is not existed then route to default service
            if (item.get('isLoaded')) {
              if (item.get('isConfigurable')) {
                router.get('mainServiceItemController').connectOutlet('mainServiceInfoConfigs', item);
              }
              else {
                // if service doesn't have configs redirect to summary
                router.transitionTo('summary');
              }
            }
            else {
              item.set('routeToConfigs', true);
              router.transitionTo('services.index');
            }
          },
          unroutePath: function (router, context) {
            var controller = router.get('mainServiceInfoConfigsController');
            if (!controller.get('forceTransition') && controller.hasUnsavedChanges()) {
              controller.showSavePopup(context);
            } else {
              this._super(router, context);
            }
          }
        }),
        audit: Em.Route.extend({
          route: '/audit',
          connectOutlets: function (router, context) {
            var item = router.get('mainServiceItemController.content');
            router.get('mainServiceItemController').connectOutlet('mainServiceInfoAudit', item);
          }
        }),
        showInfo: function (router, event) {
          var mainServiceInfoConfigsController = App.router.get('mainServiceInfoConfigsController');
          if (event.context === 'summary' && mainServiceInfoConfigsController.hasUnsavedChanges()) {
            mainServiceInfoConfigsController.showSavePopup(router.get('location.lastSetURL').replace('configs', 'summary'));
            return false;
          }
          var parent = event.view._parentView;
          parent.deactivateChildViews();
          event.view.set('active', "active");
          router.transitionTo(event.context);
        }
      }),
      showService: Em.Router.transitionTo('service'),
      addService: Em.Router.transitionTo('serviceAdd'),
      reassign: Em.Router.transitionTo('reassign')
    }),

    reassign: require('routes/reassign_master_routes'),

    serviceAdd: require('routes/add_service_routes'),

    selectService: Em.Route.transitionTo('services.service.summary'),
    selectHost: function (router, event) {
      router.get('mainHostDetailsController').set('isFromHosts', false);
      router.transitionTo('hosts.hostDetails.index', event.context);
    },
    filterHosts: function (router, component) {
      if(!component.context)
        return;
      router.get('mainHostController').filterByComponent(component.context);
      router.transitionTo('hosts.index');
    },
    showDetails: function (router, event) {
      router.get('mainHostDetailsController').set('referer', router.location.lastSetURL);
      router.get('mainHostDetailsController').set('isFromHosts', true);
      router.transitionTo('hosts.hostDetails.summary', event.context);
    }
  });
  
});
window.require.register("routes/reassign_master_routes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: '/service/reassign',

    enter: function (router) {
      console.log('in /service/reassign:enter');
      Em.run.next(function () {
        var reassignMasterController = router.get('reassignMasterController');
        App.router.get('updateController').set('isWorking', false);
        var popup = App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header: Em.I18n.t('services.reassign.header'),
          bodyClass: App.ReassignMasterView.extend({
            controller: reassignMasterController
          }),
          primary: Em.I18n.t('form.cancel'),
          showFooter: false,
          secondary: null,

          onPrimary: function () {
            this.hide();
            App.router.get('updateController').set('isWorking', true);
            App.router.transitionTo('main.services.index');
          },
          onClose: function () {
            var reassignMasterController = router.get('reassignMasterController');
            var currStep = reassignMasterController.get('currentStep');

            if (parseInt(currStep) > 3) {
              var self = this;
              App.showConfirmationPopup(function(){
                self.hide();
                reassignMasterController.setCurrentStep('1');
                router.get('reassignMasterWizardStep' + currStep + 'Controller').removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
                App.clusterStatus.setClusterStatus({
                  clusterName: router.get('reassignMasterController.content.cluster.name'),
                  clusterState: 'DEFAULT',
                  wizardControllerName: 'reassignMasterController',
                  localdb: App.db.data
                });
                router.get('updateController').set('isWorking', true);
                router.transitionTo('main.services.index')
              }, Em.I18n.t('services.reassign.closePopup').format(reassignMasterController.get('content.reassign.display_name')));
            } else {
              this.hide();
              reassignMasterController.setCurrentStep('1');
              router.get('updateController').set('isWorking', true);
              router.transitionTo('main.services.index')
            }
          },
          didInsertElement: function () {
            this.fitHeight();
          }
        });
        reassignMasterController.set('popup', popup);
        reassignMasterController.loadSecurityEnabled();
        reassignMasterController.loadComponentToReassign();
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus) {
          switch (currentClusterStatus.clusterState) {
            case 'REASSIGN_MASTER_INSTALLING' :
              reassignMasterController.setCurrentStep(currentClusterStatus.localdb.ReassignMaster.currentStep);
              break;
          }
        }
        router.transitionTo('step' + reassignMasterController.get('currentStep'));
      });
    },

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        console.log('in reassignMaster.step1:connectOutlets');
        var controller = router.get('reassignMasterController');
        controller.setCurrentStep('1');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('reassignMasterWizardStep1', controller.get('content'));
        })
      },
      next: function (router) {
        App.db.setMasterComponentHosts(undefined);
        router.transitionTo('step2');
      },

      unroutePath: function () {
        return false;
      }
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router) {
        console.log('in reassignMaster.step2:connectOutlets');
        var controller = router.get('reassignMasterController');
        controller.setCurrentStep('2');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('reassignMasterWizardStep2', controller.get('content'));
        })

      },
      back: Em.Router.transitionTo('step1'),
      next: function (router) {
        var controller = router.get('reassignMasterController');
        var reassignMasterWizardStep2 = router.get('reassignMasterWizardStep2Controller');
        controller.saveMasterComponentHosts(reassignMasterWizardStep2);
        var reassignHosts = {};
        var componentName = reassignMasterWizardStep2.get('content.reassign.component_name');
        var masterAssignmentsHosts = reassignMasterWizardStep2.get('selectedServicesMasters').filterProperty('component_name', componentName).mapProperty('selectedHost');
        var currentMasterHosts = App.HostComponent.find().filterProperty('componentName', componentName).mapProperty('host.hostName');
        masterAssignmentsHosts.forEach(function (host) {
          if (!currentMasterHosts.contains(host)) {
            reassignHosts.target = host;
          }
        }, this);
        currentMasterHosts.forEach(function (host) {
          if (!masterAssignmentsHosts.contains(host)) {
            reassignHosts.source = host;
          }
        }, this);
        controller.saveReassignHosts(reassignHosts);
        router.transitionTo('step3');
      },

      unroutePath: function () {
        return false;
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router) {
        console.log('in reassignMaster.step3:connectOutlets');
        var controller = router.get('reassignMasterController');
        controller.setCurrentStep('3');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('reassignMasterWizardStep3', controller.get('content'));
        })
      },
      back: Em.Router.transitionTo('step2'),
      next: function (router) {
        App.db.setReassignTasksStatuses(undefined);
        App.db.setReassignMasterWizardLogs(undefined);
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('reassignMasterController.content.cluster.name'),
          clusterState: 'REASSIGN_MASTER_INSTALLING',
          wizardControllerName: 'reassignMasterController',
          localdb: App.db.data
        });
        router.transitionTo('step4');
      },

      unroutePath: function () {
        return false;
      }
    }),

    step4: Em.Route.extend({
      route: '/step4',
      connectOutlets: function (router) {
        console.log('in reassignMaster.step4:connectOutlets');
        var controller = router.get('reassignMasterController');
        controller.setCurrentStep('4');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.setLowerStepsDisable(4);
          controller.connectOutlet('reassignMasterWizardStep4', controller.get('content'));
        })
      },
      next: function (router) {
        router.get('reassignMasterController').setCurrentStep('5');
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('reassignMasterController.content.cluster.name'),
          clusterState: 'REASSIGN_MASTER_INSTALLING',
          wizardControllerName: 'reassignMasterController',
          localdb: App.db.data
        });
        router.transitionTo('step5');
      },

      complete: function (router) {
        var controller = router.get('reassignMasterController');
        var reassignMasterWizardStep4 = router.get('reassignMasterWizardStep4Controller');
        if (!reassignMasterWizardStep4.get('isSubmitDisabled')) {
          controller.finish();
          controller.get('popup').hide();
          App.clusterStatus.setClusterStatus({
            clusterName: router.get('reassignMasterController.content.cluster.name'),
            clusterState: 'DEFAULT',
            wizardControllerName: 'reassignMasterController',
            localdb: App.db.data
          });
          router.transitionTo('main.index');
          location.reload();
        }
      },

      unroutePath: function () {
        return false;
      }
    }),

    step5: Em.Route.extend({
      route: '/step5',
      connectOutlets: function (router) {
        console.log('in reassignMaster.step5:connectOutlets');
        var controller = router.get('reassignMasterController');
        controller.setCurrentStep('5');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.setLowerStepsDisable(5);
          if ((controller.get('content.reassign.component_name') === 'NAMENODE') || controller.get('content.reassign.component_name') === 'SECONDARY_NAMENODE') {
            controller.usersLoading().done(function () {
              controller.connectOutlet('reassignMasterWizardStep5', controller.get('content'));
            })
          } else {
            controller.connectOutlet('reassignMasterWizardStep5', controller.get('content'));
          }
        })
      },
      next: function (router) {
        App.showConfirmationPopup(function () {
          router.transitionTo('step6');
        }, Em.I18n.t('services.reassign.step5.confirmPopup.body'));
      },

      unroutePath: function () {
        return false;
      }
    }),

    step6: Em.Route.extend({
      route: '/step6',
      connectOutlets: function (router) {
        console.log('in reassignMaster.step6:connectOutlets');
        var controller = router.get('reassignMasterController');
        controller.setCurrentStep('6');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.setLowerStepsDisable(6);
          controller.connectOutlet('reassignMasterWizardStep6', controller.get('content'));
        })
      },

      next: function (router) {
        var controller = router.get('reassignMasterController');
        var reassignMasterWizardStep6 = router.get('reassignMasterWizardStep6Controller');
        if (!reassignMasterWizardStep6.get('isSubmitDisabled')) {
          controller.finish();
          controller.get('popup').hide();
          App.clusterStatus.setClusterStatus({
            clusterName: router.get('reassignMasterController.content.cluster.name'),
            clusterState: 'DEFAULT',
            wizardControllerName: 'reassignMasterController',
            localdb: App.db.data
          });
          router.transitionTo('main.index');
          location.reload();
        }
      },

      unroutePath: function () {
        return false;
      }
    }),


    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3'),

    gotoStep4: Em.Router.transitionTo('step4'),

    gotoStep5: Em.Router.transitionTo('step5'),

    gotoStep6: Em.Router.transitionTo('step6'),

    backToServices: function (router) {
      App.router.get('updateController').set('isWorking', true);
      router.transitionTo('services');
    }

  });
  
});
window.require.register("routes/rollbackHA_routes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: '/highAvailability/rollbackHA',

    enter: function (router) {
      Em.run.next(function () {
        var rollbackHighAvailabilityWizardController = router.get('rollbackHighAvailabilityWizardController');
        App.router.get('updateController').set('isWorking', false);
        var popup = App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header: Em.I18n.t('admin.highAvailability.wizard.rollback.header.title'),
          bodyClass: App.RollbackHighAvailabilityWizardView.extend({
            controller: rollbackHighAvailabilityWizardController
          }),
          primary: Em.I18n.t('form.cancel'),
          showFooter: false,
          secondary: null,
          hideCloseButton: function () {
            var currStep = App.router.get('rollbackHighAvailabilityWizardController.currentStep');
            if(currStep == 3){
              this.set('showCloseButton', false);
            }else{
              this.set('showCloseButton', true);
            }
          }.observes('App.router.rollbackHighAvailabilityWizardController.currentStep'),

          onClose: function () {
            this.hide();
            App.router.get('rollbackHighAvailabilityWizardController').setCurrentStep('1');
            App.router.get('updateController').set('isWorking', true);
            App.router.transitionTo('main.admin.adminHighAvailability');
          },
          didInsertElement: function () {
            this.fitHeight();
          }
        });
        rollbackHighAvailabilityWizardController.set('popup', popup);
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus) {
          switch (currentClusterStatus.clusterState) {
            case 'ROLLBACK_HIGH_AVAILABILITY' :
              rollbackHighAvailabilityWizardController.setCurrentStep(currentClusterStatus.localdb.RollbackHighAvailabilityWizard.currentStep);
              break;
            default:
              var currStep = App.router.get('rollbackHighAvailabilityWizardController.currentStep');
              rollbackHighAvailabilityWizardController.setCurrentStep(currStep);
              break;
          }
        }
        router.transitionTo('step' + rollbackHighAvailabilityWizardController.get('currentStep'));
      });
    },

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        var controller = router.get('rollbackHighAvailabilityWizardController');
        controller.setCurrentStep('1');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('rollbackHighAvailabilityWizardStep1', controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('rollbackHighAvailabilityWizardController');
        controller.saveSelectedAddNN(controller.get('content.selectedAddNNHost'));
        controller.saveSelectedSNN(controller.get('content.selectedSNNHost'));
        router.transitionTo('step2');
      }
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router) {
        var controller = router.get('rollbackHighAvailabilityWizardController');
        controller.setCurrentStep('2');
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.usersLoading().done(function () {
            controller.connectOutlet('rollbackHighAvailabilityWizardStep2', controller.get('content'));
          });
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        router.transitionTo('step3');
      },
      back: function (router) {
        router.transitionTo('step1');
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router) {
        var controller = router.get('rollbackHighAvailabilityWizardController');
        controller.setCurrentStep('3');
        controller.setLowerStepsDisable(3);
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('content.cluster.name'),
          clusterState: 'ROLLBACK_HIGH_AVAILABILITY',
          wizardControllerName: 'rollbackHighAvailabilityWizardController',
          localdb: App.db.data
        });
        controller.dataLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.connectOutlet('rollbackHighAvailabilityWizardStep3',  controller.get('content'));
        })
      },
      unroutePath: function () {
        return false;
      },
      next: function (router) {
        var controller = router.get('rollbackHighAvailabilityWizardController');
        controller.clearTasksData();
        controller.clearStorageData();
        controller.finish();
        controller.get('popup').hide();
        App.router.get('updateController').set('isWorking', true);
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('content.cluster.name'),
          clusterState: 'DEFAULT',
          wizardControllerName: 'rollbackHighAvailabilityWizardController',
          localdb: App.db.data
        });
        router.transitionTo('main.index');
        location.reload();
      }
    }),

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3')

  });
  
});
window.require.register("routes/stack_upgrade", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  module.exports = App.WizardRoute.extend({
    route: 'admin/cluster/upgrade',

    enter: function (router) {
      console.log('in /admin/cluster/upgrade:enter');
      Ember.run.next(function () {
        var stackUpgradeController = router.get('stackUpgradeController');
        App.router.get('updateController').set('isWorking', false);
        App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header:Em.I18n.t('installer.stackUpgrade.header'),
          bodyClass:  App.StackUpgradeView.extend({
            controllerBinding: 'App.router.stackUpgradeController'
          }),
          showFooter: false,
          onClose: function() {
            this.hide();
            App.router.get('updateController').set('isWorking', true);
            router.transitionTo('admin.adminCluster');
          },
          didInsertElement: function(){
            this.fitHeight();
          }
        });
        var statuses = ['STOPPING_SERVICES', 'STACK_UPGRADING', 'STACK_UPGRADE_FAILED', 'STACK_UPGRADED'];
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus.localdb) App.db.data = currentClusterStatus.localdb;
        if (statuses.contains(currentClusterStatus.clusterState)) {
          stackUpgradeController.setCurrentStep(3);
        }
        router.transitionTo('step' + stackUpgradeController.get('currentStep'));
      });
    },

    step1: Em.Route.extend({
      route: '/step1',
      connectOutlets: function (router) {
        console.log('in stackUpgrade.step1:connectOutlets');
        var controller = router.get('stackUpgradeController');
        controller.setCurrentStep('1');
        controller.loadAllPriorSteps();
        controller.connectOutlet('stackUpgradeStep1', controller.get('content'));
      },
      next: Em.Router.transitionTo('step2')
    }),

    step2: Em.Route.extend({
      route: '/step2',
      connectOutlets: function (router) {
        console.log('in stackUpgrade.step2:connectOutlets');
        var controller = router.get('stackUpgradeController');
        controller.setCurrentStep('2');
        controller.loadAllPriorSteps();
        controller.connectOutlet('stackUpgradeStep2', controller.get('content'));
      },
      back: Em.Router.transitionTo('step1'),
      next: function(router){
        var controller = router.get('stackUpgradeController');
        var stepController = router.get('stackUpgradeStep3Controller');
        controller.save('upgradeOptions');
        router.transitionTo('step3');
        App.clusterStatus.setClusterStatus({
          clusterName: controller.get('content.cluster.name'),
          clusterState: 'PENDING',
          wizardControllerName: 'stackUpgradeController',
          localdb: App.db.data
        });
        stepController.stopServices();
      }
    }),

    step3: Em.Route.extend({
      route: '/step3',
      connectOutlets: function (router, context) {
        console.log('in stackUpgrade.step3:connectOutlets');
        var controller = router.get('stackUpgradeController');
        controller.setCurrentStep('3');
        controller.setLowerStepsDisable(3);
        controller.loadAllPriorSteps();
        controller.connectOutlet('stackUpgradeStep3', controller.get('content'));
      },
      finish: function (router, context) {
        var controller = router.get('stackUpgradeController');
        var stepController = router.get('stackUpgradeStep3Controller');
        controller.finish();
        $(context.currentTarget).parents("#modal").find(".close").trigger('click');
        App.clusterStatus.setClusterStatus({
          clusterName: controller.get('content.cluster.name'),
          clusterState: 'DEFAULT',
          wizardControllerName: 'stackUpgradeController',
          localdb: App.db.data
        });
        App.router.get('updateController').updateAll();
        App.set('currentStackVersion', controller.get('content.upgradeVersion'));
      }
    }),
    backToCluster: function(router, context){
      var controller = router.get('stackUpgradeController');
      controller.finish();
      $(context.currentTarget).parents("#modal").find(".close").trigger('click');
    },

    gotoStep1: Em.Router.transitionTo('step1'),

    gotoStep2: Em.Router.transitionTo('step2'),

    gotoStep3: Em.Router.transitionTo('step3')

  });
  
});
window.require.register("templates", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */



  // load templates here

  require('templates/main/service/info/summary/ganglia');
  require('templates/main/service/info/summary/oozie');
  require('templates/main/service/info/summary/zookeeper');
  require('templates/main/service/info/summary/mapreduce2');
  require('templates/main/service/info/summary/hive');
  require('templates/main/service/info/summary/hue');
  require('templates/main/service/info/summary/falcon');
  require('templates/main/service/info/summary/tez');
  require('templates/main/service/info/summary/pig');
  require('templates/main/service/info/summary/sqoop');
  require('templates/main/admin/highAvailability/progress');
  
});
window.require.register("templates/application", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <a class=\"brand cluster-name\" href=\"javascript:void(null);\" ");
    stack1 = {};
    stack2 = "clusterName";
    stack1['title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <span ");
    stack1 = depth0;
    stack2 = "showPopup";
    stack3 = {};
    stack4 = "App.router.backgroundOperationsController";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >");
    stack1 = depth0;
    stack2 = "clusterDisplayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>\n\n              ");
    stack1 = depth0;
    stack2 = "App.router.backgroundOperationsController";
    stack3 = helpers['with'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </a>\n          ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "allOperationsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  <span class=\"label operations-count\" ");
    stack1 = depth0;
    stack2 = "showPopup";
    stack3 = {};
    stack4 = "App.router.backgroundOperationsController";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "allOperationsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "ops";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  <span class=\"label\" ");
    stack1 = depth0;
    stack2 = "showPopup";
    stack3 = {};
    stack4 = "App.router.backgroundOperationsController";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "allOperationsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "ops";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <div class=\"btn-group pull-right usermenu-wrapper\">\n              <button class=\"btn btn-group dropdown-toggle\"  data-toggle=\"dropdown\">\n                ");
    stack1 = depth0;
    stack2 = "App.router.loginName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "&nbsp;<span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                <li><a href=\"\" ");
    stack1 = depth0;
    stack2 = "showAboutPopup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "app.aboutAmbari";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    stack1 = depth0;
    stack2 = "isClusterDataLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                <li class=\"break\"></li>\n                <li><a href=\"\" ");
    stack1 = depth0;
    stack2 = "logoff";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "app.signout";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              </ul>\n            </div>\n          ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                    <li><a href=\"\" ");
    stack1 = depth0;
    stack2 = "showSettingsPopup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "app.settings";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                  ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"main\">\n  <div id=\"top-nav\">\n    <div class=\"navbar navbar-static-top\">\n      <div class=\"navbar-inner\">\n        <div class=\"container main-container\">\n          <a ");
    stack1 = {};
    stack2 = "topnav.logo.href";
    stack1['href'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + " class=\"logo\" target=\"_blank\"><img src=\"/img/logo.png\" alt=\"Apache Ambari\" title=\"Apache Ambari\"></a>\n          <a class=\"brand\" ");
    stack1 = {};
    stack2 = "topnav.logo.href";
    stack1['href'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\" title=\"Apache Ambari\">");
    stack1 = depth0;
    stack2 = "app.name";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n\n          ");
    stack1 = depth0;
    stack2 = "isClusterDataLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n          ");
    stack1 = depth0;
    stack2 = "App.router.loggedIn";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n      </div>\n    </div>\n  </div>\n  <div class=\"container main-container\">\n    <div id=\"content\">\n      ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/about", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"about\">\n  <div class=\"logo\">\n    <img src=\"/img/logo.png\" alt=\"Apache Ambari\" title=\"Apache Ambari\" />\n  </div>\n  <div class=\"content\">\n    <div class=\"project\">Apache Ambari</div>\n      <br />\n      ");
    stack1 = depth0;
    stack2 = "app.aboutAmbari.version";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "view.ambariVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      <br />\n      <br />\n      <a href=\"http://ambari.apache.org/\" target=\"_blank\">");
    stack1 = depth0;
    stack2 = "app.aboutAmbari.getInvolved";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      <br />\n      <a href=\"http://www.apache.org/licenses/LICENSE-2.0\" target=\"_blank\">");
    stack1 = depth0;
    stack2 = "app.aboutAmbari.licensed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/chart/linear_time", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div ");
    stack1 = {};
    stack2 = "view.leftArrowVisible:visibleArrow :arrow-left";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "switchTimeBack";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n  <div ");
    stack1 = {};
    stack2 = "view.rightArrowVisible:visibleArrow :arrow-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "forward";
    stack3 = depth0;
    stack4 = "switchTimeForward";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n");
    return buffer;}

    data.buffer.push("\n\n<div ");
    stack1 = {};
    stack2 = "view.isReady:hide:show :screensaver :no-borders :chart-container";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n<div ");
    stack1 = {};
    stack2 = "view.isReady::hidden :time-label";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.parentView.currentTimeState.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n");
    stack1 = depth0;
    stack2 = "view.isTimePagingEnable";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div ");
    stack1 = {};
    stack2 = "view.containerId";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.isReady:show:hide view.containerClass :chart-container";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <div ");
    stack1 = {};
    stack2 = "view.yAxisId";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.yAxisClass :chart-y-axis";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
    stack1 = {};
    stack2 = "view.xAxisId";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.xAxisClass :chart-x-axis";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
    stack1 = {};
    stack2 = "view.legendId";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.legendClass :chart-legend";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
    stack1 = {};
    stack2 = "view.chartId";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.chartClass :chart";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
    stack1 = {};
    stack2 = "view.titleId";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.titleClass :chart-title";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n</div>\n");
    stack1 = depth0;
    stack2 = "view.isTimePagingEnable";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/addPropertyWindow", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <a href=\"#\" class=\"btn-padding\" ");
    stack1 = depth0;
    stack2 = "filterByKey";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "services.service.config.addPropertyWindow.filterKeyLink";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        ");
    return buffer;}

    data.buffer.push("\n  <form class=\"form-horizontal\" autocomplete=\"off\">\n    <div class=\"each-row\">\n      <label class=\"control-label\">Type</label>\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "view.serviceConfigProperty.filename";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div ");
    stack1 = {};
    stack2 = "view.serviceConfigProperty.isKeyError:error :each-row :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <label class=\"control-label\">Key</label>\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.serviceConfigProperty.name";
    stack3['valueBinding'] = stack4;
    stack4 = "span4";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "view.serviceConfigProperty.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        ");
    stack1 = depth0;
    stack2 = "view.serviceConfigProperty.showFilterLink";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    </div>\n    <div class=\"each-row\">\n      <label class=\"control-label\">Value</label>\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "Ember.TextArea";
    stack3 = {};
    stack4 = "view.serviceConfigProperty.value";
    stack3['valueBinding'] = stack4;
    stack4 = "4";
    stack3['rows'] = stack4;
    stack4 = "span6";
    stack3['classNames'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  </form>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/capacity_scheduler", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                    <div class=\"btn pull-right\" ");
    stack1 = depth0;
    stack2 = "queuePopup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.mapReduce.config.addQueue";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n                ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      <th>");
    stack1 = depth0;
    stack2 = "header";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n                    ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                  <tr ");
    stack1 = depth0;
    stack2 = "queue.name";
    stack3 = depth0;
    stack4 = "queuePopup";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                      <td><a href=\"javascript:void(0)\">\n                          <i ");
    stack1 = {};
    stack2 = "queue.color";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"marker\"></i>");
    stack1 = depth0;
    stack2 = "queue.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      </td>\n                      ");
    stack1 = depth0;
    stack2 = "queue.configs";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "config";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </tr>\n                  ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                        <td>");
    stack1 = depth0;
    stack2 = "config.value";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "isVisible";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                        <div class=\"row-fluid\">\n                            <div ");
    stack1 = {};
    stack2 = "errorMessage:error :control-label-span :span4";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                                <label>");
    stack1 = depth0;
    stack2 = "displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n                            </div>\n                            <div class=\"span8\">\n                                <div ");
    stack1 = {};
    stack2 = "errorMessage:error: :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                                  ");
    stack1 = depth0;
    stack2 = "viewClass";
    stack3 = {};
    stack4 = "this";
    stack3['serviceConfigBinding'] = stack4;
    stack4 = "view.categoryConfigs";
    stack3['categoryConfigsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                                  ");
    stack1 = depth0;
    stack2 = "view.canEdit";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                    <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n                                </div>\n                            </div>\n                        </div>\n                    ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                      ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                  ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                        ");
    stack1 = depth0;
    stack2 = "isRemovable";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                      ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                                            <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "removeProperty";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-minus-sign\"></i>");
    stack1 = depth0;
    stack2 = "common.remove";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                                        ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.category.canAddProperty";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n              <div>\n                  <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "showAddPropertyWindow";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >");
    stack1 = depth0;
    stack2 = "installer.step7.config.addProperty";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "...</a>\n              </div>\n          ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"accordion-heading\" ");
    stack1 = depth0;
    stack2 = "view.category";
    stack3 = depth0;
    stack4 = "onToggleBlock";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <i class=\"pull-left accordion-toggle\" ");
    stack1 = {};
    stack2 = "view.category.isCollapsed:icon-caret-right:icon-caret-down ";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n    <a class=\"accordion-toggle\">");
    stack1 = depth0;
    stack2 = "view.category.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n</div>\n\n<div ");
    stack1 = {};
    stack2 = "view.category.name";
    stack1['id'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"accordion-body collapse in\">\n    <div class=\"accordion-inner capacity-scheduler\">\n        <div class=\"row-fluid header\">\n            <div class=\"span1\">");
    stack1 = depth0;
    stack2 = "services.mapReduce.config.queue.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"offset9 span2\">\n                ");
    stack1 = depth0;
    stack2 = "view.canEdit";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </div>\n        </div>\n        <div>\n            <table class=\"table table-bordered table-striped\">\n                <thead>\n                <tr>\n                    ");
    stack1 = depth0;
    stack2 = "view.tableHeaders";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "header";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </tr>\n                </thead>\n                <tbody>\n                  ");
    stack1 = depth0;
    stack2 = "view.tableContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "queue";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </tbody>\n            </table>\n        </div>\n        <div class=\"row-fluid\">\n            <div class=\"span8\">\n                <div><strong>");
    stack1 = depth0;
    stack2 = "common.advanced";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong></div>\n                <form class=\"form-horizontal\" autocomplete=\"off\">\n                  ");
    stack1 = depth0;
    stack2 = "view.advancedConfigs";
    stack3 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </form>\n            </div>\n            <div class=\"span4\">\n                <div class=\"frame\">\n                  <div id=\"section_label\"></div>\n                  ");
    stack1 = depth0;
    stack2 = "view.pieChart";
    stack3 = {};
    stack4 = "view.queues";
    stack3['queuesBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </div>\n            </div>\n        </div>\n      ");
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "view.canEdit";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/overriddenProperty", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    data.buffer.push("\n    <div ");
    stack1 = {};
    stack2 = "overriddenSCP.errorMessage:error: :control-group :overrideField";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "overriddenSCP.viewClass";
    stack3 = {};
    stack4 = "overriddenSCP";
    stack3['serviceConfigBinding'] = stack4;
    stack4 = "view.categoryConfigs";
    stack3['categoryConfigsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "view.isDefaultGroupSelected";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "overriddenSCP.isEditable";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "overriddenSCP.group";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "overriddenSCP.group";
    stack3 = depth0;
    stack4 = "selectConfigGroup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "overriddenSCP.group.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	      ");
    stack1 = depth0;
    stack2 = "isNotDefaultValue";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	      <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "overriddenSCP.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n	    ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n	        <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "doRestoreDefaultValue";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-undo\"></i>");
    stack1 = depth0;
    stack2 = "common.undo";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n	      ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n	        <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "overriddenSCP";
    stack3 = depth0;
    stack4 = "removeOverride";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-minus-sign\"></i>");
    stack1 = depth0;
    stack2 = "common.remove";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.serviceConfigProperty.overrides";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "overriddenSCP";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/overrideWindow", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	        <div class=\"text-warning\">\n	          ");
    stack1 = depth0;
    stack2 = "view.parentView.warningMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n	        </div>\n        ");
    return buffer;}

  function program3(depth0,data) {
    
    
    data.buffer.push("\n                <i class='icon-ok-sign'></i>\n              ");}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n			                    <li>\n			                      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "column";
    stack3 = depth0;
    stack4 = "selectFilterColumn";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n			                        ");
    stack1 = depth0;
    stack2 = "column.selected";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n			                        ");
    stack1 = depth0;
    stack2 = "column.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n			                      </a>\n			                    </li>\n			                  ");
    return buffer;}
  function program6(depth0,data) {
    
    
    data.buffer.push("\n	                              <i class='icon-ok-sign'></i>\n	                            ");}

  function program8(depth0,data) {
    
    
    data.buffer.push("\n                                <i class='icon-placeholder'></i>\n	                            ");}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                        <li>\n                          <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "component";
    stack3 = depth0;
    stack4 = "selectFilterComponent";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                            ");
    stack1 = depth0;
    stack2 = "component.selected";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(13, program13, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	                          ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n	                        </a>\n                        </li>\n                      ");
    return buffer;}
  function program11(depth0,data) {
    
    
    data.buffer.push("\n                              <i class='icon-ok-sign'></i>\n                            ");}

  function program13(depth0,data) {
    
    
    data.buffer.push("\n                              <i class='icon-placeholder'></i>\n	                          ");}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <tr ");
    stack1 = {};
    stack2 = "entry.filtered::hidden";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <td width=\"10%\">\n                ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "entry.selected";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td width=\"45%\">\n                ");
    stack1 = depth0;
    stack2 = "entry.host.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td>\n                ");
    stack1 = depth0;
    stack2 = "entry.filterColumnValue";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
    return buffer;}

    data.buffer.push("\n\n");
    data.buffer.push("\n  <form class=\"form-horizontal\" autocomplete=\"off\">\n      <div class=\"override-controls\">\n        <div class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "view.parentView.dialogMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.warningMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        <table style=\"width: 100%;\">\n          <tr>\n            <td>\n              <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "toggleShowSelectedHosts";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.hostSelectMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
    stack1 = depth0;
    stack2 = "view.showOnlySelectedHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </td>\n            <td width=\"30%\">\n              <div class=\"row\">\n                <div class=\"span2\" id=\"filter-dropdown-div\">\n                  <!-- Filter text box with drop-down -->\n		              <div class=\"input-append\">\n		                ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.filterText";
    stack3['valueBinding'] = stack4;
    stack4 = "appendedDropdownButton";
    stack3['id'] = stack4;
    stack4 = "view.filterTextPlaceholder";
    stack3['placeholderBinding'] = stack4;
    stack4 = "span2";
    stack3['class'] = stack4;
    stack4 = "view.isDisabled";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n		                <div class=\"btn-group\">\n		                  <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" ");
    stack1 = {};
    stack2 = "view.isDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n		                    <span class=\"caret\"></span>\n		                  </button>\n		                  <ul class=\"dropdown-menu\">\n			                  ");
    stack1 = depth0;
    stack2 = "view.filterColumns";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "column";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n		                  </ul>\n		                </div>\n		              </div>\n                </div>\n                <div class=\"span2\" id=\"component-dropdown-div\">\n                  <!-- Host-components drop-down -->\n		              <div class=\"btn-group\">\n		                <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\" ");
    stack1 = {};
    stack2 = "view.isDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n		                  ");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n		                  <span class=\"caret\"></span>\n		                </button>\n		                <ul class=\"dropdown-menu\">\n		                  <!-- dropdown menu links -->\n		                  ");
    stack1 = depth0;
    stack2 = "view.filterComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n		                </ul>\n                  </div>\n                </div>\n              </div>\n            </td>\n          </tr>\n        </table>\n        <table class=\"table table-striped hosts-table\">\n          <thead>\n            <tr class=\"success\">\n              <th width=\"10%\">\n                ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.allHostsSelected";
    stack3['checkedBinding'] = stack4;
    stack4 = "view.isDisabled";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </th>\n              <th width=\"45%\">");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th width=\"45%\">");
    stack1 = depth0;
    stack2 = "view.filterColumn.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n            </tr>\n          </thead>\n         </table>\n        <div class=\"hosts-table-container\">\n          <table class=\"table table-striped hosts-table\">\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.availableHosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "entry";
    stack7 = helpers.each;
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n         </table>\n        </div>\n      </div>\n  </form>\n");
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/propertyDependence", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <tr>\n      <td>");
    stack1 = depth0;
    stack2 = "property.serviceName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "property.propertyDisplayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "property.curValue";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "property.newValue";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n  ");
    return buffer;}

    data.buffer.push("\n\n<table class=\"table table-bordered table-striped\">\n  <thead>\n  <tr>\n    <th>");
    stack1 = depth0;
    stack2 = "common.service";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
    stack1 = depth0;
    stack2 = "common.property";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
    stack1 = depth0;
    stack2 = "installer.step7.popup.currentValue";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
    stack1 = depth0;
    stack2 = "installer.step7.popup.adjustedValue";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n  </tr>\n  </thead>\n  <tbody>\n  ");
    stack1 = depth0;
    stack2 = "view.propertyChange";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "property";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </tbody>\n</table>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/propertyDependence_footer", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div class=\"pull-right\">\n  <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "onUndo";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  <button class=\"btn btn-warning\" ");
    stack1 = depth0;
    stack2 = "onIgnore";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.ignore";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  <button class=\"btn btn-success\" ");
    stack1 = depth0;
    stack2 = "onApply";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/queuePopup_body", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "isVisible";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <div class=\"rw-fluid control-group\">\n        <div ");
    stack1 = {};
    stack2 = "errorMessage:error :control-label-span :span4";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <label>");
    stack1 = depth0;
    stack2 = "displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n        </div>\n        <div ");
    stack1 = {};
    stack2 = "errorMessage:error :control-group :span8";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "viewClass";
    stack3 = {};
    stack4 = "this";
    stack3['serviceConfigBinding'] = stack4;
    stack4 = "view.categoryConfigs";
    stack3['categoryConfigsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n      </div>\n    ");
    return buffer;}

    data.buffer.push("\n\n<form class=\"form-horizontal pre-scrollable\" autocomplete=\"off\">\n  ");
    stack1 = depth0;
    stack2 = "view.parentView.content.configs";
    stack3 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</form>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/queuePopup_footer", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div class=\"pull-left\">\n    <button class=\"btn btn-danger\" ");
    stack1 = depth0;
    stack2 = "onDelete";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "view.parentView.delete";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  </div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <button type=\"button\" class=\"btn\" ");
    stack1 = depth0;
    stack2 = "onPrimary";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "view.parentView.primary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <button type=\"button\" ");
    stack1 = {};
    stack2 = "view.parentView.isError";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-success\" ");
    stack1 = depth0;
    stack2 = "onSecondary";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "view.parentView.secondary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.parentView.delete";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<p class=\"pull-right\">\n  ");
    stack1 = depth0;
    stack2 = "view.parentView.primary";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.parentView.secondary";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</p>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/saveConfigGroup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div id=\"config-group-confirm-save\">\n  <div class=\"alert alert-success\">");
    stack1 = depth0;
    stack2 = "config.group.save.confirmation.configGroup";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <em>");
    stack1 = depth0;
    stack2 = "groupName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</em> ");
    stack1 = depth0;
    stack2 = "config.group.save.confirmation.saved";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"message\">");
    stack1 = depth0;
    stack2 = "config.group.save.confirmation.msg";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/selectCreateConfigGroup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <div class=\"select-create-config-group-div\">\n      <span class=\"btn-group\">\n        <button class=\"btn\" ");
    stack1 = {};
    stack2 = "view.parentView.optionCreateConfigGroup";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.parentView.selectedConfigGroup.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</button>\n        <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" ");
    stack1 = {};
    stack2 = "view.parentView.optionCreateConfigGroup";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <span class=\"caret\"></span>\n        </button>\n        <ul class=\"dropdown-menu\">\n          <!-- available config group menu links -->\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.availableConfigGroups";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "configGroup";
    stack7 = helpers.each;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n      </span>\n      <div class=\"message\">");
    stack1 = depth0;
    stack2 = "config.group.selection.dialog.option.select.msg";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n             <li>\n               <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "configGroup";
    stack3 = depth0;
    stack4 = "doSelectConfigGroup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                 ");
    stack1 = depth0;
    stack2 = "configGroup.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n               </a>\n             </li>\n          ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"select-create-config-group-div\">\n      <div class=\"alert alert-block\">");
    stack1 = depth0;
    stack2 = "view.parentView.noGroups";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  ");
    return buffer;}

    data.buffer.push("\n<div id=\"config-group-select-create-dialog\">\n  <div class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.subTitle";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"text-warning\">\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.warningMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <label class=\"radio\">\n    ");
    stack1 = depth0;
    stack2 = "view.selectConfigGroupRadioButton";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <span ");
    stack1 = {};
    stack2 = "view.parentView.hasExistedGroups::message";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.parentView.selectExistingGroupLabel";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  </label>\n  ");
    stack1 = depth0;
    stack2 = "view.parentView.availableConfigGroups.length";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <label class=\"radio\">\n    ");
    stack1 = depth0;
    stack2 = "view.createConfigGroupRadioButton";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.createNewGroupLabel";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </label>\n  <div class=\"select-create-config-group-div\">\n    ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.parentView.newConfigGroupName";
    stack3['valueBinding'] = stack4;
    stack4 = "view.parentView.optionSelectConfigGroup";
    stack3['disabledBinding'] = stack4;
    stack4 = "255";
    stack3['maxlength'] = stack4;
    stack4 = "span5";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <div class=\"message\">");
    stack1 = depth0;
    stack2 = "view.parentView.createNewGroupDescription";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/service_config", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "selectedService.restartRequired";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n	<div class=\"alert\">\n	  <div class=\"clearfix like_pointer collapsable\" ");
    stack1 = depth0;
    stack2 = "toggleRestartMessageView";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n	      ");
    stack1 = depth0;
    stack2 = "view.isRestartMessageCollapsed";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	      <i class=\"icon-refresh\"></i>\n	      <strong>");
    stack1 = depth0;
    stack2 = "dashboard.services.configs.popup.restartService.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong> ");
    stack1 = depth0;
    stack2 = "selectedService.restartRequiredMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n	  </div>\n	  <div class=\"service-body\">\n	    <ul>\n	    ");
    stack1 = depth0;
    stack2 = "selectedService.restartRequiredHostsAndComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	    </ul>\n	  </div>\n	</div>\n	");
    return buffer;}
  function program3(depth0,data) {
    
    
    data.buffer.push("\n	        <i class=\"icon-caret-down pull-left\"></i>\n	      ");}

  function program5(depth0,data) {
    
    
    data.buffer.push("\n	        <i class=\"icon-caret-right pull-left\"></i>\n	      ");}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n	      <li>\n	        <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "host.hostData";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.hostData.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n	        <ul>\n	          ");
    stack1 = depth0;
    stack2 = "host.components";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	        </ul>\n	      </li>\n	    ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	            <li>");
    stack1 = depth0;
    stack2 = "component.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</li>\n	          ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n	<div class=\"alert alert-info\">\n	  ");
    stack1 = depth0;
    stack2 = "common.group";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "&nbsp; \n	  <span class=\"btn-group\">\n		  <button ");
    stack1 = {};
    stack2 = "controller.isHostsConfigsPage";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
    stack1 = depth0;
    stack2 = "selectedConfigGroup.displayNameHosts";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</button>\n		  <button ");
    stack1 = {};
    stack2 = "controller.isHostsConfigsPage";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n		    <span class=\"caret\"></span>\n		  </button>\n		  <ul class=\"dropdown-menu\">\n		    <!-- available config group menu links -->\n		    ");
    stack1 = depth0;
    stack2 = "configGroups";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "configGroup";
    stack7 = helpers.each;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n		  </ul>\n		</span>\n    ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n     <div class=\"pull-right\">\n       ");
    stack1 = depth0;
    stack2 = "App.FilterComboCleanableView";
    stack3 = {};
    stack4 = "view.filter";
    stack3['filterBinding'] = stack4;
    stack4 = "view.columns";
    stack3['columnsBinding'] = stack4;
    stack4 = "view.propertyFilterPopover";
    stack3['popoverDescriptionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n     </div>\n	</div>\n");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n           <li>\n             <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "configGroup";
    stack3 = depth0;
    stack4 = "selectConfigGroup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n               ");
    stack1 = depth0;
    stack2 = "configGroup.displayNameHosts";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n             </a>\n           </li>\n        ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "controller.isHostsConfigsPage";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(16, program16, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          &nbsp;<a href=\"#\" ");
    stack1 = depth0;
    stack2 = "switchHostGroup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.change";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
    return buffer;}

  function program16(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <a href=\"#\"  class=\"link-left-pad\" ");
    stack1 = depth0;
    stack2 = "manageConfigurationGroup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.service.actions.manage_configuration_groups.short";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
    return buffer;}

  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "category.isCustomView";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(22, program22, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.supports.capacitySchedulerUi";
    stack3 = helpers['if'];
    tmp1 = self.program(20, program20, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program20(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "category.customView";
    stack3 = {};
    stack4 = "category";
    stack3['categoryBinding'] = stack4;
    stack4 = "selectedService";
    stack3['serviceBinding'] = stack4;
    stack4 = "view.canEdit";
    stack3['canEditBinding'] = stack4;
    stack4 = "selectedService.configs";
    stack3['serviceConfigsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigsByCategoryView";
    stack3 = {};
    stack4 = "category";
    stack3['categoryBinding'] = stack4;
    stack4 = "view.canEdit";
    stack3['canEditBinding'] = stack4;
    stack4 = "selectedService";
    stack3['serviceBinding'] = stack4;
    stack4 = "selectedService.configs";
    stack3['serviceConfigsBinding'] = stack4;
    stack4 = "view.supportsHostOverrides";
    stack3['supportsHostOverridesBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(23, program23, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program23(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n\n            <div class=\"accordion-heading\" ");
    stack1 = depth0;
    stack2 = "category";
    stack3 = depth0;
    stack4 = "onToggleBlock";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <i ");
    stack1 = {};
    stack2 = ":pull-left :accordion-toggle category.isCollapsed:icon-caret-right:icon-caret-down";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                <a class=\"accordion-toggle\">");
    stack1 = depth0;
    stack2 = "category.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n            </div>\n\n            <div class=\"accordion-body collapse in\" ");
    stack1 = {};
    stack2 = "view.isCategoryBodyVisible";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <div class=\"accordion-inner service-config-section\">\n                <form class=\"form-horizontal\" autocomplete=\"off\">\n\n                  ");
    stack1 = depth0;
    stack2 = "view.filteredCategoryConfigs";
    stack3 = helpers.each;
    tmp1 = self.program(24, program24, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n                  ");
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(45, program45, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </form>\n              </div>\n            </div>\n\n\n    ");
    return buffer;}
  function program24(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      <div ");
    stack1 = {};
    stack2 = ":entry-row isOverridden:overridden-property";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                          ");
    stack1 = depth0;
    stack2 = "showLabel";
    stack3 = helpers['if'];
    tmp1 = self.program(25, program25, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                          <div ");
    stack1 = {};
    stack2 = "showLabel:controls";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                            ");
    data.buffer.push("\n                              <div ");
    stack1 = {};
    stack2 = "errorMessage:error: warnMessage:warning: :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                                ");
    stack1 = depth0;
    stack2 = "viewClass";
    stack3 = {};
    stack4 = "this";
    stack3['serviceConfigBinding'] = stack4;
    stack4 = "view.categoryConfigsAll";
    stack3['categoryConfigsAllBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                                ");
    stack1 = depth0;
    stack2 = "view.canEdit";
    stack3 = helpers['if'];
    tmp1 = self.program(32, program32, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                  <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n                                  <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "warnMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n                              </div>\n                            ");
    stack1 = depth0;
    stack2 = "isOverridden";
    stack3 = helpers['if'];
    tmp1 = self.program(43, program43, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                          </div>\n                      </div>\n                  ");
    return buffer;}
  function program25(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                            <span ");
    stack1 = {};
    stack2 = "errorMessage:error: :control-group :control-label-span";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                              <label class=\"control-label\">\n                                ");
    stack1 = depth0;
    stack2 = "displayName";
    foundHelper = helpers.formatWordBreak;
    stack3 = foundHelper || depth0.formatWordBreak;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatWordBreak", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                                ");
    stack1 = depth0;
    stack2 = "App.supports.secureCluster";
    stack3 = helpers['if'];
    tmp1 = self.program(26, program26, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                ");
    stack1 = depth0;
    stack2 = "view.supportsHostOverrides";
    stack3 = helpers['if'];
    tmp1 = self.program(29, program29, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                              </label>\n                            </span>\n                          ");
    return buffer;}
  function program26(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                  ");
    stack1 = depth0;
    stack2 = "isSecureConfig";
    stack3 = helpers['if'];
    tmp1 = self.program(27, program27, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                ");
    return buffer;}
  function program27(depth0,data) {
    
    
    data.buffer.push("\n                                    <a href=\"javascript:void(null);\"><i class=\"icon-lock\" rel=\"tooltip\" data-toggle=\"tooltip\" title=\"security knob\"></i></a>\n                                  ");}

  function program29(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                  ");
    stack1 = depth0;
    stack2 = "isRestartRequired";
    stack3 = helpers['if'];
    tmp1 = self.program(30, program30, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                ");
    return buffer;}
  function program30(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n                                    <i class=\"icon-refresh restart-required-property\" rel=\"tooltip\" ");
    stack1 = {};
    stack2 = "restartRequiredMessage";
    stack1['title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                                  ");
    return buffer;}

  function program32(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	                                ");
    stack1 = depth0;
    stack2 = "isPropertyOverridable";
    stack3 = helpers['if'];
    tmp1 = self.program(33, program33, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	                                ");
    stack1 = depth0;
    stack2 = "cantBeUndone";
    stack3 = helpers.unless;
    tmp1 = self.program(37, program37, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	                                ");
    stack1 = depth0;
    stack2 = "isRemovable";
    stack3 = helpers['if'];
    tmp1 = self.program(40, program40, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                ");
    return buffer;}
  function program33(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                    ");
    stack1 = depth0;
    stack2 = "view.supportsHostOverrides";
    stack3 = helpers['if'];
    tmp1 = self.program(34, program34, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	                                ");
    return buffer;}
  function program34(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                      ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(35, program35, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                                    ");
    return buffer;}
  function program35(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n	                                    <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "createOverrideProperty";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-plus-sign\"></i>");
    stack1 = depth0;
    stack2 = "common.override";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                                      ");
    return buffer;}

  function program37(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	                                  ");
    stack1 = depth0;
    stack2 = "isNotDefaultValue";
    stack3 = helpers['if'];
    tmp1 = self.program(38, program38, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	                                ");
    return buffer;}
  function program38(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n	                                      <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "doRestoreDefaultValue";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-undo\"></i>");
    stack1 = depth0;
    stack2 = "common.undo";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n	                                  ");
    return buffer;}

  function program40(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                                    ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(41, program41, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	                                ");
    return buffer;}
  function program41(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n	                                    <a class=\"action\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "removeProperty";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-minus-sign\"></i>");
    stack1 = depth0;
    stack2 = "common.remove";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                                    ");
    return buffer;}

  function program43(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                              ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigView.SCPOverriddenRowsView";
    stack3 = {};
    stack4 = "this";
    stack3['serviceConfigPropertyBinding'] = stack4;
    stack4 = "controller.selectedConfigGroup.isDefault";
    stack3['isDefaultGroupSelectedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                            ");
    return buffer;}

  function program45(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "view.canEdit";
    stack3 = helpers['if'];
    tmp1 = self.program(46, program46, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  ");
    return buffer;}
  function program46(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "category.canAddProperty";
    stack3 = helpers['if'];
    tmp1 = self.program(47, program47, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program47(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                        <div>\n                          <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "";
    stack3 = depth0;
    stack4 = "showAddPropertyWindow";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >");
    stack1 = depth0;
    stack2 = "installer.step7.config.addProperty";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "...</a>\n                        </div>\n                      ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.supportsHostOverrides";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.supportsHostOverrides";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div class=\"accordion\">\n  ");
    stack1 = depth0;
    stack2 = "selectedService.configCategories";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/configs/services_config", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "service.showConfig";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigTab";
    stack3 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n        <a href=\"#");
    stack1 = depth0;
    stack2 = "service.serviceName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" ");
    stack1 = {};
    stack2 = ":active service.selected:new";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n           data-toggle=\"tab\" ");
    stack1 = depth0;
    stack2 = "service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));
    stack1 = depth0;
    stack2 = "service.errorCount";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</a>\n      ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("<span\n                class=\"badge badge-important\">");
    stack1 = depth0;
    stack2 = "service.errorCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"alert\">");
    stack1 = depth0;
    stack2 = "installer.step7.attentionNeeded";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n");
    return buffer;}

    data.buffer.push("\n\n<ul class=\"nav nav-tabs\">\n  ");
    stack1 = depth0;
    stack2 = "controller.stepConfigs";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "service";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</ul>\n");
    stack1 = depth0;
    stack2 = "App.ServiceConfigView";
    stack3 = {};
    stack4 = "controller.isNotEditable";
    stack3['isNotEditableBinding'] = stack4;
    stack4 = "controller.filter";
    stack3['filterBinding'] = stack4;
    stack4 = "controller.filterColumns";
    stack3['columnsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    stack1 = depth0;
    stack2 = "isSubmitDisabled";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/confirmation_feedback", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <br />\n  <div class='alert'>\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.additionalWarningMsg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
    return buffer;}

    data.buffer.push("\n<div>\n  ");
    stack1 = depth0;
    stack2 = "view.parentView.statusMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
    stack1 = depth0;
    stack2 = "view.parentView.additionalWarningMsg";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/filter_combo_cleanable", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li>\n                <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "column";
    stack3 = depth0;
    stack4 = "selectFilterColumn";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <i ");
    stack1 = {};
    stack2 = "column.selected:icon-ok-sign:icon-placeholder :text-success";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                  ");
    stack1 = depth0;
    stack2 = "column.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
    return buffer;}

    data.buffer.push("\n\n  ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.filter";
    stack3['valueBinding'] = stack4;
    stack4 = "view.placeHolder";
    stack3['placeholderBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <div class=\"btn-group\">\n    <div id=\"cleanerContainer\" class=\"pull-left\">\n      <a href=\"#\" id=\"filterCleaner\" ");
    stack1 = depth0;
    stack2 = "clearFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = {};
    stack2 = "view.filterNotEmpty::hidden :ui-icon :ui-icon-circle-close";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></a>\n    </div>\n      <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n          <span class=\"caret\"></span>\n      </button>\n      <ul class=\"dropdown-menu\">\n        ");
    stack1 = depth0;
    stack2 = "view.columns";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "column";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n  </div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/filter_combobox", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li>\n                <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "column";
    stack3 = depth0;
    stack4 = "selectFilterColumn";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <i ");
    stack1 = {};
    stack2 = "column.selected:icon-ok-sign:icon-placeholder :text-success";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                  ");
    stack1 = depth0;
    stack2 = "column.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
    return buffer;}

    data.buffer.push("\n\n  ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.filter";
    stack3['valueBinding'] = stack4;
    stack4 = "view.placeHolder";
    stack3['placeholderBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <div class=\"btn-group\">\n      <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n          <span class=\"caret\"></span>\n      </button>\n      <ul class=\"dropdown-menu\">\n        ");
    stack1 = depth0;
    stack2 = "view.columns";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "column";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n  </div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/form", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "field";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.fields";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "field";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/form/checkbox", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"controls\">\n  <label class=\"checkbox\" for=\"input");
    stack1 = depth0;
    stack2 = "view.field.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">\n    ");
    stack1 = depth0;
    stack2 = "view.field.viewClass";
    stack3 = {};
    stack4 = "view.field.value";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.field.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </label>\n  <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "view.field.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/form/field", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<label class=\"control-label\"\n       for=\"input");
    stack1 = depth0;
    stack2 = "view.field.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "view.field.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n\n<div class=\"controls\">\n  ");
    stack1 = depth0;
    stack2 = "view.field.viewClass";
    stack3 = {};
    stack4 = "view.field.value";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "view.field.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/grid/filter", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n<li>\n  <label class=\"checkbox\">\n    ");
    stack1 = depth0;
    stack2 = "Em.Checkbox";
    stack3 = {};
    stack4 = "filter.checked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "filter.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </label>\n</li>\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.filters";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "filter";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/grid/header", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.filter";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <a ");
    stack1 = depth0;
    stack2 = "applyFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<i ");
    stack1 = depth0;
    stack2 = "toggleFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"icon-filter\"></i>\n");
    stack1 = depth0;
    stack2 = "view.showFilter";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/grid/pager", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n    <li ");
    stack1 = {};
    stack2 = "page.activeClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><a ");
    stack1 = depth0;
    stack2 = "page";
    stack3 = depth0;
    stack4 = "activatePage";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "page.number";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n  ");
    return buffer;}

    data.buffer.push("\n\n<ul>\n    <li ");
    stack1 = {};
    stack2 = "view.prevPageDisabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a ");
    stack1 = depth0;
    stack2 = "activatePrevPage";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "common.prev";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n  ");
    stack1 = depth0;
    stack2 = "view.pages";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "page";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <li ");
    stack1 = {};
    stack2 = "view.nextPageDisabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a ");
    stack1 = depth0;
    stack2 = "activateNextPage";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n</ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/host_progress_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n    <!-- SERVICES --->\n\n    <div ");
    stack1 = {};
    stack2 = "view.isServiceListHidden:hidden :task-list-main-warp";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"task-top-wrap\">\n        <div class=\"operation-name-top\">");
    stack1 = depth0;
    stack2 = "common.operations";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"start-time-top\">");
    stack1 = depth0;
    stack2 = "common.startTime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"duration-top\">");
    stack1 = depth0;
    stack2 = "common.duration";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"select-wrap\">\n          ");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n          ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.categories";
    stack3['contentBinding'] = stack4;
    stack4 = "content.value";
    stack3['optionValuePath'] = stack4;
    stack4 = "content.label";
    stack3['optionLabelPath'] = stack4;
    stack4 = "view.serviceCategory";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n      <div id=\"service-info\">\n        ");
    stack1 = depth0;
    stack2 = "view.isServiceEmptyList";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    </div>\n\n\n  <!-- HOSTS --->\n\n  <div ");
    stack1 = {};
    stack2 = "view.isHostListHidden:hidden :task-list-main-warp";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      ");
    stack1 = depth0;
    stack2 = "controller.isBackgroundOperations";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <span class=\"task-detail-log-rolename\" >");
    stack1 = depth0;
    stack2 = "common.hosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n      <div class=\"select-wrap\">\n        ");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n        ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.categories";
    stack3['contentBinding'] = stack4;
    stack4 = "content.value";
    stack3['optionValuePath'] = stack4;
    stack4 = "content.label";
    stack3['optionLabelPath'] = stack4;
    stack4 = "view.hostCategory";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    ");
    stack1 = depth0;
    stack2 = "view.isRequestSchedule";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <div id=\"host-info\">\n      ");
    stack1 = depth0;
    stack2 = "view.isHostEmptyList";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(16, program16, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n  </div>\n\n\n  <!-- TASKS --->\n\n  <div ");
    stack1 = {};
    stack2 = "view.isTaskListHidden:hidden :task-list-main-warp";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      <a class=\"task-detail-back-to-hosts\" href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "backToHostList";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-arrow-left\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "common.hosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      <div class=\"task-name-top task-detail-log-rolename\">");
    stack1 = depth0;
    stack2 = "common.tasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"select-wrap tasks-list-select\">\n        ");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n        ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.categories";
    stack3['contentBinding'] = stack4;
    stack4 = "content.value";
    stack3['optionValuePath'] = stack4;
    stack4 = "content.label";
    stack3['optionLabelPath'] = stack4;
    stack4 = "view.taskCategory";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div id=\"host-log\">\n      ");
    stack1 = depth0;
    stack2 = "view.isTasksEmptyList";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(21, program21, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n  </div>\n\n\n  <!-- TASK DETAILS --->\n\n  <div ");
    stack1 = {};
    stack2 = "view.isLogWrapHidden:hidden :task-detail-info";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      <a class=\"task-detail-back\" href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "backToTaskList";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-arrow-left\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "common.tasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      <div>\n        <i ");
    stack1 = {};
    stack2 = "view.openedTask.status :task-detail-status-ico view.openedTask.icon";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-status-ico\"></i>\n        <span class=\"task-detail-log-rolename\" >");
    stack1 = depth0;
    stack2 = "view.openedTask.commandDetail";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n      </div>\n      <div class=\"task-detail-ico-wrap\">\n        <a title=\"Click to Copy\" ");
    stack1 = depth0;
    stack2 = "taskInfo";
    stack3 = depth0;
    stack4 = "textTrigger";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i class=\"icon-copy\"></i> ");
    stack1 = depth0;
    stack2 = "common.copy";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        <a title=\"Open in New Window\" ");
    stack1 = depth0;
    stack2 = "openTaskLogInDialog";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-open-dialog\"><i class=\"icon-external-link\"></i> ");
    stack1 = depth0;
    stack2 = "common.open";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      </div>\n    </div>\n    <div class=\"task-detail-log-info\">\n      <div class=\"content-area\" >\n        <div class=\"task-detail-log-clipboard-wrap\" ></div>\n        <div class=\"task-detail-log-maintext\">\n          <h5>stderr: &nbsp; <span class=\"muted\">");
    stack1 = depth0;
    stack2 = "view.openedTask.errorLog";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span></h5>\n          <pre class=\"stderr\">");
    stack1 = depth0;
    stack2 = "view.openedTask.stderr";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</pre>\n          <h5>stdout: &nbsp; <span class=\"muted\"> ");
    stack1 = depth0;
    stack2 = "view.openedTask.outputLog";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span> </h5>\n          <pre class=\"stdout\">");
    stack1 = depth0;
    stack2 = "view.openedTask.stdout";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</pre>\n        </div>\n      </div>\n    </div>\n  </div>\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <div class=\"log-list-wrap\">");
    stack1 = depth0;
    stack2 = "hostPopup.noServicesToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.services";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "servicesInfo";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n            <div ");
    stack1 = {};
    stack2 = "servicesInfo.isVisible::hidden :log-list-wrap";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <div ");
    stack1 = depth0;
    stack2 = "servicesInfo";
    stack3 = depth0;
    stack4 = "gotoHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n                <div class=\"operation-name-icon-wrap\">\n                  <i ");
    stack1 = {};
    stack2 = "servicesInfo.status servicesInfo.icon";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                  <a href=\"#\">\n                    ");
    stack1 = depth0;
    stack2 = "servicesInfo.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  </a>\n                </div>\n                <div class=\"time-summary start-time-text\">");
    stack1 = depth0;
    stack2 = "servicesInfo.startTime";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n                <div class=\"time-summary duration-text\">");
    stack1 = depth0;
    stack2 = "servicesInfo.duration";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n                <div class=\"progress-bar span2\">\n                  <div ");
    stack1 = {};
    stack2 = "servicesInfo.isInProgress:progress-striped :active servicesInfo.barColor :progress";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    <div class=\"bar\" ");
    stack1 = {};
    stack2 = "servicesInfo.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n                  </div>\n                </div>\n                <div class=\"host-progress-num\">");
    stack1 = depth0;
    stack2 = "servicesInfo.progress";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "%</div>\n                <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n              </div>\n            </div>\n          ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <a class=\"task-detail-back-to-hosts\" href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "backToServiceList";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >\n          <i class=\"icon-arrow-left\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "common.operations";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "view.sourceRequestScheduleRunning";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "view.sourceRequestScheduleAborted";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n        <div class=\"alert alert-info request-schedule-abort\">\n          ");
    stack1 = depth0;
    stack2 = "hostPopup.bgop.sourceRequestSchedule.running";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <button type=\"button\" class=\"btn btn-warning pull-right\"\n            ");
    stack1 = depth0;
    stack2 = "view.sourceRequestScheduleId";
    stack3 = depth0;
    stack4 = "doAbortRequestSchedule";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "view.requestScheduleAbortLabel";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          </button>\n        </div>\n      ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"alert alert-info request-schedule-abort\">\n          ");
    stack1 = depth0;
    stack2 = "hostPopup.bgop.sourceRequestSchedule.aborted";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"log-list-wrap\">");
    stack1 = depth0;
    stack2 = "hostPopup.noHostsToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
    return buffer;}

  function program16(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.hosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "hostInfo";
    stack7 = helpers.each;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n          <div ");
    stack1 = {};
    stack2 = "hostInfo.isVisible::hidden :log-list-wrap";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
    stack1 = depth0;
    stack2 = "hostInfo";
    stack3 = depth0;
    stack4 = "gotoTasks";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n              <div class=\"host-name-icon-wrap\">\n                <i ");
    stack1 = {};
    stack2 = "hostInfo.status hostInfo.icon";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                <a href=\"#\">\n                  ");
    stack1 = depth0;
    stack2 = "hostInfo.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </div>\n              <div class=\"progress-bar span2\">\n                <div ");
    stack1 = {};
    stack2 = "hostInfo.isInProgress:progress-striped :active hostInfo.barColor :progress";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <div class=\"bar\" ");
    stack1 = {};
    stack2 = "hostInfo.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n                </div>\n              </div>\n              <div class=\"host-progress-num\">");
    stack1 = depth0;
    stack2 = "hostInfo.progress";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "%</div>\n              <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n            </div>\n          </div>\n        ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"log-list-wrap\">");
    stack1 = depth0;
    stack2 = "hostPopup.noTasksToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
    return buffer;}

  function program21(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.tasks";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "taskInfo";
    stack7 = helpers.each;
    tmp1 = self.program(22, program22, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n          <div ");
    stack1 = {};
    stack2 = "taskInfo.isVisible::hidden :log-list-wrap";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
    stack1 = depth0;
    stack2 = "taskInfo";
    stack3 = depth0;
    stack4 = "toggleTaskLog";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n              <div class=\"operation-name-icon-wrap\">\n                <i ");
    stack1 = {};
    stack2 = "taskInfo.status taskInfo.icon";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                <a href=\"#\">\n                  ");
    stack1 = depth0;
    stack2 = "taskInfo.commandDetail";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </div>\n              <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n            </div>\n          </div>\n        ");
    return buffer;}

  function program24(depth0,data) {
    
    
    data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

    data.buffer.push("\n\n<div class=\"host-component-popup-wrap\">\n\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(24, program24, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/items_list_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <p>");
    stack1 = depth0;
    stack2 = "item";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.insertedItems";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/metric", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.itemView";
    stack3 = {};
    stack4 = "metric";
    stack3['metricBinding'] = stack4;
    stack4 = "view";
    stack3['widgetBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"accordion\" id=\"metricAccordion\">\n      ");
    stack1 = depth0;
    stack2 = "view.moreMetrics";
    stack3 = helpers.each;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <div class=\"accordion-group\">\n        <div class=\"accordion-heading\">\n          <a class=\"accordion-toggle\" data-toggle=\"collapse\" data-parent=\"#metricAccordion\"\n             href=\"#");
    stack1 = depth0;
    stack2 = "code";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "Collapse\">\n            <i class=\"icon-play\"></i>");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        </div>\n        <div id=\"");
    stack1 = depth0;
    stack2 = "code";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "Collapse\" class=\"accordion-body collapse\">\n          <div class=\"accordion-inner\">\n            ");
    stack1 = depth0;
    stack2 = "items.length";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </div>\n        </div>\n      </div>\n      ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n            <ul class=\"items\">\n              ");
    stack1 = depth0;
    stack2 = "items";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "metric";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n            ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.moreItemView";
    stack3 = {};
    stack4 = "metric";
    stack3['metricBinding'] = stack4;
    stack4 = "view";
    stack3['widgetBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "metric.notFound";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"span title\">\n  ");
    stack1 = depth0;
    stack2 = "common.metrics";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-question-sign\"></i>\n</div>\n<ul class=\"display-inline-block nav nav-pills\">\n  ");
    stack1 = depth0;
    stack2 = "view.metrics";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "metric";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n  <li class=\"dropdown\">\n    <a ");
    stack1 = depth0;
    stack2 = "toggleMore";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n      ");
    stack1 = depth0;
    stack2 = "metric.more";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      <b class=\"caret\"></b>\n    </a>\n\n    ");
    stack1 = depth0;
    stack2 = "view.showMore";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </li>\n</ul>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/modal_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <a class=\"close\" ");
    stack1 = depth0;
    stack2 = "onClose";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">x</a>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "headerClass";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "header";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "bodyClass";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "encodeBody";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(12, program12, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "body";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "body";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "footerClass";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(17, program17, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "footerClass";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <div class=\"modal-footer\">\n        ");
    stack1 = depth0;
    stack2 = "view.hasFooterCheckbox";
    stack3 = helpers['if'];
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.secondary";
    stack3 = helpers['if'];
    tmp1 = self.program(20, program20, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.primary";
    stack3 = helpers['if'];
    tmp1 = self.program(22, program22, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    ");
    return buffer;}
  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <label id=\"footer-checkbox\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "checkbox";
    stack3['classNames'] = stack4;
    stack4 = "view.isNotShowBgChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " &nbsp;\n          ");
    stack1 = depth0;
    stack2 = "app.settings.notShowBgOperations";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n        ");
    return buffer;}

  function program20(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <button class=\"btn\" ");
    stack1 = {};
    stack2 = "view.disableSecondary";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "onSecondary";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.secondary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
    return buffer;}

  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <button ");
    stack1 = {};
    stack2 = "view.disablePrimary";
    stack1['disabled'] = stack2;
    stack2 = ":btn view.primaryClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "onPrimary";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.primary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
    return buffer;}

    data.buffer.push("\n\n\n<div class=\"modal-backdrop\"></div>\n<div class=\"modal\" id=\"modal\" tabindex=\"-1\" role=\"dialog\" aria-labelledby=\"modal-label\" aria-hidden=\"true\">\n  <div class=\"modal-header\">\n    ");
    stack1 = depth0;
    stack2 = "showCloseButton";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <h3 id=\"modal-label\">\n      ");
    stack1 = depth0;
    stack2 = "headerClass";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </h3>\n  </div>\n  <div class=\"modal-body\">\n    ");
    stack1 = depth0;
    stack2 = "bodyClass";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(9, program9, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  ");
    stack1 = depth0;
    stack2 = "showFooter";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/prompt_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div class=\"prompt-popup\">\n  <form>\n    <div class=\"control-group\">\n      <label class=\"control-label\">");
    stack1 = depth0;
    stack2 = "view.text";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "Em.TextField";
    stack3 = {};
    stack4 = "prompt-input";
    stack3['class'] = stack4;
    stack4 = "view.parentView.inputValue";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  </form>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/rolling_restart_view", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <p>\n        ");
    stack1 = depth0;
    stack2 = "view.maintainanceMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </p>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <p>\n        ");
    stack1 = depth0;
    stack2 = "view.suggestTurnOnMaintenanceMsg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </p>\n    ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <div class=\"alert alert-warn\">\n      <ul>\n      ");
    stack1 = depth0;
    stack2 = "view.errors";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "error";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n    </div>\n  ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <li>");
    stack1 = depth0;
    stack2 = "error";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</li>\n      ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"rolling-restart-view\">\n  <div class=\"alert alert-info\">\n    <p>\n      ");
    stack1 = depth0;
    stack2 = "view.restartMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </p>\n    ");
    stack1 = depth0;
    stack2 = "view.maintainanceMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "view.suggestTurnOnMaintenanceMsg";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <table>\n    <tr>\n      <td>");
    stack1 = depth0;
    stack2 = "common.restart";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.batchSize";
    stack3['valueBinding'] = stack4;
    stack4 = "span1";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "view.batchSizeMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n    <tr>\n      <td>");
    stack1 = depth0;
    stack2 = "rollingrestart.dialog.msg.timegap.prefix";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.interBatchWaitTimeSeconds";
    stack3['valueBinding'] = stack4;
    stack4 = "span1";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "rollingrestart.dialog.msg.timegap.suffix";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n    <tr>\n      <td>");
    stack1 = depth0;
    stack2 = "rollingrestart.dialog.msg.toleration.prefix";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.tolerateSize";
    stack3['valueBinding'] = stack4;
    stack4 = "span1";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "rollingrestart.dialog.msg.toleration.suffix";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n    <tr>\n      <td>&nbsp;</td>\n      <td colspan=\"2\">\n        ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.staleConfigsOnly";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.staleConfigsOnlyMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    </tr>\n  </table>\n  ");
    stack1 = depth0;
    stack2 = "view.errors";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/selectable_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div>\n  <div class=\"task-top-wrap\">\n    <div class=\"task-detail-ico-wrap\">\n      <a href=\"#\" title=\"Click to Copy\" ");
    stack1 = depth0;
    stack2 = "textTrigger";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i class=\"icon-copy\"></i> ");
    stack1 = depth0;
    stack2 = "common.copy";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    </div>\n  </div>\n  <div class=\"task-detail-log-info\">\n    <div class=\"content-area\" >\n      <div ");
    stack1 = {};
    stack2 = "view.textareaVisible::hidden :task-detail-log-clipboard-wrap";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <textarea class=\"task-detail-log-clipboard\"></textarea>\n      </div>\n      <div ");
    stack1 = {};
    stack2 = "view.textareaVisible:hidden :task-detail-log-maintext";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "content";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/settings", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div class=\"admin-user-settings\">\n  <div>\n    <label>\n      ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.isNotShowBgChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = "checkbox";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "app.settings.notshowBgOperationsPopup";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </label>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/common/time_range", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "view.presetView";
    stack3 = {};
    stack4 = "preset";
    stack3['presetBinding'] = stack4;
    stack4 = "view";
    stack3['widgetBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"span title\">\n  ");
    stack1 = depth0;
    stack2 = "common.timeRange";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-question-sign\"></i>\n</div>\n<ul class=\"display-inline-block nav nav-pills\">\n  ");
    stack1 = depth0;
    stack2 = "view.presets";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "preset";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "from";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.dateFromView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "to";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.dateToView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n</ul>\n<div id=\"slider\">\n  <span class=\"period-marker\">");
    stack1 = depth0;
    stack2 = "view.rangeLabel";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  <span class=\"now-marker\">");
    stack1 = depth0;
    stack2 = "view.nowLabel";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/experimental", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "App.supports";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <table class=\"table table-bordered table-striped\">\n      <thead>\n        <tr>\n          <td style=\"width: 20%; font-weight: bold;\">Experimental Functionality</td>\n          <td style=\"width: 20%; font-weight: bold;\">Enabled?</td>\n          <td style=\"width: 60%; font-weight: bold;\">Description</td>\n        </tr>\n      </thead>\n      <tbody>\n		    ");
    stack1 = depth0;
    stack2 = "view.supports";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "support";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n    </table>\n    <div class=\"control-group\" style=\"margin-bottom: 100px;\">\n	    <div class=\"controls pull-right\">\n	      <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "doCancel";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "form.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n	      <button class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "doSave";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.save";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n	    </div>\n	  </div>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n		      <tr>\n		        <td>");
    stack1 = depth0;
    stack2 = "support.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n		        <td>");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "support.selected";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n		        <td></td>\n			    </tr>\n		    ");
    return buffer;}

  function program5(depth0,data) {
    
    
    data.buffer.push("\n    <div class=\"alert alert-error\">\n      No application supports found. Please login to Ambari and revisit this page.\n    </div>\n  ");}

  function program7(depth0,data) {
    
    
    data.buffer.push("\n  <div class=\"alert alert-error\">\n    No application found. Please login to Ambari and revisit this page.\n  </div>\n");}

    data.buffer.push("\n\n<div class=\"alert alert-error\">\n  <h2>Ambari's Experimental Functionality</h2>\n  <p>\n    <span class=\"label label-important\">Important</span> <strong>Changes made here are neither tested nor supported by Ambari</strong>\n  </p>\n  <p>\n    Ambari's experimental functionality is provided <strong>as-is</strong>.<br>\n    Any changes will effect application behavior, including data corruption. Users making changes do so at their own risk.<br>\n    Experimental functionality can be modified or removed at any time without notice.\n  </p>  \n</div>\n    \n");
    stack1 = depth0;
    stack2 = "App";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(7, program7, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/installer", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div id=\"version-mismatch-title\">");
    stack1 = depth0;
    stack2 = "app.versionMismatchAlert.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"alert alert-danger\">\n    ");
    stack1 = depth0;
    stack2 = "versionConflictAlertBody";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div class=\"wizard\">\n    <div class=\"container\">\n      <div class=\"container-fluid\">\n        <div class=\"row-fluid\">\n          <div class=\"span3\">\n            <!--Sidebar content-->\n            <div class=\"well\">\n              <ul class=\"nav nav-pills nav-stacked\">\n                <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "installer.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n                <li ");
    stack1 = {};
    stack2 = "isStep0:active view.isStep0Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep0";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step0.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep4:active view.isStep4Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep5:active view.isStep5Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep5";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step5.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep6:active view.isStep6Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep6";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step6.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep7:active view.isStep7Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep7";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step7.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep8:active view.isStep8Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep8";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step8.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep9:active view.isStep9Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep9";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step9.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
    stack1 = {};
    stack2 = "isStep10:active view.isStep10Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "gotoStep10";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step10.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              </ul>\n            </div>\n          </div>\n          <div class=\"wizard-content well span9\">\n            ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          </div>\n        </div>\n      </div>\n    </div>\n  </div>\n");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "isServerClientVersionMismatch";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/login", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n<div class=\"well login span4\">\n  <h2>");
    stack1 = depth0;
    stack2 = "login.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  ");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <label>");
    stack1 = depth0;
    stack2 = "login.username";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
    stack1 = depth0;
    stack2 = "view.loginTextField";
    stack3 = {};
    stack4 = "loginName";
    stack3['valueBinding'] = stack4;
    stack4 = "span4";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <label>");
    stack1 = depth0;
    stack2 = "common.password";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
    stack1 = depth0;
    stack2 = "view.passTextField";
    stack3 = {};
    stack4 = "password";
    stack3['type'] = stack4;
    stack4 = "password";
    stack3['valueBinding'] = stack4;
    stack4 = "span4";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <button class=\"btn btn-success\" ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "login.loginButton";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n</div>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"alert alert-error\">\n    ");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "isServerClientVersionMismatch";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div id=\"version-mismatch-title\">");
    stack1 = depth0;
    stack2 = "app.versionMismatchAlert.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"alert alert-danger\">\n    ");
    stack1 = depth0;
    stack2 = "versionConflictAlertBody";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "isClusterDataLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "isClusterDataLoaded";
    stack3 = helpers.unless;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div id=\"main-nav\">\n      <div class=\"navbar\">\n        <div class=\"navbar-inner\">\n          ");
    stack1 = depth0;
    stack2 = "App.MainMenuView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n    ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <h2>");
    stack1 = depth0;
    stack2 = "app.loadingPlaceholder";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " </h2>\n    <div class=\"progress progress-striped active\">\n      <div class=\"bar\" ");
    stack1 = {};
    stack2 = "controller.clusterDataLoadedPercent";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n    </div>\n  ");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "isServerClientVersionMismatch";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.NavItemView";
    stack3 = {};
    stack4 = "category.name";
    stack3['itemBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n              <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "category.url";
    stack3 = depth0;
    stack4 = "goToAdmin";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >");
    stack1 = depth0;
    stack2 = "category.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"row-fluid\">\n  <div id=\"main-admin-menu\" class=\"well span2\">\n      <ul class=\"nav nav-list\">\n        ");
    stack1 = depth0;
    stack2 = "view.categories";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n  </div>\n  <div class=\"span10\">\n    <div class=\"row-fluid\">\n      ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/access", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div id=\"access\">\n  <div class=\"header\">\n      <strong>");
    stack1 = depth0;
    stack2 = "common.access";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n  <div class=\"acces-values\">\n      <label>\n        ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "controller.showJobs";
    stack3['checkedBinding'] = stack4;
    stack4 = "checkbox";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "admin.access.showJobs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      </label>\n  </div>\n  <div class=\"control-group\">\n      <div class=\"controls\">\n          <button type=\"submit\" class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "save";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.save";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n      </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/advanced", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<h5>");
    stack1 = depth0;
    stack2 = "common.advanced";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "<i class=\"icon-question-sign\"></i></h5>\n<div class=\"row\">\n  ");
    stack1 = depth0;
    stack2 = "admin.advanced.caution";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class=\"row\">\n  <button ");
    stack1 = depth0;
    stack2 = "uninstall";
    stack3 = {};
    stack4 = "App.router.mainAdminAdvancedController";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn\"><i class=\"icon-trash\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.advanced.button.uninstallIncludingData";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  <br/>\n  <button ");
    stack1 = depth0;
    stack2 = "view.params";
    stack3 = depth0;
    stack4 = "uninstall";
    stack5 = {};
    stack6 = "App.router.mainAdminAdvancedController";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn\"><i class=\"icon-trash\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.advanced.button.uninstallKeepData";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/advanced/uninstall", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "admin.confirmUninstall";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1));
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/audit", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "column";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "row";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.pager";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;}

    data.buffer.push("\n\n<table class=\"table table-striped\">\n  <thead>\n  <tr>\n    ");
    stack1 = depth0;
    stack2 = "view.columns";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "column";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </tr>\n  </thead>\n  <tbody>\n  ");
    stack1 = depth0;
    stack2 = "view.rows";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "row";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </tbody>\n</table>\n");
    stack1 = depth0;
    stack2 = "view.pager";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/authentication", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.primaryServer";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.secondaryServer";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.useSsl";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.bindMethod";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "view.useCredentials";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.searchBaseDn";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.usernameAttribute";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n    <h5>");
    stack1 = depth0;
    stack2 = "admin.authentication.form.configurationTest";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h5>\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.userDN";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.userPassword";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n    <button ");
    stack1 = depth0;
    stack2 = "testConfiguration";
    stack3 = {};
    stack4 = "click";
    stack3['on'] = stack4;
    stack4 = "view.form";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
    stack1 = depth0;
    stack2 = "admin.authentication.form.testConfiguration";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n\n    ");
    stack1 = depth0;
    stack2 = "view.form.testConfigurationMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.bindUser";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.password";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.passwordRetype";
    stack3['fieldBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <p ");
    stack1 = {};
    stack2 = "view.form.testConfigurationClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.form.testConfigurationMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n    ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "App.FormFieldTemplate";
    stack3 = {};
    stack4 = "view.form.field.method";
    stack3['fieldBinding'] = stack4;
    stack4 = "user-auth-method";
    stack3['id'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n");
    stack1 = depth0;
    stack2 = "view.ldapChecked";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div style=\"margin:40px 0\">\n  <button ");
    stack1 = depth0;
    stack2 = "updateValues";
    stack3 = {};
    stack4 = "view.form";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
    stack1 = depth0;
    stack2 = "form.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  <button ");
    stack1 = depth0;
    stack2 = "view.form";
    stack3 = depth0;
    stack4 = "save";
    stack5 = {};
    stack6 = "App.router.mainAdminAuthenticationController";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary\">");
    stack1 = depth0;
    stack2 = "form.save";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/cluster", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          (<a href=\"#\" ");
    stack1 = {};
    stack2 = "view.isUpgradeAvailable::inactive";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1));
    stack1 = depth0;
    stack2 = "upgradeStack";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "view.isUpgradeAvailable";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </a>)\n        ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "admin.cluster.upgradeAvailable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "upgradeVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "admin.cluster.upgradeUnavailable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "service.isHidden";
    stack3 = helpers.unless;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "service.version";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "service.description";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</td>\n      </tr>\n      ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <tr>\n            <td>");
    stack1 = depth0;
    stack2 = "repo.osType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>");
    stack1 = depth0;
    stack2 = "repo.baseUrl";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n          </tr>\n        ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"admin-cluster\">\n    <div class=\"header\">\n      <strong>");
    stack1 = depth0;
    stack2 = "admin.cluster.stackVersion";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "App.currentStackVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "App.supports.stackUpgrade";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </strong>\n    </div>\n    <table class=\"table table-bordered table-striped\">\n      <thead>\n      <tr>\n        <th>");
    stack1 = depth0;
    stack2 = "common.service";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "common.version";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "common.description";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      </tr>\n      </thead>\n      <tbody>\n      ");
    stack1 = depth0;
    stack2 = "services";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "service";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </tbody>\n    </table>\n\n    <div class=\"header\">\n      <strong>");
    stack1 = depth0;
    stack2 = "admin.cluster.repositories.repositories";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong>\n    </div>\n    <ul class=\"nav nav-tabs\">\n      <li class=\"active\">\n        <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "repositories.stackVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n      </li>\n    </ul>\n\n    <table class=\"table table-bordered table-striped\">\n      <thead>\n      <tr>\n        <th>");
    stack1 = depth0;
    stack2 = "admin.cluster.repositories.os";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "admin.cluster.repositories.baseUrl";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      </tr>\n      </thead>\n      <tbody>\n        ");
    stack1 = depth0;
    stack2 = "repositories";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "repo";
    stack7 = helpers.each;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </tbody>\n    </table>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.supports.autoRollbackHA";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <p class=\"text-success\">\n        ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.enabled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <a class=\"btn btn-padding btn-success\" ");
    stack1 = depth0;
    stack2 = "disableHighAvailability";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.button.disable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      </p>\n    ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n       <p class=\"text-success\">\n        ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.enabled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n       </p>\n    ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "controller.dataIsLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(9, program9, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <p class=\"muted\">\n        ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.disabled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <a class=\"btn btn-padding btn-success\" ");
    stack1 = depth0;
    stack2 = "enableHighAvailability";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.button.enable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      </p>\n    ");
    return buffer;}

  function program9(depth0,data) {
    
    
    data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

    data.buffer.push("\n\n<div>\n  ");
    stack1 = depth0;
    stack2 = "App.isHaEnabled";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/progress", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.taskView";
    stack3 = {};
    stack4 = "task";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n    <div class=\"item\">\n      <div ");
    stack1 = {};
    stack2 = ":pull-left view.linkClass controller.isHA:span4 controller.isRollback:span3 controller.isReassign:span5";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i ");
    stack1 = {};
    stack2 = "view.icon view.iconColor";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n        <a ");
    stack1 = {};
    stack2 = "";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "task";
    stack3 = depth0;
    stack4 = "hostsLogPopup";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >");
    stack1 = depth0;
    stack2 = "task.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n      </div>\n      <div ");
    stack1 = {};
    stack2 = "view.showProgressBar::hide :row :span5 :pull-left";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <div ");
    stack1 = {};
    stack2 = ":progress-bar controller.isRollback::span8 controller.isRollback:span3";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <div class=\"progress-striped active progress-info progress\">\n            <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n          </div>\n        </div>\n        <div class=\"span1\">");
    stack1 = depth0;
    stack2 = "task.progress";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "&#37;</div>\n      </div>\n      <div>\n      ");
    stack1 = depth0;
    stack2 = "task.showRetry";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "task.showRollback";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "task.showSkip";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    </div>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <a ");
    stack1 = depth0;
    stack2 = "retryTask";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n          <i class=\"icon-repeat icon-white\"></i>\n          ");
    stack1 = depth0;
    stack2 = "common.retry";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <a ");
    stack1 = depth0;
    stack2 = "rollback";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n          <i class=\"icon-repeat icon-white\"></i>\n          ");
    stack1 = depth0;
    stack2 = "common.rollBack";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <a ");
    stack1 = depth0;
    stack2 = "skipTask";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n          <i class=\"icon-step-forward icon-white\"></i>\n          ");
    stack1 = depth0;
    stack2 = "common.skip";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
    return buffer;}

    data.buffer.push("\n<div id=\"ha-progress-page\">\n  <h2>");
    stack1 = depth0;
    stack2 = "view.headerTitle";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div ");
    stack1 = {};
    stack2 = "view.noticeClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.notice";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</div>\n  ");
    stack1 = depth0;
    stack2 = "controller.tasks";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "task";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div class=\"btn-area\">\n    <a ");
    stack1 = {};
    stack2 = ":btn controller.isSubmitDisabled:disabled :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "done";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.submitButtonText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/rollback", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div class=\"wizard\">\n  ");
    stack1 = depth0;
    stack2 = "templates/main/admin/highAvailability/progress";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step1", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div id=\"ha-step1\">\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step1.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    <div class=\"alert alert-warn\">\n      ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step1.alert";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n  <div ");
    stack1 = {};
    stack2 = ":control-group view.showInputError:error :form-horizontal :nameserviceid-input";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <label class=\"control-label\">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step1.nameserviceid";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</label>\n\n    <div class=\"controls\">\n      ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "content.nameServiceId";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      <span  ");
    stack1 = {};
    stack2 = ":help-inline view.showInputError::hidden";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step1.nameserviceid.error";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      </span>\n    </div>\n  </div>\n  <div class=\"btn-area\">\n    <a ");
    stack1 = {};
    stack2 = ":btn controller.isNameServiceIdValid::disabled :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <div class=\"control-group\">\n        <label class=\"control-label\">\n          ");
    stack1 = depth0;
    stack2 = "isCurNameNode";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "isAddNameNode";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ":\n        </label>\n        <div class=\"controls\">\n          ");
    stack1 = depth0;
    stack2 = "App.SelectHostView";
    stack3 = {};
    stack4 = "content.host_name";
    stack3['optionValuePath'] = stack4;
    stack4 = "content.host_info";
    stack3['optionLabelPath'] = stack4;
    stack4 = "selectedHost";
    stack3['selectedHostBinding'] = stack4;
    stack4 = "component_name";
    stack3['componentNameBinding'] = stack4;
    stack4 = "host-select";
    stack3['class'] = stack4;
    stack4 = "zId";
    stack3['zIdBinding'] = stack4;
    stack4 = "isInstalled";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "showAddControl";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "showRemoveControl";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n      </div>\n      ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "common.current";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "common.additional";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.AddControlView";
    stack3 = {};
    stack4 = "component_name";
    stack3['componentNameBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.RemoveControlView";
    stack3 = {};
    stack4 = "component_name";
    stack3['componentNameBinding'] = stack4;
    stack4 = "zId";
    stack3['zIdBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"mapping-box round-corners well\">\n      <div class=\"hostString\"><span>");
    stack1 = depth0;
    stack2 = "hostInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n      ");
    stack1 = depth0;
    stack2 = "masterServices";
    stack3 = helpers.each;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <span ");
    stack1 = {};
    stack2 = "isInstalled:assignedService:newService :round-corners";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"remaining-hosts round-corners well\">\n      <span><strong>");
    stack1 = depth0;
    stack2 = "remainingHosts";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</strong> ");
    stack1 = depth0;
    stack2 = "installer.step5.attention";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n    ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step2.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class=\"assign-masters\">\n  <div class=\"select-hosts span7\">\n    <form class=\"form-horizontal\" autocomplete=\"off\">\n      <!-- View for array controller -->\n      ");
    stack1 = depth0;
    stack2 = "servicesMasters";
    stack3 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </form>\n  </div>\n\n  <div class=\"host-assignments span5\">\n    ");
    stack1 = depth0;
    stack2 = "masterHostMapping";
    stack3 = helpers.each;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    ");
    stack1 = depth0;
    stack2 = "remainingHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div style=\"clear: both;\"></div>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step3", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <li>");
    stack1 = depth0;
    stack2 = "item.hostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</li>\n            ");
    return buffer;}

    data.buffer.push("\n\n\n<h2>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.confirm.host.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n<div id=\"ha-step3-content\" class=\"well pre-scrollable\">\n  <div id=\"step8-info\">\n    <table id=\"ha-step3-review-table\">\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.curNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.curNameNode";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td></td>\n      </tr>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.secNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.secondaryNameNode";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td><span class=\"to-be-disabled-red\"><i class=\"icon-minus\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.toBeDeleted";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></td>\n      </tr>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.addNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.addNameNode";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td><span class=\"to-be-installed-green\"><i class=\"icon-plus\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></td>\n      </tr>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.journalNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          <ul>\n            ");
    stack1 = depth0;
    stack2 = "view.journalNodes";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </ul>\n        </td>\n        <td>\n          <ul>\n            <li><span class=\"to-be-installed-green\"><i\n                    class=\"icon-plus\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></li>\n            <li><span class=\"to-be-installed-green\"><i\n                    class=\"icon-plus\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></li>\n            <li><span class=\"to-be-installed-green\"><i\n                    class=\"icon-plus\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></li>\n          </ul>\n        </td>\n      </tr>\n    </table>\n  </div>\n</div>\n\n\n<div id=\"serviceConfig\">\n\n  ");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.confirm.config.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n  ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigView";
    stack3 = {};
    stack4 = "controller.isNotEditable";
    stack3['isNotEditableBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step4", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div id=\"ha-step4\">\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step4.bodyHeader";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "view.step4BodyText";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"btn-area\">\n    <a  ");
    stack1 = {};
    stack2 = "controller.isNextEnabled::disabled :btn :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "done";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <span class=\"pull-right btn-extra-info\">");
    stack1 = depth0;
    stack2 = "view.nnCheckPointText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step5", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "templates/main/admin/highAvailability/progress";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step6", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div id=\"ha-step6\">\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step6.bodyHeader";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "view.step6BodyText";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"btn-area\">\n    <a ");
    stack1 = {};
    stack2 = "controller.isNextEnabled::disabled :btn :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "done";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <span class=\"pull-right btn-extra-info\">");
    stack1 = depth0;
    stack2 = "view.jnCheckPointText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step7", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "templates/main/admin/highAvailability/progress";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step8", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div id=\"ha-step8\">\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step8.bodyHeader";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  ");
    stack1 = depth0;
    stack2 = "view.step8BodyText";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <span class=\"pull-right btn-extra-info\">");
    data.buffer.push("</span>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/step9", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "templates/main/admin/highAvailability/progress";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/highAvailability/wizard", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep4:active view.isStep4Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep5:active view.isStep5Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep5";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step5.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep6:active view.isStep6Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep6";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step6.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep7:active view.isStep7Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep7";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step7.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep8:active view.isStep8Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep8";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step8.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep9:active view.isStep9Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep9";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.step9.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/misc", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "controller.users";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n      <div class=\"span6\">\n        <table class=\"table table-bordered table-striped\">\n          <thead>\n            <tr>\n              <th>");
    stack1 = depth0;
    stack2 = "common.name";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th>");
    stack1 = depth0;
    stack2 = "common.value";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n            </tr>\n          </thead>\n          <tbody>\n            ");
    stack1 = depth0;
    stack2 = "controller.users";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "user";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tbody>\n        </table>\n      </div>\n    ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "user.isVisible";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <tr>\n                  <td>");
    stack1 = depth0;
    stack2 = "user.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                  <td>");
    stack1 = depth0;
    stack2 = "user.value";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                </tr>\n              ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <p>");
    stack1 = depth0;
    stack2 = "admin.misc.nothingToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n    ");
    return buffer;}

  function program8(depth0,data) {
    
    
    data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

    data.buffer.push("\n<div class=\"admin-misc\">\n  <div class=\"header\">\n    <strong>");
    stack1 = depth0;
    stack2 = "admin.misc.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n  ");
    stack1 = depth0;
    stack2 = "controller.dataIsLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/rollbackHA/rollback_wizard", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "admin.highAvailability.wizard.rollback.header.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.rollbackHighAvailability.wizard.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.rollbackHighAvailability.wizard.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.rollbackHighAvailability.wizard.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/rollbackHA/step1", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div>\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.rollbackHighAvailability.wizard.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    <p>\n      Select Additional NameNode host:\n      ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.addNNHosts";
    stack3['contentBinding'] = stack4;
    stack4 = "view.selectedAddNNHost";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </p>\n    <div class=\"code-snippet\">Tip: Suggested host <b>");
    stack1 = depth0;
    stack2 = "view.tipAddNNHost";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b></div>\n  </div>\n  <div class=\"alert alert-info\">\n    <p>\n      Select Secondary NameNode host:\n      ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.sNNHosts";
    stack3['contentBinding'] = stack4;
    stack4 = "view.selectedSNNHost";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </p>\n    <div class=\"code-snippet\">Tip: Suggested host <b>");
    stack1 = depth0;
    stack2 = "view.tipSNNHost";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b></div>\n  </div>\n  <div class=\"btn-area\">\n    <a ");
    stack1 = {};
    stack2 = ":btn controller.isNameServiceIdValid::disabled :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "done";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/rollbackHA/step2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div>\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.rollbackHighAvailability.wizard.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n      ");
    stack1 = depth0;
    stack2 = "view.step2BodyText";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n   <div class=\"btn-area\">\n      <a  ");
    stack1 = {};
    stack2 = ":btn :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n      <span class=\"pull-right btn-extra-info\">");
    stack1 = depth0;
    stack2 = "view.nnCheckPointText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n    </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/rollbackHA/step3", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<div>\n  <h2>");
    stack1 = depth0;
    stack2 = "admin.rollbackHighAvailability.wizard.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"btn-area\">\n    <a ");
    stack1 = {};
    stack2 = ":btn controller.isNameServiceIdValid::disabled :btn-success :pull-right";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "securityEnabled";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div>\n      <p class=\"text-success\">");
    stack1 = depth0;
    stack2 = "admin.security.enabled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <a\n                class=\"btn btn-padding btn-warning\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "notifySecurityOffPopup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.button.disable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a> <br/>\n      </p>\n    </div>\n    <div id=\"serviceConfig\">\n      ");
    stack1 = depth0;
    stack2 = "App.ServicesConfigView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div>\n      <p class=\"muted\">");
    stack1 = depth0;
    stack2 = "admin.security.disabled";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <a\n                class=\"btn btn-padding btn-success\" ");
    stack1 = depth0;
    stack2 = "addSecurity";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.button.enable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a> <br/>\n      </p>\n    </div>\n  ");
    return buffer;}

  function program6(depth0,data) {
    
    
    data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "dataIsLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/add/menu", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n\n<div class=\"wizard\" id=\"enable_security\">\n    <div class=\"container\">\n        <div class=\"container-fluid\">\n            <div class=\"row-fluid\">\n                <div class=\"span3\">\n                    <!--Sidebar content-->\n                    <div class=\"well\">\n                        <ul class=\"nav nav-pills nav-stacked\">\n                            <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "admin.addSecurity.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n                            <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            <li ");
    stack1 = {};
    stack2 = "isStep4:active view.isStep4Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                        </ul>\n                    </div>\n                </div>\n                <div class=\"wizard-content well span9\">\n                  ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </div>\n            </div>\n        </div>\n    </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/add/step1", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <li>");
    stack1 = depth0;
    stack2 = "admin.security.step1.body.instruction5";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</li>\n      ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "admin.security.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<p class=\"alert\">\n  ");
    stack1 = depth0;
    stack2 = "admin.security.step1.body.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</p>\n<div class=\"alert alert-info\">\n    <ol>\n        <li>");
    stack1 = depth0;
    stack2 = "admin.security.step1.body.instruction1";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n        <li>");
    stack1 = depth0;
    stack2 = "admin.security.step1.body.instruction2";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n        <li>");
    stack1 = depth0;
    stack2 = "admin.security.step1.body.instruction3";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n        <li>");
    stack1 = depth0;
    stack2 = "admin.security.step1.body.instruction4";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n      ");
    stack1 = depth0;
    stack2 = "view.isATSInstalled";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </ol>\n</div>\n\n\n<div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/add/step2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "admin.security.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div id=\"serviceConfig\">\n  <p class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "admin.security.step2.body.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </p>\n  ");
    stack1 = depth0;
    stack2 = "App.ServicesConfigView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n  <div class=\"btn-area\">\n    <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n\n    <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/add/step3", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <tr>\n                <td>");
    stack1 = depth0;
    stack2 = "hostComponent.host";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                <td>");
    stack1 = depth0;
    stack2 = "hostComponent.component";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                <td>");
    stack1 = depth0;
    stack2 = "hostComponent.principal";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                <td>");
    stack1 = depth0;
    stack2 = "hostComponent.keytab";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            </tr>\n        ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "admin.security.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "admin.security.step3.notice";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n<div class=\"step3\">\n    <table class=\"table table-bordered table-striped\">\n        <thead>\n        <tr>\n            <th>");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
    stack1 = depth0;
    stack2 = "common.component";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
    stack1 = depth0;
    stack2 = "admin.security.step3.table.principal";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
    stack1 = depth0;
    stack2 = "admin.security.step3.table.keytab";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n            </tr>\n        </thead>\n        <tbody>\n        ");
    stack1 = depth0;
    stack2 = "hostComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "hostComponent";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n    </table>\n</div>\n<div class=\"btn-area\">\n    <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    <div class=\"pull-right\">\n        <a class=\"btn btn-info\" ");
    stack1 = depth0;
    stack2 = "doDownloadCsv";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.security.step3.downloadCSV";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        <a class=\"btn btn-success\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/add/step4", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <p ");
    stack1 = {};
    stack2 = "view.msgColor :alert";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "admin.security.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "App.MainServiceReconfigureView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
    stack1 = {};
    stack2 = "isBackBtnDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "done";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.done";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " </a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/disable", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <p ");
    stack1 = {};
    stack2 = "view.msgColor :alert";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "App.MainServiceReconfigureView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "done";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.done";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " </a>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/security/notify_security_off_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <br />\n    ");
    stack1 = depth0;
    stack2 = "admin.security.disable.popup.body.warning";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"alert\">\n  ");
    stack1 = depth0;
    stack2 = "admin.security.disable.popup.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "view.isMapReduceInstalled";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/stack_upgrade", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div id=\"stack-upgrade\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.start";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.confirm";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.upgrade";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"content well span9\">\n          ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/user", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n<p class=\"text-info\">");
    stack1 = depth0;
    stack2 = "admin.users.ldapAuthUsed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ".</p>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n<table class=\"table table-bordered table-striped span6\">\n  <thead>\n  <tr>\n    <th style=\"width:50%\">");
    stack1 = depth0;
    stack2 = "admin.users.username";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    <th style=\"width:10%\">");
    stack1 = depth0;
    stack2 = "admin.users.privileges";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    <th style=\"width:20%\">");
    stack1 = depth0;
    stack2 = "admin.users.type";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    <th style=\"width:20%\">");
    stack1 = depth0;
    stack2 = "admin.users.action";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n  </tr>\n  </thead>\n  <tbody>\n  ");
    stack1 = depth0;
    stack2 = "view.users";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "user";
    stack7 = helpers.each;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </tbody>\n</table>\n<div class=\"span2\">\n  <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "gotoCreateUser";
    stack3 = {};
    stack4 = "click";
    stack3['on'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >\n    <i class=\"icon-plus\"></i>");
    stack1 = depth0;
    stack2 = "admin.users.addButton";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </button>\n</div>\n");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "App.MainAdminUserRowView";
    stack3 = {};
    stack4 = "user";
    stack3['userBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.ldapUser";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/user/create", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "field.isHidden";
    stack3 = helpers.unless;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div ");
    stack1 = {};
    stack2 = "field.errorMessage:error field.warnMessage:warning :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <label class=\"control-label\" for=\"input");
    stack1 = depth0;
    stack2 = "field.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "field.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n    <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "field.viewClass";
    stack3 = {};
    stack4 = "field.value";
    stack3['valueBinding'] = stack4;
    stack4 = "field.disabled";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "field.errorMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "field.warnMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n  </div>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "field.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "field.warnMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
    return buffer;}

    data.buffer.push("\n\n<form class=\"form-horizontal\" autocomplete=\"off\">\n  ");
    stack1 = depth0;
    stack2 = "view.userForm.fields";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "field";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div class=\"control-group\">\n    <div class=\"controls\">\n      <button type=\"submit\" class=\"btn\" ");
    stack1 = depth0;
    stack2 = "gotoUsers";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "form.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n      <button type=\"submit\"\n              class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "create";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">Create</button>\n    </div>\n  </div>\n</form>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/user/edit", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "field.isHidden";
    stack3 = helpers.unless;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div ");
    stack1 = {};
    stack2 = "field.errorMessage:error :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <label class=\"control-label\" for=\"input");
    stack1 = depth0;
    stack2 = "field.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "field.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n    <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "field.viewClass";
    stack3 = {};
    stack4 = "field.value";
    stack3['valueBinding'] = stack4;
    stack4 = "field.disabled";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "field.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n    </div>\n  </div>\n  ");
    return buffer;}

    data.buffer.push("\n\n<form class=\"form-horizontal\" autocomplete=\"off\">\n  ");
    stack1 = depth0;
    stack2 = "view.userForm.fields";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "field";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div class=\"control-group\">\n    <div class=\"controls\">\n      <button type=\"submit\" class=\"btn\" ");
    stack1 = depth0;
    stack2 = "gotoUsers";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "form.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n      <button type=\"submit\"\n              class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "edit";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.save";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n    </div>\n  </div>\n</form>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/admin/user/row", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<tr>\n  <td>");
    stack1 = depth0;
    stack2 = "user.userName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.disableCheckBoxes";
    stack3['disabledBinding'] = stack4;
    stack4 = "user.admin";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
    stack1 = depth0;
    stack2 = "user.type";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "user";
    stack3 = depth0;
    stack4 = "gotoEditUser";
    stack5 = {};
    stack6 = "click";
    stack5['on'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.users.edit";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>&nbsp;\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "user";
    stack3 = depth0;
    stack4 = "deleteRecord";
    stack5 = {};
    stack6 = "App.router.mainAdminUserController";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "admin.users.delete";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  </td>\n</tr>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/apps", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "controller.columnsName";
    stack3 = helpers.each;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.sortingColumns";
    stack3 = {};
    stack4 = "this";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n         <tr>\n           <td class=\"no-data\" ");
    stack1 = {};
    stack2 = "controller.columnsName.content.length";
    stack1['colspan'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "apps.filters.nothingToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n         </tr>\n       ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n         ");
    stack1 = depth0;
    stack2 = "content";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "run";
    stack7 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n       ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n           ");
    stack1 = depth0;
    stack2 = "view.containerRow";
    stack3 = {};
    stack4 = "run";
    stack3['runBinding'] = stack4;
    stack4 = "view.appTableRow";
    stack3['currentViewBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n         ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"apps\">\n  <table class=\"table table-bordered avg-table\">\n    <tbody>\n    <tr>\n      <td></td>\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.jobs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.input";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.output";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.duration";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.oldest";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.mostRecent";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n    <tr class=\"avg-info\">\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.avg";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.jobs.avg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.input.avg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.output.avg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.duration.avg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.times.oldest";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.times.youngest";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n\n    </tr>\n    <tr class=\"compare-info\">\n      <td>");
    stack1 = depth0;
    stack2 = "apps.avgTable.min";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "apps.avgTable.max";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.jobs.min";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "summary.jobs.max";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.input.min";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "summary.input.max";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.output.min";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "summary.output.max";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "summary.duration.min";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "summary.duration.max";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td></td>\n      <td></td>\n    </tr>\n    </tbody>\n  </table>\n\n  <div class=\"filter_info\">\n\n    <div class=\"span4\" id=\"filter_buttons\">");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n      <a class=\"all\" ");
    stack1 = depth0;
    stack2 = "clickViewType";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " data-view-type=\"all\">");
    stack1 = depth0;
    stack2 = "tableView.filters.all";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        (");
    stack1 = depth0;
    stack2 = "controller.paginationObject.iTotalRecords";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</a> &#124;\n      <a class=\"filtered\" ");
    stack1 = depth0;
    stack2 = "clickViewType";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " data-view-type=\"filtered\">");
    stack1 = depth0;
    stack2 = "tableView.filters.filtered";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        (");
    stack1 = depth0;
    stack2 = "controller.filterObject.filteredDisplayRecords";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</a>\n    </div>\n    <div class=\"span2 clear_filter\">\n        <a ");
    stack1 = depth0;
    stack2 = "clearFilters";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "tableView.filters.clearFilters";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    </div>\n    <div class=\"search-bar\">\n        ");
    stack1 = depth0;
    stack2 = "view.appSearchThrough";
    stack3 = {};
    stack4 = "controller.filterObject.sSearch";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n  <table class=\"table table-striped runsList\">\n    <thead>\n    ");
    stack1 = depth0;
    stack2 = "view.wrapSorting";
    stack3 = helpers.view;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <tr>\n      <th>");
    stack1 = depth0;
    stack2 = "view.appIdFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.nameFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.typeFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.userFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.jobsFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.inputFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.outputFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.durationFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.runDateFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n       ");
    stack1 = depth0;
    stack2 = "view.emptyData";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(7, program7, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n  </table>\n\n  <div class=\"page-bar\">\n      <div class=\"items-on-page\">\n          <label>");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.runPerPageSelectView";
    stack3 = {};
    stack4 = "runPerPageSelectView";
    stack3['viewName'] = stack4;
    stack4 = "controller.filterObject.iDisplayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n      </div>\n      <div class=\"info\">");
    stack1 = depth0;
    stack2 = "view.paginationInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"paging_two_button\">\n          ");
    stack1 = depth0;
    stack2 = "view.paginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.paginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/apps/item", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n    <li ");
    stack1 = {};
    stack2 = "tab.active";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "tab";
    stack3 = depth0;
    stack4 = "switchTab";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <a href=\"javascript:void(0)\">");
    stack1 = depth0;
    stack2 = "tab.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    </li>\n    ");
    return buffer;}

    data.buffer.push("\n\n<td colspan=\"10\">\n  <ul class=\"nav nav-tabs\">\n    ");
    stack1 = depth0;
    stack2 = "view.menuTabs";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "tab";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </ul>\n  <div class=\"content\">\n    ");
    stack1 = depth0;
    stack2 = "view.containerView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n </td>\n\n\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/apps/item/bar", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', foundHelper, self=this;


    data.buffer.push("\n\n<div id=\"job_tasks\"></div>\n<div class=\"clearfix\"></div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/apps/item/dag", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.hasManyJobs";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.idSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.statusSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.mapsSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.reducesSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.inputSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.outputSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.parentView.durationSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    return buffer;}
  function program2(depth0,data) {
    
    var stack1, stack2, stack3;
    stack1 = depth0;
    stack2 = "view.parentView.nameSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <tr>\n        ");
    stack1 = depth0;
    stack2 = "view.hasManyJobs";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        <td>");
    stack1 = depth0;
    stack2 = "job.id";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "job.status";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "job.maps";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "job.reduces";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "job.inputFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "job.outputFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "job.duration";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("<td>");
    stack1 = depth0;
    stack2 = "job.workflow_entity_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>");
    return buffer;}

    data.buffer.push("\n    <div id=\"dag_viewer\"></div>\n\n    <div id=\"dag_viewer_legend\">\n      <ul>\n        <li><span id=\"job\"></span>Unfinished Job</li>\n        <li><span id=\"allmap\"></span>All Maps on Cluster</li>\n        <li><span id=\"allshuffle\"></span>All Shuffles on Cluster</li>\n        <li><span id=\"allreduce\"></span>All Reduces on Cluster</li>\n        <li><span id=\"jobdone\"></span>Finished Job</li>\n        <li><span id=\"map\"></span>Maps for this App</li>\n        <li><span id=\"shuffle\"></span>Shuffles for this App</li>\n        <li><span id=\"reduce\"></span>Reduces for this App</li>\n      </ul>\n    </div>\n\n    <table class=\"table table-bordered table-striped\" id=\"innerTable\">\n      <thead>\n      <tr>\n        ");
    stack1 = depth0;
    stack2 = "view.sortView";
    stack3 = {};
    stack4 = "controller.content.jobs";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </tr>\n      </thead>\n      <tbody>\n      ");
    stack1 = depth0;
    stack2 = "controller.content.jobs";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "job";
    stack7 = helpers.each;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </tbody>\n    </table>\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/apps/list_row", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n<td rel=\"popover\" ");
    stack1 = {};
    stack2 = "apps.isRunning.popup.title";
    stack1['title'] = stack2;
    stack2 = "apps.isRunning.popup.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n    <span title=\"");
    stack1 = depth0;
    stack2 = "run.lastUpdateTimeFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "run.lastUpdateTimeFormattedShort";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n    <b class='red'>*</b></td>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n<td>\n    <span title=\"");
    stack1 = depth0;
    stack2 = "run.lastUpdateTimeFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "run.lastUpdateTimeFormattedShort";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n</td>\n");
    return buffer;}

    data.buffer.push("\n\n<td class=\"appId\" title=\"");
    stack1 = depth0;
    stack2 = "run.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "run.idFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.appName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.type";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.userName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.numJobsTotal";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.inputFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.outputFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n<td>");
    stack1 = depth0;
    stack2 = "run.duration";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n");
    stack1 = depth0;
    stack2 = "run.isRunning";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/apps/user_filter", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <li>\n      <label class=\"checkbox\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "user.checked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "user.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n    </li>\n  ");
    return buffer;}

    data.buffer.push("\n<button class=\"btn btn-info single-btn-group\" ");
    stack1 = depth0;
    stack2 = "clickFilterButton";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  ");
    stack1 = depth0;
    stack2 = "common.user";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "&nbsp;<span class=\"caret\"></span>\n</button>\n<ul class=\"dropdown-menu filter-components\">\n  ");
    stack1 = depth0;
    stack2 = "view.users";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "user";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <li>\n    <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "closeFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n    <button class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "applyFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  </li>\n</ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/charts", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/charts/heatmap", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n               <li class=\"dropdown-submenu\">\n                <a tabindex=\"-1\" >");
    stack1 = depth0;
    stack2 = "category.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                <ul class=\"dropdown-menu\">\n                  ");
    stack1 = depth0;
    stack2 = "category.items";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "metric";
    stack7 = helpers.each;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </ul>\n               </li>\n            ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                    <li>\n                      <a tabindex=\"-1\" ");
    stack1 = depth0;
    stack2 = "metric";
    stack3 = depth0;
    stack4 = "showHeatMapMetric";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "metric.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                    </li>\n                  ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n					<table class=\"legend\">\n					  ");
    stack1 = depth0;
    stack2 = "controller.selectedMetric.slotDefinitions";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "slot";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n					</table>\n	        ");
    stack1 = depth0;
    stack2 = "common.maximum";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n	        <div id=\"inputMaximum\" class=\"control-group\">\n	          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "text";
    stack3['type'] = stack4;
    stack4 = "8";
    stack3['maxlength'] = stack4;
    stack4 = "controller.inputMaximum";
    stack3['valueBinding'] = stack4;
    stack4 = "span6";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n	          ");
    stack1 = depth0;
    stack2 = "controller.selectedMetric.units";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n	        </div>\n        ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n					    <tr>\n					      <td>\n					        <div class=\"tile\" ");
    stack1 = {};
    stack2 = "slot.cssStyle";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n					      </td>\n					      <td>");
    stack1 = depth0;
    stack2 = "slot.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n					    </tr>\n					  ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n				    <div ");
    stack1 = {};
    stack2 = "controller.rackClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n				      ");
    stack1 = depth0;
    stack2 = "App.MainChartsHeatmapRackView";
    stack3 = {};
    stack4 = "rack";
    stack3['rackBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n				    </div>\n				  ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"heatmap\">\n\n  <div class=\"container-fluid\">\n	  <div class=\"row-fluid\">\n	    <div class=\"span2 legend-column\">\n		     <div class=\"btn-group\">\n				  <button class=\"btn\">");
    stack1 = depth0;
    stack2 = "charts.heatmap.selectMetric";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n				  <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n				    <span class=\"caret\"></span>\n				  </button>\n				  <ul class=\"dropdown-menu\">\n				    ");
    stack1 = depth0;
    stack2 = "controller.allMetrics";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n				  </ul>\n				</div>\n        ");
    stack1 = depth0;
    stack2 = "controller.selectedMetric";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	    </div>\n	    <div class=\"span10\">\n	      <h4 id=\"heatmap-metric-loading\">\n	        <span id=\"heatmap-metric-title\">");
    stack1 = depth0;
    stack2 = "controller.selectedMetric.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n	      </h4>\n	      <div class=\"row-fluid\">\n				  ");
    stack1 = depth0;
    stack2 = "controller.racks";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "rack";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n			  </div>\n			  ");
    stack1 = depth0;
    stack2 = "App.MainChartsHeatmapHostDetailView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n	    </div>\n	  </div>\n  </div>\n\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/charts/heatmap/heatmap_host", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div ");
    stack1 = {};
    stack2 = "view.hostClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.hostTemperatureStyle";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <a href=\"#\" class=\"heatmap-host\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></a>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/charts/heatmap/heatmap_host_detail", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.details.metricName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.metricValue";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br/>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "common.disk";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.diskUsage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "% <br/>\n");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "common.cpu";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.cpuUsage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "%<br/>\n");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "common.memory";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.memoryUsage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "%<br/>\n");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.hostComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br/>\n");
    return buffer;}

    data.buffer.push("\n\n<b>");
    stack1 = depth0;
    stack2 = "view.details.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b><br/>\n");
    stack1 = depth0;
    stack2 = "common.os";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.osType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br/>\n");
    stack1 = depth0;
    stack2 = "common.ipAddress";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.details.ip";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br/>\n");
    stack1 = depth0;
    stack2 = "view.details.metricName";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.details.diskUsage";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.details.cpuUsage";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.details.memoryUsage";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.details.hostComponents";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/charts/heatmap/heatmap_rack", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div ");
    stack1 = {};
    stack2 = "view.hostCssStyle";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "App.MainChartsHeatmapHostView";
    stack3 = {};
    stack4 = "this";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"rackHeader\">\n  <div class=\"statusName clearfix\">\n    <div ");
    stack1 = {};
    stack2 = "view.statusIndicator rack.status";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n    <div class=\"rackName\">");
    stack1 = depth0;
    stack2 = "rack.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n  </div>\n</div>\n<!--<div class=\"hostsSummary clearfix\">-->\n  <!--<div class=\"textBlock\">Total hosts: ");
    data.buffer.push("</div>-->\n  <!--");
    data.buffer.push("-->\n    <!--<div class=\"textBlock\">-->\n      <!--<div class=\"indicatorG\"></div>-->\n      <!--Hosts in live state: ");
    data.buffer.push("-->\n    <!--</div>-->\n    <!--<div class=\"textBlock\">-->\n      <!--<div class=\"indicatorY\"></div>-->\n      <!--Hosts in critical state: ");
    data.buffer.push("-->\n    <!--</div>-->\n    <!--<div class=\"textBlock\">-->\n      <!--<div class=\"indicatorR\"></div>-->\n      <!--Hosts in dead state: ");
    data.buffer.push("-->\n    <!--</div>-->\n  <!--");
    data.buffer.push("-->\n<!--</div>-->\n<div class=\"isActive hosts clearfix\">\n  ");
    stack1 = depth0;
    stack2 = "view.hosts";
    stack3 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/charts/linear_time", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-title\" class=\"chart-title\">");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"screensaver chart-container\" ");
    stack1 = {};
    stack2 = "view.isReady:hide";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >\n  <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-title\" class=\"chart-title\">");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n</div>\n<div  id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-container\" class=\"chart-container hide\" ");
    stack1 = {};
    stack2 = "view.isReady:show";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "  rel=\"ZoomInTooltip\" title=\"Click to zoom\">\n  <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-yaxis\" class=\"chart-y-axis\" ");
    stack1 = depth0;
    stack2 = "showGraphInPopup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-xaxis\" class=\"chart-x-axis\" ");
    stack1 = depth0;
    stack2 = "showGraphInPopup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-legend\" class=\"chart-legend\"></div>\n  <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-chart\" class=\"chart\"  ");
    stack1 = depth0;
    stack2 = "showGraphInPopup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  <div id=\"");
    stack1 = depth0;
    stack2 = "view.id";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-timeline\" class=\"timeline\" ");
    stack1 = depth0;
    stack2 = "showGraphInPopup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n  ");
    stack1 = depth0;
    stack2 = "view.noTitleUnderGraph";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n</div>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.isClassicDashboard";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <div class=\"row\">\n      <div class=\"span6\">\n        <div class=\"row\">\n          <div class=\"span6\">\n            <div class=\"box\">\n              <div class=\"box-header\">\n                <h4>");
    stack1 = depth0;
    stack2 = "dashboard.services";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n              </div>\n              <dl class=\"dl-horizontal services\">\n                ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </dl>\n            </div>\n          </div>\n        </div>\n      </div>\n      <div class=\"span6\">\n        <div class=\"box\">\n          <div class=\"box-header\">\n            <h4>");
    stack1 = depth0;
    stack2 = "dashboard.clusterMetrics";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n              <div class=\"btn-group\">\n                <a class=\"btn\" target=\"_blank\" rel=\"tooltip\" title=\"Go to Ganglia\" ");
    stack1 = {};
    stack2 = "view.gangliaUrl";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-link\"></i></a>\n                <a href=\"javascript:void(null)\" class=\"btn btn-success switch-button\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "switchToNew";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  ");
    stack1 = depth0;
    stack2 = "dashboard.button.switchShort";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                  <i class=\"icon-forward\"></i>\n                </a>\n              </div>\n          </div>\n          <div class=\"graphs-container\">\n            <table class=\"graphs\">\n              <tr>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "App.ChartClusterMetricsNetwork";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "App.ChartClusterMetricsLoad";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n              </tr>\n              <tr>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "App.ChartClusterMetricsMemory";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "App.ChartClusterMetricsCPU";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n              </tr>\n            </table>\n          </div>\n        </div>\n      </div>\n    </div>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "item.viewName";
    stack3 = {};
    stack4 = "item.model";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"row-fluid\">\n      <div class=\"services-menu well span2\" style=\"padding: 8px 0\">\n        ");
    stack1 = depth0;
    stack2 = "App.MainServiceMenuView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "App.AllServicesActionView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n\n      <div class=\"span10\" id=\"dashboard-widgets-container\">\n        <div class=\"box\">\n          <div class=\"box-header\">\n            <div class=\"row-fluid\">\n              <h4 class=\"span10\">");
    stack1 = depth0;
    stack2 = "dashboard.widgets";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n                <a class=\"add-widget-button span1\">");
    stack1 = depth0;
    stack2 = "view.plusButtonFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                <div class=\"btn-group\">\n                  <button class=\"btn dropdown-toggle span1 more-options-button\"  data-toggle=\"dropdown\">\n                    <i class=\"icon-cog\"></i> &nbsp;\n                    <span class= \"caret\"></span>\n                  </button>\n                  <ul class=\"dropdown-menu right-align-dropdown\">\n                    <li>\n                      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "resetAllWidgets";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <i class=\"icon-refresh\"></i>\n                        ");
    stack1 = depth0;
    stack2 = "dashboard.button.reset";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                    <li>\n                      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "switchToClassic";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <i class=\"icon-backward\"></i>\n                        ");
    stack1 = depth0;
    stack2 = "dashboard.button.switch";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                    <li>\n                      <a target=\"_blank\" ");
    stack1 = {};
    stack2 = "view.gangliaUrl";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <i class=\"icon-share\"></i>\n                        ");
    stack1 = depth0;
    stack2 = "dashboard.button.gangliaLink";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                  </ul>\n                </div>\n            </div>\n          </div>\n\n          <div id=\"dashboard-widgets\"  class=\"widgets-container\">\n            <div class=\"thumbnails row-fluid\" id=\"sortable\">\n              ");
    stack1 = depth0;
    stack2 = "view.visibleWidgets.length";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </div>\n          </div>\n\n        </div>\n      </div>\n    </div>\n  ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "view.visibleWidgets";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "widgetClass";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <div ");
    stack1 = {};
    stack2 = "widgetClass.class";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    ");
    stack1 = depth0;
    stack2 = "widgetClass";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  </div>\n                ");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.isDataLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/alert_notification_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <p>");
    stack1 = depth0;
    stack2 = "view.warnAlertsMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n    ");
    stack1 = depth0;
    stack2 = "view.warnAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <p class=\"alertsPopupLinks\">\n        <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "selectService";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.alerts.goToService";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "viewNagiosUrl";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = "mouseUp";
    stack3['on'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.alerts.goToNagios";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    </p>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <ul id='summary-alerts-list' class=\"alerts\">\n          ");
    stack1 = depth0;
    stack2 = "view.warnAlerts";
    stack3 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n    ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "App.AlertItemView";
    stack3 = {};
    stack4 = "this";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    
    data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "controller.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/edit_widget_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n         ");
    stack1 = depth0;
    stack2 = "view.configPropertyObj.isGreenOrangeRed";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n       ");
    return buffer;}
  function program2(depth0,data) {
    
    
    data.buffer.push("\n             <div class=\"progress span9\">\n                 <div class=\"bar bar-success\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-warning\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-danger\" style=\"width: 34%;\"></div>\n             </div>\n         ");}

  function program4(depth0,data) {
    
    
    data.buffer.push("\n             <div class=\"progress span9\">\n                 <div class=\"bar bar-danger\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-warning\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-success\" style=\"width: 34%;\"></div>\n             </div>\n         ");}

  function program6(depth0,data) {
    
    
    data.buffer.push("\n         <div class=\"span9\" id=\"slider-range\"></div>\n       ");}

    data.buffer.push("\n<form class=\"form-horizontal\" autocomplete=\"off\">\n    <div class=\"each-row\">\n        <div class=\"alert alert-info\">\n          ");
    stack1 = depth0;
    stack2 = "view.configPropertyObj.hintInfo";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n    </div>\n\n    <div class=\"row-fluid\" id= \"min-height-limit\">\n       ");
    stack1 = depth0;
    stack2 = "view.configPropertyObj.isIE9";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n\n    <div class=\"row-fluid\">\n        <div id=\"slider-value1\" class=\"value-on-slider span2\">0</div>\n        <div id=\"slider-value2\" ");
    stack1 = {};
    stack2 = "view.configPropertyObj.isThresh1Error:slider-error :value-on-slider :span4";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.configPropertyObj.thresh1";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "view.configPropertyObj.errorMessage1";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n        <div id=\"slider-value3\" ");
    stack1 = {};
    stack2 = "view.configPropertyObj.isThresh2Error:slider-error :value-on-slider :span4";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "view.configPropertyObj.thresh2";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "view.configPropertyObj.errorMessage2";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n        <div id=\"slider-value4\" class=\"value-on-slider span2\">");
    stack1 = depth0;
    stack2 = "view.configPropertyObj.maxValue";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n\n</form>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/plus_button_filter", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <li>\n      <ul>\n        ");
    stack1 = depth0;
    stack2 = "view.hiddenWidgets";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "widget";
    stack7 = helpers.each;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n    </li>\n    <li>\n      <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "closeFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n      <button class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "applyFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n    </li>\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <li>\n            <label class=\"checkbox\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "widget.checked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "widget.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </label>\n          </li>\n        ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <li>\n      <ul>\n      <li>");
    stack1 = depth0;
    stack2 = "dashboard.widgets.nothing";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n      </ul>\n    </li>\n    <li>\n      <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "closeFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n    </li>\n\n  ");
    return buffer;}

    data.buffer.push("\n\n<button class=\"btn single-btn-group add-button\" ");
    stack1 = depth0;
    stack2 = "clickFilterButton";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >\n    <i class=\"icon-plus\"></i> <span style=\"font-weight: bold\">");
    stack1 = depth0;
    stack2 = "common.add";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n</button>\n<ul class=\"dropdown-menu filter-components\" >\n  ");
    stack1 = depth0;
    stack2 = "view.hiddenWidgets.length";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/flume", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n<div class=\"clearfix like_pointer\" ");
    stack1 = depth0;
    stack2 = "toggleInfoView";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"name span2\">\n    ");
    stack1 = depth0;
    stack2 = "view.isCollapsed";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.summaryHeader";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n<div class=\"service-body\">\n  <div class=\"span4\">\n    <table class=\"table no-borders\">\n      <tbody>\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-down pull-left\"></i>\n    ");}

  function program4(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-right pull-left\"></i>\n    ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "channelsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n                <li> ");
    stack1 = depth0;
    stack2 = "channels";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "sourcesCount";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n                <li> ");
    stack1 = depth0;
    stack2 = "sources";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "sinksCount";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n                <li> ");
    stack1 = depth0;
    stack2 = "sinks";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    return buffer;}

  function program17(depth0,data) {
    
    
    data.buffer.push("\n      </tbody>\n    </table>\n  </div>\n</div>\n");}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <!-- Flume Agents -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.agentsLabel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n            <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.flumeServerComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.summaryHeader";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n        </td>\n      </tr>\n      <!-- Channels -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.channels";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.service.channelsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.channels";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <ul>\n            ");
    stack1 = depth0;
    stack2 = "view.service.nodes";
    stack3 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </ul>\n        </td>\n      </tr>\n      <!-- Sources -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.sources";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.service.sourcesCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.sources";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <ul>\n            ");
    stack1 = depth0;
    stack2 = "view.service.nodes";
    stack3 = helpers.each;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </ul>\n        </td>\n      </tr>\n      <!-- Sinks -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.sinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.service.sinksCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.flume.sinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <ul>\n            ");
    stack1 = depth0;
    stack2 = "view.service.nodes";
    stack3 = helpers.each;
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </ul>\n        </td>\n      </tr>\n    ");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/hbase", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n<div class=\"clearfix like_pointer\" ");
    stack1 = depth0;
    stack2 = "toggleInfoView";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"name span2\">\n    ");
    stack1 = depth0;
    stack2 = "view.isCollapsed";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.summaryHeader";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n<div class=\"service-body\">\n  <div class=\"span4\">\n    <table class=\"table no-borders\">\n      <tbody>\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-down pull-left\"></i>\n    ");}

  function program4(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-right pull-left\"></i>\n    ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.liveComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "/");
    stack1 = depth0;
    stack2 = "view.totalComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      </tbody>\n    </table>\n  </div>\n<div class=\"dashboard-mini-chart span2\">\n  ");
    stack1 = depth0;
    stack2 = "App.ChartServiceMetricsHBASE_RegionServerReadWriteRequests";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  \n  ");
    stack1 = depth0;
    stack2 = "view.service.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n</div>\n");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.service";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(12, program12, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"btn-group\">\n    <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n      ");
    stack1 = depth0;
    stack2 = "common.quickLinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      <span class=\"caret\"></span>\n    </a>\n    <ul class=\"dropdown-menu\">\n      ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </ul>\n  </div>\n  ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n      ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <!-- HBase Master Server -->\n\n      ");
    stack1 = depth0;
    stack2 = "view.dashboardMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n      <!-- RegionServers -->\n      <tr>\n        <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.regionServerComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.regionServers";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n        <td>\n          <span>\n            ");
    stack1 = depth0;
    stack2 = "view.regionServersLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </span> ");
    stack1 = depth0;
    stack2 = "services.service.summary.RegionServersLIVE";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- Regions in Transition -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.regions.transition";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.service.regionsInTransition";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- HBase Master Started Time -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.masterStarted";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.masterStartedTime";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- HBase Master Activated Time -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.masterActivated";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.masterStartedTime";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Average Load -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.averageLoad";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.averageLoad";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Master Server Heap -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.masterServerHeap";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.masterServerHeapSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n\n    ");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/hdfs", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n<div class=\"clearfix like_pointer\" ");
    stack1 = depth0;
    stack2 = "toggleInfoView";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"name span2\">\n    ");
    stack1 = depth0;
    stack2 = "view.isCollapsed";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.summaryHeader";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n<div class=\"service-body\">\n  <div class=\"span4\">\n    <table class=\"table no-borders\">\n      <tbody>\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-down pull-left\"></i>\n    ");}

  function program4(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-right pull-left\"></i>\n    ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.liveComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "/");
    stack1 = depth0;
    stack2 = "view.totalComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n        <tr>\n          <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.journalNodeComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.journalnodes";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          <td>\n                <span>\n                  ");
    stack1 = depth0;
    stack2 = "view.journalNodesLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </span>  ");
    stack1 = depth0;
    stack2 = "services.service.summary.JournalNodesLive";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n         </td>\n        </tr>\n      ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "view.liveComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "/");
    stack1 = depth0;
    stack2 = "view.totalComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      </tbody>\n    </table>\n  </div>\n<div class=\"dashboard-mini-chart span2\">\n  <div ");
    stack1 = {};
    stack2 = "view.capacity";
    stack1['title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    ");
    stack1 = depth0;
    stack2 = "view.Chart";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"chartLabel\">");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.chart.label";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
    stack1 = depth0;
    stack2 = "view.service.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n</div>\n");
    return buffer;}
  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.service";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(15, program15, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <div class=\"btn-group\">\n        <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n          ");
    stack1 = depth0;
    stack2 = "common.quickLinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <span class=\"caret\"></span>\n        </a>\n        <ul class=\"dropdown-menu\">\n          ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(16, program16, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n      </div>\n    ");
    return buffer;}
  function program16(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "disabled";
    stack3 = helpers['if'];
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(19, program19, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <li><a class=\"disabled-hdfs-quick-link\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n      ");
    stack1 = depth0;
    stack2 = "view.dashboardMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      <!-- Data Nodes -->\n      <tr>\n        <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.dataNodeComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.datanodes";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n        <td>\n          <span>\n            ");
    stack1 = depth0;
    stack2 = "view.dataNodesLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </span>  ");
    stack1 = depth0;
    stack2 = "services.service.summary.DataNodesLive";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- JournalNodes -->\n      ");
    stack1 = depth0;
    stack2 = "view.showJournalNodes";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <!-- NameNode Uptime -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nodes.uptime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.nodeUptime";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- NameNode Heap -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nodes.heap";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.nodeHeap";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Data Node Counts -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.datanodecounts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.dataNodesLive.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nodes.live";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " /\n          ");
    stack1 = depth0;
    stack2 = "view.dataNodesDead.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nodes.dead";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " /\n          ");
    stack1 = depth0;
    stack2 = "view.service.decommissionDataNodes.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nodes.decom";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- HDFS Capacity (Disk Usage)-->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.capacity.dfsUsed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dfsUsedDisk";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.capacity.nonDfsUsed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.nonDfsUsedDisk";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.capacity.remaining";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.remainingDisk";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Blocks Total -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "services.service.summary.blocksTotal";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dfsTotalBlocks";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Block Errors -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "services.service.summary.blockErrors";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.blockErrorsMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- Total Files And Directories -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.totalFilesAndDirs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dfsTotalFiles";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Upgrade Status -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "services.service.summary.pendingUpgradeStatus";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.upgradeStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- Safe Mode Status -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "services.service.summary.safeModeStatus";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.safeModeStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n\n    ");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/hive", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "component.host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>,\n    ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.titleMasters";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.clients.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.clients.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/mapreduce", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n<div class=\"clearfix like_pointer\" ");
    stack1 = depth0;
    stack2 = "toggleInfoView";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"name span2\">\n    ");
    stack1 = depth0;
    stack2 = "view.isCollapsed";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.summaryHeader";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n<div class=\"service-body\">\n  <div class=\"span4\">\n    <table class=\"table no-borders\">\n      <tbody>\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-down pull-left\"></i>\n    ");}

  function program4(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-right pull-left\"></i>\n    ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.liveComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "/");
    stack1 = depth0;
    stack2 = "view.totalComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      </tbody>\n    </table>\n  </div>\n<div class=\"dashboard-mini-chart span2\">\n  ");
    stack1 = depth0;
    stack2 = "App.ChartServiceMetricsMapReduce_TasksRunningWaiting";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  \n  ");
    stack1 = depth0;
    stack2 = "view.service.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n</div>\n");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.service";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(12, program12, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"btn-group\">\n    <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n      ");
    stack1 = depth0;
    stack2 = "common.quickLinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      <span class=\"caret\"></span>\n    </a>\n    <ul class=\"dropdown-menu\">\n      ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </ul>\n  </div>\n  ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n      ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n     ");
    stack1 = depth0;
    stack2 = "view.dashboardMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n      <!-- TaskTrackers -->\n      <tr>\n        <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.taskTrackerComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.taskTrackers";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n        <td>\n          <span>\n            ");
    stack1 = depth0;
    stack2 = "view.trackersLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </span> ");
    stack1 = depth0;
    stack2 = "services.service.summary.TrackersLive";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- Job Tracker Uptime -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.jobTrackerUptime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.jobTrackerUptime";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- TaskTracker Counts -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.taskTrackerCounts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "view.service.grayListTrackers.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.nodes.blacklist";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " /\n          ");
    stack1 = depth0;
    stack2 = "view.service.blackListTrackers.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.nodes.graylist";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " /\n          ");
    stack1 = depth0;
    stack2 = "view.service.trackersDecommissioned";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nodes.decom";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- JobTracker Heap -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.jobTrackerHeap";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.trackersHeapSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Total slots capacity -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.slotCapacity";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.slotsCapacitySummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Jobs -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.jobs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.jobsSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Map Slots -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.mapSlots";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.mapSlotsSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Reduce Slots -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.reduceSlots";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.reduceSlotsSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Tasks:Maps -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.tasks.maps";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.mapTasksSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Tasks:Reduces -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.tasks.reduces";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.reduceTasksSummary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n\n    ");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/mapreduce2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

    data.buffer.push("\n\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.titleInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/oozie", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "component.host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>,\n    ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.masters";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.clients.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.clients.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>,\n\n    <a href=\"");
    stack1 = depth0;
    stack2 = "view.webUi";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" target=\"_blank\">");
    stack1 = depth0;
    stack2 = "services.oozie.webUi";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/storm", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.liveComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "/");
    stack1 = depth0;
    stack2 = "view.totalComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.parentView.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n<tr>\n  <td>\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.superVisorComponents.0";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "dashboard.services.storm.supervisors";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </a>\n  </td>\n  <td>\n    <span>\n      ");
    stack1 = depth0;
    stack2 = "view.superVisorsLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </span> ");
    stack1 = depth0;
    stack2 = "services.service.summary.SupervisorsLIVE";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "services.service.summary.storm.freeslots";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "view.service.freeSlots";
    foundHelper = helpers.formatNull;
    stack3 = foundHelper || depth0.formatNull;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "view.service.totalSlots";
    foundHelper = helpers.formatNull;
    stack3 = foundHelper || depth0.formatNull;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " (");
    stack1 = depth0;
    stack2 = "view.freeSlotsPercentage";
    stack3 = {};
    stack4 = "0";
    stack3['empty'] = stack4;
    foundHelper = helpers.formatNull;
    stack4 = foundHelper || depth0.formatNull;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "% ");
    stack1 = depth0;
    stack2 = "common.free";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ")\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "services.service.summary.storm.tasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "view.service.totalTasks";
    foundHelper = helpers.formatNull;
    stack3 = foundHelper || depth0.formatNull;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "services.service.summary.storm.executors";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "view.service.totalExecutors";
    foundHelper = helpers.formatNull;
    stack3 = foundHelper || depth0.formatNull;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "services.service.summary.storm.topologies";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "view.service.topologies";
    foundHelper = helpers.formatNull;
    stack3 = foundHelper || depth0.formatNull;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "services.service.summary.storm.nimbus.uptime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
    stack1 = depth0;
    stack2 = "view.nimbusUptimeFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/yarn", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n<div class=\"clearfix like_pointer\" ");
    stack1 = depth0;
    stack2 = "toggleInfoView";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"name span2\">\n    ");
    stack1 = depth0;
    stack2 = "view.isCollapsed";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.summaryHeader";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n<div class=\"service-body\">\n  <div class=\"span4\">\n    <table class=\"table no-borders\">\n      <tbody>\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-down pull-left\"></i>\n    ");}

  function program4(depth0,data) {
    
    
    data.buffer.push("\n    <i class=\"icon-caret-right pull-left\"></i>\n    ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.liveComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "/");
    stack1 = depth0;
    stack2 = "view.totalComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      </tbody>\n    </table>\n    <div class=\"dashboard-mini-chart span2\">\n      ");
    stack1 = depth0;
    stack2 = "view.service.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n</div>\n</div>\n");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.service";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(12, program12, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"btn-group\">\n          <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n            ");
    stack1 = depth0;
    stack2 = "common.quickLinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"caret\"></span>\n          </a>\n          <ul class=\"dropdown-menu\">\n            ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </ul>\n        </div>\n      ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    ");
    stack1 = depth0;
    stack2 = "view.dashboardMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n      <!-- NodeManagers -->\n      <tr>\n        <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.nodeManagerComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.nodeManagers";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n        <td>\n          <span>\n            ");
    stack1 = depth0;
    stack2 = "view.nodeManagersLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </span>");
    stack1 = depth0;
    stack2 = "services.service.summary.nodeManagersLive";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- YARN Clients -->\n      <tr>\n        <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.yarnClientComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\">");
    stack1 = depth0;
    stack2 = "view.service.yarnClientNodes.length";
    stack3 = {};
    stack4 = "t:dashboard.services.yarn.client";
    stack3['singular'] = stack4;
    stack4 = "t:dashboard.services.yarn.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n        <td>\n          <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.service.yarnClientNodes.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>\n          ");
    stack1 = depth0;
    stack2 = "view.service.yarnClientNodes.length";
    stack3 = {};
    stack4 = "t:dashboard.services.yarn.client";
    stack3['singular'] = stack4;
    stack4 = "t:dashboard.services.yarn.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <!-- ResourceManager Uptime -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.resourceManager.uptime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.nodeUptime";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- ResourceManager Heap -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.resourceManager.nodes.heap";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.nodeHeap";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- NodeManagers status -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.nodeManagers.status";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.nodeManagersStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Containers -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.containers";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.containers";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Applications -->\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.apps";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.apps";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Memory -->\n      <tr>\n          <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.memory";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n          <td>");
    stack1 = depth0;
    stack2 = "view.memory";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <!-- Queues -->\n      <tr>\n          <td>");
    stack1 = depth0;
    stack2 = "dashboard.services.yarn.queues";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n          <td><span rel=\"queue-tooltip\" class=\"text-tooltip\" ");
    stack1 = {};
    stack2 = "view.service.queueFormatted";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.queues";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></td>\n      </tr>\n    ");
    stack1 = depth0;
    stack2 = "view.showOnlyRows";
    stack3 = helpers.unless;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/service/zookeeper", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "view.parentView";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a ");
    stack1 = depth0;
    stack2 = "view.service";
    stack3 = depth0;
    stack4 = "selectService";
    stack5 = {};
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "view.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
    stack1 = depth0;
    stack2 = "view.titleInfo.pre";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.titleInfo.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.titleInfo.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    stack1 = depth0;
    stack2 = "services.zookeeper.postfix";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/cluster_metrics", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"cluster-metrics\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\">");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div class=\"widget-content\" >\n        ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </li>\n  </ul>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/hbase_links", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n                <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.activeMaster.host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.masterServer";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.noMasterServer";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <a ");
    stack1 = {};
    stack2 = "view.hbaseMasterWebUrl";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.masterWebUI";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "services.service.summary.notAvailable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.model";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <div class=\"btn-group\">\n              <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n                ");
    stack1 = depth0;
    stack2 = "common.more";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                <span class=\"caret\"></span>\n              </a>\n              <ul class=\"dropdown-menu\">\n                ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </ul>\n            </div>\n          ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"links\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-remove-sign icon-large\"></i></a>\n      <div class=\"caption span10\"> ");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div class=\"widget-content\" >\n        <table>\n          <!--hbase master server-->\n          <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "view.activeMaster";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </td>\n          </tr>\n          <!--region servers-->\n          <tr>\n              <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.model.regionServers.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hbase.regionServers";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--hbase master Web UI-->\n          <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "view.activeMaster";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(7, program7, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </td>\n          </tr>\n        </table>\n      </div>\n\n      <div class=\"link-button\">\n        ");
    stack1 = depth0;
    stack2 = "view.model.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n\n    </li>\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/hdfs_links", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n        <table>\n          <!--Active NameNode-->\n          <tr>\n            ");
    stack1 = depth0;
    stack2 = "view.isActiveNNValid";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tr>\n          <!--Standby NameNodes-->\n          <tr>\n            ");
    stack1 = depth0;
    stack2 = "view.isStandbyNNValid";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(11, program11, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tr>\n          <!--Data Nodes-->\n          <tr>\n            <td>\n              <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.model.dataNodes.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.datanodes";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n            </td>\n          </tr>\n        </table>\n      ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n              <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.model.activeNameNode";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.widgets.HDFSLinks.activeNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n            ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <td><a class=\"disabled-hdfs-link\">");
    stack1 = depth0;
    stack2 = "dashboard.widgets.HDFSLinks.activeNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n            ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.isTwoStandbyNN";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(9, program9, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n                <!--Two Standby NameNodes-->\n                <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.twoStandbyComponent";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.widgets.HDFSLinks.standbyNameNodes";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n              ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n                <!--One Standby NameNode-->\n                <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.model.standbyNameNode";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.widgets.HDFSLinks.standbyNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n              ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <td><a class=\"disabled-hdfs-link\">");
    stack1 = depth0;
    stack2 = "dashboard.widgets.HDFSLinks.standbyNameNode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n            ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n        <table>\n          <!--NameNode-->\n          <tr>\n            <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.model.nameNode";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.nanmenode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--SecondaryNameNode-->\n          <tr>\n            <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.model.snameNode";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.snanmenode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--Data Nodes-->\n          <tr>\n            <td>\n              <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.model.dataNodes.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.hdfs.datanodes";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n            </td>\n          </tr>\n        </table>\n      ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.model";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(16, program16, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program16(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <div class=\"btn-group\">\n            <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n              ");
    stack1 = depth0;
    stack2 = "common.more";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"caret\"></span>\n            </a>\n            <ul class=\"dropdown-menu\">\n              ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </div>\n        ");
    return buffer;}
  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "disabled";
    stack3 = helpers['if'];
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(20, program20, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <li><a class=\"disabled-hdfs-quick-link\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    return buffer;}

  function program20(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"links\">\n  <ul>\n  <li class=\"thumbnail row\">\n    <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <i class=\"icon-remove-sign icon-large\"></i>\n    </a>\n    <div class=\"caption span10\"> ");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n\n    <div class=\"widget-content\" >\n      ");
    stack1 = depth0;
    stack2 = "view.isHAEnabled";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(13, program13, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    </div>\n\n    <div class=\"link-button\">\n      ");
    stack1 = depth0;
    stack2 = "view.model.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n  </li>\n    </ul>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/mapreduce_links", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.model";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <div class=\"btn-group\">\n              <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n                ");
    stack1 = depth0;
    stack2 = "common.more";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                <span class=\"caret\"></span>\n              </a>\n            <ul class=\"dropdown-menu\">\n              ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </ul>\n            </div>\n          ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"links\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\"> ");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div class=\"widget-content\" >\n        <table>\n          <!--jobTracker-->\n          <tr>\n            <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.model.jobTracker";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.service.summary.jobTracker";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--taskTrackers-->\n          <tr>\n            <td><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.component";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.model.taskTrackers.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "dashboard.services.mapreduce.taskTrackers";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--jobTracker Web UI-->\n          <tr>\n            <td><a ");
    stack1 = {};
    stack2 = "view.webUrl";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
    stack1 = depth0;
    stack2 = "services.service.summary.jobTrackerWebUI";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n        </table>\n      </div>\n\n      <div class=\"link-button\">\n        ");
    stack1 = depth0;
    stack2 = "view.model.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    </li>\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/mapreduce_slots", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <tr><td>");
    stack1 = depth0;
    stack2 = "line";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td></tr>");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"widget-content row-fluid\" id=\"map-reduce-slots-text\" >\n          <ul class=\"span12\">\n            <div class=\"span3\"> ");
    stack1 = depth0;
    stack2 = "dashboard.widgets.mapSlots";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"progress span5\" id=\"map-reduce-slots-bar1\">\n              <div class=\"bar bar-success\" ");
    stack1 = {};
    stack2 = "view.map_occupied";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n              <div class=\"bar bar-warning\" ");
    stack1 = {};
    stack2 = "view.map_reserved";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n              </div>\n            <div class=\"span3\" id=\"map-reduce-slots-num1\"> ");
    stack1 = depth0;
    stack2 = "view.map_display_text";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n          </ul>\n          <ul class=\"span12\">\n            <div class=\"span3\"> ");
    stack1 = depth0;
    stack2 = "dashboard.widgets.reduceSlots";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"progress span5\" id=\"map-reduce-slots-bar2\">\n              <div class=\"bar bar-success\" ");
    stack1 = {};
    stack2 = "view.reduce_occupied";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n              <div class=\"bar bar-warning\" ");
    stack1 = {};
    stack2 = "view.reduce_reserved";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n              </div>\n            <div class=\"span3\" id=\"map-reduce-slots-num2\"> ");
    stack1 = depth0;
    stack2 = "view.reduce_display_text";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n          </ul>\n        </div>\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"widget-content-isNA slots-content-isNA\" >");
    stack1 = depth0;
    stack2 = "services.service.summary.notAvailable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\" >\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-remove-sign icon-large\"></i></a>\n      <div class=\"caption span10 slots-caption\"> ");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div ");
    stack1 = {};
    stack2 = ":hidden-info-general view.hiddenInfoClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <table align=\"center\">\n          ");
    stack1 = depth0;
    stack2 = "view.hiddenInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "line";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </table>\n      </div>\n\n      ");
    stack1 = depth0;
    stack2 = "view.isViewExist";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </li>\n  </ul>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/pie_chart", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <tr><td>");
    stack1 = depth0;
    stack2 = "line";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td></tr>\n        ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <div ");
    stack1 = {};
    stack2 = ":widget-content view.hoverContentTopClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = {};
    stack4 = "view.model";
    stack3['modelBinding'] = stack4;
    stack4 = "view.thresh1";
    stack3['thresh1Binding'] = stack4;
    stack4 = "view.thresh2";
    stack3['thresh2Binding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <div class=\"widget-content-isNA\" >");
    stack1 = depth0;
    stack2 = "services.service.summary.notAvailable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n    ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\">");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "editWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-edit\"></i>\n      </a>\n\n    <div ");
    stack1 = {};
    stack2 = ":hidden-info-general view.hiddenInfoClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <table align=\"center\">\n        ");
    stack1 = depth0;
    stack2 = "view.hiddenInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "line";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </table>\n    </div>\n\n    ");
    stack1 = depth0;
    stack2 = "view.isPieExist";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </li>\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/simple_text", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <tr><td>");
    stack1 = depth0;
    stack2 = "line";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td></tr>\n          ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\" >\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\"> ");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "editWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-edit\"></i>\n      </a>\n\n      <div ");
    stack1 = {};
    stack2 = ":hidden-info-general view.hiddenInfoClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <table align=\"center\">\n          ");
    stack1 = depth0;
    stack2 = "view.hiddenInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "line";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </table>\n      </div>\n\n      <div ");
    stack1 = {};
    stack2 = ":widget-content view.hoverContentTopClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </li>\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/dashboard/widgets/uptime", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <tr>\n              <td>\n                ");
    stack1 = depth0;
    stack2 = "line";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"widget-content-isNA uptime-content-isNA\" >");
    stack1 = depth0;
    stack2 = "services.service.summary.notAvailable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"widget-content uptime-content\" >\n          ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\" >\n      <a class=\"corner-icon span1\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "deleteWidget";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span11\"> ");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </div>\n      <div ");
    stack1 = {};
    stack2 = ":hidden-info-general view.hiddenInfoClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <table align=\"center\">\n          ");
    stack1 = depth0;
    stack2 = "view.hiddenInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "line";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </table>\n      </div>\n\n      ");
    stack1 = depth0;
    stack2 = "view.isNA";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </li>\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "App.HostTableMenuView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n        <div class=\"btn-group display-inline-block\">\n          <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n            <span class=\"filters-label\">");
    stack1 = depth0;
    stack2 = "common.filter";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": </span>\n            <span>\n              <span ");
    stack1 = {};
    stack2 = "view.class";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n              ");
    stack1 = depth0;
    stack2 = "view.comboBoxLabel";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"caret\"></span>\n            </span>\n          </a>\n          <ul class=\"dropdown-menu\">\n            <li ");
    stack1 = {};
    stack2 = ":category-item view.filtersUsed::active";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a ");
    stack1 = depth0;
    stack2 = "clearFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "common.all";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " (");
    stack1 = depth0;
    stack2 = "content.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</a>\n            </li>\n            ");
    stack1 = depth0;
    stack2 = "view.categories";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n         </ul>\n       </div>\n      ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "category.isVisible";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n           ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                <li ");
    stack1 = {};
    stack2 = "aaa :category-item category.itemClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <a ");
    stack1 = depth0;
    stack2 = "category";
    stack3 = depth0;
    stack4 = "selectCategory";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                    ");
    stack1 = depth0;
    stack2 = "category.isHealthStatus";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "category.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  </a>\n               </li>\n             ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n                      <span ");
    stack1 = {};
    stack2 = ":health-status category.healthStatusValue category.class";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n                    ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n                      <span ");
    stack1 = {};
    stack2 = "category.class";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n                    ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <th class=\"first\"> </th>\n        <th> </th>\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.nameSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        <th> </th>\n        <th> </th>\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.ipSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.cpuSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.memorySort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.diskUsageSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.loadAvgSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        <th class=\"sort-view-6\">");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.HostView";
    stack3 = {};
    stack4 = "host";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(14, program14, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n\n          <td class=\"first\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "host.selected";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n\n          <td class=\"health\">\n            <span rel=\"HealthTooltip\" ");
    stack1 = {};
    stack2 = "host.healthClass host.healthIconClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "host.healthToolTip";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n          </td>\n\n          <td class=\"name\">\n            <span class=\"trim_hostname\">\n              <a title=\"");
    stack1 = depth0;
    stack2 = "host.publicHostName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.publicHostName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n            </span>\n            ");
    stack1 = depth0;
    stack2 = "host.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </td>\n          <td class=\"restart\">\n            ");
    stack1 = depth0;
    stack2 = "host.componentsWithStaleConfigsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </td>\n          <td class=\"passive-state\">\n              <span rel=\"ComponentsTooltip\" ");
    stack1 = {};
    stack2 = "view.componentsInPassiveStateMessage";
    stack1['data-original-title'] = stack2;
    stack2 = "host.componentsInPassiveStateCount:icon-medkit";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n          </td>\n          <td>");
    stack1 = depth0;
    stack2 = "host.ip";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n          <td>");
    stack1 = depth0;
    stack2 = "host.coresFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n          <td>");
    stack1 = depth0;
    stack2 = "host.memoryFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n\n          <td>\n            <div class=\"progress progress-info\" ");
    stack1 = {};
    stack2 = "host.diskInfoBar";
    stack1['title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " rel=\"UsageTooltip\">\n              <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.usageStyle";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n            </div>\n          </td>\n\n          <td>");
    stack1 = depth0;
    stack2 = "host.loadAvg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n          <td>\n            <a href=\"#\" class=\"host-components-expander\" ");
    stack1 = depth0;
    stack2 = "toggleComponents";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "> <span ");
    stack1 = {};
    stack2 = "view.isComponentsCollapsed:icon-caret-right:icon-caret-down";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n              ");
    stack1 = depth0;
    stack2 = "view.content.hostComponents.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.content.hostComponents.length";
    stack3 = {};
    stack4 = "t:common.component";
    stack3['singular'] = stack4;
    stack4 = "t:common.components";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n            <div id=\"host-");
    stack1 = depth0;
    stack2 = "host.hostName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" class=\"host-components\">\n              ");
    stack1 = depth0;
    stack2 = "view.labels";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </div>\n          </td>\n        ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n              <span class=\"label label-important alerts-count\" ");
    stack1 = depth0;
    stack2 = "host";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n              <span class=\"muted icon-refresh\" rel=\"ComponentsTooltip\" ");
    stack1 = {};
    stack2 = "view.restartRequiredComponentsMessage";
    stack1['title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n            ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <tr>\n        <td class=\"first\"> </td>\n        <td colspan=\"11\">\n          ");
    stack1 = depth0;
    stack2 = "hosts.table.noHosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n    ");
    return buffer;}

  function program21(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.statusFilter";
    stack3 = {};
    stack4 = "view.categories";
    stack3['categoriesBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(22, program22, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push(" - <a ");
    stack1 = depth0;
    stack2 = "clearSelection";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "hosts.filters.clearSelection";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
    return buffer;}
  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n          <a ");
    stack1 = depth0;
    stack2 = "view.parentView.selectedCategory";
    stack3 = depth0;
    stack4 = "selectCategory";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n            ");
    stack1 = depth0;
    stack2 = "view.parentView.selectedCategory.hostsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "view.parentView.selectedCategory.hostsCount";
    stack3 = {};
    stack4 = "t:hosts.filters.selectedHostInfo";
    stack3['singular'] = stack4;
    stack4 = "t:hosts.filters.selectedHostsInfo";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"hosts\">\n\n  <div class=\"box-header row\">\n    <div class=\"hosts-actions pull-left\">\n      ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    <div class=\"health-status-bar pull-left\">\n      ");
    stack1 = depth0;
    stack2 = "view.statusFilter";
    stack3 = {};
    stack4 = "view.categories";
    stack3['categoriesBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n  </div>\n\n  <table class=\"datatable table table-bordered table-striped\" id=\"hosts-table\">\n    <thead>\n      ");
    stack1 = depth0;
    stack2 = "view.sortView";
    stack3 = {};
    stack4 = "label-row";
    stack3['classNames'] = stack4;
    stack4 = "view.filteredContent";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <tr id=\"filter-row\">\n        <th class=\"first\"><div class=\"ember-view view-wrapper\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.selectAllHosts";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div></th>\n        <th> </th>\n        <th>");
    stack1 = depth0;
    stack2 = "view.nameFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th> </th>\n        <th> </th>\n        <th>");
    stack1 = depth0;
    stack2 = "view.ipFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "view.cpuFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "view.ramFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th> </th>\n        <th>");
    stack1 = depth0;
    stack2 = "view.loadAvgFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "view.componentsFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      </tr>\n    </thead>\n    <tbody>\n    ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(19, program19, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n  </table>\n\n  <div class=\"page-bar\">\n    <div class=\"filtered-hosts-info span4\">\n      <label>");
    stack1 = depth0;
    stack2 = "view.filteredContentInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " - <a ");
    stack1 = depth0;
    stack2 = "clearFilters";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "tableView.filters.clearAllFilters";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></label>\n    </div>\n    <div class=\"selected-hosts-info span4\">\n      ");
    stack1 = depth0;
    stack2 = "view.selectedCategory.hostsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(21, program21, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    <div class=\"items-on-page\">\n      <label>");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.rowsPerPageSelectView";
    stack3 = {};
    stack4 = "view.displayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
    stack1 = depth0;
    stack2 = "view.paginationInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      ");
    stack1 = depth0;
    stack2 = "view.paginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "view.paginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/add", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                <li ");
    stack1 = {};
    stack2 = "isStep4:active view.isStep4Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "addHost.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n\n      <!--<a class=\"btn back\" ");
    stack1 = depth0;
    stack2 = "backToHostsList";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\\u2190 Back to Hosts</a>-->\n\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "hosts.add.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step6.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    stack1 = depth0;
    stack2 = "App.supports.hostOverrides";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              <li ");
    stack1 = {};
    stack2 = "isStep5:active view.isStep5Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep5";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step8.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep6:active view.isStep6Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep6";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step9.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep7:active view.isStep7Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep7";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step10.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/addHost/step4", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <tr>\n              <td>");
    stack1 = depth0;
    stack2 = "service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n              <td>\n                ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "service.configGroupsNames";
    stack3['contentBinding'] = stack4;
    stack4 = "service.selectedConfigGroup";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
    return buffer;}

    data.buffer.push("\n<div>\n  <h2>");
    stack1 = depth0;
    stack2 = "addHost.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "addHost.step4.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"pre-scrollable\">\n      <table class=\"table\">\n        <thead>\n          <tr>\n            <th>");
    stack1 = depth0;
    stack2 = "common.service";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
    stack1 = depth0;
    stack2 = "common.conf.group";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n          </tr>\n        </thead>\n        <tbody>\n          ");
    stack1 = depth0;
    stack2 = "controller.content.serviceConfigGroups";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "service";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </tbody>\n      </table>\n  </div>\n   <div class=\"btn-area\">\n      <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/alerts_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.alerts.length";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <ul id='summary-alerts-list' class=\"alerts\">\n        ");
    stack1 = depth0;
    stack2 = "view.alerts";
    stack3 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.AlertItemView";
    stack3 = {};
    stack4 = "this";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "hosts.host.alert.noAlerts.message";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program7(depth0,data) {
    
    
    data.buffer.push("\n    <div class=\"spinner\"></div>\n");}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "controller.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(7, program7, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/bulk_operation_confirm_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"alert alert-warning\">\n    ");
    stack1 = depth0;
    stack2 = "view.warningInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br />\n      <strong>");
    stack1 = depth0;
    stack2 = "hostNamesSkipped";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n");
    return buffer;}

    data.buffer.push("\n\n<p>");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</p>\n\n<div class=\"task-top-wrap\">\n  <div class=\"task-detail-ico-wrap\">\n    <a href=\"#\" title=\"Click to Copy\" ");
    stack1 = depth0;
    stack2 = "textTrigger";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i class=\"icon-copy\"></i> ");
    stack1 = depth0;
    stack2 = "common.copy";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>\n<div class=\"task-detail-log-info\">\n  <div class=\"content-area\">\n    <div ");
    stack1 = {};
    stack2 = "view.textareaVisible::hidden :task-detail-log-clipboard-wrap";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <textarea class=\"task-detail-log-clipboard\"></textarea>\n    </div>\n    <div ");
    stack1 = {};
    stack2 = "view.textareaVisible:hidden :task-detail-log-maintext";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <pre>");
    stack1 = depth0;
    stack2 = "hostNames";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</pre>\n    </div>\n  </div>\n</div>\n");
    stack1 = depth0;
    stack2 = "hostNamesSkipped";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/bulk_operation_menu", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <li><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "addHost";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus icon-white\"></i> ");
    stack1 = depth0;
    stack2 = "hosts.host.add";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n      <li class=\"divider\"></li>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n              <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "view.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n              <ul class=\"dropdown-menu\">\n                ");
    stack1 = depth0;
    stack2 = "view.operationsInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "operation";
    stack7 = helpers.each;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </ul>\n          ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "operation.label.length";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      <li>\n                        ");
    stack1 = depth0;
    stack2 = "view.operationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "s";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      </li>\n                  ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                            <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                        ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.slaveItemView";
    stack3 = {};
    stack4 = "component";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "component.componentNameFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                <ul class=\"dropdown-menu\">\n                  ");
    stack1 = depth0;
    stack2 = "view.operationsInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "operation";
    stack7 = helpers.each;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </ul>\n            ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "operation.decommission";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(14, program14, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "view.advancedOperationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "s";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(12, program12, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "view.commonOperationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "s";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(15, program15, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n            <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "view.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n            <ul class=\"dropdown-menu\">\n              ");
    stack1 = depth0;
    stack2 = "view.operationsInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "operation";
    stack7 = helpers.each;
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          ");
    return buffer;}
  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "operation.label.length";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  <li>\n                    ");
    stack1 = depth0;
    stack2 = "view.operationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "f";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(20, program20, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </li>\n                ");
    return buffer;}
  function program20(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                    ");
    return buffer;}

  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.slaveItemView";
    stack3 = {};
    stack4 = "component";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(23, program23, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program23(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "component.componentNameFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                  <ul class=\"dropdown-menu\">\n                    ");
    stack1 = depth0;
    stack2 = "view.operationsInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "operation";
    stack7 = helpers.each;
    tmp1 = self.program(24, program24, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </ul>\n              ");
    return buffer;}
  function program24(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "operation.decommission";
    stack3 = helpers['if'];
    tmp1 = self.program(25, program25, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(28, program28, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program25(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                        ");
    stack1 = depth0;
    stack2 = "view.advancedOperationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "f";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(26, program26, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      ");
    return buffer;}
  function program26(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                        <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                        ");
    return buffer;}

  function program28(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                        ");
    stack1 = depth0;
    stack2 = "view.commonOperationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "f";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(29, program29, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      ");
    return buffer;}
  function program29(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                        <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                        ");
    return buffer;}

  function program31(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n              <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "view.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n              <ul class=\"dropdown-menu\">\n                ");
    stack1 = depth0;
    stack2 = "view.operationsInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "operation";
    stack7 = helpers.each;
    tmp1 = self.program(32, program32, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </ul>\n          ");
    return buffer;}
  function program32(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "operation.label.length";
    stack3 = helpers['if'];
    tmp1 = self.program(33, program33, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                ");
    return buffer;}
  function program33(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      <li>\n                        ");
    stack1 = depth0;
    stack2 = "view.operationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "a";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(34, program34, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      </li>\n                  ");
    return buffer;}
  function program34(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                            <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                        ");
    return buffer;}

  function program36(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.slaveItemView";
    stack3 = {};
    stack4 = "component";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(37, program37, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program37(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "component.componentNameFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                <ul class=\"dropdown-menu\">\n                  ");
    stack1 = depth0;
    stack2 = "view.operationsInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "operation";
    stack7 = helpers.each;
    tmp1 = self.program(38, program38, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </ul>\n            ");
    return buffer;}
  function program38(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "operation.decommission";
    stack3 = helpers['if'];
    tmp1 = self.program(39, program39, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(42, program42, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  ");
    return buffer;}
  function program39(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "view.advancedOperationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "a";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(40, program40, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program40(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
    return buffer;}

  function program42(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "view.commonOperationView";
    stack3 = {};
    stack4 = "operation.operationData";
    stack3['contentBinding'] = stack4;
    stack4 = "a";
    stack3['selection'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(43, program43, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program43(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "operation.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"dropdown bulk-menu\">\n  <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">");
    stack1 = depth0;
    stack2 = "common.actions";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <span class=\"caret\"></span></a>\n  <ul class=\"dropdown-menu\">\n    ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <li class=\"dropdown-submenu\">\n      <a ");
    stack1 = {};
    stack2 = "view.parentView.selectedCategory.hasHosts::disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " tabindex=\"-1\" href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "view.menuItems.s.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        (");
    stack1 = depth0;
    stack2 = "view.parentView.selectedCategory.hostsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</a>\n        <ul ");
    stack1 = {};
    stack2 = "view.parentView.selectedCategory.hasHosts::hidden :dropdown-menu";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "view.hostItemView";
    stack3 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.components";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n    </li>\n    <li class=\"dropdown-submenu\">\n      <a ");
    stack1 = {};
    stack2 = "view.parentView.hasFilteredItems::disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " tabindex=\"-1\" href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "view.menuItems.f.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        (");
    stack1 = depth0;
    stack2 = "view.parentView.filteredContent.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</a>\n        <ul ");
    stack1 = {};
    stack2 = "view.parentView.hasFilteredItems::hidden :dropdown-menu";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "view.hostItemView";
    stack3 = helpers.view;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.components";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(22, program22, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n    </li>\n    <li class=\"dropdown-submenu\">\n      <a tabindex=\"-1\" href=\"javascript:void(null);\">");
    stack1 = depth0;
    stack2 = "view.menuItems.a.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        (");
    stack1 = depth0;
    stack2 = "view.parentView.content.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</a>\n        <ul class=\"dropdown-menu\">\n          ");
    stack1 = depth0;
    stack2 = "view.hostItemView";
    stack3 = helpers.view;
    tmp1 = self.program(31, program31, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.components";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(36, program36, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n    </li>\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/component_filter", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <li>\n                <label class=\"checkbox\">\n                ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "component.checkedForHostFilter";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </label>\n              </li>\n            ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <li>\n                <label class=\"checkbox\">\n                ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "component.checkedForHostFilter";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </label>\n              </li>\n            ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <li>\n              <label class=\"checkbox\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "component.checkedForHostFilter";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </label>\n              </li>\n            ");
    return buffer;}

    data.buffer.push("\n\n<button class=\"btn single-btn-group filter-btn\" ");
    stack1 = depth0;
    stack2 = "clickFilterButton";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.filter";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-filter\"></i></button>\n    <ul class=\"dropdown-menu filter-components\">\n      <li id=\"title-bar\">");
    stack1 = depth0;
    stack2 = "apps.filters.filterComponents";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <a class=\"close\" ");
    stack1 = depth0;
    stack2 = "closeFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">X</a>\n      </li>\n      <li id=\"list-area\">\n        <ul>\n          <li>\n            <label class=\"checkbox\">\n            ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.masterComponentsChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "host.host.componentFilter.master";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n            </label>\n            <ul>\n            ");
    stack1 = depth0;
    stack2 = "masterComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </li>\n          <li>\n            <label class=\"checkbox\">\n            ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.slaveComponentsChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "host.host.componentFilter.slave";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n            </label>\n            <ul>\n            ");
    stack1 = depth0;
    stack2 = "slaveComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </li>\n          <li>\n            <label class=\"checkbox\">\n            ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.clientComponentsChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "host.host.componentFilter.client";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n            </label>\n            <ul>\n            ");
    stack1 = depth0;
    stack2 = "clientComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </li>\n        </ul>\n      </li>\n      <li id=\"button-bar\">\n        <button class=\"btn btn-inverse\" ");
    stack1 = depth0;
    stack2 = "applyFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n      </li>\n    </ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/configs", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div class=\"row-fluid\" style=\"margin-top: 7px;\">\n  <div id=\"services-menu\" class=\"well span2\">\n    ");
    stack1 = depth0;
    stack2 = "App.MainHostServiceMenuView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"span10\">\n    ");
    stack1 = depth0;
    stack2 = "service_config_outlet";
    foundHelper = helpers.outlet;
    stack3 = foundHelper || depth0.outlet;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "outlet", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/configs_service", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"clearfix\"></div>\n    ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigView";
    stack3 = {};
    stack4 = "view.isConfigsEditable";
    stack3['canEditBinding'] = stack4;
    stack4 = "controller.filter";
    stack3['filterBinding'] = stack4;
    stack4 = "controller.filterColumns";
    stack3['columnsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    
    data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

    data.buffer.push("\n\n<div id=\"serviceConfig\">\n  ");
    stack1 = depth0;
    stack2 = "dataIsLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/decommission", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n\n<li rel='decommissionTooltip' ");
    stack1 = {};
    stack2 = "view.parentView.decommissionTooltipMessage";
    stack1['data-original-title'] = stack2;
    stack2 = "view.parentView.noActionAvailable";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = {};
    stack2 = "view.parentView.isComponentDecommissionDisable:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "view.text";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </a>\n</li>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <span class=\"label label-important alerts-count\" ");
    stack1 = depth0;
    stack2 = "content";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "App.router.mainHostController";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.content.criticalAlertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <span class=\"label label-success alerts-count\" ");
    stack1 = depth0;
    stack2 = "content";
    stack3 = depth0;
    stack4 = "showAlertsPopup";
    stack5 = {};
    stack6 = "App.router.mainHostController";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "hosts.host.alert.noAlerts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n    ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <span class=\"host-maintenance-notice pull-right\"><span class=\"icon-medkit\"></span> ");
    stack1 = depth0;
    stack2 = "hosts.host.passive.mode";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n    ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.supports.deleteHost";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	  ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n				<div class=\"service-button\">\n				    <div class=\"btn-group display-inline-block\">\n				      <a href=\"javascript:void(null)\" ");
    stack1 = {};
    stack2 = ":btn :dropdown-toggle";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " data-toggle=\"dropdown\">\n                ");
    stack1 = depth0;
    stack2 = "hosts.host.details.hostActions";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n				        <span class=\"caret\"></span>\n				      </a>\n				      <ul class=\"dropdown-menu pull-right\">\n				        <!-- dropdown menu links -->\n				        ");
    stack1 = depth0;
    stack2 = "view.maintenance";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "option";
    stack7 = helpers.each;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n				      </ul>\n				    </div>\n				</div>\n			");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n				        <li ");
    stack1 = {};
    stack2 = "controller.isStopDisabled:disabled option.liClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n				          <a ");
    stack1 = depth0;
    stack2 = "option";
    stack3 = depth0;
    stack4 = "doAction";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i ");
    stack1 = {};
    stack2 = "option.cssClass option.liClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i> ");
    stack1 = depth0;
    stack2 = "option.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n				        </li>\n				        ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"host-details\">\n  <div class=\"status-info\">\n    <span rel=\"HealthTooltip\" ");
    stack1 = {};
    stack2 = "view.content.healthClass view.content.healthIconClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.content.healthToolTip";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span><span class='host-title'>");
    stack1 = depth0;
    stack2 = "view.content.publicHostName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n    ");
    stack1 = depth0;
    stack2 = "view.content.criticalAlertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "view.isActive";
    stack3 = helpers.unless;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>  \n  <div><a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-left\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></div>\n  <div class=\"content\">\n    ");
    stack1 = depth0;
    stack2 = "App.MainHostMenuView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details/addComponentPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "addComponentMsg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br /><br />\n<div class='alert'>");
    stack1 = depth0;
    stack2 = "restartNagiosMsg";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details/deleteComponentPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"alert-error row-fluid\">\n        <div class='tinyspan tinyoffset'>");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "isChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class='span10'>");
    stack1 = depth0;
    stack2 = "lastComponentError";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "deleteComponentMsg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br /><br />\n");
    stack1 = depth0;
    stack2 = "lastComponent";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div class='alert'>");
    stack1 = depth0;
    stack2 = "restartNagiosMsg";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details/doDeleteHostPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "hosts.delete.popup.unknownComponents";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  <br /><br />\n  <div class='row-fluid'>\n    <div class='tinyoffset span10'>\n      <i>");
    stack1 = depth0;
    stack2 = "unknownComponents";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</i>\n    </div>\n  </div>\n  <br />\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"alert-error row-fluid\">\n        <div class='tinyspan tinyoffset'>");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "isChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class='span10'>");
    stack1 = depth0;
    stack2 = "lastComponentError";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "hosts.delete.popup.body.msg.unknownComponents";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "unknownComponents";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<p><i class=\"icon-warning-sign\"></i> ");
    stack1 = depth0;
    stack2 = "deletePopupBody";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</p>\n");
    stack1 = depth0;
    stack2 = "lastComponent";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<br />\n<div class='alert'><strong>");
    stack1 = depth0;
    stack2 = "common.important";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</strong>\n  ");
    stack1 = depth0;
    stack2 = "unknownComponents";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "hosts.delete.popup.body.msg1";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class='alert'><strong>");
    stack1 = depth0;
    stack2 = "common.important";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</strong> ");
    stack1 = depth0;
    stack2 = "hosts.delete.popup.body.msg2";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n<div class='alert'><strong>");
    stack1 = depth0;
    stack2 = "common.important";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</strong> ");
    stack1 = depth0;
    stack2 = "hosts.delete.popup.body.msg3";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details/host_component", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "upgradeComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >\n        <i title=\"Component upgrade failed\" class=\"components-health icon-arrow-up\"></i>\n      </a>\n    ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n    <span rel='componentHealthTooltip' ");
    stack1 = {};
    stack2 = "view.statusClass view.statusIconClass :components-health";
    stack1['class'] = stack2;
    stack2 = "view.componentTextStatus";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>&nbsp;\n  ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "component.displayNameAdvanced";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program10(depth0,data) {
    
    
    data.buffer.push("\n    <span class=\"text-warning icon-refresh\"></span>\n  ");}

  function program12(depth0,data) {
    
    
    data.buffer.push("\n    <span class=\"icon-medkit\"></span>\n  ");}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"btn-group pull-right\">\n      <a ");
    stack1 = {};
    stack2 = "view.disabled :btn :dropdown-toggle";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " data-toggle=\"dropdown\">\n        ");
    stack1 = depth0;
    stack2 = "view.componentTextStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"caret pull-right\"></span>\n      </a>\n      <ul class=\"dropdown-menu\">\n        ");
    stack1 = depth0;
    stack2 = "view.isComponentDecommissionAvailable";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.isComponentRecommissionAvailable";
    stack3 = helpers['if'];
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.isRestartableComponent";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.isInstalling";
    stack3 = helpers.unless;
    tmp1 = self.program(21, program21, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.isInit";
    stack3 = helpers['if'];
    tmp1 = self.program(33, program33, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.isDeletableComponent";
    stack3 = helpers['if'];
    tmp1 = self.program(35, program35, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n    </div>\n  ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.decommissionView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.decommissionView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n          <li ");
    stack1 = {};
    stack2 = "view.isRestartComponentDisabled:hidden";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "restartComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              ");
    stack1 = depth0;
    stack2 = "common.restart";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
    return buffer;}

  function program21(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.isStart";
    stack3 = helpers['if'];
    tmp1 = self.program(22, program22, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.isStart";
    stack3 = helpers.unless;
    tmp1 = self.program(24, program24, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.isUpgradeFailed";
    stack3 = helpers['if'];
    tmp1 = self.program(27, program27, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.isInstallFailed";
    stack3 = helpers['if'];
    tmp1 = self.program(29, program29, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.isReassignable";
    stack3 = helpers['if'];
    tmp1 = self.program(31, program31, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = "view.isComponentDecommissioning:hidden view.noActionAvailable";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "stopComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "common.stop";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program24(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.isInit";
    stack3 = helpers.unless;
    tmp1 = self.program(25, program25, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program25(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n              <li ");
    stack1 = {};
    stack2 = "view.isUpgradeFailed:hidden view.isInstallFailed:hidden view.isDecommissioning:hidden view.noActionAvailable";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "startComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  ");
    stack1 = depth0;
    stack2 = "common.start";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n            ");
    return buffer;}

  function program27(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = "view.noActionAvailable";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "upgradeComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "common.reUpgrade";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program29(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = "view.noActionAvailable";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "installComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "common.reinstall";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program31(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = "view.noActionAvailable";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "moveComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "common.move";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program33(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n          <li ");
    stack1 = {};
    stack2 = "view.";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "installComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              ");
    stack1 = depth0;
    stack2 = "common.reinstall";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
    return buffer;}

  function program35(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = "view.isDeleteComponentDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "deleteComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  ");
    stack1 = depth0;
    stack2 = "common.delete";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"span7\">\n  ");
    stack1 = depth0;
    stack2 = "view.isUpgradeFailed";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "component.displayNameAdvanced";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  &nbsp;/&nbsp;\n  <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "component.service";
    stack3 = depth0;
    stack4 = "routeToService";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "component.service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n  &nbsp;\n  ");
    stack1 = depth0;
    stack2 = "component.staleConfigs";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "component.service.isInPassive";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n<div class=\"span5 pull-right\">\n  ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details/installComponentPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "installComponentMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<br /><br />\n<div class='alert'>");
    stack1 = depth0;
    stack2 = "restartNagiosMsg";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/details/raiseDeleteComponentErrorPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <div class=\"warning-details\">");
    stack1 = depth0;
    stack2 = "componentsBodyEnd";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</div>\n");
    return buffer;}

    data.buffer.push("\n\n<div class=\"dialog-delete-component\">\n  <div class=\"warning\">\n  <i class=\"icon-warning-sign\"></i> <strong>");
    stack1 = depth0;
    stack2 = "componentsBody";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n  <div class=\"row-fluid\">\n    <div class=\"tinyoffset span10 warning-list\">\n    ");
    stack1 = depth0;
    stack2 = "componentsStr";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n");
    stack1 = depth0;
    stack2 = "showBodyEnd";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/metrics", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"host-metrics\">\n    <table class=\"graphs\">\n        <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "App.ChartHostMetricsCPU";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "App.ChartHostMetricsDisk";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n        <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "App.ChartHostMetricsLoad";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "App.ChartHostMetricsMemory";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n        <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "App.ChartHostMetricsNetwork";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "App.ChartHostMetricsProcesses";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n    </table>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/host/summary", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n              <div class=\"btn-group pull-right\">\n                <button id=\"add_component\" data-toggle=\"dropdown\" ");
    stack1 = {};
    stack2 = "view.addComponentDisabled:disabled :btn :dropdown-toggle";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <span class=\"icon-plus\"></span>&nbsp;");
    stack1 = depth0;
    stack2 = "add";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                </button>\n                <ul class=\"dropdown-menu\">\n                  ");
    stack1 = depth0;
    stack2 = "view.addableComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </ul>\n              </div>\n            ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                    <li>\n                      <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "component";
    stack3 = depth0;
    stack4 = "addComponent";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                  ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n\n          ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n          ");
    stack1 = depth0;
    stack2 = "view.sortedComponents";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.content.componentsWithStaleConfigsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <div class=\"alert alert-warning clearfix\">\n                <i class=\"icon-refresh\"></i> ");
    stack1 = depth0;
    stack2 = "view.needToRestartMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  <button ");
    stack1 = {};
    stack2 = ":btn :restart-components :pull-right :btn-warning";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "restartAllStaleConfigComponents";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    ");
    stack1 = depth0;
    stack2 = "hosts.host.details.needToRestart.button";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                  </button>\n              </div>\n            ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <div class=\"row-fluid\">\n              ");
    stack1 = depth0;
    stack2 = "component.view";
    stack3 = {};
    stack4 = "component";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                  <div class=\"span3 align-right\">");
    stack1 = depth0;
    stack2 = "common.clients";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "&nbsp;/</div>\n                  <div class=\"span7\">\n                    ");
    stack1 = depth0;
    stack2 = "view.clients";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </div>\n              ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "component.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      ");
    stack1 = depth0;
    stack2 = "component.staleConfigs";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "component.service.isInPassive";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "component.isLast";
    stack3 = helpers.unless;
    tmp1 = self.program(16, program16, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    return buffer;}
  function program12(depth0,data) {
    
    
    data.buffer.push("\n                        <span class=\"text-warning icon-refresh\"></span>\n                      ");}

  function program14(depth0,data) {
    
    
    data.buffer.push("\n                        <span class=\"icon-medkit\"></span>\n                      ");}

  function program16(depth0,data) {
    
    
    data.buffer.push(",");}

  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                  <div class=\"btn-group pull-right\">\n                    <button id=\"add_component\" data-toggle=\"dropdown\" ");
    stack1 = {};
    stack2 = ":btn :dropdown-toggle controller.content.isNotHeartBeating:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                      ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      <span class=\"caret pull-right\"></span>\n                    </button>\n                    <ul class=\"dropdown-menu\">\n                      <li>\n                        <a href=\"javascript:void(null)\" ");
    stack1 = {};
    stack2 = "view.areClientWithStaleConfigs::disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "view.clients";
    stack3 = depth0;
    stack4 = "refreshConfigs";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                          ");
    stack1 = depth0;
    stack2 = "hosts.host.details.refreshConfigs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                        </a>\n                      </li>\n                    </ul>\n                  </div>\n                ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"row-fluid host-block\">\n  <div class=\"span12\">\n    <div class=\"span6\">\n      ");
    data.buffer.push("\n      <div class=\"box\">\n        <div class=\"box-header\">\n          <div class=\"pull-left\">\n            <h4>");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n          </div>\n          <div class=\"pull-right span5 row-fluid\" style=\"padding:0 10px;\">\n            ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </div>\n        </div>\n        <div class=\"host-components\">\n        ");
    stack1 = depth0;
    stack2 = "view.sortedComponents.length";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    data.buffer.push("\n          <div class=\"clients row-fluid\">\n            <div class=\"span7 row\">\n              ");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </div>\n            <div class=\"span5 row\">\n              ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </div>\n          </div>\n        </div>\n      </div>\n      <div class=\"host-configuration\">\n            <div class=\"box\">\n                <div class=\"box-header\">\n                    <h4>");
    stack1 = depth0;
    stack2 = "hosts.host.summary.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n                </div>\n                <div class=\"host-summary-content\">\n                    <dl class=\"dl-horizontal\">\n                        <dt>");
    stack1 = depth0;
    stack2 = "hosts.host.summary.hostname";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.publicHostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "common.ipAddress";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.ip";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "common.os";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.osType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "&nbsp;(");
    stack1 = depth0;
    stack2 = "view.content.osArch";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "common.cores";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.coresFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "common.disk";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.diskInfoBar";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "common.memory";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.memoryFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "common.loadAvg";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>&nbsp;");
    stack1 = depth0;
    stack2 = "view.content.loadAvg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                        <dt>");
    stack1 = depth0;
    stack2 = "hosts.host.summary.agentHeartbeat";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</dt><dd>");
    stack1 = depth0;
    stack2 = "view.timeSinceHeartBeat";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</dd>\n                    </dl>\n                </div>\n            </div>\n        </div>\n    </div>\n  ");
    data.buffer.push("\n   <div class=\"span6\">\n      <div class=\"box\">\n        <div class=\"box-header\">\n          <h4>");
    stack1 = depth0;
    stack2 = "hosts.host.summary.hostMetrics";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n          <div class=\"btn-group\">\n            <a class=\"btn\" rel=\"tooltip\" title=\"Go to Ganglia\" ");
    stack1 = depth0;
    stack2 = "showGangliaCharts";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-link\"></i></a>\n          </div>\n        </div>\n        <div>\n         ");
    stack1 = depth0;
    stack2 = "App.MainHostMetricsView";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/jobs", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <div class=\"new-jobs-link\">\n            <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "updateJobsByClick";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "jobs.new_jobs.info";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n          </div>\n        ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <th></th>\n              ");
    stack1 = depth0;
    stack2 = "view.parentView.idSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.parentView.userSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.parentView.startTimeSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.parentView.endTimeSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.parentView.durationSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <tr>\n                ");
    stack1 = depth0;
    stack2 = "controller.loaded";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </tr>\n          ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <td class=\"no-data\" ");
    stack1 = {};
    stack2 = "controller.columnsName.content.length";
    stack1['colspan'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "jobs.nothingToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n              ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    <td class=\"no-data\" ");
    stack1 = {};
    stack2 = "controller.columnsName.content.length";
    stack1['colspan'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "jobs.loadingTasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "controller.sortedContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "job";
    stack7 = helpers.each;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <tr>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "job.failed";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </td>\n                <td>\n                  <div class=\"id\">\n                    ");
    stack1 = depth0;
    stack2 = "view.jobNameView";
    stack3 = {};
    stack4 = "job";
    stack3['jobBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  </div>\n                </td>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "job.user";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "job.startTimeDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "job.endTimeDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n                <td>\n                  ");
    stack1 = depth0;
    stack2 = "job.durationDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </td>\n              </tr>\n            ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n                    <i class=\"icon-remove-sign job-link\" ");
    stack1 = {};
    stack2 = "view.jobFailMessage";
    stack1['title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                  ");
    return buffer;}

    data.buffer.push("\n\n\n  <div id=\"jobs\">\n      <div class=\"jobs_head\">\n        <div>");
    stack1 = depth0;
    stack2 = "menu.item.jobs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
    stack1 = depth0;
    stack2 = "controller.hasNewJobs";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        <div class=\"jobs-type\">\n         ");
    stack1 = depth0;
    stack2 = "jobs.type";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " : <span class=\"label label-info\">");
    stack1 = depth0;
    stack2 = "jobs.type.hive";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n      </div>\n      <table id=\"jobs-table\" class=\"table table-bordered table-striped\">\n          <thead>\n          ");
    stack1 = depth0;
    stack2 = "view.sortView";
    stack3 = {};
    stack4 = "label-row";
    stack3['classNames'] = stack4;
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n          <tr id=\"filter-row\" class=\"first\">\n              <th></th>\n              <th>");
    stack1 = depth0;
    stack2 = "view.jobsIdFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th>");
    stack1 = depth0;
    stack2 = "view.userFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th class=\"start-time\">");
    stack1 = depth0;
    stack2 = "view.startTimeFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th></th>\n              <th></th>\n          </tr>\n          </thead>\n          <tbody>\n          ");
    stack1 = depth0;
    stack2 = "view.noDataToShow";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(10, program10, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tbody>\n      </table>\n\n      <div class=\"page-bar\">\n          <div id=\"filtered-jobs\">\n            ");
    stack1 = depth0;
    stack2 = "view.filteredJobs";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " - <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "clearFilters";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "jobs.filtered.clear";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n          </div>\n          <div class=\"items-on-page\">\n              <label>");
    stack1 = depth0;
    stack2 = "jobs.show.up.to";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.rowsPerPageSelectView";
    stack3 = {};
    stack4 = "view.displayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n          </div>\n        <div class=\"paging_two_button\">\n          ");
    stack1 = depth0;
    stack2 = "view.jobsPaginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "view.jobsPaginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n  </div>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/jobs/custom_dates_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"jobs-custom-dates\">\n  <div>\n    <label>");
    stack1 = depth0;
    stack2 = "jobs.customDateFilter.startTime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n    ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "customDateFormFields.startDate";
    stack3['valueBinding'] = stack4;
    stack4 = "input-small datepicker";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.hourOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "customDateFormFields.hoursForStart";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.minuteOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "customDateFormFields.minutesForStart";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.middayPeriodOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "customDateFormFields.middayPeriodForStart";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.startDate";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n  <div>\n\n  </div>\n  <div>\n    <label>");
    stack1 = depth0;
    stack2 = "jobs.customDateFilter.endTime";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n    ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "customDateFormFields.endDate";
    stack3['valueBinding'] = stack4;
    stack4 = "input-small datepicker";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.hourOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "customDateFormFields.hoursForEnd";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.minuteOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "customDateFormFields.minutesForEnd";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.middayPeriodOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "customDateFormFields.middayPeriodForEnd";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.endDate";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/jobs/hive_job_details", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <!-- Top Bar -->\n    <div class=\"top-bar\">\n      <a ");
    stack1 = depth0;
    stack2 = "routeToJobs";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "menu.item.jobs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a> > ");
    stack1 = depth0;
    stack2 = "view.content.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      <a ");
    stack1 = depth0;
    stack2 = "toggleShowQuery";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\" id=\"toggle-query\">");
    stack1 = depth0;
    stack2 = "view.toggleShowQueryText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n      <div class=\"pull-right\">Job Type: <span class=\"label label-info\">");
    stack1 = depth0;
    stack2 = "view.content.jobType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n      <div class=\"alert alert-info query-info\">\n        <table id=\"job-more-details-table\">\n          <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "jobs.hive.query";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "view.content.queryText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n          </tr>\n          <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "jobs.hive.yarnApplication";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "view.yarnApplicationIdLink";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </td>\n          </tr>\n          <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "jobs.hive.stages";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              <ol>\n                ");
    stack1 = depth0;
    stack2 = "view.content.stages";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "stage";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </ol>\n            </td>\n          </tr>\n        </table>\n      </div>\n    </div>\n\n    <!-- Sections -->\n    <div class=\"row-fluid\">\n      <div class=\"span12 sections\">\n        <!-- Section LHS -->\n        <div id=\"tez-dag-lhs\" ");
    stack1 = {};
    stack2 = "view.isGraphMaximized:span12:span6 :sections-lhs";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <div id=\"tez-dag-section\" class=\"box\">\n            <div id=\"tez-dag-section-top-bar\"> &nbsp;\n              ");
    stack1 = depth0;
    stack2 = "jobs.hive.tez.dag.summary.metric";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              <div class=\"btn-group  display-inline-block\">\n                <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n                  ");
    stack1 = depth0;
    stack2 = "view.summaryMetricTypeDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  <span class=\"caret\"></span>\n                </a>\n                <ul class=\"dropdown-menu\">\n                  <!-- dropdown menu links -->\n                  ");
    stack1 = depth0;
    stack2 = "view.summaryMetricTypesDisplay";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "type";
    stack7 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </ul>\n              </div>\n            <div id=\"tez-dag-section-top-bar-actions\" class=\"pull-right\">\n              <div class=\"btn-group\">\n                <a id=\"tez-dag-zoom-in-button\" ");
    stack1 = {};
    stack2 = "view.canGraphZoomIn::disabled :btn";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "doGraphZoomIn";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <i class=\"icon-zoom-in\"></i>\n                </a>\n                <a id=\"tez-dag-zoom-out-button\" ");
    stack1 = {};
    stack2 = "view.canGraphZoomOut::disabled :btn";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "doGraphZoomOut";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <i class=\"icon-zoom-out\"></i>\n                </a>\n                ");
    stack1 = depth0;
    stack2 = "view.isGraphMaximized";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(12, program12, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </div>\n            </div>\n            </div>\n            <div id=\"tez-dag-section-body\">\n              ");
    stack1 = depth0;
    stack2 = "App.MainHiveJobDetailsTezDagView";
    stack3 = {};
    stack4 = "controller";
    stack3['controllerBinding'] = stack4;
    stack4 = "view.selectedVertex";
    stack3['selectedVertexBinding'] = stack4;
    stack4 = "view.summaryMetricType";
    stack3['summaryMetricTypeBinding'] = stack4;
    stack4 = "view.zoomScale";
    stack3['zoomScaleBinding'] = stack4;
    stack4 = "view.zoomScaleFrom";
    stack3['zoomScaleFromBinding'] = stack4;
    stack4 = "view.zoomScaleTo";
    stack3['zoomScaleToBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          </div>\n        </div>\n\n        <!-- Section RHS -->\n        <div id=\"tez-vertices-rhs\" ");
    stack1 = {};
    stack2 = "view.isGraphMaximized:hidden:span6 :sections-rhs";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n\n          <!-- Section RHS Vertices -->\n          <div id=\"tez-vertices-table-section\">\n            <div id=\"tez-vertices-table-container\" class=\"section\">\n              <table class=\"table table-hover table-bordered table-striped\">\n                ");
    stack1 = depth0;
    stack2 = "App.MainHiveJobDetailsVerticesTableView";
    stack3 = {};
    stack4 = "view.sortedVertices";
    stack3['contentBinding'] = stack4;
    stack4 = "view.doSelectVertex";
    stack3['doSelectVertexBinding'] = stack4;
    stack4 = "view.selectedVertex";
    stack3['selectedVertexBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(14, program14, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </table>\n            </div>\n          </div>\n\n          <!-- Section RHS Vertex -->\n          ");
    stack1 = depth0;
    stack2 = "view.selectedVertex";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n      </div>\n    </div>\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <a ");
    stack1 = {};
    stack2 = "view.yarnApplicationIdLink";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">\n                  ");
    stack1 = depth0;
    stack2 = "view.content.tezDag.yarnApplicationId";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "view.content.tezDag.yarnApplicationId";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <li>");
    stack1 = depth0;
    stack2 = "stage.id";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));
    stack1 = depth0;
    stack2 = "stage.description";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ".</li>\n                ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                    <li>\n                      <a title=\"");
    stack1 = depth0;
    stack2 = "type";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "type";
    stack3 = depth0;
    stack4 = "doSelectSummaryMetricType";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "type";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                    </li>\n                  ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  <a id=\"tez-dag-resize-small-button\" class=\"btn\" ");
    stack1 = depth0;
    stack2 = "doGraphMinimize";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    <i class=\"icon-resize-small\"></i>\n                  </a>\n                ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  <a id=\"tez-dag-resize-full-button\" class=\"btn\" ");
    stack1 = depth0;
    stack2 = "doGraphMaximize";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    <i class=\"icon-resize-full\"></i>\n                  </a>\n                ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                  <thead>\n                    ");
    stack1 = depth0;
    stack2 = "view.sortView";
    stack3 = {};
    stack4 = "view.content";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(15, program15, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </thead>\n                  <tbody>\n                    ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "vertex";
    stack7 = helpers.each;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </tbody>\n                ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "view.parentView.nameSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      ");
    stack1 = depth0;
    stack2 = "view.parentView.tasksSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      ");
    stack1 = depth0;
    stack2 = "view.parentView.inputSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      ");
    stack1 = depth0;
    stack2 = "view.parentView.outputSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      ");
    stack1 = depth0;
    stack2 = "view.parentView.durationSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                      <tr ");
    stack1 = {};
    stack2 = "vertex.isSelected:info";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <td>\n                          <a title=\"");
    stack1 = depth0;
    stack2 = "vertex.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" href=\"#\" ");
    stack1 = depth0;
    stack2 = "vertex";
    stack3 = depth0;
    stack4 = "doSelectVertex";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "vertex.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                        </td>\n                        <td>");
    stack1 = depth0;
    stack2 = "vertex.tasksNumber";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        <td>");
    stack1 = depth0;
    stack2 = "vertex.totalReadBytesDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        <td>");
    stack1 = depth0;
    stack2 = "vertex.totalWriteBytesDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        <td>");
    stack1 = depth0;
    stack2 = "vertex.durationDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      </tr>\n                    ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <div id=\"section tez-vertex-details-section\">\n              <div class=\"box\">\n                <div class=\"box-header\">\n                  <h4>");
    stack1 = depth0;
    stack2 = "view.selectedVertex.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</h4>\n                </div>\n                <div id=\"tez-vertex-details-section-body\">\n                  <table class=\"table no-borders table-condensed\">\n                    <tr>\n                      <td>");
    stack1 = depth0;
    stack2 = "common.status";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.status";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td></td>\n                    </tr>\n                    <tr>\n                      <td>");
    stack1 = depth0;
    stack2 = "common.time.start";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.started";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td></td>\n                    </tr>\n                    <tr>\n                      <td>");
    stack1 = depth0;
    stack2 = "common.time.end";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.ended";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td></td>\n                    </tr>\n                    <tr>\n                      <td>");
    stack1 = depth0;
    stack2 = "jobs.hive.tez.tasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertex.tasksCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td></td>\n                    </tr>\n                    <tr>\n                      <td>");
    stack1 = depth0;
    stack2 = "jobs.hive.tez.hdfs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.hdfs.read.ops";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.hdfs.read.bytes";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.hdfs.write.ops";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.hdfs.write.bytes";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                    </tr>\n                    <tr>\n                      <td>");
    stack1 = depth0;
    stack2 = "jobs.hive.tez.localFiles";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.file.read.ops";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.file.read.bytes";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.file.write.ops";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " / ");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.file.write.bytes";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                    </tr>\n                    <tr>\n                        <td>");
    stack1 = depth0;
    stack2 = "jobs.hive.tez.spilledRecords";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertex.spilledRecords";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                    </tr>\n                    ");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.records.read";
    stack3 = helpers['if'];
    tmp1 = self.program(20, program20, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </table>\n                </div>\n              </div>\n            </div>\n          ");
    return buffer;}
  function program20(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      <tr>\n                        <td>");
    stack1 = depth0;
    stack2 = "jobs.hive.tez.records";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.records.read";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        <td>");
    stack1 = depth0;
    stack2 = "view.selectedVertexIODisplay.records.write";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                      </tr>\n                    ");
    return buffer;}

  function program22(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"alert alert-info\">\n      <h4>");
    stack1 = depth0;
    stack2 = "app.loadingPlaceholder";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n    </div>\n  ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"hive-job-details\">\n\n  ");
    stack1 = depth0;
    stack2 = "controller.loaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(22, program22, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/jobs/hive_job_details_tez_dag", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', foundHelper, self=this;


    data.buffer.push("\n\n<div id=\"tez-dag-section-body-dag\">\n  <svg id=\"tez-dag-svg\" stroke=\"black\" strokeWidth=\"2\">\n    <defs>\n      <marker id=\"arrow\" viewBox=\"0 -5 10 10\" refX=\"4\" markerWidth=\"3\" markerHeight=\"3\" orient=\"auto\">\n        <path d=\"M0,-5L10,0L0,5\"></path>\n      </marker>\n      <filter id=\"shadow\" width=\"200%\" height=\"200%\">\n        <feGaussianBlur in=\"SourceAlpha\" stdDeviation=\"3\"/>\n        <feOffset dx=\"2\" dy=\"2\" result=\"offsetblur\"/>\n        <feMerge> \n          <feMergeNode/>\n          <feMergeNode in=\"SourceGraphic\"/>\n        </feMerge>\n      </filter>\n      <clipPath id=\"operatorClipPath\">\n        <rect x=\"0\" y=\"0\" width=\"50\" height=\"16\">\n        </rect>\n      </clipPath>\n      <font-face xmlns=\"http://www.w3.org/2000/svg\" font-family=\"FontAwesome\" unicode-range=\"U+0-7F\">\n        <font-face-src>\n          <font-face-uri xlink:href=\"font/fontawesome-webfont.svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\"/>\n        </font-face-src>\n      </font-face>\n    </defs>\n  </svg>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/jobs/hover_op_table", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <tr>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "keys.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "keys.value";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n    ");
    return buffer;}

    data.buffer.push("\n\n<p>");
    stack1 = depth0;
    stack2 = "content.operationName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n<table id=\"hover-op-table\" class=\"table table-bordered\">\n    <thead>\n    <tr>\n        <th>");
    stack1 = depth0;
    stack2 = "common.name";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "common.value";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n    ");
    stack1 = depth0;
    stack2 = "content.operatorPlanObj";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "keys";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n</table>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/jobs/user_filter", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <li>\n          <label class=\"checkbox\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "user.checked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "user.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n      </li>\n  ");
    return buffer;}

    data.buffer.push("\n<button class=\"btn single-btn-group filter-btn\" ");
    stack1 = depth0;
    stack2 = "clickFilterButton";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  ");
    stack1 = depth0;
    stack2 = "common.user";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-filter\"></i>\n</button>\n<ul class=\"dropdown-menu filter-components\">\n  <li id=\"title-bar\">");
    stack1 = depth0;
    stack2 = "apps.filters.filterComponents";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      <a class=\"close\" ");
    stack1 = depth0;
    stack2 = "closeFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">X</a>\n  </li>\n  ");
    stack1 = depth0;
    stack2 = "view.users";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "user";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    <li id=\"button-bar\">\n        <button class=\"btn btn-inverse\" ");
    stack1 = depth0;
    stack2 = "applyFilter";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.apply";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n    </li>\n</ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/menu", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n    <li ");
    stack1 = {};
    stack2 = "active";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><a ");
    stack1 = depth0;
    stack2 = "routing";
    stack3 = depth0;
    stack4 = "navigate";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n    ");
    return buffer;}

    data.buffer.push("\n\n<ul class=\"nav nav-tabs\">\n    ");
    stack1 = depth0;
    stack2 = "view.items";
    stack3 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</ul>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/menu_item", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <span class=\"label label-important alerts-count\">\n      ");
    stack1 = depth0;
    stack2 = "view.alertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </span>\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("-->\n    <!--<span class=\"label operations-count\" ");
    stack1 = depth0;
    stack2 = "showBackgroundOperationsPopup";
    stack3 = {};
    stack4 = "App.router.mainHostDetailsController";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">-->\n      <!--");
    stack1 = depth0;
    stack2 = "view.hostDetailsOperationsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "-->\n    <!--</span>-->\n  <!--");
    return buffer;}

    data.buffer.push("\n\n<a href=\"#/main/");
    stack1 = depth0;
    stack2 = "view.content.routing";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">\n  ");
    stack1 = depth0;
    stack2 = "view.content.label";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "view.alertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <!--");
    stack1 = depth0;
    stack2 = "view.hostDetailsOperationsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("-->\n</a>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n\n  ");
    stack1 = depth0;
    stack2 = "App.MainDatasetsView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring/addTargetCluster", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring/create_new_cluster", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.execute";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.execute";
    stack3['valueBinding'] = stack4;
    stack4 = "resourcemanager-host:8050";
    stack3['placeholder'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.readonly";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.readonly";
    stack3['valueBinding'] = stack4;
    stack4 = "hftp://namenode-host:50070";
    stack3['placeholder'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.workflow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.workflow";
    stack3['valueBinding'] = stack4;
    stack4 = "http://oozie-host:11000/oozie";
    stack3['placeholder'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.write";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.write";
    stack3['valueBinding'] = stack4;
    stack4 = "hdfs://namenode-host:8020";
    stack3['placeholder'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
    return buffer;}

    data.buffer.push("\n<div class=\"mirroring-manage-clusters\">\n  <form class=\"form-horizontal\">\n    <div class=\"control-group row-fluid\">\n      <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "common.name";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"span10\">\n        ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.name";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div class=\"control-group form-category\">\n      <div class=\"category-title\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.interfaces";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.executeTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.readonlyTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.workflowTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.writeTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    <div class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.locationsMessage";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"control-group form-category\">\n      <div class=\"category-title\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.locations";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"control-group row-fluid\">\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.staging";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.staging";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n      <div class=\"control-group row-fluid\">\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.working";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.working";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n      <div class=\"control-group row-fluid\">\n        <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.temp";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"span10\">\n          ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.newCluster.temp";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </form>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring/datasets", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"alert alert-error\">");
    stack1 = depth0;
    stack2 = "mirroring.dataset.services.not.started";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <div class=\"mirroring-top-nav pull-right btn-group\">\n            <button ");
    stack1 = {};
    stack2 = "controller.actionsDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
    stack1 = depth0;
    stack2 = "common.actions";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n            <button ");
    stack1 = {};
    stack2 = "controller.actionsDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n              <span class=\"caret\"></span>\n            </button>\n            <ul class=\"dropdown-menu pull-left\">\n              <li>\n                <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "addNewDataset";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <i class=\"icon-plus\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "mirroring.dataset.createDataset";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n              <li>\n                <a ");
    stack1 = depth0;
    stack2 = "manageClusters";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null);\">\n                  <i class=\"icon-cog\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "mirroring.dataset.manageClusters";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "...\n                </a>\n              </li>\n            </ul>\n          </div>\n        ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <th class=\"first\"></th>\n            ");
    stack1 = depth0;
    stack2 = "view.parentView.nameSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "view.parentView.statusSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <td class=\"first\"></td>\n            <td colspan=\"2\">\n              ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.dataset.loading.error";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n          ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "controller.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(17, program17, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(15, program15, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "dataset";
    stack7 = helpers.each;
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "view.DatasetView";
    stack3 = {};
    stack4 = "dataset";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <td class=\"first\">\n                  <span ");
    stack1 = {};
    stack2 = "dataset.healthClass dataset.healthIconClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n                </td>\n                <td class=\"name\">\n                  <a title=\"");
    stack1 = depth0;
    stack2 = "dataset.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" href=\"javascript:void(null)\">");
    stack1 = depth0;
    stack2 = "dataset.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n                </td>\n                <td><span ");
    stack1 = {};
    stack2 = "dataset.isRunning:text-info dataset.isSuspended:text-warning";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "dataset.statusFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></td>\n                ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <tr>\n                  <td class=\"first\"></td>\n                  <td colspan=\"2\">\n                    ");
    stack1 = depth0;
    stack2 = "mirroring.table.noDatasets";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                  </td>\n                </tr>\n              ");
    return buffer;}

  function program17(depth0,data) {
    
    
    data.buffer.push("\n              <tr>\n                <td class=\"first\"></td>\n                <td colspan=\"2\">\n                  <div class=\"spinner\"></div>\n                </td>\n              </tr>\n            ");}

    data.buffer.push("\n<div id=\"mirroring\">\n  ");
    stack1 = depth0;
    stack2 = "controller.isRequiredServicesStarted";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div class=\"row-fluid\">\n    <div class=\"span4\">\n      <div>\n        <div class=\"pull-left\">\n          <h4>");
    stack1 = depth0;
    stack2 = "mirroring.dataset.dataSets";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n        </div>\n        ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n      <table class=\"table table-bordered dataset-thead\">\n        <thead>\n          ");
    stack1 = depth0;
    stack2 = "view.sortView";
    stack3 = {};
    stack4 = "view.filteredContent";
    stack3['contentBinding'] = stack4;
    stack4 = "label-row";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        <tr class=\"filter-row\">\n          <th class=\"first\"></th>\n          <th>");
    stack1 = depth0;
    stack2 = "view.nameFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n          <th>");
    stack1 = depth0;
    stack2 = "view.statusFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n        </tr>\n        </thead>\n      </table>\n      <div class=\"scrollable-table\">\n        <table class=\"table\">\n          <tbody>\n          ");
    stack1 = depth0;
    stack2 = "controller.isDatasetLoadingError";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(9, program9, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tbody>\n        </table>\n      </div>\n    </div>\n    <div class=\"span8 jobs-container\">\n      ");
    stack1 = depth0;
    stack2 = "App.MainDatasetJobsView";
    stack3 = {};
    stack4 = "App.router.mainDatasetJobsController";
    stack3['controllerBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring/edit_dataset", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    
    data.buffer.push("\n  <div class=\"spinner spinner-centered\"></div>\n");}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.targetClusterSelect";
    stack3 = {};
    stack4 = "formFields.datasetTargetClusterName";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "manageClusters";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "mirroring.dataset.addTargetCluster";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n            ");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.isLoaded";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div ");
    stack1 = {};
    stack2 = ":row-fluid view.isLoaded::content-loading";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <form class=\"form-horizontal\">\n    <div class=\"add-data-set\">\n      <table>\n        <tr ");
    stack1 = {};
    stack2 = "errors.isNameError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <td class=\"percent25\" colspan=\"1\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.name";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "formFields.datasetName";
    stack3['valueBinding'] = stack4;
    stack4 = "span4";
    stack3['class'] = stack4;
    stack4 = "controller.isEdit";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n          </td>\n        </tr>\n        <tr>\n          <td class=\"spacer\" colspan=\"3\"></td>\n        </tr>\n\n        <tr>\n          <td class=\"percent25\" colspan=\"1\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.type";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.type.HDFS";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n        </tr>\n        <tr>\n          <td class=\"spacer\" colspan=\"3\"></td>\n        </tr>\n\n        <tr>\n          <td class=\"percent25\" colspan=\"1\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.sourceCluster";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            ");
    stack1 = depth0;
    stack2 = "App.clusterName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n        </tr>\n        <tr>\n          <td class=\"spacer\" colspan=\"3\"></td>\n        </tr>\n\n        <tr ");
    stack1 = {};
    stack2 = "errors.isSourceDirError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <td colspan=\"1\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.sourceDir";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "formFields.datasetSourceDir";
    stack3['valueBinding'] = stack4;
    stack4 = "span4";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.sourceDir";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n          </td>\n        </tr>\n        <tr>\n          <td class=\"spacer\" colspan=\"3\"></td>\n        </tr>\n        <tr ");
    stack1 = {};
    stack2 = "errors.isTargetClusterNameError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <td colspan=\"1\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.targetCluster";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            ");
    stack1 = depth0;
    stack2 = "view.hasTargetClusters";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.targetClusterName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n          </td>\n        </tr>\n        <tr>\n          <td class=\"spacer\" colspan=\"3\"></td>\n        </tr>\n        <tr ");
    stack1 = {};
    stack2 = "errors.isTargetDirError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <td class=\"spacer\" colspan=\"1\">\n            ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.targetDir";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "formFields.datasetTargetDir";
    stack3['valueBinding'] = stack4;
    stack4 = "span4";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.targetDir";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n          </td>\n        </tr>\n        <tr>\n          <td class=\"spacer\" colspan=\"3\"></td>\n        </tr>\n        <tr>\n          <td colspan=\"1\"> ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.schedule";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " </td>\n          <td colspan=\"2\" style=\"text-align: left\">\n            <div ");
    stack1 = {};
    stack2 = "errors.isStartDateError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "formFields.datasetStartDate";
    stack3['valueBinding'] = stack4;
    stack4 = "input-small datepicker";
    stack3['class'] = stack4;
    stack4 = "controller.isEdit";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.select";
    stack3 = {};
    stack4 = "view.hourOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.hoursForStart";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = "controller.isEdit";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.select";
    stack3 = {};
    stack4 = "view.minuteOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.minutesForStart";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = "controller.isEdit";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.select";
    stack3 = {};
    stack4 = "view.middayPeriodOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.middayPeriodForStart";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = "controller.isEdit";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.startDate";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            </div>\n            <div>\n              ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.schedule.to";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n            <div ");
    stack1 = {};
    stack2 = "errors.isEndDateError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "formFields.datasetEndDate";
    stack3['valueBinding'] = stack4;
    stack4 = "input-small datepicker";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.hourOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.hoursForEnd";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.minuteOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.minutesForEnd";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.middayPeriodOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.middayPeriodForEnd";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.endDate";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            </div>\n\n            <div ");
    stack1 = {};
    stack2 = " :each-row errors.isFrequencyError:error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.schedule.repeatEvery";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "formFields.datasetFrequency";
    stack3['valueBinding'] = stack4;
    stack4 = "input-mini";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.repeatOptions";
    stack3['contentBinding'] = stack4;
    stack4 = "formFields.repeatOptionSelected";
    stack3['selectionBinding'] = stack4;
    stack4 = "input-small";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "errorMessages.frequency";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            </div>\n          </td>\n        </tr>\n      </table>\n    </div>\n  </form>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring/jobs", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n<div class=\"dataset-details\">\n  <div class=\"top-left\">\n    <h4>");
    stack1 = depth0;
    stack2 = "view.dataset.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</h4>\n  </div>\n  <div class=\"bottom-left\">\n    <table>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "mirroring.dataset.source";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dataset.sourceClusterName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dataset.sourceDir";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <tr>\n        <td>");
    stack1 = depth0;
    stack2 = "mirroring.dataset.target";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dataset.targetClusterName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
    stack1 = depth0;
    stack2 = "view.dataset.targetDir";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n    </table>\n  </div>\n  ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div class=\"bottom-right\">");
    stack1 = depth0;
    stack2 = "common.download";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "downloadEntity";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "mirroring.dataset.entity";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ".xml</a></div>\n</div>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"top-right\">\n      <div class=\"btn-group pull-right\">\n        <button class=\"btn\">");
    stack1 = depth0;
    stack2 = "common.actions";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n        <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n          <span class=\"caret\"></span>\n        </button>\n        <ul class=\"dropdown-menu pull-right\">\n          ");
    stack1 = depth0;
    stack2 = "view.dataset.isSuspended";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.dataset.isRunning";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.dataset.isSubmitted";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          <li>\n            <a ");
    stack1 = depth0;
    stack2 = "editDataset";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null);\">\n              ");
    stack1 = depth0;
    stack2 = "common.edit";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n          <li>\n            <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "delete";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              ");
    stack1 = depth0;
    stack2 = "common.delete";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        </ul>\n      </div>\n      <span ");
    stack1 = {};
    stack2 = ":label view.dataset.isRunning:label-info  view.dataset.isSuspended:label-warning  view.dataset.isSubmitted:label-success :pull-right :dataset-status";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.dataset.statusFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n    </div>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <li>\n              <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "resume";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.resume";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <li>\n              <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "suspend";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.suspend";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <li>\n              <a href=\"javascript:void(null);\" ");
    stack1 = depth0;
    stack2 = "schedule";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.schedule";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <th class=\"first\"></th>\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.idSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.startSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.endSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.parentView.statusSort";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <td class=\"first\"></td>\n      <td colspan=\"4\">\n        ");
    stack1 = depth0;
    stack2 = "mirroring.dataset.dataset.loading.instances";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "view.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(14, program14, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(27, program27, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(25, program25, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "job";
    stack7 = helpers.each;
    tmp1 = self.program(16, program16, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program16(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.JobView";
    stack3 = {};
    stack4 = "job";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <td class=\"first\">\n              <span ");
    stack1 = {};
    stack2 = "job.healthClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n            </td>\n\n            <td>");
    stack1 = depth0;
    stack2 = "job.name";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>");
    stack1 = depth0;
    stack2 = "job.startFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>");
    stack1 = depth0;
    stack2 = "job.endFormatted";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>\n              ");
    stack1 = depth0;
    stack2 = "view.showActions";
    stack3 = helpers['if'];
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(23, program23, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </td>\n            ");
    return buffer;}
  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                <div class=\"btn-group display-inline-block\">\n                  <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n                    ");
    stack1 = depth0;
    stack2 = "job.statusFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    <span class=\"caret\"></span>\n                  </a>\n                  <ul class=\"dropdown-menu\">\n                    ");
    stack1 = depth0;
    stack2 = "job.isSuspended";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(21, program21, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    <li>\n                      <a href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "job";
    stack3 = depth0;
    stack4 = "killInstance";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "mirroring.dataset.killInstance";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                    </li>\n                  </ul>\n                </div>\n              ");
    return buffer;}
  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                      <li>\n                        <a href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "job";
    stack3 = depth0;
    stack4 = "resumeInstance";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "mirroring.dataset.resumeInstance";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      </li>\n                    ");
    return buffer;}

  function program21(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                      <li>\n                        <a href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "job";
    stack3 = depth0;
    stack4 = "suspendInstance";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "mirroring.dataset.suspendInstance";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                      </li>\n                    ");
    return buffer;}

  function program23(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "job.statusFormatted";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program25(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <tr>\n            <td class=\"first\"></td>\n            <td colspan=\"4\">\n              ");
    stack1 = depth0;
    stack2 = "mirroring.table.noJobs";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </td>\n          </tr>\n        ");
    return buffer;}

  function program27(depth0,data) {
    
    
    data.buffer.push("\n        <tr>\n          <td class=\"first\"></td>\n          <td colspan=\"4\">\n            <div class=\"spinner\"></div>\n          </td>\n        </tr>\n      ");}

  function program29(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"page-bar\">\n      <div class=\"items-on-page\">\n        <label>");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          : ");
    stack1 = depth0;
    stack2 = "view.rowsPerPageSelectView";
    stack3 = {};
    stack4 = "view.displayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n      </div>\n      <div class=\"info\">");
    stack1 = depth0;
    stack2 = "view.paginationInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"paging_two_button\">\n        ");
    stack1 = depth0;
    stack2 = "view.paginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.paginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  ");
    return buffer;}

    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.showActions";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div class=\"dataset-instances\">\n  <table class=\"table table-bordered table-striped\">\n    <thead>\n      ");
    stack1 = depth0;
    stack2 = "view.sortView";
    stack3 = {};
    stack4 = "view.filteredContent";
    stack3['contentBinding'] = stack4;
    stack4 = "label-row";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <tr class=\"filter-row\">\n      <th class=\"first\"></th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.idFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.startFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.endFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "view.statusFilterView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.controller.isDatasetLoadingError";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(13, program13, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n  </table>\n  ");
    stack1 = depth0;
    stack2 = "view.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(29, program29, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/mirroring/manage_clusters", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    
    data.buffer.push("\n    <div class=\"spinner spinner-centered\"></div>\n  ");}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.execute";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.execute";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.readonly";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.readonly";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.workflow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.workflow";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.write";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.write";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
    return buffer;}

    data.buffer.push("\n<div class=\"mirroring-manage-clusters\">\n  ");
    stack1 = depth0;
    stack2 = "controller.isLoaded";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <div ");
    stack1 = {};
    stack2 = ":row-fluid controller.isLoaded::content-loading";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"span4\">\n      ");
    stack1 = depth0;
    stack2 = "view.clusterSelect";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      <div class=\"btn-toolbar pull-right\">\n        <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "addCluster";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus\"></i></button>\n        <button class=\"btn\" ");
    stack1 = depth0;
    stack2 = "removeCluster";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.removeDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-minus\"></i></button>\n      </div>\n    </div>\n    <div class=\"span8 target-cluster-form\">\n      <form class=\"form-horizontal\">\n        <div class=\"control-group form-category\">\n          <div class=\"category-title\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.interfaces";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.executeTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.readonlyTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.workflowTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.InputContainer";
    stack3 = {};
    stack4 = "controller.writeTooltip";
    stack3['titleBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n        <div class=\"control-group form-category\">\n          <div class=\"category-title\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.locations";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          <div class=\"control-group row-fluid\">\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.staging";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.staging";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          </div>\n          <div class=\"control-group row-fluid\">\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.working";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.working";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          </div>\n          <div class=\"control-group row-fluid\">\n            <div class=\"control-label-manage-clusters span2\">");
    stack1 = depth0;
    stack2 = "mirroring.manageClusters.temp";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"span10\">\n              ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "input-fluid";
    stack3['class'] = stack4;
    stack4 = "controller.selectedCluster.temp";
    stack3['valueBinding'] = stack4;
    stack4 = "disabled";
    stack3['disabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          </div>\n        </div>\n      </form>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"row-fluid\">\n  <div class=\"services-menu well span2 service-menu-width\" style=\"padding: 8px 0\">\n    ");
    stack1 = depth0;
    stack2 = "App.MainServiceMenuView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.AllServicesActionView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"span10 summary-width\">\n    ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/add", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n\n      <!--<a class=\"btn back\" ");
    stack1 = depth0;
    stack2 = "backToServices";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">← Back to Services</a>-->\n\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "services.add.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step5.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step6.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep4:active view.isStep4Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step7.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep5:active view.isStep5Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep5";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step8.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep6:active view.isStep6Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep6";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step9.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep7:active view.isStep7Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep7";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step10.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/all_services_actions", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"service-button\" style=\"margin-top: 10px;\">\n    <div class=\"btn-group display-inline-block span11 offset1\">\n      <a class=\"btn dropdown-toggle span10\" data-toggle=\"dropdown\" href=\"#\">");
    stack1 = depth0;
    stack2 = "common.actions";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"caret\"></span>\n      </a>\n      <ul class=\"pull-left dropdown-menu\">\n        ");
    stack1 = depth0;
    stack2 = "App.supports.addServices";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "App.supports.startStopAllServices";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n    </div>\n  </div>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <li ");
    stack1 = {};
    stack2 = "controller.isAllServicesInstalled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\"\n              ");
    stack1 = {};
    stack2 = "controller.isAllServicesInstalled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "gotoAddService";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <i class=\"icon-plus icon-white\"></i> ");
    stack1 = depth0;
    stack2 = "services.service.add";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n          </li>\n          <li class=\"divider\"></li>\n        ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <li ");
    stack1 = {};
    stack2 = "controller.isStartAllDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\" data-toggle=\"modal\"\n              ");
    stack1 = {};
    stack2 = "controller.isStartAllDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "startAllService";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <i ");
    stack1 = {};
    stack2 = ":icon-play controller.isStartAllDisabled:disabled:enabled ";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
    stack1 = depth0;
    stack2 = "services.service.startAll";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n          <li ");
    stack1 = {};
    stack2 = "controller.isStopAllDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\" data-toggle=\"modal\"\n              ");
    stack1 = {};
    stack2 = "controller.isStopAllDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "stopAllService";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <i ");
    stack1 = {};
    stack2 = ":icon-stop controller.isStopAllDisabled:disabled:enabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
    stack1 = depth0;
    stack2 = "services.service.stopAll";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/configs", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "App.supports.hostOverrides";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <div class=\"clearfix\"></div>\n    ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigView";
    stack3 = {};
    stack4 = "controller.filter";
    stack3['filterBinding'] = stack4;
    stack4 = "controller.filterColumns";
    stack3['columnsBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "controller.content.isRestartRequired";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          <div>\n            <div class=\"alert alert-warning clearfix\">\n              <i class=\"icon-refresh\"></i> ");
    stack1 = depth0;
    stack2 = "view.needToRestartMessage";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push(" ");
    stack1 = depth0;
    stack2 = "services.service.config.restartService.needToRestart";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "  <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "showComponentsShouldBeRestarted";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.componentsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.componentsCount";
    stack3 = {};
    stack4 = "t:common.component";
    stack3['singular'] = stack4;
    stack4 = "t:common.components";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a> ");
    stack1 = depth0;
    stack2 = "on";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "showHostsShouldBeRestarted";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.hostsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.hostsCount";
    stack3 = {};
    stack4 = "t:common.host";
    stack3['singular'] = stack4;
    stack4 = "t:common.hosts";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              <span class=\"restart-components pull-right\">&nbsp</span>\n              <div class=\"btn-group pull-right\">\n                <button type=\"button\" class=\"btn btn-default dropdown-toggle btn-warning\" data-toggle=\"dropdown\">\n                  ");
    stack1 = depth0;
    stack2 = "hosts.host.details.needToRestart.button";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                  <span class=\"caret\"></span>\n                </button>\n                <ul class=\"dropdown-menu\">\n                  <li><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "restartAllStaleConfigComponents";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "restart.service.all";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                  ");
    stack1 = depth0;
    stack2 = "view.rollingRestartSlaveComponentName";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </ul>\n              </div>\n            </div>\n          </div>\n        ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                    <li><a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.rollingRestartSlaveComponentName";
    stack3 = depth0;
    stack4 = "rollingRestartStaleConfigSlaveComponents";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.rollingRestartActionName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                  ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <p class=\"pull-right\">\n        <!--<input class=\"btn btn-primary\" type=\"button\" value=\"Save\" ");
    data.buffer.push(" />-->\n        <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "doCancel";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        <a class=\"btn btn-primary\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "restartServicePopup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.save";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      </p>\n    ");
    return buffer;}

  function program9(depth0,data) {
    
    
    data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

    data.buffer.push("\n\n<div id=\"serviceConfig\">\n  ");
    stack1 = depth0;
    stack2 = "dataIsLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(9, program9, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/configs_save_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n  <div class=\"pre-scrollable\" style=\"max-height: 250px;\">\n    <ul>\n    ");
    stack1 = depth0;
    stack2 = "view.getDisplayMessage";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "val";
    stack7 = helpers.each;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </ul>\n  </div>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <li>\n        ");
    stack1 = depth0;
    stack2 = "val";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </li>\n    ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n  <i class=\"icon-warning-sign\"></i>  ");
    stack1 = depth0;
    stack2 = "view.runningHostsMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <table class=\"table-striped running-host-components-table\">\n    <tr><th>");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th><th>");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th></tr>\n    ");
    stack1 = depth0;
    stack2 = "view.runningHosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </table>\n");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <tr><td>");
    stack1 = depth0;
    stack2 = "host.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td><td>");
    stack1 = depth0;
    stack2 = "host.components";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td></tr>\n    ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n  <i class=\"icon-warning-sign\"></i>  ");
    stack1 = depth0;
    stack2 = "view.unknownHostsMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <table class=\"table-striped running-host-components-table\">\n    <tr><th>");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th><th>");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th></tr>\n    ");
    stack1 = depth0;
    stack2 = "view.unknownHosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </table>\n");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <tr><td>");
    stack1 = depth0;
    stack2 = "host.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td><td>");
    stack1 = depth0;
    stack2 = "host.components";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td></tr>\n    ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"");
    stack1 = depth0;
    stack2 = "view.messageClass";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\">");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n");
    stack1 = depth0;
    stack2 = "view.flag";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.runningHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.unknownHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/save_popup_footer", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n<div class=\"modal-footer\">\n  <button type=\"button\" class=\"btn btn\" ");
    stack1 = depth0;
    stack2 = "onCancel";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  <button type=\"button\" class=\"btn btn\" ");
    stack1 = depth0;
    stack2 = "onDiscard";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.discard";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n  <button type=\"button\" class=\"btn btn-success\" ");
    stack1 = {};
    stack2 = "view.isSaveDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "onSave";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.save";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHdfsView";
    stack3 = {};
    stack4 = true;
    stack3['showOnlyRows'] = stack4;
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceMapreduceView";
    stack3 = {};
    stack4 = true;
    stack3['showOnlyRows'] = stack4;
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceYARNView";
    stack3 = {};
    stack4 = true;
    stack3['showOnlyRows'] = stack4;
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/mapreduce2";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHbaseView";
    stack3 = {};
    stack4 = true;
    stack3['showOnlyRows'] = stack4;
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/zookeeper";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/oozie";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/ganglia";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/hive";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program21(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/hue";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program23(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceFlumeView";
    stack3 = {};
    stack4 = true;
    stack3['showOnlyRows'] = stack4;
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program25(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/falcon";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program27(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceStormView";
    stack3 = {};
    stack4 = true;
    stack3['showOnlyRows'] = stack4;
    stack4 = "view.service";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program29(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/tez";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program31(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/pig";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program33(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "templates/main/service/info/summary/sqoop";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program35(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <div class=\"box\">\n        <div class=\"box-header\">\n          <h4>");
    stack1 = depth0;
    stack2 = "services.alerts.headingOfList";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n          ");
    stack1 = depth0;
    stack2 = "controller.isNagiosInstalled";
    stack3 = helpers['if'];
    tmp1 = self.program(36, program36, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n        <ul id='summary-alerts-list' class=\"alerts\">\n          ");
    stack1 = depth0;
    stack2 = "view.alertsController.isLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(38, program38, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(47, program47, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n      </div>\n    ");
    return buffer;}
  function program36(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n            <div class=\"btn-group\">\n              <a class=\"btn\" target=\"_blank\" rel=\"tooltip\" title=\"Go to Nagios\" ");
    stack1 = {};
    stack2 = "controller.nagiosUrl";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-link\"></i></a>\n            </div>\n          ");
    return buffer;}

  function program38(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.alerts.length";
    stack3 = helpers['if'];
    tmp1 = self.program(39, program39, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(42, program42, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program39(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "view.alerts";
    stack3 = helpers.each;
    tmp1 = self.program(40, program40, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program40(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "App.AlertItemView";
    stack3 = {};
    stack4 = "this";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program42(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "controller.isNagiosInstalled";
    stack3 = helpers['if'];
    tmp1 = self.program(43, program43, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(45, program45, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program43(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <div class=\"alert alert-info\">\n                  ");
    stack1 = depth0;
    stack2 = "services.service.info.summary.nagios.noAlerts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                </div>\n              ");
    return buffer;}

  function program45(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <div class=\"alert\">\n                  ");
    stack1 = depth0;
    stack2 = "services.service.info.summary.nagios.alerts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                </div>\n              ");
    return buffer;}

  function program47(depth0,data) {
    
    
    data.buffer.push("\n            <div class=\"spinner\"></div>\n          ");}

  function program49(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n<div class=\"row-fluid\">\n  <div class=\"span12\">\n    <div class=\"box\">\n      <div class=\"box-header\">\n        <h4>");
    stack1 = depth0;
    stack2 = "controller.content.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.metrics";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n        ");
    stack1 = depth0;
    stack2 = "controller.isGangliaInstalled";
    stack3 = helpers['if'];
    tmp1 = self.program(50, program50, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n      \n      <div class=\"\">\n        <table class=\"graphs\">\n          ");
    stack1 = depth0;
    stack2 = "view.serviceMetricGraphs";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "graphs";
    stack7 = helpers.each;
    tmp1 = self.program(52, program52, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </table>\n      </div>\n      \n    </div>\n  </div>\n</div>\n");
    return buffer;}
  function program50(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n          <div class=\"btn-group\">\n            <a class=\"btn\" target=\"_blank\" rel=\"tooltip\" title=\"Go to Ganglia\" ");
    stack1 = {};
    stack2 = "view.gangliaUrl";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-link\"></i></a>\n          </div>\n        ");
    return buffer;}

  function program52(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n            <tr>\n              ");
    stack1 = depth0;
    stack2 = "graphs";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "graph";
    stack7 = helpers.each;
    tmp1 = self.program(53, program53, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </tr>\n          ");
    return buffer;}
  function program53(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <td>\n                  <div class=\"\">\n                    ");
    stack1 = depth0;
    stack2 = "graph";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  </div>\n                </td>\n              ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"row-fluid service-block\">\n  <div class=\"span6\">\n\n    <div class=\"box\">\n      <div class=\"box-header\">\n        <h4>");
    stack1 = depth0;
    stack2 = "controller.content.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "services.service.info.menu.summary";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h4>\n      </div>\n      <div class=\"service-content\">\n        <table id=\"summary-info\" class=\"table no-borders table-condensed\">\n          <tbody>\n            ");
    stack1 = depth0;
    stack2 = "view.noTemplateService";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.hdfs";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.mapreduce";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.yarn";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.mapreduce2";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.hbase";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.zookeeper";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.oozie";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.ganglia";
    stack3 = helpers['if'];
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.hive";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.hue";
    stack3 = helpers['if'];
    tmp1 = self.program(21, program21, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.flume";
    stack3 = helpers['if'];
    tmp1 = self.program(23, program23, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.falcon";
    stack3 = helpers['if'];
    tmp1 = self.program(25, program25, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.storm";
    stack3 = helpers['if'];
    tmp1 = self.program(27, program27, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.tez";
    stack3 = helpers['if'];
    tmp1 = self.program(29, program29, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.pig";
    stack3 = helpers['if'];
    tmp1 = self.program(31, program31, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.serviceStatus.sqoop";
    stack3 = helpers['if'];
    tmp1 = self.program(33, program33, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tbody>\n        </table>\n      </div>\n    </div>\n  </div>\n  <div class=\"span6\">\n    ");
    stack1 = depth0;
    stack2 = "view.isNoAlertsService";
    stack3 = helpers.unless;
    tmp1 = self.program(35, program35, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n</div>\n");
    stack1 = depth0;
    stack2 = "view.serviceMetricGraphs.length";
    stack3 = helpers['if'];
    tmp1 = self.program(49, program49, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/falcon", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n");
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n<tr>\n    <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.falcon.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.falcon.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n    <td>\n        <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.falcon.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.falcon.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </td>\n</tr>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/ganglia", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.monitors";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    return buffer;}

    data.buffer.push("\n\n\n\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n<tr>\n  <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.monitorsObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "services.ganglia.monitors";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n  <td>\n    <span>\n      ");
    stack1 = depth0;
    stack2 = "view.monitorsLiveTextView";
    stack3 = helpers.view;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </span> ");
    stack1 = depth0;
    stack2 = "services.service.summary.GangliaMonitorsLIVE";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/hive", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n<tr>\n    <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.hive.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.hive.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n    <td>\n        <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.hive.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.hive.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </td>\n</tr>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/hue", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/mapreduce2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n<tr>\n    <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.service.summary.mapreduce2.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.service.summary.mapreduce2.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n    <td>\n        <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.service.summary.mapreduce2.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.service.summary.mapreduce2.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </td>\n</tr>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/master_components", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n  <tr>\n     <td>\n      <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "masterComp.host";
    stack3 = depth0;
    stack4 = "showDetails";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " title=\"");
    stack1 = depth0;
    stack2 = "masterComp.host.publicHostName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" rel=\"UsageTooltip\">\n       ");
    stack1 = depth0;
    stack2 = "masterComp.displayNameAdvanced";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(4, program4, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </a>\n     </td>\n   <td>\n     <span rel='healthTooltip' ");
    stack1 = {};
    stack2 = "masterComp.statusClass masterComp.statusIconClass";
    stack1['class'] = stack2;
    stack2 = "masterComp.passiveTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></span>\n     ");
    stack1 = depth0;
    stack2 = "masterComp.componentTextStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n   </td>\n  </tr>\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n         ");
    stack1 = depth0;
    stack2 = "masterComp.displayNameAdvanced";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n       ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n         ");
    stack1 = depth0;
    stack2 = "masterComp.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n       ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.mastersComp";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "masterComp";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/oozie", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n<tr>\n    <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.oozie.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.oozie.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n    <td>\n        <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.oozie.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.oozie.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </td>\n</tr>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/pig", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n<tr>\n  <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.pig.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.pig.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n  <td>\n    <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.pig.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.pig.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/sqoop", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n<tr>\n  <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.sqoop.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.sqoop.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n  <td>\n    <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.sqoop.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.sqoop.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/tez", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n<tr>\n  <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.tez.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.tez.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n  <td>\n    <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.tez.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.tez.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary/zookeeper", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


    data.buffer.push("\n\n\n\n");
    stack1 = depth0;
    stack2 = "view.sumMasterComponentView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n<tr>\n    <td class=\"summary-label\"><a ");
    stack1 = depth0;
    stack2 = "view.clientObj";
    stack3 = depth0;
    stack4 = "filterHosts";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.zookeeper.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.zookeeper.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</a></td>\n    <td>\n        <span class=\"green-live\">");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span>");
    stack1 = depth0;
    stack2 = "view.clients.length";
    stack3 = {};
    stack4 = "t:services.zookeeper.client";
    stack3['singular'] = stack4;
    stack4 = "t:services.zookeeper.clients";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.installed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </td>\n</tr>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/info/summary_alert", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    
    data.buffer.push("\n              <i class=\"icon-ok icon-large\"></i>\n          ");}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	          ");
    stack1 = depth0;
    stack2 = "isWarning";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program4(depth0,data) {
    
    
    data.buffer.push("\n	              <i class=\"icon-warning-sign icon-large\"></i>\n	          ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "isCritical";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(9, program9, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n	          ");
    return buffer;}
  function program7(depth0,data) {
    
    
    data.buffer.push("\n                  <i class=\"icon-remove icon-large\"></i>\n              ");}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "isPassive";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(12, program12, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program10(depth0,data) {
    
    
    data.buffer.push("\n                    <i class=\"icon-medkit icon-large\"></i>\n                ");}

  function program12(depth0,data) {
    
    
    data.buffer.push("\n                    <i class=\"icon-question-sign icon-large\"></i>\n                ");}

    data.buffer.push("\n\n<div class=\"container-fluid\">\n    <div class=\"row-fluid\">\n        <div class=\"span1 status-icon\">\n          ");
    stack1 = depth0;
    stack2 = "isOk";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n        <div class=\"span11\">\n            <div class=\"row-fluid\">\n                <div class=\"span7 title\">");
    stack1 = depth0;
    stack2 = "title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </div>\n                <div rel=\"tooltip\" ");
    stack1 = {};
    stack2 = "timeSinceAlertDetails";
    stack1['data-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " data-placement=\"right\" class=\"span5 date-time\">");
    stack1 = depth0;
    stack2 = "timeSinceAlert";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n            </div>\n            <div class=\"row-fluid message\">");
    stack1 = depth0;
    stack2 = "message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n        </div>\n    </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/item", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.QuickViewLinks";
    stack3 = {};
    stack4 = "view.service";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <ul class=\"nav nav-pills move\">\n          <li class=\"dropdown\">\n            <a class=\"dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">");
    stack1 = depth0;
    stack2 = "common.quickLinks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "<b class=\"caret\"></b></a>\n            <ul class=\"dropdown-menu\">\n              ");
    stack1 = depth0;
    stack2 = "view.quickLinks";
    stack3 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </li>\n        </ul>\n      ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "disabled";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <li><a class=\"disabled-hdfs-quick-link\">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <li><a ");
    stack1 = {};
    stack2 = "url";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "view.linkTarget";
    stack1['target'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n<div class=\"service-button\">\n  ");
    stack1 = depth0;
    stack2 = "view.isMaintenanceActive";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <div class=\"btn-group display-inline-block\">\n      <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n        ");
    stack1 = depth0;
    stack2 = "services.service.actions.serviceActions";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"caret\"></span>\n      </a>\n      <ul class=\"pull-right dropdown-menu\">\n        <!-- dropdown menu links -->\n\n        <!-- Start/Stop service actions -->\n        ");
    stack1 = depth0;
    stack2 = "controller.content.isClientsOnly";
    stack3 = helpers.unless;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n        <!-- Other service actions -->\n        ");
    stack1 = depth0;
    stack2 = "view.maintenance";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "option";
    stack7 = helpers.each;
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n    </div>\n  ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "controller.isServiceRestartable";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = "controller.isStartDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" ");
    stack1 = {};
    stack2 = "controller.isStartDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                ");
    stack1 = depth0;
    stack2 = "startService";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <i ");
    stack1 = {};
    stack2 = ":icon-play controller.isStartDisabled:disabled:enabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                ");
    stack1 = depth0;
    stack2 = "services.service.start";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n            <li ");
    stack1 = {};
    stack2 = "controller.isStopDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" ");
    stack1 = {};
    stack2 = "controller.isStopDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "stopService";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <i ");
    stack1 = {};
    stack2 = ":icon-stop controller.isStopDisabled:disabled:enabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                ");
    stack1 = depth0;
    stack2 = "services.service.stop";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n        <li ");
    stack1 = {};
    stack2 = "option.disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <a ");
    stack1 = depth0;
    stack2 = "option";
    stack3 = depth0;
    stack4 = "doAction";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = true;
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i ");
    stack1 = {};
    stack2 = "option.cssClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n");
    stack1 = depth0;
    stack2 = "option.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n        </li>\n        ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "App.MainServiceInfoMenuView";
    stack3 = {};
    stack4 = "view.hasConfigTab";
    stack3['configTabBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  <div class=\"span3 quick-links-wrapper\">\n    ");
    stack1 = depth0;
    stack2 = "view.service.quickLinks.length";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n\n");
    stack1 = depth0;
    stack2 = "App.isAdmin";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/manage_configuration_groups_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

  function program3(depth0,data) {
    
    
    data.buffer.push("\n              &nbsp;\n          ");}

    data.buffer.push("\n<div class=\"alert alert-info margin-bottom-5\">");
    stack1 = depth0;
    stack2 = "services.service.config_groups_popup.notice";
    stack3 = {};
    stack4 = "view.displayName";
    stack3['serviceNameBinding'] = stack4;
    foundHelper = helpers['t'];
    stack4 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n<div class=\"row-fluid  manage-configuration-group-content\">\n  <div class=\"span12\">\n    <div class=\"row-fluid\">\n      <div class=\"span4\">\n        <span>&nbsp;</span>\n        ");
    stack1 = depth0;
    stack2 = "Em.Select";
    stack3 = {};
    stack4 = "configGroups";
    stack3['contentBinding'] = stack4;
    stack4 = "content.displayNameHosts";
    stack3['optionLabelPath'] = stack4;
    stack4 = "view.selectedConfigGroup";
    stack3['selectionBinding'] = stack4;
    stack4 = "multiple";
    stack3['multiple'] = stack4;
    stack4 = "group-select";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        <div class=\"btn-toolbar pull-right\">\n          <button rel=\"button-info\" class=\"btn\" ");
    stack1 = {};
    stack2 = "view.addButtonTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "addConfigGroup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus\"></i></button>\n          <button rel=\"button-info\" class=\"btn\" ");
    stack1 = {};
    stack2 = "view.removeButtonTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = "view.isRemoveButtonDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "confirmDelete";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-minus\"></i></button>\n          <div class=\"btn-group\">\n            <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n              <i class=\"icon-cog\"></i>&nbsp;<span class=\"caret\"></span>\n            </button>\n            <ul class=\"dropdown-menu\">\n              <li ");
    stack1 = {};
    stack2 = "view.isRenameButtonDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <a href=\"\" rel=\"button-info-dropdown\" ");
    stack1 = {};
    stack2 = "view.renameButtonTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "renameConfigGroup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.service.config_groups_popup.rename";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              </li>\n              <li ");
    stack1 = {};
    stack2 = "view.isDuplicateButtonDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <a href=\"\" rel=\"button-info-dropdown\" ");
    stack1 = {};
    stack2 = "view.duplicateButtonTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "duplicateConfigGroup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.service.config_groups_popup.duplicate";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              </li>\n            </ul>\n          </div>\n        </div>\n      </div>\n      <div class=\"span8\">\n        <span>&nbsp;</span>\n        <div class=\"row-fluid\">\n          <div class=\"span12 pull-right\">\n            ");
    stack1 = depth0;
    stack2 = "Em.Select";
    stack3 = {};
    stack4 = "selectedConfigGroup.hosts";
    stack3['contentBinding'] = stack4;
    stack4 = "multiple";
    stack3['multiple'] = stack4;
    stack4 = "group-select";
    stack3['class'] = stack4;
    stack4 = "selectedHosts";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          </div>\n          <div class=\"button-group pull-right\">\n            <a rel=\"button-info\" class=\"btn\" ");
    stack1 = {};
    stack2 = "view.addHostTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = "selectedConfigGroup.isAddHostsDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "addHosts";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-plus\"></i></a>\n            <a rel=\"button-info\" class=\"btn\" ");
    stack1 = {};
    stack2 = "view.removeHostTooltip";
    stack1['data-original-title'] = stack2;
    stack2 = "isDeleteHostsDisabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "deleteHosts";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-minus\"></i></a>\n          </div>\n        </div>\n        <div class=\"row-fluid\">\n          <div class=\"span2\">");
    stack1 = depth0;
    stack2 = "common.overrides";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          <div class=\"span10\">\n            <a href=\"\" class=\"properties-link\" ");
    stack1 = depth0;
    stack2 = "showProperties";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = {};
    stack2 = "selectedConfigGroup.propertiesList";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "selectedConfigGroup.properties.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "common.properties";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n          </div>\n        </div>\n        <div class=\"clearfix\"></div>\n        <div class=\"row-fluid\">\n          <div class=\"span2\"><span class=\"prrl\">");
    stack1 = depth0;
    stack2 = "common.description";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n          <div class=\"span10 group-description\">");
    stack1 = depth0;
    stack2 = "selectedConfigGroup.description";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n        </div>\n      </div>\n      <div class=\"clearfix\"></div>\n      <div class=\"row-fluid\">\n        <div class=\"span12 text-error\" id=\"manage-config-group-error-div\">\n          ");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/menu_item", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <span class=\"label operations-count\">\n      ");
    stack1 = depth0;
    stack2 = "view.alertsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </span>\n  ");
    return buffer;}

    data.buffer.push("\n\n<a ");
    stack1 = {};
    stack2 = "view.link";
    stack1['href'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  ");
    stack1 = depth0;
    stack2 = "App.MainDashboardServiceHealthView";
    stack3 = {};
    stack4 = "service-health";
    stack3['class'] = stack4;
    stack4 = "view.content";
    stack3['serviceBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "&nbsp;\n  <span>");
    stack1 = depth0;
    stack2 = "view.content.displayName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  ");
    stack1 = depth0;
    stack2 = "view.alertsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  <i rel=\"tooltip\" ");
    stack1 = {};
    stack2 = ":icon-refresh :restart-required-service view.content.isRestartRequired::hidden";
    stack1['class'] = stack2;
    stack2 = "view.restartRequiredMessage";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n</a>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/new_config_group", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "view.parentView.warningMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    
    data.buffer.push("\n     &nbsp;\n    ");}

    data.buffer.push("\n\n<div class=\"new-config-group-div\">\n  <div class=\"text-warning\">\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.warningMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <table>\n    <tr>\n      <td>");
    stack1 = depth0;
    stack2 = "services.service.config_groups_popup.group_name_lable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</td>\n      <td>");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "255";
    stack3['maxlength'] = stack4;
    stack4 = "configGroupName";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n    <tr>\n      <td>");
    stack1 = depth0;
    stack2 = "services.service.config_groups_popup.group_desc_lable";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</td>\n      <td class=\"row-fluid\">\n        ");
    stack1 = depth0;
    stack2 = "Ember.TextArea";
    stack3 = {};
    stack4 = "span12";
    stack3['classNames'] = stack4;
    stack4 = "configGroupDesc";
    stack3['valueBinding'] = stack4;
    stack4 = "4";
    stack3['rows'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    </tr>\n  </table>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reassign", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <li ");
    stack1 = {};
    stack2 = "isStep5:active view.isStep5Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.reassign.step5.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep6:active view.isStep6Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.reassign.step6.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
    stack1 = depth0;
    stack2 = "services.reassign.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
    stack1 = {};
    stack2 = "isStep1:active view.isStep1Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep1";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.reassign.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep2:active view.isStep2Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep2";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.reassign.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep3:active view.isStep3Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep3";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.reassign.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
    stack1 = {};
    stack2 = "isStep4:active view.isStep4Disabled:disabled";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
    stack1 = depth0;
    stack2 = "gotoStep4";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "services.reassign.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
    stack1 = depth0;
    stack2 = "controller.content.hasManualSteps";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
    stack1 = depth0;
    stack2 = "outlet";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reassign/step1", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "message";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "services.reassign.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "message";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reassign/step3", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "services.reassign.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "services.reassign.step3.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n<div id=\"step8-content\" class=\"well pre-scrollable\">\n  <div id=\"printReview\">\n    <a class=\"btn btn-info pull-right\" ");
    stack1 = depth0;
    stack2 = "printReview";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.print";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a> <br/>\n  </div>\n  <div id=\"step8-info\">\n    <p><b>");
    stack1 = depth0;
    stack2 = "services.reassign.step3.component";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</b> ");
    stack1 = depth0;
    stack2 = "controller.content.reassign.display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n\n    <p><b>");
    stack1 = depth0;
    stack2 = "services.reassign.step3.sourceHost";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</b> ");
    stack1 = depth0;
    stack2 = "view.sourceHost";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n\n    <p><b>");
    stack1 = depth0;
    stack2 = "services.reassign.step3.targetHost";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</b> ");
    stack1 = depth0;
    stack2 = "view.targetHost";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n  </div>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = {};
    stack4 = "true";
    stack3['href'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\"\n     id=\"spinner\" ");
    stack1 = {};
    stack2 = "controller.isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.deploy";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reassign/step4", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "templates/main/admin/highAvailability/progress";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reassign/step5", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.securityNotice";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}

    data.buffer.push("\n<h2>");
    stack1 = depth0;
    stack2 = "services.reassign.step5.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n");
    stack1 = depth0;
    stack2 = "view.manualCommands";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.securityNotice";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reassign/step6", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "templates/main/admin/highAvailability/progress";
    stack3 = helpers.template;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/main/service/reconfigure", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "command.isVisible";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n      <tr>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "App.StageLabelView";
    stack3 = {};
    stack4 = "command";
    stack3['commandBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "command.isCompleted";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(11, program11, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </td>\n        <td>\n          ");
    stack1 = depth0;
    stack2 = "command.isError";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </td>\n      </tr>\n    ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <p ");
    stack1 = {};
    stack2 = "view.isStarted::faintText";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "command.index";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ".&nbsp");
    stack1 = depth0;
    stack2 = "command.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "command.isSuccess";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "App.StageSuccessView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "command.isError";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "App.StageFailureView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "command.isStarted";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "command.isPolling";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                <div class=\"progress-bar pull-left\">\n                  ");
    stack1 = depth0;
    stack2 = "App.StageInProgressView";
    stack3 = {};
    foundHelper = helpers.command;
    stack4 = foundHelper || depth0.command;
    stack3['commandBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </div>\n                <div class=\"progress-percentage pull-left\">");
    stack1 = depth0;
    stack2 = "command.progress";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "%</div>\n              ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            <a class=\"btn btn-primary\"\n               href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "retry";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i\n                    class=\"icon-repeat icon-white\"></i>\n              ");
    stack1 = depth0;
    stack2 = "common.retry";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          ");
    return buffer;}

    data.buffer.push("\n\n<table id=\"security-stages\" class=\"table no-borders\">\n  <tbody>\n  ");
    stack1 = depth0;
    stack2 = "controller.commands";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "command";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </tbody>\n</table>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/utils/ajax", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <br />\n  <br />\n  <pre><strong>");
    stack1 = depth0;
    stack2 = "utils.ajax.errorMessage";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": </strong><span class=\"text-error\">");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></pre>\n");
    return buffer;}

    data.buffer.push("\n\n<span class=\"text-error\">");
    stack1 = depth0;
    stack2 = "view.statusCode";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span> <span>");
    stack1 = depth0;
    stack2 = "view.api";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n");
    stack1 = depth0;
    stack2 = "view.showMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/utils/config_launch_switch_config_group_of_host", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "installer.controls.slaveComponentGroups";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "&#58;&nbsp;\n");
    stack1 = depth0;
    stack2 = "Em.Select";
    stack3 = {};
    stack4 = "view.parentView.configGroups";
    stack3['contentBinding'] = stack4;
    stack4 = "content.displayName";
    stack3['optionLabelPath'] = stack4;
    stack4 = "view.parentView.selectedConfigGroup";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/controls_service_config_radio_buttons", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "option.hidden";
    stack3 = helpers.unless;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <label class=\"radio\">\n      ");
    stack1 = depth0;
    stack2 = "App.ServiceConfigRadioButton";
    stack3 = {};
    stack4 = "view.name";
    stack3['nameBinding'] = stack4;
    stack4 = "option.displayName";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "option.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " &nbsp;\n    </label>\n  ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '';
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.options";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "option";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/controls_service_config_textfield_with_unit", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "App.ServiceConfigTextField";
    stack3 = {};
    stack4 = "view.serviceConfig";
    stack3['serviceConfigBinding'] = stack4;
    stack4 = "false";
    stack3['isPopoverEnabled'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n<span class=\"add-on\">");
    stack1 = depth0;
    stack2 = "view.serviceConfig.unit";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/controls_slave_component_groups_menu", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <span class=\"badge badge-important\">");
    stack1 = depth0;
    stack2 = "view.errorCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n  ");
    return buffer;}

    data.buffer.push("\n\n<a ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "showSlaveComponentGroup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n  ");
    stack1 = depth0;
    stack2 = "view.content.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "view.errorCount";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</a>\n<i ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "removeSlaveComponentGroup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"icon-remove\"></i>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/master_hosts", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "installer.noHostsAssigned";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "value";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "showHosts";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "value.firstObject";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "and";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.otherLength";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.hasNoHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.hasOneHost";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.hasMultipleHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/master_hosts_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <li>");
    stack1 = depth0;
    stack2 = "host";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</li>\n  ");
    return buffer;}

    data.buffer.push("\n\n<ul>\n  ");
    stack1 = depth0;
    stack2 = "view.serviceConfig.value";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/slave_component_hosts", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "none";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " -\n<a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.serviceConfig.category";
    stack3 = depth0;
    stack4 = "showEditSlaveComponentGroups";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  ");
    stack1 = depth0;
    stack2 = "installer.slaveComponentHosts.selectHosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</a>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n<a href=\"#\" ");
    stack1 = depth0;
    stack2 = "view.serviceConfig.category";
    stack3 = depth0;
    stack4 = "showEditSlaveComponentGroups";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  ");
    stack1 = depth0;
    stack2 = "view.hasMultipleHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</a>\n  ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.hosts.firstObject.hostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "and";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.otherLength";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.hosts.firstObject.hostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.hasNoHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/slave_component_hosts_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <tr>\n      <td>\n        <label>");
    stack1 = depth0;
    stack2 = "host.hostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n      </td>\n      <td>\n        ");
    stack1 = depth0;
    stack2 = "App.SlaveComponentDropDownGroupView";
    stack3 = {};
    stack4 = "host";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </td>\n    </tr>\n  ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n        <select ");
    stack1 = depth0;
    stack2 = "changeGroup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = "change";
    stack3['on'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "controller.getGroupsForDropDown";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "groupName";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </select>\n        ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "view.optionTag";
    stack3 = {};
    stack4 = "groupName";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <option value=\"");
    stack1 = depth0;
    stack2 = "groupName";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" ");
    stack1 = {};
    stack2 = "view.selected";
    stack1['selected'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                ");
    stack1 = depth0;
    stack2 = "groupName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </option>\n            ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"slave-hosts-popup\" class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "header";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n<table class=\"table table-striped\">\n  <thead>\n    <tr>\n      <th>");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "common.group";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n  </thead>\n  <tbody>\n  ");
    stack1 = depth0;
    stack2 = "hosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </tbody>\n</table>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/slave_hosts", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "installer.noHostsAssigned";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.hasMultipleHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "showHosts";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "value.firstObject";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "and";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.otherLength";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n  ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "view.value.0";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.hasNoHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/stack_upgrade/step1", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "service.isHidden";
    stack3 = helpers.unless;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <tr>\n            <td>");
    stack1 = depth0;
    stack2 = "service.displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>");
    stack1 = depth0;
    stack2 = "service.version";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>");
    stack1 = depth0;
    stack2 = "service.newVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n            <td>");
    stack1 = depth0;
    stack2 = "service.description";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n          </tr>\n        ");
    return buffer;}

    data.buffer.push("\n\n<div class=\"box alert-info\">\n  <div>\n  ");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step1.description";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <br/>\n  <div>\n    <div>");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step1.installedStackVersion";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "App.currentStackVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    <div>");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step1.upgradeStackVersion";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "content.upgradeVersion";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n  </div>\n</div>\n<div>\n    <table class=\"table table-bordered table-striped\">\n      <thead>\n      <tr>\n        <th>");
    stack1 = depth0;
    stack2 = "common.service";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step1.installedVersion";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step1.newVersion";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
    stack1 = depth0;
    stack2 = "common.description";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      </tr>\n      </thead>\n      <tbody>\n        ");
    stack1 = depth0;
    stack2 = "content.servicesInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "service";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </tbody>\n    </table>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "backToCluster";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.cancel";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n</div>\n\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/stack_upgrade/step2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<div class=\"box alert-info\">\n  <div>\n  ");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.notice.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <br/>\n  <ul class=\"unstyled\">\n      <li>1&#41;&nbsp;");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.notice.first";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n      <li>2&#41;&nbsp;");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.notice.second";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n      <li>3&#41;&nbsp;");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.notice.third";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</li>\n  </ul>\n  <div>\n      <strong>");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.notice.complete";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n</div>\n<div class=\"control-group\">\n    <label><strong>\n      ");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.advancedOption";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      <i href=\"javascript:void(null)\" class=\"icon-question-sign\"\n         rel=\"popover\"\n         ");
    stack1 = {};
    stack2 = "installer.step2.localRepo.tooltip.title";
    stack1['title'] = stack2;
    stack2 = "installer.step2.localRepo.tooltip.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n      </i>\n    </strong></label>\n    <label class=\"checkbox\">\n      ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "content.upgradeOptions.localRepo";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "installer.stackUpgrade.step2.localRepository";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </label>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "upgradeAction";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.upgrade";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/stack_upgrade/step3", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.processView";
    stack3 = {};
    stack4 = "process";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <div class=\"item\">\n                <i ");
    stack1 = {};
    stack2 = "view.icon view.iconColor";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                <a href=\"javascript:void(0)\" ");
    stack1 = depth0;
    stack2 = "process.name";
    stack3 = depth0;
    stack4 = "hostsLogPopup";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.status";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n            </div>\n          ");
    stack1 = depth0;
    stack2 = "process.isRetry";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            <div ");
    stack1 = {};
    stack2 = "view.inProgress::hide :row :span12";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <div class=\"progress-bar span4\">\n                    <div ");
    stack1 = {};
    stack2 = "view.isProcessCompleted::progress-striped view.isProcessCompleted::active view.barColor :progress";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n                    </div>\n                </div>\n                <div class=\"span1\">");
    stack1 = depth0;
    stack2 = "process.progress";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "&#37;</div>\n                <div class=\"span7\">");
    stack1 = depth0;
    stack2 = "process.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n            </div>\n        ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n              <a class=\"btn btn-primary\" ");
    stack1 = depth0;
    stack2 = "process.name";
    stack3 = depth0;
    stack4 = "retry";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = "true";
    stack5['href'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  <i class=\"icon-repeat icon-white\"></i>\n                ");
    stack1 = depth0;
    stack2 = "view.retryMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "finish";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "submitButton";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
    return buffer;}

    data.buffer.push("\n<div id=\"deploy\">\n    <div ");
    stack1 = {};
    stack2 = "view.statusClass :alert";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.overallStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"service\">\n      ");
    stack1 = depth0;
    stack2 = "processes";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "process";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    <div class=\"btn-area\">\n      ");
    stack1 = depth0;
    stack2 = "submitButton";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step0", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "installer.step0.body.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<p class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "installer.step0.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</p>\n<div ");
    stack1 = {};
    stack2 = "view.onError:error :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n  <label class=\"control-label\" for=\"cluster-name\">");
    stack1 = depth0;
    stack2 = "installer.step0.clusterName";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    <a href=\"javascript:void(null)\"\n       rel=\"popover\"\n      ");
    stack1 = {};
    stack2 = "installer.step0.clusterName.tooltip.title";
    stack1['title'] = stack2;
    stack2 = "installer.step0.clusterName.tooltip.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.learnMore";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  </label>\n\n  <div class=\"controls\">\n    ");
    stack1 = depth0;
    stack2 = "App.WizardStep0ViewClusterNameInput";
    stack3 = {};
    stack4 = "content.cluster.name";
    stack3['valueBinding'] = stack4;
    stack4 = "cluster name";
    stack3['placeholder'] = stack4;
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    <p class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "clusterNameError";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n  </div>\n</div>\n\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "invalidClusterName";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n\n\n\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step1", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <label class=\"radio\">");
    stack1 = depth0;
    stack2 = "view.stackRadioButton";
    stack3 = {};
    stack4 = "stack";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "stack.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n<div class=\"accordion\" id=\"advancedRepoAccordion\">\n  <div class=\"accordion-group\">\n    <div class=\"accordion-heading\" ");
    stack1 = depth0;
    stack2 = "onToggleBlock";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <i ");
    stack1 = {};
    stack2 = ":pull-left :accordion-toggle view.isRLCollapsed:icon-caret-right:icon-caret-down";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n      <a class=\"accordion-toggle\">\n        ");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "view.isSubmitDisabled";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </a>\n\n    </div>\n      <div  class=\"accordion-body collapse in\">\n        <div class=\"accordion-inner\">\n          <div class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.message";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          <div class=\"alert alert-warning\">");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.importantMassage";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n\n          <table class=\"table\">\n            <thead>\n            <tr>\n              <th></th>\n              <th class=\"os\">");
    stack1 = depth0;
    stack2 = "common.os";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th class=\"validations\"></th>\n              <th class=\"url\">");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.localRepo.column.baseUrl";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n              <th class=\"actions\"></th>\n            </tr>\n            </thead>\n            <tbody>\n              ");
    stack1 = depth0;
    stack2 = "view.allRepositoriesGroup";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "repoGroup";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </tbody>\n          </table>\n          <div id=\"skip-validation\">\n            <label>");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.skipValidationChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = "checkbox";
    stack3['class'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.skipValidation.message";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n              <i class=\"icon-question-sign\" rel=\"skip-validation-tooltip\" data-toggle=\"tooltip\" ");
    stack1 = {};
    stack2 = "installer.step1.advancedRepo.skipValidation.tooltip";
    stack1['title'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + "></i></label>\n          </div>\n          ");
    stack1 = depth0;
    stack2 = "view.emptyRepoExist";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.invalidUrlExist";
    stack3 = helpers['if'];
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.allRepoUnchecked";
    stack3 = helpers['if'];
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n      </div>\n  </div>\n</div>\n");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <span class=\"badge badge-important\">");
    stack1 = depth0;
    stack2 = "view.totalErrorCnt";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                <tr>\n                  <td class=\"group-checkbox\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "repoGroup.checked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                  <td class=\"os-td\">\n                    <table ");
    stack1 = {};
    stack2 = "repoGroup.checked::disabled-label";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                      <tbody>\n                      ");
    stack1 = depth0;
    stack2 = "repoGroup";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "repo";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      </tbody>\n                    </table>\n                  </td>\n                  <td class=\"validation-results\">\n                    ");
    stack1 = depth0;
    stack2 = "repoGroup.validation";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </td>\n                  <td  ");
    stack1 = {};
    stack2 = ":url-results repoGroup.checked::disabled-textfield repoGroup.empty-error:textfield-error repoGroup.invalid-error:textfield-error";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    ");
    stack1 = depth0;
    stack2 = "Ember.TextField";
    stack3 = {};
    stack4 = "repoGroup.baseUrl";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  </td>\n                  <td class=\"clearAll-icon\">\n                    ");
    stack1 = depth0;
    stack2 = "repoGroup.clearAll";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </td>\n                  <td class=\"action-results\">\n                    ");
    stack1 = depth0;
    stack2 = "repoGroup.undo";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </td>\n                </tr>\n              ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                        <tr>\n                          <td  ");
    stack1 = {};
    stack2 = ":os repo.empty-error:label-error repo.invalid-error:label-error ";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "repo.osType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n                        </tr>\n                      ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "view.popoverView";
    stack3 = {};
    stack4 = "repoGroup";
    stack3['repoGroupBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                      <a ");
    stack1 = depth0;
    stack2 = "repoGroup";
    stack3 = depth0;
    stack4 = "clearGroupLocalRepository";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <i class=\"icon-remove-sign\"></i>\n                      </a>\n                    ");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                      <a ");
    stack1 = depth0;
    stack2 = "repoGroup";
    stack3 = depth0;
    stack4 = "undoGroupLocalRepository";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                        <i class=\"icon-undo\"></i>");
    stack1 = depth0;
    stack2 = "common.undo";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <div class=\"alert\">");
    stack1 = depth0;
    stack2 = "installer.step1.attentionNeeded";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <div class=\"alert\">");
    stack1 = depth0;
    stack2 = "installer.step1.invalidURLAttention";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
    return buffer;}

  function program19(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n            <div class=\"alert\">");
    stack1 = depth0;
    stack2 = "installer.step1.checkAtLeastOneAttention";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
    return buffer;}

  function program21(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n   <a class=\"btn btn-success pull-right\" disabled=\"disabled\">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n");
    return buffer;}

  function program23(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n   <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n");
    return buffer;}

    data.buffer.push("\n<h2>");
    stack1 = depth0;
    stack2 = "installer.step1.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<p class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "installer.step1.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</p>\n<p><b>");
    stack1 = depth0;
    stack2 = "common.stacks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</b></p>\n<form autocomplete=\"off\">\n  ");
    stack1 = depth0;
    stack2 = "view.stacks";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "stack";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</form>\n\n");
    stack1 = depth0;
    stack2 = "App.supports.localRepositories";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n<a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n");
    stack1 = depth0;
    stack2 = "view.isSubmitDisabled";
    stack3 = helpers['if'];
    tmp1 = self.program(21, program21, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(23, program23, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step10", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"alert alert-danger\">\n    ");
    stack1 = depth0;
    stack2 = "installer.step10.nagiosRestartRequired";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n      <li>\n        <span ");
    stack1 = {};
    stack2 = "item.color";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "item.displayStatement";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        <ul>\n        ");
    stack1 = depth0;
    stack2 = "item.status";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "status";
    stack7 = helpers.each;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n      </li>\n    ");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n          <li>\n            <span ");
    stack1 = {};
    stack2 = "status.color";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "status.displayStatement";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            <ul>\n              ");
    stack1 = depth0;
    stack2 = "status.statements";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "statement";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n          </li>\n        ");
    return buffer;}
  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <li>\n                <span ");
    stack1 = {};
    stack2 = "status.color";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "statement.displayStatement";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n              </li>\n              ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "installer.step10.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n");
    stack1 = depth0;
    stack2 = "isNagiosRestartRequired";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "installer.step10.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div id=\"step10-content\" class=\"well pre-scrollable\">\n  <ul>\n    ");
    stack1 = depth0;
    stack2 = "clusterInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </ul>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "complete";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.complete";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step1_addLocalRepository", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n					    <li>\n  					    <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "os";
    stack3 = depth0;
    stack4 = "selectOS";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  ");
    stack1 = depth0;
    stack2 = "os.osType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n				    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n	    <div class=\"alert\">\n			  <strong>");
    stack1 = depth0;
    stack2 = "common.warning";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</strong>  ");
    stack1 = depth0;
    stack2 = "view.errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n			</div>\n	  ");
    return buffer;}

  function program5(depth0,data) {
    
    
    data.buffer.push("\n	    <div class=\"alert alert-info\">\n        Provide location of the local repository for the selected OS\n      </div>\n		");}

    data.buffer.push("\n  <form class=\"form-horizontal\" id=\"addLocalRepositoryPopup\" autocomplete=\"off\">\n    <div class=\"each-row\">\n      <label class=\"control-label\">");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.localRepo.label.os";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        <div class=\"btn-group\">\n				  <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n				    ");
    stack1 = depth0;
    stack2 = "view.selectedOS.osType";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n				    <span class=\"caret\"></span>\n				  </a>\n				  <ul class=\"dropdown-menu\">\n  				  ");
    stack1 = depth0;
    stack2 = "view.oses";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "os";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n				  </ul>\n				</div>\n      </div>\n    </div>\n    <div class=\"each-row\">\n      <label class=\"control-label\">");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.localRepo.label.baseUrl";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "Em.TextField";
    stack3 = {};
    stack4 = "view.enteredUrl";
    stack3['valueBinding'] = stack4;
    stack4 = "span6";
    stack3['classNames'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div class=\"each-row\">\n      <label class=\"control-label\">");
    stack1 = depth0;
    stack2 = "installer.step1.advancedRepo.localRepo.label.stack";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "view.stackName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div style=\"height:100px;\"></div>\n    ");
    stack1 = depth0;
    stack2 = "view.errorMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(5, program5, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </form>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step2", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <p class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "hostsError";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n        ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "App.SshKeyFileUploader";
    stack3 = {};
    stack4 = "view.sshKeyState";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "sshKeyError";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <span class=\"help-inline\">");
    stack1 = depth0;
    stack2 = "sshUserError";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n            ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <h5>");
    stack1 = depth0;
    stack2 = "installer.step2.advancedOptions.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h5>\n    <label ");
    stack1 = {};
    stack2 = ":checkbox";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "content.installOptions.localRepo";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n\n      ");
    stack1 = depth0;
    stack2 = "installer.step2.localRepo.label_use";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n\n      <a href=\"javascript:void(null)\"\n         rel=\"popover\"\n        ");
    stack1 = {};
    stack2 = "installer.step2.localRepo.tooltip.title";
    stack1['title'] = stack2;
    stack2 = "installer.step2.localRepo.tooltip.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "installer.step2.localRepo.tooltip.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
    stack1 = depth0;
    stack2 = "installer.step2.localRepo.label_instead";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </label>\n    ");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"installOptions\">\n  <h2>");
    stack1 = depth0;
    stack2 = "installer.step2.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "installer.step2.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n\n  <div id=\"targetHosts\">\n    <h5>");
    stack1 = depth0;
    stack2 = "installer.step2.targetHosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h5>\n\n    <div ");
    stack1 = {};
    stack2 = "hostsError:error :control-group :target-hosts-input";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <p>");
    stack1 = depth0;
    stack2 = "installer.step2.targetHosts.info";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ". ");
    stack1 = depth0;
    stack2 = "installer.step2.orUse";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"javascript:void(null)\"\n           rel=\"popover\"\n          ");
    stack1 = {};
    stack2 = "installer.step2.hostPattern.tooltip.title";
    stack1['title'] = stack2;
    stack2 = "installer.step2.hostPattern.tooltip.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n          ");
    stack1 = depth0;
    stack2 = "installer.step2.hostPattern.tooltip.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      </p>\n\n      <div class=\"controls\">\n        ");
    stack1 = depth0;
    stack2 = "Ember.TextArea";
    stack3 = {};
    stack4 = "span6";
    stack3['class'] = stack4;
    stack4 = "content.installOptions.hostNames";
    stack3['valueBinding'] = stack4;
    stack4 = "5";
    stack3['rows'] = stack4;
    stack4 = "host names";
    stack3['placeholder'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "hostsError";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    </div>\n  </div>\n\n  <div id=\"hostConnectivity\">\n    <div class=\"ambari-agents\">\n      <h5>");
    stack1 = depth0;
    stack2 = "installer.step2.sshKey";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h5>\n\n      <label class=\"radio\">\n        ");
    stack1 = depth0;
    stack2 = "view.providingSSHKeyRadioButton";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "installer.step2.useSsh.provide";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <a href=\"javascript:void(null)\"\n             rel=\"popover\"\n            ");
    stack1 = {};
    stack2 = "installer.step2.useSsh.tooltip.title";
    stack1['title'] = stack2;
    stack2 = "installer.step2.useSsh.tooltip.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "installer.step2.useSsh.tooltip.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
    stack1 = depth0;
    stack2 = "installer.step2.useSsh.provide_id_rsa";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      </label>\n      <div class=\"ssh-key-input\">\n        ");
    stack1 = depth0;
    stack2 = "view.isFileApi";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        <div ");
    stack1 = {};
    stack2 = "sshKeyError:error :controls :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "Ember.TextArea";
    stack3 = {};
    stack4 = "span6";
    stack3['class'] = stack4;
    stack4 = "3";
    stack3['rows'] = stack4;
    stack4 = "sshKey";
    stack3['id'] = stack4;
    stack4 = "ssh private key";
    stack3['placeholder'] = stack4;
    stack4 = "view.sshKeyState";
    stack3['disabledBinding'] = stack4;
    stack4 = "content.installOptions.sshKey";
    stack3['valueBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "sshKeyError";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n        <div>\n            <label class=\"ssh-user pull-left\">\n              ");
    stack1 = depth0;
    stack2 = "installer.step2.sshUser";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                <a href=\"javascript:void(null)\"\n                   rel=\"popover\"\n                  ");
    stack1 = {};
    stack2 = "installer.step2.sshUser.link";
    stack1['title'] = stack2;
    stack2 = "installer.step2.sshUser.toolTip";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n                  ");
    stack1 = depth0;
    stack2 = "installer.step2.sshUser.link";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
    stack1 = depth0;
    stack2 = "installer.step2.sshUser.account";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n            </label>\n          <div ");
    stack1 = {};
    stack2 = "sshUserError:error :control-group";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "view.textFieldView";
    stack3 = {};
    stack4 = "content.installOptions.sshUser";
    stack3['valueBinding'] = stack4;
    stack4 = "content.installOptions.useSsh";
    stack3['isEnabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "sshUserError";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </div>\n        </div>\n      </div>\n\n      <label class=\"radio\">\n        ");
    stack1 = depth0;
    stack2 = "view.manualRegistrationRadioButton";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        ");
    stack1 = depth0;
    stack2 = "installer.step2.manualInstall.perform";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          <a href=\"javascript:void(null)\"\n             rel=\"popover\"\n            ");
    stack1 = {};
    stack2 = "installer.step2.manualInstall.tooltip.title";
    stack1['title'] = stack2;
    stack2 = "installer.step2.manualInstall.tooltip.content";
    stack1['data-content'] = stack2;
    foundHelper = helpers.translateAttr;
    stack2 = foundHelper || depth0.translateAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
    else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
    else { stack1 = stack2; }
    data.buffer.push(escapeExpression(stack1) + ">\n            ");
    stack1 = depth0;
    stack2 = "installer.step2.manualInstall.tooltip.title";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
    stack1 = depth0;
    stack2 = "installer.step2.manualInstall.perform_on_hosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n      </label>\n\n    </div>\n  </div>\n\n  <div class=\"advancedOptions\">\n    ");
    stack1 = depth0;
    stack2 = "App.supports.localRepositories";
    stack3 = helpers.unless;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"btn-area\">\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.controller.hideBackButton";
    stack3 = helpers.unless;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "evaluateStep";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      ");
    stack1 = depth0;
    stack2 = "installer.step2.registerAndConfirm";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step2ManualInstallPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<p>");
    stack1 = depth0;
    stack2 = "installer.step2.manualInstall.popup.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step2_installed_hosts_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


    data.buffer.push("\n\n<p>");
    stack1 = depth0;
    stack2 = "installer.step2.evaluateStep.installedHosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n<p>");
    stack1 = depth0;
    stack2 = "view.inputtedAgainHostNames";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n<p>");
    stack1 = depth0;
    stack2 = "installer.step2.evaluateStep.continueConfirm";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step3", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <a class=\"btn btn-primary decommission\"\n           href=\"#\" ");
    stack1 = depth0;
    stack2 = "retrySelectedHosts";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i\n          class=\"icon-repeat icon-white\"></i>\n          ");
    stack1 = depth0;
    stack2 = "installer.step3.retryFailed";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n        ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n              <li ");
    stack1 = {};
    stack2 = ":filter-status category.itemClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                <a ");
    stack1 = depth0;
    stack2 = "category";
    stack3 = depth0;
    stack4 = "selectCategory";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                  ");
    stack1 = depth0;
    stack2 = "category.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n              ");
    stack1 = depth0;
    stack2 = "category.last";
    stack3 = helpers.unless;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program4(depth0,data) {
    
    
    data.buffer.push("\n                <li class=\"divider\">|</li>\n              ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n                ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "App.WizardHostView";
    stack3 = {};
    stack4 = "controller.category";
    stack3['categoryBinding'] = stack4;
    stack4 = "host";
    stack3['hostInfoBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n                      <td>\n                        ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "host.isChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      </td>\n                      <td>\n                        ");
    stack1 = depth0;
    stack2 = "host.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                      </td>\n                      <td>\n                          <div ");
    stack1 = {};
    stack2 = "host.bootBarColor host.isBootDone::progress-striped host.isBootDone::active :progress";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                              <div class=\"bar\" style=\"width:100%\">\n                              </div>\n                          </div>\n                      </td>\n                      <td>\n                          <a href=\"javascript:void(null)\"\n                             data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "host";
    stack3 = depth0;
    stack4 = "hostLogPopup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><span  ");
    stack1 = {};
    stack2 = "host.bootStatusColor";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.bootStatusForDisplay";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></a>\n                      </td>\n                      <td>\n                        ");
    stack1 = depth0;
    stack2 = "view.isRemovable";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                        ");
    stack1 = depth0;
    stack2 = "view.isRetryable";
    stack3 = helpers['if'];
    tmp1 = self.program(11, program11, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                      </td>\n                  ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("<a class=\"btn btn-mini\" ");
    stack1 = depth0;
    stack2 = "remove";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i\n                                class=\"icon-trash\"></i>\n                          ");
    stack1 = depth0;
    stack2 = "common.remove";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>");
    return buffer;}

  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("<a class=\"btn btn-mini\" ");
    stack1 = depth0;
    stack2 = "retry";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i\n                                class=\"icon-repeat\"></i>\n                          ");
    stack1 = depth0;
    stack2 = "common.retry";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>");
    return buffer;}

  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <tr>\n                      <td colspan=\"5\">\n                        ");
    stack1 = depth0;
    stack2 = "hosts.table.noHosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      </td>\n                  </tr>\n              ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                  <a ");
    stack1 = depth0;
    stack2 = "selectedHostsPopup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                    ");
    stack1 = depth0;
    stack2 = "view.selectedHostsCount";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    stack1 = depth0;
    stack2 = "view.selectedHostsCount";
    stack3 = {};
    stack4 = "t:hosts.filters.selectedHostInfo";
    stack3['singular'] = stack4;
    stack4 = "t:hosts.filters.selectedHostsInfo";
    stack3['plural'] = stack4;
    foundHelper = helpers.pluralize;
    stack4 = foundHelper || depth0.pluralize;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
    else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
    else { stack1 = stack4; }
    data.buffer.push(escapeExpression(stack1) + "\n                  </a>\n                  -\n                  <a ");
    stack1 = depth0;
    stack2 = "unSelectAll";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "hosts.filters.clearSelection";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <div ");
    stack1 = {};
    stack2 = ":alert alert-warn";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "registeredHostsPopup";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.registeredHostsMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n    ");
    return buffer;}

  function program19(depth0,data) {
    
    
    data.buffer.push("\n          <div class=\"spinner\"></div>\n        ");}

    data.buffer.push("\n\n<div id=\"confirm-hosts\">\n  <h2>");
    stack1 = depth0;
    stack2 = "installer.step3.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "installer.step3.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n  <div class=\"box\">\n    <div class=\"box-header\">\n      <div class=\"button-section\">\n        <button class=\"btn btn-primary\" ");
    stack1 = {};
    stack2 = "view.noHostsSelected";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n           ");
    stack1 = depth0;
    stack2 = "removeSelectedHosts";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i\n                class=\"icon-trash icon-white\"></i>\n          ");
    stack1 = depth0;
    stack2 = "installer.step3.removeSelected";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </button>\n        ");
    stack1 = depth0;
    stack2 = "isRetryDisabled";
    stack3 = helpers.unless;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n        <div id=\"host-filter\" class=\"pull-right\">\n          <ul class=\"clearfix\">\n            <li class=\"first\">");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</li>\n            ");
    stack1 = depth0;
    stack2 = "view.categories";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </ul>\n        </div>\n      </div>\n    </div>\n\n      <div class=\"pre-scrollable\" style=\"max-height: 440px;\">\n          <table class=\"table table-bordered table-striped\">\n              <thead>\n              <tr>\n                  <th class=\"span1\">\n                      <div class=\"btn-group\">\n                          <a class=\"btn\">\n                            ");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "view.pageChecked";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                          </a><button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n                              <span class=\"caret\"></span>\n                          </button>\n                          <ul class=\"dropdown-menu\">\n                              <li>\n                                  <a ");
    stack1 = depth0;
    stack2 = "selectAll";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "installer.step3.hostsTable.selectAll";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                              </li>\n                          </ul>\n                      </div>\n                  </th>\n                  <th class=\"span3\">");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n                  <!-- retrieved from local storage initially -->\n                  <th class=\"span3\">");
    stack1 = depth0;
    stack2 = "common.progress";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n                  <th class=\"span2\">");
    stack1 = depth0;
    stack2 = "common.status";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n                  <!-- given by the parsing function that parses data from bootstrap call, dynamically assign the color -->\n                  <th class=\"span3\">");
    stack1 = depth0;
    stack2 = "common.action";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n                  <!-- trash icon -->\n                  <!-- retry icon -->\n              </tr>\n              </thead>\n              <tbody>\n              ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(13, program13, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </tbody>\n          </table>\n      </div>\n      <div id=\"hosts\">\n          <div class=\"page-bar\">\n              <div class=\"selected-hosts-info pull-left\">\n                ");
    stack1 = depth0;
    stack2 = "view.selectedHostsCount";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </div>\n              <div class=\"items-on-page\">\n                  <label>");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.rowsPerPageSelectView";
    stack3 = {};
    stack4 = "view.displayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n              </div>\n              <div class=\"info\">");
    stack1 = depth0;
    stack2 = "view.paginationInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n              <div class=\"paging_two_button\">\n                ");
    stack1 = depth0;
    stack2 = "view.paginationFirst";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                ");
    stack1 = depth0;
    stack2 = "view.paginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                ");
    stack1 = depth0;
    stack2 = "view.paginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                ");
    stack1 = depth0;
    stack2 = "view.paginationLast";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </div>\n          </div>\n      </div>\n  </div>\n    ");
    stack1 = depth0;
    stack2 = "hasMoreRegisteredHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <div ");
    stack1 = {};
    stack2 = ":alert view.status isWarningsBoxVisible::hidden";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        ");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "warnings";
    stack3 = depth0;
    stack4 = "hostWarningsPopup";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.linkText";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
    stack1 = depth0;
    stack2 = "isWarningsLoaded";
    stack3 = helpers.unless;
    tmp1 = self.program(19, program19, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n  <div class=\"btn-area\">\n    <button class=\"btn pull-left\" ");
    stack1 = {};
    stack2 = "isRegistrationInProgress";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</button>\n    <button class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step3_host_log_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div>\n      ");
    stack1 = depth0;
    stack2 = "view.textArea";
    stack3 = {};
    stack4 = "view.host.bootLog";
    stack3['contentBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <pre class=\"bootLog\">");
    stack1 = depth0;
    stack2 = "view.host.bootLog";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</pre>\n    ");
    return buffer;}

    data.buffer.push("\n\n\n<div id=\"host-log\">\n  <div class=\"content-area\">\n      <div class=\"textTrigger\">");
    stack1 = depth0;
    stack2 = "popup.highlight";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n    ");
    stack1 = depth0;
    stack2 = "view.isTextArea";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step3_host_warning_popup_footer", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2;
    data.buffer.push("\n    <div class=\"progress-info active progress\">\n      <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.progressWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n    </div>\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <label ");
    stack1 = {};
    stack2 = "view.updateStatusClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.updateStatus";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <button type=\"button\" class=\"btn btn-info\" ");
    stack1 = {};
    stack2 = "view.isUpdateInProgress";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "onSecondary";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-repeat\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "view.parentView.secondary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</button>\n");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n  <button type=\"button\" class=\"btn\" ");
    stack1 = depth0;
    stack2 = "onPrimary";
    stack3 = {};
    stack4 = "view.parentView";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.parentView.primary";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</button>\n");
    return buffer;}

    data.buffer.push("\n\n<div class=\"update-progress pull-left\">\n  ");
    stack1 = depth0;
    stack2 = "view.isUpdateInProgress";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n");
    stack1 = depth0;
    stack2 = "view.parentView.secondary";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    stack1 = depth0;
    stack2 = "view.parentView.primary";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step3_host_warnings_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <div class=\"span3 offset2\">\n      <a href=\"javascript:void(null)\" title=\"Show Details\" ");
    stack1 = depth0;
    stack2 = "openWarningsInDialog";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n         class=\"task-detail-open-dialog\"><i\n              class=\"icon-external-link\"></i> ");
    stack1 = depth0;
    stack2 = "installer.step3.hostWarningsPopup.report";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    </div>\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n    <div class=\"accordion-group block\">\n      <div class=\"accordion-heading\" ");
    stack1 = depth0;
    stack2 = "category";
    stack3 = depth0;
    stack4 = "onToggleBlock";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n        <i ");
    stack1 = {};
    stack2 = ":pull-left :accordion-toggle category.isCollapsed:icon-caret-right:icon-caret-down";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n        ");
    stack1 = depth0;
    stack2 = "category.warnings.length";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        <a class=\"accordion-toggle\">\n          ");
    stack1 = depth0;
    stack2 = "category.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " (");
    stack1 = depth0;
    stack2 = "category.warnings.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ")\n        </a>\n      </div>\n      <div id=\"");
    stack1 = depth0;
    stack2 = "category.category";
    stack3 = helpers.unbound;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\" class=\"accordion-body collapse in\" style=\"display: none\">\n        <div class=\"accordion-inner\">\n          ");
    stack1 = depth0;
    stack2 = "category.warnings.length";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(16, program16, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </div>\n      </div>\n    </div>\n    ");
    return buffer;}
  function program4(depth0,data) {
    
    
    data.buffer.push("\n        <i class=\"pull-right accordion-toggle icon-warning-sign\"></i>\n        ");}

  function program6(depth0,data) {
    
    
    data.buffer.push("\n        <i class=\"pull-right accordion-toggle icon-ok\"></i>\n        ");}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n          <table>\n            <thead>\n            <tr>\n              <th colspan=\"2\">");
    stack1 = depth0;
    stack2 = "category.message";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</th>\n            </tr>\n            <tr>\n              <th colspan=\"2\"><b>");
    stack1 = depth0;
    stack2 = "category.type";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b></th>\n            </tr>\n            </thead>\n            <tbody>\n            ");
    stack1 = depth0;
    stack2 = "category.warnings";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "warning";
    stack7 = helpers.each;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </tbody>\n          </table>\n          ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n            <tr>\n              <td class=\"warning-name\" ");
    stack1 = {};
    stack2 = "warning.command";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " >");
    stack1 = depth0;
    stack2 = "warning.name";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</td>\n              ");
    stack1 = depth0;
    stack2 = "warning.version";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              <td>");
    stack1 = depth0;
    stack2 = "category.action";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                <a href=\"javascript:void(null);\" rel='HostsListTooltip' ");
    stack1 = {};
    stack2 = "warning.hostsList";
    stack1['data-original-title'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "warning.hosts";
    stack3 = depth0;
    stack4 = "showHostsPopup";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                  ");
    stack1 = depth0;
    stack2 = "warning.hosts.length";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  ");
    stack1 = depth0;
    stack2 = "warning.onSingleHost";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(14, program14, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </a>\n              </td>\n            </tr>\n            ");
    return buffer;}
  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  <td class=\"package-version\">");
    stack1 = depth0;
    stack2 = "warning.version";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n              ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "installer.step3.hostWarningsPopup.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                  ");
    return buffer;}

  function program14(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                  ");
    stack1 = depth0;
    stack2 = "installer.step3.hostWarningsPopup.hosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                  ");
    return buffer;}

  function program16(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "installer.step3.hostWarningsPopup.emptyMessage";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "category.emptyName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    return buffer;}

    data.buffer.push("\n\n\n<div id=\"host-warnings\">\n  <div class=\"notice\">\n    <span>");
    stack1 = depth0;
    stack2 = "installer.step3.hostWarningsPopup.checks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " <b>");
    stack1 = depth0;
    stack2 = "view.warningsNotice";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b>.<br>");
    stack1 = depth0;
    stack2 = "installer.step3.hostWarningsPopup.notice";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n  <div class=\"row-fluid\">\n    <div class=\"span7\">\n      ");
    stack1 = depth0;
    stack2 = "common.hosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "&nbsp;");
    stack1 = depth0;
    stack2 = "view.hostSelectView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    ");
    stack1 = depth0;
    stack2 = "view.categoryWarnings.length";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"accordion warnings-list\" id=\"accordion2\">\n    ");
    stack1 = depth0;
    stack2 = "view.content";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step3_registered_hosts_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <li>");
    stack1 = depth0;
    stack2 = "host";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</li>\n  ");
    return buffer;}

    data.buffer.push("\n\n<p>");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n<ul>\n  ");
    stack1 = depth0;
    stack2 = "view.registeredHosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</ul>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step4", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "isHidden";
    stack3 = helpers.unless;
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n    <tr ");
    stack1 = {};
    stack2 = "isSelected:success:";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <td><label\n        class=\"checkbox\">");
    stack1 = depth0;
    stack2 = "Ember.Checkbox";
    stack3 = {};
    stack4 = "isDisabled";
    stack3['disabledBinding'] = stack4;
    stack4 = "isSelected";
    stack3['checkedBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1));
    stack1 = depth0;
    stack2 = "displayName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n      </td>\n      <td>");
    stack1 = depth0;
    stack2 = "version";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
    stack1 = depth0;
    stack2 = "description";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</td>\n    </tr>\n    ");
    return buffer;}

  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </a>\n    ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"step4\">\n  <h2>");
    stack1 = depth0;
    stack2 = "installer.step4.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
    stack1 = depth0;
    stack2 = "installer.step4.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <table class=\"table table-striped\">\n    <thead>\n    <tr>\n      <th class=\"span3\">");
    stack1 = depth0;
    stack2 = "common.service";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        <span style=\"margin-left:10px\">\n          <a href=\"#\" ");
    stack1 = depth0;
    stack2 = "selectAll";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "isAll:selected:deselected";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "all";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n                |\n                <a\n                  href=\"#\" ");
    stack1 = depth0;
    stack2 = "selectMinimum";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = {};
    stack2 = "isMinimum:selected:deselected";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "none";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        </span>\n      </th>\n      <th>");
    stack1 = depth0;
    stack2 = "common.version";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
    stack1 = depth0;
    stack2 = "common.description";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n    ");
    stack1 = depth0;
    stack2 = "controller";
    stack3 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n  </table>\n\n  <div class=\"btn-area\">\n    ");
    stack1 = depth0;
    stack2 = "view.parentView.controller.hideBackButton";
    stack3 = helpers.unless;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "> ");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step5", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <br>\n    ");
    stack1 = depth0;
    stack2 = "installer.step5.body.hive";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <div class=\"span12 control-group mlc\">\n          <div class=\"row-fluid\">\n            <div class=\"span4\"><span class=\"pull-right control-label\">");
    stack1 = depth0;
    stack2 = "services.reassign.step2.currentHost";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n            <div class=\"span8\"><span>");
    stack1 = depth0;
    stack2 = "currentHostId";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n          </div>\n        </div>\n      ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              <div class=\"row-fluid\">\n                <div class=\"span4\">\n                  <div class=\"control-group\">\n                    <label class=\"pts pull-right\">\n                      ");
    stack1 = depth0;
    stack2 = "controller.isReassignHive";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(8, program8, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    </label>\n                  </div>\n                </div>\n                <div class=\"span8\">\n                  ");
    stack1 = depth0;
    stack2 = "isHiveCoHost";
    stack3 = helpers['if'];
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(12, program12, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                </div>\n              </div>\n            ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                        ");
    stack1 = depth0;
    stack2 = "installer.step5.hiveGroup";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n                      ");
    return buffer;}

  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                        ");
    stack1 = depth0;
    stack2 = "display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ":\n                      ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                    <label class=\"host-name\">");
    stack1 = depth0;
    stack2 = "selectedHost";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "<i class=\"icon-asterisks\">&#10037;</i></label>\n                  ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "App.SelectHostView";
    stack3 = {};
    stack4 = "content.host_name";
    stack3['optionValuePath'] = stack4;
    stack4 = "content.host_info";
    stack3['optionLabelPath'] = stack4;
    stack4 = "selectedHost";
    stack3['selectedHostBinding'] = stack4;
    stack4 = "component_name";
    stack3['componentNameBinding'] = stack4;
    stack4 = "host-select";
    stack3['class'] = stack4;
    stack4 = "zId";
    stack3['zIdBinding'] = stack4;
    stack4 = "isInstalled";
    stack3['disabledBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    stack1 = depth0;
    stack2 = "showAddControl";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                    ");
    stack1 = depth0;
    stack2 = "showRemoveControl";
    stack3 = helpers['if'];
    tmp1 = self.program(15, program15, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "App.AddControlView";
    stack3 = {};
    stack4 = "component_name";
    stack3['componentNameBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    return buffer;}

  function program15(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "App.RemoveControlView";
    stack3 = {};
    stack4 = "component_name";
    stack3['componentNameBinding'] = stack4;
    stack4 = "zId";
    stack3['zIdBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    return buffer;}

  function program17(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"mapping-box round-corners well\">\n      <div class=\"hostString\"><span>");
    stack1 = depth0;
    stack2 = "hostInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n      ");
    stack1 = depth0;
    stack2 = "masterServices";
    stack3 = helpers.each;
    tmp1 = self.program(18, program18, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n    ");
    return buffer;}
  function program18(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <span ");
    stack1 = {};
    stack2 = "isInstalled:assignedService:newService :round-corners";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
    return buffer;}

  function program20(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"remaining-hosts round-corners well\">\n      <span><strong>");
    stack1 = depth0;
    stack2 = "remainingHosts";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</strong> ");
    stack1 = depth0;
    stack2 = "installer.step5.attention";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</span></div>\n    ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "view.body";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    stack1 = depth0;
    stack2 = "hasHiveServer";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n</div>\n<div class=\"assign-masters row-fluid\">\n  <div class=\"select-hosts span7\">\n    <div class=\"row-fluid\">\n      ");
    stack1 = depth0;
    stack2 = "showCurrentHost";
    stack3 = helpers['if'];
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      <div class=\"clearfix\"></div>\n      <div class=\"row-fluid\">\n        <div class=\"span12 control-group\">\n          <form class=\"form-horizontal\" autocomplete=\"off\">\n            <!-- View for array controller -->\n            ");
    stack1 = depth0;
    stack2 = "servicesMasters";
    stack3 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </form>\n        </div>\n      </div>\n    </div>\n  </div>\n\n  <div class=\"host-assignments span5\">\n    ");
    stack1 = depth0;
    stack2 = "masterHostMapping";
    stack3 = helpers.each;
    tmp1 = self.program(17, program17, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    ");
    stack1 = depth0;
    stack2 = "remainingHosts";
    stack3 = helpers['if'];
    tmp1 = self.program(20, program20, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </div>\n  <div class=\"clearfix\"></div>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn pull-left\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = {};
    stack4 = "true";
    stack3['href'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step6", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <div class=\"alert alert-error\">");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n\n              <th>\n                <a href=\"#\" ");
    stack1 = {};
    stack2 = "header.allChecked:selected:deselected";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  ");
    stack1 = depth0;
    stack2 = "header";
    stack3 = depth0;
    stack4 = "selectAllNodes";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "all";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a> |\n\n                <a href=\"#\" ");
    stack1 = {};
    stack2 = "header.noChecked:selected:deselected";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  ");
    stack1 = depth0;
    stack2 = "header";
    stack3 = depth0;
    stack4 = "deselectAllNodes";
    stack5 = {};
    stack6 = "controller";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "none";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n              </th>\n\n          ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      ");
    return buffer;}
  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n          <tr>\n            ");
    stack1 = depth0;
    stack2 = "App.WizardStep6HostView";
    stack3 = {};
    stack4 = "host";
    stack3['hostBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "host.checkboxes";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "checkbox";
    stack7 = helpers.each;
    tmp1 = self.program(10, program10, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          </tr>\n        ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n              ");
    stack1 = depth0;
    stack2 = "host.hostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "host.hasMaster";
    stack3 = helpers['if'];
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            ");
    return buffer;}
  function program8(depth0,data) {
    
    
    data.buffer.push("\n                <i class=icon-asterisks>&#10037;</i>\n              ");}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <td>\n                  <label class=\"checkbox\">\n                    ");
    stack1 = depth0;
    stack2 = "view.checkboxView";
    stack3 = {};
    stack4 = "checkbox";
    stack3['checkboxBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(11, program11, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n                  </label>\n              </td>\n            ");
    return buffer;}
  function program11(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                      ");
    stack1 = depth0;
    stack2 = "checkbox.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                    ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"step6\">\n  <h2>");
    stack1 = depth0;
    stack2 = "view.title";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "view.label";
    stack3 = {};
    stack4 = "true";
    stack3['unescaped'] = stack4;
    stack4 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("</div>\n  ");
    stack1 = depth0;
    stack2 = "errorMessage";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n  <div class=\"pre-scrollable\">\n    <table class=\"table table-striped\" id=\"component_assign_table\">\n      <thead>\n        <tr>\n          <th>");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n          ");
    stack1 = depth0;
    stack2 = "controller.headers";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "header";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </tr>\n      </thead>\n      <tbody>\n      ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = helpers['if'];
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </tbody>\n    </table>\n  </div>\n    <div id=\"hosts\">\n        <div class=\"page-bar\">\n            <div class=\"items-on-page\">\n                <label>");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.rowsPerPageSelectView";
    stack3 = {};
    stack4 = "view.displayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n            </div>\n            <div class=\"info\">");
    stack1 = depth0;
    stack2 = "view.paginationInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n            <div class=\"paging_two_button\">\n              ");
    stack1 = depth0;
    stack2 = "view.paginationFirst";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.paginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.paginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              ");
    stack1 = depth0;
    stack2 = "view.paginationLast";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            </div>\n        </div>\n    </div>\n  <div class=\"btn-area\">\n    <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\" ");
    stack1 = depth0;
    stack2 = "next";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step7", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "App.ServicesConfigView";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n    ");
    return buffer;}

  function program3(depth0,data) {
    
    
    data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

    data.buffer.push("\n\n<div id=\"serviceConfig\">\n    <h2>");
    stack1 = depth0;
    stack2 = "installer.step7.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n    <div class=\"alert alert-info\">\n      ");
    stack1 = depth0;
    stack2 = "installer.step7.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n    </div>\n\n    ");
    stack1 = depth0;
    stack2 = "isAdvancedConfigLoaded";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n\n    <div class=\"btn-area\">\n        <a class=\"btn\" ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n\n        <a\n                class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    </div>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step7_custom_config_error", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      ");
    stack1 = depth0;
    stack2 = "val.siteProperties";
    stack3 = helpers['if'];
    tmp1 = self.program(2, program2, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    ");
    return buffer;}
  function program2(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n        <li>\n          ");
    stack1 = depth0;
    stack2 = "val.serviceName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            <ul>\n              ");
    stack1 = depth0;
    stack2 = "val.siteProperties";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n            </ul>\n        </li>\n      ");
    return buffer;}
  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <li>\n                  ");
    stack1 = depth0;
    stack2 = "item.displayMsg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </li>\n              ");
    return buffer;}

    data.buffer.push("\n\n<h5>");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</h5>\n<br />\n<div class=\"pre-scrollable\" style=\"max-height: 250px;\">\n  <ul>\n    ");
    stack1 = depth0;
    stack2 = "view.customConfig";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "val";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n  </ul>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step8", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div class=\"alert alert-error\">\n    ");
    stack1 = depth0;
    stack2 = "installer.step8.securityWarning";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <p>\n        <b>");
    stack1 = depth0;
    stack2 = "item.display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b> : ");
    stack1 = depth0;
    stack2 = "item.config_value";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </p>\n    ");
    return buffer;}

  function program5(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <li>\n            <p><span class=\"text text-info\">");
    stack1 = depth0;
    stack2 = "item.os_type";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " : </span>");
    stack1 = depth0;
    stack2 = "item.base_url";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n          </li>\n        ");
    return buffer;}

  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <p><b>");
    stack1 = depth0;
    stack2 = "menu.item.services";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</b></p>\n        <ul>\n          ");
    stack1 = depth0;
    stack2 = "controller.services";
    stack3 = helpers.each;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n      ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n            <li>\n              <em><b>");
    stack1 = depth0;
    stack2 = "display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b></em>\n              <ul>\n              ");
    stack1 = depth0;
    stack2 = "service_components";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "component";
    stack7 = helpers.each;
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n              </ul>\n            </li>\n          ");
    return buffer;}
  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n                <li>\n                  <span class=\"text text-info\">");
    stack1 = depth0;
    stack2 = "component.display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " : </span>");
    stack1 = depth0;
    stack2 = "component.component_value";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </li>\n              ");
    return buffer;}

    data.buffer.push("\n\n<h2>");
    stack1 = depth0;
    stack2 = "installer.step8.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div class=\"alert alert-info\">\n  ");
    stack1 = depth0;
    stack2 = "installer.step8.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n");
    stack1 = depth0;
    stack2 = "controller.securityEnabled";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n<div id=\"step8-content\" class=\"well pre-scrollable\">\n  <div id=\"printReview\">\n      <a class=\"btn btn-info pull-right\" ");
    stack1 = depth0;
    stack2 = "printReview";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.print";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a> <br/>\n  </div>\n  <div id=\"step8-info\">\n    ");
    stack1 = depth0;
    stack2 = "controller.clusterInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n\n    <p><b>");
    stack1 = depth0;
    stack2 = "controller.clusterInfo.repoInfo.display_name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</b>:</p>\n    <div>\n      <ul>\n        ");
    stack1 = depth0;
    stack2 = "controller.clusterInfo.repoInfo";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "item";
    stack7 = helpers.each;
    tmp1 = self.program(5, program5, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </ul>\n    </div>\n\n    <div>\n      ");
    stack1 = depth0;
    stack2 = "controller.services.length";
    stack3 = helpers['if'];
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </div>\n  </div>\n</div>\n<div class=\"btn-area\">\n    <a class=\"btn pull-left\" ");
    stack1 = {};
    stack2 = "controller.isBackBtnDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "back";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">&larr; ");
    stack1 = depth0;
    stack2 = "common.back";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\"\n       id=\"spinner\" ");
    stack1 = {};
    stack2 = "controller.isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.deploy";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step8_log_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


    data.buffer.push("\n<p>");
    stack1 = depth0;
    stack2 = "view.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</p>\n<div class=\"progress log_popup\">\n    <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n    </div>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step9", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n        <a class=\"btn btn-primary\"\n           href=\"#\" ");
    stack1 = depth0;
    stack2 = "retry";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-repeat icon-white\"></i>\n          ");
    stack1 = depth0;
    stack2 = "common.retry";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n        </a>\n        ");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
    data.buffer.push("\n            <li ");
    stack1 = {};
    stack2 = ":filter-status category.itemClass";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n              <a ");
    stack1 = depth0;
    stack2 = "category";
    stack3 = depth0;
    stack4 = "selectCategory";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                ");
    stack1 = depth0;
    stack2 = "category.label";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n            ");
    stack1 = depth0;
    stack2 = "category.last";
    stack3 = helpers.unless;
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program4(depth0,data) {
    
    
    data.buffer.push("\n              <li class=\"divider\">|</li>\n            ");}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n          ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n            ");
    stack1 = depth0;
    stack2 = "App.HostStatusView";
    stack3 = {};
    stack4 = "host";
    stack3['objBinding'] = stack4;
    stack4 = "controller";
    stack3['controllerBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = self.program(8, program8, data);
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n          ");
    return buffer;}
  function program8(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n              <td>\n                ");
    stack1 = depth0;
    stack2 = "host.name";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td>\n                <div class=\"progress-bar pull-left\">\n                  <div ");
    stack1 = {};
    stack2 = "view.isHostCompleted::progress-striped view.isHostCompleted::active view.barColor :progress";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n                    <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></div>\n                  </div>\n                </div>\n                <div class=\"progress-percentage pull-left\">");
    stack1 = depth0;
    stack2 = "host.progress";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "%</div>\n              </td>\n              <td>\n                <a ");
    stack1 = {};
    stack2 = "view.isFailed:text-error view.isSuccess:text-success view.isWarning:text-warning";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  href=\"javascript:void(null)\"\n                  data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "hostLogPopup";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "host.message";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</a>\n              </td>\n            ");
    return buffer;}

  function program10(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <tr>\n            <td colspan=\"3\"><p>");
    stack1 = depth0;
    stack2 = "installer.step3.hosts.noHosts";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p></td>\n          </tr>\n        ");
    return buffer;}

  function program12(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n      <p ");
    stack1 = {};
    stack2 = "view.resultMsgColor :alert";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "view.resultMsg";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      ");
    stack1 = depth0;
    stack2 = "view.isHostHeartbeatLost";
    stack3 = helpers['if'];
    tmp1 = self.program(13, program13, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </p>\n    ");
    return buffer;}
  function program13(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4;
    data.buffer.push("\n        <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
    stack1 = depth0;
    stack2 = "hostWithInstallFailed";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.showDetails";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"deploy\">\n\n  <h2>");
    stack1 = depth0;
    stack2 = "installer.step9.header";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">");
    stack1 = depth0;
    stack2 = "installer.step9.body";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</p>\n  <div id=\"overallProgress\">\n    <div class=\"row-fluid\">\n      <div class=\"span10\">\n        <div ");
    stack1 = {};
    stack2 = "view.isStepCompleted::progress-striped view.isStepCompleted::active view.barColor :progress";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          <div class=\"bar\" ");
    stack1 = {};
    stack2 = "view.barWidth";
    stack1['style'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n          </div>\n        </div>\n      </div>\n      <div class=\"span2\">");
    stack1 = depth0;
    stack2 = "view.progressMessage";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  </div>\n\n  <div class=\"box\">\n    <div class=\"box-header\">\n      <div class=\"pull-left\">\n        ");
    stack1 = depth0;
    stack2 = "controller.showRetry";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n      <!-- filter by host level -->\n      <div id=\"host-filter\" class=\"pull-right\">\n        <ul class=\"clearfix\">\n          <li class=\"first\">");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":</li>\n          ");
    stack1 = depth0;
    stack2 = "view.categories";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "category";
    stack7 = helpers.each;
    tmp1 = self.program(3, program3, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </ul>\n      </div>\n    </div>\n    <div class=\"pre-scrollable\">\n      <table id=\"deploy-status-by-host\" class=\"table table-bordered table-striped\">\n        <thead>\n        <tr>\n          <th class=\"host\">\n            ");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n          </th>\n          <th class=\"status\">");
    stack1 = depth0;
    stack2 = "common.status";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n          <!--  given by the parsing function that parses data from bootstrap call -->\n          <th class=\"message\">");
    stack1 = depth0;
    stack2 = "common.message";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n          <!-- retrieved from local storage initially -->\n        </tr>\n        </thead>\n\n        <tbody>\n        ");
    stack1 = depth0;
    stack2 = "view.pageContent";
    stack3 = helpers['if'];
    tmp1 = self.program(6, program6, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(10, program10, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        </tbody>\n      </table>\n    </div>\n      <div id=\"hosts\">\n        <div class=\"page-bar\">\n          <div class=\"selected-hosts-info pull-left\">\n            ");
    stack1 = depth0;
    stack2 = "view.filteredHostsInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            -\n            <a ");
    stack1 = depth0;
    stack2 = "showAllHosts";
    stack3 = {};
    stack4 = "view";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
    stack1 = depth0;
    stack2 = "tableView.filters.showAll";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n          </div>\n          <div class=\"items-on-page\">\n              <label>");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ": ");
    stack1 = depth0;
    stack2 = "view.rowsPerPageSelectView";
    stack3 = {};
    stack4 = "view.displayLength";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</label>\n          </div>\n          <div class=\"info\">");
    stack1 = depth0;
    stack2 = "view.paginationInfo";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</div>\n          <div class=\"paging_two_button\">\n            ");
    stack1 = depth0;
    stack2 = "view.paginationFirst";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "view.paginationLeft";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "view.paginationRight";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n            ");
    stack1 = depth0;
    stack2 = "view.paginationLast";
    stack3 = helpers.view;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n          </div>\n        </div>\n      </div>\n  </div>\n\n  <div>\n    ");
    stack1 = depth0;
    stack2 = "view.resultMsg";
    stack3 = helpers['if'];
    tmp1 = self.program(12, program12, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    <div class=\"btn-area\">\n      <button class=\"btn btn-success pull-right\" ");
    stack1 = {};
    stack2 = "isSubmitDisabled";
    stack1['disabled'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "submit";
    stack3 = {};
    stack4 = "controller";
    stack3['target'] = stack4;
    stack4 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">");
    stack1 = depth0;
    stack2 = "common.next";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n    </div>\n  </div>\n\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step9HostTasksLogPopup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  <div>\n    ");
    stack1 = depth0;
    stack2 = "installer.step9.host.heartbeat_lost_popup";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
    return buffer;}

  function program3(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n  ");
    stack1 = depth0;
    stack2 = "view.isNoTasksScheduled";
    stack3 = helpers['if'];
    tmp1 = self.program(4, program4, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(6, program6, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;}
  function program4(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    ");
    stack1 = depth0;
    stack2 = "installer.step9.host.status.noTasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "\n  ");
    return buffer;}

  function program6(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
    data.buffer.push("\n    <div ");
    stack1 = {};
    stack2 = "view.isLogWrapHidden::hidden :task-list-main-warp";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"task-top-wrap\">\n        <div class=\"task-name-top task-detail-log-rolename\">");
    stack1 = depth0;
    stack2 = "common.tasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"select-wrap\">\n          ");
    stack1 = depth0;
    stack2 = "common.show";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + ":\n          ");
    stack1 = depth0;
    stack2 = "Ember.Select";
    stack3 = {};
    stack4 = "view.categories";
    stack3['contentBinding'] = stack4;
    stack4 = "content.value";
    stack3['optionValuePath'] = stack4;
    stack4 = "content.label";
    stack3['optionLabelPath'] = stack4;
    stack4 = "view.category";
    stack3['selectionBinding'] = stack4;
    stack4 = helpers.view;
    tmp1 = {};
    tmp1.hash = stack3;
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack4.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n      <div id=\"host-log\">\n        ");
    stack1 = depth0;
    stack2 = "view.tasks";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "taskInfo";
    stack7 = helpers.each;
    tmp1 = self.program(7, program7, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n        ");
    stack1 = depth0;
    stack2 = "view.isEmptyList";
    stack3 = helpers['if'];
    tmp1 = self.program(9, program9, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n      </div>\n    </div>\n\n    <div ");
    stack1 = {};
    stack2 = "view.isLogWrapHidden:hidden :task-detail-info";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"task-top-wrap\">\n        <a class=\"task-detail-back\" href=\"javascript:void(null)\" ");
    stack1 = depth0;
    stack2 = "backToTaskList";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ><i\n                class=\"icon-arrow-left\"></i>&nbsp;");
    stack1 = depth0;
    stack2 = "common.tasks";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n\n        <div>\n          <i ");
    stack1 = {};
    stack2 = "view.openedTask.status :task-detail-status-ico view.openedTask.icon";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                  class=\"task-detail-status-ico\"></i>\n          <span class=\"task-detail-log-rolename\">");
    stack1 = depth0;
    stack2 = "view.openedTask.role";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "view.openedTask.command";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n        <div class=\"task-detail-ico-wrap\">\n          <a title=\"Click to Copy\" ");
    stack1 = depth0;
    stack2 = "taskInfo";
    stack3 = depth0;
    stack4 = "textTrigger";
    stack5 = {};
    stack6 = "view";
    stack5['target'] = stack6;
    stack6 = helpers.action;
    tmp1 = {};
    tmp1.hash = stack5;
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack6.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i\n                  class=\"icon-copy\"></i> ");
    stack1 = depth0;
    stack2 = "common.copy";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n          <a title=\"Open in New Window\" ");
    stack1 = depth0;
    stack2 = "openTaskLogInDialog";
    stack3 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-open-dialog\"><i\n                  class=\"icon-external-link\"></i> ");
    stack1 = depth0;
    stack2 = "common.open";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n      </div>\n      <div class=\"task-detail-log-info\">\n        <div class=\"content-area\">\n          <div class=\"task-detail-log-clipboard-wrap\"></div>\n          <div class=\"task-detail-log-maintext\">\n            <h5>stderr: &nbsp; <span class=\"muted\">");
    stack1 = depth0;
    stack2 = "view.openedTask.errorLog";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span></h5>\n            <pre class=\"stderr\">");
    stack1 = depth0;
    stack2 = "view.openedTask.stderr";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</pre>\n            <h5>stdout: &nbsp; <span class=\"muted\"> ");
    stack1 = depth0;
    stack2 = "view.openedTask.outputLog";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " </span></h5>\n            <pre class=\"stdout\">");
    stack1 = depth0;
    stack2 = "view.openedTask.stdout";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "</pre>\n          </div>\n        </div>\n      </div>\n    </div>\n  ");
    return buffer;}
  function program7(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3, stack4, stack5;
    data.buffer.push("\n          <div ");
    stack1 = {};
    stack2 = "taskInfo.isVisible::hidden :log-list-wrap";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
    stack1 = depth0;
    stack2 = "taskInfo";
    stack3 = depth0;
    stack4 = "toggleTaskLog";
    stack5 = helpers.action;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack5.call(depth0, stack4, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n              <div class=\"operation-name-icon-wrap\">\n                <i ");
    stack1 = {};
    stack2 = "taskInfo.status taskInfo.icon";
    stack1['class'] = stack2;
    stack2 = helpers.bindAttr;
    tmp1 = {};
    tmp1.hash = stack1;
    tmp1.contexts = [];
    tmp1.data = data;
    stack1 = stack2.call(depth0, tmp1);
    data.buffer.push(escapeExpression(stack1) + "></i>\n                <a href=\"#\">\n                  ");
    stack1 = depth0;
    stack2 = "taskInfo.role";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + " ");
    stack1 = depth0;
    stack2 = "taskInfo.command";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </div>\n              <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n            </div>\n          </div>\n        ");
    return buffer;}

  function program9(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n          <div class=\"log-list-wrap\">");
    stack1 = depth0;
    stack2 = "installer.step9.hostLog.popup.noTasksToShow";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
    return buffer;}

    data.buffer.push("\n\n");
    stack1 = depth0;
    stack2 = "view.isHeartbeatLost";
    stack3 = helpers['if'];
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.program(3, program3, data);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("templates/wizard/step9_install_host_popup", function(exports, require, module) {
  Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
  helpers = helpers || Ember.Handlebars.helpers;
    var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

  function program1(depth0,data) {
    
    var buffer = '', stack1, stack2, stack3;
    data.buffer.push("\n    <tr>\n      <td>\n        ");
    stack1 = depth0;
    stack2 = "host.hostName";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </td>\n      <td>\n        ");
    stack1 = depth0;
    stack2 = "host.componentNames";
    stack3 = helpers._triageMustache;
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    stack1 = stack3.call(depth0, stack2, tmp1);
    data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    </tr>\n  ");
    return buffer;}

    data.buffer.push("\n\n<div id=\"install-fail-popup\" class=\"pre-scrollable\">\n  <table class=\"table table-striped\">\n    <thead>\n    <tr>\n      <th class=\"span3\">");
    stack1 = depth0;
    stack2 = "common.host";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n      <th class=\"span3\">");
    stack1 = depth0;
    stack2 = "common.components";
    foundHelper = helpers['t'];
    stack3 = foundHelper || depth0['t'];
    tmp1 = {};
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack1);
    tmp1.data = data;
    if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
    else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
    else { stack1 = stack3; }
    data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n  ");
    stack1 = depth0;
    stack2 = "view.failedHosts";
    stack3 = depth0;
    stack4 = "in";
    stack5 = depth0;
    stack6 = "host";
    stack7 = helpers.each;
    tmp1 = self.program(1, program1, data);
    tmp1.hash = {};
    tmp1.contexts = [];
    tmp1.contexts.push(stack5);
    tmp1.contexts.push(stack3);
    tmp1.contexts.push(stack1);
    tmp1.fn = tmp1;
    tmp1.inverse = self.noop;
    tmp1.data = data;
    stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
    if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
    data.buffer.push("\n    </tbody>\n  </table>\n</div>");
    return buffer;
  });
  module.exports = module.id;
});
window.require.register("utils/ajax", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');

  /**
   * Config for each ajax-request
   *
   * Fields example:
   *  mock - testMode url
   *  real - real url (without API prefix)
   *  type - request type (also may be defined in the format method)
   *  format - function for processing ajax params after default formatRequest. May be called with one or two parameters (data, opt). Return ajax-params object
   *  testInProduction - can this request be executed on production tests (used only in tests)
   *
   * @type {Object}
   */
  var urls = {
    'alerts.get_by_service': {
      'real': '/clusters/{clusterName}/services/{serviceName}?fields=alerts',
      'mock': '/data/alerts/HDP2/service_alerts.json'
    },
    'alerts.get_by_host': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=alerts',
      'mock': '/data/alerts/HDP2/host_alerts.json'
    },
    'background_operations.get_most_recent': {
      'real': '/clusters/{clusterName}/requests?to=end&page_size=10&fields=Requests',
      'mock': '/data/background_operations/list_on_start.json',
      'testInProduction': true
    },
    'background_operations.get_by_request': {
      'real': '/clusters/{clusterName}/requests/{requestId}?fields=*,tasks/Tasks/command,tasks/Tasks/command_detail,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/exit_code,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,tasks/Tasks/status&minimal_response=true',
      'mock': '/data/background_operations/task_by_request{requestId}.json',
      'testInProduction': true,
      'format': function (data) {
        return {
          async: !data.sync
        };
      }
    },
    'background_operations.get_by_task': {
      'real': '/clusters/{clusterName}/requests/{requestId}/tasks/{taskId}',
      'mock': '/data/background_operations/list_on_start.json',
      'testInProduction': true,
      'format': function (data) {
        return {
          async: !data.sync
        };
      }
    },
    'service.item.start_stop': {
      'real': '/clusters/{clusterName}/services/{serviceName}',
      'mock': '/data/wizard/deploy/poll_1.json',
      'format': function (data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              "context": data.requestInfo
            },
            Body: {
              ServiceInfo: {
                state: data.state
              }
            }
          })
        };
      }
    },
    'service.item.smoke': {
      'real': '/clusters/{clusterName}/requests',
      'mock': '/data/wizard/deploy/poll_1.json',
      'format': function (data) {
        return {
          'type': 'POST',
          data: JSON.stringify({
            "RequestInfo": {
              "context": data.displayName + " Service Check",
              "command" : data.actionName
            },
            "Requests/resource_filters": [{"service_name" : data.serviceName}]
          })
        };
      }
    },
    'service.item.passive': {
      'real': '/clusters/{clusterName}/services/{serviceName}',
      'mock': '',
      'format': function (data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              "context": data.requestInfo
            },
            Body: {
              ServiceInfo: {
                maintenance_state: data.passive_state
              }
            }
          })
        };
      }
    },
    'service.stale_host_components.start_stop': {
      'real': '/clusters/{clusterName}/host_components?' +
              'HostRoles/stale_configs=true&HostRoles/component_name.in({componentNames})',
      'mock': '/data/wizard/deploy/poll_1.json',
      'format': function (data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              "context": data.requestInfo
            },
            Body: {
              HostRoles: {
                state: data.state
              }
            }
          })
        };
      }
    },
    'service.load_config_groups': {
      'real': '/clusters/{clusterName}/config_groups?ConfigGroup/tag={serviceName}&fields=*',
      'mock': '/data/configurations/config_group.json'
    },
    'reassign.stop_services': {
      'real': '/clusters/{clusterName}/services',
      'mock': '',
      'format': function() {
        return {
          type: 'PUT',
          data: JSON.stringify({
            "RequestInfo": {
              "context": "Stop all services"
            },
            "Body": {
              "ServiceInfo": {
                "state": "INSTALLED"
              }
            }
          })
        }
      }
    },
    'reassign.stop_YMR2_services': {
      'real': '/clusters/{clusterName}/services?ServiceInfo/service_name.in({servicesList})',
      'mock': '',
      'format': function() {
        return {
          type: 'PUT',
          data: JSON.stringify({
            "RequestInfo": {
              "context": "Stop without HDFS"
            },
            "Body": {
              "ServiceInfo": {
                "state": "INSTALLED"
              }
            }
          })
        }
      }
    },
    'reassign.start_services': {
      'real': '/clusters/{clusterName}/services?params/run_smoke_test=true',
      'mock': '',
      'format': function() {
        return {
          type: 'PUT',
          data: JSON.stringify({
            "RequestInfo": {
              "context": "Start all services"
            },
            "Body": {
              "ServiceInfo": {
                "state": "STARTED"
              }
            }
          })
        }
      }
    },
    'reassign.start_YMR2_services': {
      'real': '/clusters/{clusterName}/services/?ServiceInfo/service_name.in({servicesList})',
      'mock': '',
      'format': function() {
        return {
          type: 'PUT',
          data: JSON.stringify({
            "RequestInfo": {
              "context": "Start without HDFS"
            },
            "Body": {
              "ServiceInfo": {
                "state": "STARTED"
              }
            }
          })
        }
      }
    },
    'reassign.maintenance_mode': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function () {
        return {
          data: JSON.stringify(
            {
              "HostRoles": {
                "state": "DISABLED"
              }
            }
          )
        }
      }
    },
    'reassign.remove_component': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'DELETE'
    },

    'reassign.load_configs': {
      'real': '/clusters/{clusterName}/configurations?{urlParams}',
      'mock': ''
    },

    'reassign.save_configs': {
      'real': '/clusters/{clusterName}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          async: false,
          data: JSON.stringify({
            Clusters: {
              desired_config: {
                "type": data.siteName,
                "tag": 'version' + (new Date).getTime(),
                "properties": data.properties
              }
            }
          })
        }
      }
    },

    'host_component.passive': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function(data) {
        return {
          data: JSON.stringify({
            RequestInfo: {
              "context": data.requestInfo
            },
            Body: {
              HostRoles: {
                maintenance_state: data.passive_state
              }
            }
          })
        };
      }
    },

    'config.cluster_configuration.put': {
      'real': '/clusters/{cluster}',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          async: false,
          dataType: 'text',
          data: data.data
        }
      }
    },

    'config.advanced': {
      'real': '{stack2VersionUrl}/stackServices/{serviceName}/configurations?fields=*',
      'mock': '/data/wizard/stack/hdp/version{stackVersion}/{serviceName}.json',
      'format': function (data) {
        return {
          async: !data.sync
        };
      }
    },
    'config.advanced.global': {
      'real': '{stack2VersionUrl}/stackServices?fields=configurations/StackConfigurations/type',
      'mock': '/data/wizard/stack/hdp/version1.3.0/global.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'config.tags': {
      'real': '/clusters/{clusterName}?fields=Clusters/desired_configs',
      'mock': '/data/clusters/cluster.json'
    },
    'config.tags_and_groups': {
      'real': '/clusters/{clusterName}?fields=Clusters/desired_configs,config_groups/*',
      'mock': '/data/clusters/tags_and_groups.json'
    },
    'config.tags.sync': {
      'real': '/clusters/{clusterName}?fields=Clusters/desired_configs',
      'mock': '/data/clusters/cluster.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'config_groups.all_fields': {
      'real': '/clusters/{clusterName}/config_groups?fields=*',
      'mock': ''
    },
    'config_groups.get_config_group_by_id': {
      'real': '/clusters/{clusterName}/config_groups/{id}',
      'mock': ''
    },
    'config_groups.update_config_group': {
      'real': '/clusters/{clusterName}/config_groups/{id}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          async: false,
          data: JSON.stringify(
            [
              data.configGroup
            ]
          )
        }
      }
    },
    'config_groups.delete_config_group': {
      'real': '/clusters/{clusterName}/config_groups/{id}',
      'mock': '',
      'type': 'DELETE'
    },
    'config.on_site': {
      'real': '/clusters/{clusterName}/configurations?{params}',
      'mock': '/data/configurations/cluster_level_configs.json?{params}',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'config.host_overrides': {
      'real': '/clusters/{clusterName}/configurations?{params}',
      'mock': '/data/configurations/host_level_overrides_configs.json?{params}',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'host.host_component.action': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          data: JSON.stringify({
            RequestInfo: {
              "context": data.context
            },
            Body: {
              "HostRoles": {
                "state": data.state
              }
            }
          })
        }
      }
    },
    'host.host_component.slave_desired_admin_state': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}/?fields=HostRoles/desired_admin_state',
      'mock': ''
    },
    'host.host_component.decommission_status': {
      'real': '/clusters/{clusterName}/services/{serviceName}/components/{componentName}/?fields=ServiceComponentInfo',
      'mock': ''
    },
    'host.host_component.decommission_status_datanode': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}?fields=metrics/dfs/namenode',
      'mock': ''
    },
    'host.host_component.decommission_slave': {
      'real' : '/clusters/{clusterName}/requests',
      'mock' : '',
      'format' : function(data) {
        return {
          type : 'POST',
          data : JSON.stringify({
            RequestInfo: {
              'context': data.context,
              'command': data.command,
              'parameters': {
                'slave_type': data.slaveType,
                'excluded_hosts': data.hostName
              }
            },
            "Requests/resource_filters": [{"service_name" : data.serviceName, "component_name" : data.componentName}]
          })
        }
      }
    },
    'host.host_component.recommission_and_restart': {
      'real': '/clusters/{clusterName}/request_schedules',
      'mock': '',
      'format' : function(data) {
        return {
          type : 'POST',
          data : JSON.stringify([ {
            "RequestSchedule" : {
              "batch" : [ {
                "requests" : data.batches
              }, {
                "batch_settings" : {
                  "batch_separation_in_seconds" : data.intervalTimeSeconds,
                  "task_failure_tolerance" : data.tolerateSize
                }
              } ]
            }
          } ])
        }
      }
    },
    'host.host_component.recommission_slave' : {
      'real' : '/clusters/{clusterName}/requests',
      'mock' : '',
      'format' : function(data) {
        return {
          type : 'POST',
          data : JSON.stringify({
            RequestInfo: {
              context: data.context,
              command: data.command,
              parameters: {
                slave_type: data.slaveType,
                included_hosts: data.hostName
              }
            },
            "Requests/resource_filters": [{"service_name" : data.serviceName, "component_name" : data.componentName}]
          })
        }
      }
    },
    'host.stale_host_components.start_stop': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components?HostRoles/stale_configs=true&' +
              'HostRoles/component_name.in({componentNames})',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          data: JSON.stringify({
            RequestInfo: {
              "context": data.context
            },
            Body: {
              "HostRoles": {
                "state": data.state
              }
            }
          })
        }
      }
    },
    'host.delete': {
      'real': '/clusters/{clusterName}/hosts/{hostName}',
      'mock': '',
      'type': 'DELETE',
      'format': function() {
        return {
          async: false
        }
      }
    },
    'hosts.metrics': {
      'real': '/clusters/{clusterName}/hosts?fields={metricName}',
      'mock': '/data/cluster_metrics/cpu_1hr.json'
    },
    'hosts.metrics.host_component': {
      'real': '/clusters/{clusterName}/services/{serviceName}/components/{componentName}?fields=host_components/{metricName}',
      'mock': '/data/cluster_metrics/cpu_1hr.json'
    },
    'service.service_component': {
      'real': '/clusters/{clusterName}/services/{serviceName}/components/{componentName}',
      'mock': '',
      'format': function(data) {
        return {
          async: data.async
        };
      }
    },
    'service.metrics.flume.channel_fill_percent': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/flume/flume/CHANNEL/*/ChannelFillPercentage[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/channelFillPct.json',
      'testInProduction': true
    },
    'service.metrics.flume.channel_size': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/flume/flume/CHANNEL/*/ChannelSize[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/channelSize.json',
      'testInProduction': true
    },
    'service.metrics.flume.sink_drain_success': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/flume/flume/SINK/*/EventDrainSuccessCount[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/sinkDrainSuccessCount.json',
      'testInProduction': true
    },
    'service.metrics.flume.sink_connection_failed': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/flume/flume/SINK/*/ConnectionFailedCount[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/sinkConnectionFailedCount.json',
      'testInProduction': true
    },
    'service.metrics.flume.gc': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/jvmGcTime.json',
      'testInProduction': true
    },
    'service.metrics.flume.jvm_heap_used': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/jvmMemHeapUsedM.json',
      'testInProduction': true
    },
    'service.metrics.flume.jvm_threads_runnable': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/jvmThreadsRunnable.json',
      'testInProduction': true
    },
    'service.metrics.flume.cpu_user': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/cpu/cpu_user[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '',
      'testInProduction': true
    },
    'service.metrics.flume.source_accepted': {
      'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_SERVER?fields=host_components/metrics/flume/flume/SOURCE/*/EventAcceptedCount[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/flume/sourceEventAccepted.json',
      'testInProduction': true
    },
    'service.metrics.hbase.cluster_requests': {
      'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_MASTER?fields=metrics/hbase/master/cluster_requests[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hbase/cluster_requests.json',
      'testInProduction': true
    },
    'service.metrics.hbase.hlog_split_size': {
      'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_MASTER?fields=metrics/hbase/master/splitSize_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hbase/hlog_split_size.json',
      'testInProduction': true
    },
    'service.metrics.hbase.hlog_split_time': {
      'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_MASTER?fields=metrics/hbase/master/splitTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hbase/hlog_split_time.json',
      'testInProduction': true
    },
    'service.metrics.hbase.regionserver_queuesize': {
      'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_REGIONSERVER?fields=metrics/hbase/regionserver/flushQueueSize[{fromSeconds},{toSeconds},{stepSeconds}],metrics/hbase/regionserver/compactionQueueSize[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hbase/regionserver_queuesize.json',
      'testInProduction': true
    },
    'service.metrics.hbase.regionserver_regions': {
      'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_REGIONSERVER?fields=metrics/hbase/regionserver/regions[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hbase/regionserver_regions.json',
      'testInProduction': true
    },
    'service.metrics.hbase.regionserver_rw_requests': {
      'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_REGIONSERVER?fields=metrics/hbase/regionserver/readRequestsCount[{fromSeconds},{toSeconds},{stepSeconds}],metrics/hbase/regionserver/writeRequestsCount[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hbase/regionserver_rw_requests.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.gc': {
      'real': '/clusters/{clusterName}/hosts/{jobTrackerNode}/host_components/JOBTRACKER?fields=metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/gc.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.jobs_status': {
      'real': '/clusters/{clusterName}/services/MAPREDUCE/components/JOBTRACKER?fields=metrics/mapred/jobtracker/jobs_completed[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/jobs_preparing[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/jobs_failed[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/jobs_submitted[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/jobs_failed[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/jobs_running[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/jobs_status.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.jobs_heap': {
      'real': '/clusters/{clusterName}/hosts/{jobTrackerNode}/host_components/JOBTRACKER?fields=metrics/jvm/memNonHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memNonHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/jvm_heap.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.jobs_threads': {
      'real': '/clusters/{clusterName}/hosts/{jobTrackerNode}/host_components/JOBTRACKER?fields=metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsBlocked[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsWaiting[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsTimedWaiting[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/jvm_threads.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.map_slots': {
      'real': '/clusters/{clusterName}/services/MAPREDUCE/components/JOBTRACKER?fields=metrics/mapred/jobtracker/occupied_map_slots[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/reserved_map_slots[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/map_slots.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.reduce_slots': {
      'real': '/clusters/{clusterName}/services/MAPREDUCE/components/JOBTRACKER?fields=metrics/mapred/jobtracker/occupied_reduce_slots[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/reserved_reduce_slots[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/reduce_slots.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.rpc': {
      'real': '/clusters/{clusterName}/hosts/{jobTrackerNode}/host_components/JOBTRACKER?fields=metrics/rpc/RpcQueueTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/rpc.json',
      'testInProduction': true
    },
    'service.metrics.mapreduce.tasks_running_waiting': {
      'real': '/clusters/{clusterName}/services/MAPREDUCE/components/JOBTRACKER?fields=metrics/mapred/jobtracker/running_maps[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/running_reduces[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/waiting_maps[{fromSeconds},{toSeconds},{stepSeconds}],metrics/mapred/jobtracker/waiting_reduces[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/mapreduce/tasks_running_waiting.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.block_status': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/dfs/FSNamesystem/PendingReplicationBlocks[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/UnderReplicatedBlocks[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/block_status.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.file_operations': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/dfs/namenode/FileInfoOps[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/namenode/CreateFileOps[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/file_operations.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.gc': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/gc.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.io': {
      'real': '/clusters/{clusterName}/services/HDFS/components/DATANODE?fields=metrics/dfs/datanode/bytes_written[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/datanode/bytes_read[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/io.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.jvm_heap': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/jvm/memNonHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memNonHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/jvm_heap.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.jvm_threads': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsBlocked[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsWaiting[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsTimedWaiting[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/jvm_threads.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.rpc': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/rpc/RpcQueueTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/rpc.json',
      'testInProduction': true
    },
    'service.metrics.hdfs.space_utilization': {
      'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/dfs/FSNamesystem/CapacityRemainingGB[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/CapacityUsedGB[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/CapacityTotalGB[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/hdfs/space_utilization.json',
      'testInProduction': true
    },
    'service.start_stop': {
      'real': '/clusters/{clusterName}/services?params/run_smoke_test=true',
      'mock': '/data/mirroring/poll/poll_6.json',
      'format': function (data) {
        return {
          type: 'PUT',
          async: false,
          data: data.data
        };
      }
    },
    'service.metrics.yarn.gc': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/yarn/gc.json',
      'testInProduction': true
    },
    'service.metrics.yarn.jobs_threads': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsBlocked[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsWaiting[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsTimedWaiting[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/yarn/jvm_threads.json',
      'testInProduction': true
    },
    'service.metrics.yarn.rpc': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/rpc/RpcQueueTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/yarn/rpc.json',
      'testInProduction': true
    },
    'service.metrics.yarn.jobs_heap': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/jvm/memNonHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memNonHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/services/metrics/yarn/jvm_heap.json',
      'testInProduction': true
    },
    'service.metrics.yarn.queue.allocated': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AvailableMB[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/PendingMB[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AllocatedMB[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '',
      'testInProduction': true
    },
    'service.metrics.yarn.queue.allocated.container': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AllocatedContainers[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/ReservedContainers[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/PendingContainers[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '',
      'testInProduction': true
    },
    'service.metrics.yarn.node.manager.statuses': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/ClusterMetrics/NumActiveNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumDecommissionedNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumLostNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumRebootedNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumUnhealthyNMs[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '',
      'testInProduction': true
    },
    'service.metrics.yarn.queue.memory.resource': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=',
      'mock': '',
      'format': function (data, opt) {
        var field1 = 'metrics/yarn/Queue/{queueName}/AllocatedMB[{fromSeconds},{toSeconds},{stepSeconds}]';
        var field2 = 'metrics/yarn/Queue/{queueName}/AvailableMB[{fromSeconds},{toSeconds},{stepSeconds}]';
        if (opt.url != null && data.queueNames != null && data.queueNames.length > 0) {
          data.queueNames.forEach(function (q) {
            data.queueName = q;
            opt.url += (formatUrl(field1, data) + ",");
            opt.url += (formatUrl(field2, data) + ",");
          });
        } else {
          opt.url += (formatUrl(field1, data) + ",");
          opt.url += (formatUrl(field2, data) + ",");
        }
      },
      'testInProduction': true
    },
    'service.metrics.yarn.queue.apps.states.current': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AppsPending[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsRunning[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '',
      'testInProduction': true
    },
    'service.metrics.yarn.queue.apps.states.finished': {
      'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AppsKilled[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsFailed[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsSubmitted[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsCompleted[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '',
      'testInProduction': true
    },
    'service.metrics.storm.nimbus': {
      'real': '/clusters/{clusterName}/services/STORM/components/NIMBUS?fields={metricsTemplate}',
      'mock': ''
    },
    'dashboard.cluster_metrics.cpu': {
      'real': '/clusters/{clusterName}/?fields=metrics/cpu[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/cluster_metrics/cpu_1hr.json',
      'testInProduction': true
    },
    'dashboard.cluster_metrics.load': {
      'real': '/clusters/{clusterName}/?fields=metrics/load[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/cluster_metrics/load_1hr.json',
      'testInProduction': true
    },
    'dashboard.cluster_metrics.memory': {
      'real': '/clusters/{clusterName}/?fields=metrics/memory[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/cluster_metrics/memory_1hr.json',
      'testInProduction': true
    },
    'dashboard.cluster_metrics.network': {
      'real': '/clusters/{clusterName}/?fields=metrics/network[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/cluster_metrics/network_1hr.json',
      'testInProduction': true
    },
    'host.metrics.cpu': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/cpu/cpu_user[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_wio[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_nice[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_aidle[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_system[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_idle[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/hosts/metrics/cpu.json',
      'testInProduction': true
    },
    'host.metrics.disk': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/disk/disk_total[{fromSeconds},{toSeconds},{stepSeconds}],metrics/disk/disk_free[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/hosts/metrics/disk.json',
      'testInProduction': true
    },
    'host.metrics.load': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/load/load_fifteen[{fromSeconds},{toSeconds},{stepSeconds}],metrics/load/load_one[{fromSeconds},{toSeconds},{stepSeconds}],metrics/load/load_five[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/hosts/metrics/load.json',
      'testInProduction': true
    },
    'host.metrics.memory': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/memory/swap_free[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_shared[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_free[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_cached[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_buffers[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/hosts/metrics/memory.json',
      'testInProduction': true
    },
    'host.metrics.network': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/network/bytes_in[{fromSeconds},{toSeconds},{stepSeconds}],metrics/network/bytes_out[{fromSeconds},{toSeconds},{stepSeconds}],metrics/network/pkts_in[{fromSeconds},{toSeconds},{stepSeconds}],metrics/network/pkts_out[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/hosts/metrics/network.json',
      'testInProduction': true
    },
    'host.metrics.processes': {
      'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/process/proc_total[{fromSeconds},{toSeconds},{stepSeconds}],metrics/process/proc_run[{fromSeconds},{toSeconds},{stepSeconds}]',
      'mock': '/data/hosts/metrics/processes.json',
      'testInProduction': true
    },
    'host.service_config_hosts_overrides': {
      'real': '/clusters/{clusterName}/configurations?{urlParams}',
      'mock': '',
      'format': function() {
        return {
          async: false,
          timeout: 10000
        };
      }
    },
    'admin.security_status': {
      'real': '/clusters/{clusterName}?fields=Clusters/desired_configs',
      'mock': '',
      'format': function() {
        return {
          timeout: 10000
        };
      }
    },
    'settings.get.user_pref': {
      'real': '/persist/{key}',
      'mock': '/data/user_settings/{key}.json',
      'format': function(data) {
        return {
          async: data.async
        };
      }
    },
    'settings.post.user_pref': {
      'real': '/persist',
      'mock': '',
      'type': 'POST',
      'format': function (data) {
        return {
          async: data.async,
          data: JSON.stringify(data.keyValuePair)
        }
      }
    },
    'cluster.load_cluster_name': {
      'real': '/clusters',
      'mock': '/data/clusters/info.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'cluster.state': {
      'type': 'POST',
      'real': '/persist/',
      'mock': '',
      'format': function(data) {
        return {
          async: false,
          data: JSON.stringify(data.keyValuePair)
        };
      }
    },
    'cluster.update_upgrade_version': {
      'real': '/stacks2/HDP/versions?fields=stackServices/StackServices,Versions',
      'mock': '/data/wizard/stack/stacks.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'cluster.load_repositories': {
      'real': '/stacks2/{stackName}/versions/{stackVersion}/operatingSystems?fields=repositories/*',
      'mock': '',
      'type': 'GET',
      'format': function (data) {
        return {
          data: data.data
        };
      }
    },
    'admin.high_availability.stop_all_services': {
      'real': '/clusters/{clusterName}/services',
      'mock': '',
      'format': function() {
        return {
          type: 'PUT',
          data: JSON.stringify({
            "RequestInfo": {
              "context": "Stop all services"
            },
            "Body": {
              "ServiceInfo": {
                "state": "INSTALLED"
              }
            }
          })
        }
      }
    },
    'admin.high_availability.start_all_services': {
      'real': '/clusters/{clusterName}/services',
      'mock': '',
      'format': function() {
        return {
          type: 'PUT',
          data: JSON.stringify({
            "RequestInfo": {
              "context": "Start all services"
            },
            "Body": {
              "ServiceInfo": {
                "state": "STARTED"
              }
            }
          })
        }
      }
    },
    'admin.high_availability.polling': {
      'real': '/clusters/{clusterName}/requests/{requestId}?fields=tasks/*',
      'mock': '',
      'type': 'GET'
    },
    'admin.high_availability.getNnCheckPointStatus': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/NAMENODE',
      'mock': '',
      'type': 'GET'
    },
    'admin.high_availability.getJnCheckPointStatus': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/JOURNALNODE?fields=metrics',
      'mock': ''
    },
    'admin.high_availability.getHostComponent': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': ''
    },
    'admin.high_availability.create_component': {
      'real': '/clusters/{clusterName}/hosts?Hosts/host_name={hostName}',
      'mock': '',
      'type': 'POST',
      'format': function (data) {
        return {
          data: JSON.stringify({
            "host_components": [
              {
                "HostRoles": {
                  "component_name": data.componentName
                }
              }
            ]
          })
        }
      }
    },
    'admin.high_availability.create_journalnode': {
      'real': '/clusters/{clusterName}/services?ServiceInfo/service_name=HDFS',
      'mock': '',
      'type': 'POST',
      'format': function() {
        return {
          data: JSON.stringify({
            "components": [
              {
                "ServiceComponentInfo": {
                  "component_name": "JOURNALNODE"
                }
              }
            ]
          })
        }
      }
    },
    'admin.high_availability.create_zkfc': {
      'real': '/clusters/{clusterName}/services?ServiceInfo/service_name=HDFS',
      'mock': '',
      'type': 'POST',
      'format': function() {
        return {
          data: JSON.stringify({
            "components": [
              {
                "ServiceComponentInfo": {
                  "component_name": "ZKFC"
                }
              }
            ]
          })
        }
      }
    },
    'admin.high_availability.install_component': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          data: JSON.stringify({
            RequestInfo: {
              "context": "Install " + data.displayName
            },
            Body: {
              "HostRoles": {
                "state": "INSTALLED"
              }
            }
          })
        }
      }
    },
    'admin.high_availability.start_component': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          data: JSON.stringify({
            RequestInfo: {
              "context": "Start " + data.displayName
            },
            Body: {
              "HostRoles": {
                "state": "STARTED"
              }
            }
          })
        }
      }
    },
    'admin.high_availability.maintenance_mode': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function () {
        return {
          data: JSON.stringify({
            "HostRoles": {
              "state": "DISABLED"
            }
          })
        }
      }
    },
    'admin.high_availability.stop_component': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          data: JSON.stringify({
            RequestInfo: {
              "context": "Stop " + data.displayName
            },
            Body: {
              "HostRoles": {
                "state": "INSTALLED"
              }
            }
          })
        }
      }
    },
    'admin.high_availability.load_configs': {
      'real': '/clusters/{clusterName}/configurations?(type=core-site&tag={coreSiteTag})|(type=hdfs-site&tag={hdfsSiteTag})',
      'mock': '',
      'type': 'GET'
    },
    'admin.high_availability.save_configs': {
      'real': '/clusters/{clusterName}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          async: false,
          data: JSON.stringify({
            Clusters: {
              desired_config: {
                "type": data.siteName,
                "tag": 'version' + (new Date).getTime(),
                "properties": data.properties
              }
            }
          })
        }
      }
    },
    'admin.high_availability.load_hbase_configs': {
      'real': '/clusters/{clusterName}/configurations?type=hbase-site&tag={hbaseSiteTag}',
      'mock': '',
      'type': 'GET'
    },
    'admin.delete_component': {
      'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
      'mock': '',
      'type': 'DELETE'
    },
    'admin.security.cluster_configs': {
      'real': '/clusters/{clusterName}',
      'mock': '',
      'format': function() {
        return {
          timeout: 10000
        };
      }
    },
    'admin.delete_host': {
      'real': '/clusters/{clusterName}/hosts/{hostName}',
      'mock': '',
      'type': 'DELETE'
    },
    'admin.get.all_configurations': {
      'real': '/clusters/{clusterName}/configurations?{urlParams}',
      'mock': '',
      'format': function() {
        return {
          timeout: 10000
        };
      }
    },
    'admin.security.apply_configurations': {
      'real': '/clusters/{clusterName}',
      'mock': '',
      'format': function (data) {
        return {
          type: 'PUT',
          timeout: 10000,
          data: data.configData
        };
      }
    },
    'admin.security.apply_configuration': {
      'real': '/clusters/{clusterName}',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          async: false,
          timeout: 5000,
          data: JSON.stringify(data.clusterData)
        };
      }
    },
    'admin.security.add.cluster_configs': {
      'real': '/clusters/{clusterName}' + '?fields=Clusters/desired_configs',
      'mock': '',
      'format': function() {
        return {
          timeout: 10000
        };
      }
    },
    'admin.stack_upgrade.run_upgrade': {
      'real': '/clusters/{clusterName}',
      'mock': '',
      'format': function (data) {
        return {
          type: 'PUT',
          async: false,
          data: data.data
        };
      }
    },
    'admin.stack_upgrade.stop_services': {
      'real': '/clusters/{clusterName}/services?ServiceInfo/state=STARTED',
      'mock': '',
      'format': function (data) {
        return {
          type: 'PUT',
          async: false,
          data: data.data
        };
      }
    },
    'admin.stack_upgrade.do_poll': {
      'real': '/clusters/{cluster}/requests/{requestId}?fields=tasks/*',
      'mock': '/data/wizard/{mock}'
    },
    'wizard.advanced_repositories.valid_url': {
      'real': '/stacks2/{stackName}/versions/{stackVersion}/operatingSystems/{osType}/repositories/{nameVersionCombo}',
      'mock': '',
      'type': 'PUT',
      'format': function (data) {
        return {
          async: true,
          data: JSON.stringify(data.data)
        }
      }
    },
    'wizard.install_services.add_host_controller.is_retry': {
      'real': '/clusters/{cluster}/host_components',
      'mock': '',
      'format': function (data) {
        return {
          type: 'PUT',
          async: false,
          data: data.data
        };
      }
    },
    'wizard.install_services.add_host_controller.not_is_retry': {
      'real': '/clusters/{cluster}/host_components',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          async: false,
          data: data.data
        };
      }
    },
    'wizard.install_services.installer_controller.is_retry': {
      'real': '/clusters/{cluster}/host_components?HostRoles/state=INSTALLED',
      'mock': '/data/wizard/deploy/2_hosts/poll_1.json',
      'type': 'PUT',
      'format': function (data) {
        return {
          async: false,
          data: data.data
        };
      }
    },
    'wizard.install_services.installer_controller.not_is_retry': {
      'real': '/clusters/{cluster}/services?ServiceInfo/state=INIT',
      'mock': '/data/wizard/deploy/2_hosts/poll_1.json',
      'type': 'PUT',
      'format': function (data) {
        return {
          async: false,
          data: data.data
        };
      }
    },
    'wizard.install_services.add_service_controller.get_failed_host_components': {
      'real': '/clusters/{clusterName}/host_components?fields=HostRoles/state,component/ServiceComponentInfo/service_name',
      'mock': '',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'wizard.service_components': {
      'real': '{stackUrl}/stackServices?fields=StackServices',
      'mock': '/data/wizard/stack/hdp/version/{stackVersion}.json',
      'format': function() {
        return {
          timeout: 10000,
          async: false
        };
      }
    },
    'wizard.step9.installer.get_host_status': {
      'real': '/clusters/{cluster}/hosts?fields=Hosts/host_state,host_components/HostRoles/state',
      'mock': '/data/wizard/deploy/5_hosts/get_host_state.json',
      'format': function () {
        return {
          async: false
        };
      }
    },
    'wizard.step9.installer.launch_start_services': {
      'real': '/clusters/{cluster}/services?ServiceInfo/state=INSTALLED&params/run_smoke_test=true&params/reconfigure_client=false',
      'mock': '/data/wizard/deploy/5_hosts/poll_6.json',
      'format': function (data) {
        var d = {
          type: 'PUT',
          async: false,
          data: data.data
        };
        if (App.testMode) {
          d.type = 'GET';
        }
        return d;
      }
    },
    'wizard.step9.add_host.launch_start_services': {
      'real': '/clusters/{cluster}/host_components',
      'mock': '/data/wizard/deploy/5_hosts/poll_6.json',
      'format': function (data) {
        return {
          type: 'PUT',
          async: false,
          data: data.data
        };
      }
    },
    'wizard.step8.delete_cluster': {
      'real': '/clusters/{name}',
      'mock': '',
      'format': function() {
        return {
          type: 'DELETE',
          async: false
        };
      }
    },
    'wizard.step8.existing_cluster_names': {
      'real': '/clusters',
      'mock': '',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'wizard.step3.host_info': {
      'real': '/hosts?fields=Hosts/total_mem,Hosts/cpu_count,Hosts/disk_info,Hosts/last_agent_env,Hosts/host_name,Hosts/os_type,Hosts/os_arch,Hosts/ip',
      'mock': '/data/wizard/bootstrap/two_hosts_information.json',
      'format': function() {
        return {
          contentType: 'application/json'
        };
      }
    },
    'wizard.step3.rerun_checks': {
      'real': '/hosts?fields=Hosts/last_agent_env',
      'mock': '/data/wizard/bootstrap/two_hosts_information.json',
      'format': function() {
        return {
          contentType: 'application/json'
        };
      }
    },
    'wizard.step3.bootstrap': {
      'real': '/bootstrap/{bootRequestId}',
      'mock': '/data/wizard/bootstrap/poll_{numPolls}.json'
    },
    'wizard.step3.is_hosts_registered': {
      'real': '/hosts',
      'mock': '/data/wizard/bootstrap/single_host_registration.json'
    },
    'wizard.stacks': {
      'real': '/stacks2',
      'mock': '/data/wizard/stack/stacks2.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'wizard.stacks_versions': {
      'real': '/stacks2/{stackName}/versions?fields=Versions,operatingSystems/repositories/Repositories',
      'mock': '/data/wizard/stack/{stackName}_versions.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'wizard.launch_bootstrap': {
      'real': '/bootstrap',
      'mock': '/data/wizard/bootstrap/bootstrap.json',
      'type': 'POST',
      'format': function (data) {
        return {
          async: false,
          contentType: 'application/json',
          data: data.bootStrapData
        }
      }
    },
    'router.login': {
      'real': '/users/{loginName}',
      'mock': '/data/users/user_{usr}.json',
      'format': function (data) {
        var statusCode = jQuery.extend({}, require('data/statusCodes'));
        statusCode['403'] = function () {
          console.log("Error code 403: Forbidden.");
        };
        return {
          statusCode: statusCode
        };
      }
    },
    'router.login2': {
      'real': '/clusters',
      'mock': '/data/clusters/info.json'
    },
    'router.logoff': {
      'real': '/logout',
      'mock': ''
    },
    'router.set_ambari_stacks': {
      'real': '/stacks',
      'mock': '/data/wizard/stack/stacks.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'router.authentication': {
      'real': '/clusters',
      'mock': '/data/clusters/info.json',
      'format': function() {
        return {
          async: false
        };
      }
    },
    'ambari.service.load_server_version': {
      'real': '/services/AMBARI/components/AMBARI_SERVER?fields=RootServiceComponents/component_version',
      'mock': ''
    },
    'ambari.service': {
      'real': '/services/AMBARI/components/AMBARI_SERVER',
      'mock': '/data/services/ambari_server.json'
    },
    'ambari.service.load_server_clock': {
      'real': '/services/AMBARI/components/AMBARI_SERVER?fields=RootServiceComponents/server_clock',
      'mock': ''
    },

    'config_groups.create': {
      'real': '/clusters/{clusterName}/config_groups',
      'mock': '',
      'format': function (data) {
        return {
          async: false,
          type: 'POST',
          data: JSON.stringify([{
            "ConfigGroup": {
              "group_name": data.group_name,
              "tag": data.service_id,
              "description": data.description,
              "desired_configs": data.desired_configs,
              "hosts": data.hosts
            }
          }])
        }
      }
    },
    'config_groups.update': {
      'real': '/clusters/{clusterName}/config_groups/{id}',
      'mock': '',
      'format': function (data) {
        return {
          async: true,
          type: 'PUT',
          data: JSON.stringify(data.data)
        }
      }
    },
    'rolling_restart.post': {
      'real': '/clusters/{clusterName}/request_schedules',
      'mock': '',
      'format' : function(data) {
        return {
          type : 'POST',
          data : JSON.stringify([ {
            "RequestSchedule" : {
              "batch" : [ {
                "requests" : data.batches
              }, {
                "batch_settings" : {
                  "batch_separation_in_seconds" : data.intervalTimeSeconds,
                  "task_failure_tolerance" : data.tolerateSize
                }
              } ]
            }
          } ])
        }
      }
    },
    'request_schedule.delete': {
      'real': '/clusters/{clusterName}/request_schedules/{request_schedule_id}',
      'mock': '',
      'format' : function() {
        return {
          type : 'DELETE'
        }
      }
    },
    'request_schedule.get': {
      'real': '/clusters/{clusterName}/request_schedules/{request_schedule_id}',
      'mock': ''
    },
    'restart.hostComponents': {
      'real':'/clusters/{clusterName}/requests',
      'mock':'',
      'format': function(data) {
        return {
          type : 'POST',
          data : JSON.stringify({
            "RequestInfo": {
              "command": "RESTART",
              "context": data.context
            },
            "Requests/resource_filters": data.resource_filters
          })
        }
      }
    },

    'mirroring.get_all_entities': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/list/{type}?fields=status&user.name=ambari-qa',
      'mock': '/data/mirroring/{type}s.xml',
      'apiPrefix': '',
      'format': function () {
        return {
          dataType: 'xml'
        }
      }
    },

    'mirroring.get_definition': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/definition/{type}/{name}?user.name=ambari-qa',
      'mock': '/data/mirroring/{name}_definition.xml',
      'apiPrefix': '',
      'format': function () {
        return {
          cache: true,
          dataType: 'xml'
        }
      }
    },

    'mirroring.dataset.get_all_instances': {
      'real': '/proxy?url=http://{falconServer}:15000/api/instance/status/feed/{dataset}?start={start}&end={end}&user.name=ambari-qa',
      'mock': '/data/mirroring/{dataset}_instances.json',
      'apiPrefix': ''
    },


    'mirroring.create_new_dataset': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/submitAndSchedule/feed?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST',
      'format': function (data) {
        return {
          contentType: 'text/xml',
          dataType: 'xml',
          data: data.dataset,
          headers: {
            'AmbariProxy-Content-Type': 'text/xml'
          }
        }
      }
    },

    'mirroring.submit_entity': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/submit/{type}?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST',
      'format': function (data) {
        return {
          contentType: 'text/xml',
          dataType: 'xml',
          data: data.entity,
          headers: {
            'AmbariProxy-Content-Type': 'text/xml'
          }
        }
      }
    },

    'mirroring.update_entity': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/update/{type}/{name}?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST',
      'format': function (data) {
        return {
          contentType: 'text/xml',
          dataType: 'xml',
          data: data.entity,
            headers: {
          'AmbariProxy-Content-Type': 'text/xml'
        }
        }
      }
    },

    'mirroring.delete_entity': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/delete/{type}/{name}?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'DELETE',
      'format': function () {
        return {
          dataType: 'xml'
        }
      }
    },

    'mirroring.suspend_entity': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/suspend/{type}/{name}?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST',
      'format': function (data) {
        return {
          dataType: 'xml',
          data: data.entity
        }
      }
    },

    'mirroring.resume_entity': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/resume/{type}/{name}?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST',
      'format': function () {
        return {
          dataType: 'xml'
        }
      }
    },

    'mirroring.schedule_entity': {
      'real': '/proxy?url=http://{falconServer}:15000/api/entities/schedule/{type}/{name}?user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST',
      'format': function () {
        return {
          dataType: 'xml'
        }
      }
    },

    'mirroring.suspend_instance': {
      'real': '/proxy?url=http://{falconServer}:15000/api/instance/suspend/feed/{feed}?start={name}&user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST'
    },

    'mirroring.resume_instance': {
      'real': '/proxy?url=http://{falconServer}:15000/api/instance/resume/feed/{feed}?start={name}&user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST'
    },

    'mirroring.kill_instance': {
      'real': '/proxy?url=http://{falconServer}:15000/api/instance/kill/feed/{feed}?start={name}&user.name=ambari-qa',
      'mock': '/data/mirroring/succeeded.json',
      'apiPrefix': '',
      'type': 'POST'
    },

    'bulk_request.host_components': {
      'real': '/clusters/{clusterName}/host_components',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              context: data.requestInfo,
              query: 'HostRoles/component_name=' + data.componentName + '&HostRoles/host_name.in(' + data.hostNames + ')&HostRoles/maintenance_state=OFF'
            },
            Body: {
              HostRoles: {
                state: data.state
              }
            }
          })
        }
      }
    },

    'bulk_request.hosts.all_components': {
      'real': '/clusters/{clusterName}/host_components',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              context: data.requestInfo,
              query: data.query
            },
            Body: {
              HostRoles: {
                state: data.state
              }
            }
          })
        }
      }
    },

    'bulk_request.decommission': {
      'real' : '/clusters/{clusterName}/requests',
      'mock' : '',
      'format': function(data) {
        return {
          type: 'POST',
          data: JSON.stringify({
            'RequestInfo': {
              'context': data.context,
              'command': 'DECOMMISSION',
              'parameters': data.parameters
            },
            "Requests/resource_filters": [{"service_name" : data.serviceName, "component_name" : data.componentName}]
          })
        }
      }
    },

    'bulk_request.hosts.passive_state': {
      'real': '/clusters/{clusterName}/hosts',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              context: data.requestInfo,
              query: 'Hosts/host_name.in(' + data.hostNames + ')'
            },
            Body: {
              Hosts: {
                maintenance_state: data.passive_state
              }
            }
          })
        }
      }
    },

    'bulk_request.hosts.all_components.passive_state': {
      'real': '/clusters/{clusterName}/host_components',
      'mock': '',
      'format': function(data) {
        return {
          type: 'PUT',
          data: JSON.stringify({
            RequestInfo: {
              context: data.requestInfo,
              query: data.query
            },
            Body: {
              HostRoles: {
                maintenance_state: data.passive_state
              }
            }
          })
        }
      }
    },

    'jobs.lastID': {
      'real': '/proxy?url=http://{historyServerHostName}:{ahsWebPort}/ws/v1/timeline/HIVE_QUERY_ID?limit=1&secondaryFilter=tez:true',
      'mock': 'data/jobs/hive-queries.json',
      'apiPrefix': ''
    },

    'jobs.tezDag.NametoID': {
      'real': '/proxy?url=http://{historyServerHostName}:{ahsWebPort}/ws/v1/timeline/TEZ_DAG_ID?primaryFilter=dagName:{tezDagName}',
      'mock': '/data/jobs/tezDag-name-to-id.json',
      'apiPrefix': ''
    },
    'jobs.tezDag.tezDagId': {
      'real': '/proxy?url=http://{historyServerHostName}:{ahsWebPort}/ws/v1/timeline/TEZ_DAG_ID/{tezDagId}?fields=relatedentities,otherinfo',
      'mock': '/data/jobs/tezDag.json',
      'apiPrefix': ''
    },
    'jobs.tezDag.tezDagVertexId': {
      'real': '/proxy?url=http://{historyServerHostName}:{ahsWebPort}/ws/v1/timeline/TEZ_VERTEX_ID/{tezDagVertexId}?fields=otherinfo',
      'mock': '/data/jobs/tezDagVertex.json',
      'apiPrefix': ''
    }
  };
  /**
   * Replace data-placeholders to its values
   *
   * @param {String} url
   * @param {Object} data
   * @return {String}
   */
  var formatUrl = function (url, data) {
    if (!url) return null;
    var keys = url.match(/\{\w+\}/g);
    keys = (keys === null) ? [] : keys;
    if (keys) {
      keys.forEach(function (key) {
        var raw_key = key.substr(1, key.length - 2);
        var replace;
        if (!data || !data[raw_key]) {
          replace = '';
        }
        else {
          replace = data[raw_key];
        }
        url = url.replace(new RegExp(key, 'g'), replace);
      });
    }
    return url;
  };

  /**
   * this = object from config
   * @return {Object}
   */
  var formatRequest = function (data) {
    var opt = {
      type: this.type || 'GET',
      timeout: App.timeout,
      dataType: 'json',
      statusCode: require('data/statusCodes')
    };
    if (App.testMode) {
      opt.url = formatUrl(this.mock ? this.mock : '', data);
      opt.type = 'GET';
    }
    else {
      var prefix = this.apiPrefix != null ? this.apiPrefix : App.apiPrefix;
      opt.url = prefix + formatUrl(this.real, data);
    }

    if (this.format) {
      jQuery.extend(opt, this.format(data, opt));
    }
    return opt;
  };

  /**
   * Wrapper for all ajax requests
   *
   * @type {Object}
   */
  var ajax = Em.Object.extend({
    /**
     * Send ajax request
     *
     * @param {Object} config
     * @return {$.ajax} jquery ajax object
     *
     * config fields:
     *  name - url-key in the urls-object *required*
     *  sender - object that send request (need for proper callback initialization) *required*
     *  data - object with data for url-format
     *  beforeSend - method-name for ajax beforeSend response callback
     *  success - method-name for ajax success response callback
     *  error - method-name for ajax error response callback
     *  callback - callback from <code>App.updater.run</code> library
     */
    send: function (config) {

      console.warn('============== ajax ==============', config.name, config.data);

      if (!config.sender) {
        console.warn('Ajax sender should be defined!');
        return null;
      }

      // default parameters
      var params = {
        clusterName: App.get('clusterName')
      };

      // extend default parameters with provided
      if (config.data) {
        jQuery.extend(params, config.data);
      }

      var opt = {};
      if (!urls[config.name]) {
        console.warn('Invalid name provided!');
        return null;
      }
      opt = formatRequest.call(urls[config.name], params);

      opt.context = this;

      // object sender should be provided for processing beforeSend, success and error responses
      opt.beforeSend = function (xhr) {
        if (config.beforeSend) {
          config.sender[config.beforeSend](opt, xhr, params);
        }
      };
      opt.success = function (data) {
        console.log("TRACE: The url is: " + opt.url);
        if (config.success) {
          config.sender[config.success](data, opt, params);
        }
      };
      opt.error = function (request, ajaxOptions, error) {
        if (config.error) {
          config.sender[config.error](request, ajaxOptions, error, opt, params);
        } else {
          this.defaultErrorHandler(request, opt.url, opt.type);
        }
      };
      opt.complete = function () {
        if (config.callback) {
          config.callback();
        }
      };
      if ($.mocho) {
        opt.url = 'http://' + $.hostName + opt.url;
      }
      return $.ajax(opt);
    },

    // A single instance of App.ModalPopup view
    modalPopup: null,
    /**
     * defaultErrorHandler function is referred from App.ajax.send function and App.HttpClient.defaultErrorHandler function
     * @jqXHR {jqXHR Object}
     * @url {string}
     * @method {String} Http method
     * @showStatus {number} HTTP response code which should be shown. Default is 500.
     */
    defaultErrorHandler: function (jqXHR, url, method, showStatus) {
      method = method || 'GET';
      var self = this;
      var api = " received on " + method + " method for API: " + url;
      try {
        var json = $.parseJSON(jqXHR.responseText);
        var message = json.message;
      } catch (err) {
      }
      if (!showStatus) {
        showStatus = 500;
      }
      var statusCode = jqXHR.status + " status code";
      if (jqXHR.status === showStatus && !this.modalPopup) {
        this.modalPopup = App.ModalPopup.show({
          header: jqXHR.statusText,
          secondary: false,
          onPrimary: function () {
            this.hide();
            self.modalPopup = null;
          },
          bodyClass: Ember.View.extend({
            classNames: ['api-error'],
            templateName: require('templates/utils/ajax'),
            api: api,
            statusCode: statusCode,
            message: message,
            showMessage: !!message
          })
        });
      }
    }

  });

  /**
   * Add few access-methods for test purposes
   */
  if ($.mocho) {
    ajax.reopen({
      /**
       * Don't use it anywhere except tests!
       * @returns {Array}
       */
      fakeGetUrlNames: function() {
        return Em.keys(urls);
      },

      /**
       * Don't use it anywhere except tests!
       * @param name
       * @returns {*}
       */
      fakeGetUrl: function(name) {
        return urls[name];
      },

      /**
       * Don't use it anywhere except tests!
       * @param url
       * @param data
       * @returns {String}
       */
      fakeFormatUrl: function(url, data) {
        return formatUrl(url, data);
      },

      /**
       * Don't use it anywhere except tests!
       * @param urlObj
       * @param data
       * @returns {Object}
       */
      fakeFormatRequest: function(urlObj, data) {
        return formatRequest.call(urlObj, data);
      }
    });
  }

  App.ajax = ajax.create({});
  App.formatRequest = formatRequest;
  App.urls = urls;
  
});
window.require.register("utils/base64", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  (function () {

    var
      object = typeof window != 'undefined' ? window : exports,
      chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=',
      INVALID_CHARACTER_ERR = (function () {
        // fabricate a suitable error object
        try {
          document.createElement('$');
        }
        catch (error) {
          return error;
        }
      }());

    // encoder
    // [https://gist.github.com/999166] by [https://github.com/nignag]
    object.btoa || (
      object.btoa = function (input) {
        for (
          // initialize result and counter
          var block, charCode, idx = 0, map = chars, output = '';
          // if the next input index does not exist:
          // change the mapping table to "="
          // check if d has no fractional digits
          input.charAt(idx | 0) || (map = '=', idx % 1);
          // "8 - idx % 1 * 8" generates the sequence 2, 4, 6, 8
          output += map.charAt(63 & block >> 8 - idx % 1 * 8)
          ) {
          charCode = input.charCodeAt(idx += 3 / 4);
          if (charCode > 0xFF) throw INVALID_CHARACTER_ERR;
          block = block << 8 | charCode;
        }
        return output;
      });

    // decoder
    // [https://gist.github.com/1020396] by [https://github.com/atk]
    object.atob || (
      object.atob = function (input) {
        input = input.replace(/=+$/, '');
        if (input.length % 4 == 1) throw INVALID_CHARACTER_ERR;
        for (
          // initialize result and counters
          var bc = 0, bs, buffer, idx = 0, output = '';
          // get next character
          buffer = input.charAt(idx++);
          // character found in table? initialize bit storage and add its ascii value;
          ~buffer && (bs = bc % 4 ? bs * 64 + buffer : buffer,
            // and if not first of each 4 characters,
            // convert the first 8 bits to one ascii character
            bc++ % 4) ? output += String.fromCharCode(255 & bs >> (-2 * bc & 6)) : 0
          ) {
          // try to find character in table (0-63, not found => -1)
          buffer = chars.indexOf(buffer);
        }
        return output;
      });

  }());
});
window.require.register("utils/batch_scheduled_requests", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * Default success callback for ajax-requests in this module
   * @type {Function}
   */
  var defaultSuccessCallback = function(data, ajaxOptions, params) {
    App.router.get('applicationController').dataLoading().done(function(initValue) {
      params.query && params.query.set('status', 'SUCCESS');
      if (initValue) {
        App.router.get('backgroundOperationsController').showPopup();
      }
    });
  };
  /**
   * Default error callback for ajax-requests in this module
   * @param {Object} xhr
   * @param {String} textStatus
   * @param {String} error
   * @param {Object} opt
   * @type {Function}
   */
  var defaultErrorCallback = function(xhr, textStatus, error, opt, params) {
    params.query && params.query.set('status', 'FAIL');
    App.ajax.defaultErrorHandler(xhr, opt.url, 'POST', xhr.status);
  };

  /**
   * Contains helpful utilities for handling batch and scheduled requests.
   */
  module.exports = {

    /**
     * Some services have components which have a need for rolling restarts. This
     * method returns the name of the host-component which supports rolling
     * restarts for a service.
     * @param {String} serviceName
     */
    getRollingRestartComponentName: function(serviceName) {
      var rollingRestartComponents = {
        HDFS: 'DATANODE',
        YARN: 'NODEMANAGER',
        MAPREDUCE: 'TASKTRACKER',
        HBASE: 'HBASE_REGIONSERVER',
        STORM: 'SUPERVISOR'
      };
      return rollingRestartComponents[serviceName] ? rollingRestartComponents[serviceName] : null;
    },

    /**
     * Facade-function for restarting host components of specific service
     * @param {String} serviceName for which service hostComponents should be restarted
     * @param {bool} staleConfigsOnly restart only hostComponents with <code>staleConfig</code> true
     */
    restartAllServiceHostComponents: function(serviceName, staleConfigsOnly, query) {
      var service = App.Service.find(serviceName);
      var context = staleConfigsOnly ? Em.I18n.t('rollingrestart.context.allWithStaleConfigsForSelectedService').format(serviceName) : Em.I18n.t('rollingrestart.context.allForSelectedService').format(serviceName);
      if (service) {
        var hostComponents = service.get('hostComponents').filterProperty('host.passiveState','OFF');

        // HCatalog components are technically owned by Hive.
        if (serviceName == 'HIVE') {
          var hcatService = App.Service.find('HCATALOG');
          if (hcatService != null && hcatService.get('isLoaded')) {
            var hcatHcs = hcatService.get('hostComponents').filterProperty('host.passiveState', 'OFF');
            if (hcatHcs != null) {
              hostComponents.pushObjects(hcatHcs);
            }
          }
        }

        if (staleConfigsOnly) {
          hostComponents = hostComponents.filterProperty('staleConfigs', true);
        }
        this.restartHostComponents(hostComponents, context, query);
      }
    },

    /**
     * Restart list of host components
     * @param {Ember.Enumerable} hostComponentsList list of host components should be restarted
     * @param {String} context message to show in BG popup
     */
    restartHostComponents: function(hostComponentsList, context, query) {
      context = context || Em.I18n.t('rollingrestart.context.default');
      /**
       * Format: {
       *  'DATANODE': ['host1', 'host2'],
       *  'NAMENODE': ['host1', 'host3']
       *  ...
       * }
       */
      var componentToHostsMap = {};
      var componentServiceMap = App.QuickDataMapper.componentServiceMap();
      hostComponentsList.forEach(function(hc) {
        var componentName = hc.get('componentName');
        if (!componentToHostsMap[componentName]) {
          componentToHostsMap[componentName] = [];
        }
        componentToHostsMap[componentName].push(hc.get('host.hostName'));
      });
      var resource_filters = [];
      for (var componentName in componentToHostsMap) {
        if (componentToHostsMap.hasOwnProperty(componentName)) {
          resource_filters.push({
            service_name:  componentServiceMap[componentName],
            component_name: componentName,
            hosts: componentToHostsMap[componentName].join(",")
          });
        }
      }
      if (resource_filters.length) {
        App.ajax.send({
          name: 'restart.hostComponents',
          sender: {
            successCallback: defaultSuccessCallback,
            errorCallback: defaultErrorCallback
          },
          data: {
            context: context,
            resource_filters: resource_filters,
            query: query
          },
          success: 'successCallback',
          error: 'errorCallback'
        });
      }
    },

    /**
     * Makes a REST call to the server requesting the rolling restart of the
     * provided host components.
     * @param {Array} restartHostComponents list of host components should be restarted
     * @param {Number} batchSize size of each batch
     * @param {Number} intervalTimeSeconds delay between two batches
     * @param {Number} tolerateSize task failure tolerance
     * @param {callback} successCallback
     * @param {callback} errorCallback
     */
    _doPostBatchRollingRestartRequest: function(restartHostComponents, batchSize, intervalTimeSeconds, tolerateSize, successCallback, errorCallback) {
      successCallback = successCallback || defaultSuccessCallback;
      errorCallback = errorCallback || defaultErrorCallback;
      if (!restartHostComponents.length) {
        console.log('No batch rolling restart if no restartHostComponents provided!');
        return;
      }
      App.ajax.send({
        name: 'rolling_restart.post',
        sender: {
          successCallback: successCallback,
          errorCallback: errorCallback
        },
        data: {
          intervalTimeSeconds: intervalTimeSeconds,
          tolerateSize: tolerateSize,
          batches: this.getBatchesForRollingRestartRequest(restartHostComponents, batchSize)
        },
        success: 'successCallback',
        error: 'errorCallback'
      });
    },

    /**
     * Create list of batches for rolling restart request
     * @param {Array} restartHostComponents list host components should be restarted
     * @param {Number} batchSize size of each batch
     * @returns {Array} list of batches
     */
    getBatchesForRollingRestartRequest: function(restartHostComponents, batchSize) {
      var hostIndex = 0,
          batches = [],
          batchCount = Math.ceil(restartHostComponents.length / batchSize),
          sampleHostComponent = restartHostComponents.objectAt(0),
          componentName = sampleHostComponent.get('componentName'),
          serviceName = sampleHostComponent.get('service.serviceName');

      for ( var count = 0; count < batchCount; count++) {
        var hostNames = [];
        for ( var hc = 0; hc < batchSize && hostIndex < restartHostComponents.length; hc++) {
          hostNames.push(restartHostComponents.objectAt(hostIndex++).get('host.hostName'));
        }
        if (hostNames.length > 0) {
          batches.push({
            "order_id" : count + 1,
            "type" : "POST",
            "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
            "RequestBodyInfo" : {
              "RequestInfo" : {
                "context" : "_PARSE_.ROLLING-RESTART." + componentName + "." + (count + 1) + "." + batchCount,
                "command" : "RESTART"
              },
              "Requests/resource_filters": [{
                "service_name" : serviceName,
                "component_name" : componentName,
                "hosts" : hostNames.join(",")
              }]
            }
          });
        }
      }
      return batches;
    },

    /**
     * Launches dialog to handle rolling restarts of host components.
     *
     * Rolling restart is supported only for components listed in <code>getRollingRestartComponentName</code>
     * @see getRollingRestartComponentName
     * @param {String} hostComponentName
     *           Type of host-component to restart across cluster
     *          (ex: DATANODE)
     * @param {bool} staleConfigsOnly
     *           Pre-select host-components which have stale
     *          configurations
     */
    launchHostComponentRollingRestart: function(hostComponentName, serviceName, isMaintenanceModeOn, staleConfigsOnly, skipMaintenance) {
      if (App.get('components.rollinRestartAllowed').contains(hostComponentName)) {
        this.showRollingRestartPopup(hostComponentName, serviceName, isMaintenanceModeOn, staleConfigsOnly, null, skipMaintenance);
      }
      else {
        this.showWarningRollingRestartPopup(hostComponentName);
      }
    },

    /**
     * Show popup with rolling restart dialog
     * @param {String} hostComponentName name of the host components that should be restarted
     * @param {bool} staleConfigsOnly restart only components with <code>staleConfigs</code> = true
     * @param {App.hostComponent[]} hostComponents list of hostComponents that should be restarted (optional).
     * Using this parameter will reset hostComponentName
     */
    showRollingRestartPopup: function(hostComponentName, serviceName, isMaintenanceModeOn, staleConfigsOnly, hostComponents, skipMaintenance) {
      hostComponents = hostComponents || [];
      var componentDisplayName = App.format.role(hostComponentName);
      if (!componentDisplayName) {
        componentDisplayName = hostComponentName;
      }
      var title = Em.I18n.t('rollingrestart.dialog.title').format(componentDisplayName);
      var viewExtend = {
        staleConfigsOnly : staleConfigsOnly,
        hostComponentName : hostComponentName,
        skipMaintenance: skipMaintenance,
        serviceName: serviceName,
        isServiceInMM: isMaintenanceModeOn,
        didInsertElement : function() {
          this.set('parentView.innerView', this);
          this.initialize();
        }
      };
      if (hostComponents.length) {
        viewExtend.allHostComponents = hostComponents;
      }

      var self = this;
      App.ModalPopup.show({
        header : title,
        hostComponentName : hostComponentName,
        serviceName: serviceName,
        isServiceInMM: isMaintenanceModeOn,
        staleConfigsOnly : staleConfigsOnly,
        skipMaintenance: skipMaintenance,
        innerView : null,
        bodyClass : App.RollingRestartView.extend(viewExtend),
        classNames : [ 'rolling-restart-popup' ],
        primary : Em.I18n.t('rollingrestart.dialog.primary'),
        onPrimary : function() {
          var dialog = this;
          var restartComponents = this.get('innerView.restartHostComponents');
          var batchSize = this.get('innerView.batchSize');
          var waitTime = this.get('innerView.interBatchWaitTimeSeconds');
          var tolerateSize = this.get('innerView.tolerateSize');
          self._doPostBatchRollingRestartRequest(restartComponents, batchSize, waitTime, tolerateSize, function(data, ajaxOptions, params) {
            dialog.hide();
            defaultSuccessCallback(data, ajaxOptions, params);
          });
        },
        updateButtons : function() {
          var errors = this.get('innerView.errors');
          this.set('disablePrimary', (errors != null && errors.length > 0))
        }.observes('innerView.errors')
      });
    },

    /**
     * Show warning popup about not supported host components
     * @param {String} hostComponentName
     */
    showWarningRollingRestartPopup: function(hostComponentName) {
      var componentDisplayName = App.format.role(hostComponentName);
      if (!componentDisplayName) {
        componentDisplayName = hostComponentName;
      }
      var title = Em.I18n.t('rollingrestart.dialog.title').format(componentDisplayName);
      var msg = Em.I18n.t('rollingrestart.notsupported.hostComponent').format(componentDisplayName);
      console.log(msg);
      App.ModalPopup.show({
        header : title,
        secondary : false,
        msg : msg,
        bodyClass : Em.View.extend({
          template : Em.Handlebars.compile('<div class="alert alert-warning">{{msg}}</div>')
        })
      });
    },

    /**
     * Warn user that alerts will be updated in few minutes
     * @param {String} hostComponentName
     */
    infoPassiveState: function(passiveState) {
      var enabled = passiveState == 'OFF' ? 'enabled' : 'suppressed';
      App.ModalPopup.show({
        header: Em.I18n.t('common.information'),
        secondary: null,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<p>{{view.message}}</p>'),
          message: function() {
            return Em.I18n.t('hostPopup.warning.alertsTimeOut').format(passiveState.toLowerCase(), enabled);
          }.property()
        })
      });
    },

    /**
     * Retrieves the latest information about a specific request schedule
     * identified by 'requestScheduleId'
     *
     * @param {Number} requestScheduleId ID of the request schedule to get
     * @param {Function} successCallback Called with request_schedule data from server. An
     *          empty object returned for invalid ID.
     * @param {Function} errorCallback Optional error callback. Default behavior is to
     *          popup default error dialog.
     */
    getRequestSchedule: function(requestScheduleId, successCallback, errorCallback) {
      if (requestScheduleId != null && !isNaN(requestScheduleId) && requestScheduleId > -1) {
        errorCallback = errorCallback ? errorCallback : defaultErrorCallback;
        App.ajax.send({
          name : 'request_schedule.get',
          sender : {
            successCallbackFunction : function(data) {
              successCallback(data);
            },
            errorCallbackFunction : function(xhr, textStatus, error, opt) {
              errorCallback(xhr, textStatus, error, opt);
            }
          },
          data : {
            request_schedule_id : requestScheduleId
          },
          success : 'successCallbackFunction',
          error : 'errorCallbackFunction'
        });
      } else {
        successCallback({});
      }
    },

    /**
     * Attempts to abort a specific request schedule identified by 'requestScheduleId'
     *
     * @param {Number} requestScheduleId ID of the request schedule to get
     * @param {Function} successCallback Called when request schedule successfully aborted
     * @param {Function} errorCallback Optional error callback. Default behavior is to
     *          popup default error dialog.
     */
    doAbortRequestSchedule: function(requestScheduleId, successCallback, errorCallback) {
      if (requestScheduleId != null && !isNaN(requestScheduleId) && requestScheduleId > -1) {
        errorCallback = errorCallback || defaultErrorCallback;
        App.ajax.send({
          name : 'request_schedule.delete',
          sender : {
            successCallbackFunction : function(data) {
              successCallback(data);
            },
            errorCallbackFunction : function(xhr, textStatus, error, opt) {
              errorCallback(xhr, textStatus, error, opt);
            }
          },
          data : {
            request_schedule_id : requestScheduleId
          },
          success : 'successCallbackFunction',
          error : 'errorCallbackFunction'
        });
      } else {
        successCallback({});
      }
    }
  };
});
window.require.register("utils/component", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Here will be stored slave functions related to components
   * @type {Object}
   */
  module.exports = {

    /**
     * Return list of installed components. Syntax is:
     *
     * [{
     *    id : 'DATANODE',
     *    displayName : 'DataNode',
     *    isMaster : true,
     *    isSlave : false,
     *    isClient : false
     * }]
     *
     */
    getInstalledComponents : function(){
      var components = App.HostComponent.find();
      var names = components.mapProperty('componentName').uniq();
      var result = [];

      names.forEach(function(componentName){
        var component = components.findProperty('componentName', componentName);
        result.push(Ember.Object.create({
          id: componentName,
          isMaster: component.get('isMaster'),
          isSlave: component.get('isSlave'),
          isClient: component.get('isClient'),
          displayName: component.get('displayName'),
          serviceName: component.get('service.id')
        }));
      });

      return result;
    }
  };
  
});
window.require.register("utils/config", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  var categotyConfigs = require('data/service_configs');
  var configGroupsByTag = [];

  App.config = Em.Object.create({
    /**
     * XML characters which should be escaped in values
     * http://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references#Predefined_entities_in_XML
     */
    xmlEscapeMap: {
      "&": "&amp;",
      "<": "&lt;",
      ">": "&gt;",
      '"': '&quot;',
      "'": '&apos;'
    },
    xmlUnEscapeMap: {
      "&amp;": "&",
      "&lt;": "<",
      "&gt;": ">",
      "&quot;": '"',
      "&apos;": "'"
    },
    
    CONFIG_GROUP_NAME_MAX_LENGTH: 18,

    /**
     * filename exceptions used to support substandard sitenames which don't have "xml" extension
     */
    filenameExceptions: ['zoo.cfg'],

    /**
     * Since values end up in XML files (core-sit.xml, etc.), certain
     * XML sensitive characters should be escaped. If not we will have
     * an invalid XML document, and services will fail to start. 
     * 
     * Special characters in XML are defined at
     * http://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references#Predefined_entities_in_XML
     */
    escapeXMLCharacters: function(value, toXML) {
      var self = this;
      // To prevent double/triple replacing '&gt;' to '&amp;gt;' to '&amp;amp;gt;', we need
      // to first unescape all XML chars, and then escape them again.
      var newValue = String(value).replace(/(&amp;|&lt;|&gt;|&quot;|&apos;)/g, function (s) {
        return self.xmlUnEscapeMap[s];
      });
      if (toXML) {
        return String(newValue).replace(/[&<>"']/g, function (s) {
          return self.xmlEscapeMap[s];
        });
      } else {
        return newValue;
      }
    },
    preDefinedServiceConfigs: function () {
      var configs = this.get('preDefinedGlobalProperties');
      var services = [];
      $.extend(true, [], require('data/service_configs')).forEach(function (service) {
        service.configs = configs.filterProperty('serviceName', service.serviceName);
        services.push(service);
      });
      return services;
    }.property('preDefinedGlobalProperties'),
    configMapping: function () {
      if (App.get('isHadoop2Stack')) {
        return $.extend(true, [], require('data/HDP2/config_mapping'));
      }
      return $.extend(true, [], require('data/config_mapping'));
    }.property('App.isHadoop2Stack'),
    preDefinedGlobalProperties: function () {
      if (App.get('isHadoop2Stack')) {
        return $.extend(true, [], require('data/HDP2/global_properties').configProperties);
      }
      return $.extend(true, [], require('data/global_properties').configProperties);
    }.property('App.isHadoop2Stack'),
    preDefinedSiteProperties: function () {
      if (App.get('isHadoop2Stack')) {
        return $.extend(true, [], require('data/HDP2/site_properties').configProperties);
      }
      return $.extend(true, [], require('data/site_properties').configProperties);
    }.property('App.isHadoop2Stack'),
    preDefinedCustomConfigs: function () {
      if (App.get('isHadoop2Stack')) {
        return $.extend(true, [], require('data/HDP2/custom_configs'));
      }
      return $.extend(true, [], require('data/custom_configs'));
    }.property('App.isHadoop2Stack'),
    //categories which contain custom configs
    categoriesWithCustom: ['CapacityScheduler'],
    //configs with these filenames go to appropriate category not in Advanced
    customFileNames: function () {
      if (App.supports.capacitySchedulerUi) {
        if (App.get('isHadoop2Stack')) {
          return ['capacity-scheduler.xml'];
        }
        return ['capacity-scheduler.xml', 'mapred-queue-acls.xml'];
      } else {
        return [];
      }
    }.property('App.isHadoop2Stack'),

    /**
     * Function should be used post-install as precondition check should not be done only after installer wizard
     * @param siteNames
     * @returns {Array}
     */
    getBySitename: function (siteNames) {
      var computedConfigs = this.get('configMapping').computed();
      var siteProperties = [];
      if (typeof siteNames === "string") {
        siteProperties = computedConfigs.filterProperty('filename', siteNames);
      } else if (siteNames instanceof Array) {
        siteNames.forEach(function (_siteName) {
          siteProperties = siteProperties.concat(computedConfigs.filterProperty('filename', _siteName));
        }, this);
      }
      return siteProperties;
    },


    /**
     * Cache of loaded configurations. This is useful in not loading
     * same configuration multiple times. It is populated in multiple
     * places.
     *
     * Example:
     * {
     *  'global_version1': {...},
     *  'global_version2': {...},
     *  'hdfs-site_version3': {...},
     * }
     */
    loadedConfigurationsCache: {},

    /**
     * Array of global "service/desired_tag/actual_tag" strings which
     * indicate different configurations. We cache these so that
     * we dont have to recalculate if two tags are difference.
     */
    differentGlobalTagsCache: [],

    identifyCategory: function (config) {
      var category = null;
      var serviceConfigMetaData = this.get('preDefinedServiceConfigs').findProperty('serviceName', config.serviceName);
      if (serviceConfigMetaData) {
        serviceConfigMetaData.configCategories.forEach(function (_category) {
          if (_category.siteFileNames && Array.isArray(_category.siteFileNames) && _category.siteFileNames.contains(config.filename)) {
            category = _category;
          }
        });
        category = (category == null) ? serviceConfigMetaData.configCategories.findProperty('siteFileName', config.filename) : category;
      }
      return category;
    },
    /**
     * additional handling for special properties such as
     * checkbox and digital which values with 'm' at the end
     * @param config
     */
    handleSpecialProperties: function (config) {
      if (config.displayType === 'int' && /\d+m$/.test(config.value)) {
        config.value = config.value.slice(0, config.value.length - 1);
        config.defaultValue = config.value;
      }
      if (config.displayType === 'checkbox') {
        config.value = (config.value === 'true') ? config.defaultValue = true : config.defaultValue = false;
      }
    },
    /**
     * calculate config properties:
     * category, filename, isRequired, isUserProperty
     * @param config
     * @param isAdvanced
     * @param advancedConfigs
     */
    calculateConfigProperties: function (config, isAdvanced, advancedConfigs) {
      if (!isAdvanced || this.get('customFileNames').contains(config.filename)) {
        var categoryMetaData = this.identifyCategory(config);
        if (categoryMetaData != null) {
          config.category = categoryMetaData.get('name');
          if (!isAdvanced) config.isUserProperty = true;
        }
      } else {
        config.category = config.category ? config.category : 'Advanced';
        config.description = isAdvanced && advancedConfigs.findProperty('name', config.name).description;
        config.isRequired = true;
      }
    },
    capacitySchedulerFilter: function () {
      var yarnRegex = /^yarn\.scheduler\.capacity\.root\.(?!unfunded)([a-z]([\_\-a-z0-9]{0,50}))\.(acl_administer_jobs|acl_submit_jobs|state|user-limit-factor|maximum-capacity|capacity)$/i;
      var self = this;
      if (App.get('isHadoop2Stack')) {
        return function (_config) {
          return (yarnRegex.test(_config.name));
        }
      } else {
        return function (_config) {
          return (_config.name.indexOf('mapred.capacity-scheduler.queue.') !== -1) ||
            (/^mapred\.queue\.[a-z]([\_\-a-z0-9]{0,50})\.(acl-administer-jobs|acl-submit-job)$/i.test(_config.name));
        }
      }
    }.property('App.isHadoop2Stack'),
    /**
     * return:
     *   configs,
     *   globalConfigs,
     *   mappingConfigs
     *
     * @param configGroups
     * @param advancedConfigs
     * @param tags
     * @param serviceName
     * @return {object}
     */
    mergePreDefinedWithLoaded: function (configGroups, advancedConfigs, tags, serviceName) {
      var configs = [];
      var globalConfigs = [];
      var preDefinedConfigs = this.get('preDefinedGlobalProperties').concat(this.get('preDefinedSiteProperties'));
      var mappingConfigs = [];
      var filenameExceptions = this.get('filenameExceptions');
      var selectedServiceNames = App.Service.find().mapProperty('serviceName');
      tags.forEach(function (_tag) {
        var isAdvanced = null;
        var filename = (filenameExceptions.contains(_tag.siteName)) ? _tag.siteName : _tag.siteName + ".xml";
        var properties = configGroups.filter(function (serviceConfigProperties) {
          return _tag.tagName === serviceConfigProperties.tag && _tag.siteName === serviceConfigProperties.type;
        });

        properties = (properties.length) ? properties.objectAt(0).properties : {};
        for (var index in properties) {
          var configsPropertyDef =  null;
          var preDefinedConfig = [];
          if (_tag.siteName === 'global') {
          // Unlike other site where one site maps to ones service, global site contains configurations for multiple services
          // So Global Configuration should not be filtered out with serviceName.
            preDefinedConfig = preDefinedConfigs.filterProperty('name', index);
            preDefinedConfig.forEach(function(_preDefinedConfig){
              var isServiceInstalled = selectedServiceNames.contains(_preDefinedConfig.serviceName);
                if ( isServiceInstalled || _preDefinedConfig.serviceName === 'MISC') {
                  configsPropertyDef = _preDefinedConfig;
                }
            },this);
          } else {
            configsPropertyDef = preDefinedConfigs.filterProperty('name',index).findProperty('filename',filename);
            if (!configsPropertyDef) {
              configsPropertyDef = preDefinedConfigs.filterProperty('name',index).findProperty('serviceName', serviceName);
            }
          }

          var serviceConfigObj = App.ServiceConfig.create({
            name: index,
            value: properties[index],
            defaultValue: properties[index],
            filename: filename,
            isUserProperty: false,
            isOverridable: true,
            showLabel: true,
            serviceName: serviceName,
            belongsToService: []
          });

          if (configsPropertyDef) {
            this.setServiceConfigUiAttributes(serviceConfigObj, configsPropertyDef);
          }
          if (_tag.siteName === 'global') {
            if (configsPropertyDef) {
              if (configsPropertyDef.isRequiredByAgent === false) {
                continue;
              }
              this.handleSpecialProperties(serviceConfigObj);
            } else {
              serviceConfigObj.isVisible = false;  // if the global property is not defined on ui metadata global_properties.js then it shouldn't be a part of errorCount
            }
            serviceConfigObj.id = 'puppet var';
            serviceConfigObj.displayName = configsPropertyDef ? configsPropertyDef.displayName : null;
            serviceConfigObj.options = configsPropertyDef ? configsPropertyDef.options : null;
            globalConfigs.push(serviceConfigObj);
          } else if (!this.getBySitename(serviceConfigObj.get('filename')).someProperty('name', index)) {
            isAdvanced = advancedConfigs.someProperty('name', index);
            serviceConfigObj.id = 'site property';
            if (!configsPropertyDef) {
              serviceConfigObj.displayType = stringUtils.isSingleLine(serviceConfigObj.value) ? 'advanced' : 'multiLine';
            }
            serviceConfigObj.displayName = configsPropertyDef ? configsPropertyDef.displayName : index;
            this.calculateConfigProperties(serviceConfigObj, isAdvanced, advancedConfigs);
            if(serviceConfigObj.get('displayType') == 'directories'
              && (serviceConfigObj.get('category') == 'DataNode'
              || serviceConfigObj.get('category') == 'NameNode')) {
              var dirs = serviceConfigObj.get('value').split(',').sort();
              serviceConfigObj.set('value', dirs.join(','));
              serviceConfigObj.set('defaultValue', dirs.join(','));
            }
            if(serviceConfigObj.get('displayType') == 'directory'
              && serviceConfigObj.get('category') == 'SNameNode') {
              var dirs = serviceConfigObj.get('value').split(',').sort();
              serviceConfigObj.set('value', dirs[0]);
              serviceConfigObj.set('defaultValue', dirs[0]);
            }
            if (serviceConfigObj.get('displayType') == 'masterHosts') {
              if (typeof(serviceConfigObj.get('value')) == 'string') {
                var value = serviceConfigObj.get('value').replace(/\[|]|'|&apos;/g, "").split(',');
                serviceConfigObj.set('value', value);
              }
            }
            configs.push(serviceConfigObj);
          } else {
            mappingConfigs.push(serviceConfigObj);
          }
        }
      }, this);
      return {
        configs: configs,
        globalConfigs: globalConfigs,
        mappingConfigs: mappingConfigs
      }
    },

    /**
     * @param serviceConfigObj : Object
     * @param configsPropertyDef : Object
     */
    setServiceConfigUiAttributes: function (serviceConfigObj, configsPropertyDef) {
      serviceConfigObj.displayType = configsPropertyDef.displayType;
      serviceConfigObj.isRequired = (configsPropertyDef.isRequired !== undefined) ? configsPropertyDef.isRequired : true;
      serviceConfigObj.isRequiredByAgent = (configsPropertyDef.isRequiredByAgent !== undefined) ? configsPropertyDef.isRequiredByAgent : true;
      serviceConfigObj.isReconfigurable = (configsPropertyDef.isReconfigurable !== undefined) ? configsPropertyDef.isReconfigurable : true;
      serviceConfigObj.isVisible = (configsPropertyDef.isVisible !== undefined) ? configsPropertyDef.isVisible : true;
      serviceConfigObj.unit = (configsPropertyDef.unit !== undefined) ? configsPropertyDef.unit : undefined;
      serviceConfigObj.description = (configsPropertyDef.description !== undefined) ? configsPropertyDef.description : undefined;
      serviceConfigObj.isOverridable = configsPropertyDef.isOverridable === undefined ? true : configsPropertyDef.isOverridable;
      serviceConfigObj.serviceName = configsPropertyDef ? configsPropertyDef.serviceName : null;
      serviceConfigObj.index = configsPropertyDef.index;
      serviceConfigObj.isSecureConfig = configsPropertyDef.isSecureConfig === undefined ? false : configsPropertyDef.isSecureConfig;
      serviceConfigObj.belongsToService = configsPropertyDef.belongsToService;
      serviceConfigObj.category = configsPropertyDef.category;
      serviceConfigObj.showLabel = configsPropertyDef.showLabel !== false;
    },

    /**
     * synchronize order of config properties with order, that on UI side
     * @param configSet
     * @return {Object}
     */
    syncOrderWithPredefined: function (configSet) {
      var globalConfigs = configSet.globalConfigs,
        siteConfigs = configSet.configs,
        globalStart = [],
        siteStart = [];

      this.get('preDefinedGlobalProperties').mapProperty('name').forEach(function (name) {
        var _global = globalConfigs.findProperty('name', name);
        if (_global) {
          globalStart.push(_global);
          globalConfigs = globalConfigs.without(_global);
        }
      }, this);

      this.get('preDefinedSiteProperties').mapProperty('name').forEach(function (name) {
        var _site = siteConfigs.filterProperty('name', name);
        if (_site.length == 1) {
          siteStart.push(_site[0]);
          siteConfigs = siteConfigs.without(_site[0]);
        } else if (_site.length >1) {
          _site.forEach(function(site){
            siteStart.push(site);
            siteConfigs = siteConfigs.without(site);
          }, this);
        }
      }, this);

      return {
        globalConfigs: globalStart.concat(globalConfigs.sortProperty('name')),
        configs: siteStart.concat(siteConfigs.sortProperty('name')),
        mappingConfigs: configSet.mappingConfigs
      }
    },

    /**
     * merge stored configs with pre-defined
     * @param storedConfigs
     * @param advancedConfigs
     * @return {*}
     */
    mergePreDefinedWithStored: function (storedConfigs, advancedConfigs, selectedServiceNames) {
      var mergedConfigs = [];
      var preDefinedConfigs = $.extend(true, [], this.get('preDefinedGlobalProperties').concat(this.get('preDefinedSiteProperties')));

      storedConfigs = (storedConfigs) ? storedConfigs : [];

      var preDefinedNames = preDefinedConfigs.mapProperty('name');
      var storedNames = storedConfigs.mapProperty('name');
      var names = preDefinedNames.concat(storedNames).uniq();
      names.forEach(function (name) {
        var storedCfgs = storedConfigs.filterProperty('name', name);
        var preDefinedCfgs = [];
        var preDefinedConfig = preDefinedConfigs.filterProperty('name', name);
        preDefinedConfig.forEach(function (_preDefinedConfig) {
          if (selectedServiceNames.contains(_preDefinedConfig.serviceName) || _preDefinedConfig.serviceName === 'MISC') {
            preDefinedCfgs.push(_preDefinedConfig);
          }
        }, this);

        var configData = {};
        var isAdvanced = advancedConfigs.someProperty('name', name);
        if (storedCfgs.length <= 1 && preDefinedCfgs.length <= 1) {
          var stored = storedCfgs[0];
          var preDefined = preDefinedCfgs[0];
          if (preDefined && stored) {
            configData = preDefined;
            configData.value = stored.value;
            configData.defaultValue = stored.defaultValue;
            configData.overrides = stored.overrides;
            configData.filename = stored.filename;
            configData.description = stored.description;
            configData.isRequiredByAgent = (configData.isRequiredByAgent !== undefined) ? configData.isRequiredByAgent : true;
            configData.showLabel = stored.showLabel !== false;
          } else if (!preDefined && stored) {

            configData = {
              id: stored.id,
              name: stored.name,
              displayName: stored.name,
              serviceName: stored.serviceName,
              value: stored.value,
              defaultValue: stored.defaultValue,
              displayType: stringUtils.isSingleLine(stored.value) ? 'advanced' : 'multiLine',
              filename: stored.filename,
              category: 'Advanced',
              isUserProperty: stored.isUserProperty === true,
              isOverridable: true,
              overrides: stored.overrides,
              isRequired: true,
              showLabel: stored.showLabel !== false
            };
            this.calculateConfigProperties(configData, isAdvanced, advancedConfigs);
          } else if (preDefined && !stored) {
            configData = preDefined;
            configData.isRequiredByAgent = (configData.isRequiredByAgent !== undefined) ? configData.isRequiredByAgent : true;
            if (isAdvanced) {
              var advanced = advancedConfigs.findProperty('name', configData.name);
              this.setPropertyFromStack(configData,advanced);
            }
          }
          if (configData.displayType === 'checkbox') {
            configData.value = configData.value === 'true'; // convert {String} value to {Boolean}
            configData.defaultValue = configData.value;
          }
          mergedConfigs.push(configData);
        } else {
          preDefinedCfgs.forEach(function (cfg) {
            configData = cfg;
            configData.isRequiredByAgent = (configData.isRequiredByAgent !== undefined) ? configData.isRequiredByAgent : true;
            var storedCfg = storedCfgs.findProperty('filename', cfg.filename);
            if (storedCfg) {
              configData.value = storedCfg.value;
              configData.defaultValue = storedCfg.defaultValue;
              configData.overrides = storedCfg.overrides;
              configData.filename = storedCfg.filename;
              configData.description = storedCfg.description;
              configData.description = storedCfg.showLabel !== false;
            } else if (isAdvanced){
                advanced = advancedConfigs.filterProperty('filename', configData.filename).findProperty('name', configData.name);
                this.setPropertyFromStack(configData,advanced);
            }
            mergedConfigs.push(configData);
          }, this);
        }
      }, this);
      return mergedConfigs;
    },

    /**
     *
     * @param configData {Object} Configs that will be binded to the view on step-7 of installer wizard
     * @param advanced {Object} Config property loaded from Server side stack definition
     */
    setPropertyFromStack: function(configData,advanced) {

      // Password fields should be made blank by default in installer wizard
      // irrespective of whatever value is sent from stack definition.
      // This forces the user to fill the password field.
      configData.value = configData.displayType == "password" ? '' : advanced ? advanced.value : configData.value;
      configData.defaultValue = configData.value;
      configData.filename = advanced ? advanced.filename : configData.filename;
      configData.description = advanced ? advanced.description : configData.description;
    },


    /**
     * look over advanced configs and add missing configs to serviceConfigs
     * filter fetched configs by service if passed
     * @param serviceConfigs
     * @param advancedConfigs
     * @param serviceName
     */
    addAdvancedConfigs: function (serviceConfigs, advancedConfigs, serviceName) {
      var configsToVerifying = (serviceName) ? serviceConfigs.filterProperty('serviceName', serviceName) : serviceConfigs.slice();
      advancedConfigs.forEach(function (_config) {
        var configCategory = 'Advanced';
        var categoryMetaData = null;
        if (_config) {
          if (this.get('configMapping').computed().someProperty('name', _config.name)) {
          } else if (!(configsToVerifying.someProperty('name', _config.name))) {
            if (this.get('customFileNames').contains(_config.filename)) {
              categoryMetaData = this.identifyCategory(_config);
              if (categoryMetaData != null) {
                configCategory = categoryMetaData.get('name');
              }
            }
            _config.id = "site property";
            _config.category = configCategory;
            _config.displayName = _config.name;
            _config.defaultValue = _config.value;
            // make all advanced configs optional and populated by default
            /*
             * if (/\${.*}/.test(_config.value) || (service.serviceName !==
             * 'OOZIE' && service.serviceName !== 'HBASE')) { _config.isRequired =
             * false; _config.value = ''; } else if
             * (/^\s+$/.test(_config.value)) { _config.isRequired = false; }
             */
            _config.isRequired = true;
            _config.displayType = stringUtils.isSingleLine(_config.value) ? 'advanced' : 'multiLine';
            serviceConfigs.push(_config);
          }
        }
      }, this);
    },
    /**
     * Render a custom conf-site box for entering properties that will be written in *-site.xml files of the services
     */
    addCustomConfigs: function (configs) {
      var preDefinedCustomConfigs = $.extend(true, [], this.get('preDefinedCustomConfigs'));
      var stored = configs.filter(function (_config) {
        return this.get('categoriesWithCustom').contains(_config.category);
      }, this);
      if (App.supports.capacitySchedulerUi) {
        var queueProperties = stored.filter(this.get('capacitySchedulerFilter'));
        if (queueProperties.length) {
          queueProperties.setEach('isQueue', true);
        }
      }
    },

    miscConfigVisibleProperty: function (configs, serviceToShow) {
      configs.forEach(function (item) {
        item.set("isVisible", item.belongsToService.some(function (cur) {
          return serviceToShow.contains(cur)
        }));
      });
      return configs;
    },

    /**
     * render configs, distribute them by service
     * and wrap each in ServiceConfigProperty object
     * @param configs
     * @param storedConfigs
     * @param allSelectedServiceNames
     * @param installedServiceNames
     * @param localDB
     * @return {Array}
     */
    renderConfigs: function (configs, storedConfigs, allSelectedServiceNames, installedServiceNames, localDB) {
      var renderedServiceConfigs = [];
      var services = [];

      this.get('preDefinedServiceConfigs').forEach(function (serviceConfig) {
        if (allSelectedServiceNames.contains(serviceConfig.serviceName) || serviceConfig.serviceName === 'MISC') {
          console.log('pushing ' + serviceConfig.serviceName, serviceConfig);
          if (!installedServiceNames.contains(serviceConfig.serviceName) || serviceConfig.serviceName === 'MISC') {
            serviceConfig.showConfig = true;
          }
          services.push(serviceConfig);
        }
      });
      services.forEach(function (service) {
        var configsByService = [];
        var serviceConfigs = configs.filterProperty('serviceName', service.serviceName);
        serviceConfigs.forEach(function (_config) {
          _config.isOverridable = (_config.isOverridable === undefined) ? true : _config.isOverridable;
          var serviceConfigProperty = App.ServiceConfigProperty.create(_config);
          this.updateHostOverrides(serviceConfigProperty, _config);
          if (!storedConfigs) {
            serviceConfigProperty.initialValue(localDB);
          }
          this.tweakDynamicDefaults(localDB, serviceConfigProperty, _config);
          serviceConfigProperty.validate();
          configsByService.pushObject(serviceConfigProperty);
        }, this);
        var serviceConfig = this.createServiceConfig(service.serviceName);
        serviceConfig.set('showConfig', service.showConfig);

        // Use calculated default values for some configs
        var recommendedDefaults = {};
        if (!storedConfigs && service.defaultsProviders) {
          service.defaultsProviders.forEach(function (defaultsProvider) {
            var defaults = defaultsProvider.getDefaults(localDB);
            for (var name in defaults) {
              var config = configsByService.findProperty('name', name);
              if (!config) {
                continue;
              }
              if (!!defaults[name]) {
                recommendedDefaults[name] = defaults[name];
                config.set('value', defaults[name]);
                config.set('defaultValue', defaults[name]);
              } else {
                recommendedDefaults[name] = config.get('defaultValue');
              }
            }
          });
        }
        if (service.configsValidator) {
          service.configsValidator.set('recommendedDefaults', recommendedDefaults);
          var validators = service.configsValidator.get('configValidators');
          for (var validatorName in validators) {
            var c = configsByService.findProperty('name', validatorName);
            if (c) {
              c.set('serviceValidator', service.configsValidator);
            }
          }
        }

        serviceConfig.set('configs', configsByService);
        renderedServiceConfigs.push(serviceConfig);
      }, this);
      return renderedServiceConfigs;
    },
    /**
     Takes care of the "dynamic defaults" for the GLUSTERFS configs.  Sets
     some of the config defaults to previously user-entered data.
     **/
    tweakDynamicDefaults: function (localDB, serviceConfigProperty, config) {
      var firstHost = null;
      for (var host in localDB.hosts) {
        firstHost = host;
        break;
      }
      try {
        if (typeof(config.defaultValue) == "string" && config.defaultValue.indexOf("{firstHost}") >= 0) {
          serviceConfigProperty.set('value', serviceConfigProperty.value.replace(new RegExp("{firstHost}"), firstHost));
          serviceConfigProperty.set('defaultValue', serviceConfigProperty.defaultValue.replace(new RegExp("{firstHost}"), firstHost));
        }
      } catch (err) {
        // Nothing to worry about here, most likely trying indexOf on a non-string
      }
    },
    /**
     * create new child configs from overrides, attach them to parent config
     * override - value of config, related to particular host(s)
     * @param configProperty
     * @param storedConfigProperty
     */
    updateHostOverrides: function (configProperty, storedConfigProperty) {
      if (storedConfigProperty.overrides != null && storedConfigProperty.overrides.length > 0) {
        var overrides = [];
        storedConfigProperty.overrides.forEach(function (overrideEntry) {
          // create new override with new value
          var newSCP = App.ServiceConfigProperty.create(configProperty);
          newSCP.set('value', overrideEntry.value);
          newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
          newSCP.set('parentSCP', configProperty);
          var hostsArray = Ember.A([]);
          overrideEntry.hosts.forEach(function (host) {
            hostsArray.push(host);
          });
          newSCP.set('selectedHostOptions', hostsArray);
          overrides.pushObject(newSCP);
        });
        configProperty.set('overrides', overrides);
      }
    },
    /**
     * create new ServiceConfig object by service name
     * @param serviceName
     */
    createServiceConfig: function (serviceName) {
      var preDefinedServiceConfig = App.config.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName);
      var serviceConfig = App.ServiceConfig.create({
        filename: preDefinedServiceConfig.filename,
        serviceName: preDefinedServiceConfig.serviceName,
        displayName: preDefinedServiceConfig.displayName,
        configCategories: preDefinedServiceConfig.configCategories,
        configs: [],
        configGroups: []
      });
      serviceConfig.configCategories.filterProperty('isCustomView', true).forEach(function (category) {
        switch (category.name) {
          case 'CapacityScheduler':
            if (App.supports.capacitySchedulerUi) {
              category.set('customView', App.ServiceConfigCapacityScheduler);
            } else {
              category.set('isCustomView', false);
            }
            break;
        }
      }, this);
      return serviceConfig;
    },
    /**
     * GETs all cluster level sites in one call.
     *
     * @return Array of all site configs
     */
    loadConfigsByTags: function (tags) {
      var urlParams = [];
      tags.forEach(function (_tag) {
        urlParams.push('(type=' + _tag.siteName + '&tag=' + _tag.tagName + ')');
      });
      var params = urlParams.join('|');
      App.ajax.send({
        name: 'config.on_site',
        sender: this,
        data: {
          params: params
        },
        success: 'loadConfigsByTagsSuccess'
      });
      return configGroupsByTag;
    },

    loadConfigsByTagsSuccess: function (data) {
      if (data.items) {
        configGroupsByTag = [];
        data.items.forEach(function (item) {
          this.loadedConfigurationsCache[item.type + "_" + item.tag] = item.properties;
          configGroupsByTag.push(item);
        }, this);
      }
    },
    /**
     * Generate serviceProperties save it to localDB
     * called form stepController step6WizardController
     *
     * @param serviceName
     * @return {*}
     */
    loadAdvancedConfig: function (serviceName, callback, sync) {
      App.ajax.send({
        name: 'config.advanced',
        sender: this,
        data: {
          serviceName: serviceName,
          stack2VersionUrl: App.get('stack2VersionURL'),
          stackVersion: App.get('currentStackVersionNumber'),
          callback: callback,
          sync: sync
        },
        success: 'loadAdvancedConfigSuccess',
        error: 'loadAdvancedConfigError'
      });
    },

    loadAdvancedConfigSuccess: function (data, opt, params) {
      console.log("TRACE: In success function for the loadAdvancedConfig; url is ", opt.url);
      var properties = [];
      if (data.items.length) {
        data.items.forEach(function (item) {
          item = item.StackConfigurations;
          item.isVisible = item.type !== 'global.xml';
          var serviceName = item.service_name;
          var fileName = item.type;
          var isHDP2 = App.get('isHadoop2Stack');
          /**
           * Properties from mapred-queue-acls.xml are ignored unless App.supports.capacitySchedulerUi is true
           * Properties from capacity-scheduler.xml are ignored unless HDP stack version is 2.x or
           * HDP stack version is 1.x and App.supports.capacitySchedulerUi is true.
            */
          if ((fileName !== 'mapred-queue-acls.xml' || App.supports.capacitySchedulerUi) &&
              (fileName !== 'capacity-scheduler.xml' || isHDP2 || App.supports.capacitySchedulerUi)) {
            properties.push({
              serviceName: serviceName,
              name: item.property_name,
              value: item.property_value,
              description: item.property_description,
              isVisible: item.isVisible,
              filename: item.filename || fileName
            });
          }
        }, this);
      }
      params.callback(properties);
    },

    loadAdvancedConfigError: function (request, ajaxOptions, error, opt, params) {
      console.log('ERROR: failed to load stack configs for', params.serviceName);
      params.callback([]);
    },

    /**
     * Get properties from server by type and tag with properties, that belong to group
     * push them to common {serviceConfigs}
     */
    loadServiceConfigGroupOverrides: function (serviceConfigs, loadedGroupToOverrideSiteToTagMap, configGroups) {
      var configKeyToConfigMap = {};
      serviceConfigs.forEach(function (item) {
        configKeyToConfigMap[item.name] = item;
      });
      var typeTagToGroupMap = {};
      var urlParams = [];
      for (var group in loadedGroupToOverrideSiteToTagMap) {
        var overrideTypeTags = loadedGroupToOverrideSiteToTagMap[group];
        for (var type in overrideTypeTags) {
          var tag = overrideTypeTags[type];
          typeTagToGroupMap[type + "///" + tag] = configGroups.findProperty('name', group);
          urlParams.push('(type=' + type + '&tag=' + tag + ')');
        }
      }
      var params = urlParams.join('|');
      if (urlParams.length) {
        App.ajax.send({
          name: 'config.host_overrides',
          sender: this,
          data: {
            params: params,
            configKeyToConfigMap: configKeyToConfigMap,
            typeTagToGroupMap: typeTagToGroupMap
          },
          success: 'loadServiceConfigGroupOverridesSuccess'
        });
      }
    },
    loadServiceConfigGroupOverridesSuccess: function (data, opt, params) {
      data.items.forEach(function (config) {
        App.config.loadedConfigurationsCache[config.type + "_" + config.tag] = config.properties;
        var group = params.typeTagToGroupMap[config.type + "///" + config.tag];
        var properties = config.properties;
        for (var prop in properties) {
          var serviceConfig = params.configKeyToConfigMap[prop];
          var hostOverrideValue = properties[prop];
          if (serviceConfig && serviceConfig.displayType === 'int') {
            if (/\d+m$/.test(hostOverrideValue)) {
              hostOverrideValue = hostOverrideValue.slice(0, hostOverrideValue.length - 1);
            }
          } else if (serviceConfig && serviceConfig.displayType === 'checkbox') {
            switch (hostOverrideValue) {
              case 'true':
                hostOverrideValue = true;
                break;
              case 'false':
                hostOverrideValue = false;
                break;
            }
          }
          if (serviceConfig) {
            // Value of this property is different for this host.
            var overrides = 'overrides';
            if (!(overrides in serviceConfig)) {
              serviceConfig.overrides = [];
            }
            if (!serviceConfig.overrides) {
             serviceConfig.set('overrides', []);
            }
            console.log("loadServiceConfigGroupOverridesSuccess(): [" + group + "] OVERRODE(" + serviceConfig.name + "): " + serviceConfig.value + " -> " + hostOverrideValue);
            serviceConfig.overrides.push({value: hostOverrideValue, group: group});
          }
        }
      });
    },

    /**
     * Set all site property that are derived from other site-properties
     */
    setConfigValue: function (mappedConfigs, allConfigs, config, globalConfigs) {
      var globalValue;
      if (config.value == null) {
        return;
      }
      var fkValue = config.value.match(/<(foreignKey.*?)>/g);
      var fkName = config.name.match(/<(foreignKey.*?)>/g);
      var templateValue = config.value.match(/<(templateName.*?)>/g);
      if (fkValue) {
        fkValue.forEach(function (_fkValue) {
          var index = parseInt(_fkValue.match(/\[([\d]*)(?=\])/)[1]);
          if (mappedConfigs.someProperty('name', config.foreignKey[index])) {
            globalValue = mappedConfigs.findProperty('name', config.foreignKey[index]).value;
            config.value = config.value.replace(_fkValue, globalValue);
          } else if (allConfigs.someProperty('name', config.foreignKey[index])) {
            if (allConfigs.findProperty('name', config.foreignKey[index]).value === '') {
              globalValue = allConfigs.findProperty('name', config.foreignKey[index]).defaultValue;
            } else {
              globalValue = allConfigs.findProperty('name', config.foreignKey[index]).value;
            }
            config.value = config.value.replace(_fkValue, globalValue);
          }
        }, this);
      }

      // config._name - formatted name from original config name
      if (fkName) {
        fkName.forEach(function (_fkName) {
          var index = parseInt(_fkName.match(/\[([\d]*)(?=\])/)[1]);
          if (mappedConfigs.someProperty('name', config.foreignKey[index])) {
            globalValue = mappedConfigs.findProperty('name', config.foreignKey[index]).value;
            config._name = config.name.replace(_fkName, globalValue);
          } else if (allConfigs.someProperty('name', config.foreignKey[index])) {
            if (allConfigs.findProperty('name', config.foreignKey[index]).value === '') {
              globalValue = allConfigs.findProperty('name', config.foreignKey[index]).defaultValue;
            } else {
              globalValue = allConfigs.findProperty('name', config.foreignKey[index]).value;
            }
            config._name = config.name.replace(_fkName, globalValue);
          }
        }, this);
      }

      //For properties in the configMapping file having foreignKey and templateName properties.
      if (templateValue) {
        templateValue.forEach(function (_value) {
          var index = parseInt(_value.match(/\[([\d]*)(?=\])/)[1]);
          if (globalConfigs.someProperty('name', config.templateName[index])) {
            var globalValue = globalConfigs.findProperty('name', config.templateName[index]).value;
            config.value = config.value.replace(_value, globalValue);
          } else {
            config.value = null;
          }
        }, this);
      }
    },
    complexConfigs: [
      {
        "id": "site property",
        "name": "capacity-scheduler",
        "displayName": "Capacity Scheduler",
        "value": "",
        "defaultValue": "",
        "description": "Capacity Scheduler properties",
        "displayType": "custom",
        "isOverridable": true,
        "isRequired": true,
        "isVisible": true,
        "serviceName": "YARN",
        "filename": "capacity-scheduler.xml",
        "category": "CapacityScheduler"
      }
    ],

    /**
     * transform set of configs from file
     * into one config with textarea content:
     * name=value
     * @param configs
     * @param filename
     * @return {*}
     */
    fileConfigsIntoTextarea: function (configs, filename) {
      var fileConfigs = configs.filterProperty('filename', filename);
      var value = '';
      var defaultValue = '';
      var complexConfig = this.get('complexConfigs').findProperty('filename', filename);
      if (complexConfig) {
        fileConfigs.forEach(function (_config) {
          value += _config.name + '=' + _config.value + '\n';
          defaultValue += _config.name + '=' + _config.defaultValue + '\n';
        }, this);
        complexConfig.value = value;
        complexConfig.defaultValue = defaultValue;
        configs = configs.filter(function (_config) {
          return _config.filename !== filename;
        });
        configs.push(complexConfig);
      }
      return configs;
    },

    /**
     * transform one config with textarea content
     * into set of configs of file
     * @param configs
     * @param filename
     * @return {*}
     */
    textareaIntoFileConfigs: function (configs, filename) {
      var complexConfigName = this.get('complexConfigs').findProperty('filename', filename).name;
      var configsTextarea = configs.findProperty('name', complexConfigName);
      if (configsTextarea) {
        var properties = configsTextarea.get('value').split('\n');

        properties.forEach(function (_property) {
          var name, value;
          if (_property) {
            _property = _property.split('=');
            name = _property[0];
            value = (_property[1]) ? _property[1] : "";
            configs.push(Em.Object.create({
              id: configsTextarea.get('id'),
              name: name,
              value: value,
              defaultValue: value,
              serviceName: configsTextarea.get('serviceName'),
              filename: filename
            }));
          }
        });
        return configs.without(configsTextarea);
      }
      console.log('ERROR: textarea config - ' + complexConfigName + ' is missing');
      return configs;
    },

    /**
     * trim trailing spaces for all properties.
     * trim both trailing and leading spaces for host displayType and hive/oozie datebases url.
     * for directory or directories displayType format string for further using.
     * for password and values with spaces only do nothing.
     * @param {Object} property
     * @param {Boolean} isEmberObject
     * @returns {*}
     */
    trimProperty: function (property, isEmberObject) {
      var displayType = (isEmberObject) ? property.get('displayType') : property.displayType;
      var value = (isEmberObject) ? property.get('value') : property.value;
      var name = (isEmberObject) ? property.get('name') : property.name;
      var rez;
      switch (displayType) {
        case 'directories':
        case 'directory':
          rez = value.trim().split(/\s+/g).join(',');
          break;
        case 'host':
          rez = value.trim();
          break;
        case 'password':
          break;
        case 'advanced':
          if (name == 'javax.jdo.option.ConnectionURL' || name == 'oozie.service.JPAService.jdbc.url') {
            rez = value.trim();
          }
        default:
          rez = (typeof value == 'string') ? value.replace(/(\s+$)/g, '') : value;
      }
      return ((rez == '') || (rez == undefined)) ? value : rez;
    },

    OnNnHAHideSnn: function (ServiceConfig) {
      var configCategories = ServiceConfig.get('configCategories');
      var snCategory = configCategories.findProperty('name', 'SNameNode');
      var isSnnPresent = !!App.HDFSService.find('HDFS').get('snameNode');
      if (snCategory && !isSnnPresent) {
        configCategories.removeObject(snCategory);
      }
    },
    
    /**
     * Launches a dialog where an existing config-group can be selected, or a new
     * one can be created. This is different than the config-group management
     * dialog where host membership can be managed.
     *
     * The callback will be passed the created/selected config-group in the form
     * of {id:2, name:'New hardware group'}. In the case of dialog being cancelled,
     * the callback is provided <code>null</code>
     *
     * @param {String} groupName
     *  is closed, cancelled or OK is pressed.
     */

    saveGroupConfirmationPopup: function(groupName) {
      App.ModalPopup.show({
        header: Em.I18n.t('config.group.save.confirmation.header'),
        secondary: Em.I18n.t('config.group.save.confirmation.manage.button'),
        groupName: groupName,
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/configs/saveConfigGroup')
        }),
        onSecondary: function() {
          App.router.get('mainServiceInfoConfigsController').manageConfigurationGroups();
          this.hide();
        }
      });
    },

    //Persist config groups created in step7 wizard controller
    persistWizardStep7ConfigGroups: function () {
      var installerController = App.router.get('installerController');
      var step7Controller = App.router.get('wizardStep7Controller');
      if (App.supports.hostOverridesInstaller) {
        installerController.saveServiceConfigGroups(step7Controller);
        App.clusterStatus.setClusterStatus({
          localdb: App.db.data
        });
      }
    },
    /**
     * exclude configs that depends on services which are uninstalled
     * if config doesn't have serviceName or dependent service is installed then
     * config not excluded
     */
    excludeUnsupportedConfigs: function (configs, installedServices) {
      return configs.filter(function (config) {
        return !(config.serviceName && !installedServices.contains(config.serviceName));
      });
    },

    launchConfigGroupSelectionCreationDialog : function(serviceId, configGroups, configProperty, callback, isInstaller) {
      var self = this;
      var availableConfigGroups = configGroups.slice();
      // delete Config Groups, that already have selected property overridden
      var alreadyOverriddenGroups = [];
      if (configProperty.get('overrides')) {
        alreadyOverriddenGroups = configProperty.get('overrides').mapProperty('group.name');
      }
      var result = [];
      availableConfigGroups.forEach(function (group) {
        if (!group.get('isDefault') && (!alreadyOverriddenGroups.length || !alreadyOverriddenGroups.contains(group.name))) {
          result.push(group);
        }
      }, this);
      availableConfigGroups = result;
      var selectedConfigGroup = availableConfigGroups && availableConfigGroups.length > 0 ?
          availableConfigGroups[0] : null;
      var serviceName = App.Service.DisplayNames[serviceId];
      App.ModalPopup.show({
        classNames: [ 'sixty-percent-width-modal' ],
        header: Em.I18n.t('config.group.selection.dialog.title').format(serviceName),
        subTitle: Em.I18n.t('config.group.selection.dialog.subtitle').format(serviceName),
        selectExistingGroupLabel: Em.I18n.t('config.group.selection.dialog.option.select').format(serviceName),
        noGroups: Em.I18n.t('config.group.selection.dialog.no.groups').format(serviceName),
        createNewGroupLabel: Em.I18n.t('config.group.selection.dialog.option.create').format(serviceName),
        createNewGroupDescription: Em.I18n.t('config.group.selection.dialog.option.create.msg').format(serviceName),
        warningMessage: '&nbsp;',
        isWarning: false,
        optionSelectConfigGroup: true,
        optionCreateConfigGroup: function(){
          return !this.get('optionSelectConfigGroup');
        }.property('optionSelectConfigGroup'),
        hasExistedGroups: function() {
          return !!this.get('availableConfigGroups').length;
        }.property('availableConfigGroups'),
        availableConfigGroups: availableConfigGroups,
        selectedConfigGroup: selectedConfigGroup,
        newConfigGroupName: '',
        disablePrimary: function () {
          return !(this.get('optionSelectConfigGroup') || (this.get('newConfigGroupName').trim().length > 0 && !this.get('isWarning')));
        }.property('newConfigGroupName', 'optionSelectConfigGroup', 'warningMessage'),
        onPrimary: function () {
          if (this.get('optionSelectConfigGroup')) {
            var selectedConfigGroup = this.get('selectedConfigGroup');
            this.hide();
            callback(selectedConfigGroup);
          } else {
            var newConfigGroupName = this.get('newConfigGroupName').trim();
            var newConfigGroup = App.ConfigGroup.create({
              id: null,
              name: newConfigGroupName,
              description: Em.I18n.t('config.group.description.default').format(new Date().toDateString()),
              isDefault: false,
              parentConfigGroup: null,
              service: (isInstaller) ? Em.Object.create({id: serviceId}) : App.Service.find().findProperty('serviceName', serviceId),
              hosts: [],
              configSiteTags: [],
              properties: []
            });
            self.postNewConfigurationGroup(newConfigGroup);
            if (newConfigGroup) {
              newConfigGroup.set('parentConfigGroup', configGroups.findProperty('isDefault'));
              configGroups.pushObject(newConfigGroup);
              if (isInstaller) {
                self.persistWizardStep7ConfigGroups();
              } else {
                self.saveGroupConfirmationPopup(newConfigGroupName);
              }
              this.hide();
              callback(newConfigGroup);
            }
          }
        },
        onSecondary: function () {
          this.hide();
          callback(null);
        },
        doSelectConfigGroup: function (event) {
          var configGroup = event.context;
          console.log(configGroup);
          this.set('selectedConfigGroup', configGroup);
        },
        validate: function () {
          var msg = '&nbsp;';
          var isWarning = false;
          var optionSelect = this.get('optionSelectConfigGroup');
          if (!optionSelect) {
            var nn = this.get('newConfigGroupName');
            if (nn && configGroups.mapProperty('name').contains(nn.trim())) {
              msg = Em.I18n.t("config.group.selection.dialog.err.name.exists");
              isWarning = true;
            }
          }
          this.set('warningMessage', msg);
          this.set('isWarning', isWarning);
        }.observes('newConfigGroupName', 'optionSelectConfigGroup'),
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/configs/selectCreateConfigGroup'),
          controllerBinding: 'App.router.mainServiceInfoConfigsController',
          selectConfigGroupRadioButton: Ember.Checkbox.extend({
            tagName: 'input',
            attributeBindings: ['type', 'checked', 'disabled'],
            checked: function () {
              return this.get('parentView.parentView.optionSelectConfigGroup');
            }.property('parentView.parentView.optionSelectConfigGroup'),
            type: 'radio',
            disabled: false,
            click: function () {
              this.set('parentView.parentView.optionSelectConfigGroup', true);
            },
            didInsertElement: function () {
              if (!this.get('parentView.parentView.hasExistedGroups')) {
                this.set('disabled', true);
                this.set('parentView.parentView.optionSelectConfigGroup', false);
              }
            }
          }),
          createConfigGroupRadioButton: Ember.Checkbox.extend({
            tagName: 'input',
            attributeBindings: ['type', 'checked'],
            checked: function () {
              return !this.get('parentView.parentView.optionSelectConfigGroup');
            }.property('parentView.parentView.optionSelectConfigGroup'),
            type: 'radio',
            click: function () {
              this.set('parentView.parentView.optionSelectConfigGroup', false);
            }
          })
        })
      });
    },
    /**
     * launch dialog where can be assigned another group to host
     * @param selectedGroup
     * @param configGroups
     * @param hostName
     * @param callback
     */
    launchSwitchConfigGroupOfHostDialog: function (selectedGroup, configGroups, hostName, callback) {
      var self = this;
      App.ModalPopup.show({
        header: Em.I18n.t('config.group.host.switch.dialog.title'),
        configGroups: configGroups,
        selectedConfigGroup: selectedGroup,
        disablePrimary: function () {
          return !(this.get('selectedConfigGroup.name') !== selectedGroup.get('name'));
        }.property('selectedConfigGroup'),
        onPrimary: function () {
          var newGroup = this.get('selectedConfigGroup');
          selectedGroup.get('hosts').removeObject(hostName);
          if (!selectedGroup.get('isDefault')) {
            self.updateConfigurationGroup(selectedGroup, function () {
            }, function () {});
          }
          newGroup.get('hosts').pushObject(hostName);
          callback(newGroup);
          if (!newGroup.get('isDefault')) {
            self.updateConfigurationGroup(newGroup, function () {
            }, function () {});
          }
          this.hide();
        },
        bodyClass: Ember.View.extend({
          templateName: require('templates/utils/config_launch_switch_config_group_of_host')
        })
      });
    },
    /**
     * Create a new config-group for a service.
     *
     * @param newConfigGroupData   config group to post to server
     * @param callback    Callback function for Success or Error handling
     * @return  Returns the created config-group
     */
    postNewConfigurationGroup: function (newConfigGroupData, callback) {
      var dataHosts = [];
      newConfigGroupData.get('hosts').forEach(function (_host) {
        dataHosts.push({
          host_name: _host
        });
      }, this);
      var sendData = {
        name: 'config_groups.create',
        data: {
          'group_name': newConfigGroupData.get('name'),
          'service_id': newConfigGroupData.get('service.id'),
          'description': newConfigGroupData.get('description'),
          'hosts': dataHosts
        },
        success: 'successFunction',
        error: 'errorFunction',
        successFunction: function (response) {
          newConfigGroupData.set('id', response.resources[0].ConfigGroup.id);
          if (callback) {
            callback();
          }
        },
        errorFunction: function (xhr, text, errorThrown) {
          if (callback) {
            callback(xhr, text, errorThrown);
          }
          console.error('Error in creating new Config Group');
        }
      };
      sendData.sender = sendData;
      App.ajax.send(sendData);
      return newConfigGroupData;
    },

    /**
     * PUTs the new configuration-group on the server.
     * Changes possible here are the name, description and
     * host memberships of the configuration-group.
     * 
     * @param {App.ConfigGroup} configGroup Configuration group to update
     * @param {Function} successCallback
     * @param {Function} errorCallback
     */
    updateConfigurationGroup: function (configGroup, successCallback, errorCallback) {
      var putConfigGroup = {
        ConfigGroup: {
          group_name: configGroup.get('name'),
          description: configGroup.get('description'),
          tag: configGroup.get('service.id'),
          hosts: [],
          desired_configs: []
        }  
      };
      configGroup.get('hosts').forEach(function(h){
        putConfigGroup.ConfigGroup.hosts.push({
          host_name: h
        });
      });
      configGroup.get('configSiteTags').forEach(function(cst){
        putConfigGroup.ConfigGroup.desired_configs.push({
          type: cst.get('site'),
          tag: cst.get('tag')
        });
      });
      
      var sendData = {
        name: 'config_groups.update',
        data: {
          id: configGroup.get('id'),
          data: putConfigGroup
        },
        success: 'successFunction',
        error: 'errorFunction',
        successFunction: function () {
          if(successCallback) {
            successCallback();
          }
        },
        errorFunction: function (xhr, text, errorThrown) {
          if(errorCallback) {
            errorCallback(xhr, text, errorThrown);
          }
        }
      };
      sendData.sender = sendData;
      App.ajax.send(sendData);
    },

    clearConfigurationGroupHosts: function (configGroup, successCallback, errorCallback) {
      configGroup = jQuery.extend({}, configGroup);
      configGroup.set('hosts', []);
      this.updateConfigurationGroup(configGroup, successCallback, errorCallback);
    },

    deleteConfigGroup: function (configGroup, successCallback, errorCallback) {
      var sendData = {
        name: 'config_groups.delete_config_group',
        sender: this,
        data: {
          id: configGroup.get('id')
        },
        success: 'successFunction',
        error: 'errorFunction',
        successFunction: function () {
          if(successCallback) {
            successCallback();
          }
        },
        errorFunction: function (xhr, text, errorThrown) {
          if(errorCallback) {
            errorCallback(xhr, text, errorThrown);
          }
        }
      };
      sendData.sender = sendData;
      App.ajax.send(sendData);
    },

    /**
     * Gets all the configuration-groups for the given service.
     * 
     * @param serviceId
     *          (string) ID of the service. Ex: HDFS
     */
    getConfigGroupsForService: function (serviceId) {

    },

    /**
     * Gets all the configuration-groups for a host.
     *
     * @param hostName
     *          (string) host name used to register
     */
    getConfigGroupsForHost: function (hostName) {

    }

  });
  
});
window.require.register("utils/configs/defaults_providers/defaultsProvider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.DefaultsProvider = Em.Object.extend({

    /**
     * Look at cluster setup, the provided properties, and provide an object where keys are property-names, and values are the recommended defaults
     * @param {App.ServiceConfigProperty} serviceConfigProperty
     */
    getDefaults: function(serviceConfigProperty) {

    },

    /**
     * Cluster info used to calculate properties values
     * Should be redeclared in the child providers
     */
    getClusterData: function() {

    }
  });
  
});
window.require.register("utils/configs/defaults_providers/hive_defaults_provider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/defaults_providers/yarn_defaults_provider');

  App.HiveDefaultsProvider = App.YARNDefaultsProvider.extend({

    configsTemplate: {
      'hive.tez.container.size': null,
      'hive.tez.java.opts': null,
      'hive.auto.convert.join.noconditionaltask.size': null
    },

    getDefaults : function(localDB) {
      var configs = this._super(localDB);
      if (configs['yarn.scheduler.maximum-allocation-mb'] != null && configs['mapreduce.map.memory.mb'] != null
          && configs['mapreduce.reduce.memory.mb'] != null) {
        var containerSize = configs['mapreduce.map.memory.mb'] > 2048 ? configs['mapreduce.map.memory.mb'] : configs['mapreduce.reduce.memory.mb'];
        containerSize = Math.min(configs['yarn.scheduler.maximum-allocation-mb'], containerSize);
        configs['hive.auto.convert.join.noconditionaltask.size'] = Math.round(containerSize / 3) * 1048576; // MB to Bytes
        configs['hive.tez.java.opts'] = "-server -Xmx" + Math.round(0.8 * containerSize) + "m -Djava.net.preferIPv4Stack=true -XX:NewRatio=8 -XX:+UseNUMA -XX:+UseParallelGC";
        configs['hive.tez.container.size'] = containerSize;
      } else {
        jQuery.extend(configs, this.get('configsTemplate'));
      }
      return configs;
    }

  });
  
});
window.require.register("utils/configs/defaults_providers/oozie_defaults_provider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/defaults_providers/defaultsProvider');

  App.OOZIEDefaultsProvider = App.DefaultsProvider.extend({

    clusterData: null,

    /**
     * List of the configs that should be calculated
     */
    configsTemplate: {
      'oozie.services.ext': null
    },


    getDefaults: function (localDB) {
      this._super();
      this.getClusterData(localDB);
      var configs = {};
      jQuery.extend(configs, this.get('configsTemplate'));
      if (localDB.masterComponentHosts.findProperty('component','FALCON_SERVER')) {
        configs['oozie.services.ext'] = "org.apache.oozie.service.JMSAccessorService,org.apache.oozie.service.PartitionDependencyManagerService,org.apache.oozie.service.HCatAccessorService";
        return configs;
      }
    }
  });
  
});
window.require.register("utils/configs/defaults_providers/storm_defaults_provider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/defaults_providers/defaultsProvider');

  App.STORMDefaultsProvider = App.DefaultsProvider.extend({

    clusterData: null,

    /**
     * List of the configs that should be calculated
     */
    configsTemplate: {
      'drpc.childopts': null,
      'ui.childopts': null,
      'logviewer.childopts': null
    },

    // @todo fill with correct values
    getDefaults: function (localDB) {
      this._super();
      this.getClusterData(localDB);
      var configs = {};
      jQuery.extend(configs, this.get('configsTemplate'));
      if (!this.clusterDataIsValid()) {
        return configs;
      }
      return configs;
    },

    /**
     * Verify <code>clusterData</code> - check if all properties are defined
     */
    clusterDataIsValid: function () {
      return true;
    }
  });
  
});
window.require.register("utils/configs/defaults_providers/tez_defaults_provider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/defaults_providers/yarn_defaults_provider');

  App.TezDefaultsProvider = App.YARNDefaultsProvider.extend({

    configsTemplate: {
      'tez.am.resource.memory.mb': null,
      'tez.am.java.opts': null
    },

    getDefaults : function(localDB) {
      var configs = this._super(localDB);
      if (configs['yarn.app.mapreduce.am.resource.mb'] != null) {
        configs['tez.am.resource.memory.mb'] = configs['yarn.app.mapreduce.am.resource.mb'];
        configs['tez.am.java.opts'] = '-server -Xmx' + Math.round(0.8 * configs['tez.am.resource.memory.mb'])
            + 'm -Djava.net.preferIPv4Stack=true -XX:+UseNUMA -XX:+UseParallelGC';
      } else {
        jQuery.extend(configs, this.get('configsTemplate'));
      }
      return configs;
    }

  });
  
});
window.require.register("utils/configs/defaults_providers/user_defaults_provider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  require('utils/configs/defaults_providers/defaultsProvider');

  App.userDefaultsProvider = App.DefaultsProvider.extend({

    clusterData: null,

    /**
     * List of the configs that should be calculated
     */
    configsTemplate: {
      'hdfs_user': null,
      'mapred_user': null,
      'yarn_user': null,
      'hbase_user': null,
      'hive_user': null,
      'hcat_user': null,
      'webhcat_user': null,
      'oozie_user': null,
      'falcon_user': null,
      'storm_user': null,
      'zk_user': null,
      'gmetad_user': null,
      'gmond_user': null,
      'nagios_user': null,
      'smokeuser': null
    },

    // @todo fill with correct values
    getDefaults: function (localDB) {
      this._super();
      var configs = {};
      jQuery.extend(configs, this.get('configsTemplate'));
      return configs;
    }
  });

  
});
window.require.register("utils/configs/defaults_providers/yarn_defaults_provider", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/defaults_providers/defaultsProvider');

  App.YARNDefaultsProvider = App.DefaultsProvider.extend({

    /**
     * List of the configs that should be calculated
     */
    configsTemplate: {
      'yarn.nodemanager.resource.memory-mb': null,
      'yarn.scheduler.minimum-allocation-mb': null,
      'yarn.scheduler.maximum-allocation-mb': null,
      'mapreduce.map.memory.mb': null,
      'mapreduce.reduce.memory.mb': null,
      'mapreduce.map.java.opts': null,
      'mapreduce.reduce.java.opts': null,
      'mapreduce.task.io.sort.mb': null,
      'yarn.app.mapreduce.am.resource.mb': null,
      'yarn.app.mapreduce.am.command-opts': null
    },

    /**
     * Information about ram, disk count, cpu count and hbase availability
     * Example:
     * {
     *   disk: 12,
     *   ram: 48,
     *   cpu: 12,
     *   hBaseInstalled: false
     * }
     */
    clusterData: null,

    /**
     * Reserved for system memory
     */
    reservedRam: null,

    /**
     * Reserved for HBase memory
     */
    hBaseRam: null,

    GB: 1024,
    /**
     *  Minimum container size (in RAM).
     *  This value is dependent on the amount of RAM available, as in smaller memory nodes the minimum container size should also be smaller
     *
     *  Value in MB!
     */
    recommendedMinimumContainerSize: function () {
      if (!this.clusterDataIsValid()) return null;
      var ram = this.get('clusterData.ram');
      switch(true) {
        case (ram <= 4): return 256;
        case (ram > 4 && ram <= 8): return 512;
        case (ram > 8 && ram <= 24): return 1024;
        case (ram > 24):
        default: return 2048;
      }
    }.property('clusterData.ram'),

    /**
     * Maximum number of containers allowed per node
     * max(3, min (2*cores,min (1.8*DISKS,(Total available RAM) / MIN_CONTAINER_SIZE))))
     */
    containers: function () {
      if (!this.clusterDataIsValid()) return null;
      var cpu = this.get('clusterData.cpu');
      var disk = this.get('clusterData.disk');
      var ram = this.get('clusterData.ram');
      var containerSize = this.get('recommendedMinimumContainerSize');
      cpu *= 2;
      disk = Math.ceil(disk * 1.8);
      ram -= this.get('reservedRam');
      if (this.get('clusterData.hBaseInstalled')) {
        ram -= this.get('hBaseRam');
      }
      // On low memory systems, memory left over after
      // removing reserved-RAM and HBase might be
      // less than 2GB (even negative). If so, we force
      // a 2GB value relying on virtual memory.
      if (ram < 2) {
        ram = 2;
      }
      ram *= this.get('GB');
      ram /= containerSize;
      return Math.round(Math.max(3, Math.min(cpu, Math.min(disk, ram))));
    }.property('clusterData.cpu', 'clusterData.ram', 'clusterData.hBaseInstalled', 'clusterData.disk', 'reservedRam', 'hBaseRam', 'recommendedMinimumContainerSize'),

    /**
     * Amount of RAM per container.
     * Calculated to be max(2GB, RAM - reservedRam - hBaseRam) / containers
     *
     * @return Memory per container in MB. If greater than 1GB, 
     *          value will be in multiples of 512. 
     */
    ramPerContainer: function () {
      var containers = this.get('containers');
      if (!containers) {
        return null;
      }
      var ram = this.get('clusterData.ram');
      ram = (ram - this.get('reservedRam'));
      if (this.get('clusterData.hBaseInstalled')) {
        ram -= this.get('hBaseRam');
      }
      // On low memory systems, memory left over after
      // removing reserved-RAM and HBase might be
      // less than 2GB (even negative). If so, we force
      // a 2GB value relying on virtual memory.
      if (ram < 2) {
        ram = 2;
      }
      ram *= this.get('GB');
      var container_ram = Math.abs(ram / containers);
      // If container memory is greater than 1GB, 
      // we use multiples of 512 as value
      return container_ram > this.get('GB') ? (Math.floor(container_ram / 512) * 512) : container_ram;
    }.property('containers', 'clusterData.ram', 'clusterData.hBaseInstalled', 'hBaseRam', 'reservedRam'),

    mapMemory: function () {
      return Math.floor(this.get('ramPerContainer'));
    }.property('ramPerContainer'),

    reduceMemory: function () {
      return this.get('ramPerContainer');
    }.property('ramPerContainer'),

    amMemory: function () {
      return Math.max(this.get('mapMemory'), this.get('reduceMemory'));
    }.property('mapMemory', 'reduceMemory'),

    /**
     * Reserved for HBase and system memory is based on total available memory
     */
    reservedMemoryRecommendations: function() {
      var table = [
        {os:1,hbase:1},
        {os:2,hbase:1},
        {os:2,hbase:2},
        {os:4,hbase:4},
        {os:6,hbase:8},
        {os:8,hbase:8},
        {os:8,hbase:8},
        {os:12,hbase:16},
        {os:24,hbase:24},
        {os:32,hbase:32},
        {os:64,hbase:64}
      ];
      var ram = this.get('clusterData.ram');
      var index = 0;
      switch (true) {
        case (ram <= 4): index = 0; break;
        case (ram > 4 && ram <= 8): index = 1; break;
        case (ram > 8 && ram <= 16): index = 2; break;
        case (ram > 16 && ram <= 24): index = 3; break;
        case (ram > 24 && ram <= 48): index = 4; break;
        case (ram > 48 && ram <= 64): index = 5; break;
        case (ram > 64 && ram <= 72): index = 6; break;
        case (ram > 72 && ram <= 96): index = 7; break;
        case (ram > 96 && ram <= 128): index = 8; break;
        case (ram > 128 && ram <= 256): index = 9; break;
        case (ram > 256 && ram <= 512): index = 10; break;
        default: index = 10; break;
      }
      this.set('reservedRam', table[index].os);
      this.set('hBaseRam', table[index].hbase);
    }.observes('clusterData.ram'),

    /**
     * Provide an object where keys are property-names and values are the recommended defaults
     * @param {object} localDB Object with information about hosts and master/slave components
     * Example:
     *  <code>
     *    {
     *       "hosts": {
     *           "host1": {
     *               "name": "host1",
     *               "cpu": 1,
     *               "memory": "6123683.00",
     *               "disk_info": [{
     *                   ....
     *               },...]
     *           },...
     *       },
     *       "masterComponentHosts": [{
     *           "component": "NAMENODE",
     *           "hostName": "host1",
     *           "serviceId": "HDFS"
     *       },...],
     *       "slaveComponentHosts": [{
     *           "componentName": "DATANODE",
     *           "hosts": [{
     *               "hostName": "host2"
     *           }]
     *       },...]
     *   }
     *  </code>
     * @return {object}
     */
    getDefaults: function (localDB) {
      this._super();
      this.getClusterData(localDB);
      var configs = {};
      jQuery.extend(configs, this.get('configsTemplate'));
      if (!this.clusterDataIsValid()) {
        return configs;
      }
      configs['yarn.nodemanager.resource.memory-mb'] = Math.round(this.get('containers') * this.get('ramPerContainer'));
      configs['yarn.scheduler.minimum-allocation-mb'] = Math.floor(this.get('ramPerContainer'));
      configs['yarn.scheduler.maximum-allocation-mb'] = Math.round(this.get('containers') * this.get('ramPerContainer'));
      configs['yarn.app.mapreduce.am.resource.mb'] = Math.floor(this.get('amMemory'));
      configs['yarn.app.mapreduce.am.command-opts'] = "-Xmx" + Math.round(0.8 * this.get('amMemory')) + "m";
      configs['mapreduce.map.memory.mb'] = Math.floor(this.get('mapMemory'));
      configs['mapreduce.reduce.memory.mb'] = Math.floor(this.get('reduceMemory'));
      configs['mapreduce.map.java.opts'] = "-Xmx" + Math.round(0.8 * this.get('mapMemory')) + "m";
      configs['mapreduce.reduce.java.opts'] = "-Xmx" + Math.round(0.8 * this.get('reduceMemory')) + "m";
      configs['mapreduce.task.io.sort.mb'] = Math.round(Math.min(0.4 * this.get('mapMemory'), 1024));
      return configs;
    },

    /**
     * Calculate needed cluster data (like disk count, cpu count, ram (in MB!) and hbase availability)
     * @param {object} localDB Object with information about hosts and master/slave components
     */
    getClusterData: function (localDB) {
      this._super();
      var components = ['NODEMANAGER'];
      var hosts = [];
      if (!localDB.hosts || !(localDB.masterComponentHosts || localDB.slaveComponentHosts)) return;
      var hBaseInstalled = !!localDB.masterComponentHosts.filterProperty('component', 'HBASE_MASTER').length;
      components.forEach(function (component) {
        var mc = localDB.masterComponentHosts.findProperty('component', component);
        if (mc) {
          if (!hosts.contains(mc.hostName)) {
            hosts.push(mc.hostName);
          }
        }
        else {
          var sc = localDB.slaveComponentHosts.findProperty('componentName', component);
          if (sc) {
            sc.hosts.map(function (host) {
              if (!hosts.contains(host.hostName)) {
                hosts.push(host.hostName);
              }
            });
          }
        }
      });
      var clusterData = {
        cpu: 0,
        disk: 0,
        ram: 0,
        hBaseInstalled: hBaseInstalled
      };
      var host = hosts[0] && localDB.hosts[hosts[0]];
      if (host) {
        clusterData.cpu = parseInt(host.cpu);
        var length = 0;
        host.disk_info.forEach(function(disk) {
          //invalid mountpoints
          if (!(disk.mountpoint.startsWith('/home/') || disk.mountpoint.startsWith('/homes/') || 
          		disk.mountpoint.startsWith('/dev/') || disk.mountpoint.startsWith('/tmp/'))) {
            length++;
          }
        },this);
        clusterData.disk = length;
        clusterData.ram = Math.round(parseFloat(host.memory) / (1024 * 1024));
      }
      this.set('clusterData', clusterData);
    },

    /**
     * Verify <code>clusterData</code> - check if all properties are defined
     */
    clusterDataIsValid: function () {
      if (!this.get('clusterData')) return false;
      return !(this.get('clusterData.ram') == null || this.get('clusterData.cpu') == null || this.get('clusterData.disk') == null || this.get('clusterData.hBaseInstalled') == null);
    }

  });
  
});
window.require.register("utils/configs/validators/hive_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/validators/service_configs_validator');

  App.HiveConfigsValidator = App.ServiceConfigsValidator.create({
    /**
     * List of the configs that should be validated
     */
    configValidators: {
      'hive.tez.container.size': 'hiveTezContainerMb',
      'hive.tez.java.opts': 'hiveTezJavaOpts',
      'hive.auto.convert.join.noconditionaltask.size': 'hiveTezJoinNoConditionalBytes'
    },

    hiveTezContainerMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    hiveTezJavaOpts: function(config) {
      return this.validateXmxValue(config);
    },

    hiveTezJoinNoConditionalBytes: function(config) {
      return this.validatorLessThenDefaultValue(config);
    }
  });
  
});
window.require.register("utils/configs/validators/mapreduce2_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/validators/service_configs_validator');

  App.MapReduce2ConfigsValidator = App.ServiceConfigsValidator.create({

    /**
     * List of the configs that should be validated
     */
    configValidators: {
      'mapreduce.map.java.opts': 'mapreduceMapJavaOpts',
      'mapreduce.reduce.java.opts': 'mapreduceReduceJavaOpts',
      'mapreduce.task.io.sort.mb': 'mapreduceTaskIoSortMb',
      'mapreduce.map.memory.mb': 'mapreduceMapMemoryMb',
      'mapreduce.reduce.memory.mb': 'mapreduceReduceMemoryMb',
      'yarn.app.mapreduce.am.resource.mb': 'yarnAppMapreduceAmResourceMb',
      'yarn.app.mapreduce.am.command-opts': 'yarnAppMapreduceAmCommandOpts'
    },

    mapreduceMapJavaOpts: function(config) {
      return this.validateXmxValue(config);
    },

    mapreduceReduceJavaOpts: function(config) {
      return this.validateXmxValue(config);
    },

    mapreduceTaskIoSortMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    mapreduceMapMemoryMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    mapreduceReduceMemoryMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    yarnAppMapreduceAmResourceMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    yarnAppMapreduceAmCommandOpts: function(config) {
      return this.validateXmxValue(config);
    }

  });
  
});
window.require.register("utils/configs/validators/service_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.ServiceConfigsValidator = Em.Object.extend({

    /**
     * Defaults recommended for various properties. This is to be used 
     * by this validator to validate a property. Generally this value should be 
     * set to values given by the defaults provider of the service.
     * 
     * The key is the property name, and the value is the recommended default.
     */
    recommendedDefaults: {},
    
    /**
     * Per property validators where key is config name and value 
     * is the validation function. This field is expected to be
     * overridden by extending classes.
     */
    configValidators: {},
    
    /**
     * Validate the given config property with the available  
     *
     * @param config  {App.ServiceConfigProperty}
     * @return {string}  No validation issues when <code>null</code> returned.
     */
    validateConfig: function(config) {
      var validatorFunction = this.get('configValidators')[config.get('name')];
      if (validatorFunction) {
        return this[validatorFunction](config);
      }
      return null;
    },
    
    /**
     * Check if provided <code>config.value</code> is less than <code>recommendedDefaults</code>
     * @param {object} config - configProperty name
     * @return {string|null}
     */
    validatorLessThenDefaultValue: function(config) {
      var defaultValue = this.get('recommendedDefaults')[config.get('name')];
      var currentValue = parseInt(config.get('value').toString().replace( /\D+/g, ''));
      if (!defaultValue) {
        return null;
      }
      defaultValue = parseInt(defaultValue.toString().replace( /\D+/g, ''));
      if (defaultValue && currentValue &&  currentValue < defaultValue) {
        return "Value is less than the recommended default of " + defaultValue;
      }
      return null;
    },

    /**
     * Check if provided <code>config.value</code> is less than <code>recommendedDefaults</code> or <code>config.defaultValue</code>
     * Value looks like "-Xmx****m"
     * @param {object} config
     * @return {string|null}
     */
    validateXmxValue: function(config) {
      var recomendedDefault = this.get('recommendedDefaults')[config.get('name')];
      var defaultValueRaw = Em.isNone(recomendedDefault) ? config.get('defaultValue') : recomendedDefault;
      Em.assert('validateXmxValue: Config\'s default value can\'t be null or undefined', !Em.isNone(defaultValueRaw));
      var currentValueRaw = config.get('value');
      if (!this._checkXmxValueFormat(currentValueRaw)) {
        return 'Invalid value format';
      }
      var currentValue = this._formatXmxSizeToBytes(this._getXmxSize(currentValueRaw));
      var defaultValueXmx = this._getXmxSize(defaultValueRaw);
      var defaultValue = this._formatXmxSizeToBytes(defaultValueXmx);
      if (currentValue < defaultValue) {
        return "Value is less than the recommended default of -Xmx" + defaultValueXmx;
      }
      return null;
    },
    /**
     * Verify if provided value has proper format (should be like "-Xmx***(b|k|m|g|p|t|B|K|M|G|P|T)")
     * @param  {string} value
     * @returns {bool}
     * @private
     */
    _checkXmxValueFormat: function(value) {
      var regex = /(^|\s)\-Xmx(\d+)(b|k|m|g|p|t|B|K|M|G|P|T)?(\s|$)/;
      if (!regex.test(value)) {
        return false;
      }
      // "-Xmx" can be only one
      return value.match(/\-Xmx/ig).length == 1;
    },
    /**
     * Parse Xmx size from raw value
     * @param {string} value
     * @returns {string}
     * @private
     */
    _getXmxSize: function(value) {
      var regex = /\-Xmx(\d+)(.?)/;
      var result = regex.exec(value);
      if (result.length > 1) {
        // result[2] - is a space or size formatter (b|k|m|g etc)
        return result[1] + result[2].toLowerCase();
      }
      return result[1];
    },
    /**
     * Calculate bytes size from value
     * @param {string} value
     * @returns {int}
     * @private
     */
    _formatXmxSizeToBytes: function(value) {
      value = value.toLowerCase();
      if (value.length == 0) {
        return 0;
      }
      var modifier = value[value.length - 1];
      if (modifier == ' ' || "0123456789".indexOf(modifier) != -1) {
        modifier = 'b';
      }
      var m = 1;
      switch (modifier) {
        case 'b': m = 1; break;
        case 'k': m = 1024; break;
        case 'm': m = 1024 * 1024; break;
        case 'g': m = 1024 * 1024 * 1024; break;
        case 't': m = 1024 * 1024 * 1024 * 1024; break;
        case 'p': m = 1024 * 1024 * 1024 * 1024 * 1024; break;
      }
      var result = parseInt(value.replace(modifier, '').trim());
      result *= m;
      return result;
    }

  });
  
});
window.require.register("utils/configs/validators/storm_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/validators/service_configs_validator');

  App.STORMConfigsValidator = App.ServiceConfigsValidator.create({
    /**
     * List of the configs that should be validated
     */
    configValidators: {
      'drpc.childopts': 'drpcChildOpts',
      'ui.childopts': 'uiChildOpts',
      'logviewer.childopts': 'logviewerChildOpts'
    },

    drpcChildOpts: function(config) {
      return this.validateXmxValue(config);
    },

    uiChildOpts: function(config) {
      return this.validateXmxValue(config);
    },

    logviewerChildOpts: function(config) {
      return this.validateXmxValue(config);
    }
  });
  
});
window.require.register("utils/configs/validators/tez_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/validators/service_configs_validator');

  App.TezConfigsValidator = App.ServiceConfigsValidator.create({
    /**
     * List of the configs that should be validated
     */
    configValidators: {
      'tez.am.resource.memory.mb': 'tezAMResourceMb',
      'tez.am.java.opts': 'tezAMJavaOpts'
    },

    tezAMResourceMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    tezAMJavaOpts: function(config) {
      return this.validateXmxValue(config);
    }

  });
  
});
window.require.register("utils/configs/validators/user_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');
  require('utils/configs/validators/service_configs_validator');

  App.userConfigsValidator = App.ServiceConfigsValidator.create({
    /**
     * List of the configs that should be validated
     */
    configValidators: {
      'hdfs_user': 'validateUserValue',
      'mapred_user': 'validateUserValue',
      'yarn_user': 'validateUserValue',
      'hbase_user': 'validateUserValue',
      'hive_user': 'validateUserValue',
      'hcat_user': 'validateUserValue',
      'webhcat_user': 'validateUserValue',
      'oozie_user': 'validateUserValue',
      'falcon_user': 'validateUserValue',
      'storm_user': 'validateUserValue',
      'zk_user': 'validateUserValue',
      'gmetad_user': 'validateUserValue',
      'gmond_user': 'validateUserValue',
      'nagios_user': 'validateUserValue',
      'smokeuser': 'validateUserValue'
    },

    validateUserValue: function(config) {
      var recomendedDefault = this.get('recommendedDefaults')[config.get('name')];
      var defaultValue = Em.isNone(recomendedDefault) ? config.get('defaultValue') : recomendedDefault;
      Em.assert('validateUserValue: User\'s default value can\'t be null or undefined', !Em.isNone(defaultValue));
      var currentValue = config.get('value');
      if (!validator.isValidUserName(currentValue)) {
        return Em.I18n.t('users.userName.validationFail');
      }
      return null;
    }
  });
  
});
window.require.register("utils/configs/validators/yarn_configs_validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('utils/configs/validators/service_configs_validator');

  App.YARNConfigsValidator = App.ServiceConfigsValidator.create({
    /**
     * List of the configs that should be validated
     */
    configValidators: {
      'yarn.nodemanager.resource.memory-mb': 'yarnNodemanagerResourceMemoryMb',
      'yarn.scheduler.minimum-allocation-mb': 'yarnSchedulerMinimumAllocationMb',
      'yarn.scheduler.maximum-allocation-mb': 'yarnSchedulerMaximumAllocationMb'
    },

    yarnNodemanagerResourceMemoryMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    yarnSchedulerMinimumAllocationMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    },

    yarnSchedulerMaximumAllocationMb: function(config) {
      return this.validatorLessThenDefaultValue(config);
    }

  });
  
});
window.require.register("utils/date", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var validator = require('utils/validator');
  var App = require('app');

  module.exports = {
    dateMonths:['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
    dateDays:['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
    dateFormatZeroFirst:function (time) {
      if (time < 10) return '0' + time;
      return time;
    },
    /**
     * Convert timestamp to date-string 'DAY_OF_THE_WEEK, MONTH DAY, YEAR HOURS:MINUTES'
     * @param timestamp
     * @return string date
     */
    dateFormat:function (timestamp, showSeconds, showMilliseconds) {
      if (!validator.isValidInt(timestamp)) return timestamp;
      var date = new Date(timestamp);
      var months = this.dateMonths;
      var days = this.dateDays;
      var formattedDate = days[date.getDay()] + ', ' + months[date.getMonth()] + ' ' + this.dateFormatZeroFirst(date.getDate()) + ', ' + date.getFullYear() + ' ' + this.dateFormatZeroFirst(date.getHours()) + ':' + this.dateFormatZeroFirst(date.getMinutes());
      if (showSeconds) {
        formattedDate += ':' + this.dateFormatZeroFirst(date.getSeconds());
        if (showMilliseconds) {
          formattedDate += '.' + this.dateFormatZeroFirst(date.getMilliseconds());
        };
      };
      return formattedDate;
    },
    /**
     * Convert timestamp to date-string 'DAY_OF_THE_WEEK MONTH DAY YEAR'
     * @param timestamp
     * @return {*}
     */
    dateFormatShort: function(timestamp) {
      if (!validator.isValidInt(timestamp)) return timestamp;

      var date = new Date(timestamp);
      var today = new Date();
      if (date.toDateString() === today.toDateString()) {
        return 'Today ' + date.toLocaleTimeString();
      }
      return date.toDateString();
    },
    /**
     * Convert starTimestamp to 'DAY_OF_THE_WEEK, MONTH DAY, YEAR HOURS:MINUTES', except for the case: year equals 1969
     * @param startTimestamp
     * @return string startTimeSummary
     */
    startTime: function (startTimestamp) {
      if (!validator.isValidInt(startTimestamp)) return '';
      var startDate = new Date(startTimestamp);
      var months = this.dateMonths;
      var days = this.dateDays;
      // generate start time
      if (startDate.getFullYear() == 1969 || startTimestamp < 1) {
        return 'Not started';
      }
      var startTimeSummary = '';
      if (new Date(startTimestamp).setHours(0, 0, 0, 0) == new Date().setHours(0, 0, 0, 0) ) { //today
        startTimeSummary = 'Today ' + this.dateFormatZeroFirst(startDate.getHours()) + ':' + this.dateFormatZeroFirst(startDate.getMinutes());
      } else {
        startTimeSummary =  days[startDate.getDay()] + ' ' + months[startDate.getMonth()] + ' ' + this.dateFormatZeroFirst(startDate.getDate()) + ' ' + startDate.getFullYear() + ' '
          + this.dateFormatZeroFirst(startDate.getHours()) + ':' + this.dateFormatZeroFirst(startDate.getMinutes());
      }
      return startTimeSummary;
    },
    /**
     * Provides the duration between the given start and end timestamp. If start time
     * not valid, duration will be ''. If end time is not valid, duration will
     * be till now, showing 'Lasted for xxx secs'.
     * @param startTimestamp
     * @param endTimestamp
     * @return string durationSummary
     */
    durationSummary: function (startTimestamp, endTimestamp) {
      // generate duration
      var durationSummary = '';
      var startDate = new Date(startTimestamp);
      var endDate = new Date(endTimestamp);
      var self = this;
      if (startDate.getFullYear() == 1969 || startTimestamp < 1) {
        // not started
        return Em.I18n.t('common.na');
      }
      if (endDate.getFullYear() != 1969 && endTimestamp > 0) {
        return '' + this.timingFormat(endTimestamp - startTimestamp, 1); //lasted for xx secs
      } else {
        // still running, duration till now
        var time =  (App.dateTime() - startTimestamp) < 0? 0 : (App.dateTime() - startTimestamp) ;
        durationSummary = '' + this.timingFormat( time , 1);
      }
      return durationSummary;
    },

    /**
     * Convert time in mseconds to
     * 30 ms = 30 ms
     * 300 ms = 300 ms
     * 999 ms = 999 ms
     * 1000 ms = 1.00 secs
     * 3000 ms = 3.00 secs
     * 35000 ms = 35.00 secs
     * 350000 ms = 350.00 secs
     * 999999 ms = 999.99 secs
     * 1000000 ms = 16.66 mins
     * 3500000 secs = 58.33 mins
     * @param time
     * @param zeroValid for the case to show 0 when time is 0, not null
     * @return string formatted date
     */
    timingFormat:function (time, /* optional */ zeroValid) {
      var intTime  = parseInt(time);
      if (zeroValid && intTime == 0) return 0 + ' secs';
      if (!intTime) return null;
      var timeStr = intTime.toString();
      var lengthOfNumber = timeStr.length;
      var oneMinMs = 60000;
      var oneHourMs = 3600000;
      var oneDayMs = 86400000;

      if (lengthOfNumber < 4) {
        return time + ' ms';
      } else if (lengthOfNumber < 7) {
        time = (time / 1000).toFixed(2);
        return time + ' secs';
      } else if (time < oneHourMs) {
        time = (time / oneMinMs).toFixed(2);
        return time + ' mins';
      } else if (time < oneDayMs) {
        time = (time / oneHourMs).toFixed(2);
        return time + ' hours';
      } else {
        time = (time / oneDayMs).toFixed(2);
        return time + ' days';
      }
    },
    /**
     * Provides the duration between the given start and end time. If start time
     * is not given, duration will be 0. If end time is not given, duration will
     * be till now.
     *
     * @param {Number} startTime Start time from epoch
     * @param {Number} endTime End time from epoch
     * @return {Number} duration
     */
    duration : function(startTime, endTime) {
      var duration = 0;
      if (startTime && startTime > 0) {
        if (!endTime || endTime < 1) {
          endTime = App.dateTime();
        }
        duration = endTime - startTime;
      }
      return duration;
    }
  };
  
});
window.require.register("utils/db", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');
  App.db = {};
  var InitialData =  {
    'app': {
      'loginName': '',
      'authenticated': false,
      'configs': [],
      'tables': {
        'filterConditions': {},
        'displayLength': {},
        'startIndex': {},
        'sortingConditions': {}
      }
    },

    'Installer' : {},
    'AddHost' : {},
    'AddService' : {},
    'StackUpgrade' : {},
    'ReassignMaster' : {},
    'AddSecurity': {},
    'HighAvailabilityWizard': {},
    'RollbackHighAvailabilityWizard': {}

  };

  if (typeof Storage !== 'undefined') {
    Storage.prototype.setObject = function (key, value) {
      this.setItem(key, JSON.stringify(value));
    };

    Storage.prototype.getObject = function (key) {
      var value = this.getItem(key);
      return value && JSON.parse(value);
    };
  } else {
    // stub for unit testing purposes
    window.localStorage = {};
    localStorage.setItem = function (key, val) {
      this[key] = val;
    };
    localStorage.getItem = function (key) {
      return this[key];
    };
    window.localStorage.setObject = function (key, value) {
      this[key] = value;
    };
    window.localStorage.getObject = function (key, value) {
      return this[key];
    };
  }

  App.db.cleanUp = function () {
    console.log('TRACE: Entering db:cleanup function');
    App.db.data = InitialData;
    console.log("In cleanup./..");
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.updateStorage = function() {
    App.db.data = localStorage.getObject('ambari');
    if (App.db.data && App.db.data.app && App.db.data.app.tables && App.db.data.app.configs) {
      return true;
    }
    console.warn("local storage is deprecated!");
    App.db.cleanUp();
    return false;
  };

  /*
    Initialize wizard namespaces if they are not initialized on login.
    This will be required during upgrade.
   */
  App.db.mergeStorage = function() {
    if (localStorage.getObject('ambari') == null) {
      console.log('doing a cleanup');
      App.db.cleanUp();
    } else {
      localStorage.setObject('ambari', $.extend(true,{}, InitialData, App.db.data));
    }
  };

  // called whenever user logs in
  if (localStorage.getObject('ambari') == null) {
    console.log('doing a cleanup');
    App.db.cleanUp();
  }

  App.db.get = function (namespace, key) {
    console.log('TRACE: Entering db:get' + key);
    App.db.data = localStorage.getObject('ambari');
    return App.db.data[namespace][key];
  };

  App.db.set = function (namespace, key, value) {
    console.log('TRACE: Entering db:set' + key + ';value: ', value);
    App.db.data = localStorage.getObject('ambari');
    App.db.data[namespace][key] = value;
    localStorage.setObject('ambari', App.db.data);
  };
  /*
   * setter methods
   */

  App.db.setLoginName = function (name) {
    console.log('TRACE: Entering db:setLoginName function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.app.loginName = name;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setAmbariStacks = function (stacks) {
    console.log('TRACE: Entering db:setAmbariStacks function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.app.stacks = stacks;
    localStorage.setObject('ambari', App.db.data);
  };

  /**
   * Set user model to db
   * @param user
   */
  App.db.setUser = function (user) {
    console.log('TRACE: Entering db:setUser function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.app.user = user;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setAuthenticated = function (authenticated) {
    console.log('TRACE: Entering db:setAuthenticated function');

    App.db.data = localStorage.getObject('ambari');
    console.log('present value of authentication is: ' + App.db.data.app.authenticated);
    console.log('desired value of authentication is: ' + authenticated);
    App.db.data.app.authenticated = authenticated;
    localStorage.setObject('ambari', App.db.data);
    App.db.data = localStorage.getObject('ambari');
    console.log('Now present value of authentication is: ' + App.db.data.app.authenticated);
  };

  App.db.setFilterConditions = function(name, filterConditions) {
    console.log('TRACE: Entering db:setFilterConditions function');
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.app.tables.filterConditions) {
      App.db.data.app.tables.filterConditions = {};
    }
    App.db.data.app.tables.filterConditions[name] = filterConditions;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setDisplayLength = function(name, displayLength) {
    console.log('TRACE: Entering db:setDisplayLength function');
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.app.tables.displayLength) {
      App.db.data.app.tables.displayLength = {};
    }
    App.db.data.app.tables.displayLength[name] = displayLength;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setStartIndex = function(name, startIndex) {
    console.log('TRACE: Entering db:setStartIndex function');
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.app.tables.startIndex) {
      App.db.data.app.tables.startIndex = {};
    }
    App.db.data.app.tables.startIndex[name] = startIndex;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSortingStatuses = function(name, sortingConditions) {
    console.log('TRACE: Entering db:setSortingConditions function');
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.app.tables.sortingConditions) {
      App.db.data.app.tables.sortingConditions = {};
    }
    App.db.data.app.tables.sortingConditions[name] = sortingConditions;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setAllHostNames = function (hostNames) {
    console.log('TRACE: Entering db:setAllHostNames function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.hostNames = hostNames;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setAllHostNamesPattern = function (hostNames) {
    console.log('TRACE: Entering db:setAllHostNamesPattern function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.hostNamesPattern = hostNames;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHosts = function (hostInfo) {
    console.log('TRACE: Entering db:setHosts function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.hostInfo = hostInfo;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setInstallOptions = function (installOptions) {
    console.log('TRACE: Entering db:setInstallOptions function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.installOptions = installOptions;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setBootStatus = function (status) {
    console.log('TRACE: Entering db:setBootStatus function');
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.bootStatus = status;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setService = function (serviceInfo) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.serviceInfo = serviceInfo;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSelectedServiceNames = function (serviceNames) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.selectedServiceNames = serviceNames;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setClientsForSelectedServices = function (clientInfo) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.clientInfo = clientInfo;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setMasterComponentHosts = function (masterComponentHosts) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.masterComponentHosts = masterComponentHosts;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSlaveComponentHosts = function (slaveComponentHosts) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.slaveComponentHosts = slaveComponentHosts;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setServiceConfigs = function (serviceConfigs) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.serviceConfigs = serviceConfigs;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setAdvancedServiceConfig = function (serviceConfigs) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.advanceServiceConfigs = serviceConfigs;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setServiceConfigProperties = function (configProperties) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.configProperties = configProperties;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setCluster = function (status) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.Installer.clusterStatus = status;
    console.log('db.setClusterStatus called: ' + JSON.stringify(status));
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setUpgradeOptions = function (upgradeOptions) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.StackUpgrade.upgradeOptions = upgradeOptions;
    console.log('db.setUpgradeOptions called: ' + JSON.stringify(upgradeOptions));
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setMasterToReassign = function (masterComponent) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.ReassignMaster.masterComponent = masterComponent;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setReassignTasksStatuses = function (tasksStatuses) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.ReassignMaster.tasksStatuses = tasksStatuses;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setStacks = function (stacks) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.app.stacksVersions = stacks;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setServices = function (services) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.app.services = services;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setConfigs = function (configs) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.app.configs = configs;
    localStorage.setObject('ambari', App.db.data);
  };

  /**
   * Set current step value for specified Wizard Type
   * @param wizardType
   * @param currentStep
   */
  App.db.setWizardCurrentStep = function (wizardType, currentStep) {
    console.log('TRACE: Entering db:setWizardCurrentStep function');
    App.db.data[wizardType.capitalize()].currentStep = currentStep;
    localStorage.setObject('ambari', App.db.data);
  };

  /**
   * Set localStorage with data from server
   */
  App.db.setLocalStorage = function () {
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSecurityWizardStatus = function (status) {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    App.db.data.AddSecurity.status = status;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setDisableSecurityStatus = function (status) {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    App.db.data.AddSecurity.disableSecurityStatus = status;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSecurityDeployCommands = function (commands) {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    App.db.data.AddSecurity.securityDeployCommands = commands;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSecureConfigProperties  = function (secureConfigs) {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    App.db.data.AddSecurity.secureConfigProperties = secureConfigs;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setSecureUserInfo  = function (userInfo) {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    App.db.data.AddSecurity.secureUserInfo = userInfo;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setIsNameNodeHa = function (haStatus) {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    App.db.data.AddSecurity.haStatus = haStatus;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardConfigTag = function (tag) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard[tag.name] = tag.value;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardHdfsClientHosts = function (hostNames) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard.hdfsClientHostNames = hostNames;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardTasksStatuses = function (tasksStatuses) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard.tasksStatuses = tasksStatuses;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardHdfsUser = function (hdfsUser) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard.hdfsUser = hdfsUser;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardRequestIds = function (requestIds) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard.requestIds = requestIds;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardLogs = function (logs) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard.logs = logs;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setHighAvailabilityWizardNameServiceId = function (nameServiceId) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.HighAvailabilityWizard.nameServiceId = nameServiceId;
    localStorage.setObject('ambari', App.db.data);
  };


  App.db.setRollBackHighAvailabilityWizardAddNNHost = function (host) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.addNNHost = host;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setRollBackHighAvailabilityWizardSNNHost = function (host) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.sNNHost = host;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setRollBackHighAvailabilityWizardSelectedAddNN = function (host) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.selectedAddNN = host;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setRollBackHighAvailabilityWizardSelectedSNN = function (host) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.selectedSNNH = host;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setRollbackHighAvailabilityWizardTasksStatuses = function (tasksStatuses) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.tasksStatuses = tasksStatuses;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setRollbackHighAvailabilityWizardRequestIds = function (requestIds) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.requestIds = requestIds;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setRollbackHighAvailabilityWizardLogs = function (logs) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.RollbackHighAvailabilityWizard.logs = logs;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setReassignMasterWizardRequestIds = function (requestIds) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.ReassignMaster.requestIds = requestIds;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setReassignMasterWizardLogs = function (logs) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.ReassignMaster.logs = logs;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setReassignMasterWizardComponentDir = function (componentDir) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.ReassignMaster.componentDir = componentDir;
    localStorage.setObject('ambari', App.db.data);
  };

  App.db.setReassignMasterWizardReassignHosts = function (reassignHosts) {
    App.db.data = localStorage.getObject('ambari');
    App.db.data.ReassignMaster.reassignHosts = reassignHosts;
    localStorage.setObject('ambari', App.db.data);
  };

  /*
   *  getter methods
   */

  /**
   * Get user model from db
   * @return {*}
   */
  App.db.getUser = function () {
    console.log('TRACE: Entering db:getUser function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.user;
  };

  App.db.getLoginName = function () {
    console.log('Trace: Entering db:getLoginName function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.loginName;
  };

  App.db.getAuthenticated = function () {
    console.log('Trace: Entering db:getAuthenticated function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.authenticated;
  };

  App.db.getAmbariStacks = function () {
    console.log('TRACE: Entering db:setAmbariStacks function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.stacks;
  };

  App.db.getFilterConditions = function(name) {
    console.log('TRACE: Entering db:getFilterConditions function');
    App.db.data = localStorage.getObject('ambari');
    if (App.db.data.app.tables.filterConditions[name]) {
      return App.db.data.app.tables.filterConditions[name];
    }
    return null;
  };

  App.db.getDisplayLength = function(name) {
    console.log('TRACE: Entering db:getDisplayLength function');
    App.db.data = localStorage.getObject('ambari');
    if (App.db.data.app.tables.displayLength[name]) {
      return App.db.data.app.tables.displayLength[name];
    }
    return null;
  };

  App.db.getStartIndex = function(name) {
    console.log('TRACE: Entering db:getStartIndex function');
    App.db.data = localStorage.getObject('ambari');
    if (App.db.data.app.tables.startIndex[name]) {
      return App.db.data.app.tables.startIndex[name];
    }
    return null;
  };

  App.db.getSortingStatuses = function(name) {
    console.log('TRACE: Entering db:getSortingConditions function');
    App.db.data = localStorage.getObject('ambari');
    if (App.db.data.app.tables.sortingConditions[name]) {
      return App.db.data.app.tables.sortingConditions[name];
    }
    return null;
  };

  /**
   * Return current step for specified Wizard Type
   * @param wizardType
   * @return {*}
   */
  App.db.getWizardCurrentStep = function (wizardType) {
    console.log('Trace: Entering db:getWizardCurrentStep function for ', wizardType);
    if (App.db.data[wizardType.capitalize()]) {
      return App.db.data[wizardType.capitalize()].currentStep;
    }
    return 0;
  };

  App.db.getAllHostNames = function () {
    console.log('TRACE: Entering db:getHostNames function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.hostNames;
  };

  App.db.getAllHostNamesPattern = function () {
    console.log('TRACE: Entering db:getHostNamesPattern function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.hostNamesPattern;
  };
  App.db.getInstallOptions = function () {
    console.log('TRACE: Entering db:getInstallOptions function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.installOptions;
  };

  App.db.isCompleted = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.completed;
  };

  App.db.getHosts = function () {
    console.log('TRACE: Entering db:getHosts function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.hostInfo;
  };

  App.db.getBootStatus = function () {
    console.log('TRACE: Entering db:getBootStatus function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.bootStatus;
  };

  App.db.getService = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.serviceInfo;
  };

  App.db.getSelectedServiceNames = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.selectedServiceNames;
  };

  App.db.getClientsForSelectedServices = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.clientInfo;
  };

  App.db.getMasterComponentHosts = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.masterComponentHosts;
  };

  App.db.getSlaveComponentHosts = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.slaveComponentHosts;
  };

  App.db.getServiceConfigs = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.serviceConfigs;
  };

  App.db.getAdvancedServiceConfig = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.advanceServiceConfigs;
  };

  App.db.getServiceConfigProperties = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.configProperties;
  };

  App.db.getCluster = function () {
    console.log('TRACE: Entering db:getClusterStatus function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.Installer.clusterStatus;
  };

  App.db.getMasterToReassign = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.ReassignMaster.masterComponent;
  };

  App.db.getReassignTasksStatuses = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.ReassignMaster.tasksStatuses;
  };

  App.db.getUpgradeOptions = function () {
    console.log('TRACE: Entering db:getUpgradeOptions function');
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.StackUpgrade.upgradeOptions;
  };

  App.db.getSecurityWizardStatus = function () {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    return App.db.data.AddSecurity.status;
  };

  App.db.getDisableSecurityStatus = function () {
    App.db.data = localStorage.getObject('ambari');
    if (!App.db.data.AddSecurity) {
      App.db.data.AddSecurity = {};
    }
    return App.db.data.AddSecurity.disableSecurityStatus;
  };

  App.db.getSecurityDeployCommands = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.AddSecurity.securityDeployCommands;
  };

  App.db.getSecureConfigProperties = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.AddSecurity.secureConfigProperties;
  };

  App.db.getSecureUserInfo  = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.AddSecurity.secureUserInfo;
  };

  App.db.getIsNameNodeHa = function (haStatus) {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.AddSecurity.haStatus;
  };

  App.db.getStacks = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.stacksVersions;
  };

  App.db.getServices = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.services;
  };

  App.db.getHighAvailabilityWizardHdfsUser = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.hdfsUser;
  };

  App.db.getHighAvailabilityWizardTasksStatuses = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.tasksStatuses;
  };

  App.db.getHighAvailabilityWizardFailedTask = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.failedTask;
  };

  App.db.getHighAvailabilityWizardHdfsClientHosts = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.hdfsClientHostNames;
  };

  App.db.getHighAvailabilityWizardConfigTag = function (tag) {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard[tag];
  };

  App.db.getHighAvailabilityWizardRequestIds = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.requestIds;
  };

  App.db.getHighAvailabilityWizardLogs = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.logs;
  };

  App.db.getHighAvailabilityWizardNameServiceId = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.HighAvailabilityWizard.nameServiceId;
  };

  App.db.getRollbackHighAvailabilityWizardTasksStatuses = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.RollbackHighAvailabilityWizard.tasksStatuses;
  };

  App.db.getRollbackHighAvailabilityWizardRequestIds = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.RollbackHighAvailabilityWizard.requestIds;
  };

  App.db.getRollbackHighAvailabilityWizardLogs = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.RollbackHighAvailabilityWizard.logs;
  };

  App.db.getRollBackHighAvailabilityWizardAddNNHost = function (host) {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.RollbackHighAvailabilityWizard.addNNHost;
  };

  App.db.getRollBackHighAvailabilityWizardSNNHost = function (host) {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.RollbackHighAvailabilityWizard.sNNHost;
  };

  App.db.getReassignMasterWizardRequestIds = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.ReassignMaster.requestIds;
  };

  App.db.getReassignMasterWizardLogs = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.ReassignMaster.logs;
  };

  App.db.getReassignMasterWizardComponentDir = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.ReassignMaster.componentDir;
  };

  App.db.getConfigs = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.app.configs;
  };

  App.db.getReassignMasterWizardReassignHosts = function () {
    App.db.data = localStorage.getObject('ambari');
    return App.db.data.ReassignMaster.reassignHosts;
  };

  module.exports = App.db;
  
});
window.require.register("utils/ember_reopen", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   Merge the contents of two objects together into the first object.

   ```javascript
   Ember.merge({first: 'Tom'}, {last: 'Dale'}); // {first: 'Tom', last: 'Dale'}
   var a = {first: 'Yehuda'}, b = {last: 'Katz'};
   Ember.merge(a, b); // a == {first: 'Yehuda', last: 'Katz'}, b == {last: 'Katz'}
   ```

   @method merge
   @for Ember
   @param {Object} original The object to merge into
   @param {Object} updates The object to copy properties from
   @return {Object}
   */
  Ember.merge = function(original, updates) {
    for (var prop in updates) {
      if (!updates.hasOwnProperty(prop)) { continue; }
      original[prop] = updates[prop];
    }
    return original;
  };

  /**
   Returns true if the passed value is null or undefined. This avoids errors
   from JSLint complaining about use of ==, which can be technically
   confusing.

   ```javascript
   Ember.isNone();              // true
   Ember.isNone(null);          // true
   Ember.isNone(undefined);     // true
   Ember.isNone('');            // false
   Ember.isNone([]);            // false
   Ember.isNone(function() {});  // false
   ```

   @method isNone
   @for Ember
   @param {Object} obj Value to test
   @return {Boolean}
   */
  Ember.isNone = function(obj) {
    return obj === null || obj === undefined;
  };

  /**
   Verifies that a value is `null` or an empty string, empty array,
   or empty function.

   Constrains the rules on `Ember.isNone` by returning false for empty
   string and empty arrays.

   ```javascript
   Ember.isEmpty();                // true
   Ember.isEmpty(null);            // true
   Ember.isEmpty(undefined);       // true
   Ember.isEmpty('');              // true
   Ember.isEmpty([]);              // true
   Ember.isEmpty('Adam Hawkins');  // false
   Ember.isEmpty([0,1,2]);         // false
   ```

   @method isEmpty
   @for Ember
   @param {Object} obj Value to test
   @return {Boolean}
   */
  Ember.isEmpty = function(obj) {
    return Ember.isNone(obj) || (obj.length === 0 && typeof obj !== 'function') || (typeof obj === 'object' && Ember.get(obj, 'length') === 0);
  };

  /**
   A value is blank if it is empty or a whitespace string.

   ```javascript
   Ember.isBlank();                // true
   Ember.isBlank(null);            // true
   Ember.isBlank(undefined);       // true
   Ember.isBlank('');              // true
   Ember.isBlank([]);              // true
   Ember.isBlank('\n\t');          // true
   Ember.isBlank('  ');            // true
   Ember.isBlank({});              // false
   Ember.isBlank('\n\t Hello');    // false
   Ember.isBlank('Hello world');   // false
   Ember.isBlank([1,2,3]);         // false
   ```

   @method isBlank
   @for Ember
   @param {Object} obj Value to test
   @return {Boolean}
   */
  Ember.isBlank = function(obj) {
    return Ember.isEmpty(obj) || (typeof obj === 'string' && obj.match(/\S/) === null);
  };

  Em.View.reopen({
    /**
     * overwritten set method of Ember.View to avoid uncaught errors
     * when trying to set property of destroyed view
     */
    set: function(attr, value){
      if(!this.get('isDestroyed') && !this.get('isDestroying')){
        this._super(attr, value);
      } else {
        console.debug('Calling set on destroyed view');
      }
    }
  });
  
});
window.require.register("utils/graph", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  module.exports = {
    durationFormatter:function(d) {
        if (d==0) { return "0" }
        var seconds = Math.floor(parseInt(d) / 1000);
        if ( seconds < 60 )
          return seconds + "s";
        var minutes = Math.floor(seconds / 60);
        if ( minutes < 60 ) {
          var x = seconds - 60*minutes;
          return minutes + "m" + (x==0 ? "" : " " + x + "s");
        }
        var hours = Math.floor(minutes / 60);
        if ( hours < 24 ) {
          var x = minutes - 60*hours;
          return hours + "h" + (x==0 ? "" : " " + x + "m");
        }
        var days = Math.floor(hours / 24);
        if ( days < 7 ) {
          var x = hours - 24*days;
          return days + "d " + (x==0 ? "" : " " + x + "h");
        }
        var weeks = Math.floor(days / 7);
        var x = days - 7*weeks;
        return weeks + "w " + (x==0 ? "" : " " + x + "d");
    },
    bytesFormatter:function(y) {
      if (y >= 1125899906842624)  { return Math.floor(10 * y / 1125899906842624)/10 + " PB" }
      else if (y >= 1099511627776){ return Math.floor(10 * y / 1099511627776)/10 + " TB" }
      else if (y >= 1073741824)   { return Math.floor(10 * y / 1073741824)/10 + " GB" }
      else if (y >= 1048576)      { return Math.floor(10 * y / 1048576)/10 + " MB" }
      else if (y >= 1024)         { return Math.floor(10 * y / 1024)/10 + " KB" }
      else                        { return y + " B"}
    },
    addSeries:function(svgg,series,color,xscale,yscale,margin,startTime,dotInfo) {
      if (series.length==0) return;
      var self = this;
      var g = svgg.append("svg:g").selectAll("g")
        .data(series)
        .enter().append("svg:g")
        .attr("transform", "translate(0,"+margin+")");
      g.append("svg:circle")
        .attr("r",function(d) {return d.r;})
        .attr("cx",function(d) {return xscale(d.x);})
        .attr("cy",function(d) {return yscale(d.y);})
        .style("fill",color)
        .style("fill-opacity",0.8)
        .style("stroke",d3.interpolateRgb(color, 'black')(0.125))
        .append("title")
        .text(function(d) { return dotInfo[Math.round(xscale(d.x))][Math.round(yscale(d.y))]; });
      g.append("svg:line")
        .attr("x1", function(d) { return xscale(d.x)+d.r; } )
        .attr("x2", function(d) { return xscale(d.x+d.y); } )
        .attr("y1", function(d) { return yscale(d.y); } )
        .attr("y2", function(d) { return yscale(d.y); } )
        .style("stroke",d3.interpolateRgb(color, 'black')(0.125))
        .style("stroke-width",2)
        .append("title")
        .text(function(d) { return dotInfo[Math.round(xscale(d.x))][Math.round(yscale(d.y))]; });
    },
    /**
     *
     * @param mapNodeLocal
     * @param mapRackLocal
     * @param mapOffSwitch
     * @param reduceOffSwitch
     * @param startTime
     * @param endTime
     * @param svgw
     * @param svgh
     * @param element
     */
    drawJobTasks:function (mapNodeLocal, mapRackLocal, mapOffSwitch, reduceOffSwitch, startTime, endTime, svgw, svgh, element) {
      var rmax = 24; // default value
      var axisHeight = 24;
      var margin = {"vertical":10, "horizontal":50};
      var w = svgw - 2*margin.horizontal;
      var h = svgh - 2*margin.vertical;
      var x = d3.time.scale.utc()
        .domain([startTime, endTime])
        .range([0, w]);
      var xrel = d3.time.scale()
        .domain([0, endTime-startTime])
        .range([0, w]);
      // create axes
      var topAxis = d3.svg.axis()
        .scale(x)
        .orient("bottom");
      var self = this;
      var bottomAxis = d3.svg.axis()
        .scale(xrel)
        .orient("bottom")
        .tickFormat(function(d) {return self.durationFormatter(d.getTime())});

      var svg = d3.select("div#" + element).append("svg:svg")
        .attr("width", svgw+"px")
        .attr("height", svgh+"px");
      var svgg = svg.append("g")
        .attr("transform", "translate("+margin.horizontal+","+margin.vertical+")");

      svgg.append("g")
        .attr("class", "x axis top")
        .call(topAxis);
      svgg.append("g")
        .attr("class", "x axis bottom")
        .call(bottomAxis)
        .attr("transform", "translate(0,"+(h-axisHeight)+")");

      var ymax = 0;
      if (mapNodeLocal.length > 0)
        ymax = Math.max(ymax, d3.max(mapNodeLocal, function(d) { return d.y; } ));
      if (mapRackLocal.length > 0)
        ymax = Math.max(ymax, d3.max(mapRackLocal, function(d) { return d.y; } ));
      if (mapOffSwitch.length > 0)
        ymax = Math.max(ymax, d3.max(mapOffSwitch, function(d) { return d.y; } ));
      if (reduceOffSwitch.length > 0)
        ymax = Math.max(ymax, d3.max(reduceOffSwitch, function(d) { return d.y; } ));

      var y = d3.scale.linear()
        .domain([0, ymax])
        .range([h-2*axisHeight-rmax, 0]);

      var yAxis = d3.svg.axis()
        .scale(y)
        .orient("left")
        .tickFormat(self.durationFormatter);
   
      svgg.append("svg:g")
        .attr("class", "y axis")
        .call(yAxis)
        .attr("transform", "translate(0,"+(axisHeight+rmax)+")")
        .append("text")
        .attr("transform", "rotate(-90)")
        .attr("x", -(h-2*axisHeight-rmax)/2)
        .attr("y", -margin.horizontal + 11)
        .attr("class", "axislabel")
        .text("Task Attempt Duration");


      var dotInfo = [];
      var mapDotInfo = function(d) {
        var thisx = Math.round(x(d.x));
        var thisy = Math.round(y(d.y));
        if (!(thisx in dotInfo))
          dotInfo[thisx] = [];
        var existing = dotInfo[thisx][thisy];
        var newInfo = d.label + "  \n" +
            'Run-time: ' + self.durationFormatter(d.y) + '  \nWait-time: ' + self.durationFormatter(d.x-startTime) +
            '  \nI/O: ' + self.bytesFormatter(d.io) + '  \nStatus: ' + d.status;
        if (existing)
          dotInfo[thisx][thisy] = existing + "  \n" + newInfo;
        else
          dotInfo[thisx][thisy] = newInfo;
      };

      mapNodeLocal.forEach(mapDotInfo);
      mapRackLocal.forEach(mapDotInfo);
      mapOffSwitch.forEach(mapDotInfo);
      reduceOffSwitch.forEach(mapDotInfo);

      this.addSeries(svgg, mapNodeLocal, "green", x, y, axisHeight+rmax, startTime, dotInfo);
      this.addSeries(svgg, mapRackLocal,'#66B366', x, y, axisHeight+rmax, startTime, dotInfo);
      this.addSeries(svgg, mapOffSwitch, 'brown', x, y, axisHeight+rmax, startTime, dotInfo);
      this.addSeries(svgg, reduceOffSwitch, 'steelblue', x, y, axisHeight+rmax, startTime, dotInfo);
    }
  };
  
});
window.require.register("utils/heatmap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  module.exports = {
    mappers: Em.Mixin.create({
      metricMapperWithTransform: function (json, metricName, transformValueFunction) {
        var hostToValueMap = {};
        //var metricName = this.get('defaultMetric');
        if (json.host_components) {
          var props = metricName.split('.');
          //var transformValueFunction = this.get('transformValue');
          json.host_components.forEach(function (hc) {
            var value = hc;
            props.forEach(function (prop) {
              if (value != null && prop in value) {
                value = value[prop];
              } else {
                value = null;
              }
            });
            if (value != null) {
              if (transformValueFunction) {
                value = transformValueFunction(value);
              }
              var hostName = hc.HostRoles.host_name;
              hostToValueMap[hostName] = value;
            }
          });
        }
        return hostToValueMap;
      }
    })
  };
  
});
window.require.register("utils/helper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  String.prototype.trim = function () {
    return this.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
  };

  String.prototype.endsWith = function(suffix) {
    return this.indexOf(suffix, this.length - suffix.length) !== -1;
  };

  String.prototype.startsWith = function (prefix){
    return this.indexOf(prefix) == 0;
  };

  String.prototype.contains = function(substring) {
    return this.indexOf(substring) != -1;
  };

  String.prototype.capitalize = function () {
    return this.charAt(0).toUpperCase() + this.slice(1);
  };

  /**
   * Finds the value in an object where this string is a key.
   * Optionally, the index of the key can be provided where the
   * value of the nth key in the hierarchy is returned.
   *
   * Example:
   *  var tofind = 'smart';
   *  var person = {'name': 'Bob Bob', 'smart': 'no', 'age': '28', 'personality': {'smart': 'yes', 'funny': 'yes', 'emotion': 'happy'} };
   *  tofind.findIn(person); // 'no'
   *  tofind.findIn(person, 0); // 'no'
   *  tofind.findIn(person, 1); // 'yes'
   *  tofind.findIn(person, 2); // null
   *
   *  @param multi  Object
   *  @param index  Occurrence count of this key
   *  @return Value of key at given index
   */
  String.prototype.findIn = function(multi, index, _foundValues) {
    if (!index) {
      index = 0;
    }
    if (!_foundValues) {
      _foundValues = [];
    }
    multi = multi || '';
    var value = null;
    var str = this.valueOf();
    if (typeof multi == 'object') {
      for ( var key in multi) {
        if (value != null) {
          break;
        }
        if (key == str) {
          _foundValues.push(multi[key]);
        }
        if (_foundValues.length - 1 == index) {
          // Found the value
          return _foundValues[index];
        }
        if (typeof multi[key] == 'object') {
          value = value || this.findIn(multi[key], index, _foundValues);
        }
      }
    }
    return value;
  };

  /**
   * Replace {i} with argument. where i is number of argument to replace with
   * @return {String}
   */
  String.prototype.format = function () {
    var args = arguments;
    return this.replace(/{(\d+)}/g, function (match, number) {
      return typeof args[number] != 'undefined' ? args[number] : match;
    });
  };

  String.prototype.highlight = function (words, highlightTemplate) {
    var self = this;
    highlightTemplate = highlightTemplate ? highlightTemplate : "<b>{0}</b>";

    words.forEach(function (word) {
      var searchRegExp = new RegExp("\\b" + word + "\\b", "gi");
      self = self.replace(searchRegExp, function (found) {
        return highlightTemplate.format(found);
      });
    });

    return self;
  };

  Number.prototype.toDaysHoursMinutes = function () {
    var formatted = {},
      dateDiff = this,
      secK = 1000, //ms
      minK = 60 * secK, // sec
      hourK = 60 * minK, // sec
      dayK = 24 * hourK;

    dateDiff = parseInt(dateDiff);
    formatted.d = Math.floor(dateDiff / dayK);
    dateDiff -= formatted.d * dayK;
    formatted.h = Math.floor(dateDiff / hourK);
    dateDiff -= formatted.h * hourK;
    formatted.m = (dateDiff / minK).toFixed(2);

    return formatted;
  };

  /**
   Sort an array by the key specified in the argument.
   Handle only native js objects as element of array, not the Ember's object.

   Can be used as alternative to sortProperty method of Ember library
   in order to speed up executing on large data volumes

   @method sortBy
   @param {String} path name(s) to sort on
   @return {Array} The sorted array.
   */
  Array.prototype.sortPropertyLight = function (path) {
    var realPath = (typeof path === "string") ? path.split('.') : [];
    this.sort(function (a, b) {
      var aProperty = a;
      var bProperty = b;
      realPath.forEach(function (key) {
        aProperty = aProperty[key];
        bProperty = bProperty[key];
      });
      if (aProperty > bProperty) return 1;
      if (aProperty < bProperty) return -1;
      return 0;
    });
    return this;
  };

  Em.CoreObject.reopen({
    t:function (key, attrs) {
      return Em.I18n.t(key, attrs)
    }
  });

  Em.Handlebars.registerHelper('log', function (variable) {
    console.log(variable);
  });

  Em.Handlebars.registerHelper('warn', function (variable) {
    console.warn(variable);
  });

  Em.Handlebars.registerHelper('highlight', function (property, words, fn) {
    var context = (fn.contexts && fn.contexts[0]) || this;
    property = Em.Handlebars.getPath(context, property, fn);

    words = words.split(";");

  //  if (highlightTemplate == undefined) {
    var highlightTemplate = "<b>{0}</b>";
  //  }

    words.forEach(function (word) {
      var searchRegExp = new RegExp("\\b" + word + "\\b", "gi");
      property = property.replace(searchRegExp, function (found) {
        return highlightTemplate.format(found);
      });
    });

    return new Em.Handlebars.SafeString(property);
  });

  App = require('app');

  /**
   * Certain variables can have JSON in string
   * format, or in JSON format itself.
   */
  App.parseJSON = function (value) {
    if (typeof value == "string") {
      return jQuery.parseJSON(value);
    }
    return value;
  };
  /**
   * Check for empty <code>Object</code>, built in Em.isEmpty()
   * doesn't support <code>Object</code> type
   *
   * @params obj {Object}
   *
   * @return {Boolean}
   */
  App.isEmptyObject = function(obj) {
    var empty = true;
    for (var prop in obj) { if (obj.hasOwnProperty(prop)) {empty = false; break;} }
    return empty;
  }

  App.format = {

    /**
     * @type Object
     */
    components: {
      'APP_TIMELINE_SERVER': 'App Timeline Server',
      'DATANODE': 'DataNode',
      'DECOMMISSION_DATANODE': 'Update Exclude File',
      'DRPC_SERVER': 'DRPC Server',
      'FALCON': 'Falcon',
      'FALCON_CLIENT': 'Falcon Client',
      'FALCON_SERVER': 'Falcon Server',
      'FALCON_SERVICE_CHECK': 'Falcon Service Check',
      'FLUME_SERVER': 'Flume Agent',
      'GANGLIA_MONITOR': 'Ganglia Monitor',
      'GANGLIA_SERVER': 'Ganglia Server',
      'GLUSTERFS_CLIENT': 'GLUSTERFS Client',
      'GLUSTERFS_SERVICE_CHECK': 'GLUSTERFS Service Check',
      'GMETAD_SERVICE_CHECK': 'Gmetad Service Check',
      'GMOND_SERVICE_CHECK': 'Gmond Service Check',
      'HADOOP_CLIENT': 'Hadoop Client',
      'HBASE_CLIENT': 'HBase Client',
      'HBASE_MASTER': 'HBase Master',
      'HBASE_REGIONSERVER': 'HBase RegionServer',
      'HBASE_SERVICE_CHECK': 'HBase Service Check',
      'HCAT': 'HCat',
      'HCAT_SERVICE_CHECK': 'HCat Service Check',
      'HDFS_CLIENT': 'HDFS Client',
      'HDFS_SERVICE_CHECK': 'HDFS Service Check',
      'HISTORYSERVER': 'History Server',
      'HIVE_CLIENT': 'Hive Client',
      'HIVE_METASTORE': 'Hive Metastore',
      'HIVE_SERVER': 'HiveServer2',
      'HIVE_SERVICE_CHECK': 'Hive Service Check',
      'HUE_SERVER': 'Hue Server',
      'JAVA_JCE': 'Java JCE',
      'JOBTRACKER': 'JobTracker',
      'JOBTRACKER_SERVICE_CHECK': 'JobTracker Service Check',
      'JOURNALNODE': 'JournalNode',
      'KERBEROS_ADMIN_CLIENT': 'Kerberos Admin Client',
      'KERBEROS_CLIENT': 'Kerberos Client',
      'KERBEROS_SERVER': 'Kerberos Server',
      'MAPREDUCE2_CLIENT': 'MapReduce2 Client',
      'MAPREDUCE2_SERVICE_CHECK': 'MapReduce2 Service Check',
      'MAPREDUCE_CLIENT': 'MapReduce Client',
      'MAPREDUCE_SERVICE_CHECK': 'MapReduce Service Check',
      'MYSQL_SERVER': 'MySQL Server',
      'NAGIOS_SERVER': 'Nagios Server',
      'NAMENODE': 'NameNode',
      'NAMENODE_SERVICE_CHECK': 'NameNode Service Check',
      'NIMBUS': 'Nimbus',
      'NODEMANAGER': 'NodeManager',
      'OOZIE_CLIENT': 'Oozie Client',
      'OOZIE_SERVER': 'Oozie Server',
      'OOZIE_SERVICE_CHECK': 'Oozie Service Check',
      'PIG': 'Pig',
      'PIG_SERVICE_CHECK': 'Pig Service Check',
      'RESOURCEMANAGER': 'ResourceManager',
      'SECONDARY_NAMENODE': 'SNameNode',
      'SQOOP': 'Sqoop',
      'SQOOP_SERVICE_CHECK': 'Sqoop Service Check',
      'STORM_REST_API': 'Storm REST API Server',
      'STORM_SERVICE_CHECK': 'Storm Service Check',
      'STORM_UI_SERVER': 'Storm UI Server',
      'SUPERVISOR': 'Supervisor',
      'TASKTRACKER': 'TaskTracker',
      'TEZ_CLIENT': 'Tez Client',
      'WEBHCAT_SERVER': 'WebHCat Server',
      'WEBHCAT_SERVICE_CHECK': 'WebHCat Service Check',
      'YARN_CLIENT': 'YARN Client',
      'YARN_SERVICE_CHECK': 'YARN Service Check',
      'ZKFC': 'ZKFailoverController',
      'ZOOKEEPER_CLIENT': 'ZooKeeper Client',
      'ZOOKEEPER_QUORUM_SERVICE_CHECK': 'ZK Quorum Service Check',
      'ZOOKEEPER_SERVER': 'ZooKeeper Server',
      'ZOOKEEPER_SERVICE_CHECK': 'ZooKeeper Service Check',
      'CLIENT': 'client'
    },

    /**
     * @type Object
     */
    command: {
      'INSTALL': 'Install',
      'UNINSTALL': 'Uninstall',
      'START': 'Start',
      'STOP': 'Stop',
      'EXECUTE': 'Execute',
      'ABORT': 'Abort',
      'UPGRADE': 'Upgrade',
      'RESTART': 'Restart',
      'SERVICE_CHECK': 'Check',
      'Excluded:': 'Decommission:',
      'Included:': 'Recommission:'
    },

    /**
     * convert role to readable string
     * @param role
     */
    role:function (role) {
      return this.components[role] ? this.components[role] : '';
    },

    /**
     * convert command_detail to readable string, show the string for all tasks name
     * @param command_detail
     */
    commandDetail: function (command_detail) {
      var detailArr = command_detail.split(' ');
      var self = this;
      var result = '';
      detailArr.forEach( function(item) {
        // if the item has the pattern SERVICE/COMPONENT, drop the SERVICE part
        if (item.contains('/')) {
          item = item.split('/')[1];
        }
        // ignore 'DECOMMISSION', command came from 'excluded/included'
        if (item == 'DECOMMISSION,') {
          item = '';
        }
        if (self.components[item]) {
          result = result + ' ' + self.components[item];
        } else if (self.command[item]) {
          result = result + ' ' + self.command[item];
        } else {
          result = result + ' ' + item;
        }
      });
      if (result === ' nagios_update_ignore ACTIONEXECUTE') {
         result = Em.I18n.t('common.maintenance.task');
      }
      return result;
    },

    /**
     * PENDING - Not queued yet for a host
     * QUEUED - Queued for a host
     * IN_PROGRESS - Host reported it is working
     * COMPLETED - Host reported success
     * FAILED - Failed
     * TIMEDOUT - Host did not respond in time
     * ABORTED - Operation was abandoned
     */
    taskStatus:function (_taskStatus) {
      return _taskStatus.toLowerCase();
    }
  };

  /**
   * wrapper to bootstrap popover
   * fix issue when popover stuck on view routing
   * @param self
   * @param options
   */
  App.popover = function(self, options) {
    self.popover(options);
    self.on("remove", function () {
      $(this).trigger('mouseleave');
    });
  };

  /**
   * wrapper to bootstrap tooltip
   * fix issue when tooltip stuck on view routing
   * @param self - DOM element
   * @param options
   */
  App.tooltip = function(self, options) {
    self.tooltip(options);
    self.on("remove", function () {
      $(this).trigger('mouseleave');
    });
  };

  /**
   * wrapper to Date().getTime()
   * fix issue when client clock and server clock not sync
   * @return timeStamp of current server clock
   */
  App.dateTime = function() {
    return new Date().getTime() + App.clockDistance;
  };

  /*
   * Helper function for bound property helper registration
   * @params name {String} - name of helper
   * @params view {Em.View} - view
   */
  App.registerBoundHelper = function(name, view) {
    Em.Handlebars.registerHelper(name, function(property, options) {
      options.hash.contentBinding = property;
      return Em.Handlebars.helpers.view.call(this, view, options);
    });
  };

  /*
   * Return singular or plural word based on Em.I18n property key.
   *
   *  Example: {{pluralize hostsCount singular="t:host" plural="t:hosts"}}
   */
  App.registerBoundHelper('pluralize', Em.View.extend({
    tagName: 'span',
    template: Em.Handlebars.compile('{{view.wordOut}}'),

    wordOut: function() {
      var count, singular, plural;
      count = this.get('content');
      singular = this.get('singular');
      plural = this.get('plural');
      return this.getWord(count, singular, plural);
    }.property('content'),

    getWord: function(count, singular, plural) {
      singular = this.tDetect(singular);
      plural = this.tDetect(plural);
      if (singular && plural) {
        if (count > 1) {
          return plural;
        } else {
          return singular;
        }
      }
      return '';
    },
    /*
     * Detect for Em.I18n.t reference call
     * @params word {String}
     * return {String}
    */
    tDetect: function(word) {
      var splitted = word.split(':');
      if (splitted.length > 1 && splitted[0] == 't') {
        return Em.I18n.t(splitted[1]);
      } else {
        return splitted[0];
      }
    }
    })
  );
  /**
   * Return defined string instead of empty if value is null/undefined
   * by default is `n/a`.
   *
   * @param empty {String} - value instead of empty string (not required)
   *  can be used with Em.I18n pass value started with't:'
   *
   * Examples:
   *
   * default value will be returned
   * {{formatNull service.someValue}}
   *
   * <code>empty<code> will be returned
   * {{formatNull service.someValue empty="I'm empty"}}
   *
   * Em.I18n translation will be returned
   * {{formatNull service.someValue empty="t:my.key.to.translate"
   */
  App.registerBoundHelper('formatNull', Em.View.extend({
    tagName: 'span',
    template: Em.Handlebars.compile('{{view.result}}'),

    result: function() {
      var emptyValue = this.get('empty') ? this.get('empty') : Em.I18n.t('services.service.summary.notAvailable');
      emptyValue = emptyValue.startsWith('t:') ? Em.I18n.t(emptyValue.substr(2, emptyValue.length)) : emptyValue;
      return (this.get('content') || this.get('content') == 0) ? this.get('content') : emptyValue;
    }.property('content')
  }));

  /**
   * Return formatted string with inserted <code>wbr</code>-tag after each dot
   *
   * @param {String} content
   *
   * Examples:
   *
   * returns 'apple'
   * {{formatWordBreak 'apple'}}
   *
   * returns 'apple.<wbr />banana'
   * {{formatWordBreak 'apple.banana'}}
   *
   * returns 'apple.<wbr />banana.<wbr />uranium'
   * {{formatWordBreak 'apple.banana.uranium'}}
   */
  App.registerBoundHelper('formatWordBreak', Em.View.extend({
    tagName: 'span',
    template: Em.Handlebars.compile('{{{view.result}}}'),

    /**
     * @type {string}
     */
    result: function() {
      return this.get('content').replace(/\./g, '.<wbr />');
    }.property('content')
  }));

  /**
   * Ambari overrides the default date transformer.
   * This is done because of the non-standard data
   * sent. For example Nagios sends date as "12345678".
   * The problem is that it is a String and is represented
   * only in seconds whereas Javascript's Date needs
   * milliseconds representation.
   */
  DS.attr.transforms.date = {
    from: function (serialized) {
      var type = typeof serialized;
      if (type === Em.I18n.t('common.type.string')) {
        serialized = parseInt(serialized);
        type = typeof serialized;
      }
      if (type === Em.I18n.t('common.type.number')) {
        if (!serialized ){  //serialized timestamp = 0;
          return 0;
        }
        // The number could be seconds or milliseconds.
        // If seconds, then the length is 10
        // If milliseconds, the length is 13
        if (serialized.toString().length < 13) {
          serialized = serialized * 1000;
        }
        return new Date(serialized);
      } else if (serialized === null || serialized === undefined) {
        // if the value is not present in the data,
        // return undefined, not null.
        return serialized;
      } else {
        return null;
      }
    },
    to: function (deserialized) {
      if (deserialized instanceof Date) {
        return deserialized.getTime();
      } else if (deserialized === undefined) {
        return undefined;
      } else {
        return null;
      }
    }
  };

  DS.attr.transforms.object = {
    from: function(serialized) {
      return Ember.none(serialized) ? null : Object(serialized);
    },

    to: function(deserialized) {
      return Ember.none(deserialized) ? null : Object(deserialized);
    }
  };

  /**
   * Allows EmberData models to have array properties.
   *
   * Declare the property as <code>
   *  operations: DS.attr('array'),
   * </code> and
   * during load provide a JSON array for value.
   *
   * This transform simply assigns the same array in both directions.
   */
  DS.attr.transforms.array = {
    from : function(serialized) {
      return serialized;
    },
    to : function(deserialized) {
      return deserialized;
    }
  };
  
});
window.require.register("utils/host_progress_popup", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var batchUtils = require('utils/batch_scheduled_requests');
  var date = require('utils/date');

  /**
   * App.HostPopup is for the popup that shows up upon clicking already-performed or currently-in-progress operations
   */
  App.HostPopup = Em.Object.create({

    servicesInfo: null,
    hosts: null,
    inputData: null,

    /**
     * @type {string}
     */
    serviceName: "",

    /**
     * @type {Number}
     */
    currentServiceId: null,
    previousServiceId: null,

    /**
     * @type {string}
     */
    popupHeaderName: "",

    /**
     * @type {App.Controller}
     */
    dataSourceController: null,

    /**
     * @type {bool}
     */
    isBackgroundOperations: false,

    /**
     * @type {string}
     */
    currentHostName: null,

    /**
     * @type {App.ModalPopup}
     */
    isPopup: null,

    /**
     * Entering point of this component
     * @param {String} serviceName
     * @param {App.Controller} controller
     * @param {bool} isBackgroundOperations
     */
    initPopup: function (serviceName, controller, isBackgroundOperations) {
      if (!isBackgroundOperations) {
        this.clearHostPopup();
        this.set("popupHeaderName", serviceName);
      }
      this.set("serviceName", serviceName);
      this.set("dataSourceController", controller);
      this.set("isBackgroundOperations", isBackgroundOperations);
      this.set("inputData", this.get("dataSourceController.services"));
      if(isBackgroundOperations){
        this.onServiceUpdate();
      } else {
        this.onHostUpdate();
      }
      return this.createPopup();
    },

    /**
     * clear info popup data
     */
    clearHostPopup: function () {
      this.set('servicesInfo', null);
      this.set('hosts', null);
      this.set('inputData', null);
      this.set('serviceName', "");
      this.set('currentServiceId', null);
      this.set('previousServiceId', null);
      this.set('popupHeaderName', "");
      this.set('dataSourceController', null);
      this.set('currentHostName', null);
      this.get('isPopup') ? this.get('isPopup').remove() : null;
    },

    /**
     * Depending on tasks status
     * @param {Array} tasks
     * @return {Array} [Status, Icon type, Progressbar color, is IN_PROGRESS]
     */
    getStatus: function(tasks){
      var isCompleted = true;
      var status;
      var tasksLength = tasks.length;
      var isFailed = false;
      var isAborted = false;
      var isTimedout = false;
      var isInProgress = false;
      for (var i = 0; i < tasksLength; i++) {
        if (tasks[i].Tasks.status !== 'COMPLETED') {
          isCompleted = false;
        }
        if(tasks[i].Tasks.status === 'FAILED'){
          isFailed = true;
        }
        if (tasks[i].Tasks.status === 'ABORTED') {
          isAborted = true;
        }
        if (tasks[i].Tasks.status === 'TIMEDOUT') {
          isTimedout = true;
        }
        if (tasks[i].Tasks.status === 'IN_PROGRESS') {
          isInProgress = true;
        }
      }
      if (isFailed) {
        status = ['FAILED', 'icon-exclamation-sign', 'progress-danger', false];
      } else if (isAborted) {
        status = ['CANCELLED', 'icon-minus', 'progress-warning', false];
      } else if (isTimedout) {
        status = ['TIMEDOUT', 'icon-time', 'progress-warning', false];
      } else if (isInProgress) {
        status = ['IN_PROGRESS', 'icon-cogs', 'progress-info', true];
      }
      if(status){
        return status;
      } else if(isCompleted){
        return ['SUCCESS', 'icon-ok', 'progress-success', false];
      } else {
        return ['PENDING', 'icon-cog', 'progress-info', true];
      }
    },

    /**
     * Progress of host or service depending on tasks status
     * @param {Array} tasks
     * @return {Number} percent of completion
     */
    getProgress: function (tasks) {
      var completedActions = 0;
      var queuedActions = 0;
      var inProgressActions = 0;
      tasks.forEach(function(task) {
        if(['COMPLETED', 'FAILED', 'ABORTED', 'TIMEDOUT'].contains(task.Tasks.status)){
          completedActions++;
        } else if(task.Tasks.status === 'QUEUED'){
          queuedActions++;
        } else if(task.Tasks.status === 'IN_PROGRESS'){
          inProgressActions++;
        }
      });
      return Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / tasks.length * 100);
    },

    /**
     * Count number of operations for select box options
     * @param {Object[]} obj
     * @param {Object[]} categories
     */
    setSelectCount: function (obj, categories) {
      if (!obj) return;
      var countAll = obj.length;
      var countPending = 0;
      var countInProgress = 0;
      var countFailed = 0;
      var countCompleted = 0;
      var countAborted = 0;
      var countTimedout = 0;
      obj.forEach(function(item){
        switch (item.status){
          case 'pending':
            countPending++;
            break;
          case 'queued':
            countPending++;
            break;
          case 'in_progress':
            countInProgress++;
            break;
          case 'failed':
            countFailed++;
            break;
          case 'success':
            countCompleted++;
            break;
          case 'completed':
            countCompleted++;
            break;
          case 'aborted':
            countAborted++;
            break;
          case 'timedout':
            countTimedout++;
            break;
        }
      });

      categories.findProperty("value", 'all').set("count", countAll);
      categories.findProperty("value", 'pending').set("count", countPending);
      categories.findProperty("value", 'in_progress').set("count", countInProgress);
      categories.findProperty("value", 'failed').set("count", countFailed);
      categories.findProperty("value", 'completed').set("count", countCompleted);
      categories.findProperty("value", 'aborted').set("count", countAborted);
      categories.findProperty("value", 'timedout').set("count", countTimedout);
    },

    /**
     * For Background operation popup calculate number of running Operations, and set popup header
     * @param {bool} isServiceListHidden
     */
    setBackgroundOperationHeader: function (isServiceListHidden) {
      if (this.get('isBackgroundOperations') && !isServiceListHidden) {
        var numRunning =  App.router.get('backgroundOperationsController.allOperationsCount');
        this.set("popupHeaderName", numRunning + Em.I18n.t('hostPopup.header.postFix'));
      }
    },

    /**
     * Create services obj data structure for popup
     * Set data for services
     * @param {bool} isServiceListHidden
     */
    onServiceUpdate: function (isServiceListHidden) {
      if (this.get('isBackgroundOperations') && this.get("inputData")) {
        var self = this;
        var allNewServices = [];
        var statuses = {
          'FAILED': ['FAILED', 'icon-exclamation-sign', 'progress-danger', false],
          'ABORTED': ['CANCELLED', 'icon-minus', 'progress-warning', false],
          'TIMEDOUT': ['TIMEDOUT', 'icon-time', 'progress-warning', false],
          'IN_PROGRESS': ['IN_PROGRESS', 'icon-cogs', 'progress-info', true],
          'COMPLETED': ['SUCCESS', 'icon-ok', 'progress-success', false]
        };
        var pendingStatus = ['PENDING', 'icon-cog', 'progress-info', true];
        this.set("servicesInfo", null);
        this.get("inputData").forEach(function (service) {
          var status = statuses[service.status] || pendingStatus;
          var newService = Ember.Object.create({
            id: service.id,
            displayName: service.displayName,
            progress: service.progress,
            status: App.format.taskStatus(status[0]),
            isRunning: service.isRunning,
            name: service.name,
            isVisible: true,
            startTime: date.startTime(service.startTime),
            duration: date.durationSummary(service.startTime, service.endTime),
            icon: status[1],
            barColor: status[2],
            isInProgress: status[3],
            barWidth: "width:" + service.progress + "%;",
            sourceRequestScheduleId: service.get('sourceRequestScheduleId'),
            contextCommand: service.get('contextCommand')
          });
          allNewServices.push(newService);
        });
        self.set('servicesInfo', allNewServices);
        this.setBackgroundOperationHeader(isServiceListHidden);
      }
    },

    /**
     * create task Ember object
     * @param {Object} _task
     * @return {Em.Object}
     */
    createTask: function (_task) {
      return Em.Object.create({
        id: _task.Tasks.id,
        hostName: _task.Tasks.host_name,
        command: ( _task.Tasks.command.toLowerCase() != 'service_check') ? _task.Tasks.command.toLowerCase() : '',
        commandDetail: App.format.commandDetail(_task.Tasks.command_detail),
        status: App.format.taskStatus(_task.Tasks.status),
        role: App.format.role(_task.Tasks.role),
        outputLog: Em.I18n.t('common.hostLog.popup.logDir.path') + Em.I18n.t('common.hostLog.popup.outputLog.value').format(_task.Tasks.id),
        errorLog: Em.I18n.t('common.hostLog.popup.logDir.path') + Em.I18n.t('common.hostLog.popup.errorLog.value').format(_task.Tasks.id),
        stderr: _task.Tasks.stderr,
        stdout: _task.Tasks.stdout,
        isVisible: true,
        startTime: date.startTime(_task.Tasks.start_time),
        duration: date.durationSummary(_task.Tasks.start_time, _task.Tasks.end_time),
        icon: function () {
          var statusIconMap = {
            'pending': 'icon-cog',
            'queued': 'icon-cog',
            'in_progress': 'icon-cogs',
            'completed': 'icon-ok',
            'failed': 'icon-exclamation-sign',
            'aborted': 'icon-minus',
            'timedout': 'icon-time'
          };
          return statusIconMap[this.get('status')] || 'icon-cog';
        }.property('status')
      });
    },

    /**
     * Create hosts and tasks data structure for popup
     * Set data for hosts and tasks
     */
    onHostUpdate: function () {
      var self = this;
      var inputData = this.get("inputData");
      if (inputData) {
        var hostsArr = [];
        var hostsData;
        var hostsMap = {};

        if(this.get('isBackgroundOperations') && this.get("currentServiceId")){
          //hosts popup for Background Operations
          hostsData = inputData.findProperty("id", this.get("currentServiceId"));
        } else if (this.get("serviceName")) {
          //hosts popup for Wizards
          hostsData = inputData.findProperty("name", this.get("serviceName"));
        }
        if (hostsData) {
          if (hostsData.hostsMap) {
            //hosts data come from Background Operations as object map
            hostsMap = hostsData.hostsMap;
          } else if (hostsData.hosts) {
            //hosts data come from Wizard as array
            hostsData.hosts.forEach(function (_host) {
              hostsMap[_host.name] = _host;
            });
          }
        }
        var existedHosts = self.get('hosts');

        if (existedHosts && (existedHosts.length > 0) && this.get('currentServiceId') === this.get('previousServiceId')) {
          existedHosts.forEach(function (host) {
            var newHostInfo = hostsMap[host.get('name')];
            //update only hosts with changed tasks or currently opened tasks of host
            if (newHostInfo && (newHostInfo.isModified || this.get('currentHostName') === host.get('name'))) {
              var hostStatus = self.getStatus(newHostInfo.logTasks);
              var hostProgress = self.getProgress(newHostInfo.logTasks);
              host.set('status', App.format.taskStatus(hostStatus[0]));
              host.set('icon', hostStatus[1]);
              host.set('barColor', hostStatus[2]);
              host.set('isInProgress', hostStatus[3]);
              host.set('progress', hostProgress);
              host.set('barWidth', "width:" + hostProgress + "%;");
              host.set('logTasks', newHostInfo.logTasks);
              var existTasks = host.get('tasks');
              if (existTasks) {
                newHostInfo.logTasks.forEach(function (_task) {
                  var existTask = existTasks.findProperty('id', _task.Tasks.id);
                  if (existTask) {
                    existTask.set('status', App.format.taskStatus(_task.Tasks.status));
                    existTask.set('stdout', _task.Tasks.stdout);
                    existTask.set('stderr', _task.Tasks.stderr);
                    existTask.set('startTime', date.startTime(_task.Tasks.start_time));
                    existTask.set('duration', date.durationSummary(_task.Tasks.start_time, _task.Tasks.end_time));
                  } else {
                    existTasks.pushObject(this.createTask(_task));
                  }
                }, this);
              }
            }
          }, this);
        } else {
          for (var hostName in hostsMap) {
            var _host = hostsMap[hostName];
            var tasks = _host.logTasks;
            var hostInfo = Ember.Object.create({
              name: hostName,
              publicName: _host.publicName,
              displayName: function () {
                return this.get('name').length < 43 ? this.get('name') : (this.get('name').substr(0, 40) + '...');
              }.property('name'),
              progress: 0,
              status: App.format.taskStatus("PENDING"),
              serviceName: _host.serviceName,
              isVisible: true,
              icon: "icon-cog",
              barColor: "progress-info",
              barWidth: "width:0%;"
            });

            if (tasks.length) {
              tasks = tasks.sortProperty('Tasks.id');
              var hostStatus = self.getStatus(tasks);
              var hostProgress = self.getProgress(tasks);
              hostInfo.set('status', App.format.taskStatus(hostStatus[0]));
              hostInfo.set('icon', hostStatus[1]);
              hostInfo.set('barColor', hostStatus[2]);
              hostInfo.set('isInProgress', hostStatus[3]);
              hostInfo.set('progress', hostProgress);
              hostInfo.set('barWidth', "width:" + hostProgress + "%;");
            }
            hostInfo.set('logTasks', tasks);
            hostsArr.push(hostInfo);
          }

          hostsArr = hostsArr.sortProperty('name');
          hostsArr.setEach("serviceName", this.get("serviceName"));
          self.set("hosts", hostsArr);
          self.set('previousServiceId', this.get('currentServiceId'));
        }
      }
    },

    /**
     * Show popup
     * @return {App.ModalPopup} PopupObject For testing purposes
     */
    createPopup: function () {
      var self = this;
      var hostsInfo = this.get("hosts");
      var servicesInfo = this.get("servicesInfo");
      var isBackgroundOperations = this.get('isBackgroundOperations');
      var categoryObject = Em.Object.extend({
        value: '',
        count: 0,
        labelPath: '',
        label: function(){
          return Em.I18n.t(this.get('labelPath')).format(this.get('count'));
        }.property('count')
      });
      self.set('isPopup', App.ModalPopup.show({
        /**
         * no need to track is it loaded when popup contain only list of hosts
         * @type {bool}
         */
        isLoaded: !isBackgroundOperations,

        /**
         * is BG-popup opened
         * @type {bool}
         */
        isOpen: false,

        didInsertElement: function(){
          this._super();
          this.set('isOpen', true);
        },

        /**
         * @type {Em.View}
         */
        headerClass: Em.View.extend({
          controller: this,
          template: Ember.Handlebars.compile('{{popupHeaderName}}')
        }),

        /**
         * @type {String[]}
         */
        classNames: ['sixty-percent-width-modal', 'host-progress-popup'],

        /**
         * for the checkbox: do not show this dialog again
         * @type {bool}
         */
        hasFooterCheckbox: true,

        /**
         * Auto-display BG-popup
         * @type {bool}
         */
        isNotShowBgChecked : null,

        /**
         * Save user pref about auto-display BG-popup
         */
        updateNotShowBgChecked: function () {
          var curVal = !this.get('isNotShowBgChecked');
          var key = App.router.get('applicationController').persistKey();
          if (!App.testMode) {
            App.router.get('applicationController').postUserPref(key, curVal);
          }
        }.observes('isNotShowBgChecked'),

        autoHeight: false,
        closeModelPopup: function () {
          this.set('isOpen', false);
          if(isBackgroundOperations){
            $(this.get('element')).detach();
            App.router.get('backgroundOperationsController').set('levelInfo.name', 'REQUESTS_LIST');
          } else {
            this.hide();
            self.set('isPopup', null);
          }
        },
        onPrimary: function () {
          this.closeModelPopup();
        },
        onClose: function () {
          this.closeModelPopup();
        },
        secondary: null,

        bodyClass: Em.View.extend({
          templateName: require('templates/common/host_progress_popup'),
          isLogWrapHidden: true,
          isTaskListHidden: true,
          isHostListHidden: true,
          isServiceListHidden: false,
          showTextArea: false,
          isServiceEmptyList: true,
          isHostEmptyList: true,
          isTasksEmptyList: true,
          controller: this,
          sourceRequestScheduleId: -1,
          sourceRequestScheduleRunning: false,
          sourceRequestScheduleAborted: false,
          sourceRequestScheduleCommand: null,
          hosts: self.get('hosts'),
          services: self.get('servicesInfo'),

          currentHost: function () {
            return this.get('hosts') && this.get('hosts').findProperty('name', this.get('controller.currentHostName'));
          }.property('controller.currentHostName'),

          tasks: function () {
            var currentHost = this.get('currentHost');
            if (currentHost) {
              return currentHost.get('tasks');
            }
            return [];
          }.property('currentHost.tasks', 'currentHost.tasks.@each.status'),

          /**
           * Preset values on init
           */
          setOnStart: function () {
            if (this.get("controller.isBackgroundOperations")) {
              this.get('controller').setSelectCount(this.get("services"), this.get('categories'));
              this.updateHostInfo();
            } else {
              this.set("isHostListHidden", false);
              this.set("isServiceListHidden", true);
            }
          },

          /**
           * force popup to show list of operations
           */
          resetState: function(){
            if(this.get('parentView.isOpen')){
              this.set('isLogWrapHidden', true);
              this.set('isTaskListHidden', true);
              this.set('isHostListHidden', true);
              this.set('isServiceListHidden', false);
              this.get("controller").setBackgroundOperationHeader(false);
              this.setOnStart();
            }
          }.observes('parentView.isOpen'),

          /**
           * When popup is opened, and data after polling has changed, update this data in component
           */
          updateHostInfo: function () {
            if(!this.get('parentView.isOpen')) return;
            this.set('parentView.isLoaded', false);
            this.get("controller").set("inputData", this.get("controller.dataSourceController.services"));
            this.get("controller").onServiceUpdate(this.get('isServiceListHidden'));
            this.get("controller").onHostUpdate();
            this.set('parentView.isLoaded', true);
            //push hosts into view when none or all hosts are loaded
            if(this.get('hosts') == null || this.get('hosts').length === this.get("controller.hosts").length){
              this.set("hosts", this.get("controller.hosts"));
            }
            this.set("services", this.get("controller.servicesInfo"));
          }.observes("controller.dataSourceController.serviceTimestamp"),

          /**
           * Depending on service filter, set which services should be shown
           */
          visibleServices: function () {
            if (this.get("services")) {
              this.set("isServiceEmptyList", true);
              if (this.get('serviceCategory.value')) {
                var filter = this.get('serviceCategory.value');
                var services = this.get('services');
                this.set("isServiceEmptyList", this.setVisibility(filter, services));
              }
            }
          }.observes('serviceCategory', 'services', 'services.@each.status'),

          /**
           * Depending on hosts filter, set which hosts should be shown
           */
          visibleHosts: function () {
            this.set("isHostEmptyList", true);
            if (this.get('hostCategory.value') && this.get('hosts')) {
              var filter = this.get('hostCategory.value');
              var hosts = this.get('hosts');
              this.set("isHostEmptyList", this.setVisibility(filter, hosts));
            }
          }.observes('hostCategory', 'hosts', 'hosts.@each.status'),

          /**
           * Depending on tasks filter, set which tasks should be shown
           */
          visibleTasks: function () {
            this.set("isTasksEmptyList", true);
            if (this.get('taskCategory.value') && this.get('tasks')) {
              var filter = this.get('taskCategory.value');
              var tasks = this.get('tasks');
              this.set("isTasksEmptyList", this.setVisibility(filter, tasks));
            }
          }.observes('taskCategory', 'tasks', 'tasks.@each.status'),

          /**
           * Depending on selected filter type, set object visibility value
           * @param filter
           * @param obj
           * @return {bool} isEmptyList
           */
          setVisibility: function (filter, obj) {
            var isEmptyList = true;
            if (filter == "all") {
              obj.setEach("isVisible", true);
              isEmptyList = !(obj.length > 0);
            } else {
              obj.forEach(function(item){
                if (filter == "pending") {
                  item.set('isVisible', ["pending", "queued"].contains(item.status));
                } else if (filter == "in_progress") {
                  item.set('isVisible', ["in_progress", "upgrading"].contains(item.status));
                } else if (filter == "failed") {
                  item.set('isVisible', (item.status === "failed"));
                } else if (filter == "completed") {
                  item.set('isVisible', ["completed", "success"].contains(item.status));
                } else if (filter == "aborted") {
                  item.set('isVisible', (item.status === "aborted"));
                } else if (filter == "timedout") {
                  item.set('isVisible', (item.status === "timedout"));
                }
                isEmptyList = (isEmptyList) ? !item.get('isVisible') : false;
              })
            }
            return isEmptyList;
          },

          /**
           * Select box, display names and values
           */
          categories: [
            categoryObject.create({value: 'all', labelPath: 'hostPopup.status.category.all'}),
            categoryObject.create({value: 'pending', labelPath: 'hostPopup.status.category.pending'}),
            categoryObject.create({value: 'in_progress', labelPath: 'hostPopup.status.category.inProgress'}),
            categoryObject.create({value: 'failed', labelPath: 'hostPopup.status.category.failed'}),
            categoryObject.create({value: 'completed', labelPath: 'hostPopup.status.category.success'}),
            categoryObject.create({value: 'aborted', labelPath: 'hostPopup.status.category.aborted'}),
            categoryObject.create({value: 'timedout', labelPath: 'hostPopup.status.category.timedout'})
          ],

          /**
           * Selected option is binded to this values
           */
          serviceCategory: null,
          hostCategory: null,
          taskCategory: null,

          /**
           * Depending on currently viewed tab, call setSelectCount function
           */
          updateSelectView: function () {
            if (!this.get('isHostListHidden')) {
              //since lazy loading used for hosts, we need to get hosts info directly from controller, that always contains entire array of data
              this.get('controller').setSelectCount(this.get("controller.hosts"), this.get('categories'));
            } else if (!this.get('isTaskListHidden')) {
              this.get('controller').setSelectCount(this.get("tasks"), this.get('categories'));
            } else if (!this.get('isServiceListHidden')) {
              this.get('controller').setSelectCount(this.get("services"), this.get('categories'));
            }
          }.observes('tasks.@each.status', 'hosts.@each.status', 'isTaskListHidden', 'isHostListHidden', 'services.length', 'services.@each.status'),

          /**
           * control data uploading, depending on which display level is showed
           * @param levelName
           */
          switchLevel: function (levelName) {
            if (this.get("controller.isBackgroundOperations")) {
              var BGController = App.router.get('backgroundOperationsController');
              var levelInfo = BGController.get('levelInfo');
              levelInfo.set('taskId', this.get('openedTaskId'));
              levelInfo.set('requestId', this.get('controller.currentServiceId'));
              levelInfo.set('name', levelName);
              if (levelName === 'HOSTS_LIST') {
                levelInfo.set('sync', (this.get('controller.hosts').length === 0));
                BGController.requestMostRecent();
              } else if (levelName === 'TASK_DETAILS') {
                levelInfo.set('sync', true);
                BGController.requestMostRecent();
              } else if (levelName === 'REQUESTS_LIST') {
                this.get('controller.hosts').clear();
                BGController.requestMostRecent();
              }
            }
          },
          /**
           * Onclick handler for button <-Tasks
           */
          backToTaskList: function () {
            this.destroyClipBoard();
            this.set("openedTaskId", 0);
            this.set("isLogWrapHidden", true);
            this.set("isTaskListHidden", false);
            this.switchLevel("TASKS_LIST");
          },

          /**
           * Onclick handler for button <-Hosts
           */
          backToHostList: function () {
            this.set("isHostListHidden", false);
            this.set("isTaskListHidden", true);
            this.get("controller").set("popupHeaderName", this.get("controller.serviceName"));
            this.switchLevel("HOSTS_LIST");
          },

          /**
           * Onclick handler for button <-Services
           */
          backToServiceList: function () {
            this.get("controller").set("serviceName", "");
            this.set("isHostListHidden", true);
            this.set("isServiceListHidden", false);
            this.set("isTaskListHidden", true);
            this.set("hosts", null);
            this.get("controller").setBackgroundOperationHeader(false);
            this.switchLevel("REQUESTS_LIST");
          },

          /**
           * Onclick handler for selected Service
           * @param {Object} event
           */
          gotoHosts: function (event) {
            this.get("controller").set("serviceName", event.context.get("name"));
            this.get("controller").set("currentServiceId", event.context.get("id"));
            this.get("controller").set("currentHostName", null);
            this.get("controller").onHostUpdate();
            this.switchLevel("HOSTS_LIST");
            var servicesInfo = this.get("controller.hosts");
            if (servicesInfo.length) {
              this.get("controller").set("popupHeaderName", event.context.get("name"));
            }
            //apply lazy loading on cluster with more than 100 nodes
            if (servicesInfo.length > 100) {
              this.set('hosts', servicesInfo.slice(0, 50));
            } else {
              this.set('hosts', servicesInfo);
            }
            this.set("isServiceListHidden", true);
            this.set("isHostListHidden", false);
            $(".modal").scrollTop(0);
            $(".modal-body").scrollTop(0);
            if (servicesInfo.length > 100) {
              Em.run.next(this, function(){
                this.set('hosts', this.get('hosts').concat(servicesInfo.slice(50, servicesInfo.length)));
              });
            }
            // Determine if source request schedule is present
            this.set('sourceRequestScheduleId', event.context.get("sourceRequestScheduleId"));
            this.set('sourceRequestScheduleCommand', event.context.get('contextCommand'));
            this.refreshRequestScheduleInfo();
          },

          isRequestSchedule : function() {
            var id = this.get('sourceRequestScheduleId');
            return id != null && !isNaN(id) && id > -1;
          }.property('sourceRequestScheduleId'),

          refreshRequestScheduleInfo : function() {
            var self = this;
            var id = this.get('sourceRequestScheduleId');
            batchUtils.getRequestSchedule(id, function(data) {
              if (data != null && data.RequestSchedule != null &&
                  data.RequestSchedule.status != null) {
                switch (data.RequestSchedule.status) {
                case 'DISABLED':
                  self.set('sourceRequestScheduleRunning', false);
                  self.set('sourceRequestScheduleAborted', true);
                  break;
                case 'COMPLETED':
                  self.set('sourceRequestScheduleRunning', false);
                  self.set('sourceRequestScheduleAborted', false);
                  break;
                case 'SCHEDULED':
                  self.set('sourceRequestScheduleRunning', true);
                  self.set('sourceRequestScheduleAborted', false);
                  break;
                }
              } else {
                self.set('sourceRequestScheduleRunning', false);
                self.set('sourceRequestScheduleAborted', false);
              }
            }, function(xhr, textStatus, error, opt) {
              console.log("Error getting request schedule information: ", textStatus, error, opt);
              self.set('sourceRequestScheduleRunning', false);
              self.set('sourceRequestScheduleAborted', false);
            });
          }.observes('sourceRequestScheduleId'),

          /**
           * Attempts to abort the current request schedule
           */
          doAbortRequestSchedule: function(event){
            var self = this;
            var id = event.context;
            console.log("Aborting request schedule: ", id);
            batchUtils.doAbortRequestSchedule(id, function(){
              self.refreshRequestScheduleInfo();
            });
          },

          requestScheduleAbortLabel : function() {
            var label = Em.I18n.t("common.abort");
            var command = this.get('sourceRequestScheduleCommand');
            if (command != null && "ROLLING-RESTART" == command) {
              label = Em.I18n.t("hostPopup.bgop.abort.rollingRestart");
            }
            return label;
          }.property('sourceRequestScheduleCommand'),

          /**
           * Onclick handler for selected Host
           * @param {Object} event
           */
          gotoTasks: function (event) {
            var tasksInfo = [];
            event.context.logTasks.forEach(function (_task) {
              tasksInfo.pushObject(this.get("controller").createTask(_task));
            }, this);
            if (tasksInfo.length) {
              this.get("controller").set("popupHeaderName", event.context.publicName);
              this.get("controller").set("currentHostName", event.context.publicName);
            }
            this.switchLevel("TASKS_LIST");
            this.set('currentHost.tasks', tasksInfo);
            this.set("isHostListHidden", true);
            this.set("isTaskListHidden", false);
            $(".modal").scrollTop(0);
            $(".modal-body").scrollTop(0);
          },

          /**
           * Onclick handler for selected Task
           */
          openTaskLogInDialog: function () {
            if ($(".task-detail-log-clipboard").length > 0) {
              this.destroyClipBoard();
            }
            var newWindow = window.open();
            var newDocument = newWindow.document;
            newDocument.write($(".task-detail-log-info").html());
            newDocument.close();
          },

          openedTaskId: 0,

          /**
           * Return task detail info of opened task
           */
          openedTask: function () {
            if (!(this.get('openedTaskId') && this.get('tasks'))) {
              return Ember.Object.create();
            }
            return this.get('tasks').findProperty('id', this.get('openedTaskId'));
          }.property('tasks', 'tasks.@each.stderr', 'tasks.@each.stdout', 'openedTaskId'),

          /**
           * Onclick event for show task detail info
           * @param {Object} event
           */
          toggleTaskLog: function (event) {
            var taskInfo = event.context;
            this.set("isLogWrapHidden", false);
            if ($(".task-detail-log-clipboard").length > 0) {
              this.destroyClipBoard();
            }
            this.set("isHostListHidden", true);
            this.set("isTaskListHidden", true);
            this.set('openedTaskId', taskInfo.id);
            this.switchLevel("TASK_DETAILS");
            $(".modal").scrollTop(0);
            $(".modal-body").scrollTop(0);
          },

          /**
           * Onclick event for copy to clipboard button
           */
          textTrigger: function () {
            $(".task-detail-log-clipboard").length > 0 ? this.destroyClipBoard() : this.createClipBoard();
          },

          /**
           * Create Clip Board
           */
          createClipBoard: function () {
            var logElement = $(".task-detail-log-maintext");
            $(".task-detail-log-clipboard-wrap").html('<textarea class="task-detail-log-clipboard"></textarea>');
            $(".task-detail-log-clipboard")
              .html("stderr: \n" + $(".stderr").html() + "\n stdout:\n" + $(".stdout").html())
              .css("display", "block")
              .width(logElement.width())
              .height(logElement.height())
              .select();
            logElement.css("display", "none")
          },

          /**
           * Destroy Clip Board
           */
          destroyClipBoard: function () {
            $(".task-detail-log-clipboard").remove();
            $(".task-detail-log-maintext").css("display", "block");
          }

        })
      }));
      return self.get('isPopup');
    }

  });
});
window.require.register("utils/hosts", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var lazyloading = require('utils/lazy_loading');

  module.exports = {

    /**
     * Launches a dialog to select hosts from the provided available hosts. 
     * 
     * Once the user clicks OK or Cancel, the callback is called with the 
     * array of hosts (App.Host[]) selected. If the dialog was cancelled
     * or closed, <code>null</code> is provided to the callback. Else
     * an array (maybe empty) will be provided to the callback.
     * 
     * @param availableHosts  {App.Host[]} List of hosts to pick from
     * @param selectedHosts {App.Host[]} List of hosts already selected from the available hosts
     * @param selectAtleastOneHost  {boolean} If true atleast one host has to be selected
     * @param validComponents {App.HostComponent[]} List of host-component types to pick from.
     * @param callback  Callback function which is invoked when dialog
     * @param popupDescription {Object} Consist header and message for popup
     *   Example: {header: 'header', dialogMessage: 'message'}
     *  is closed, cancelled or OK is pressed.
     */
    launchHostsSelectionDialog : function(initialHosts, selectedHosts,
        selectAtleastOneHost, validComponents, callback, popupDescription) {
      // set default popup description
      var defaultPopupDescription = {
        header: Em.I18n.t('hosts.selectHostsDialog.title'),
        dialogMessage: Em.I18n.t('hosts.selectHostsDialog.message')
      };
      if (popupDescription !== null) {
        popupDescription = $.extend(true, defaultPopupDescription, popupDescription);
      }
      App.ModalPopup.show({
        classNames: [ 'sixty-percent-width-modal' ],
        header: popupDescription.header,
        dialogMessage: popupDescription.dialogMessage,
        warningMessage: null,
        availableHosts: [],
        onPrimary: function () {
          this.set('warningMessage', null);
          var arrayOfSelectedHosts = this.get('availableHosts').filterProperty('selected', true).mapProperty('host.id');
          if (selectAtleastOneHost && arrayOfSelectedHosts.length < 1) {
            this.set('warningMessage', Em.I18n.t('hosts.selectHostsDialog.message.warning'));
            return;
          }
          callback(arrayOfSelectedHosts);
          console.debug('(new-selectedHosts)=', arrayOfSelectedHosts);
          this.hide();
        },
        disablePrimary: function () {
          return !this.get('isLoaded');
        }.property('isLoaded'),
        onSecondary: function () {
          callback(null);
          this.hide();
        },
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/configs/overrideWindow'),
          controllerBinding: 'App.router.mainServiceInfoConfigsController',
          filterText: '',
          filterTextPlaceholder: Em.I18n.t('hosts.selectHostsDialog.filter.placeHolder'),
          filterColumn: null,
          filterColumns: Ember.A([
            Ember.Object.create({id: 'ip', name: 'IP Address', selected: true}),
            Ember.Object.create({id: 'cpu', name: 'CPU', selected: false}),
            Ember.Object.create({id: 'memory', name: 'RAM', selected: false}),
            Ember.Object.create({id: 'osArch', name: 'OS Architecture', selected: false}),
            Ember.Object.create({id: 'osType', name: 'OS Type', selected: false}),
            Ember.Object.create({id: 'diskTotal', name: 'Total Disks Capacity', selected: false}),
            Ember.Object.create({id: 'disksMounted', name: '# of Disk Mounts', selected: false})
          ]),
          showOnlySelectedHosts: false,
          filterComponents: validComponents,
          filterComponent: null,
          isDisabled: function () {
            return !this.get('parentView.isLoaded');
          }.property('parentView.isLoaded'),
          didInsertElement: function(){
            var defaultFilterColumn = this.get('filterColumns').findProperty('selected');
            this.set('filterColumn', defaultFilterColumn);
            this.initContent();
          },
          initContent: function () {
            initialHosts.setEach('filtered', true);
            if (initialHosts.length > 100) {
              lazyloading.run({
                destination: this.get('parentView.availableHosts'),
                source: initialHosts,
                context: this.get('parentView'),
                initSize: 50,
                chunkSize: 100,
                delay: 50
              });
            } else {
              this.set('parentView.availableHosts', initialHosts);
              this.set('parentView.isLoaded', true);
            }
          },
          filterHosts: function () {
            var filterText = this.get('filterText');
            var showOnlySelectedHosts = this.get('showOnlySelectedHosts');
            var filterComponent = this.get('filterComponent');
            var filterColumn = this.get('filterColumn');

            this.get('parentView.availableHosts').forEach(function (host) {
              var skip = showOnlySelectedHosts && !host.get('selected');
              var value = host.get('host').get(filterColumn.id);
              var hostComponentNames = host.get('hostComponentNames');

              host.set('filterColumnValue', value);

              if (!skip && filterText) {
                if ((value == null || !value.toString().match(filterText)) && !host.get('host.publicHostName').match(filterText)) {
                  skip = true;
                }
              }
              if (!skip && filterComponent) {
                if (hostComponentNames.length > 0) {
                  skip = !hostComponentNames.contains(filterComponent.get('componentName'));
                }
              }
              host.set('filtered', !skip);
            }, this);
          }.observes('parentView.availableHosts', 'filterColumn', 'filterText', 'filterComponent', 'filterComponent.componentName', 'showOnlySelectedHosts'),
          hostSelectMessage: function () {
            var hosts = this.get('parentView.availableHosts');
            var selectedHosts = hosts.filterProperty('selected', true);
            return this.t('hosts.selectHostsDialog.selectedHostsLink').format(selectedHosts.get('length'), hosts.get('length'))
          }.property('parentView.availableHosts.@each.selected'),
          selectFilterColumn: function (event) {
            if (event != null && event.context != null && event.context.id != null) {
              var filterColumn = this.get('filterColumn');
              if (filterColumn != null) {
                filterColumn.set('selected', false);
              }
              event.context.set('selected', true);
              this.set('filterColumn', event.context);
            }
          },
          selectFilterComponent: function (event) {
            if (event != null && event.context != null && event.context.componentName != null) {
              var currentFilter = this.get('filterComponent');
              if (currentFilter != null) {
                currentFilter.set('selected', false);
              }
              if (currentFilter != null && currentFilter.componentName === event.context.componentName) {
                // selecting the same filter deselects it.
                this.set('filterComponent', null);
              } else {
                this.set('filterComponent', event.context);
                event.context.set('selected', true);
              }
            }
          },
          allHostsSelected: false,
          toggleSelectAllHosts: function (event) {
            this.get('parentView.availableHosts').filterProperty('filtered').setEach('selected', this.get('allHostsSelected'));
          }.observes('allHostsSelected'),
          toggleShowSelectedHosts: function () {
            var currentFilter = this.get('filterComponent');
            if (currentFilter != null) {
              currentFilter.set('selected', false);
            }
            this.set('filterComponent', null);
            this.set('filterText', null);
            this.set('showOnlySelectedHosts', !this.get('showOnlySelectedHosts'));
          }
        })
      });
    }
  };
  
});
window.require.register("utils/http_client", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * App.HttpClient perform an ajax request
   */
  App.HttpClient = Em.Object.create({

    /**
     *
     * @param jqXHR
     * @param textStatus
     * @param errorThrown
     * @param url api that invoked this callback function
     */
    defaultErrorHandler: function (jqXHR, textStatus, errorThrown, url) {
      try {
        var json = $.parseJSON(jqXHR.responseText);
      } catch (err) {
      }
      App.ajax.defaultErrorHandler(jqXHR, url);
      if (json) {
        Em.assert("HttpClient:", json);
      } else {
        if (!$.mocho) { // don't use this assert on tests
          Em.assert("HttpClient:", errorThrown);
        }
      }
    },

    /**
     * @param {string} url
     * @param {Object} ajaxOptions
     * @param {App.ServerDataMapper} mapper - json processor
     * @param {callback} errorHandler
     */
    request: function (url, ajaxOptions, mapper, errorHandler) {

      if (!errorHandler) {
        errorHandler = this.defaultErrorHandler;
      }

      var xhr = new XMLHttpRequest();
      var curTime = App.dateTime();

      xhr.open('GET', url + (url.indexOf('?') >= 0 ? '&_=' : '?_=') + curTime, true);
      xhr.send(null);

      this.onReady(xhr, "", ajaxOptions, mapper, errorHandler, url);
    },

    /*
     This function checks if we get response from server
     Not using onreadystatechange cuz of possible closure
     */
    onReady: function (xhr, tm, tmp_val, mapper, errorHandler, url) {
      var self = this;
      clearTimeout(tm);
      var timeout = setTimeout(function () {
        if (xhr.readyState == 4) {
          if (xhr.status == 200) {
            try {
              App.store.commit();
            } catch (err) {
              console.warn('App.store.commit error:', err);
            }
            mapper.map($.parseJSON(xhr.responseText));
            tmp_val.complete.call(self);
            xhr.abort();
          } else {
            errorHandler(xhr, "error", xhr.statusText, url);
          }

          tmp_val = null;
          xhr = null;
          clearTimeout(timeout);
          timeout = null;

        } else {
          self.onReady(xhr, timeout, tmp_val, mapper, errorHandler, url);
        }
      }, 10);
    },

    /**
     * @param {string} url
     * @param {App.ServerDataMapper} mapper - json processor
     * @param {Object} data - ajax data property
     * @param {callback} errorHandler
     * @param {number} interval - frequency request
     */
    get: function (url, mapper, data, errorHandler, interval) {
      if (!errorHandler && data.error) {
        errorHandler = data.error;
      }
      var client = this;
      var request = function () {
        client.request(url, data, mapper, errorHandler);
        url = null;
        data = null;
        mapper = null;
        errorHandler = null;
      };

      interval = "" + interval;
      if (interval.match(/\d+/)) {
        $.periodic({period: interval}, request);
      } else {
        request();
      }
    },

    /**
     * @param {string} url
     * @param {Object} data - ajax data property
     * @param {App.ServerDataMapper} mapper - json processor
     * @param {callback} errorHandler
     * @param {number} interval - frequecy request
     */
    post: function (url, data, mapper, errorHandler, interval) {
      this.get(url, data, mapper, errorHandler, interval);
    }
  });
  
});
window.require.register("utils/jobs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  module.exports = {

    /**
     * Refreshes the latest information for a given job
     *
     * @param {App.AbstractJob}
     *          job
     * @param {Function}
     *          successCallback
     * @param {Funtion}
     *          errorCallback(errorId) Called in error cases where there is no
     *          data from server. 'errorId' can be one of
     *          <ul>
     *            <li>job.dag.noId</li>
     *            <li>job.dag.noname</li>
     *            <li>job.dag.id.noDag</li>
     *            <li>job.dag.id.loaderror</li>
     *            <li>job.dag.name.loaderror</li>
     *          </ul>
     */
    refreshJobDetails : function(job, successCallback, errorCallback) {
      this.refreshHiveJobDetails(job, successCallback, errorCallback);
    },

    /**
     * Refreshes latest information of a Hive Job.
     *
     * @param {App.HiveJob}
     *          hiveJob
     * @param {Function}
     *          successCallback
     * @param {Function}
     *          errorCallback @see #refreshJobDetails()
     */
    refreshHiveJobDetails : function(hiveJob, successCallback, errorCallback) {
      var self = this;
      // TODO - to be changed to history server when implemented in stack.
      var yarnService = App.YARNService.find().objectAt(0);
      var historyServerHostName = yarnService.get('appTimelineServerNode.hostName');
      var ahsWebPort = yarnService.get('ahsWebPort');
      var hiveJobId = hiveJob.get('id');
      // First refresh query
      var hiveQueriesUrl = App.testMode ? "/data/jobs/hive-query-2.json" : "/proxy?url=http://" + historyServerHostName
          + ":" + ahsWebPort + "/ws/v1/timeline/HIVE_QUERY_ID/" + hiveJob.get('id') + "?fields=otherinfo";
      App.HttpClient.get(hiveQueriesUrl, App.hiveJobMapper, {
        complete : function(jqXHR, textStatus) {
          // Now get the Tez DAG ID from the DAG name
          var hiveRecord = App.HiveJob.find(hiveJobId);
          App.router.get('mainHiveJobDetailsController').set('job', hiveRecord);
          var tezDagName = hiveRecord.get('tezDag.name');
          if (tezDagName != null) {
            var sender = {
              dagNameToIdSuccess : function(data) {
                if (data && data.entities && data.entities.length > 0) {
                  var dagId = data.entities[0].entity;
                  App.TezDag.find(tezDagName).set('instanceId', dagId);
                  self.refreshTezDagDetails(tezDagName, successCallback, errorCallback);
                }else{
                  console.log('No Tez DAG was found for DAG name of ' + tezDagName);
                  errorCallback('job.dag.noId');
                }
              },
              dagNameToIdError : function(jqXHR, url, method, showStatus) {
                App.ajax.defaultErrorHandler(jqXHR, url, method, showStatus);
                errorCallback('job.dag.name.loaderror');
              }
            }
            App.ajax.send({
              name : 'jobs.tezDag.NametoID',
              sender : sender,
              data : {
                historyServerHostName : historyServerHostName,
                tezDagName : tezDagName,
                ahsWebPort: ahsWebPort
              },
              success : 'dagNameToIdSuccess',
              error : 'dagNameToIdError'
            });
          } else {
            console.log('No Tez DAG name to ask ID for. Job ID: ' + hiveJobId);
            errorCallback('job.dag.noname');
          }
        }
      });
    },

    /**
     * Refreshes runtime information of a Tez DAG based on events generated. The
     * instance ID of the Tez DAG should be set.
     *
     * @param {string}
     *          tezDagId ID of the Tez DAG. Example: 'HIVE-Q2:1'
     * @param {Function}
     *          successCallback
     * @param {Function}
     *          errorCallback @see #refreshJobDetails()
     */
    refreshTezDagDetails : function(tezDagId, successCallback, errorCallback) {
      var self = this;
      var yarnService = App.YARNService.find().objectAt(0);
      var ahsWebPort = yarnService.get('ahsWebPort');
      var historyServerHostName = yarnService.get('appTimelineServerNode.hostName');
      var tezDag = App.TezDag.find(tezDagId);
      if (tezDag) {
        var tezDagInstanceId = tezDag.get('instanceId');
        var sender = {
          loadTezDagSuccess : function(data) {
            if (data) {
              if (data.otherinfo && data.otherinfo.applicationId) {
                tezDag.set('yarnApplicationId', data.otherinfo.applicationId);
              }
              if (data.relatedentities && data.relatedentities.TEZ_VERTEX_ID != null) {
                var count = data.relatedentities.TEZ_VERTEX_ID.length;
                data.relatedentities.TEZ_VERTEX_ID.forEach(function(v) {
                  self.refreshTezDagVertex(tezDagId, v, function() {
                    if (--count <= 0) {
                      // all vertices succeeded
                      successCallback();
                    }
                  });
                });
              }
            }
          },
          loadTezDagError : function(jqXHR, url, method, showStatus) {
            App.ajax.defaultErrorHandler(jqXHR, url, method, showStatus);
            errorCallback('job.dag.id.loaderror');
          }
        }
        App.ajax.send({
          name : 'jobs.tezDag.tezDagId',
          sender : sender,
          data : {
            historyServerHostName : historyServerHostName,
            tezDagId : tezDagInstanceId,
            ahsWebPort: ahsWebPort
          },
          success : 'loadTezDagSuccess',
          error : 'loadTezDagError'
        });
      } else {
        console.log('Unable to find DAG for ID: ' + tezDagId);
        errorCallback('job.dag.id.noDag');
      }
    },

    /**
     * Refreshes runtime information of the given vertex.
     *
     * @param {string}
     *          tezDagId ID of the Tez DAG. Exmaple: 'HIVE-Q2:1'
     * @param {string}
     *          tezVertexInstanceID Instance ID of the vertex to refresh. Example
     *          'vertex_1390516007863_0001_1_00'
     * @param {Function}
     *          successCallback
     */
    refreshTezDagVertex : function(tezDagId, tezVertexInstanceId, successCallback) {
      var yarnService = App.YARNService.find().objectAt(0);
      var ahsWebPort = yarnService.get('ahsWebPort');
      var historyServerHostName = yarnService.get('appTimelineServerNode.hostName');
      var sender = {
        loadTezDagVertexSuccess : function(data) {
          if (data && data.otherinfo) {
            var vertexRecord = App.TezDagVertex.find(tezDagId + "/" + data.otherinfo.vertexName);
            if (vertexRecord != null) {
              vertexRecord.set('startTime', data.otherinfo.startTime);
              vertexRecord.set('endTime', data.otherinfo.endTime);
              vertexRecord.set('tasksCount', data.otherinfo.numTasks);
              vertexRecord.set('state', data.otherinfo.status);
              if (data.otherinfo.counters && data.otherinfo.counters.counterGroups) {
                data.otherinfo.counters.counterGroups.forEach(function(cGroup) {
                  var cNameToPropetyMap = {};
                  switch (cGroup.counterGroupName) {
                  case 'org.apache.tez.common.counters.FileSystemCounter':
                    cNameToPropetyMap = {
                      'FILE_BYTES_READ' : 'fileReadBytes',
                      'FILE_BYTES_WRITTEN' : 'fileWriteBytes',
                      'FILE_READ_OPS' : 'fileReadOps',
                      'FILE_WRITE_OPS' : 'fileWriteOps',
                      'HDFS_BYTES_READ' : 'hdfsReadBytes',
                      'HDFS_BYTES_WRITTEN' : 'hdfsWriteBytes',
                      'HDFS_READ_OPS' : 'hdfsReadOps',
                      'HDFS_WRITE_OPS' : 'hdfsWriteOps'
                    };
                    break;
                  case 'org.apache.tez.common.counters.TaskCounter':
                    cNameToPropetyMap = {
                       'SPILLED_RECORDS' : 'spilledRecords'
                      };
                    break;
                  case 'HIVE':
                    cNameToPropetyMap = {
                      'RECORDS_READ' : 'recordReadCount',
                      'RECORDS_WRITE' : 'recordWriteCount'
                    };
                    break;
                  default:
                    break;
                  }
                  if (cGroup.counters) {
                    cGroup.counters.forEach(function(counter) {
                      var prop = cNameToPropetyMap[counter.counterName];
                      if (prop != null) {
                        vertexRecord.set(prop, counter.counterValue);
                      }
                    });
                  }
                });
              }
              successCallback();
            }
          }
        },
        loadTezDagVertexError : function(jqXHR, url, method, showStatus) {
          App.ajax.defaultErrorHandler(jqXHR, url, method, showStatus);
        }
      }
      App.ajax.send({
        name : 'jobs.tezDag.tezDagVertexId',
        sender : sender,
        data : {
          historyServerHostName : historyServerHostName,
          tezDagVertexId : tezVertexInstanceId,
          ahsWebPort: ahsWebPort
        },
        success : 'loadTezDagVertexSuccess',
        error : 'loadTezDagVertexError'
      });
    }
  };
  
});
window.require.register("utils/lazy_loading", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports = {
    /**
     * Divide source array in chunks, then push each chunk into destination array
     * with delay time one by one. So then destination array gets more and more items
     * till all items are loaded.
     * @param options
     * options.initSize - number of items which will be pushed in array immediately
     * options.chunkSize - number of items which will be pushed after defined delay
     * options.delay - interval between each chunk push
     * options.destination - array where items will be pushed
     * options.source - source of items
     * options.context - the object that should know when data is completely loaded,
     * lazy loading will define "isLoaded" property for context object and update it
     */
    run: function (options) {
      var initSize = options.initSize || 25,
        chunkSize = options.chunkSize || 50,
        delay = options.delay || 300,
        destination = options.destination,
        source = options.source,
        context = options.context,
        chunks;
      if (Array.isArray(destination) && Array.isArray(source)) {
        destination.pushObjects(source.slice(0, initSize));
        if(source.length > initSize) {
          chunks = this.divideIntoChunks(source.slice(initSize, source.length), chunkSize);
          this.pushChunk(chunks, 0, delay, destination, context);
        } else {
          context.set('isLoaded', true);
        }
      } else {
        console.error('Lazy loading: source or destination has incorrect value');
      }
    },

    /**
     * push chunks into destination array in delay time
     * @param chunks
     * @param index
     * @param delay
     * @param destination
     * @param context
     */
    pushChunk: function (chunks, index, delay, destination, context) {
      var self = this;
      setTimeout(function () {
        destination.pushObjects(chunks[index]);
        if (chunks.length === (++index)) {
          context.set('isLoaded', true);
        } else {
          self.pushChunk(chunks, index, delay, destination, context);
        }
      }, delay);
    },

    /**
     * divide source array into chunks
     * @param source
     * @param chunkSize
     * @return {Array}
     */
    divideIntoChunks: function (source, chunkSize) {
      var chunk = [];
      var chunks = [];
      var counter = 0;
      source.forEach(function (item) {
        counter++;
        chunk.push(item);
        if (counter === chunkSize) {
          chunks.push(chunk);
          chunk = [];
          counter = 0;
        }
      });
      if (chunk.length > 0) {
        chunks.push(chunk);
      }
      return chunks;
    }
  };
  
});
window.require.register("utils/misc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  module.exports = {
    /**
     * Convert value from bytes to appropriate measure
     */
    formatBandwidth: function (value) {
      if (value) {
        if (value < 1024) {
          value = '<1KB';
        } else {
          if (value < 1048576) {
            value = (value / 1024).toFixed(1) + 'KB';
          } else  if (value >= 1048576 && value < 1073741824){
            value = (value / 1048576).toFixed(1) + 'MB';
          } else {
            value = (value / 1073741824).toFixed(2) + 'GB';
          }
        }
      }
      return value;
    },
    /**
     * Convert ip address to integer
     * @param ip
     * @return integer
     */
    ipToInt: function(ip){
      // *     example 1: ipToInt('192.0.34.166');
      // *     returns 1: 3221234342
      // *     example 2: ipToInt('255.255.255.256');
      // *     returns 2: false
      // Verify IP format.
      if (!/^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/.test(ip)) {
        return false; // Invalid format.
      }
      // Reuse ip variable for component counter.
      var d = ip.split('.');
      return ((((((+d[0])*256)+(+d[1]))*256)+(+d[2]))*256)+(+d[3]);
    },

    sortByOrder: function (sortOrder, array) {
      var sorted = [];
      for (var i = 0; i < sortOrder.length; i++)
        for (var j = 0; j < array.length; j++) {
          if (sortOrder[i] == Em.get(array[j], 'id')) {
            sorted.push(array[j]);
          }
        }
      return sorted;
    },

    /**
     * Convert XML document to Object
     * @param xml
     * @return {Object}
     */
    xmlToObject: function (xml) {
      var obj = {};
      if (xml.nodeType == 1) {
        if (xml.attributes.length > 0) {
          obj["@attributes"] = {};
          for (var j = 0; j < xml.attributes.length; j++) {
            var attribute = xml.attributes.item(j);
            obj["@attributes"][attribute.nodeName] = attribute.nodeValue;
          }
        }
      } else if (xml.nodeType == 3) {
        obj = xml.nodeValue;
      }
      if (xml.hasChildNodes()) {
        for (var i = 0; i < xml.childNodes.length; i++) {
          var item = xml.childNodes.item(i);
          var nodeName = item.nodeName;
          if (typeof (obj[nodeName]) == "undefined") {
            obj[nodeName] = this.xmlToObject(item);
          } else {
            if (typeof (obj[nodeName].push) == "undefined") {
              var old = obj[nodeName];
              obj[nodeName] = [];
              obj[nodeName].push(old);
            }
            obj[nodeName].push(this.xmlToObject(item));
          }
        }
      }
      return obj;
    }

  };
  
});
window.require.register("utils/number_utils", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */
  module.exports = {

    /**
     * Convert byte size to other metrics.
     * 
     * @param {Number} bytes to convert to string
     * @param {Number} precision Number to adjust precision of return value. Default is 0.
     * @param {String} parseType
     *           JS method name for parse string to number. Default is "parseInt".
     * @param {Number} multiplyBy bytes by this number if given. This is needed
     *          as <code>null * 1024 = 0</null>
     * @remarks The parseType argument can be "parseInt" or "parseFloat".
     * @return {String} Returns converted value with abbreviation.
     */
    bytesToSize: function (bytes, precision, parseType, multiplyBy) {
      if (bytes === null || bytes === undefined) {
        return 'n/a';
      } else {
        if (arguments[2] === undefined) {
          parseType = 'parseInt';
        }
        if (arguments[3] === undefined) {
          multiplyBy = 1;
        }
        var value = bytes * multiplyBy;
        var sizes = [ 'Bytes', 'KB', 'MB', 'GB', 'TB', 'PB' ];
        var posttxt = 0;
        while (value >= 1024) {
          posttxt++;
          value = value / 1024;
        }
        if (value === 0) {
          precision = 0;
        }
        var parsedValue = window[parseType](value);
        return parsedValue.toFixed(precision) + " " + sizes[posttxt];
      }
    },

    /**
     * Validates if the given string or number is an integer between the
     * values of min and max (inclusive). The minimum and maximum
     * checks are ignored if their valid is NaN.
     */
    validateInteger : function(str, min, max) {
      if (str==null || str==undefined || (str + "").trim().length < 1) {
        return Em.I18n.t('number.validate.empty');
      } else {
        str = (str + "").trim();
        var number = parseInt(str);
        if (isNaN(number)) {
          return Em.I18n.t('number.validate.notValidNumber');
        } else {
          if (str.length != (number + "").length) {
            // parseInt("1abc") returns 1 as integer
            return Em.I18n.t('number.validate.notValidNumber');
          }
          if (!isNaN(min) && number < min) {
            return Em.I18n.t('number.validate.lessThanMinumum').format(min);
          }
          if (!isNaN(max) && number > max) {
            return Em.I18n.t('number.validate.moreThanMaximum').format(max);
          }
        }
      }
      return null;
    }
  };
  
});
window.require.register("utils/object_utils", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var stringUtils = require('utils/string_utils');

  module.exports = {

    isChild: function(obj)
    {
      for (var k in obj) {
        if (obj.hasOwnProperty(k)) {
          if (obj[k] instanceof Object) {
            return false;
          }
        }
      }
      return true;
    },

    recursiveKeysCount: function(obj) {
      if (!(obj instanceof Object)) {
        return null;
      }
      var self = this;
      function r(obj) {
        var count = 0;
        for (var k in obj) {
          if(self.isChild(obj[k])){
            count++;
          } else {
            count += r(obj[k]);
          }
        }
        return count;
      }
      return r(obj);
    },

    recursiveTree: function(obj) {
      if (!(obj instanceof Object)) {
        return null;
      }
      var self = this;
      function r(obj,parent) {
        var leaf = '';
        for (var k in obj) {
          if(self.isChild(obj[k])){
            leaf += k + ' ('+parent+')' + '<br/>';
          } else {
            leaf += r(obj[k],parent +'/' + k);
          }
        }
        return leaf;
      }
      return r(obj,'');
    },
    
    /**
     * Gets value of property path.
     *
     * @param {Object} object
     * @param {String} propertyPath
     *          Format is 'a.b.c'
     * @return Returns <code>undefined</code> when path does not exist.
     */
    getProperty: function (object, propertyPath) {
      var props = propertyPath.split('.');
      for ( var c = 0; c < props.length - 1 && object; c++) {
        object = object[props[c]];
        if (object === null) {
          break;
        }
      }
      if (object != null) {
        return object[props[props.length - 1]];
      }
      return undefined;
    }

  };
  
});
window.require.register("utils/polling", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  App.Poll = Em.Object.extend({
    name: '',
    stage: '',
    label: '',
    isVisible: true,
    isStarted: false,
    isPolling: true,
    clusterName: null,
    requestId: undefined,
    temp: false,
    progress: 0,
    url: null,
    testUrl: null,
    data: null,
    isError: false,
    isSuccess: false,
    POLL_INTERVAL: 4000,
    polledData: [],
    numPolls: 0,
    mockDataPrefix: '/data/wizard/deploy/5_hosts',

    barWidth: function () {
      return 'width: ' + this.get('progress') + '%;';
    }.property('progress'),

    isCompleted: function () {
      return (this.get('isError') || this.get('isSuccess'));
    }.property('isError', 'isSuccess'),

    showLink: function () {
      return (this.get('isPolling') === true && this.get('isStarted') === true);
    }.property('isPolling', 'isStarted'),

    start: function () {
      if (this.get('requestId') === undefined) {
        this.setRequestId();
      } else {
        this.startPolling();
      }
    },

    setRequestId: function () {
      if (App.testMode) {
        this.set('requestId', '1');
        this.doPolling();
        return;
      }
      var self = this;
      var url = this.get('url');
      var method = 'PUT';
      var data = this.get('data');

      $.ajax({
        type: method,
        url: url,
        async: false,
        data: data,
        dataType: 'text',
        timeout: App.timeout,
        success: function (data) {
          var jsonData = jQuery.parseJSON(data);
          console.log("TRACE: Polling -> value of the url is: " + url);
          console.log("TRACE: Polling-> value of the sent data is: " + self.get('data'));
          console.log("TRACE: Polling-> value of the received data is: " + jsonData);
          if (jsonData === null) {
            self.set('isSuccess', true);
            self.set('isError', false);
            self.set('requestId',undefined);
          } else {
            var requestId = jsonData.Requests.id;
            self.set('requestId', requestId);
            self.doPolling();
            console.log('requestId is: ' + requestId);
          }
        },

        error: function () {
          console.log("ERROR");
          self.set('isError', true);
          self.set('isSuccess', false);
        },

        statusCode: require('data/statusCodes')
      });
    },

    doPolling: function () {
      if (this.get('requestId')) {
        this.startPolling();
      }
    },

    startPolling: function () {
      if (!this.get('requestId')) {
        return;
      }
      var self = this;
      var url = App.apiPrefix + '/clusters/' + App.router.getClusterName() + '/requests/' + this.get('requestId') + '?fields=tasks/*';
      if (App.testMode) {
        this.set('POLL_INTERVAL', 1);
        this.numPolls++;
        url = this.get('mockDataPrefix') + '/poll_' + this.get('numPolls') + '.json';
      }

      $.ajax({
        type: 'GET',
        url: url,
        async: true,
        dataType: 'text',
        timeout: App.timeout,
        success: function (data) {
          console.log("TRACE: In success function for the GET logs data");
          console.log("TRACE: The value is: ", jQuery.parseJSON(data));
          var result = self.parseInfo(jQuery.parseJSON(data));
          if (result !== true) {
            window.setTimeout(function () {
              self.startPolling();
            }, self.POLL_INTERVAL);
          } else {
            self.set('requestId', undefined);
          }
        },

        error: function (request, ajaxOptions, error) {
          console.log("TRACE: In error function for the GET data");
          console.log("TRACE: value of the url is: " + url);
          console.log("TRACE: error code status is: " + request.status);
          if (!self.get('isSuccess')) {
            self.set('isError', true);
          }
        },

        statusCode: require('data/statusCodes')
      }).retry({times: App.maxRetries, timeout: App.timeout}).then(null,
        function () {
          App.showReloadPopup();
          console.log('Install services all retries failed');
        }
      );
    },

    stopPolling: function () {
      //this.set('isSuccess', true);
    },

    replacePolledData: function (polledData) {
      this.polledData.clear();
      this.set('polledData', polledData);
    },


    calculateProgressByTasks: function (tasksData) {
      var queuedTasks = tasksData.filterProperty('Tasks.status', 'QUEUED').length;
      var completedTasks = tasksData.filter(function (task) {
        return ['COMPLETED', 'FAILED', 'ABORTED', 'TIMEDOUT'].contains(task.Tasks.status);
      }).length;
      var inProgressTasks = tasksData.filterProperty('Tasks.status', 'IN_PROGRESS').length;
      return Math.ceil(((queuedTasks * 0.09) + (inProgressTasks * 0.35) + completedTasks ) / tasksData.length * 100)
    },

    isPollingFinished: function (polledData) {
      var runningTasks;
      runningTasks = polledData.filterProperty('Tasks.status', 'QUEUED').length;
      runningTasks += polledData.filterProperty('Tasks.status', 'IN_PROGRESS').length;
      runningTasks += polledData.filterProperty('Tasks.status', 'PENDING').length;
      if (runningTasks === 0) {
        if (polledData.everyProperty('Tasks.status', 'COMPLETED')) {
          this.set('isSuccess', true);
          this.set('isError', false);
        } else if (polledData.someProperty('Tasks.status', 'FAILED') || polledData.someProperty('Tasks.status', 'TIMEDOUT') || polledData.someProperty('Tasks.status', 'ABORTED')) {
          this.set('isSuccess', false);
          this.set('isError', true);
        }
        return true;
      } else {
        return false;
      }
    },


    parseInfo: function (polledData) {
      console.log('TRACE: Entering task info function');
      var self = this;
      var totalProgress = 0;
      var tasksData = polledData.tasks;
      console.log("The value of tasksData is: ", tasksData);
      if (!tasksData) {
        console.log("ERROR: NO tasks available to process");
      }
      var requestId = this.get('requestId');
      if (polledData.Requests && polledData.Requests.id && polledData.Requests.id != requestId) {
        // We dont want to use non-current requestId's tasks data to
        // determine the current install status.
        // Also, we dont want to keep polling if it is not the
        // current requestId.
        return false;
      }
      this.replacePolledData(tasksData);
      var totalProgress = this.calculateProgressByTasks(tasksData);
      this.set('progress', totalProgress.toString());
      console.log("INFO: right now the progress is: " + this.get('progress'));
      return this.isPollingFinished(tasksData);
    }

  });

  
});
window.require.register("utils/string_utils", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports = {

    pad: function(str, len, pad, dir) {

      var STR_PAD_LEFT = 1;
      var STR_PAD_RIGHT = 2;
      var STR_PAD_BOTH = 3;

      if (typeof(len) == "undefined") { var len = 0; }
      if (typeof(pad) == "undefined") { var pad = ' '; }
      if (typeof(dir) == "undefined") { var dir = STR_PAD_RIGHT; }

      if (len + 1 >= str.length) {

        switch (dir){

          case STR_PAD_LEFT:
            str = Array(len + 1 - str.length).join(pad) + str;
            break;

          case STR_PAD_BOTH:
            var padlen = len - str.length;
            var right = Math.ceil((padlen) / 2);
            var left = padlen - right;
            str = Array(left+1).join(pad) + str + Array(right+1).join(pad);
            break;

          default:
            str = str + Array(len + 1 - str.length).join(pad);
            break;

        } // switch

      }
      return str;

    },
    underScoreToCamelCase: function(name){
      function replacer(str, p1, p2, offset, s) {
        return str[1].toUpperCase();
      }
      return name.replace(/_\w/g,replacer);
    },

    /**
     * Forces given string into upper camel-case representation. The first
     * character of each word will be capitalized with the rest in lower case.
     */
    getCamelCase : function(name) {
      if (name != null) {
        return name.toLowerCase().replace(/(\b\w)/g, function(f) {
          return f.toUpperCase();
        })
      }
      return name;
    },

    /**
     * Compare two versions by following rules:
     * first higher than second then return 1
     * first lower than second then return -1
     * first equal to second then return 0
     * @param first {string}
     * @param second {string}
     * @return {number}
     */
    compareVersions: function(first, second){
      if (!(typeof first === 'string' && typeof second === 'string')) {
        return false;
      }
      if (first === '' || second === '') {
        return false;
      }
      var firstNumbers = first.split('.');
      var secondNumbers = second.split('.');
      var length = 0;
      var i = 0;
      var result = false;
      if(firstNumbers.length === secondNumbers.length) {
        length = firstNumbers.length;
      } else if(firstNumbers.length < secondNumbers.length){
        length = secondNumbers.length;
      } else {
        length = firstNumbers.length;
      }

      while(i < length && !result){
        firstNumbers[i] = (firstNumbers[i] === undefined) ? 0 : window.parseInt(firstNumbers[i]);
        secondNumbers[i] = (secondNumbers[i] === undefined) ? 0 : window.parseInt(secondNumbers[i]);
        if(firstNumbers[i] > secondNumbers[i]){
          result = 1;
          break;
        } else if(firstNumbers[i] === secondNumbers[i]){
          result = 0;
        } else if(firstNumbers[i] < secondNumbers[i]){
          result = -1;
          break;
        }
        i++;
      }
      return result;
    },

    isSingleLine: function(string){
      return String(string).trim().indexOf("\n") == -1;
    },
    /**
     * transform array of objects into CSV format content
     * @param array
     * @return {Array}
     */
    arrayToCSV: function(array){
      var content = "";
      array.forEach(function(item){
        var row = [];
        for(var i in item){
          if(item.hasOwnProperty(i)){
            row.push(item[i]);
          }
        }
        content += row.join(',') + '\n';
      });
      return content;
    },

    /**
     * Extracts filename from linux/unix path
     * @param path
     * @return {string}: filename
     */
    getFileFromPath: function(path) {
      if (!path || typeof path !== 'string') {
        return '';
      }
      return path.replace(/^.*[\/]/, '');
    },

    getPath: function(path) {
      if (!path || typeof path !== 'string' || path[0] != '/') {
        return '';
      }
      var last_slash = path.lastIndexOf('/');
      return (last_slash!=0)?path.substr(0,last_slash):'/';
    }
  };
  
});
window.require.register("utils/ui_effects", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports = {
    /**
     *
     * @param node - DOM element which blinking
     * @param delay - overall time of blinking
     * @param callback
     * @param interval - change frequence of blinking
     */
    pulsate: function (node, delay, callback, interval) {
      var self = this;
      /**
       * execute single blink
       * @param interval - time of single blink
       * @param callback
       * @param opacity
       * @param iteration - current iteration(default amount of iterations: 10)
       * @param isReverse - flag, that mean opacity increase or decrease
       */
      var blink = function (interval, callback, opacity, iteration, isReverse) {
        var iterations = 10;
        opacity = opacity || 1;
        iteration = (iteration !== undefined) ? iteration : 10;
        if (iteration > 0) {
          iteration--;
          setTimeout(function () {
            isReverse = isReverse || (opacity <= 1 / (iterations / 2));
            opacity = (isReverse) ? opacity + (1 / (iterations / 2)) : opacity - (1 / (iterations / 2));
            node.css('opacity', opacity);
            blink(interval, callback, opacity, iteration, isReverse);
          }, interval / iterations);
        } else {
          node.css('opacity', 1);
          callback();
        }
      };
      interval = interval || 200;
      if (delay > 0) {
        delay -= interval;
        setTimeout(function () {
          blink(interval, function () {
            self.pulsate(node, delay, callback, interval);
          });
        }, interval);
      } else {
        callback();
      }
    }
  };
});
window.require.register("utils/updater", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');

  var states = {};

  function update(obj, name, isWorking, interval){
    if(typeof isWorking == 'string' && !obj.get(isWorking)){
      return false;
    }

    var state = states[name];

    if(!state){
      var callback = function(){
        update(obj, name, isWorking, interval);
      };
      states[name] = state = {
        timeout: null,
        func: function(){
          if(typeof isWorking == 'string' && !obj.get(isWorking)){
            return false;
          }
          obj[name](callback);
          return true;
        },
        callback: callback
      };
    }

    clearTimeout(state.timeout);

    state.timeout = setTimeout(state.func, interval);
    return true;
  }

  function rerun(name){
    var state = states[name];
    if(state){
      clearTimeout(state.timeout);
      state.func();
    }
  }

  App.updater = {

    /**
     * Run function periodically with <code>App.contentUpdateInterval</code> delay.
     * Example 1(wrong way, will not be working):
     *    var obj = {
     *      method: function(callback){
     *        //do something
     *      }
     *    };
     *    App.updater.run(obj, 'method');
     *
     * Will be called only once, because <code>callback</code> will never execute. Below is right way:
     *
     * Example 2:
     *    var obj = {
     *      method: function(callback){
     *        //do something
     *        callback();
     *      }
     *    };
     *    App.updater.run(obj, 'method');
     *
     * Method will always be called.
     *
     * Example 3:
     *    var obj = {
     *      method: function(callback){
     *          //do something
     *          callback();
     *      },
     *      isWorking: true
     *    };
     *    App.updater.run(obj, 'method', 'isWorking');
     *
     * <code>obj.method</code> will be called automatically.
     * Warning: You should call <code>callback</code> parameter when function finished iteration.
     * Otherwise nothing happened next time.
     * If <code>isWorking</code> provided, library will check <code>obj.isWorking</code> before iteration and
     * stop working when it equals to false. Otherwise method will always be called.
     *
     *
     *
     * @param obj Object
     * @param name Method name
     * @param isWorking Property, which will be checked as a rule for working
     * @param interval Interval between calls
     * @return {*}
     */
    run: function(obj, name, isWorking, interval){
      interval = interval || App.contentUpdateInterval;
      return update(obj, name, isWorking, interval);
    },

    /**
     * Immediate run function, which is periodically running using <code>run</code> method
     * Example:
     *    App.updater.run(obj, 'clickIt');
     *    App.updater.immediateRun('clickIt');
     *
     * <code>clickIt</code> will be executed immediately and proceed executing periodically
     *
     * @param name Method name, which was used as a parameter in <code>run</code> method
     * @return {*}
     */
    immediateRun: function(name){
      return rerun(name);
    }

  };
  
});
window.require.register("utils/validator", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  module.exports = {

    isValidEmail: function(value) {
      var emailRegex = /^((([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))$/i;
      return emailRegex.test(value);
    },

    isValidInt: function(value) {
      var intRegex = /^-?\d+$/;
      return intRegex.test(value);
    },

    isValidUNIXUser: function(value){
      var regex = /^[a-z_][a-z0-9_-]{0,31}$/;
      return regex.test(value);
    },

    isValidFloat: function(value) {
      if (typeof value === 'string' && value.trim() === '') {
        return false;
      }
      var floatRegex = /^-?(?:\d+|\d{1,3}(?:,\d{3})+)?(?:\.\d+)?$/;
      return floatRegex.test(value);
    },
    /**
     * validate directory with slash at the start
     * @param value
     * @return {Boolean}
     */
    isValidDir: function(value){
      var floatRegex = /^\/[0-9a-z]*/;
      var dirs = value.replace(/,/g,' ').trim().split(new RegExp("\\s+", "g"));
      for(var i = 0; i < dirs.length; i++){
        if(!floatRegex.test(dirs[i])){
          return false;
        }
      }
      return true;
    },

    /**
     * validate directory doesn't start "home" or "homes"
     * @param value
     * @returns {boolean}
     */
    isAllowedDir: function(value) {
      var dirs = value.replace(/,/g,' ').trim().split(new RegExp("\\s+", "g"));
      for(var i = 0; i < dirs.length; i++){
        if(dirs[i].startsWith('/home') || dirs[i].startsWith('/homes')) {
          return false;
        }
      }
      return true;
    },

    /**
     * validate ip address with port
     * @param value
     * @return {Boolean}
     */
    isIpAddress: function(value) {
      var ipRegex = /^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}(\:[0-9]{1,5})?$/;
      return ipRegex.test(value);
    },

    /**
     * validate hostname
     * @param value
     * @return {Boolean}
     */
    isHostname: function(value) {
      var regex = /(?=^.{3,254}$)(^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])(\.([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9]))*(\.[a-zA-Z]{1,62})$)/;
      return regex.test(value);
    },

    hasSpaces: function(value) {
      var regex = /(\s+)/;
      return regex.test(value);
    },

    isNotTrimmed: function(value) {
      var regex = /(^\s+|\s+$)/;
      return regex.test(value);
    },
    /**
     * validate domain name with port
     * @param value
     * @return {Boolean}
     */
    isDomainName: function(value) {
      var domainRegex = /^([a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]{2,6}$/;
      return domainRegex.test(value);
    },

    /**
     * validate username
     * @param value
     * @return {Boolean}
     */
    isValidUserName: function(value) {
      var usernameRegex = /^[a-z]([-a-z0-9]{0,30})$/;
      return usernameRegex.test(value);
    },

    /**
     * validate key of configurations
     * @param value
     * @return {Boolean}
     */
    isValidConfigKey: function(value) {
      var configKeyRegex = /^[0-9a-z_\-\.\*]+$/i;
      return configKeyRegex.test(value);
    },

    empty:function (e) {
      switch (e) {
        case "":
        case 0:
        case "0":
        case null:
        case false:
        case typeof this == "undefined":
          return true;
        default :
          return false;
      }
  }
  };
  
});
window.require.register("views", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  // load all views here

  require('views/application');
  require('views/common/chart');
  require('views/common/chart/pie');
  require('views/common/chart/linear');
  require('views/common/chart/linear_time');
  require('views/common/modal_popup');
  require('views/common/rolling_restart_view');
  require('views/common/metric');
  require('views/common/time_range');
  require('views/common/form/field');
  require('views/common/quick_view_link_view');
  require('views/common/configs/services_config');
  require('views/common/configs/overriddenProperty_view');
  require('views/common/filter_combobox');
  require('views/common/filter_combo_cleanable');
  require('views/common/table_view');
  require('views/login');
  require('views/main');
  require('views/main/menu');
  require('views/main/charts');
  require('views/main/host');
  require('views/main/host/hosts_table_menu_view');
  require('views/main/host/details');
  require('views/main/host/details/host_component_view');
  require('views/main/host/details/host_component_views/datanode_view');
  require('views/main/host/details/host_component_views/nodemanager_view');
  require('views/main/host/details/host_component_views/regionserver_view');
  require('views/main/host/details/host_component_views/tasktracker_view');
  require('views/main/host/menu');
  require('views/main/host/summary');
  require('views/main/host/configs');
  require('views/main/host/configs_service');
  require('views/main/host/configs_service_menu');
  require('views/main/host/metrics');
  require('views/main/host/add_view');
  require('views/main/host/metrics/cpu');
  require('views/main/host/metrics/disk');
  require('views/main/host/metrics/load');
  require('views/main/host/metrics/memory');
  require('views/main/host/metrics/network');
  require('views/main/host/metrics/processes');
  require('views/main/host/addHost/step4_view');
  require('views/main/admin');
  require('views/main/admin/highAvailability_view');
  require('views/main/admin/highAvailability/wizard_view');
  require('views/main/admin/highAvailability/progress_view');
  require('views/main/admin/highAvailability/rollback_view');
  require('views/main/admin/highAvailability/step1_view');
  require('views/main/admin/highAvailability/step2_view');
  require('views/main/admin/highAvailability/step3_view');
  require('views/main/admin/highAvailability/step4_view');
  require('views/main/admin/highAvailability/step5_view');
  require('views/main/admin/highAvailability/step6_view');
  require('views/main/admin/highAvailability/step7_view');
  require('views/main/admin/highAvailability/step8_view');
  require('views/main/admin/highAvailability/step9_view');
  require('views/main/admin/rollbackHA/step1_view');
  require('views/main/admin/rollbackHA/step2_view');
  require('views/main/admin/rollbackHA/step3_view');
  require('views/main/admin/rollbackHA/rollback_wizard_view');
  require('views/main/admin/cluster');
  require('views/main/admin/misc_view');
  require('views/main/admin/stack_upgrade');
  require('views/main/admin/advanced');
  require('views/main/admin/access_view');
  require('views/main/admin/advanced/password');
  require('views/main/admin/audit');
  require('views/main/admin/authentication');
  require('views/main/admin/menu');
  require('views/main/admin/user');
  require('views/main/admin/user/create');
  require('views/main/admin/user/edit');
  require('views/main/admin/user/row');
  require('views/main/admin/security');
  require('views/main/admin/security/disable');
  require('views/main/admin/security/add/menu');
  require('views/main/admin/security/add/step1');
  require('views/main/admin/security/add/step2');
  require('views/main/admin/security/add/step3');
  require('views/main/admin/security/add/step4');

  require('views/main/dashboard');
  require('views/main/dashboard/service');
  require('views/main/dashboard/service/hdfs');
  require('views/main/dashboard/service/yarn');
  require('views/main/dashboard/service/mapreduce');
  require('views/main/dashboard/service/mapreduce2');
  require('views/main/dashboard/service/hbase');
  require('views/main/dashboard/service/hive');
  require('views/main/dashboard/service/zookeeper');
  require('views/main/dashboard/service/oozie');
  require('views/main/dashboard/service/flume');
  require('views/main/dashboard/service/storm');

  require('views/main/dashboard/cluster_metrics/cpu');
  require('views/main/dashboard/cluster_metrics/load');
  require('views/main/dashboard/cluster_metrics/memory');
  require('views/main/dashboard/cluster_metrics/network');

  require('views/main/dashboard/widget');
  require('views/main/dashboard/widgets/text_widget');
  require('views/main/dashboard/widgets/uptime_text_widget');
  require('views/main/dashboard/widgets/links_widget');
  require('views/main/dashboard/widgets/pie_chart_widget');
  require('views/main/dashboard/widgets/cluster_metrics_widget');
  require('views/main/dashboard/widgets/namenode_heap');
  require('views/main/dashboard/widgets/namenode_cpu');
  require('views/main/dashboard/widgets/hdfs_capacity');
  require('views/main/dashboard/widgets/jobtracker_heap');
  require('views/main/dashboard/widgets/jobtracker_cpu');
  require('views/main/dashboard/widgets/datanode_live');
  require('views/main/dashboard/widgets/tasktracker_live');
  require('views/main/dashboard/widgets/namenode_rpc');
  require('views/main/dashboard/widgets/jobtracker_rpc');
  require('views/main/dashboard/widgets/mapreduce_slots');
  require('views/main/dashboard/widgets/metrics_memory');
  require('views/main/dashboard/widgets/metrics_network');
  require('views/main/dashboard/widgets/metrics_cpu');
  require('views/main/dashboard/widgets/metrics_load');
  require('views/main/dashboard/widgets/namenode_uptime');
  require('views/main/dashboard/widgets/jobtracker_uptime');
  require('views/main/dashboard/widgets/hdfs_links');
  require('views/main/dashboard/widgets/mapreduce_links');
  require('views/main/dashboard/widgets/hbase_links');
  require('views/main/dashboard/widgets/hbase_master_heap');
  require('views/main/dashboard/widgets/hbase_average_load');
  require('views/main/dashboard/widgets/hbase_regions_in_transition');
  require('views/main/dashboard/widgets/hbase_master_uptime');
  require('views/main/dashboard/widgets/resource_manager_heap');
  require('views/main/dashboard/widgets/resource_manager_uptime');
  require('views/main/dashboard/widgets/node_managers_live');
  require('views/main/dashboard/widgets/yarn_memory');
  require('views/main/dashboard/widgets/supervisor_live');


  require('views/main/service');
  require('views/main/service/all_services_actions');
  require('views/main/service/menu');
  require('views/main/service/item');
  require('views/main/service/reconfigure');
  require('views/main/service/info/menu');
  require('views/main/service/info/summary');
  require('views/main/service/info/configs');
  require('views/main/service/info/metrics/hdfs/jvm_threads');
  require('views/main/service/info/metrics/hdfs/jvm_heap');
  require('views/main/service/info/metrics/hdfs/io');
  require('views/main/service/info/metrics/hdfs/rpc');
  require('views/main/service/info/metrics/hdfs/file_operations');
  require('views/main/service/info/metrics/hdfs/gc');
  require('views/main/service/info/metrics/hdfs/space_utilization');
  require('views/main/service/info/metrics/hdfs/block_status');
  require('views/main/service/info/metrics/yarn/gc');
  require('views/main/service/info/metrics/yarn/jvm_threads');
  require('views/main/service/info/metrics/yarn/jvm_heap');
  require('views/main/service/info/metrics/yarn/rpc');
  require('views/main/service/info/metrics/yarn/allocated');
  require('views/main/service/info/metrics/yarn/allocated_container');
  require('views/main/service/info/metrics/yarn/apps_current_states');
  require('views/main/service/info/metrics/yarn/apps_finished_states');
  require('views/main/service/info/metrics/yarn/nms');
  require('views/main/service/info/metrics/yarn/qmr');
  require('views/main/service/info/metrics/mapreduce/gc');
  require('views/main/service/info/metrics/mapreduce/jvm_threads');
  require('views/main/service/info/metrics/mapreduce/jvm_heap');
  require('views/main/service/info/metrics/mapreduce/rpc');
  require('views/main/service/info/metrics/mapreduce/tasks_running_waiting');
  require('views/main/service/info/metrics/mapreduce/jobs_status');
  require('views/main/service/info/metrics/mapreduce/map_slots');
  require('views/main/service/info/metrics/mapreduce/reduce_slots');
  require('views/main/service/info/metrics/hbase/cluster_requests');
  require('views/main/service/info/metrics/hbase/regionserver_rw_requests');
  require('views/main/service/info/metrics/hbase/regionserver_regions');
  require('views/main/service/info/metrics/hbase/regionserver_queuesize');
  require('views/main/service/info/metrics/hbase/hlog_split_time');
  require('views/main/service/info/metrics/hbase/hlog_split_size');
  require('views/main/service/info/metrics/flume/channel_fill_pct');
  require('views/main/service/info/metrics/flume/channel_size');
  require('views/main/service/info/metrics/flume/sink_drain_success');
  require('views/main/service/info/metrics/flume/source_accepted');
  require('views/main/service/info/metrics/flume/sink_connection_failed');
  require('views/main/service/info/metrics/flume/gc');
  require('views/main/service/info/metrics/flume/jvm_heap');
  require('views/main/service/info/metrics/flume/jvm_threads_runnable');
  require('views/main/service/info/metrics/flume/cpu_user');
  require('views/main/service/info/metrics/storm/slots_number_metric');
  require('views/main/service/info/metrics/storm/executors_metric');
  require('views/main/service/info/metrics/storm/tasks_metric');
  require('views/main/service/info/metrics/storm/topologies_metric');

  require('views/main/service/add_view');
  require('views/main/service/reassign_view');
  require('views/main/service/reassign/step1_view');
  require('views/main/service/reassign/step2_view');
  require('views/main/service/reassign/step3_view');
  require('views/main/service/reassign/step4_view');
  require('views/main/service/reassign/step5_view');
  require('views/main/service/reassign/step6_view');
  require('views/main/service/manage_config_groups_view');
  require('views/main/charts/menu');
  require('views/main/charts/heatmap');
  require('views/main/charts/heatmap/heatmap_rack');
  require('views/main/charts/heatmap/heatmap_host');
  require('views/main/charts/heatmap/heatmap_host_detail');
  require('views/main/apps_view');
  require('views/main/jobs_view');
  require('views/main/jobs/select_custom_date_view');
  require('views/main/jobs/hive_job_details_view');
  require('views/main/jobs/hive_job_details_tez_dag_view');
  require('views/main/apps/item_view');
  require('views/main/apps/item/bar_view');
  require('views/main/apps/item/dag_view');
  require('views/main/mirroring_view');
  require('views/main/mirroring/edit_dataset_view');
  require('views/main/mirroring/datasets_view');
  require('views/main/mirroring/jobs_view');
  require('views/main/mirroring/manage_clusters_view');

  require('views/main/views');

  require('views/installer');
  require('views/wizard/controls_view');
  require('views/wizard/step0_view');
  require('views/wizard/step1_view');
  require('views/wizard/step2_view');
  require('views/wizard/step3_view');
  require('views/wizard/step4_view');
  require('views/wizard/step5_view');
  require('views/wizard/step6_view');
  require('views/wizard/step7_view');
  require('views/wizard/step8_view');
  require('views/wizard/step9_view');
  require('views/wizard/step10_view');
  require('views/wizard/stack_upgrade/step1_view');
  require('views/wizard/stack_upgrade/step2_view');
  require('views/wizard/stack_upgrade/step3_view');
  require('views/loading');

  require('views/experimental');
  
});
window.require.register("views/application", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ApplicationView = Em.View.extend({
      templateName: require('templates/application')
  });
});
window.require.register("views/common/chart", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartView = Em.View.extend({
    dateFormat:'dd/mm/yy',
    timeFormat:'h:m',
    w:900,
    p:30, // axis padding
    shift:30,
    ticksCount:10,
    pointsLimit:300,
    areaHeight:30, // px
    axis:false,
    x:false,
    y:false,

    init:function () {
      this._super();
      var renderer = this;

      this.x = d3.time.scale().domain([renderer.getMinDate({}), renderer.getMaxDate({})]).range([0, this.get('w')]);
      this.y = d3.scale.linear().domain([0, 50]).range([this.get('h'), 0]);
      this.axis = d3.svg.axis().orient("top").scale(this.x).ticks(this.get('ticksCount'));
    },

    h: function(){
      return this.get('p') + this.get('nodeAttributes').length * this.get('areaHeight'); //default: 160
    }.property('nodeAttributes', 'p'),

    activeH: function(){
      return this.get('nodeAttributes').length * this.get('areaHeight'); // default 160;
    }.property('h'),

    ruleHeight: function(){
      return this.get('nodeAttributes').length * this.get('areaHeight');
    }.property('nodeAttributes'),

    updateY: function(){
      this.y = d3.scale.linear().domain([0, 50]).range([this.get('h'), 0]);
    }.observes('h'),

    getMinDate:function (data) {
      if (data.length)
        return new Date(Date.parse(data[0]['date']));

      return new Date();
    },

    getMaxDate:function (data) {
      if (data.length)
        return new Date(Date.parse(data[data.length - 1]['date']));

      return new Date();
    },

    area:function () {
      var renderer = this;
      var area = d3.svg.area().x(function (d) {
        return renderer.x(renderer.getDate(d));
      });

      area.y1(function (d) {
        return renderer.get('h') - (renderer.get('h') - renderer.y(d[$(this).attr("getter")])) / renderer.get('koef');
      });

      area.y0(function (d) {
        return renderer.get('h');
      });

      return area;
    },

    line:function () {
      var renderer = this;
      var area = d3.svg.line().x(function (d) {
        return renderer.x(renderer.getDate(d));
      })
        .interpolate("basis");

      area.y(function (d) {
        return renderer.get('h');
      });

      return area;
    },

    /**
     * @todo: calculate this
     * coefficient of compression
     * @param shift
     * @return {Number}
     */

    koef:function () {
      // max value divide on area height;
      return 2 * (this.get('nodeAttributes').length + 1);
    }.property('h'),

    getDate:function (d) {
      return new Date(Date.parse(d.date));
    },

    dateTimeToDateObject:function (string) {
      var ren = this;
      return new Date($.datepicker.parseDateTime(ren.dateFormat, ren.timeFormat, string));
    },

    getDefaultShift:function () {
      return  -1 * this.get('areaHeight') * (this.get('nodeAttributes').length - 1);
    },

    percentScaleXDefaultTranslate:function () {
      return this.w + 3
    },

    clearPlot: function(){
      d3.select(this.get('chartContainerSelector')).selectAll("*").remove();
    },

    drawPlot:function () {
      this.clearPlot();

      var renderer = this;
      this.x.domain([renderer.getMinDate({}), renderer.getMaxDate({})]);

      var rule = $('<div></div>').addClass("rule").css('height', renderer.get('ruleHeight')).mouseenter(function () { $(this).hide(); });
      $(this.get('chartContainerSelector')).prepend(rule);

      var vis = d3.select(this.get('chartContainerSelector'))
        .append("svg:svg")
        .attr("width", renderer.get('w') + 5)
        .attr("height", renderer.get('h'))
        .attr("rendererId", this.get('elementId'))
        .on("mousemove", function () {

          var area = d3.select(this).select("path.line");
          var d = area.data()[0];
          var x = d3.mouse(this)[0];

          var renderer = Em.View.views[d3.select(this).attr('rendererId')];
          var container = $(this).parent();
          var scale = renderer.x;

          // first move rule
          var rule = $(container).children("div.rule");
          rule.css("left", (168 + x) + "px"); // 168 - left container margin
          rule.show();

          x = x + 5; // some correction
          var selectedDate = scale.invert(x);

          // search date between this coordinates
          var prevVal = false;
          var nextVal = d[0];

          $.each(d, function (i, point) {
            if (renderer.getDate(point).getTime() <= selectedDate.getTime()) {
              prevVal = nextVal;
              nextVal = point;
            }
          });

          var len1 = Math.abs(x - scale(renderer.getDate(prevVal)));
          var len2 = Math.abs(x - scale(renderer.getDate(nextVal)));

          var clearing = 5;
          var pointToShow = false;
          // if the first point if closer
          if ((len1 < len2) && (len1 <= clearing)) {
            pointToShow = prevVal;
          } else if (len2 <= clearing) { // the second point is closer
            pointToShow = nextVal;
          }

          $.each(renderer.get('nodeAttributes'), function (i, v) {
            var value = !pointToShow ? "" : pointToShow[v] + "%";
            $(rule).children("div." + v).html(value);
          });
        });

      vis.append("svg:g")
        .attr("class", "axis")
        .attr("transform", "translate(0," + this.get('p') + ")")
        .call(renderer.axis);

      $.each(this.get('nodeAttributes'), function (i, v) {
        var element = $('<div></div>').addClass(v).addClass("stateValue").html("");
        rule.append(element);
      });

      var shift = this.getDefaultShift();
      vis.append("svg:path")
        .attr("class", "horizontal-line")
        .data([
        {}
      ])
        .attr("transform", "translate(0," + (shift - this.get('areaHeight')) + ")")
        .attr("d", renderer.line())
        .style("stroke", "#000");

      $.each(this.get('nodeAttributes'), function (i, v) {
        vis.append("svg:path").data([
          {}
        ])
          .attr("class", "line")
          .attr("getter", v)
          .attr("transform", "translate(0, " + shift + ")")
          .attr("d", renderer.area())
          .style("fill", function () {
            return "#31a354";
          });

        vis.append("svg:path")
          .attr("class", "horizontal-line")
          .data([
          {}
        ])
          .attr("transform", "translate(0," + shift + ")")
          .attr("d", renderer.line())
          .style("stroke", "#000");

        shift += renderer.get('areaHeight');
      });
    },

    getData:function (containerId) {
      return (d3.select(containerId + " path.line").data())[0];
    },

    drawChart:function () {
      var containerSel = this.get('chartContainerSelector');
      var data = this.get('data');

      while (data.length > this.get('pointsLimit')) {
        data.shift();
      }

      var renderer = this;
      var minDate = this.getMinDate(data);
      var maxDate = this.getMaxDate(data);

      this.x.domain([minDate, maxDate]);

      var ticks = data.length > 10 ? 10 : data.length;
      this.axis.scale(renderer.x).ticks(ticks);

  //        remove dots axis
      $(containerSel + " svg g.axis g").remove();
      d3.select(containerSel + " svg g.axis")
        .call(this.axis);

      $.each(this.get('nodeAttributes'), function (i, v) {
        d3.select(containerSel + " path.line[getter='" + v + "']")
          .data([data])
          .transition()
          .attr("d", renderer.area());
      });

      // lines between charts
      $(containerSel + " path.horizontal-line").each(
        function (i, path) {
          d3.select(path).data([
            [
              {date:minDate},
              {date:maxDate}
            ]
          ]).attr("d", renderer.line());
        }
      );
    }
  });
});
window.require.register("views/common/chart/linear", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartLinearView = Em.View.extend({
    w:90,
    h:90,
    margin:2,

    classNames: ['linear'],

    init:function () {
      this._super();

      this.x = d3.scale.linear().domain([0, this.get('data').length]).range([0 + this.get('margin'), this.get('w') - this.get('margin')]);
      this.y = d3.scale.linear().domain([0, d3.max(this.get('data'))]).range([0 + this.get('margin'), this.get('h') - this.get('margin')]);
      this.line = d3.svg.line().x(function (d, i) { return this.x(i); }).y(function (d) {return -1 * this.y(d); })
    },

    didInsertElement:function () {
      this._super();
      this.appendSvg();
    },

    selector:function () {
      return '#' + this.get('elementId');
    }.property('elementId'),

    appendSvg:function () {
      var thisChart = this;

      this.set('svg', d3.select(this.get('selector'))
        .append("svg:svg")
        .attr("width", thisChart.get('w'))
        .attr("height", thisChart.get('h')));

      this.set('g', thisChart.get('svg').append("svg:g").attr("transform", "translate(0, " + thisChart.get('h') + ")"));
      this.get('g').append("svg:path").attr("d", thisChart.line(thisChart.get('data')));


      // axis
      this.get('g').append("svg:line")
        .attr("x1", thisChart.x(0))
        .attr("y1", -1 * thisChart.y(0))
        .attr("x2", thisChart.x(this.get('data').length))
        .attr("y2", -1 * thisChart.y(0));

      this.get('g').append("svg:line")
        .attr("x1", thisChart.x(0))
        .attr("y1", -1 * thisChart.y(0))
        .attr("x2", thisChart.x(0))
        .attr("y2", -1 * thisChart.y(d3.max(thisChart.get('data'))))
    }
  });
});
window.require.register("views/common/chart/linear_time", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var string_utils = require('utils/string_utils');

  /**
   * @class
   * 
   * This is a view which GETs data from a URL and shows it as a time based line
   * graph. Time is shown on the X axis with data series shown on Y axis. It
   * optionally also has the ability to auto refresh itself over a given time
   * interval.
   * 
   * This is an abstract class which is meant to be extended.
   * 
   * Extending classes should override the following:
   * <ul>
   * <li>url - from where the data can be retrieved
   * <li>title - Title to be displayed when showing the chart
   * <li>id - which uniquely identifies this chart in any page
   * <li>#transformToSeries(jsonData) - function to map server data into graph
   * series
   * </ul>
   * 
   * Extending classes could optionally override the following:
   * <ul>
   * <li>#colorForSeries(series) - function to get custom colors per series
   * </ul>
   * 
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartLinearTimeView = Ember.View.extend({
    templateName: require('templates/main/charts/linear_time'),

    /**
     * The URL from which data can be retrieved.
     *
     * This property must be provided for the graph to show properly.
     *
     * @type String
     * @default null
     */
    url: null,

    /**
     * A unique ID for this chart.
     *
     * @type String
     * @default null
     */
    id: null,

    /**
     * Title to be shown under the chart.
     *
     * @type String
     * @default null
     */
    title: null,

    /**
     * @private
     *
     * @type Rickshaw.Graph
     * @default null
     */
    _graph: null,

    /**
     * Array of classnames for each series (in widget)
     * @type Rickshaw.Graph
     */
    _popupGraph: null,

    /**
     * Array of classnames for each series
     * @type Array
     */
    _seriesProperties: null,

    /**
     * Array of classnames for each series (in widget)
     * @type Array
     */
    _seriesPropertiesWidget: null,

    /**
     * Renderer type
     * See <code>Rickshaw.Graph.Renderer</code> for more info
     * @type String
     */
    renderer: 'area',

    /**
     * Suffix used in DOM-elements selectors
     * @type String
     */
    popupSuffix: '-popup',

    /**
     * Is popup for current graph open
     * @type Boolean
     */
    isPopup: false,

    /**
     * Is graph ready
     * @type Boolean
     */
    isReady: false,

    /**
     * Is popup-graph ready
     * @type Boolean
     */
    isPopupReady: false,

    /**
     * Is data for graph available
     * @type Boolean
     */
    hasData: true,

    didInsertElement: function () {
      this.loadData();
      this.registerGraph();
    },

    registerGraph: function() {
      var graph = {
        name: this.get('title'),
        id: this.get('elementId'),
        popupId: this.get('id')
      };
      App.router.get('updateController.graphs').push(graph);
    },

    loadData: function() {
      App.ajax.send({
        name: this.get('ajaxIndex'),
        sender: this,
        data: this.getDataForAjaxRequest(),
        success: '_refreshGraph',
        error: 'loadDataErrorCallback'
      });
    },

    getDataForAjaxRequest: function() {
      var toSeconds = Math.round(App.dateTime() / 1000);
      var hostName = (this.get('content')) ? this.get('content.hostName') : "";

      var HDFSService = App.HDFSService.find().objectAt(0);
      var nameNodeName = HDFSService ? HDFSService.get('nameNode.hostName') : "";
      var MapReduceService = App.MapReduceService.find().objectAt(0);
      var jobTrackerNode = MapReduceService ? MapReduceService.get('jobTracker.hostName') : "";
      var YARNService = App.YARNService.find().objectAt(0);
      var resourceManager = YARNService ? YARNService.get('resourceManagerNode.hostName') : "";
      var timeUnit = this.get('timeUnitSeconds');
      return {
        toSeconds: toSeconds,
        fromSeconds: toSeconds - timeUnit,
        stepSeconds: 15,
        hostName: hostName,
        nameNodeName: nameNodeName,
        jobTrackerNode: jobTrackerNode,
        resourceManager: resourceManager
      };
    },

    loadDataErrorCallback: function(xhr, textStatus, errorThrown) {
      this.set('isReady', true);
      if (xhr.readyState == 4 && xhr.status) {
        textStatus = xhr.status + " " + textStatus;
      }
      this._showMessage('warn', this.t('graphs.error.title'), this.t('graphs.error.message').format(textStatus, errorThrown));
      this.set('isPopup', false);
      this.set('hasData', false);
    },

    /**
     * Shows a yellow warning message in place of the chart.
     *
     * @param type  Can be any of 'warn', 'error', 'info', 'success'
     * @param title Bolded title for the message
     * @param message String representing the message
     * @type: Function
     */
    _showMessage: function(type, title, message) {
      var chartOverlay = '#' + this.id;
      var chartOverlayId = chartOverlay + '-chart';
      var chartOverlayY = chartOverlay + '-yaxis';
      var chartOverlayX = chartOverlay + '-xaxis';
      var chartOverlayLegend = chartOverlay + '-legend';
      var chartOverlayTimeline = chartOverlay + '-timeline';
      if (this.get('isPopup')) {
        chartOverlayId += this.get('popupSuffix');
        chartOverlayY += this.get('popupSuffix');
        chartOverlayX += this.get('popupSuffix');
        chartOverlayLegend += this.get('popupSuffix');
        chartOverlayTimeline += this.get('popupSuffix');
      }
      var typeClass;
      switch (type) {
        case 'error':
          typeClass = 'alert-error';
          break;
        case 'success':
          typeClass = 'alert-success';
          break;
        case 'info':
          typeClass = 'alert-info';
          break;
        default:
          typeClass = '';
          break;
      }
      $(chartOverlayId+', '+chartOverlayY+', '+chartOverlayX+', '+chartOverlayLegend+', '+chartOverlayTimeline).html('');
      $(chartOverlayId).append('<div class=\"alert '+typeClass+'\"><strong>'+title+'</strong> '+message+'</div>');
    },

    /**
     * Transforms the JSON data retrieved from the server into the series
     * format that Rickshaw.Graph understands.
     *
     * The series object is generally in the following format: [ { name :
     * "Series 1", data : [ { x : 0, y : 0 }, { x : 1, y : 1 } ] } ]
     *
     * Extending classes should override this method.
     *
     * @param seriesData
     *          Data retrieved from the server
     * @param displayName
     *          Graph title
     * @type: Function
     *
     */
    transformData: function (seriesData, displayName) {
      var seriesArray = [];
      if (seriesData != null) {
        // Is it a string?
        if ("string" == typeof seriesData) {
          seriesData = JSON.parse(seriesData);
        }
        // Is it a number?
        if ("number" == typeof seriesData) {
          // Same number applies to all time.
          var number = seriesData;
          seriesData = [];
          seriesData.push([number, App.dateTime()-(60*60)]);
          seriesData.push([number, App.dateTime()]);
        }
        // We have valid data
        var series = {};
        series.name = displayName;
        series.data = [];
        for ( var index = 0; index < seriesData.length; index++) {
          series.data.push({
            x: seriesData[index][1],
            y: seriesData[index][0]
          });
        }
        return series;
      }
      return null;
    },

    /**
     * Provides the formatter to use in displaying Y axis.
     *
     * Uses the App.ChartLinearTimeView.DefaultFormatter which shows 10K,
     * 300M etc.
     *
     * @type Function
     */
    yAxisFormatter: function(y) {
      return App.ChartLinearTimeView.DefaultFormatter(y);
    },

    /**
     * Provides the color (in any HTML color format) to use for a particular
     * series.
     * May be redefined in child views
     *
     * @param series
     *          Series for which color is being requested
     * @return color String. Returning null allows this chart to pick a color
     *         from palette.
     * @default null
     * @type Function
     */
    colorForSeries: function (series) {
      return null;
    },

    /**
    * Check whether seriesData is correct data for chart drawing
    * @param {Array} seriesData
    * @return {Boolean}
    */
    checkSeries : function(seriesData) {
      if(!seriesData || !seriesData.length) {
        return false;
      }
      var result = true;
      seriesData.forEach(function(item) {
        if(!item.data || !item.data.length || !item.data[0] || typeof item.data[0].x === 'undefined') {
          result = false;
        }
      });
      return result;
    },

    /**
     * @private
     *
     * Refreshes the graph with the latest JSON data.
     *
     * @type Function
     */
    _refreshGraph: function (jsonData) {
      if(this.get('isDestroyed')){
        return;
      }
      var seriesData = this.transformToSeries(jsonData);

        //if graph opened as modal popup
        var popup_path = $("#" + this.get('id') + "-container" + this.get('popupSuffix'));
        var graph_container = $("#" + this.get('id') + "-container");
        if(popup_path.length) {
          popup_path.children().each(function () {
            $(this).children().remove();
          });
          this.set('isPopup', true);
        }
        else {
          graph_container.children().each(function (index, value) {
            $(value).children().remove();
          });
        }
      if (this.checkSeries(seriesData)) {
        // Check container exists (may be not, if we go to another page and wait while graphs loading)
        if (graph_container.length) {
          this.draw(seriesData);
          this.set('hasData', true);
            //move yAxis value lower to make them fully visible
          $("#" + this.id + "-container").find('.y_axis text').attr('y',8);
        }
      }
      else {
        this.set('isReady', true);
        //if Axis X time interval is default(60 minutes)
        if(this.get('timeUnitSeconds') === 3600){
          this._showMessage('info', this.t('graphs.noData.title'), this.t('graphs.noData.message'));
          this.set('hasData', false);
        }
        else {
          this._showMessage('info', this.t('graphs.noData.title'), this.t('graphs.noDataAtTime.message'));
        }
        this.set('isPopup', false);
      }
    },

    /**
     * Returns a custom time unit, that depends on X axis interval length, for the graph's X axis.
     * This is needed as Rickshaw's default time X axis uses UTC time, which can be confusing
     * for users expecting locale specific time.
     *
     * If <code>null</code> is returned, Rickshaw's default time unit is used.
     *
     * @type Function
     * @return Rickshaw.Fixtures.Time
     */
    localeTimeUnit: function(timeUnitSeconds) {
      var timeUnit = new Rickshaw.Fixtures.Time();
      switch (timeUnitSeconds){
        case 604800:
          timeUnit = timeUnit.unit('day');
          break;
        case 2592000:
          timeUnit = timeUnit.unit('week');
          break;
        case 31104000:
          timeUnit = timeUnit.unit('month');
          break;
        default:
          timeUnit = {
            name: timeUnitSeconds / 240 + ' minute',
            seconds: timeUnitSeconds / 4,
            formatter: function (d) {
              return d.toLocaleString().match(/(\d+:\d+):/)[1];
            }
          };
      }
      return timeUnit;
    },

    /**
     * temporary fix for incoming data for graph
     * to shift data time to correct time point
     * @param {Array} data
     */
    dataShiftFix: function(data) {
      var nowTime = Math.round(App.dateTime() / 1000);
      data.forEach(function(series){
        var l = series.data.length;
        var shiftDiff = nowTime - series.data[l - 1].x;
        if(shiftDiff > 3600){
          for(var i = 0;i < l;i++){
            series.data[i].x = series.data[i].x + shiftDiff;
          }
          series.data.unshift({
            x: nowTime - this.get('timeUnitSeconds'),
            y: 0
          });
        }
      }, this);
    },

    /**
     * calculate statistic data for popup legend and set proper colors for series
     * @param {Array} data
     */
    dataPreProcess: function(data) {
      var self = this;
      var palette = new Rickshaw.Color.Palette({ scheme: 'munin'});
      // Format series for display
      var series_min_length = 100000000;
      data.forEach(function (series, index) {
        var seriesColor = self.colorForSeries(series);
        if (seriesColor == null) {
          seriesColor = palette.color();
        }
        series.color = seriesColor;
        series.stroke = 'rgba(0,0,0,0.3)';
        if (this.get('isPopup')) {
          // calculate statistic data for popup legend
          var avg = 0;
          var min = Number.MAX_VALUE;
          var max = Number.MIN_VALUE;
          for (var i = 0; i < series.data.length; i++) {
            avg += series.data[i]['y'];
            if (series.data[i]['y'] < min) {
              min = series.data[i]['y'];
            }
            else {
              if (series.data[i]['y'] > max) {
                max = series.data[i]['y'];
              }
            }
          }
          series.name = string_utils.pad(series.name, 30, '&nbsp;', 2) +
                        string_utils.pad('min', 5, '&nbsp;', 3) +
                        string_utils.pad(this.get('yAxisFormatter')(min), 12, '&nbsp;', 3) +
                        string_utils.pad('avg', 5, '&nbsp;', 3) +
                        string_utils.pad(this.get('yAxisFormatter')(avg/series.data.length), 12, '&nbsp;', 3) +
                        string_utils.pad('max', 12, '&nbsp;', 3) +
                        string_utils.pad(this.get('yAxisFormatter')(max), 5, '&nbsp;', 3);
        }
        if (series.data.length < series_min_length) {
          series_min_length = series.data.length;
        }
      }.bind(this));

      // All series should have equal length
      data.forEach(function(series, index) {
        if (series.data.length > series_min_length) {
          series.data.length = series_min_length;
        }
      });
    },

    draw: function(seriesData) {
      var self = this;
      var isPopup = this.get('isPopup');
      var p = isPopup ? this.get('popupSuffix') : '';

      this.dataShiftFix(seriesData);
      this.dataPreProcess(seriesData);

      var chartId = "#" + this.get('id') + "-chart" + p;
      var chartOverlayId = "#" + this.get('id') + "-container" + p;
      var xaxisElementId = "#" + this.get('id') + "-xaxis" + p;
      var yaxisElementId = "#" + this.get('id') + "-yaxis" + p;
      var legendElementId = "#" + this.get('id') + "-legend" + p;

      var chartElement = document.querySelector(chartId);
      var overlayElement = document.querySelector(chartOverlayId);
      var xaxisElement = document.querySelector(xaxisElementId);
      var yaxisElement = document.querySelector(yaxisElementId);
      var legendElement = document.querySelector(legendElementId);

      var height = 150;
      var width = 400;
      var diff = 32;

      if(this.get('inWidget')) {
        height = 105; // for widgets view
        diff = 22;
      }
      if (isPopup) {
        height = 180;
        width = 670;
      }
      else {
        // If not in popup, the width could vary.
        // We determine width based on div's size.
        var thisElement = this.get('element');
        if (thisElement!=null) {
          var calculatedWidth = $(thisElement).width();
          if (calculatedWidth > diff) {
            width = calculatedWidth - diff;
          }
        }
      }

      var _graph = new Rickshaw.Graph({
        height: height,
        width: width,
        element: chartElement,
        series: seriesData,
        interpolation: 'step-after',
        stroke: true,
        renderer: this.get('renderer'),
        strokeWidth: (this.get('renderer') != 'area' ? 2 : 1)
      });

      if (this.get('renderer') === 'area') {
        _graph.renderer.unstack = false;
      }

      new Rickshaw.Graph.Axis.Time({
        graph: _graph,
        timeUnit: this.localeTimeUnit(this.get('timeUnitSeconds'))
      });

      new Rickshaw.Graph.Axis.Y({
        tickFormat: this.yAxisFormatter,
        element: yaxisElement,
        orientation: (isPopup ? 'left' : 'right'),
        graph: _graph
      });

      var legend = new Rickshaw.Graph.Legend({
        graph: _graph,
        element: legendElement
      });

      new Rickshaw.Graph.Behavior.Series.Toggle({
        graph: _graph,
        legend: legend
      });

      new Rickshaw.Graph.Behavior.Series.Order({
        graph: _graph,
        legend: legend
      });

      if (!isPopup) {
        overlayElement.addEventListener('mousemove', function () {
          $(xaxisElement).removeClass('hide');
          $(legendElement).removeClass('hide');
          $(chartElement).children("div").removeClass('hide');
        });
        overlayElement.addEventListener('mouseout', function () {
          $(legendElement).addClass('hide');
        });
        _graph.onUpdate(function () {
          $(legendElement).addClass('hide');
        });
      }

      //show the graph when it's loaded
      _graph.onUpdate(function() {
        self.set('isReady', true);
      });
      _graph.render();

      if (isPopup) {
        new Rickshaw.Graph.HoverDetail({
          graph: _graph,
          yFormatter:function (y) {
            return self.yAxisFormatter(y);
          },
          xFormatter:function (x) {
            return (new Date(x * 1000)).toLocaleTimeString();
          },
          formatter:function (series, x, y, formattedX, formattedY, d) {
            return formattedY + '<br />' + formattedX;
          }
        });
      }

      _graph = this.updateSeriesInGraph(_graph);
      if (isPopup) {
        //show the graph when it's loaded
        _graph.onUpdate(function() {
          self.set('isPopupReady', true);
        });
        _graph.update();

        var selector = '#'+this.get('id')+'-container'+this.get('popupSuffix');
        $(selector + ' li.line').click(function() {
          var series = [];
          $(selector + ' a.action').each(function(index, v) {
            series[index] = v.parentNode.classList;
          });
          self.set('_seriesProperties', series);
        });

        this.set('_popupGraph', _graph);
      }
      else {
        _graph.update();
        var selector = '#'+this.get('id')+'-container';
        $(selector + ' li.line').click(function() {
          var series = [];
          $(selector + ' a.action').each(function(index, v) {
            series[index] = v.parentNode.classList;
          });
          self.set('_seriesPropertiesWidget', series);
        });

        this.set('_graph', _graph);
      }
    },

    /**
     *
     * @param {Rickshaw.Graph} graph
     * @returns {Rickshaw.Graph}
     */
    updateSeriesInGraph: function(graph) {
      var id = this.get('id');
      var isPopup = this.get('isPopup');
      var popupSuffix = this.get('popupSuffix');
      var _series = isPopup ? this.get('_seriesProperties') : this.get('_seriesPropertiesWidget');
      graph.series.forEach(function(series, index) {
        if (_series !== null && _series[index] !== null && _series[index] !== undefined ) {
          if(_series[_series.length - index - 1].length > 1) {
            var s = '#' + id + '-container' + (isPopup ? popupSuffix : '') + ' a.action:eq(' + (_series.length - index - 1) + ')';
            $(s).parent('li').addClass('disabled');
            series.disable();
          }
        }
      });
      return graph;
    },

    showGraphInPopup: function() {
      if(!this.get('hasData')) {
        return;
      }

      this.set('isPopup', true);
      var self = this;

      App.ModalPopup.show({
        bodyClass: Em.View.extend({

          containerId: null,
          containerClass: null,
          yAxisId: null,
          yAxisClass: null,
          xAxisId: null,
          xAxisClass: null,
          legendId: null,
          legendClass: null,
          chartId: null,
          chartClass: null,
          titleId: null,
          titleClass: null,

          isReady: function() {
            return this.get('parentView.graph.isPopupReady');
          }.property('parentView.graph.isPopupReady'),

          didInsertElement: function() {
            $('#modal').addClass('modal-graph-line');
            var popupSuffix = this.get('parentView.graph.popupSuffix');
            var id = this.get('parentView.graph.id');
            var idTemplate = id + '-{element}' + popupSuffix;

            this.set('containerId', idTemplate.replace('{element}', 'container'));
            this.set('containerClass', 'chart-container' + popupSuffix);
            this.set('yAxisId', idTemplate.replace('{element}', 'yaxis'));
            this.set('yAxisClass', this.get('yAxisId').replace(popupSuffix, ''));
            this.set('xAxisId', idTemplate.replace('{element}', 'xaxis'));
            this.set('xAxisClass', this.get('xAxisId').replace(popupSuffix, ''));
            this.set('legendId', idTemplate.replace('{element}', 'legend'));
            this.set('legendClass', this.get('legendId').replace(popupSuffix, ''));
            this.set('chartId', idTemplate.replace('{element}', 'chart'));
            this.set('chartClass', this.get('chartId').replace(popupSuffix, ''));
            this.set('titleId', idTemplate.replace('{element}', 'title'));
            this.set('titleClass', this.get('titleId').replace(popupSuffix, ''));
          },

          templateName: require('templates/common/chart/linear_time'),
          /**
           * check is time paging feature is enable for graph
           */
          isTimePagingEnable: function() {
            return !self.get('isTimePagingDisable');
          }.property(),
          rightArrowVisible: function() {
            return (this.get('isReady') && (this.get('parentView.currentTimeIndex') != 0));
          }.property('isReady', 'parentView.currentTimeIndex'),
          leftArrowVisible: function() {
            return (this.get('isReady') && (this.get('parentView.currentTimeIndex') != 7));
          }.property('isReady', 'parentView.currentTimeIndex')
        }),
        header: this.get('title'),
        /**
         * App.ChartLinearTimeView
         */
        graph: self,
        secondary: null,
        onPrimary: function() {
          this.hide();
          self.set('isPopup', false);
          self.set('timeUnitSeconds', 3600);
        },
        onClose: function() {
          this.onPrimary();
        },
        /**
         * move graph back by time
         * @param event
         */
        switchTimeBack: function(event) {
          var index = this.get('currentTimeIndex');
          // 7 - number of last time state
          if(index < 7) {
            this.reloadGraphByTime(++index);
          }
        },
        /**
         * move graph forward by time
         * @param event
         */
        switchTimeForward: function(event) {
          var index = this.get('currentTimeIndex');
          if(index > 0) {
            this.reloadGraphByTime(--index);
          }
        },
        /**
         * reload graph depending on the time
         * @param index
         */
        reloadGraphByTime: function(index) {
          this.set('currentTimeIndex', index);
          self.set('timeUnitSeconds', this.get('timeStates')[index].seconds);
          self.loadData();
        },
        timeStates: [
          {name: Em.I18n.t('graphs.timeRange.hour'), seconds: 3600},
          {name: Em.I18n.t('graphs.timeRange.twoHours'), seconds: 7200},
          {name: Em.I18n.t('graphs.timeRange.fourHours'), seconds: 14400},
          {name: Em.I18n.t('graphs.timeRange.twelveHours'), seconds: 43200},
          {name: Em.I18n.t('graphs.timeRange.day'), seconds: 86400},
          {name: Em.I18n.t('graphs.timeRange.week'), seconds: 604800},
          {name: Em.I18n.t('graphs.timeRange.month'), seconds: 2592000},
          {name: Em.I18n.t('graphs.timeRange.year'), seconds: 31104000}
        ],
        currentTimeIndex: 0,
        currentTimeState: function() {
          return this.get('timeStates').objectAt(this.get('currentTimeIndex'));
        }.property('currentTimeIndex')
      });
      Ember.run.next(function() {
        self.loadData();
        self.set('isPopupReady', false);
      });
    },
    //60 minute interval on X axis.
    timeUnitSeconds: 3600
  });

  /**
   * A formatter which will turn a number into computer storage sizes of the
   * format '23 GB' etc.
   * 
   * @type {Function}
   */
  App.ChartLinearTimeView.BytesFormatter = function (y) {
    if (y == 0) return '0 B';
    var value = Rickshaw.Fixtures.Number.formatBase1024KMGTP(y);
    if (!y || y.length < 1) {
      value = '0 B';
    }
    else {
      if ("number" == typeof value) {
        value = String(value);
      }
      if ("string" == typeof value) {
        value = value.replace(/\.\d(\d+)/, function($0, $1){ // Remove only 1-digit after decimal part
          return $0.replace($1, '');
        }); 
        // Either it ends with digit or ends with character
        value = value.replace(/(\d$)/, '$1 '); // Ends with digit like '120'
        value = value.replace(/([a-zA-Z]$)/, ' $1'); // Ends with character like
        // '120M'
        value = value + 'B'; // Append B to make B, MB, GB etc.
      }
    }
    return value;
  };

  /**
   * A formatter which will turn a number into percentage display like '42%'
   * 
   * @type {Function}
   */
  App.ChartLinearTimeView.PercentageFormatter = function (percentage) {
    var value = percentage;
    if (!value || value.length < 1) {
      value = '0 %';
    } else {
      value = value.toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + '%';
    }
    return value;
  };

  /**
   * A formatter which will turn elapsed time into display time like '50 ms',
   * '5s', '10 m', '3 hr' etc. Time is expected to be provided in milliseconds.
   * 
   * @type {Function}
   */
  App.ChartLinearTimeView.TimeElapsedFormatter = function (millis) {
    var value = millis;
    if (!value || value.length < 1) {
      value = '0 ms';
    } else if ("number" == typeof millis) {
      var seconds = millis > 1000 ? Math.round(millis / 1000) : 0;
      var minutes = seconds > 60 ? Math.round(seconds / 60) : 0;
      var hours = minutes > 60 ? Math.round(minutes / 60) : 0;
      var days = hours > 24 ? Math.round(hours / 24) : 0;
      if (days > 0) {
        value = days + ' d';
      } else if (hours > 0) {
        value = hours + ' hr';
      } else if (minutes > 0) {
        value = minutes + ' m';
      } else if (seconds > 0) {
        value = seconds + ' s';
      } else if (millis > 0) {
        value = millis.toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + ' ms';
      } else {
        value = millis.toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + ' ms';
      }
    }
    return value;
  };

  /**
   * The default formatter which uses Rickshaw.Fixtures.Number.formatKMBT 
   * which shows 10K, 300M etc.
   *
   * @type {Function}
   */
  App.ChartLinearTimeView.DefaultFormatter = function(y) {
    if(isNaN(y)){
      return 0;
    }
    var value = Rickshaw.Fixtures.Number.formatKMBT(y);
    if (value == '') return '0';
    value = String(value);
    var c = value[value.length - 1];
    if (!isNaN(parseInt(c))) {
      // c is digit
      value = parseFloat(value).toFixed(3).replace(/0+$/, '').replace(/\.$/, '');
    }
    else {
      // c in not digit
      value = parseFloat(value.substr(0, value.length - 1)).toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + c;
    }
    return value;
  };


  /**
   * Creates and returns a formatter that can convert a 'value' 
   * to 'value units/s'. 
   * 
   * @param unitsPrefix Prefix which will be used in 'unitsPrefix/s'
   * @param valueFormatter  Value itself will need further processing 
   *        via provided formatter. Ex: '10M requests/s'. Generally
   *        should be App.ChartLinearTimeView.DefaultFormatter. 
   * @return {Function}
   */
  App.ChartLinearTimeView.CreateRateFormatter = function (unitsPrefix, valueFormatter) {
    var suffix = " "+unitsPrefix+"/s";
    return function (value) {
      value = valueFormatter(value) + suffix;
      return value;
    };
  };
});
window.require.register("views/common/chart/pie", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartPieView = Em.View.extend({
    w:90,
    h:90,
    data:[300, 500],
    id:null,
    palette: new Rickshaw.Color.Palette({ scheme: 'munin'}),
    stroke: 'black',
    strokeWidth: 2,
    donut:d3.layout.pie().sort(null),
    existCenterText: false,
    centerTextColor: 'black',

    r:function () {
      return Math.min(this.get('w'), this.get('h')) / 2 - this.get('strokeWidth');
    }.property('w', 'h'),

    outerR:function () {
      return this.get('r'); // - 10;
    }.property('r'),

    innerR:function () {
      return 0; // this.get('r') - 20;
    }.property('r'),

    arc:function () {
      return d3.svg.arc().innerRadius(this.get('innerR')).outerRadius(this.get('outerR'));
    }.property(),

    didInsertElement:function () {
      this._super();
      this.appendSvg();
    },

    selector:function () {
      return '#' + this.get('elementId');
    }.property('elementId'),

    appendSvg:function () {

      var thisChart = this;
      var svg = d3.select(thisChart.get('selector')).append("svg:svg")
        .attr("id", thisChart.get('id'))
        .attr("width", thisChart.get('w'))
        .attr("height", thisChart.get('h'))
        .attr("stroke", thisChart.get('stroke'))
        .attr("stroke-width", thisChart.get('strokeWidth'));

      // set percentage data in center if there exist a center text
      if(thisChart.get('existCenterText')){
        this.set('svg', svg
          .append("svg:g")
          .attr("render-order", 1)
          .append("svg:text")
          .attr("stroke", thisChart.get('centerTextColor'))
          .attr("font-size", 17)
          .attr("transform", "translate(" + thisChart.get('w') / 2 + "," + ((thisChart.get('h') / 2) + 3) + ")")
          .attr("text-anchor", "middle")
          .text(function(d) {
                   return thisChart.get('data')[0] + '%';
                })
           );
      }

      this.set('svg', svg
        .append("svg:g")
        .attr("transform", "translate(" + thisChart.get('w') / 2 + "," + thisChart.get('h') / 2 + ")"));

      this.set('arcs', thisChart.get('svg').selectAll("path")
        .data(thisChart.donut(thisChart.get('data')))
        .enter().append("svg:path")
        .attr("fill", function (d, i) {
          return thisChart.palette.color(i);
        })
        .attr("d", thisChart.get('arc'))

      );

    }

  });
});
window.require.register("views/common/configs/overriddenProperty_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */
  // SCP means ServiceConfigProperty

  var App = require('app');

  App.ServiceConfigView.SCPOverriddenRowsView = Ember.View.extend({
    templateName: require('templates/common/configs/overriddenProperty'),
    serviceConfigProperty: null, // is passed dynamically at runtime where ever
    // we are declaring this from configs.hbs ( we are initializing this from UI )
    categoryConfigs: null, // just declared as viewClass need it
    
    removeOverride: function (event) {
      // arg 1 SCP means ServiceConfigProperty
      var scpToBeRemoved = event.contexts[0];
      var overrides = this.get('serviceConfigProperty.overrides');
      // remove override property from selectedService on installer 7-th step
      if (this.get('controller.name') == 'wizardStep7Controller') {
        var controller = this.get('controller');
        var group = controller.get('selectedService.configGroups').findProperty('name', controller.get('selectedConfigGroup.name'));
        group.get('properties').removeObject(scpToBeRemoved);
      }
      overrides = overrides.without(scpToBeRemoved);
      this.set('serviceConfigProperty.overrides', overrides);
    }
  });
  
});
window.require.register("views/common/configs/services_config", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');
  var stringUtils = require('utils/string_utils');
  var hostsUtils = require('utils/hosts');

  App.ServicesConfigView = Em.View.extend({
    templateName: require('templates/common/configs/services_config'),
    didInsertElement: function () {
      var controller = this.get('controller');
      controller.loadStep();
    }
  });

  App.ServiceConfigView = Em.View.extend({
    templateName: require('templates/common/configs/service_config'),
    isRestartMessageCollapsed: false,
    filter: '', //from template
    columns: [], //from template
    propertyFilterPopover: [Em.I18n.t('services.service.config.propertyFilterPopover.title'), Em.I18n.t('services.service.config.propertyFilterPopover.content')],
    canEdit: true, // View is editable or read-only?
    supportsHostOverrides: function () {
      switch (this.get('controller.name')) {
        case 'wizardStep7Controller':
          return App.supports.hostOverridesInstaller && (this.get('controller.selectedService.serviceName') !== 'MISC');
        case 'mainServiceInfoConfigsController':
          return App.supports.hostOverrides;
        case 'mainHostServiceConfigsController':
          return App.supports.hostOverridesHost;
        default:
          return false;
      }
    }.property('controller.name', 'controller.selectedService'),
    toggleRestartMessageView: function () {
      this.$('.service-body').toggle('blind', 200);
      this.set('isRestartMessageCollapsed', !this.get('isRestartMessageCollapsed'));
    },
    didInsertElement: function () {
      if(this.get('isNotEditable') === true) {
        this.set('canEdit', false);
      }
      this.$('.service-body').hide();
      App.tooltip($(".restart-required-property"), {html: true});
      App.tooltip($(".icon-lock"), {placement: 'right'});
      this.checkCanEdit();
    },

    /**
     * Check if we should show Custom Property category
     */
    checkCanEdit: function () {
      var controller = this.get('controller');
      if (!controller.get('selectedService.configCategories')) {
        return;
      }
      var canAddProperty = true;
      if (controller.get('selectedConfigGroup') && !controller.get('selectedConfigGroup').isDefault) {
        canAddProperty = false;
      }
      controller.get('selectedService.configCategories').filterProperty('siteFileName').forEach(function (config) {
        if (config.get('canAddProperty') !== false) {
          config.set('canAddProperty', canAddProperty);
        }
      });
    }.observes(
      'App.router.mainServiceInfoConfigsController.selectedConfigGroup.name',
      'App.router.wizardStep7Controller.selectedConfigGroup.name'
    )
  });


  App.ServiceConfigsByCategoryView = Ember.View.extend({

    classNames: ['accordion-group', 'common-config-category'],
    classNameBindings: ['category.name', 'isShowBlock::hidden'],

    content: null,
    category: null,
    service: null,
    canEdit: true, // View is editable or read-only?
    serviceConfigs: null, // General, Advanced, NameNode, SNameNode, DataNode, etc.
    // total number of
    // hosts (by
    // default,
    // cacheable )
    categoryConfigs: function () {
      return this.get('serviceConfigs').filterProperty('category', this.get('category.name')).filterProperty('isVisible', true);
    }.property('serviceConfigs.@each', 'categoryConfigsAll.@each.isVisible').cacheable(),

    /**
     * This method provides all the properties which apply
     * to this category, irrespective of visibility. This
     * is helpful in Oozie/Hive database configuration, where
     * MySQL etc. database options don't show up, because
     * they were not visible initially.
     */
    categoryConfigsAll: function () {
      return this.get('serviceConfigs').filterProperty('category', this.get('category.name'));
    }.property('serviceConfigs.@each').cacheable(),

    /**
     * Warn/prompt user to adjust Service props when changing user/groups in Misc
     * Is triggered when user ended editing text field
     */
    miscConfigChange: function (manuallyChangedProperty) {
      var changedProperty;
      if(manuallyChangedProperty.get("id")){
        changedProperty = [manuallyChangedProperty];
      }else{
        changedProperty = this.get("serviceConfigs").filterProperty("editDone", true);
      }

      if (changedProperty.length > 0) {
        changedProperty = changedProperty.objectAt(0);
      } else {
        return;
      }
      if (this.get('controller.selectedService.serviceName') == 'MISC') {
        var newValue = changedProperty.get("value");
        var stepConfigs = this.get("controller.stepConfigs");
        this.affectedProperties = [];
        var curConfigs = "";
        var affectedPropertyName = App.get('isHadoop2Stack') ? "dfs.permissions.superusergroup" : "dfs.permissions.supergroup";
        if (changedProperty.get("name") == "hdfs_user") {
          curConfigs = stepConfigs.findProperty("serviceName", "HDFS").get("configs");
          if (newValue != curConfigs.findProperty("name", affectedPropertyName).get("value")) {
            this.affectedProperties.push(
              {
                serviceName: "HDFS",
                propertyName: affectedPropertyName,
                propertyDisplayName: affectedPropertyName,
                newValue: newValue,
                curValue: curConfigs.findProperty("name", affectedPropertyName).get("value"),
                changedPropertyName: "hdfs_user"
              }
            );
          }
          if ($.trim(newValue) != $.trim(curConfigs.findProperty("name", "dfs.cluster.administrators").get("value"))) {
            this.affectedProperties.push(
              {
                serviceName: "HDFS",
                propertyName: "dfs.cluster.administrators",
                propertyDisplayName: "dfs.cluster.administrators",
                newValue: " " + $.trim(newValue),
                curValue: curConfigs.findProperty("name", "dfs.cluster.administrators").get("value"),
                changedPropertyName: "hdfs_user"
              }
            );
          }
        } else if (changedProperty.get("name") == "hbase_user" && !App.get('isHadoop2Stack')) {
          curConfigs = stepConfigs.findProperty("serviceName", "HDFS").get("configs");
          if (newValue != curConfigs.findProperty("name", "dfs.block.local-path-access.user").get("value")) {
            this.affectedProperties.push(
              {
                serviceName: "HDFS",
                propertyName: "dfs.block.local-path-access.user",
                propertyDisplayName: "dfs.block.local-path-access.user",
                newValue: newValue,
                curValue: curConfigs.findProperty("name", "dfs.block.local-path-access.user").get("value"),
                changedPropertyName: "hbase_user"
              }
            );
          }
          var hbaseCurConfigs = stepConfigs.findProperty("serviceName", "HBASE").get("configs");
          if (newValue != hbaseCurConfigs.findProperty("name", "hbase.superuser").get("value")) {
            this.affectedProperties.push(
              {
                serviceName: "HBASE",
                propertyName: "hbase.superuser",
                propertyDisplayName: "hbase.superuser",
                newValue: newValue,
                curValue: hbaseCurConfigs.findProperty("name", "hbase.superuser").get("value"),
                changedPropertyName: "hbase_user"
              }
            );
          }
        } else if (changedProperty.get("name") == "user_group") {
          if (!((this.get("controller.selectedServiceNames").indexOf("MAPREDUCE") >= 0) || (this.get("controller.selectedServiceNames").indexOf("YARN") >= 0))) {
            return;
          }
          if(this.get("controller.selectedServiceNames").indexOf("MAPREDUCE") >= 0) {
            curConfigs = stepConfigs.findProperty("serviceName", "MAPREDUCE").get("configs");
            if (newValue != curConfigs.findProperty("name", "mapreduce.tasktracker.group").get("value")) {
              this.affectedProperties.push(
                {
                  serviceName: "MAPREDUCE",
                  propertyName: "mapreduce.tasktracker.group",
                  propertyDisplayName: "mapreduce.tasktracker.group",
                  newValue: newValue,
                  curValue: curConfigs.findProperty("name", "mapreduce.tasktracker.group").get("value"),
                  changedPropertyName: "user_group"
                }
              )
            }
            if ($.trim(newValue) != $.trim(curConfigs.findProperty("name", "mapreduce.cluster.administrators").get("value"))) {
              this.affectedProperties.push(
                {
                  serviceName: "MAPREDUCE",
                  propertyName: "mapreduce.cluster.administrators",
                  propertyDisplayName: "mapreduce.cluster.administrators",
                  newValue: " " + $.trim(newValue),
                  curValue: curConfigs.findProperty("name", "mapreduce.cluster.administrators").get("value"),
                  changedPropertyName: "user_group"
                }
              );
            }
          }
          if(this.get("controller.selectedServiceNames").indexOf("MAPREDUCE2") >= 0) {
            curConfigs = stepConfigs.findProperty("serviceName", "MAPREDUCE2").get("configs");
            if ($.trim(newValue) != $.trim(curConfigs.findProperty("name", "mapreduce.cluster.administrators").get("value"))) {
              this.affectedProperties.push(
                {
                  serviceName: "MAPREDUCE2",
                  propertyName: "mapreduce.cluster.administrators",
                  propertyDisplayName: "mapreduce.cluster.administrators",
                  newValue: " " + $.trim(newValue),
                  curValue: curConfigs.findProperty("name", "mapreduce.cluster.administrators").get("value"),
                  changedPropertyName: "user_group"
                }
              );
            }
          }
          if (this.get("controller.selectedServiceNames").indexOf("YARN") >= 0) {
          curConfigs = stepConfigs.findProperty("serviceName", "YARN").get("configs");
          if (newValue != curConfigs.findProperty("name", "yarn.nodemanager.linux-container-executor.group").get("value")) {
            this.affectedProperties.push(
              {
                serviceName: "YARN",
                propertyName: "yarn.nodemanager.linux-container-executor.group",
                propertyDisplayName: "yarn.nodemanager.linux-container-executor.group",
                newValue: newValue,
                curValue: curConfigs.findProperty("name", "yarn.nodemanager.linux-container-executor.group").get("value"),
                changedPropertyName: "user_group"
              }
            )
          }
          }
        }
        if (this.affectedProperties.length > 0 && !this.get("controller.miscModalVisible")) {
          this.newAffectedProperties = this.affectedProperties;
          var self = this;
          return App.ModalPopup.show({
            classNames: ['modal-690px-width'],
            showCloseButton: false,
            header: "Warning: you must also change these Service properties",
            onApply: function () {
              self.get("newAffectedProperties").forEach(function (item) {
                self.get("controller.stepConfigs").findProperty("serviceName", item.serviceName).get("configs")
                  .findProperty("name", item.propertyName).set("value", item.newValue);
              });
              self.get("controller").set("miscModalVisible", false);
              this.hide();
            },
            onIgnore: function () {
              self.get("controller").set("miscModalVisible", false);
              this.hide();
            },
            onUndo: function () {
              var affected = self.get("newAffectedProperties").objectAt(0);
              self.get("controller.stepConfigs").findProperty("serviceName", "MISC").get("configs")
                .findProperty("name", affected.changedPropertyName).set("value", $.trim(affected.curValue));
              self.get("controller").set("miscModalVisible", false);
              this.hide();
            },
            footerClass: Ember.View.extend({
              classNames: ['modal-footer'],
              templateName: require('templates/common/configs/propertyDependence_footer')
            }),
            bodyClass: Ember.View.extend({
              templateName: require('templates/common/configs/propertyDependence'),
              controller: this,
              propertyChange: self.get("newAffectedProperties"),
              didInsertElement: function () {
                self.get("controller").set("miscModalVisible", true);
              }
            })
          });
        }
      }
    }.observes('categoryConfigs.@each.editDone'),

    /**
     * When the view is in read-only mode, it marks
     * the properties as read-only.
     */
    updateReadOnlyFlags: function () {
      var configs = this.get('serviceConfigs');
      var canEdit = this.get('canEdit');
      if (!canEdit && configs) {
        configs.forEach(function(c){
          c.set('isEditable', false);
          var overrides = c.get('overrides');
          if (overrides!=null) {
            overrides.setEach('isEditable', false);
          }
        });
      }
    },

    /**
     * Filtered <code>categoryConfigs</code> array. Used to show filtered result
     */
    filteredCategoryConfigs: function () {
      $('.popover').remove();
      var filter = this.get('parentView.filter');
      var columns = this.get('parentView.columns');
      if (filter != null) {
        filter = filter.toLowerCase();
      }
      //var isOnlyModified = this.get('parentView.columns').length && this.get('parentView.columns')[1].get('selected');
      var isOnlyOverridden = columns!=null ? (columns.length && columns[0].get('selected')) : false;
      //var isOnlyRestartRequired = this.get('parentView.columns').length && this.get('parentView.columns')[2].get('selected');
      var filteredResult = this.get('categoryConfigs').filter(function (config) {

       /* if (isOnlyModified && !config.get('isNotDefaultValue')) {
          return false;
        }

        if (isOnlyRestartRequired && !config.get('isRestartRequired')) {
          return false;
        }*/

        if (isOnlyOverridden && !config.get('isOverridden')) {
          return false;
        }

        var searchString = config.get('defaultValue') + config.get('description') +
          config.get('displayName') + config.get('name');

       if (config.get('overrides')) {
         config.get('overrides').forEach(function(overriddenConf){
           searchString += overriddenConf.get('value') + overriddenConf.get('group.name');
         });
       }

       if (filter != null) {
         return searchString.toLowerCase().indexOf(filter) > -1;
       } else {
         return true;
       }
      });
      filteredResult = this.sortByIndex(filteredResult);
      var categoryBlock = $('.' + this.get('category.name') + '>.accordion-body');
      filteredResult.length && !this.get('category.isCollapsed') ? categoryBlock.show() : categoryBlock.hide();
      return filteredResult;
    }.property('categoryConfigs', 'parentView.filter', 'parentView.columns.@each.selected'),

    /**
     * sort configs in current category by index
     * @param configs
     * @return {*}
     */
    sortByIndex: function (configs) {
      var sortedConfigs = [];
      var unSorted = [];
      if (!configs.someProperty('index')) {
        return configs;
      }
      configs.forEach(function (config) {
        var index = config.get('index');
        if ((index !== null) && isFinite(index)) {
          sortedConfigs[index] ? sortedConfigs.splice(index, 0, config) : sortedConfigs[index] = config;
        } else {
          unSorted.push(config);
        }
      });
      // remove undefined elements from array
      sortedConfigs = sortedConfigs.filter(function (config) {
        return config !== undefined;
      });
      return sortedConfigs.concat(unSorted);
    },
    /**
     * Onclick handler for Config Group Header. Used to show/hide block
     */
    onToggleBlock: function () {
      this.$('.accordion-body').toggle('blind', 500);
      this.set('category.isCollapsed', !this.get('category.isCollapsed'));
    },

    /**
     * Should we show config group or not
     */
    isShowBlock: function () {
      return this.get('category.canAddProperty') || this.get('filteredCategoryConfigs').length > 0;
    }.property('category.canAddProperty', 'filteredCategoryConfigs.length'),

    didInsertElement: function () {
      var isCollapsed = this.get('category.isCollapsed') == undefined ? (this.get('category.name').indexOf('Advanced') != -1 || this.get('category.name').indexOf('CapacityScheduler') != -1) : this.get('category.isCollapsed');
      var self = this;
      this.set('category.isCollapsed', isCollapsed);
      if (isCollapsed) {
        this.$('.accordion-body').hide();
      } else {
        this.$('.accordion-body').show();
      }
      this.updateReadOnlyFlags();
      Em.run.next(function() {
        self.updateReadOnlyFlags();
      });
    },

    /**
     * If added/removed a serverConfigObject, this property got updated.
     * Without this property, all serviceConfigs Objects will show up even if some was collapsed before.
     */
    isCategoryBodyVisible: function () {
      return this.get('category.isCollapsed')? "display: none;" : "display: block;"
    }.property('serviceConfigs.length'),

    childView: App.ServiceConfigsOverridesView,
    changeFlag: Ember.Object.create({
      val: 1
    }),
    isOneOfAdvancedSections: function () {
      var category = this.get('category');
      return category.indexOf("Advanced") != -1;
    },
    showAddPropertyWindow: function (event) {
      var configsOfFile = this.get('service.configs').filterProperty('filename', this.get('category.siteFileName'));
      var self =this;
      var serviceConfigObj = Ember.Object.create({
        name: '',
        value: '',
        defaultValue: null,
        filename: '',
        isUserProperty: true,
        isKeyError: false,
        showFilterLink: false,
        isNotSaved: true,
        errorMessage: "",
        observeAddPropertyValue: function () {
          var name = this.get('name');
          if (name.trim() != "") {
            if (validator.isValidConfigKey(name)) {
              var configMappingProperty = App.config.get('configMapping').all().filterProperty('filename',self.get('category.siteFileName')).findProperty('name', name);
              if ((configMappingProperty == null) && (!configsOfFile.findProperty('name', name))) {
                this.set("isKeyError", false);
                this.set("errorMessage", "");
              } else {
                this.set("showFilterLink", true);
                this.set("isKeyError", true);
                this.set("errorMessage", Em.I18n.t('services.service.config.addPropertyWindow.error.derivedKey'));
              }
            } else {
              this.set("isKeyError", true);
              this.set("errorMessage", Em.I18n.t('form.validator.configKey'));
            }
          } else {
            this.set("isKeyError", true);
            this.set("errorMessage", Em.I18n.t('services.service.config.addPropertyWindow.errorMessage'));
          }
        }.observes("name")
      });

      var category = this.get('category');
      serviceConfigObj.displayType = "advanced";
      serviceConfigObj.category = category.get('name');

      var serviceName = this.get('service.serviceName');
      var serviceConfigsMetaData = App.config.get('preDefinedServiceConfigs');
      var serviceConfigMetaData = serviceConfigsMetaData.findProperty('serviceName', serviceName);
      var categoryMetaData = serviceConfigMetaData == null ? null : serviceConfigMetaData.configCategories.findProperty('name', category.get('name'));
      if (categoryMetaData != null) {
        serviceConfigObj.filename = categoryMetaData.siteFileName;
      }

      var self = this;
      App.ModalPopup.show({
        // classNames: ['big-modal'],
        classNames: [ 'sixty-percent-width-modal'],
        header: "Add Property",
        primary: 'Add',
        secondary: 'Cancel',
        onPrimary: function () {
          serviceConfigObj.observeAddPropertyValue();
          /**
           * For the first entrance use this if (serviceConfigObj.name.trim() != "")
           */
          if (!serviceConfigObj.isKeyError) {
            serviceConfigObj.displayName = serviceConfigObj.name;
            serviceConfigObj.id = 'site property';
            serviceConfigObj.serviceName = serviceName;
            serviceConfigObj.displayType = stringUtils.isSingleLine(serviceConfigObj.get('value')) ? 'advanced' : 'multiLine';
            var serviceConfigProperty = App.ServiceConfigProperty.create(serviceConfigObj);
            self.get('controller.secureConfigs').filterProperty('filename', self.get('category.siteFileName')).forEach(function (_secureConfig) {
              if (_secureConfig.name === serviceConfigProperty.get('name')) {
                serviceConfigProperty.set('isSecureConfig', true);
              }
            }, this);
            self.get('serviceConfigs').pushObject(serviceConfigProperty);
            this.hide();
          }
        },
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/configs/addPropertyWindow'),
          controllerBinding: 'App.router.mainServiceInfoConfigsController',
          serviceConfigProperty: serviceConfigObj,
          filterByKey: function(event) {
            var controller = (App.router.get('currentState.name') != 'configs')
              ? App.router.get('wizardStep7Controller')
              : App.router.get('mainServiceInfoConfigsController');
            this.get('parentView').onClose();
            controller.set('filter', event.view.get('serviceConfigProperty.name'));
          }
        })
      });

    },

    /**
     * Removes the top-level property from list of properties.
     * Should be only called on user properties.
     */
    removeProperty: function (event) {
      var serviceConfigProperty = event.contexts[0];
      this.get('serviceConfigs').removeObject(serviceConfigProperty);
    },

    /**
     * Restores given property's value to be its default value.
     * Does not update if there is no default value.
     */
    doRestoreDefaultValue: function (event) {
      var serviceConfigProperty = event.contexts[0];
      var value = serviceConfigProperty.get('value');
      var dValue = serviceConfigProperty.get('defaultValue');
      if (dValue != null) {
        if (serviceConfigProperty.get('displayType') === 'password') {
          serviceConfigProperty.set('retypedPassword', dValue);
        }
        serviceConfigProperty.set('value', dValue);
      }
      this.miscConfigChange(serviceConfigProperty);
    },

    createOverrideProperty: function (event) {
      var serviceConfigProperty = event.contexts[0];
      var serviceConfigController = this.get('controller');
      var selectedConfigGroup = serviceConfigController.get('selectedConfigGroup');
      var isInstaller = (this.get('controller.name') === 'wizardStep7Controller');
      var configGroups = (isInstaller) ? serviceConfigController.get('selectedService.configGroups') : serviceConfigController.get('configGroups');

      //user property is added, and it has not been saved, not allow override
      if(serviceConfigProperty.get('isUserProperty') && serviceConfigProperty.get('isNotSaved') && !isInstaller){
        App.ModalPopup.show({
          header: Em.I18n.t('services.service.config.configOverride.head'),
          body: Em.I18n.t('services.service.config.configOverride.body'),
          secondary: false
        });
        return;
      }
      if (selectedConfigGroup.get('isDefault')) {
        // Launch dialog to pick/create Config-group
        App.config.launchConfigGroupSelectionCreationDialog(this.get('service.serviceName'),
          configGroups, serviceConfigProperty, function (selectedGroupInPopup) {
            console.log("launchConfigGroupSelectionCreationDialog(): Selected/Created:", selectedGroupInPopup);
            if (selectedGroupInPopup) {
              serviceConfigController.set('overrideToAdd', serviceConfigProperty);
              serviceConfigController.set('selectedConfigGroup', selectedGroupInPopup);
            }
          }, isInstaller);
      } else {
        serviceConfigController.addOverrideProperty(serviceConfigProperty);
      }
    }
  });

  App.ServiceConfigTab = Ember.View.extend({

    tagName: 'li',

    selectService: function (event) {
      this.set('controller.selectedService', event.context);
    },

    didInsertElement: function () {
      var serviceName = this.get('controller.selectedService.serviceName');
      this.$('a[href="#' + serviceName + '"]').tab('show');
    }
  });

  /**
   * custom view for capacity scheduler category
   * @type {*}
   */
  App.ServiceConfigCapacityScheduler = App.ServiceConfigsByCategoryView.extend({
    templateName: require('templates/common/configs/capacity_scheduler'),
    category: null,
    service: null,
    serviceConfigs: null,
    customConfigs: function(){
      return App.config.get('preDefinedCustomConfigs');
    }.property('App.config.preDefinedCustomConfigs'),
    /**
     * configs filtered by capacity-scheduler category
     */
    categoryConfigs: function () {
      return this.get('serviceConfigs').filterProperty('category', this.get('category.name'));
    }.property('queueObserver', 'serviceConfigs.@each'),
    /**
     * rewrote method to avoid incompatibility with parent
     */
    filteredCategoryConfigs: function () {
      return this.get('categoryConfigs');
    }.property(),
    advancedConfigs: function () {
      return this.get('categoryConfigs').filterProperty('isQueue', undefined) || [];
    }.property('categoryConfigs.@each'),
    didInsertElement: function () {
      this._super();
      this.createEmptyQueue(this.get('customConfigs').filterProperty('isQueue'));
    },
    //list of fields which will be populated by default in a new queue
    fieldsToPopulate: function(){
      if(App.get('isHadoop2Stack')){
        return ["yarn.scheduler.capacity.root.<queue-name>.user-limit-factor",
        "yarn.scheduler.capacity.root.<queue-name>.state"];
      }
      return [
        "mapred.capacity-scheduler.queue.<queue-name>.minimum-user-limit-percent",
        "mapred.capacity-scheduler.queue.<queue-name>.user-limit-factor",
        "mapred.capacity-scheduler.queue.<queue-name>.supports-priority",
        "mapred.capacity-scheduler.queue.<queue-name>.maximum-initialized-active-tasks",
        "mapred.capacity-scheduler.queue.<queue-name>.maximum-initialized-active-tasks-per-user",
        "mapred.capacity-scheduler.queue.<queue-name>.init-accept-jobs-factor"
      ];
    }.property('App.isHadoop2Stack'),
    /**
     * create empty queue
     * take some queue then copy it and set all config values to null
     * @param customConfigs
     */
    createEmptyQueue: function (customConfigs) {
      var emptyQueue = {
        name: '<queue-name>',
        configs: []
      };
      var fieldsToPopulate = this.get('fieldsToPopulate');
      customConfigs.forEach(function (config) {
        var newConfig = $.extend({}, config);
        if (fieldsToPopulate.contains(config.name)) {
          newConfig.value = config.defaultValue;
        }
        newConfig = App.ServiceConfigProperty.create(newConfig);
        newConfig.validate();
        emptyQueue.configs.push(newConfig);
      });
      this.set('emptyQueue', emptyQueue);
    },
    deriveQueueNames: function(configs){
      var queueNames = [];
      configs.mapProperty('name').forEach(function(name){
        var queueName;
        if(App.get('isHadoop2Stack')){
          queueName = /^yarn\.scheduler\.capacity\.root\.(.*?)\./.exec(name);
        } else {
          queueName = /^mapred\.capacity-scheduler\.queue\.(.*?)\./.exec(name);
        }
        if(queueName){
          queueNames.push(queueName[1]);
        }
      });
      return queueNames.uniq();
    },
    queues: function(){
      var configs = this.get('categoryConfigs').filterProperty('isQueue', true);
      var queueNames = this.deriveQueueNames(configs);
      var queues = [];

      queueNames.forEach(function(queueName){
        queues.push({
          name: queueName,
          color: this.generateColor(queueName),
          configs: this.groupConfigsByQueue(queueName, configs)
        })
      }, this);
      return queues;
    }.property('queueObserver'),
    /**
     * group configs by queue
     * @param queueName
     * @param configs
     */
    groupConfigsByQueue: function (queueName, configs) {
      var customConfigs = [];
      var queue = [];
      this.get('customConfigs').forEach(function(_config){
        var copy = $.extend({}, _config);
        copy.name = _config.name.replace('<queue-name>', queueName);
        customConfigs.push(copy);
      });
      configs.forEach(function (config) {
        var customConfig = customConfigs.findProperty('name', config.get('name'));
        if (customConfig) {
          config.set('description', customConfig.description);
          config.set('displayName', customConfig.displayName);
          config.set('isRequired', customConfig.isRequired);
          config.set('unit', customConfig.unit);
          config.set('displayType', customConfig.displayType);
          config.set('valueRange', customConfig.valueRange);
          config.set('isVisible', customConfig.isVisible);
          config.set('inTable', customConfig.inTable);
          config.set('index', customConfig.index);
          queue.push(config);
        }
      });
      if(queue.length < customConfigs.length){
        this.addMissingProperties(queue, customConfigs);
      }
      return queue;
    },
    /**
     * add missing properties to queue when they don't come from server
     * @param queue
     * @param customConfigs
     */
    addMissingProperties: function(queue, customConfigs){
      customConfigs.forEach(function(_config){
        var serviceConfigProperty;
        if(!queue.someProperty('name', _config.name)){
          _config.value = _config.defaultValue;
          serviceConfigProperty = App.ServiceConfigProperty.create(_config);
          serviceConfigProperty.validate();
          queue.push(serviceConfigProperty);
        }
      }, this);
    },
    /**
     * convert queues to table content
     */
    tableContent: function () {
      var result = [];
      this.get('queues').forEach(function (queue) {
        var usersAndGroups = queue.configs.findProperty('name', this.getUserAndGroupNames(queue.name)[0]).get('value');
        usersAndGroups = (usersAndGroups) ? usersAndGroups.split(' ') : [''];
        if (usersAndGroups.length == 1) {
          usersAndGroups.push('');
        }
        var queueObject = {
          name: queue.name,
          color: 'background-color:' + queue.color + ';',
          configs: this.sortByIndex(queue.configs.filterProperty('inTable'))
        };
        //push acl_submit_jobs users
        queueObject.configs.unshift({
          value: usersAndGroups[1],
          inTable: true,
          displayName: Em.I18n.t('common.users')
        });
        //push acl_submit_jobs groups
        queueObject.configs.unshift({
          value: usersAndGroups[0],
          inTable: true,
          displayName: Em.I18n.t('services.mapReduce.config.queue.groups')
        });
        result.push(queueObject);
      }, this);
      return result;
    }.property('queues'),
    /**
     * create headers depending on existed properties in queue
     */
    tableHeaders: function(){
      var headers = [
        Em.I18n.t('services.mapReduce.config.queue.name')
      ];
      return (this.get('tableContent').length) ?
        headers.concat(this.get('tableContent').objectAt(0).configs.filterProperty('inTable').mapProperty('displayName')):
        headers;
    }.property('tableContent'),
    queueObserver: null,
    /**
     * uses as template for adding new queue
     */
    emptyQueue: {},
    /**
     * get capacities sum of queues except of current
     * @param queueName
     * @return {Number}
     */
    getQueuesCapacitySum: function(queueName){
      var capacitySum = 0;
      this.get('queues').filter(function(queue){
        return queue.name !== queueName;
      }).forEach(function(queue){
          capacitySum = capacitySum + window.parseInt(queue.configs.find(function(config){
            return config.get('name').substr(-9, 9) === '.capacity';
          }).get('value'));
        });
      return capacitySum;
    },
    /**
     * get names of configs, for users and groups, which have different names in HDP1 and HDP2
     * @param queueName
     * @return {Array}
     */
    getUserAndGroupNames: function(queueName){
      queueName = queueName || '<queue-name>';
      if(App.get('isHadoop2Stack') && this.get('controller.selectedService')) {
        if (this.get('controller.selectedService.serviceName') == "YARN") {
          return ['yarn.scheduler.capacity.root.' + queueName + '.acl_submit_jobs',
            'yarn.scheduler.capacity.root.' + queueName + '.acl_administer_jobs']
        }
        return ['mapred.queue.' + queueName + '.acl-submit-job',
          'mapred.queue.' + queueName + '.acl-administer-jobs']
      }
    },
    generateColor: function (str) {
      var hash = 0;
      for (var i = 0; i < str.length; i++) {
        hash = str.charCodeAt(i) + ((hash << 5) - hash);
      }
      return '#' + Number(Math.abs(hash)).toString(16).concat('00000').substr(0, 6);
    },
    /**
     * add new queue
     * add created configs to serviceConfigs with current queue name
     * @param queue
     */
    addQueue: function (queue) {
      var serviceConfigs = this.get('serviceConfigs');
      var admin = [];
      var submit = [];
      var submitConfig;
      var adminConfig;
      queue.name = queue.configs.findProperty('name', 'queueName').get('value');
      queue.configs.forEach(function (config) {
        var adminName = this.getUserAndGroupNames()[1];
        var submitName = this.getUserAndGroupNames()[0];
        if(config.name == adminName){
          if (config.type == 'USERS') {
            admin[0] = config.value;
          }
          if (config.type == 'GROUPS') {
            admin[1] = config.value;
          }
          if (config.isQueue) {
            adminConfig = config;
          }
        }
        if(config.name == submitName){
          if (config.type == 'USERS') {
            submit[0] = config.value;
          }
          if (config.type == 'GROUPS') {
            submit[1] = config.value;
          }
          if (config.isQueue) {
            submitConfig = config;
          }
        }
        config.set('name', config.get('name').replace('<queue-name>', queue.name));
        config.set('value', config.get('value').toString());
        if (config.isQueue) {
          serviceConfigs.push(config);
        }
      }, this);
      adminConfig.set('value', admin.join(' '));
      submitConfig.set('value', submit.join(' '));
      this.set('queueObserver', App.dateTime());
    },
    /**
     * delete queue
     * delete configs from serviceConfigs which have current queue name
     * @param queue
     */
    deleteQueue: function (queue) {
      var serviceConfigs = this.get('serviceConfigs');
      var configNames = queue.configs.filterProperty('isQueue').mapProperty('name');
      for (var i = 0, l = serviceConfigs.length; i < l; i++) {
        if (configNames.contains(serviceConfigs[i].name)) {
          serviceConfigs.splice(i, 1);
          l--;
          i--;
        }
      }
      this.set('queueObserver', App.dateTime());
    },
    /**
     * save changes that was made to queue
     * edit configs from serviceConfigs which have current queue name
     * @param queue
     */
    editQueue: function (queue) {
      var serviceConfigs = this.get('serviceConfigs');
      var configNames = queue.configs.filterProperty('isQueue').mapProperty('name');
      serviceConfigs.forEach(function (_config) {
        var configName = _config.get('name');
        var admin = [];
        var submit = [];
        //comparison executes including 'queue.<queue-name>' to avoid false matches
        var queueNamePrefix = App.get('isHadoop2Stack') ? 'root.' : 'queue.';
        if (configNames.contains(_config.get('name'))) {
          if(configName == this.getUserAndGroupNames(queue.name)[0]){
            submit = queue.configs.filterProperty('name', configName);
            submit = submit.findProperty('type', 'USERS').get('value') + ' ' + submit.findProperty('type', 'GROUPS').get('value');
            _config.set('value', submit);
          } else if(configName == this.getUserAndGroupNames(queue.name)[1]){
            admin = queue.configs.filterProperty('name', configName);
            admin = admin.findProperty('type', 'USERS').get('value') + ' ' + admin.findProperty('type', 'GROUPS').get('value');
            _config.set('value', admin);
          } else {
            _config.set('value', queue.configs.findProperty('name', _config.get('name')).get('value').toString());
          }
          _config.set('name', configName.replace(queueNamePrefix + queue.name, queueNamePrefix + queue.configs.findProperty('name', 'queueName').get('value')));
        }
      }, this);
      this.set('queueObserver', App.dateTime());
    },
    pieChart: App.ChartPieView.extend({
      w: 200,
      h: 200,
      queues: null,
      didInsertElement: function () {
        this.update();
      },
      data: [{"label":"default", "value":100}],
      update: function () {
        var self = this;
        var data = [];
        var queues = this.get('queues');
        var capacitiesSum = 0;
        queues.forEach(function (queue) {
          var value = window.parseInt(queue.configs.find(function(_config){
            return _config.get('name').substr(-9, 9) === '.capacity';
          }).get('value'));
          data.push({
            label: queue.name,
            value: value,
            color: queue.color
          })
        });

        data.mapProperty('value').forEach(function (capacity) {
          capacitiesSum += capacity;
        });
        if (capacitiesSum < 100) {
          data.push({
            label: Em.I18n.t('common.empty'),
            value: (100 - capacitiesSum),
            color: 'transparent',
            isEmpty: true
          })
        }
        $(d3.select(this.get('selector'))[0]).children().remove();
        this.set('data', data);
        this.set('palette', new Rickshaw.Color.Palette({
          scheme: data.mapProperty('color')
        }));
        this.appendSvg();

        this.get('arcs')
          .on("click",function (d, i) {
            var event = {context: d.data.label};
            if (d.data.isEmpty !== true) self.get('parentView').queuePopup(event);
          }).on('mouseover', function (d, i) {
            var position = d3.svg.mouse(this);
            var label = $('#section_label');
            label.css('left', position[0] + 100);
            label.css('top', position[1] + 100);
            label.text(d.data.label);
            label.show();
          })
          .on('mouseout', function (d, i) {
            $('#section_label').hide();
          })

      }.observes('queues'),
      donut: d3.layout.pie().sort(null).value(function (d) {
        return d.value;
      })
    }),
    /**
     * open popup with chosen queue
     * @param event
     */
    queuePopup: function (event) {
      //if queueName was handed that means "Edit" mode, otherwise "Add" mode
      var queueName = event.context || null;
      var self = this;
      App.ModalPopup.show({
        didInsertElement: function () {
          if (queueName) {
            this.set('header', Em.I18n.t('services.mapReduce.config.editQueue'));
            this.set('secondary', Em.I18n.t('common.save'));
            if (self.get('queues').length > 1 && self.get('canEdit')) {
              this.set('delete', Em.I18n.t('common.delete'));
            }
          }
        },
        header: Em.I18n.t('services.mapReduce.config.addQueue'),
        secondary: Em.I18n.t('common.add'),
        primary: Em.I18n.t('common.cancel'),
        delete: null,
        isError: function () {
          if (!self.get('canEdit')) {
            return true;
          }
          var content = this.get('content');
          var configs = content.configs.filter(function (config) {
            return !(config.name == self.getUserAndGroupNames(content.name)[0] ||
              config.name == self.getUserAndGroupNames(content.name)[1] &&
                config.isQueue);
          });
          return configs.someProperty('isValid', false);
        }.property('content.configs.@each.isValid'),
        onDelete: function () {
          var view = this;
          App.ModalPopup.show({
            header: Em.I18n.t('popup.confirmation.commonHeader'),
            body: Em.I18n.t('hosts.delete.popup.body'),
            primary: Em.I18n.t('yes'),
            onPrimary: function () {
              self.deleteQueue(view.get('content'));
              view.hide();
              this.hide();
            }
          });
        },
        onSecondary: function () {
          if (queueName) {
            self.editQueue(this.get('content'));
          } else {
            self.addQueue(this.get('content'));
          }
          this.hide();
        },
        /**
         * Queue properties order:
         * 1. Queue Name
         * 2. Capacity
         * 3. Max Capacity
         * 4. Users
         * 5. Groups
         * 6. Admin Users
         * 7. Admin Groups
         * 8. Support Priority
         * ...
         */
        content: function () {
          var content = (queueName) ? self.get('queues').findProperty('name', queueName) : self.get('emptyQueue');
          var configs = [];
          // copy of queue configs
          content.configs.forEach(function (config, index) {
            if(config.get('name').substr(-9, 9) === '.capacity') {
              //added validation function for capacity property
              config.reopen({
                validate: function () {
                  var value = this.get('value');
                  var isError = false;
                  var capacitySum = self.getQueuesCapacitySum(content.name);
                  if (value == '') {
                    if (this.get('isRequired')) {
                      this.set('errorMessage', 'This is required');
                      isError = true;
                    } else {
                      return;
                    }
                  }
                  if (!isError) {
                    if (!validator.isValidInt(value)) {
                      this.set('errorMessage', 'Must contain digits only');
                      isError = true;
                    }
                  }
                  if (!isError) {
                    if ((capacitySum + parseInt(value)) > 100) {
                      isError = true;
                      this.set('errorMessage', 'The sum of capacities more than 100');
                    }
                    if (!isError) {
                      this.set('errorMessage', '');
                    }
                  }
                }.observes('value')
              });
            }
            if (config.name == 'mapred.capacity-scheduler.queue.' + content.name + '.supports-priority') {
              if (config.get('value') == 'true' || config.get('value') === true) {
                config.set('value', true);
              } else {
                config.set('value', false);
              }
            }
            if(config.name === 'yarn.scheduler.capacity.root.' + content.name + '.state'){
              config.reopen({
                validate: function(){
                  var value = this.get('value');
                  this._super();
                  if(!this.get('errorMessage')){
                    if(!(value === 'STOPPED' || value === 'RUNNING')){
                      this.set('errorMessage', 'State value should be RUNNING or STOPPED');
                    }
                  }
                }.observes('value')
              })
            }
            configs[index] = App.ServiceConfigProperty.create(config);
          });
          content = {
            name: content.name,
            configs: configs
          };
          content = this.insertExtraConfigs(content);
          content.configs = self.sortByIndex(content.configs);
          return content;
        }.property(),
        footerClass: Ember.View.extend({
          classNames: ['modal-footer', 'host-checks-update'],
          templateName: require('templates/common/configs/queuePopup_footer')
        }),
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/configs/queuePopup_body')
        }),
        /**
         * Insert extra config in popup according to queue
         *
         * the mapred.queue.default.acl-administer-jobs turns into two implicit configs:
         * "Admin Users" field and "Admin Groups" field
         * the mapred.queue.default.acl-submit-job turns into two implicit configs:
         * "Users" field and "Groups" field
         * Add implicit config that contain "Queue Name"
         * @param content
         * @return {*}
         */
        insertExtraConfigs: function (content) {
          var that = this;
          var admin = content.configs.findProperty('name', self.getUserAndGroupNames(content.name)[1]).get('value');
          var submit = content.configs.findProperty('name', self.getUserAndGroupNames(content.name)[0]).get('value');
          admin = (admin) ? admin.split(' ') : [''];
          submit = (submit) ? submit.split(' ') : [''];
          if (admin.length < 2) {
            admin.push('');
          }
          if (submit.length < 2) {
            submit.push('');
          }
          var nameField = App.ServiceConfigProperty.create({
            name: 'queueName',
            displayName: Em.I18n.t('services.mapReduce.extraConfig.queue.name'),
            description: Em.I18n.t('services.mapReduce.description.queue.name'),
            value: (content.name == '<queue-name>') ? '' : content.name,
            validate: function () {
              var queueNames = self.get('queues').mapProperty('name');
              var value = this.get('value');
              var isError = false;
              var regExp = /^[a-z]([\_\-a-z0-9]{0,50})\$?$/i;
              if (value == '') {
                if (this.get('isRequired')) {
                  this.set('errorMessage', 'This is required');
                  isError = true;
                } else {
                  return;
                }
              }
              if (!isError) {
                if ((queueNames.indexOf(value) !== -1) && (value != content.name)) {
                  this.set('errorMessage', 'Queue name is already used');
                  isError = true;
                }
              }
              if (!isError) {
                if (!regExp.test(value)) {
                  this.set('errorMessage', 'Incorrect input');
                  isError = true;
                }
              }
              if (!isError) {
                this.set('errorMessage', '');
              }
            }.observes('value'),
            isRequired: true,
            isVisible: true,
            isEditable: self.get('canEdit'),
            index: 0
          });
          nameField.validate();
          content.configs.unshift(nameField);

          var submitUser = App.ServiceConfigProperty.create({
            name: self.getUserAndGroupNames(content.name)[0],
            displayName: Em.I18n.t('common.users'),
            value: submit[0],
            description: Em.I18n.t('services.mapReduce.description.queue.submit.user'),
            isRequired: true,
            isVisible: true,
            type: 'USERS',
            displayType: "UNIXList",
            isEditable: self.get('canEdit'),
            index: 3
          });

          var submitGroup = App.ServiceConfigProperty.create({
            name: self.getUserAndGroupNames(content.name)[0],
            displayName: Em.I18n.t('services.mapReduce.config.queue.groups'),
            description: Em.I18n.t('services.mapReduce.description.queue.submit.group'),
            value: submit[1],
            isRequired: true,
            isVisible: true,
            "displayType": "UNIXList",
            type: 'GROUPS',
            isEditable: self.get('canEdit'),
            index: 4
          });

          var adminUser = App.ServiceConfigProperty.create({
            name: self.getUserAndGroupNames(content.name)[1],
            displayName: Em.I18n.t('services.mapReduce.config.queue.adminUsers'),
            description: Em.I18n.t('services.mapReduce.description.queue.admin.user'),
            value: admin[0],
            isRequired: true,
            isVisible: true,
            type: 'USERS',
            displayType: "UNIXList",
            isEditable: self.get('canEdit'),
            index: 5
          });

          var adminGroup = App.ServiceConfigProperty.create({
            name: self.getUserAndGroupNames(content.name)[1],
            displayName: Em.I18n.t('services.mapReduce.config.queue.adminGroups'),
            value: admin[1],
            description: Em.I18n.t('services.mapReduce.description.queue.admin.group'),
            isRequired: true,
            isVisible: true,
            "displayType": "UNIXList",
            type: 'GROUPS',
            isEditable: self.get('canEdit'),
            index: 6
          });

          submitUser.reopen({
            validate: function () {
              that.userGroupValidation(this, submitGroup);
            }.observes('value')
          });
          submitGroup.reopen({
            validate: function () {
              that.userGroupValidation(this, submitUser);
            }.observes('value')
          });
          adminUser.reopen({
            validate: function () {
              that.userGroupValidation(this, adminGroup);
            }.observes('value')
          });
          adminGroup.reopen({
            validate: function () {
              that.userGroupValidation(this, adminUser);
            }.observes('value')
          });

          submitUser.validate();
          adminUser.validate();
          content.configs.push(submitUser);
          content.configs.push(submitGroup);
          content.configs.push(adminUser);
          content.configs.push(adminGroup);

          return content;
        },
        /**
         * Validate by follow rules:
         * Users can be blank. If it is blank, Groups must not be blank.
         * Groups can be blank. If it is blank, Users must not be blank.
         * @param context
         * @param boundConfig
         */
        userGroupValidation: function (context, boundConfig) {
          if (context.get('value') == '') {
            if (boundConfig.get('value') == '') {
              context._super();
            } else {
              boundConfig.validate();
            }
          } else {
            if (boundConfig.get('value') == '') {
              boundConfig.set('errorMessage', '');
            }
            context._super();
          }
        }
      })
    }
  });
  
});
window.require.register("views/common/filter_combo_cleanable", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');


  /**
   * Filter component with custom dropdown and cross-circled cleaner. Build on Twitter Bootstrap styles
   * @type {*}
   */
  App.FilterComboCleanableView = Ember.View.extend({
    templateName: require('templates/common/filter_combo_cleanable'),

    classNames: ['filter-combobox', 'input-append'],

    didInsertElement: function() {
      App.popover(this.$("input[type=text]"), {
        title: this.get('popoverDescription')[0],
        content: this.get('popoverDescription')[1],
        placement: 'bottom',
        trigger: 'hover'
      });
    },

    placeHolder: function(){
      return this.t('common.combobox.placeholder');
    }.property(),

    /**
     * Onclick handler for dropdown menu
     * @param event
     */
    selectFilterColumn: function(event){
      var column = event.context;
      column.set('selected', !column.get('selected'));
    },

    filterNotEmpty: function(target){
      return (this.get('filter').length > 0);
    }.property('filter'),

    /**
     * clear Filter textfield
     */
    clearFilter: function() {
      this.set('filter', '');
    }
  });
  
});
window.require.register("views/common/filter_combobox", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');


  /**
   * Filter component with custom dropdown. Build on Twitter Bootstrap styles
   * @type {*}
   */
  App.FilterComboboxView = Ember.View.extend({
    templateName: require('templates/common/filter_combobox'),

    classNames: ['filter-combobox', 'input-append'],

    //filter: '', binding outside
    //columns: [], binding outside

    placeHolder: function(){
      return this.t('common.combobox.placeholder');
    }.property(),

    /**
     * Onclick handler for dropdown menu
     * @param event
     */
    selectFilterColumn: function(event){
      var column = event.context;
      column.set('selected', !column.get('selected'));
    }

  });
  
});
window.require.register("views/common/filter_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Wrapper View for all filter components. Layout template and common actions are located inside of it.
   * Logic specific for data component(input, select, or custom multi select, which fire any changes on interface) are
   * located in inner view - <code>filterView</code>.
   *
   * If we want to have input filter, put <code>textFieldView</code> to it.
   * All inner views implemented below this view.
   * @type {*}
   */

  var App = require('app');

  var wrapperView = Ember.View.extend({
    classNames: ['view-wrapper'],
    layout: Ember.Handlebars.compile('<a href="#" {{action "clearFilter" target="view"}} class="ui-icon ui-icon-circle-close"></a> {{yield}}'),
    template: Ember.Handlebars.compile(
      '{{#if view.fieldId}}<input type="hidden" id="{{unbound view.fieldId}}" value="" />{{/if}}' +
      '{{view view.filterView}}' +
      '{{#if view.showApply}}<button {{action "setValueOnApply" target="view"}} class="apply-btn btn"><span>Apply</span></button>{{/if}} '
    ),

    value: null,

    /**
     * Column index
     */
    column: null,

    /**
     * If this field is exists we dynamically create hidden input element and set value there.
     * Used for some cases, where this values will be used outside of component
     */
    fieldId: null,

    clearFilter: function(){
      this.set('value', this.get('emptyValue'));
      if(this.get('setPropertyOnApply')){
        this.setValueOnApply();
      }
      return false;
    },

    setValueOnApply: function() {
      if(this.get('value') == null){
        this.set('value', '')
      }
      this.set(this.get('setPropertyOnApply'), this.get('value'));
      return false;
    },

    /**
     * Use to determine whether filter is clear or not. Also when we want to set empty value
     */
    emptyValue: '',

    /**
     * Whether our <code>value</code> is empty or not
     * @return {Boolean}
     */
    isEmpty: function(){
      if(this.get('value') === null){
        return true;
      }
      return this.get('value').toString() === this.get('emptyValue').toString();
    },

    /**
     * Show/Hide <code>Clear filter</code> button.
     * Also this method updates computed field related to <code>fieldId</code> if it exists.
     * Call <code>onChangeValue</code> callback when everything is done.
     */
    showClearFilter: function(){
      if(!this.get('parentNode')){
        return;
      }
      // get the sort view element in the same column to current filter view to highlight them together
      var relatedSort = $(this.get('element')).parents('thead').find('.sort-view-' + this.get('column'));
      if(this.isEmpty()){
        this.get('parentNode').removeClass('active-filter');
        this.get('parentNode').addClass('notActive');
        relatedSort.removeClass('active-sort');
      } else {
        this.get('parentNode').removeClass('notActive');
        this.get('parentNode').addClass('active-filter');
        relatedSort.addClass('active-sort');
      }

      if(this.get('fieldId')){
        this.$('> input').eq(0).val(this.get('value'));
      }

      this.onChangeValue();
    }.observes('value'),

    /**
     * Callback for value changes
     */
    onChangeValue: function(){

    },

    /**
     * Filter components is located here. Should be redefined
     */
    filterView: Em.View,

    /**
     * Update class of parentNode(hide clear filter button) on page load
     */
    didInsertElement: function(){
      var parent = this.$().parent();
      this.set('parentNode', parent);
      parent.addClass('notActive');
    }
  });

  /**
   * Simple input control for wrapperView
   */
  var textFieldView = Ember.TextField.extend({
    type:'text',
    placeholder: Em.I18n.t('any'),
    valueBinding: "parentView.value"
  });

  /**
   * Simple multiselect control for wrapperView.
   * Used to render blue button and popup, which opens on button click.
   * All content related logic should be implemented manually outside of it
   */
  var componentFieldView = Ember.View.extend({
    classNames: ['btn-group'],
    classNameBindings: ['isFilterOpen:open:'],

    /**
     * Whether popup is shown or not
     */
    isFilterOpen: false,

    /**
     * We have <code>value</code> property similar to inputs <code>value</code> property
     */
    valueBinding: 'parentView.value',

    /**
     * Clear filter to initial state
     */
    clearFilter: function(){
      this.set('value', '');
    },

    /**
     * Onclick handler for <code>cancel filter</code> button
     */
    closeFilter:function () {
      $(document).unbind('click');
      this.set('isFilterOpen', false);
    },

    /**
     * Onclick handler for <code>apply filter</code> button
     */
    applyFilter:function() {
      this.closeFilter();
    },

    /**
     * Onclick handler for <code>show component filter</code> button.
     * Also this function is used in some other places
     */
    clickFilterButton:function () {
      var self = this;
      this.set('isFilterOpen', !this.get('isFilterOpen'));
      if (this.get('isFilterOpen')) {

        var dropDown = this.$('.filter-components');
        var firstClick = true;
        $(document).bind('click', function (e) {
          if (!firstClick && $(e.target).closest(dropDown).length == 0) {
            self.set('isFilterOpen', false);
            $(document).unbind('click');
          }
          firstClick = false;
        });
      }
    }
  });

  /**
   * Simple select control for wrapperView
   */
  var selectFieldView = Ember.Select.extend({
    selectionBinding: 'parentView.value',
    contentBinding: 'parentView.content'
  });

  /**
   * Result object, which will be accessible outside
   * @type {Object}
   */
  module.exports = {
    /**
     * You can access wrapperView outside
     */
    wrapperView : wrapperView,

    /**
     * And also controls views if need it
     */
    textFieldView : textFieldView,
    selectFieldView: selectFieldView,
    componentFieldView: componentFieldView,

    /**
     * Quick create input filters
     * @param config parameters of <code>wrapperView</code>
     */
    createTextView : function(config){
      config.fieldType = config.fieldType || 'input-medium';
      config.filterView = textFieldView.extend({
        classNames : [ config.fieldType ]
      });

      return wrapperView.extend(config);
    },

    /**
     * Quick create multiSelect filters
     * @param config parameters of <code>wrapperView</code>
     */
    createComponentView : function(config){
      config.clearFilter = function(){
        this.forEachChildView(function(item){
          if(item.clearFilter){
            item.clearFilter();
          }
        });
        return false;
      };

      return wrapperView.extend(config);
    },

    /**
     * Quick create select filters
     * @param config parameters of <code>wrapperView</code>
     */
    createSelectView: function(config){

      config.fieldType = config.fieldType || 'input-medium';
      config.filterView = selectFieldView.extend({
        classNames : [ config.fieldType ],
        attributeBindings: ['disabled','multiple'],
        disabled: false
      });
      config.emptyValue = Em.I18n.t('any');

      return wrapperView.extend(config);
    },
    /**
     * returns the filter function, which depends on the type of property
     * @param type
     * @param isGlobal check is search global
     * @return {Function}
     */
    getFilterByType: function(type, isGlobal){
      switch (type){
        case 'ambari-bandwidth':
          return function(rowValue, rangeExp){
            var compareChar = isNaN(rangeExp.charAt(0)) ? rangeExp.charAt(0) : false;
            var compareScale = rangeExp.charAt(rangeExp.length - 1);
            var compareValue = compareChar ? parseFloat(rangeExp.substr(1, rangeExp.length)) : parseFloat(rangeExp.substr(0, rangeExp.length));
            var match = false;
            if (rangeExp.length == 1 && compareChar !== false) {
              // User types only '=' or '>' or '<', so don't filter column values
              match = true;
              return match;
            }
            switch (compareScale) {
              case 'g':
                compareValue *= 1073741824;
                break;
              case 'm':
                compareValue *= 1048576;
                break;
              case 'k':
                compareValue *= 1024;
                break;
              default:
                //default value in GB
                compareValue *= 1073741824;
            }
            rowValue = (jQuery(rowValue).text()) ? jQuery(rowValue).text() : rowValue;

            var convertedRowValue;
            if (rowValue === '<1KB') {
              convertedRowValue = 1;
            } else {
              var rowValueScale = rowValue.substr(rowValue.length - 2, 2);
              switch (rowValueScale) {
                case 'KB':
                  convertedRowValue = parseFloat(rowValue)*1024;
                  break;
                case 'MB':
                  convertedRowValue = parseFloat(rowValue)*1048576;
                  break;
                case 'GB':
                  convertedRowValue = parseFloat(rowValue)*1073741824;
                  break;
              }
            }

            switch (compareChar) {
              case '<':
                if (compareValue > convertedRowValue) match = true;
                break;
              case '>':
                if (compareValue < convertedRowValue) match = true;
                break;
              case false:
              case '=':
                if (compareValue == convertedRowValue) match = true;
                break;
            }
            return match;
          };
          break;
        case 'duration':
          return function (rowValue, rangeExp) {
            var compareChar = isNaN(rangeExp.charAt(0)) ? rangeExp.charAt(0) : false;
            var compareScale = rangeExp.charAt(rangeExp.length - 1);
            var compareValue = compareChar ? parseFloat(rangeExp.substr(1, rangeExp.length)) : parseFloat(rangeExp.substr(0, rangeExp.length));
            var match = false;
            if (rangeExp.length == 1 && compareChar !== false) {
              // User types only '=' or '>' or '<', so don't filter column values
              match = true;
              return match;
            }
            switch (compareScale) {
              case 's':
                compareValue *= 1000;
                break;
              case 'm':
                compareValue *= 60000;
                break;
              case 'h':
                compareValue *= 3600000;
                break;
              default:
                compareValue *= 1000;
            }
            rowValue = (jQuery(rowValue).text()) ? jQuery(rowValue).text() : rowValue;

            switch (compareChar) {
              case '<':
                if (compareValue > rowValue) match = true;
                break;
              case '>':
                if (compareValue < rowValue) match = true;
                break;
              case false:
              case '=':
                if (compareValue == rowValue) match = true;
                break;
            }
            return match;
          };
          break;
        case 'date':
          return function (rowValue, rangeExp) {
            var match = false;
            var timePassed = App.dateTime() - rowValue;
            switch (rangeExp) {
              case 'Past 1 hour':
                match = timePassed <= 3600000;
                break;
              case 'Past 1 Day':
                match = timePassed <= 86400000;
                break;
              case 'Past 2 Days':
                match = timePassed <= 172800000;
                break;
              case 'Past 7 Days':
                match = timePassed <= 604800000;
                break;
              case 'Past 14 Days':
                match = timePassed <= 1209600000;
                break;
              case 'Past 30 Days':
                match = timePassed <= 2592000000;
                break;
              case 'Any':
                match = true;
                break;
            }
            return match;
          };
          break;
        case 'number':
          return function(rowValue, rangeExp){
            var compareChar = rangeExp.charAt(0);
            var compareValue;
            var match = false;
            if (rangeExp.length == 1) {
              if (isNaN(parseInt(compareChar))) {
                // User types only '=' or '>' or '<', so don't filter column values
                match = true;
                return match;
              }
              else {
                compareValue = parseFloat(parseFloat(rangeExp).toFixed(2));
              }
            }
            else {
              if (isNaN(parseInt(compareChar))) {
                compareValue = parseFloat(parseFloat(rangeExp.substr(1, rangeExp.length)).toFixed(2));
              }
              else {
                compareValue = parseFloat(parseFloat(rangeExp.substr(0, rangeExp.length)).toFixed(2));
              }
            }
            rowValue = parseFloat((jQuery(rowValue).text()) ? jQuery(rowValue).text() : rowValue);
            match = false;
            switch (compareChar) {
              case '<':
                if (compareValue > rowValue) match = true;
                break;
              case '>':
                if (compareValue < rowValue) match = true;
                break;
              case '=':
                if (compareValue == rowValue) match = true;
                break;
              default:
                if (rangeExp == rowValue) match = true;
            }
            return match;
          };
          break;
        case 'multiple':
          return function(origin, compareValue){
            var options = compareValue.split(',');
            if(typeof (origin) === "string"){
              var rowValue = origin;
            }else{
              var rowValue = origin.mapProperty('componentName').join(" ");
            }
            var str = new RegExp(compareValue, "i");
            for (var i = 0; i < options.length; i++) {
              if(!isGlobal) {
                str = new RegExp('(\\W|^)' + options[i] + '(\\W|$)');
              }
              if (rowValue.search(str) !== -1) {
                return true;
              }
            }
            return false;
          };
          break;
        case 'boolean':
          return function (origin, compareValue){
            return origin === compareValue;
          };
          break;
        case 'string':
        default:
          return function(origin, compareValue){
            var regex = new RegExp(compareValue,"i");
            return regex.test(origin);
          }
      }
    }
  };
  
});
window.require.register("views/common/form/field", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.FormFieldTemplate = Em.View.extend({
    templateDir: 'templates/common/form/',
    defaultTemplate: 'field',
    classNames: ["control-group"],
    init: function(){
      this._super();
      this.updateTemplate();
    },
    updateTemplate: function(){
      var field = this.get('field');
      if(field) {
        var templateName = '';
        switch(field.get('displayType')) {
          case 'checkbox':
            templateName = field.get('displayType');
            break;
          default:
            templateName = this.get('defaultTemplate');
            break;
        }

        this.set('templateName', require(this.get('templateDir') + templateName));
      }
    }.observes('field')
  });
});
window.require.register("views/common/grid", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var validator = require('utils/validator');

  App.GridFilterObject = Em.Object.extend({
    checked:false
  });

  App.GridFilter = Em.View.extend({
    tagName:"ul",
    classNames:['filter'],
    templateName:require('templates/common/grid/filter'),
    attributeBindings:['style'],
    getHeader:function () {
      return this.get('header')
    },
    filters:function () {
      return this.get('header._filters');
    }.property('header._filters')
  });

  App.GridHeader = Em.View.extend({
    templateName:require('templates/common/grid/header'),
    tagName:'th',
    filterable:true,
    showFilter:false,
    getGrid:function () {
      return this.get('grid');
    },
    _filters:[],
    doFilter:function () {
      console.log(this.get('grid'));
    },
    toggleFilter:function () {
      this.set('showFilter', 1 - this.get('showFilter'));
    },
    applyFilter:function () {
      console.warn('APPLYING FILTERS');

      var filters = this.get('_filters');
      var filterValues = [];
      $.each(filters, function(){
        if(this.get('checked')) {
          filterValues.push(this.get('value'));
        }
      });

      var grid = this.get('grid');
      grid.addFilters(this.get('name'), filterValues);
      this.set('showFilter', false);
    },
    init:function () {
      this._super();
      if (!this.get('_filters').length) {
        this.filterValues();
        var thisHeader = this;
        this.set('filter', App.GridFilter.extend({ header:thisHeader }));
      }
    },

    filterValues:function () {
      var gridFilters = this.get('grid._filters');
      if (gridFilters && gridFilters[this.get('name')]) {
        var filters = this.get('grid._filters')[this.get('name')];
        // there should be something like filter preparing
        var newFilters = [];
        $.each(filters, function (i, v) {
          newFilters.push(App.GridFilterObject.create({label:v, value:v}));
        });

        this.set('_filters', newFilters);
      }
    }.observes('grid._filters')
  });

  App.GridRow = Em.View.extend({
    tagName:'tr',
    init:function (options) {
      var object = this.get('object');
      var grid = this.get('grid');
      var fieldNames = grid.get('fieldNames');
      var template = '';

      if (fieldNames) {
        $.each(grid.get('fieldNames'), function (i, field) {
          template += "<td>" + object.get(field) + "</td>";
        });

        this.set('template', Em.Handlebars.compile(template));
      }
      return this._super();
    }
  });

  App.GridPage = Em.Object.extend({
    activeClass:function () {
      return this.get('active') ? "active" : "";
    }.property('active'),
    active:function () {
      return parseInt(this.get('number')) == parseInt(this.get('pager.grid.currentPage'));
    }.property('pager.grid.currentPage')
  });

  App.GridPager = Em.View.extend({

    pages:[],
    templateName:require('templates/common/grid/pager'),
    classNames:['pagination'],

    activatePrevPage:function () {
      var current = this.get('grid.currentPage');
      if (current > 1) this.set('grid.currentPage', current - 1);
    },
    activateNextPage:function () {
      var current = this.get('grid.currentPage');
      if (current < this.get('pages').length) this.set('grid.currentPage', current + 1);
    },

    prevPageDisabled:function () {
      return this.get('grid.currentPage') > 1 ? false : "disabled";
    }.property('grid.currentPage'),

    nextPageDisabled:function () {
      return this.get('grid.currentPage') < this.get('pages').length ? false : "disabled";
    }.property('grid.currentPage'),

    init:function () {
      this._super();
      this.clearPages();
      this.pushPages();
    },

    activatePage:function (event) {
      var page = event.context;
      this.get('grid').set('currentPage', parseInt(event.context.get('number')));
    },

    clearPages:function () {
      this.set('pages', []);
    },

    pushPages:function () {
      var thisPager = this;
      var pages = this.get('grid._pager.pages');
      $.each(pages, function () {
        var thisNumber = this;
        thisPager.get('pages').push(App.GridPage.create({
          number:thisNumber,
          pager:thisPager
        }));
      })
    }.observes('grid._pager')
  });

  App.Grid = Em.View.extend({
    _columns:{}, // not used
    _filters:{}, // prepared filters from data values
    _pager:{pages:[1, 2, 3, 4, 5]}, // observed by pager to config it

    _collection:{className:false, staticOptions:{}}, // collection config
    currentPage:1,
    fieldNames:[],
    appliedFilters:{},
    filteredArray:[],
    columns:[],
    collection:[],
    initComleted:false,
    rows:[],
    templateName:require('templates/main/admin/audit'),

    init:function () {
      console.warn("  Grid INIT  ");
      this._super();
      this.prepareColumns(); // should be the 1
      this.prepareCollection();
      this.preparePager();
    },

    preparePager:function () {
  //    this.set('pager', App.GridPager.extend({ grid:this })); ask to hide
    },

    addFilters: function(field, values){
      var filters = this.get('appliedFilters');
      filters[field] = values;

      var collection = this.get('_collection.className');
      collection = collection.find();
      var arrayCollection = collection.filter(function(data) {
        var oneFilterFail = false;
        $.each(filters, function(fieldname, values){
          if(values.length && values.indexOf(data.get(fieldname)) == -1) {
            return oneFilterFail = true;
          }
        });
        return !oneFilterFail;
      });

      this.set('filteredArray', arrayCollection);
    },

    prepareCollection:function () {
      if (validator.empty(this.get('_collection.className'))) {
        throw "_collection.className field is not defined";
      }
      var collection = this.get('_collection.className');
      this.set('collection', collection.find(this.get('_collection.staticOptions')));
    },

    addColumn:function (options) {
      options.grid = this;
      if (validator.empty(options.name)) {
        throw "define column name";
      }

      if (this.get('_columns.' + options.name)) {
        throw "column with this '" + options.name + "' already exists";
      }

      var field = App.GridHeader.extend(options);
      this.columns.push(field);

      if (field.filterable || 1) { // .filterable - field not working :(
        this.fieldNames.push(options.name);
      }
    },

    clearColumns:function () {
      this.set('_columns', {});
      this.set('columns', []);
      this.set('fieldNames', []);
    },

    prepareColumns:function () {
      this.clearColumns();
    },

    prepareFilters:function () {
      var thisGrid = this;
      var collection = this.get('collection');
      var fieldNames = this.get('fieldNames');
      var options = {};

      if (collection && collection.content) {
        collection.forEach(function (object, i) {
          $.each(fieldNames, function (j, field) {
            if (!options[field]) {
              options[field] = [];
            }

            var filter = object.get(field);
            if (options[field].indexOf(filter) == -1) {
              options[field].push(filter);
            }
          });
        });

        thisGrid.set('_filters', options);
      }
    }.observes('collection.length'),


    clearRows:function () {
      this.set('rows', [])
    },

    prepareRows:function () {
      var collection = this.get('collection');
      var thisGrid = this;
      this.clearRows();
      console.warn("PREPARE ROWS LEN:", collection.get('length'));
      var i=1;

      if (collection && collection.content) {
        collection.forEach(function (object, i) {
          var row = App.GridRow.extend({grid:thisGrid, object:object});
          thisGrid.rows.push(row);
        });
      }
    }.observes('collection.length'),

    filteredRows:function () {
      var collection = this.get('filteredArray');
      var thisGrid = this;
      this.clearRows();

      collection.forEach(function (object) {
        var row = App.GridRow.extend({grid:thisGrid, object:object});
        thisGrid.rows.push(row);
      });
    }.observes('filteredArray')
  });
});
window.require.register("views/common/metric", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * use: {{view App.MetricFilteringWidget controllerBinding="App.router.mainChartsController"}}
   * set controller.metric field with metric value
   * widget assign itself to controller like metricWidget (controller.get('metricWidget'))
   * @type {*}
   */
  App.MetricFilteringWidget = Em.View.extend({
    classNames:['metric-filtering-widget'],
    /**
     * chosen metric value
     */
    chosenMetric:null,
    chosenMoreMetric:null,
    showMore:0, // toggle more metrics indicator
    /**
     * metrics
     */
    metrics:[
      Em.Object.create({ label:Em.I18n.t('metric.default'), value:null}),
      Em.Object.create({ label:Em.I18n.t('metric.cpu'), value:'cpu'}),
      Em.Object.create({ label:Em.I18n.t('metric.memory'), value:'memory'}),
      Em.Object.create({ label:Em.I18n.t('metric.network'), value:'network'}),
      Em.Object.create({ label:Em.I18n.t('metric.io'), value:'io'})
    ],


    moreMetrics:[
      Em.Object.create({ label:Em.I18n.t('metric.more.cpu'), code:'cpu', items:[
        Em.Object.create({value:"cpu_nice"}),
        Em.Object.create({value:"cpu_wio"}),
        Em.Object.create({value:"cpu_user"}),
        Em.Object.create({value:"cpu_idle"}),
        Em.Object.create({value:"cpu_system"}),
        Em.Object.create({value:"cpu_aidle"})
      ] }),

      Em.Object.create({ label:Em.I18n.t('metric.more.disk'), code:'disk',
        items:[
          Em.Object.create({value:'disk_free'}),
          Em.Object.create({value:'disk_total'}),
          Em.Object.create({value:'part_max_used'})
        ]
      }),

      Em.Object.create({ label:Em.I18n.t('metric.more.load'), code:'load',
        items:[
          Em.Object.create({value:'load_one'}),
          Em.Object.create({value:'load_five'}),
          Em.Object.create({value:'load_fifteen'})
        ]
      }),

      Em.Object.create({ label:Em.I18n.t('metric.more.memory'), code:'memory',
        items:[
          Em.Object.create({value:'swap_free'}),
          Em.Object.create({value:'cpu'})
        ]
      }),

      Em.Object.create({ label:Em.I18n.t('metric.more.network'), code:'network',
        items:[
          Em.Object.create({value:'bytes_out'}),
          Em.Object.create({value:'bytes_in'}),
          Em.Object.create({value:'pkts_in'}),
          Em.Object.create({value:'pkts_out'})
        ]
      }),

      Em.Object.create({ label:Em.I18n.t('metric.more.process'), code:'process',
        items:[
          Em.Object.create({value:'proc_run'}),
          Em.Object.create({value:'proc_total'})
        ]
      })

    ],

    /**
     * return array of chosen metrics
     */
    chosenMetrics:function () {
      return this.get('chosenMetric') ? [this.get('chosenMetric')] : this.get('defaultMetrics');
    }.property('chosenMetric'),

    /**
     * metric item view
     */
    itemView:Em.View.extend({
      tagName:'li',
      classNameBindings:['disabled'],
      disabled:function () {
        return this.get('isActive') ? "disabled" : false;
      }.property('isActive'),
      isActive:function () {
        return this.get('metric.value') == this.get('widget.chosenMetric');
      }.property('widget.chosenMetric'),
      label:function () {
        return this.get('metric.label');
      }.property('metric.label'),
      template:Em.Handlebars.compile('<a {{action activate view.metric.value target="view.widget" href="#" }}>{{unbound view.label}}</a>')
    }),

    moreItemView:function () {
      return this.get('itemView').extend({
        label:function () {
          return this.get('metric.value');
        }.property('metric.value')
      });
    }.property(),

    /**
     * return default selected metrics (currently - all)
     */
    defaultMetrics:function () {
      var values = [];
      $.each(this.get('metrics'), function () {
        if (this.value) {
          values.push(this.value);
        }
      });
      return values;
    }.property(),

    bindToController:function () {
      var thisW = this;
      var controller = this.get('controller');
      controller.set('metricWidget', thisW);
    },

    toggleMore:function () {
      this.set('showMore', 1 - this.get('showMore'));
    },

    /**
     * assign this widget to controller, prepare items by metricsConfig
     */
    init:function () {
      this._super();
      this.bindToController();
    },

    /**
     * write active metric to widget
     * @param event
     */
    activate:function (event) {
      this.set('chosenMetric', event.context);
    },

    templateName:require('templates/common/metric')
  });
});
window.require.register("views/common/modal_popup", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ModalPopup = Ember.View.extend({

    viewName: 'modalPopup',
    templateName: require('templates/common/modal_popup'),
    header: '&nbsp;',
    body: '&nbsp;',
    encodeBody: true,
    // define bodyClass which extends Ember.View to use an arbitrary Handlebars template as the body
    primary: Em.I18n.t('ok'),
    secondary: Em.I18n.t('common.cancel'),
    autoHeight: true,
    disablePrimary: false,
    disableSecondary: false,
    primaryClass: 'btn-success',
    onPrimary: function () {
      this.hide();
    },

    onSecondary: function () {
      this.hide();
    },

    onClose: function () {
      this.hide();
    },

    hide: function () {
      this.destroy();
    },

    showFooter: true,

    /**
     * Hide or show 'X' button for closing popup
     */
    showCloseButton: true,

    didInsertElement: function () {
      if (this.autoHeight) {
        var block = this.$().find('#modal > .modal-body').first();
        block.css('max-height', $(window).height() - block.offset().top - 300 + $(window).scrollTop()); // fix popup height
      }
      // If popup is opened from another popup it should be displayed above
      var existedPopups = $(document).find('.modal-backdrop');
      if (existedPopups) {
        var maxZindex = 1;
        existedPopups.each(function(index, popup) {
          if ($(popup).css('z-index') > maxZindex) {
            maxZindex = $(popup).css('z-index');
          }
        });
        this.$().find('.modal-backdrop').css('z-index', maxZindex * 2);
        this.$().find('.modal').css('z-index', maxZindex * 2 + 1);
      }
    },

    fitHeight: function () {
      var popup = this.$().find('#modal');
      var block = this.$().find('#modal > .modal-body');
      var wh = $(window).height();

      var top = wh * .05;
      popup.css({
        'top': top + 'px',
        'marginTop': 0
      });

      block.css('max-height', $(window).height() - top * 2 - (popup.height() - block.height()));
    }
  });

  App.ModalPopup.reopenClass({

    show: function (options) {
      var popup = this.create(options);
      popup.appendTo('#wrapper');
      return popup;
    }

  });

  App.showReloadPopup = function () {
    return App.ModalPopup.show({
      primary: null,
      secondary: null,
      showFooter: false,
      header: this.t('app.reloadPopup.header'),
      body: "<div id='reload_popup' class='alert alert-info'><div class='spinner'><span>" + this.t('app.reloadPopup.text') + "</span></div></div><div><a href='#' onclick='location.reload();'>" + this.t('app.reloadPopup.link') + "</a></div>",
      encodeBody: false
    });
  };

  /**
   * Show confirmation popup
   *
   * @param {Function} primary - "OK" button click handler
   * @param {String} body - additional text constant. Will be placed in the popup-body
   * @param {Function} secondary
   * @return {*}
   */
  App.showConfirmationPopup = function (primary, body, secondary) {
    if (!primary) {
      return false;
    }
    return App.ModalPopup.show({
      encodeBody: false,
      header: Em.I18n.t('popup.confirmation.commonHeader'),
      body: body || Em.I18n.t('question.sure'),
      onPrimary: function () {
        this.hide();
        primary();
      },
      onSecondary: function () {
        this.hide();
        if (secondary) {
          secondary();
        }
      }
    });
  };

  /**
   * Show confirmation popup
   * After sending command watch status of query,
   * and in case of failure provide ability to retry to launch an operation.
   *
   * @param {Function} primary - "OK" button click handler
   * @param {Object} bodyMessage - confirmMsg:{String},
                                   confirmButton:{String},
                                   additionalWarningMsg:{String},
   * @param {Function} secondary - "Cancel" button click handler
   * @return {*}
   */
  App.showConfirmationFeedBackPopup = function (primary, bodyMessage, secondary) {
    if (!primary) {
      return false;
    }
    return App.ModalPopup.show({
      header: Em.I18n.t('popup.confirmation.commonHeader'),
      bodyClass: Em.View.extend({
        templateName: require('templates/common/confirmation_feedback')
      }),
      query: Em.Object.create({status: "INIT"}),
      primary: function () {
        return bodyMessage? bodyMessage.confirmButton : Em.I18n.t('ok');
      }.property('bodyMessage'),
      onPrimary: function () {
        this.set('query.status', "INIT");
        this.set('disablePrimary', true);
        this.set('disableSecondary', true);
        this.set('statusMessage', Em.I18n.t('popup.confirmationFeedBack.sending'));
        primary(this.get('query'));
      },
      statusMessage: function () {
        return bodyMessage? bodyMessage.confirmMsg : Em.I18n.t('question.sure');
      }.property('bodyMessage'),
      additionalWarningMsg: function () {
        return bodyMessage? bodyMessage.additionalWarningMsg : null;
      }.property('bodyMessage'),
      watchStatus: function() {
        if (this.get('query.status') === "SUCCESS") {
          this.hide();
        } else if(this.get('query.status') === "FAIL") {
          this.set('primaryClass', 'btn-primary');
          this.set('primary', Em.I18n.t('common.retry'));
          this.set('disablePrimary', false);
          this.set('disableSecondary', false);
          this.set('statusMessage', Em.I18n.t('popup.confirmationFeedBack.query.fail'));
        }
      }.observes('query.status'),
      onSecondary: function () {
        this.hide();
        if (secondary) {
          secondary();
        }
      }
    });
  };

  /**
   * Show alert popup
   *
   * @param {String} header - header of the popup
   * @param {String} body - body of the popup
   * @param {Function} primary - function to call upon clicking the OK button
   * @return {*}
   */
  App.showAlertPopup = function (header, body, primary) {
    return App.ModalPopup.show({
      primary: Em.I18n.t('ok'),
      secondary: null,
      header: header,
      body: body,
      onPrimary: function () {
        this.hide();
        if (primary) {
          primary();
        }
      }
    });
  };

  /**
   * Show prompt popup
   *
   * @param {String} text - additional text constant. Will be placed on the top of the input field
   * @param {Function} primary - "OK" button click handler
   * @param {String} defaultValue - additional text constant. Will be default value for input field
   * @param {Function} secondary
   * @return {*}
   */
  App.showPromptPopup = function (text, primary, defaultValue, secondary) {
    if (!primary) {
      return false;
    }
    return App.ModalPopup.show({
      header: Em.I18n.t('popup.prompt.commonHeader'),
      bodyClass: Em.View.extend({
        templateName: require('templates/common/prompt_popup'),
        text: text
      }),
      inputValue: defaultValue || '',
      onPrimary: function () {
        this.hide();
        primary(this.get('inputValue'));
      },
      onSecondary: function () {
        this.hide();
        if (secondary) {
          secondary();
        }
      }
    });
  };
});
window.require.register("views/common/quick_view_link_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.QuickViewLinks = Em.View.extend({

    loadTags: function () {
      App.ajax.send({
        name: 'config.tags.sync',
        sender: this,
        success: 'loadTagsSuccess'
      });
    },

    loadTagsSuccess: function (data) {
      var tags = [];
      for (var prop in data.Clusters.desired_configs) {
        tags.push(Em.Object.create({
          siteName: prop,
          tagName: data.Clusters.desired_configs[prop]['tag']
        }));
      }
      this.set('actualTags', tags);
      this.setConfigProperties();
    },

    actualTags: [],

    configProperties: [],

    /**
     * list of files that contains properties for enabling/disabling ssl
     */
    requiredSiteNames: ['global','core-site', 'hdfs-site', 'hbase-site', 'oozie-site', 'yarn-site', 'mapred-site'],

    setConfigProperties: function () {
      this.set('configProperties', []);
      var requiredSiteNames = this.get('requiredSiteNames');
      var tags = this.get('actualTags').filter(function (tag) {
        return requiredSiteNames.contains(tag.siteName);
      });
      var data = App.router.get('configurationController').getConfigsByTags(tags);
      this.set('configProperties', data);

    },

    ambariProperties: function () {
      return App.router.get('clusterController.ambariProperties');
    },
    /**
     * Updated quick links. Here we put correct hostname to url
     */
    quickLinks: [],

    didInsertElement: function () {
      this.setQuickLinks();
    },

    findComponentHost: function (componentName) {
      var components = this.get('content.hostComponents');
      return App.singleNodeInstall ? App.singleNodeAlias : components.findProperty('componentName', componentName).get('host.publicHostName')
    },

    setQuickLinks: function () {
      this.loadTags();
      var serviceName = this.get('content.serviceName');
      var components = this.get('content.hostComponents');
      var host;
      var self = this;
      var version = App.get('currentStackVersionNumber');
      var quickLinks = [];
      switch (serviceName) {
        case "HDFS":
          if (this.get('content.snameNode')) { // not HA
            host = this.findComponentHost('NAMENODE');
          } else {
            // HA
            if (this.get('content.activeNameNode')) {
              host = this.get('content.activeNameNode.publicHostName');
            } else {
              host = 'noActiveNN';
            }
          }
          break;
        case "MAPREDUCE":
        case "OOZIE":
        case "GANGLIA":
        case "NAGIOS":
        case "HUE":
          host = App.singleNodeInstall ? App.singleNodeAlias : components.findProperty('isMaster', true).get("host").get("publicHostName");
          break;
        case "HBASE":
          var component;
          if (App.supports.multipleHBaseMasters) {
            component = components.filterProperty('componentName', 'HBASE_MASTER').findProperty('haStatus', 'true');
          } else {
            component = components.findProperty('componentName', 'HBASE_MASTER');
          }
          if (component) {
            if (App.singleNodeInstall) {
              host = App.singleNodeAlias;
            } else {
              host = component.get('host.publicHostName');
            }
          } else {
            host = 'noActiveHbaseMaster';
          }
          break;
        case "YARN":
          host = this.findComponentHost('RESOURCEMANAGER');
          break;
        case "MAPREDUCE2":
          host = this.findComponentHost('HISTORYSERVER');
          break;
        case "FALCON":
          host = this.findComponentHost('FALCON_SERVER');
          break;
        case "STORM":
          host = this.findComponentHost('NIMBUS');
          break;
      }
      if (!host) {
        quickLinks = [
          {
            label: this.t('quick.links.error.label'),
            url: 'javascript:alert("' + this.t('contact.administrator') + '");return false;'
          }
        ];
      } else {
        quickLinks = this.get('content.quickLinks').map(function (item) {
          if (host == 'noActiveNN' || host == 'noActiveHbaseMaster') {
            item.set('disabled', true);
          } else {
            item.set('disabled', false);
            var protocol = self.setProtocol(item.get('service_id'));
            if (item.get('template')) {
              var port = item.get('http_config') && self.setPort(item, protocol, version);
              item.set('url', item.get('template').fmt(protocol, host, port));
            }
          }
          return item;
        });
      }
      this.set('quickLinks', quickLinks);
    }.observes('App.currentStackVersionNumber', 'App.singleNodeInstall'),

    setProtocol: function (service_id) {
      var properties = this.ambariProperties();
      var configProperties = this.get('configProperties');
      var hadoopSslEnabled = false;
      if (configProperties) {
        var site = configProperties.findProperty('type', 'core-site');
        site.properties['hadoop.ssl.enabled'] && site.properties['hadoop.ssl.enabled'] === 'true' ? hadoopSslEnabled = true : null;
      }
      switch (service_id) {
        case "GANGLIA":
          return (properties && properties.hasOwnProperty('ganglia.https') && properties['ganglia.https']) ? "https" : "http";
          break;
        case "NAGIOS":
          return (properties && properties.hasOwnProperty('nagios.https') && properties['nagios.https']) ? "https" : "http";
          break;
        case "HDFS":
        case "HBASE":
        case "MAPREDUCE":
          return hadoopSslEnabled ? "https" : "http";
          break;
        case "YARN":
          var yarnProperties = configProperties.findProperty('type', 'yarn-site');
          if (yarnProperties && yarnProperties.properties) {
            if (yarnProperties.properties['yarn.http.policy'] === 'HTTPS_ONLY') {
              return "https";
            } else if (yarnProperties.properties['yarn.http.policy'] === 'HTTP_ONLY') {
              return "http";
            }
          }
          return hadoopSslEnabled ? "https" : "http";
          break;
        case "MAPREDUCE2":
          var mapred2Properties = configProperties.findProperty('type', 'mapred-site');
          if (mapred2Properties && mapred2Properties.properties) {
            if (mapred2Properties.properties['mapreduce.jobhistory.http.policy'] === 'HTTPS_ONLY') {
              return "https";
            } else if (mapred2Properties.properties['mapreduce.jobhistory.http.policy'] === 'HTTP_ONLY') {
              return "http";
            }
          }
          return hadoopSslEnabled ? "https" : "http";
          break;
        default:
          return "http";
      }
    },

    setPort: function (item, protocol, version) {
      var service_id = item.get('service_id');
      var configProperties = this.get('configProperties');
      var config = item.get('http_config');
      var defaultPort = item.get('default_http_port');
      if (protocol === 'https' && item.get('https_config')) {
        config = item.get('https_config');
        if (item.get('default_https_port')) {
          defaultPort = item.get('default_https_port');
        }
      }
      var site = configProperties.findProperty('type', item.get('site'));
      var propertyValue = site && site.properties[config];
      if (!propertyValue) {
        return defaultPort;
      }

      var re = new RegExp(item.get('regex'));

      var portValue = propertyValue.match(re);
      return  portValue[1];
    },

    linkTarget: function () {
      switch (this.get('content.serviceName').toLowerCase()) {
        case "hdfs":
        case "yarn":
        case "mapreduce2":
        case "mapreduce":
        case "hbase":
        case "oozie":
        case "ganglia":
        case "nagios":
        case "hue":
        case "storm":
        case "falcon":
          return "_blank";
          break;
        default:
          return "";
          break;
      }
    }.property('service')

  });
  
});
window.require.register("views/common/rolling_restart_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  /**
   * View content of the rolling restart dialog.
   *
   * Callers provide the context in which this dialog is invoked.
   */
  App.RollingRestartView = Em.View.extend({
    templateName : require('templates/common/rolling_restart_view'),

    /**
     * Component name for components that should be restarted
     * @type {String}
     */
    hostComponentName : null,

    /**
     * Service name for components that should be restarted
     * @type {String}
     */
    serviceName : null,

    /**
     * If service is in Maintenance Mode
     * @type {bool}
     */
    isServiceInMM: false,


    /**
     * Restart only components with <code>staleConfigs</code>
     * @type {bool}
     */
    staleConfigsOnly : false,

    /**
     * We should do rolling restart for components if we run
     * restart for service and service is in Maintenance mode
     * @type {bool}
     */
    skipMaintenance: false,

    /**
     * Count of host components in one batch
     * @type {Number}
     */
    batchSize : -1,

    /**
     * Delay between batches
     * @type {Number}
     */
    interBatchWaitTimeSeconds : -1,

    /**
     * @type {Number}
     */
    tolerateSize : -1,

    /**
     * List of error in batch-request properties
     * @type {Array}
     */
    errors : [],
    /**
     * Set initial values for batch-request properties
     */
    initialize : function() {
      if (this.get('batchSize') == -1 && this.get('interBatchWaitTimeSeconds') == -1 && this.get('tolerateSize') == -1) {
        var restartCount = this.get('restartHostComponents.length');
        var batchSize = 1;
        if (restartCount > 10) {
          batchSize = Math.ceil(restartCount / 10);
        }
        var tolerateCount = batchSize;
        this.set('batchSize', batchSize);
        this.set('tolerateSize', tolerateCount);
        this.set('interBatchWaitTimeSeconds', 120);
      }
    },

    /**
     * Validate batch-request properties
     * List of errors is saved to <code>errors</code>
     */
    validate : function() {
      var displayName = this.get('hostComponentDisplayName');
      var totalCount = this.get('restartHostComponents.length');
      var bs = this.get('batchSize');
      var ts = this.get('tolerateSize');
      var wait = this.get('interBatchWaitTimeSeconds');
      var errors = [];
      if (totalCount < 1) {
        errors.push(Em.I18n.t('rollingrestart.dialog.msg.noRestartHosts').format(displayName));
      } else {
        var bsError = numberUtils.validateInteger(bs, 1, totalCount);
        var tsError = numberUtils.validateInteger(ts, 0, totalCount);
        if (bsError != null) {
          errors.push(Em.I18n.t('rollingrestart.dialog.err.invalid.batchsize').format(bsError));
        }
        if (tsError != null) {
          errors.push(Em.I18n.t('rollingrestart.dialog.err.invalid.toleratesize').format(tsError));
        }
      }
      var waitError = numberUtils.validateInteger(wait, 0, NaN);
      if (waitError != null) {
        errors.push(Em.I18n.t('rollingrestart.dialog.err.invalid.waitTime').format(waitError));
      }
      this.set('errors', errors);
    }.observes('batchSize', 'interBatchWaitTimeSeconds', 'tolerateSize', 'restartHostComponents', 'hostComponentDisplayName'),

    /**
     * Formatted <code>hostComponentName</code>
     * @type {String}
     */
    hostComponentDisplayName : function() {
      return App.format.role(this.get('hostComponentName'));
    }.property('hostComponentName'),

    /**
     * List of all host components
     * @type {Array}
     */
    allHostComponents : function() {
      return App.HostComponent.find().filterProperty('componentName', this.get('hostComponentName'));
    }.property('hostComponentName'),

    /**
     * List of host components without components in out-of-service state
     * @type {Array}
     */
    nonMaintainanceHostComponents : function() {
      return this.get('allHostComponents').filter(function(item) {
        return item.get('passiveState') == 'OFF';
      });
    }.property('allHostComponents', 'allHostComponents.@each.passiveState'),

    /**
     * List of host components with host in Maintenance mode
     * @type {Array}
     */
    componentsWithMaintenanceHost: function() {
      return this.get('allHostComponents').filterProperty('host.passiveState','ON');
    }.property('allHostComponents', 'allHostComponents.@each.passiveState'),
    /**
     * List of host components without components in out-of-service state
     * If <code>staleConfigsOnly</code> is true, components with <code>staleConfigs</code> = false are also filtered
     * @type {Array}
     */
    restartHostComponents : function() {
      var hostComponents = null;
      if (this.get('skipMaintenance')) {
        hostComponents = this.get('allHostComponents').filterProperty('host.passiveState','OFF');
      } else {
        hostComponents = this.get('nonMaintainanceHostComponents');
      }
      if (this.get('staleConfigsOnly')) {
        hostComponents = hostComponents.filterProperty('staleConfigs', true);
      }
      return hostComponents;
    }.property('nonMaintainanceHostComponents', 'staleConfigsOnly'),

    /**
     * @type {String}
     */
    suggestTurnOnMaintenanceMsg : function() {
      if (!this.get('isServiceInMM')) {
        return Em.I18n.t('rollingrestart.dialog.msg.serviceNotInMM').format(this.get('serviceName'));
      } else {
        return null;
      }
    }.property('isServiceInMM'),

    /**
     * @type {String}
     */
    restartMessage : function() {
      return Em.I18n.t('rollingrestart.dialog.msg.restart').format(this.get('hostComponentDisplayName'))
    }.property('hostComponentDisplayName'),

    /**
     * @type {String}
     */
    maintainanceMessage : function() {
      var count = this.get('componentsWithMaintenanceHost.length');
      if (count > 0) {
        var name = this.get('hostComponentDisplayName');
        if (count > 1) {
          return Em.I18n.t('rollingrestart.dialog.msg.maintainance.plural').format(count, name)
        }
        return Em.I18n.t('rollingrestart.dialog.msg.maintainance').format(count, name)
      }
      return null;
    }.property('componentsWithMaintenanceHost', 'hostComponentDisplayName'),

    /**
     * @type {String}
     */
    batchSizeMessage : function() {
      return Em.I18n.t('rollingrestart.dialog.msg.componentsAtATime').format(this.get('hostComponentDisplayName'));
    }.property('hostComponentDisplayName'),

    /**
     * @type {String}
     */
    staleConfigsOnlyMessage : function() {
      return Em.I18n.t('rollingrestart.dialog.msg.staleConfigsOnly').format(this.get('hostComponentDisplayName'));
    }.property('hostComponentDisplayName')
  });
  
});
window.require.register("views/common/sort_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var misc = require('utils/misc');
  var App = require('app');

  /**
   * Wrapper View for all sort components. Layout template and common actions are located inside of it.
   * Logic specific for sort fields
   * located in inner view - <code>fieldView</code>.
   *
   * @type {*}
   */
  var wrapperView = Em.View.extend({
    tagName: 'tr',

    classNames: ['sort-wrapper'],

    willInsertElement:function () {
      if(this.get('parentView.tableFilteringComplete')){
        this.get('parentView').set('filteringComplete', true);
      }
    },

    /**
     * Load sort statuses from local storage
     * Works only after finish filtering in the parent View
     */
    loadSortStatuses: function() {
      var statuses = App.db.getSortingStatuses(this.get('controller.name'));
      if (!this.get('parentView.filteringComplete')) return;
      if (statuses) {
        var childViews = this.get('childViews');
        var self = this;
        statuses.forEach(function(st) {
          if (st.status != 'sorting') {
            var sortOrder = false;
            if(st.status == 'sorting_desc') {
              sortOrder = true;
            }
            self.sort(childViews.findProperty('name', st.name), sortOrder);
            childViews.findProperty('name', st.name).set('status', (sortOrder)?'sorting_desc':'sorting_asc');
            self.get('controller').set('sortingColumn', childViews.findProperty('name', st.name));
          }
          else {
            childViews.findProperty('name', st.name).set('status', st.status);
          }
        });
      }
      this.get('parentView').showProperPage();
    }.observes('parentView.filteringComplete'),

    /**
     * Save sort statuses to local storage
     * Works only after finish filtering in the parent View
     */
    saveSortStatuses: function() {
      if (!this.get('parentView.filteringComplete')) return;
      var statuses = [];
      this.get('childViews').forEach(function(childView) {
        statuses.push({
          name: childView.get('name'),
          status: childView.get('status')
        });
      });
      App.db.setSortingStatuses(this.get('controller.name'), statuses);
    }.observes('childViews.@each.status'),

    /**
     * sort content by property
     * @param property
     * @param order true - DESC, false - ASC
     */
    sort: function(property, order, returnSorted){
      returnSorted = returnSorted ? true : false;
      var content = this.get('content').toArray();
      var sortFunc = this.getSortFunc(property, order);
      this.resetSort();
      content.sort(sortFunc);
      if(returnSorted){
        return content;
      }else{
        this.set('content', content);
      }
    },

    isSorting: false,

    onContentChange: function () {
      if (!this.get('isSorting') && this.get('content.length')) {
        this.get('childViews').forEach(function (view) {
          if (view.status !== 'sorting') {
            var status = view.get('status');
            this.set('isSorting', true);
            this.sort(view, status == 'sorting_desc');
            this.set('isSorting', false);
            view.set('status', status);
          }
        }, this);
      }
    }.observes('content.length'),

    /**
     * reset all sorts fields
     */
    resetSort: function(){
      this.get('childViews').setEach('status', 'sorting');
    },
    /**
     * determines sort function depending on the type of sort field
     * @param property
     * @param order
     * @return {*}
     */
    getSortFunc: function(property, order){
      var func;
      switch (property.get('type')){
        case 'ip':
          func = function (a, b) {
            var a = misc.ipToInt(a.get(property.get('name')));
            var b = misc.ipToInt(b.get(property.get('name')));
            if (order) {
              return b - a;
            } else {
              return a - b;
            }
          };
          break;
        case 'number':
          func = function (a, b) {
            var a = parseFloat(a.get(property.get('name')));
            var b = parseFloat(b.get(property.get('name')));
            if (order) {
              return b - a;
            } else {
              return a - b;
            }
          };
          break;
        default:
          func = function(a,b){
            if(order){
              if (a.get(property.get('name')) > b.get(property.get('name')))
                return -1;
              if (a.get(property.get('name')) < b.get(property.get('name')))
                return 1;
              return 0;
            } else {
              if (a.get(property.get('name')) < b.get(property.get('name')))
                return -1;
              if (a.get(property.get('name')) > b.get(property.get('name')))
                return 1;
              return 0;
            }
          }
      }
      return func;
    }
  });
  /**
   * particular view that contain sort field properties:
   * name - name of property in content table
   * type(optional) - specific type to sort
   * displayName - label to display
   * @type {*}
   */
  var fieldView = Em.View.extend({
    template:Em.Handlebars.compile('<span {{bindAttr class="view.status :column-name"}}>{{view.displayName}}</span>'),
    classNameBindings: ['viewNameClass'],
    tagName: 'th',
    name: null,
    displayName: null,
    status: 'sorting',
    viewNameClass: function () {
      return 'sort-view-' + this.get('column');
    }.property(),
    type: null,
    column: 0,
    /**
     * callback that run sorting and define order of sorting
     * @param event
     */
    click: function(event){
      if(this.get('status') === 'sorting_desc'){
        this.get('parentView').sort(this, false);
        this.set('status', 'sorting_asc');
      }
      else {
        this.get('parentView').sort(this, true);
        this.set('status', 'sorting_desc');
      }
      this.get('controller').set('sortingColumn', this);
    }
  });

  /**
   * Result object, which will be accessible outside
   * @type {Object}
   */
  module.exports = {
    wrapperView: wrapperView,
    fieldView: fieldView
  };
});
window.require.register("views/common/table_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');
  var sort = require('views/common/sort_view');

  App.TableView = Em.View.extend(App.UserPref, {

    /**
     * Shows if all data is loaded and filtered
     * @type {Boolean}
     */
    filteringComplete: false,

    /**
     * intermediary for filteringComplete
     * @type {Boolean}
     */
    tableFilteringComplete: false,

    /**
     * Loaded from local storage startIndex value
     * @type {Number}
     */
    startIndexOnLoad: null,

    /**
     * Loaded from server persist value
     * @type {Number}
     */
    displayLengthOnLoad: null,

    /**
     * The number of rows to show on every page
     * The value should be a number converted into string type in order to support select element API
     * Example: "10", "25"
     * @type {String}
     */
    displayLength: null,

    /**
     * default value of display length
     * The value should be a number converted into string type in order to support select element API
     * Example: "10", "25"
     */
    defaultDisplayLength: "10",

    /**
     * Do filtering, using saved in the local storage filter conditions
     */
    willInsertElement:function () {
      var self = this;
      var name = this.get('controller.name');

      this.set('startIndexOnLoad', App.db.getStartIndex(name));
      if (!this.get('displayLength')) {
        if (App.db.getDisplayLength(name)) {
          this.set('displayLength', App.db.getDisplayLength(name));
        } else {
          self.dataLoading().done(function (initValue) {
            self.set('displayLength', initValue);
          });
        }
      }

      var filterConditions = App.db.getFilterConditions(name);
      if (filterConditions) {
        this.set('filterConditions', filterConditions);

        var childViews = this.get('childViews');

        filterConditions.forEach(function(condition) {
          var view = childViews.findProperty('column', condition.iColumn);
          if (view) {
            view.set('value', condition.value);
            Em.run.next(function() {
              view.showClearFilter();
            });
          }
        });
      } else {
        this.clearFilters();
      }
      self.set('tableFilteringComplete', true);
    },

    /**
     * Load user preference value on hosts page from server
     */
    dataLoading: function() {
      var dfd = $.Deferred();
      var self = this;
      this.getUserPref(this.displayLengthKey()).done(function () {
        var curLength = self.get('displayLengthOnLoad');
        self.set('displayLengthOnLoad', null);
        dfd.resolve(curLength);
      });
      return dfd.promise();
    },

    /**
     * Persist-key of current table displayLength property
     * @param {String} loginName current user login name
     * @returns {String}
     */
    displayLengthKey: function (loginName) {
      if (App.get('testMode')) {
        return 'pagination_displayLength';
      }
      loginName = loginName ? loginName : App.router.get('loginName');
      return this.get('controller.name') + '-pagination-displayLength-' + loginName;
    },

    /**
     * Set received from server value to <code>displayLengthOnLoad</code>
     * @param {Number} response
     * @param {Object} request
     * @param {Object} data
     * @returns {*}
     */
    getUserPrefSuccessCallback: function (response, request, data) {
      console.log('Got DisplayLength value from server with key ' + data.key + '. Value is: ' + response);
      this.set('displayLengthOnLoad', response);
      return response;
    },

    /**
     * Set default value to <code>displayLengthOnLoad</code> (and send it on server) if value wasn't found on server
     * @returns {Number}
     */
    getUserPrefErrorCallback: function () {
      // this user is first time login
      console.log('Persist did NOT find the key');
      var displayLengthDefault = this.get('defaultDisplayLength');
      this.set('displayLengthOnLoad', displayLengthDefault);
      if (App.get('isAdmin')) {
        this.postUserPref(this.displayLengthKey(), displayLengthDefault);
      }
      return displayLengthDefault;
    },

    /**
     * Do pagination after filtering and sorting
     * Don't call this method! It's already used where it's need
     */
    showProperPage: function() {
      var self = this;
      Em.run.next(function() {
        Em.run.next(function() {
          if(self.get('startIndexOnLoad')) {
            self.set('startIndex', self.get('startIndexOnLoad'));
          }
        });
      });
    },

    /**
     * Return pagination information displayed on the page
     * @type {String}
     */
    paginationInfo: function () {
      return this.t('tableView.filters.paginationInfo').format(this.get('startIndex'), this.get('endIndex'), this.get('filteredContent.length'));
    }.property('displayLength', 'filteredContent.length', 'startIndex', 'endIndex'),

    paginationLeft: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-arrow-left"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if (this.get("parentView.startIndex") > 1) {
          return "paginate_previous";
        }
        return "paginate_disabled_previous";
      }.property("parentView.startIndex", 'filteredContent.length'),

      click: function () {
        if (this.get('class') === "paginate_previous") {
          this.get('parentView').previousPage();
        }
      }
    }),

    paginationRight: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-arrow-right"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if ((this.get("parentView.endIndex")) < this.get("parentView.filteredContent.length")) {
          return "paginate_next";
        }
        return "paginate_disabled_next";
      }.property("parentView.endIndex", 'filteredContent.length'),

      click: function () {
        if (this.get('class') === "paginate_next") {
          this.get('parentView').nextPage();
        }
      }
    }),

    paginationFirst: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-step-backward"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if ((this.get("parentView.endIndex")) > parseInt(this.get("parentView.displayLength"))) {
          return "paginate_previous";
        }
        return "paginate_disabled_previous";
      }.property("parentView.endIndex", 'filteredContent.length'),

      click: function () {
        if (this.get('class') === "paginate_previous") {
          this.get('parentView').firstPage();
        }
      }
    }),

    paginationLast: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-step-forward"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if (this.get("parentView.endIndex") !== this.get("parentView.filteredContent.length")) {
          return "paginate_next";
        }
        return "paginate_disabled_next";
      }.property("parentView.endIndex", 'filteredContent.length'),

      click: function () {
        if (this.get('class') === "paginate_next") {
          this.get('parentView').lastPage();
        }
      }
    }),

    /**
     * Select View with list of "rows-per-page" options
     * @type {Ember.View}
     */
    rowsPerPageSelectView: Em.Select.extend({
      content: ['10', '25', '50', '100'],
      change: function () {
        this.get('parentView').saveDisplayLength();
      }
    }),

    /**
     * Start index for displayed content on the page
     */
    startIndex: 1,

    /**
     * Calculate end index for displayed content on the page
     */
    endIndex: function () {
      return Math.min(this.get('filteredContent.length'), this.get('startIndex') + parseInt(this.get('displayLength')) - 1);
    }.property('startIndex', 'displayLength', 'filteredContent.length'),

    /**
     * Onclick handler for previous page button on the page
     */
    previousPage: function () {
      var result = this.get('startIndex') - parseInt(this.get('displayLength'));
      this.set('startIndex', (result < 2) ? 1 : result);
    },

    /**
     * Onclick handler for next page button on the page
     */
    nextPage: function () {
      var result = this.get('startIndex') + parseInt(this.get('displayLength'));
      if (result - 1 < this.get('filteredContent.length')) {
        this.set('startIndex', result);
      }
    },
    /**
     * Onclick handler for first page button on the page
     */
    firstPage: function () {
      this.set('startIndex', 1);
    },
    /**
     * Onclick handler for last page button on the page
     */
    lastPage: function () {
      var pagesCount = this.get('filteredContent.length') / parseInt(this.get('displayLength'));
      var startIndex = (this.get('filteredContent.length') % parseInt(this.get('displayLength')) === 0) ?
        (pagesCount - 1) * parseInt(this.get('displayLength')) :
        Math.floor(pagesCount) * parseInt(this.get('displayLength'));
      this.set('startIndex', ++startIndex);
    },

    /**
     * Calculates default value for startIndex property after applying filter or changing displayLength
     */
    updatePaging: function (controller, property) {
      var displayLength = this.get('displayLength');
      var filteredContentLength = this.get('filteredContent.length');
      if (property == 'displayLength') {
        this.set('startIndex', Math.min(1, filteredContentLength));
      } else if (!filteredContentLength) {
        this.set('startIndex', 0);
      } else if (this.get('startIndex') > filteredContentLength) {
        this.set('startIndex', Math.floor((filteredContentLength - 1) / displayLength) * displayLength + 1);
      } else if (!this.get('startIndex')) {
        this.set('startIndex', 1);
      }
    }.observes('displayLength', 'filteredContent.length'),

    /**
     * Apply each filter to each row
     *
     * @param {Number} iColumn number of column by which filter
     * @param {Object} value
     * @param {String} type
     */
    updateFilter: function (iColumn, value, type) {
      var filterCondition = this.get('filterConditions').findProperty('iColumn', iColumn);
      if (filterCondition) {
        filterCondition.value = value;
      }
      else {
        filterCondition = {
          iColumn: iColumn,
          value: value,
          type: type
        };
        this.get('filterConditions').push(filterCondition);
      }
      this.saveFilterConditions();
      this.filtersUsedCalc();
      this.filter();
    },

    saveFilterConditions: function() {
      App.db.setFilterConditions(this.get('controller.name'), this.get('filterConditions'));
    },

    saveDisplayLength: function() {
      var self = this;
      Em.run.next(function() {
        App.db.setDisplayLength(self.get('controller.name'), self.get('displayLength'));
        if (!App.testMode) {
          if (App.get('isAdmin')) {
            self.postUserPref(self.displayLengthKey(), self.get('displayLength'));
          }
        }
      });
    },

    saveStartIndex: function() {
      if (this.get('filteringComplete')) {
        App.db.setStartIndex(this.get('controller.name'), this.get('startIndex'));
      }
    }.observes('startIndex'),

    clearFilterCondition: function() {
      App.db.setFilterConditions(this.get('controller.name'), null);
    },

    clearStartIndex: function() {
      App.db.setStartIndex(this.get('controller.name'), null);
    },

    /**
     * Contain filter conditions for each column
     * @type {Array}
     */
    filterConditions: [],

    /**
     * Contains content after implementing filters
     * @type {Array}
     */
    filteredContent: [],

    /**
     * Determine if <code>filteredContent</code> is empty or not
     * @type {Boolean}
     */
    hasFilteredItems: function() {
      return !!this.get('filteredContent.length');
    }.property('filteredContent.length'),

    /**
     * Contains content to show on the current page of data page view
     * @type {Array}
     */
    pageContent: function () {
      return this.get('filteredContent').slice(this.get('startIndex') - 1, this.get('endIndex'));
    }.property('filteredContent.length', 'startIndex', 'endIndex'),

    /**
     * Filter table by filterConditions
     */
    filter: function () {
      var content = this.get('content');
      var filterConditions = this.get('filterConditions').filterProperty('value');
      var result;
      var assoc = this.get('colPropAssoc');
      if (filterConditions.length) {
        result = content.filter(function (item) {
          var match = true;
          filterConditions.forEach(function (condition) {
            var filterFunc = filters.getFilterByType(condition.type, false);
            if (match) {
              match = filterFunc(item.get(assoc[condition.iColumn]), condition.value);
            }
          });
          return match;
        });
        this.set('filteredContent', result);
      } else {
        this.set('filteredContent', content.toArray());
      }
    }.observes('content.length'),

    /**
     * Does any filter is used on the page
     * @type {Boolean}
     */
    filtersUsed: false,

    /**
     * Determine if some filters are used on the page
     * Set <code>filtersUsed</code> value
     */
    filtersUsedCalc: function() {
      var filterConditions = this.get('filterConditions');
      if (!filterConditions.length) {
        this.set('filtersUsed', false);
        return;
      }
      var filtersUsed = false;
      filterConditions.forEach(function(filterCondition) {
        if (filterCondition.value.toString() !== '') {
          filtersUsed = true;
        }
      });
      this.set('filtersUsed', filtersUsed);
    },

    /**
     * Run <code>clearFilter</code> in the each child filterView
     */
    clearFilters: function() {
      this.set('filterConditions', []);
      this.get('_childViews').forEach(function(childView) {
        if (childView['clearFilter']) {
          childView.clearFilter();
        }
      });
    }

  });
});
window.require.register("views/common/time_range", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * use: {{view App.TimeRangeWidget controllerBinding="App.router.mainChartsController"}}
   * set controller.preset field with preset value
   * widget assign itself to controller like presetWidget (controller.get('presetWidget'))
   * @type {*}
   */
  App.TimeRangeWidget = Em.View.extend({
    classNames:['time-range-widget'],
    templateName:require('templates/common/time_range'),
    dateFrom: null,
    dateTo: null,

    /**
     * presets
     */
    presets:[
      Em.Object.create({ label:Em.I18n.t('timeRange.presets.1hour'), value:'1h', period: 3600000, step: 300000}),
      Em.Object.create({ label:Em.I18n.t('timeRange.presets.12hour'), value:'12h', period: 43200000, step: 3600000}),
      Em.Object.create({ label:Em.I18n.t('timeRange.presets.1day'), value:'1d', period: 86400000, step: 3600000}),
      Em.Object.create({ label:Em.I18n.t('timeRange.presets.1week'), value:'1wk', period: 604800000, step: 86400000}),
      Em.Object.create({ label:Em.I18n.t('timeRange.presets.1month'), value:'1mo', period: 2592000000, step: 86400000}),
      Em.Object.create({ label:Em.I18n.t('timeRange.presets.1year'), value:'1yr', period: 31536000000, step: 2592000000})
    ],
    /**
     * chosen preset value
     */
    chosenPreset: null,

    /**
     * return array of chosen presets
     */
    chosenPresets:function () {
      return this.get('chosenPreset') ? [this.get('chosenPreset')] : this.get('defaultPresets');
    }.property('chosenPreset'),

    /**
     * preset item view
     */
    presetView:Em.View.extend({
      tagName:'li',
      classNameBindings:['disabled'],
      disabled:function () {
        return this.get('isActive') ? "disabled" : false;
      }.property('isActive'),
      isActive:function () {
        return this.get('preset.value') == this.get('widget.chosenPreset.value');
      }.property('widget.chosenPreset'),
      template:Em.Handlebars.compile('<a {{action activate view.preset target="view.widget" href="true" }}>{{unbound view.preset.label}}</a>')
    }),

    /**
     * return default selected presets (currently - all)
     */
    defaultPresets:function () {
      var values = [];
      $.each(this.get('presets'), function () {
        if (this.value) {
          values.push(this.value);
        }
      });
      return values;
    }.property(),

    bindToController:function () {
      var thisW = this;
      var controller = this.get('controller');
      controller.set('presetWidget', thisW);
    },

    /**
     * assign this widget to controller, prepare items by presetsConfig
     */
    init:function () {
      this._super();
      this.bindToController();
    },

    /**
     * write active preset to widget
     * @param event
     */
    activate:function (event) {
      if (event.context == this.get('chosenPreset')) {
        this.set('chosenPreset', null);
      } else {
        this.set('chosenPreset', event.context);
      }
    },

    dateFromView: Ember.TextField.extend({
      elementId: 'timeRangeFrom',
      classNames: 'timeRangeFrom',
      attributeBindings:['readonly'],
      readonly: true,
      didInsertElement: function() {
        var self = this;
        this.$().datetimepicker({
          dateFormat: 'dd/mm/yy',
          timeFormat: 'hh:mm',
          maxDate: new Date(),
          onClose:function (dateText, inst) {
            var endDateTextBox = $('#timeRangeTo');
            if (endDateTextBox.val() != '') {
              var testStartDate = new Date(dateText);
              var testEndDate = new Date(endDateTextBox.val());
              if (testStartDate > testEndDate)
                endDateTextBox.val(dateText);
            } else {
              endDateTextBox.val(dateText);
            }
            self.set('dateFrom', dateText);
          },
          onSelect:function (selectedDateTime) {
            var start = $(this).datetimepicker('getDate');
            $('#timeRangeTo').datetimepicker('option', 'minDate', new Date(start.getTime()));
          }
        });
        self.set('dateFrom', this.get('value'));
      }
    }),

    dateToView: Ember.TextField.extend({
      elementId: 'timeRangeTo',
      classNames: 'timeRangeTo',
      attributeBindings:['readonly'],
      readonly: true,
      didInsertElement: function() {
        var self = this;
        this.$().datetimepicker({
          dateFormat: 'dd/mm/yy',
          timeFormat: 'hh:mm',
          maxDate: new Date(),
          onClose:function (dateText, inst) {
            var startDateTextBox = $('#timeRangeFrom');
            if (startDateTextBox.val() != '') {
              var testStartDate = new Date(startDateTextBox.val());
              var testEndDate = new Date(dateText);
              if (testStartDate > testEndDate)
                startDateTextBox.val(dateText);
            } else {
              startDateTextBox.val(dateText);
            }
            self.set('dateTo', dateText);
          },
          onSelect:function (selectedDateTime) {
            var end = $(this).datetimepicker('getDate');
            $('#timeRangeFrom').datetimepicker('option', 'maxDate', new Date(end.getTime()));
          }
        });
        self.set('dateTo', this.get('value'));
      }
    }),

    sliderOptions: Ember.Object.extend({
      end: null,
      period: null,
      start: function() {
        return this.get('end') - this.get('period');
      }.property('end', 'period')
    }),
    nowLabel: null,
    rangeLabel: null,
    buildSlider: function() {
      if (this.get('chosenPreset')) {
        var sliderOptions = this.sliderOptions.create({
          end: function() {
            var endDate = new Date();
            return endDate.getTime();
          }.property(),
          period: this.get('chosenPreset.period'),
          step: this.get('chosenPreset.step'),
          countTimeAgo: function(stepValue) {
            var msAgo = this.get('end') - stepValue;
            return msAgo.toDaysHoursMinutes();
          }
        });
        this.set('nowLabel', 'Now');
        this.set('rangeLabel', new Date(sliderOptions.get('start')));

        var self = this;
        $('#slider').slider({
          range: "max",
          min: sliderOptions.get('start'),
          max: sliderOptions.get('end'),
          value: sliderOptions.get('start'),
          step: sliderOptions.get('step'),
          stop: function(event, ui) {
            self.set('rangeLabel', new Date(ui.value));
  //          self.set('rangeLabel', sliderOptions.countTimeAgo(ui.value).h);
          },
          slide: function(event, ui){
            self.set('rangeLabel', new Date(ui.value));
  //          self.set('rangeLabel', sliderOptions.countTimeAgo(ui.value).h);
          }
        });
      } else {
        console.log(this.get('chosenPreset'));
        $("#slider").slider("destroy");
      }
    }.observes('chosenPreset')
  });
});
window.require.register("views/experimental", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.ExperimentalView = Em.View.extend({
    templateName: require('templates/experimental'),
    supports: function () {
      var supports = [];
      for ( var sup in App.supports) {
        supports.push(Ember.Object.create({
          name: sup,
          selected: App.supports[sup]
        }));
      }
      return supports;
    }.property('App.supports'),

    doSave: function () {
      var supports = this.get('supports');
      supports.forEach(function(s){
        var propName = 'App.supports.' + s.get('name');
        var propValue = s.get('selected');
        console.log(">>>>>>>> " + propName + " = "+ propValue);
        Ember.set(propName, propValue);
      });
      App.router.transitionTo('root.index');
    },

    doCancel: function () {
      App.router.transitionTo('root.index');
    }
  });
});
window.require.register("views/installer", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.InstallerView = Em.View.extend({

    templateName: require('templates/installer'),

    isStep0Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',0).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep1Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',1).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',2).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',3).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep4Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',4).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep5Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',5).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep6Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',6).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep7Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',7).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep8Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',8).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep9Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',9).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep10Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',10).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable()

  });
});
window.require.register("views/loading", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.LoadingView = Em.View.extend({
      tagName: 'h2',
      template: Ember.Handlebars.compile('{{t app.loadingPlaceholder}}')
  });
});
window.require.register("views/login", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.LoginView = Em.View.extend({

    templateName: require('templates/login'),
    loginTextField: Em.TextField.extend({
      didInsertElement: function(){
        this._super();
        this.$().focus();
      }
    }),
    passTextField : Em.TextField.extend({
      insertNewline: function(){
        this.get("controller").submit();
      }
    })
  });
});
window.require.register("views/main", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.MainView = Em.View.extend({
    templateName:require('templates/main')
  });
});
window.require.register("views/main/admin", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminView = Em.View.extend({
    templateName: require('templates/main/admin'),
    selectedBinding: 'controller.category',
    categories: function() {
      var items = [{
        name: 'user',
        url: 'admin.index',
        label: Em.I18n.t('common.users')
      }];
      if (App.get('isHadoop2Stack') && App.supports.highAvailability) {
        items.push({
          name: 'highAvailability',
          url: 'adminHighAvailability',
          label: Em.I18n.t('admin.highAvailability')
        });
      }
      if (App.supports.secureCluster) {
        items.push({
          name: 'security',
          url: 'adminSecurity.index',
          label: Em.I18n.t('common.security')
        });
      }
      items.push({
        name: 'cluster',
        url: 'adminCluster',
        label: Em.I18n.t('common.cluster')
      });
      items.push({
        name: 'misc',
        url: 'adminMisc',
        label: Em.I18n.t('common.misc')
      });
      if (this.get('controller.isAccessAvailable')) {
        items.push({
          name: 'access',
          url: 'adminAccess',
          label: Em.I18n.t('common.access')
        });
      }
      return items;
    }.property(''),

    NavItemView: Ember.View.extend({
      tagName: 'li',
      classNameBindings: 'isActive:active'.w(),
      isActive: function () {
        return this.get('item') === this.get('parentView.selected');
      }.property('item', 'parentView.selected')
    })
  });

  
});
window.require.register("views/main/admin/access_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAccessView = Em.View.extend({
    templateName: require('templates/main/admin/access')
  });
  
});
window.require.register("views/main/admin/advanced", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAdvancedView = Em.View.extend({
    templateName: require('templates/main/admin/advanced'),
    params: {keepData: 1}
  });
});
window.require.register("views/main/admin/advanced/password", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAdvancedPasswordView = App.Form.extend({
    fieldsOptions:[
      { name:"password", displayName:Em.I18n.t('common.password'), displayType:"password", disableRequiredOnExistent:true },
      { name:"passwordRetype", displayName:Em.I18n.t('form.passwordRetype'), displayType:"passwordRetype", disableRequiredOnExistent:true }
    ],
    fields: []
  });
});
window.require.register("views/main/admin/audit", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  require('views/common/grid');

  App.MainAdminAuditView = App.Grid.extend({
    _collection: {className: App.ServiceAudit},
    prepareColumns:function () {
      this._super();

      this.addColumn({
        name:"date",
        label:Em.I18n.t("admin.audit.grid.date")
      });
      this.addColumn({
        name:"service.label",
        label:Em.I18n.t("admin.audit.grid.service")
      });
      this.addColumn({
        name:"operationName",
        label:Em.I18n.t("admin.audit.grid.operationName")
      });
      this.addColumn({
        name:"user.userName",
        label:Em.I18n.t("admin.audit.grid.performedBy")
      });
    }
  });
});
window.require.register("views/main/admin/authentication", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminAuthenticationView = Em.View.extend({
    templateName:require('templates/main/admin/authentication'),
    form:App.AuthenticationForm.create({}),
    pushAuthenticationToForm:function () {
      var auth = App.Authentication.find(1);
      App.router.set('mainAdminAuthenticationController.content', auth);
      this.form.set('object', auth);
    },
    didInsertElement: function (){
      this._super();
      this.pushAuthenticationToForm();
    },

    ldapChecked:function () {
      return this.get('form.field.method.value');
    }.property('form.field.method.value'),

    useCredentials:function () {
      return this.get('form.field.bindMethod.value');
    }.property('form.field.bindMethod.value')
  });
});
window.require.register("views/main/admin/cluster", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var stringUtils = require('utils/string_utils');

  App.MainAdminClusterView = Em.View.extend({
    templateName: require('templates/main/admin/cluster'),

    isUpgradeAvailable: function(){
      return stringUtils.compareVersions(this.get('controller.upgradeVersion').replace(/HDP(Local)?-/, ''), App.get('currentStackVersionNumber')) === 1;
    }.property('controller.upgradeVersion', 'App.currentStackVersion')
  });
});
window.require.register("views/main/admin/highAvailability/progress_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityProgressPageView = Em.View.extend({

    didInsertElement: function () {
      this.get('controller').loadStep();
    },

    headerTitle: function () {
      var currentStep = App.router.get('highAvailabilityWizardController.currentStep');
      if(currentStep == 1) {
        return  Em.I18n.t('admin.highAvailability.wizard.rollback.header.title');
      }else {
        return  Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.header.title');
      }
    }.property(),

    submitButtonText: Em.I18n.t('common.next'),

    noticeCompleted: Em.I18n.t('admin.highAvailability.wizard.progressPage.notice.completed'),

    noticeFailed: Em.I18n.t('admin.highAvailability.wizard.progressPage.notice.failed'),

    noticeInProgress: function () {
      var currentStep = App.router.get('highAvailabilityWizardController.currentStep');
      if(currentStep == 1) {
        return  Em.I18n.t('admin.highAvailability.rollback.notice.inProgress');
      }else {
        return  Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.notice.inProgress');
      }
    }.property(),

    notice: Em.I18n.t('admin.highAvailability.wizard.progressPage.notice.inProgress'),

    noticeClass: 'alert alert-info',

    onStatusChange: function () {
      var status = this.get('controller.status');
      if (status === 'COMPLETED') {
        this.set('notice', this.get('noticeCompleted'));
        this.set('noticeClass', 'alert alert-success');
      } else if (status === 'FAILED') {
        this.set('notice', this.get('noticeFailed'));
        this.set('noticeClass', 'alert alert-error');
      } else {
        this.set('notice', this.get('noticeInProgress'));
        this.set('noticeClass', 'alert alert-info');
      }
    }.observes('controller.status'),

    taskView: Em.View.extend({
      icon: '',
      iconColor: '',
      linkClass: '',

      didInsertElement: function () {
        this.onStatus();
      },

      barWidth: function () {
        return 'width: ' + this.get('content.progress') + '%;';
      }.property('content.progress'),

      onStatus: function () {
        var linkClass = !!this.get('content.hosts.length') ? 'active-link' : 'active-text';
        this.set('linkClass', linkClass);
        if (this.get('content.status') === 'IN_PROGRESS') {
          this.set('icon', 'icon-cog');
          this.set('iconColor', 'text-info');
        } else if (this.get('content.status') === 'FAILED') {
          this.set('icon', 'icon-exclamation-sign');
          this.set('iconColor', 'text-error');
        } else if (this.get('content.status') === 'COMPLETED') {
          this.set('icon', 'icon-ok');
          this.set('iconColor', 'text-success');
        } else {
          this.set('icon', 'icon-cog');
          this.set('iconColor', '');
          this.set('linkClass', 'not-active-link');
        }
      }.observes('content.status', 'content.hosts.length'),

      showProgressBar: function () {
        return this.get('content.status') === "IN_PROGRESS";
      }.property('content.status'),

      /**
       * open popup with list of hosts, that associated to service
       * @param event
       */
      hostsLogPopup: function(event){
        if(!!this.get('content.hosts.length')){
          var serviceName = event.contexts[0].title;
          var controller = this.get("controller");
          App.router.get('applicationController').dataLoading().done(function (initValue) {
            var popupView = App.HostPopup.initPopup(serviceName, controller);
            popupView.set ('isNotShowBgChecked', !initValue);
          })
        }
      }
    })
  });
  
});
window.require.register("views/main/admin/highAvailability/rollback_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityRollbackView = App.HighAvailabilityProgressPageView.extend({
    templateName: require('templates/main/admin/highAvailability/rollback'),

    notice: Em.I18n.t('admin.highAvailability.rollback.notice.inProgress'),
    submitButtonText: Em.I18n.t('common.done')


  });
  
});
window.require.register("views/main/admin/highAvailability/step1_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep1View = Em.View.extend({

    templateName: require('templates/main/admin/highAvailability/step1'),

    didInsertElement: function() {
      App.popover($("div.controls input[type=text]"), {'placement': 'right', 'trigger': 'hover', 'title': this.t('admin.highAvailability.wizard.step1.nameserviceid.tooltip.title'), 'content': this.t('admin.highAvailability.wizard.step1.nameserviceid.tooltip.content')});
    },

    showInputError: function () {
      return !this.get('controller.isNameServiceIdValid') && this.get('controller.content.nameServiceId').length != 0 ;
    }.property('controller.isNameServiceIdValid', 'controller.nameServiceId')

  });
  
});
window.require.register("views/main/admin/highAvailability/step2_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  require('views/wizard/step5_view');

  App.HighAvailabilityWizardStep2View = App.WizardStep5View.extend({

    templateName: require('templates/main/admin/highAvailability/step2')

  });
  
});
window.require.register("views/main/admin/highAvailability/step3_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep3View = Em.View.extend({

    templateName: require('templates/main/admin/highAvailability/step3'),
    didInsertElement: function () {
      this.get('controller').loadStep();
    },
    curNameNode: function () {
      var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
      return nN.hostName;
    }.property('controller.content.masterComponentHosts'),
    addNameNode: function () {
      var addNN = this.get('controller.content.masterComponentHosts').findProperty('isAddNameNode', true);
      return addNN.hostName;
    }.property('controller.content.masterComponentHosts'),
    secondaryNameNode: function () {
      var sn = this.get('controller.content.masterComponentHosts').findProperty('component', "SECONDARY_NAMENODE");
      return sn.hostName;
    }.property('controller.content.masterComponentHosts'),
    journalNodes: function () {
      return this.get('controller.content.masterComponentHosts').filterProperty('component', "JOURNALNODE");
    }.property('controller.content.masterComponentHosts')

  });
  
});
window.require.register("views/main/admin/highAvailability/step4_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep4View = Em.View.extend({

    templateName: require('templates/main/admin/highAvailability/step4'),

    didInsertElement: function() {
      this.get('controller').pullCheckPointStatus();
    },

    step4BodyText: function () {
      var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
      return Em.I18n.t('admin.highAvailability.wizard.step4.body').format(this.get('controller.content.hdfsUser'), nN.hostName);
    }.property('controller.content.masterComponentHosts'),

    nnCheckPointText: function () {
      var curStatus = this.get('controller.isNextEnabled');
      if(curStatus){
        return Em.I18n.t('admin.highAvailability.wizard.step4.ckCreated');
      }else{
        return Em.I18n.t('admin.highAvailability.wizard.step4.ckNotCreated');
      }
    }.property('controller.isNextEnabled')

  });
  
});
window.require.register("views/main/admin/highAvailability/step5_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep5View = App.HighAvailabilityProgressPageView.extend({

    templateName: require('templates/main/admin/highAvailability/step5')
  });
  
});
window.require.register("views/main/admin/highAvailability/step6_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep6View = Em.View.extend({

    templateName: require('templates/main/admin/highAvailability/step6'),

    didInsertElement: function() {
      this.get('controller').pullCheckPointStatus();
    },

    step6BodyText: function () {
      var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
      return Em.I18n.t('admin.highAvailability.wizard.step6.body').format(this.get('controller.content.hdfsUser'), nN.hostName);
    }.property('controller.content.masterComponentHosts'),

    jnCheckPointText: function () {
      var curStatus = this.get('controller.isNextEnabled');
      if (curStatus) {
        return Em.I18n.t('admin.highAvailability.wizard.step6.jsInit');
      } else {
        return Em.I18n.t('admin.highAvailability.wizard.step6.jsNoInit');
      }
    }.property('controller.isNextEnabled')

  });
  
});
window.require.register("views/main/admin/highAvailability/step7_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep7View = App.HighAvailabilityProgressPageView.extend({

    templateName: require('templates/main/admin/highAvailability/step7')
  });
  
});
window.require.register("views/main/admin/highAvailability/step8_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep8View = Em.View.extend({

    templateName: require('templates/main/admin/highAvailability/step8'),

    step8BodyText: function () {
      var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
      var addNN = this.get('controller.content.masterComponentHosts').findProperty('isAddNameNode', true);
      return Em.I18n.t('admin.highAvailability.wizard.step8.body').format(this.get('controller.content.hdfsUser'), nN.hostName, addNN.hostName);
    }.property('controller.content.masterComponentHosts')

  });
  
});
window.require.register("views/main/admin/highAvailability/step9_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardStep9View = App.HighAvailabilityProgressPageView.extend({

    templateName: require('templates/main/admin/highAvailability/step9'),

    submitButtonText: Em.I18n.t('common.done'),

    noticeCompleted: Em.I18n.t('admin.highAvailability.wizard.step9.notice.completed')
  });
  
});
window.require.register("views/main/admin/highAvailability/wizard_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.HighAvailabilityWizardView = Em.View.extend({

    didInsertElement: function() {
      var currentStep = this.get('controller.currentStep');
      if (currentStep > 4) {
        this.get('controller').setLowerStepsDisable(currentStep);
      }
    },

    templateName: require('templates/main/admin/highAvailability/wizard'),

    isStep1Disabled: function () {
      return this.isStepDisabled(1);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.isStepDisabled(2);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.isStepDisabled(3);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep4Disabled: function () {
      return this.isStepDisabled(4);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep5Disabled: function () {
      return this.isStepDisabled(5);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep6Disabled: function () {
      return this.isStepDisabled(6);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep7Disabled: function () {
      return this.isStepDisabled(7);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep8Disabled: function () {
      return this.isStepDisabled(8);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep9Disabled: function () {
      return this.isStepDisabled(9);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStepDisabled: function (index) {
      return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
    }

  });
  
});
window.require.register("views/main/admin/highAvailability_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminHighAvailabilityView = Em.View.extend({
    templateName: require('templates/main/admin/highAvailability'),

    didInsertElement: function () {
      this.get('controller').setSecurityStatus();
    }
  });
  
});
window.require.register("views/main/admin/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  // This logic is substituted by MainAdminView for now.
  App.MainAdminMenuView = Em.CollectionView.extend({
    //contentBinding: 'controller',
    /*content: [
      {
        route:'user',
        label:'Users'
      },
      {
        route:'security',
        label:'Security'
      },
      {
        route:'cluster',
        label:'Cluster'
      }
      /*,
      {
        route:'authentication',
        label:'Authentication'
      },

  {
        route: 'user',
        label: 'Users'

      },
      {
        route: 'security',
        label: 'Security'
      }/*,
       {
       route:'authentication',
       label:'Authentication'
       },

       {
       route:'audit',
       label:'Audit'
       }*/
      /*,
       {
       route:'advanced',
       label:'Advanced'
       }

    ],
    tagName: "ul",
    classNames: ["nav", "nav-list"],

    init: function () {
      this._super();
      this.activateView(); // default selected menu
    },

    activateView: function () {
      var route = App.get('router.mainAdminController.category');
      $.each(this._childViews, function () {
        this.set('active', (this.get('content.route') == route ? "active" : ""));
      });
    }.observes('App.router.mainAdminController.category'),

    itemViewClass:Em.View.extend({
      classNameBindings:["active"],
      active:"",
      template:Ember.Handlebars.compile('<a class="text-center" {{action adminNavigate view.content.route }} href="#"> {{unbound view.content.label}}</a>')
    })
  */
  });
});
window.require.register("views/main/admin/misc_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminMiscView = Em.View.extend({
    templateName: require('templates/main/admin/misc'),
    didInsertElement: function() {
      this.get('controller').loadUsers();
    }
  });
  
});
window.require.register("views/main/admin/rollbackHA/rollback_wizard_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.RollbackHighAvailabilityWizardView = Em.View.extend({

    didInsertElement: function() {
      var currentStep = this.get('controller.currentStep');
      if (currentStep > 4) {
        this.get('controller').setLowerStepsDisable(currentStep);
      }
    },

    templateName: require('templates/main/admin/rollbackHA/rollback_wizard'),

    isStep1Disabled: function () {
      return this.isStepDisabled(1);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.isStepDisabled(2);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.isStepDisabled(3);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStepDisabled: function (index) {
      return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
    }

  });
  
});
window.require.register("views/main/admin/rollbackHA/step1_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.RollbackHighAvailabilityWizardStep1View = Em.View.extend({

    templateName: require('templates/main/admin/rollbackHA/step1'),

    addNNHosts: null,
    sNNHosts: null,
    selectedSNNHost: null,
    selectedAddNNHost: null,

    didInsertElement: function() {
      var addNNHosts = App.HostComponent.find().filterProperty('componentName','NAMENODE');
      this.secondaryNNHosts = [];

      this.set('selectedSNNHost', this.get('controller.content.sNNHost'));
      this.set('selectedAddNNHost', this.get('controller.content.addNNHost'));

      if(addNNHosts.length == 2){
        this.set('addNNHosts', addNNHosts.mapProperty('host.hostName'));
      }
      App.Host.find().forEach(function(host){
        this.secondaryNNHosts.push(host.get('id'));
      },this);
      this.set('sNNHosts', this.secondaryNNHosts);
    },

    tipAddNNHost: function () {
      return this.get('controller.content.addNNHost');
    }.property('controller.content.addNNHost'),

    tipSNNHost: function () {
      return this.get('controller.content.sNNHost');
    }.property('controller.content.sNNHost'),

    done: function () {
      this.get('controller.content').set('selectedSNNHost', this.get('selectedSNNHost'));
      this.get('controller.content').set('selectedAddNNHost', this.get('selectedAddNNHost'));
      App.router.send('next');
    }

  });
  
});
window.require.register("views/main/admin/rollbackHA/step2_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.RollbackHighAvailabilityWizardStep2View = App.HighAvailabilityWizardStep4View.extend({

    templateName: require('templates/main/admin/rollbackHA/step2'),

    step2BodyText: function () {
      var activeNN = App.HostComponent.find().findProperty('displayNameAdvanced','Active NameNode');
      if(!activeNN){
        activeNN = App.HostComponent.find().findProperty('componentName','NAMENODE');
      }
      activeNN = activeNN.get('host.hostName');
      this.get('controller.content').set('activeNNHost', activeNN);
      return Em.I18n.t('admin.highAvailability.rollback.step2.body').format(this.get('controller.content.hdfsUser'), activeNN);
    }.property()

  });
  
});
window.require.register("views/main/admin/rollbackHA/step3_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.RollbackHighAvailabilityWizardStep3View = Em.View.extend({

    templateName: require('templates/main/admin/rollbackHA/step3'),

    didInsertElement: function() {

    }

  });
  
});
window.require.register("views/main/admin/security", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityView = Em.View.extend({
    templateName: require('templates/main/admin/security'),
    didInsertElement: function() {
      this.get('controller').setSecurityStatus();
    },

    configProperties: function () {
      var configProperties = [];
      var stepConfigs = this.get('controller.stepConfigs');
      if (stepConfigs) {
        this.get('controller.stepConfigs').mapProperty('configs').forEach(function (_stepProperties) {
          _stepProperties.forEach(function (_stepConfigProperty) {
            configProperties.pushObject(_stepConfigProperty);
          }, this);
        }, this);
      }
      return configProperties;
    }.property('controller.stepConfigs.@each.configs'),

    realmName: function () {
      return this.get('configProperties').findProperty('name', 'kerberos_domain');
    }.property('configProperties'),

    onRealmNameChange: function () {
      this.get('configProperties').forEach(function (_globalProperty) {
        if (/principal_name?$/.test(_globalProperty.get('name'))) {
          _globalProperty.set('unit', '@' + this.get('realmName.value'));
        }
      }, this);
    }.observes('realmName.value')

  });
});
window.require.register("views/main/admin/security/add/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityAddMenuView = Em.View.extend({

    templateName: require('templates/main/admin/security/add/menu'),

    isStep1Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',1).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',2).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',3).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep4Disabled: function () {
      return this.get('controller.isStepDisabled').findProperty('step',4).get('value');
    }.property('controller.isStepDisabled.@each.value').cacheable()

  });


  
});
window.require.register("views/main/admin/security/add/step1", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityAddStep1View = Em.View.extend({

    templateName: require('templates/main/admin/security/add/step1'),
    isATSInstalled: false,
    didInsertElement: function(){
      var isATSInstalled = this.get('controller').isATSInstalled();
      this.set('isATSInstalled',isATSInstalled);
    }

  });



  
});
window.require.register("views/main/admin/security/add/step2", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityAddStep2View = Em.View.extend({

    templateName: require('templates/main/admin/security/add/step2'),

    configProperties: function () {
      var configProperties = [];
      var stepConfigs = this.get('controller.stepConfigs');
      if (stepConfigs) {
        this.get('controller.stepConfigs').mapProperty('configs').forEach(function (_stepProperties) {
          _stepProperties.forEach(function (_stepConfigProperty) {
            configProperties.pushObject(_stepConfigProperty);
          }, this);
        }, this);
      }
      return configProperties;
    }.property('controller.stepConfigs.@each.configs'),

    realmName: function () {
      return this.get('configProperties').findProperty('name', 'kerberos_domain');
    }.property('configProperties'),

    onRealmNameChange: function () {
      this.get('configProperties').forEach(function (_globalProperty) {
        if (/principal_name?$/.test(_globalProperty.get('name')) || _globalProperty.get('name') == 'namenode_principal_name_falcon') {
          _globalProperty.set('unit', '@' + this.get('realmName.value'));
        }
      }, this);
    }.observes('realmName.value')

  });
  
});
window.require.register("views/main/admin/security/add/step3", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityAddStep3View = Em.View.extend({
    templateName: require('templates/main/admin/security/add/step3'),
    didInsertElement: function(){
      this.get('controller').loadStep();
    }
  });

  
});
window.require.register("views/main/admin/security/add/step4", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityAddStep4View = Em.View.extend({

    templateName: require('templates/main/admin/security/add/step4'),
    didInsertElement: function () {
      this.get('controller').loadStep();
    },
    msgColor: 'alert-info',
    message: Em.I18n.t('admin.security.step4.body.header'),
    onResult: function () {
      var stopServiceCommand = this.get('controller.commands').findProperty('name', 'STOP_SERVICES');
      var applyConfigCommand = this.get('controller.commands').findProperty('name', 'APPLY_CONFIGURATIONS');
        if (applyConfigCommand && applyConfigCommand.get('isSuccess') === true ) {
          this.set('message', Em.I18n.t('admin.security.step4.body.success.header'));
          this.set('msgColor','alert-success');
        } else if ((stopServiceCommand && stopServiceCommand.get('isError') === true) || (applyConfigCommand && applyConfigCommand.get('isError') === true)) {
          this.set('message', Em.I18n.t('admin.security.step4.body.failure.header'));
          this.set('msgColor','alert-error');
        } else {
          this.set('message', Em.I18n.t('admin.security.step4.body.header'));
          this.set('msgColor','alert-info');
        }
    }.observes('controller.commands.@each.isCompleted')

  });

  App.StageStatusView = Em.View.extend({
    tagName: 'tr',
    hasStarted: null,
    classNameBindings: ['faintText']
  });
  
});
window.require.register("views/main/admin/security/disable", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminSecurityDisableView = Em.View.extend({

    templateName: require('templates/main/admin/security/disable'),

    didInsertElement: function () {
      this.get('controller').loadStep();
    },
    msgColor: 'alert-info',
    message: Em.I18n.t('admin.security.disable.body.header'),
    onResult: function () {
      var stopServiceCommand = this.get('controller.commands').findProperty('name', 'STOP_SERVICES');
      var applyConfigCommand = this.get('controller.commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      if (applyConfigCommand && applyConfigCommand.get('isSuccess') === true ) {
        this.set('message', Em.I18n.t('admin.security.disable.body.success.header'));
        this.set('msgColor', 'alert-success');
      } else if ((stopServiceCommand && stopServiceCommand.get('isError') === true) || (applyConfigCommand && applyConfigCommand.get('isError') === true)) {
        this.set('message', Em.I18n.t('admin.security.disable.body.failure.header'));
        this.set('msgColor', 'alert-error');
      } else {
        this.set('message', Em.I18n.t('admin.security.disable.body.header'));
        this.set('msgColor', 'alert-info');
      }
    }.observes('controller.commands.@each.isCompleted')
  });
  
});
window.require.register("views/main/admin/stack_upgrade", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.StackUpgradeView = Em.View.extend({
    templateName: require('templates/main/admin/stack_upgrade'),

    isStep1Disabled: function () {
      return this.isStepDisabled(1);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.isStepDisabled(2);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.isStepDisabled(3);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStepDisabled: function (index) {
      return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
    }
  });
  
});
window.require.register("views/main/admin/user", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserView = Em.View.extend({
    templateName: require('templates/main/admin/user'),
    users: App.User.find(),
    delete: function(event, context){
      console.log("EVENT:");
      console.log(event);

      console.log("CONTEXT:");
      console.log(context);
    },
    ldapUser: function(){
      var auth = App.router.get('mainAdminAuthenticationController.content');
      return auth.get('method');
    }.property('App.router.mainAdminAuthenticationController.content')


  });
});
window.require.register("views/main/admin/user/create", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserCreateView = Em.View.extend({
    templateName: require('templates/main/admin/user/create'),
    userId: false,
    isPasswordDirty: false,

    create: function(event){
      var parent_controller=this.get("controller").controllers.mainAdminUserController;
      var form = this.get("userForm");
      if(form.isValid()) {
        form.getField("userName").set('value', form.getField("userName").get('value').toLowerCase());
        if(form.getField("admin").get('value') === "" || form.getField("admin").get('value') == true) {
          form.getField("roles").set("value","admin,user");
          form.getField("admin").set("value","true");
        } else{
          form.getField("roles").set("value","user");
        }
        parent_controller.sendCommandToServer('/users/' + form.getField("userName").get('value'), "POST" , {
          Users: {
            password: form.getField("password").get('value'),
            roles: form.getField("roles").get('value')
          }
        }, function (success) {

          if (!success) {
            App.ModalPopup.show({
              header: Em.I18n.t('admin.users.addButton'),
              body: Em.I18n.t('admin.users.createError'),
              primary: Em.I18n.t('ok'),
              secondary: null,
              onPrimary: function() {
                this.hide();
              }
            });
            return;
          }
          App.ModalPopup.show({
            header: Em.I18n.t('admin.users.addButton'),
            body: Em.I18n.t('admin.users.createSuccess'),
            primary: Em.I18n.t('ok'),
            secondary: null,
            onPrimary: function() {
              this.hide();
            }
          });
          var persists = App.router.get('applicationController').persistKey(form.getField("userName").get('value'));
          App.router.get('applicationController').postUserPref(persists,true);

          form.save();

          App.router.transitionTo("allUsers");
        })
      }
    },

    userForm: App.CreateUserForm.create({}),

    keyPress: function(event) {
      if (event.keyCode === 13) {
        this.create();
        return false;
      }
    },

    passwordValidation: function() {
      var passwordValue = this.get('userForm').getField('password').get('value');
      if (passwordValue && !this.get('isPasswordDirty')) {
        this.set('isPasswordDirty', true);
      }
      if (this.get('isPasswordDirty')) {
        this.get('userForm').isValid();
        this.get('userForm').isWarn();
      }
    }.observes('userForm.fields.@each.value'),

    didInsertElement: function(){
      this.get('userForm').propertyDidChange('object');
    }
  });
  
});
window.require.register("views/main/admin/user/edit", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserEditView = Em.View.extend({
    templateName: require('templates/main/admin/user/edit'),
    userId: false,
    edit: function(event){
      var form = this.get("userForm");
      if(form.isValid()) {
        var Users={};
        if(form.getField("admin").get('value') === "" || form.getField("admin").get('value') == true) {
          form.getField("roles").set("value","admin,user");
          form.getField("admin").set("value", true);
        } else{
          form.getField("roles").set("value","user");
        }

        Users.roles = form.getField("roles").get('value');

        if(form.getField("new_password").get('value') != "" && form.getField("old_password").get('value') != "") {
          Users.password = form.getField("new_password").get('value');
          Users.old_password = form.getField("old_password").get('value');
        }

        this.get("controller").sendCommandToServer('/users/' + form.getField("userName").get('value'), "PUT" , {
         Users:Users
        }, function (success, message) {
          if (!success) {
            App.ModalPopup.show({
              header: Em.I18n.t('admin.users.editButton'),
              body: message,
              primary: Em.I18n.t('ok'),
              secondary: null,
              onPrimary: function() {
                this.hide();
              }
            });
            return;
          }

          form.save();

          App.router.transitionTo("allUsers");
        })
      }
    },

    keyPress: function(event) {
      if (event.keyCode === 13) {
        this.edit();
        return false;
      }
    },

    userForm: App.EditUserForm.create({}),

    didInsertElement: function() {
      var form = this.get('userForm');
      if (form.getField("isLdap").get("value")) {
        form.getField("old_password").set("disabled", true);
        form.getField("new_password").set("disabled", true);
        form.getField("new_passwordRetype").set("disabled", true);
      } else {
        form.getField("old_password").set("disabled", false);
        form.getField("new_password").set("disabled", false);
        form.getField("new_passwordRetype").set("disabled", false);
      }
      form.propertyDidChange('object');
    }
  });
});
window.require.register("views/main/admin/user/row", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAdminUserRowView = Em.View.extend({
    templateName: require('templates/main/admin/user/row'),
    disableCheckBoxes: "disabled"
  });
});
window.require.register("views/main/apps/item/bar_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var graph = require('utils/graph');

  App.MainAppsItemBarView = Em.View.extend({
    elementId:'bars',
    templateName:require('templates/main/apps/item/bar'),
    content:function () {
      return this.get('controller.content.jobs');
    }.property('controller.content.jobs'),
    onLoad:function () {
      if (!this.get('controller.content.loadAllJobs') || this.get('activeJob')) {
        return;
      }

      var self = this;
      Ember.run.next(function(){
        self.updateTasksView();
      });
    }.observes('controller.content.loadAllJobs'),
    didInsertElement:function () {
      this.onLoad();
    },
    draw:function () {
    },

    mapNodeLocal:false,
    mapRackLocal:false,
    mapOffSwitch:false,
    reduceOffSwitch:false,
    submit:false,
    finish:false,

    updateTasksView:function () {
      var url = App.testMode ? '/data/apps/jobs/taskview.json' : App.apiPrefix + "/jobhistory/tasklocality?workflowId=" + this.get('controller.content.id');
      var mapper = App.jobTasksMapper;
      mapper.set('model', this);
      var self = this;
      App.HttpClient.get(url, mapper,{
        complete:function(jqXHR, textStatus) {
          self.set('loadJobTasks', true);
        }
      });
    },

    drawJobTasks:function () {
      if (!this.get('mapNodeLocal') || !this.get('mapRackLocal') || !this.get('mapOffSwitch') || !this.get('reduceOffSwitch')) {return;}
      graph.drawJobTasks(this.get('mapNodeLocal'), this.get('mapRackLocal'), this.get('mapOffSwitch'), this.get('reduceOffSwitch'), this.get('controller.content.startTime'), this.get('controller.content.startTime')+this.get('controller.content.elapsedTime'), this.$().width(), 300, 'job_tasks');
    }.observes('loadJobTasks')

  });
  
});
window.require.register("views/main/apps/item/dag_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var sort = require('views/common/sort_view');

  App.MainAppsItemDagView = Em.View.extend({
    templateName: require('templates/main/apps/item/dag'),
    elementId : 'jobs',
    content:function(){
      return this.get('controller.content.jobs');
    }.property('controller.content.jobs'),

    classNames:['table','dataTable'],
    /**
     * convert content to special jobs object for DagViewer
     */
    jobs: function(){
      var c = this.get('content');
      var result = [];
      c.forEach(function(item, index){
        result[index] = new Object({
          'name' : item.get('id'),
          'entityName' : item.get('workflow_entity_name'),
          'status' : item.get('status'),
          'input' : item.get('input'),
          'output' : item.get('output'),
          'submitTime' : item.get('submit_time'),
          'elapsedTime' : item.get('elapsed_time')
        })
      });
      return result;
    }.property('content'),

    loaded : false,

    hasManyJobs: function(){
      return (this.get('content') && this.get('content').length > 1);
    }.property('content'),

    onLoad:function (){
      if(!this.get('controller.content.loadAllJobs') || this.get('loaded')){
        return;
      }

      this.set('loaded', true);

      var self = this;

      Ember.run.next(function(){
        self.draw();
        self.updateTimeline();
      });

    }.observes('controller.content.loadAllJobs'),

    resizeModal: function () {
      var modal = $('.modal');
      var body = $('body');
      modal.find('.modal-body').first().css('max-height', 'none');
      var modalHeight = modal.height() + 300;
      var bodyHeight = body.height();
      if (modalHeight > bodyHeight) {
        modal.css('top', '20px');
        $('.modal-body').height(bodyHeight - 180);
      } else {
        modal.css('top', (bodyHeight - modalHeight) / 2 + 'px');
      }

      var modalWidth = modal.width();
      var bodyWidth = body.width();
      if (modalWidth > bodyWidth) {
        modal.css('left', '10px');
        modal.width(bodyWidth - 20);
      } else {
        modal.css('left', (bodyWidth - modalWidth) / 2 + 'px');
      }
    },

    didInsertElement: function(){
      this.onLoad();
    },

    loadTaskTimeline:false,
    loadJobTimeline:false,
    map:false,
    shuffle:false,
    reduce:false,
    allmap:false,
    allshuffle:false,
    allreduce:false,

    updateTimeline:function () {
      var url = App.testMode ? '/data/apps/jobs/timeline.json' : App.apiPrefix + "/jobhistory/task?workflowId=" + this.get('controller.content.id') + "&width=" + this.$().width() + "&startTime=" + this.get('controller.content.startTime') + "&endTime=" + (this.get('controller.content.startTime')+this.get('controller.content.elapsedTime'));
      var mapper = App.jobTimeLineMapper;
      mapper.set('model', this);
      var self = this;
      App.HttpClient.get(url, mapper,{
        complete:function(jqXHR, textStatus) {
          self.set('loadJobTimeline', true);
        }
      });
      url = App.testMode ? '/data/apps/jobs/timeline.json' : App.apiPrefix + "/jobhistory/task?width=" + this.$().width() + "&startTime=" + this.get('controller.content.startTime') + "&endTime=" + (this.get('controller.content.startTime')+this.get('controller.content.elapsedTime'));
      var mapper = App.taskTimeLineMapper;
      mapper.set('model', this);
      var self = this;
      App.HttpClient.get(url, mapper,{
        complete:function(jqXHR, textStatus) {
          self.set('loadTaskTimeline', true);
        }
      });
    },

    drawJobTimeline:function () {
      if (this.get('loadJobTimeline') && this.get('loadTaskTimeline')) {
        this.daggraph.addTimeSeries([{"name":"allmap","color":"green","values":this.get('allmap')},
          {"name":"map","color":"green","values":this.get('map')}], 0, "Maps");
        this.daggraph.addTimeSeries([
          {"name":"allshuffle","color":"lightblue","values":this.get('allshuffle')},
          {"name":"allreduce","color":"steelblue","values":this.get("allreduce")},
          {"name":"shuffle","color":"lightblue","values":this.get('shuffle')},
          {"name":"reduce","color":"steelblue","values":this.get("reduce")}], 1, "Reduces");
      }
    }.observes('loadJobTimeline', 'loadTaskTimeline'),

    daggraph:false,

    draw: function(){
      var dagSchema = this.get('controller.content.workflowContext');
      var jobs = this.get('jobs');
      this.resizeModal();
      this.daggraph = new DagViewer('dag_viewer')
          .setData(dagSchema, jobs)
          .drawDag(this.$().width(), 300, 20);
    },
    sortView: sort.wrapperView,
    nameSort: sort.fieldView.extend({
      name:'workflow_entity_name',
      displayName: Em.I18n.t('apps.item.dag.job')
    }),
    idSort: sort.fieldView.extend({
      name:'id',
      displayName: Em.I18n.t('apps.item.dag.jobId')
    }),
    statusSort: sort.fieldView.extend({
      name:'status',
      displayName: Em.I18n.t('apps.item.dag.status')
    }),
    mapsSort: sort.fieldView.extend({
      name:'maps',
      displayName: Em.I18n.t('apps.item.dag.maps')
    }),
    reducesSort: sort.fieldView.extend({
      name:'reduces',
      displayName: Em.I18n.t('apps.item.dag.reduces')
    }),
    inputSort: sort.fieldView.extend({
      name:'input',
      displayName: Em.I18n.t('apps.item.dag.input')
    }),
    outputSort: sort.fieldView.extend({
      name:'output',
      displayName: Em.I18n.t('apps.item.dag.output')
    }),
    durationSort: sort.fieldView.extend({
      name:'elapsed_time',
      displayName: Em.I18n.t('apps.item.dag.duration')
    })
  });
  
});
window.require.register("views/main/apps/item_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainAppsItemView = Em.View.extend({
    tagName: 'tr',
    classNames : ['containerRow'],
    templateName:require('templates/main/apps/item'),
    menuTabs:[
      Em.Object.create({
        label:Em.I18n.t('apps.dagCharts.popup.dag'),
        active:'active',
        content:'App.MainAppsItemDagView'
      }),
      Em.Object.create({
        label:Em.I18n.t('apps.dagCharts.popup.tasks'),
        active:'',
        content:'App.MainAppsItemBarView'
      })
    ],
    activeTab:null,
    switchTab:function(event){
      var tabs = this.get('menuTabs');
      for(var i = 0; i < tabs.length; i++){
        if(tabs[i] === event.context){
          tabs[i].set('active', 'active');
        }
        else {
          tabs[i].set('active', '');
        }
      }
      this.set('activeTab', event.context);
    },
    didInsertElement: function(){
      var tabs = this.get('menuTabs');
      tabs[0].set('active', 'active');
      tabs[1].set('active', '');
      this.set('activeTab', tabs[0]);
    },
    containerView: Em.ContainerView.extend({
      onchange:function(){

        if(this.get('childViews').length){
          this.get('childViews').get('firstObject').destroy();
        }

        var view = this.get('parentView.activeTab.content').split('.')[1];
        view = App[view].create({
          controllerBinding: 'App.router.mainAppsItemController'
        });
        this.get('childViews').pushObject(view);
      }.observes('parentView.activeTab')
    })
  });
  
});
window.require.register("views/main/apps_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');

  App.MainAppsView = Em.View.extend({
    templateName: require('templates/main/apps'),
    paginationInfo: function() {
      return this.t('tableView.filters.paginationInfo').format(this.get('controller.paginationObject.startIndex'), this.get('controller.paginationObject.endIndex'), this.get('controller.paginationObject.iTotalDisplayRecords'));
    }.property('controller.paginationObject.startIndex', 'controller.paginationObject.endIndex', 'controller.paginationObject.iTotalDisplayRecords'),
    //Pagination left/right buttons css class
    paginationLeft: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-arrow-left"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if (parseInt(this.get("controller.paginationObject.startIndex")) > 1) {
          return "paginate_previous";
        } else {
          return "paginate_disabled_previous";
        }
      }.property("controller.paginationObject.startIndex"),
      click: function (event) {
        if (this.get('class') == "paginate_previous") {
          var startIndex = parseInt(this.get("controller.paginationObject.startIndex")) - 1;
          var showRows = parseInt(this.get("controller.filterObject.iDisplayLength"));
          var startDisplayValue = Math.max(0, startIndex - showRows);
          this.set("controller.filterObject.iDisplayStart", startDisplayValue);
        }
      }
    }),
    paginationRight: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-arrow-right"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if ((parseInt(this.get("controller.paginationObject.endIndex"))) < parseInt(this.get("controller.paginationObject.iTotalDisplayRecords"))) {
          return "paginate_next";
        } else {
          return "paginate_disabled_next";
        }
      }.property("controller.paginationObject.endIndex"),
      click: function (event) {
        if (this.get('class') == "paginate_next") {
          var startDisplayValue = parseInt(this.get("controller.paginationObject.endIndex"));
          this.set("controller.filterObject.iDisplayStart", startDisplayValue);
        }
      }
    }),

    /**
     * If there are table rows with runs.
     */
    emptyData:true,

    /*
     If no runs to display set emptyData to true and reset Avg table data, else to set emptyData to false.
     */
    emptyDataObserver:function(){
      if(this.get("controller.paginationObject.iTotalRecords") != null && this.get("controller.paginationObject.iTotalDisplayRecords")>0){
        this.set("emptyData",false);
      }else{
        this.set("emptyData",true);
        this.set("controller.serverData",null);
      }
    }.observes("controller.paginationObject.iTotalDisplayRecords","controller.paginationObject.iTotalRecords"),


    /**
     * View for RunPerPage select component
     */
    runPerPageSelectView: Em.Select.extend({
      selected: '10',
      content: ['10', '25', '50', '100']
    }),

    wrapSorting: Ember.View.extend({
      tagName: 'tr'
    }),

    sortingColumns: Ember.View.extend({
      tagName: 'th',
      classNameBindings: ['class', 'widthClass'],
      class: "sorting",
      widthClass: "",
      content: null,
      defaultColumn: 8,

      didInsertElement: function () {
        this.set("widthClass", "col" + this.get('content.index'));
        if (this.get('content.index') == this.get('defaultColumn')) {
          this.setControllerObj(this.content.index, "DESC");
          this.set("class", "sorting_desc");
        }
      },
      click: function (event) {
        console.log(this.get('class'));
        if (this.get('class') == "sorting") {
          this.resetSortClass();
          this.setControllerObj(this.get('content.index'), "ASC");
          this.set("class", "sorting_asc");
        } else if (this.get('class') == "sorting_asc") {
          this.setControllerObj(this.get('content.index'), "DESC");
          this.set("class", "sorting_desc");
        } else if (this.get('class') == "sorting_desc") {
          this.setControllerObj(this.get('content.index'), "ASC");
          this.set("class", "sorting_asc");
        }
      },
      resetSortClass: function () {
        this.get("parentView.childViews").map(function (a, e) {
          a.get("childViews")[0].set("class", "sorting")
        });
      },
      setControllerObj: function (col, dir) {
        this.set("controller.filterObject.iSortCol_0", col);
        this.set("controller.filterObject.sSortDir_0", dir);
      }
    }),

    /**
     * Filter-field for Search
     */
    appSearchThrough: Em.TextField.extend({
      classNames: ['input-medium'],
      type: 'text',
      placeholder: Em.I18n.t('common.search')
    }),
    /**
     * Filter-field for App ID.
     * Based on <code>filters</code> library
     */
    appIdFilterView: filters.createTextView({
      valueBinding: "controller.filterObject.sSearch_0"
    }),
    /**
     * Filter-field for name.
     * Based on <code>filters</code> library
     */
    nameFilterView: filters.createTextView({
      valueBinding: "controller.filterObject.sSearch_1",
      fieldType: 'input-small'
    }),
    /**
     * Filter-field for type.
     * Based on <code>filters</code> library
     */
    typeFilterView: filters.createSelectView({
      fieldType: 'input-small',
      valueBinding: "controller.filterObject.runType",
      content: ['Any', 'Pig', 'Hive', 'MapReduce']
    }),

    /**
     * Filter-list for User.
     * Based on <code>filters</code> library
     */
    userFilterView: filters.createComponentView({
      /**
       * Inner FilterView. Used just to render component. Value bind to <code>mainview.value</code> property
       * Base methods was implemented in <code>filters.componentFieldView</code>
       */
      filterView: filters.componentFieldView.extend({
        templateName:require('templates/main/apps/user_filter'),

        usersBinding: 'controller.users',

        allComponentsChecked:false,
        toggleAllComponents:function () {
          var checked = this.get('allComponentsChecked');
          this.get('users').setEach('checked', checked);
        }.observes('allComponentsChecked'),

        clearFilter:function() {
          this.set('allComponentsChecked', false);

          this.get('users').setEach('checked', false);

          this._super();
        },

        applyFilter:function() {
          this._super();

          var chosenUsers = this.get('users').filterProperty('checked', true).mapProperty('name');
          this.set('value', chosenUsers.toString());
        }
      }),

      valueBinding: 'controller.filterObject.sSearch_3'
    }),
    /**
     * Filter-field for jobs.
     * Based on <code>filters</code> library
     */
    jobsFilterView: filters.createTextView({
      fieldType: 'input-super-mini',
      valueBinding: "controller.filterObject.jobs"
    }),
    /**
     * Filter-field for Input.
     * Based on <code>filters</code> library
     */
    inputFilterView: filters.createTextView({
      fieldType: 'input-super-mini',
      valueBinding: "controller.filterObject.input"
    }),
    /**
     * Filter-field for Output.
     * Based on <code>filters</code> library
     */
    outputFilterView: filters.createTextView({
      fieldType: 'input-super-mini',
      valueBinding: "controller.filterObject.output"
    }),
    /**
     * Filter-field for Duration.
     * Based on <code>filters</code> library
     */
    durationFilterView: filters.createTextView({
      fieldType: 'input-super-mini',
      valueBinding: "controller.filterObject.duration"
    }),
    /**
     * Filter-field for RunDate.
     * Based on <code>filters</code> library
     */
    runDateFilterView: filters.createSelectView({
      fieldType: 'input-medium',
      valueBinding: "controller.filterObject.runDate",
      content: ['Any', 'Past 1 Day', 'Past 2 Days', 'Past 7 Days', 'Past 14 Days', 'Past 30 Days']
    }),

    /**
     * Onclick handler for Show All/Filtered buttons
     */
    clickViewType: function (event) {
      this.set("controller.filterObject.viewTypeClickEvent", true);
      if ($(event.target).hasClass("filtered") || $(event.target.parentNode).hasClass("filtered")) {
        this.set("controller.filterObject.viewType", "filtered");
      } else {
        this.set("controller.filterObject.allFilterActivated", true);
        this.set("controller.filterObject.viewType", "all");
      }
    },
    /**
     * Clears up last job ID when coming in fresh to page.
     * Not doing this will result in failure to load job
     * data, and subsequently the popup dialog.
     */
    didInsertElement: function(){
      var self = this;
      Em.run.next(function() {
        self.get('_childViews').forEach(function(childView) {
          if(childView['showClearFilter']) {
            childView.showClearFilter();
          }
        });
      });
      this.get('controller').set('lastJobId', null);
      this.onChangeViewType();
    },
    /**
     *
     */
    onChangeViewType: function () {
      var tmpViewType = this.get('controller.filterObject.viewType');
      var filterButtons = $("#filter_buttons").children();
      filterButtons.each(function (index, element) {
        $(element).removeClass('selected');
        if (tmpViewType == $(element).data('view-type')) {
          $(element).addClass('selected');
        }
      });
      this.set("controller.filterObject.viewTypeClickEvent", false);
    }.observes("controller.filterObject.viewType"),

    /**
     * This Container View is used to render static table row(appTableRow) and additional dynamic content
     */
    containerRow: Em.ContainerView.extend({

      /**
       * Unique row id
       */
      id: function () {
        return this.get('run.id');
      }.property("run.id"),

      /**
       * Show/hide additional content appropriated for this row
       */
      expandToggle: function () {
        //App.router.get('mainAppsItemController').set('jobsLoaded', false);
        App.router.get('mainAppsItemController').set('content', this.get('run'));
        App.ModalPopup.show({
          classNames: ['big-modal'],
          header: Em.I18n.t('apps.dagCharts.popup'),
          bodyClass: App.MainAppsItemView.extend({
            controllerBinding: 'App.router.mainAppsItemController'
          }),
          secondary: null
        });
      }
    }),
    /**
     * Table-row view
     */
    appTableRow: Em.View.extend({
      templateName: require('templates/main/apps/list_row'),
      classNames: ['app-table-row'],
      tagName: "tr",
      onLoad: function() {
        var run = this.get('parentView.run');
        if (run.index) {
          var strip = (run.index % 2) ? 'even' : 'odd';
          this.$().addClass(strip);
        }
      }.observes('parentView.run'),

      didInsertElement: function() {
        this.onLoad();
      },
      mouseEnter: function (event, view) {
        $(event.currentTarget).addClass("hover")
      },
      mouseLeave: function (event, view) {
        $(event.currentTarget).removeClass("hover");
      },
      click: function (event, view) {
        this.get('parentView').expandToggle();
      }

    })

  });
});
window.require.register("views/main/charts", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainChartsView = Em.View.extend({
      templateName: require('templates/main/charts')
  });
});
window.require.register("views/main/charts/heatmap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MainChartsHeatmapView = Em.View.extend({
    templateName: require('templates/main/charts/heatmap'),
    spinner: null,
    didInsertElement: function () {
      this._super();
      // set default metric
      this.set('controller.selectedMetric', this.get('controller.allMetrics')[0].get('items')[0]);
      this.get('controller.racks').setEach('isLoaded', false);
      $("#heatmapDetailsBlock").hide();
    },
    showLoading: function () {
      if (this.get('controller.selectedMetric.loading') || !this.get('controller.racks').everyProperty('isLoaded')) {
        var e = document.getElementById("heatmap-metric-loading");
        if (e) {
          $(e).children('div.spinner').remove();
          var spinOpts = {
            lines: 9,
            length: 4,
            width: 2,
            radius: 3,
            top: '0',
            left: '0'
          };
          this.set('spinner', new Spinner(spinOpts).spin(e));
        }
      } else {
        var spinner = this.get('spinner');
        if (spinner) {
          spinner.stop();
        }
        this.set('spinner', null);
      }
    }.observes('controller.selectedMetric.loading', 'controller.racks.@each.isLoaded')
  });
});
window.require.register("views/main/charts/heatmap/heatmap_host", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var date = require('utils/date');

  var App = require('app');

  App.MainChartsHeatmapHostView = Em.View.extend({
    templateName: require('templates/main/charts/heatmap/heatmap_host'),
    /** @private */
    hostClass: 'hostBlock',

    /**
     * show Host details block and move it near the cursor
     * 
     * @param {Object} e
     * @this App.MainChartsHeatmapHostView
     */
    mouseEnter: function (e) {
      var host = this.get('content');
      var view = App.MainChartsHeatmapHostDetailView.create();
      $.each(view.get('details'), function(i){
        var val = host.get(i);
        if (i == 'diskUsage') {
          if (val == undefined || isNaN(val) || val == Infinity || val == -Infinity) {
            val = null;
          } else {
            val = val.toFixed(1);
          }
        } else if (i == 'cpuUsage') {
          if (val == undefined || isNaN(val)) {
            val = null;
          } else {
            val = val.toFixed(1);
          }
        } else if (i == 'memoryUsage') {
          if (val == undefined || isNaN(val) || val == Infinity || val == -Infinity) {
            val = null;
          } else {
            val = val.toFixed(1);
          }
        } else if (i == 'hostComponents') {
          if (val == undefined) {
            val = null;
          } else {
            val = val.filterProperty('isMaster').concat(val.filterProperty('isSlave')).mapProperty('displayName').join(', ');
          }
        }
        view.set('details.' + i, val);
      });
      var selectedMetric = this.get('controller.selectedMetric');
      if (selectedMetric) {
        var metricName = selectedMetric.get('name');
        var h2vMap = selectedMetric.get('hostToValueMap');
        if (h2vMap && metricName) {
          var value = h2vMap[host.get('hostName')];
          if (value == undefined || value == null) {
            value = this.t('charts.heatmap.unknown');
          } else {
            if (metricName == 'Garbage Collection Time') {
              value = date.timingFormat(parseInt(value));
            } else {
              if (isNaN(value)) {
                value = this.t('charts.heatmap.unknown');
              } else {
                value = value + selectedMetric.get('units');
              }
            }
          }
          view.set('details.metricName', metricName);
          view.set('details.metricValue', value);
        }
      }
      var detailsBlock = $("#heatmapDetailsBlock");
      detailsBlock.css('top', e.pageY + 10);
      detailsBlock.css('left', e.pageX + 10);
      detailsBlock.show();
    },

    /**
     * hide Host details block
     * 
     * @param {Object} e
     * @this App.MainChartsHeatmapHostView
     */
    mouseLeave: function (e) {
      $("#heatmapDetailsBlock").hide();
    },

    hostTemperatureStyle: function () {
      var controller = this.get('controller');
      var h2sMap = controller.get('hostToSlotMap');
      if (h2sMap) {
        var hostname = this.get('content.hostName');
        if (hostname) {
          var slot = h2sMap[hostname];
          if (slot > -1) {
            var slotDefs = controller.get('selectedMetric.slotDefinitions');
            if (slotDefs && slotDefs.length > slot) {
              return slotDefs[slot].get('cssStyle');
            }
          }
        }
      }
      return '';
    }.property('controller.hostToSlotMap')
  });
});
window.require.register("views/main/charts/heatmap/heatmap_host_detail", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainChartsHeatmapHostDetailView = Em.View.extend({
    templateName: require('templates/main/charts/heatmap/heatmap_host_detail'),
    /** @private */ classNames:['heatmap_host_details'],
    /** @private */ elementId:'heatmapDetailsBlock',
    /** @private */ details:{
      hostName:'test node',
      publicHostName:'test node',
      osType: 'OS',
      ip: '192.168.0.0',
      metricName: 'metric-name',
      metricValue: 'metric-value',
      diskUsage: '10',
      cpuUsage: '10',
      memoryUsage: '10',
      hostComponents: 'host components'
    }
  });
});
window.require.register("views/main/charts/heatmap/heatmap_rack", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var lazyloading = require('utils/lazy_loading');

  App.MainChartsHeatmapRackView = Em.View.extend({
    templateName: require('templates/main/charts/heatmap/heatmap_rack'),
    classNames: ['rack'],
    classNameBindings: ['visualSchema'],

    /** rack status block class */
    statusIndicator:'statusIndicator',
    /** loaded hosts of rack */
    hosts: [],

    willInsertElement: function () {
      this.set('hosts', []);
    },

    didInsertElement: function () {
      var rackHosts = this.get('rack.hosts').toArray();
      if (rackHosts.length > 100) {
        lazyloading.run({
          destination: this.get('hosts'),
          source: rackHosts,
          context: this.get('rack'),
          initSize: 25,
          chunkSize: 100,
          delay: 25
        });
      } else {
        this.set('hosts', rackHosts);
        this.set('rack.isLoaded', true);
      }
    },
    /**
     * Provides the CSS style for an individual host.
     * This can be used as the 'style' attribute of element.
     */
    hostCssStyle: function () {
      var rack = this.get('rack');
      var widthPercent = 100;
      var hostCount = rack.get('hosts.length');
      if (hostCount && hostCount < 11) {
        widthPercent = (100 / hostCount) - 0.5;
      } else {
        widthPercent = 10; // max out at 10%
      }
      return "width:" + widthPercent + "%;float:left;";
    }.property('rack')
  });
});
window.require.register("views/main/charts/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainChartsMenuView = Em.CollectionView.extend({
    tagName: 'ul',
    classNames: ["nav", "nav-tabs"],
    content:[
      /*{ label:'Heatmap', routing:'heatmap', active:"active"},
      { label:'Horizon Chart', routing:'horizon_chart'}*/
    ],

    init: function(){ this._super(); this.activateView(); },

    activateView:function () {
      $.each(this._childViews, function () {
        this.set('active', (this.get('content.routing') == 'heatmap' ? "active" : ""));
      });
    },

    deactivateChildViews: function() {
      $.each(this._childViews, function(){
        this.set('active', "");
      });
    },

    itemViewClass: Em.View.extend({
      classNameBindings: ["active"],
      active: "",
      template: Ember.Handlebars.compile('<a {{action showChart view.content.routing }} href="#"> {{unbound view.content.label}}</a>')
    })
  });
});
window.require.register("views/main/dashboard", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');

  App.MainDashboardView = Em.View.extend(App.UserPref, {
    templateName:require('templates/main/dashboard'),
    didInsertElement:function () {
      this.services();
      this.setWidgetsDataModel();
      this.setInitPrefObject();
      this.setOnLoadVisibleWidgets();
      this.set('isDataLoaded',true);
      Ember.run.next(this, 'makeSortable');
    },
    content:[],
    isDataLoaded: false,
    isClassicDashboard: false,

    makeSortable: function () {
      var self = this;
      $( "#sortable" ).sortable({
        items: "> div",
        //placeholder: "sortable-placeholder",
        cursor: "move",
        update: function (event, ui) {
          if (!App.testMode) {
            // update persist then translate to real
            var widgetsArray = $('div[viewid]'); // get all in DOM
            self.getUserPref(self.get('persistKey'));
            var oldValue = self.get('currentPrefObject');
            var newValue = Em.Object.create({
              dashboardVersion: oldValue.dashboardVersion,
              visible: [],
              hidden: oldValue.hidden,
              threshold: oldValue.threshold
            });
            var size = oldValue.visible.length;
            for(var j = 0; j <= size -1; j++){
              var viewID = widgetsArray.get(j).getAttribute('viewid');
              var id = viewID.split("-").get(1);
              newValue.visible.push(id);
            }
            self.postUserPref(self.get('persistKey'), newValue);
            //self.translateToReal(newValue);
          }
        }
      }).disableSelection();
    },

    setWidgetsDataModel: function () {
      var services = App.Service.find();
      var self = this;
      services.forEach(function (item) {
        switch (item.get('serviceName')) {
          case "HDFS":
            self.set('hdfs_model',  App.HDFSService.find(item.get('id')) || item);
            break;
          case "YARN":
            self.set('yarn_model', App.YARNService.find(item.get('id')) || item);
            break;
          case "MAPREDUCE":
            self.set('mapreduce_model', App.MapReduceService.find(item.get('id')) || item);
            break;
          case "HBASE":
            self.set('hbase_model', App.HBaseService.find(item.get('id')) || item);
            break;
          case "STORM":
            self.set('storm_model', item);
            break;
        }
      }, this);
    },

    setInitPrefObject: function() {
      //in case of some service not installed
      var visibleFull = [
        '2', '4', '8', '10',
        '17', '11', '12', '13', '14',
        '18', '1', '6', '5', '9',
        '3', '7', '15', '16', '20',
        '19', '21', '23',
        '24', '25', '26', '27',// all yarn
        '28' // storm
      ]; // all in order
      var hiddenFull = [['22','Region In Transition']];
      if (this.get('hdfs_model') == null) {
        var hdfs= ['1', '2', '3', '4', '5', '15', '17'];
        hdfs.forEach ( function (item) {
          visibleFull = visibleFull.without(item);
        }, this);
      }
      if (this.get('mapreduce_model') == null) {
        var map = ['6', '7', '8', '9', '10', '16', '18'];
        map.forEach ( function (item) {
          visibleFull = visibleFull.without(item);
        }, this);
      }
      if (this.get('hbase_model') == null) {
        var hbase = ['19', '20', '21', '23'];
        hbase.forEach ( function (item) {
          visibleFull = visibleFull.without(item);
        }, this);
        hiddenFull = [];
      }
      if (this.get('yarn_model') == null) {
        var yarn = ['24', '25', '26', '27'];
        yarn.forEach ( function (item) {
          visibleFull = visibleFull.without(item);
        }, this);
      }
      if (this.get('storm_model') == null) {
        var storm = ['28'];
        storm.forEach(function(item) {
          visibleFull = visibleFull.without(item);
        }, this);
      }
      var obj = this.get('initPrefObject');
      obj.set('visible', visibleFull);
      obj.set('hidden', hiddenFull);
    },

    hdfs_model: null,
    mapreduce_model: null,
    mapreduce2_model: null,
    yarn_model: null,
    hbase_model: null,
    storm_model: null,
    visibleWidgets: [],
    hiddenWidgets: [], // widget child view will push object in this array if deleted

    plusButtonFilterView: filters.createComponentView({
      /**
       * Base methods was implemented in <code>filters.componentFieldView</code>
       */
      hiddenWidgetsBinding: 'parentView.hiddenWidgets',
      visibleWidgetsBinding: 'parentView.visibleWidgets',
      layout: null,

      filterView: filters.componentFieldView.extend({
        templateName:require('templates/main/dashboard/plus_button_filter'),
        hiddenWidgetsBinding: 'parentView.hiddenWidgets',
        visibleWidgetsBinding: 'parentView.visibleWidgets',
        valueBinding: '',
        applyFilter:function() {
          this._super();
          var parent = this.get('parentView').get('parentView');
          var hiddenWidgets = this.get('hiddenWidgets');
          var checkedWidgets = hiddenWidgets.filterProperty('checked', true);

          if (App.testMode) {
            var visibleWidgets = this.get('visibleWidgets');
            checkedWidgets.forEach(function(item){
              var newObj = parent.widgetsMapper(item.id);
              visibleWidgets.pushObject(newObj);
              hiddenWidgets.removeObject(item);
            }, this);
          } else {
            //save in persist
            parent.getUserPref(parent.get('persistKey'));
            var oldValue = parent.get('currentPrefObject');
            var newValue = Em.Object.create({
              dashboardVersion: oldValue.dashboardVersion,
              visible: oldValue.visible,
              hidden: [],
              threshold: oldValue.threshold
            });
            checkedWidgets.forEach(function(item){
              newValue.visible.push(item.id);
              hiddenWidgets.removeObject(item);
            }, this);
            hiddenWidgets.forEach(function(item){
              newValue.hidden.push([item.id, item.displayName]);
            }, this);

            parent.postUserPref(parent.get('persistKey'), newValue);
            parent.translateToReal(newValue);
          }
        }
      })
    }),

    /**
     * translate from Json value got from persist to real widgets view
     */
    translateToReal: function (value) {
      var version = value.dashboardVersion;
      var visible = value.visible;
      var hidden = value.hidden;
      var threshold = value.threshold;

      if (version == 'classic') {
        this.set('isClassicDashboard', true);
      } else if (version == 'new') {
        this.set('isClassicDashboard', false);
        var visibleWidgets = [];
        var hiddenWidgets = [];
        // re-construct visibleWidgets and hiddenWidgets
        for (var j = 0; j <= visible.length -1; j++) {
          var id = visible[j];
          var widgetClass = this.widgetsMapper(id);
          //override with new threshold
          if (threshold[id].length > 0) {
            widgetClass.reopen({
              thresh1: threshold[id][0],
              thresh2: threshold[id][1]
            });
          }
          visibleWidgets.pushObject(widgetClass);
        }
        for (var j = 0; j <= hidden.length -1; j++) {
          var id = hidden[j][0];
          var title = hidden[j][1];
          hiddenWidgets.pushObject(Em.Object.create({displayName:title , id: id, checked: false}));
        }
        this.set('visibleWidgets', visibleWidgets);
        this.set('hiddenWidgets', hiddenWidgets);
      }
    },

    setOnLoadVisibleWidgets: function () {
      if (App.testMode) {
        this.translateToReal(this.get('initPrefObject'));
      } else {
        // called when first load/refresh/jump back page
        this.getUserPref(this.get('persistKey'));
        var currentPrefObject = this.get('currentPrefObject');
        if (currentPrefObject) { // fit for no dashboard version
          if (!currentPrefObject.dashboardVersion) {
            currentPrefObject.dashboardVersion = 'new';
            this.postUserPref(this.get('persistKey'), currentPrefObject);
          }
          this.set('currentPrefObject', this.checkServicesChange(currentPrefObject));
          this.translateToReal(this.get('currentPrefObject'));
        } else {
          // post persist then translate init object
          if(App.get('isAdmin')) {
            this.postUserPref(this.get('persistKey'), this.get('initPrefObject'));
          }
          this.translateToReal(this.get('initPrefObject'));
        }
      }
    },
    removeWidget: function (value, itemToRemove) {
      value.visible = value.visible.without(itemToRemove);
      for (var j = 0; j <= value.hidden.length -1; j++) {
        if (value.hidden[j][0] == itemToRemove) {
          value.hidden.splice(j, 1);
        }
      }
      return value;
    },
    containsWidget: function (value, item) {
      var flag = value.visible.contains (item);
      for (var j = 0; j <= value.hidden.length -1; j++) {
        if ( !flag && value.hidden[j][0] == item) {
          flag = true;
          break;
        }
      }
      return flag;
    },
    /**
     * check if stack has upgraded from HDP 1.0 to 2.0 OR add/delete services.
     * Update the value on server if true.
     */
    checkServicesChange: function (currentPrefObject) {
      var toDelete = $.extend(true, {}, currentPrefObject);
      var toAdd = [];
      var self = this;

      // check each service, find out the newly added service and already deleted service
      if (this.get('hdfs_model') != null) {
        var hdfsAndMetrics= ['1', '2', '3', '4', '5', '15', '17', '11', '12', '13', '14'];
        hdfsAndMetrics.forEach ( function (item) {
          toDelete = self.removeWidget(toDelete, item);
        }, this);
      }
      if (this.get('mapreduce_model') != null) {
        var map = ['6', '7', '8', '9', '10', '16', '18'];
        var flag = self.containsWidget(toDelete, map[0]);
        if (flag) {
          map.forEach ( function (item) {
            toDelete = self.removeWidget(toDelete, item);
          }, this);
        } else {
          toAdd = toAdd.concat(map);
        }
      }
      if (this.get('hbase_model') != null) {
        var hbase = ['19', '20', '21', '22', '23'];
        var flag = self.containsWidget(toDelete, hbase[0]);
        if (flag) {
          hbase.forEach ( function (item) {
            toDelete = self.removeWidget(toDelete, item);
          }, this);
        } else {
          toAdd = toAdd.concat(hbase);
        }
      }
      if (this.get('yarn_model') != null) {
        var yarn = ['24', '25', '26', '27'];
        var flag = self.containsWidget(toDelete, yarn[0]);
        if (flag) {
          yarn.forEach ( function (item) {
            toDelete = self.removeWidget(toDelete, item);
          }, this);
        } else {
          toAdd = toAdd.concat(yarn);
        }
      }
      if (this.get('storm_model') != null) {
        var storm = ['28'];
        var flag = self.containsWidget(toDelete, storm[0]);
        var self = this;
        if (flag) {
          storm.forEach ( function (item) {
            toDelete = self.removeWidget(toDelete, item);
          }, this);
        } else {
          toAdd = toAdd.concat(storm);
        }
      }
      var value = currentPrefObject;
      if (toDelete.visible.length || toDelete.hidden.length) {
        toDelete.visible.forEach ( function (item) {
          value = self.removeWidget(value, item);
        }, this);
        toDelete.hidden.forEach ( function (item) {
          value = self.removeWidget(value, item[0]);
        }, this);
      }
      if (toAdd.length) {
        value.visible = value.visible.concat(toAdd);
        var allThreshold = this.get('initPrefObject').threshold;
        // add new threshold OR override with default value
        toAdd.forEach ( function (item) {
          value.threshold[item] = allThreshold[item];
        }, this);
      }
      return value;
    },

    widgetsMapper: function (id) {
      switch(id){
        case '1': return App.NameNodeHeapPieChartView;
        case '2': return App.NameNodeCapacityPieChartView;
        case '3': return App.NameNodeCpuPieChartView;
        case '4': return App.DataNodeUpView;
        case '5': return App.NameNodeRpcView;
        case '6': return App.JobTrackerHeapPieChartView;
        case '7': return App.JobTrackerCpuPieChartView;
        case '8': return App.TaskTrackerUpView;
        case '9': return App.JobTrackerRpcView;
        case '10': return App.MapReduceSlotsView;
        case '11': return App.ChartClusterMetricsMemoryWidgetView;
        case '12': return App.ChartClusterMetricsNetworkWidgetView;
        case '13': return App.ChartClusterMetricsCPUWidgetView;
        case '14': return App.ChartClusterMetricsLoadWidgetView;
        case '15': return App.NameNodeUptimeView;
        case '16': return App.JobTrackerUptimeView;
        case '17': return App.HDFSLinksView;
        case '18': return App.MapReduceLinksView;
        case '19': return App.HBaseLinksView;
        case '20': return App.HBaseMasterHeapPieChartView;
        case '21': return App.HBaseAverageLoadView;
        case '22': return App.HBaseRegionsInTransitionView;
        case '23': return App.HBaseMasterUptimeView;
        case '24': return App.ResourceManagerHeapPieChartView;
        case '25': return App.ResourceManagerUptimeView;
        case '26': return App.NodeManagersLiveView;
        case '27': return App.YARNMemoryPieChartView;
        case '28': return App.SuperVisorUpView;
      }
    },

    currentPrefObject: null,
    initPrefObject: Em.Object.create({
      dashboardVersion: 'new',
      visible: [],
      hidden: [],
      threshold: {1: [80, 90], 2: [85, 95], 3: [90, 95], 4: [80, 90], 5: [1000, 3000], 6: [70, 90], 7: [90, 95], 8: [50, 75], 9: [30000, 120000],
        10: [], 11: [], 12: [], 13: [], 14: [], 15: [], 16: [], 17: [], 18: [], 19: [], 20: [70, 90], 21: [10, 19.2], 22: [3, 10], 23: [],
        24: [70, 90], 25: [], 26: [50, 75], 27: [50, 75], 28: [85, 95]} // id:[thresh1, thresh2]
    }),
    persistKey: function () {
      var loginName = App.router.get('loginName');
      return 'user-pref-' + loginName + '-dashboard';
    }.property(''),

    makeRequestAsync: false,

    getUserPrefSuccessCallback: function (response, request, data) {
      if (response) {
        console.log('Got persist value from server with key ' + data.key + '. Value is: ' + response);
        this.set('currentPrefObject', response);
      }
    },

    getUserPrefErrorCallback: function (request, ajaxOptions, error) {
      // this user is first time login
      if (request.status == 404) {
        console.log('Persist did NOT find the key');
        return null;
      }
    },

    resetAllWidgets: function(){
      var self = this;
      App.showConfirmationPopup(function() {
        if(!App.testMode){
          self.postUserPref(self.get('persistKey'), self.get('initPrefObject'));
        }
        self.translateToReal(self.get('initPrefObject'));
      });
    },

    switchToClassic: function () {
      if(!App.testMode){
        this.getUserPref(this.get('persistKey'));
        var oldValue = this.get('currentPrefObject');
        oldValue.dashboardVersion = 'classic';
        this.postUserPref(this.get('persistKey'), oldValue);
      }else{
        var oldValue = this.get('initPrefObject');
        oldValue.dashboardVersion = 'classic';
      }
      this.translateToReal(oldValue);
    },

    switchToNew: function () {
      if(!App.testMode){
        this.getUserPref(this.get('persistKey'));
        var oldValue = this.get('currentPrefObject');
        oldValue.dashboardVersion = 'new';
        this.postUserPref(this.get('persistKey'), oldValue);
        this.didInsertElement();
      }else{
        var oldValue = this.get('initPrefObject');
        oldValue.dashboardVersion = 'new';
        this.translateToReal(oldValue);
      }
    },

    updateServices: function(){
      var services = App.Service.find();
      services.forEach(function (item) {
        var view;
        switch (item.get('serviceName')) {
          case "HDFS":
            view = this.get('content').filterProperty('viewName', App.MainDashboardServiceHdfsView);
            view.objectAt(0).set('model', App.HDFSService.find(item.get('id')));
            break;
          case "MAPREDUCE":
            view = this.get('content').filterProperty('viewName', App.MainDashboardServiceMapreduceView);
            view.objectAt(0).set('model', App.MapReduceService.find(item.get('id')));
            break;
          case "HBASE":
            view = this.get('content').filterProperty('viewName', App.MainDashboardServiceHbaseView);
            view.objectAt(0).set('model', App.HBaseService.find(item.get('id')));
        }
      }, this);
    }.observes('App.router.updateController.isUpdate'),

    services: function () {
      var services = App.Service.find();
      if (this.get('content').length > 0) {
        return false
      }
      services.forEach(function (item) {
        var vName;
        var item2;
        switch (item.get('serviceName')) {
          case "HDFS":
            vName = App.MainDashboardServiceHdfsView;
            item2 = App.HDFSService.find(item.get('id'));
            break;
          case "YARN":
            vName = App.MainDashboardServiceYARNView;
            item2 = App.YARNService.find(item.get('id'));
            break;
          case "MAPREDUCE":
            vName = App.MainDashboardServiceMapreduceView;
            item2 = App.MapReduceService.find(item.get('id'));
            break;
          case "MAPREDUCE2":
            vName = App.MainDashboardServiceMapreduce2View;
            break;
          case "HBASE":
            vName = App.MainDashboardServiceHbaseView;
            item2 = App.HBaseService.find(item.get('id'));
            break;
          case "HIVE":
            vName = App.MainDashboardServiceHiveView;
            break;
          case "ZOOKEEPER":
            vName = App.MainDashboardServiceZookeperView;
            break;
          case "OOZIE":
            vName = App.MainDashboardServiceOozieView;
            break;
          default:
            vName = Em.View;
        }
        this.get('content').pushObject(Em.Object.create({
          viewName: vName,
          model: item2 || item
        }))
      }, this);
    },

    gangliaUrl: function () {
      return App.router.get('clusterController.gangliaUrl') + "/?r=hour&cs=&ce=&m=&s=by+name&c=HDPSlaves&tab=m&vn=";
    }.property('App.router.clusterController.gangliaUrl'),

    showAlertsPopup: function (event) {
      var service = event.context;
      App.router.get('mainAlertsController').loadAlerts(service.get('serviceName'), "SERVICE");
      App.ModalPopup.show({
        header: this.t('services.alerts.headingOfList'),
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/alert_notification_popup'),
          service: service,
          controllerBinding: 'App.router.mainAlertsController',
          warnAlerts: function () {
            return this.get('controller.alerts').filterProperty('isOk', false).filterProperty('ignoredForServices', false);
          }.property('controller.alerts'),

          warnAlertsCount: function () {
            return this.get('warnAlerts').length;
          }.property('warnAlerts'),

          warnAlertsMessage: function() {
            return Em.I18n.t('services.alerts.head').format(this.get('warnAlertsCount'));
          }.property('warnAlertsCount'),

          nagiosUrl: function () {
            return App.router.get('clusterController.nagiosUrl');
          }.property('App.router.clusterController.nagiosUrl'),

          closePopup: function () {
            this.get('parentView').hide();
          },

          viewNagiosUrl: function () {
            window.open(this.get('nagiosUrl'), "_blank");
            this.closePopup();
          },

          selectService: function () {
            App.router.transitionTo('services.service.summary', service);
            this.closePopup();
          }
        }),
        primary: Em.I18n.t('common.close'),
        secondary : null,
        didInsertElement: function () {
          this.$().find('.modal-footer').addClass('align-center');
          this.$().children('.modal').css({'margin-top': '-350px'});
        }
      });
      event.stopPropagation();
    }
  });
  
});
window.require.register("views/main/dashboard/cluster_metrics/cpu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartClusterMetricsCPU = App.ChartLinearTimeView.extend({
    id: "cluster-metrics-cpu",

    ajaxIndex: 'dashboard.cluster_metrics.cpu',

    title: Em.I18n.t('dashboard.clusterMetrics.cpu'),
    yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,
    isTimePagingDisable: true,
    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.cpu) {
        var cpu_idle;
        for ( var name in jsonData.metrics.cpu) {
          var seriesData = jsonData.metrics.cpu[name];
          if (seriesData) {
            var s = this.transformData(seriesData, name);
            if (Em.I18n.t('dashboard.clusterMetrics.cpu.displayNames.idle') == s.name) {
              cpu_idle = s;
            }
            else {
              seriesArray.push(s);
            }
          }
        }
        seriesArray.push(cpu_idle);
      }
      return seriesArray;
    },
    
    colorForSeries: function (series) {
      if (Em.I18n.t('dashboard.clusterMetrics.cpu.displayNames.idle') == series.name){
        return '#CFECEC';
      }
      return null;
    }
  });
});
window.require.register("views/main/dashboard/cluster_metrics/load", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster load
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartClusterMetricsLoad = App.ChartLinearTimeView.extend({
    id: "cluster-metrics-load",

    ajaxIndex: 'dashboard.cluster_metrics.load',
    isTimePagingDisable: true,
    renderer: 'line',
    title: Em.I18n.t('dashboard.clusterMetrics.load'),
    
    transformToSeries: function(jsonData){
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.load) {
        for ( var name in jsonData.metrics.load) {
          var displayName = name;
          var seriesData = jsonData.metrics.load[name];
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/dashboard/cluster_metrics/memory", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster memory metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartClusterMetricsMemory = App.ChartLinearTimeView.extend({
    id: "cluster-metrics-memory",

    ajaxIndex: 'dashboard.cluster_metrics.memory',

    isTimePagingDisable: true,
    title: Em.I18n.t('dashboard.clusterMetrics.memory'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',
    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.memory) {
        for ( var name in jsonData.metrics.memory) {
          var displayName = name;
          var seriesData = jsonData.metrics.memory[name];
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/dashboard/cluster_metrics/network", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster network metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartClusterMetricsNetwork = App.ChartLinearTimeView.extend({
    id: "cluster-metrics-network",

    ajaxIndex: 'dashboard.cluster_metrics.network',

    isTimePagingDisable: true,
    title: Em.I18n.t('dashboard.clusterMetrics.network'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    transformToSeries : function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.network) {
        for ( var name in jsonData.metrics.network) {
          var displayName = name;
          var seriesData = jsonData.metrics.network[name];
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/dashboard/service", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var uiEffects = require('utils/ui_effects');

  require('models/alert');

  App.MainDashboardServiceHealthView = Em.View.extend({
    classNameBindings: ["healthStatus", "healthStatusClass"],
    //template: Em.Handlebars.compile(""),
    blink: false,
    tagName: 'span',
    attributeBindings:['rel', 'title','data-original-title'],
    rel: 'HealthTooltip',
    'data-original-title': '',

    updateToolTip: function () {
      this.set('data-original-title', this.get('service.toolTipContent'));
    }.observes('service.toolTipContent'),

    /**
     * When set to true, extending classes should
     * show only tabular rows as they will be 
     * embedded into other tables.
     */
    showOnlyRows: false,

    startBlink: function () {
      this.set('blink', true);
    },

    doBlink: function () {
      var self = this;
      if (this.get('blink') && (this.get("state") == "inDOM")) {
        uiEffects.pulsate(self.$(), 1000, function(){
          self.doBlink();
        });
      }
    }.observes('blink'),

    stopBlink: function () {
      this.set('blink', false);
    },

    healthStatus: function () {
      if (this.get('service.isClientsOnly')) {
        return 'icon-laptop';
      }
      if (this.get('service.passiveState') != 'OFF') {
        return 'icon-medkit';
      }
      var status = this.get('service.healthStatus');
      switch (status) {
        case 'green':
          status = App.Service.Health.live;
          this.stopBlink();
          break;
        case 'green-blinking':
          status = App.Service.Health.live;
          this.startBlink();
          break;
        case 'red-blinking':
          status = App.Service.Health.dead;
          this.startBlink();
          break;
        case 'yellow':
          status = App.Service.Health.unknown;
          break;
        default:
          status = App.Service.Health.dead;
          this.stopBlink();
          break;
      }

      return 'health-status-' + status;
    }.property('service.healthStatus','service.passiveState','service.isClientsOnly'),

    healthStatusClass: function () {
      switch (this.get('healthStatus')) {
        case 'health-status-LIVE':
          return App.healthIconClassGreen;
          break;
        case 'health-status-DEAD-RED':
          return App.healthIconClassRed;
          break;
        case 'health-status-DEAD-YELLOW':
          return App.healthIconClassYellow;
          break;
        case 'health-status-DEAD-ORANGE':
          return App.healthIconClassOrange;
          break;
        default:
          return "";
          break;
      }
    }.property('healthStatus'),

    didInsertElement: function () {
      this.updateToolTip();
      App.tooltip($("[rel='HealthTooltip']"));
      this.doBlink(); // check for blink availability
    }
  });

  App.ComponentLiveTextView =  Em.View.extend({
    classNameBindings: ['color:service-summary-component-red-dead:service-summary-component-green-live'],
    service: function() { return this.get("parentView").get("service")}.property("parentView.service"),
    liveComponents: function() {
    }.property(),
    totalComponents: function() {
    }.property(),
    color: function() {
      return this.get("liveComponents") == 0;
    }.property("liveComponents")
  });

  App.MainDashboardServiceView = Em.View.extend({
    classNames: ['service', 'clearfix'],

    data: function () {
      return this.get('controller.data.' + this.get('serviceName'));
    }.property('controller.data'),

    dashboardMasterComponentView : Em.View.extend({
      didInsertElement: function() {
        App.tooltip($('[rel=healthTooltip]'));
      },
      templateName: require('templates/main/service/info/summary/master_components'),
      mastersComp : function(){
       return this.get('parentView.service.hostComponents').filterProperty('isMaster', true);
      }.property("service")
    }),

    formatUnavailable: function(value){
      return (value || value == 0) ? value : this.t('services.service.summary.notAvailable');
    },

    criticalAlertsCount: function () {
      return this.get('service.criticalAlertsCount');
    }.property('service.criticalAlertsCount'),

    isCollapsed: false,

    toggleInfoView: function () {
      this.$('.service-body').toggle('blind', 200);
      this.set('isCollapsed', !this.isCollapsed);
    },

    masters: function(){
      return this.get('service.hostComponents').filterProperty('isMaster', true);
    }.property('service'),

    clients: function(){
      var clients = this.get('service.hostComponents').filterProperty('isClient', true);
      var len = clients.length;
      var template = 'dashboard.services.{0}.client'.format(this.get('serviceName').toLowerCase());
      if(len > 1){
        template += 's';
      }

      return {
        title: this.t(template).format(len),
        component: clients.objectAt(0)
      };
    }.property('service')

  });
  
});
window.require.register("views/main/dashboard/service/flume", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.MainDashboardServiceFlumeView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/flume'),
    serviceName: 'flume',
    
    summaryHeader: function () {
      var count = this.get('service.nodes.length');
      if(count==1)
        return this.t("dashboard.services.flume.summary.single").format(count);
      return this.t("dashboard.services.flume.summary.multiple").format(count);
    }.property('service.nodes'),

    flumeServerComponent: function () {
      return App.HostComponent.find().findProperty('componentName', 'FLUME_SERVER');
    }.property()

  });
  
});
window.require.register("views/main/dashboard/service/hbase", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');
  var numberUtils = require('utils/number_utils');

  App.MainDashboardServiceHbaseView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/hbase'),
    serviceName: 'hbase',
    /**
     * All master components
     */
    masters: function () {
      return this.get('service.hostComponents').filterProperty('isMaster', true);
    }.property('service.hostComponents.length'),
    /**
     * Passive master components
     */
    passiveMasters: function () {
      if(App.supports.multipleHBaseMasters){
        return this.get('masters').filterProperty('haStatus', 'false');
      }
      return [];
    }.property('masters'),


    liveRegionServers: function () {
      return this.get('service.regionServers').filterProperty("workStatus", "STARTED");
    }.property('service.regionServers.@each.workStatus'),

    regionServesText: function () {
      if (this.get('service.regionServers.length') == 0) {
        return '';
      } else if (this.get('service.regionServers.length') > 1) {
        return Em.I18n.t('services.service.summary.viewHosts');
      } else {
        return Em.I18n.t('services.service.summary.viewHost');
      }
    }.property("service"),

    regionServersLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function() {
        return this.get('service.regionServers').filterProperty("workStatus","STARTED").get('length');
      }.property('service.regionServers.@each.workStatus'),
      totalComponents: function() {
        return this.get("service.regionServers.length");
      }.property("service.regionServers.length")
    }),

    /**
     * One(!) active master component
     */
    activeMaster: function () {
      if(App.supports.multipleHBaseMasters){
        return this.get('masters').findProperty('haStatus', 'true');
      } else {
        return this.get('masters')[0];
      }
    }.property('masters'),

    activeMasterTitle: function(){
      if(App.supports.multipleHBaseMasters){
        return this.t('service.hbase.activeMaster');
      } else {
        return this.get('activeMaster.host.publicHostName');
      }
    }.property('activeMaster'),

    masterServerHeapSummary: function () {
      var heapUsed = this.get('service').get('heapMemoryUsed');
      var heapMax = this.get('service').get('heapMemoryMax');
      var percent = heapMax > 0 ? 100 * heapUsed / heapMax : 0;
      var heapString = numberUtils.bytesToSize(heapUsed, 1, "parseFloat");
      var heapMaxString = numberUtils.bytesToSize(heapMax, 1, "parseFloat");
      return this.t('dashboard.services.hbase.masterServerHeap.summary').format(heapString, heapMaxString, percent.toFixed(1));
    }.property('service.heapMemoryUsed', 'service.heapMemoryMax'),

    summaryHeader: function () {
      var avgLoad = this.get('service.averageLoad');
      if (isNaN(avgLoad)) {
        avgLoad = this.t("services.service.summary.unknown");
      }
      return this.t("dashboard.services.hbase.summary").format(this.get('service.regionServers.length'), avgLoad);
    }.property('service.regionServers', 'service.averageLoad'),

    hbaseMasterWebUrl: function () {
      if (this.get('activeMaster.host') && this.get('activeMaster.host').get('publicHostName')) {
        return "http://" + (App.singleNodeInstall ? App.singleNodeAlias : this.get('activeMaster.host').get('publicHostName')) + ":60010";
      }
    }.property('activeMaster'),

    averageLoad: function () {
      var avgLoad = this.get('service.averageLoad');
      if (isNaN(avgLoad)) {
        avgLoad = this.t('services.service.summary.notAvailable');
      }
      return this.t('dashboard.services.hbase.averageLoadPerServer').format(avgLoad);
    }.property("service.averageLoad"),

    masterStartedTime: function () {
      var uptime = this.get('service').get('masterStartTime');
      if (uptime && uptime > 0) {
        var diff = App.dateTime() - uptime;
        if (diff < 0) {
          diff = 0;
        }
        var formatted = date.timingFormat(diff);
        return this.t('dashboard.services.uptime').format(formatted);
      }
      return this.t('services.service.summary.notRunning');
    }.property("service.masterStartTime"),

    masterActivatedTime: function () {
      var uptime = this.get('service').get('masterActiveTime');
      if (uptime && uptime > 0) {
        var diff = App.dateTime() - uptime;
        if (diff < 0) {
          diff = 0;
        }
        var formatted = date.timingFormat(diff);
        return this.t('dashboard.services.uptime').format(formatted);
      }
      return this.t('services.service.summary.notRunning');
    }.property("service.masterActiveTime"),

    regionServerComponent: function () {
      return this.get('service.regionServers').objectAt(0);
    }.property()

  });
});
window.require.register("views/main/dashboard/service/hdfs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');
  var numberUtils = require('utils/number_utils');

  App.MainDashboardServiceHdfsView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/hdfs'),
    serviceName: 'HDFS',
    Chart: App.ChartPieView.extend({
      service: null,
      color: '#0066B3',
      stroke: '#0066B3',
      palette: new Rickshaw.Color.Palette({
        scheme: [ 'rgba(0,102,179,0)', 'rgba(0,102,179,1)'].reverse()
      }),
      data: function () {
        var total = this.get('service.capacityTotal') + 0;
        var remaining = (this.get('service.capacityRemaining') + 0);
        var used = total - remaining;
        return [ used, remaining ];
      }.property('service.capacityUsed', 'service.capacityTotal')
    }),

    dashboardMasterComponentView: Em.View.extend({
      didInsertElement: function() {
        App.tooltip($('[rel=healthTooltip]'));
      },
      templateName: require('templates/main/service/info/summary/master_components'),
      mastersComp : function() {
        return this.get('parentView.service.hostComponents').filter(function(comp){
          return comp.get('isMaster') && comp.get('componentName') !== 'JOURNALNODE';
        });
      }.property('parentView.service.hostComponents')
    }),

    dataNodesLive: function () {
      return this.get('service.dataNodes').filterProperty("workStatus", "STARTED");
    }.property('service.dataNodes.@each.workStatus'),
    dataNodesDead: function () {
      return this.get('service.dataNodes').filterProperty("workStatus", "INSTALLED");
    }.property('service.dataNodes.@each.workStatus'),

    showJournalNodes: function () {
      return this.get('service.journalNodes.length') > 0;
    }.property('service.journalNodes.length'),

    dataNodesLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function () {
        return this.get('service.dataNodes').filterProperty("workStatus", "STARTED").get("length");
      }.property("service.dataNodes.@each.workStatus"),
      totalComponents: function() {
        return this.get("service.dataNodes.length");
      }.property("service.dataNodes.length")
    }),

    journalNodesLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function () {
        return this.get('service.journalNodes').filterProperty("workStatus", "STARTED").get("length");
      }.property("service.journalNodes.@each.workStatus"),
      totalComponents: function () {
        return this.get('service.journalNodes.length');
      }.property("service.journalNodes.length")
    }),

    dfsTotalBlocks: function(){
      return this.formatUnavailable(this.get('service.dfsTotalBlocks'));
    }.property('service.dfsTotalBlocks'),
    dfsTotalFiles: function(){
      return this.formatUnavailable(this.get('service.dfsTotalFiles'));
    }.property('service.dfsTotalFiles'),
    dfsCorruptBlocks: function(){
      return this.formatUnavailable(this.get('service.dfsCorruptBlocks'));
    }.property('service.dfsCorruptBlocks'),
    dfsMissingBlocks: function(){
      return this.formatUnavailable(this.get('service.dfsMissingBlocks'));
    }.property('service.dfsMissingBlocks'),
    dfsUnderReplicatedBlocks: function(){
      return this.formatUnavailable(this.get('service.dfsUnderReplicatedBlocks'));
    }.property('service.dfsUnderReplicatedBlocks'),

    blockErrorsMessage: function() {
      return Em.I18n.t('dashboard.services.hdfs.blockErrors').format(this.get('dfsCorruptBlocks'), this.get('dfsMissingBlocks'), this.get('dfsUnderReplicatedBlocks'));
    }.property('dfsCorruptBlocks','dfsMissingBlocks','dfsUnderReplicatedBlocks'),

    nodeUptime: function () {
      var uptime = this.get('service').get('nameNodeStartTime');
      if (uptime && uptime > 0){
        var diff = App.dateTime() - uptime;
        if (diff < 0) {
          diff = 0;
        }
        var formatted = date.timingFormat(diff);
        return this.t('dashboard.services.uptime').format(formatted);
      }
      return this.t('services.service.summary.notRunning');
    }.property("service.nameNodeStartTime"),

    nodeWebUrl: function () {
      return "http://" + (App.singleNodeInstall ? App.singleNodeAlias :  this.get('service').get('nameNode').get('publicHostName')) + ":50070";
    }.property('service.nameNode'),

    nodeHeap: function () {
      var memUsed = this.get('service').get('jvmMemoryHeapUsed');
      var memMax = this.get('service').get('jvmMemoryHeapMax');
      var percent = memMax > 0 ? ((100 * memUsed) / memMax) : 0;
      return this.t('dashboard.services.hdfs.nodes.heapUsed').format(
        numberUtils.bytesToSize(memUsed, 1, 'parseFloat'),
        numberUtils.bytesToSize(memMax, 1, 'parseFloat'),
        percent.toFixed(1));
    }.property('service.jvmMemoryHeapUsed', 'service.jvmMemoryHeapMax'),

    summaryHeader: function () {
      var text = this.t("dashboard.services.hdfs.summary");
      var service = this.get('service');
      var liveCount = service.get('dataNodes').filterProperty("workStatus", "STARTED").length;
      var totalCount = service.get('dataNodes').get('length');
      var total = service.get('capacityTotal') + 0;
      var remaining = service.get('capacityRemaining') + 0;
      var used = total - remaining;
      var percent = total > 0 ? ((used * 100) / total).toFixed(1) : 0;
      if (percent == "NaN" || percent < 0) {
        percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      return text.format(liveCount, totalCount, percent);
    }.property('service.dataNodes.@each.workStatus', 'service.capacityUsed', 'service.capacityTotal'),

    dfsUsedDisk: function () {
      var text = this.t("dashboard.services.hdfs.capacityUsed");
      var total = this.get('service.capacityTotal');
      var dfsUsed = this.get('service.capacityUsed');
      var percent = total > 0 ? ((dfsUsed * 100) / total).toFixed(2) : 0;
      if (percent == "NaN" || percent < 0) {
        percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      return text.format(numberUtils.bytesToSize(dfsUsed, 1, 'parseFloat'), numberUtils.bytesToSize(total, 1, 'parseFloat'), percent);
    }.property('service.capacityUsed', 'service.capacityTotal'),
    nonDfsUsedDisk: function () {
      var text = this.t("dashboard.services.hdfs.capacityUsed");
      var total = this.get('service.capacityTotal');
      var remaining = this.get('service.capacityRemaining');
      var dfsUsed = this.get('service.capacityUsed');
      var nonDfsUsed = total !== null && remaining !== null && dfsUsed !== null ? total - remaining - dfsUsed : null;
      var percent = total > 0 ? ((nonDfsUsed * 100) / total).toFixed(2) : 0;
      if (percent == "NaN" || percent < 0) {
        percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      return text.format(numberUtils.bytesToSize(nonDfsUsed, 1, 'parseFloat'), numberUtils.bytesToSize(total, 1, 'parseFloat'), percent);
    }.property('service.capacityUsed', 'service.capacityRemaining', 'service.capacityTotal'),
    remainingDisk: function () {
      var text = this.t("dashboard.services.hdfs.capacityUsed");
      var total = this.get('service.capacityTotal');
      var remaining = this.get('service.capacityRemaining');
      var percent = total > 0 ? ((remaining * 100) / total).toFixed(2) : 0;
      if (percent == "NaN" || percent < 0) {
        percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      return text.format(numberUtils.bytesToSize(remaining, 1, 'parseFloat'), numberUtils.bytesToSize(total, 1, 'parseFloat'), percent);
    }.property('service.capacityRemaining', 'service.capacityTotal'),

    dataNodeComponent: function () {
      return this.get('service.dataNodes').objectAt(0);
    }.property(),

    journalNodeComponent: function () {
      return this.get('service.journalNodes').objectAt(0);
    }.property(),

    safeModeStatus: function () {
      var safeMode = this.get('service.safeModeStatus');
      if (safeMode == null) {
        return Em.I18n.t("services.service.summary.notAvailable");
      } else if (safeMode.length == 0) {
        return Em.I18n.t("services.service.summary.safeModeStatus.notInSafeMode");
      } else {
        return Em.I18n.t("services.service.summary.safeModeStatus.inSafeMode");
      }
    }.property('service.safeModeStatus'),
    upgradeStatus: function () {
      var upgradeStatus = this.get('service.upgradeStatus');
      var healthStatus = this.get('service.healthStatus');
      if (upgradeStatus) {
        return Em.I18n.t('services.service.summary.pendingUpgradeStatus.notPending');
      } else if (healthStatus == 'green') {
        return Em.I18n.t('services.service.summary.pendingUpgradeStatus.pending');
      } else {
        return Em.I18n.t("services.service.summary.notAvailable");
      }
    }.property('service.upgradeStatus', 'service.healthStatus')
  });
  
});
window.require.register("views/main/dashboard/service/hive", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainDashboardServiceHiveView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/hive'),
    serviceName: 'hive',

    titleMasters: function(){
      var masters = this.get('masters');
      return [masters.findProperty('componentName', 'HIVE_SERVER'), masters.findProperty('componentName', 'HIVE_METASTORE')];
    }.property('service')
  });
});
window.require.register("views/main/dashboard/service/mapreduce", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');
  var numberUtils = require('utils/number_utils');

  App.MainDashboardServiceMapreduceView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/mapreduce'),
    serviceName: 'MAPREDUCE',
    jobTrackerWebUrl: function () {
      return "http://" + (App.singleNodeInstall ? App.singleNodeAlias : this.get('service').get('jobTracker').get('publicHostName')) + ":50030";
    }.property('service.jobTracker'),

    Chart: App.ChartLinearView.extend({
      data: function () {
        return this.get('_parentView.data.chart');
      }.property('_parentView.data.chart')
    }),

    jobTrackerUptime: function () {
      var uptime = this.get('service').get('jobTrackerStartTime');
      if (uptime && uptime > 0){
        var diff = App.dateTime() - uptime;
        if (diff < 0) {
          diff = 0;
        }
        var formatted = date.timingFormat(diff);
        return this.t('dashboard.services.uptime').format(formatted);

      }
      return this.t('services.service.summary.notRunning');
    }.property("service.jobTrackerStartTime"),

    summaryHeader: function () {
      var svc = this.get('service');
      var liveCount = svc.get('aliveTrackers').get('length');
      var allCount = svc.get('taskTrackers').get('length');
      var runningCount = svc.get('jobsRunning');
      if (runningCount === null) {
        runningCount = 'n/a';
      }
      var template = this.t('dashboard.services.mapreduce.summary');
      return template.format(liveCount, allCount, runningCount);
    }.property('service.aliveTrackers', 'service.taskTrackers', 'service.jobsRunning'),

    trackersText: function () {
      if (this.get('service').get('taskTrackers.length') == 0) {
        return '';
      } else if (this.get('service').get('taskTrackers.length') > 1){
        return Em.I18n.t('services.service.summary.viewHosts');
      }else{
        return Em.I18n.t('services.service.summary.viewHost');
      }
    }.property("service.taskTrackers.length"),

    trackersSummary: function () {
      var svc = this.get('service');
      var liveCount = svc.get('taskTrackers').filterProperty("workStatus", "STARTED").length;
      var totalCount = svc.get('taskTrackers').get('length');
      var template = this.t('dashboard.services.mapreduce.trackersSummary');
      return template.format(liveCount, totalCount);
    }.property('service.aliveTrackers.length', 'service.taskTrackers.@each.workStatus'),

    trackersLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function () {
        return this.get('service.taskTrackers').filterProperty("workStatus", "STARTED").get("length");
      }.property("service.taskTrackers.@each.workStatus"),
      totalComponents: function() {
        return this.get("service.taskTrackers.length");
      }.property('service.taskTrackers.length')
    }),

    trackersHeapSummary: function () {
      var heapUsed = this.get('service').get('jobTrackerHeapUsed');
      var heapMax = this.get('service').get('jobTrackerHeapMax');
      var percent = heapMax > 0 ? 100 * heapUsed / heapMax : 0;
      return this.t('dashboard.services.mapreduce.jobTrackerHeapSummary').format(numberUtils.bytesToSize(heapUsed, 1, "parseFloat"), numberUtils.bytesToSize(heapMax, 1, "parseFloat"), percent.toFixed(1));
    }.property('service.jobTrackerHeapUsed', 'service.jobTrackerHeapMax'),

    jobsSummary: function () {
      var svc = this.get('service');
      var template = this.t('dashboard.services.mapreduce.jobsSummary');
      return template.format(this.formatUnavailable(svc.get('jobsSubmitted')), this.formatUnavailable(svc.get('jobsCompleted')));
    }.property('service.jobsSubmitted', 'service.jobsCompleted'),

    mapSlotsSummary: function () {
      var svc = this.get('service');
      var template = this.t('dashboard.services.mapreduce.mapSlotsSummary');
      return template.format(this.formatUnavailable(svc.get('mapSlotsOccupied')), this.formatUnavailable(svc.get('mapSlotsReserved')));
    }.property('service.mapSlotsOccupied', 'service.mapSlotsReserved'),

    reduceSlotsSummary: function () {
      var svc = this.get('service');
      var template = this.t('dashboard.services.mapreduce.reduceSlotsSummary');
      return template.format(this.formatUnavailable(svc.get('reduceSlotsOccupied')), this.formatUnavailable(svc.get('reduceSlotsReserved')));
    }.property('service.reduceSlotsOccupied', 'service.reduceSlotsReserved'),

    mapTasksSummary: function () {
      var svc = this.get('service');
      var template = this.t('dashboard.services.mapreduce.tasksSummary');
      return template.format(this.formatUnavailable(svc.get('mapsRunning')), this.formatUnavailable(svc.get('mapsWaiting')));
    }.property('service.mapsRunning', 'service.mapsWaiting'),

    reduceTasksSummary: function () {
      var svc = this.get('service');
      var template = this.t('dashboard.services.mapreduce.tasksSummary');
      return template.format(this.formatUnavailable(svc.get('reducesRunning')), this.formatUnavailable(svc.get('reducesWaiting')));
    }.property('service.reducesRunning', 'service.reducesWaiting'),

    slotsCapacitySummary: function () {
      var mapSlots = this.get('service').get('mapSlots');
      var reduceSlots = this.get('service').get('reduceSlots');
      var liveNodeCount = this.get('service').get('aliveTrackers').get('length');
      if(liveNodeCount != 0){
        var avg = (mapSlots + reduceSlots) / liveNodeCount;
      }else{
        avg = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      return this.t('dashboard.services.mapreduce.slotCapacitySummary').format(mapSlots, reduceSlots, avg);
    }.property('service.mapSlots', 'service.reduceSlots', 'service.aliveTrackers'),

    taskTrackerComponent: function () {
      return this.get('service.taskTrackers').objectAt(0);
    }.property()
  });
  
});
window.require.register("views/main/dashboard/service/mapreduce2", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MainDashboardServiceMapreduce2View = App.MainDashboardServiceView.extend({
      templateName: require('templates/main/dashboard/service/mapreduce2'),
      serviceName: 'MAPREDUCE2',

      titleInfo: function(){
        var hs = this.get('service.hostComponents').findProperty('componentName', 'HISTORYSERVER');
        if (!hs) {
          return this.t('services.mapreduce2.history.unknown');
        } else if (hs.get('workStatus') === App.HostComponentStatus.started) {
          return this.t('services.mapreduce2.history.running');
        } else {
          return this.t('services.mapreduce2.history.stopped');
        }
      }.property('service')
    });
});
window.require.register("views/main/dashboard/service/oozie", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainDashboardServiceOozieView = App.MainDashboardServiceView.extend({
    serviceName: 'oozie',
    templateName: require('templates/main/dashboard/service/oozie'),

    webUi: function () {
      var hostName = App.singleNodeInstall ? App.singleNodeAlias : this.get('service.hostComponents').findProperty('componentName', 'OOZIE_SERVER').get('host.publicHostName');
      return "http://{0}:11000/oozie".format(hostName);
    }.property('service')
  });
});
window.require.register("views/main/dashboard/service/storm", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.MainDashboardServiceStormView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/storm'),
    serviceName: 'storm',

    superVisorComponents: function() {
      return this.get('service.hostComponents').filterProperty('componentName', 'SUPERVISOR');
    }.property('service.hostComponents.@each'),

    superVisorsLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function () {
        return this.get('parentView.superVisorComponents').filterProperty('workStatus','STARTED').length;
      }.property('parentView.superVisorComponents.@each.workStatus'),
      totalComponents: function() {
        return this.get('parentView.superVisorComponents').length;
      }.property('parentView.superVisorComponents.length')
    }),

    freeSlotsPercentage: function() {
      return Math.round(this.get('service.freeSlots')/this.get('service.totalSlots')*100);
    }.property('service.freeSlots', 'service.totalSlots'),

    nimbusUptimeFormatted: function() {
      if (this.get('service.nimbusUptime') > 0) {
        return date.timingFormat(this.get('service.nimbusUptime')*1000);
      }
      return Em.I18n.t('services.service.summary.notRunning');
    }.property('service.nimbusUptime')
  });
  
});
window.require.register("views/main/dashboard/service/yarn", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');
  var numberUtils = require('utils/number_utils');

  App.MainDashboardServiceYARNView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/yarn'),
    serviceName: 'YARN',

    nodeHeap: function () {
      var memUsed = this.get('service').get('jvmMemoryHeapUsed');
      var memMax = this.get('service').get('jvmMemoryHeapMax');
      var percent = memMax > 0 ? ((100 * memUsed) / memMax) : 0;
      return this.t('dashboard.services.hdfs.nodes.heapUsed').format(
        numberUtils.bytesToSize(memUsed, 1, 'parseFloat'),
        numberUtils.bytesToSize(memMax, 1, 'parseFloat'),
        percent.toFixed(1));
    }.property('service.jvmMemoryHeapUsed', 'service.jvmMemoryHeapMax'),

    summaryHeader: function () {
      var text = this.t("dashboard.services.yarn.summary");
      var svc = this.get('service');
      var totalCount = svc.get('nodeManagerNodes').get('length');
      return text.format(totalCount, totalCount);
    }.property('service.nodeManagerNodes'),
    
    nodeManagerComponent: function () {
      return this.get('service.nodeManagerNodes').objectAt(0);
    }.property(),
    
    yarnClientComponent: function () {
      return this.get('service.hostComponents').findProperty('componentName', 'YARN_CLIENT');
    }.property(),

    hasManyYarnClients: function () {
      return (this.get('service.yarnClientNodes.length') > 1);
    }.property('service.yarnClientNodes.length'),

    nodeUptime: function () {
      var uptime = this.get('service').get('resourceManagerStartTime');
      if (uptime && uptime > 0){
        var diff = App.dateTime() - uptime;
        if (diff < 0) {
          diff = 0;
        }
        var formatted = date.timingFormat(diff);
        return this.t('dashboard.services.uptime').format(formatted);
      }
      return this.t('services.service.summary.notRunning');
    }.property("service.resourceManagerStartTime"),

    nodeManagerText: function () {
      if(!this.get("nodeManagerComponent") || this.get("nodeManagerComponent.length") == 0){
        return '';
      } else if(this.get("nodeManagerComponent.length") > 1){
        return Em.I18n.t('services.service.summary.viewHosts');
      } else {
        return Em.I18n.t('services.service.summary.viewHost');
      }
    }.property("nodeManagerComponent"),

    nodeManagersLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function () {
        return this.get('service.nodeManagerNodes').filterProperty("workStatus", "STARTED").get("length");
      }.property("service.nodeManagerNodes.@each.workStatus"),
      totalComponents: function () {
        return this.get('service.nodeManagerNodes.length');
      }.property("service.nodeManagerNodes.length")
    }),

    nodeManagersStatus: function () {
      var nmActive = this.get('service.nodeManagersCountActive');
      var nmLost = this.get('service.nodeManagersCountLost');
      var nmUnhealthy = this.get('service.nodeManagersCountUnhealthy');
      var nmRebooted = this.get('service.nodeManagersCountRebooted');
      var nmDecom = this.get('service.nodeManagersCountDecommissioned');
      return this.t('dashboard.services.yarn.nodeManagers.status.msg').format(
        this.formatUnavailable(nmActive),
        this.formatUnavailable(nmLost),
        this.formatUnavailable(nmUnhealthy),
        this.formatUnavailable(nmRebooted),
        this.formatUnavailable(nmDecom)
      );
    }.property('service.nodeManagersCountActive', 'service.nodeManagersCountLost', 
        'service.nodeManagersCountUnhealthy', 'service.nodeManagersCountRebooted', 'service.nodeManagersCountDecommissioned'),

    containers: function () {
      var allocated = this.get('service.containersAllocated');
      var pending = this.get('service.containersPending');
      var reserved = this.get('service.containersReserved');
      return this.t('dashboard.services.yarn.containers.msg').format(
        this.formatUnavailable(allocated),
        this.formatUnavailable(pending),
        this.formatUnavailable(reserved)
      );
    }.property('service.containersAllocated', 'service.containersPending', 'service.containersReserved'),

    apps: function () {
      var appsSubmitted = this.get('service.appsSubmitted');
      var appsRunning = this.get('service.appsRunning');
      var appsPending = this.get('service.appsPending');
      var appsCompleted = this.get('service.appsCompleted');
      var appsKilled = this.get('service.appsKilled');
      var appsFailed = this.get('service.appsFailed');
      return this.t('dashboard.services.yarn.apps.msg').format(
        this.formatUnavailable(appsSubmitted),
        this.formatUnavailable(appsRunning),
        this.formatUnavailable(appsPending),
        this.formatUnavailable(appsCompleted),
        this.formatUnavailable(appsKilled),
        this.formatUnavailable(appsFailed));
    }.property('service.appsSubmitted', 'service.appsRunning', 'service.appsPending', 'service.appsCompleted', 'service.appsKilled', 'service.appsFailed'),

    memory: function () {
      return Em.I18n.t('dashboard.services.yarn.memory.msg').format(
          numberUtils.bytesToSize(this.get('service.allocatedMemory'), 1, 'parseFloat', 1024 * 1024), 
          numberUtils.bytesToSize(this.get('service.reservedMemory'), 1, 'parseFloat', 1024 * 1024), 
          numberUtils.bytesToSize(this.get('service.availableMemory'), 1, 'parseFloat', 1024 * 1024));
    }.property('service.allocatedMemory', 'service.reservedMemory', 'service.availableMemory'),

    queues: function() {
      return Em.I18n.t('dashboard.services.yarn.queues.msg').format(this.formatUnavailable(this.get('service.queuesCount')));
    }.property('service.queuesCount'),
    
    didInsertElement: function(){
      App.tooltip($("[rel='queue-tooltip']"), {html: true, placement: "right"});
    }

  });
  
});
window.require.register("views/main/dashboard/service/zookeeper", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainDashboardServiceZookeperView = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/dashboard/service/zookeeper'),
    serviceName: 'zookeeper',

    titleInfo: function(){
      var components = this.get('service.hostComponents').filterProperty('componentName', 'ZOOKEEPER_SERVER');
      var running = 0;
      components.forEach(function(item){
        if(item.get('workStatus') === App.HostComponentStatus.started){
          running += 1;
        }
      });

      return {
        pre: this.t('services.zookeeper.prefix').format(running),
        title: this.t('services.zookeeper.title').format(components.length),
        component: components.objectAt(0)
      };
    }.property('service')
  });
});
window.require.register("views/main/dashboard/widget", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.DashboardWidgetView = Em.View.extend({

    title: null,
    templateName: null, // each has specific template

    /**
     * Setup model for widget by `model_type`. Usually `model_type` is a lowercase service name,
     * for example `hdfs`, `yarn`, etc. You need to set `model_type` in extended object View, for example
     * look App.DataNodeUpView.
     * @return {Object} - model that set up in App.MainDashboardView.setWidgetsDataModel()
     */
    model : function () {
      if (!this.get('model_type')) return {};
      return this.get('parentView').get(this.get('model_type') + '_model');
    }.property(), //data bind from parent view

    id: null, // id 1-10 used to identify
    viewID: function(){ // used by re-sort
      return 'widget-' + this.get('id');
    }.property('id'),  //html id bind to view-class: widget-(1)
    attributeBindings: ['viewID'],

    isPieChart: false,
    isText: false,
    isProgressBar: false,
    isLinks: false,
    content: null, // widget content pieChart/ text/ progress bar/links/ metrics. etc
    hiddenInfo: null, // more info details
    hiddenInfoClass: "hidden-info-two-line",

    thresh1: null, //@type {Number}
    thresh2: null, //@type {Number}

    didInsertElement: function () {
      App.tooltip(this.$("[rel='ZoomInTooltip']"), {placement : 'left'});
    },

    deleteWidget: function (event) {
      var parent = this.get('parentView');
      if (App.testMode) {
        //update view on dashboard
        var objClass = parent.widgetsMapper(this.id);
        parent.get('visibleWidgets').removeObject(objClass);
        parent.get('hiddenWidgets').pushObject(Em.Object.create({displayName: this.get('title'), id: this.get('id'), checked: false}));
      } else {
        //reconstruct new persist value then post in persist
        parent.getUserPref(parent.get('persistKey'));
        var oldValue = parent.get('currentPrefObject');
        var deletedId = this.get('id');
        var newValue = Em.Object.create({
          dashboardVersion: oldValue.dashboardVersion,
          visible: [],
          hidden: oldValue.hidden,
          threshold: oldValue.threshold
        });
        for (var i = 0; i <= oldValue.visible.length - 1; i++) {
          if (oldValue.visible[i] != deletedId) {
            newValue.visible.push(oldValue.visible[i]);
          }
        }
        newValue.hidden.push([deletedId, this.get('title')]);
        parent.postUserPref(parent.get('persistKey'), newValue);
        parent.translateToReal(newValue);
      }
    },

    editWidget: function (event) {
      var self = this;
      var max_tmp =  parseFloat(self.get('maxValue'));
      var configObj = Ember.Object.create({
        thresh1: self.get('thresh1') + '',
        thresh2: self.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.common').format(max_tmp),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: max_tmp,
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else if (this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }

          // update the slider handles and color
          if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
            $("#slider-range").slider('values', 0 , parseFloat(thresh1));
            $("#slider-range").slider('values', 1 , parseFloat(thresh2));
          }
        }.observes('thresh1', 'thresh2')

      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget' ],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function() {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            self.set('thresh1', parseFloat(configObj.get('thresh1')) );
            self.set('thresh2', parseFloat(configObj.get('thresh2')) );

            if (!App.testMode) {
              // save to persist
              var parent = self.get('parentView');
              parent.getUserPref(parent.get('persistKey'));
              var oldValue = parent.get('currentPrefObject');
              oldValue.threshold[parseInt(self.get('id'))] = [configObj.get('thresh1'), configObj.get('thresh2')];
              parent.postUserPref(parent.get('persistKey'), oldValue);
            }

            this.hide();
          }
        },

        didInsertElement: function () {
          var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
          var colors = ['#95A800', '#FF8E00', '#B80000']; //color green, orange ,red

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', true);
            $("#slider-range").slider({
              range: true,
              min: 0,
              max: max_tmp,
              values: handlers,
              create: function (event, ui) {
                updateColors(handlers);
              },
              slide: function (event, ui) {
                updateColors(ui.values);
                configObj.set('thresh1', ui.values[0] + '');
                configObj.set('thresh2', ui.values[1] + '');
              },
              change: function (event, ui) {
                updateColors(ui.values);
              }
            });

            function updateColors(handlers) {
              var colorstops = colors[0] + ", "; // start with the first color
              for (var i = 0; i < handlers.length; i++) {
                colorstops += colors[i] + " " + handlers[i]*100/max_tmp + "%,";
                colorstops += colors[i+1] + " " + handlers[i]*100/max_tmp + "%,";
              }
              colorstops += colors[colors.length - 1];
              var sliderElement = $('#slider-range');
              var css1 = '-webkit-linear-gradient(left,' + colorstops + ')'; // chrome & safari
              sliderElement.css('background-image', css1);
              var css2 = '-ms-linear-gradient(left,' + colorstops + ')'; // IE 10+
              sliderElement.css('background-image', css2);
              //$('#slider-range').css('filter', 'progid:DXImageTransform.Microsoft.gradient( startColorStr= ' + colors[0] + ', endColorStr= ' + colors[2] +',  GradientType=1 )' ); // IE 10-
              var css3 = '-moz-linear-gradient(left,' + colorstops + ')'; // Firefox
              sliderElement.css('background-image', css3);

              sliderElement.find('.ui-widget-header').css({'background-color': '#FF8E00', 'background-image': 'none'}); // change the  original ranger color
            }
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', true);
          }
        }
      });
    },

    getInternetExplorerVersion: function (){
      var rv = -1; //return -1 for other browsers
      if (navigator.appName == 'Microsoft Internet Explorer') {
        var ua = navigator.userAgent;
        var re  = new RegExp("MSIE ([0-9]{1,}[\.0-9]{0,})");
        if (re.exec(ua) != null)
          rv = parseFloat( RegExp.$1 ); // IE version 1-10
      }
      var isFirefox = typeof InstallTrigger !== 'undefined';   // Firefox 1.0+
      if (isFirefox) {
        return -2;
      }else{
        return rv;
      }
    },

    /**
     * for widgets has hidden info(hover info),
     * calculate the hover content top number
     * based on how long the hiddenInfo is
     */
    hoverContentTopClass: function () {
      var lineNum = this.get('hiddenInfo.length');
      if (lineNum == 2) {
        return "content-hidden-two-line";
      } else if (lineNum == 3) {
        return "content-hidden-three-line";
      } else if (lineNum == 4) {
        return "content-hidden-four-line";
      } else if (lineNum == 5) {
        return "content-hidden-five-line";
      } else if (lineNum == 6) {
        return "content-hidden-six-line";
      }
      return '';
    }.property('hiddenInfo.length')

  });


  App.DashboardWidgetView.reopenClass({
    class: 'span2p4'
  });
  
});
window.require.register("views/main/dashboard/widgets/cluster_metrics_widget", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ClusterMetricsDashboardWidgetView = App.DashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/cluster_metrics')

  });
  
});
window.require.register("views/main/dashboard/widgets/datanode_live", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.DataNodeUpView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.DataNodeUp'),
    id: '4',

    isPieChart: false,
    isText: true,
    isProgressBar: false,
    model_type: 'hdfs',

    hiddenInfo: function () {
      var result = [];
      result.pushObject(this.get('dataNodesLive').length + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.live'));
      result.pushObject(this.get('dataNodesDead').length + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.dead'));
      result.pushObject(this.get('model.decommissionDataNodes.length')+ ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.decom'));
      return result;
    }.property('dataNodesLive', 'dataNodesDead', 'model.decommissionDataNodes.length'),
    hiddenInfoClass: "hidden-info-three-line",

    thresh1: 40,
    thresh2: 70,
    maxValue: 100,

    dataNodesLive: function () {
      return this.get('model.dataNodes').filterProperty("workStatus", "STARTED");
    }.property('model.dataNodes.@each.workStatus'),
    dataNodesDead: function () {
      return this.get('model.dataNodes').filterProperty("workStatus", "INSTALLED");
    }.property('model.dataNodes.@each.workStatus'),

    data: function () {
      if ( !this.get('model.dataNodes.length')) {
        return -1;
      } else {
        return ((this.get('dataNodesLive').length / this.get('model.dataNodes.length')).toFixed(2)) * 100;
      }
    }.property('model.dataNodes.length', 'dataNodesLive'),

    content: function () {
      return this.get('dataNodesLive').length + "/" + this.get('model.dataNodes.length');
    }.property('model.dataNodes.length', 'dataNodesLive'),

    editWidget: function (event) {
      var parent = this;
      var max_tmp =  parseFloat(parent.get('maxValue'));
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: max_tmp,
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }

          // update the slider handles and color
          if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
            $("#slider-range").slider('values', 0 , parseFloat(thresh1));
            $("#slider-range").slider('values', 1 , parseFloat(thresh2));
          }
        }.observes('thresh1', 'thresh2')

      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function () {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }
            this.hide();
          }
        },

        didInsertElement: function () {
          var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
          var colors = ['#B80000', '#FF8E00', '#95A800']; //color red, orange, green

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', false);
            $("#slider-range").slider({
              range: true,
              min: 0,
              max: max_tmp,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              },
              slide: function (event, ui) {
                parent.updateColors(ui.values, colors);
                configObj.set('thresh1', ui.values[0] + '');
                configObj.set('thresh2', ui.values[1] + '');
              },
              change: function (event, ui) {
                parent.updateColors(ui.values, colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', false);
          }
        }
      });
    }
  });
  
});
window.require.register("views/main/dashboard/widgets/hbase_average_load", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HBaseAverageLoadView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.HBaseAverageLoad'),
    id: '21',

    model_type: 'hbase',
    hiddenInfo: function () {
      var avgLoad = this.get('model.averageLoad');
      if (isNaN(avgLoad)) {
        avgLoad = Em.I18n.t('services.service.summary.notAvailable');
      }
      var result = [];
      result.pushObject(Em.I18n.t('dashboard.services.hbase.averageLoadPerServer').format(avgLoad));
      return result;
    }.property("model.averageLoad"),

    isGreen: function () {
      return this.get('data') <= this.get('thresh1');
    }.property('data','thresh1','thresh2'),
    isRed: function () {
      return this.get('data') > this.get('thresh2');
    }.property('data','thresh1','thresh2'),
    isNA: function (){
      return this.get('data') === null || isNaN(this.get('data'));
    }.property('data'),

    thresh1: 0.5,
    thresh2: 2,
    maxValue: 'infinity',

    data: function () {
      return this.get('model.averageLoad');
    }.property("model.averageLoad"),

    content: function (){
      if(this.get('data') || this.get('data') == 0){
        return this.get('data') + "";
      }else{
        return Em.I18n.t('services.service.summary.notAvailable');
      }
    }.property('model.averageLoad'),

    editWidget: function (event) {
      var parent = this;
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint2'),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: 'infinity',
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
            } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }

        }.observes('thresh1', 'thresh2')

      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show( {
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function () {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              //save to persist
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }

            this.hide();
          }
        },

        didInsertElement: function () {
          var colors = ['#95A800', '#FF8E00', '#B80000']; //color green, orange ,red
          var handlers = [33, 66]; //fixed value

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', true);
            $("#slider-range").slider({
              range:true,
              disabled:true, //handlers cannot move
              min: 0,
              max: 100,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', true);
          }
        }
      });
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/hbase_links", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HBaseLinksView = App.LinkDashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/hbase_links'),
    title: Em.I18n.t('dashboard.widgets.HBaseLinks'),
    id: '19',

    model_type: 'hbase',

    port: '60010',

    componentName: 'HBASE_REGIONSERVER',

    /**
     * All master components
     */
    masters: function () {
      return this.get('model.hostComponents').filterProperty('isMaster', true);
    }.property('model.hostComponents.@each'),
    /**
     * Passive master components
     */
    passiveMasters: function () {
      if (App.supports.multipleHBaseMasters) {
        return this.get('masters').filterProperty('haStatus', 'false');
      }
      return [];
    }.property('masters'),
    /**
     * One(!) active master component
     */
    activeMaster: function () {
      if(App.supports.multipleHBaseMasters) {
        return this.get('masters').findProperty('haStatus', 'true');
      } else {
        return this.get('masters')[0];
      }
    }.property('masters'),

    activeMasterTitle: function(){
      if (App.supports.multipleHBaseMasters) {
        return this.t('service.hbase.activeMaster');
      } else {
        return this.get('activeMaster.host.publicHostName');
      }
    }.property('activeMaster'),

    hbaseMasterWebUrl: function () {
      if (this.get('activeMaster.host') && this.get('activeMaster.host').get('publicHostName')) {
        return "http://" + this.get('activeMaster.host').get('publicHostName') + ':' + this.get('port');
      }
      return '';
    }.property('activeMaster'),

    calcWebUrl: function() {
      return '';
    }

  });
});
window.require.register("views/main/dashboard/widgets/hbase_master_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.HBaseMasterHeapPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.HBaseMasterHeap'),
    id: '20',

    model_type: 'hbase',
    modelFieldMax: 'heapMemoryMax',
    modelFieldUsed: 'heapMemoryUsed',
    widgetHtmlId: 'widget-hbase-heap',

    didInsertElement: function() {
      this._super();
      this.calc();
    },

    getUsed: function() {
      return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
    },

    getMax: function() {
      return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
    }
  });
});
window.require.register("views/main/dashboard/widgets/hbase_master_uptime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HBaseMasterUptimeView = App.UptimeTextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.HBaseMasterUptime'),
    id: '23',

    model_type: 'hbase',

    component: 'Hbase Master',
    modelField: 'masterStartTime',

    didInsertElement: function() {
      this._super();
      this.calc();
    }

  });
});
window.require.register("views/main/dashboard/widgets/hbase_regions_in_transition", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HBaseRegionsInTransitionView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.HBaseRegionsInTransition'),
    id: '22',

    model_type: 'hbase',
    hiddenInfo: function () {
      var result = [];
      result.pushObject(this.get("model.regionsInTransition") + " regions");
      result.pushObject("in transition");
      return result;
    }.property("model.regionsInTransition"),

    classNameBindings: ['isRed', 'isOrange', 'isGreen', 'isNA'],
    isGreen: function () {
      return this.get('data') <= this.get('thresh1');
    }.property('data','thresh1'),
    isOrange: function () {
      return (this.get('data') <= this.get('thresh2') && this.get('data') > this.get('thresh1') );
    }.property('data','thresh1','thresh2'),
    isRed: function () {
      return this.get('data') > this.get('thresh2');
    }.property('data','thresh2'),
    isNA: function () {
      return this.get('data') === null;
    }.property('data'),

    thresh1: 0.5,
    thresh2: 2,
    maxValue: 'infinity',

    data: function () {
      return this.get('model.regionsInTransition');
    }.property("model.regionsInTransition"),

    content: function (){
      return this.get('data') + "";
    }.property('model.regionsInTransition'),

    editWidget: function (event) {
      var parent = this;
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint2'),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: 'infinity',
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
            } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }
        }.observes('thresh1', 'thresh2')

      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function () {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              //save to persist
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }

            this.hide();
          }
        },

        didInsertElement: function () {
          var colors = ['#95A800', '#FF8E00', '#B80000']; //color green, orange ,red
          var handlers = [33, 66]; //fixed value

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', true);
            $("#slider-range").slider({
              range:true,
              disabled:true, //handlers cannot move
              min: 0,
              max: 100,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', true);
          }
        }
      });
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/hdfs_capacity", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.NameNodeCapacityPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.HDFSDiskUsage'),
    id: '2',

    model_type: 'hdfs',
    modelFieldMax: 'capacityTotal',
    /**
     * HDFS model has 'remaining' value, but not 'used'
     */
    modelFieldUsed: 'capacityRemaining',
    widgetHtmlId: 'widget-nn-capacity',
    hiddenInfoClass: "hidden-info-six-line",

    didInsertElement: function() {
      this._super();
      this.calc();
    },

    calcHiddenInfo: function () {
      var total = this.get('model').get(this.get('modelFieldMax')) + 0;
      var remaining = this.get('model.capacityRemaining');
      var dfsUsed = this.get('model.capacityUsed');
      var nonDfsUsed = total !== null && remaining !== null && dfsUsed !== null ? total - remaining - dfsUsed : null;
      var dfsPercent = total > 0 ? ((dfsUsed * 100) / total).toFixed(2) : 0;
      var nonDfsPercent = total > 0 ? ((nonDfsUsed * 100) / total).toFixed(2) : 0;
      var remainingPercent = total > 0 ? ((remaining * 100) / total).toFixed(2) : 0;
      if (dfsPercent == "NaN" || dfsPercent < 0) {
        dfsPercent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      if (nonDfsPercent == "NaN" || nonDfsPercent < 0) {
        nonDfsPercent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      if (remainingPercent == "NaN" || remainingPercent < 0) {
        remainingPercent = Em.I18n.t('services.service.summary.notAvailable') + " ";
      }
      var result = [];
      result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.DFSused'));
      result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.info').format(numberUtils.bytesToSize(dfsUsed, 1, 'parseFloat'), dfsPercent));
      result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.nonDFSused'));
      result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.info').format(numberUtils.bytesToSize(nonDfsUsed, 1, 'parseFloat'), nonDfsPercent));
      result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.remaining'));
      result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.info').format(numberUtils.bytesToSize(remaining, 1, 'parseFloat'), remainingPercent));
      return result;
    },

    calcDataForPieChart: function() {
      var total = this.get('model').get(this.get('modelFieldMax')) * 1024 * 1024;
      var used = total - this.get('model').get(this.get('modelFieldUsed')) * 1024 * 1024;
      var percent = total > 0 ? ((used)*100 / total).toFixed() : 0;
      var percent_precise = total > 0 ? ((used)*100 / total).toFixed(1) : 0;
      return [percent, percent_precise];
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/hdfs_links", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HDFSLinksView = App.LinkDashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/hdfs_links'),
    title: Em.I18n.t('dashboard.widgets.HDFSLinks'),
    id: '17',

    model_type: 'hdfs',

    port: '50070',

    componentName: 'DATANODE',

    modelField: 'nameNode',

    didInsertElement: function() {
      this._super();
      this.calc();
    },

    isHAEnabled: function() {
      return !this.get('model.snameNode');
    }.property('model.snameNode'),
    isActiveNNValid: function () {
      return this.get('model.activeNameNode') != null;
    }.property('model.activeNameNode'),
    isStandbyNNValid: function () {
      return this.get('model.standbyNameNode') != null;
    }.property('model.standbyNameNode'),
    isTwoStandbyNN: function () {
      return (this.get('model.standbyNameNode') != null && this.get('model.standbyNameNode2') != null);
    }.property('model.standbyNameNode', 'model.standbyNameNode2'),
    twoStandbyComponent: function () {
      return App.HostComponent.find().findProperty('componentName', 'NAMENODE');
    }.property()
  });
});
window.require.register("views/main/dashboard/widgets/jobtracker_cpu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.JobTrackerCpuPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.JobTrackerCpu'),
    id: '7',

    model_type: 'mapreduce',

    widgetHtmlId: 'widget-jt-cpu',

    modelFieldUsed: 'jobTrackerCpu',

    didInsertElement: function() {
      this._super();
      this.calc();
    },

    calcHiddenInfo: function() {
      var value = this.get('model').get(this.get('modelFieldUsed'));
      var obj1;
      if( value == null) {
        obj1 = Em.I18n.t('services.service.summary.notAvailable');
      }
      else {
        value = value >= 100 ? 100: value;
        obj1 = (value + 0).toFixed(2) + '%';
      }
      var result = [];
      result.pushObject(obj1);
      result.pushObject('CPU wait I/O');
      return result;
    },

    calcIsPieExists: function() {
      return (this.get('model').get(this.get('modelFieldUsed')) != null);
    },

    calcDataForPieChart: function() {
      var value = this.get('model').get(this.get('modelFieldUsed'));
      value = value >= 100 ? 100: value;
      var percent = (value + 0).toFixed(1);
      var percent_precise = (value + 0).toFixed(2);
      return [ percent, percent_precise];
    }

  });
});
window.require.register("views/main/dashboard/widgets/jobtracker_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.JobTrackerHeapPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.JobTrackerHeap'),
    id: '6',

    model_type: 'mapreduce',
    modelFieldMax: 'jobTrackerHeapMax',
    modelFieldUsed: 'jobTrackerHeapUsed',

    widgetHtmlId: 'widget-jt-heap',

    getUsed: function() {
      return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
    },

    getMax: function() {
      return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
    },

    didInsertElement: function() {
      this._super();
      this.calc();
    }
  });
});
window.require.register("views/main/dashboard/widgets/jobtracker_rpc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.JobTrackerRpcView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.JobTrackerRpc'),
    id:'9',

    model_type: 'mapreduce',
    hiddenInfo: function (){
      var result = [];
      result.pushObject(this.get('content') + ' average RPC');
      result.pushObject('queue wait time');
      return result;
    }.property('model.jobTrackerRpc'),

    thresh1: 0.5,
    thresh2: 2,
    maxValue: 'infinity',

    isGreen: function () {
      return (this.get('data') <= this.get('thresh1'));
    }.property('data','thresh1'),

    isRed: function () {
      return (this.get('data') > this.get('thresh2'));
    }.property('data','thresh2'),

    data: function (){
      if (this.get('model.jobTrackerRpc')) {
        return (this.get('model.jobTrackerRpc')).toFixed(2);
      } else {
        if (this.get('model.jobTrackerRpc') == 0) {
          return 0;
        } else {
          return null;
        }
      }
    }.property('model.jobTrackerRpc'),

    content: function () {
      if (this.get('data') || this.get('data') == 0) {
        return this.get('data') + " ms";
      } else {
        return Em.I18n.t('services.service.summary.notAvailable');
      }
    }.property('model.jobTrackerRpc'),

    editWidget: function (event) {
      var parent = this;
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint3'),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: 'infinity',
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
            } else if (this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else{
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }
        }.observes('thresh1', 'thresh2')
      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function () {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              // save to persist
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }

            this.hide();
          }
        },

        didInsertElement: function () {
          var colors = ['#95A800', '#FF8E00', '#B80000']; //color green, orange ,red
          var handlers = [33, 66]; //fixed value

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', true);
            $("#slider-range").slider({
              range:true,
              disabled:true,
              min: 0,
              max: 100,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', true);
          }
        }
      });
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/jobtracker_uptime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.JobTrackerUptimeView = App.UptimeTextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.JobTrackerUptime'),
    id: '16',

    model_type: 'mapreduce',

    component: 'JobTracker',
    modelField: 'jobTrackerStartTime',

    didInsertElement: function() {
      this._super();
      this.calc();
    }

  });
});
window.require.register("views/main/dashboard/widgets/links_widget", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.LinkDashboardWidgetView = App.DashboardWidgetView.extend({

    componentName: null,

    component: function() {
      return App.HostComponent.find().findProperty('componentName', this.get('componentName'));
    }.property(),

    port: null,

    modelField: null,

    webUrl: null,

    didInsertElement: function() {
      this._super();
      this.addObserver('model.' + this.get('modelField'), this, this.calc);
    },
    calc: function() {
      this.set('webUrl', this.calcWebUrl());
    },
    calcWebUrl: function() {
      if (this.get('model')) {
        return "http://" + this.get('model').get(this.get('modelField')).get('publicHostName') + ':' + this.get('port');
      }
      return '';
    }
  });
  
});
window.require.register("views/main/dashboard/widgets/mapreduce_links", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MapReduceLinksView = App.LinkDashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/mapreduce_links'),
    title: Em.I18n.t('dashboard.widgets.MapReduceLinks'),
    id: '18',

    model_type: 'mapreduce',

    port: '50030',

    componentName: 'TASKTRACKER',

    modelField: 'jobTracker',

    didInsertElement: function() {
      this._super();
      this.calc();
    }

  });
});
window.require.register("views/main/dashboard/widgets/mapreduce_slots", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MapReduceSlotsView = App.DashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/mapreduce_slots'),
    title: Em.I18n.t('dashboard.widgets.MapReduceSlots'),
    id:'10',

    isProgressBar: true,
    model_type: 'mapreduce',
    hiddenInfo: function () {
      var result = [];
      if(this.get('isViewExist')) {
        var line1 = "Map: " + this.get('model.mapSlotsOccupied') + " Occupied / " + this.get('model.mapSlotsReserved') + " Reserved / " + this.get('model.mapSlots') + " Total";
        result.pushObject(line1);
        var line2 = "Reduce: " + this.get('model.reduceSlotsOccupied') + " Occupied / " + this.get('model.reduceSlotsReserved') + " Reserved / " + this.get('model.reduceSlots') + " Total";
        result.pushObject(line2);
      }
      else {
        result.pushObject('MapReduce Not Started');
      }
      return result;
    }.property('isViewExist', 'map_display_text', 'reduce_display_text'),
    isViewExist: function () {
      return this.get('model.mapSlotsOccupied') != null && this.get('model.mapSlotsReserved') != null && this.get('model.reduceSlotsOccupied') != null && this.get('model.reduceSlotsReserved') != null;
    }.property('model.mapSlotsReserved', 'model.mapSlotsOccupied', 'model.reduceSlotsReserved', 'model.reduceSlotsOccupied'),

    map_occupied: function () {
      if (this.get('model.mapSlotsOccupied')) {
        return "width: " + ((this.get('model.mapSlotsOccupied'))*100/(this.get('model.mapSlots'))).toString() + "%";
      } else {
        return "width: 0%";
      }
    }.property('model.mapSlotsOccupied','model.mapSlots'),
    map_reserved: function () {
      if (this.get('model.mapSlotsReserved')) {
        return "width: " + ((this.get('model.mapSlotsReserved'))*100/(this.get('model.mapSlots'))).toString() + "%";
      } else {
        return "width: 0%";
      }
    }.property('model.mapSlotsReserved','model.mapSlots'),
    map_display_text: function () {
      return this.get('model.mapSlotsOccupied') + "/" + this.get('model.mapSlotsReserved') + "/" + this.get('model.mapSlots');
    }.property('model.mapSlotsReserved','model.mapSlotsOccupied','model.mapSlots'),


    reduce_occupied: function () {
      if (this.get('model.reduceSlotsOccupied')) {
        return "width: " + ((this.get('model.reduceSlotsOccupied'))*100/(this.get('model.reduceSlots'))).toString() + "%";
      } else {
        return "width: 0%";
      }
    }.property('model.reduceSlotsOccupied','model.reduceSlots'),
    reduce_reserved: function () {
      if (this.get('model.reduceSlotsReserved')) {
        return "width: " + ((this.get('model.reduceSlotsReserved'))*100/(this.get('model.reduceSlots'))).toString() + "%";
      } else {
        return "width: 0%";
      }
    }.property('model.reduceSlotsReserved','model.reduceSlots'),
    reduce_display_text: function () {
      return this.get('model.reduceSlotsOccupied') + "/" + this.get('model.reduceSlotsReserved') + "/" + this.get('model.reduceSlots');
    }.property('model.reduceSlotsReserved','model.reduceSlotsOccupied','model.reduceSlots')

  });

  App.MapReduceSlotsView.reopenClass({
    class: 'span4p8'
  });
});
window.require.register("views/main/dashboard/widgets/metrics_cpu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartClusterMetricsCPUWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.clusterMetrics.cpu'),
    id: '13',

    content: App.ChartClusterMetricsCPU.extend({
      noTitleUnderGraph: true,
      inWidget: true
    })
  });
});
window.require.register("views/main/dashboard/widgets/metrics_load", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartClusterMetricsLoadWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.clusterMetrics.load'),
    id: '14',

    content: App.ChartClusterMetricsLoad.extend({
      noTitleUnderGraph: true,
      inWidget: true
    })
  });
});
window.require.register("views/main/dashboard/widgets/metrics_memory", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartClusterMetricsMemoryWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.clusterMetrics.memory'),
    id: '11',

    content: App.ChartClusterMetricsMemory.extend({
      noTitleUnderGraph: true,
      inWidget: true
    })
  });
});
window.require.register("views/main/dashboard/widgets/metrics_network", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.ChartClusterMetricsNetworkWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.clusterMetrics.network'),
    id: '12',

    content: App.ChartClusterMetricsNetwork.extend({
      noTitleUnderGraph: true,
      inWidget: true
    })
  });
});
window.require.register("views/main/dashboard/widgets/namenode_cpu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.NameNodeCpuPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.NameNodeCpu'),
    id: '3',

    model_type: 'hdfs',
    widgetHtmlId: 'widget-nn-cpu',
    modelFieldUsed: 'nameNodeCpu',

    didInsertElement: function() {
      this._super();
      this.calc();
    },

    calcHiddenInfo: function() {
      var value = this.get('model').get(this.get('modelFieldUsed'));
      var obj1;
      if( value == null) {
        obj1 = Em.I18n.t('services.service.summary.notAvailable');
      }
      else {
        value = value >= 100 ? 100: value;
        obj1 = (value + 0).toFixed(2) + '%';
      }
      var result = [];
      result.pushObject(obj1);
      result.pushObject('CPU wait I/O');
      return result;
    },

    calcIsPieExists: function() {
      return (this.get('model').get(this.get('modelFieldUsed')) != null);
    },

    calcDataForPieChart: function() {
      var value = this.get('model').get(this.get('modelFieldUsed'));
      value = value >= 100 ? 100: value;
      var percent = (value + 0).toFixed(1);
      var percent_precise = (value + 0).toFixed(2);
      return [ percent, percent_precise];
    }
  });
});
window.require.register("views/main/dashboard/widgets/namenode_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.NameNodeHeapPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.NameNodeHeap'),
    id: '1',

    model_type: 'hdfs',
    modelFieldMax: 'jvmMemoryHeapMax',
    modelFieldUsed: 'jvmMemoryHeapUsed',
    widgetHtmlId: 'widget-nn-heap',

    getUsed: function() {
      return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
    },

    getMax: function() {
      return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
    },

    didInsertElement: function() {
      this._super();
      this.calc();
    }
  });
});
window.require.register("views/main/dashboard/widgets/namenode_rpc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.NameNodeRpcView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.NameNodeRpc'),
    id: '5',

    model_type: 'hdfs',
    hiddenInfo: function () {
      var result = [];
      result.pushObject(this.get('content') + ' average RPC');
      result.pushObject('queue wait time');
      return result;
    }.property('model.nameNodeRpc'),

    thresh1: 0.5,
    thresh2: 2,
    maxValue: 'infinity',

    isGreen: function () {
      return (this.get('data') <= this.get('thresh1'));
    }.property('data','thresh1'),

    isRed: function () {
      return (this.get('data') > this.get('thresh2'));
    }.property('data','thresh2'),

    data: function () {
      if (this.get('model.nameNodeRpc')) {
        return (this.get('model.nameNodeRpc')).toFixed(2);
      } else {
        if (this.get('model.nameNodeRpc') == 0) {
          return 0;
        } else {
          return null;
        }
      }
    }.property('model.nameNodeRpc'),

    content: function () {
      if (this.get('data') || this.get('data') == 0) {
        return this.get('data') + " ms";
      }
      else {
        return Em.I18n.t('services.service.summary.notAvailable');
      }
    }.property('model.nameNodeRpc'),

    editWidget: function (event) {
      var parent = this;
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint3'),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: 'infinity',
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
            } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }
        }.observes('thresh1', 'thresh2')

      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: ['sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function () {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              //save to persist
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }

            this.hide();
          }
        },

        didInsertElement: function () {
          var colors = ['#95A800', '#FF8E00', '#B80000']; //color green, orange ,red
          var handlers = [33, 66]; //fixed value

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', true);
            $("#slider-range").slider({
              range:true,
              disabled:true, //handlers cannot move
              min: 0,
              max: 100,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', true);
          }
        }
      });
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/namenode_uptime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.NameNodeUptimeView = App.UptimeTextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.NameNodeUptime'),
    id: '15',

    model_type: 'hdfs',

    component: 'NameNode',
    modelField: 'nameNodeStartTime',

    didInsertElement: function() {
      this._super();
      this.calc();
    }

  });
});
window.require.register("views/main/dashboard/widgets/node_managers_live", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.NodeManagersLiveView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.NodeManagersLive'),
    id: '26',

    model_type: 'yarn',

    hiddenInfo: function () {
      var nmActive = this.get('model.nodeManagersCountActive') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountActive');
      var nmLost = this.get('model.nodeManagersCountLost') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountLost');
      var nmUnhealthy = this.get('model.nodeManagersCountUnhealthy') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountUnhealthy');
      var nmRebooted = this.get('model.nodeManagersCountRebooted') == null ? Em.I18n.t('services.service.summary.notAvailable'): this.get('model.nodeManagersCountRebooted');
      var nmDecom = this.get('model.nodeManagersCountDecommissioned') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountDecommissioned');
      var result = [];
        result.pushObject(nmActive + " active");
        result.pushObject(nmLost + " lost");
        result.pushObject(nmUnhealthy + " unhealthy");
        result.pushObject(nmRebooted + " rebooted");
        result.pushObject(nmDecom + " decommissioned");
      return result;
    }.property('model.nodeManagersCountActive', 'model.nodeManagersCountLost',
      'model.nodeManagersCountUnhealthy', 'model.nodeManagersCountRebooted', 'model.nodeManagersCountDecommissioned'),
    hiddenInfoClass: "hidden-info-five-line",

    thresh1: 40,
    thresh2: 70,
    maxValue: 100,

    nodeManagersLive: function () {
      return this.get('model.nodeManagerNodes').filterProperty("workStatus", "STARTED");
    }.property('model.nodeManagerNodes.@each.workStatus'),

    data: function () {
      var nodeManagers = this.get('model.nodeManagerNodes.length');
      var nodeManagersLive = this.get('nodeManagersLive').length;
      if (!nodeManagers) {
        return -1;
      } else {
        return (nodeManagersLive / nodeManagers).toFixed(2) * 100;
      }
    }.property('model.nodeManagerNodes.length', 'nodeManagersLive'),

    content: function () {
      var nodeManagers = this.get('model.nodeManagerNodes.length');
      var nodeManagersLive = this.get('nodeManagersLive').length;
      return nodeManagersLive + '/' + nodeManagers;
    }.property('model.nodeManagerNodes.length', 'nodeManagersLive'),

    editWidget: function (event) {
      var parent = this;
      var max_tmp =  parseFloat(parent.get('maxValue'));
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: max_tmp,
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else if (this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }

          // update the slider handles and color
          if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
            $("#slider-range").slider('values', 0 , parseFloat(thresh1));
            $("#slider-range").slider('values', 1 , parseFloat(thresh2));
          }
        }.observes('thresh1', 'thresh2')
      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: 'Customize Widget',
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function() {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              //save to persit
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }
            this.hide();
          }
        },

        didInsertElement: function () {
          var self = this;
          var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
          var colors = ['#B80000', '#FF8E00', '#95A800']; //color red, orange, green

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', false);
            $("#slider-range").slider({
              range: true,
              min: 0,
              max: max_tmp,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers,colors);
              },
              slide: function (event, ui) {
                parent.updateColors(ui.values,colors);
                configObj.set('thresh1', ui.values[0] + '');
                configObj.set('thresh2', ui.values[1] + '');
              },
              change: function (event, ui) {
                parent.updateColors(ui.values,colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', false);
          }
        }
      });
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/pie_chart_widget", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.PieChartDashboardWidgetView = App.DashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/pie_chart'),

    maxValue: 100,

    modelFieldMax: null,
    modelFieldUsed: null,

    hiddenInfo: null,

    isPieExist: null,

    dataForPieChart: null,

    widgetHtmlId: null,

    getUsed: function() {
      return this.get('model').get(this.get('modelFieldUsed')) || 0;
    },

    getMax: function() {
      return this.get('model').get(this.get('modelFieldMax')) || 0;
    },

    calcHiddenInfo: function() {
      var used = this.getUsed();
      var max = this.getMax();
      var percent = max > 0 ? 100 * used / max : 0;
      var result = [];
      result.pushObject(percent.toFixed(1) + '% used');
      result.pushObject(numberUtils.bytesToSize(used, 1, 'parseFloat', 1024 * 1024) + ' of ' + numberUtils.bytesToSize(max, 1, 'parseFloat', 1024 * 1024));
      return result;
    },

    calcIsPieExists: function() {
      return (this.get('model').get(this.get('modelFieldMax')) > 0);
    },

    calcDataForPieChart: function() {
      var used = this.getUsed();
      var total = this.getMax();
      var percent = total > 0 ? ((used)*100 / total).toFixed() : 0;
      var percent_precise = total > 0 ? ((used)*100 / total).toFixed(1) : 0;
      return [percent, percent_precise];
    },

    calc: function() {
      this.set('hiddenInfo', this.calcHiddenInfo());
      this.set('isPieExist', this.calcIsPieExists());
      this.set('dataForPieChart', this.calcDataForPieChart());
    },

    didInsertElement: function() {
      this._super();
      this.addObserver('model.' + this.get('modelFieldMax'), this, this.calc);
      this.addObserver('model.' + this.get('modelFieldUsed'), this, this.calc);
    },

    content: App.ChartPieView.extend({
      model: null,  //data bind here
      id: function() {
        return this.get('parentView.widgetHtmlId');
      }.property('parentView.widgetHtmlId'), // html id
      stroke: '#D6DDDF', //light grey
      thresh1: null, //bind from parent
      thresh2: null,
      innerR: 25,

      existCenterText: true,
      centerTextColor: function () {
        return this.get('contentColor');
      }.property('contentColor'),

      palette: new Rickshaw.Color.Palette({
        scheme: [ '#FFFFFF', '#D6DDDF'].reverse()
      }),

      data: function() {
        var ori_data = this.get('parentView.dataForPieChart');
        return [ ori_data[0], 100 - ori_data[0]];
      }.property(),

      setData: function() {
        this.set('data', this.get('parentView.dataForPieChart'));
      }.observes('parentView.dataForPieChart'),

      contentColor: function () {
        var used = parseFloat(this.get('parentView.dataForPieChart')[1]);
        var thresh1 = parseFloat(this.get('thresh1'));
        var thresh2 = parseFloat(this.get('thresh2'));
        var color_green = '#95A800';
        var color_red = '#B80000';
        var color_orange = '#FF8E00';
        if (used <= thresh1) {
          this.set('palette', new Rickshaw.Color.Palette({
            scheme: [ '#FFFFFF', color_green  ].reverse()
          }));
          return color_green;
        } else if (used <= thresh2) {
          this.set('palette', new Rickshaw.Color.Palette({
            scheme: [ '#FFFFFF', color_orange  ].reverse()
          }));
          return color_orange;
        } else {
          this.set('palette', new Rickshaw.Color.Palette({
            scheme: [ '#FFFFFF', color_red  ].reverse()
          }));
          return color_red;
        }
      }.property('data', 'thresh1', 'thresh2'),

      // refresh text and color when data in model changed
      refreshSvg: function () {
        // remove old svg
        var old_svg =  $("#" + this.get('id'));
        if(old_svg){
          old_svg.remove();
        }

        // draw new svg
        this.appendSvg();
      }.observes('data', 'thresh1', 'thresh2')
    })
  });
  
});
window.require.register("views/main/dashboard/widgets/resource_manager_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.ResourceManagerHeapPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.ResourceManagerHeap'),
    id: '24',

    model_type: 'yarn',
    modelFieldMax: 'jvmMemoryHeapMax',
    modelFieldUsed: 'jvmMemoryHeapUsed',
    widgetHtmlId: 'widget-rm-heap',

    getUsed: function() {
      return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
    },

    getMax: function() {
      return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
    },

    didInsertElement: function() {
      this._super();
      this.calc();
    }

  });
});
window.require.register("views/main/dashboard/widgets/resource_manager_uptime", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.ResourceManagerUptimeView = App.UptimeTextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.ResourceManagerUptime'),
    id: '25',

    model_type: 'yarn',

    component: 'ResourceManager',
    modelField: 'resourceManagerStartTime',

    didInsertElement: function() {
      this._super();
      this.calc();
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/supervisor_live", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.SuperVisorUpView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.SuperVisorUp'),
    id: '28',

    isPieChart: false,
    isText: true,
    isProgressBar: false,
    model_type: 'storm',

    hiddenInfo: function () {
      var result = [];
      result.pushObject(this.get('superVisorsLive').length + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.live'));
      result.pushObject(this.get('superVisorsDead').length + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.dead'));
      return result;
    }.property('superVisorsLive', 'superVisorsDead'),
    hiddenInfoClass: "hidden-info-two-line",

    thresh1: 40,
    thresh2: 70,
    maxValue: 100,

    superVisorComponents: function() {
      return this.get('model.hostComponents').filterProperty('componentName', 'SUPERVISOR');
    }.property('model.hostComponents.length'),

    superVisorsLive: function () {
      return this.get('superVisorComponents').filterProperty("workStatus", "STARTED");
    }.property('superVisorComponents.@each.workStatus'),

    superVisorsDead: function () {
      return this.get('superVisorComponents').filterProperty("workStatus", "INSTALLED");
    }.property('superVisorComponents.@each.workStatus'),


    data: function () {
      if ( !this.get('superVisorComponents.length')) {
        return -1;
      } else {
        return ((this.get('superVisorsLive').length / this.get('model.hostComponents').filterProperty('componentName', 'SUPERVISOR').length).toFixed(2)) * 100;
      }
    }.property('model.hostComponents.length', 'superVisorsLive'),

    content: function () {
      return this.get('superVisorsLive').length + "/" + this.get('superVisorComponents').length;
    }.property('superVisorComponents.length', 'superVisorsLive'),

    editWidget: function (event) {
      var parent = this;
      var max_tmp =  parseFloat(parent.get('maxValue'));
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: max_tmp,
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }

          // update the slider handles and color
          if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
            $("#slider-range").slider('values', 0 , parseFloat(thresh1));
            $("#slider-range").slider('values', 1 , parseFloat(thresh2));
          }
        }.observes('thresh1', 'thresh2')

      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function () {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }
            this.hide();
          }
        },

        didInsertElement: function () {
          var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
          var colors = ['#B80000', '#FF8E00', '#95A800']; //color red, orange, green

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', false);
            $("#slider-range").slider({
              range: true,
              min: 0,
              max: max_tmp,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              },
              slide: function (event, ui) {
                parent.updateColors(ui.values, colors);
                configObj.set('thresh1', ui.values[0] + '');
                configObj.set('thresh2', ui.values[1] + '');
              },
              change: function (event, ui) {
                parent.updateColors(ui.values, colors);
              }
            });
          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', false);
          }
        }
      });
    }
  });
  
});
window.require.register("views/main/dashboard/widgets/tasktracker_live", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.TaskTrackerUpView = App.TextDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.TaskTrackerUp'),
    id: '8',

    model_type: 'mapreduce',

    hiddenInfo: function () {
      var svc = this.get('model');
      var liveCount = this.get('taskTrackersLive').length;
      var totalCount = svc.get('taskTrackers').get('length');
      var result = [];
      result.pushObject(liveCount + " live");
      result.pushObject(totalCount + " total");
      return result;
    }.property('model.taskTrackers.length', 'taskTrackersLive'),

    thresh1: 40,
    thresh2: 70,
    maxValue: 100,

    taskTrackersLive: function () {
      return this.get('model.taskTrackers').filterProperty("workStatus", "STARTED");
    }.property('model.taskTrackers.@each.workStatus'),

    data: function () {
      if (!this.get('model.taskTrackers.length')) {
        return -1;
      } else {
        return (this.get('taskTrackersLive').length / this.get('model.taskTrackers.length')).toFixed(2) * 100;
      }
    }.property('model.taskTrackers.length', 'taskTrackersLive'),

    content: function () {
      return this.get('taskTrackersLive').length + "/" + this.get('model.taskTrackers.length');
    }.property('model.taskTrackers.length', 'taskTrackersLive'),

    editWidget: function (event) {
      var parent = this;
      var max_tmp =  parseFloat(parent.get('maxValue'));
      var configObj = Ember.Object.create({
        thresh1: parent.get('thresh1') + '',
        thresh2: parent.get('thresh2') + '',
        hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
        isThresh1Error: false,
        isThresh2Error: false,
        errorMessage1: "",
        errorMessage2: "",
        maxValue: max_tmp,
        observeNewThresholdValue: function () {
          var thresh1 = this.get('thresh1');
          var thresh2 = this.get('thresh2');
          if (thresh1.trim() != "") {
            if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0) {
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else if (this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
              this.set('isThresh1Error', true);
              this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
            } else {
              this.set('isThresh1Error', false);
              this.set('errorMessage1', '');
            }
          } else {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'This is required');
          }

          if (thresh2.trim() != "") {
            if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
              this.set('isThresh2Error', true);
              this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
            } else {
              this.set('isThresh2Error', false);
              this.set('errorMessage2', '');
            }
          } else {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'This is required');
          }

          // update the slider handles and color
          if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
            $("#slider-range").slider('values', 0 , parseFloat(thresh1));
            $("#slider-range").slider('values', 1 , parseFloat(thresh2));
          }
        }.observes('thresh1', 'thresh2')
      });

      var browserVerion = this.getInternetExplorerVersion();
      App.ModalPopup.show({
        header: Em.I18n.t('dashboard.widgets.popupHeader'),
        classNames: [ 'sixty-percent-width-modal-edit-widget'],
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/dashboard/edit_widget_popup'),
          configPropertyObj: configObj
        }),
        primary: Em.I18n.t('common.apply'),
        onPrimary: function() {
          configObj.observeNewThresholdValue();
          if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
            parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
            parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
            if (!App.testMode) {
              //save to persit
              var big_parent = parent.get('parentView');
              big_parent.getUserPref(big_parent.get('persistKey'));
              var oldValue = big_parent.get('currentPrefObject');
              oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
              big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
            }
            this.hide();
          }
        },

        didInsertElement: function () {
          var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
          var colors = ['#B80000', '#FF8E00', '#95A800']; //color red, orange, green

          if (browserVerion == -1 || browserVerion > 9) {
            configObj.set('isIE9', false);
            configObj.set('isGreenOrangeRed', false);
            $("#slider-range").slider({
              range: true,
              min: 0,
              max: max_tmp,
              values: handlers,
              create: function (event, ui) {
                parent.updateColors(handlers, colors);
              },
              slide: function (event, ui) {
                parent.updateColors(ui.values, colors);
                configObj.set('thresh1', ui.values[0] + '');
                configObj.set('thresh2', ui.values[1] + '');
              },
              change: function (event, ui) {
                parent.updateColors(ui.values, colors);
              }
            });


          } else {
            configObj.set('isIE9', true);
            configObj.set('isGreenOrangeRed', false);
          }
        }
      });
    }

  });
  
});
window.require.register("views/main/dashboard/widgets/text_widget", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.TextDashboardWidgetView = App.DashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/simple_text'),

    classNameBindings: ['isRed', 'isOrange', 'isGreen', 'isNA'],

    isRed: function () {
      return (this.get('data') <= this.get('thresh1'));
    }.property('data','thresh1'),

    isOrange: function () {
      return (this.get('data') <= this.get('thresh2') && this.get('data') > this.get('thresh1'));
    }.property('data','thresh1','thresh2'),

    isGreen: function () {
      return (this.get('data') > this.get('thresh2'));
    }.property('data','thresh2'),

    isNA: function () {
      return this.get('data') === null;
    }.property('data'),

    hiddenInfo: [],
    thresh1: null,
    thresh2: null,
    maxValue: null,
    updateColors: function(handlers, colors) {
      var colorstops = colors[0] + ", "; // start with the first color
      for (var i = 0; i < handlers.length; i++) {
        colorstops += colors[i] + " " + handlers[i] + "%,";
        colorstops += colors[i+1] + " " + handlers[i] + "%,";
      }
      colorstops += colors[colors.length - 1];
      var cssForChromeAndSafari = '-webkit-linear-gradient(left,' + colorstops + ')'; // chrome & safari
      var slider = $('#slider-range');
      slider.css('background-image', cssForChromeAndSafari);
      var cssForIE = '-ms-linear-gradient(left,' + colorstops + ')'; // IE 10+
      slider.css('background-image', cssForIE);
      //$('#slider-range').css('filter', 'progid:DXImageTransform.Microsoft.gradient( startColorStr= ' + colors[0] + ', endColorStr= ' + colors[2] +',  GradientType=1 )' ); // IE 10-
      var cssForFireFox = '-moz-linear-gradient(left,' + colorstops + ')'; // Firefox
      slider.css('background-image', cssForFireFox);

      slider.find('.ui-widget-header').css({'background-color': '#FF8E00', 'background-image': 'none'}); // change the  original ranger color
    }
  });
  
});
window.require.register("views/main/dashboard/widgets/uptime_text_widget", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.UptimeTextDashboardWidgetView = App.TextDashboardWidgetView.extend({

    templateName: require('templates/main/dashboard/widgets/uptime'),

    timeUnit: null,

    hiddenInfoClass: "hidden-info-three-line",

    thresh1: 5,
    thresh2: 10,
    maxValue: 'infinity',

    component: null,
    /**
     * Model's field that used to calculate data and content
     * Should be defined in every child
     */
    modelField: null,

    data: null,
    content: null,

    isGreen: function () {
      return this.get('data') != null;
    }.property('data'),

    isOrange: function () {
      return false;
    }.property('data'),

    isRed: function () {
      return false;
    }.property('data'),

    timeConverter: function (timestamp) {
      var origin = new Date(timestamp);
      origin = origin.toString();
      var result = [];
      var start = origin.indexOf('GMT');
      if (start == -1) { // ie
        var arr = origin.split(" ");
        result.pushObject(arr[0] + " " + arr[1] + " " + arr[2] + " " + arr[3]);
        var second = '';
        for (var i = 4; i < arr.length; i++) {
          second = second + " " + arr[i];
        }
        result.pushObject(second);
      }
      else { // other browsers
        var end = origin.indexOf(" ", start);
        result.pushObject(origin.slice(0, start-10));
        result.pushObject(origin.slice(start-9));
      }
      return result;
    },

    /**
     * All children should have such code
     * <code>
     * didInsertElement: function() {
     *   this._super();
     *   this.calc();
     * }
     * </code>
     */
    didInsertElement: function() {
      this._super();
      this.addObserver('model.' + this.get('modelField'), this, this.calc);
    },

    calc: function() {
      this.set('data', this.calcData());
      this.set('content', this.calcContent());
    },

    uptimeProcessing: function(uptime) {
      var uptimeString = this.timeConverter(uptime);
      var diff = App.dateTime() - uptime;
      if (diff < 0) {
        diff = 0;
      }
      var formatted = date.timingFormat(diff); //17.67 days
      var timeUnit = null;
      if (formatted != null) {
        switch (formatted.split(" ")[1]) {
          case 'secs':
            timeUnit = 's';
            break;
          case 'hours':
            timeUnit = 'hr';
            break;
          case 'days':
            timeUnit = 'd';
            break;
          case 'mins':
            timeUnit = 'min';
            break;
          default:
            timeUnit = formatted.split(" ")[1];
        }
        this.set('timeUnit', timeUnit);
        this.set('hiddenInfo', []);
        this.get('hiddenInfo').pushObject(formatted);
        this.get('hiddenInfo').pushObject(uptimeString[0]);
        this.get('hiddenInfo').pushObject(uptimeString[1]);
      }
      return formatted;
    },

    calcData: function () {
      var field = this.get('modelField');
      var uptime = this.get('model').get(field);
      if (uptime && uptime > 0) {
        var formatted = this.uptimeProcessing(uptime);
        if (formatted != null) {
          return parseFloat(formatted.split(" ")[0]);
        }
      }
      this.set('hiddenInfo', [this.get('component'),'Not Running']);
      return null;
    },

    calcContent: function () {
      var data = this.get('data');
      if (data) {
        return data.toFixed(1) + ' '+ this.get('timeUnit');
      }
      else {
        return Em.I18n.t('services.service.summary.notAvailable');
      }
    }
  });
  
});
window.require.register("views/main/dashboard/widgets/yarn_memory", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var numberUtils = require('utils/number_utils');

  App.YARNMemoryPieChartView = App.PieChartDashboardWidgetView.extend({

    title: Em.I18n.t('dashboard.widgets.YARNMemory'),
    id: '27',

    widgetHtmlId: 'widget-yarn-memory',
    model_type: 'yarn',
    modelFieldUsed: 'allocatedMemory',
    modelFieldMax: 'maxMemory',

    didInsertElement: function() {
      this._super();
      this.calc();
    }
  });
});
window.require.register("views/main/host", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');
  var sort = require('views/common/sort_view');
  var date = require('utils/date');

  App.MainHostView = App.TableView.extend({
    templateName:require('templates/main/host'),
    /**
     * List of hosts in cluster
     * @type {Array}
     */
    content:function () {
      return this.get('controller.content');
    }.property('controller.content.length'),

    clearFiltersObs: function() {
      var self = this;
      Em.run.next(function() {
        if (self.get('controller.clearFilters')) {
          self.clearFilters();
          self.clearStartIndex();
        }
      });
    },

    didInsertElement: function() {
      this.addObserver('controller.clearFilters', this, this.clearFiltersObs);
      this.clearFiltersObs();
      this.addObserver('selectAllHosts', this, this.toggleAllHosts);
    },

    willDestroyElement: function() {
      this.get('categories').forEach(function(c) {
        if (c.get('observes')) {
          c.removeObserver(c.get('observes'), c, c.updateHostsCount);
        }
      });
    },

    /**
     * return filtered number of all content number information displayed on the page footer bar
     * @returns {String}
     */
    filteredContentInfo: function () {
      return this.t('hosts.filters.filteredHostsInfo').format(this.get('filteredContent.length'), this.get('content').get('length'));
    }.property('content.length', 'filteredContent.length'),

    /**
     * Select/deselect all visible hosts flag
     * @property {bool}
     */
    selectAllHosts: false,

    /**
     * Set <code>selected</code> property for each App.Host
     */
    toggleAllHosts: function() {
      this.get('pageContent').setEach('selected', this.get('selectAllHosts'));
    },

    /**
     * Trigger updating <code>selectedHostsCount</code> only 1 time
     */
    selectedHostsObserver: function() {
      Ember.run.once(this, 'updateCheckedFlags');
    }.observes('pageContent.@each.selected'),

    /**
     * Update <code>selectAllHosts</code> value
     */
    updateCheckedFlags: function() {
      this.removeObserver('selectAllHosts', this, this.toggleAllHosts);
      if (this.get('pageContent').length) {
        this.set('selectAllHosts', this.get('pageContent').everyProperty('selected', true));
      }
      else {
        this.set('selectAllHosts', false);
      }
      this.addObserver('selectAllHosts', this, this.toggleAllHosts);
    },

    /**
     * Clear selectedFilter
     * Set <code>selected</code> to false for each host
     */
    clearSelection: function() {
      this.get('pageContent').setEach('selected', false);
      this.set('selectAllHosts', false);
      this.clearFilters();
    },

    /**
     * Confirmation Popup for bulk Operations
     */
    bulkOperationConfirm: function(operationData, selection) {
      var hosts = [];
      var self = this;
      switch(selection) {
        case 's':
          hosts = this.get('content').filterProperty('selected');
          break;
        case 'f':
          hosts = this.get('filteredContent');
          break;
        case 'a':
          hosts = this.get('content').toArray();
          break;
      }
      // no hosts - no actions
      if (!hosts.length) {
        console.log('No bulk operation if no hosts selected.');
        return;
      }
      var hostNames = hosts.mapProperty('hostName');
      var hostsToSkip = [];
      if (operationData.action == "DECOMMISSION") {
        hostsToSkip = hosts.filter(function(host) {
          var invalidStateComponents = host.get('hostComponents').filter(function(component) {
            return component.get('componentName') == operationData.realComponentName && component.get('workStatus') == 'INSTALLED';
          });
          return invalidStateComponents.length > 0;
        });
      }
      var hostNamesSkipped = hostsToSkip.mapProperty('hostName');
      var message;
      if (operationData.componentNameFormatted) {
        message = Em.I18n.t('hosts.bulkOperation.confirmation.hostComponents').format(operationData.message, operationData.componentNameFormatted, hostNames.length);
      }
      else {
        message = Em.I18n.t('hosts.bulkOperation.confirmation.hosts').format(operationData.message, hostNames.length);
      }
      App.ModalPopup.show({
        header: Em.I18n.t('hosts.bulkOperation.confirmation.header'),
        hostNames: hostNames.join("\n"),
        hostNamesSkipped: function() {
          if (hostNamesSkipped.length) {
            return hostNamesSkipped.join("<br/>");
          }
          return false;
        }.property(),
        onPrimary: function() {
          self.get('controller').bulkOperation(operationData, hosts);
          this._super();
        },
        bodyClass: Em.View.extend({
          templateName: require('templates/main/host/bulk_operation_confirm_popup'),
          message: message,
          warningInfo: Em.I18n.t('hosts.bulkOperation.warningInfo.body'),
          textareaVisible: false,
          textTrigger: function() {
            this.set('textareaVisible', !this.get('textareaVisible'));
          },
          putHostNamesToTextarea: function() {
            var hostNames = this.get('parentView.hostNames');
            if (this.get('textareaVisible')) {
              var wrapper = $(".task-detail-log-maintext");
              $('.task-detail-log-clipboard').html(hostNames).width(wrapper.width()).height(250);
              Em.run.next(function() {
                $('.task-detail-log-clipboard').select();
              });
            }
          }.observes('textareaVisible')
        })
      });
    },

    sortView: sort.wrapperView,
    nameSort: sort.fieldView.extend({
      column: 1,
      name:'publicHostName',
      displayName: Em.I18n.t('common.name')
    }),
    ipSort: sort.fieldView.extend({
      column: 2,
      name:'ip',
      displayName: Em.I18n.t('common.ipAddress'),
      type: 'ip'
    }),
    cpuSort: sort.fieldView.extend({
      column: 3,
      name:'cpu',
      displayName: Em.I18n.t('common.cores'),
      type: 'number'
    }),
    memorySort: sort.fieldView.extend({
      column: 4,
      name:'memory',
      displayName: Em.I18n.t('common.ram'),
      type: 'number'
    }),
    diskUsageSort: sort.fieldView.extend({
      name:'diskUsage',
      displayName: Em.I18n.t('common.diskUsage')
    }),
    loadAvgSort: sort.fieldView.extend({
      column: 5,
      name:'loadAvg',
      displayName: Em.I18n.t('common.loadAvg'),
      type: 'number'
    }),

    HostView:Em.View.extend({
      content:null,
      tagName: 'tr',
      didInsertElement: function(){
        App.tooltip(this.$("[rel='HealthTooltip'], [rel='UsageTooltip'], [rel='ComponentsTooltip']"));
        this.set('isComponentsCollapsed', true);
      },

      toggleComponents: function(event) {
        this.set('isComponentsCollapsed', !this.get('isComponentsCollapsed'));
        this.$('.host-components').toggle();
      },

      /**
       * Tooltip message for "Restart Required" icon
       * @returns {String}
       */
      restartRequiredComponentsMessage: function() {
        var restartRequiredComponents = this.get('content.hostComponents').filterProperty('staleConfigs', true);
        var count = restartRequiredComponents.length;
        if (count <= 5) {
          var word = (count == 1) ? Em.I18n.t('common.component') : Em.I18n.t('common.components');
          return Em.I18n.t('hosts.table.restartComponents.withNames').format(restartRequiredComponents.getEach('displayName').join(', ')) + ' ' + word.toLowerCase();
        }
        return Em.I18n.t('hosts.table.restartComponents.withoutNames').format(count);
      }.property('content.hostComponents.@each.staleConfigs'),

      /**
       * Tooltip message for "Maintenance" icon
       * @returns {String}
       */
      componentsInPassiveStateMessage: function() {
        var componentsInPassiveState = this.get('content.hostComponents').filter(function(component) {
          return component.get('passiveState') !== 'OFF';
        });
        var count = componentsInPassiveState.length;
        if (count <= 5) {
          return Em.I18n.t('hosts.table.componentsInPassiveState.withNames').format(componentsInPassiveState.getEach('displayName').join(', '));
        }
        return Em.I18n.t('hosts.table.componentsInPassiveState.withoutNames').format(count);
      }.property('content.hostComponents.@each.passiveState'),

      /**
       * String with list of host components <code>displayName</code>
       * @returns {String}
       */
      labels: function() {
        return this.get('content.hostComponents').getEach('displayName').join("<br />");
      }.property('content.hostComponents.length'),

      /**
       * CSS value for disk usage bar
       * @returns {String}
       */
      usageStyle:function () {
        return "width:" + this.get('content.diskUsage') + "%";
      }.property('content.diskUsage')

    }),

    /**
     * Category view for all hosts
     * @type {Object}
     */
    //@TODO maybe should be separated to two types (basing on <code>isHealthStatus</code>)
    categoryObject: Em.Object.extend({

      /**
       * Text used with <code>hostsCount</code> in category label
       * @type {String}
       */
      value: null,
      /**
       * Is category based on host health status
       * @type {bool}
       */
      isHealthStatus: true,
      /**
       * host health status (used if <code>isHealthStatus</code> is true)
       * @type {String}
       */
      healthStatusValue: '',
      /**
       * Should category be displayed on the top of the hosts table
       * @type {bool}
       */
      isVisible: true,
      /**
       * Is category selected now
       * @type {bool}
       */
      isActive: false,
      /**
       * String with path that category should observe
       * @type {String}
       */
      observes: null,
      /**
       * CSS-class for span in the category-link (used if <code>isHealthStatus</code> is false)
       * @type {String}
       */
      class: null,
      /**
       * Associated column number
       * @type {Number}
       */
      column: null,
      /**
       * Type of filter value (string, number, boolean)
       * @type {String}
       */
      type: null,
      /**
       * @type {String|Number|bool}
       */
      filterValue: null,
      /**
       * <code>App.Host</code> property that should be used to calculate <code>hostsCount</code> (used if <code>isHealthStatus</code> is false)
       * @type {String}
       */
      hostProperty: null,

      /**
       * Number of host in current category
       * @type {Number}
       */
      hostsCount: 0,

      /**
       * Determine if category has hosts
       * @type {bool}
       */
      hasHosts: false,

      /**
       * Add "active" class for category span-wrapper if current category is selected
       * @type {String}
       */
      itemClass: function() {
        return this.get('isActive') ? 'active' : '';
      }.property('isActive'),

      /**
       * Trigger updating <code>hostsCount</code> only 1 time
       */
      updateHostsCount: function() {
        Em.run.once(this, 'updateOnce');
      },
      /**
       * Update <code>hostsCount</code> in current category
       */
      updateOnce: function() {
        //skip update when view is destroyed
        if(!this.get('view.content')) return;
        var statusString = this.get('healthStatusValue');
        if (this.get('isHealthStatus')) {
          if (statusString == "") {
            this.set('hostsCount', this.get('view.content').get('length'));
          }
          else {
            this.set('hostsCount', this.get('view.content').filterProperty('healthClass', statusString).get('length'));
          }
        }
        else {
          this.set('hostsCount', this.get('view.content').filterProperty(this.get('hostProperty')).get('length'));
        }
        this.set('hasHosts', !!this.get('hostsCount'));
      },

      /**
       * Text shown on the right of category icon
       * @type {String}
       */
      label: function () {
        return "%@ (%@)".fmt(this.get('value'), this.get('hostsCount'));
      }.property('hostsCount')
    }),

    /**
     * List of categories used to filter hosts
     * @type {Array}
     */
    categories: function () {
      var self = this;
      self.categoryObject.reopen({
        view: self
      });

      var category_mocks = require('data/host/categories');

      return category_mocks.map(function(category_mock) {
        var c = self.categoryObject.create(category_mock);
        if (c.get('observes')) {
          c.addObserver(c.get('observes'), c, c.updateHostsCount);
          c.updateHostsCount();
        }
        return c;
      });
    }.property(),

    /**
     * Category for <code>selected</code> property of each App.Host
     */
    selectedCategory: function() {
      return this.get('categories').findProperty('selected', true);
    }.property('categories.@each.selected'),

    statusFilter: Em.View.extend({
      column: 0,
      categories: [],
      value: null,
      class: "",
      comboBoxLabel: function(){
        var selected = this.get('categories').findProperty('isActive',true);
        if (!this.get('value') || !selected) {
          return "%@ (%@)".fmt(Em.I18n.t('common.all'), this.get('parentView.content.length'));
        } else {
          return "%@ (%@)".fmt(selected.get('value'), selected.get('hostsCount'))
        }
      }.property('value'),
      /**
       * switch active category label
       */
      onCategoryChange: function() {
        this.get('categories').setEach('isActive', false);
        var selected = this.get('categories').findProperty('healthStatusValue', this.get('value'));
        selected.set('isActive', true);
        this.set('class', selected.get('class') + ' ' + this.get('value'));
      }.observes('value'),

      showClearFilter: function(){
        var mockEvent = {
          context: this.get('categories').findProperty('healthStatusValue', this.get('value'))
        };
        this.selectCategory(mockEvent);
      },
      /**
       * Trigger on Category click
       * @param {Object} event
       */
      selectCategory: function(event){
        var category = event.context;
        var self = this;
        this.set('value', category.get('healthStatusValue'));
        if (category.get('isHealthStatus')) {
          this.get('parentView').updateFilter(0, category.get('healthStatusValue'), 'string');
          this.get('categories').filterProperty('isHealthStatus', false).forEach(function(c) {
            self.get('parentView').updateFilter(c.get('column'), '', c.get('type'));
          });
        }
        else {
          this.get('parentView').updateFilter(0, '', 'string');
          this.get('categories').filterProperty('isHealthStatus', false).forEach(function(c) {
            if (c.get('column') === category.get('column')) {
              self.get('parentView').updateFilter(category.get('column'), category.get('filterValue'), category.get('type'));
            }
            else {
              self.get('parentView').updateFilter(c.get('column'), '', c.get('type'));
            }
          });
        }
      },
      clearFilter: function() {
        this.get('categories').setEach('isActive', false);
        this.set('value', '');
        this.set('class', '');
        this.showClearFilter();
      }
    }),

    /**
     * view of the alert filter implemented as a category of host statuses
     */
    alertFilter: Em.View.extend({
      column: 7,
      value: null,
      classNames: ['noDisplay'],
      showClearFilter: function(){
        var mockEvent = {
          context: this.get('parentView.categories').findProperty('healthStatusValue', 'health-status-WITH-ALERTS')
        };
        if(this.get('value')) {
          this.get('parentView.childViews').findProperty('column', 0).selectCategory(mockEvent);
        }
      }
    }),

    /**
     * view of the staleConfigs filter implemented as a category of host statuses
     */
    restartFilter: Em.View.extend({
      column: 8,
      value: null,
      classNames: ['noDisplay'],
      showClearFilter: function(){
        var mockEvent = {
          context: this.get('parentView.categories').findProperty('healthStatusValue', 'health-status-RESTART')
        };
        if(this.get('value')) {
          this.get('parentView.childViews').findProperty('column', 0).selectCategory(mockEvent);
        }
      }
    }),

    /**
     * view of the maintenance filter implemented as a category of host statuses
     */
    passiveStateFilter: Em.View.extend({
      column: 9,
      value: null,
      classNames: ['noDisplay'],
      showClearFilter: function(){
        var mockEvent = {
          context: this.get('parentView.categories').findProperty('healthStatusValue', 'health-status-PASSIVE_STATE')
        };
        if(this.get('value')) {
          this.get('parentView.childViews').findProperty('column', 0).selectCategory(mockEvent);
        }
      }
    }),

    /**
     * view of the "selected" filter implemented as a category of host statuses
     */
    selectedFilter: Em.View.extend({
      column: 10,
      value: null,
      class: ['noDisplay'],
      showClearFilter: function(){
        var mockEvent = {
          context: this.get('parentView.categories').findProperty('healthStatusValue', 'health-status-SELECTED')
        };
        if(this.get('value')) {
          this.get('parentView.childViews').findProperty('column', 0).selectCategory(mockEvent);
        }
      }
    }),

    /**
     * Filter view for name column
     * Based on <code>filters</code> library
     */
    nameFilterView: filters.createTextView({
      column: 1,
      fieldType: 'filter-input-width',
      onChangeValue: function(){
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
      }
    }),

    /**
     * Filter view for ip column
     * Based on <code>filters</code> library
     */
    ipFilterView: filters.createTextView({
      column: 2,
      fieldType: 'filter-input-width',
      onChangeValue: function(){
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
      }
    }),

    /**
     * Filter view for Cpu column
     * Based on <code>filters</code> library
     */
    cpuFilterView: filters.createTextView({
      fieldType: 'filter-input-width',
      fieldId: 'cpu_filter',
      column: 3,
      onChangeValue: function(){
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'number');
      }
    }),

    /**
     * Filter view for LoadAverage column
     * Based on <code>filters</code> library
     */
    loadAvgFilterView: filters.createTextView({
      fieldType: 'filter-input-width',
      fieldId: 'load_avg_filter',
      column: 5,
      onChangeValue: function(){
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'number');
      }
    }),

    /**
     * Filter view for Ram column
     * Based on <code>filters</code> library
     */
    ramFilterView: filters.createTextView({
      fieldType: 'filter-input-width',
      fieldId: 'ram_filter',
      column: 4,
      onChangeValue: function(){
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'ambari-bandwidth');
      }
    }),

    /**
     * Filter view for HostComponents column
     * Based on <code>filters</code> library
     */
    componentsFilterView: filters.createComponentView({

      column: 6,

      /**
       * Inner FilterView. Used just to render component. Value bind to <code>mainview.value</code> property
       * Base methods was implemented in <code>filters.componentFieldView</code>
       */
      filterView: filters.componentFieldView.extend({
        templateName: require('templates/main/host/component_filter'),

        /**
         * Next three lines bind data to this view
         */
        masterComponentsBinding: 'controller.masterComponents',
        slaveComponentsBinding: 'controller.slaveComponents',
        clientComponentsBinding: 'controller.clientComponents',

        /**
         * Checkbox for quick selecting/deselecting of master components
         */
        masterComponentsChecked:false,
        toggleMasterComponents:function () {
          this.get('masterComponents').setEach('checkedForHostFilter', this.get('masterComponentsChecked'));
        }.observes('masterComponentsChecked'),

        /**
         * Checkbox for quick selecting/deselecting of slave components
         */
        slaveComponentsChecked:false,
        toggleSlaveComponents:function () {
          this.get('slaveComponents').setEach('checkedForHostFilter', this.get('slaveComponentsChecked'));
        }.observes('slaveComponentsChecked'),

        /**
         * Checkbox for quick selecting/deselecting of client components
         */
        clientComponentsChecked: false,
        toggleClientComponents: function() {
          this.get('clientComponents').setEach('checkedForHostFilter', this.get('clientComponentsChecked'));
        }.observes('clientComponentsChecked'),

        /**
         * Clear filter.
         * Called by parent view, when user clicks on <code>x</code> button(clear button)
         */
        clearFilter:function() {
          this.set('masterComponentsChecked', false);
          this.set('slaveComponentsChecked', false);
          this.set('clientComponentsChecked', false);

          this.get('masterComponents').setEach('checkedForHostFilter', false);
          this.get('slaveComponents').setEach('checkedForHostFilter', false);
          this.get('clientComponents').setEach('checkedForHostFilter', false);

          this._super();
        },

        /**
         * Onclick handler for <code>Apply filter</code> button
         */
        applyFilter:function() {
          this._super();
          var self = this;
          var chosenComponents = [];

          this.get('masterComponents').filterProperty('checkedForHostFilter', true).forEach(function(item){
            chosenComponents.push(item.get('id'));
          });
          this.get('slaveComponents').filterProperty('checkedForHostFilter', true).forEach(function(item){
            chosenComponents.push(item.get('id'));
          });
          this.get('clientComponents').filterProperty('checkedForHostFilter', true).forEach(function(item){
            chosenComponents.push(item.get('id'));
          });
          Em.run.next(function() {
            self.set('value', chosenComponents.toString());
          });
        },

        /**
         * Verify that checked checkboxes are equal to value stored in hidden field (components ids list)
         */
        checkComponents: function() {
          var components = this.get('value').split(',');
          var self = this;
          if (components) {
            components.forEach(function(componentId) {
              if(!self.tryCheckComponent(self, 'masterComponents', componentId)) {
                if(!self.tryCheckComponent(self, 'slaveComponents', componentId)) {
                  self.tryCheckComponent(self, 'clientComponents', componentId);
                }
              }
            });
          }
        }.observes('value'),

        tryCheckComponent: function(self, category, componentId) {
          var c = self.get(category).findProperty('id', componentId);
          if (c) {
            if (!c.get('checkedForHostFilter')) {
              c.set('checkedForHostFilter', true);
              return true;
            }
          }
          return false;
        }

      }),
      onChangeValue: function(){
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'multiple');
      }
    }),

    /**
     * associations between host property and column index
     * @type {Array}
     */
    colPropAssoc: function(){
      var associations = [];
      associations[0] = 'healthClass';
      associations[1] = 'publicHostName';
      associations[2] = 'ip';
      associations[3] = 'cpu';
      associations[4] = 'memoryFormatted';
      associations[5] = 'loadAvg';
      associations[6] = 'hostComponents';
      associations[7] = 'criticalAlertsCount';
      associations[8] = 'componentsWithStaleConfigsCount';
      associations[9] = 'componentsInPassiveStateCount';
      associations[10] = 'selected';
      return associations;
    }.property()
  });
  
});
window.require.register("views/main/host/addHost/step4_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.AddHostStep4View = Em.View.extend({

    templateName: require('templates/main/host/addHost/step4'),


    didInsertElement: function() {

    }

  });
  
});
window.require.register("views/main/host/add_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.AddHostView = Em.View.extend({

    templateName: require('templates/main/host/add'),

    isStep1Disabled: function () {
      return this.isStepDisabled(1);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.isStepDisabled(2);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.isStepDisabled(3);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep4Disabled: function () {
      return this.isStepDisabled(4);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep5Disabled: function () {
      return this.isStepDisabled(5);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep6Disabled: function () {
      return this.isStepDisabled(6);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep7Disabled: function () {
      return this.isStepDisabled(7);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

   isStepDisabled: function (index) {
      return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
    }

  });
  
});
window.require.register("views/main/host/configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.MainHostConfigsView = Em.View.extend({
    templateName: require('templates/main/host/configs'),

    content: function(){
      return App.router.get('mainHostDetailsController.content');
    }.property('App.router.mainHostDetailsController.content')

  });
  
});
window.require.register("views/main/host/configs_service", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.MainHostServiceConfigsView = Em.View.extend({
    templateName: require('templates/main/host/configs_service'),
    didInsertElement: function () {
      var controller = this.get('controller');
      controller.loadStep();
    },
    isConfigsEditable: false,
    content: function () {
      return App.router.get('mainHostDetailsController.content');
    }.property('App.router.mainHostDetailsController.content')
  });
  
});
window.require.register("views/main/host/configs_service_menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var misc = require('utils/misc');

  App.MainHostServiceMenuView = Em.CollectionView.extend({
    content:function () {
      var host = this.get('host');
      var hostComponents = host.get('hostComponents');
      var services = Em.A([]);
      if (hostComponents) {
        hostComponents.forEach(function (hc) {
          var service = hc.get('service');
          if (service) {
          var serviceName = service.get('serviceName');
            if(!['PIG', 'SQOOP', 'HCATALOG', 'GANGLIA'].contains(serviceName)){
              if (!services.findProperty('serviceName', serviceName)) {
                services.push(service);
              }
            }
          } else {
            console.warn("service not found for " + hc.get('componentName'));
          }
        });
      }
      return misc.sortByOrder(App.Service.servicesSortOrder, services);
    }.property('host'),
    
    host: function(){
      return App.router.get('mainHostDetailsController.content');
    }.property('App.router.mainHostDetailsController.content'),
    
    selectedService: null,
    
    showHostService: function(event){
      var service = event.contexts[0];
      if(service != null) {
        this.set('selectedService', service);
        var context = service;
        context.host = this.get('host');
        this.get('controller').connectOutlet('service_config_outlet', 'mainHostServiceConfigs', context);
      }
      else {
        this.get('controller').connectOutlet('service_config_outlet', Em.View.extend({
          template: Ember.Handlebars.compile('<i class="message">{{t hosts.host.serviceNotAvailable}}</i>')
        }))
      }
    },

    didInsertElement:function () {
      var event = {
          contexts: [this.get('content').objectAt(0)]
      };
      this.showHostService(event);
    },

    activeServiceId:null,

    tagName:'ul',
    classNames:["nav", "nav-list", "nav-services"],

    itemViewClass:Em.View.extend({
      classNameBindings:["active", "clients"],
      active:function () {
        return this.get('content.serviceName') == this.get('parentView.selectedService.serviceName') ? 'active' : '';
      }.property('parentView.selectedService.serviceName'),
      template: Ember.Handlebars.compile('<a href="#" {{action showHostService view.content target="view.parentView"}} >{{view.content.displayName}}</a>')
    })
  });
});
window.require.register("views/main/host/details", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.MainHostDetailsView = Em.View.extend({
    templateName: require('templates/main/host/details'),

    content: function(){
      return App.router.get('mainHostDetailsController.content');
    }.property('App.router.mainHostDetailsController.content'),

    isActive: function() {
      return this.get('controller.content.passiveState') === "OFF";
    }.property('controller.content.passiveState'),

    maintenance: function(){
      var onOff = this.get('isActive') ? "On" : "Off";
      return [
        {action: 'startAllComponents', liClass: (this.get('controller.content.isNotHeartBeating')?'disabled':'enabled'), cssClass: 'icon-play', 'label': this.t('hosts.host.details.startAllComponents')},
        {action: 'stopAllComponents', liClass: (this.get('controller.content.isNotHeartBeating')?'disabled':'enabled'), cssClass: 'icon-stop', 'label': this.t('hosts.host.details.stopAllComponents')},
        {action: 'restartAllComponents', liClass: (this.get('controller.content.isNotHeartBeating')?'disabled':'enabled'), cssClass: 'icon-repeat', 'label': this.t('hosts.host.details.restartAllComponents')},
        {action: 'onOffPassiveModeForHost', liClass:'', cssClass: 'icon-medkit', active:this.get('isActive'), 'label': this.t('passiveState.turn' + onOff)},
        {action: 'deleteHost', liClass:'', cssClass: 'icon-remove', 'label': this.t('hosts.host.details.deleteHost')}];
    }.property('controller.content','isActive', 'controller.content.isNotHeartBeating'),
    didInsertElement: function() {
      //if host is not existed then route to list of hosts
      if (!this.get('content.isLoaded')) {
        App.router.transitionTo('main.hosts.index');
        return;
      }
      App.tooltip($("[rel='HealthTooltip']"));
    }
  });
  
});
window.require.register("views/main/host/details/host_component_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var uiEffects = require('utils/ui_effects');

  App.HostComponentView = Em.View.extend({

    templateName: require('templates/main/host/details/host_component'),

    /**
     * @type {App.HostComponent}
     */
    content: null,

    /**
     * @type {App.HostComponent}
     */
    hostComponent: function () {
      var hostComponent = null;
      var serviceComponent = this.get('content');
      var host = App.router.get('mainHostDetailsController.content');
      if (host) {
        hostComponent = host.get('hostComponents').findProperty('componentName', serviceComponent.get('componentName'));
      }
      return hostComponent;
    }.property('content', 'App.router.mainHostDetailsController.content'),
    /**
     * @type {String}
     */
    workStatus: function () {
      var workStatus = this.get('content.workStatus');
      var hostComponent = this.get('hostComponent');
      if (hostComponent) {
        workStatus = hostComponent.get('workStatus');
      }
      return workStatus;
    }.property('content.workStatus', 'hostComponent.workStatus'),

    /**
     * Return host component text status
     * @type {String}
     */
    componentTextStatus: function () {
      var componentTextStatus = this.get('content.componentTextStatus');
      var hostComponent = this.get('hostComponent');
      if (hostComponent) {
        componentTextStatus = hostComponent.get('componentTextStatus');
      }
      return componentTextStatus;
    }.property('content.passiveState','workStatus'),

    /**
     * CSS-class for host component status
     * @type {String}
     */
    statusClass: function () {
      //Class when install failed
      if (this.get('workStatus') === App.HostComponentStatus.install_failed) {
        return 'health-status-color-red icon-cog';
      }

      //Class when installing
      if (this.get('workStatus') === App.HostComponentStatus.installing) {
        return 'health-status-color-blue icon-cog';
      }

      //For all other cases
      return 'health-status-' + App.HostComponentStatus.getKeyName(this.get('workStatus'));

    }.property('workStatus'),
    /**
     * CSS-icon-class for host component status
     * @type {String}
     */
    statusIconClass: function () {
      switch (this.get('statusClass')) {
        case 'health-status-started':
        case 'health-status-starting':
          return App.healthIconClassGreen;
          break;
        case 'health-status-installed':
        case 'health-status-stopping':
          return App.healthIconClassRed;
          break;
        case 'health-status-unknown':
          return App.healthIconClassYellow;
          break;
        case 'health-status-DEAD-ORANGE':
          return App.healthIconClassOrange;
          break;
        default:
          return "";
          break;
      }
    }.property('statusClass'),


    /**
     * CSS-class for disabling drop-down menu with list of host component actions
     * Disabled if host's <code>healthClass</code> is health-status-DEAD-YELLOW (lost heartbeat)
     * @type {String}
     */
    disabled: function () {
      return (this.get('parentView.content.healthClass') === "health-status-DEAD-YELLOW") ? 'disabled' : '';
    }.property('parentView.content.healthClass'),

    /**
     * For Upgrade failed state
     * @type {bool}
     */
    isUpgradeFailed: function () {
      return App.HostComponentStatus.getKeyName(this.get('workStatus')) === "upgrade_failed";
    }.property("workStatus"),

    /**
     * For Install failed state
     * @type {bool}
     */
    isInstallFailed: function () {
      return App.HostComponentStatus.getKeyName(this.get('workStatus')) === "install_failed";
    }.property("workStatus"),

    /**
     * For Started and Starting states
     * @type {bool}
     */
    isStart: function () {
      return [App.HostComponentStatus.started, App.HostComponentStatus.starting].contains(this.get('workStatus'));
    }.property('workStatus'),

    /**
     * For Installed state
     * @type {bool}
     */
    isStop: function () {
      return (this.get('workStatus') == App.HostComponentStatus.stopped);
    }.property('workStatus'),

    /**
     * For Installing state
     * @type {bool}
     */
    isInstalling: function () {
      return (this.get('workStatus') == App.HostComponentStatus.installing);
    }.property('workStatus'),

    /**
     * For Init state
     * @type {bool}
     */
    isInit: function() {
      return this.get('workStatus') == App.HostComponentStatus.init;
    }.property('workStatus'),

    /**
     * No action available while component is starting/stopping/unknown
     * @type {String}
     */
    noActionAvailable: function () {
      var workStatus = this.get('workStatus');
      if ([App.HostComponentStatus.starting, App.HostComponentStatus.stopping, App.HostComponentStatus.unknown, App.HostComponentStatus.disabled].contains(workStatus)) {
        return "hidden";
      }else{
        return "";
      }
    }.property('workStatus'),

    /**
     * For Stopping or Starting states
     * @type {bool}
     */
    isInProgress: function () {
      return (this.get('workStatus') === App.HostComponentStatus.stopping ||
        this.get('workStatus') === App.HostComponentStatus.starting);
    }.property('workStatus'),

    /**
     * For OFF <code>passiveState</code> of host component
     * @type {bool}
     */
    isActive: function () {
      return (this.get('content.passiveState') == "OFF");
    }.property('content.passiveState'),

    /**
     * For PASSIVE <code>passiveState</code> of host or service
     * @type {bool}
     */
    isImplied: function() {
      return (this.get('parentView.content.passiveState') === 'ON' || this.get('content.service.passiveState') === 'ON');
    }.property('parentView.content.passiveState', 'content.service.passiveState'),

    /**
     * Shows whether we need to show Delete button
     * @type {bool}
     */
    isDeletableComponent: function () {
      return App.get('components.deletable').contains(this.get('content.componentName'));
    }.property('content'),

    /**
     * Host component with some <code>workStatus</code> can't be deleted (so, disable such action in the dropdown list)
     * @type {bool}
     */
    isDeleteComponentDisabled: function () {
      return !(this.get('isComponentRecommissionAvailable') || [App.HostComponentStatus.stopped, App.HostComponentStatus.unknown, App.HostComponentStatus.install_failed, App.HostComponentStatus.upgrade_failed, App.HostComponentStatus.init].contains(this.get('workStatus')));
    }.property('workStatus','isComponentRecommissionAvailable'),

    /**
     * Check if component may be reassinged to another host
     * @type {bool}
     */
    isReassignable: function () {
      return App.supports.reassignMaster && App.get('components.reassignable').contains(this.get('content.componentName')) && App.Host.find().content.length > 1;
    }.property('content.componentName'),

    /**
     * Check if component is restartable
     * @type {bool}
     */
    isRestartableComponent: function() {
      return App.get('components.restartable').contains(this.get('content.componentName'));
    }.property('content'),

    /**
     * Host component with some <code>workStatus</code> can't be restarted (so, disable such action in the dropdown list)
     * @type {bool}
     */
    isRestartComponentDisabled: function() {
      return ![App.HostComponentStatus.started].contains(this.get('workStatus'));
    }.property('workStatus'),

    didInsertElement: function () {
      App.tooltip($('[rel=componentHealthTooltip]'));
      App.tooltip($('[rel=passiveTooltip]'));
      if (this.get('isInProgress')) {
        this.doBlinking();
      }
    },

    /**
     * Do blinking for 1 minute
     */
    doBlinking: function () {
      var workStatus = this.get('workStatus');
      var self = this;
      var pulsate = [ App.HostComponentStatus.starting, App.HostComponentStatus.stopping, App.HostComponentStatus.installing].contains(workStatus);
      if (pulsate && !self.get('isBlinking')) {
        self.set('isBlinking', true);
        uiEffects.pulsate(self.$('.components-health'), 1000, function () {
          self.set('isBlinking', false);
          self.doBlinking();
        });
      }
    },

    /**
     * Start blinking when host component is starting/stopping
     */
    startBlinking: function () {
      this.$('.components-health').stop(true, true);
      this.$('.components-health').css({opacity: 1.0});
      this.doBlinking();
    }.observes('workStatus')

  });
  
});
window.require.register("views/main/host/details/host_component_views/datanode_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.DataNodeComponentView = App.HostComponentView.extend(App.Decommissionable, {

    componentForCheckDecommission: 'NAMENODE',

    /**
     * Get component decommission status from server
     * @returns {$.ajax}
     */
    getDNDecommissionStatus: function() {
      // always get datanode decommission statue from active namenode (if NN HA enabled)
      var hdfs = App.HDFSService.find().objectAt(0);
      var activeNNHostName = (!hdfs.get('snameNode') && hdfs.get('activeNameNode')) ? hdfs.get('activeNameNode.hostName'): hdfs.get('nameNode.hostName');
      return App.ajax.send({
        name: 'host.host_component.decommission_status_datanode',
        sender: this,
        data: {
          hostName: activeNNHostName,
          componentName: this.get('componentForCheckDecommission')
        },
        success: 'getDNDecommissionStatusSuccessCallback',
        error: 'getDNDecommissionStatusErrorCallback'
      });
    },

    /**
     * Set received value or null to <code>decommissionedStatusObject</code>
     * @param {Object} response
     * @returns {Object|null}
     */
    getDNDecommissionStatusSuccessCallback: function (response) {
      var statusObject = Em.get(response, 'metrics.dfs.namenode');
      if ( statusObject != null) {
        this.set('decommissionedStatusObject', statusObject);
        return statusObject;
      }
      return null;
    },

    /**
     * Set null to <code>decommissionedStatusObject</code> if server returns error
     * @returns {null}
     */
    getDNDecommissionStatusErrorCallback: function () {
      this.set('decommissionedStatusObject', null);
      return null;
    },

    /**
     * load Recommission/Decommission status from adminState of each live node
     */
    loadComponentDecommissionStatus: function () {
      var hostName = this.get('content.host.hostName');
      var dfd = $.Deferred();
      var self = this;
      this.getDNDecommissionStatus().done(function () {
        var curObj = self.get('decommissionedStatusObject');
        self.set('decommissionedStatusObject', null);
        // HDP-2 stack
        if (App.get('isHadoop2Stack')) {
          if (curObj) {
            var liveNodesJson = App.parseJSON(curObj.LiveNodes);
            if (liveNodesJson && liveNodesJson[hostName] ) {
              switch(liveNodesJson[hostName].adminState) {
                case "In Service":
                  self.set('isComponentRecommissionAvailable', false);
                  self.set('isComponentDecommissioning', false);
                  self.set('isComponentDecommissionAvailable', self.get('isStart'));
                  break;
                case "Decommission In Progress":
                  self.set('isComponentRecommissionAvailable', true);
                  self.set('isComponentDecommissioning', true);
                  self.set('isComponentDecommissionAvailable', false);
                  break;
                case "Decommissioned":
                  self.set('isComponentRecommissionAvailable', true);
                  self.set('isComponentDecommissioning', false);
                  self.set('isComponentDecommissionAvailable', false);
                  break;
              }
            } else {
              // if namenode is down, get desired_admin_state to decide if the user had issued a decommission
              var deferred = $.Deferred();
              self.getDesiredAdminState().done(function () {
                var desired_admin_state = self.get('desiredAdminState');
                self.set('desiredAdminState', null);
                switch(desired_admin_state) {
                  case "INSERVICE":
                    self.set('isComponentRecommissionAvailable', false);
                    self.set('isComponentDecommissioning', false);
                    self.set('isComponentDecommissionAvailable', self.get('isStart'));
                    break;
                  case "DECOMMISSIONED":
                    self.set('isComponentRecommissionAvailable', true);
                    self.set('isComponentDecommissioning', false);
                    self.set('isComponentDecommissionAvailable', false);
                    break;
                }
                deferred.resolve(desired_admin_state);
              });
            }
          }
        }
        else {
          if (curObj) {
            var liveNodesJson = App.parseJSON(curObj.LiveNodes);
            var decomNodesJson = App.parseJSON(curObj.DecomNodes);
            var deadNodesJson = App.parseJSON(curObj.DeadNodes);
            if (decomNodesJson && decomNodesJson[hostName] ) {
              self.set('isComponentRecommissionAvailable', true);
              self.set('isComponentDecommissioning', true);
              self.set('isComponentDecommissionAvailable', false);
            } else if (deadNodesJson && deadNodesJson[hostName] ) {
              self.set('isComponentRecommissionAvailable', true);
              self.set('isComponentDecommissioning', false);
              self.set('isComponentDecommissionAvailable', false);
            } else if (liveNodesJson && liveNodesJson[hostName] ) {
              self.set('isComponentRecommissionAvailable', false);
              self.set('isComponentDecommissioning', false);
              self.set('isComponentDecommissionAvailable', self.get('isStart'));
            } else {
              // if namenode is down, get desired_admin_state to decide if the user had issued a decommission
              var deferred = $.Deferred();
              self.getDesiredAdminState().done( function () {
                var desired_admin_state = self.get('desiredAdminState');
                self.set('desiredAdminState', null);
                switch(desired_admin_state) {
                  case "INSERVICE":
                    self.set('isComponentRecommissionAvailable', false);
                    self.set('isComponentDecommissioning', false);
                    self.set('isComponentDecommissionAvailable', self.get('isStart'));
                    break;
                  case "DECOMMISSIONED":
                    self.set('isComponentRecommissionAvailable', true);
                    self.set('isComponentDecommissioning', false);
                    self.set('isComponentDecommissionAvailable', false);
                    break;
                }
                deferred.resolve(desired_admin_state);
              });
            }
          }
        }
        dfd.resolve(curObj);
      });
      return dfd.promise();
    }


  });
  
});
window.require.register("views/main/host/details/host_component_views/nodemanager_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.NodeManagerComponentView = App.HostComponentView.extend(App.Decommissionable, {

    componentForCheckDecommission: 'RESOURCEMANAGER',

    /**
     * load Recommission/Decommission status for nodeManager from nodeManagers list
     */
    loadComponentDecommissionStatus: function () {
      var hostName = this.get('content.host.hostName');
      var dfd = $.Deferred();
      var self = this;
      this.getDesiredAdminState().done( function () {
        var desired_admin_state = self.get('desiredAdminState');
        self.set('desiredAdminState', null);
        switch(desired_admin_state) {
          case "INSERVICE":
            // can be decommissioned if already started
            self.set('isComponentRecommissionAvailable', false);
            self.set('isComponentDecommissioning', false);
            self.set('isComponentDecommissionAvailable', self.get('isStart'));
            break;
          case "DECOMMISSIONED":
            var deferred = $.Deferred();
            self.getDecommissionStatus().done( function() {
              var curObj = self.get('decommissionedStatusObject');
              self.set('decommissionedStatusObject', null);
              if (curObj && curObj.rm_metrics) {
                var nodeManagersArray = App.parseJSON(curObj.rm_metrics.cluster.nodeManagers);
                if (nodeManagersArray.findProperty('HostName', hostName)){
                  // decommisioning ..
                  self.set('isComponentRecommissionAvailable', true);
                  self.set('isComponentDecommissioning', true);
                  self.set('isComponentDecommissionAvailable', false);
                } else {
                  // decommissioned ..
                  self.set('isComponentRecommissionAvailable', true);
                  self.set('isComponentDecommissioning', false);
                  self.set('isComponentDecommissionAvailable', false);
                }
              }
              deferred.resolve(curObj);
            });
            break;
        }
        dfd.resolve(desired_admin_state);
      });
      return dfd.promise();
    }

  });
  
});
window.require.register("views/main/host/details/host_component_views/regionserver_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.RegionServerComponentView = App.HostComponentView.extend(App.Decommissionable, {

    componentForCheckDecommission: 'HBASE_MASTER',
    /**
     * load Recommission/Decommission status of RegionServer
     */
    loadComponentDecommissionStatus: function () {
      var hostName = this.get('content.host.hostName');
      var slaveType = 'HBASE_REGIONSERVER';
      var self = this;
      var deferred = $.Deferred();
      self.getDesiredAdminState().done( function () {
        var desired_admin_state = self.get('desiredAdminState');
        self.set('desiredAdminState', null);
        switch(desired_admin_state) {
          case "INSERVICE":
            self.set('isComponentRecommissionAvailable', false);
            self.set('isComponentDecommissioning', false);
            self.set('isComponentDecommissionAvailable', self.get('isStart'));
            break;
          case "DECOMMISSIONED":
            self.set('isComponentRecommissionAvailable', true);
            self.set('isComponentDecommissioning', self.get('isStart'));
            self.set('isComponentDecommissionAvailable', false);
            break;
        }
        deferred.resolve(desired_admin_state);
      });
      return deferred.promise();
    }

  });
  
});
window.require.register("views/main/host/details/host_component_views/tasktracker_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.TaskTrackerComponentView = App.HostComponentView.extend(App.Decommissionable, {

    componentForCheckDecommission: 'JOBTRACKER',

    /**
     * load Recommission/Decommission status for TaskTracker from JobTracker/AliveNodes list
     */
    loadComponentDecommissionStatus: function () {
      var hostName = this.get('content.host.hostName');
      var dfd = $.Deferred();
      var self = this;
      this.getDesiredAdminState().done( function () {
        var desired_admin_state = self.get('desiredAdminState');
        self.set('desiredAdminState', null);
        switch(desired_admin_state) {
          case "INSERVICE":
            // can be decommissioned if already started
            self.set('isComponentRecommissionAvailable', false);
            self.set('isComponentDecommissioning', false);
            self.set('isComponentDecommissionAvailable', self.get('isStart'));
            break;
          case "DECOMMISSIONED":
            var deferred = $.Deferred();
            self.getDecommissionStatus().done( function() {
              var curObj = self.get('decommissionedStatusObject');
              self.set('decommissionedStatusObject', null);
              if (curObj) {
                var aliveNodesArray = App.parseJSON(curObj.AliveNodes);
                if (aliveNodesArray != null) {
                  if (aliveNodesArray.findProperty('hostname', hostName)){
                    //decommissioning ..
                    self.set('isComponentRecommissionAvailable', true);
                    self.set('isComponentDecommissioning', true);
                    self.set('isComponentDecommissionAvailable', false);
                  } else {
                    //decommissioned
                    self.set('isComponentRecommissionAvailable', true);
                    self.set('isComponentDecommissioning', false);
                    self.set('isComponentDecommissionAvailable', false);
                  }
                }

              }
              deferred.resolve(curObj);
            });
            break;
        }
        dfd.resolve(desired_admin_state);
      });
      return dfd.promise();
    }

  });
  
});
window.require.register("views/main/host/hosts_table_menu_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.HostTableMenuView = Em.View.extend({

    templateName: require('templates/main/host/bulk_operation_menu'),

    menuItems: function () {
      return {
        s: {label: Em.I18n.t('hosts.table.menu.l1.selectedHosts')},
        f: {label: Em.I18n.t('hosts.table.menu.l1.filteredHosts')},
        a: {label: Em.I18n.t('hosts.table.menu.l1.allHosts')}
      };
    }.property('App.router.clusterController.isLoaded'),

    components: function(){
      var menuItems = [
      Em.Object.create({
        serviceName: 'HDFS',
        componentName: 'DATANODE',
        masterComponentName: 'NAMENODE',
        componentNameFormatted: Em.I18n.t('dashboard.services.hdfs.datanodes')
      }),
      Em.Object.create({
        serviceName: 'YARN',
        componentName: 'NODEMANAGER',
        masterComponentName: 'RESOURCEMANAGER',
        componentNameFormatted: Em.I18n.t('dashboard.services.yarn.nodeManagers')
      }),
      Em.Object.create({
        serviceName: 'HBASE',
        componentName: 'HBASE_REGIONSERVER',
        masterComponentName: 'HBASE_MASTER',
        componentNameFormatted: Em.I18n.t('dashboard.services.hbase.regionServers')
      }),
      Em.Object.create({
        serviceName: 'MAPREDUCE',
        componentName: 'TASKTRACKER',
        masterComponentName: 'JOBTRACKER',
        componentNameFormatted: Em.I18n.t('dashboard.services.mapreduce.taskTrackers')
      }),
      Em.Object.create({
        serviceName: 'STORM',
        componentName: 'SUPERVISOR',
        masterComponentName: 'SUPERVISOR',
        componentNameFormatted: Em.I18n.t('dashboard.services.storm.supervisors')
      })];

      return menuItems.filter(function(item){
        return App.Service.find().findProperty('serviceName',item.serviceName);
      });
    }.property(),


    /**
     * slaveItemView build second-level menu
     * for slave component
     */

    slaveItemView: Em.View.extend({

      tagName: 'li',

      classNames: ['dropdown-submenu'],

      /**
       * Get third-level menu items ingo for slave components
       * operationData format:
       * <code>
       *  {
       *    action: 'STARTED|INSTALLED|RESTART|DECOMMISSION|DECOMMISSION_OFF', // action for selected host components
       *    message: 'some text', // just text to BG popup
       *    componentName: 'DATANODE|NODEMANAGER...', //component name that should be processed
       *    realComponentName: 'DATANODE|NODEMANAGER...', // used only for decommission(_off) actions
       *    serviceName: 'HDFS|YARN|HBASE...', // service name of the processed component
       *    componentNameFormatted: 'DataNodes|NodeManagers...' // "user-friendly" string with component name (used in BG popup)
       *  }
       *  </code>
       *
       * @returns {Array}
       */

      operationsInfo: function () {
        var content = this.get('content');
        var menuItems = Em.A([
          Em.Object.create({
            label: Em.I18n.t('common.start'),
            operationData: Em.Object.create({
              action: App.HostComponentStatus.started,
              message: Em.I18n.t('common.start'),
              componentName: content.componentName,
              serviceName: content.serviceName,
              componentNameFormatted: content.componentNameFormatted
            })
          }),
          Em.Object.create({
            label: Em.I18n.t('common.stop'),
            operationData: Em.Object.create({
              action: App.HostComponentStatus.stopped,
              message: Em.I18n.t('common.stop'),
              componentName: content.componentName,
              serviceName: content.serviceName,
              componentNameFormatted: content.componentNameFormatted
            })
          }),
          Em.Object.create({
            label: Em.I18n.t('common.restart'),
            operationData: Em.Object.create({
              action: 'RESTART',
              message: Em.I18n.t('common.restart'),
              componentName: content.componentName,
              serviceName: content.serviceName,
              componentNameFormatted: content.componentNameFormatted
            })
          })
        ]);
        if (App.get('components.decommissionAllowed').contains(content.componentName)) {
          menuItems.pushObject(Em.Object.create({
            label: Em.I18n.t('common.decommission'),
            decommission: true,
            operationData: Em.Object.create({
              action: 'DECOMMISSION',
              message: Em.I18n.t('common.decommission'),
              componentName: content.masterComponentName,
              realComponentName: content.componentName,
              serviceName: content.serviceName,
              componentNameFormatted: content.componentNameFormatted
            })
          }));
          menuItems.pushObject(Em.Object.create({
            label: Em.I18n.t('common.recommission'),
            decommission: true,
            operationData: Em.Object.create({
              action: 'DECOMMISSION_OFF',
              message: Em.I18n.t('common.recommission'),
              componentName: content.masterComponentName,
              realComponentName: content.componentName,
              serviceName: content.serviceName,
              componentNameFormatted: content.componentNameFormatted
            })
          }));
        }
        return menuItems;
      }.property("content"),

      /**
       * commonOperationView is used for third-level menu items
       * for simple operations ('START','STOP','RESTART')
       */
      commonOperationView: Em.View.extend({
        tagName: 'li',

        /**
         * click function use
         * App.MainHostView as a thirl level parent
         * and runs it's function
         */
        click: function () {
          this.get('parentView.parentView.parentView').bulkOperationConfirm(this.get('content'), this.get('selection'));
        }
      }),

      /**
       * advancedOperationView is used for third level menu item
       * for advanced operations ('RECOMMISSION','DECOMMISSION')
       */
      advancedOperationView: Em.View.extend({
        tagName: 'li',
        rel: 'menuTooltip',
        classNameBindings: ['disabledElement'],
        attributeBindings: ['tooltipMsg:data-original-title'],

        service: function () {
          return App.router.get('mainServiceController.content').findProperty('serviceName', this.get('content.serviceName'))
        }.property('App.router.mainServiceController.content.@each', 'content'),

        tooltipMsg: function () {
          return (this.get('disabledElement') == 'disabled') ?
             Em.I18n.t('hosts.decommission.tooltip.warning').format(this.get('content.message'),  App.format.role(this.get('content.componentName'))) : '';
        }.property('disabledElement','content.componentName'),

        disabledElement: function () {
          return (this.get('service.workStatus') != 'STARTED') ? 'disabled' : '';
        }.property('service.workStatus'),

        /**
         * click function use
         * App.MainHostView as a thirl level parent
         * and runs it's function
         */
        click: function () {
          if (this.get('disabledElement') == 'disabled') {
            return;
          }
          this.get('parentView.parentView.parentView').bulkOperationConfirm(this.get('content'), this.get('selection'));
        },

        didInsertElement: function () {
          App.tooltip($(this.get('element')));
        }
      })
    }),

    /**
     * hostItemView build second-level menu
     * for host
     */

    hostItemView: Em.View.extend({

      tagName: 'li',

      classNames: ['dropdown-submenu'],

      label: Em.I18n.t('common.hosts'),

      /** Get third-level menu items for Hosts
       * operationData format:
       * <code>
       *  {
       *    action: 'STARTED|INSTALLED|RESTART..', // action for selected hosts (will be applied for each host component in selected hosts)
       *    actionToCheck: 'INSTALLED|STARTED..' // state to filter host components should be processed
       *    message: 'some text', // just text to BG popup
       *  }
       *  </code>
       * @returns {Array}
       */
      operationsInfo: function () {
        return Em.A([
          Em.Object.create({
            label: Em.I18n.t('hosts.host.details.startAllComponents'),
            operationData: Em.Object.create({
              action: 'STARTED',
              actionToCheck: 'INSTALLED',
              message: Em.I18n.t('hosts.host.details.startAllComponents')
            })
          }),
          Em.Object.create({
            label: Em.I18n.t('hosts.host.details.stopAllComponents'),
            operationData: Em.Object.create({
              action: 'INSTALLED',
              actionToCheck: 'STARTED',
              message: Em.I18n.t('hosts.host.details.stopAllComponents')
            })
          }),
          Em.Object.create({
            label: Em.I18n.t('hosts.table.menu.l2.restartAllComponents'),
            operationData: Em.Object.create({
              action: 'RESTART',
              message: Em.I18n.t('hosts.table.menu.l2.restartAllComponents')
            })
          }),
          Em.Object.create({
            label: Em.I18n.t('passiveState.turnOn'),
            operationData: Em.Object.create({
              state: 'ON',
              action: 'PASSIVE_STATE',
              message: Em.I18n.t('passiveState.turnOnFor').format('hosts')
            })
          }),
          Em.Object.create({
            label: Em.I18n.t('passiveState.turnOff'),
            operationData: Em.Object.create({
              state: 'OFF',
              action: 'PASSIVE_STATE',
              message: Em.I18n.t('passiveState.turnOffFor').format('hosts')
            })
          })
        ]);
      }.property(),

      /**
       * commonOperationView is used for third-level menu items
       * for all operations for host
       */
      operationView: Em.View.extend({
        tagName: 'li',

        /**
         * click function use
         * App.MainHostView as a thirl level parent
         * and runs it's function
         */
        click: function () {
          this.get('parentView.parentView.parentView').bulkOperationConfirm(this.get('content'), this.get('selection'));
        }
      })
    })
  });
  
});
window.require.register("views/main/host/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainHostMenuView = Em.CollectionView.extend({
    tagName: 'ul',
    classNames: ["nav", "nav-tabs"],
    content: function () {
      var array = [ {
        label: 'Summary',
        routing: 'summary'
      }
      /* { label:'Audit', routing:'audit'} */
      ];
      if (App.supports.hostOverridesHost) {
        array.push({
          label: 'Configs',
          routing: 'configs'
        });
      }
      return array;
    }.property(''),

    init: function(){ this._super(); this.activateView(); },

    activateView:function () {
      $.each(this._childViews, function () {
        this.set('active', (this.get('content.routing') == 'summary' ? "active" : ""));
      });
    },

    deactivateChildViews: function() {
      $.each(this._childViews, function(){
        this.set('active', "");
      });
    },

    itemViewClass: Em.View.extend({
      classNameBindings: ["active"],
      active: "",
      template: Ember.Handlebars.compile('<a {{action hostNavigate view.content.routing }} href="#"> {{unbound view.content.label}}</a>')
    })
  });
});
window.require.register("views/main/host/metrics", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainHostMetricsView = Em.View.extend({
    templateName: require('templates/main/host/metrics'),
    content:function(){
      return App.router.get('mainHostDetailsController.content');
    }.property('App.router.mainHostDetailsController.content')
  });
});
window.require.register("views/main/host/metrics/cpu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing Host CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartHostMetricsCPU = App.ChartLinearTimeView.extend({
    id: "host-metrics-cpu",
    title: Em.I18n.t('hosts.host.metrics.cpu'),
    yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,

    ajaxIndex: 'host.metrics.cpu',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.cpu) {
        var cpu_idle;
        for ( var name in jsonData.metrics.cpu) {
          var displayName;
          var seriesData = jsonData.metrics.cpu[name];
          switch (name) {
            case "cpu_wio":
              displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_wio');
              break;
            case "cpu_idle":
              displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_idle');
              break;
            case "cpu_nice":
              displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_nice');
              break;
            case "cpu_aidle":
              displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_aidle');
              break;
            case "cpu_system":
              displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_system');
              break;
            case "cpu_user":
              displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_user');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            if (Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_idle') == s.name) {
              cpu_idle = s;
            }
            else {
              seriesArray.push(s);
            }
          }
        }
        seriesArray.push(cpu_idle);
      }
      return seriesArray;
    },

    colorForSeries: function (series) {
      if (Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_idle') == series.name) {
        return '#CFECEC';
      }
      return null;
    }
  });
});
window.require.register("views/main/host/metrics/disk", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing host disk usage
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartHostMetricsDisk = App.ChartLinearTimeView.extend({
    id: "host-metrics-disk",
    title: Em.I18n.t('hosts.host.metrics.disk'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'host.metrics.disk',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var GB = Math.pow(2, 30);
      if (jsonData && jsonData.metrics && jsonData.metrics.disk) {
        if(jsonData.metrics.part_max_used){
          jsonData.metrics.disk.part_max_used = jsonData.metrics.part_max_used;
        }
        for ( var name in jsonData.metrics.disk) {
          var displayName;
          var seriesData = jsonData.metrics.disk[name];
          switch (name) {
            case "disk_total":
              displayName = Em.I18n.t('hosts.host.metrics.disk.displayNames.disk_total');
              break;
            case "disk_free":
              displayName = Em.I18n.t('hosts.host.metrics.disk.displayNames.disk_free');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= GB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/host/metrics/load", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing host load
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartHostMetricsLoad = App.ChartLinearTimeView.extend({
    id: "host-metrics-load",
    title: Em.I18n.t('hosts.host.metrics.load'),
    renderer: 'line',

    ajaxIndex: 'host.metrics.load',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.load) {
        for ( var name in jsonData.metrics.load) {
          var displayName;
          var seriesData = jsonData.metrics.load[name];
          switch (name) {
            case "load_fifteen":
              displayName = Em.I18n.t('hosts.host.metrics.load.displayNames.load_fifteen');
              break;
            case "load_one":
              displayName = Em.I18n.t('hosts.host.metrics.load.displayNames.load_one');
              break;
            case "load_five":
              displayName = Em.I18n.t('hosts.host.metrics.load.displayNames.load_five');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/host/metrics/memory", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing host memory metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartHostMetricsMemory = App.ChartLinearTimeView.extend({
    id: "host-metrics-memory",
    title: Em.I18n.t('hosts.host.metrics.memory'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'host.metrics.memory',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var KB = Math.pow(2, 10);
      if (jsonData && jsonData.metrics && jsonData.metrics.memory) {
        for ( var name in jsonData.metrics.memory) {
          var displayName;
          var seriesData = jsonData.metrics.memory[name];
          switch (name) {
            case "mem_shared":
              displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_shared');
              break;
            case "swap_free":
              displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.swap_free');
              break;
            case "mem_buffers":
              displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_buffers');
              break;
            case "mem_free":
              displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_free');
              break;
            case "mem_cached":
              displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_cached');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= KB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/host/metrics/network", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing host network metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartHostMetricsNetwork = App.ChartLinearTimeView.extend({
    id: "host-metrics-network",
    title: Em.I18n.t('hosts.host.metrics.network'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'host.metrics.network',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.network) {
        for ( var name in jsonData.metrics.network) {
          var displayName;
          var seriesData = jsonData.metrics.network[name];
          switch (name) {
            case "pkts_out":
              displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.pkts_out');
              break;
            case "bytes_in":
              displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.bytes_in');
              break;
            case "bytes_out":
              displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.bytes_out');
              break;
            case "pkts_in":
              displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.pkts_in');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/host/metrics/processes", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing host process counts
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartHostMetricsProcesses = App.ChartLinearTimeView.extend({
    id: "host-metrics-processes",
    title: Em.I18n.t('hosts.host.metrics.processes'),
    renderer: 'line',

    ajaxIndex: 'host.metrics.processes',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.process) {
        for ( var name in jsonData.metrics.process) {
          var displayName;
          var seriesData = jsonData.metrics.process[name];
          switch (name) {
            case "proc_total":
              displayName = Em.I18n.t('hosts.host.metrics.processes.displayNames.proc_total');
              break;
            case "proc_run":
              displayName = Em.I18n.t('hosts.host.metrics.processes.displayNames.proc_run');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/host/summary", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainHostSummaryView = Em.View.extend({

    templateName: require('templates/main/host/summary'),

    /**
     * List of custom view for some host components
     * @type {Em.Object}
     * Format:
     *  <code>
     *    {
     *      COMPONENT_NAME1: VIEW1,
     *      COMPONENT_NAME2: VIEW2
     *      ....
     *    }
     *  </code>
     */
    hostComponentViewMap: Em.Object.create({
      'DATANODE': App.DataNodeComponentView,
      'NODEMANAGER': App.NodeManagerComponentView,
      'HBASE_REGIONSERVER': App.RegionServerComponentView,
      'TASKTRACKER': App.TaskTrackerComponentView
    }),

    /**
     * @type {bool}
     */
    isStopCommand:true,

    /**
     * @type {App.Host}
     */
    content: function () {
      return App.router.get('mainHostDetailsController.content');
    }.property('App.router.mainHostDetailsController.content'),

    showGangliaCharts: function () {
      var name = this.get('content.hostName');
      var gangliaMobileUrl = App.router.get('clusterController.gangliaUrl') + "/mobile_helper.php?show_host_metrics=1&h=" + name + "&c=HDPNameNode&r=hour&cs=&ce=";
      window.open(gangliaMobileUrl);
    },

    /**
     * Message for "restart" block
     * @type {String}
     */
    needToRestartMessage: function() {
      var componentsCount, word;
      componentsCount = this.get('content.componentsWithStaleConfigsCount');
      if (componentsCount > 1) {
        word = Em.I18n.t('common.components').toLowerCase();
      } else {
        word = Em.I18n.t('common.component').toLowerCase();
      }
      return Em.I18n.t('hosts.host.details.needToRestart').format(this.get('content.componentsWithStaleConfigsCount'), word);
    }.property('content.componentsWithStaleConfigsCount'),

    /**
     * Reset <code>sortedComponents</code>
     * Used when some component was deleted from host
     */
    redrawComponents: function() {
      if (App.router.get('mainHostDetailsController.redrawComponents')) {
        this.set('sortedComponents', []);
        this.sortedComponentsFormatter();
        App.router.set('mainHostDetailsController.redrawComponents', false);
      }
    }.observes('App.router.mainHostDetailsController.redrawComponents'),

    willInsertElement: function() {
      this.set('sortedComponents', []);
      this.sortedComponentsFormatter();
    },

    didInsertElement: function () {
      this.addToolTip();
    },

    /**
     * Create tooltip for "Add" button if nothing to add to the current host
     */
    addToolTip: function() {
      if (this.get('addComponentDisabled')) {
        App.tooltip($('#add_component'), {title: Em.I18n.t('services.nothingToAdd')});
      }
    }.observes('addComponentDisabled'),

    /**
     * List of installed services
     * @type {String[]}
     */
    installedServices: function () {
      return App.Service.find().mapProperty('serviceName');
    }.property('App.router.clusterController.dataLoadList.serviceMetrics'),

    /**
     * List of installed masters and slaves
     * Masters first, then slaves
     * @type {App.HostComponent[]}
     */
    sortedComponents: [],

    /**
     * Update <code>sortedComponents</code>
     * Master components first, then slaves
     */
    sortedComponentsFormatter: function() {
      var updatebleProperties = Em.A(['workStatus', 'passiveState', 'staleConfigs', 'haStatus']);
      var self = this;
      var hostComponentViewMap = this.get('hostComponentViewMap');
      // Remove deleted components
      this.get('sortedComponents').forEach(function(sortedComponent, index) {
        if (!self.get('content.hostComponents').findProperty('id', sortedComponent.get('id'))) {
          self.get('sortedComponents').removeAt(index, 1);
        }
      });

      this.get('content.hostComponents').forEach(function (component) {
        if (component.get('isMaster') || component.get('isSlave')) {
          var obj = this.get('sortedComponents').findProperty('id', component.get('id'));
          if (obj) {
            // Update existing component
            updatebleProperties.forEach(function(property) {
              obj.set(property, component.get(property));
            });
          }
          else {
            // Add new component
            component.set('view', hostComponentViewMap[component.get('componentName')] ? hostComponentViewMap[component.get('componentName')] : App.HostComponentView);
            if (component.get('isMaster')) {
              // Masters should be before slaves
              var lastMasterIndex = 0, atLeastOneMasterExists = false;
              this.get('sortedComponents').forEach(function(sortedComponent, index) {
                if (sortedComponent.get('isMaster')) {
                  lastMasterIndex = index;
                  atLeastOneMasterExists = true;
                }
              });
              this.get('sortedComponents').insertAt(atLeastOneMasterExists ? lastMasterIndex + 1 : 0, component);
            }
            else {
              // it is slave 100%
              this.get('sortedComponents').pushObject(component);
            }
          }
        }
      }, this);
    }.observes('content.hostComponents.length'),

    /**
     * List of installed clients
     * @type {App.HostComponent[]}
     */
    clients: function () {
      var clients = [];
      this.get('content.hostComponents').forEach(function (component) {
        if (!component.get('componentName')) {
          //temporary fix because of different data in hostComponents and serviceComponents
          return;
        }
        if (!component.get('isSlave') && !component.get('isMaster')) {
          if (clients.length) {
            clients[clients.length - 1].set('isLast', false);
          }
          component.set('isLast', true);
          clients.push(component);
        }
      }, this);
      return clients;
    }.property('content.hostComponents.length'),

    /**
     * Check if some clients have stale configs
     * @type {bool}
     */
    areClientWithStaleConfigs: function() {
      return !!this.get('clients').filter(function(component) {
        return component.get('staleConfigs');
      }).length;
    }.property('clients.@each.staleConfigs'),

    /**
     * Template for addable component
     * @type {Em.Object}
     */
    addableComponentObject: Em.Object.extend({
      componentName: '',
      subComponentNames: null,
      displayName: function () {
        if (this.get('componentName') === 'CLIENTS') {
          return this.t('common.clients');
        }
        return App.format.role(this.get('componentName'));
      }.property('componentName')
    }),

    /**
     * If host lost heartbeat, components can't be added on it
     * @type {bool}
     */
    isAddComponent: function () {
      return this.get('content.healthClass') !== 'health-status-DEAD-YELLOW';
    }.property('content.healthClass'),

    /**
     * Disable "Add" button if components can't be added to the current host
     * @type {bool}
     */
    addComponentDisabled: function() {
      return (!this.get('isAddComponent')) || (this.get('addableComponents.length') == 0);
    }.property('isAddComponent', 'addableComponents.length'),

    /**
     * List of client's that may be installed to the current host
     * @type {String[]}
     */
    installableClientComponents: function() {
      if (!App.supports.deleteHost) {
        return [];
      }
      var componentServiceMap = App.QuickDataMapper.componentServiceMap();
      var allClients = App.get('components.clients');
      var installedServices = this.get('installedServices');
      var installedClients = this.get('clients').mapProperty('componentName');
      return allClients.filter(function(componentName) {
        // service for current client is installed but client isn't installed on current host
        return installedServices.contains(componentServiceMap[componentName]) && !installedClients.contains(componentName);
      });
    }.property('content.hostComponents.length', 'installedServices.length', 'App.supports.deleteHost'),

    /**
     * List of components that may be added to the current host
     * @type {Em.Object[]}
     */
    addableComponents: function () {
      var components = [];
      var self = this;
      var installableClients = this.get('installableClientComponents');
      var installedComponents = this.get('content.hostComponents').mapProperty('componentName');
      var addableToHostComponents = App.get('components.addableToHost');
      var installedServices = this.get('installedServices');
      var componentServiceMap = App.QuickDataMapper.componentServiceMap();

      addableToHostComponents.forEach(function(addableComponent) {
        if(installedServices.contains(componentServiceMap[addableComponent]) && !installedComponents.contains(addableComponent)) {
          components.pushObject(self.addableComponentObject.create({'componentName': addableComponent}));
        }
      });
      if (installableClients.length > 0) {
        components.pushObject(this.addableComponentObject.create({ 'componentName': 'CLIENTS', subComponentNames: installableClients }));
      }

      return components;
    }.property('content.hostComponents.length', 'installableClientComponents'),

    /**
     * Formatted with <code>$.timeago</code> value of host's last heartbeat
     * @type {String}
     */
    timeSinceHeartBeat: function () {
      var d = this.get('content.lastHeartBeatTime');
      if (d) {
        return $.timeago(d);
      }
      return "";
    }.property('content.lastHeartBeatTime')
  });
  
});
window.require.register("views/main/jobs/hive_job_details_tez_dag_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');
  var numberUtils = require('utils/number_utils');
  var stringUtils = require('utils/string_utils');

  App.MainHiveJobDetailsTezDagView = Em.View.extend({
    templateName : require('templates/main/jobs/hive_job_details_tez_dag'),
    selectedVertex : null,
    summaryMetricType: null,
    svgVerticesLayer : null, // The contents of the <svg> element.
    svgTezRoot: null,
    svgWidth : -1,
    svgHeight : -1,

    // zoomScaleFom: -1, // Bound from parent view
    // zoomScaleTo: -1, // Bound from parent view
    // zoomScale: -1, // Bound from parent view
    zoomTranslate: [0, 0],
    zoomBehavior: null,
    svgCreated: false,

    content : null,

    /**
     * Populated by #drawTezDag() 
     * 
     * {
     *   "nodes": [
     *     {
     *       "id": "Map2",
     *       "name": "Map 2",
     *       "type": App.TezDagVertexType.MAP,
     *       "operations": [
     *         "TableScan",
     *         "File Output"
     *       ],
     *       "depth": 1,
     *       "parents": [],
     *       "children": [],
     *       "x": 0,
     *       "y": 0,
     *       "metricDisplay": "100MB",
     *       "metricPercent": 64,
     *       "metricType": "Input",
     *       "selected": true,
     *       "fixed": true,
     *       "metrics": {
     *         "input": 40022,
     *         "output": 224344,
     *         "recordsRead": 200,
     *         "recordsWrite": 122,
     *         "tezTasks": 2
     *       }
     *     }
     *   ],
     *   "links": [
     *     {
     *       "source": {},
     *       "target": {},
     *       "edgeType": "BROADCAST"
     *     }
     *   ]
     * }
     */
    dagVisualModel : {
      nodes : [],
      links : [],
      maxMetrics : {},
      minMetrics: {}
    },

    didInsertElement : function() {
      this._super();
      this.createSvg();
    },

    willDestroyElement : function() {
      $('.svg-tooltip').tooltip('destroy');
    },

    createSvg : function() {
      var self = this;
      var dagVisualModel = this.get('dagVisualModel');
      dagVisualModel.nodes.clear();
      dagVisualModel.links.clear();
      dagVisualModel.maxMetrics = {};
      dagVisualModel.minMetrics = {};

      this.set('content', this.get('controller.content'));
      var svg = d3.select("#tez-dag-svg");
      d3.selectAll(".tez-dag-canvas").remove();
      var tezRoot = svg.append("svg:g").attr("class", "tez-root");
      this.set('svgTezRoot', tezRoot);
      var tezRootRect = tezRoot.append("rect").attr("class", "tez-root-rect");
      this.set('svgVerticesLayer', tezRoot.append("svg:g").attr("class", "tez-dag-canvas"));
      this.adjustGraphHeight();
      var canvasSize = this.drawTezDag();
      var minScale = Math.min(this.get('svgHeight') / canvasSize.height, this.get('svgWidth') / canvasSize.width);
      if (minScale > 1) {
        minScale = 1;
      }
      tezRootRect.attr("width", canvasSize.width).attr("height", canvasSize.height);
      var zoom = d3.behavior.zoom().scaleExtent([ minScale, 2 ]).on("zoom", function() {
        tezRoot.attr("transform", "translate(" + (d3.event.translate) + ")scale(" + d3.event.scale + ")");
        self.set('zoomScale', d3.event.scale);
        self.set('zoomTranslate', d3.event.translate);
      });
      svg.call(zoom);
      this.set('zoomBehavior', zoom);
      this.set('zoomTranslate', [0, 0]);
      this.set('zoomScaleFrom', minScale);
      this.set('zoomScaleTo', 2);
      this.set('zoomScale', minScale);
      this.set('svgCreated', true);
    },

    zoomScaleObserver : function() {
      var tezRoot = this.get("svgTezRoot");
      var newScale = this.get('zoomScale');
      var newScaleFrom = this.get('zoomScaleFrom');
      var newScaleTo = this.get('zoomScaleTo');
      var zoomTranslate = this.get('zoomTranslate');
      var zoomBehavior = this.get('zoomBehavior');
      if (d3.event == null && this.get('svgCreated')) {
        // Values were set from actions instead of UI events
        // We need to center in on selected vertex if available.
        var selectedNode = null;
        var dagVisualModel = this.get('dagVisualModel');
        if (dagVisualModel && dagVisualModel.nodes && dagVisualModel.nodes.length > 0) {
          dagVisualModel.nodes.every(function(node) {
            if (node.selected) {
              selectedNode = node;
              return false;
            }
            return true;
          })
        }
        if (selectedNode != null) {
          var cX = selectedNode.x + selectedNode.width / 2;
          var cY = selectedNode.y + selectedNode.height / 2;
          var mX = (cX * zoomBehavior.scale()) + zoomTranslate[0];
          var mY = (cY * zoomBehavior.scale()) + zoomTranslate[1];
          var pX = (cX * newScale) + zoomTranslate[0];
          var pY = (cY * newScale) + zoomTranslate[1];
          var nX = (mX - pX);
          var nY = (mY - pY);
          zoomTranslate[0] += nX;
          zoomTranslate[1] += nY;
          this.set('zoomTranslate', zoomTranslate);
        }
      }
      console.debug("zoomScaleObserver(): New scale = " + newScale + ", Range = [" + newScaleFrom + ", " + newScaleTo + "]. Translate = ", zoomTranslate.join(','));
      zoomBehavior.scale(newScale);
      zoomBehavior.translate(zoomTranslate);
      tezRoot.attr("transform", "translate(" + zoomTranslate + ")scale(" + newScale + ")");
    }.observes('zoomScale', 'zoomScaleFrom', 'zoomScaleTo', 'zoomTranslate'),

    /**
     * We have to make the height of the DAG section match the height of the
     * Summary section.
     */
    adjustGraphHeight : function() {
      var rhsDiv = document.getElementById('tez-vertices-rhs');
      var lhsDiv = document.getElementById('tez-dag-section');
      if (lhsDiv && rhsDiv) {
        var rhsHeight = rhsDiv.clientHeight - 26; // box boundary
        var currentWidth = lhsDiv.clientWidth;
        var currentHeight = lhsDiv.clientHeight;
        $(lhsDiv).attr('style', "height:" + rhsHeight + "px;");
        var svgHeight = rhsHeight - 20;
        d3.select("#tez-dag-svg").attr('height', svgHeight).attr('width', '100%');
        this.set('svgWidth', currentWidth);
        this.set('svgHeight', svgHeight);
        console.log("SVG Width=", currentWidth, ", Height=", svgHeight);
      }
    },

    vertexSelectionUpdated : function() {
      var vertexId = this.get('selectedVertex.id');
      console.log("vertexSelectionUpdated(): Selected ",vertexId);
      var dagVisualModel = this.get('dagVisualModel');
      if (dagVisualModel && dagVisualModel.nodes && dagVisualModel.nodes.length > 0) {
        dagVisualModel.nodes.forEach(function(node) {
          node.selected = node.id == vertexId;
          console.log("vertexSelectionUpdated(): Updated  ",node.id," to ",node.selected);
        })
      }
      this.refreshGraphUI();
    }.observes('selectedVertex'),

    summaryMetricTypeUpdated : function() {
      var summaryMetricType = this.get('summaryMetricType');
      var dagVisualModel = this.get('dagVisualModel');
      var min = dagVisualModel.minMetrics[summaryMetricType];
      var max = dagVisualModel.maxMetrics[summaryMetricType];
      dagVisualModel.nodes.forEach(function(node) {
        var value = node.metrics[summaryMetricType];
        var percent = -1;
        if (numberUtils.validateInteger(value)==null && value >= 0) {
          if (numberUtils.validateInteger(min) == null && numberUtils.validateInteger(max) == null) {
            if (max > min && value >= 0) {
              percent = Math.round((value - min) * 100 / (max - min));
            }
          }
        } else {
          value = '';
        }
        switch (summaryMetricType) {
        case "input":
        case "output":
          value = numberUtils.bytesToSize(value);
          break;
        default:
          break;
        }
        node.metricType = Em.I18n.t('jobs.hive.tez.metric.' + summaryMetricType);
        node.metricDisplay = value;
        node.metricPercent = percent;
      });
      this.refreshGraphUI();
    }.observes('summaryMetricType'),

    /**
     * Observes metrics of all vertices.
     */
    vertexMetricsUpdated : function() {
      var dagVisualModel = this.get('dagVisualModel');
      dagVisualModel.minMetrics = {
        input : Number.MAX_VALUE,
        output : Number.MAX_VALUE,
        recordsRead : Number.MAX_VALUE,
        recordsWrite : Number.MAX_VALUE,
        tezTasks : Number.MAX_VALUE,
        spilledRecords : Number.MAX_VALUE
      };
      dagVisualModel.maxMetrics = {
        input : 0,
        output : 0,
        recordsRead : 0,
        recordsWrite : 0,
        tezTasks : 0,
        spilledRecords : 0
      };
      if (dagVisualModel.nodes) {
        dagVisualModel.nodes.forEach(function(node) {
          var vertex = App.TezDagVertex.find(node.id);
          if (vertex) {
            node.metrics['input'] = vertex.get('fileReadBytes') + vertex.get('hdfsReadBytes');
            node.metrics['output'] = vertex.get('fileWriteBytes') + vertex.get('hdfsWriteBytes');
            node.metrics['recordsRead'] = vertex.get('recordReadCount');
            node.metrics['recordsWrite'] = vertex.get('recordWriteCount');
            node.metrics['tezTasks'] = vertex.get('tasksCount');
            node.metrics['spilledRecords'] = vertex.get('spilledRecords');
            node.state = vertex.get('state');
            // Min metrics
            dagVisualModel.minMetrics.input = Math.min(dagVisualModel.minMetrics.input, node.metrics.input);
            dagVisualModel.minMetrics.output = Math.min(dagVisualModel.minMetrics.output, node.metrics.output);
            dagVisualModel.minMetrics.recordsRead = Math.min(dagVisualModel.minMetrics.recordsRead, node.metrics.recordsRead);
            dagVisualModel.minMetrics.recordsWrite = Math.min(dagVisualModel.minMetrics.recordsWrite, node.metrics.recordsWrite);
            dagVisualModel.minMetrics.tezTasks = Math.min(dagVisualModel.minMetrics.tezTasks, node.metrics.tezTasks);
            dagVisualModel.minMetrics.spilledRecords = Math.min(dagVisualModel.minMetrics.spilledRecords, node.metrics.spilledRecords);
            // Max metrics
            dagVisualModel.maxMetrics.input = Math.max(dagVisualModel.maxMetrics.input, node.metrics.input);
            dagVisualModel.maxMetrics.output = Math.max(dagVisualModel.maxMetrics.output, node.metrics.output);
            dagVisualModel.maxMetrics.recordsRead = Math.max(dagVisualModel.maxMetrics.recordsRead, node.metrics.recordsRead);
            dagVisualModel.maxMetrics.recordsWrite = Math.max(dagVisualModel.maxMetrics.recordsWrite, node.metrics.recordsWrite);
            dagVisualModel.maxMetrics.tezTasks = Math.max(dagVisualModel.maxMetrics.tezTasks, node.metrics.tezTasks);
            dagVisualModel.maxMetrics.spilledRecords = Math.max(dagVisualModel.maxMetrics.spilledRecords, node.metrics.spilledRecords);
          }
        });
      }
      Ember.run.once(this, 'summaryMetricTypeUpdated');
    }.observes('content.tezDag.vertices.@each.fileReadBytes', 'content.tezDag.vertices.@each.fileWriteBytes', 
        'content.tezDag.vertices.@each.hdfsReadBytes', 'content.tezDag.vertices.@each.hdfsWriteBytes', 
        'content.tezDag.vertices.@each.recordReadCount', 'content.tezDag.vertices.@each.recordWriteCount',
        'content.tezDag.vertices.@each.state', 'content.tezDag.vertices.@each.spilledRecords'),

    createOperationPlanObj: function (vertexName, op, opIndex) {
      var operatorPlanObj = [];
      var text = this.get('content.tezDag.vertices').findProperty('name', vertexName).get('operationPlan');
      text = text.replace(/:"/g,'"').replace(/([:,])(?=\S)/g,'$1 ');
      var jsonText =  $.parseJSON(text);
      if (!opIndex) {
        opIndex = 0;
      } else {
        opIndex = parseInt(opIndex) - 1;
      }
      var jsonText = op.findIn(jsonText, opIndex);
      if (jsonText!=null) {
        for (var key in jsonText) {
          if (jsonText.hasOwnProperty(key) && typeof(jsonText[key]) == "string") {
            operatorPlanObj.push(
                {
                  name: key,
                  value: jsonText[key]
                }
            );
          }
        }
      }
      return operatorPlanObj;
    },

    /**
     * Determines layout and creates Tez graph. In the process it populates the
     * visual model into 'dagVisualModel' field.
     *
     * Terminology: 'vertices' and 'edges' are Tez terms. 'nodes' and 'links' are
     * visual (d3) terms.
     */
    drawTezDag : function() {
      var self = this;
      var width = this.get('svgWidth');
      var svgLayer = this.get('svgVerticesLayer');
      var vertices = this.get('content.tezDag.vertices');
      var edges = this.get('content.tezDag.edges');
      var constants = this.get('constants');
      var vertexIdToNode = {};
      var depthToNodes = []; // Array of id arrays
      var dagVisualModel = this.get('dagVisualModel');
      var selectedVertex = this.get('selectedVertex');
      var minVertexDuration = Number.MAX_VALUE;

      //
      // CALCULATE DEPTH - BFS to get correct graph depth
      //
      var visitEdges = [];
      var maxRowLength = 0;
      var maxRowDepth = 0;
      vertices.forEach(function(v) {
        if (v.get('incomingEdges.length') < 1) {
          visitEdges.push({
            depth : 0,
            parent : null,
            toVertex : v
          });
        }
      });
      function getNodeFromEdge(edgeObj) {
        var vertex = edgeObj.toVertex;
        var pName = edgeObj.parent ? edgeObj.parent.name : null;
        var cName = edgeObj.toVertex ? edgeObj.toVertex.get('name') : null;
        console.debug("Processing vertex ", edgeObj, " (",pName, " > ", cName,")");
        if(edgeObj.parent && edgeObj.depth < edgeObj.parent.depth + 1){
          console.debug("Updating child edge to " + (edgeObj.parent.depth + 1));
          edgeObj.depth = edgeObj.parent.depth + 1;
        }
        var node = vertexIdToNode[vertex.get('id')];
        for ( var k = depthToNodes.length; k <= edgeObj.depth; k++) {
          depthToNodes.push([]);
        }
        if (!node) {
          // New node
          node = {
            id : vertex.get('id'),
            name : vertex.get('name'),
            state : vertex.get('state'),
            type : vertex.get('type'),
            operations : vertex.get('operations'),
            depth : edgeObj.depth,
            parents : [],
            children : [],
            x : 0,
            y : 0,
            metricType : null,
            metricDisplay : null,
            metricPercent : -1,
            selected : selectedVertex != null ? selectedVertex.get('id') == vertex.get('id') : false,
            fixed : true,
            metrics : {
              input : -1,
              output : -1,
              recordsRead : -1,
              recordsWrite : -1,
              tezTasks : -1
            },
            duration: vertex.get('duration')
          }
          if (node.duration < minVertexDuration && node.duration > 0) {
            minVertexDuration = node.duration;
          }
          vertexIdToNode[vertex.get('id')] = node;
          depthToNodes[node.depth].push(node);
        } else {
          // Existing node
          if (edgeObj.depth > node.depth) {
            function moveNodeToDepth(node, newDepth) {
              console.debug("Moving " + node.name + " from depth " + node.depth + " to " + newDepth);
              var oldIndex = depthToNodes[node.depth].indexOf(node);
              depthToNodes[node.depth].splice(oldIndex, 1);
              node.depth = newDepth;
              if (!depthToNodes[node.depth]) {
                depthToNodes[node.depth] = [];
              }
              depthToNodes[node.depth].push(node);
              if (node.children) {
                // Move children down depth
                node.children.forEach(function(c) {
                  moveNodeToDepth(c, node.depth + 1);
                })
              }
            }
            moveNodeToDepth(node, edgeObj.depth);
          }
        }
        if (depthToNodes[node.depth].length > maxRowLength) {
          maxRowLength = depthToNodes[node.depth].length;
          maxRowDepth = node.depth;
        }
        if (edgeObj.parent != null) {
          node.parents.push(edgeObj.parent);
          edgeObj.parent.children.push(node);
        }
        return node;
      }
      var edgeObj;
      var visitedVertexMap = {};
      while (edgeObj = visitEdges.shift()) {
        var node = getNodeFromEdge(edgeObj);
        if (!visitedVertexMap[edgeObj.toVertex.get('id')]) {
          visitedVertexMap[edgeObj.toVertex.get('id')] = true;
          var outEdges = edgeObj.toVertex.get('outgoingEdges');
          outEdges.forEach(function(oe) {
            var childVertex = oe.get('toVertex');
            visitEdges.push({
              depth : node.depth + 1,
              parent : node,
              toVertex : childVertex
            });
          });
        }
      }
      edges.forEach(function(e) {
        dagVisualModel.links.push({
          source : vertexIdToNode[e.get('fromVertex.id')],
          target : vertexIdToNode[e.get('toVertex.id')],
          edgeType : e.get('edgeType')
        });
      });
      // Sort nodes so that parents stay together
      for ( var depth = 0; depth < depthToNodes.length; depth++) {
        var nodes = depthToNodes[depth];
        nodes.sort(function(n1, n2) {
          var ck1 = '';
          var ck2 = '';
          if (n1.children) {
            n1.children.forEach(function(c) {
              ck1 += c.name;
            });
          }
          if (n2.children) {
            n2.children.forEach(function(c) {
              ck2 += c.name;
            });
          }
          if (ck1 < ck2) {
            return -1
          }
          if (ck1 > ck2) {
            return 1
          }
          return 0
        });
        depthToNodes[depth] = nodes;
      }

      //
      // LAYOUT - Now with correct depth, we calculate layouts
      //
      // When a node's effective width changes, all its parent nodes are updated.
      var updateNodeEffectiveWidth = function(node, newEffectiveWidth) {
        console.debug("Updating effective width of (" + node.id + ") to " + newEffectiveWidth);
        if (numberUtils.validateInteger(node.effectiveWidth) != null) {
          node.effectiveWidth = newEffectiveWidth;
        }
        var diff = newEffectiveWidth - node.effectiveWidth;
        if (diff > 0) {
          var oldEffectiveWidth = node.effectiveWidth;
          node.effectiveWidth = newEffectiveWidth;
          if (node.parents != null) {
            node.parents.forEach(function(parent) {
              updateNodeEffectiveWidth(parent, parent.effectiveWidth + diff);
            })
          }
        }
      }
      var xGap = 20;
      var yGap = 70;
      var currentY = 40;
      // First pass - calculate layout widths, and Y coordinates
      for ( var depth = 0; depth < depthToNodes.length; depth++) {
        var nodes = depthToNodes[depth];
        var maxNodeHeight = 0;
        for ( var nodeIndex = 0; nodeIndex < nodes.length; nodeIndex++) {
          var node = nodes[nodeIndex];
          var nodeDim = this.getNodeCalculatedDimensions(node, minVertexDuration);
          node.drawWidth = nodeDim.drawWidth;
          node.drawHeight = nodeDim.drawHeight;
          node.scale = nodeDim.scale;
          node.width = nodeDim.width;
          node.height = nodeDim.height;
          if (maxNodeHeight < node.height) {
            maxNodeHeight = node.height;
          }
          if (depth == 0) {
            // Top nodes - position uniformly
            updateNodeEffectiveWidth(node, xGap + node.width);
          }
          if (node.children && node.children.length > 0) {
            // There can be dedicated or shared children.
            // Dedicated children increase effective width of parent by their
            // width.
            // Shared children increase effective width of parent only by the
            // fraction of parentage
            var childrenWidth = 0;
            node.children.forEach(function(child) {
              var childDim = self.getNodeCalculatedDimensions(child, minVertexDuration);
              childrenWidth += ((childDim.width + xGap) / child.parents.length);
            });
            updateNodeEffectiveWidth(node, Math.max(childrenWidth, (node.width+xGap)));
          } else {
            updateNodeEffectiveWidth(node, xGap + node.width);
          }
          node.y = currentY;
          node.incomingY = node.y;
          node.outgoingY = node.incomingY + node.height;
        }
        currentY += maxNodeHeight;
        currentY += yGap;
      }
      // Second pass - determine actual X coordinates
      var maxX = 0;
      for ( var depth = 0; depth < depthToNodes.length; depth++) {
        var nodes = depthToNodes[depth];
        var currentX = -1;
        var parentCurrentXMap = {};
        for ( var nodeIndex = 0; nodeIndex < nodes.length; nodeIndex++) {
          var node = nodes[nodeIndex];
          var parentsKey = null;
          if (node.parents != null && node.parents.length > 0) {
            var parentMidX = 0;
            var parentsKey = '';
            var childrenEffectiveWidth = -1;
            node.parents.forEach(function(parent) {
              parentMidX += (parent.x + (parent.width / 2));
              parentsKey += (parent.id + '//');
              if (childrenEffectiveWidth < 0) {
                parent.children.forEach(function(c){
                  childrenEffectiveWidth += (c.effectiveWidth);
                });
              }
            });
            parentMidX = parentMidX / node.parents.length;
            var parentCurrentX = parentCurrentXMap[parentsKey];
            if (parentCurrentX == null || parentCurrentX == undefined) {
              parentCurrentX = parentMidX - (childrenEffectiveWidth/2);
              parentCurrentXMap[parentsKey] = parentCurrentX;
            }
            currentX = parentCurrentX;
          } else {
            if (currentX < 0) {
              currentX = 0;
            }
          }
          node.x = (currentX + (node.effectiveWidth - node.width) / 2);
          node.outgoingX = (node.x + node.width / 2);
          node.incomingX = node.outgoingX;
          console.log("drawTezDag(). Layout Node: ", node);
          dagVisualModel.nodes.push(node);
          if (parentsKey != null) {
            parentCurrentXMap[parentsKey] = currentX + node.effectiveWidth;
          } else {
            currentX += node.effectiveWidth;
          }
          if ((node.x + node.width) > maxX) {
            maxX = node.x + node.width;
          }
        }
      }
      var canvasHeight = currentY;
      var canvasWidth = maxX + (xGap << 1);

      //
      // Draw SVG
      //
      var force = d3.layout.force().nodes(dagVisualModel.nodes).links(dagVisualModel.links).start();
      var nodeDragData = {
        nodeRelativeX : 0,
        nodeRelativeY : 0
      };
      var nodeDrag = d3.behavior.drag().on('dragstart', function(node){
        d3.event.sourceEvent.stopPropagation();
        var rc = d3.mouse(this);
        nodeDragData.nodeRelativeX = (rc[0] * node.scale);
        nodeDragData.nodeRelativeY = (rc[1] * node.scale);
      }).on('drag', function(node){
        var nx = d3.event.x - nodeDragData.nodeRelativeX;
        var ny = d3.event.y - nodeDragData.nodeRelativeY;
        self.dragVertex(d3.select(this), node, [nx, ny], diagonal);
      }).on('dragend', function(){
        nodeDragData.nodeRelativeX = 0;
        nodeDragData.nodeRelativeY = 0;
      });
      // Create Links
      var diagonal = d3.svg.diagonal().source(function(d) {
        return {
          x : d.source.outgoingX,
          y : d.source.outgoingY
        };
      }).target(function(d) {
        return {
          x : d.target.incomingX,
          y : d.target.incomingY - 12
        }
      });
      var link = svgLayer.selectAll(".link-g").data(dagVisualModel.links).enter().append("g").attr("class", "link-g").attr("marker-end", "url(#arrow)");
      link.append("path").attr("class", function(l) {
        var classes = "link svg-tooltip ";
        if (l.edgeType) {
          classes += ("type-" + l.edgeType.toLowerCase() + " ");
        } else {
          classes += "type-unknown ";
        }
        return classes;
      }).attr("d", diagonal).attr("title", function(l) {
        var lower = l.edgeType ? l.edgeType.toLowerCase() : '';
        return Em.I18n.t("jobs.hive.tez.edge." + lower);
      });
      // Create Nodes
      var node = svgLayer.selectAll(".node").data(dagVisualModel.nodes).enter().append("g").attr("class", "node").call(nodeDrag);
      node.append("rect").attr("class", "background").attr("width", function(n) {
        return n.drawWidth;
      }).attr("height", function(n) {
        return n.drawHeight;
      }).attr("rx", "10").attr("filter", "url(#shadow)").on('mousedown', function(n) {
        var vertex = App.TezDagVertex.find(n.id);
        if (vertex != null) {
          self.get('parentView').doSelectVertex({
            context : vertex
          });
        }
      });
      node.each(function(n, nodeIndex) {
        var ops = n.operations;
        var opCount = {};
        if (ops != null && ops.length > 0) {
          var opGroups = d3.select(this).selectAll(".operation").data(ops).enter().append("g").attr("class", "operation").attr("transform", function(op, opIndex) {
            var row = Math.floor(opIndex / 3);
            var column = opIndex % 3;
            return "translate(" + (10 + column * 55) + "," + (37 + row * 20) + ")";
          }).attr("clip-path", "url(#operatorClipPath)").attr("opIndex", function(op){
            if(!opCount[op]) {
              opCount[op] = 1;
            } else {
              opCount[op] = opCount[op]+1;
            }
            return opCount[op];
          }).on('mouseover', function(op) {
            var viewContent = {
              operationName: op,
              operatorPlanObj: []
            };
            var opIndex = this.getAttribute('opIndex');
            var operatorPlanObj = self.createOperationPlanObj(n.name, op, opIndex);
            viewContent.operatorPlanObj = operatorPlanObj;
            var template = App.HoverOpTable.create({content: viewContent}) ;
            $(this).find('.svg-tooltip').attr('title', template.renderToBuffer().string()).tooltip('fixTitle').tooltip('show');
            })

          opGroups.append("rect").attr("class", "operation svg-tooltip ").attr("width", "50").attr("height", "16");
          opGroups.append("text").attr("x", "2").attr("dy", "1em").text(function(op) {
            return op != null ? op.split(' ')[0] : '';
          });
        }
      });
      var metricNodes = node.append("g").attr("class", "metric").attr("transform", "translate(112,7)");
      metricNodes.append("rect").attr("width", function(n) {
        if (n.type == App.TezDagVertexType.UNION) {
          return 0;
        }
        return 60;
      }).attr("height", function(n) {
        if (n.type == App.TezDagVertexType.UNION) {
          return 0;
        }
        return 18;
      }).attr("rx", "3").attr("class", "metric-title svg-tooltip");
      metricNodes.append("text").attr("class", "metric-text").attr("x", "2").attr("dy", "1em");
      node.append("text").attr("x", "1.9em").attr("dy", "1.5em").text(function(d) {
        return d.name;
      });
      var iconContainer = node.append('g').attr('class', 'vertex-icon-container').attr('transform', 'translate(10,10)');
      iconContainer.append('rect').attr('width', '1em').attr('height', '1em').attr('class', 'vertex-icon-rect  svg-tooltip ');
      iconContainer.append('text').attr('dy', '10px').attr("font-family", "FontAwesome").attr('class', 'vertex-icon-text');
      node.attr("transform", function(d) {
        return "translate(" + d.x + "," + d.y + ") scale("+d.scale+") ";
      });
      this.vertexMetricsUpdated();
      $('.svg-tooltip').each(function() {
        var item = $(this);
        if (item.prop('tagName') == 'path') {
          item.hover(function(e) {
            var offset = $(this).offset();
            item.prop('offsetWidth', function() {
              return 2 * (e.pageX - offset.left);
            });
            item.prop('offsetHeight', function() {
              return 2 * (e.pageY - offset.top);
            });
          });
        };
        if (item.prop('offsetWidth') == undefined) {
          item.prop('offsetWidth', function() {
            return item.width();
          });
        };
        if (item.prop('offsetHeight') == undefined) {
          item.prop('offsetHeight', function() {
            return item.height();
          });
        };
      });
      App.tooltip($('.svg-tooltip'), {
        placement : 'left',
        template: '<div class="tooltip jobs-tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>'
      });

      if (App.supports.debugJobsDag) {
        // Draws node bounding box - for debug purposes
        node.append("rect").attr("width", function(n) {
          return n.effectiveWidth;
        }).attr("height", function(n) {
          return n.height;
        }).attr("x", function(n) {
          return -1 * ((n.effectiveWidth - n.width) / 2);
        }).attr("y", function(n) {
          return 0;
        }).attr("style", "opacity: 0.2;fill:yellow;");
      }

      // Position in center
      var translateX = Math.round((width - canvasWidth) / 2);
      if (translateX > 0) {
        svgLayer.attr("transform", "translate("+translateX+",0)");
      }
      return {
        width : canvasWidth,
        height : canvasHeight
      }
    },

    dragVertex: function(d3Vertex, node, newPosition, diagonal){
      console.debug("Dragging vertex [", node.name, "] to (", newPosition[0], ",", newPosition[1], ")");
      // Move vertex
      node.x = newPosition[0];
      node.y = newPosition[1];
      node.incomingX = newPosition[0] + (node.width/2);
      node.incomingY = newPosition[1];
      node.outgoingX = newPosition[0] + (node.width/2);
      node.outgoingY = newPosition[1] + node.height;
      d3Vertex.attr('transform', 'translate(' + newPosition[0] + ',' + newPosition[1] + ') scale('+node.scale+') ');
      // Move links
      d3.selectAll('.link').filter(function(l) {
        if (l && (l.source === node || l.target === node)) {
          return this
        }
        return null;
      }).attr('d', diagonal);
    },

    /**
     * Refreshes UI of the Tez graph with latest values
     */
    refreshGraphUI: function () {
      var svgLayer = this.get('svgVerticesLayer');
      if (svgLayer!=null) {
        var self = this;
        var metricNodes = svgLayer.selectAll(".metric");
        var metricNodeTexts = svgLayer.selectAll(".metric-text");
        var metricNodeTitles = svgLayer.selectAll(".metric-title");
        var nodeBackgrounds =svgLayer.selectAll(".background");
        var vertexIconTexts = svgLayer.selectAll(".vertex-icon-text");
        var vertexIconRects = svgLayer.selectAll(".vertex-icon-rect");
        metricNodes.attr("class", function(node) {
          var classes = "metric ";
          var percent = node.metricPercent;
          if (numberUtils.validateInteger(percent) == null && percent >= 0) {
            if (percent <= 20) {
              classes += "heat-0-20 ";
            } else if (percent <= 40) {
              classes += "heat-20-40 ";
            } else if (percent <= 60) {
              classes += "heat-40-60 ";
            } else if (percent <= 80) {
              classes += "heat-60-80 ";
            } else if (percent <= 100) {
              classes += "heat-80-100 ";
            } else {
              classes += "heat-none";
            }
          } else {
            classes += "heat-none";
          }
          return classes;
        });
        metricNodeTexts.text(function(node){
          if (node.type == App.TezDagVertexType.UNION) {
            return '';
          }
          return node.metricDisplay;
        });
        metricNodeTitles.attr("title", function(node){
          return node.metricType;
        }).attr("data-original-title", function(node){
          return node.metricType;
        });
        nodeBackgrounds.attr("class", function(n) {
          var classes = "background ";
          if (n.type) {
            classes += (n.type.toLowerCase() + " ");
          } else {
            classes += "unknown-vertex-type ";
          }
          if (n.selected) {
            classes += "selected ";
          }
          return classes;
        });
        vertexIconRects.attr('title', function(node) {
          return stringUtils.getCamelCase(node.state);
        }).attr('data-original-title', function(node) {
          return stringUtils.getCamelCase(node.state);
        });
        vertexIconTexts.text(function(n) {
          return self.getVertexIcon(n)
        }).attr('class', function(n) {
          var classes = 'vertex-icon-text ';
          if (n.state != null) {
            classes += n.state.toLowerCase();
          }
          return classes;
        });
      }
    },

    getVertexIcon : function(node){
      var icon = "";
      switch (node.state) {
      case App.TezDagVertexState.NEW:
        icon = '\uF10C'; //icon-circle-blank
      case App.TezDagVertexState.RUNNING:
      case App.TezDagVertexState.FAILED:
        icon = '\uF111'; //icon-circle
        break;
      case App.TezDagVertexState.SUCCEEDED:
        icon = '\uF00C'; //icon-ok
        break;
      case App.TezDagVertexState.KILLED:
      case App.TezDagVertexState.ERROR:
        icon = '\uF057'; //icon-remove-sign
        break;
      case App.TezDagVertexState.INITED:
      case App.TezDagVertexState.INITIALIZING:
      case App.TezDagVertexState.TERMINATING:
        icon = '\uF141'; //icon-ellipsis-horizontal
        break;
      }
      return icon;
    },

    /**
     * Determines the size of a node by taking into account its duration and
     * number of operations performed.
     *
     * @return {Object} Provides various metrics necessary in drawing a node.
     * <code>
     * {
     *  width: 360, // Scaled width of the node
     *  height: 80, // Scaled height of the node
     *  scale: 2, // Scale used on vertex dimensions. Quickest vertex is scaled to 1.
     *  drawWidth: 180, // Width of actual drawing (that will be scaled)
     *  drawHeight: 40 // Height of actual drawing (that will be scaled)
     * }
     * </code>
     */
    getNodeCalculatedDimensions : function(node, minVertexDuration) {
      var size = {
        width : 180,
        height : 40,
        drawWidth : 180,
        drawHeight : 40,
        scale : 1
      }
      if (node.operations && node.operations.length > 0) {
        var opsHeight = Math.ceil(node.operations.length / 3);
        size.drawHeight += (opsHeight * 20);
      }
      size.scale = (minVertexDuration < Number.MAX_VALUE && node.duration > 0) ? (node.duration / minVertexDuration) : 1;
      if (size.scale < 1) {
        size.scale = 1;
      }
      size.scale = Math.sqrt(size.scale);
      size.width = size.drawWidth * size.scale;
      size.height = size.drawHeight * size.scale;
      return size;
    }

  });

  App.HoverOpTable = Ember.View.extend({
    templateName : require('templates/main/jobs/hover_op_table')
  });
});
window.require.register("views/main/jobs/hive_job_details_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var date = require('utils/date');
  var numberUtils = require('utils/number_utils');
  var dateUtils = require('utils/date');
  var stringUtils = require('utils/string_utils');
  var sort = require('views/common/sort_view');

  App.MainHiveJobDetailsView = Em.View.extend({
    templateName : require('templates/main/jobs/hive_job_details'),

    selectedVertex : null,
    content : null,
    zoomScaleFrom : 1,
    zoomScaleTo: 2,
    zoomScale : 1,
    showQuery : false,

    zoomStep : function() {
      var zoomStep = 0.01;
      var zoomFrom = this.get('zoomScaleFrom');
      var zoomTo = this.get('zoomScaleTo');
      if (zoomFrom < zoomTo) {
        zoomStep = (zoomTo - zoomFrom) / 5;
      }
      return zoomStep;
    }.property('zoomScaleFrom', 'zoomScaleTo'),
    isGraphMaximized: false,

    toggleShowQuery : function () {
      this.toggleProperty('showQuery');
      var queryBlock =  $('.query-info');
      if (this.get('showQuery')) {
        queryBlock.slideDown();
      } else {
        queryBlock.slideUp();
      };
    },

    toggleShowQueryText : function () {
      return this.get('showQuery') ? Em.I18n.t('jobs.hive.less') : Em.I18n.t('jobs.hive.more');
    }.property('showQuery'),

    summaryMetricType: 'input',
    summaryMetricTypesDisplay : [
      Em.I18n.t('jobs.hive.tez.metric.input'),
      Em.I18n.t('jobs.hive.tez.metric.output'),
     /* Em.I18n.t('jobs.hive.tez.metric.recordsRead'),
      Em.I18n.t('jobs.hive.tez.metric.recordsWrite'), */
      Em.I18n.t('jobs.hive.tez.metric.tezTasks'),
      Em.I18n.t('jobs.hive.tez.metric.spilledRecords')
    ],
    summaryMetricTypeDisplay: function(){
      return Em.I18n.t('jobs.hive.tez.metric.'+this.get('summaryMetricType'));
    }.property('summaryMetricType'),

    sortedVertices : function() {
      var sortColumn = this.get('controller.sortingColumn');
      if(sortColumn && sortColumn.get('status')){
        var sortColumnStatus = sortColumn.get('status');
        var sorted = sortColumn.get('parentView').sort(sortColumn, sortColumnStatus === "sorting_desc", true);
        sortColumn.set('status', sortColumnStatus);
        return sorted;
      }
      var vertices = this.get('controller.content.tezDag.vertices');
      if (vertices != null) {
        vertices = vertices.toArray();
        return vertices;
      }
      return vertices;
    }.property('content.tezDag.vertices','controller.sortingColumn'),

    initialDataLoaded : function() {
      var loaded = this.get('controller.loaded');
      if (loaded) {
        this.set('content', this.get('controller.content'));
      }
    }.observes('controller.loaded'),

    yarnApplicationIdLink : function() {
      var yarnService = App.YARNService.find().objectAt(0);
      var appId = this.get('content.tezDag.yarnApplicationId');
      if (yarnService != null) {
        var quickLinksView = App.QuickViewLinks.create();
        try {
          quickLinksView.set('content', yarnService);
          quickLinksView.setQuickLinks();
          var links = quickLinksView.get('quickLinks');
          if (links && links.length > 0) {
            var rmUILink = links.findProperty('label', 'ResourceManager UI');
            if (rmUILink != null) {
              return rmUILink.get('url') + '/cluster/app/' + appId;
            }
          }
        } finally {
          quickLinksView.destroy();
          quickLinksView = null;
        }
      }
      return null;
    }.property('content.tezDag.yarnApplicationId', 'App.YARNService.resourceManagerNode.hostName'),

    jobObserver : function() {
      var content = this.get('content');
      var selectedVertex = this.get('selectedVertex');
      if (selectedVertex == null && content != null) {
        var vertices = content.get('tezDag.vertices');
        if (vertices) {
          vertices.setEach('isSelected', false);
          this.doSelectVertex({
            context : vertices.objectAt(0)
          });
        }
      }
    }.observes('selectedVertex', 'content.tezDag.vertices.@each.id'),

    doSelectVertex : function(event) {
      var newVertex = event.context;
      var currentVertex = this.get('selectedVertex');
      if (currentVertex != null) {
        currentVertex.set('isSelected', false);
      }
      newVertex.set('isSelected', true);
      this.set('selectedVertex', newVertex);
    },

    doSelectSummaryMetricType: function(event) {
      var summaryType = event.context;
      switch (summaryType) {
      case Em.I18n.t('jobs.hive.tez.metric.input'):
        summaryType = 'input';
        break;
      case Em.I18n.t('jobs.hive.tez.metric.output'):
        summaryType = 'output';
        break;
      case Em.I18n.t('jobs.hive.tez.metric.recordsRead'):
        summaryType = 'recordsRead';
        break;
      case Em.I18n.t('jobs.hive.tez.metric.recordsWrite'):
        summaryType = 'recordsWrite';
        break;
      case Em.I18n.t('jobs.hive.tez.metric.tezTasks'):
        summaryType = 'tezTasks';
        break;
      case Em.I18n.t('jobs.hive.tez.metric.spilledRecords'):
        summaryType = 'spilledRecords';
        break;
      default:
        break;
      }
      this.set('summaryMetricType', summaryType);
    },

    /**
     * Provides display information for vertex I/O.
     * 
     * {
     *  'file': {
     *    'read': {
     *      'ops': '100 reads',
     *      'bytes': '10 MB'
     *    }
     *    'write: {
     *      'ops': '200 writes',
     *      'bytes': '20 MB'
     *    }
     *  },
     *  'hdfs': {
     *    'read': {
     *      'ops': '100 reads',
     *      'bytes': '10 MB'
     *    }
     *    'write: {
     *      'ops': '200 writes',
     *      'bytes': '20 MB'
     *    }
     *  },
     *  'records': {
     *    'read': '100 records',
     *    'write': '123 records'
     *  },
     *  'started': 'Feb 12, 2014 10:30am',
     *  'ended': 'Feb 12, 2014 10:35am',
     *  'status': 'Running'
     * }
     */
    selectedVertexIODisplay : function() {
      var v = this.get('selectedVertex');
      var status = v.get('state');
      if (status) {
        status = stringUtils.getCamelCase(status);
      }
      var fileReadOps = v.get('fileReadOps');
      var fileWriteOps = v.get('fileWriteOps');
      var hdfsReadOps = v.get('hdfsReadOps');
      var hdfsWriteOps = v.get('hdfsWriteOps');
      var naString = Em.I18n.t('common.na');
      if (fileReadOps === null) {
        fileReadOps = naString;
      }
      if (fileWriteOps === null) {
        fileWriteOps = naString;
      }
      if (hdfsReadOps === null) {
        hdfsReadOps = naString;
      }
      if (hdfsWriteOps === null) {
        hdfsWriteOps = naString;
      }
      return {
        file : {
          read : {
            ops : Em.I18n.t('jobs.hive.tez.reads').format(fileReadOps),
            bytes : numberUtils.bytesToSize(v.get('fileReadBytes'))
          },
          write : {
            ops : Em.I18n.t('jobs.hive.tez.writes').format(fileWriteOps),
            bytes : numberUtils.bytesToSize(v.get('fileWriteBytes'))
          }
        },
        hdfs : {
          read : {
            ops : Em.I18n.t('jobs.hive.tez.reads').format(hdfsReadOps),
            bytes : numberUtils.bytesToSize(v.get('hdfsReadBytes'))
          },
          write : {
            ops : Em.I18n.t('jobs.hive.tez.writes').format(hdfsWriteOps),
            bytes : numberUtils.bytesToSize(v.get('hdfsWriteBytes'))
          }
        },
        records : {
          read : v.get('recordReadCount') == null ? null : Em.I18n.t('jobs.hive.tez.records.count').format(v.get('recordReadCount')),
          write : v.get('recordWriteCount') == null ? null : Em.I18n.t('jobs.hive.tez.records.count').format(v.get('recordWriteCount'))
        },
        started: v.get('startTime') ? dateUtils.dateFormat(v.get('startTime'), true, true) : '',
        ended: v.get('endTime') ? dateUtils.dateFormat(v.get('endTime'), true, true) : '',
        status: status
      };
    }.property('selectedVertex.fileReadOps', 'selectedVertex.fileWriteOps', 'selectedVertex.hdfsReadOps', 'selectedVertex.hdfdWriteOps',
        'selectedVertex.fileReadBytes', 'selectedVertex.fileWriteBytes', 'selectedVertex.hdfsReadBytes', 'selectedVertex.hdfdWriteBytes',
        'selectedVertex.recordReadCount', 'selectedVertex.recordWriteCount', 'selectedVertex.status'),

    canGraphZoomIn : function() {
      var zoomTo = this.get('zoomScaleTo');
      var zoomScale = this.get('zoomScale');
      console.debug("canGraphZoomIn? : ", (zoomScale < zoomTo), " (scaleTo=", zoomTo,", scale=",zoomScale,")");
      return zoomScale < zoomTo;
    }.property('zoomScale', 'zoomScaleTo'),

    canGraphZoomOut : function() {
      var zoomFrom = this.get('zoomScaleFrom');
      var zoomScale = this.get('zoomScale');
      console.debug("canGraphZoomOut? : ", (zoomScale > zoomFrom), " (scaleFrom=", zoomFrom,", scale=",zoomScale,")");
      return zoomScale > zoomFrom;
    }.property('zoomScale', 'zoomScaleFrom'),

    doGraphZoomIn: function() {
      var zoomTo = this.get('zoomScaleTo');
      var zoomScale = this.get('zoomScale');
      var zoomStep = this.get('zoomStep');
      if (zoomScale < zoomTo) {
        var step = Math.min(zoomStep, (zoomTo - zoomScale));
        zoomScale += step;
        console.debug("doGraphZoomIn(): New scale = ", zoomScale);
        this.set('zoomScale', zoomScale);
      }
    },

    doGraphZoomOut: function() {
      var zoomFrom = this.get('zoomScaleFrom');
      var zoomScale = this.get('zoomScale');
      var zoomStep = this.get('zoomStep');
      if (zoomScale > zoomFrom) {
        var step = Math.min(zoomStep, (zoomScale - zoomFrom));
        zoomScale -= step;
        console.debug("doGraphZoomOut(): New scale = ", zoomScale);
        this.set('zoomScale', zoomScale);
      }
     },

     doGraphMaximize: function() {
       this.set('isGraphMaximized', true);
     },

     doGraphMinimize: function() {
       this.set('isGraphMaximized', false);
     }
  });

  App.MainHiveJobDetailsVerticesTableView = App.TableView.extend({
    sortView: sort.wrapperView,

    didInsertElement: function () {
      if(!this.get('controller.sortingColumn')){
        var columns = this.get('childViews')[0].get('childViews');
        if(columns && columns.findProperty('name', 'name')){
          columns.findProperty('name','name').set('status', 'sorting_asc');
          this.get('controller').set('sortingColumn', columns.findProperty('name','name'))
        }
      }
    },

    nameSort: sort.fieldView.extend({
      column: 0,
      name: 'name',
      displayName: Em.I18n.t('common.name'),
      type: 'string'
    }),
    tasksSort: sort.fieldView.extend({
      column: 1,
      name: 'tasksNumber',
      displayName: Em.I18n.t('common.tasks'),
      type: 'number'
    }),
    inputSort: sort.fieldView.extend({
      column: 2,
      name: 'totalReadBytes',
      displayName: Em.I18n.t('apps.item.dag.input'),
      type: 'number'
    }),
    outputSort: sort.fieldView.extend({
      column: 3,
      name: 'totalWriteBytes',
      displayName: Em.I18n.t('apps.item.dag.output'),
      type: 'number'
    }),
    durationSort: sort.fieldView.extend({
      column: 4,
      name: 'duration',
      displayName: Em.I18n.t('apps.item.dag.duration'),
      type: 'number'
    })
  });
});
window.require.register("views/main/jobs/select_custom_date_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.JobsCustomDatesSelectView = Em.View.extend({
    name: 'jobsCustomDatesSelectView',
    templateName: require('templates/main/jobs/custom_dates_popup'),

    middayPeriodOptions: [Em.I18n.t('jobs.table.custom.date.am'), Em.I18n.t('jobs.table.custom.date.pm')],

    hourOptions: ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12'],

    minuteOptions: ['00', '05', '10', '15', '20', '25', '30', '35', '40', '45', '50', '55'],

    didInsertElement: function () {
      $('.datepicker').datepicker({
        format: 'mm/dd/yyyy'
      });
    }
  });
  
});
window.require.register("views/main/jobs_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');
  var sort = require('views/common/sort_view');

  App.MainJobsView = App.TableView.extend({
    templateName: require('templates/main/jobs'),

    content: [],


    /**
     * If no jobs table rows to show.
     */
    noDataToShow:true,

    filterCondition:[],

    /*
     If no jobs to display set noDataToShow to true, else set emptyData to false.
     */
    noDataToShowObserver: function () {
      if(this.get("controller.content.length") > 0){
        this.set("noDataToShow",false);
      }else{
        this.set("noDataToShow",true);
      }
    }.observes("controller.content.length"),

    willInsertElement: function () {
      var self = this;
      var name = this.get('controller.name');
      var colPropAssoc = this.get('colPropAssoc');
      var filterConditions = App.db.getFilterConditions(name);
      if (filterConditions) {
        this.set('filterConditions', filterConditions);
        var childViews = this.get('childViews');

        filterConditions.forEach(function(condition) {
          var view = childViews.findProperty('column', condition.iColumn);
          if (view) {
            //self.get('controller.filterObject').set(colPropAssoc[condition.iColumn], condition.value);
            view.set('value', condition.value);
            if(view.get('setPropertyOnApply')){
              view.setValueOnApply();
            }
            Em.run.next(function() {
              view.showClearFilter();
            });
          }
        });
      } else {
        this.clearFilters();
      }
      this.onApplyIdFilter();
      this.set('tableFilteringComplete', true);
    },

    didInsertElement: function () {
      if(!this.get('controller.sortingColumn')){
        var columns = this.get('childViews')[0].get('childViews')
        if(columns && columns.findProperty('name', 'startTime')){
          columns.findProperty('name','startTime').set('status', 'sorting_desc');
          this.get('controller').set('sortingColumn', columns.findProperty('name','startTime'))
        }
      }
    },

    onApplyIdFilter: function() {
      var isIdFilterApplied = this.get('controller.filterObject.isIdFilterApplied');
      this.get('childViews').forEach(function(childView) {
        if (childView['clearFilter'] && childView.get('column') != 1) {
          if(isIdFilterApplied){
            childView.clearFilter();
          }
          var childOfChild = childView.get('childViews')[0];
          if(childOfChild){
            Em.run.next(function() {
              childOfChild.set('disabled', isIdFilterApplied);
            })
          }
        }
      });
    }.observes('controller.filterObject.isIdFilterApplied'),

    saveFilter: function () {
      if(this.get('tableFilteringComplete')){
        this.updateFilter(1, this.get('controller.filterObject.id'), 'string');
        this.updateFilter(2, this.get('controller.filterObject.user'), 'string');
        this.updateFilter(4, this.get('controller.filterObject.windowEnd'), 'date');
      }
    }.observes(
      'controller.filterObject.id',
      'controller.filterObject.user',
      'controller.filterObject.windowEnd'
    ),

    sortView: sort.wrapperView,
    idSort: sort.fieldView.extend({
      column: 1,
      name: 'id',
      displayName: Em.I18n.t('jobs.column.id'),
      type: 'string'
    }),
    userSort: sort.fieldView.extend({
      column: 2,
      name: 'user',
      displayName: Em.I18n.t('jobs.column.user'),
      type: 'string'
    }),
    startTimeSort: sort.fieldView.extend({
      column: 3,
      name: 'startTime',
      displayName: Em.I18n.t('jobs.column.start.time'),
      type: 'number'
    }),
    endTimeSort: sort.fieldView.extend({
      column: 4,
      name: 'endTime',
      displayName: Em.I18n.t('jobs.column.end.time'),
      type: 'number'
    }),
    durationSort: sort.fieldView.extend({
      column: 5,
      name: 'duration',
      displayName: Em.I18n.t('jobs.column.duration'),
      type: 'number'
    }),

    /**
     * Select View with list of "rows-per-page" options
     * @type {Ember.View}
     */
    rowsPerPageSelectView: Em.Select.extend({
      content: ['10', '25', '50', '100', "250", "500"],
      valueBinding: "controller.filterObject.jobsLimit",
      attributeBindings: ['disabled'],
      disabled: false,
      disabledObserver: function () {
        if (this.get("parentView.hasBackLinks")) {
          this.set('disabled', true);
        }else{
          this.set('disabled', false);
        }
      }.observes('parentView.hasBackLinks'),
      change: function () {
        this.get('controller').set('navIDs.nextID', '');
        this.get('parentView').saveDisplayLength();
      }
    }),

    /**
     * return filtered number of all content number information displayed on the page footer bar
     * @returns {String}
     */
    filteredJobs: function () {
      return Em.I18n.t('jobs.filtered.jobs').format(this.get('controller.content.length'));
    }.property('controller.content.length', 'controller.totalOfJobs'),

    pageContentObserver: function () {
      if (!this.get('controller.loading')) {
        if ($('.tooltip').length) {
          Ember.run.later(this, function() {
            if ($('.tooltip').length > 1) {
              $('.tooltip').first().remove();
            };
          }, 500);
        };
      };
    }.observes('controller.loading'),

    init: function() {
      this._super();
      App.tooltip($('body'), {
        selector: '[rel="tooltip"]',
        template: '<div class="tooltip jobs-tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>',
        placement: 'bottom'
      });
    },

    willDestroyElement : function() {
      $('.tooltip').remove();
    },

    /**
     * Filter-field for Jobs ID.
     * Based on <code>filters</code> library
     */
    jobsIdFilterView: filters.createTextView({
      column: 1,
      showApply: true,
      setPropertyOnApply: 'controller.filterObject.id'
    }),

    /**
     * Filter-list for User.
     * Based on <code>filters</code> library
     */
    userFilterView: filters.createTextView({
      column: 2,
      fieldType: 'input-small',
      showApply: true,
      setPropertyOnApply: 'controller.filterObject.user'
    }),

    /**
     * Filter-field for Start Time.
     * Based on <code>filters</code> library
     */
    startTimeFilterView: filters.createSelectView({
      fieldType: 'input-120',
      column: 3,
      content: ['Any', 'Past 1 hour',  'Past 1 Day', 'Past 2 Days', 'Past 7 Days', 'Past 14 Days', 'Past 30 Days', 'Custom'],
      valueBinding: "controller.filterObject.startTime",
      onChangeValue: function () {
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'date');
      }
    }),

    jobNameView: Em.View.extend({
      classNames: ['job-link'],
      tagName: 'a',
      classNameBindings: ['isLink'],
      attributeBindings: ['rel', 'job.queryText:data-original-title', 'href'],
      rel: 'tooltip',
      href: '#',
      isLink: 'is-not-link',
      isLinkObserver: function () {
        if (this.get('job.hasTezDag')) {
          this.set('isLink', "");
        }else{
          this.set('isLink', "is-not-link");
        }
      }.observes('controller.sortingDone'),
      template: Ember.Handlebars.compile('{{job.name}}'),
      click: function(event) {
        if (this.get('job.hasTezDag')) {
          App.router.transitionTo('main.jobs.jobDetails', this.get('job'));
        };
        return false;
      }
    }),

    /**
     * associations between content (jobs list) property and column index
     */
    colPropAssoc: function () {
      var associations = [];
      associations[1] = 'id';
      associations[2] = 'user';
      associations[3] = 'startTime';
      associations[4] = 'endTime';
      return associations;
    }.property(),

    clearFilters: function() {
      this.get('childViews').forEach(function(childView) {
        if (childView['clearFilter']) {
          childView.clearFilter();
        }
      });
    },

    jobFailMessage: function() {
      return Em.I18n.t('jobs.table.job.fail');
    }.property(),

    jobsPaginationLeft: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-arrow-left"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if (this.get("parentView.hasBackLinks") && !this.get('controller.filterObject.isAnyFilterApplied')) {
          return "paginate_previous";
        }
        return "paginate_disabled_previous";
      }.property('parentView.hasBackLinks', 'controller.filterObject.isAnyFilterApplied'),

      click: function () {
        if (this.get("parentView.hasBackLinks") && !this.get('controller.filterObject.isAnyFilterApplied')) {
          this.get('controller').navigateBack();
        }
      }
    }),

    jobsPaginationRight: Ember.View.extend({
      tagName: 'a',
      template: Ember.Handlebars.compile('<i class="icon-arrow-right"></i>'),
      classNameBindings: ['class'],
      class: function () {
        if (this.get("parentView.hasNextJobs") && !this.get('controller.filterObject.isAnyFilterApplied')) {
          return "paginate_next";
        }
        return "paginate_disabled_next";
      }.property("parentView.hasNextJobs", 'controller.filterObject.isAnyFilterApplied'),

      click: function () {
        if (this.get("parentView.hasNextJobs") && !this.get('controller.filterObject.isAnyFilterApplied')) {
          this.get('controller').navigateNext();
        }
      }
    }),

    hasNextJobs: function() {
      return (this.get("controller.navIDs.nextID.length") > 0);
    }.property('controller.navIDs.nextID'),

    hasBackLinks: function() {
      return (this.get("controller.navIDs.backIDs").length > 1);
    }.property('controller.navIDs.backIDs.[].length')

  })
  
});
window.require.register("views/main/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * this menu extended by other with modifying content and itemViewClass.template
   * @type {*}
   */
  App.MainMenuView = Em.CollectionView.extend({
    tagName:'ul',
    classNames:['nav'],
    content:function(){
      var result = [
        { label:Em.I18n.t('menu.item.dashboard'), routing:'dashboard', active:'active'},
        { label:Em.I18n.t('menu.item.heatmaps'), routing:'charts'},
        { label:Em.I18n.t('menu.item.services'), routing:'services'},
        { label:Em.I18n.t('menu.item.hosts'), routing:'hosts'}
      ];

      if (App.supports.mirroring && App.Service.find().findProperty('serviceName', 'FALCON')) {
        result.push({ label:Em.I18n.t('menu.item.mirroring'), routing:'mirroring'});
      }

      if (!App.get('isHadoop2Stack')) {
        result.push({ label:Em.I18n.t('menu.item.jobs'), routing:'apps'});
      } else if( App.router.get('mainAdminController.isAccessAvailable') && App.supports.jobs
        && (App.router.get('mainAdminAccessController.showJobs') || App.get('isAdmin'))) {
        result.push({ label:Em.I18n.t('menu.item.jobs'), routing:'jobs'});
      }

      if (App.get('isAdmin')) {
        result.push({ label:Em.I18n.t('menu.item.admin'), routing:'admin'});
      }

      if (App.supports.views) {
        result.push({ label:Em.I18n.t('menu.item.views'), routing:'views'});
      }
      return result;
    }.property(),
      /**
       *    Adds observer on lastSetURL and calls navigation sync procedure
       */
    didInsertElement:function () {
      App.router.location.addObserver('lastSetURL', this, 'renderOnRoute');
      this.renderOnRoute();
    },

    /**
     *    Syncs navigation menu with requested URL
     */
    renderOnRoute:function () {
      var last_url = App.router.location.lastSetURL || location.href.replace(/^[^#]*#/, '');
      if (last_url.substr(1, 4) !== 'main' || !this._childViews) {
        return;
      }
      var reg = /^\/main\/([a-z]+)/g;
      var sub_url = reg.exec(last_url);
      var chunk = (null != sub_url) ? sub_url[1] : 'dashboard';
      $.each(this._childViews, function () {
        this.set('active', this.get('content.routing') == chunk ? "active" : "");
      });
    },

    itemViewClass:Em.View.extend({

      classNameBindings:['active', ':span2'],
      active:'',

      alertsCount:function () {
        if (this.get('content').routing == 'hosts') {
          return App.router.get('mainHostController.content').mapProperty('criticalAlertsCount')
            .reduce(function(pv, cv) { return pv + parseInt(cv); }, 0);
        }
      }.property('App.router.mainHostController.content.@each.criticalAlertsCount'),

      templateName: require('templates/main/menu_item')
    })
  });
  
});
window.require.register("views/main/mirroring/datasets_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');
  var sort = require('views/common/sort_view');

  App.MainDatasetsView = App.TableView.extend({
    name: 'mainDatasetsView',
    templateName: require('templates/main/mirroring/datasets'),
    content: function () {
      return this.get('controller.datasets');
    }.property('controller.datasets.length'),

    didInsertElement: function () {
      this.filter();
    },

    endIndex: function () {
      return this.get('filteredContent.length');
    }.property('filteredContent.length'),

    sortView: sort.wrapperView,
    nameSort: sort.fieldView.extend({
      column: '1',
      name: 'name',
      displayName: Em.I18n.t('common.name')
    }),

    statusSort: sort.fieldView.extend({
      column: '2',
      name: 'status',
      displayName: Em.I18n.t('mirroring.table.datasetStatus')
    }),

    /**
     * Filter view for name column
     * Based on <code>filters</code> library
     */
    nameFilterView: filters.createTextView({
      fieldType: 'input-small',
      column: 1,
      onChangeValue: function () {
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
      }
    }),

    statusFilterView: filters.createSelectView({
      fieldType: 'input-small',
      column: 2,
      content: ['Any', 'Scheduled', 'Suspended', 'Running'],
      onClearValue: function () {
        if (this.get('value') === '') {
          this.set('value', 'Any');
        }
      }.observes('value'),
      onChangeValue: function () {
        var value = this.get('value');
        if (value === 'Any') {
          value = '';
        }
        this.get('parentView').updateFilter(this.get('column'), value, 'string');
      }
    }),

    DatasetView: Em.View.extend({
      content: null,
      tagName: 'tr',

      click: function () {
        App.router.send('gotoShowJobs', this.get('content'));
      },

      classNameBindings: ['selectedClass', ':dataset-item'],

      selectedClass: function () {
        return this.get('controller.selectedDataset.id') === this.get('content.id') ? 'dataset-selected' : '';
      }.property('controller.selectedDataset'),

      lastDurationFormatted: function () {
        var milliseconds = this.get('content.lastDuration');
        var h = Math.floor(milliseconds / 3600000);
        var m = Math.floor((milliseconds % 3600000) / 60000);
        var s = Math.floor(((milliseconds % 360000) % 60000) / 1000);
        return (h == 0 ? '' : h + 'hr ') + (m == 0 ? '' : m + 'mins ') + (s == 0 ? '' : s + 'secs ');
      }.property('content.lastDuration'),

      lastSucceededDateFormatted: function () {
        if (this.get('content.lastSucceededDate')) {
          return $.timeago(this.get('content.lastSucceededDate'));
        }
      }.property('content.lastSucceededDate'),

      lastFailedDateFormatted: function () {
        if (this.get('content.lastFailedDate')) {
          return $.timeago(this.get('content.lastFailedDate'));
        }
      }.property('content.lastFailedDate')
    }),

    /**
     * associations between dataset property and column index
     */
    colPropAssoc: function () {
      var associations = [];
      associations[1] = 'name';
      associations[2] = 'status';
      return associations;
    }.property()

  });
  
});
window.require.register("views/main/mirroring/edit_dataset_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');
  var sort = require('views/common/sort_view');
  var date = require('utils/date');

  App.MainMirroringEditDataSetView = Em.View.extend({
    name: 'mainMirroringEditDataSetView',
    templateName: require('templates/main/mirroring/edit_dataset'),

    hasTargetClusters: false,

    targetClusters: App.TargetCluster.find(),

    targetClusterSelect: Em.Select.extend({
      classNames: ['target-cluster-select'],

      content: function () {
        if (!this.get('parentView.isLoaded')) return [];
        return this.get('parentView.targetClusters').mapProperty('name').without(App.get('clusterName')).concat(Em.I18n.t('mirroring.dataset.addTargetCluster'));
      }.property('parentView.isLoaded', 'parentView.targetClusters.length'),

      change: function () {
        if (this.get('selection') === Em.I18n.t('mirroring.dataset.addTargetCluster')) {
          this.set('selection', this.get('content')[0]);
          this.get('parentView').manageClusters();
        }
        this.set('parentView.controller.formFields.datasetTargetClusterName', this.get('selection'))
      }
    }),

    onTargetClustersChange: function () {
      if (this.get('isLoaded') && this.get('targetClusters.length') > 1) {
        this.set('hasTargetClusters', true);
      } else {
        this.set('hasTargetClusters', false);
        this.set('controller.formFields.datasetTargetClusterName', null);
      }
    }.observes('isLoaded', 'targetClusters.length'),

    repeatOptions: [Em.I18n.t('mirroring.dataset.repeat.minutes'), Em.I18n.t('mirroring.dataset.repeat.hours'), Em.I18n.t('mirroring.dataset.repeat.days'), Em.I18n.t('mirroring.dataset.repeat.months')],

    middayPeriodOptions: [Em.I18n.t('mirroring.dataset.middayPeriod.am'), Em.I18n.t('mirroring.dataset.middayPeriod.pm')],

    hourOptions: ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12'],

    minuteOptions: ['00', '05', '10', '15', '20', '25', '30', '35', '40', '45', '50', '55'],

    isLoaded: function () {
      return App.router.get('mainMirroringController.isLoaded');
    }.property('App.router.mainMirroringController.isLoaded'),

    manageClusters: function () {
      App.router.get('mainMirroringController').manageClusters();
    },

    fillForm: function () {
      var isEdit = this.get('controller.isEdit');
      if (this.get('isLoaded')  && isEdit) {
        var controller = this.get('controller');
        var dataset = App.Dataset.find().findProperty('id', controller.get('datasetIdToEdit'));
        var scheduleStartDate = new Date(dataset.get('scheduleStartDate'));
        var scheduleEndDate = new Date(dataset.get('scheduleEndDate'));
        var formFields = controller.get('formFields');
        formFields.set('datasetName', dataset.get('name'));
        formFields.set('datasetSourceDir', dataset.get('sourceDir'));
        formFields.set('datasetTargetDir', dataset.get('targetDir'));
        formFields.set('datasetTargetClusterName', dataset.get('targetClusterName'));
        formFields.set('datasetFrequency', dataset.get('frequency'));
        formFields.set('repeatOptionSelected', dataset.get('frequencyUnit'));
        formFields.set('datasetStartDate', controller.addZero(scheduleStartDate.getMonth() + 1) + '/' + controller.addZero(scheduleStartDate.getDate()) + '/' + controller.addZero(scheduleStartDate.getFullYear()));
        formFields.set('datasetEndDate', controller.addZero(scheduleEndDate.getMonth() + 1) + '/' + controller.addZero(scheduleEndDate.getDate()) + '/' + controller.addZero(scheduleEndDate.getFullYear()));
        var startHours = scheduleStartDate.getHours();
        var endHours = scheduleEndDate.getHours();
        formFields.set('hoursForStart', controller.toAMPMHours(startHours));
        formFields.set('hoursForEnd', controller.toAMPMHours(endHours));
        formFields.set('minutesForStart', controller.addZero(scheduleStartDate.getMinutes()));
        formFields.set('minutesForEnd', controller.addZero(scheduleEndDate.getMinutes()));
        formFields.set('middayPeriodForStart', startHours > 11 ? 'PM' : 'AM');
        formFields.set('middayPeriodForEnd', endHours > 11 ? 'PM' : 'AM');
      }
    }.observes('isLoaded', 'App.router.mainMirroringController.selectedDataset', 'controller.isEdit'),

    select: Em.Select.extend({
      attributeBindings: ['disabled']
    }),

    didInsertElement: function () {
      // Initialize datepicker
      $('.datepicker').datepicker({
        format: 'mm/dd/yyyy'
      }).on('changeDate', function (ev) {
            $(this).datepicker('hide');
          });

      // Set default value for Repeat every combo box
      this.set('controller.formFields.repeatOptionSelected', this.get('repeatOptions')[2]);

      this.fillForm();
      this.onTargetClustersChange();
    },

    willDestroyElement: function () {
      var controller = this.get('controller');
      Em.keys(this.get('controller.formFields')).forEach(function (key) {
        controller.removeObserver('formFields.' + key, controller, 'validate');
      }, this);
      this._super();
    },

    init: function () {
      this.get('controller').clearStep();
      this._super();
    }
  });
  
});
window.require.register("views/main/mirroring/jobs_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var filters = require('views/common/filter_view');
  var sort = require('views/common/sort_view');

  App.MainDatasetJobsView = App.TableView.extend({
    templateName: require('templates/main/mirroring/jobs'),

    isLoaded: function () {
      return App.router.get('mainMirroringController.isLoaded');
    }.property('App.router.mainMirroringController.isLoaded'),

    content: function () {
      var datasetName = this.get('controller.content.name');
      return (this.get('isLoaded') && datasetName) ?
          App.Dataset.find().findProperty('name', datasetName).get('datasetJobs') : [];
    }.property('controller.content', 'isLoaded'),

    dataset: function () {
      return this.get('controller.content');
    }.property('controller.content'),

    showActions: function () {
      return this.get('isLoaded') && this.get('dataset.status') && !this.get('parentView.controller.isDatasetLoadingError');
    }.property('isLoaded', 'dataset.status', 'parentView.controller.isDatasetLoadingError'),

    sortView: sort.wrapperView.extend({
      loadSortStatuses: function(){
        this._super();
        var statuses = App.db.getSortingStatuses(this.get('controller.name'));
        if (!statuses || statuses.everyProperty('status', 'sorting')) {
          var sorting = this.get('childViews').findProperty('name', 'startDate');
          this.sort(sorting, true);
          sorting.set('status', 'sorting_desc');
        }
      }.observes('parentView.filteringComplete')
    }),
    idSort: sort.fieldView.extend({
      column: 1,
      name: 'name',
      displayName: Em.I18n.t('mirroring.table.jobId'),
      type: 'string'
    }),
    startSort: sort.fieldView.extend({
      column: 2,
      name: 'startDate',
      displayName: Em.I18n.t('mirroring.table.start'),
      type: 'number'
    }),
    endSort: sort.fieldView.extend({
      column: 3,
      name: 'endDate',
      displayName: Em.I18n.t('mirroring.table.end'),
      type: 'number'
    }),
    statusSort: sort.fieldView.extend({
      column: 4,
      name: 'statusFormatted',
      displayName: Em.I18n.t('mirroring.table.status'),
      type: 'string'
    }),

    /**
     * Filter view for name column
     * Based on <code>filters</code> library
     */
    idFilterView: filters.createTextView({
      fieldType: 'input-small',
      column: 1,
      onChangeValue: function () {
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
      }
    }),

    startFilterView: filters.createSelectView({
      fieldType: 'input-small',
      column: 2,
      content: ['Any', 'Past 1 Day', 'Past 2 Days', 'Past 7 Days', 'Past 14 Days', 'Past 30 Days'],
      onChangeValue: function () {
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'date');
      }
    }),

    endFilterView: filters.createSelectView({
      fieldType: 'input-small',
      column: 3,
      content: ['Any', 'Past 1 Day', 'Past 2 Days', 'Past 7 Days', 'Past 14 Days', 'Past 30 Days'],
      onChangeValue: function () {
        this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'date');
      }
    }),

    statusFilterView: filters.createSelectView({
      fieldType: 'input-small',
      column: 4,
      content: ['Any', 'Waiting', 'Running', 'Suspended', 'Killed', 'Failed', 'Succeeded', 'Error'],
      onClearValue: function () {
        if (this.get('value') === '') {
          this.set('value', 'Any');
        }
      }.observes('value'),
      onChangeValue: function () {
        var value = this.get('value');
        if (value === 'Any') {
          value = '';
        }
        this.get('parentView').updateFilter(this.get('column'), value, 'string');
      }
    }),

    JobView: Em.View.extend({
      content: null,
      tagName: 'tr',

      showActions: function () {
        return ['RUNNING', 'SUSPENDED'].contains(this.get('content.status')) && App.get('isAdmin');
      }.property('content.status')
    }),

    /**
     * associations between dataset property and column index
     */
    colPropAssoc: function () {
      var associations = [];
      associations[1] = 'name';
      associations[2] = 'startDate';
      associations[3] = 'endDate';
      associations[4] = 'statusFormatted';
      return associations;
    }.property()
  });
  
});
window.require.register("views/main/mirroring/manage_clusters_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainMirroringManageClusterstView = Em.View.extend({
    name: 'mainMirroringManageClustersView',
    templateName: require('templates/main/mirroring/manage_clusters'),

    clusterSelect: Ember.Select.extend({
      classNames: ['cluster-select'],
      multiple: true,
      content: function () {
        return this.get('controller.clusters');
      }.property('controller.clusters.@each'),

      optionLabelPath: 'content.name',

      onSelect: function () {
        if (this.get('selection.length')) {
          if (this.get('selection').length === 1) {
            this.set('controller.selectedCluster', this.get('selection')[0]);
          } else {
            this.set('selection', [this.get('controller.selectedCluster')]);
          }
        } else {
          this.set('selection', [this.get('content')[0]]);
        }
      }.observes('selection'),

      // set selected option in Cluster Select after adding or removing an option
      onContentChange: function () {
        if (this.get('controller.isLoaded')) {
          var content = this.get('content');
          if (content.length < this.get('element.options.length')) {
            this.set('selection', [content[0]]);
          } else {
            this.set('selection', [content[content.length - 1]]);
          }
        }
      }.observes('content.length', 'controller.isLoaded')
    }),

    removeDisabled: function () {
      var selectedCluster = this.get('controller.selectedCluster');
      return !selectedCluster || selectedCluster.get('name') === App.get('clusterName');
    }.property('controller.selectedCluster', 'App.clusterName'),

    didInsertElement: function () {
      this.get('controller').onLoad();
    }
  });

  App.InputContainer = Em.View.extend({
    tagName: 'div',
    classNames: ['control-group', 'row-fluid'],
    attributeBindings: ['title'],
    didInsertElement: function () {
      $(this.get('element')).tooltip({placement: 'right'});
    }
  });


  
});
window.require.register("views/main/mirroring_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainMirroringView = Em.View.extend({

    templateName: require('templates/main/mirroring'),

    didInsertElement: function () {
      var controller = this.get('controller');
      controller.set('isDatasetsLoaded', false);
      controller.set('isTargetClustersLoaded', false);
      this.loadDataPeriodically();
    },

    // updata datasets data, when Mirroring page is opened
    loadDataPeriodically: function () {
      if (this.get('isVisible')) {
        var self = this;
        this.get('controller').loadData();
        setTimeout(function () {
          self.loadDataPeriodically();
        }, App.contentUpdateInterval);
      }
    }
  });
  
});
window.require.register("views/main/service", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainServiceView = Em.View.extend({
    templateName:require('templates/main/service'),
    didInsertElement: function() {
      this.addToolTip();
    },
    addToolTip: function() {
      if (this.get('controller.isAllServicesInstalled')) {
        App.tooltip($('.add-service-button a'), {title: Em.I18n.t('services.nothingToAdd')});
      }
    }.observes('controller.isAllServicesInstalled')
  });
});
window.require.register("views/main/service/add_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.AddServiceView = Em.View.extend({

    templateName: require('templates/main/service/add'),

    isStep1Disabled: function () {
      return this.isStepDisabled(1);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.isStepDisabled(2);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.isStepDisabled(3);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep4Disabled: function () {
      return this.isStepDisabled(4);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep5Disabled: function () {
      return this.isStepDisabled(5);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep6Disabled: function () {
      return this.isStepDisabled(6);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep7Disabled: function () {
      return this.isStepDisabled(7);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStepDisabled: function (index) {
      return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
    }

  });
  
});
window.require.register("views/main/service/all_services_actions", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.AllServicesActionView = Em.View.extend({
    templateName: require('templates/main/service/all_services_actions')
  });
  
});
window.require.register("views/main/service/info/configs", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var batchUtils = require('utils/batch_scheduled_requests');

  App.MainServiceInfoConfigsView = Em.View.extend({
    templateName: require('templates/main/service/info/configs'),
    didInsertElement: function () {
      var controller = this.get('controller');
      controller.loadStep();
    },

    componentsCount: null,
    hostsCount: null,
    isStopCommand:true,

    updateComponentInformation: function() {
      var hc = this.get('controller.content.restartRequiredHostsAndComponents');
      var hostsCount = 0;
      var componentsCount = 0;
      for (var host in hc) {
        hostsCount++;
        componentsCount += hc[host].length;
      }
      this.set('componentsCount', componentsCount);
      this.set('hostsCount', hostsCount);
    }.observes('controller.content.restartRequiredHostsAndComponents'),

    rollingRestartSlaveComponentName : function() {
      return batchUtils.getRollingRestartComponentName(this.get('controller.content.serviceName'));
    }.property('controller.content.serviceName'),

    rollingRestartActionName : function() {
      var label = null;
      var componentName = this.get('rollingRestartSlaveComponentName');
      if (componentName) {
        label = Em.I18n.t('rollingrestart.dialog.title').format(App.format.role(componentName));
      }
      return label;
    }.property('rollingRestartSlaveComponentName')
  });
  
});
window.require.register("views/main/service/info/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainServiceInfoMenuView = Em.CollectionView.extend({
    tagName: 'ul',
    classNames: ["nav", "nav-tabs"],
    content:function(){
      var menuItems = [
        { label: Em.I18n.t('services.service.info.menu.summary'), routing:'summary', active:"active"}
        //{ label:'Audit', routing:'audit'}
      ];
      if(this.get('configTab')) menuItems.push({ label: Em.I18n.t('services.service.info.menu.configs'), routing:'configs'});
      return menuItems;
    }.property(),

    init: function(){ this._super(); this.activateView(); },

    activateView:function () {
      $.each(this._childViews, function () {
        this.set('active', (document.URL.endsWith(this.get('content.routing')) ? "active" : ""));
      });
    }.observes('App.router.location.lastSetURL'),

    deactivateChildViews: function() {
      $.each(this._childViews, function(){
        this.set('active', "");
      });
    },

    itemViewClass: Em.View.extend({
      classNameBindings: ["active"],
      active: "",
      template: Ember.Handlebars.compile('<a {{action showInfo view.content.routing }} href="#"> {{unbound view.content.label}}</a>')
    })
  });
});
window.require.register("views/main/service/info/metrics/flume/channel_fill_pct", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_ChannelFillPercent = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-channel-fill-percent",
    title: Em.I18n.t('services.service.info.metrics.flume.channelFillPercent'),
    yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,

    ajaxIndex: 'service.metrics.flume.channel_fill_percent',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.CHANNEL) {
            for ( var cname in hc.metrics.flume.flume.CHANNEL) {
              var seriesName = Em.I18n.t('services.service.info.metrics.flume.channelName').format(cname + " (" + hostName + ")");
              var seriesData = hc.metrics.flume.flume.CHANNEL[cname]['ChannelFillPercentage'];
              if (seriesData) {
                seriesArray.push(self.transformData(seriesData, seriesName));
              }
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/channel_size", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_ChannelSize = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-channel-size",
    title: Em.I18n.t('services.service.info.metrics.flume.channelSize'),

    ajaxIndex: 'service.metrics.flume.channel_size',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.CHANNEL) {
            for ( var cname in hc.metrics.flume.flume.CHANNEL) {
              var seriesName = Em.I18n.t('services.service.info.metrics.flume.channelName').format(cname + " (" + hostName + ")");
              var seriesData = hc.metrics.flume.flume.CHANNEL[cname]['ChannelSize'];
              if (seriesData) {
                seriesArray.push(self.transformData(seriesData, seriesName));
              }
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/cpu_user", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_CPUUser = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-cpu-user",
    title: Em.I18n.t('services.service.info.metrics.flume.cpu.user'),
    yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,
    renderer: 'line',
    
    ajaxIndex: 'service.metrics.flume.cpu_user',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.cpu && hc.metrics.cpu.cpu_user) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
            var seriesData = hc.metrics.cpu.cpu_user;
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/gc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_GarbageCollection = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-jvm-gc",
    title: Em.I18n.t('services.service.info.metrics.flume.gc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.flume.gc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.jvm && hc.metrics.jvm.gcTimeMillis) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
            var seriesData = hc.metrics.jvm.gcTimeMillis;
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/jvm_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_JVMHeapUsed = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-jvm-heap-used",
    title: Em.I18n.t('services.service.info.metrics.flume.jvmHeapUsed'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'service.metrics.flume.jvm_heap_used',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      var MB = Math.pow(2, 20);
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.jvm && hc.metrics.jvm.memHeapUsedM) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
            var seriesData = hc.metrics.jvm.memHeapUsedM;
            if (seriesData) {
              var s = self.transformData(seriesData, seriesName);
              for (var i = 0; i < s.data.length; i++) {
                s.data[i].y *= MB;
              }
              seriesArray.push(s);
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/jvm_threads_runnable", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_JVMThreadsRunnable = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-jvm-threads-runnable",
    title: Em.I18n.t('services.service.info.metrics.flume.jvmThreadsRunnable'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.flume.jvm_threads_runnable',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      var MB = Math.pow(2, 20);
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.jvm && hc.metrics.jvm.threadsRunnable) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
            var seriesData = hc.metrics.jvm.threadsRunnable;
            if (seriesData) {
              var s = self.transformData(seriesData, seriesName);
              seriesArray.push(s);
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/sink_connection_failed", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_SinkConnectionFailedCount = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-sink-connection-failed",
    title: Em.I18n.t('services.service.info.metrics.flume.sinkConnectionFailed'),

    ajaxIndex: 'service.metrics.flume.sink_connection_failed',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.SINK) {
            for ( var cname in hc.metrics.flume.flume.SINK) {
              var seriesName = Em.I18n.t('services.service.info.metrics.flume.sinkName').format(cname + " (" + hostName + ")");
              var seriesData = hc.metrics.flume.flume.SINK[cname]['ConnectionFailedCount'];
              if (seriesData) {
                seriesArray.push(self.transformData(seriesData, seriesName));
              }
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/sink_drain_success", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_SinkDrainSuccessCount = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-sink-drain-success",
    title: Em.I18n.t('services.service.info.metrics.flume.sinkDrainSuccess'),

    ajaxIndex: 'service.metrics.flume.sink_drain_success',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.SINK) {
            for ( var cname in hc.metrics.flume.flume.SINK) {
              var seriesName = Em.I18n.t('services.service.info.metrics.flume.sinkName').format(cname + " (" + hostName + ")");
              var seriesData = hc.metrics.flume.flume.SINK[cname]['EventDrainSuccessCount'];
              if (seriesData) {
                seriesArray.push(self.transformData(seriesData, seriesName));
              }
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/flume/source_accepted", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsFlume_SourceAcceptedCount = App.ChartLinearTimeView.extend({
    id: "service-metrics-flume-source-accepted",
    title: Em.I18n.t('services.service.info.metrics.flume.sourceAccepted'),

    ajaxIndex: 'service.metrics.flume.source_accepted',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var self = this;
      if (jsonData && jsonData.host_components) {
        jsonData.host_components.forEach(function (hc) {
          var hostName = hc.HostRoles.host_name;
          var host = App.Host.find(hostName);
          if (host && host.get('publicHostName')) {
            hostName = host.get('publicHostName');
          }
          if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.SOURCE) {
            for ( var cname in hc.metrics.flume.flume.SOURCE) {
              var seriesName = Em.I18n.t('services.service.info.metrics.flume.sourceName').format(cname + " (" + hostName + ")");
              var seriesData = hc.metrics.flume.flume.SOURCE[cname]['EventAcceptedCount'];
              if (seriesData) {
                seriesArray.push(self.transformData(seriesData, seriesName));
              }
            }
          }
        });
      }
      return seriesArray;
    }
  });
  
});
window.require.register("views/main/service/info/metrics/hbase/cluster_requests", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHBASE_ClusterRequests = App.ChartLinearTimeView.extend({
    id: "service-metrics-hbase-cluster-requests",
    title: Em.I18n.t('services.service.info.metrics.hbase.clusterRequests'),

    ajaxIndex: 'service.metrics.hbase.cluster_requests',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.master) {
        for ( var name in jsonData.metrics.hbase.master) {
          var displayName;
          var seriesData = jsonData.metrics.hbase.master[name];
          switch (name) {
            case "cluster_requests":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.clusterRequests.displayNames.requestCount');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hbase/hlog_split_size", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHBASE_HlogSplitSize = App.ChartLinearTimeView.extend({
    id: "service-metrics-hbase-hlog-split-size",
    title: Em.I18n.t('services.service.info.metrics.hbase.hlogSplitSize'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,

    ajaxIndex: 'service.metrics.hbase.hlog_split_size',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.master) {
        for ( var name in jsonData.metrics.hbase.master) {
          var displayName;
          var seriesData = jsonData.metrics.hbase.master[name];
          switch (name) {
            case "splitSize_avg_time":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.hlogSplitSize.displayNames.splitSize');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hbase/hlog_split_time", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHBASE_HlogSplitTime = App.ChartLinearTimeView.extend({
    id: "service-metrics-hbase-hlog-split-time",
    title: Em.I18n.t('services.service.info.metrics.hbase.hlogSplitTime'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.hbase.hlog_split_time',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.master) {
        for ( var name in jsonData.metrics.hbase.master) {
          var displayName;
          var seriesData = jsonData.metrics.hbase.master[name];
          switch (name) {
            case "splitTime_avg_time":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.hlogSplitTime.displayNames.splitTime');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hbase/regionserver_queuesize", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHBASE_RegionServerQueueSize = App.ChartLinearTimeView.extend({
    id: "service-metrics-hbase-regionserver-queuesize",
    title: Em.I18n.t('services.service.info.metrics.hbase.regionServerQueueSize'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.hbase.regionserver_queuesize',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.regionserver) {
        for ( var name in jsonData.metrics.hbase.regionserver) {
          var displayName;
          var seriesData = jsonData.metrics.hbase.regionserver[name];
          switch (name) {
            case "compactionQueueSize":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerQueueSize.displayNames.compactionQueueSize');
              break;
            case "flushQueueSize":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerQueueSize.displayNames.flushQueueSize');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hbase/regionserver_regions", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHBASE_RegionServerRegions = App.ChartLinearTimeView.extend({
    id: "service-metrics-hbase-regionserver-regions",
    title: Em.I18n.t('services.service.info.metrics.hbase.regionServerRegions'),

    ajaxIndex: 'service.metrics.hbase.regionserver_regions',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.regionserver) {
        for ( var name in jsonData.metrics.hbase.regionserver) {
          var displayName;
          var seriesData = jsonData.metrics.hbase.regionserver[name];
          switch (name) {
            case "regions":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerRegions.displayNames.regions');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hbase/regionserver_rw_requests", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing HBase Cluster Requests
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHBASE_RegionServerReadWriteRequests = App.ChartLinearTimeView.extend({
    id: "service-metrics-hbase-regionserver-rw-requests",
    renderer: 'line',
    ajaxIndex: 'service.metrics.hbase.regionserver_rw_requests',
    yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('req', 
        App.ChartLinearTimeView.DefaultFormatter),
    title: function () {
      return App.get('isHadoop2Stack') ? 
          Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests.2') : 
            Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests');
    }.property('App.isHadoop2Stack'),

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.regionserver) {
        for ( var name in jsonData.metrics.hbase.regionserver) {
          var displayName;
          var seriesData = jsonData.metrics.hbase.regionserver[name];
          switch (name) {
            case "writeRequestsCount":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests.displayNames.writeRequests');
              break;
            case "readRequestsCount":
              displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests.displayNames.readRequests');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/block_status", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_BlockStatus = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-block-status",
    title: Em.I18n.t('services.service.info.metrics.hdfs.blockStatus'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.hdfs.block_status',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.FSNamesystem) {
        for ( var name in jsonData.metrics.dfs.FSNamesystem) {
          var displayName;
          var seriesData = jsonData.metrics.dfs.FSNamesystem[name];
          switch (name) {
            case "PendingReplicationBlocks":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.blockStatus.displayNames.pendingReplicationBlocks');
              break;
            case "UnderReplicatedBlocks":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.blockStatus.displayNames.underReplicatedBlocks');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/file_operations", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_FileOperations = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-file-operations",
    title: Em.I18n.t('services.service.info.metrics.hdfs.fileOperations'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.hdfs.file_operations',
    yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('ops', 
        App.ChartLinearTimeView.DefaultFormatter),

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.namenode) {
        for ( var name in jsonData.metrics.dfs.namenode) {
          var displayName;
          var seriesData = jsonData.metrics.dfs.namenode[name];
          switch (name) {
            case "FileInfoOps":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.fileOperations.displayNames.fileInformationOperations');
              break;
            case "DeleteFileOps":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.fileOperations.displayNames.deleteFileOperations');
              break;
            case "CreateFileOps":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.fileOperations.displayNames.createFileOperations');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/gc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_GC = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-gc",
    title: Em.I18n.t('services.service.info.metrics.hdfs.gc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.hdfs.gc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "gcTimeMillis":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.gc.displayNames.gcTimeMillis');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/io", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_IO = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-io",
    title: Em.I18n.t('services.service.info.metrics.hdfs.io'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'service.metrics.hdfs.io',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.datanode) {
        for ( var name in jsonData.metrics.dfs.datanode) {
          var displayName;
          var seriesData = jsonData.metrics.dfs.datanode[name];
          switch (name) {
            case "bytes_written":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.io.displayNames.bytesWritten');
              break;
            case "bytes_read":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.io.displayNames.bytesRead');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/jvm_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_JVMHeap = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-jvm-heap",
    title: Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'service.metrics.hdfs.jvm_heap',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var MB = Math.pow(2, 20);
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "memHeapCommittedM":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapCommittedM');
              break;
            case "memNonHeapUsedM":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapUsedM');
              break;
            case "memHeapUsedM":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapUsedM');
              break;
            case "memNonHeapCommittedM":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapCommittedM');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= MB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/jvm_threads", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_JVMThreads = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-jvm-threads",
    title: Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.hdfs.jvm_threads',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "threadsBlocked":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsBlocked');
              break;
            case "threadsWaiting":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsWaiting');
              break;
            case "threadsTimedWaiting":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsTimedWaiting');
              break;
            case "threadsRunnable":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsRunnable');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/rpc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_RPC = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-rpc",
    title: Em.I18n.t('services.service.info.metrics.hdfs.rpc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.hdfs.rpc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.rpc) {
        for ( var name in jsonData.metrics.rpc) {
          var displayName;
          var seriesData = jsonData.metrics.rpc[name];
          switch (name) {
            case "RpcQueueTime_avg_time":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.rpc.displayNames.rpcQueueTimeAvgTime');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/hdfs/space_utilization", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsHDFS_SpaceUtilization = App.ChartLinearTimeView.extend({
    id: "service-metrics-hdfs-space-utilization",
    title: Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',
    ajaxIndex: 'service.metrics.hdfs.space_utilization',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var GB = Math.pow(2, 30);
      if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.FSNamesystem) {
        for ( var name in jsonData.metrics.dfs.FSNamesystem) {
          var displayName;
          var seriesData = jsonData.metrics.dfs.FSNamesystem[name];
          switch (name) {
            case "CapacityRemainingGB":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityRemainingGB');
              break;
            case "CapacityUsedGB":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityUsedGB');
              break;
            case "CapacityTotalGB":
              displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityTotalGB');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= GB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/gc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_GC = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-gc",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.gc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.mapreduce.gc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "gcTimeMillis":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.gc.displayNames.gcTimeMillis');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/jobs_status", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_JobsStatus = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-jobs-status",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.jobsStatus'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.mapreduce.jobs_status',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.mapred && jsonData.metrics.mapred.jobtracker) {
        for ( var name in jsonData.metrics.mapred.jobtracker) {
          var displayName;
          var seriesData = jsonData.metrics.mapred.jobtracker[name];
          switch (name) {
            case "jobs_running":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsRunning');
              break;
            case "jobs_failed":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsFailed');
              break;
            case "jobs_completed":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsCompleted');
              break;
            case "jobs_preparing":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsPreparing');
              break;
            case "jobs_submitted":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jobsStatus.displayNames.jobsSubmitted');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/jvm_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_JVMHeap = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-jvm-heap",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'service.metrics.mapreduce.jobs_heap',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var MB = Math.pow(2, 20);
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "memHeapCommittedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memHeapCommittedM');
              break;
            case "memNonHeapUsedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memNonHeapUsedM');
              break;
            case "memHeapUsedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memHeapUsedM');
              break;
            case "memNonHeapCommittedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memNonHeapCommittedM');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= MB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/jvm_threads", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_JVMThreads = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-jvm-threads",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.mapreduce.jobs_threads',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "threadsBlocked":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsBlocked');
              break;
            case "threadsWaiting":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsWaiting');
              break;
            case "threadsTimedWaiting":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsTimedWaiting');
              break;
            case "threadsRunnable":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsRunnable');
              break;
            default:
              break;
          }

          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/map_slots", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_MapSlots = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-map-slots",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.mapSlots'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.mapreduce.map_slots',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.mapred && jsonData.metrics.mapred.jobtracker) {
        for ( var name in jsonData.metrics.mapred.jobtracker) {
          var displayName;
          var seriesData = jsonData.metrics.mapred.jobtracker[name];
          switch (name) {
            case "reserved_map_slots":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.mapSlots.displayNames.reservedMapSlots');
              break;
            case "occupied_map_slots":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.mapSlots.displayNames.occupiedMapSlots');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/reduce_slots", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_ReduceSlots = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-reduce-slots",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.reduceSlots'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.mapreduce.reduce_slots',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.mapred && jsonData.metrics.mapred.jobtracker) {
        for ( var name in jsonData.metrics.mapred.jobtracker) {
          var displayName;
          var seriesData = jsonData.metrics.mapred.jobtracker[name];
          switch (name) {
            case "reserved_reduce_slots":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.reduceSlots.displayNames.reservedReduceSlots');
              break;
            case "occupied_reduce_slots":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.reduceSlots.displayNames.occupiedReduceSlots');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/rpc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_RPC = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-rpc",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.rpc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.mapreduce.rpc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.rpc) {
        for ( var name in jsonData.metrics.rpc) {
          var displayName;
          var seriesData = jsonData.metrics.rpc[name];
          switch (name) {
            case "RpcQueueTime_avg_time":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.rpc.displayNames.RpcQueueTimeAvgTime');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/mapreduce/tasks_running_waiting", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsMapReduce_TasksRunningWaiting = App.ChartLinearTimeView.extend({
    id: "service-metrics-mapreduce-tasks-running-waiting",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.tasksRunningWaiting'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.mapreduce.tasks_running_waiting',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.mapred && jsonData.metrics.mapred.jobtracker) {
        for ( var name in jsonData.metrics.mapred.jobtracker) {
          var displayName;
          var seriesData = jsonData.metrics.mapred.jobtracker[name];
          switch (name) {
            case "running_maps":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.runningMaps');
              break;
            case "running_reduces":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.runningReduces');
              break;
            case "waiting_maps":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.waitingMaps');
              break;
            case "waiting_reduces":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.tasksRunningWaiting.displayNames.waitingReduces');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/storm/executors_metric", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('mixins/common/chart/storm_linear_time');

  App.ChartServiceMetricsSTORM_Executors = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
    id: "service-metrics-storm-executors",
    title: Em.I18n.t('services.storm.executors.metrics.title'),
    renderer: 'line',
    yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

    stormChartDefinition: [
      {
        name: Em.I18n.t('services.storm.executors.metrics.total'),
        field: 'totalexecutors'
      }
    ]

  });
  
});
window.require.register("views/main/service/info/metrics/storm/slots_number_metric", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('mixins/common/chart/storm_linear_time');

  App.ChartServiceMetricsSTORM_SlotsNumber = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
    id: "service-metrics-storm-supervisor-allocated",
    title: Em.I18n.t('services.storm.slots.metrics.title'),
    renderer: 'line',
    yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

    stormChartDefinition: [
      {
        name: Em.I18n.t('services.storm.slots.metrics.total'),
        field: 'totalslots'
      },
      {
        name: Em.I18n.t('services.storm.slots.metrics.free'),
        field: 'freeslots'
      },
      {
        name: Em.I18n.t('services.storm.slots.metrics.used'),
        field: 'usedslots'
      }
    ]

  });
  
});
window.require.register("views/main/service/info/metrics/storm/tasks_metric", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('mixins/common/chart/storm_linear_time');

  App.ChartServiceMetricsSTORM_Tasks = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
    id: "service-metrics-storm-tasks",
    title: Em.I18n.t('services.storm.tasks.metrics.title'),
    renderer: 'line',
    yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

    stormChartDefinition: [
      {
        name: Em.I18n.t('services.storm.tasks.metrics.total'),
        field: 'totaltasks'
      }
    ]

  });
  
});
window.require.register("views/main/service/info/metrics/storm/topologies_metric", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  require('mixins/common/chart/storm_linear_time');

  App.ChartServiceMetricsSTORM_Topologies = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
    id: "service-metrics-storm-topologies",
    title: Em.I18n.t('services.storm.topology.metrics.title'),
    renderer: 'line',
    yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

    stormChartDefinition: [
      {
        name: Em.I18n.t('services.storm.topology.metrics.total'),
        field: 'topologies'
      }
    ]

  });
  
});
window.require.register("views/main/service/info/metrics/yarn/allocated", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   *
   * This is a view for showing cluster CPU metrics
   *
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_AllocatedMemory = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-queue-allocated",
    title: Em.I18n.t('services.service.info.metrics.yarn.allocated.memory'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.yarn.queue.allocated',
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var MB = Math.pow(2, 20);
      if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
        for (var name in jsonData.metrics.yarn.Queue.root) {
          var displayName = null;
          var seriesData = jsonData.metrics.yarn.Queue.root[name];
          switch (name) {
            case "AvailableMB":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.memory.displayNames.available');
              break;
            case "PendingMB":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.memory.displayNames.pending');
              break;
            case "AllocatedMB":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.memory.displayNames.allocated');
              break;
            default:
              break;
          }
          if (seriesData && displayName) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= MB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/allocated_container", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   *
   * This is a view for showing cluster CPU metrics
   *
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_AllocatedContainer = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-queue-allocated-container",
    title: Em.I18n.t('services.service.info.metrics.yarn.allocated.container'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.yarn.queue.allocated.container',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
        for (var name in jsonData.metrics.yarn.Queue.root) {
          var displayName = null;
          var seriesData = jsonData.metrics.yarn.Queue.root[name];
          switch (name) {
            case "ReservedContainers":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.container.displayNames.reserved');
              break;
            case "PendingContainers":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.container.displayNames.pending');
              break;
            case "AllocatedContainers":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.container.displayNames.allocated');
              break;
            default:
              break;
          }
          if (seriesData != null && displayName) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/apps_current_states", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   *
   * This is a view for showing cluster CPU metrics
   *
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_ApplicationCurrentStates = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-apps-current-states",
    title: Em.I18n.t('services.service.info.metrics.yarn.apps.states.current.title'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.yarn.queue.apps.states.current',
        
    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
        for (var name in jsonData.metrics.yarn.Queue.root) {
          var displayName = null;
          var seriesData = jsonData.metrics.yarn.Queue.root[name];
          switch (name) {
            case "AppsPending":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.pending');
              break;
            case "AppsRunning":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.running');
              break;
            default:
              break;
          }
          if (seriesData != null && displayName) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/apps_finished_states", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   *
   * This is a view for showing cluster CPU metrics
   *
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_ApplicationFinishedStates = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-apps-finished-states",
    title: Em.I18n.t('services.service.info.metrics.yarn.apps.states.finished.title'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.yarn.queue.apps.states.finished',
    yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('apps', 
        App.ChartLinearTimeView.DefaultFormatter),

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
        for (var name in jsonData.metrics.yarn.Queue.root) {
          var displayName = null;
          var seriesData = jsonData.metrics.yarn.Queue.root[name];
          switch (name) {
            case "AppsCompleted":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.completed');
              break;
  //          case "AppsFailed":
  //            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.failed');
  //            break;
            case "AppsKilled":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.killed');
              break;
            case "AppsSubmitted":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.submitted');
              break;
            default:
              break;
          }
          if (seriesData != null && displayName) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/gc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_GC = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-gc",
    title: Em.I18n.t('services.service.info.metrics.yarn.gc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.yarn.gc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "gcTimeMillis":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.gc.displayNames.gcTimeMillis');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/jvm_heap", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_JVMHeap = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-jvm-heap",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap'),
    yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
    renderer: 'line',

    ajaxIndex: 'service.metrics.yarn.jobs_heap',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      var MB = Math.pow(2, 20);
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "memHeapCommittedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memHeapCommittedM');
              break;
            case "memNonHeapUsedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memNonHeapUsedM');
              break;
            case "memHeapUsedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memHeapUsedM');
              break;
            case "memNonHeapCommittedM":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmHeap.displayNames.memNonHeapCommittedM');
              break;
            default:
              break;
          }
          if (seriesData) {
            var s = this.transformData(seriesData, displayName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= MB;
            }
            seriesArray.push(s);
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/jvm_threads", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_JVMThreads = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-jvm-threads",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads'),
    renderer: 'line',

    ajaxIndex: 'service.metrics.yarn.jobs_threads',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
        for ( var name in jsonData.metrics.jvm) {
          var displayName;
          var seriesData = jsonData.metrics.jvm[name];
          switch (name) {
            case "threadsBlocked":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsBlocked');
              break;
            case "threadsWaiting":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsWaiting');
              break;
            case "threadsTimedWaiting":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsTimedWaiting');
              break;
            case "threadsRunnable":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.jvmThreads.displayNames.threadsRunnable');
              break;
            default:
              break;
          }

          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/nms", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   *
   * This is a view for showing cluster CPU metrics
   *
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_NMS = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-node-manager-statuses",
    title: Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.yarn.node.manager.statuses',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.yarn && jsonData.metrics.yarn.ClusterMetrics) {
        for (var name in jsonData.metrics.yarn.ClusterMetrics) {
          var displayName;
          var seriesData = jsonData.metrics.yarn.ClusterMetrics[name];
          switch (name) {
            case "NumActiveNMs":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.active');
              break;
            case "NumDecommissionedNMs":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.decommissioned');
              break;
            case "NumLostNMs":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.lost');
              break;
            case "NumRebootedNMs":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.rebooted');
              break;
            case "NumUnhealthyNMs":
              displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.unhealthy');
              break;
            default:
              break;
          }
          if (seriesData != null) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/qmr", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');
  var objUtils = require('utils/object_utils');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_QMR = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-queue-memory-resource",
    title: Em.I18n.t('services.service.info.metrics.yarn.queueMemoryResource'),
    renderer: 'line',
    ajaxIndex: 'service.metrics.yarn.queue.memory.resource',
    yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,

    getDataForAjaxRequest: function () {
      var data = this._super();
      var svc = App.YARNService.find().objectAt(0);
      var queueNames = [];
      if (svc != null) {
        queueNames = svc.get('childQueueNames');
      }
      data.queueNames = queueNames;
      return data;
    },

    transformToSeries: function (jsonData) {
      var self = this;
      var seriesArray = [];
      var MB = Math.pow(2, 20);
      var svc = App.YARNService.find().objectAt(0);
      var queueNames = [];
      if (svc != null) {
        queueNames = svc.get('childQueueNames');
      }
      if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue) {
        queueNames.forEach(function (qName) {
          var qPath = qName.replace(/\//g, '.');
          var displayName;
          var allocatedData = objUtils.getProperty(jsonData.metrics.yarn.Queue, qPath + '.AllocatedMB');
          var availableData = objUtils.getProperty(jsonData.metrics.yarn.Queue, qPath + '.AvailableMB');
          displayName = Em.I18n.t('services.service.info.metrics.yarn.queueMemoryResource.displayName').format(qName);
          var seriesData = null;
          if (allocatedData != null && availableData != null) {
            if (typeof allocatedData == "number" && typeof availableData == "number") {
              seriesData = (availableData != 0) ? (allocatedData * 100) / availableData : 0;
            } else if (allocatedData.length > 0 && availableData.length > 0) {
              seriesData = [];
              for ( var c = 0; c < Math.min(availableData.length, allocatedData.length); c++) {
                var allocDivAvail =  (availableData[c][0] != 0) ? (allocatedData[c][0] * 100) / availableData[c][0] : 0;
                seriesData.push([allocDivAvail, allocatedData[c][1] ]);
              }
            } else {
              console.log("Skipping data series for Queue " + qName);
            }
          }
          if (seriesData != null) {
            seriesArray.push(self.transformData(seriesData, displayName));
          }
        });
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/metrics/yarn/rpc", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  /**
   * @class
   * 
   * This is a view for showing cluster CPU metrics
   * 
   * @extends App.ChartLinearTimeView
   * @extends Ember.Object
   * @extends Ember.View
   */
  App.ChartServiceMetricsYARN_RPC = App.ChartLinearTimeView.extend({
    id: "service-metrics-yarn-rpc",
    title: Em.I18n.t('services.service.info.metrics.mapreduce.rpc'),
    yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

    ajaxIndex: 'service.metrics.yarn.rpc',

    transformToSeries: function (jsonData) {
      var seriesArray = [];
      if (jsonData && jsonData.metrics && jsonData.metrics.rpc) {
        for ( var name in jsonData.metrics.rpc) {
          var displayName;
          var seriesData = jsonData.metrics.rpc[name];
          switch (name) {
            case "RpcQueueTime_avg_time":
              displayName = Em.I18n.t('services.service.info.metrics.mapreduce.rpc.displayNames.RpcQueueTimeAvgTime');
              break;
            default:
              break;
          }
          if (seriesData) {
            seriesArray.push(this.transformData(seriesData, displayName));
          }
        }
      }
      return seriesArray;
    }
  });
});
window.require.register("views/main/service/info/summary", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   * http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.AlertItemView = Em.View.extend({
    tagName:"li",
    templateName: require('templates/main/service/info/summary_alert'),
    classNameBindings: ["status"],
    status: function () {
      return "status-" + this.get("content.status");
    }.property('content'),
    didInsertElement: function () {
      // Tooltips for alerts need to be enabled.
      App.tooltip($("div[rel=tooltip]"));
      $(".tooltip").remove();
    }
  });

  App.MainServiceInfoSummaryView = Em.View.extend({
    templateName: require('templates/main/service/info/summary'),
    attributes:null,
    serviceStatus:{
      hdfs:false,
      yarn:false,
      mapreduce:false,
      mapreduce2:false,
      hbase:false,
      zookeeper:false,
      oozie:false,
      hive:false,
      ganglia:false,
      nagios:false,
      hue: false,
      flume: false,
      falcon: false,
      storm: false,
      tez: false,
      pig :false,
      sqoop: false
    },

    servicesHaveClients: ["OOZIE", "ZOOKEEPER", "HIVE", "MAPREDUCE2", "TEZ", "SQOOP", "PIG","FALCON"],

    sumMasterComponentView : Em.View.extend({
      didInsertElement: function() {
        App.tooltip($('[rel=healthTooltip]'));
      },
      templateName: require('templates/main/service/info/summary/master_components'),
      mastersComp : function(){
        return this.get('parentView.service.hostComponents').filterProperty('isMaster', true);
      }.property("service")
    }),

    alertsControllerBinding: 'App.router.mainAlertsController',
    alerts: function () {
      return this.get('alertsController.alerts');
    }.property('alertsController.alerts'),

    noTemplateService: function () {
      var serviceName = this.get("service.serviceName");
      //services with only master components
      return serviceName == "WEBHCAT" || serviceName == "NAGIOS";
    }.property('controller.content'),

    clients: function () {
      var service = this.get('controller.content');
      if (this.get('servicesHaveClients').contains(service.get("id"))) {
        return service.get('hostComponents').filterProperty('isClient');
      }
      return [];
    }.property('controller.content'),

    hasManyServers: function () {
      return this.get('servers').length > 1;
    }.property('servers'),

    clientsHostText: function () {
      if (this.get('clients').length == 0) {
        return '';
      } else if (this.get("hasManyClients")) {
        return Em.I18n.t('services.service.summary.viewHosts');
      } else {
        return Em.I18n.t('services.service.summary.viewHost');
      }
    }.property("hasManyClients"),

    hasManyClients: function () {
      return this.get('clients').length > 1;
    }.property('clients'),

    servers: function () {
      var result = [];
      var service = this.get('controller.content');
      if (service.get("id") == "ZOOKEEPER" || service.get("id") == "FLUME") {
        var servers = service.get('hostComponents').filterProperty('isMaster');
        if (servers.length > 0) {
          result = [{
            'host': servers[0].get('displayName'),
            'isComma': false,
            'isAnd': false
          }];
        }
        if (servers.length > 1) {
          result[0].isComma = true;
          result.push({
            'host': servers[1].get('displayName'),
            'isComma': false,
            'isAnd': false
          });
        }
        if (servers.length > 2) {
          result[1].isAnd = true;
          result.push({
            'host': Em.I18n.t('services.service.info.summary.serversHostCount').format(servers.length - 2),
            'isComma': false,
            'isAnd': false
          });
        }
      }
      return result;
    }.property('controller.content'),

    historyServerUI: function(){
      var service=this.get('controller.content');
      return (App.singleNodeInstall ? "http://" + App.singleNodeAlias + ":19888" : "http://" + service.get("hostComponents").findProperty('isMaster', true).get("host").get("publicHostName")+":19888");
    }.property('controller.content'),

    monitors: function () {
      var result = '';
      var service = this.get('controller.content');
      if (service.get("id") == "GANGLIA") {
        var monitors = service.get('hostComponents').filterProperty('isMaster', false);
        var liveMonitors = monitors.filterProperty("workStatus","STARTED").length;
        this.set("liveMonitors", liveMonitors);
        if (monitors.length) {
          result = Em.I18n.t('services.service.info.summary.hostsRunningMonitor').format(liveMonitors, monitors.length);
        }
      }
      return result;
    }.property('controller.content'),

    monitorsLiveTextView: App.ComponentLiveTextView.extend({
      liveComponents: function () {
        var service = this.get('service');
        if (service.get("id") == "GANGLIA") {
          return service.get('hostComponents').filterProperty('isMaster', false).filterProperty("workStatus","STARTED").length;
        } else {
          return null;
        }
      }.property('service.hostComponents.@each'),
      monitors: function () {
        var result = '';
        var service = this.get('parentView.controller.content');
        if (service.get("id") == "GANGLIA") {
          var monitors = service.get('hostComponents').filterProperty('isMaster', false);
          var liveMonitors = monitors.filterProperty("workStatus","STARTED").length;
          if (monitors.length) {
            result = Em.I18n.t('services.service.info.summary.hostsRunningMonitor').format(liveMonitors, monitors.length);
          }
        }
        return result;
      }.property('service.hostComponents.@each')
    }),

    hasManyMonitors: function () {
      var service = this.get('controller.content');
      if (service.get("id") == "GANGLIA") {
        var monitors = service.get('hostComponents').filterProperty('isMaster', false);
        if (monitors.length == 0) {
          return '';
        } else if (monitors.length > 1) {
          return Em.I18n.t('services.service.summary.viewHosts');
        } else {
          return Em.I18n.t('services.service.summary.viewHost');
        }
      }
      return result;
    }.property('controller.content'),

    /**
     * Property related to GANGLIA service, is unused for other services
     * @return {Object}
     */
    monitorsObj: function(){
      var service = this.get('controller.content');
      if (service.get("id") == "GANGLIA") {
        var monitors = service.get('hostComponents').filterProperty('isMaster', false);
        if (monitors.length) {
          return monitors[0];
        }
      }
      return {};
    }.property('controller.content'),

    /**
     * Property related to ZOOKEEPER service, is unused for other services
     * @return {Object}
     */
    serversHost: function() {
      var service = this.get('controller.content');
      if (service.get("id") == "ZOOKEEPER" || service.get("id") == "FLUME") {
        var servers = service.get('hostComponents').filterProperty('isMaster');
        if (servers.length > 0) {
          return servers[0];
        }
      }
      return {};
    }.property('controller.content'),

    /**
     * Property related to OOZIE and ZOOKEEPER services, is unused for other services
     * HIVE is supported too
     * @return {Object}
     */
    clientObj: function() {
      var service = this.get('controller.content');
      if (this.get('servicesHaveClients').contains(service.get("id"))) {
        var clients = service.get('hostComponents').filterProperty('isClient', true);
        if (clients.length > 0) {
          return clients[0];
        }
      }
      return {};
    }.property('controller.content'),

    data:{
      hive:{
        "database":"PostgreSQL",
        "databaseName":"hive",
        "user":"hive"
      }
    },

    /**
     * Wrapper for displayName. used to render correct display name for mysql_server
     */
    componentNameView: Ember.View.extend({
      template: Ember.Handlebars.compile('{{view.displayName}}'),
      comp : null,
      displayName: function(){
        if(this.get('comp.componentName') == 'MYSQL_SERVER'){
          return this.t('services.hive.databaseComponent');
        }
        return this.get('comp.displayName');
      }.property('comp')
    }),

    service:function () {
      var svc = this.get('controller.content');
      var svcName = svc.get('serviceName');
      if (svcName) {
        switch (svcName.toLowerCase()) {
          case 'hdfs':
            svc = App.HDFSService.find().objectAt(0);
            break;
          case 'yarn':
            svc = App.YARNService.find().objectAt(0);
            break;
          case 'mapreduce':
            svc = App.MapReduceService.find().objectAt(0);
            break;
          case 'hbase':
            svc = App.HBaseService.find().objectAt(0);
            break;
          case 'flume':
            svc = App.FlumeService.find().objectAt(0);
            break;
          case 'storm':
            svc = App.StormService.find().objectAt(0);
            break;
          default:
            break;
        }
      }
      return svc;
    }.property('controller.content.serviceName').volatile(),

    isHide:true,
    moreStatsView:Em.View.extend({
      tagName:"a",
      template:Ember.Handlebars.compile('{{t services.service.summary.moreStats}}'),
      attributeBindings:[ 'href' ],
      classNames:[ 'more-stats' ],
      click:function (event) {
        this._parentView._parentView.set('isHide', false);
        this.remove();
      },
      href:'javascript:void(null)'
    }),

    serviceName:function () {
      return this.get('service.serviceName');
    }.property('service'),

    oldServiceName:'',

    /**
     * Contains graphs for this particular service
     */
    serviceMetricGraphs:function () {
      var svcName = this.get('service.serviceName');
      var graphs = [];
      if (svcName) {
        switch (svcName.toLowerCase()) {
          case 'hdfs':
            graphs = [ [App.ChartServiceMetricsHDFS_SpaceUtilization.extend(),
              App.ChartServiceMetricsHDFS_FileOperations.extend(),
              App.ChartServiceMetricsHDFS_BlockStatus.extend(),
              App.ChartServiceMetricsHDFS_IO.extend()],
              [App.ChartServiceMetricsHDFS_RPC.extend(),
              App.ChartServiceMetricsHDFS_GC.extend(),
              App.ChartServiceMetricsHDFS_JVMHeap.extend(),
              App.ChartServiceMetricsHDFS_JVMThreads.extend()]];
            break;
          case 'yarn':
            graphs = [[App.ChartServiceMetricsYARN_AllocatedMemory.extend(),
                App.ChartServiceMetricsYARN_QMR.extend(),
                App.ChartServiceMetricsYARN_AllocatedContainer.extend(),
                App.ChartServiceMetricsYARN_NMS.extend()],
              [App.ChartServiceMetricsYARN_ApplicationCurrentStates.extend(),
               App.ChartServiceMetricsYARN_ApplicationFinishedStates.extend(),
               App.ChartServiceMetricsYARN_RPC.extend(),
              App.ChartServiceMetricsYARN_GC.extend()
              ],
              [App.ChartServiceMetricsYARN_JVMThreads.extend(),
               App.ChartServiceMetricsYARN_JVMHeap.extend()]];
            break;
          case 'mapreduce':
            graphs = [ [App.ChartServiceMetricsMapReduce_JobsStatus.extend(),
              App.ChartServiceMetricsMapReduce_TasksRunningWaiting.extend(),
              App.ChartServiceMetricsMapReduce_MapSlots.extend(),
              App.ChartServiceMetricsMapReduce_ReduceSlots.extend()],
              [App.ChartServiceMetricsMapReduce_GC.extend(),
              App.ChartServiceMetricsMapReduce_RPC.extend(),
              App.ChartServiceMetricsMapReduce_JVMHeap.extend(),
              App.ChartServiceMetricsMapReduce_JVMThreads.extend()]];
            break;
          case 'hbase':
            graphs = [  [App.ChartServiceMetricsHBASE_ClusterRequests.extend(),
              App.ChartServiceMetricsHBASE_RegionServerReadWriteRequests.extend(),
              App.ChartServiceMetricsHBASE_RegionServerRegions.extend(),
              App.ChartServiceMetricsHBASE_RegionServerQueueSize.extend()],
              [App.ChartServiceMetricsHBASE_HlogSplitTime.extend(),
              App.ChartServiceMetricsHBASE_HlogSplitSize.extend()]];
            break;
          case 'flume':
            graphs = [[App.ChartServiceMetricsFlume_ChannelFillPercent.extend(),
               App.ChartServiceMetricsFlume_ChannelSize.extend(),
               App.ChartServiceMetricsFlume_SourceAcceptedCount.extend(),
               App.ChartServiceMetricsFlume_SinkDrainSuccessCount.extend()],
               [App.ChartServiceMetricsFlume_SinkConnectionFailedCount.extend(),
                //App.ChartServiceMetricsFlume_GarbageCollection.extend(),
                //App.ChartServiceMetricsFlume_JVMHeapUsed.extend(),
                //App.ChartServiceMetricsFlume_JVMThreadsRunnable.extend(),
                App.ChartServiceMetricsFlume_CPUUser.extend()]];
            break;
          case 'storm':
            graphs = [
              [
                App.ChartServiceMetricsSTORM_SlotsNumber.extend(),
                App.ChartServiceMetricsSTORM_Executors.extend(),
                App.ChartServiceMetricsSTORM_Topologies.extend(),
                App.ChartServiceMetricsSTORM_Tasks.extend()
              ]
            ];
            break;
          default:
            break;
        }
      }
      return graphs;
    }.property(''),

    loadServiceSummary:function (serviceName) {

      var serviceName = this.get('serviceName');
      if (!serviceName) {
        return;
      }

      if (this.get('oldServiceName')) {
        // do not delete it!
        return;
      }

      var summaryView = this;
      var serviceStatus = summaryView.get('serviceStatus');
      $.each(serviceStatus, function (key, value) {
        if (key.toUpperCase() == serviceName) {
          summaryView.set('serviceStatus.' + key, true);
        } else {
          summaryView.set('serviceStatus.' + key, false);
        }
      });

      console.log('load ', serviceName, ' info');
      this.set('oldServiceName', serviceName);
      serviceName = serviceName.toLowerCase();
    }.observes('serviceName'),

    /*
     * Alerts panel not display for PIG, SQOOP and TEZ Service
     */
    isNoAlertsService: function () {
      var serviceName = this.get('service.serviceName');
      if (!serviceName) {
        return false;
      }
      var noAlertsServices = ['PIG', 'SQOOP', 'TEZ'];
      return noAlertsServices.indexOf(serviceName) > -1;
    }.property(''),

    gangliaUrl:function () {
      var gangliaUrl = App.router.get('clusterController.gangliaUrl');
      if (!gangliaUrl) return null;
      var svcName = this.get('service.serviceName');
      if (svcName) {
        switch (svcName.toLowerCase()) {
          case 'hdfs':
            gangliaUrl += "/?r=hour&cs=&ce=&m=&s=by+name&c=HDPNameNode&tab=m&vn=";
            break;
          case 'mapreduce':
            gangliaUrl += "/?r=hour&cs=&ce=&m=&s=by+name&c=HDPJobTracker&tab=m&vn=";
            break;
          case 'hbase':
            gangliaUrl += "?r=hour&cs=&ce=&m=&s=by+name&c=HDPHBaseMaster&tab=m&vn=";
            break;
          default:
            break;
        }
      }
      return gangliaUrl;
    }.property('App.router.clusterController.gangliaUrl', 'service.serviceName'),

    didInsertElement:function () {
      var alertsController = this.get('alertsController');
      alertsController.loadAlerts(this.get('service.serviceName'), "SERVICE");
      alertsController.set('isUpdating', true);
      //TODO delegate style calculation to css
      // We have to make the height of the Alerts section
      // match the height of the Summary section.
      var summaryTable = document.getElementById('summary-info');
      var alertsList = document.getElementById('summary-alerts-list');
      if (summaryTable && alertsList) {
        var rows = $(summaryTable).find('tr');
        if (rows != null && rows.length > 0) {
          var minimumHeightSum = 20;
          var summaryActualHeight = summaryTable.clientHeight;
          // for summary window
          if (summaryActualHeight <= minimumHeightSum) {
            $(summaryTable).attr('style', "height:" + minimumHeightSum + "px;");
          }
        } else if (alertsList.clientHeight > 0) {
          $(summaryTable).append('<tr><td></td></tr>');
          $(summaryTable).attr('style', "height:" + alertsList.clientHeight + "px;");
        }
      }
    },
    willDestroyElement: function(){
      this.get('alertsController').set('isUpdating', false);
    },

    setAlertsWindowSize: function() {
      // for alerts window
      var summaryTable = document.getElementById('summary-info');
      var alertsList = document.getElementById('summary-alerts-list');
      var alertsNum = this.get('alerts').length;
      if (summaryTable && alertsList && alertsNum != null) {
        var summaryActualHeight = summaryTable.clientHeight;
        var alertsActualHeight = alertsNum * 60;
        var alertsMinHeight = 58;
        if (alertsNum == 0) {
          $(alertsList).attr('style', "height:" + alertsMinHeight + "px;");
        } else if ( alertsNum <= 4) {
          // set window size to actual alerts height
          $(alertsList).attr('style', "height:" + alertsActualHeight + "px;");
        } else {
          // set window size : sum of first 4 alerts height
          $(alertsList).attr('style', "height:" + 240 + "px;");
        }
        var curSize = alertsList.clientHeight;
        if ( summaryActualHeight >= curSize) {
          $(alertsList).attr('style', "height:" + summaryActualHeight + "px;");
        }
      }
    }.observes('alertsController.isLoaded'),

    clientHosts:App.Host.find(),

    clientHostsLength:function () {
      var text = this.t('services.service.summary.clientCount');
      var self = this;
      return text.format(self.get('clientHosts.length'));
    }.property('clientHosts'),

    clientComponents:function () {
      return App.HostComponent.find().filterProperty('isClient', true);
    }.property(),

    clientComponentsString:function () {
      var components = this.get('clientComponents');
      var names = [];
      components.forEach(function (component) {
        if (names.indexOf(component.get('displayName')) == -1) {
          names.push(component.get('displayName'));
        }
      });

      return names.length ? names.join(', ') : false;
    }.property('clientComponents')
  });
  
});
window.require.register("views/main/service/item", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var batchUtils = require('utils/batch_scheduled_requests');

  App.MainServiceItemView = Em.View.extend({
    templateName: require('templates/main/service/item'),
    maintenance: function(){
      var options = [];
      var service = this.get('controller.content');
      var hosts = App.Host.find().content.length;
      var allMasters = this.get('controller.content.hostComponents').filterProperty('isMaster').mapProperty('componentName').uniq();
      var disabled = this.get('controller.isStopDisabled');
      var serviceName = service.get('serviceName');

      if (service.get('isClientsOnly')) {
        var disableRefreshConfgis = !service.get('isRestartRequired');
        if (serviceName != 'TEZ') {
          options.push({action: 'runSmokeTest', cssClass: 'icon-thumbs-up-alt', 'label': Em.I18n.t('services.service.actions.run.smoke')});
        }
        options.push({action: 'refreshConfigs', cssClass: 'icon-refresh', 'label': Em.I18n.t('hosts.host.details.refreshConfigs'), disabled: disableRefreshConfgis});
      } else {
        // Restart All action
        options.push({action:'restartAllHostComponents', cssClass: 'icon-repeat', context: serviceName, 'label': Em.I18n.t('restart.service.all'), disabled: false});
        // Rolling Restart action
        var rrComponentName = batchUtils.getRollingRestartComponentName(serviceName);
        if (rrComponentName) {
          var label = Em.I18n.t('rollingrestart.dialog.title').format(App.format.role(rrComponentName));
          options.push({action:'rollingRestart', cssClass: 'icon-time', context: rrComponentName, 'label': label, disabled: false});
        }
        // Service Check and Reassign Master actions
        switch (serviceName) {
          case 'GANGLIA':
          case 'NAGIOS':
            break;
          case 'YARN':
          case 'HDFS':
          case 'MAPREDUCE':
            if (App.supports.reassignMaster && hosts > 1) {
              allMasters.forEach(function (hostComponent) {
                if (App.get('components.reassignable').contains(hostComponent)) {
                  options.push({action: 'reassignMaster', context: hostComponent, cssClass: 'icon-share-alt',
                    'label': Em.I18n.t('services.service.actions.reassign.master').format(App.format.role(hostComponent)), disabled: false});
                }
              })
            }
          default:
            options.push({action: 'runSmokeTest', cssClass: 'icon-thumbs-up-alt', 'label': Em.I18n.t('services.service.actions.run.smoke'), disabled:disabled});
        }
        var requestLabel = service.get('passiveState') === "OFF" ?
            Em.I18n.t('passiveState.turnOnFor').format(App.Service.DisplayNames[serviceName]) :
            Em.I18n.t('passiveState.turnOffFor').format(App.Service.DisplayNames[serviceName]);
        var passiveLabel = service.get('passiveState') === "OFF" ?
            Em.I18n.t('passiveState.turnOn') :
            Em.I18n.t('passiveState.turnOff');
        options.push({action:'turnOnOffPassive', cssClass: 'icon-medkit', context:requestLabel, 'label':passiveLabel , disabled: false});
      }
      return options;
    }.property('controller.content', 'controller.isStopDisabled'),
    isMaintenanceActive: function() {
      return this.get('maintenance').length !== 0;
    }.property('maintenance'),
    hasConfigTab: function(){
      return this.get("controller.content.isConfigurable");
    }.property('controller.content.isConfigurable'),

    didInsertElement: function () {
      this.get('controller').setStartStopState();
    },
    service:function () {
      var svc = this.get('controller.content');
      var svcName = svc.get('serviceName');
      if (svcName) {
        switch (svcName.toLowerCase()) {
          case 'hdfs':
            svc = App.HDFSService.find().objectAt(0);
            break;
          case 'yarn':
            svc = App.YARNService.find().objectAt(0);
            break;
          case 'mapreduce':
            svc = App.MapReduceService.find().objectAt(0);
            break;
          case 'hbase':
            svc = App.HBaseService.find().objectAt(0);
            break;
          case 'flume':
            svc = App.FlumeService.find().objectAt(0);
            break;
          default:
            break;
        }
      }
      return svc;
    }.property('controller.content.serviceName').volatile()
  });
  
});
window.require.register("views/main/service/manage_config_groups_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainServiceManageConfigGroupView = Em.View.extend({

    templateName: require('templates/main/service/manage_configuration_groups_popup'),

    selectedConfigGroup: null,

    isRemoveButtonDisabled: true,

    isRenameButtonDisabled: true,

    isDuplicateButtonDisabled: true,

    //Disable actions remove and rename for Default config group
    buttonObserver: function () {
      var selectedConfigGroup = this.get('controller.selectedConfigGroup');
      if(selectedConfigGroup.isDefault){
        this.set('isRemoveButtonDisabled', true);
        this.set('isRenameButtonDisabled', true);
        this.set('isDuplicateButtonDisabled', false);
      }else{
        this.set('isRemoveButtonDisabled', false);
        this.set('isRenameButtonDisabled', false);
        this.set('isDuplicateButtonDisabled', false);
      }
    }.observes('controller.selectedConfigGroup'),

    onGroupSelect: function () {
      var selectedConfigGroup = this.get('selectedConfigGroup');
      // to unable user select more than one config group at a time
      if (selectedConfigGroup.length) {
        this.set('controller.selectedConfigGroup', selectedConfigGroup[selectedConfigGroup.length - 1]);
      }
      if (selectedConfigGroup.length > 1) {
        this.set('selectedConfigGroup', selectedConfigGroup[selectedConfigGroup.length - 1]);
      }
      this.set('controller.selectedHosts', []);
    }.observes('selectedConfigGroup'),

    onLoad: function () {
      if (this.get('controller.isLoaded')) {
        this.set('selectedConfigGroup', this.get('controller.configGroups')[0])
      }
    }.observes('controller.isLoaded', 'controller.configGroups'),

    didInsertElement: function () {
      this.set('controller.isLoaded', false);
      this.get('controller').loadConfigGroups(this.get('serviceName'));
      this.selectDefaultGroup();
      App.tooltip($('.properties-link'));
      App.tooltip($("[rel='button-info']"));
      App.tooltip($("[rel='button-info-dropdown']"), {placement: 'left'});
    },

    selectDefaultGroup: function () {
      if (this.get('controller.isLoaded')) {
        this.set('selectedConfigGroup', [this.get('controller.configGroups').findProperty('isDefault')]);
      }
    }.observes('controller.isLoaded'),

    addButtonTooltip: function () {
      return  Em.I18n.t('services.service.config_groups_popup.addButton');
    }.property(),
    removeButtonTooltip: function () {
      return  Em.I18n.t('services.service.config_groups_popup.removeButton');
    }.property(),
    renameButtonTooltip: function () {
      return  Em.I18n.t('services.service.config_groups_popup.renameButton');
    }.property(),
    duplicateButtonTooltip: function () {
      return  Em.I18n.t('services.service.config_groups_popup.duplicateButton');
    }.property(),
    addHostTooltip: function () {
      if (!this.get('controller.selectedConfigGroup.isDefault') && this.get('controller.selectedConfigGroup.isAddHostsDisabled')) {
        return Em.I18n.t('services.service.config_groups_popup.addHostDisabled');
      } else {
        return  Em.I18n.t('services.service.config_groups_popup.addHost');
      }
    }.property('controller.selectedConfigGroup.isDefault', 'controller.selectedConfigGroup.isAddHostsDisabled'),
    removeHostTooltip: function () {
      return  Em.I18n.t('services.service.config_groups_popup.removeHost');
    }.property()

  });
  
});
window.require.register("views/main/service/menu", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var misc = require('utils/misc');

  App.MainServiceMenuView = Em.CollectionView.extend({
    disabledServices: ['HCATALOG'],

    content:function () {
      var items = App.router.get('mainServiceController.content').filter(function(item){
        return !this.get('disabledServices').contains(item.get('id'));
      }, this);
      return misc.sortByOrder(App.Service.servicesSortOrder, items);
    }.property('App.router.mainServiceController.content', 'App.router.mainServiceController.content.length'),

    didInsertElement:function () {
      App.router.location.addObserver('lastSetURL', this, 'renderOnRoute');
      this.renderOnRoute();
      App.tooltip($(".restart-required-service"), {html:true, placement:"right"});
    },

    activeServiceId:null,

    /**
     *    Syncs navigation menu with requested URL
     */
    renderOnRoute:function () {
      var last_url = App.router.location.lastSetURL || location.href.replace(/^[^#]*#/, '');
      if (last_url.substr(1, 4) !== 'main' || !this._childViews) {
        return;
      }
      var reg = /^\/main\/services\/(\S+)\//g;
      var sub_url = reg.exec(last_url);
      var service_id = (null != sub_url) ? sub_url[1] : 1;
      this.set('activeServiceId', service_id);

    },

    tagName:'ul',
    classNames:["nav", "nav-list", "nav-services"],

    itemViewClass:Em.View.extend({

      classNameBindings:["active", "clients"],
      templateName:require('templates/main/service/menu_item'),
      restartRequiredMessage: null,

      shouldBeRestarted: function() {
        return this.get('content.hostComponents').someProperty('staleConfigs', true);
      }.property('content.hostComponents.@each.staleConfigs'),

      active:function () {
        return this.get('content.id') == this.get('parentView.activeServiceId') ? 'active' : '';
      }.property('parentView.activeServiceId'),

      alertsCount: function () {
        return this.get('content.criticalAlertsCount');
      }.property('content.criticalAlertsCount'),

      link: function() {
        var stateName = (['summary','configs'].contains(App.router.get('currentState.name')))
          ? this.get('content.isConfigurable') ?  App.router.get('currentState.name') : 'summary'
          : 'summary';
        return "#/main/services/" + this.get('content.id') + "/" + stateName;
      }.property('App.router.currentState.name', 'parentView.activeServiceId'),

      refreshRestartRequiredMessage: function() {
        var restarted, componentsCount, hostsCount, message, tHosts, tComponents;
        restarted = this.get('content.restartRequiredHostsAndComponents');
        componentsCount = 0;
        hostsCount = 0;
        message = "";
        for (var host in restarted) {
          hostsCount++;
          componentsCount += restarted[host].length;
        }
        if (hostsCount > 1) {
          tHosts = Em.I18n.t('common.hosts');
        } else {
          tHosts = Em.I18n.t('common.host');
        }
        if (componentsCount > 1) {
          tComponents = Em.I18n.t('common.components');
        } else {
          tComponents = Em.I18n.t('common.component');
        }
        message += componentsCount + ' ' + tComponents + ' ' + Em.I18n.t('on') + ' ' +
          hostsCount + ' ' + tHosts + ' ' + Em.I18n.t('services.service.config.restartService.needToRestartEnd');
        this.set('restartRequiredMessage', message);
      }.observes('content.restartRequiredHostsAndComponents')
    })

  });
  
});
window.require.register("views/main/service/reassign/step1_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterWizardStep1View = Em.View.extend({

    message: function () {
      var messages = [
        Em.I18n.t('services.reassign.step1.message1').format(this.get('controller.content.reassign.display_name')),
        Em.I18n.t('services.reassign.step1.message3').format(this.get('controller.content.reassign.display_name'))
      ];
      if (this.get('controller.content.hasManualSteps')) {
        messages.splice(1,0, Em.I18n.t('services.reassign.step1.message2').format(this.get('controller.content.reassign.display_name')));
      }
      return messages;
    }.property('controller.content.reassign.display_name','controller.content.hasManualSteps'),

    templateName: require('templates/main/service/reassign/step1')

  });
  
});
window.require.register("views/main/service/reassign/step2_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterWizardStep2View = App.WizardStep5View.extend({
    body: function () {
      if (this.get('controller.content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
        return Em.I18n.t('services.reassign.step2.body.namenodeHA').format(this.get('controller.content.reassign.display_name'));
      }
      return Em.I18n.t('services.reassign.step2.body').format(this.get('controller.content.reassign.display_name'));
    }.property('controller.content.reassign.component_name', 'controller.content.reassign.display_name')
  });
  
});
window.require.register("views/main/service/reassign/step3_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterWizardStep3View = Em.View.extend({

    templateName: require('templates/main/service/reassign/step3'),

    sourceHost: function(){
      return this.get('controller.content.reassignHosts.source');
    }.property('controller.content.reassignHosts.source'),

    targetHost: function(){
      return this.get('controller.content.reassignHosts.target');
    }.property('controller.content.reassignHosts.target'),

    printReview: function () {
      $("#step8-info").jqprint();
    }
  });
  
});
window.require.register("views/main/service/reassign/step4_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterWizardStep4View = App.HighAvailabilityProgressPageView.extend({

    headerTitle: Em.I18n.t('services.reassign.step4.header'),

    noticeInProgress: function () {
      return Em.I18n.t('services.reassign.step4.status.info').format(App.format.role(this.get('controller.content.reassign.component_name')))
    }.property('controller.content.reassign.component_name'),

    noticeFailed: function () {
      return Em.I18n.t('services.reassign.step4.status.failed').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'));
    }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

    noticeCompleted: function () {
      if (this.get('controller.content.hasManualSteps')) {
        return Em.I18n.t('services.reassign.step4.status.success.withManualSteps').format(App.format.role(this.get('controller.content.reassign.component_name')));
      } else {
        return Em.I18n.t('services.reassign.step4.status.success').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'));
      }
    }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

    submitButtonText: function () {
      if (this.get('controller.content.hasManualSteps')) {
        return Em.I18n.t('common.next') + ' &rarr;';
      } else {
        return Em.I18n.t('common.complete');
      }
    }.property('controller.content.hasManualSteps'),

    templateName: require('templates/main/service/reassign/step4')
  });
  
});
window.require.register("views/main/service/reassign/step5_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterWizardStep5View = Em.View.extend({

    templateName: require('templates/main/service/reassign/step5'),

    manualCommands: function () {
      if (!this.get('controller.content.componentsWithManualCommands').contains(this.get('controller.content.reassign.component_name'))) {
        return '';
      }
      var componentDir = this.get('controller.content.componentDir');
      var sourceHost = this.get('controller.content.reassignHosts.source');
      var targetHost = this.get('controller.content.reassignHosts.target');
      var ha = '';
      if (this.get('controller.content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
        ha = '_ha';
        var nnStartedHost = this.get('controller.content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName').without(this.get('controller.content.reassignHosts.target'));
      }
      return  Em.I18n.t('services.reassign.step5.body.' + this.get('controller.content.reassign.component_name').toLowerCase() + ha).format(componentDir, sourceHost, targetHost, this.get('controller.content.hdfsUser'), nnStartedHost,this.get('controller.content.group'));
    }.property('controller.content.reassign.component_name', 'controller.content.componentDir', 'controller.content.masterComponentHosts', 'controller.content.reassign.host_id', 'controller.content.hdfsUser'),

    securityNotice: function () {
      var secureConfigs = this.get('controller.content.secureConfigs');
      var proceedMsg = Em.I18n.t('services.reassign.step5.body.proceedMsg');
      if (!this.get('controller.content.securityEnabled') || !secureConfigs.length) {
        return proceedMsg;
      }
      var formattedText = '<ul>';
      secureConfigs.forEach(function (config) {
        formattedText += '<li>' + Em.I18n.t('services.reassign.step5.body.securityConfigsList').format(config.keytab,
            config.principal.replace('_HOST', this.get('controller.content.reassignHosts.target')), this.get('controller.content.reassignHosts.target')) + '</li>';
      }, this);
      formattedText += '</ul>';
      return Em.I18n.t('services.reassign.step5.body.securityNotice').format(formattedText) + proceedMsg;
    }.property('controller.content.securityEnabled', 'controller.content.secureConfigs', 'controller.content.reassignHosts.target')
  });
  
});
window.require.register("views/main/service/reassign/step6_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterWizardStep6View = App.HighAvailabilityProgressPageView.extend({

    headerTitle: Em.I18n.t('services.reassign.step6.header'),

    noticeInProgress: function () {
      return Em.I18n.t('services.reassign.step6.status.info').format(App.format.role(this.get('controller.content.reassign.component_name')))
    }.property('controller.content.reassign.component_name'),

    noticeFailed: function () {
      return Em.I18n.t('services.reassign.step6.status.failed').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'))
    }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

    noticeCompleted: function () {
      return Em.I18n.t('services.reassign.step6.status.success').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'))
    }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

    submitButtonText: Em.I18n.t('common.complete'),

    templateName: require('templates/main/service/reassign/step6')
  });
  
});
window.require.register("views/main/service/reassign_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.ReassignMasterView = Em.View.extend({

    templateName: require('templates/main/service/reassign'),

    isStep1Disabled: function () {
      return this.isStepDisabled(1);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep2Disabled: function () {
      return this.isStepDisabled(2);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep3Disabled: function () {
      return this.isStepDisabled(3);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep4Disabled: function () {
      return this.isStepDisabled(4);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep5Disabled: function () {
      return this.isStepDisabled(5);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStep6Disabled: function () {
      return this.isStepDisabled(6);
    }.property('controller.isStepDisabled.@each.value').cacheable(),

    isStepDisabled: function (index) {
      return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
    }

  });
  
});
window.require.register("views/main/service/reconfigure", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  var App = require('app');

  App.MainServiceReconfigureView = Em.View.extend({

    templateName: require('templates/main/service/reconfigure')

  });

  App.StageLabelView = Em.View.extend({
    tagName: 'a',
    classNameBindings: ['removeLink'],
    attributeBindings: ['href'],
    href: '#',
    removeLink: null,
    didInsertElement: function() {
     this.onLink();
    },
    onLink: function() {
     if (this.get('showLink') === true) {
       this.set('removeLink',null);
     } else {
       this.set('removeLink','remove-link');
     }
    }.observes('showLink'),
    command: null,
    click: function () {
      if (this.get('command') && this.get('showLink')) {
        this.showHostPopup(this.get('command.label'));
      }
    },

    showHostPopup: function (label) {
      var serviceName = label;
      var controller = this.get("controller");
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        var popupView = App.HostPopup.initPopup(serviceName, controller);
        popupView.set ('isNotShowBgChecked', !initValue);
      })
    },

    isStarted: function () {
      return  (this.get('command') && this.get('command.isStarted'));
    }.property('command.isStarted'),

    showLink: function () {
      return (this.get('command') && this.get('command.showLink'));
    }.property('command.showLink')
  });

  App.StageSuccessView = Em.View.extend({
    layout: Ember.Handlebars.compile('<i class="icon-ok icon-large"></i> {{t common.done}}')
  });

  App.StageFailureView = Em.View.extend({
    layout: Ember.Handlebars.compile('<i class="icon-remove icon-large"></i> {{t common.failed}}')
  });

  App.StageInProgressView = Em.View.extend({
    command: null,
    classNames: ['progress-striped', 'active', 'progress'],
    template: Ember.Handlebars.compile('<div class="bar" {{bindAttr style="command.barWidth"}}></div>'),

    isStageCompleted: function () {
      return this.get('obj.progress') == 100 || this.get('controller.isStepCompleted');
    }.property('controller.isStepCompleted', 'obj.progress')

  });
});
window.require.register("views/main/views", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  App.MainViewsView = Em.View.extend({

    tagName: "iframe",

    classNames: ["views_sizes"],

    attributeBindings: ['src','seamless'],

    seamless: "seamless",

    src: function() {
      return window.location.origin + this.get('views')[1].link;
    }.property("views"),

    views: [{
        name: "filebrowser",
        link: "/views/FILE_BROWSER/FILEBROWSER_1/#/?path=/"
      },
      {
        name: "pig",
        link: "/views/PIG/PIG_1/"
    }]

  });
});
window.require.register("views/wizard/controls_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');

  /**
   * Abstract view for config fields.
   * Add popover support to control
   */
  App.ServiceConfigPopoverSupport = Ember.Mixin.create({

    /**
     * Config object. It will instance of App.ServiceConfigProperty
     */
    serviceConfig: null,
    attributeBindings:['readOnly'],
    isPopoverEnabled: true,

    didInsertElement: function () {
      // if description for this serviceConfig not exist, then no need to show popover
      if (this.get('isPopoverEnabled') !== 'false' && this.get('serviceConfig.description')) {
        App.popover(this.$(), {
          title: Em.I18n.t('installer.controls.serviceConfigPopover.title').format(
            this.get('serviceConfig.displayName'),
            (this.get('serviceConfig.displayName') == this.get('serviceConfig.name')) ? '' : this.get('serviceConfig.name')
          ),
          content: this.get('serviceConfig.description'),
          placement: 'right',
          trigger: 'hover'
        });
      }
    },

    willDestroyElement: function() {
      this.$().popover('destroy');
    },

    readOnly: function () {
      return !this.get('serviceConfig.isEditable');
    }.property('serviceConfig.isEditable')
  });

  /**
   * Default input control
   * @type {*}
   */
  App.ServiceConfigTextField = Ember.TextField.extend(App.ServiceConfigPopoverSupport, {

    valueBinding: 'serviceConfig.value',
    classNameBindings: 'textFieldClassName',
    placeholderBinding: 'serviceConfig.defaultValue',

    keyPress: function (event) {
      if (event.keyCode == 13) {
        return false;
      }
    },
    //Set editDone true for last edited config text field parameter
    focusOut: function (event) {
      this.get('serviceConfig').set("editDone", true);
    },
    //Set editDone false for all current category config text field parameter
    focusIn: function (event) {
      if (!this.get('serviceConfig.isOverridden')) {
        this.get("parentView.categoryConfigsAll").setEach("editDone", false);
      }
    },

    textFieldClassName: function () {
      if (this.get('serviceConfig.unit')) {
        return ['input-small'];
      } else if (this.get('serviceConfig.displayType') === 'principal') {
        return ['span12'];
      } else {
        return ['span9'];
      }
    }.property('serviceConfig.displayType', 'serviceConfig.unit')

  });

  /**
   * Customized input control with Utits type specified
   * @type {*}
   */
  App.ServiceConfigTextFieldWithUnit = Ember.View.extend(App.ServiceConfigPopoverSupport, {
    valueBinding: 'serviceConfig.value',
    classNames: ['input-append', 'with-unit'],
    placeholderBinding: 'serviceConfig.defaultValue',

    templateName: require('templates/wizard/controls_service_config_textfield_with_unit')
  });

  /**
   * Password control
   * @type {*}
   */
  App.ServiceConfigPasswordField = Ember.TextField.extend({

    serviceConfig: null,
    type: 'password',
    attributeBindings:['readOnly'],
    valueBinding: 'serviceConfig.value',
    classNames: [ 'span4' ],
    placeholder: Em.I18n.t('form.item.placeholders.typePassword'),

    template: Ember.Handlebars.compile('{{view view.retypePasswordView}}'),

    keyPress: function (event) {
      if (event.keyCode == 13) {
        return false;
      }
    },

    retypePasswordView: Ember.TextField.extend({
      placeholder: Em.I18n.t('form.passwordRetype'),
      attributeBindings:['readOnly'],
      type: 'password',
      classNames: [ 'span4', 'retyped-password' ],
      keyPress: function (event) {
        if (event.keyCode == 13) {
          return false;
        }
      },
      valueBinding: 'parentView.serviceConfig.retypedPassword',
      readOnly: function () {
        return !this.get('parentView.serviceConfig.isEditable');
      }.property('parentView.serviceConfig.isEditable')
    }),

    readOnly: function () {
      return !this.get('serviceConfig.isEditable');
    }.property('serviceConfig.isEditable')

  });

  /**
   * Textarea control
   * @type {*}
   */
  App.ServiceConfigTextArea = Ember.TextArea.extend(App.ServiceConfigPopoverSupport, {

    valueBinding: 'serviceConfig.value',
    rows: 4,
    classNames: ['span9', 'directories']
  });

  /**
   * Textarea control for content type
   * @type {*}
   */
  App.ServiceConfigTextAreaContent = Ember.TextArea.extend(App.ServiceConfigPopoverSupport, {

    valueBinding: 'serviceConfig.value',
    rows: 20,
    classNames: ['span10']
  });

  /**
   * Textarea control with bigger height
   * @type {*}
   */
  App.ServiceConfigBigTextArea = App.ServiceConfigTextArea.extend({
    rows: 10
  });

  /**
   * Checkbox control
   * @type {*}
   */
  App.ServiceConfigCheckbox = Ember.Checkbox.extend(App.ServiceConfigPopoverSupport, {

    checkedBinding: 'serviceConfig.value',

    disabled: function () {
      return !this.get('serviceConfig.isEditable');
    }.property('serviceConfig.isEditable')
  });

  App.ServiceConfigRadioButtons = Ember.View.extend({
    templateName: require('templates/wizard/controls_service_config_radio_buttons'),

    didInsertElement: function () {
      // on page render, automatically populate JDBC URLs only for default database settings
      // so as to not lose the user's customizations on these fields
      if (['addServiceController', 'installerController'].contains(App.clusterStatus.wizardControllerName) && ['New MySQL Database', 'New Derby Database'].contains(this.get('serviceConfig.value'))) {
        this.onOptionsChange();
      }
    },

    configs: function () {
      return this.get('categoryConfigsAll').filterProperty('isObserved', true);
    }.property('categoryConfigsAll'),

    serviceConfig: null,
    categoryConfigsAll: null,

    onOptionsChange: function () {
      // The following if condition will be satisfied only for installer wizard flow
      if (this.get('configs').length) {
        var connectionUrl = this.get('connectionUrl');
        var dbClass = this.get('dbClass');
        if (connectionUrl) {
          if (this.get('serviceConfig.serviceName') === 'HIVE') {
            switch (this.get('serviceConfig.value')) {
              case 'New MySQL Database':
              case 'Existing MySQL Database':
                connectionUrl.set('value', "jdbc:mysql://" + this.get('hostName') + "/" + this.get('databaseName') + "?createDatabaseIfNotExist=true");
                dbClass.set('value', "com.mysql.jdbc.Driver");
                break;
              case 'Existing Oracle Database':
                connectionUrl.set('value', "jdbc:oracle:thin:@//" + this.get('hostName') + ":1521/" + this.get('databaseName'));
                dbClass.set('value', "oracle.jdbc.driver.OracleDriver");
                break;
            }
          } else if (this.get('serviceConfig.serviceName') === 'OOZIE') {
            switch (this.get('serviceConfig.value')) {
              case 'New Derby Database':
                connectionUrl.set('value', "jdbc:derby:${oozie.data.dir}/${oozie.db.schema.name}-db;create=true");
                dbClass.set('value', "org.apache.derby.jdbc.EmbeddedDriver");
                break;
              case 'Existing MySQL Database':
                connectionUrl.set('value', "jdbc:mysql://" + this.get('hostName') + "/" + this.get('databaseName'));
                dbClass.set('value', "com.mysql.jdbc.Driver");
                break;
              case 'Existing Oracle Database':
                connectionUrl.set('value', "jdbc:oracle:thin:@//" + this.get('hostName') + ":1521/" + this.get('databaseName'));
                dbClass.set('value', "oracle.jdbc.driver.OracleDriver");
                break;
            }
          }
          connectionUrl.set('defaultValue', connectionUrl.get('value'));
        }
      }
    }.observes('databaseName', 'hostName', 'connectionUrl','dbClass'),

    nameBinding: 'serviceConfig.radioName',

    databaseName: function () {
      switch (this.get('serviceConfig.serviceName')) {
        case 'HIVE':
          return this.get('categoryConfigsAll').findProperty('name', 'ambari.hive.db.schema.name').get('value');
        case 'OOZIE':
          return this.get('categoryConfigsAll').findProperty('name', 'oozie.db.schema.name').get('value');
        default:
          return null;
      }
    }.property('configs.@each.value', 'serviceConfig.serviceName'),


    hostName: function () {
      var value = this.get('serviceConfig.value');
      var returnValue;
      var hostname;
      if (this.get('serviceConfig.serviceName') === 'HIVE') {
        switch (value) {
          case 'New MySQL Database':
            hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_ambari_host');
            break;
          case 'Existing MySQL Database':
            hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_mysql_host');
            break;
          case 'Existing Oracle Database':
            hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_oracle_host');
            break;
        }
        if (hostname) {
          returnValue = hostname.get('value');
        } else {
          returnValue = this.get('categoryConfigsAll').findProperty('name', 'hive_hostname').get('value');
        }

      } else if (this.get('serviceConfig.serviceName') === 'OOZIE') {
        switch (value) {
          case 'New Derby Database':
            hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_ambari_host');
            break;
          case 'Existing MySQL Database':
            hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_mysql_host');
            break;
          case 'Existing Oracle Database':
            hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_oracle_host');
            break;
        }
        if (hostname) {
          returnValue = hostname.get('value');
        } else {
          returnValue = this.get('categoryConfigsAll').findProperty('name', 'oozie_hostname').get('value');
        }
      }
      return returnValue;
    }.property('serviceConfig.serviceName', 'serviceConfig.value', 'configs.@each.value'),

    connectionUrl: function () {
      if (this.get('serviceConfig.serviceName') === 'HIVE') {
        return this.get('categoryConfigsAll').findProperty('name', 'javax.jdo.option.ConnectionURL');
      } else {
        return this.get('categoryConfigsAll').findProperty('name', 'oozie.service.JPAService.jdbc.url');
      }
    }.property('serviceConfig.serviceName'),

    dbClass: function () {
      if (this.get('serviceConfig.serviceName') === 'HIVE') {
        return this.get('categoryConfigsAll').findProperty('name', 'javax.jdo.option.ConnectionDriverName');
      } else {
        return this.get('categoryConfigsAll').findProperty('name', 'oozie.service.JPAService.jdbc.driver');
      }
    }.property('serviceConfig.serviceName'),

    optionsBinding: 'serviceConfig.options'
  });

  App.ServiceConfigRadioButton = Ember.Checkbox.extend({
    tagName: 'input',
    attributeBindings: ['type', 'name', 'value', 'checked'],
    checked: false,
    type: 'radio',
    name: null,
    value: null,

    didInsertElement: function () {
      console.debug('App.ServiceConfigRadioButton.didInsertElement');
      if (this.get('parentView.serviceConfig.value') === this.get('value')) {
        console.debug(this.get('name') + ":" + this.get('value') + ' is checked');
        this.set('checked', true);
      }
    },

    click: function () {
      this.set('checked', true);
      console.debug('App.ServiceConfigRadioButton.click');
      this.onChecked();
    },

    onChecked: function () {
      // Wrapping the call with Ember.run.next prevents a problem where setting isVisible on component
      // causes JS error due to re-rendering.  For example, this occurs when switching the Config Group
      // in Service Config page
      Em.run.next(this, function() {
        console.debug('App.ServiceConfigRadioButton.onChecked');
        this.set('parentView.serviceConfig.value', this.get('value'));
        var components = this.get('parentView.serviceConfig.options');
        if (components) {
          components.forEach(function (_component) {
            if (_component.foreignKeys) {
              _component.foreignKeys.forEach(function (_componentName) {
                if (this.get('parentView.categoryConfigsAll').someProperty('name', _componentName)) {
                  var component = this.get('parentView.categoryConfigsAll').findProperty('name', _componentName);
                  component.set('isVisible', _component.displayName === this.get('value'));
                }
              }, this);
            }
          }, this);
        }
      });
    }.observes('checked'),

    disabled: function () {
      return !this.get('parentView.serviceConfig.isEditable');
    }.property('parentView.serviceConfig.isEditable')
  });

  App.ServiceConfigComboBox = Ember.Select.extend(App.ServiceConfigPopoverSupport, {
    contentBinding: 'serviceConfig.options',
    selectionBinding: 'serviceConfig.value',
    placeholderBinding: 'serviceConfig.defaultValue',
    classNames: [ 'span3' ]
  });


  /**
   * Base component for host config with popover support
   */
  App.ServiceConfigHostPopoverSupport = Ember.Mixin.create({

    /**
     * Config object. It will instance of App.ServiceConfigProperty
     */
    serviceConfig: null,

    didInsertElement: function () {
      App.popover(this.$(), {
        title: this.get('serviceConfig.displayName'),
        content: this.get('serviceConfig.description'),
        placement: 'right',
        trigger: 'hover'
      });
    }
  });

  /**
   * Master host component.
   * Show hostname without ability to edit it
   * @type {*}
   */
  App.ServiceConfigMasterHostView = Ember.View.extend(App.ServiceConfigHostPopoverSupport, {

    classNames: ['master-host', 'span6'],
    valueBinding: 'serviceConfig.value',

    template: Ember.Handlebars.compile('{{value}}')

  });

  /**
   * Base component to display Multiple hosts
   * @type {*}
   */
  App.ServiceConfigMultipleHostsDisplay = Ember.Mixin.create(App.ServiceConfigHostPopoverSupport, {

    hasNoHosts: function () {
      console.log('view', this.get('viewName')); //to know which View cause errors
      console.log('controller', this.get('controller').name); //should be slaveComponentGroupsController
      if (!this.get('value')) {
        return true;
      }
      return this.get('value').length === 0;
    }.property('value'),

    hasOneHost: function () {
      return this.get('value').length === 1;
    }.property('value'),

    hasMultipleHosts: function () {
      return this.get('value').length > 1;
    }.property('value'),

    otherLength: function () {
      var len = this.get('value').length;
      if (len > 2) {
        return Em.I18n.t('installer.controls.serviceConfigMultipleHosts.others').format(len - 1);
      } else {
        return Em.I18n.t('installer.controls.serviceConfigMultipleHosts.other');
      }
    }.property('value')

  });


  /**
   * Multiple master host component.
   * Show hostnames without ability to edit it
   * @type {*}
   */
  App.ServiceConfigMasterHostsView = Ember.View.extend(App.ServiceConfigMultipleHostsDisplay, {

    viewName: "serviceConfigMasterHostsView",
    valueBinding: 'serviceConfig.value',

    classNames: ['master-hosts', 'span6'],
    templateName: require('templates/wizard/master_hosts'),

    /**
     * Onclick handler for link
     */
    showHosts: function () {
      var serviceConfig = this.get('serviceConfig');
      App.ModalPopup.show({
        header: Em.I18n.t('installer.controls.serviceConfigMasterHosts.header').format(serviceConfig.category),
        bodyClass: Ember.View.extend({
          serviceConfig: serviceConfig,
          templateName: require('templates/wizard/master_hosts_popup')
        }),
        secondary: null
      });
    }

  });

  /**
   * Show tabs list for slave hosts
   * @type {*}
   */
  App.SlaveComponentGroupsMenu = Em.CollectionView.extend({

    content: function () {
      return this.get('controller.componentGroups');
    }.property('controller.componentGroups'),

    tagName: 'ul',
    classNames: ["nav", "nav-tabs"],

    itemViewClass: Em.View.extend({
      classNameBindings: ["active"],

      active: function () {
        return this.get('content.active');
      }.property('content.active'),

      errorCount: function () {
        return this.get('content.properties').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
      }.property('content.properties.@each.isValid', 'content.properties.@each.isVisible'),

      templateName: require('templates/wizard/controls_slave_component_groups_menu')
    })
  });

  /**
   * <code>Add group</code> button
   * @type {*}
   */
  App.AddSlaveComponentGroupButton = Ember.View.extend({

    tagName: 'span',
    slaveComponentName: null,

    didInsertElement: function () {
      App.popover(this.$(), {
        title: Em.I18n.t('installer.controls.addSlaveComponentGroupButton.title').format(this.get('slaveComponentName')),
        content: Em.I18n.t('installer.controls.addSlaveComponentGroupButton.content').format(this.get('slaveComponentName'), this.get('slaveComponentName'), this.get('slaveComponentName')),
        placement: 'right',
        trigger: 'hover'
      });
    }

  });

  /**
   * Multiple Slave Hosts component
   * @type {*}
   */
  App.ServiceConfigSlaveHostsView = Ember.View.extend(App.ServiceConfigMultipleHostsDisplay, {

    viewName: 'serviceConfigSlaveHostsView',

    classNames: ['slave-hosts', 'span6'],

    valueBinding: 'serviceConfig.value',

    templateName: require('templates/wizard/slave_hosts'),

    /**
     * Onclick handler for link
     */
    showHosts: function () {
      var serviceConfig = this.get('serviceConfig');
      App.ModalPopup.show({
        header: Em.I18n.t('installer.controls.serviceConfigMasterHosts.header').format(serviceConfig.category),
        bodyClass: Ember.View.extend({
          serviceConfig: serviceConfig,
          templateName: require('templates/wizard/master_hosts_popup')
        }),
        secondary: null
      });
    }

  });

  /**
   * properties for present active slave group
   * @type {*}
   */
  App.SlaveGroupPropertiesView = Ember.View.extend({

    viewName: 'serviceConfigSlaveHostsView',

    group: function () {
      return this.get('controller.activeGroup');
    }.property('controller.activeGroup'),

    groupConfigs: function () {
      console.log("************************************************************************");
      console.log("The value of group is: " + this.get('group'));
      console.log("************************************************************************");
      return this.get('group.properties');
    }.property('group.properties.@each').cacheable(),

    errorCount: function () {
      return this.get('group.properties').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
    }.property('configs.@each.isValid', 'configs.@each.isVisible')
  });

  /**
   * DropDown component for <code>select hosts for groups</code> popup
   * @type {*}
   */
  App.SlaveComponentDropDownGroupView = Ember.View.extend({

    viewName: "slaveComponentDropDownGroupView",

    /**
     * On change handler for <code>select hosts for groups</code> popup
     * @param event
     */
    changeGroup: function (event) {
      var host = this.get('content');
      var groupName = $('#' + this.get('elementId') + ' select').val();
      this.get('controller').changeHostGroup(host, groupName);
    },

    optionTag: Ember.View.extend({

      /**
       * Whether current value(OptionTag value) equals to host value(assigned to SlaveComponentDropDownGroupView.content)
       */
      selected: function () {
        return this.get('parentView.content.group') === this.get('content');
      }.property('content')
    })
  });

  /**
   * Show info about current group
   * @type {*}
   */
  App.SlaveComponentChangeGroupNameView = Ember.View.extend({

    contentBinding: 'controller.activeGroup',
    classNames: ['control-group'],
    classNameBindings: 'error',
    error: false,
    setError: function () {
      this.set('error', false);
    }.observes('controller.activeGroup'),
    errorMessage: function () {
      return this.get('error') ? Em.I18n.t('installer.controls.slaveComponentChangeGroupName.error') : '';
    }.property('error'),

    /**
     * Onclick handler for saving updated group name
     * @param event
     */
    changeGroupName: function (event) {
      var inputVal = $('#' + this.get('elementId') + ' input[type="text"]').val();
      if (inputVal !== this.get('content.name')) {
        var result = this.get('controller').changeSlaveGroupName(this.get('content'), inputVal);
        this.set('error', result);
      }
    }
  });

  
});
window.require.register("views/wizard/stack_upgrade/step1_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.StackUpgradeStep1View = Em.View.extend({
    templateName: require('templates/wizard/stack_upgrade/step1')
  });
  
});
window.require.register("views/wizard/stack_upgrade/step2_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.StackUpgradeStep2View = Em.View.extend({
    templateName: require('templates/wizard/stack_upgrade/step2'),
    didInsertElement: function(){
      App.popover($("[rel=popover]"), {'placement': 'right', 'trigger': 'hover'});
    }
  });
  
});
window.require.register("views/wizard/stack_upgrade/step3_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.StackUpgradeStep3View = Em.View.extend({
    templateName: require('templates/wizard/stack_upgrade/step3'),
    overallStatus: Em.I18n.t('installer.stackUpgrade.step3.status.info'),
    statusClass: 'alert-info',
    didInsertElement: function(){
      this.get('controller').resumeStep();
      this.onStatus();
    },
    onStatus: function(){
      var currentProcess = this.get('controller.processes').findProperty('isRunning', true);
      if(currentProcess && (currentProcess.get('name') == 'UPGRADE_SERVICES')){
        switch (currentProcess.get('status')){
          case 'SUCCESS':
            this.set('overallStatus', Em.I18n.t('installer.stackUpgrade.step3.status.success').format(this.get('controller.content.upgradeVersion')));
            this.set('statusClass', 'alert-success');
            break;
          case 'FAILED':
            this.set('overallStatus', Em.I18n.t('installer.stackUpgrade.step3.status.failed'));
            this.set('statusClass', 'alert-error');
            break;
          case 'WARNING':
            this.set('overallStatus', Em.I18n.t('installer.stackUpgrade.step3.status.warning').format(this.get('controller.content.upgradeVersion')));
            this.set('statusClass', 'alert-block');
            break;
          case 'IN_PROGRESS':
          default:
            this.set('overallStatus', Em.I18n.t('installer.stackUpgrade.step3.status.info'));
            this.set('statusClass', 'alert-info');
        }
      } else {
        this.set('overallStatus', Em.I18n.t('installer.stackUpgrade.step3.status.info'));
        this.set('statusClass', 'alert-info');
      }
    }.observes('controller.processes.@each.status'),
    processView: Em.View.extend({
      barColor: '',
      icon:'',
      iconColor:'',
      status: function(){
        var process = this.get('content');
        if(process.get('name') == 'STOP_SERVICES'){
          switch(process.get('status')){
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.stackUpgrade.step3.service.stopping');
            case 'FAILED':
              return Em.I18n.t('installer.stackUpgrade.step3.service.stopFail');
            case 'SUCCESS':
              return Em.I18n.t('installer.stackUpgrade.step3.service.stopped');
            default:
              return Em.I18n.t('installer.stackUpgrade.step3.service.stopPending');
          }
        } else {
          switch(process.get('status')){
            case 'IN_PROGRESS':
              return Em.I18n.t('installer.stackUpgrade.step3.service.upgrading');
            case 'WARNING':
              return Em.I18n.t('installer.stackUpgrade.step3.service.upgraded');
            case 'FAILED':
              return Em.I18n.t('installer.stackUpgrade.step3.service.failedUpgrade');
            case 'SUCCESS':
              return Em.I18n.t('installer.stackUpgrade.step3.service.upgraded');
            default:
              return Em.I18n.t('installer.stackUpgrade.step3.service.pending');
          }
        }
      }.property('content.status'),
      didInsertElement: function(){
        this.onStatus();
      },
      retryMessage: function(){
        if(this.get('content.name') == 'STOP_SERVICES'){
          return Em.I18n.t('installer.stackUpgrade.step3.retry.services');
        }
        return Em.I18n.t('installer.stackUpgrade.step3.retry.upgrade');
      }.property('content.name'),
      isProcessCompleted: function(){
        return this.get('content.status') == 'SUCCESS' || this.get('content.status') === 'WARNING';
      }.property('content.status'),
      barWidth: function(){
        return 'width: ' + this.get('content.progress') + '%;';
      }.property('content.progress'),
      /**
       * change service appearance(icon, progress-bar color,) depending on the service status
       */
      onStatus:function () {
        if (this.get('content.status') === 'IN_PROGRESS') {
          this.set('barColor', 'progress-info');
          this.set('icon', 'icon-cog');
          this.set('iconColor', 'text-info');
        } else if (this.get('content.status') === 'WARNING') {
          this.set('barColor', 'progress-warning');
          this.set('icon', 'icon-warning-sign');
          this.set('iconColor', 'text-warning');
        } else if (this.get('content.status') === 'FAILED') {
          this.set('barColor', 'progress-danger');
          this.set('icon', 'icon-exclamation-sign');
          this.set('iconColor', 'text-error');
        } else if (this.get('content.status') === 'SUCCESS') {
          this.set('barColor', 'progress-success');
          this.set('icon', 'icon-ok');
          this.set('iconColor', 'text-success');
        } else {
          this.set('barColor', 'progress-info');
          this.set('icon', 'icon-cog');
          this.set('iconColor', '');
        }
      }.observes('content.status'),
      inProgress: function(){
        return this.get('content.isRunning') && !this.get('content.isRetry');
      }.property('content.status'),
      /**
       * open popup with list of hosts, that associated to service
       * @param event
       */
      hostsLogPopup: function(event){
        var serviceName = event.contexts[0];
        var controller = this.get("controller");
        App.router.get('applicationController').dataLoading().done(function (initValue) {
          var popupView = App.HostPopup.initPopup(serviceName, controller);
          popupView.set ('isNotShowBgChecked', !initValue);
        })
      }
    })
  });
});
window.require.register("views/wizard/step0_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep0View = Em.View.extend({

    tagName: "form", //todo: why form?
    attributeBindings: ['autocomplete'],
    autocomplete: 'off',
    templateName: require('templates/wizard/step0'),

    //todo: create property for placeholder(go to template)

    didInsertElement: function () {
      App.popover($("[rel=popover]"), {'placement': 'right', 'trigger': 'hover'});
      this.get('controller').loadStep();
    },

    //todo: rename it to class or write comments
    onError: function () {
      return this.get('controller.clusterNameError') !== '';
    }.property('controller.clusterNameError')

  });

  App.WizardStep0ViewClusterNameInput = Em.TextField.extend({
    keyPress: function(event) {
      if (event.keyCode == 13) {
        this.get('parentView.controller').submit();
        return false;
      }
    }
  });
  
});
window.require.register("views/wizard/step10_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep10View = Em.View.extend({

    templateName: require('templates/wizard/step10'),

    didInsertElement: function () {
      var controller = this.get('controller');
      controller.loadStep();
    }

  });
});
window.require.register("views/wizard/step1_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one or more
   * contributor license agreements. See the NOTICE file distributed with this
   * work for additional information regarding copyright ownership. The ASF
   * licenses this file to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   * 
   * http://www.apache.org/licenses/LICENSE-2.0
   * 
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
   * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
   * License for the specific language governing permissions and limitations under
   * the License.
   */

  var App = require('app');

  App.WizardStep1View = Em.View.extend({

    templateName: require('templates/wizard/step1'),

    /**
     * List of available stacks
     * @type {Em.Object[]}
     */
    stacks: function () {
      var stacks = [];
      this.get('controller.content.stacks').forEach(function (stack) {
        stacks.pushObject(Em.Object.create({
          name: stack.get('name').replace('-', ' '),
          isSelected: stack.get('isSelected')
        }));
      });
      return stacks;
    }.property('controller.content.stacks.@each.isSelected'),

    /**
     * Checkbox for each stack
     * @type {Em.Checkbox}
     */
    stackRadioButton: Em.Checkbox.extend({
      tagName: 'input',
      attributeBindings: [ 'type', 'checked' ],
      checked: function () {
        return this.get('content.isSelected');
      }.property('content.isSelected'),
      type: 'radio',

      click: function () {
        this.get('controller.content.stacks').setEach('isSelected', false);
        this.get('controller.content.stacks').findProperty('name', this.get('content.name').replace(' ', '-')).set('isSelected', true);
      }
    }),

    /**
     * List of all repo-groups
     * @type {Object[][]}
     */
    allRepositoriesGroup: [[],[],[]],

    /**
     * Verify if some repo has empty base-url
     * @type {bool}
     */
    emptyRepoExist: function () {
      return (this.get('allRepositoriesGroup').filterProperty('empty-error', true).length != 0);
    }.property('allRepositoriesGroup.@each.empty-error'),

    /**
     * Disable submit button flag
     * @type {bool}
     */
    isSubmitDisabled: function() {
      return this.get('emptyRepoExist') || this.get('allRepoUnchecked') || this.get('invalidUrlExist') ;
    }.property('emptyRepoExist', 'allRepoUnchecked', 'invalidUrlExist'),

    /**
     * Verify if some invalid repo-urls exist
     * @type {bool}
     */
    invalidUrlExist: function () {
      var selectedStack = this.get('controller.content.stacks').findProperty('isSelected', true);
      var invalidExist = this.get('allRepositoriesGroup').filterProperty('validation', 'icon-exclamation-sign').length != 0;
      return (selectedStack.get('invalidCnt') > 0) && invalidExist;
    }.property('controller.content.stacks.@each.invalidCnt', 'allRepositoriesGroup.@each.validation'),

    /**
     * If all repo links are unchecked
     * @type {bool}
     */
    allRepoUnchecked: function () {
      return (!this.get('allRepositoriesGroup').filterProperty('checked', true).length);
    }.property('allRepositoriesGroup.@each.checked'),

    /**
     * Overall errors count
     * @type {number}
     */
    totalErrorCnt: function () {
      var emptyCnt = this.get('allRepositoriesGroup').filterProperty('empty-error', true).length;
      var invalidCnt = this.get('allRepositoriesGroup').filterProperty('validation', 'icon-exclamation-sign').length;
      if (this.get('allRepoUnchecked')) {
        return 1;
      } else if ( emptyCnt || invalidCnt) {
        return emptyCnt + invalidCnt;
      } else {
        return 0;
      }
    }.property('allRepositoriesGroup.@each.empty-error', 'allRepoUnchecked', 'allRepositoriesGroup.@each.validation'),

    /**
     * Onclick handler for Config Group Header. Used to show/hide block
     */
    onToggleBlock: function () {
      this.$('.accordion-body').toggle('blind', 500);
      this.set('isRLCollapsed', !this.get('isRLCollapsed'));
    },

    /**
     * Is Repositories Accordion collapsed
     * @type {bool}
     */
    isRLCollapsed: true,

    didInsertElement: function () {
      if (this.get('isRLCollapsed')) {
        this.$('.accordion-body').hide();
      }
      $("[rel=skip-validation-tooltip]").tooltip({ placement: 'right'});
    },

    /**
     * Popover for repo-url error indicator
     * @type {Em.View}
     */
    popoverView: Em.View.extend({
      tagName: 'i',
      classNameBindings: ['repoGroup.validation'],
      attributeBindings: ['repoGroup.errorTitle:title', 'repoGroup.errorContent:data-content'],
      didInsertElement: function() {
        App.popover($(this.get('element')), {'trigger': 'hover'});
      }
    }),

    /**
     * Format repo-group values and set it to <code>allRepositoriesGroup</code>
     */
    loadRepositories: function () {
      var selectedStack = this.get('controller.content.stacks').findProperty('isSelected', true);
      var reposGroup = [[],[],[]];
      if (App.supports.ubuntu) reposGroup.push([]); // @todo: remove after Ubuntu support confirmation
      var self = this;
      if (selectedStack && selectedStack.operatingSystems) {
        selectedStack.operatingSystems.forEach(function (os) {
          var cur_repo = Em.Object.create({
            baseUrl: os.baseUrl
          });
          switch(os.osType) {
            case 'redhat5':
              cur_repo.set('osType', 'Red Hat 5');
              reposGroup[0][0] = cur_repo;
              // set group 0 properties by redhat5 (any of the three is ok)
              self.setGroupByOs(reposGroup[0], os, 0);
              break;
            case 'centos5':
              cur_repo.set('osType', 'CentOS 5');
              reposGroup[0][1] = cur_repo;
              break;
            case 'oraclelinux5':
              cur_repo.set('osType', 'Oracle Linux 5');
              reposGroup[0][2] = cur_repo;
              break;
            case 'redhat6':
              cur_repo.set('osType', 'Red Hat 6');
              reposGroup[1][0] = cur_repo;
              // set group 1 properties by redhat6 (any of the three is ok)
              self.setGroupByOs(reposGroup[1], os, 1);
              break;
            case 'centos6':
              cur_repo.set('osType', 'CentOS 6');
              reposGroup[1][1] = cur_repo;
              break;
            case 'oraclelinux6':
              cur_repo.set('osType', 'Oracle Linux 6');
              reposGroup[1][2] = cur_repo;
              break;
            case 'sles11':
              cur_repo.set('osType', 'SLES 11');
              reposGroup[2][0] = cur_repo;
              // set group 2 properties by sles11 (any of the twe is ok)
              self.setGroupByOs(reposGroup[2], os, 2);
              break;
            case 'suse11':
              cur_repo.set('osType', 'SUSE 11');
              reposGroup[2][1] = cur_repo;
              break;
            case 'ubuntu12':
              cur_repo.set('osType','Ubuntu 12');
              reposGroup[3][0] = cur_repo;
              self.setGroupByOs(reposGroup[3], os, 3);
              break;
          }
        });
      }
      this.set('allRepositoriesGroup', reposGroup);
    }.observes('controller.content.stacks.@each.isSelected', 'controller.content.stacks.@each.reload'),

    /**
     * Set group parameters according to operation system
     * @param {Em.Object} group
     * @param {Object} os
     * @param {number} groupNumber
     */
    setGroupByOs: function (group, os, groupNumber) {
      var isChecked = this.get('allGroupsCheckbox')[groupNumber];
      group.set('checked', isChecked);
      group.set('baseUrl', os.baseUrl);
      group.set('latestBaseUrl', os.latestBaseUrl);
      group.set('defaultBaseUrl', os.defaultBaseUrl);
      group.set('empty-error', !os.baseUrl);
      group.set('invalid-error', os.validation == 'icon-exclamation-sign');
      group.set('validation', os.validation);
      group.set('undo', os.baseUrl != os.latestBaseUrl);
      group.set('clearAll', os.baseUrl);
      group.set('errorTitle', os.errorTitle);
      group.set('errorContent', os.errorContent);
      group.set('group-number', groupNumber);
    },

    /**
     * Onclick handler for checkbox of each repo group
     */
    updateByCheckbox: function () {
      //upload to content
      var groups = this.get('allRepositoriesGroup');
      var self = this;
      var selectedStack = this.get('controller.content.stacks').findProperty('isSelected', true);
      if (selectedStack && selectedStack.operatingSystems) {
        selectedStack.operatingSystems.forEach(function (os) {
          var groupNumber = self.osTypeToGroup(os.osType);
          var targetGroup = groups.findProperty('group-number', groupNumber);
          if (!targetGroup.get('checked')) {
            os.baseUrl = os.latestBaseUrl;
            os.validation = null;
            os.selected = false;
            targetGroup.set('baseUrl', os.latestBaseUrl);
            targetGroup.set('latestBaseUrl', os.latestBaseUrl);
            targetGroup.set('undo', targetGroup.get('baseUrl') != targetGroup.get('latestBaseUrl'));
            targetGroup.set('invalid-error', false);
            targetGroup.set('validation', null);
            targetGroup.set('clearAll', false);
            targetGroup.set('empty-error',!targetGroup.get('baseUrl'));
            self.get('allGroupsCheckbox')[groupNumber] = false;
            self.set('allGroupsCheckbox', self.get('allGroupsCheckbox'));
          } else {
            os.selected = true;
            os.skipValidation = self.get('skipValidationChecked');
            if (os.skipValidation) {
              targetGroup.set('validation', null);
              targetGroup.set('invalid-error', false);
            }
            targetGroup.set('clearAll', targetGroup.get('baseUrl'));
            targetGroup.set('empty-error',!targetGroup.get('baseUrl'));
            self.get('allGroupsCheckbox')[groupNumber] = true;
          }
        });
      }
    }.observes('allRepositoriesGroup.@each.checked', 'skipValidationChecked'),

    /**
     * Checked flags for each repo-checkbox
     * @type {bool[]}
     */
    allGroupsCheckbox: [true, true, true],

    /**
     * Skip repo-validation
     * @type {bool}
     */
    skipValidationChecked: false,

    /**
     * Onclick handler for undo action of each repo group
     */
    undoGroupLocalRepository: function (event) {
      var group = event.context;
      var osTypes = this.groupToOsType(group.get('group-number'));
      var selectedStack = this.get('controller.content.stacks').findProperty('isSelected', true);
      osTypes.forEach( function (os) {
        var cos = selectedStack.operatingSystems.findProperty('osType', os );
        cos.baseUrl = cos.latestBaseUrl;
        cos.validation = null;
      });
      this.loadRepositories();
    },

    /**
     * Handler for clear icon click
     * @param {object} event
     */
    clearGroupLocalRepository: function (event) {
      var group = event.context;
      var osTypes = this.groupToOsType(group.get('group-number'));
      var selectedStack = this.get('controller.content.stacks').findProperty('isSelected', true);
      osTypes.forEach( function (os) {
        var cos = selectedStack.operatingSystems.findProperty('osType', os );
        cos.baseUrl = '';
        cos.validation = null;
      });
      this.loadRepositories();
    },

    /**
     * Handler when editing any repo group BaseUrl
     * @param {object} event
     */
    editGroupLocalRepository: function (event) {
      //upload to content
      var groups = this.get('allRepositoriesGroup');
      var self = this;
      var selectedStack = this.get('controller.content.stacks').findProperty('isSelected', true);
      if (selectedStack && selectedStack.operatingSystems) {
        selectedStack.operatingSystems.forEach(function (os) {
          var targetGroup = groups.findProperty('group-number', self.osTypeToGroup(os.osType));
          if (os.baseUrl != targetGroup.get('baseUrl')) {
            os.baseUrl = targetGroup.get('baseUrl');
            os.validation = null;
            targetGroup.set('undo', targetGroup.get('baseUrl') != targetGroup.get('latestBaseUrl'));
            targetGroup.set('invalid-error', false);
            targetGroup.set('validation', null);
            targetGroup.set('clearAll', os.baseUrl);
            targetGroup.set('empty-error',!targetGroup.get('baseUrl'));
          }
        });
      }
    }.observes('allRepositoriesGroup.@each.baseUrl'),

    /**
     * Get list of OS for provided group number
     * @param {number} groupNumber
     * @returns {Array}
     */
    groupToOsType: function (groupNumber) {
      switch (groupNumber) {
        case 0:
          return ['redhat5', 'centos5', 'oraclelinux5'];
        case 1:
          return ['redhat6', 'centos6', 'oraclelinux6'];
        case 2:
          return ['sles11', 'suse11'];
        case 3:
          return ['ubuntu12'];
      }
      return [];
    },

    /**
     * Get group number for provided OS
     * @param  {string} osType
     * @returns {number}
     */
    osTypeToGroup: function (osType) {
      switch(osType) {
        case 'redhat5':
        case 'centos5':
        case 'oraclelinux5':
          return 0;
        case 'redhat6':
        case 'centos6':
        case 'oraclelinux6':
          return 1;
        case 'sles11':
        case 'suse11':
          return 2;
        case 'ubuntu12':
          return 3;
      }
      return -1;
    }
  });
  
});
window.require.register("views/wizard/step2_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * 'License'); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an 'AS IS' BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.SshKeyFileUploader = Ember.View.extend({
    //TODO: rewrite it using tagName and attribute binding
    //TODO: rewrite it as independent component and place it somewhere in utils
    // alternative is to move it to App.WizardStep2View
    template:Ember.Handlebars.compile('<input type="file" {{bindAttr disabled="view.disabled"}} />'),
    classNames: ['ssh-key-input-indentation'],

    change: function (e) {
      var self=this;
      if (e.target.files && e.target.files.length == 1) {
        var file = e.target.files[0];
        var reader = new FileReader();

        reader.onload = (function(theFile) {
          return function(e) {
            $('#sshKey').html(e.target.result);
            self.get("controller").setSshKey(e.target.result);
          };
        })(file);
        reader.readAsText(file);
      }
    }
  });

  App.WizardStep2View = Em.View.extend({

    templateName: require('templates/wizard/step2'),

    didInsertElement: function () {
      //TODO: move it to separate function in Ember.View using reopenClass
      App.popover($("[rel=popover]"), {'placement': 'right', 'trigger': 'hover'});

      //todo: move them to conroller
      this.set('controller.hostsError',null);
      this.set('controller.sshKeyError',null);
    },

    sshKeyState: function(){
      return this.get("controller.content.installOptions.manualInstall");
    }.property("controller.content.installOptions.manualInstall"),

    //TODO: incupsulate it inside of App.SshKeyFileUploader
    isFileApi: function () {
      return (window.File && window.FileReader && window.FileList) ? true : false ;
    }.property(),

    //TODO: replace next 2 properties with new one used in both places
    providingSSHKeyRadioButton: Ember.Checkbox.extend({
      tagName: 'input',
      attributeBindings: ['type', 'checked'],
      checked: function () {
        return this.get('controller.content.installOptions.useSsh');
      }.property('controller.content.installOptions.useSsh'),
      type: 'radio',

      click: function () {
        this.set('controller.content.installOptions.useSsh', true);
        this.set('controller.content.installOptions.manualInstall', false);
      }
    }),

    manualRegistrationRadioButton: Ember.Checkbox.extend({
      tagName: 'input',
      attributeBindings: ['type', 'checked'],
      checked: function () {
        return this.get('controller.content.installOptions.manualInstall');
      }.property('controller.content.installOptions.manualInstall'),
      type: 'radio',

      click: function () {
        this.set('controller.content.installOptions.manualInstall', true);
        this.set('controller.content.installOptions.useSsh', false);
      }
    }),

    textFieldView: Ember.TextField.extend({
      disabled: function(){
        return !this.get('isEnabled');
      }.property('isEnabled')
    })
  });


  
});
window.require.register("views/wizard/step3_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep3View = App.TableView.extend({

    templateName: require('templates/wizard/step3'),

    content:function () {
      return this.get('controller.hosts');
    }.property('controller.hosts.length'),

    message:'',
    linkText: '',
    status: '',

    selectedCategory: function() {
      return this.get('categories').findProperty('isActive');
    }.property('categories.@each.isActive'),

    registeredHostsMessage: '',

    displayLength: "25",

    didInsertElement: function () {
      this.get('controller').loadStep();
    },

    pageChecked: false,

    /**
     * select checkboxes of hosts on page
     */
    onPageChecked: function () {
      if (this.get('selectionInProgress')) return;
      this.get('pageContent').setEach('isChecked', this.get('pageChecked'));
    }.observes('pageChecked'),

    /**
     * select checkboxes of all hosts
     */
    selectAll: function () {
      this.get('content').setEach('isChecked', true);
    },

    /**
     * reset checkbox of all hosts
     */
    unSelectAll: function() {
      this.get('content').setEach('isChecked', false);
    },

    watchSelectionOnce: function () {
      Em.run.once(this, 'watchSelection');
    }.observes('content.@each.isChecked', 'pageContent'),

    /**
     * watch selection and calculate such flags as:
     * - noHostsSelected
     * - selectedHostsCount
     * - pageChecked
     */
    watchSelection: function() {
      this.set('selectionInProgress', true);
      this.set('pageChecked', !!this.get('pageContent.length') && this.get('pageContent').everyProperty('isChecked', true));
      this.set('selectionInProgress', false);
      var noHostsSelected = true;
      var selectedHostsCount = 0;
      this.get('content').forEach(function(host){
        selectedHostsCount += ~~host.get('isChecked');
        noHostsSelected = (noHostsSelected) ? !host.get('isChecked') : noHostsSelected;
      });
      this.set('noHostsSelected', noHostsSelected);
      this.set('selectedHostsCount', selectedHostsCount);
    },

    setRegisteredHosts: function(){
      this.set('registeredHostsMessage',Em.I18n.t('installer.step3.warning.registeredHosts').format(this.get('controller.registeredHosts').length));
    }.observes('controller.registeredHosts'),

    categoryObject: Em.Object.extend({
      hostsCount: 0,
      label: function () {
        return "%@ (%@)".fmt(this.get('value'), this.get('hostsCount'));
      }.property('value', 'hostsCount'),
      isActive: false,
      itemClass: function () {
        return this.get('isActive') ? 'active' : '';
      }.property('isActive')
    }),

    categories: function () {
      return [
        this.categoryObject.create({value: Em.I18n.t('common.all'), hostsBootStatus: 'ALL', isActive: true}),
        this.categoryObject.create({value: Em.I18n.t('installer.step3.hosts.status.installing'), hostsBootStatus: 'RUNNING'}),
        this.categoryObject.create({value: Em.I18n.t('installer.step3.hosts.status.registering'), hostsBootStatus: 'REGISTERING'}),
        this.categoryObject.create({value: Em.I18n.t('common.success'), hostsBootStatus: 'REGISTERED' }),
        this.categoryObject.create({value: Em.I18n.t('common.fail'), hostsBootStatus: 'FAILED', last: true })
      ];
    }.property(),

    hostBootStatusObserver: function(){
      Ember.run.once(this, 'countCategoryHosts');
      Ember.run.once(this, 'filter');
      Ember.run.once(this, 'monitorStatuses');
    }.observes('content.@each.bootStatus'),

    countCategoryHosts: function () {
      var counters = {
        "RUNNING": 0,
        "REGISTERING": 0,
        "REGISTERED": 0,
        "FAILED": 0
      };
      this.get('content').forEach(function (host) {
        if (counters[host.get('bootStatus')] !== undefined) {
          counters[host.get('bootStatus')]++;
        }
      }, this);
      counters["ALL"] = this.get('content.length');
      this.get('categories').forEach(function(category) {
        category.set('hostsCount', counters[category.get('hostsBootStatus')]);
      }, this);
    },


    /**
     * filter hosts by category
     */
    filter: function () {
      var self = this;
      Em.run.next(function () {
        var result = [];
        var selectedCategory = self.get('selectedCategory');
        if (!selectedCategory || selectedCategory.get('hostsBootStatus') === 'ALL') {
          result = self.get('content');
        } else {
          result = self.get('content').filterProperty('bootStatus', self.get('selectedCategory.hostsBootStatus'));
        }
        self.set('filteredContent', result);
      });
    }.observes('selectedCategory'),
    /**
     * Trigger on Category click
     * @param {Object} event
     */
    selectCategory: function (event) {
      var categoryStatus = event.context.get('hostsBootStatus');
      var self = this;
      this.get('categories').forEach(function (category) {
        category.set('isActive', (category.get('hostsBootStatus') === categoryStatus));
      });
      this.watchSelection();
    },

    /**
     * Select "All" hosts category
     * run registration of failed hosts again
     */
    retrySelectedHosts: function () {
      var eventObject = {context: Em.Object.create({hostsBootStatus: 'ALL'})};
      this.selectCategory(eventObject);
      this.get('controller').retrySelectedHosts();
    },

    monitorStatuses: function() {
      var hosts = this.get('controller.bootHosts');
      var failedHosts = hosts.filterProperty('bootStatus', 'FAILED').length;

      if (hosts.length === 0) {
        this.set('status', 'alert-warn');
        this.set('linkText', '');
        this.set('message', Em.I18n.t('installer.step3.warnings.missingHosts'));
      } else if (!this.get('controller.isWarningsLoaded')) {
        this.set('status', 'alert-info');
        this.set('linkText', '');
        this.set('message', Em.I18n.t('installer.step3.warning.loading'));
      } else if (this.get('controller.isHostHaveWarnings') || this.get('controller.repoCategoryWarnings.length') || this.get('controller.diskCategoryWarnings.length')) {
        this.set('status', 'alert-warn');
        this.set('linkText', Em.I18n.t('installer.step3.warnings.linkText'));
        this.set('message', Em.I18n.t('installer.step3.warnings.fails').format(hosts.length - failedHosts));
      } else {
        this.set('status', 'alert-success');
        this.set('linkText', Em.I18n.t('installer.step3.noWarnings.linkText'));
        if (failedHosts == 0) {
          // all are ok
          this.set('message', Em.I18n.t('installer.step3.warnings.noWarnings').format(hosts.length));
        } else if (failedHosts == hosts.length) {
          // all failed
          this.set('status', 'alert-warn');
          this.set('linkText', '');
          this.set('message', Em.I18n.t('installer.step3.warnings.allFailed').format(failedHosts));
        } else {
          // some failed
          this.set('message', Em.I18n.t('installer.step3.warnings.someWarnings').format((hosts.length - failedHosts), failedHosts));
        }
      }
    }.observes('controller.isWarningsLoaded', 'controller.isHostHaveWarnings', 'controller.repoCategoryWarnings', 'controller.diskCategoryWarnings')
  });

  //todo: move it inside WizardStep3View
  App.WizardHostView = Em.View.extend({

    tagName: 'tr',
    classNameBindings: ['hostInfo.bootStatus'],
    hostInfo: null,

    remove: function () {
      this.get('controller').removeHost(this.get('hostInfo'));
    },

    retry: function() {
      this.get('controller').retryHost(this.get('hostInfo'));
    },

    isRemovable: function () {
      return true;
    }.property(),

    isRetryable: function() {
      // return ['FAILED'].contains(this.get('hostInfo.bootStatus'));
      return false;
    }.property('hostInfo.bootStatus')

  });


  
});
window.require.register("views/wizard/step4_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep4View = Em.View.extend({

    templateName: require('templates/wizard/step4')

  });
});
window.require.register("views/wizard/step5_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');
  var lazyloading = require('utils/lazy_loading');

  App.WizardStep5View = Em.View.extend({

    templateName:require('templates/wizard/step5'),

    didInsertElement:function () {
      this.get('controller').loadStep();
    },

    body: Em.I18n.t('installer.step5.body')

  });

  App.SelectHostView = Em.Select.extend({
    content:[],
    zId:null,
    selectedHost:null,
    componentName:null,
    attributeBindings:['disabled'],
    isLoaded: false,
    isLazyLoading: false,

    change: function () {
      this.get('controller').assignHostToMaster(this.get("componentName"), this.get("value"), this.get("zId"));
      this.set('selectedHost', this.get('value'));
      this.get('controller').set('componentToRebalance', this.get("componentName"));
      this.get('controller').incrementProperty('rebalanceComponentHostsCounter');
    },

    /**
     * recalculate available hosts
     */
    rebalanceComponentHosts: function () {
      if (this.get('componentName') === this.get('controller.componentToRebalance')) {
        this.get('content').clear();
        this.set('isLoaded', false);
        this.initContent();
      }
    }.observes('controller.rebalanceComponentHostsCounter'),

    /**
     * get available hosts
     * @multipleComponents component can be assigned to multiple hosts,
     * shared hosts among the same component should be filtered out
     * @return {*}
     */
    getAvailableHosts: function () {
      var hosts = this.get('controller.hosts').slice();
      var componentName = this.get('componentName');
      var multipleComponents = this.get('controller.multipleComponents');
      var occupiedHosts = this.get('controller.selectedServicesMasters')
        .filterProperty('component_name', componentName)
        .mapProperty('selectedHost').without(this.get('selectedHost'));
      if (multipleComponents.contains(componentName)) {
        return hosts.filter(function (host) {
          return !occupiedHosts.contains(host.get('host_name'));
        }, this);
      }
      return hosts;
    },
    /**
     * on click start lazy loading
     */
    click: function () {
      var source = [];
      var componentName = this.get('componentName');
      var availableHosts = this.getAvailableHosts();
      var selectedHost = this.get('selectedHost');

      if (!this.get('isLoaded') && this.get('isLazyLoading')) {
        //filter out hosts, which already pushed in select
        source = availableHosts.filter(function(_host){
          return !this.get('content').someProperty('host_name', _host.host_name);
        }, this).slice();
        lazyloading.run({
          destination: this.get('content'),
          source: source,
          context: this,
          initSize: 30,
          chunkSize: 200,
          delay: 50
        });
      }
    },

    didInsertElement: function () {
      //The lazy loading for select elements supported only by Firefox and Chrome
      var isBrowserSupported = $.browser.mozilla || ($.browser.safari && navigator.userAgent.indexOf('Chrome') !== -1);
      var isLazyLoading = isBrowserSupported && this.get('controller.hosts').length > 100;
      this.set('isLazyLoading', isLazyLoading);
      this.initContent();
      this.set("value", this.get("selectedHost"));
    },
    /**
     * extract hosts from controller,
     * filter out available to selection and
     * push them into Em.Select content
     */
    initContent: function () {
      var hosts = this.getAvailableHosts();
      if (this.get('isLazyLoading')) {
        //select need at least 30 hosts to have scrollbar
        var initialHosts = hosts.slice(0, 30);
        if (!initialHosts.someProperty('host_name', this.get('selectedHost'))) {
          initialHosts.unshift(hosts.findProperty('host_name', this.get('selectedHost')));
        }
        this.set("content", initialHosts);
      } else {
        this.set("content", hosts);
      }
    }
  });

  App.AddControlView = Em.View.extend({
    componentName:null,
    tagName:"span",
    classNames:["badge", "badge-important"],
    template:Ember.Handlebars.compile('+'),

    click:function () {
      this.get('controller').addComponent(this.get('componentName'));
    }
  });

  App.RemoveControlView = Em.View.extend({
    zId:null,
    componentName:null,
    tagName:"span",
    classNames:["badge", "badge-important"],
    template:Ember.Handlebars.compile('-'),

    click:function () {
      this.get('controller').removeComponent(this.get('componentName'), this.get("zId"));
    }
  });
  
});
window.require.register("views/wizard/step6_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep6View = App.TableView.extend({

    templateName: require('templates/wizard/step6'),

    title: '',

    displayLength: "25",

    content: function () {
      return this.get('controller.hosts');
    }.property('controller.hosts'),

    filteredContent: function () {
      return this.get('content');
    }.property('content'),

    didInsertElement: function () {
      var controller = this.get('controller');
      if (controller.get('isMasters')) {
        this.set('label', Em.I18n.t('installer.step6.addHostWizard.body'));
        this.set('title', Em.I18n.t('installer.step5.header'));
      }
      else {
        this.set('title', Em.I18n.t('installer.step6.header'));
        this.setLabel();
      }
      App.tooltip($('body'), {selector: '[rel=tooltip]'});
      controller.loadStep();
    },

    setLabel: function () {
      var label = Em.I18n.t('installer.step6.body');
      var clients = this.get('controller.content.clients');
      clients.forEach(function (_client) {
        if (clients.length === 1) {
          label = label + ' ' + _client.display_name;
        }
        else
          if (_client !== clients[clients.length - 1]) {           // [clients.length - 1]
            label = label + ' ' + _client.display_name;
            if(_client !== clients[clients.length - 2]) {
              label = label + ',';
            }
          }
          else {
            label = label + ' ' + Em.I18n.t('and') + ' ' + _client.display_name + '.';
          }
      }, this);
      this.set('label', label);
    },
    /**
     * Binding host property with dynamic name
     * @type {*}
     */
    checkboxView: Em.Checkbox.extend({
      /**
       * Header object with host property name
       */
      checkbox: null,

      //if setAll true there is no need to check every checkbox whether all checked or not

      checkedBinding: 'checkbox.checked',

      disabledBinding: 'checkbox.isInstalled',

      click: function () {
        if ($.browser.mozilla) {
          this.toggleProperty('checkbox.checked');
        }
        this.get('controller').checkCallback(this.get('checkbox.component'));
      }
    })
  });

  App.WizardStep6HostView = Em.View.extend({

    host: null,
    tagName: 'td',

    didInsertElement: function () {
      if (!this.get('controller.isMasters')) {
        var components = this.get('controller').getMasterComponentsForHost(this.get('host.hostName'));
        if (components && components.length > 0) {
          components = components.map(function(_component) {
            return App.format.role(_component);
          });
          components = components.join(" /\n");
          App.popover(this.$(), {
            title: Em.I18n.t('installer.step6.wizardStep6Host.title').format(this.get('host.hostName')),
            content: components,
            placement: 'right',
            trigger: 'hover'
          });
        }
      }
    }
  });
});
window.require.register("views/wizard/step7_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep7View = Em.View.extend({

    templateName: require('templates/wizard/step7')

  });
  
});
window.require.register("views/wizard/step8_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  var App = require('app');

  App.WizardStep8View = Em.View.extend({

    templateName: require('templates/wizard/step8'),

    didInsertElement: function () {
      var controller = this.get('controller');
      controller.loadStep();
    },

    spinner : null,

    printReview: function() {
      var o = $("#step8-info");
      o.jqprint();
    },

    ajaxQueueLength: function() {
      return this.get('controller.ajaxQueueLength');
    }.property('controller.ajaxQueueLength'),

    ajaxQueueLeft: function() {
      return this.get('controller.ajaxQueueLeft');
    }.property('controller.ajaxQueueLeft'),

    // reference to modalPopup to make sure only one instance is created
    modalPopup: null,

    showLoadingIndicator: function() {
      if (!this.get('controller.isSubmitDisabled') || App.testMode) {
        if (this.get('modalPopup')) {
          this.get('modalPopup').hide();
          this.set('modalPopup', null);
        }
        return;
      }
      // don't create popup if it already exists
      if (this.get('modalPopup')) {
        return;
      }
      this.set('modalPopup', App.ModalPopup.show({
        header: '',

        showFooter: false,

        showCloseButton: false,

        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step8_log_popup'),

          message: function() {
            return Em.I18n.t('installer.step8.deployPopup.message').format(this.get('ajaxQueueComplete'), this.get('ajaxQueueLength'));
          }.property('ajaxQueueComplete', 'ajaxQueueLength'),

          controllerBinding: 'App.router.wizardStep8Controller',

          ajaxQueueLength: function() {
            return this.get('controller.ajaxQueueLength');
          }.property(),

          ajaxQueueComplete: function() {
            return this.get('ajaxQueueLength') - this.get('controller.ajaxQueueLeft');
          }.property('controller.ajaxQueueLeft', 'ajaxQueueLength'),

          barWidth: function () {
            return 'width: ' + (this.get('ajaxQueueComplete') / this.get('ajaxQueueLength') * 100) + '%;';
          }.property('ajaxQueueComplete', 'ajaxQueueLength'),

          autoHide: function() {
            if (this.get('controller.servicesInstalled')) {
              this.get('parentView').hide();
            }
          }.observes('controller.servicesInstalled')
        })
      }));
    }.observes('controller.isSubmitDisabled')
  });

  
});
window.require.register("views/wizard/step9_view", function(exports, require, module) {
  /**
   * Licensed to the Apache Software Foundation (ASF) under one
   * or more contributor license agreements.  See the NOTICE file
   * distributed with this work for additional information
   * regarding copyright ownership.  The ASF licenses this file
   * to you under the Apache License, Version 2.0 (the
   * "License"); you may not use this file except in compliance
   * with the License.  You may obtain a copy of the License at
   *
   *     http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  var App = require('app');
  var date = require('utils/date');

  App.WizardStep9View = App.TableView.extend({

    templateName: require('templates/wizard/step9'),
    barColor: '',
    resultMsg: '',
    resultMsgColor: '',
    isStepCompleted: function() {
     return (this.get('controller.progress') === '100');
    }.property('controller.progress'),

    displayLength: "25",

    content: function () {
      return this.get('controller.hosts');
    }.property('controller.hosts'),

    filteredContent: [],

    selectedCategory: function() {
      return this.get('categories').findProperty('isActive');
    }.property('categories.@each.isActive'),

    /**
     * Ember Object category. This object also contains
     * <code>
     *   hostStatus: {String} A valid status of a host.
     *               Used to filter hosts in that status.
     *   hostsCount: {Int} Dynamic count of hosts displayed in the category label
     *   label : {String} status and hosts in that status displayed which consists as a category on the page
     *   isActive: {boolean} Gets set when the category is selected/clicked by the user
     *   itemClass: {computed property} Binds the category link to active class when user clicks on the link
     * </code>
     */
    categoryObject: Em.Object.extend({
      hostsCount: 0,
      label: function () {
        return "%@ (%@)".fmt(this.get('value'), this.get('hostsCount'));
      }.property('value', 'hostsCount'),
      isActive: false,
      itemClass: function () {
        return this.get('isActive') ? 'active' : '';
      }.property('isActive')
    }),
    /**
     * computed property creates the category objects on the load of the page and sets 'All' as the active category
     * @Returns: All created categories which are binded and iterated in the template
     */
    categories: function () {
      return [
        this.categoryObject.create({value: Em.I18n.t('common.all'), hostStatus: 'all', isActive: true}),
        this.categoryObject.create({value: Em.I18n.t('installer.step9.hosts.status.label.inProgress'), hostStatus: 'inProgress'}),
        this.categoryObject.create({value: Em.I18n.t('installer.step9.hosts.status.label.warning'), hostStatus: 'warning'}),
        this.categoryObject.create({value: Em.I18n.t('common.success'), hostStatus: 'success'}),
        this.categoryObject.create({value: Em.I18n.t('common.fail'), hostStatus: 'failed', last: true })
      ];
    }.property(),

    hostStatusObserver: function(){
      Ember.run.once(this, 'countCategoryHosts');
      Ember.run.once(this, 'filter');
    }.observes('content.@each.status'),

    /**
     * count each category hosts to update label
     */
    countCategoryHosts: function () {
      var counters = {
        "info": 0,
        "pending": 0,
        "in_progress": 0,
        "heartbeat_lost": 0,
        "warning": 0,
        "success": 0,
        "failed": 0
      };

      this.get('content').forEach(function (host) {
        if (counters[host.get('status')] !== undefined) {
          counters[host.get('status')]++;
        }
      }, this);
      counters["all"] = this.get('content.length');
      counters["inProgress"] = counters["info"] + counters["pending"] + counters["in_progress"];
      counters["failed"] += counters["heartbeat_lost"];
      this.get('categories').forEach(function(category) {
        category.set('hostsCount', counters[category.get('hostStatus')]);
      }, this);
    },

    /**
     * filter hosts by category
     */
    filter: function () {
      var self = this;
      Em.run.next(function () {
        var result = [];
        var content = self.get('content');
        var selectedCategory = self.get('selectedCategory');
        if (!selectedCategory || selectedCategory.get('hostStatus') === 'all') {
          result = content;
        } else if (selectedCategory.get('hostStatus') == 'inProgress') {
          result = content.filter(function (_host) {
            return (_host.get('status') == 'info' || _host.get('status') == 'pending' || _host.get('status') == 'in_progress');
          });
        } else if (selectedCategory.get('hostStatus') == 'failed') {
          result = content.filter(function (_host) {
            return (_host.get('status') == 'failed' || _host.get('status') == 'heartbeat_lost');
          });
        } else {
          result = content.filterProperty('status', selectedCategory.get('hostStatus'));
        }
        self.set('filteredContent', result);
        self.set('filteredHostsInfo', Em.I18n.t('installer.step9.hosts.filteredHostsInfo').format(result.get('length'), content.get('length')));
      });
    }.observes('selectedCategory'),

    /**
     * filter hosts info shown up on bottom of the box. Set by filter function, when 'seletedCategory' changed
     */
    filteredHostsInfo: '',

    /**
     * On click handler for 'show all' link
     */
    showAllHosts: function () {
      this.get('categories').forEach(function (category) {
        category.set('isActive', (category.get('hostStatus') === 'all'));
      });
    },

    /**
     * Trigger on Category click
     * @param {Object} event
     */
    selectCategory: function (event) {
      var categoryStatus = event.context.get('hostStatus');
      this.get('categories').forEach(function (category) {
        category.set('isActive', (category.get('hostStatus') === categoryStatus));
      });
    },

    didInsertElement: function () {
      var controller = this.get('controller');
      this.onStatus();
      controller.navigateStep();
    },

    isHostHeartbeatLost: function () {
      return (this.get('controller.hostsWithHeartbeatLost').length > 0);
    }.property('controller.hostsWithHeartbeatLost.@each'),

    barWidth: function () {
      var controller = this.get('controller');
      return 'width: ' + controller.get('progress') + '%;';
    }.property('controller.progress'),

    progressMessage: function () {
      return Em.I18n.t('installer.step9.overallProgress').format(this.get('controller.progress'));
    }.property('controller.progress'),

    onStatus: function () {
      if (this.get('controller.status') === 'info') {
        this.set('resultMsg', '');
        this.set('barColor', 'progress-info');
      } else if (this.get('controller.status') === 'warning') {
        this.set('barColor', 'progress-warning');
        this.set('resultMsg', Em.I18n.t('installer.step9.status.warning'));
        this.set('resultMsgColor', 'alert-warning');
      } else if (this.get('controller.status') === 'failed') {
        this.set('barColor', 'progress-danger');
        this.set('resultMsgColor', 'alert-error');
        console.log('TRACE: Inside error view step9');
        if (this.get('isHostHeartbeatLost')) {
          // When present requests succeeds but some host components are in UNKNOWN or INSTALL_FAILED state and
          // hosts are in HEARTBEAT_LOST state
          this.set('resultMsg', Em.I18n.t('installer.step9.status.hosts.heartbeat_lost').format(this.get('controller.hostsWithHeartbeatLost').length));
        } else if (this.get('controller.startCallFailed')) {
          this.set('resultMsg', Em.I18n.t('installer.step9.status.start.services.failed'));
        } else {
          this.set('resultMsg', Em.I18n.t('installer.step9.status.failed'));
        }
      } else if (this.get('controller.status') === 'success') {
        console.log('TRACE: Inside success view step9');
        this.set('barColor', 'progress-success');
        this.set('resultMsg', Em.I18n.t('installer.step9.status.success'));
        this.set('resultMsgColor', 'alert-success');
      }
    }.observes('controller.status', 'controller.startCallFailed','isHostHeartbeatLost'),


    hostWithInstallFailed: function (event, context) {
      var controller = this.get('controller');
      App.ModalPopup.show({
        header: Em.I18n.t('installer.step9.host.heartbeat_lost.header'),
        classNames: ['sixty-percent-width-modal'],
        autoHeight: false,
        secondary: null,

        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step9_install_host_popup'),
          c: controller,
          failedHosts: function () {
            return controller.get('hostsWithHeartbeatLost');
          }.property()
        })
      });
    }
  });

  App.HostStatusView = Em.View.extend({
    tagName: 'tr',
    /**
     * Current host
     */
    obj: null,
    /**
     * wizardStep9Controller
     */
    controller: null,
    barColor: '',

    didInsertElement: function () {
      this.onStatus();
    },

    barWidth: function () {
      return 'width: ' + this.get('obj.progress') + '%;';
    }.property('obj.progress'),

    onStatus: function () {
      if (this.get('obj.status') === 'info') {
        this.set('barColor', 'progress-info');
      } else if (this.get('obj.status') === 'warning') {
        this.set('barColor', 'progress-warning');
        if (this.get('obj.progress') === '100') {
          this.set('obj.message', Em.I18n.t('installer.step9.host.status.warning'));
        }
      } else if (this.get('obj.status') === 'failed') {
        this.set('barColor', 'progress-danger');
        if (this.get('obj.progress') === '100') {
          this.set('obj.message', Em.I18n.t('installer.step9.host.status.failed'));
        }
      } else if (this.get('obj.status') === 'heartbeat_lost') {
        this.set('barColor', 'progress-danger');
        if (this.get('obj.progress') === '100') {
          this.set('obj.message', Em.I18n.t('installer.step9.host.heartbeat_lost'));
        }
      } else if (this.get('obj.status') === 'success' && this.get('isHostCompleted') && parseInt(this.get('controller.progress')) > 34) {
          this.set('barColor', 'progress-success');
          this.set('obj.message', Em.I18n.t('installer.step9.host.status.success'));
      }
    }.observes('obj.status', 'obj.progress', 'controller.progress'),

    isFailed: function () {
      return (this.get('isHostCompleted') && (this.get('obj.status') === 'failed' || this.get('obj.status') === 'heartbeat_lost'));
    }.property('obj.status', 'isHostCompleted'),

    isSuccess: function () {
      return (this.get('isHostCompleted') && this.get('obj.status') === 'success');
    }.property('obj.status', 'isHostCompleted'),

    isWarning: function () {
      return(this.get('isHostCompleted') && this.get('obj.status') === 'warning');
    }.property('obj.status', 'isHostCompleted'),

    isHostCompleted: function () {
      return this.get('obj.progress') == 100;
    }.property('obj.progress'),

    hostLogPopup: function (event, context) {
      var controller = this.get('controller');
      var host = this.get('obj');

      App.ModalPopup.show({
        header: host.get('name'),
        classNames: ['sixty-percent-width-modal'],
        autoHeight: false,
        secondary: null,
        /**
         * Current host
         */
        host: host,
        /**
         * wizardStep9Controller
         */
        c: controller,

        onClose: function () {
          this.set('c.currentOpenTaskId', 0);
          this.hide();
        },

        bodyClass: Ember.View.extend({
          templateName: require('templates/wizard/step9HostTasksLogPopup'),
          isHeartbeatLost: function() {
            return (host.get('status') === 'heartbeat_lost');
          }.property('host.status'),
          isNoTasksScheduled: function() {
            return host.get('isNoTasksForInstall');
          }.property('host.isNoTasksForInstall'),

          isLogWrapHidden: true,

          showTextArea: false,

          isEmptyList: true,

          visibleTasks: function () {
            this.set("isEmptyList", true);
            if (this.get('category.value')) {
              var filter = this.get('category.value');
              var tasks = this.get('tasks');
              tasks.setEach("isVisible", false);

              if (filter == "all") {
                tasks.setEach("isVisible", true);
              }
              else if (filter == "pending") {
                tasks.filterProperty("status", "pending").setEach("isVisible", true);
                tasks.filterProperty("status", "queued").setEach("isVisible", true);
              }
              else if (filter == "in_progress") {
                tasks.filterProperty("status", "in_progress").setEach("isVisible", true);
              }
              else if (filter == "failed") {
                tasks.filterProperty("status", "failed").setEach("isVisible", true);
              }
              else if (filter == "completed") {
                tasks.filterProperty("status", "completed").setEach("isVisible", true);
              }
              else if (filter == "aborted") {
                tasks.filterProperty("status", "aborted").setEach("isVisible", true);
              }
              else if (filter == "timedout") {
                tasks.filterProperty("status", "timedout").setEach("isVisible", true);
              }

              if (tasks.filterProperty("isVisible", true).length > 0) {
                this.set("isEmptyList", false);
              }
            }
          }.observes('category', 'tasks'),

          categories: [
            Ember.Object.create({value: 'all', label: Em.I18n.t('installer.step9.hostLog.popup.categories.all') }),
            Ember.Object.create({value: 'pending', label: Em.I18n.t('installer.step9.hostLog.popup.categories.pending')}),
            Ember.Object.create({value: 'in_progress', label: Em.I18n.t('installer.step9.hostLog.popup.categories.in_progress')}),
            Ember.Object.create({value: 'failed', label: Em.I18n.t('installer.step9.hostLog.popup.categories.failed') }),
            Ember.Object.create({value: 'completed', label: Em.I18n.t('installer.step9.hostLog.popup.categories.completed') }),
            Ember.Object.create({value: 'aborted', label: Em.I18n.t('installer.step9.hostLog.popup.categories.aborted') }),
            Ember.Object.create({value: 'timedout', label: Em.I18n.t('installer.step9.hostLog.popup.categories.timedout') })
          ],

          category: null,

          tasks: function () {
            var tasksArr = [];
            var host = this.get('parentView.host');
            var tasks = this.getStartedTasks(host);
            tasks = tasks.sortProperty('Tasks.id');
            if (tasks.length) {
              tasks.forEach(function (_task) {
                var taskInfo = Ember.Object.create({});
                taskInfo.set('id', _task.Tasks.id);
                taskInfo.set('requestId', _task.Tasks.request_id);
                taskInfo.set('command', _task.Tasks.command.toLowerCase() === 'service_check' ? '' : _task.Tasks.command.toLowerCase());
                taskInfo.set('status', App.format.taskStatus(_task.Tasks.status));
                taskInfo.set('role', App.format.role(_task.Tasks.role));
                taskInfo.set('stderr', _task.Tasks.stderr);
                taskInfo.set('stdout', _task.Tasks.stdout);
                taskInfo.set('startTime',  date.startTime(_task.Tasks.start_time));
                taskInfo.set('duration', date.durationSummary(_task.Tasks.start_time, _task.Tasks.end_time));
                taskInfo.set('isVisible', true);
                taskInfo.set('icon', '');
                taskInfo.set('hostName', _task.Tasks.host_name);
                taskInfo.set('outputLog', Em.I18n.t('common.hostLog.popup.logDir.path') + Em.I18n.t('common.hostLog.popup.outputLog.value').format(_task.Tasks.id));
                taskInfo.set('errorLog', Em.I18n.t('common.hostLog.popup.logDir.path') + Em.I18n.t('common.hostLog.popup.errorLog.value').format(_task.Tasks.id));
                if (taskInfo.get('status') == 'pending' || taskInfo.get('status') == 'queued') {
                  taskInfo.set('icon', 'icon-cog');
                } else if (taskInfo.get('status') == 'in_progress') {
                  taskInfo.set('icon', 'icon-cogs');
                } else if (taskInfo.get('status') == 'completed') {
                  taskInfo.set('icon', ' icon-ok');
                } else if (taskInfo.get('status') == 'failed') {
                  taskInfo.set('icon', 'icon-exclamation-sign');
                } else if (taskInfo.get('status') == 'aborted') {
                  taskInfo.set('icon', 'icon-minus');
                } else if (taskInfo.get('status') == 'timedout') {
                  taskInfo.set('icon', 'icon-time');
                }
                tasksArr.push(taskInfo);
              }, this);
            }
            return tasksArr;
          }.property('parentView.c.logTasksChangesCounter'),

          backToTaskList: function (event, context) {
            this.destroyClipBoard();
            this.set("isLogWrapHidden", true);
          },

          getStartedTasks: function (host) {
            return host.logTasks.filter(function (task) {
              return task.Tasks.status;
            });
          },

          openTaskLogInDialog: function () {
            var newwindow = window.open();
            var newdocument = newwindow.document;
            newdocument.write($(".task-detail-log-info").html());
            newdocument.close();
          },

          openedTask: function () {
            return this.get('tasks').findProperty('id', this.get('parentView.c.currentOpenTaskId'))
          }.property('parentView.c.currentOpenTaskId', 'tasks.@each'),

          toggleTaskLog: function (event, context) {
            if (this.get('isLogWrapHidden')) {
              var taskInfo = event.context;
              this.set("isLogWrapHidden", false);
              this.set('parentView.c.currentOpenTaskId', taskInfo.id);
              this.set('parentView.c.currentOpenTaskRequestId', taskInfo.requestId);
              this.get('parentView.c').loadCurrentTaskLog();
              $(".modal").scrollTop(0);
              $(".modal-body").scrollTop(0);
            }
            else {
              this.set("isLogWrapHidden", true);
              this.set('parentView.c.currentOpenTaskId', 0);
              this.set('parentView.c.currentOpenTaskRequestId', 0);
            }
          },

          textTrigger: function (event) {
            if ($(".task-detail-log-clipboard").length > 0) {
              this.destroyClipBoard();
            }
            else {
              this.createClipBoard();
            }
          },

          createClipBoard: function () {
            var log = $(".task-detail-log-maintext");
            $(".task-detail-log-clipboard-wrap").html('<textarea class="task-detail-log-clipboard"></textarea>');
            $(".task-detail-log-clipboard")
              .html("stderr: \n" + $(".stderr").html() + "\n stdout:\n" + $(".stdout").html())
              .css("display", "block")
              .width(log.width())
              .height(log.height())
              .select();
            log.css("display", "none")
          },

          destroyClipBoard: function () {
            $(".task-detail-log-clipboard").remove();
            $(".task-detail-log-maintext").css("display", "block");
          }
        })
      });
    }

  });
});
